//
//	VisualAttentionTest.cc		This file is a part of the IKAROS project
// 				<Short description of the module>
//
//    Copyright (C) 2005 Mikael Asker
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//    
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//    
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Created: 050429
//
//	<Additional description of the module>

#include "VisualAttentionTest.h"


TRMProxy::TRMProxy()
    : myOutputMatrix(0),
      myImageWidth(0),
      myImageHeight(0),
      myMapWidth(0),
      myMapHeight(0),
      myMinimumStdDev(0),
      myMaximumStdDev(0),
      iAmCompletelyOpen(false),
      myCurrentStdDev(0),
      myCurrentVariance(0)
{}


void 
TRMProxy::init(unsigned int  theImageWidth, 
               unsigned int  theImageHeight,
               float**       theOutputMatrix)
{
    // Valid matrix && has not been initialized before
    assert(theOutputMatrix && ! myOutputMatrix);

    myOutputMatrix = theOutputMatrix;
    myImageWidth = theImageWidth;
    myImageHeight = theImageHeight;
    myMapWidth = (myImageWidth >> aMapScale);
    myMapHeight = (myImageHeight >> aMapScale);

    const float aSmallestImageSide = min(myImageWidth, myImageHeight);
    //cout << aSmallestImageSide << endl;
    myMinimumStdDev = aSmallestImageSide / aMaximumDivisor;
    myMaximumStdDev = aSmallestImageSide / aMinimumDivisor;

    iAmCompletelyOpen = true;
    this->generateOutput(0, 0);
}


void 
TRMProxy::reduce(unsigned int x, unsigned int y)
{
    assert(myOutputMatrix);   // Has been initialized
    assert(x < myImageWidth && y < myImageHeight);   // Inside window

    iAmCompletelyOpen = false;
    myCurrentStdDev = myMinimumStdDev;
    myCurrentVariance = myCurrentStdDev * myCurrentStdDev;

    cout << "reduce (" << x << "," << y << "):" << myCurrentStdDev << " \n";
    this->generateOutput(x, y);
}


void 
TRMProxy::move(unsigned int x, unsigned int y)
{
    assert(myOutputMatrix);   // Has been initialized
    assert(x < myImageWidth && y < myImageHeight);   // Inside window

    cout << "move (" << x << "," << y << "):" << myCurrentStdDev << " \n";
    this->generateOutput(x, y);
}


void 
TRMProxy::enlarge(unsigned int x, unsigned int y)
{
    assert(myOutputMatrix);   // Has been initialized
    assert(x < myImageWidth && y < myImageHeight);   // Inside window

    myCurrentStdDev *= 1.08;
    myCurrentVariance = myCurrentStdDev * myCurrentStdDev;
    iAmCompletelyOpen = (myCurrentStdDev > myMaximumStdDev);    

    cout << "enlarge (" << x << "," << y << "):" << myCurrentStdDev << " \n";
    this->generateOutput(x, y);
}


void 
TRMProxy::open()
{
    assert(myOutputMatrix);   // Has been initialized

    cout << "open" << " \n";
    iAmCompletelyOpen = true;
    this->generateOutput(0, 0);
}


void
TRMProxy::generateOutput(unsigned int x, unsigned int y)
{
    assert(myOutputMatrix);   // Has been initialized
    assert(x < myImageWidth && y < myImageHeight);   // Inside window
    //cout << "generateOutput (" << x << "," << y << "):" << myCurrentStdDev << " \n";

    if (iAmCompletelyOpen)
    {
        for (unsigned int i = 0; i < myMapWidth; i++)
        {
            for (unsigned int j = 0; j < myMapHeight; j++)
            {
                myOutputMatrix[j][i] = aFullRelevance;
            }
        }
    }
    else   // Not completely open
    {    
        //const unsigned int aReducedX = x >> 4;
        //const unsigned int aReducedY = y >> 4;
        //const unsigned int aReducedCurrentRadius = myCurrentRadius >> 4;

        //cout << "(" << aReducedX << "," << aReducedY << "):" << aReducedCurrentRadius << " \n";

        for (unsigned int i = 0; i < myMapWidth; i++)   // Sweep X coordinates
        {
            const unsigned int anImageI = i << aMapScale | 1 << aMapScale - 1;
            
            for (unsigned int j = 0; j < myMapHeight; j++)
            {
                const unsigned int anImageJ = j << aMapScale | 1 << aMapScale - 1;

                myOutputMatrix[j][i] = 
                    TRMProxy::weight(x, y, anImageI, anImageJ) 
                    * aFullRelevance;

                //cout << "(" << i << "," << j << ") = " << myOutputMatrix[j][i] << endl;
            }
        }
    }
}


float
TRMProxy::weight(unsigned int x, unsigned int y, unsigned int i, unsigned int j)
{
    const float dx = (int)i - (int)x;
    const float dy = (int)j - (int)y;
    static const float c = 1 / sqrt(2 * /*PI*/ 3.14159267) / 0.399 * 1.5;
    return c * exp(- (dx * dx + dy * dy) / myCurrentVariance / 2);
}


Module * 
VisualAttentionTest::Create(char* theName, Parameter* theParameters)
{
    return new VisualAttentionTest(theName, theParameters);
}


VisualAttentionTest::VisualAttentionTest(char* theName, Parameter* theParameters) :
    Module(theName),
    myTask(anUndefinedTask),
    myLearningCoordinates(aNotACoordinateValue, aNotACoordinateValue),
    mySearchCategory(GetIntValue(theParameters, "search", VisualCategories::aNotACategoryIdentifier)),
    myDetectCategory(GetIntValue(theParameters, "detect", VisualCategories::aNotACategoryIdentifier)),
    myFollowCategory(GetIntValue(theParameters, "follow", VisualCategories::aNotACategoryIdentifier)),
    myFollowState(aDetectingState),
    myBiasFlag((mySearchCategory != VisualCategories::aNotACategoryIdentifier ||
                myDetectCategory != VisualCategories::aNotACategoryIdentifier ||
                myFollowCategory != VisualCategories::aNotACategoryIdentifier) &&
               ! GetIntValueFromList(theParameters, "bias", "yes/no")),
    myFOAInput(0),
    myCategoryInput(0),
    myCategoryOutput(0),
    myIOROutput(0),
    myLearnOutput(0),
    myTickCounter(0),
    myLastXPos(0),
    myLastYPos(0),
    myEnlargeDelay(0)
{
    const char * const aLearningTargetString = GetValue(theParameters, "learn");

    if (aLearningTargetString)
    {
        myTask = aLearnTask;
        
        const char * anErrorMessage = 0;
        
        const bool aSuccessFlag = this->learningTarget(aLearningTargetString,
                                                       anErrorMessage, 
                                                       myLearningCoordinates,
                                                       myLearningRadius);
        
        if (! aSuccessFlag)
        {
            this->Notify(msg_fatal_error, "VisualAttentionTest: %s\n", anErrorMessage);
            return;
        }
    }

    if (mySearchCategory != VisualCategories::aNotACategoryIdentifier)
    {
        if (myTask != anUndefinedTask)
        {
            this->Notify(msg_fatal_error, "VisualAttentionTest: more than one task\n");
            return;
        }
        
        myTask = aSearchTask;
    }

    if (myDetectCategory != VisualCategories::aNotACategoryIdentifier)
    {
        if (myTask != anUndefinedTask)
        {
            this->Notify(msg_fatal_error, "VisualAttentionTest: more than one task\n");
            return;
        }
        
        myTask = aDetectTask;
    }

    if (myFollowCategory != VisualCategories::aNotACategoryIdentifier)
    {
        if (myTask != anUndefinedTask)
        {
            this->Notify(msg_fatal_error, "VisualAttentionTest: more than one task\n");
            return;
        }
        
        myTask = aFollowTask;
        myFollowState = aDetectingState;
    }

    // AddInput here declares one input for the module. This is the same
    // name used in the experiment configuration file.
    AddInput("INPUT_RED");
    AddInput("FOA_INPUT");
    AddInput("CATEGORY_INPUT");

    // Here we declare an output and tell the simulator that its size will be
    // figured out later. The parameter is not strictly necessary as it is the
    // default value, but it is a good idea to include it for clarity.

    AddOutput("CATEGORY_OUTPUT");
    AddOutput("TRM_OUTPUT");
    AddOutput("IOR_OUTPUT");
    AddOutput("LEARN_OUTPUT");

    SetOutputSize("CATEGORY_OUTPUT", 1);
    SetOutputSize("TRM_OUTPUT", 10, 10);   // SIZE DEPENDENCIES !
    SetOutputSize("IOR_OUTPUT", 1);
    SetOutputSize("LEARN_OUTPUT", 1);
}


void
VisualAttentionTest::SetSizes()
{
    const int anImageSizeX = GetInputSizeX("INPUT_RED");
    const int anImageSizeY = GetInputSizeY("INPUT_RED");

    if (anImageSizeX == unknown_size || anImageSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: the red input size is not set\n");
        return;
    }
    else if (anImageSizeX == 0 || anImageSizeY == 0)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: a red input size is zero\n");
        return;
    }

    //cout << "anImageSizeX = " << anImageSizeX << 
    //        ", anImageSizeY = " << anImageSizeY << endl;
            
    myImageWidth = anImageSizeX;
    myImageHeight = anImageSizeY;

    // Find out the size of the FOA input.

    const int aFOAInputSizeX = GetInputSizeX("FOA_INPUT");
    const int aFOAInputSizeY = GetInputSizeY("FOA_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (aFOAInputSizeX == unknown_size || aFOAInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: the FOA input size is not set\n");
        return;
    }

    if (aFOAInputSizeX != 2 || aFOAInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: wrong size of FOA input\n");
        return;
    }


    // Find out the size of the category input.

    const int aCategoryInputSizeX = GetInputSizeX("CATEGORY_INPUT");
    const int aCategoryInputSizeY = GetInputSizeY("CATEGORY_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (aCategoryInputSizeX == unknown_size || aCategoryInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: the category input size is not set\n");
        return;
    }

    if (aCategoryInputSizeX != 1 || aCategoryInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualAttentionTest: the category input is not scalar\n");
        return;
    }
}


void
VisualAttentionTest::Init()
{
    // Here we could add any number of consistency checks and do any memory
    // allocations and other initializations that wer need to do.

    // Set the variables that will be associated with the inputs and outputs
    // during the simulation. Each time the Tick method is entered, myInput
    // will point to the input vector, and myOutput will point to
    // the output vector. 

    myFOAInput = GetInputMatrix("FOA_INPUT");
    myCategoryInput = GetInputMatrix("CATEGORY_INPUT");

    myCategoryOutput = GetOutputMatrix("CATEGORY_OUTPUT");
    myTRM.init(myImageWidth, myImageHeight, GetOutputMatrix("TRM_OUTPUT"));
    myIOROutput = GetOutputMatrix("IOR_OUTPUT");
    myLearnOutput = GetOutputMatrix("LEARN_OUTPUT");

    if (! myBiasFlag)
    {
        **myCategoryOutput = VisualCategories::aNotACategoryIdentifier;
    }
    else
    {
        switch (myTask)
        {
            case aSearchTask :
                **myCategoryOutput = mySearchCategory;
                break;

            case aDetectTask :
                **myCategoryOutput = myDetectCategory;
                break;

            case aFollowTask :
                **myCategoryOutput = myFollowCategory;
                break;
                
            default :
                assert(false);
        }
    }
    
    switch (myTask)
    {
        case aLearnTask :
            **myIOROutput = 1;   // Enable IOR
            break;

        case aSearchTask :
            **myIOROutput = 1;   // Enable IOR
            break;

        case aDetectTask :
            **myIOROutput = 1;   // Enable IOR
            break;

        case aFollowTask :
            **myIOROutput = 0;   // Disable IOR
            break;
                
        default :
            assert(false);
    }

    **myLearnOutput = 0;   // Don't learn until the learning target has been reached
                 
    this->Notify(msg_print, 
                 "VisualAttentionTest::Init : myBiasFlag = %d.\n", 
                 (int)myBiasFlag);
}


void
VisualAttentionTest::Tick()
{
    //this->Notify(msg_print, "VisualAttentionTest::Tick begins.\n");

    // VisualAttentionTest in action. Input is a pointer to the current 
    // input vector and output points to the output vector that will be sent
    // along to whatever modules it is connected to.

    switch (myTask)
    {
        case aLearnTask :
        {
            if (**myLearnOutput)   // The learning target has been reached
            {
                static int aDelayCountLeft = 1;

                if (aDelayCountLeft > 0)   // A delay tick
                {        
                    aDelayCountLeft--;
                }
                else   // The delay is over
                {
                    this->Notify(msg_fatal_error,
                                 "VisualAttentionTest: completed - category %.0f learned.\n",
                                 **myCategoryInput);
                    return;
                }
            }
            // The learning target has not been reached

            if (Point2D(lround((*myFOAInput)[0]),
                        lround((*myFOAInput)[1])).distance(myLearningCoordinates) < 
                myLearningRadius)
            {
                **myLearnOutput = 1;   // First time => positive edge
            }

            this->Notify(msg_print,
                         "VisualAttentionTest::Tick :\n"
                         "          myFOAInput = (%.0f,%.0f),\n"
                         "          myLearningCoordinates = (%d,%d),\n"
                         "          myLearnOutput = %.0f\n",
                         (*myFOAInput)[0],
                         (*myFOAInput)[1],
                         myLearningCoordinates.i,
                         myLearningCoordinates.j,
                         **myLearnOutput);
        }
        break;

        case aSearchTask :

            this->Notify(msg_print, 
                         "VisualAttentionTest::Tick : **myCategoryInput = %.0f\n", 
                         **myCategoryInput);

            if (**myCategoryInput == mySearchCategory)   // Completed
            {
                this->Notify(msg_fatal_error,
                             "VisualAttentionTest: completed - category %.0f found.\n",
                             **myCategoryInput);
                return;
            }
                
            break;

        case aDetectTask :

            //cout << "VisualAttentionTest::Tick : **myCategoryInput = " << **myCategoryInput << endl;

            if (**myCategoryInput != VisualCategories::aNotACategoryIdentifier)   // Found something
            {
                this->Notify(msg_print, 
                             "%u !!! CATEGORY %.0f FOUND !!!\n", 
                             myTickCounter, 
                             **myCategoryInput);
            }
                
            break;

        case aFollowTask :

            switch (myFollowState)
            {
                case aDetectingState :
                    if (**myCategoryInput != VisualCategories::aNotACategoryIdentifier)   // Found
                    {
                        myFollowCategory = **myCategoryInput;
                        
                        this->Notify(msg_print, 
                                     "%u !!! CATEGORY %.0f FOUND - STARTING !!!\n", 
                                     myTickCounter, 
                                     myFollowCategory);

                        myLastXPos = (*myFOAInput)[0];
                        myLastYPos = (*myFOAInput)[1];

                        this->startTrackers();

                        float aNextXPosEstimate = 0;
                        float aNextYPosEstimate = 0;
                        this->estimatePosition(aNextXPosEstimate, aNextYPosEstimate);
                        myTRM.reduce(lround(aNextXPosEstimate), lround(aNextYPosEstimate));   

                        if (! myBiasFlag)   // Bias filters if not already done
                        {
                            **myCategoryOutput = myFollowCategory;
                        }

                        myFollowState = aFollowingState;
                    }
                    break;

                case aFollowingState :
                {
                    const float aCurrentXPos = (*myFOAInput)[0];
                    const float aCurrentYPos = (*myFOAInput)[1];
                    
                    if ((aCurrentXPos != myLastXPos || aCurrentYPos != myLastYPos) &&
                        **myCategoryInput == myFollowCategory)   // New measurement
                    {
                        this->Notify(msg_print, 
                                     "%u !!! NEW MEASUREMENT FOR CATEGORY %.0f !!!\n", 
                                     myTickCounter, 
                                     myFollowCategory);

                        myLastXPos = aCurrentXPos;
                        myLastYPos = aCurrentYPos;
                        
                        // Update trackers
                            myXTracker.update(aCurrentXPos);
                            myYTracker.update(aCurrentYPos);
                    }
                    else   // We have no new measurement
                    {
                        // Update trackers blindly
                            myXTracker.update();
                            myYTracker.update();
                    }

                    float aNextXPosEstimate;
                    float aNextYPosEstimate;
                    this->estimatePosition(aNextXPosEstimate, aNextYPosEstimate);

                    if (aNextXPosEstimate < 0 || aNextXPosEstimate > myImageWidth - 1 ||
                        aNextYPosEstimate < 0 || aNextYPosEstimate > myImageHeight - 1)   // Lost
                    {
                        this->Notify(msg_print, 
                                     "%u !!! CATEGORY %.0f ESCAPED FROM WINDOW - DETECTING !!!\n", 
                                     myTickCounter, 
                                     myFollowCategory);

                        myFollowCategory = VisualCategories::aNotACategoryIdentifier;

                        // Stop trackers
                        
                        myTRM.open();

                        // Unbias filters if they were unbiased before
                            if (! myBiasFlag)
                            {
                            }

                        myFollowState = aDetectingState;
                    }
                    else if (**myCategoryInput != myFollowCategory)   // Lost
                    {
                        this->Notify(msg_print, 
                                     "%u !!! CATEGORY %.0f LOST - SEARCHING !!!\n", 
                                     myTickCounter, 
                                     myFollowCategory);

                        // Zoom out one step
                            if (myEnlargeDelay > 0)
                            {
                                myEnlargeDelay--;
                            }
                            else
                            {
                                myTRM.enlarge(lround(aNextXPosEstimate), lround(aNextYPosEstimate));
                                myEnlargeDelay = anEnlargeTime;
                            }

                        myFollowState = aSearchingState;
                    }
                    else   // Not lost
                    {
                        myTRM.move(lround(aNextXPosEstimate), lround(aNextYPosEstimate));
                    }
                }
                break;

                case aSearchingState :
                    
                    if (**myCategoryInput != myFollowCategory)   // Still lost
                    {
                        // Update trackers blindly - we have no measurement
                            myXTracker.update();
                            myYTracker.update();
                            const float aNextXPosEstimate = myXTracker.getEstimate();
                            const float aNextYPosEstimate = myYTracker.getEstimate();

                        if (aNextXPosEstimate < 0 || aNextXPosEstimate > myImageWidth - 1 ||
                            aNextYPosEstimate < 0 || aNextYPosEstimate > myImageHeight - 1)   // Lost
                        {
                            this->Notify(msg_print, 
                                         "%u !!! CATEGORY %.0f ESCAPED FROM WINDOW - DETECTING !!!\n", 
                                         myTickCounter, 
                                         myFollowCategory);

                            myFollowCategory = VisualCategories::aNotACategoryIdentifier;

                            // Stop trackers
                        
                            myTRM.open();

                            // Unbias filters if they were unbiased before
                                if (! myBiasFlag)
                                {
                                }

                            myFollowState = aDetectingState;
                        }
                        else   // Inside window
                        {
                            // Zoom out one step
                                if (myEnlargeDelay > 0)
                                {
                                    myEnlargeDelay--;
                                }
                                else
                                {
                                    myTRM.enlarge(lround(aNextXPosEstimate), lround(aNextYPosEstimate));
                                    myEnlargeDelay = anEnlargeTime;
                                }

                            if (myTRM.isCompletelyOpen())   // Zoomed out completely
                            {
                                this->Notify(msg_print, 
                                             "%u !!! GIVING UP SEARCH FOR CATEGORY %.0f - DETECTING !!!\n", 
                                             myTickCounter, 
                                             myFollowCategory);

                                myFollowCategory = VisualCategories::aNotACategoryIdentifier;

                                // Stop trackers

                                if (! myBiasFlag)   // Unbias filters if they were unbiased before
                                {
                                    **myCategoryOutput = VisualCategories::aNotACategoryIdentifier;
                                }

                                myFollowState = aDetectingState;
                            }
                        }
                    }
                    else   // Found
                    {
                        this->Notify(msg_print, 
                                     "%u !!! CATEGORY %.0f FOUND - RESTARTING !!!\n", 
                                     myTickCounter, 
                                     myFollowCategory);

                        // Update trackers
                            const float aCurrentXPos = (*myFOAInput)[0];
                            myXTracker.update(aCurrentXPos);
                            const float aCurrentYPos = (*myFOAInput)[1];
                            myYTracker.update(aCurrentYPos);

                        float aNextXPosEstimate = 0;
                        float aNextYPosEstimate = 0;
                        this->estimatePosition(aNextXPosEstimate, aNextYPosEstimate);

                        if (aNextXPosEstimate < 0 || aNextXPosEstimate > myImageWidth - 1 ||
                            aNextYPosEstimate < 0 || aNextYPosEstimate > myImageHeight - 1)   // Lost
                        {
                            this->Notify(msg_print, 
                                         "%u !!! CATEGORY %.0f ESCAPED FROM WINDOW - DETECTING !!!\n", 
                                         myTickCounter, 
                                         myFollowCategory);

                            myFollowCategory = VisualCategories::aNotACategoryIdentifier;

                            // Stop trackers
                        
                            myTRM.open();

                            // Unbias filters if they were unbiased before
                                if (! myBiasFlag)
                                {
                                }

                            myFollowState = aDetectingState;
                        }
                        else   // Not lost
                        {
                            myTRM.reduce(lround(aNextXPosEstimate), lround(aNextYPosEstimate));   
                            myFollowState = aFollowingState;
                        }
                    }

                    break;

                default :
                    this->Notify(msg_fatal_error, "VisualAttentionTest: impossible myFollowState value\n");
                    return;
            }

            break;

        default :
            this->Notify(msg_fatal_error, "VisualAttentionTest: impossible myTask value\n");
            return;
    }

    //myOldLostInput = **myLostInput;
    myTickCounter++;
    //this->Notify(msg_print, "VisualAttentionTest::Tick ends.\n");
}


/*static*/ bool 
VisualAttentionTest::learningTarget(const char *    theTargetString,
                                    const char * &  theErrorMessage,
                                    Point2D &       theTargetCoordinates,
                                    int&            theTargetRadius)
{
    const char * aCurrentPos = strchr(theTargetString, '(');
    if (! aCurrentPos)
    {
        theErrorMessage = "no '(' in learn parameter value\n";
        return false;
    }
    aCurrentPos++;
        
    char * aPosAfterLong;
    const long anXValue = strtol(aCurrentPos, &aPosAfterLong, 10);
    if (aPosAfterLong == aCurrentPos)
    {
        theErrorMessage = "no first coordinate in learn parameter value\n";
        return false;
    }
    aCurrentPos = aPosAfterLong;

    aCurrentPos = strchr(aCurrentPos, ',');
    if (! aCurrentPos)
    {
        theErrorMessage = "no first ',' in learn parameter value\n";
        return false;
    }
    aCurrentPos++;

    const long anYValue = strtol(aCurrentPos, &aPosAfterLong, 10);
    if (aPosAfterLong == aCurrentPos)
    {
        theErrorMessage = "no second coordinate in learn parameter value\n";
        return false;
    }
    aCurrentPos = aPosAfterLong;

    aCurrentPos = strchr(aCurrentPos, ')');
    if (! aCurrentPos)
    {
        theErrorMessage = "no ')' in learn parameter value\n";
        return false;
    }
    aCurrentPos++;

    aCurrentPos = strchr(aCurrentPos, ',');
    if (! aCurrentPos)
    {
        theErrorMessage = "no second ',' in learn parameter value\n";
        return false;
    }
    aCurrentPos++;

    theTargetRadius = strtol(aCurrentPos, &aPosAfterLong, 10);
    if (aPosAfterLong == aCurrentPos)
    {
        theErrorMessage = "no radius in learn parameter value\n";
        return false;
    }
    aCurrentPos = aPosAfterLong;
        
    theTargetCoordinates = Point2D(anXValue, anYValue);
    return true;
}


void 
VisualAttentionTest::startTrackers(/*float& theNextXPosEstimate, 
                                   float& theNextYPosEstimate*/)
{
    const float aCurrentXPos = (*myFOAInput)[0];
    const float aCurrentYPos = (*myFOAInput)[1];

    // Estimate speed from features
    const float aFakedFirstXPos = aCurrentXPos;
    const float aFakedFirstYPos = aCurrentYPos;

    myXTracker.init(aFakedFirstXPos, (float)aProcessNoiseStdDev, (float)aMeasurementNoiseStdDev);
    myYTracker.init(aFakedFirstYPos, (float)aProcessNoiseStdDev, (float)aMeasurementNoiseStdDev);

    myXTracker.update(aCurrentXPos);
    myYTracker.update(aCurrentYPos);

    //theNextXPosEstimate = myXTracker.getEstimate();
    //theNextYPosEstimate = myYTracker.getEstimate();
}


void 
VisualAttentionTest::estimatePosition(float& theXPosEstimate, float& theYPosEstimate)
{
    const Image<float> anXStateVector = myXTracker.getStateVector();
    //cout << "TEST: x = " << anXStateVector.getVal(0,0) << ", v = " << anXStateVector.getVal(0,1) <<
    //        ", a = " << anXStateVector.getVal(0,2) << endl;
    const float x = anXStateVector.getVal(0,0);
    const float vx = anXStateVector.getVal(0,1);
    const float ax = anXStateVector.getVal(0,2);
    theXPosEstimate = x + aDelay * (vx + aDelay * ax / 2);

    //cout << "TEST: theXPosEstimate = " << theXPosEstimate <<
    //        ", myXTracker.getEstimate() = " << myXTracker.getEstimate() << endl;

    const Image<float> anYStateVector = myYTracker.getStateVector();
    const float y = anYStateVector.getVal(0,0);
    const float vy = anYStateVector.getVal(0,1);
    const float ay = anYStateVector.getVal(0,2);
    theYPosEstimate = y + aDelay * (vy + aDelay * ay / 2);

    //cout << "TEST: theYPosEstimate = " << theYPosEstimate <<
    //        ", myYTracker.getEstimate() = " << myYTracker.getEstimate() << endl;
}
