//
//	VisualCategories.cc		This file is a part of the IKAROS project
// 				<Short description of the module>
//
//    Copyright (C) 2005 Mikael Asker
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//    
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//    
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Created: 050427
//
//	<Additional description of the module>

#include "VisualCategories.h"

#include <fstream>


CategorySet::CategorySet(const char * theFileName) :
    myFileName(theFileName ? theFileName : "/CategorySet.dat")
{
    ifstream aStream;
    //cout << "Before open: aStream.isOpen() = " << aStream.is_open() << endl;
    aStream.open(myFileName.c_str());
    //cout << "After open: aStream.isOpen() = " << aStream.is_open() << endl;

    if (! aStream.is_open())
    {
        cout << "CategorySet::CategorySet : No input file \"" << myFileName << "\" found" << endl;
    }
    else
    {
        static const unsigned aBufferSize = 100;
        static char aBuffer[aBufferSize];
        strcpy(aBuffer, "");

        while (! aStream.eof() && strcmp(aBuffer, "END") != 0)
        {
            VisualFeaturesAndPositions::FeatureValueVector aVector;
            unsigned i = 0;
        
            while (i++ < VisualFeaturesAndPositions::aNumberOfFeatureValues && ! aStream.eof())
            {
                aStream.getline(aBuffer, aBufferSize);
                aVector.push_back(atof(aBuffer));
                //cout << i << ": " << aBuffer << endl;
            }

            assert(aVector.size() == VisualFeaturesAndPositions::aNumberOfFeatureValues);
            myVectors.push_back(aVector);

            assert(! aStream.eof());

            if (! aStream.eof())
            {
                aStream.getline(aBuffer, aBufferSize);
                //cout << "END: " << aBuffer << endl;
            }
        }

        aStream.close();
    }
    
    //cout << "CategorySet::CategorySet : myVectors.size() = " << myVectors.size() << endl;
}


CategorySet::CategoryIdentifier 
CategorySet::newCategory(const VisualFeaturesAndPositions::FeatureValueVector& theFeatureValues)
{
    assert(VisualFeaturesAndPositions::isAFeatureValueVector(theFeatureValues));
    myVectors.push_back(theFeatureValues);
    this->save();
    return myVectors.size() - 1;
}


CategorySet::CategoryIdentifier 
CategorySet::categorizeFeatures(const VisualFeaturesAndPositions::FeatureValueVector& theFeatureValues,
                                unsigned                                              theTickCounter,
                                Module *                                              theIkarosModule) const
{
    assert(VisualFeaturesAndPositions::isAFeatureValueVector(theFeatureValues));

    float aMinimumFeatureDistance = CategorySet::aFeatureDistanceThreshold + 1E6;
    CategoryIdentifier aBestMatch = VisualCategories::aNotACategoryIdentifier;
    const int aNumberOfVectors = myVectors.size();
    
    for (CategoryIdentifier i = 0; i < aNumberOfVectors; i++)
    {
        const float aDistance = CategorySet::featureDistance(myVectors[i], theFeatureValues);

        /*theIkarosModule->Notify(msg_print,
                                "%u CategorySet::categorizeFeatures : i = %u, aDistance = %f.\n",
                                theTickCounter,
                                i,
                                aDistance);*/

        if (aDistance < aMinimumFeatureDistance)
        {
            aMinimumFeatureDistance = aDistance;
            aBestMatch = i;
        }
    }

    static float anOldMinimumFeatureDistance = -1;
    if (aMinimumFeatureDistance != anOldMinimumFeatureDistance)
    {
        theIkarosModule->Notify(msg_print,
                                "%u CategorySet::categorizeFeatures : cat = %u, dist = %f.\n",
                                theTickCounter,
                                aBestMatch,
                                aMinimumFeatureDistance);

        anOldMinimumFeatureDistance = aMinimumFeatureDistance;
    }

    if (aMinimumFeatureDistance < CategorySet::aFeatureDistanceThreshold)
    {
        return aBestMatch;
    }
    else
    {
        return VisualCategories::aNotACategoryIdentifier;
    }
}


VisualFeaturesAndPositions::FeatureValueVector 
CategorySet::featuresOfCategory(CategoryIdentifier theCategory) const
{
    assert(0 <= theCategory && theCategory < (int)myVectors.size());
    return myVectors[theCategory];
}


/*virtual*/
CategorySet::~CategorySet() 
{}


float
CategorySet::featureDistance(const VisualFeaturesAndPositions::FeatureValueVector& theFeatureValues1,
                             const VisualFeaturesAndPositions::FeatureValueVector& theFeatureValues2)
{
    assert(VisualFeaturesAndPositions::isAFeatureValueVector(theFeatureValues1) && 
           VisualFeaturesAndPositions::isAFeatureValueVector(theFeatureValues2));

   int aFeatureIndex = 0;

    VisualFeaturesAndPositions::FeatureValueVector::const_iterator j = theFeatureValues2.begin();
    float aSum = 0;
    
    for (VisualFeaturesAndPositions::FeatureValueVector::const_iterator i = theFeatureValues1.begin(); 
         i != theFeatureValues1.end(); 
         i++)
    {
        if (CategorySet::compareFeature(aFeatureIndex))
        {
            const float aDifference = *i - *j;
            aSum += aDifference * aDifference;
        }

        j++;
        aFeatureIndex++;
    }
    
    return sqrt(aSum);
}


void
CategorySet::save() 
{
    ofstream aStream;
    //cout << "Before open: aStream.isOpen() = " << aStream.is_open() << endl;
    aStream.open(myFileName.c_str());
    //cout << "After open: aStream.isOpen() = " << aStream.is_open() << endl;

    for (FeatureValueVectorVector::iterator i = myVectors.begin(); i != myVectors.end(); i++)
    {
        if (i != myVectors.begin())
        {
            aStream << endl;
        }

        VisualFeaturesAndPositions::FeatureValueVector aVector = *i;

        for (VisualFeaturesAndPositions::FeatureValueVector::iterator j = aVector.begin(); 
             j != aVector.end(); 
             j++)
        {
            aStream << *j << endl;
        }
    }
    
    aStream << "END" << endl;
    aStream.close();
}


Module * 
VisualCategories::Create(char* theName, Parameter* theParameters)
{
    return new VisualCategories(theName, theParameters);
}


VisualCategories::VisualCategories(char* theName, Parameter* theParameters) :
    Module(theName),
    myCategorySet(GetValue(theParameters, "file")),
    myCategoryInput(0),
    myFeatureValuesInput(0),
    myLearnInput(0),
    myFilterBiasOutput(0),
    myCategoryOutput(0),
    myTickCounter(0),
    myLastLearnInputValue(1)   // > 0.5 => the first edge becomes negative
{
    // AddInput here declares one input for the module. This is the same
    // name used in the experiment configuration file.
    AddInput("CATEGORY_INPUT");
    AddInput("FEATURE_VALUES_INPUT");
    AddInput("LEARN_INPUT");

    // Here we declare an output and tell the simulator that its size will be
    // figured out later. The parameter is not strictly necessary as it is the
    // default value, but it is a good idea to include it for clarity.

    AddOutput("FILTER_BIAS_OUTPUT");
    AddOutput("CATEGORY_OUTPUT");

    SetOutputSize("FILTER_BIAS_OUTPUT", VisualFeaturesAndPositions::aNumberOfFeatureValues + 1, 2);
    SetOutputSize("CATEGORY_OUTPUT", 1);
}


void
VisualCategories::SetSizes()
{
    // Find out the size of the category input.

    const int aCategoryInputSizeX = GetInputSizeX("CATEGORY_INPUT");
    const int aCategoryInputSizeY = GetInputSizeY("CATEGORY_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (aCategoryInputSizeX == unknown_size || aCategoryInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualCategories: the category input size is not set\n");
        return;
    }

    if (aCategoryInputSizeX != 1 || aCategoryInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualCategories: the category input is not scalar\n");
        return;
    }


    // Find out the size of the feature values input.

    const int aFeatureValuesInputSizeX = GetInputSizeX("FEATURE_VALUES_INPUT");
    const int aFeatureValuesInputSizeY = GetInputSizeY("FEATURE_VALUES_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (aFeatureValuesInputSizeX == unknown_size || aFeatureValuesInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualCategories: the feature values input size is not set\n");
        return;
    }

    if (aFeatureValuesInputSizeX != VisualFeaturesAndPositions::aNumberOfFeatureValues || 
        aFeatureValuesInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualCategories: wrong size of feature values input\n");
        return;
    }


    // Find out the size of the learn input.

    const int aLearnInputSizeX = GetInputSizeX("LEARN_INPUT");
    const int aLearnInputSizeY = GetInputSizeY("LEARN_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (aLearnInputSizeX == unknown_size || aLearnInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualCategories: the learn input size is not set\n");
        return;
    }

    if (aLearnInputSizeX != 1 || aLearnInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualCategories: the learn input is not scalar\n");
        return;
    }
}


void
VisualCategories::Init()
{
    // Here we could add any number of consistency checks and do any memory
    // allocations and other initializations that wer need to do.

    // Set the variables that will be associated with the inputs and outputs
    // during the simulation. Each time the Tick method is entered, myInput
    // will point to the input vector, and myOutput will point to
    // the output vector. 

    myCategoryInput = GetInputMatrix("CATEGORY_INPUT");
    myFeatureValuesInput = GetInputMatrix("FEATURE_VALUES_INPUT");
    myLearnInput = GetInputMatrix("LEARN_INPUT");

    myFilterBiasOutput = GetOutputMatrix("FILTER_BIAS_OUTPUT");
    myCategoryOutput = GetOutputMatrix("CATEGORY_OUTPUT");
    
    **myCategoryOutput = aNotACategoryIdentifier;
}


VisualCategories::~VisualCategories()
{}


void
VisualCategories::Tick()
{
    //this->Notify(msg_print, "VisualCategories::Tick begins.\n");

    // IttiKoch in action. Input is a pointer to the current input
    // vector and output points to the output vector that will be 
    // sent along to whatever modules it is connected to.

    static int anInitCountLeft = 3;

    if (anInitCountLeft > 0)   // An init tick
    {        
        anInitCountLeft--;
    }
    else   // Not an init tick
    {
        if (**myCategoryInput == aNotACategoryIdentifier)   // Don't bias features
        {
            //this->Notify(msg_print, "VisualCategories::Tick : don't bias features.\n");
            myFilterBiasOutput[0][VisualFeaturesAndPositions::aNumberOfFeatureValues] = 0;
        }
        else   // Bias features
        {
            //this->Notify(msg_print, "VisualCategories::Tick : bias features.\n");

            const VisualFeaturesAndPositions::FeatureValueVector someFeatureValues =
                myCategorySet.featuresOfCategory(lround(**myCategoryInput));

            for (unsigned aFeatureIndex = 0; 
                 aFeatureIndex < VisualFeaturesAndPositions::aNumberOfFeatureValues; 
                 aFeatureIndex++)
            {
                myFilterBiasOutput[0][aFeatureIndex] = someFeatureValues[aFeatureIndex];

                if (CategorySet::compareFeature(aFeatureIndex))
                {
                    myFilterBiasOutput[1][aFeatureIndex] = 0;
                }
                else
                {
                    myFilterBiasOutput[1][aFeatureIndex] = 10;
                }
            }                

            myFilterBiasOutput[0][VisualFeaturesAndPositions::aNumberOfFeatureValues] = 1;
        }

        if (myLastLearnInputValue < 0.5 && **myLearnInput > 0.5)   // Positive edge
        {
            this->Notify(msg_print, "VisualCategories::Tick : positive edge detected on LEARN_INPUT.\n");

            **myCategoryOutput = myCategorySet.newCategory(
                VisualFeaturesAndPositions::featureValuesInputToVector(myFeatureValuesInput));

            this->Notify(msg_print,
                         "VisualCategories::Tick : newCategory returned %.0f.\n", 
                         **myCategoryOutput);
        }
        else
        {
            **myCategoryOutput = myCategorySet.categorizeFeatures(
                VisualFeaturesAndPositions::featureValuesInputToVector(myFeatureValuesInput), 
                myTickCounter, 
                this);
        }
        
        myLastLearnInputValue = **myLearnInput;
    }

    myTickCounter++;
    //this->Notify(msg_print, "VisualCategories::Tick ends.\n");
}
