/*!@file ArrayData.H The guts of the 2D ref-counted copy-on-write array
    implementation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ArrayData.H 4029 2004-11-12 07:14:58Z rjpeters $
//

#ifndef ARRAYDATA_H_DEFINED
#define ARRAYDATA_H_DEFINED


// Uncomment the line below to allow printing of memory alloc/free info:
//#define ARRAYDATA_DEBUG_ALLOC

#ifdef ARRAYDATA_DEBUG_ALLOC
#  include "log.H"
#endif
#include "saliency.H" // for ASSERT()
#include "TypeTraits.H"
#include "Dims.H"

#include <cstring> // for memset() and memcpy()
#include <new> // for placement new, in ArrayData's ctors

// For now we only have inline assembly for atomic integer operations
// for ix86 cpus; would be nice to eventually have analogous
// operations for ppc.
#ifdef INVT_CPU_IX86
#  define REFCOUNT_THREADSAFE
#endif

#ifdef REFCOUNT_THREADSAFE
#  include "atomic.H"
//! In threadsafe mode, the RefCountType is an integral type with atomic operations.
typedef atomic_int_t RefCountType;
//! Set ref-count \a x to the given value \a v.
#define RC_SET(x, v)          atomic_set(&x, v)
//! Increment the ref-count \a x by one.
#define RC_INCR(x)            atomic_inc(&x)
//! Decrement the ref-count \a x by one; return true if the new value is zero.
#define RC_DECR_TEST_ZERO(x)  bool(atomic_dec_and_test(&x))
                              // atomic_dec_and_test returns true if the
                              // new count is zero
//! Get the integral value of the ref-count \a x.
#define RC_GET(x)             atomic_read(&x)
#else
//! In non-threadsafe mode, the RefCountType is plain int.
typedef int RefCountType;
//! Set ref-count \a x to the given value \a v.
#define RC_SET(x, v)          (x = v)
//! Increment the ref-count \a x by one.
#define RC_INCR(x)            (++x)
//! Decrement the ref-count \a x; return true if the new value is zero.
#define RC_DECR_TEST_ZERO(x)  bool(--x == 0)
//! Get the integral value of the ref-count \a x.
#define RC_GET(x)             (x)
#endif



//! Specifies what to do with the T* used to init an ArrayData.
enum StoragePolicy
  {
    //! Make an initial copy of the data, then copy-on-write as usual.
    MAKE_OWN_COPY,

    //! Borrow the data, with write-through.
    /*! Writing to the Array will modify the T[] that initialized it. */
    WRITE_THRU,
  };

//! Specifies how to initialize the pixels of a new image.
enum InitPolicy
  {
    //! All elements are initialized to the pixel type's zero or default value.
    ZEROS,

    //! Pixels are only minimally initialized.
    /*! For trivial types, this means the pixel values are not initialized
        at all; for non-trivial user-defined types, this means each pixel
        is copy-constructed from the type's default value. */
    NO_INIT
  };


//! Implements a ref-counted 2-D array, should only be used via ArrayHandle.
/*! ArrayData implements a 2-D array, by storing:
      - a Dims representing the width+height of the array,
      - a pointer to the dynamically allocated memory where the array data
        are stored,
      - a reference count, and
      - a StoragePolicy, indicating whether the ArrayData object "owns" the
        memory and should delete[] it eventually (in the case of
        MAKE_OWN_COPY), or whether someone else "owns" the memory and will
        eventually delete[] it themselves (in the case of WRITE_THRU)

    All of these data members are immutable, with the exception of the
    reference count. This reflects the semantics that an ArrayData object is a
    fixed-size block of data; if a higher-level clients wants to resize(), it
    does so by creating a new ArrayData of the desired size, and releasing the
    old ArrayData. This simplifies the memory management: within ArrayData,
    memory is new[]'ed only in the constructor(s), and delete[]'ed only in the
    destructor.

    ArrayData provides the basics needed to use STL-style iterator algorithms.
    In particular it offers two pairs of functions to get half-open iterator
    ranges:
      - data() and end() for read-only iteration, and
      - dataw() and endw() for read-write iteration (where the "w" means
        "write" or "writable").
*/
template <class T>
class ArrayData
{
private:
  ArrayData(const ArrayData&); // not allowed; use clone() instead
  ArrayData& operator=(const ArrayData&); // not allowed; use clone() instead

public:

  // ############################################################
  // ##### Constructors
  // ############################################################

  //! Default construct an empty (i.e. 0-by-0) array.
  inline ArrayData();

  //! Construct with the specified dimensions and InitPolicy.
  inline ArrayData(const Dims& dims, const InitPolicy ip = NO_INIT);

  //! Construct with the specified dimensions.
  /*! If the passed-in T* array is non-null, copy that data into the
      constructed object. */
  inline ArrayData(const Dims& dims, const T* aa);

  //! Construct with the specified dimensions and storage policy.
  /*! If the passed-in T* array is non-null, then if the storage policy is
      MAKE_OWN_COPY we make a copy of the data, but if the storage policy is
      WRITE_THRU then we "borrow" the data. */
  inline ArrayData(const Dims& dims, T* aa, const StoragePolicy s);

  //! Return a new'ed copy of *this.
  /*! The cloned copy will always be MAKE_OWN_COPY, even if *this is
      WRITE_THRU. */
  inline ArrayData* clone() const;

  //! Release the pointed-to data if our storage policy is MAKE_OWN_COPY.
  inline ~ArrayData();

  // ############################################################
  // ##### Accessors
  // ############################################################

  //! Returns a read-only pointer to the data array.
  /*! With end(), forms an STL-style half-open const_iterator range. */
  inline const T* data() const throw();

  //! Returns a read-only pointer to one-past-the-end of the data array.
  /*! With data(), forms an STL-style half-open const_iterator range. */
  inline const T* end() const throw();

  //! Returns a read-write pointer to the data array (the "w" means "write").
  /*! With endw(), forms an STL-style half-open iterator range. */
  inline T* dataw() throw();

  //! Returns a read-write pointer to one-past-the-end of the data array.
  /*! With dataw(), forms an STL-style half-open iterator range. */
  inline T* endw() throw();

  //! Returns the dimensions (i.e. width, height) of the 2-D data array.
  inline const Dims& dims() const throw();

  //! Returns the width of the data array (i.e. the number of columns).
  inline int w() const throw();

  //! Returns the height of the data array (i.e. the number of rows).
  inline int h() const throw();

  // ############################################################
  // ##### Reference count manipulation
  // ############################################################

  //! Increment the reference count.
  inline void acquire() throw();

  //! Decrement the reference count and delete self if the count goes to zero.
  inline void release() throw();

  //! Query whether this object is shared (i.e. ref count > 1).
  inline bool isShared() const throw();

  //! For debugging/testing only.
  /*! Returns the current reference count. */
  inline int refCount() const throw();

private:
  RefCountType itsRefCount;
  StoragePolicy const itsStoragePolicy;

  Dims const itsDims;           // width+height of data array
  T* const itsData;             // data array
};


//! This class provides ref-counting and copy-on-write for ArrayData's.
/*! NOTE: Most clients will not need to use ArrayHandle<T> directly, but can
    rather use Image<T>, which wraps and expands the ArrayHandle
    interface. Nevertheless, these notes may be of interest because the
    ref-counting and copy-on-write semantics that ArrayHandle<T> offers are
    transferred through to Image<T>.

    As long as clients access ArrayData<T> through an ArrayHandle<T>, the
    proper copy-on-write semantics are guaranteed. This is by virtue of the
    fact that the only way to extract a non-const ArrayData<T>& from an
    ArrayHandle<T> is to call uniq(), which in turn checks to see if the
    ArrayData is shared (i.e. ref count > 1). If it is shared, then that
    ArrayData is clone()'d before returning a reference to it.

    The only caveat: you should not store non-const references to an ArrayData
    or to the elements of an ArrayData object for "long" periods of time. In
    this case, "long" means "longer than you are sure you have a unique
    reference", which in practice means "you can keep such references up until
    a new copy of the ArrayHandle<T> is made". This goes along with sensible
    coding practices anyway; a typical usage pattern would be:

      -# you have an ArrayHandle<T> (or an Image<T>, which holds an
         ArrayHandle<T>) that you want to modify
      -# get a pair of iterator's for that object so that you can perform your
         series of write operations
      -# don't use those iterators anymore, and then you can safely make
         copies of the freshly-modified ArrayHandle<T> or Image<T> object (for
         example to return a copy of it as the return value from a function)
*/
template<class T>
class ArrayHandle
{
public:
  //! Default construct with an empty array (i.e. 0-by-0).
  inline ArrayHandle();

  //! Construct with a new ArrayData, which the ArrayHandle now owns.
  inline explicit ArrayHandle(const ArrayData<T>* p) throw();

  //! Copy construct; no deep-copy is incurred, just bumps the ref-count.
  inline ArrayHandle(const ArrayHandle& r) throw();

  //! Destruct; delete's the ArrayData<T> if its ref count goes to zero.
  inline ~ArrayHandle() throw();

  //! Swap contents with another ArrayHandle.
  inline void swap(ArrayHandle<T>& other) throw();

  //! Assignment; no deep-copy is incurred, just bumps the ref-count.
  inline ArrayHandle& operator=(const ArrayHandle& r) throw();

  //! Get a read-only reference to the ArrayData object.
  /*! This is a "cheap" operation, since the ArrayData will never be copied in
      this operation. */
  inline const ArrayData<T>& get() const throw();

  //! Get a unique read-write reference to the ArrayData object.
  /*! This is a potentially "expensive" operation, since in order to get a
      unique ArrayData object, the current one may have to be clone()'d if it
      is shared (i.e. ref-count > 1). */
  inline ArrayData<T>& uniq();

  //! See if the ArrayData object is shared (i.e. ref-count > 1).
  inline bool isShared() const throw();

  //! For testing/debugging only.
  /*! See if we are pointing to the same ArrayData<T> as is the other
      ArrayHandle<T>. */
  inline bool hasSameData(const ArrayHandle<T>& b) const throw();

  //! For testing/debugging only.
  /*! Returns the current reference count of the our ArrayData<T>. */
  inline int refCount() const throw();

private:
  ArrayData<T>* px;     // contained pointer
};



// ######################################################################
// ######################################################################
// INLINE FUNCTIONS for ArrayData:
// ######################################################################
// ######################################################################

//! This is a helper template struct for ArrayData's constructors.
/*! It exposes functions for initializing raw memory in different
    ways. These functions can then be specialized for builtin or trivial
    types (i.e., types for which TypeTraits<T>::isTrivial is true) for
    maximum efficiency. We get this specialization by having an extra bool
    isTrivial template parameter for ArrayHelper, which allows us to select
    at compile-time based on the complexity of T. This is much better than
    using a run-time "if" statement in the constructors, since that incurs
    both a time cost (to evaluate an "if" statement) and a code space cost
    (since we're then including code that won't be used). */
template <class T, bool isTrivial = TypeTraits<T>::isTrivial>
struct ArrayHelper
{
  //! Construct \a n copies of \a val at \a space using placement new.
  /*! For general types, the minimal initialization that we can do is to
      zero_initialize() (i.e., unlike for builtin or trivial types, we
      can't just leave the memory sitting as it is, uninitialized). */
  static void minimal_initialize(T* space, int n)
  {
    zero_initialize(space, n);
  }

  //! Initialize the memory to T's default or zero value.
  /*! For general types, the best way to do this is to make one default T
      object, and then copy construct each element of the array from that
      object using placement new. */
  static void zero_initialize(T* space, int n)
  {
    T zero = T();
    T* cur = space;
    try
      {
        for (; n > 0; --n, ++cur)
          new (cur) T(zero); // placement new
      }
    // if a constructor call throws, we have to clean up manually:
    catch (...)
      {
        while (cur != space) (cur--)->~T(); // explicit destructor call
        ::operator delete(space);
        throw; // rethrow the exception
      }
  }

  //! Initialize the memory by copying elements from the \a src array.
  /*! For general types, we just copy construct each element in the target
      array from the corresponding element in the \a src array, using
      placement new. */
  static void copy_initialize(T* space, int n, const T* src)
  {
    T* cur = space;
    try
      {
        for (; n > 0; --n, ++cur)
          new (cur) T(*src++); // placement new
      }
    // if a constructor call throws, we have to clean up manually:
    catch (...)
      {
        while (cur != space) (cur--)->~T(); // explicit destructor call
        ::operator delete(space);
        throw; // rethrow the exception
      }
  }

  //! Call destructors on the \a n objects stored in \a space.
  static void destruct(T* space, int n)
  {
    while (--n >= 0)
      (space+n)->~T(); // explicit destructor call
  }
};

//! Specialization of ArrayHelper for builtin/trivial types.
/*! We can do things much faster with these kinds of types, using
    e.g. memset() and memcpy(), and we don't have to do default
    initialization at all. */
template <class T>
struct ArrayHelper<T, true>
{
  //! For trivial types, we don't have to do anything here.
  static void minimal_initialize(T* /*space*/, int /*n*/) {}

  //! For trivial types, we just memset() all the memory to zero.
  static void zero_initialize(T* space, int n)
  {
    memset(space, 0, n * sizeof(T));
  }

  //! For trivial types, we just memcpy() into the destination.
  static void copy_initialize(T* space, int n, const T* src)
  {
    memcpy(space, src, n * sizeof(T));
  }

  //! For trivial types, no destruction is needed.
  static void destruct(T* /*space*/, int /*n*/) {}
};

// ######################################################################
template <class T> inline
ArrayData<T>::ArrayData() :
  itsRefCount(),
  itsStoragePolicy(MAKE_OWN_COPY),
  itsDims(0, 0),
  itsData(static_cast<T*>(::operator new(1))) // just ask for one byte
{
  RC_SET(itsRefCount, 0);

#ifdef ARRAYDATA_DEBUG_ALLOC
  LDEBUG("N %lx", itsData);
#endif
}

// ######################################################################
template <class T> inline
ArrayData<T>::ArrayData(const Dims& dims, const InitPolicy ip) :
  itsRefCount(),
  itsStoragePolicy(MAKE_OWN_COPY),
  itsDims(dims),
  itsData(static_cast<T*>(::operator new(itsDims.sz() * sizeof(T))))
{
  if (ip == ZEROS)
    ArrayHelper<T>::zero_initialize(itsData, itsDims.sz());
  else
    ArrayHelper<T>::minimal_initialize(itsData, itsDims.sz());

  RC_SET(itsRefCount, 0);

#ifdef ARRAYDATA_DEBUG_ALLOC
  LDEBUG("N %lx", itsData);
#endif
}

// ######################################################################
template <class T> inline
ArrayData<T>::ArrayData(const Dims& dims, const T* aa) :
  itsRefCount(),
  itsStoragePolicy(MAKE_OWN_COPY),
  itsDims(dims),
  itsData(static_cast<T*>(::operator new(itsDims.sz() * sizeof(T))))
{
  ArrayHelper<T>::copy_initialize(itsData, itsDims.sz(), aa);

  RC_SET(itsRefCount, 0);

#ifdef ARRAYDATA_DEBUG_ALLOC
  LDEBUG("N %lx", itsData);
#endif
}

// ######################################################################
template <class T> inline
ArrayData<T>::ArrayData(const Dims& dims, T* aa, const StoragePolicy s) :
  itsRefCount(),
  itsStoragePolicy(s),
  itsDims(dims),
  itsData(s == MAKE_OWN_COPY
          ? static_cast<T*>(::operator new(itsDims.sz() * sizeof(T)))
          : aa)
{
  if (s == MAKE_OWN_COPY)
    {
      if (itsDims.sz() != 0 && aa != 0)
        ArrayHelper<T>::copy_initialize(itsData, itsDims.sz(), aa);
      else
        ArrayHelper<T>::minimal_initialize(itsData, itsDims.sz());
    }

  RC_SET(itsRefCount, 0);

#ifdef ARRAYDATA_DEBUG_ALLOC
  LDEBUG("N %lx", itsData);
#endif
}

// ######################################################################
template <class T> inline
ArrayData<T>* ArrayData<T>::clone() const
{
  ArrayData<T>* ret = new ArrayData(itsDims, itsData);
#ifdef ARRAYDATA_DEBUG_ALLOC
  LDEBUG("New %lx", ret);
#endif
  return ret;
}

// ######################################################################
template <class T> inline
ArrayData<T>::~ArrayData()
{
  if (itsStoragePolicy == MAKE_OWN_COPY)
    {
#ifdef ARRAYDATA_DEBUG_ALLOC
      LDEBUG("D %lx", itsData);
#endif
      ArrayHelper<T>::destruct(itsData, itsDims.sz());
      ::operator delete(itsData);
    }
}

// ######################################################################
template <class T> inline
const T* ArrayData<T>::data() const throw()
{ return itsData; }

// ######################################################################
template <class T> inline
const T* ArrayData<T>::end() const throw()
{ return itsData + itsDims.sz(); }

// ######################################################################
template <class T> inline
T* ArrayData<T>::dataw() throw()
{ return itsData; }

// ######################################################################
template <class T> inline
T* ArrayData<T>::endw() throw()
{ return itsData + itsDims.sz(); }

// ######################################################################
template <class T> inline
const Dims& ArrayData<T>::dims() const throw()
{ return itsDims; }

// ######################################################################
template <class T> inline
int ArrayData<T>::w() const throw()
{ return itsDims.w(); }

// ######################################################################
template <class T> inline
int ArrayData<T>::h() const throw()
{ return itsDims.h(); }

// ######################################################################
template <class T> inline
void ArrayData<T>::acquire() throw()
{
  RC_INCR(itsRefCount);
}

// ######################################################################
template <class T> inline
void ArrayData<T>::release() throw()
{
  if ( RC_DECR_TEST_ZERO(itsRefCount) ) delete this;
}

// ######################################################################
template <class T> inline
bool ArrayData<T>::isShared() const throw()
{
  return (RC_GET(itsRefCount) > 1);
}

// ######################################################################
template <class T> inline
int ArrayData<T>::refCount() const throw()
{
  return RC_GET(itsRefCount);
}

// ######################################################################
// ######################################################################
// INLINE FUNCTIONS for ArrayHandle:
// ######################################################################
// ######################################################################

// ######################################################################
template <class T> inline
ArrayHandle<T>::ArrayHandle() : px(new ArrayData<T>)
{
  px->acquire();
}

// ######################################################################
template <class T> inline
ArrayHandle<T>::ArrayHandle(const ArrayData<T>* p) throw() :
  // this cast is OK because we maintain const-correctness through
  // copy-on-write
  px(const_cast<ArrayData<T>*>(p))
{
  px->acquire();
}

// ######################################################################
template <class T> inline
ArrayHandle<T>::ArrayHandle(const ArrayHandle& r) throw() :
  px(r.px)
{
  px->acquire();
}

// ######################################################################
template <class T> inline
ArrayHandle<T>::~ArrayHandle() throw()
{
  px->release();
}

// ######################################################################
template <class T> inline
void ArrayHandle<T>::swap(ArrayHandle<T>& other) throw()
{
  ArrayData<T>* other_px = other.px;
  other.px = this->px;
  this->px = other_px;
}

// ######################################################################
template <class T> inline
ArrayHandle<T>& ArrayHandle<T>::operator=(const ArrayHandle& r) throw()
{
  ArrayHandle tmp(r); this->swap(tmp); return *this;
}

// ######################################################################
template <class T> inline
const ArrayData<T>& ArrayHandle<T>::get() const throw()
{ return *px; }

// ######################################################################
template <class T> inline
ArrayData<T>& ArrayHandle<T>::uniq()
{
  if (px->isShared())
    {
      ArrayHandle copy(px->clone());
      this->swap(copy);
    }

  return *px;
}

// ######################################################################
template <class T> inline
bool ArrayHandle<T>::isShared() const throw()
{ return px->isShared(); }

// ######################################################################
template <class T> inline
bool ArrayHandle<T>::hasSameData(const ArrayHandle<T>& b) const throw()
{ return px==b.px; }

// ######################################################################
template <class T> inline
int ArrayHandle<T>::refCount() const throw()
{ return px->refCount(); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !ARRAY_IMPL_H_DEFINED
