/*!@file AttentionGuidanceMap.C Implementation for task-relevance map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: AttentionGuidanceMap.C 3092 2004-02-06 09:29:35Z itti $
//

#include "saliency.H"
#include "Image.H"
#include "Image_ShapeOps.H"  // for rescale()
#include "AttentionGuidanceMap.H"
#include "log.H"
#include "ChannelBase.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "FrameSeries.H"
#include "ShapeEstimator.H"
#include "Brain.H"

// ######################################################################
// ######################################################################
// ########## AttentionGuidanceMap implementation
// ######################################################################
// ######################################################################

AttentionGuidanceMap::AttentionGuidanceMap(ModelManager& mgr,
                                   const std::string& descrName,
                                   const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSaveResults(this, "AGMsaveResults", false), // see ModelOptionDefs.C
  itsAGMFactor(this, "AGMfactor", 50000.0F),
  itsBrain(NULL)
{ }

// ######################################################################
AttentionGuidanceMap::~AttentionGuidanceMap()
{ }

// ######################################################################
void AttentionGuidanceMap::exportOptions(const int whatToExport,
                                     const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsAGMFactor);

  if (whatToExport & OPTEXP_SAVE)
    manager->requestOption(this, itsSaveResults);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void AttentionGuidanceMap::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  if (itsSaveResults.getVal())
    ofs->writeFloat(getV(true), PNM, -1, true, "AGM");
}

// ######################################################################
void AttentionGuidanceMap::setBrain(const SharedPtr<Brain>& brain)
{ itsBrain = brain; }

// ######################################################################
// ######################################################################
// ########## AttentionGuidanceMapConfigurator implementation
// ######################################################################
// ######################################################################
AttentionGuidanceMapConfigurator::
AttentionGuidanceMapConfigurator(ModelManager& mgr,
                                 const std::string& descrName,
                                 const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsAGMtype(this, "AttentionGuidanceMapType", "None"),
  itsAGM(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
AttentionGuidanceMapConfigurator::~AttentionGuidanceMapConfigurator()
{  }

// ######################################################################
void AttentionGuidanceMapConfigurator::exportOptions(const int whatToExport,
                                                     const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsAGMtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<AttentionGuidanceMap>
AttentionGuidanceMapConfigurator::getAGM() const
{ return itsAGM; }

// ######################################################################
bool AttentionGuidanceMapConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsAGMtype.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsAGM.get()) { removeSubComponent(itsAGM); itsAGM.reset(NULL); }

    // instantiate a SM of the appropriate type:
    if (value.compare("None") == 0)              // no TRM
      itsAGM.reset(NULL);
    else if (value.compare("Std") == 0)          // standard
      itsAGM.reset(new AttentionGuidanceMapStd(*manager));
    else
      LFATAL("Unknown AGM type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsAGM.get()) {
      addSubComponent(itsAGM);

      // tell the controller to export its options:
      itsAGM->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected AGM of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ########## AttentionGuidanceMapStd implementation
// ######################################################################
// ######################################################################

// ######################################################################
AttentionGuidanceMapStd::
AttentionGuidanceMapStd(ModelManager& mgr, const std::string& descrName,
                        const std::string& tagName) :
  AttentionGuidanceMap(mgr, descrName, tagName),
  itsBUmap(), itsTDmap()
{  }

// ######################################################################
AttentionGuidanceMapStd::~AttentionGuidanceMapStd()
{ }

// ######################################################################
void AttentionGuidanceMapStd::reset()
{ itsBUmap.freeMem(); itsTDmap.freeMem(); }

// ######################################################################
void AttentionGuidanceMapStd::inputBU(const Image<float>& current)
{ itsBUmap = current; }

// ######################################################################
void AttentionGuidanceMapStd::inputTD(const Image<float>& current)
{ itsTDmap = current; }

// ######################################################################
Image<float> AttentionGuidanceMapStd::getV(const bool normalize) const
{
  if (itsBUmap.isSameSize(itsTDmap) == false)
    LFATAL("Bottom-up %dx%d vs. top-down %dx%d dims mismatch -- ABORT",
           itsBUmap.getWidth(), itsBUmap.getHeight(),
           itsTDmap.getWidth(), itsTDmap.getHeight());
  Image<float> ret = itsBUmap * itsTDmap;
  if (normalize)
    {
      if (itsAGMFactor.getVal() > 0.0F) ret *= itsAGMFactor.getVal();
      else ret.normalize(0.0F, 255.0F);
    }
  return ret;
}

// ######################################################################
void AttentionGuidanceMapStd::evolve(const double t)
{ }

// ######################################################################
float AttentionGuidanceMapStd::getV(const Point2D& p) const
{ return itsBUmap.getVal(p) * itsTDmap.getVal(p); }

// ######################################################################
int AttentionGuidanceMapStd::getWidth() const
{ return itsBUmap.getWidth(); }

// ######################################################################
int AttentionGuidanceMapStd::getHeight() const
{ return itsBUmap.getHeight(); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
