/*!@file AttentionGuidanceMap.H Class declarations for task-relevance map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: AttentionGuidanceMap.H 2891 2003-10-08 20:30:08Z itti $
//

#ifndef ATTENTIONGUIDANCEMAP_H_DEFINED
#define ATTENTIONGUIDANCEMAP_H_DEFINED

#include "Image.H"
#include "ModelComponent.H"
class ModelManager;
class OutputFrameSeries;
class Brain;

// ######################################################################
//! The task-relevance map base class
// ######################################################################
/*! This is a 2D task-relevance map. It is just a base class with
  virtual function definitions. Various implementations are available
  below.  The TRM is a topographic map where a value of zero indicates
  that a location is not relevant to the task at hand, while a value
  of 1.0 indicates neutral relevance and values higher than 1.0
  indicate high relevance. */ 
class AttentionGuidanceMap : public ModelComponent
{
 public:
  // ######################################################################
  //! @name Constructor, destructor, and reset
  //@{

  //! Ininitialized constructor
  /*! The map will be resized and initialized the first time input() is
    called */
  AttentionGuidanceMap(ModelManager& mgr,
                   const std::string& descrName = "Attention Guidance Map",
                   const std::string& tagName = "AttentionGuidanceMap");
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. A base
    implementation is provided that will only export the base class
    options (see the ModelParam definitions in the protected section
    below). */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);
  //! Destructor
  virtual ~AttentionGuidanceMap();

  //! Reset to initial state just after construction
  virtual void reset() = 0;

  //! Setup a hook back to our brain
  void setBrain(const SharedPtr<Brain>& brain);

  //@}

  // ######################################################################
  //! @name Input, evolve and output
  //@{

  //! Set new bottom-up input
  /*! This should initialize and resize the map if it is currently
      uninitialized (e.g., just after construction or reset()). */
  virtual void inputBU(const Image<float>& in) = 0;

  //! Set new top-down input
  /*! This should initialize and resize the map if it is currently
      uninitialized (e.g., just after construction or reset()). */
  virtual void inputTD(const Image<float>& in) = 0;

  //! Get width of the map
  virtual int getWidth() const = 0;

  //! Get height of the map
  virtual int getHeight() const = 0;

  //! Return all our relevance values as an Image<float>
  /*! The current values are possibly normalized to 0..255. The
    normalization will use itsTRMFactor if it is non-null */
  virtual Image<float> getV(const bool normalize = false) const = 0;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const = 0;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t) = 0;

  //@}

  // ######################################################################
  //! @name Results
  //@{

  //! Save our various results
  /*! A default implementation is provided, which just calls
      getV() and saves that map with a file name prefix of "SM".
      @param ofs will figure out file name/format and save our
      results. */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

  // ######################################################################
protected:
  //! Save our internals when saveResults() is called?
  ModelParam<bool> itsSaveResults;

  //! Multiply by a factor rather than normalize when doing getV()?
  ModelParam<float> itsAGMFactor;

  //! Our brain, or SharedPtr<NULL> if it has not been set with setBrain()
  SharedPtr<Brain> itsBrain;

private:
  // forbid assignment and copy-construction:
  AttentionGuidanceMap& operator=(const AttentionGuidanceMap& sm);
  AttentionGuidanceMap(const AttentionGuidanceMap& sm);
};

// ######################################################################
//! AttentionGuidanceMap configurator
// ######################################################################
/*! This will export the --agm-type=XX command-line option and will
  instantiate an AGM of the desired type as the option gets assigned a
  value. As this happens, new options may become available in the
  command-line. To see them, use --help AFTER you have chosen the type
  to use. The current AGM may be retrieved using getAGM(). */
class AttentionGuidanceMapConfigurator : public ModelComponent
{
public:
  //! Constructor
  AttentionGuidanceMapConfigurator(ModelManager& mgr,
                               const std::string& descrName =
                               "Attention Guidance Map Configurator",
                               const std::string& tagName =
                               "AttentionGuidanceMapConfigurator");

  //! destructor
  virtual ~AttentionGuidanceMapConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SM once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen AGM
  /*! You should call this during start() of the ModelComponent that
    needs the AGM. CAUTION: this may be a SharedPtr(NULL) if no
    AGM type has been chosen. */
  SharedPtr<AttentionGuidanceMap> getAGM() const;

protected:
  ModelParam<std::string> itsAGMtype; //!< type of map

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<AttentionGuidanceMap> itsAGM; // the map
  int itsWhatToExport;
};


// ######################################################################
//! The standard task-relevance map
// ######################################################################
/*! This is our current standard AGM implementation. It just takes the
  pointwise product between bottom-up and top-down inputs. */
class AttentionGuidanceMapStd : public AttentionGuidanceMap
{
 public:
  //! Uninitialized constructor
  AttentionGuidanceMapStd(ModelManager& mgr, const std::string& descrName =
                      "Task-Relevance Map Std",
                      const std::string& tagName =
                      "AttentionGuidanceMapStd");

  //! Destructor
  virtual ~AttentionGuidanceMapStd();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new bottom-up input
  virtual void inputBU(const Image<float>& in);

  //! Set new top-down input
  virtual void inputTD(const Image<float>& in);

  //! Get width of the map
  virtual int getWidth() const;

  //! Get height of the map
  virtual int getHeight() const;

  //! Return all our relevance values as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t);

private:
  Image<float> itsBUmap;
  Image<float> itsTDmap;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
