/*!@file Beowulf.C Simple interfacing to a Beowulf cluster */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Beowulf.C 3515 2004-08-18 23:23:29Z itti $
//

#include "Beowulf.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include <cmath>    // for floor()
#include <cstdio>   // for fseek()

// ######################################################################
Beowulf::Beowulf(ModelManager& mgr, const std::string& descrName,
                 const std::string& tagName, const bool ismaster) :
  ModelComponent(mgr, descrName, tagName),
  itsSlaveNames(this, "BeowulfSlaveNames", ""),
  isMaster(this, "BeowulfMaster", ismaster),
  selfqlen(this, "BeowulfSelfQlen", 0),
  selfdroplast(this, "BeowulfSelfDropLast", false),
  com(new TCPcommunicator(mgr)), initialized(false),
  nbnode(0), nodename(NULL), node2fd(NULL), fd2node(NULL), master(0), me(0),
  nextAvailableNode(0), slaveETI(NULL), slaveETIreceived(NULL),
  tim(1000000),  // microsecond accuracy
  selfmsg()
{
  // initialize our self-addressed message queue mutex:
  pthread_mutex_init(&mutselfmsg, NULL);

  // make our TCPcommunicator a subcomponent of us:
  addSubComponent(com);
}

// ######################################################################
Beowulf::~Beowulf()
{ pthread_mutex_destroy(&mutselfmsg); }

// ######################################################################
void Beowulf::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    if (isMaster.getVal()) manager->requestOption(this, itsSlaveNames);
    manager->requestOption(this, isMaster, true); // use my value
    manager->requestOption(this, selfqlen);
    manager->requestOption(this, selfdroplast);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
bool Beowulf::setModelParamString(const std::string& name,
                                  const std::string& value,
                                  const bool recurse, const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of masterhood?
  if (name.compare(isMaster.getName()) == 0)
    {
      if (isMaster.getVal())
        { me = -1; manager->requestOption(this, itsSlaveNames); }
      else
        { me = 0; /* should un-request the option? */ }
    }

  return ret;
}

// ######################################################################
void Beowulf::start2()
{
  if (isMaster.getVal()) masterInit(itsSlaveNames.getVal().c_str());
  else slaveInit();
}

// ######################################################################
void Beowulf::stop1()
{
  resetConnections(-1);
}

// ######################################################################
int Beowulf::getNbSlaves() const
{ return nbnode; }

// ######################################################################
int Beowulf::getNodeNumber() const
{ return me; }

// ######################################################################
const char* Beowulf::nodeName(const int nb) const
{
  // is it the master?
  if (nb == -1) return "BeoMaster";
  if (nb >= nbnode)
    LFATAL("Node number %d exceeds number of nodes: %d", nb, nbnode);
  return nodename[nb];
}

// ######################################################################
int Beowulf::requestNode()
{
  if (started() == false) LFATAL("I am not started -- ABORT");
  if (nextAvailableNode >= nbnode)
    { LERROR("I am out of nodes -- RETURNING -2"); return -2; }

  return nextAvailableNode ++;
}

// ######################################################################
void Beowulf::resetConnections(const int masterfd)
{
  if (nodename)
    {
      for (int i = 0; i < nbnode; i ++) delete [] nodename[i];
      delete [] nodename;
      nodename = NULL;
    }
  if (node2fd) { delete [] node2fd; node2fd = NULL; }
  if (fd2node) { delete [] fd2node; fd2node = NULL; }
  if (slaveETI) { delete [] slaveETI; slaveETI = NULL; }
  if (slaveETIreceived) { delete [] slaveETIreceived; slaveETIreceived=NULL; }
  if (masterfd == -1) com->terminateAll();
  else com->terminateAllButOne(masterfd);
  initialized = false; tim.reset(); selfmsg.clear();
}

// ######################################################################
void Beowulf::masterInit(const int nb_nodes, char** node_names)
{
  resetConnections(-1);  // reset possible existing communications
  master = -1;  // I have no master, so I am the master
  me = -1;      // yes, I am the master

  // setup table of all node nostnames:
  nbnode = nb_nodes;
  typedef char* charptr;
  nodename = new charptr[nbnode];
  for (int i = 0; i < nbnode; i ++)
    {
      nodename[i] = new char[strlen(node_names[i]) + 1];
      strcpy(nodename[i], node_names[i]);
    }

  // setup initial communication with all nodes:
  node2fd = new int[nbnode];
  fd2node = new int[FD_SETSIZE];
  for (int i = 0; i < FD_SETSIZE; i ++) fd2node[i] = -1;

  for (int i = 0; i < nbnode; i ++)
    {
      node2fd[i] = com->contact(nodename[i]);
      //LINFO("Linked Node FD %d node %d", node2fd[i], i);
      if (node2fd[i] == -1) LFATAL("Failed contacting %s", nodename[i]);
      fd2node[node2fd[i]] = i;
    }

  // setup initial arrays of ETIs. We assume that all nodes are idle:
  slaveETI = new float[nbnode];
  slaveETIreceived = new float[nbnode];
  for (int i = 0; i < nbnode; i ++)
    { slaveETI[i] = 0.0F; slaveETIreceived[i] = 0.0F; }
  tim.reset();

  // now, send an initial message to each node that tells them what the
  // array of slaves is, so that they can allocate memory for it and
  // initialize it:
  for (int i = 0; i < nbnode; i ++)
    {
      // send a message with the names of all nodes and the index of the
      // node of interest:
      TCPmessage msg(0, BEO_INIT);
      char buf[2048]; sprintf(buf, "%d %d ", nbnode, i);
      for (int j = 0; j < nbnode; j ++)
        { strcat(buf, nodename[j]); strcat(buf, " "); }
      msg.addString(buf);
      com->send(node2fd[i], msg);
    }

  // wait for a message of type INIT from each node, acknowledging that they
  // have initialized their array of peers:
  int nodeok = 0;
  while(nodeok < nbnode)
    {
      TCPmessage msg; int rfd = -1;  // receive from any node
      if (com->receive(rfd, msg, 5))  // block up to 5ms
        {
          if (msg.getAction() != BEO_INIT)
            LERROR("Bogus message [%d, %d] ignored!",
                   msg.getID(), msg.getAction());
          else
            nodeok ++;
        }
    }

  // now, send an order to each node to contact its peers:
  TCPmessage msg(0, BEO_INIT2);
  for (int i = 0; i < nbnode; i ++) com->send(node2fd[i], msg);

  // wait for a message of type INIT3 from each node, acknowledging that they
  // have contacted their peers:
  nodeok = 0;
  while(nodeok < nbnode)
    {
      TCPmessage msg; int rfd = -1;  // receive from any node
      if (com->receive(rfd, msg, 5))  // block up to 5ms
        {
          if (msg.getAction() != BEO_INIT3)
            LERROR("Bogus message [%d, %d] ignored!",
                   msg.getID(), msg.getAction());
          else
            {
              nodeok ++;
              LINFO("Slave node %d [%s] is configured.", msg.getID(),
                    nodename[msg.getID()]);
            }
        }
    }

  // ready to transmit/receive messages:
  initialized = true;
  LINFO("Initialization complete -- all slaves ready!");
}

// ######################################################################
void Beowulf::masterInit(const char* node_names)
{
  ASSERT(node_names);
  int nbn = 0;
  typedef char* charptr; char **nn;

  if (node_names[0] == '/') // file that contains the node list
    {
      // count how many nodes:
      LINFO("Using Beowulf slaves from file %s", node_names);
      char line[1024];
      FILE *fil = fopen(node_names, "r");
      if (fil == NULL) LFATAL("Cannot open slave node file %s", node_names);
      while(fgets(line, 1024, fil)) nbn ++;

      // create list of node names:
      fseek(fil, 0, SEEK_SET); // back to start of file
      int no = 0;
      nn = new charptr[nbn];
      while(fgets(line, 1024, fil))
        {
          line[strlen(line) - 1] = '\0';  // eliminate LF
          nn[no] = new char[strlen(line) + 1];
          strcpy(nn[no], line);
          no ++;
        }
      fclose(fil);
    }
  else
    {                       // node list as n01:9567,n02:9567, ...
      LINFO("Using Beowulf slaves %s", node_names);
      // count how many nodes:
      int idx = 0, slen = strlen(node_names);
      while(idx < slen) {
        if (node_names[idx] == ' ' || node_names[idx] == ',' ||
            idx == slen - 1) nbn ++;
        idx ++;
      }
      
      // create list of node names:
      nn = new charptr[nbn];
      idx = 0; int idx2 = 0, no = 0;
      while(idx <= slen) {
        if (idx == slen || node_names[idx] == ' ' || node_names[idx] == ',') {
          nn[no] = new char[idx - idx2 + 1];
          strncpy(nn[no], node_names + idx2, idx - idx2);
          nn[no][idx - idx2] = '\0';
          no ++; idx2 = idx + 1;
        }
        idx ++;
      }
    }

  // do the initialization:
  masterInit(nbn, nn);

  // delete local vars:
  for (int i = 0; i < nbn; i ++) delete [] nn[i];
  delete [] nn;
}

// ######################################################################
void Beowulf::slaveInit()
{
  int nodeok = 1;
  LINFO("waiting for master...");

  // wait for my master to contact me, then contact other nodes and tell
  // master when I am ready:
  while(initialized == false)
    {
      TCPmessage msg; int rfd = -1;  // receive from any open socket
      if (com->receive(rfd, msg, 5))  // block up to 5ms
        {
          int32 id = msg.getID();
          int32 action = msg.getAction();
          //LDEBUG("Received message [%d, %d] from fd=%d", id, action, rfd);
          switch(action)
            {
            case BEO_INIT: // ##############################
              {
                // this message comes from our master:
                master = rfd;

                // start the (re)-initialization:
                slaveReInit(msg);
              }
              break;
            case BEO_INIT2:
              {
                // ok, now all the slaves should know what the array
                // of slaves is.  Let's contact our peers: send a
                // message of type INIT3 to all nodes with number
                // above mine:
                TCPmessage msg2(me, BEO_INIT3); nodeok = 1;
                for (int i = me + 1; i < nbnode; i ++)
                  {
                    node2fd[i] = com->contact(nodename[i]);
                    if (node2fd[i] == -1)
                      LFATAL("Failed contacting %s -- ABORT", nodename[i]);
                    fd2node[node2fd[i]] = i;
                    com->send(node2fd[i], msg2);
                    nodeok ++; // node[i] is initialized and ok
                  }
                // have we contacted everybody we needed to?
                if (nodeok == nbnode) initialized = true;
              }
              break;
            case BEO_INIT3: // ##############################
              {
                // we received a message that will allow us to fill up
                // blanks in our node[] array:
                node2fd[id] = rfd; fd2node[rfd] = id;
                nodeok ++;

                // have we contacted everybody we needed to?
                if (nodeok == nbnode) initialized = true;
              }
              break;
            default: // ##############################
              LERROR("Bogus action %d -- IGNORING.", action);
              break;
            }
        }
    }

  // ok, we are ready; tell master:
  TCPmessage msg(me, BEO_INIT3);
  com->send(master, msg);
  LINFO("Initialization complete -- all connections ready!");

  char xx[1024], yy[256]; xx[0] = '\0';
  for (int i = 0; i < nbnode; i ++)
    if (i == me) strcat(xx, "me ");
    else { sprintf(yy, "%02d ", node2fd[i]); strcat(xx, yy); }
  LINFO("NODES = [ %s]", xx);
}

// ######################################################################
void Beowulf::slaveReInit(TCPmessage& rmsg)
{
  // kill all existing communications except towards master:
  resetConnections(master);

  // decode the message:
  char *buf, *buf2, *buf3; int32 typ;
  rmsg.getElement((void **)&buf, typ);
  ASSERT(typ == TCPMSG_STRING);
  nbnode = strtol(buf, &buf2, 10); buf2 ++;
  me = strtol(buf2, &buf3, 10); buf3 ++;
  typedef char* charptr;
  nodename = new charptr[nbnode];
  node2fd = new int[nbnode];
  fd2node = new int[FD_SETSIZE];
  for (int i = 0; i < FD_SETSIZE; i ++) fd2node[i] = -1;

  for (int i = 0; i < nbnode; i ++)
    {
      node2fd[i] = -1;  // no known node here yet...
      buf2 = buf3; while(*buf3 != ' ' && *buf3 != '\0') buf3 ++;
      *buf3++ = '\0'; nodename[i] = new char[strlen(buf2) + 1];
      strcpy(nodename[i], buf2);
    }

  // remember to deallocate contents of decoded TCPmessage:
  delete [] buf;

  // send an acknowledgment to master:
  LINFO("INIT with %d nodes, me = %d [%s]", nbnode, me, nodename[me]);
  TCPmessage msg(me, BEO_INIT);
  com->send(master, msg);
}

// ######################################################################
void Beowulf::send(const int node_nb, TCPmessage& msg)
{
  ASSERT(initialized); ASSERT(node_nb >= -1 && node_nb < nbnode);
  if (node_nb == me)       // send to ourselves?
    {
      LDEBUG("Sending msg [%d, %d] to myself", msg.getID(), msg.getAction());
      if (selfqlen.getVal() > 0 && int(selfmsg.size()) >= selfqlen.getVal())
        {
          if (selfdroplast.getVal())
            LERROR("Self-message queue full -- DROPPING MOST RECENT MESSAGE");
          else
            {
              LERROR("Self-message queue full -- DROPPING LEAST RECENT "
                     "MESSAGE");
              pthread_mutex_lock(&mutselfmsg);
              selfmsg.pop_front();
              selfmsg.push_back(msg);
              pthread_mutex_unlock(&mutselfmsg);
            }
        }
      else
        {
          pthread_mutex_lock(&mutselfmsg);
          selfmsg.push_back(msg);
          pthread_mutex_unlock(&mutselfmsg);
        }
    }
  else if (node_nb == -1)  // want to send to my master
    {
      LDEBUG("Sending msg [%d, %d] to master", msg.getID(), msg.getAction());
      com->send(master, msg);
    }
  else
    {
      LDEBUG("Sending msg [%d, %d] to node %d [%s] [%d]", msg.getID(),
             msg.getAction(), node_nb, nodename[node_nb], node2fd[node_nb]);
      com->send(node2fd[node_nb], msg);
    }
}

// ######################################################################
void Beowulf::send(TCPmessage& msg)
{
  ASSERT(initialized);
  // this is the load-balanced send. Only works if we are master:
  ASSERT(me == -1);

  // find out least ETI delta; what counts here is slaveETIreceived +
  // slaveETI:
  int minnode[nbnode]; // nodes that have the smallest ETI
  int nmin = 0;  // how many nodes have the smallest ETI?
  float mindiff = 1.0e30F;
  //char load[1000]; load[0] = '\0';
  for (int i = 0; i < nbnode; i ++)
    {
      if (slaveETIreceived[i] < 0.0F)
        { /*strcat(load, "- ");*/ continue; } // don't know about this node
      float diff = slaveETI[i] + slaveETIreceived[i];
      if (diff < mindiff) { mindiff = diff; nmin = 1; minnode[0] = i; }
      else if (diff == mindiff) minnode[nmin++] = i;
      //char d[20]; sprintf(d, "%.1f ", diff * 1000.0F); strcat(load, d);
    }
  //LINFO("LOAD = [ %s]", load);

  int node_nb = -1;
  if (nmin == 0)  // don't know about load of any node -> random pick
    node_nb = int(floor(nbnode * drand48()));
  else            // pick one of our mindiff nodes and send to it
    node_nb = minnode[int(floor(nmin * drand48()))];

  LDEBUG("Sending msg [%d, %d] to least-loaded node %d [%s] [%d]", msg.getID(),
         msg.getAction(), node_nb, nodename[node_nb], node2fd[node_nb]);
  com->send(node2fd[node_nb], msg);
  slaveETIreceived[node_nb] = -1.0F; // don't know about this node anymore
}

// ######################################################################
bool Beowulf::receive(int& node_nb, TCPmessage& msg, int32& frame,
                      int32& action, const int timeout)
{
  ASSERT(initialized);

  // our self-sent messages have priority, let's deal with them first:
  if (node_nb == me || node_nb == -1) // receive from myself or anybody
    {
      if (selfmsg.size() > 0)
        {
          pthread_mutex_lock(&mutselfmsg);
          msg = selfmsg.front(); selfmsg.pop_front();
          pthread_mutex_unlock(&mutselfmsg);
          frame = msg.getID(); action = msg.getAction(); node_nb = me;
          LDEBUG("Received msg [%d, %d] from myself", frame, action);
          return true;
        }
      else if (node_nb == me && me != -1)
        return false; // nothing received from myself and I am not master
    }

  int rfd;
  if (node_nb == -1) rfd = -1;  // receive from any node
  else rfd = node2fd[node_nb];  // receive from given node
  bool rec = com->receive(rfd, msg, timeout);
  if (rec)
    {
      frame = msg.getID(); action = msg.getAction(); node_nb = fd2node[rfd];
      if (node_nb == -1)
        LDEBUG("Received msg [%d, %d] from master", frame, action);
      else
        LDEBUG("Received msg [%d, %d] from node %d [%s]", frame, action,
               node_nb, nodename[node_nb]);

      // is this message an order for re-initialization?
      if (action == BEO_INIT && initialized)
        {
          LINFO("RE-INIT order received -- Starting re-initialization...");
          // whoever re-inits me is my new master:
          master = rfd;
          // start a re-init based on message we just received:
          slaveReInit(msg);
          // finish up re-init:
          slaveInit();
          // we still return the message to let our user know we got re-inited
        }

      // if we are the master, update that node's ETI:
      if (me == -1)
        {
          if (node_nb < 0 || node_nb >= nbnode)
            LERROR("Bogus node number %d - IGNORED", node_nb);
          else
            {
              slaveETI[node_nb] = msg.getETI();
              slaveETIreceived[node_nb] = tim.getSecs();
            }
        }
    }
  else
    { frame = -1; action = -1; }

  return rec;
}

// ######################################################################
int Beowulf::nbReceived(const int node_nb)
{
  if (node_nb == -2)  // any node
    return com->nbReceived(-1);
  if (node_nb == -1)  // only master
    {
      if (me == -1) LFATAL("Hey, I am the master!");
      return com->nbReceived(master);
    }
  if (node_nb < 0 || node_nb >= nbnode)
    LFATAL("Node number %d out of range [-2 .. %d]", node_nb, nbnode);
  return com->nbReceived(node2fd[node_nb]);
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
