/*!@file Brain.C The whole brain */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Brain.C 4181 2005-02-12 22:40:44Z itti $
//

// MA
#include <iostream>
using namespace std;

#include "Brain.H"

#include "Image_ColorOps.H"
#include "Image_FilterOps.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "Image_Transforms.H"
#include "Pyramid_Ops.H"
#include "Retina.H"
#include "VisualCortex.H"
#include "VisualCortexEyeMvt.H"
#include "Point2DT.H"
#include "Channels.H"
#include "ShapeEstimator.H"
#include "SimulationViewer.H"
#include "ModelManager.H"
#include "SaccadeController.H"
#include "SaliencyMap.H"
#include "TaskRelevanceMap.H"
#include "AttentionGuidanceMap.H"
#include "WinnerTakeAll.H"
#include "Raster.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"
#include "Keypoint.H"
#include "InferoTemporal.H"
#include "GistEstimator.H"

#include <cstdio>

using namespace iNVT;

// ######################################################################
Brain::Brain(ModelManager& mgr, const std::string& descrName,
             const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsIORtype(this, "IORtype", IORauto),   // see ModelOptionDefs.C
  itsUseRandom(this, "UseRandom", true),  // see ModelOptionDefs.C
  itsFOAradius(this, "FOAradius", -1),  // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", -1),  // see ModelOptionDefs.C
  itsTimeStep(this, "SimulationTimeStep", 0.0001), // see ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()),    // see ModelOptionDefs.C
  itsBoringDelay(this, "BrainBoringDelayInMs", 200.0),
  itsBoringSMmv(this, "BrainBoringSMmv", 3.0F),
  itsMaxWinMv(this, "BrainMaxWinMv", 5.0F),
  itsBlankBlink(this, "BlankBlink", false),
  itsSaveObjMask(this, "BrainSaveObjMask", false), // see ModelOptionDefs.C
  itsTooManyShifts(this, "BrainTooManyShifts", 0), // see ModelOptionDefs.C
  itsTooMuchTime(this, "BrainTooMuchTime", 0.0f), // see ModelOptionDefs.C
  itsSaveWinnerFeatures(this, "BrainSaveWinnerFeatures", false),
  itsRET(NULL), itsVC(NULL), itsSC(NULL), itsSV(NULL), itsSE(NULL),
  itsSM(NULL), itsTRM(NULL), itsAGM(NULL), itsWTA(NULL),
  itsIT(NULL), itsGE(NULL),
  itsRawInput(),
  itsRetina(),
  itsClipMask(),
  itsTargetMask(),
  itsNumTargetsRemaining(0),
  itsTime(0.0),
  itsVisualField(),
  itsObjectMask(),
  itsNumAttentionShifts(0)
{  }

// ######################################################################
Brain::~Brain()
{  }

// ######################################################################
void Brain::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsIORtype);
    manager->requestOption(this, itsUseRandom);
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsFOAradius);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsBlankBlink);
    manager->requestOption(this, itsTooManyShifts);
    manager->requestOption(this, itsTooMuchTime);
    manager->requestOption(this, itsTimeStep);
  }
  if (whatToExport & OPTEXP_SAVE) {
    manager->requestOption(this, itsSaveObjMask);
    manager->requestOption(this, itsSaveWinnerFeatures);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void Brain::start1()
{
  // set the IOR type according to ShapeEstimator if it isn't set explicitely
  if (itsIORtype.getVal() == IORauto) {
    if (itsSE.get() == NULL) // no shape estimator? fall-back to disc IOR
      itsIORtype.setVal(IORdisc);
    else {
      ShapeEstimatorMode mode;
      itsSE->getModelParamVal("ShapeEstimatorMode", mode);

      switch(mode)
        {
        case SEMfeatureMap:
          itsIORtype.setVal(IORshapeEst); break;
        case SEMconspicuityMap:
          itsIORtype.setVal(IORshapeEstCM); break;
        case SEMsaliencyMap:
          itsIORtype.setVal(IORshapeEstSM); break;
        case SEMnone:
        default:
          itsIORtype.setVal(IORdisc);
        }
    }
  }

  // for ShapeEstimator IOR set the ShapeEstimator to the correct mode
  if (itsSE.get())
    switch(itsIORtype.getVal())
      {
      case IORshapeEst:
        itsSE->setModelParamVal("ShapeEstimatorMode", SEMfeatureMap); break;
      case IORshapeEstCM:
        itsSE->setModelParamVal("ShapeEstimatorMode", SEMconspicuityMap);break;
      case IORshapeEstSM:
        itsSE->setModelParamVal("ShapeEstimatorMode", SEMsaliencyMap); break;
      default:
        break;
      }
}

// ######################################################################
void Brain::addModule(SharedPtr<ModelComponent> mod, const char *modtype,
                      const bool makeSub)
{
  if (strcmp(modtype, "Retina") == 0)
    {
      if (itsRET.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsRET.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsRET.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "VisualCortex") == 0)
    {
      if (itsVC.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsVC.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsVC.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "SaccadeController") == 0)
    {
      if (itsSC.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsSC.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsSC.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "SimulationViewer") == 0)
    {
      if (itsSV.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsSV.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsSV.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "ShapeEstimator") == 0)
    {
      if (itsSE.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsSE.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsSE.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "SaliencyMap") == 0)
    {
      if (itsSM.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsSM.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsSM.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "TaskRelevanceMap") == 0)
    {
      if (itsTRM.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsTRM.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsTRM.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "AttentionGuidanceMap") == 0)
    {
      if (itsAGM.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsAGM.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsAGM.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "WinnerTakeAll") == 0)
    {
      if (itsWTA.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsWTA.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsWTA.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "InferoTemporal") == 0)
    {
      if (itsIT.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsIT.dynCastFrom(mod); // will return NULL if incorrect type
      if (itsIT.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if(makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else if (strcmp(modtype, "GistEstimator") == 0)
    {
      if (itsGE.get()) LFATAL("I already have a %s -- ABORT", modtype);
      itsGE.dynCastFrom(mod);  // will return NULL if incorrect type
      if (itsGE.get() == NULL) LFATAL("Incorrect type for %s", modtype);
      if (makeSub) { mod->setTagName(modtype); addSubComponent(mod); }
    }
  else
    LFATAL("Invalid module type %s -- ABORT", modtype);

}

// ######################################################################
void Brain::removeModule(const char *modtype)
{
  if (strcmp(modtype, "Retina") == 0)
    {
      if (itsRET.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsRET)) removeSubComponent(itsRET);
      itsRET.reset(NULL);
    }
  else if (strcmp(modtype, "VisualCortex") == 0)
    {
      if (itsVC.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsVC)) removeSubComponent(itsVC); itsVC.reset(NULL);
    }
  else if (strcmp(modtype, "SaccadeController") == 0)
    {
      if (itsSC.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsSC)) removeSubComponent(itsSC); itsSC.reset(NULL);
    }
  else if (strcmp(modtype, "SimulationViewer") == 0)
    {
      if (itsSV.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsSV)) removeSubComponent(itsSV); itsSV.reset(NULL);
    }
  else if (strcmp(modtype, "ShapeEstimator") == 0)
    {
      if (itsSE.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsSE)) removeSubComponent(itsSE); itsSE.reset(NULL);
    }
  else if (strcmp(modtype, "SaliencyMap") == 0)
    {
      if (itsSM.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsSM)) removeSubComponent(itsSM); itsSM.reset(NULL);
    }
  else if (strcmp(modtype, "TaskRelevanceMap") == 0)
    {
      if (itsTRM.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsTRM)) removeSubComponent(itsTRM);
      itsTRM.reset(NULL);
    }
  else if (strcmp(modtype, "AttentionGuidanceMap") == 0)
    {
      if (itsAGM.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsAGM)) removeSubComponent(itsAGM);
      itsAGM.reset(NULL);
    }
  else if (strcmp(modtype, "WinnerTakeAll") == 0)
    {
      if (itsWTA.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsWTA)) removeSubComponent(itsWTA);
      itsWTA.reset(NULL);
    }
  else if (strcmp(modtype, "InferoTemporal") == 0)
    {
      if (itsIT.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsIT)) removeSubComponent(itsIT);
      itsIT.reset(NULL);
    }
  else if (strcmp(modtype, "GistEstimator") == 0)
    {
      if (itsGE.get() == NULL) LFATAL("I don't have a %s -- ABORT", modtype);
      if (hasSubComponent(itsGE)) removeSubComponent(itsGE);
      itsGE.reset(NULL);
    }
  else
    LFATAL("Invalid module type %s -- ABORT", modtype);
}

// ######################################################################
SharedPtr<ModelComponent> Brain::module(const char *modtype) const
{

  if (strcmp(modtype, "Retina")) return itsRET;
  if (strcmp(modtype, "VisualCortex")) return itsVC;
  if (strcmp(modtype, "SaccadeController")) return itsSC;
  if (strcmp(modtype, "SimulationViewer")) return itsSV;
  if (strcmp(modtype, "ShapeEstimator")) return itsSE;
  if (strcmp(modtype, "SaliencyMap")) return itsSM;
  if (strcmp(modtype, "TaskRelevanceMap")) return itsTRM;
  if (strcmp(modtype, "AttentionGuidanceMap")) return itsAGM;
  if (strcmp(modtype, "WinnerTakeAll")) return itsWTA;
  if (strcmp(modtype, "InferoTemporal")) return itsIT;
  if (strcmp(modtype, "GistEstimator")) return itsGE;
  LFATAL("Invalid module type %s -- ABORT", modtype);
  return SharedPtr<ModelComponent>(NULL); // keep compiler happy
}

// ######################################################################
SharedPtr<Retina> Brain::getRET() const
{
  if (itsRET.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsRET;
}

// ######################################################################
SharedPtr<VisualCortex> Brain::getVC() const
{
  if (itsVC.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsVC;
}

// ######################################################################
SharedPtr<SaccadeController> Brain::getSC() const
{
  if (itsSC.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsSC;
}

// ######################################################################
SharedPtr<SimulationViewer> Brain::getSV() const
{
  if (itsSV.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsSV;
}

// ######################################################################
SharedPtr<ShapeEstimator> Brain::getSE() const
{
  if (itsSE.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsSE;
}

// ######################################################################
SharedPtr<SaliencyMap> Brain::getSM() const
{
  if (itsSM.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsSM;
}

// ######################################################################
SharedPtr<TaskRelevanceMap> Brain::getTRM() const
{
  if (itsTRM.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsTRM;
}

// ######################################################################
SharedPtr<AttentionGuidanceMap> Brain::getAGM() const
{
  if (itsAGM.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsAGM;
}

// ######################################################################
SharedPtr<WinnerTakeAll> Brain::getWTA() const
{
  if (itsWTA.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsWTA;
}


// ######################################################################
SharedPtr<InferoTemporal> Brain::getIT() const
{
  if (itsIT.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsIT;
}

// ######################################################################
SharedPtr<GistEstimator> Brain::getGE() const
{
  if (itsGE.get() == NULL) LERROR("I don't have one -- RETURNING NULL");
  return itsGE;
}

// ######################################################################
bool Brain::hasModule(const char *modtype) const
{
  if (strcmp(modtype, "Retina") == 0)
    { if (itsRET.get()) return true; }
  else if (strcmp(modtype, "VisualCortex") == 0)
    { if (itsVC.get()) return true; }
  else if (strcmp(modtype, "SaccadeController") == 0)
    { if (itsSC.get()) return true; }
  else if (strcmp(modtype, "SimulationViewer") == 0)
    { if (itsSV.get()) return true; }
  else if (strcmp(modtype, "ShapeEstimator") == 0)
    { if (itsSE.get()) return true; }
  else if (strcmp(modtype, "SaliencyMap") == 0)
    { if (itsSM.get()) return true; }
  else if (strcmp(modtype, "TaskRelevanceMap") == 0)
    { if (itsTRM.get()) return true; }
  else if (strcmp(modtype, "AttentionGuidanceMap") == 0)
    { if (itsAGM.get()) return true; }
  else if (strcmp(modtype, "WinnerTakeAll") == 0)
    { if (itsWTA.get()) return true; }
  else if (strcmp(modtype, "InferoTemporal") == 0)
    { if (itsIT.get()) return true; }
  else if (strcmp(modtype, "GistEstimator") == 0)
    { if (itsGE.get()) return true; }
  else
    LFATAL("Invalid module type %s -- ABORT", modtype);

  // if we get here, modtype is correct but we don't have the module:
  return false;
}

// ######################################################################
void Brain::reset(const bool recurse)
{
  // reset a lot of stuff in Brain
  itsRawInput.freeMem();
  itsRetina.freeMem();
  itsTime = 0.0;
  itsVisualField.freeMem();
  itsObjectMask.freeMem();
  itsNumAttentionShifts = 0;

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void Brain::input(const Image< PixRGB<byte> >& inp, const Image<byte>& targets,
                  const Image<byte>& clipmask)
{
  processInput(inp,targets,clipmask);
  obtainOutput();
}

// ######################################################################
void Brain::processInput(const Image< PixRGB<byte> >& inp,
                         const Image<byte>& targets,
                         const Image<byte>& clipmask)
{
  // before we actually deal with this new input, give the SV a chance
  // to do some last updates on our current frame:
  if (itsSV.get() && itsRetina.initialized()) itsSV->newInputReady();

  // keep a copy of clipmask (may need it later, e.g., to refoveate):
  itsClipMask = clipmask;

  // keep a copy of the raw input:
  itsRawInput = inp;

  // do we have a Retina module? if so, use it, otherwise just use the
  // raw input:
  if (itsRET.get()) { itsRET->input(inp); itsRetina = itsRET->getOutput(); }
  else itsRetina = inp;

  // initialize targetmask:
  if (targets.initialized())
    {
      itsTargetMask = targets;
      itsTargetMask.normalize(0, 255);
      itsNumTargetsRemaining = itsTargetMask.countParticles(255);
      LINFO("Counting targets... %d detected", itsNumTargetsRemaining);
      manager->textLog("Targets = %d",  itsNumTargetsRemaining);
    }
  else
    itsTargetMask.freeMem();

  // process image through visual cortex:
  if (itsVC.get() == NULL)
    LFATAL("I am useless without a VisualCortex module -- ABORT");

  // if we are in a blink, assume that inputs are blanked out;
  // otherwise, process new input through visual cortex and build new
  // saliency map; during saccades, inputs are not blanked out here,
  // but saccadic suppression will hit the saliency map:
  if (itsBlankBlink.getVal() && itsSC.get() && itsSC->isInBlink())
    {
      LINFO("#### Visual input blanked out while in blink ####");
      Image<PixRGB<byte> > blank(itsRetina.getDims(), ZEROS);
      itsVC->input(blank, itsTime, itsClipMask);
    }
  else
    itsVC->input(itsRetina, itsTime, itsClipMask);

  LINFO ("visual cortex input is ok");

  // process the gist of image if we have a GistEstimator:
  if (itsGE.get())
    {
      itsGE->compute();
      LINFO ("Gist features are computed");
    }
}

// ######################################################################
void Brain::obtainOutput()
{
  Image<float> sminput = itsVC->getOutput();

  // send output of visual cortex to the saliency map:
  if (itsSM.get() == NULL)
    LFATAL("I am useless without a SaliencyMap module -- ABORT");
  itsSM->input(sminput);

  // initialize visual field if necessary:
  if (itsVisualField.isSameSize(itsRetina) == false)
    itsVisualField = Image<byte>(itsRetina.getDims(), ZEROS);

  // initialize task-relevance map (itsTRM) if necessary
  if (itsTRM.get() && sminput.isSameSize(*itsTRM) == false) {
    Image<float> blank(sminput.getDims(), NO_INIT); blank.clear(1.0F);
    itsTRM->input(blank);
  }

  // if we have a SimulationViewer, pass the retina on to it:
  if (itsSV.get()) itsSV->newInputFrame(itsRetina, itsTime);
}

// ######################################################################
void Brain::evolve(double& stime, Point2D& winner, int& ntarg, float& winV)
{

  // did we just find all the targets that were outlined in
  // itsTargetMask (if it has been provided to Brain::input()) at the
  // previous iteration and want to quit?
  if (itsTargetMask.initialized() && itsNumTargetsRemaining <= 0)
    {
      LINFO("##### All targets found -- EXIT #####");
      manager->textLog("AllTargetsFound");
      winner.i = -1; winner.j = -1; ntarg = -1; winV = 0.0;
      return;
    }

  // #################### main evolve ####################
  Point2D win(-1, -1); winner.i = -1; winner.j = -1;
  bool saccading1 = false, blinking1 = false, headsaccading1 = false;

  // too much time has passed?
  if (itsTooMuchTime.getVal() > 0.0f && stime > itsTooMuchTime.getVal())
    {
      LINFO("#### Too much time elapsed -- DONE ####");
      ntarg = -1;   // that should signal our caller that it's time to quit
      return;
    }

  if (itsSC.get()) {
    // we will use those at the end of evolve() to determine whether
    // to turn on/off saccadic suppression and blink suppression:
    saccading1 = itsSC->isInSaccade();  // saccade status before evolve()
    blinking1 = itsSC->isInBlink();     // blink status before evolve()
    headsaccading1 = itsSC->isInSaccadeHead();  // head saccade status
  }

  // increment local time one step. To avoid tiny drifts of the clock
  // due to rounding errors, we add one time step to our time counter
  // in the following manner: 1) convert the current time to a number
  // of time steps since the epoch; 2) add one; 3) convert back to
  // time. This way, we avoid rounding problems:
  uint64 nstep = uint64(itsTime / itsTimeStep.getVal() + 0.25) + 1;
  itsTime = itsTimeStep.getVal() * double(nstep);

  // if we have a VisualCortex that needs to evolve(), let's do
  // it. Currently, only VisualCortexEyeMvt supports evolve():
  VisualCortexEyeMvt *vc = dynamic_cast<VisualCortexEyeMvt *>(itsVC.get());
  if (vc) vc->evolve(itsTime);

  // let the dynamical saliency map simulation evolve one step:
  itsSM->evolve(itsTime);

  // get a snapshot of current saliency map:
  Image<float> currsm = itsSM->getV();

  // if we have a TRM and an AGM, we will combine bottom-up inputs
  // from the SM and top-down inputs from the TRM into the
  // AGM. Otherwise, we are a purely bottom-up model, driven by the SM
  // only:
  Image<float> curragm;
  if (itsTRM.get() && itsAGM.get())
    {
      // let the TRM evolve and get the current:
      itsTRM->evolve(itsTime);
      Image<float> currtrm = itsTRM->getV();

      // feed SM as bottom-up input to the AGM, and TRM as top-down input:
      itsAGM->inputBU(currsm);
      itsAGM->inputTD(currtrm);

      // let the AGM evolve:
      itsAGM->evolve(itsTime);

      // get the current AGM:
      curragm = itsAGM->getV();
    }
  else
    curragm = currsm;

  // let the WTA evolve to find most salient and relevant object:
  if (itsWTA.get()) {
    itsWTA->input(curragm);
    itsWTA->evolve(itsTime, win);
  }

  // evolve the saccade controller one time step
  if (itsSC.get()) itsSC->evolve(itsTime);

  // update the SimulationViewer to new time
  if (itsSV.get()) itsSV->evolve(itsTime);

  // ########## do we have a new COVERT attention shift?
  if (win.i >= 0)
    {
      // increment our count of attention shifts so far:
      itsNumAttentionShifts ++;

      // get SM and AGM voltages at winner location:
      //double sv = itsSM->getV(win), av = itsAGM->getV(win);
      double sv, av;
      sv = itsSM->getV(win);
      if (itsAGM.get()) av = itsAGM->getV(win);
      else av = sv;

      // compute coords of winner in original image and build WTAwinner:
      WTAwinner newwin;
      newwin.buildFromSMcoords(win, itsLevelSpec.getVal().mapLevel(),
                               itsUseRandom.getVal(),
                               itsTime, sv, av, false);

      // in case we don't have an SC (see below), we'll return the
      // scaled covert attention coordinates in 'winner':
      winner = newwin.p;

      // if the last covert attention shift was slow or the SM voltage
      // was low, mark the covert attention shift as boring: NOTE:
      // this feature relies on the SaccadeController to keep track of
      // the shifts of covert attention, so it will not work if we
      // don't have a SaccadeController. Using a
      // TrivialSaccadeController will do, though:

      if (itsSC.get() &&
          itsSC->havePercepts() &&
          (newwin.t - itsSC->getPreviousPercept(0).t >
           itsBoringDelay.getVal() * 0.001 ||
           newwin.sv < itsBoringSMmv.getVal() * 0.001) )
        newwin.boring = true;

      LINFO("##### WTA Winner (%d,%d) at %fms %s#####",
            newwin.p.i, newwin.p.j, itsTime * 1000.0,
            newwin.boring ? "[boring] ":"");

      // save vector of feature values at winning location?
      if (itsSaveWinnerFeatures.getVal())
        {
          std::vector<double> f; itsVC->getFeatures(newwin.p, f);
          char txt[2048]; strcpy(txt, "WinnerFeatures ="); char tmp[100];
          for (uint ii = 0; ii < f.size(); ii ++)
            { sprintf(tmp, " %g", f[ii]); strcat(txt, tmp); }
          manager->textLog(txt);
        }

      // update the shapeEstimator data at the winner location; then
      // update our object mask accordingly:
      if (itsSE.get()) {
        itsSE->compute(newwin.p);
        itsObjectMask = itsSE->getSmoothMask() * 255.0F;
      } else
        itsObjectMask.freeMem();  // no object mask available

      // ### detect pedestrian:
      if (itsVC.get() && itsVC->hasChannel("pedestrian"))
        {
          SharedPtr<PedestrianChannel> pch;
          pch.dynCastFrom(itsVC->channelPtr("pedestrian"));

          Rectangle prect(newwin.p.j - itsFOAradius.getVal() * 9 / 4,
                          newwin.p.i - itsFOAradius.getVal() * 9 / 4,
                          newwin.p.j + itsFOAradius.getVal() * 9 / 4,
                          newwin.p.i + itsFOAradius.getVal() * 9 / 4);
          Point2D pos; int lev; float score;
          pch->findBestMatch(prect, pos, lev, score);

          // let SimulationViewer know:
          if (itsSV.get()) itsSV->channelUpdate(itsVC);
          //FIXME: channelUpdate should be called in a more general
          //manner, not only for the pedestrianchannel (though it's
          //the only one that has drawings at the moment)
       }

      // let the SimulationViewer know about this event:
      if (itsSV.get())
        itsSV->attentionShift(newwin, itsObjectMask);

      // tell saccade controller about new winner:
      if (itsSC.get())
        itsSC->setPercept(newwin);

      // show SM voltage at winning location:
      LINFO("SM voltage at winner: %fmV above rest", 1000.0 * sv);

      // ### try to recognize object
      if (itsIT.get() && ((itsSE.get() && itsSE->getObjectArea() > 0) ||
                          (!itsIT->getUseAttention())))
        {
          // Get the segmented object image
          Image<PixRGB<byte> > objImage = itsRawInput;
          if(itsIT->getUseAttention())
            objImage = itsSE->getObjectImage(objImage);
          
          // Get the image features at the attended point
          std::vector<double> f; itsVC->getFeatures(newwin.p, f);
              
          // send the image and features to the InferiorTemporalCortex for 
          // object recognition
          itsIT->input(objImage, newwin.p, f, itsNumAttentionShifts);
        }

      // if current winning voltage is too high, calm things down with
      // some mild broad inhibition:
      if (1000.0 * sv > itsMaxWinMv.getVal())
        {
          LINFO("Triggering global inhibition to avoid map explosion.");
          itsSM->depress(win, 0.00005 * (sv - itsMaxWinMv.getVal() / 1000.0),
                         0.0, 1000.0 * double(itsFOAradius.getVal()), 100.0); // MA
        }
      // fire-up inhibition-of-return at winning location:
      switch (itsIORtype.getVal())
        {
        case IORdisc:       ntarg = discIOR(win); break;
        case IORshapeEst:   shapeEstimatorIOR(win); ntarg = 0; break;
        case IORshapeEstCM: shapeEstimatorIOR(win); ntarg = 0; break;
        case IORshapeEstSM: shapeEstimatorIOR(win); ntarg = 0; break;
        case IORnone: ntarg = 0; break; // do nothing
        default:
          LFATAL("encountered unknown IOR type - don't know what to do");
        }

      // log what happened:
      manager->textLog("WTAwinner = %d %d %.2f %d %.3f",
                       newwin.p.i, newwin.p.j, itsTime * 1000.0,
                       ntarg, sv * 1000.0F);

      // did we hit any targets, as defined by our itsTragetMask (if
      // it was provided to Brain::input())?
      if (ntarg > 0)
        {
          LINFO("###### %d new target(s) hit at t=%fms ######",
                ntarg, itsTime * 1000.0);
          itsNumTargetsRemaining -= ntarg;
          // next time we are called, we will quit if we have no
          // target remaining (see above)
        }

      // too many shifts of attention?
      if (itsTooManyShifts.getVal() > 0 &&
          itsNumAttentionShifts > itsTooManyShifts.getVal())
        {
          LINFO("#### Too many attention shifts -- DONE ####");
          ntarg = -1;   // that should signal our caller that it's time to quit
        }
    }

  // ########## compute saccade target (eye position) and head
  // position, if we have a SaccadeController; otherwise (see above)
  // we have made the eye position equal to the focus of attention and
  // the head position invalid; also figure out whether we are just
  // starting/ending an eye saccade or a blink:
  Point2D headpos(-1, -1);
  bool saccading2 = false, blinking2 = false, headsaccading2 = false;
  if (itsSC.get())
    {
      winner = itsSC->getDecision(itsTime);
      headpos = itsSC->getDecisionHead(itsTime);
      saccading2 = itsSC->isInSaccade(); // saccade status after evolve
      blinking2 = itsSC->isInBlink(); // blink status after evolve
      headsaccading2 = itsSC->isInSaccadeHead(); // head saccade status
    }

  // ########## are we just beginning/ending an eye saccade?
  if (saccading1 == true && saccading2 == false)
    {
      // end of eye saccade; turn off saccadic suppression and let
      // SimulationViewer know:
      saccadicSuppression(false);
      if (itsSV.get()) itsSV->eyeSaccade(false);
    }
  if (saccading1 == false && saccading2 == true)
    {
      // beginning of eye saccade; turn on saccadic suppression and
      // let SimulationViewer know:
      saccadicSuppression(true);
      if (itsSV.get()) itsSV->eyeSaccade(true);
    }
  // ########## are we just beginning/ending an eye blink?
  if (blinking1 == true && blinking2 == false)
    {
      // end of eye blink; turn off blink suppression and let
      // SimulationViewer know:
      blinkSuppression(false);
      if (itsSV.get()) itsSV->eyeBlink(false);
    }
  if (blinking1 == false && blinking2 == true)
    {
      // beginning of eye blink; turn on blink suppression and
      // let SimulationViewer know:
      blinkSuppression(true);
      if (itsSV.get()) itsSV->eyeBlink(true);
    }
  // ########## are we just beginning/ending a head saccade?
  if (headsaccading1 == true && headsaccading2 == false)
    {
      // end of head saccade; let SimulationViewer know:
      if (itsSV.get()) itsSV->headSaccade(false);
    }
  if (headsaccading1 == false && headsaccading2 == true)
    {
      // beginning of head saccade; let SimulationViewer know:
      if (itsSV.get()) itsSV->headSaccade(true);
    }

  // ########## do we have a new OVERT attention shift (eye movement)?
  if (itsSC.get() && winner.i != -1 && winner.j != -1)
    {
      // for winV, we return the AGM voltage at last percept:
      winV = itsSC->getPreviousPercept(0).av;

      // let the SimulationViewer know about this event:
      if (itsSV.get())
        itsSV->eyeMovement(Point2DT(winner, itsTime), Image<byte>());
    }
  else  // no new saccade target this time
    winV = 0.0f;

  // ########## do we have a new HEAD movement?
  if (headpos.i >= 0)
    {
      // let the SimulationViewer know about this event:
      if (itsSV.get())
        itsSV->headMovement(Point2DT(headpos, itsTime), Image<byte>());
    }
  // return current time:
  stime = itsTime;

}

// ######################################################################
double Brain::resetTime()
{ itsTime = 0.0; return itsTime; }

// ######################################################################
Image<byte> Brain::getObjectMask() const
{ return itsObjectMask; }

// ######################################################################
const Dims& Brain::getInputDims() const
{ return itsRetina.getDims(); }

// ######################################################################
Image< PixRGB<byte> > Brain::getRawInput() const
{ return itsRawInput; }

// ######################################################################
Image<byte> Brain::getVisualField() const
{ return itsVisualField; }

// ######################################################################
void Brain::findFixation(Point2D& fixation, float& value) const
{ itsAGM->getV().findMax(fixation, value); }

// ######################################################################
void Brain::saccadicSuppression(const bool on)
{
  // first, pass this through to the SM:
  itsSM->saccadicSuppression(on);

  // also to the WTA:
  if (itsWTA.get()) itsWTA->saccadicSuppression(on);
}

// ######################################################################
void Brain::blinkSuppression(const bool on)
{
  // first, pass this through to the SM:
  //FIXME: used to be off.....   itsSM->blinkSuppression(on);

  // also the WTA:
  if (itsWTA.get()) itsWTA->blinkSuppression(on);
}

// ######################################################################
void Brain::shapeEstimatorIOR(const Point2D win)
{
  // revert to discIOR we don't have a shape estimator:
  if (itsSE.get() == NULL) { discIOR(win); return; }

  // what does the shape estimator think of this?
  Image<byte> iormask = itsSE->getIORMask();

  // revert to discIOR if the shape estimator doesn't know:
  if (iormask.initialized() == false) { discIOR(win); return; }

  itsSM->depress(win, itsSM->getV(win), iormask);
  LINFO("ShapeEstimator-based inhibition of return fired.");
}

// ######################################################################
int Brain::discIOR(const Point2D win)
{
  // fire up inhibition of return:
  float winV = itsSM->getV(win);
  // compute the parameters of the local depression, and depress
  // amplitudes for M-hat:
  double pAmpl = 0.1 * ((double)winV); // strongly inhibit winner
  double mAmpl = 1e-4 * pAmpl;          // mildly excite surround
  // standard deviations for M_hat: inhibit of the foa size
  double pSdev = 0.3 * double(itsFOAradius.getVal()) /
    double(1 << itsLevelSpec.getVal().mapLevel());  //foa inhib
  double mSdev = 4.0 * pSdev;    // wide mild excitation outside foa
// MA
LDEBUG("pSdev = %f, mSDev = %f", (float)pSdev, (float)mSdev);
  itsSM->depress(win, pAmpl, mAmpl, pSdev, mSdev);
  LINFO("Inhibition of return fired.");

  // draw a disk in visual field. Check for targets still unhit
  // (val 255) and if found, flood them with val 250:
  Point2D newwin(win.i * itsVisualField.getWidth() / itsSM->getWidth(),
                 win.j * itsVisualField.getHeight() / itsSM->getHeight());

  // also draw in the visual field, and count the new targets hit if we
  // have a target mask:
  if (itsTargetMask.initialized())
    return itsVisualField.
      drawDiskCheckTarget(itsTargetMask, newwin, itsFOAradius.getVal(),
                          255, 255, 250);

  // if no target mask, just draw in the visual field:
  itsVisualField.drawDisk(newwin, itsFOAradius.getVal(), 255);
  return 0;
}

// ######################################################################
void Brain::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // first do our own:
  if (itsSaveObjMask.getVal() && itsObjectMask.initialized())
    ofs->writeGray(itsObjectMask, PNM, -1, "OBJ");

  // now our modules:
  if (itsRET.get()) itsRET->saveResults(ofs);
  if (itsVC.get()) itsVC->saveResults(ofs);
  //if (itsSC.get()) itsSC->saveResults(ofs);
  if (itsSV.get()) itsSV->saveResults(ofs);
  //if (itsSE.get()) itsSE->saveResults(ofs);
  if (itsSM.get()) itsSM->saveResults(ofs);
  if (itsTRM.get()) itsTRM->saveResults(ofs);
  if (itsAGM.get()) itsAGM->saveResults(ofs);
  if (itsWTA.get()) itsWTA->saveResults(ofs);
  //if(itsITC.get()) itsITC->saveResults(ofs);
  //if (itsGE.get()) itsGE->saveResults(ofs);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
