/*!@file Brain.H The whole brain */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Brain.H 4170 2005-02-10 05:09:56Z everist $
//

#ifndef BRAIN_H_DEFINED
#define BRAIN_H_DEFINED

#include "saliency.H"
#include "Point2DT.H"
#include "Image.H"
#include "SharedPtr.H"
#include "ImageSet.H"
#include "ModelComponent.H"
#include "IORtypes.H"
#include "LevelSpec.H"

class Retina;
class VisualCortex;
class SaccadeController;
class SimulationViewer;
class ShapeEstimator;

namespace iNVT 
{
  class SaliencyMap;
}

class TaskRelevanceMap;
class AttentionGuidanceMap;
class WinnerTakeAll;
class OutputFrameSeries;
class InferoTemporal;
class GistEstimator;

//! This is the brain, a holder for various vision objects.
/*! It contains a retina (input image), a VisualCortex, a SaliencyMap,
  a WinnerTakeAll, a visual field (image that keeps track of attended
  locations), and lots of interconnections between those.  */
class Brain : public ModelComponent
{
public:
  // ######################################################################
  /*! @name Constructors, destructors, and run-time "modules"

  Typically, you instantiate a Brain and then add various plug-in
  "modules" to it by using the addModule() function. If you decide so,
  each module you add may become a subcomponent of your Brain object;
  if so, you should not do a ModelManager::registerComponent() on
  them, but just do it once for all on your Brain object
  instead. Modules are a restricted set of ModelComponent objects, see
  addModule() for details of which ones are supported. */
  //@{

  //! Constructor
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage */
  Brain(ModelManager& mgr, const std::string& descrName = "Brain",
        const std::string& tagName = "Brain");

  //! Virtual destructor for safe inheritance.
  virtual ~Brain();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Add a module
  /*! Supported modules are, with their modtype names:
    - "Retina" - derived from class Retina
    - "VisualCortex" - derived from class VisualCortex
    - "SaccadeController" - derived from class SaccadeController
    - "SimulationViewer" - derived from class SimulationViewer
    - "ShapeEstimator" - derived from class ShapeEstimator   
    - "SaliencyMap" - derived from class SaliencyMap
    - "TaskRelevanceMap" - derived from class TaskRelevanceMap
    - "AttentionGuidanceMap" - derived from class AttentionGuidanceMap
    - "WinnerTakeAll" - derived from class WinnerTakeAll
    - "InferoTemporal" - derived from class InferoTemporal
    - "GistEstimator" - derived from class GistEstimator   

    @param mod a <TT>SharedPtr</TT> to a ModelComponent derivative
    object. For example, to add a module derived from VisualCortex, a
    typical call would be: <TT>brain.addModule(makeShared(new
    VCderived(...)), "VisualCortex"); </TT>

    @param modtype one of the module type names specified
    above. Anything else will throw a fatal error. Some type checking
    will be done to enforce that the passed ModelComponent indeed
    derives from the expected base class (e.g., if you add a module
    with modtype "VisualCortex" then we will check that your mod
    derives from VisualCortex. A fatal error will be thrown if that
    test fails. 

    @param makeSubComp If true, the object passed as argument will
    become a subcomponent of us, thus you will not need to separately
    register it with ModelManager. Yields a fatal error if we already
    have one module of that type (name). Typically, you want to set
    this flag to true in models where the Brain is your top-level
    (root) component; in models that contains objects for whom the
    Brain is a subcomponent, typically they will also have our modules
    as subcomponents, so we should not set this flag to true */
  void addModule(SharedPtr<ModelComponent> mod, const char *modtype,
                 const bool makeSubComp = false);

  //! Remove a previously added module
  /*! Yields a fatal error if no such module was previously added */
  void removeModule(const char *modtype);

  //! Reset Brain and propagate the reset signal to the sub-components
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! Get a module; see shorthands below
  /*! Yields a SharedPtr to NULL if the corresponding module has not
    been added using addModule(), and throws a fatal error if modtype
    is unknown */
  SharedPtr<ModelComponent> module(const char *modtype) const;

  //! Get the Retina
  SharedPtr<Retina> getRET() const;

  //! Get the VisualCortex
  SharedPtr<VisualCortex> getVC() const;

  //! Get the SaccadeController
  SharedPtr<SaccadeController> getSC() const;

  //! Get the SimulationViewer
  SharedPtr<SimulationViewer> getSV() const;

  //! Get the ShapeEstimator
  SharedPtr<ShapeEstimator> getSE() const;

  //! Get the SaliencyMap
  SharedPtr<iNVT::SaliencyMap> getSM() const;

  //! Get the TaskRelevanceMap
  SharedPtr<TaskRelevanceMap> getTRM() const;

  //! Get the AttentionGuidanceMap
  SharedPtr<AttentionGuidanceMap> getAGM() const;

  //! Get the WinnerTakeAll
  SharedPtr<WinnerTakeAll> getWTA() const;

  //! Get the InferoTemporal
  SharedPtr<InferoTemporal> getIT() const;

  //! Get the GistEstimator
  SharedPtr<GistEstimator> getGE() const;

  //! Returns true if the corresponding module has been added with addModule()
  bool hasModule(const char *modtype) const;

  //@}

  // ######################################################################
  /*! @name Brain simulation functions */
  //@{

  //! Load a new frame and pass it through visual cortex
  /*! This will yield a new input to the saliency map.
    @param inp input color image
    @param targets binary map of the same size as the input, in which
    targets for the focus of attention are 255 and the rest is
    zero. Pass an uninitialized image if you have no targets to look
    for.
    @param clipMask binary map of the same size as the input, in which
    areas that are to be considered for processing for attention are
    255, and areas that are to be "clipped" from the Saliency Map are
    0.  An uninitialized image means: don't use a mask. */
  void input(const Image< PixRGB<byte> >& inp,
             const Image<byte>& targets,
             const Image<byte>& clipmask = Image<byte>());

  //! Deliver output to VC & Co (called from input, see there for details)
  void processInput(const Image< PixRGB<byte> >& inp,
                    const Image<byte>& targets,
                    const Image<byte>& clipmask = Image<byte>());

  //! Get the output from VisualCortex & Co. (called from input)
  void obtainOutput();

  //! Evolve saliency map and winner-take-all
  /*! @param stime current simulation time, in seconds
    @param winner coordinates of new attended location; winner coords
    are -1 if there was no new winner (i.e., no shift of attention) If
    we have a SaccadeController module, winner will contain the
    current eye position rather than attention position.
    @param ntarg number of new targets hit (or -1 if whole visual
    field visited)
    @param winV saliency map voltage at winning location. */
  void evolve(double& stime, Point2D& winner, int& ntarg, float& winV);

  //! Reset the internal time
  double resetTime();

  //! Turn saccadic suppression on/off
  /*! For now, this is a pass-through to SM::saccadicSuppression(), so
    see there for usage. In adition, we reset our static-detection
    buffer and also the WTA. If we have a SaccadeController module, we
    will call this internally during our evolve(), as appropriate;
    this method is made public just so that saccadic suppression may
    be forced externally too. */
  void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  /*! For now, we only reset the WTA. If we have a SaccadeController
    module, we will call this internally during our evolve() as
    appropriate; this method is made public just so that blink
    suppression may be forced externally too. */
  void blinkSuppression(const bool on);

  //@}

  // ##### Access functions:

  //! Get input dimensions
  const Dims& getInputDims() const;

  //! Get whatever raw input we have received before passing it through Retina
  Image< PixRGB<byte> > getRawInput() const;

  //! Returns binary mask of currently attended object
  Image<byte> getObjectMask() const;

  //! Returns current visual field.
  /*! The visual field starts empty and fills in with each FOA shift */
  Image<byte> getVisualField() const;

  //! Finds the current fixation in AGM
  /*! @param fixation enter a reference to current fixation
    ie a dummy that returns the current fixation point
    @param relevance, enter a reference to value at current fixation
    ie a dummy that returns the value at current fixation  */
  void findFixation(Point2D& fixation, float& value) const;

  //! Save our internals and outputs from our modules
  void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

protected:
  //! type of IOR to use. See IORtypes.H
  ModelParam<IORtype> itsIORtype;

  //! add jitter to winning location?
  ModelParam<bool> itsUseRandom;

  //! radius of the FOA, in pixels
  ModelParam<int> itsFOAradius;

  //! radius of the Fovea, in pixels
  ModelParam<int> itsFoveaRadius;

  //! Simulation time step, in seconds
  ModelParam<double> itsTimeStep;

  //! Specification of our image processing center-surround levels
  /*! The only thing that we care about here is the saliency map level */
  ModelParam<LevelSpec> itsLevelSpec;

  //! Inter-covert-shift delay (in ms) that makes a shift of attention boring
  ModelParam<double> itsBoringDelay;

  /*! Minimum saliency map depolarization (in mV) to make a covert
    attention shift not boring */
  ModelParam<float> itsBoringSMmv;

  /*! If the saliency map voltage at winning location gets higher than
    this (in mV), trigger some global inhibition */
  ModelParam<float> itsMaxWinMv;

  //! Blank out visual inputs during blinks?
  ModelParam<bool> itsBlankBlink;

  //! Save our internals when saveResults() is called?
  ModelParam<bool> itsSaveObjMask;

  //! Exit after a given number of attention shifts (if not zero)
  ModelParam<int> itsTooManyShifts;

  //! Exit after a given time in seconds (if not zero)
  ModelParam<float> itsTooMuchTime;

  //! Save vector of features at each attended location?
  ModelParam<bool> itsSaveWinnerFeatures;

  // a "Module" is a subcomponent for which we explicitly keep here a
  // SharedPtr with the correct type, to accelerate access (avoiding
  // name lookups and dynamic casting from SharedPtr<ModelComponent>,
  // as would be necessary if we were doing a straight
  // ModelComponent::subComponent() call each time we need access to
  // the module). Thus, not anybody can be a module, as we need to
  // have an itsXX member for it here. We keep those members in the
  // protected section so that classes that derive from Brain can
  // directly access the modules.
  SharedPtr<Retina> itsRET;               //!< Retina module
  SharedPtr<VisualCortex> itsVC;          //!< VisualCortex module
  SharedPtr<SaccadeController> itsSC;     //!< SaccadeController module
  SharedPtr<SimulationViewer> itsSV;      //!< SimulationViewer module
  SharedPtr<ShapeEstimator> itsSE;        //!< ShapeEstimator Module
  SharedPtr<iNVT::SaliencyMap> itsSM;           //!< SaliencyMap Module
  SharedPtr<TaskRelevanceMap> itsTRM;     //!< TaskRelevanceMap Module
  SharedPtr<AttentionGuidanceMap> itsAGM; //!< AttentionGuidanceMap Module
  SharedPtr<WinnerTakeAll> itsWTA;        //!< WinnerTakeAll Module
  SharedPtr<InferoTemporal> itsIT;        //!< InferoTemporal Module
  SharedPtr<GistEstimator> itsGE;         //!< GistEstimator Module

  //! do some additional config at start time; see ModelComponent.H
  virtual void start1();

//private:
  // copy of the last raw input received, before it went through retina
  Image< PixRGB<byte> > itsRawInput;

  // retina, may be foveated or copy of input
  Image< PixRGB<byte> > itsRetina;

  // mask to eliminate parts of the scene; initialized only if a
  // clipmask was given to Brain::input()
  Image<byte> itsClipMask;

  // binary mask indicating where targets are; if one was given to
  // Brain::input(), that's a copy of it with the targets found so far
  // painted in a different color, otherwise it's an initialized image
  Image<byte> itsTargetMask;

  // number of targets that still need to be found
  int itsNumTargetsRemaining;

  // current simulation time, in seconds
  double itsTime;

  // This is a visual memory of which locations in the visual field
  // have been covertly attended to; it starts as an empty image of
  // same dims as itsRetina, but FOAs get painted into it each time a
  // new shift of covert attention is made. NOTE: this concept works
  // well with a static input, but determining when the visual field
  // should be erased or shifted because the viewpoint changed should
  // come from the SaccadeController in a future implementation. The
  // same applies to itsSM and itsWTA, for that matter.
  Image<byte> itsVisualField;

  // binary mask, at the resolution of the retina, of the currently
  // attended object (255 inside object, 0 outside, possibly
  // intermediary values on fuzzy boundaries).
  Image<byte> itsObjectMask;

  // attention shift number (to decide whether we have done too many)
  int itsNumAttentionShifts;

  // shape-based inhibition of return
  void shapeEstimatorIOR(const Point2D win);

  // disc-based inhibition of return
  int discIOR(const Point2D win);
};

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
