/*!@file ChannelBase.C The base class for all channel classes */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ChannelBase.C 4188 2005-02-13 07:01:15Z itti $
//

#include "ChannelBase.H"
#include "MathFunctions.H"
#include "Image.H"
#include "ParamMap.H"
#include "VisualFeatures.H"
#include "fancynorm.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"

#include <cmath>  // for fabs()
#include <string>

// ######################################################################
// ######################################################################
// ChannelListener member definitions:
// ######################################################################
// ######################################################################

ChannelListener::~ChannelListener() {}

// ######################################################################
// ######################################################################
// ChannelBase member definitions:
// ######################################################################
// ######################################################################

ChannelBase::ChannelBase(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName, const VisualFeature vs) :
  ModelComponent(mgr, descrName, tagName),
  itsOutputRangeMin(this, "ChannelOutputRangeMin", MAXNORMMIN),
  itsOutputRangeMax(this, "ChannelOutputRangeMax", MAXNORMMAX),
  itsTotalWeight(this, "TotalWeight", 1.0),
  itsVisFeature(this, "VisualFeature", vs),
  itsHomogeneous(this, "Homogeneous", true)
{ }

// ######################################################################
ChannelBase::~ChannelBase()
{ }

// ######################################################################
void ChannelBase::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsOutputRangeMin);
    manager->requestOption(this, itsOutputRangeMax);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
VisualFeature ChannelBase::visualFeature() const
{ return itsVisFeature.getVal(); }

// ######################################################################
void ChannelBase::setVisualFeature(const VisualFeature v)
{ itsVisFeature.setVal(v); }

// ######################################################################
bool ChannelBase::isHomogeneous() const
{ return itsHomogeneous.getVal(); }

// ######################################################################
void ChannelBase::setHomogeneous(const bool h)
{ itsHomogeneous.setVal(h); }

// ######################################################################
void ChannelBase::readFrom(const ParamMap& pmap)
{
  killCaches();
  double wt = getTotalWeight();

  if (pmap.queryDoubleParam("weight", wt) == ParamMap::MISSING)
    setTotalWeight(1.0);
  else
    setTotalWeight(wt);
}

// ######################################################################
void ChannelBase::writeTo(ParamMap& pmap) const
{
  pmap.putDoubleParam("weight", getTotalWeight());
}

// ######################################################################
void ChannelBase::input(const Image<float>& bwimg, const double t,
                        const Image<byte>& clipMask)
{
  killCaches();
  doInput(0, &bwimg, t, clipMask);
}

// ######################################################################
void ChannelBase::input(const Image<PixRGB<float> >& colimg, const double t,
                        const Image<byte>& clipMask)
{
  killCaches();
  doInput(&colimg, 0, t, clipMask);
}

// ######################################################################
void ChannelBase::input(const Image<PixRGB<float> >& colimg,
                        const Image<float>& bwimg, const double t,
                        const Image<byte>& clipMask)
{
  killCaches();
  doInput(&colimg, &bwimg, t, clipMask);
}

// ######################################################################
void ChannelBase::setTotalWeight(const double wt)
{
  killCaches();
  itsTotalWeight.setVal(wt);
}

// ######################################################################
double ChannelBase::getTotalWeight() const
{ return itsTotalWeight.getVal(); }

// ######################################################################
void ChannelBase::setWeightPerChan(const double wt)
{
  killCaches();
  itsTotalWeight.setVal(wt * numSubmaps());
}

// ######################################################################
void ChannelBase::clampWeight(const double cmin, const double cmax)
{
  killCaches();
  setWeightPerChan(CLAMP(weightPerChan(), cmin, cmax));
}

// ######################################################################
void ChannelBase::cleanWeight(const double epsilon)
{
  killCaches();
  if (fabs(weightPerChan()) <= epsilon) itsTotalWeight.setVal(0.0);
}

// ######################################################################
double ChannelBase::weightPerChan() const
{ return itsTotalWeight.getVal() / numSubmaps(); }

// ######################################################################
bool ChannelBase::isZeroWeight() const
{ return (itsTotalWeight.getVal() == 0.0); }

// ######################################################################
void ChannelBase::setOutputRange(const float mini, const float maxi)
{
  itsOutputRangeMin.setVal(mini);
  itsOutputRangeMax.setVal(maxi);
}

// ######################################################################
void ChannelBase::drawResults(Image< PixRGB<byte> >& /*traj*/)
{
  /* do nothing */
}


// ######################################################################
void ChannelBase::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // default is a no-op
}

// ######################################################################
void ChannelBase::killCaches()
{
  /* no caches to worry about here */
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
