/*!@file ChannelBase.H The base class for all channel classes */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ChannelBase.H 4228 2005-02-23 17:19:25Z vidhya $
//

#ifndef CHANNELBASE_H_DEFINED
#define CHANNELBASE_H_DEFINED

#include "saliency.H"
#include "VisualFeatures.H"
#include "Image.H"
#include <vector>
#include "ModelComponent.H"

class Dims;
class JetSpec;
template <class T> class Jet;
class ParamMap;
class Point2D;
class Rectangle;
class OutputFrameSeries;

template <class T> class PixRGB;
template <class T> class SharedPtr;

class ChannelBase;

// ######################################################################
//! A callback class for use with ChannelBase and its derivatives.
/*! ChannelListener gets called back from Channel::getRawOutput() for each
    center/surround pairing. */
class ChannelListener
{
public:
  virtual ~ChannelListener();

  //! Subclasses of ChannelBase use this interface to callback to listeners.
  /*! Subclasses should pass *this as the first argument to onEachCS(), so
      that the client can figure out which channel triggered the
      callback. */
  virtual void onEachCS(const ChannelBase& channel,
                        const uint lev, const uint del,
                        const Image<float>& cs) = 0;
};

// ######################################################################
//! ChannelBase represents the various computational modules in VisualCortex.
class ChannelBase : public ModelComponent
{
public:
  //! Default constructor. See ModelComponent.H
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage
      @param vs The VisualFeature implemented by the channel */
  ChannelBase(ModelManager& mgr, const std::string& descrName,
              const std::string& tagName, const VisualFeature vs);

  //! Virtual destructor ensures proper destruction of derived classes.
  virtual ~ChannelBase();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  // ######################################################################
  /*! @name VisualFeature functions */
  //@{

  //! Get the visual feature
  VisualFeature visualFeature() const;

  //! Set the visual feature
  void setVisualFeature(const VisualFeature v);

  //! Is the channel homogeneous?
  /*! A homogeneous channel is one whose subchannels (and their subs)
    have identical VisualFeatures, and those are also identical to our
    VisualFeature */
  bool isHomogeneous() const;

  //! Set our homogeneous flag
  void setHomogeneous(const bool h);
  //@}

  // ######################################################################
  /*! @name ParamMap functions

      Subclasses should be sure to call ChannelBase's version of these
      functions within their own implementations, so that ChannelBase's
      parameters (such as its weights) are properly handled.

      This interface is different from that provided by
      ModelComponent, as here we only save a subset of all of the
      Channel's parameters, that are useful, e.g., for object
      recognition or feature-based attention biasing. Some of the
      values saved here may not even be ModelParam parameters. */
  //@{

  //! Read params from the ParamMap.
  virtual void readFrom(const ParamMap& pmap);

  //! Write params to the ParamMap.
  virtual void writeTo(ParamMap& pmap) const;

  //@}

  // ######################################################################
  /*! @name Weights+coeffs functions

      These functions manage the weights and coefficients maintained by the
      channel. In general, the coefficients determine the relative
      weighting of any submaps that are maintained by the channel, while
      the weight determines the overall level of output of the
      channel. Thus, the coefficients are for controlling
      <i>intra</i>-channel competition, while the weights are for
      controlling <i>inter</i>-channel competition. */
  //@{

  //! Set the total weight allocated to this channel.
  /*! This weight will be divided amongst any subchannels managed by this
      channel. So, for example, if the total weight is set to 2.0, and the
      channel has four subchannels, then each subchannel will get an
      effective weight of 0.5. The initial setting is 1.0 for most
      channels. But, If you are handling b/w images, for instance, you may
      set the weight for the color channel to 0.0 and save computations. */
  void setTotalWeight(const double wt);

  //! return the value set for total weight
  double getTotalWeight() const;

  //! Set the weight allocated per subchannel.
  /*! The total weight will then be the per-channel weight times the number
      of subchannels. */
  void setWeightPerChan(const double wt);

  //! Clamp the per-subchannel weight to the given range.
  void clampWeight(const double cmin, const double cmax);

  //! If the per-subchannel weight is less than \a eps, then it is set to zero.
  void cleanWeight(const double eps);

  //! Returns the weight allocated per subchannel.
  double weightPerChan() const;

  //! Check if the channel's weight is zero.
  /*! If true, then the user of the channel may find it more efficient to
      simply ignore the output from this channel. */
  bool isZeroWeight() const;

  //! Set any coefficients less than eps to zero.
  virtual void cleanCoeffs(const double eps) = 0;

  //! Clamp the weighting coefficients to the specified range.
  virtual void clampCoeffs(const double cmin, const double cmax) = 0;

  //! Return the sum of the absolute values of the weighting coefficients.
  virtual double absSumCoeffs() const = 0;

  //! Normalize the weighting coefficients by dividing by div.
  virtual void normalizeCoeffs(const double div) = 0;

  //! Adjust the weighting coefficients according to the target map.
  virtual void learn(const Image<byte>& dmap, const double eta,
                     const bool softmask = false,
                     const int inthresh = 20, const int outthresh = 255) = 0;

  //@}

  // ######################################################################
  /*! @name Input functions

      There are several overloaded input() functions. These are
      provided to allow maximum efficiency, so that only the minimum
      required information must be passed to the channel. These
      functions delegate to the virtual function doInput(), passing
      null pointers in place of missing inputs. The size of the input
      image(s) can later be queried with getInputDims(). The timestamp
      parameter (t, in seconds) is used to compute transient responses
      to flickering stimuli. */
  //@{
  //! Send monochrome (luminance) input to the channel.
  void input(const Image<float>& bwimg, const double t,
             const Image<byte>& clipMask = Image<byte>());

  //! Send multispectral (color) input to the channel.
  void input(const Image<PixRGB<float> >& colimg, const double t,
             const Image<byte>& clipMask = Image<byte>());

  //! Send luminance and color input to the channel.
  void input(const Image<PixRGB<float> >& colimg,
             const Image<float>& bwimg, const double t,
             const Image<byte>& clipMask = Image<byte>());
  //@}


  // ######################################################################
  /*! @name Access functions */
  //@{

  //! Set a fixed range of values for the raw output
  /*! By default, the range is set to [MAXNORMMIN .. MAXNORMMAX] at
    construction. If non-zero values are provided here, in getOutput()
    we will cumulate our various submaps, apply the provided range,
    apply spatial competition for salience, apply our total weight,
    and return the result. If a range [0.0 .. 0.0] is provided here,
    only spatial competition and weights will be applied, but the
    first step of applying the range will be skipped. This must be
    called before start(). */
  void setOutputRange(const float mini, const float maxi);

  //! Query whether the channel is ready to give yield valid output.
  /*! In some cases (e.g. with motion channels), it may take several input
      frames before the channel is fully initialized. */
  virtual bool outputAvailable() const = 0;

  //! Return the dimensions of the output image.
  virtual const Dims& getMapDims() const = 0;

  //! Return the dimensions of the most recent input image.
  virtual const Dims& getInputDims() const = 0;

  //! Fill the components of the given Jet that relate to this channel.
  /*! All relevant components of the Jet are set to this channel's
      values at the specified location across all pyramid levels. The
      channel will store its data starting at the given index in the
      jet, and will return the number of entries filled. The index
      vector will be modified as the various Jet components are
      explored, but will be restored to its original state upon
      completion. */
  virtual void fillJet(const Point2D& loc, Jet<float>& j,
                       std::vector<int>& index) const = 0;

  //! Fill the components of the given Jet that relate to this channel.
  /*! This version does not attempt to use interpolation */
  virtual void fillJetNI(const Point2D& loc, Jet<float>& j,
                         std::vector<int>& index) const = 0;

  //! Fill the specification of a Jet with data that relates to this channel
  virtual void fillJetSpec(JetSpec &js) const = 0;

  //! Return the number of feature maps represented by this channel.
  virtual uint numSubmaps() const = 0;

  //! Get the feature map at the given index.
  /*! This function is intended to be implemented by subclasses so that it
      recurses into any nested channels, mapping nested submaps into a
      linear sequence of indices. */
  virtual Image<float> getSubmap(const uint index) const = 0;

  //! Get the name/description of the feature map at the given index.
  /*! This function is intended to be implemented by subclasses so that it
      recurses into any nested channels, mapping nested submaps into a
      linear sequence of indices. The output of this function is intended
      only to be human-readable (e.g. for labeling image displays), and
      does not have to follow any particular parseable syntax. */
  virtual const char* getSubmapName(const uint index) const = 0;

  //! Get the coefficients (gain factors) for perceptual feature grouping
  virtual void getCoeff(std::vector<double>& coeff) = 0;

  //! Set the coefficient for submap idx for perceptual feature grouping
  virtual void setGroupingCoeff(std::vector<double>& coeff, uint& idx) = 0;

  //! Compute the percept by grouping features
  virtual Image<float> perceptualGrouping() = 0;

  //! Bias the features at submap idx with mean, intra-class std. dev
  //! and inter-sub-class std. dev
  virtual void biasFeatures(const double* mean, const double* sigma,
                            int& idx)  = 0;

  virtual void getFeatures(const Point2D& locn,
                           std::vector<double>& mean) const = 0;

  virtual void getFeaturesBatch(std::vector<Point2D*> *locn,
                                std::vector<std::vector<float> > *mean,
                                int *count) const = 0;

  //! Weight of the channel is the maximum weight of its subchannels
  /*! used for biasing to find targets */
  virtual void findWeight() = 0;

  //! Recognise this location in the channel by checking if it is obj
  virtual bool reco(const ParamMap& pmap, double& mismatch) = 0;

  //! what is the probability that the fixation belongs to this object?
  virtual void findLikelihood(const ParamMap& pmap, double& likelihood) = 0;

  //! The output will be the size of the feature map level in the pyramid.
  /*! (I.e., the size of the output is given by getMapDims(). */
  virtual Image<float> getOutput() = 0;

  //! Draws something channel-specific in the trajectory image.
  /*! The default implementation is a no-op; subclasses can override to do
      something interesting (see for example TemplateMatchChannel). */
  virtual void drawResults(Image< PixRGB<byte> >& traj);

  //! Save our various maps using an OutputFrameSeries
  /*! The default implementation is a no-op; subclasses can override to do
      something interesting */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //! Set a ChannelListener for this channel.
  /*! The ChannelListener which will get called once for each
      center/surround pairing. */
  virtual void setListener(SharedPtr<ChannelListener> listener) = 0;

  //! number of subchannels for ComplexChannel, 0 for SingleChannel
  virtual uint numChans() const = 0;

  //@}

protected:
  //! Subclasses override this implementation of the input() function.
  /*! This will be called by one of the input() methods. For efficiency
      reasons (e.g. network bandwidth in parallel code), it is possible
      that the color or luminance input may be null, although the code
      should be wired up properly so that given subclasses do not receive a
      null input that they in fact need. Subclasses should be sure to
      ASSERT that any used inputs are non-null. */
  virtual void doInput(const Image<PixRGB<float> >* colimg,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask) = 0;

  //! Subclasses override this to clear any caches they may maintain.
  /*! <b>Subclasses must be sure to explicitly call their base class's
      version of this function!</b>. This function will be called by
      ChannelBase when a new input image is received, or when new
      parameters are read from a ParamMap, etc. Subclasses may also call
      this function when they know that caches may have become invalid. */
  virtual void killCaches();

  ModelParam<float> itsOutputRangeMin, itsOutputRangeMax;

private:
  ChannelBase(const ChannelBase&); // not allowed
  ChannelBase& operator=(const ChannelBase&); // not allowed

  ModelParam<double> itsTotalWeight;
  ModelParam<VisualFeature> itsVisFeature;
  ModelParam<bool> itsHomogeneous;
};


/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !CHANNELBASE_H_DEFINED
