/*!@file Channels.C Feature channels for use in visual cortex classes */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Channels.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "Channels.H"
#include "MathFunctions.H"
#include "ModelManager.H"
#include "Image_All.H"
#include "ImageSet.H"
#include "Jet.H"
#include "LevelSpec.H"
#include "log.H"
#include "saliency.H"
#include "ParamMap.H"
#include "PyrBuilder.H"
#include "Pyramid_Ops.H"
#include "ModelOptionDefs.H"
#include "StringConversions.H"
#include "Image_Transforms.H"
#include "Image_ColorOps.H"  // for infoMeasure()
#include "Image_FilterOps.H"
#include "colorDefs.H"      // for HueChannel
#include "ImageCache.H"

#include <cstdio>
#include <fstream>
#include <iomanip>
#include <iostream> // std::cerr doesn't compile for some reason
#include <limits>
#include <algorithm> // for std::count(), etc.
#include <cmath>
#include <numeric> // for std::accumulate(), etc.

// ######################################################################
// ######################################################################
// SingleChannel subclass member definitions:
// ######################################################################
// ######################################################################


// ######################################################################
// IntensityChannel member definitions:
// ######################################################################

IntensityChannel::IntensityChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Intensity", "intensity", INTENS,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
}

IntensityChannel::~IntensityChannel() {}

// ######################################################################
// ConvolveChannel member definitions:
// ######################################################################

ConvolveChannel::ConvolveChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Convolve", "convolve", CONVOLVE,
                SharedPtr< PyrBuilder<float> >(NULL))
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
}

void ConvolveChannel::setFilter(const Image<float>& filter)
{
  setPyramid(makeShared(new ConvolvePyrBuilder<float>(filter)));
}

ConvolveChannel::~ConvolveChannel() {}

// ######################################################################
// RGBConvolveChannel member definitions:
// ######################################################################

RGBConvolveChannel::RGBConvolveChannel(ModelManager& mgr) :
  SingleChannel(mgr, "RGBConvolve", "rgbconvolve", RGBCONVOLVE,
                SharedPtr< PyrBuilder<float> >(NULL))
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
}

void RGBConvolveChannel::setFilters(const Image<float>& rfilter,
                                    const Image<float>& gfilter,
                                    const Image<float>& bfilter)
{
  setPyramid(makeShared(new RGBConvolvePyrBuilder<float>(rfilter,
                                                         gfilter,
                                                         bfilter)));
}

RGBConvolveChannel::~RGBConvolveChannel() {}

void RGBConvolveChannel::doInput(const Image<PixRGB<float> > *colimg,
                                 const Image<float> *, const double t,
                                 const Image<byte>& clipMask)
{
  ASSERT(colimg != NULL);

  setClipPyramid(clipMask);
  storePyramid(computePyramid(*colimg), t);
}

ImageSet<float> RGBConvolveChannel::computePyramid(const
                                                   Image< PixRGB<float> >& img)
{
  SharedPtr< RGBConvolvePyrBuilder<float> > pb;
  pb.dynCastFrom(itsPyrBuilder);

  // Compute our pyramid and return it
  return pb->build(img, itsLevelSpec.getVal().maxDepth());
}

// ######################################################################
// RedGreenChannel member definitions:
// ######################################################################

RedGreenChannel::RedGreenChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Red/Green", "rg", RG,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

RedGreenChannel::~RedGreenChannel() {}

// ######################################################################
// BlueYellowChannel member definitions:
// ######################################################################

BlueYellowChannel::BlueYellowChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Blue/Yellow", "by", BY,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

BlueYellowChannel::~BlueYellowChannel() {}

// ######################################################################
// Single Opponent RedGreenChannel member definitions:
// ######################################################################

SORedGreenChannel::SORedGreenChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Red/Green", "r-g", SORG, 
                SharedPtr<PyrBuilder<float> >(NULL))
{
  useSingleOpponence(true);
}

SORedGreenChannel::~SORedGreenChannel() {}

// ######################################################################
// Single Opponent GreenRedChannel member definitions:
// ######################################################################

SOGreenRedChannel::SOGreenRedChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Green/Red", "g-r", SOGR, 
                SharedPtr<PyrBuilder<float> >(NULL))
{
  useSingleOpponence(true);
}

SOGreenRedChannel::~SOGreenRedChannel() {}

// ######################################################################
// Single Opponent BlueYellowChannel member definitions:
// ######################################################################

SOBlueYellowChannel::SOBlueYellowChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Blue/Yellow", "b-y", SOBY, 
                SharedPtr<PyrBuilder<float> >(NULL))
{
  useSingleOpponence(true);
}

SOBlueYellowChannel::~SOBlueYellowChannel() {}

// ######################################################################
// Single Opponent YellowBlueChannel member definitions:
// ######################################################################

SOYellowBlueChannel::SOYellowBlueChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Yellow/Blue", "y-b", SOYB, 
                SharedPtr<PyrBuilder<float> >(NULL))
{
  useSingleOpponence(true);
}

SOYellowBlueChannel::~SOYellowBlueChannel() {}

// ######################################################################
// RedChannel member definitions:
// ######################################################################
RedChannel::RedChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Red", "r", R,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

RedChannel::~RedChannel() {}

// ######################################################################
// GreenChannel member definitions:
// ######################################################################
GreenChannel::GreenChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Green", "g", G,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

GreenChannel::~GreenChannel() {}

// ######################################################################
// BlueChannel member definitions:
// ######################################################################
BlueChannel::BlueChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Blue", "b", B,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

BlueChannel::~BlueChannel() {}

// ######################################################################
// YellowChannel member definitions:
// ######################################################################
YellowChannel::YellowChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Yellow", "y", Y,
                makeShared(new GaussianPyrBuilder<float>(5)))
{
  itsTakeAbs.setVal(true);
}

YellowChannel::~YellowChannel() {}

// ######################################################################
// GaborChannel member definitions:
// ######################################################################

// ######################################################################
GaborChannel::GaborChannel(ModelManager& mgr, const uint oriIndex,
                           const double ang) :
  SingleChannel(mgr, "", "", ORI, SharedPtr< PyrBuilder<float> >(NULL)),
  itsOriIndex(this, "GaborChannelOrientationIndex", oriIndex),
  itsOrientation(this, "GaborChannelOrientation", ang),
  itsGaborIntens(this, "GaborChannelIntensity", 10.0),
  itsOriCompType(this, "OrientComputeType", ORISteerable)
{
  itsNormalizeOutput.setVal(true);

  char name[256];
  snprintf(name, 256, "Gabor(%d)", int(ang));
  setDescriptiveName(name);

  snprintf(name, 256, "ori_%d", oriIndex);
  setTagName(name);
}

// ######################################################################
void GaborChannel::start1()
{
  SingleChannel::start1();
  resetPyramid();
}

// ######################################################################
GaborChannel::~GaborChannel()
{ }

// ######################################################################
void GaborChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsOriCompType);
    manager->requestOption(this, itsGaborIntens);
  }

  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void GaborChannel::readFrom(const ParamMap& pmap)
{
  SingleChannel::readFrom(pmap);
  double gi = itsGaborIntens.getVal();
  int otype = (int)itsOriCompType.getVal();
  bool doPyrReset = false;


  if (pmap.queryDoubleParam("gaborIntens", gi) == ParamMap::CHANGED)
    {
      itsGaborIntens.setVal(gi);
      doPyrReset = true;
    }

  if (pmap.queryIntParam("oriCompType", otype) == ParamMap::CHANGED)
    {
      itsOriCompType.setVal((OrientComputeType)otype);
      doPyrReset = true;
    }

  if (doPyrReset) resetPyramid();
}

// ######################################################################
void GaborChannel::writeTo(ParamMap& pmap) const
{
  SingleChannel::writeTo(pmap);
  pmap.putDoubleParam("gaborIntens", itsGaborIntens.getVal());
  pmap.putIntParam("oriCompType", (int)itsOriCompType.getVal());
}

// ######################################################################
double GaborChannel::angle() const
{ return itsOrientation.getVal(); }

// ######################################################################
void GaborChannel::resetPyramid()
{
  switch(itsOriCompType.getVal())
    { 
    case ORISteerable:
      setPyramid(makeShared(new OrientedPyrBuilder<float>
                            (9, itsOrientation.getVal(),
                             itsGaborIntens.getVal())));
      break;

    case ORIGabor:
      setPyramid(makeShared(new GaborPyrBuilder<float>
                            (itsOrientation.getVal(),7,1,9,0)));
       break;

    case ORIGaborEnergyNorm:
      setPyramid(makeShared(new GaborPyrBuilder<float>
                            (itsOrientation.getVal(),7,1,9,
                             DO_ENERGY_NORM)));
       break;

    default:
      LFATAL("Unknown OrientComputeType");
    }
}


// ######################################################################
// FlickerChannel member definitions:
// ######################################################################

// ######################################################################
FlickerChannel::FlickerChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Flicker", "flicker", FLICKER,
                makeShared(new GaussianPyrBuilder<float>(5))),
  itsTimeConst(this, "FlickerChannelTimeConst", 0.03),
  itsPrevInput(),
  itsPrevTime(0.0),
  itsOutput()
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
}

// ######################################################################
FlickerChannel::~FlickerChannel() {}

// ######################################################################
void FlickerChannel::learn(const Image<byte>& /*dmap*/, double /*eta*/)
{
  // do nothing; we can't "learn" flicker from a single input image
}

// ######################################################################
void FlickerChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  ASSERT(bwimg != 0);

  // If this is the first time the flicker channel has seen input, then
  // itsPrevInput will be uninitialized; obviously we can't compute any
  // flicker with only one frame, so we just store the current input as the
  // next iteration's previous input
  if (!itsPrevInput.initialized())
    {
      itsPrevInput = *bwimg;
      itsPrevTime = t;
    }
  else
    {
      // take simple abs difference between current and previous frame:
      Image<float> fli = absDiff(*bwimg, itsPrevInput);

      SingleChannel::doInput(0, &fli, t, clipMask);

      // update the previous image:
      if (t - itsPrevTime >= itsTimeConst.getVal())
        {
          itsPrevInput = *bwimg;
          itsPrevTime = t;
        }
      else
        {
          float fac = float( (t - itsPrevTime) / itsTimeConst.getVal() );
          itsPrevInput = itsPrevInput * (1.0F - fac) + (*bwimg) * fac;
          itsPrevTime = t;
        }
    }

  // invalidate the previously computed intermediate results, which will
  // force recomputation next time we want to access them
  itsOutput.freeMem();
}

// ######################################################################
Image<float> FlickerChannel::getOutput()
{
  if ( !this->outputAvailable() )
    {
      LERROR("No flicker channel yet! -- IGNORING.");
      int lev = itsLevelSpec.getVal().mapLevel();
      if (lev == 0) return Image<float>(itsPrevInput.getDims(), ZEROS);

      return Image<float>(Dims(itsPrevInput.getDims().w() / (1 << lev),
                               itsPrevInput.getDims().h() / (1 << lev)),
                          ZEROS);
    }

  // already computed?
  if (!itsOutput.initialized())
    {
      itsOutput = combineSubMaps();
      if (itsNormalizeOutput.getVal())
        itsOutput = maxNormalize(itsOutput,
                                 itsOutputRangeMin.getVal(),
                                 itsOutputRangeMax.getVal(),
                                 itsNormType.getVal());
    }

  return itsOutput;
}

// ######################################################################
void FlickerChannel::reset(const bool recurse)
{
  // reset some stuff in FlickerChannel
  itsPrevInput.freeMem();
  itsOutput.freeMem();

  // call reset in SingleChannel
  SingleChannel::reset(recurse);
}

// ######################################################################
// ColorBandChannel member definitions:
// ######################################################################
ColorBandChannel::ColorBandChannel(ModelManager& mgr, const uint bandidx) :
  SingleChannel(mgr, "", "", COLBAND,
                makeShared(new GaussianPyrBuilder<float>(5))),
  itsBandIndex(this, "ColorBandChannelIndex", bandidx)
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
  
  char name[256];
  snprintf(name, 256, "ColorBand(%d)", bandidx);
  setDescriptiveName(name);

  snprintf(name, 256, "cband_%d", bandidx);
  setTagName(name);
}

// ######################################################################
ColorBandChannel::~ColorBandChannel()
{ }


// ######################################################################
// DisparityChannel member definitions:
// ######################################################################

// ######################################################################
DisparityChannel::DisparityChannel(ModelManager& mgr, const uint dIndex,
                                       const double ang, const double pha) :
  SingleChannel(mgr, "", "", DISPARITY, SharedPtr< PyrBuilder<float> >(NULL)),
  itsIndex(this, "DisparityChannelIndex", dIndex),
  itsOrientation(this, "DisparityChannelOrientation", ang),
  itsPhase(this, "DisparityChannelPhase", pha)
{
  //  itsNormalizeOutput.setVal(true);

  char name[256];
  snprintf(name, 256, "Disparity(%d,%d)", int(ang), int(pha));
  setDescriptiveName(name);

  snprintf(name, 256, "d_%d", dIndex);
  setTagName(name);

}

// ######################################################################
DisparityChannel::~DisparityChannel()
{ }

// ######################################################################
void DisparityChannel::exportOptions(const int whatToExport, 
                                     const bool recurse)
{
   
  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);

} 

// ######################################################################
void DisparityChannel::start1()
{
  SingleChannel::start1();

  // initilialize the size of the disparity map
  depth = itsLevelSpec.getVal().maxDepth();
  dispMap.reset(depth);
 
}

// ######################################################################
double DisparityChannel::angle() const
{ 
  return itsOrientation.getVal(); 
}

// ######################################################################
double DisparityChannel::phase() const
{ 
  return itsPhase.getVal(); 
}

// ######################################################################
void DisparityChannel::setRawFilteredImages(Image<float> ***fImgLE, 
                                            Image<float> ***fImgRI)
{
  fImgL = fImgLE;
  fImgR = fImgRI;
}

// ######################################################################
void DisparityChannel::getRawFilteredImages(Image<float> ****fImgLE,
                                            Image<float> ****fImgRI)
{
  *fImgLE = fImgL; 
  *fImgRI = fImgR;
}

// ######################################################################
void DisparityChannel::getDispMap(ImageSet<float> *dMap)
{  
  *dMap = dispMap; 
}

// ######################################################################
void DisparityChannel::normalizeDispMap(ImageSet<float> tDispMap, int nChan)
{float min,max;
  for(uint d = 0; d < depth; d++)
  {
    dispMap[d] /= tDispMap[d];
    dispMap[d].getMinMax(min,max);
    printf("MIN:%f, MAX:%f\n\n\n",min,max);

    // FIXME: get the numer of channels from the
    dispMap[d] -= (float)(1.0F/(float)(nChan));
     dispMap[d].getMinMax(min,max);
    printf("-- MIN:%f, MAX:%f\n\n\n",min,max);
   dispMap[d].rectify();
    dispMap[d].getMinMax(min,max);
    printf("RR MIN:%f, MAX:%f\n\n\n",min,max);

  }
}

// ######################################################################
void DisparityChannel::setNumPhase(int n)
{ 
  nPhase = n;
} 

// ######################################################################
void DisparityChannel::setNumTheta(int t)
{ 
  nTheta = t;
} 

// ######################################################################
void DisparityChannel::doInput(const Image< PixRGB<float> >*,
                               const Image<float>* bwimg, const double t,
                               const Image<byte>& clipMask)
{
  // for now only evaluate the horizontal disparity 
  // and make sure the phase is normalized to 0 to 360
  //int t  = 0; //(int)(round((itsOrientation.getVal()/180.0)*nTheta));
  //int dp = (int)(round((itsPhase.getVal()/360.0)*nPhase));
  float dt = 1.2;//itsPhase.getVal();
  if(itsPhase.getVal() > 180.0) dt = -1.2;//dt -= 360.0;
  if(itsPhase.getVal() == 0.0) dt = 0.0;
  float  dx = dt;
  //float dx = period * dt/360.0; 
  //printf("dx = %f \n",dx);

  int width  = fImgL[0][0][0].getWidth(); 
  int height = fImgL[0][0][0].getHeight();

  Image<float> sumL(width,height,ZEROS);
  Image<float> sumR(width,height,ZEROS);
   
  for(uint d = 0; d < 1; d++)
  {    
    dispMap[d].resize(width,height,true);
    /*  
    // add the phase shift portion
    for(int p = 0; p < nPhase; p++)
      dispMap[d] +=
       fImgL[d][0][p] * fImgR[d][0][(p+dp)%nPhase];
    */  
    
    //int t = 0;
    // add the position shift portion
    // the equation is already derived
    for(uint t = 0; t < nTheta; t++)
      for(uint p = 0; p < nPhase; p++)
      {
        Image<float> tmp = 
          fImgL[d][t][p] * shiftImage(fImgR[d][t][p],dx,0);
        tmp.rectify();
        dispMap[d] += squared(tmp);
     }
    
    if(width  > 1) width  /= 2;
    if(height > 1) height /= 2;

  }

  // build the feature map pyramid and store it
  dispMap = buildPyrGaussian((dispMap[0]),depth,5); 
  SingleChannel::storePyramid(dispMap,t);
}

// store the disparity map as feature map  
void DisparityChannel::storePyramid(double t)
{    
  SingleChannel::storePyramid(dispMap, t);
}


// ######################################################################
// MultiConvolveChannel member definitions:
// ######################################################################
MultiConvolveChannel::MultiConvolveChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "MultiConvolve", "multiconv", CONVOLVE),
  itsFilterFname(this, "MultiConvolveChannelFilterFname", ""),
  itsLumThresh(this, "MultiConvolveChannelLuminanceThreshold", 25.5F)
{
  // since we only support 4 subchannels, let's build and name them,
  // though we don't know what their filters are yet; we'll assign a
  // filter to them in start1():
  for (int n = 0; n < 4; n ++)
    {
      SharedPtr<ConvolveChannel> cc(new ConvolveChannel(*manager));
      
      // let's change our subchan's name:
      char name[100];
      sprintf(name, "Convolve%d", n); cc->setDescriptiveName(name);
      sprintf(name, "conv%d", n); cc->setTagName(name);
      
      // add cc as one of our subchannels:
      addSubChan(cc);
    }
}

// ######################################################################
MultiConvolveChannel::~MultiConvolveChannel()
{ }

// ######################################################################
void MultiConvolveChannel::start1()
{
  // let's configure our subchannels by reading their kernel from
  // text file:
  const char *fname = itsFilterFname.getVal().c_str();
  FILE *f = fopen(fname, "r");
  if (f == NULL) LFATAL("Cannot open %s -- ABORT", fname);

  int w, h, d;
  if (fscanf(f, "%d %d %d\n", &w, &h, &d) != 3)
    LFATAL("Bogus first line in %s -- ABORT", fname);

  if (d != 4) LFATAL("Only 4 subchannels are supported -- ABORT");

  LINFO("Building %dx%dx%d kernel from '%s'", w, h, d, fname);
  for (int n = 0; n < d; n ++)
    {
      Image<float> ker(w, h, NO_INIT);
      for (int j = 0; j < h; j ++)
        for (int i = 0; i < w; i ++) {
          float coeff;
          if (fscanf(f, "%f\n", &coeff) != 1)
            LFATAL("Bogus coeff in %s at (%d, %d) -- ABORT", fname, i, j);
          ker.setVal(i, j, coeff);
        }

      // ok, let's assign this filter to one of our subchannels:
      (dynamic_cast<ConvolveChannel&>(subChan(n))).setFilter(ker);
    }
  fclose(f);
}

// ######################################################################
void MultiConvolveChannel::doInput(const Image<PixRGB<float> >* colimg,
                                   const Image<float>*, const double t,
                                   const Image<byte>& clipMask)
{
  ASSERT(colimg != 0);

  Image<float> r, g, b, y;
  getRGBY(*colimg, r, g, b, y, itsLumThresh.getVal());

  ASSERT(numChans() == 4); // only 4 subchannels currently supported

  subChan(0).input(r, t, clipMask);
  subChan(1).input(g, t, clipMask);
  subChan(2).input(b, t, clipMask);
  subChan(3).input(y, t, clipMask);
}

// ######################################################################
// Composite Color Channel member definitions: contains single and double
// opponent color channels
// ######################################################################
CompositeColorChannel::CompositeColorChannel(ModelManager& mgr,
                                             const double dw, const double sw):
  ComplexChannel(mgr, "CompositeColor", "composite-color", COMPOSITECOLOR),
  itsDoubleOpponent(dw),
  itsSingleOpponent(sw)
{
  // let's build our member channels
  if (dw > 0.0)
    {
      uint i = addSubChan(makeShared(new ColorChannel(*manager)));
      channel(i).setTotalWeight(dw);
    }
  if (sw > 0.0)
    {
      uint i = addSubChan(makeShared(new SOColorChannel(*manager)));
      channel(i).setTotalWeight(sw);
    }
}
// ######################################################################
CompositeColorChannel::~CompositeColorChannel() {}

// ######################################################################
ChannelBase& CompositeColorChannel::channel(const uint idx) const
{ return dynamic_cast<ChannelBase&>(*(subComponent(idx))); }

// ######################################################################
ChannelBase& CompositeColorChannel::channel(const char* name) const
{ return dynamic_cast<ChannelBase&>(*(subComponent(name))); }

// ######################################################################
void CompositeColorChannel::doInput(const Image<PixRGB<float> >* inp,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  for (uint i = 0; i < numChans(); ++i)
    {
      channel(i).input(*inp, *bwimg, t, clipMask);
    }
  LINFO("CompositeColor channel ok.");
}

// ######################################################################
// Double Opponent ColorChannel member definitions:
// ######################################################################

ColorChannel::ColorChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "Color", "color", COLOR),
  itsLumThresh(this, "ColorChannelLuminanceThreshold", 25.5F),
  itsMethod(this,"ColorComputeType",COLstandard)
{
  itsRGidx = addSubChan(makeShared(new RedGreenChannel(mgr)));
  itsBYidx = addSubChan(makeShared(new BlueYellowChannel(mgr)));
}

// ######################################################################
RedGreenChannel& ColorChannel::rg() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<RedGreenChannel&>(subChan(itsRGidx));
}

// ######################################################################
BlueYellowChannel& ColorChannel::by() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<BlueYellowChannel&>(subChan(itsBYidx));
}

// ######################################################################
ColorChannel::~ColorChannel()
{ }

// ######################################################################
void ColorChannel::doInput(const Image<PixRGB<float> >* colimg,
                           const Image<float>*, const double t,
                           const Image<byte>& clipMask)
{
  ASSERT(colimg != 0);

  Image<float> rg, by;
  switch(itsMethod.getVal())
    {
    case COLstandard:
      LINFO("Using 'standard' method to compute color opponencies.");
      getRGBY(*colimg, rg, by, itsLumThresh.getVal());
      break;

    case COLsimple:
      LINFO("Using 'simple' method to compute color opponencies.");
      getRGBYsimple(*colimg, rg, by, itsLumThresh.getVal());
      break;

    default:
      LFATAL("Unknown ColorComputeType!");
      break;
    }
  subChan(itsRGidx).input(rg, t, clipMask);
  subChan(itsBYidx).input(by, t, clipMask);
  LINFO("Double Opponent Color channel ok.");
}

// ######################################################################
void ColorChannel::exportOptions(const int whatToExport,
                                 const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    {
      manager->requestOption(this, itsMethod);
    }

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
// Single Opponent ColorChannel member definitions:
// ######################################################################
SOColorChannel::SOColorChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "SOColor", "single-opponent-color", SOCOLOR),
  itsLumThresh(this, "ColorChannelLuminanceThreshold", 25.5F)
{
  itsRGidx = addSubChan(makeShared(new SORedGreenChannel(mgr)));
  itsGRidx = addSubChan(makeShared(new SOGreenRedChannel(mgr)));
  itsBYidx = addSubChan(makeShared(new SOBlueYellowChannel(mgr)));
  itsYBidx = addSubChan(makeShared(new SOYellowBlueChannel(mgr)));

  // the following channels are computed to obtain the pyramids for
  // the above channels
  itsRidx = addSubChan(makeShared(new RedChannel(mgr)));
  itsGidx = addSubChan(makeShared(new GreenChannel(mgr)));
  itsBidx = addSubChan(makeShared(new BlueChannel(mgr)));
  itsYidx = addSubChan(makeShared(new YellowChannel(mgr)));
  
  // these pure color channels are not considered for saliency
  r().setTotalWeight(0.0);
  g().setTotalWeight(0.0);
  b().setTotalWeight(0.0);
  y().setTotalWeight(0.0);  
}

// ######################################################################
SORedGreenChannel& SOColorChannel::rg() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<SORedGreenChannel&>(subChan(itsRGidx));
}

// ######################################################################
SOGreenRedChannel& SOColorChannel::gr() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<SOGreenRedChannel&>(subChan(itsGRidx));
}

// ######################################################################
SOBlueYellowChannel& SOColorChannel::by() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<SOBlueYellowChannel&>(subChan(itsBYidx));
}

// ######################################################################
SOYellowBlueChannel& SOColorChannel::yb() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<SOYellowBlueChannel&>(subChan(itsYBidx));
}
// ######################################################################
RedChannel& SOColorChannel::r() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<RedChannel&>(subChan(itsRidx));
}

// ######################################################################
GreenChannel& SOColorChannel::g() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<GreenChannel&>(subChan(itsGidx));
}

// ######################################################################
BlueChannel& SOColorChannel::b() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<BlueChannel&>(subChan(itsBidx));
}

// ######################################################################
YellowChannel& SOColorChannel::y() const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<YellowChannel&>(subChan(itsYidx));
}
// ######################################################################
SOColorChannel::~SOColorChannel()
{ }

// ######################################################################
void SOColorChannel::doInput(const Image<PixRGB<float> >* colimg,
                             const Image<float>*, const double t,
                             const Image<byte>& clipMask)
{
  ASSERT(colimg != 0);

  Image<float> red, green, blue, yellow;
  getRGBY(*colimg, red, green, blue, yellow, itsLumThresh.getVal());
  // first find the pyramids for each color
  subChan(itsRidx).input(red, t, clipMask);
  subChan(itsGidx).input(green, t, clipMask);  
  subChan(itsBidx).input(blue, t, clipMask);
  subChan(itsYidx).input(yellow, t, clipMask);

  // then compute the single opponent center-surround using the above
  // pyramids
  rg().singleOpponentInput(r().pyramid(0), g().pyramid(0), t, clipMask);
  gr().singleOpponentInput(g().pyramid(0), r().pyramid(0), t, clipMask);
  by().singleOpponentInput(b().pyramid(0), y().pyramid(0), t, clipMask);
  yb().singleOpponentInput(y().pyramid(0), b().pyramid(0), t, clipMask);
  LINFO("Single Opponent Color channel ok.");    
}


// ######################################################################
// OrientationChannel member definitions:
// ######################################################################

// ######################################################################
OrientationChannel::OrientationChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "Orientation", "orientation", ORI),
  itsNumOrients(this, "NumOrientations", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE),
  itsInteractString(this, "OriInteraction", "None"),
  itsInteractType(NONE),
  itsDidInteractions(false)
{
  // let's build our channels; we may have to re-build them if
  // itsNumOrient get changed on us before we start():
  buildSubChans();
}

// ######################################################################
void OrientationChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  // let's instantiate our Gabor subchannels now that we know how many
  // we want. They will inherit the current values (typically
  // post-command-line parsing) of all their options as they are
  // constructed:
  LINFO("Using %d orientations spanning [0..180]deg", itsNumOrients.getVal());
  for (uint ori = 0; ori < itsNumOrients.getVal(); ++ori)
    addSubChan(makeShared
               (new GaborChannel(*manager, ori, 180.0 * double(ori) /
                                 double(itsNumOrients.getVal()))));
  // Here, we just use GaborChannel's default tag names

  // if we have been requested to export some options, let's ask it
  // again on our newly built channels:
  if (itsExportOpts != OPTEXP_NONE)
    exportOptions(itsExportOpts, true);
  //ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
void OrientationChannel::parseInteractString(const std::string& value)
{

  // need to clear the coefficients in any case
  itsInteractCoeffs.clear();

  if (value.compare("None") == 0)
    {
      itsInteractType = NONE;
      return;
    }

  if (value.compare("SubtractMean") == 0)
    {
      itsInteractType = SUB_MEAN;
      return;
    }

  itsInteractType = CUSTOM;

  // format here is "c.c,...,c.c"
  int curpos = 0, len = value.length();
  while (curpos < len)
    {
      // get end of next number
      int nextpos = value.find_first_not_of("-.0123456789eE",curpos);
      if (nextpos == -1) nextpos = len;

      // no number characters found -> bummer
      if (nextpos == curpos) 
        LFATAL("Error parsing the OriInteract string '%s' - found '%c' "
               "instead of a number.",value.data(),value[curpos]);

      // now let's see - can we get a number here?
      float coeff;
      int rep = sscanf(value.substr(curpos,nextpos-curpos).data(),"%g",&coeff);

      // couldn't read a number -> bummer
      if (rep != 1)
        LFATAL("Error parsing OriInteract string '%s' - found '%s' instead of "
               "a number.", value.data(),
               value.substr(curpos,nextpos-curpos).data());

      // yeah! found a number -> store it
      itsInteractCoeffs.push_back(coeff);

      LDEBUG("coeff = %g; value[nextpos] = '%c'",coeff,value[nextpos]);

      // not a comma -> bummer
      if ((nextpos < len) && (value[nextpos] != ',')) 
        LFATAL("Error parsing the OriInteract string '%s' - found '%c' "
               "instead of ','.",value.data(),value[nextpos]);

      // the character right after the comma should be a number again
      curpos = nextpos + 1;
    }

  // end of string, done
  return;
}

// ######################################################################
bool OrientationChannel::setModelParamString(const std::string& name,
                                             const std::string& value,
                                             const bool recurse,
                                             const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumOrients.getName()) == 0 &&
      numChans() != itsNumOrients.getVal())
    buildSubChans();

  // if the param is our OriInteraction, then parse the string
  if (name.compare(itsInteractString.getName()) == 0)
    parseInteractString(value);

  return ret;
}

// ######################################################################
OrientationChannel::~OrientationChannel()
{ }

// ######################################################################
void OrientationChannel::exportOptions(const int whatToExport,
                                       const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    {
      manager->requestOption(this, itsNumOrients);
      manager->requestOption(this, itsInteractString);
    }

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
GaborChannel& OrientationChannel::gabor(const uint idx) const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<GaborChannel&>(ComplexChannel::subChan(idx));
}

// ######################################################################
void OrientationChannel::doInput(const Image<PixRGB<float> >*,
                                 const Image<float>* bwimg, const double t,
                                 const Image<byte>& clipMask)
{
  ASSERT(bwimg != 0);

  for (uint i = 0; i < numChans(); ++i)
    {
      gabor(i).input(*bwimg, t, clipMask);
      LINFO("Orientation pyramid (%d/%d) ok.", i+1, numChans());
    }
  itsDidInteractions = false;
}

// ######################################################################
void OrientationChannel::doInteractions()
{
  LINFO("OriInteractionType is %s.",itsInteractString.getVal().data());
  switch(itsInteractType)
    {
    case NONE:
      {
        // compute oriented gabor pyramids in several basis directions:
        break;
      }

    case SUB_MEAN:
      {
        // compute the average and subtract it from each GaborPyramid
        // this is actually a special case of the CUSTOM one, but
        // this method here is more efficient
        ImageSet<float> mean;

        // compute the mean
        for (uint i = 0; i < numChans(); ++i)
          {
            gabor(i).getOutput(); // make sure that gabor(i) has a pyramid
            if (mean.isEmpty())
              mean = gabor(i).itsPq.back().pyr;
            else
              mean += gabor(i).itsPq.back().pyr;
          }
        mean /= (float)numChans();

        // clampdiff each pyramid with the mean and feed them back
        // into the channel:
        for (uint i = 0; i < numChans(); ++i)
          {
            gabor(i).getOutput(); // make sure that gabor(i) has a pyramid
            gabor(i).itsPq.back().pyr =
              clampedDiff(gabor(i).itsPq.back().pyr,mean);
            LINFO("Orientation pyramid interactions (%d/%d) ok.",
                  i+1, numChans());
          }

        // done
        break;
      }

    case CUSTOM:
      {
        // In this case we have arbitrary linear interactions between the 
        // orientations - takes a bit of computation

        // make sure we have the correct number of coefficients
        ASSERT(itsInteractCoeffs.size() == numChans());
        std::vector< ImageSet<float> > resPyrs(numChans());

        // compute the pyramid interactions
        for (uint i = 0; i < numChans(); ++i)
          {
            ImageSet<float> curPyr = gabor(i).itsPq.back().pyr;
            uint coeff_ptr = i;

            // add this new pyramid with the correct coefficients
            for (uint j = 0; j < numChans(); ++j)
              {
                if (resPyrs[j].isEmpty())
                  resPyrs[j] =  (curPyr * itsInteractCoeffs[coeff_ptr]);
                else
                  resPyrs[j] += (curPyr * itsInteractCoeffs[coeff_ptr]);

                // count down the coeff_ptr and wrap it around
                // We count down here, because at parsing the input string
                // we have used push_back and hence reversed the order.
                coeff_ptr = (coeff_ptr - 1) % numChans();
              }
          }

        // now clamp these pyramids and feed them back to the channel
        for (uint i = 0; i < numChans(); ++i)
          {
            doLowThresh(resPyrs[i],0.0f,0.0f);
            gabor(i).itsPq.back().pyr = resPyrs[i];
            LINFO("Orientation pyramid interactions (%d/%d) ok.",
                  i+1, numChans());
          }

        // done
        break;
      }
    default: LFATAL("Unknown orientation interaction type: %d",
                    itsInteractType);
    } // end switch
}

// ######################################################################
Image<float> OrientationChannel::getOutput()
{
  if (!itsDidInteractions)
    {
      doInteractions();
      itsDidInteractions = true;
    }
  return ComplexChannel::getOutput();
}

// ######################################################################
OrientationChannel::InteractType OrientationChannel::getInteractType()
{ return itsInteractType; }

// ######################################################################
void OrientationChannel::setInteractType(OrientationChannel::InteractType type)
{ itsInteractType = type; }

// ######################################################################
std::vector<float> OrientationChannel::getInteractCoeffs()
{ return itsInteractCoeffs; }

// ######################################################################
void OrientationChannel::setInteractCoeffs(std::vector<float>& coeffs)
{ itsInteractCoeffs = coeffs; }



// ######################################################################
// MultiColorBandChannel member definitions:
// ######################################################################

// ######################################################################
MultiColorBandChannel::MultiColorBandChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "MultiColorBand", "multicolorband", COLBAND),
  itsNumBands(this, "NumColorBands", 3),  // see ModelOptionDefs.C
  itsSatBands(this, "NumSatBands", 3),  // see ModelOptionDefs.C
  itsHueSigma(this, "HueBandWidth", 30),  // see ModelOptionDefs.C
  itsSatSigma(this, "SatBandWidth", 0.3),  // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE)
{
  // let's create our subchannels (may be reconfigured if itsNumBands
  // gets changed):
  buildSubChans();
}

// ######################################################################
void MultiColorBandChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  // let's instantiate our subchannels now that we know how many we
  // want. They will inherit the current values (typically
  // post-command-line parsing) of all their options as they are
  // constructed:
  LINFO("Using %d hue bands", itsNumBands.getVal());
  for (uint i = 0; i < itsNumBands.getVal(); i ++)
    addSubChan(makeShared(new ColorBandChannel(*manager, i)));
  LINFO("Using %d saturation bands", itsSatBands.getVal());
  for (uint i = 0; i < itsSatBands.getVal(); i ++)
    addSubChan(makeShared(new ColorBandChannel(*manager, 
                                               itsNumBands.getVal()+i)));
  // Here, we just use our subchannels' default tag names

  // if we have been requested to export some options, let's ask it
  // again on our newly built channels:
  if (itsExportOpts != OPTEXP_NONE)
    ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
bool MultiColorBandChannel::setModelParamString(const std::string& name,
                                                const std::string& value,
                                                const bool recurse,
                                                const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumBands.getName()) == 0 &&
      numChans() != (itsNumBands.getVal() + itsSatBands.getVal()))
    buildSubChans();

  if (name.compare(itsSatBands.getName()) == 0 &&
      numChans() != (itsNumBands.getVal() + itsSatBands.getVal()))
    buildSubChans();

  return ret;
}

// ######################################################################
MultiColorBandChannel::~MultiColorBandChannel()
{ }

// ######################################################################
void MultiColorBandChannel::exportOptions(const int whatToExport,
                                          const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE){
    manager->requestOption(this, itsNumBands);
    manager->requestOption(this, itsSatBands);
    manager->requestOption(this, itsHueSigma); 
    manager->requestOption(this, itsSatSigma); 
  }

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
ColorBandChannel& MultiColorBandChannel::band(const uint idx) const
{
  // Since we are dynamic_cast'ing a reference, this operation will either
  // succeed or throw an exception.
  return dynamic_cast<ColorBandChannel&>(ComplexChannel::subChan(idx));
}

// ######################################################################
void MultiColorBandChannel::doInput(const Image<PixRGB<float> > *colimg,
                                    const Image<float>*, const double t,
                                    const Image<byte>& clipMask)
{
  ASSERT(colimg != 0);
  /*
    in the old implementation, we only supported 3 bands: r, g, b. in
    the new implementation, we allow several bands, i.e., distributed
    coding of color.  ASSERT(numChans() == 3);
  */
  // create bands with different ranges of hues and saturation
  uint numHue = itsNumBands.getVal(), numSat = itsSatBands.getVal();
  uint numChans = numHue + numSat ;
  Image<float> input[numChans]; 
  Image<float>::iterator bptr[numChans];
  // initialize the pointers
  for (uint i = 0; i < numChans; i++){
    input[i].resize(colimg->getDims(), true);
    bptr[i] = input[i].beginw();
  }
  // some constants for the loop
  float hue_range = 360/numHue; // range of each band
  float c1 = 1 / (sqrt(6.28) * itsHueSigma.getVal());
  float c2 = 2.0f * itsHueSigma.getVal() * itsHueSigma.getVal();
  float c3 = 1 / (sqrt(6.28) * itsSatSigma.getVal());  
  float c4 = 2.0f * itsSatSigma.getVal() * itsSatSigma.getVal();
  // get the hue and saturation at all pixels
  Image<PixRGB<float> >::const_iterator aptr = colimg->begin();
  Image<PixRGB<float> >::const_iterator stop = colimg->end();
  while (aptr != stop)
    {
      float h, s, v;
      aptr->getHSV(h, s, v);
      aptr++;
      // what is the response of each band to this hue?
      for (uint i = 0; i < numHue; i++){
        if (v == 0.0f || s == 0.0f)
          *(bptr[i]) = 0.0f; // response to black and white is zero
        else {
          float mean = i * hue_range;
          float dist = h-mean;
          if (dist > 180) dist = 360 - dist;
          else if (dist < -180) dist = dist + 360;
          *(bptr[i]) = c1 * exp(-1.0f * dist * dist / c2);
        }
        bptr[i] =  bptr[i] + 1;
      }
      // what is the response of each band to this saturation?
      for (uint i = 0; i < numSat; i++){
        if (v == 0.0f)
          *(bptr[numHue+i]) = 0.0f; // response to black is zero
        else {
          float mean = (i+0.5)/numSat;
          float dist = s-mean;
          *(bptr[numHue+i]) = c3 * exp(-1.0f * dist * dist / c4);
        }
        bptr[numHue+i] =  bptr[numHue+i] + 1;
      }
    }
  for (uint i = 0; i < numChans; i++) 
    subChan(i).input(input[i], t, clipMask);
  
}

// ######################################################################
// DirectionChannel member definitions:
// ######################################################################

// ######################################################################
DirectionChannel::DirectionChannel(ModelManager& mgr, const uint dirIndex,
                                   const double direction,
                                   const PyramidType type) :
  SingleChannel(mgr, "", "", MOTION,
                makeShared(new ReichardtPyrBuilder<float>
                           (cos(direction * M_PI / 180.0),
                            -sin(direction * M_PI / 180.0),
                            type, direction + 90.0))),
  itsIndex(this, "DirectionChannelIndex", dirIndex),
  itsDirection(this, "DirectionChannelDirection", direction),
  itsTakeSqrt(this, "DirectionChannelTakeSqrt", false),
  itsThresh(this, "DirectionChannelLowThresh", 3.0)
{
  itsTakeAbs.setVal(true);
  itsNormalizeOutput.setVal(true);
  itsScaleNoiseToMax.setVal(true);
  itsLowThresh.setVal(itsThresh.getVal());  // kill small values in pyramid
  itsRectifyPyramid.setVal(true); // kill negative values in pyramid

  char name[256];
  snprintf(name, 256, "Direction(%d)", int(direction));
  setDescriptiveName(name);

  snprintf(name, 256, "dir_%d", dirIndex);
  setTagName(name);
}

// ######################################################################
void DirectionChannel::exportOptions(const int whatToExport,
                                     const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsTakeSqrt);
    manager->requestOption(this, itsThresh);
  }

  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void DirectionChannel::start2()
{
  itsLowThresh.setVal(itsThresh.getVal());  // kill small values in pyramid
}

// ######################################################################
DirectionChannel::~DirectionChannel()
{ }

// ######################################################################
void DirectionChannel::doInput(const Image<PixRGB<float> >*,
                               const Image<float>* bwimg, const double t,
                               const Image<byte>& clipMask)
{
  SingleChannel::doInput(0, bwimg, t, clipMask);
  ImageSet<float>& p = pyrMut(0);
  if (itsTakeSqrt.getVal()) doSqrt(p);
}

// ######################################################################
// MotionChannel member definitions:
// ######################################################################

// ######################################################################
MotionChannel::MotionChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "Motion", "motion", MOTION),
  itsPyrType(this, "MotionChannelPyramidType", Gaussian5),
  itsNumDirs(this, "NumDirections", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE)
{
  // let's create our subchannels (may be reconfigured later if our
  // number of directions changes):
  buildSubChans();
}

// ######################################################################
MotionChannel::~MotionChannel()
{  }

// ######################################################################
void MotionChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumDirs);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
DirectionChannel& MotionChannel::dirChan(const uint idx) const
{ return dynamic_cast<DirectionChannel&>(ComplexChannel::subChan(idx)); }

// ######################################################################
void MotionChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  // let's instantiate our subchannels now that we know how many we
  // want. They will inherit the current values (typically
  // post-command-line parsing) of all their options as they are
  // constructed:
  LINFO("Using %d directions spanning [0..360]deg", itsNumDirs.getVal());
  for (uint i = 0; i < itsNumDirs.getVal(); i ++)
    addSubChan(makeShared
               (new DirectionChannel(*manager, i, 360.0 * double(i) / 
                                     double(itsNumDirs.getVal()),
                                     itsPyrType.getVal())));
  // Here, we just use our subchannels' default tag names

  // if we have been requested to export some options, let's ask it
  // again on our newly built channels:
  if (itsExportOpts != OPTEXP_NONE)
    ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
bool MotionChannel::setModelParamString(const std::string& name,
                                                const std::string& value,
                                                const bool recurse,
                                                const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumDirs.getName()) == 0 &&
      numChans() != itsNumDirs.getVal())
    buildSubChans();

  return ret;
}

// ######################################################################
void MotionChannel::doInput(const Image< PixRGB<float> >*,
                            const Image<float>* bwimg, const double t,
                            const Image<byte>& clipMask)
{
  ASSERT(bwimg != 0);

  // compute Reichardt motion detection into several directions
  for (uint dir = 0; dir < numChans(); ++dir)
    {
      subChan(dir).input(*bwimg, t, clipMask);
      LINFO("Motion pyramid (%d/%d) ok.", dir+1, numChans());
    }
}

// ######################################################################
// TemplateMatchChannel member definitions:
// ######################################################################

// ######################################################################
TemplateMatchChannel::TemplateMatchChannel(ModelManager& mgr,
                                           const Image<float>& filt) :
  SingleChannel(mgr, "Template Match", "template", TEMPLATE,
                makeShared(new TemplateMatchPyrBuilder(filt))),
  itsBestMatchPos(-1, -1),
  itsBestMatchLev(-1),
  itsBestMatchScore(-1.0e30F),
  itsFiltDims(filt.getDims())
{  }

// ######################################################################
TemplateMatchChannel::~TemplateMatchChannel() {}

// ######################################################################
void TemplateMatchChannel::start1()
{
  SingleChannel::start1();
}

// ######################################################################
void TemplateMatchChannel::findBestMatch(const Rectangle& r,
                                         Point2D& matchpos,
                                         int& match_lev, float& score)
{
  itsBestMatchScore = -1.0e30F;
  for (uint i = itsLevelSpec.getVal().levMin();
       i <= itsLevelSpec.getVal().levMax(); ++i)
    {
      Image<float> zone = rescale(this->getImage(i), this->getInputDims());

      zone = crop(zone, r, true);   // search zone

      Point2D p; float val;
      zone.findMax(p, val);
      if (val > itsBestMatchScore)
        {
          itsBestMatchPos.i = p.i + r.left();
          itsBestMatchPos.j = p.j + r.top();
          itsBestMatchLev = i;
          itsBestMatchScore = val;
        }
    }
  LINFO("best match: (%d, %d) at scale %d [%f]", itsBestMatchPos.i,
        itsBestMatchPos.j, itsBestMatchLev, itsBestMatchScore);

  // return our results:
  matchpos = itsBestMatchPos; match_lev = itsBestMatchLev;
  score = itsBestMatchScore;
}

// ######################################################################
void TemplateMatchChannel::drawResults(Image< PixRGB<byte> >& traj)
{
  if (itsBestMatchLev < 0) return; // nothing to draw
  const int fw = itsFiltDims.w(), fh = itsFiltDims.h();

  int pfac = (1 << itsBestMatchLev) / 2;
  Rectangle pedrect(std::max(itsBestMatchPos.j - fh*pfac, 0),
                    std::max(itsBestMatchPos.i - fw*pfac, 0),
                    std::min(itsBestMatchPos.j + fh*pfac, traj.getHeight()-1),
                    std::min(itsBestMatchPos.i + fw*pfac, traj.getWidth()-1));

  // display bounding box around pedestrian:
  PixRGB<byte> redPix(255, 0, 0);
  traj.drawRect(pedrect, redPix, 1);
}

// ######################################################################
// PedestrianChannel member definitions:
// ######################################################################

namespace
{
  const int pedX = 9;
  const int pedY = 15;
  const float pedFILT[pedX*pedY] = {
    -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0,
    -3.0, -3.0, -3.0,  0.0,  1.0,  0.0, -3.0, -3.0, -3.0,
    -3.0, -3.0,  0.0,  2.0,  3.0,  2.0,  0.0, -3.0, -3.0,
    -3.0,  0.0,  0.0,  1.0,  3.0,  1.0,  0.0,  0.0, -3.0,
    -3.0,  1.0,  2.0,  3.0,  3.0,  3.0,  2.0,  1.0, -3.0,
    -3.0,  1.0,  2.0,  3.0,  3.0,  3.0,  2.0,  1.0, -3.0,
    -3.0,  1.0,  2.0,  3.0,  3.0,  3.0,  2.0,  1.0, -3.0,
    -3.0,  1.0,  2.0,  3.0,  3.0,  3.0,  2.0,  1.0, -3.0,
    -3.0,  1.0,  2.0,  3.0,  3.0,  3.0,  2.0,  1.0, -3.0,
    -3.0,  0.0,  2.0,  3.0,  2.0,  3.0,  2.0,  0.0, -3.0,
    -3.0,  0.0,  2.0,  3.0,  0.0,  3.0,  2.0,  0.0, -3.0,
    -3.0,  0.0,  2.0,  3.0,  0.0,  3.0,  2.0,  0.0, -3.0,
    -3.0,  0.0,  2.0,  3.0,  0.0,  3.0,  2.0,  0.0, -3.0,
    -3.0,  0.0,  2.0,  3.0,  0.0,  3.0,  2.0,  0.0, -3.0,
    -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0, -3.0
  };
}

PedestrianChannel::PedestrianChannel(ModelManager& mgr) :
  TemplateMatchChannel(mgr, Image<float>(pedFILT, pedX, pedY)),
  itsMinScale(this, "PedestrianChannelMinScale", 1),
  itsMaxScale(this, "PedestrianChannelMaxScale", 3)
{  }

// ######################################################################
PedestrianChannel::~PedestrianChannel()
{  }

// ######################################################################
void PedestrianChannel::start1()
{
  TemplateMatchChannel::start1();

  // refuse to have our LevelSpec changed by the command line, because
  // we are special:
  itsLevelSpec.setVal(LevelSpec(itsMinScale.getVal(), itsMaxScale.getVal(),
                                0, 1, 2));
  LINFO("Using scales [%d..%d]", itsMinScale.getVal(), itsMaxScale.getVal());
}

// ######################################################################
// junction Channel member definitions:
// ######################################################################
JunctionChannel::JunctionChannel(ModelManager& mgr, VisualCortex* vcx,
                                 const VisualFeature vs,
                                 const int r0, const int r1, const int r2, 
                                 const int r3, const int r4, const int r5, 
                                 const int r6, const int r7) :
  SingleChannel(mgr, "JunctionChannel", "JunctionChannel", vs, 
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsFull(this,
          vs == LJUNCTION ? "LFullImplementation" :
          vs == TJUNCTION ? "TFullImplementation" :
          vs == XJUNCTION ? "XFullImplementation" :
          vs == ENDPOINT ? "EFullImplementation" :
          "BugJunctionFullImplementation", false),  // see ModelOptionDefs.C
  itsVCX(vcx),
  R0(r0), R1(r1), R2(r2), R3(r3), R4(r4), R5(r5), R6(r6), R7(r7)
{
  itsTakeAbs.setVal(true);
  
  char name[256];
  snprintf(name, 256, "Junction(%d%d%d%d%d%d%d%d)",
           r0, r1, r2, r3, r4, r5, r6, r7);
  setDescriptiveName(name);

  snprintf(name, 256, "junction_%d%d%d%d%d%d%d%d",
           r0, r1, r2, r3, r4, r5, r6, r7);
  setTagName(name);

  setVisualFeature(vs);
}

// ######################################################################
JunctionChannel::~JunctionChannel()
{ }

// ######################################################################
void JunctionChannel::exportOptions(const int whatToExport,
                                    const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsFull);

  // the parent class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void JunctionChannel::doInput(const Image<PixRGB<float> >*,
                              const Image<float>* bwimg, const double t,
                              const Image<byte>& clipMask)
{
  // access the orientation channel
  /* since subChan method is declared only in ComplexChannel, we use
     dynamic cast to downcast from the base class ChannelBase to the
     derived class ComplexChannel */
  ASSERT(itsVCX->hasChannel("orientation"));
  ComplexChannel& ori = dynamic_cast<ComplexChannel&>
    (itsVCX->channel("orientation"));
  uint num = 0; ori.getModelParamVal("NumOrientations", num);
  ASSERT(num == 4);
  
  // access the gabor pyramids at the relevant orientations: 0, 45,
  // 90, 135
  const ImageSet<float>& ori0_pyr =  dynamic_cast<SingleChannel&> 
    (ori.subChan(0)).pyramid(0);
  const ImageSet<float>& ori45_pyr =  dynamic_cast<SingleChannel&> 
    (ori.subChan(1)).pyramid(0);
  const ImageSet<float>& ori90_pyr = dynamic_cast<SingleChannel&> 
    (ori.subChan(2)).pyramid(0);
  const ImageSet<float>& ori135_pyr =  dynamic_cast<SingleChannel&> 
    (ori.subChan(3)).pyramid(0);
  
  ImageSet<float> result(ori0_pyr.size());
  SingleChannel::killCaches();
  SingleChannel::setClipPyramid(clipMask);

  const bool R[8] = { R0, R1, R2, R3, R4, R5, R6, R7 };

  // combine the orientation pyramids to form the L pyramid
  for (uint i = 0; i < ori0_pyr.size(); i ++)
    if (itsFull.getVal())
      result[i] = junctionFilterFull(ori0_pyr[i], ori45_pyr[i],
                                     ori90_pyr[i], ori135_pyr[i],
                                     R, 6, 6);
    else
      result[i] = junctionFilterPartial(ori0_pyr[i], ori45_pyr[i],
                                        ori90_pyr[i], ori135_pyr[i],
                                        R, 6, 6);

   // store the result
   SingleChannel::storePyramid(result, t);
}

// ######################################################################
// L Junction Channel member definitions:
// ######################################################################
LJunctionChannel::LJunctionChannel(ModelManager& mgr,
                                   VisualCortex* vcx) :
  ComplexChannel(mgr, "LJunction", "l-junction", LJUNCTION),
  itsNumOrients(this, "NumLOrients", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE),  
  itsVCX(vcx)
{
  // let's build our channels
  buildSubChans();
}

// ######################################################################
LJunctionChannel::~LJunctionChannel() {}

// ######################################################################
void LJunctionChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  LINFO("Using %d orientations spanning [0..360]deg", itsNumOrients.getVal());
  for (uint ori = 0; ori < 8; )
    {
      switch (ori)
        {
        case 0:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,1,0,0,0,0,0 )));
          break;
        case 1:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,1,0,0,0,0 )));
          break;
        case 2:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,1,0,1,0,0,0 )));
          break;
        case 3:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,1,0,1,0,0 )));
          break;
        case 4:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,1,0,1,0 )));
          break;
        case 5:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,0,1,0,1 )));
          break;
        case 6:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,0,0,0,0,1,0 )));
          break;
        case 7:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,0,0,0,0,1 )));
          break;
        default:
          break;
        }
      ori += 8 / itsNumOrients.getVal();
    }
}

// ######################################################################
bool LJunctionChannel::setModelParamString(const std::string& name,
                                           const std::string& value,
                                           const bool recurse,
                                           const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);
  
  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumOrients.getName()) == 0 &&
      numChans() != itsNumOrients.getVal())
    buildSubChans();
  
  return ret;
}

// ######################################################################
void LJunctionChannel::exportOptions(const int whatToExport,
                                     const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumOrients);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void LJunctionChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  ASSERT(itsVCX->hasChannel("orientation"));
  for (uint i = 0; i < numChans(); ++i)
    {
      JunctionChannel& l = dynamic_cast<JunctionChannel&>
        (ComplexChannel::subChan(i));
      l.input(*bwimg, t, clipMask);
      LINFO("LJunction pyramid (%d/%d) ok.", i+1, numChans());
    }
}

// ######################################################################
// T Junction Channel member definitions:
// ######################################################################
TJunctionChannel::TJunctionChannel(ModelManager& mgr, VisualCortex* vcx) :
  ComplexChannel(mgr, "TJunction", "t-junction", TJUNCTION),
  itsNumOrients(this, "NumTOrients", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE),  
  itsVCX(vcx)
{
  // let's build our channels
  buildSubChans();
}

// ######################################################################
TJunctionChannel::~TJunctionChannel()
{ }

// ######################################################################
void TJunctionChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  LINFO("Using %d orientations spanning [0..360]deg", itsNumOrients.getVal());
  for (uint ori = 0; ori < 8; )
    {
      switch (ori)
        {
        case 0:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,0,0,1,0,1,0 )));
          break;
        case 1:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,0,0,1,0,1 )));
          break;
        case 2:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,1,0,0,0,1,0 )));
          break;
        case 3:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,1,0,0,0,1 )));
          break;
        case 4:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,1,0,1,0,0,0 )));
          break;
        case 5:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,1,0,1,0,0 )));
          break;
        case 6:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,1,0,1,0,1,0 )));
          break;
        case 7:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,1,0,1,0,1 )));
          break;
        default:
          break;
        }
      ori += 8 / itsNumOrients.getVal();
    }
}

// ######################################################################
bool TJunctionChannel::setModelParamString(const std::string& name,
                                           const std::string& value,
                                           const bool recurse,
                                           const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);
  
  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumOrients.getName()) == 0 &&
      numChans() != itsNumOrients.getVal())
    buildSubChans();
  
  return ret;
}

// ######################################################################
void TJunctionChannel::exportOptions(const int whatToExport,
                                     const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumOrients);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void TJunctionChannel::doInput(const Image<PixRGB<float> >*,
                               const Image<float>* bwimg, const double t,
                               const Image<byte>& clipMask)
{
  ASSERT(itsVCX->hasChannel("orientation"));
  for (uint i = 0; i < numChans(); ++i)
    {
      JunctionChannel& l = dynamic_cast<JunctionChannel&>
        (ComplexChannel::subChan(i));
      l.input(*bwimg, t, clipMask);
      LINFO("TJunction pyramid (%d/%d) ok.", i+1, numChans());
    }
}

// ######################################################################
// X Junction Channel member definitions:
// ######################################################################
XJunctionChannel::XJunctionChannel(ModelManager& mgr, VisualCortex* vcx) :
  ComplexChannel(mgr, "XJunction", "x-junction", XJUNCTION),
  itsNumOrients(this, "NumXOrients", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE),  
  itsVCX(vcx)
{
  // let's build our channels
  buildSubChans();
}

// ######################################################################
XJunctionChannel::~XJunctionChannel()
{ }

// ######################################################################
void XJunctionChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  LINFO("Using %d orientations spanning [0..90]deg", itsNumOrients.getVal());
  if (itsNumOrients.getVal() == 1)
    addSubChan(makeShared
               (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                    1,0,1,0,1,0,1,0)));
  else if (itsNumOrients.getVal() == 2)
    addSubChan(makeShared
               (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                    0,1,0,1,0,1,0,1)));
}

// ######################################################################
bool XJunctionChannel::setModelParamString(const std::string& name,
                                        const std::string& value,
                                        const bool recurse,
                                        const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);
  
  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumOrients.getName()) == 0 &&
      numChans() != itsNumOrients.getVal())
    buildSubChans();
  
  return ret;
}

// ######################################################################
void XJunctionChannel::exportOptions(const int whatToExport,
                                       const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumOrients);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void XJunctionChannel::doInput(const Image<PixRGB<float> >*,
                               const Image<float>* bwimg, const double t,
                               const Image<byte>& clipMask)
{
  ASSERT(itsVCX->hasChannel("orientation"));
  for (uint i = 0; i < numChans(); ++i)
    {
      JunctionChannel& l = dynamic_cast<JunctionChannel&>
        (ComplexChannel::subChan(i));
      l.input(*bwimg, t, clipMask);
      LINFO("XJunction pyramid (%d/%d) ok.", i+1, numChans());
    }
}

// ######################################################################
// End point Channel member definitions:
// ######################################################################
EndPointChannel::EndPointChannel(ModelManager& mgr, VisualCortex* vcx) :
  ComplexChannel(mgr, "EndPoint", "end-point", ENDPOINT),
  itsNumOrients(this, "NumEOrients", 4), // see ModelOptionDefs.C
  itsExportOpts(OPTEXP_NONE),  
  itsVCX(vcx)
{
  // let's build our channels
  buildSubChans();
}

// ######################################################################
EndPointChannel::~EndPointChannel()
{ }

// ######################################################################
void EndPointChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  LINFO("Using %d orientations spanning [0..360]deg", itsNumOrients.getVal());
  for (uint ori = 0; ori < 8; )
    {
      switch (ori)
        {
        case 0:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           1,0,0,0,0,0,0,0 )));
          break;
        case 1:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,1,0,0,0,0,0,0 )));
          break;
        case 2:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,1,0,0,0,0,0 )));
          break;
        case 3:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,1,0,0,0,0 )));
          break;
        case 4:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,1,0,0,0 )));
          break;
        case 5:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,0,1,0,0 )));
          break;
        case 6:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,0,0,1,0 )));
          break;
        case 7:
          addSubChan(makeShared
                     (new JunctionChannel(*manager, itsVCX, visualFeature(),
                                           0,0,0,0,0,0,0,1 )));
          break;
        default:
          break;
        }
      ori += 8 / itsNumOrients.getVal();
    }
}

// ######################################################################
bool EndPointChannel::setModelParamString(const std::string& name,
                                        const std::string& value,
                                        const bool recurse,
                                        const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);
  
  // if the param is our number of orientations and it has become
  // different from our number of channels, let's reconfigure:
  if (name.compare(itsNumOrients.getName()) == 0 &&
      numChans() != itsNumOrients.getVal())
    buildSubChans();
  
  return ret;
}

// ######################################################################
void EndPointChannel::exportOptions(const int whatToExport,
                                       const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumOrients);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
void EndPointChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  ASSERT(itsVCX->hasChannel("orientation"));
  for (uint i = 0; i < numChans(); ++i)
    {
      JunctionChannel& l = dynamic_cast<JunctionChannel&>
        (ComplexChannel::subChan(i));
      l.input(*bwimg, t, clipMask);
      LINFO("EndPoint pyramid (%d/%d) ok.", i+1, numChans());
    }
}

// ######################################################################
// StereoChannel member definitions:
// ######################################################################

// ######################################################################
StereoChannel::StereoChannel(ModelManager& mgr) :
  ComplexChannel(mgr, "Stereo", "stereo", STEREO),
  itsPyrType(this, "StereoChannelPyramidType", Gaussian3),
  itsNumTheta(this, "NumTheta", 1), // make sure the ModelOptionsDef.C 
                                    // also changed
  itsNumPhase(this, "NumPhase", nPhase/2 + 1),
  itsExportOpts(OPTEXP_NONE)
{
  fImgL = NULL; fImgR = NULL;

  // let's create our subchannels (may be reconfigured later if our
  // number of orientations changes):
  buildSubChans();
}

// ######################################################################
StereoChannel::~StereoChannel()
{  }

// ######################################################################
void StereoChannel::start1()
{

  // get the depth of the feature maps
  itsDepth = dispChan(0,0).getLevelSpec().maxDepth();

  // create the phased gabor pyramids
  createPhasedGaborPyramids();
}

// ######################################################################
void StereoChannel::exportOptions(const int whatToExport, 
                                  const bool recurse)
{
  
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsNumTheta);

  // keep track of what to export, in case we rebuild our subchans:
  itsExportOpts = whatToExport;

  // the parent class will propagate this call to our subcomponents:
  ComplexChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
DisparityChannel& StereoChannel::dispChan(const uint idO, 
                                          const uint idP) const
{ 
  uint idx= idO * itsNumPhase.getVal() + idP;  
  return dynamic_cast<DisparityChannel&>(ComplexChannel::subChan(idx)); 
}

// ######################################################################
void StereoChannel::buildSubChans()
{
  // kill any subchans we may have had...
  removeAllSubComponents();
  setHomogeneous(true);

  // let's instantiate our subchannels now that we know how many we
  // want. They will inherit the current values (typically
  // post-command-line parsing) of all their options as they are
  // constructed:
  LINFO("Using %d orientations spanning [0..180]deg and %d phase spanning [0..360]deg", 
    itsNumTheta.getVal(), itsNumPhase.getVal());
  
  double theta; 
  double phase; 

  for (uint i = 0; i < itsNumTheta.getVal(); i ++)
    for (uint j = 0; j < itsNumPhase.getVal(); j ++)
    {
      theta = 180.0 * double(i) / double(itsNumTheta.getVal());
      // the phase range is -PI/2 to PI/2 but -PI/2 to 0 is normalized. 
      // the number of thetas is always odd
      // first half is 0 to PI/2
      if(j <= itsNumPhase.getVal()/2)  
        phase = 90.0 * double(j)   / double(itsNumPhase.getVal()/2);
      // the second half is -PI/2 to 0
      else
        phase = 90.0 * double(j-1) / double(itsNumPhase.getVal()/2) + 
                180.0;
      
      addSubChan(makeShared
                 (new DisparityChannel(*manager, i*itsNumPhase.getVal()
                                       +j,theta,phase )));
 
      dispChan(i,j).setNumTheta(nTheta);
      dispChan(i,j).setNumPhase(nPhase);

    }

  // Here, we just use our subchannels' default tag names

  // if we have been requested to export some options, let's ask it
  // again on our newly built channels:
  if (itsExportOpts != OPTEXP_NONE)
    ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
void StereoChannel::createPhasedGaborPyramids()
{
  // create the phased gabor pyramids
  if(fImgL != NULL || fImgR != NULL)
  {
    for(uint i = 0; i < itsDepth; i++)
    {
      for(uint j = 0; j < nTheta; j++)
      {
        delete(fImgL[i][j]);
        delete(fImgR[i][j]);
      }
    } 

    for(uint i = 0; i < itsDepth; i++)
    {
      delete(fImgL[i]);
      delete(fImgR[i]);
    } 
   
    delete(fImgL);
    delete(fImgR);
  }

  fImgL = new Image<float>**[itsDepth];
  fImgR = new Image<float>**[itsDepth];

  for(uint i = 0; i < itsDepth; i++)
  {
    fImgL[i] = new Image<float>*[nTheta];
    fImgR[i] = new Image<float>*[nTheta];
  } 

  for(uint i = 0; i < itsDepth; i++)
  {
    for(uint j = 0; j < nTheta; j++)
    {
      fImgL[i][j] = new Image<float>[nPhase];
      fImgR[i][j] = new Image<float>[nPhase];
    }
  } 
}

// ######################################################################
bool StereoChannel::setModelParamString(const std::string& name,
                                        const std::string& value,
                                        const bool recurse,
                                        const bool warn)
{
  // let the ModelParam value change:
  bool ret = ModelComponent::setModelParamString(
               name, value, recurse, warn);

  // if the param is our  number of channel and it has 
  // become different from our number of orientations * number of phase , 
  // let's reconfigure:
  if (name.compare(itsNumTheta.getName()) == 0 &&
      //      name.compare(itsNumPhase.getName()) == 0 &&
      numChans() != itsNumPhase.getVal() * itsNumTheta.getVal())
    buildSubChans();

  return ret;
}

// ######################################################################
void StereoChannel::doInput(const Image< PixRGB<float> >*,
                            const Image<float>* bwimg, const double t,
                            const Image<byte>& clipMask)
{ 
  imgL = *bwimg;
  // IMPORTANT
  // at this point the secondary view image should already 
  // be set by Stereo Vision.

  // create an image pyramid for both images
  pyrL = buildPyrGaussian(imgL, itsDepth, 5);
  pyrR = buildPyrGaussian(imgR, itsDepth, 5); 
  
  // apply the gabor filter at various angle and phase on both eyes
  applyGabor();

  // for normalization
  ImageSet<float> tDispMap(itsDepth);
  ImageSet<float> tempDM;

  for(uint d = 0; d < itsDepth; d++)
  {
    tDispMap[d].resize(pyrL[d].getWidth(),pyrL[d].getHeight(),true);
    //ispMap[d] += 1.0F;
  }

  // pass the response array to each subchannel
  // so they can calculate each disparity correspondence
  for (uint i = 0; i < itsNumTheta.getVal(); i ++)
    for (uint j = 0; j < itsNumPhase.getVal(); j ++)
    {
      dispChan(i,j).setRawFilteredImages(fImgL,fImgR);
      dispChan(i,j).doInput(0, bwimg, t, clipMask);

      //dispChan(i,j).getDispMap(&tempDM);        

      // add the results for normalizing the output
      //for (uint d = 0; d< itsDepth; d++)
      //  tDispMap[d] += tempDM[d];

    }

  // normalize the values to [0,  1] - confidence level
  // and store the pyramid for the interaction
  /*  for (uint i = 0; i < itsNumTheta.getVal(); i ++)
    for (uint j = 0; j < itsNumPhase.getVal(); j ++)
    {
      dispChan(i,j).normalizeDispMap(tDispMap, 
        itsNumTheta.getVal()*itsNumPhase.getVal());
      dispChan(i,j).storePyramid(t);        
    } 
  */


}

// ######################################################################
void StereoChannel::setSecondImage(const Image<float>* bwimg)
{ 
  imgR = *bwimg;
}

// ######################################################################
void StereoChannel::getRawFilteredImages(Image<float> ****fImgLE,
                                         Image<float> ****fImgRI)
{
  *fImgLE = fImgL; 
  *fImgRI = fImgR;
}

// ######################################################################
void StereoChannel::applyGabor()
{ 
  Image<float> gaborF;Image<float> gF;
  Point2D a(3,3); Dims b(5,5);
  for(uint d = 0; d < itsDepth; d++)
  {
    for(uint t = 0; t < nTheta; t++)
    {   
      for(uint p = 0; p < nPhase; p++)
      { 
        gaborF = gaborFilter<float>(stddev,period,p*dPhase,t*dTheta);
        ASSERT(gaborF.getWidth() == 11 && gaborF.getHeight() == 11);
        gaborF = crop(gaborF,a,b,false);
       
        fImgL[d][t][p] = convolve(pyrL[d],gaborF);  
        //fImgL[d][t][p].rectify();  
        fImgR[d][t][p] = convolve(pyrR[d],gaborF);
        //fImgR[d][t][p].rectify(); 
      }
    }
  }
}

// ######################################################################
// Entropy channel member definitions
// ######################################################################
EntropyChannel::EntropyChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Entropy", "entropy", ENTROPY,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsQstep(this, "EntropyChannelQstep", 4),
  itsMap()
{ }

// ######################################################################
EntropyChannel::~EntropyChannel()
{ }

// ######################################################################
bool EntropyChannel::outputAvailable() const
{ return itsMap.initialized(); }

// ######################################################################
void EntropyChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(bwimg);

  uint lev = ls.mapLevel();
  
  // figure out the tile and map sizes:
  int siz = 1 << lev; int w = bwimg->getWidth(), h = bwimg->getHeight();
  itsMap.resize(w >> lev, h >> lev);
  float siz2 = float(siz * siz);

  // prepare a histogram for our quantized patch values:
  int nsteps = 256 / itsQstep.getVal();
  int histo[nsteps];
  Image<byte> quant = (*bwimg) / nsteps;  // quantize
  quant = highThresh(quant, byte(nsteps - 1), byte(nsteps - 1)); // saturate

  // let's loop over the tiles and compute entropy for each:
  Image<float>::iterator dest = itsMap.beginw();

  for (int j = 0; j <= h-siz; j += siz)
    {
      int jmax = std::min(j + siz, h);
      for (int i = 0; i <= w-siz; i += siz)
        {
          int imax = std::min(i + siz, w);
          Image<byte>::const_iterator src = quant.begin() + i + j * w;
          memset(histo, 0, nsteps * sizeof(int));

          // accumulate the histogram of values:
          for (int jj = j; jj < jmax; jj ++)
            {
              for (int ii = i; ii < imax; ii ++)
                histo[*src++] ++;

              // skip to next input row:
              src += w - imax + i;
           }
          
          // now compute the entropy:
          float e = 0.0f;
          for (int k = 0; k < nsteps; k ++)
            if (histo[k])
              {
                float freq = float(histo[k]) / siz2;
                e -= freq * logf(freq);
              }
          // store entropy in output:
          *dest++ = e;
        }
    }
}

// ######################################################################
Image<float> EntropyChannel::getOutput()
{ return itsMap; }


// ######################################################################
// Variance channel member definitions
// ######################################################################
VarianceChannel::VarianceChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Variance", "variance", VARIANCE,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsMap()
{ }

// ######################################################################
VarianceChannel::~VarianceChannel()
{ }

// ######################################################################
bool VarianceChannel::outputAvailable() const
{ return itsMap.initialized(); }

// ######################################################################
void VarianceChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(bwimg);

  uint lev = ls.mapLevel();
  
  // figure out the tile and map sizes:
  int siz = 1 << lev; int w = bwimg->getWidth(), h = bwimg->getHeight();
  itsMap.resize(w >> lev, h >> lev);

  // let's loop over the tiles and compute variance for each: We use
  // the well-known identity E[(X-E[X])^2] = E[X^2] - E[X]^2 and we
  // also use the small-sample approximation to the variance (divide
  // by n-1 instead of n):
  Image<float>::iterator dest = itsMap.beginw();
  for (int j = 0; j <= h-siz; j += siz)
    {
      int jmax = std::min(j + siz, h);
      for (int i = 0; i <= w-siz; i += siz)
        {
          int imax = std::min(i + siz, w);
          double sum = 0.0, sumsq = 0.0;
          Image<float>::const_iterator src = bwimg->begin() + i + j * w;

          // compute sum and sum of squares:
          for (int jj = j; jj < jmax; jj ++)
            {
              for (int ii = i; ii < imax; ii ++)
                {
                  double val = *src++;
                  sum += val; sumsq += val * val;
                }
              // skip to next input row:
              src += w - imax + i;
           }
          
          // now compute the variance:
          double npix = double((jmax - j) * (imax - i));
          double mean = sum / npix;
          double var = (sumsq - mean * mean * npix) / (npix - 1.0);
          *dest++ = sqrtf(float(var)) / 255.0F;
        }
    }
}

// ######################################################################
Image<float> VarianceChannel::getOutput()
{ return itsMap; }


// ######################################################################
// Tcorr channel member definitions
// ######################################################################
TcorrChannel::TcorrChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Tcorr", "tcorr", TCORR,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsFrameLag(this, "TcorrChannelFrameLag", 1), // see ModelOptionDefs.C
  itsCache(), itsMap()
{ }

// ######################################################################
void TcorrChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsFrameLag);

  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
TcorrChannel::~TcorrChannel()
{ }

// ######################################################################
bool TcorrChannel::outputAvailable() const
{ return (itsCache.size() == itsCache.getMaxSize()); }

// ######################################################################
void TcorrChannel::doInput(const Image<PixRGB<float> >*,
                           const Image<float>* bwimg, const double t,
                           const Image<byte>& clipMask)
{
  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(bwimg);

  // configure our cache if not done yet. We use one more frame in the
  // cache than the given lag, to hold our current frame:
  if (itsCache.getMaxSize() == 0)
    itsCache.setMaxSize(itsFrameLag.getVal() + 1);

  // just push our incoming input. All computations are done in getOutput():
  itsCache.push_back(*bwimg);
}

// ######################################################################
Image<float> TcorrChannel::getOutput()
{
  // if cache not full yet, we should not be called since
  // outputAvailable() returns false:
  if (itsCache.size() < itsCache.getMaxSize())
    {
      LERROR("I don't have an output yet -- RETURNING EMPTY");
      return itsMap; // should be uninitialized
    }

  // figure out the tile and map sizes:
  Image<float> cur = itsCache.front(), old = itsCache.back();

  LevelSpec ls = itsLevelSpec.getVal();
  uint lev = ls.mapLevel();
  int siz = 1 << lev; int w = cur.getWidth(), h = cur.getHeight();
  itsMap.resize(w >> lev, h >> lev);

  // let's loop over the tiles and compute correlations:
  Image<float>::iterator dest = itsMap.beginw();
  for (int j = 0; j <= h-siz; j += siz)
    {
      const int ph = std::min(j + siz, h) - j;
      for (int i = 0; i <= w-siz; i += siz)
        {
          const int pw = std::min(i + siz, w) - i;

          // use corrpatch function of Image_MathOps, which returns
          // values in [-1 .. 1]. For this channel, we want high
          // 'salience' for patches that are highly decorrelated from
          // frame to frame. So in our map we store 1-correlation,
          // i.e., some measure of decorrelation:
          const Point2D topleft(i, j); const Dims patchdims(pw, ph);

          *dest++ = 1.0 - corrpatch(cur, topleft, patchdims, old, topleft);
        }
    }

  return itsMap;
}


// ######################################################################
// Scorr channel member definitions
// ######################################################################
ScorrChannel::ScorrChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Scorr", "scorr", SCORR,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsRadius(this, "ScorrChannelRadius", 1), // see ModelOptionDefs.C
  itsMap()
{ }

// ######################################################################
void ScorrChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsRadius);

  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
ScorrChannel::~ScorrChannel()
{ }

// ######################################################################
bool ScorrChannel::outputAvailable() const
{ return itsMap.initialized(); }

// accumulate correlation from a point pp:
#define ACCUMCORR if (itsMap.coordsOk(pp) && itsMask.getVal(pp) == 0) \
{ ++n; itsMask.setVal(pp, 255); corr += 1.0 - \
corrpatch(*bwimg, topleft1, patchdims, *bwimg, \
Point2D(pp.i << lev, pp.j << lev)); }

// ######################################################################
void ScorrChannel::doInput(const Image<PixRGB<float> >*,
                           const Image<float>* bwimg, const double t,
                           const Image<byte>& clipMask)
{
  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(bwimg); ASSERT(itsRadius.getVal() >= 1);

  uint lev = ls.mapLevel();
  int siz = 1 << lev;
  int mw = bwimg->getWidth() >> lev, mh = bwimg->getHeight() >> lev;
  itsMap.resize(mw, mh); Image<byte> itsMask(mw, mh, NO_INIT);
  const int radius = itsRadius.getVal(), r2 = radius * radius;

  // FIXME: this channel will not work with images whose dims are not
  // multiple of patch dims...
  Dims patchdims(siz, siz);

  // loop over the destination and compute spatial correlations. Code
  // here is similar to that in Image::drawDisk():
  Image<float>::iterator dest = itsMap.beginw();
  for (int j = 0; j < mh; j ++)
    for (int i = 0; i < mw; i ++)
      {
        double corr = 0.0, n = 0; itsMask.clear();

        // our center patch's top-left corner:
        const Point2D topleft1(i << lev, j << lev);

        // go over a circle and get cross-correlations. First do the
        // two vertical extremes so that we won't count them twice:
        Point2D pp(i, j);
        pp.j -= radius; ACCUMCORR;
        pp.j += radius + radius; ACCUMCORR;

        // now we draw one quarter of the circle and symmetrize it.
        // NOTE: in this algo like in Image::drawCircle() there is
        // some repetition (points that get drawn twice). Have a look at
        // http://www.cs.unc.edu/~mcmillan/comp136/Lecture7/circle.html
        // for possibly better algos. Here we don't want to count
        // those correlations several times, so we just use a mask to
        // keep track of those we already have counted:
        int bound1 = radius, bound2;
        for (int y = 1; y <= radius; y ++)
          {
            bound2 = bound1;
            bound1 = int(sqrtf(float( r2 - SQR(y) )));
            for (int x = bound1; x <= bound2; x ++)
              {
                pp.j = j - y;
                pp.i = i - x; ACCUMCORR;
                pp.i = i + x; ACCUMCORR;
                pp.j = j + y; ACCUMCORR;
                pp.i = i - x; ACCUMCORR;
              }
          }

        if (n > 0) *dest++ = corr / double(n); else *dest++ = 0.0;
      }
}

// ######################################################################
Image<float> ScorrChannel::getOutput()
{ return itsMap; }


// ######################################################################
// Information channel member definitions
// ######################################################################
InformationChannel::InformationChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Information", "information", INFORMATION,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsEps(this, "InformationChannelEps", 0.01f),
  itsMap()
{ }

// ######################################################################
InformationChannel::~InformationChannel()
{ }

// ######################################################################
bool InformationChannel::outputAvailable() const
{ return itsMap.initialized(); }

// ######################################################################
void InformationChannel::doInput(const Image<PixRGB<float> >* colimg,
                                 const Image<float>* , const double t,
                                 const Image<byte>& clipMask)
{
  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(colimg);

  uint lev = ls.mapLevel();
  
  // figure out the tile size and compute map:
  int siz = 1 << lev;
  itsMap = infoMeasure(*colimg, itsEps.getVal(), siz); // Image_ColorOps.C
}

// ######################################################################
Image<float> InformationChannel::getOutput()
{ return itsMap; }

// ######################################################################
// Dummy channel member definitions
// ######################################################################
DummyChannel::DummyChannel(ModelManager& mgr) :
  SingleChannel(mgr, "Dummy", "dummy", DUMMY,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsFactor(this, "DummyChannelFactor", 0.0f),
  itsMap()
{ }

// ######################################################################
DummyChannel::~DummyChannel()
{ }

// ######################################################################
void DummyChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsFactor);

  // the base class will propagate this call to our subcomponents:
  SingleChannel::exportOptions(whatToExport, recurse);
}

// ######################################################################
bool DummyChannel::outputAvailable() const
{ return itsMap.initialized(); }

// ######################################################################
const Dims& DummyChannel::getMapDims() const
{ return itsMap.getDims(); }

// ######################################################################
void DummyChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
  if (itsFactor.getVal()) itsMap = (*bwimg) * itsFactor.getVal();
  else itsMap = *bwimg;
}

// ######################################################################
Image<float> DummyChannel::centerSurround(const uint cntrlev,
                                          const uint surrlev) const
{ 
  ASSERT(itsLevelSpec.getVal().csOK(cntrlev, surrlev));
  return itsMap;
}


// ######################################################################
// PN03contrast channel member definitions
// ######################################################################
PN03contrastChannel::PN03contrastChannel(ModelManager& mgr) :
  SingleChannel(mgr, "PN03contrast", "pn03contrast", PN03CONTRAST,
                SharedPtr< PyrBuilder<float> >(NULL)),
  itsMap()
{ }

// ######################################################################
PN03contrastChannel::~PN03contrastChannel()
{ }

// ######################################################################
bool PN03contrastChannel::outputAvailable() const
{ return itsMap.initialized(); }

// ######################################################################
void PN03contrastChannel::doInput(const Image<PixRGB<float> >*,
                             const Image<float>* bwimg, const double t,
                             const Image<byte>& clipMask)
{
//    LERROR("\n START doInput of PN03contrast \n");

  LevelSpec ls = itsLevelSpec.getVal();
  ASSERT(ls.levMin() == ls.levMax());
  ASSERT(ls.delMin() == 0 && ls.delMax() == 0);
  ASSERT(ls.levMin() == ls.mapLevel());
  ASSERT(bwimg);

  uint lev = ls.mapLevel();
  
  // figure out the tile and map sizes:
  int siz = 1 << lev; int w = bwimg->getWidth(), h = bwimg->getHeight();
  itsMap.resize(w >> lev, h >> lev);
  float siz2 = float(siz * siz);

  // let's loop over the tiles and compute contrast for each:
  Image<float>::iterator dest = itsMap.beginw();
  for (int j = 0; j < h; j += siz)
  {
      int jmax = std::min(j + siz, h);
      for (int i = 0; i < w; i += siz)
      {          
           int imax = std::min(i + siz, w);
           float patch_sum=0;
           double patch_sumsq=0;
           for (int jj = j; jj < jmax; jj ++)
           {
               for (int ii = i; ii < imax; ii ++) {
                   float pix_val = bwimg->getVal(ii,jj);
                   patch_sum+=pix_val;
                   patch_sumsq+=pix_val*pix_val;
               }
           }
//           patch_sum = std::accumulate(bwimg->begin() + i+j*w, bwimg->begin() + imax+jmax*w, 0.0); // includes out-of-patch pixels!
           double patch_mean =  patch_sum / double(siz2);
           double patch_var = patch_sumsq / double(siz2) - patch_mean*patch_mean;
           
               // calculate the sd of pixel intensities (for long_patch_var)
//            double patch_diffsum=0;
//            if (i==0 && j==0) LERROR("patch mean: %.2f\n", patch_mean);
//            if (i==0 && j==0) LERROR( "[ " );
//            for (int jj = j; jj < jmax; jj ++)
//            {
//                for (int ii = i; ii < imax; ii ++) {
//                    double pix_val = bwimg->getVal(ii,jj);
//                    if (i==0 && j==0) std::cerr << pix_val << " ";
//                    double pix_diff = pix_val-patch_mean;
//                    double pix_var = double(pix_diff*pix_diff);
//                    patch_diffsum+=pix_var;
//                }
//                if (i==0 && j==0) std::cerr << "\n  ";
//            }
//            if (i==0 && j==0) LERROR(" ]\n");

          // now compute the contrast:

//           if (i==0 && j==0) LERROR("short patch var %.2f\n", patch_var);
//            double long_patch_var=patch_diffsum/double(siz2);
//            if (i==0 && j==0) LERROR("long patch var %.2f\n", long_patch_var);

           double patch_cntrst = sqrt(patch_var);
           if (i==0 && j==0) LERROR( "\nfirst %ix%i patch contrast: %.2f\n\n", siz, siz, patch_cntrst);
          *dest++ = patch_cntrst;
        }
    }
//  LERROR("\n END doInput of PN03contrast \n");
}

// ######################################################################
Image<float> PN03contrastChannel::getOutput()
{ return itsMap; }


// ######################################################################
HueChannel::HueChannel(ModelManager& mgr, 
                       float muR, float muG, 
                       float sigR, float sigG, float rho) :
  SingleChannel(mgr, "hue", "hue", HUE,
                makeShared(new GaussianPyrBuilder<float>(5))),
  itsMuR(muR),
  itsMuG(muG),
  itsSigR(sigR),
  itsSigG(sigG),
  itsRho(rho)
{}

HueChannel::HueChannel(ModelManager& mgr,
                       const PixRGB<byte> hue,
                       float sig) :
  SingleChannel(mgr, "hue", "hue", HUE,
                makeShared(new GaussianPyrBuilder<float>(5))),
  itsSigR(sig),
  itsSigG(sig),
  itsRho(0.0f)
{
  float intens,cr,cg;
  RGBtoCIE(hue,cr,cg,intens);
  itsMuR = cr;
  itsMuG = cg;
}

HueChannel::~HueChannel() {}


// ######################################################################
void HueChannel::doInput(const Image<PixRGB<float> > *colimg,
                         const Image<float>*, const double t,
                         const Image<byte>& clipMask)
{
  ASSERT(colimg != 0);

  Image<float> tmp = hueDistance(*colimg,itsMuR,itsMuG,
                                 itsSigR,itsSigG,itsRho);
  SingleChannel::doInput(0,&tmp,t,clipMask);
}

// ######################################################################
Image<float> HueChannel::combineSubMaps()
{
  Image<float> output = getImage(itsLevelSpec.getVal().mapLevel());

  LDEBUG("%s: Normalizing output: %s(%f .. %f)", tagName().c_str(),
         maxNormTypeName(itsNormType.getVal()), itsOutputRangeMin.getVal(),
         itsOutputRangeMax.getVal());

  output = maxNormalize(output, itsOutputRangeMin.getVal(),
                        itsOutputRangeMax.getVal(), itsNormType.getVal());

  return output;
}

// ######################################################################
uint HueChannel::numSubmaps() const
{ return 1; }

// ######################################################################
Image<float> HueChannel::getSubmap(const uint idx)
{
  return getOutput();
}

// ######################################################################
const char* HueChannel::getSubmapName(const uint idx) const
{
  return "Hue";
}

// ######################################################################
SkinHueChannel::SkinHueChannel(ModelManager& mgr) :
  HueChannel(mgr,COL_SKIN_MUR,COL_SKIN_MUG,
             COL_SKIN_SIGR,COL_SKIN_SIGG,COL_SKIN_RHO)
{}
/*
  itsMean[0] = 0.434904;
  itsMean[1] = 0.301983;
  itsInvCov[0] = 533.839800;
  itsInvCov[1] = 684.827315;
  itsInvCov[2] = 2565.283958;
*/

SkinHueChannel::~SkinHueChannel() {}


/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

