/*!@file Channels.H Feature channels for use in visual cortex classes */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Channels.H 4228 2005-02-23 17:19:25Z vidhya $
//

#ifndef CHANNELS_H_DEFINED
#define CHANNELS_H_DEFINED

#include "ChannelBase.H"
#include "ComplexChannel.H"
#include "SingleChannel.H"
#include "Image.H"
#include "PyrBuilder.H"
#include "ImageSet_Ops.H"
#include "OrientComputeTypes.H"
#include "VisualCortex.H"
#include "VisualObject.H"
#include "ColorComputeTypes.H"
#include "ImageCache.H"

// ######################################################################
// ######################################################################
// #################### SINGLE CHANNELS
// ######################################################################
// ######################################################################

// ######################################################################
//! An intensity channel.
class IntensityChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  IntensityChannel(ModelManager& mgr);
 
  //! Destructor
  virtual ~IntensityChannel();
};

// ######################################################################
//! A generic convolution channel.
class ConvolveChannel : public SingleChannel
{
public:
  //! Construct with standard params. Will need to set a filter before using
  ConvolveChannel(ModelManager& mgr);

  //! Set the filter to use for convolutions
  void setFilter(const Image<float>& filter);
  
  //! Destructor
  virtual ~ConvolveChannel();
};

// ######################################################################
//! A generic convolution channel using color filters.
class RGBConvolveChannel : public SingleChannel
{
public:
  //! Construct with standard params. Will need to set the 3 filters.
  RGBConvolveChannel(ModelManager& mgr);
  
  //! Set the 3 filters to use for convolutions
  void setFilters(const Image<float>& rfilter,
                  const Image<float>& gfilter,
                  const Image<float>& bfilter);

  //! Destructor
  virtual ~RGBConvolveChannel();

protected:
  //! SingleChannel requires only luminance input
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! compute from a color image
  virtual ImageSet<float> computePyramid(const Image< PixRGB<float> >& img);
};

// ######################################################################
//! A red-green double-opponent color channel.
class RedGreenChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  RedGreenChannel(ModelManager& mgr);

  //! Destructor
  virtual ~RedGreenChannel();
};

// ######################################################################
//! A blue-yellow double-opponent color channel.
class BlueYellowChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  BlueYellowChannel(ModelManager& mgr);

  //! Destructor
  virtual ~BlueYellowChannel();
};

// ######################################################################
//! Single Opponent RedGreenChannel member definitions:
class SORedGreenChannel: public SingleChannel
{
public:
  //! Construct with standard params.
  SORedGreenChannel(ModelManager& mgr);
 
  //! Destructor
  virtual ~SORedGreenChannel();
};

// ######################################################################
//! Single Opponent GreenRedChannel member definitions:
class SOGreenRedChannel: public SingleChannel
{
public:
  //! Construct with standard params.
  SOGreenRedChannel(ModelManager& mgr);
 
  //! Destructor
  virtual ~SOGreenRedChannel();
};

// ######################################################################
//! Single Opponent BlueYellowChannel member definitions:
class SOBlueYellowChannel: public SingleChannel
{
public:
  //! Construct with standard params.
  SOBlueYellowChannel(ModelManager& mgr);
 
  //! Destructor
  virtual ~SOBlueYellowChannel();
};

// ######################################################################
//! Single Opponent YellowBlueChannel member definitions:
class SOYellowBlueChannel: public SingleChannel
{
public:
  //! Construct with standard params.
  SOYellowBlueChannel(ModelManager& mgr);
 
  //! Destructor
  virtual ~SOYellowBlueChannel();
};
// ######################################################################
//! A red color channel.
class RedChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  RedChannel(ModelManager& mgr);

  //! Destructor
  virtual ~RedChannel();
};
// ######################################################################
//! A green color channel.
class GreenChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  GreenChannel(ModelManager& mgr);

  //! Destructor
  virtual ~GreenChannel();
};

// ######################################################################
//! A blue color channel.
class BlueChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  BlueChannel(ModelManager& mgr);

  //! Destructor
  virtual ~BlueChannel();
};

// ######################################################################
//! A yellow double-opponent color channel.
class YellowChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  YellowChannel(ModelManager& mgr);

  //! Destructor
  virtual ~YellowChannel();
};

// ######################################################################
//! An orientation-tuned gabor channel.
class GaborChannel : public SingleChannel
{
public:
  //! Construct with standard params plus an orientation index for this object.
  /*! @param oriIndex is used for various indexing such as with ParamMap
    and Jet interfacing.
    @param ang is the preferred orientation for this channel, in
      degrees (0..180). */
  GaborChannel(ModelManager& mgr, const uint oriIndex, const double ang);

  //! Destructor
  virtual ~GaborChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  virtual void readFrom(const ParamMap& pmap);

  virtual void writeTo(ParamMap& pmap) const;

  //! Get the preferred orientation (in degrees) of this channel.
  double angle() const;

protected:
  ModelParam<uint> itsOriIndex;
  ModelParam<double> itsOrientation;
  ModelParam<double> itsGaborIntens;

  //! determines the type of computation used to compute orientation
  /*! see OrientCOmputeTypes.H for possible values. */
  ModelParam<OrientComputeType> itsOriCompType;

  virtual void start1();
  void resetPyramid();
};

// ######################################################################
//! A temporal flicker channel.
/*! This channel just keeps a memory trace of the previous frame and
  then does a basic center-surround processing on the difference
  between current and memorized input. The time constant determines
  how the memorized input should be updated when a new input
  arrives. */
class FlickerChannel : public SingleChannel
{
public:
  //! Construct with standard specs.
  FlickerChannel(ModelManager& mgr);

  //! Destructor
  virtual ~FlickerChannel();

  //! This operation does nothing.
  /*! That's because FlickerChannel cannot learn() from a single input. */
  virtual void learn(const Image<byte>& dmap, double eta);

  //! get the output
  virtual Image<float> getOutput();

  //! Reset FlickerChannel
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

protected:
  //! FlickerChannel requires only luminance input.
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  /*! Time constant of the channel, in seconds. This
    is used to determine how to update our memory of the previous
    frame as a new frame arrives. If the previous frame is more than a
    time constant older than the new one, it will be replaced by the
    new one for the next step of processing; otherwise a mix between
    both will be used, depending on the difference between old and new
    times. */
  ModelParam<double> itsTimeConst;

  Image<float> itsPrevInput;
  double       itsPrevTime;
  Image<float> itsOutput;
};

// ######################################################################
//! A color band channel
class ColorBandChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  ColorBandChannel(ModelManager& mgr, const uint bandidx);

  //! Destructor
  virtual ~ColorBandChannel();

protected:
  ModelParam<uint> itsBandIndex;
};

// ######################################################################
//! a motion sensitive channel with direction selectivity
class DirectionChannel : public SingleChannel
{
public:
  //! Constructor with some parameteres
  /*! @param dirIndex is the channel's index, used in Param and Jet access
    @param direction is the preferred direction in degrees (0..360)
    @param type the type of the pyramid used for computing
    the Reichardt motion. */
  DirectionChannel(ModelManager& mgr, const uint dirIndex,
                   const double direction, const PyramidType type);

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~DirectionChannel();

protected:
  //! ReichardtChannel requires only luminance input
  /*! After sending the input image to the ReichardtPyramid,
    rectify and threshold the Reichardt maps. */
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  void start2(); //!< get started

  ModelParam<uint> itsIndex;
  ModelParam<double> itsDirection;
  ModelParam<bool> itsTakeSqrt;
  ModelParam<float> itsThresh;
};

// ######################################################################
//! Convolves each level of its pyramid with a feature template.
/*! This type of channel is useful for searching for a specific feature
    shape at multiple scales. For example, it is used by VisualCortex to
    search for possible pedestrians in an input image. */
class TemplateMatchChannel : public SingleChannel
{
public:
  //! Construct with the standard params, plus a filter specification.
  /*! The filter is specified by a C array, which will be interpreted to
      have dimensions \a filt_x and \a filt_y. A private copy of the filter
      will be made, so the \a filt pointer does not have to be
      persistent. */
  TemplateMatchChannel(ModelManager& mgr, const Image<float>& filt);

  //! Virtual destructor.
  virtual ~TemplateMatchChannel();

  //! Look for the most likely template match in a given rectangular region.
  /*! The position of the match in 0-level coordinates is returned in
      \a matchpos, the pyramid level at which the match was found is
      returned in \a match_lev, and a score rating the goodness of the
      match is returned in \a score. */
  void findBestMatch(const Rectangle& r, Point2D& matchpos,
                     int& match_lev, float& score);

  //! Draws a box in \a traj around the last best template match found
  virtual void drawResults(Image< PixRGB<byte> >& traj);

protected:
  virtual void start1();

  Point2D itsBestMatchPos;
  int itsBestMatchLev;
  float itsBestMatchScore;
  Dims itsFiltDims;         // only used by drawResults()
};

// ######################################################################
//! A template-matching channel for detecting pedestrians.
class PedestrianChannel : public TemplateMatchChannel
{
public:
  //! Constructor
  PedestrianChannel(ModelManager& mgr);

  //! Virtual destructor.
  virtual ~PedestrianChannel();

protected:
  ModelParam<int> itsMinScale;  //!< smallest scale to use in template matching
  ModelParam<int> itsMaxScale;  //!< largest scale to use in template matching
  virtual void start1();
};

//
//######################################################################
//! A junction channel.
class JunctionChannel : public SingleChannel
{
public:
  /*! Construct with the manager, visual cortex, visual feature and
    relevance of the different oriented features (0 => 0 deg., 1 => 45
    deg, ... , 7 => 315 deg., and whether we want a full or partial
    implementation of the junction */
  JunctionChannel(ModelManager& mgr, VisualCortex* vcx,
                  const VisualFeature vs = UNKNOWN,
                  const int R0 = 0,  const int R1 = 0,  const int R2 = 0,
                  const int R3 = 0,  const int R4 = 0,  const int R5 = 0, 
                  const int R6 = 0,  const int R7 = 0 );

  //! Destructor
  virtual ~JunctionChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

protected:
  //! Computes the corners from the gabor channels accessed via the
  //visual cortex
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);
  /*! do we want a slow, full implementation of this junction which
  checks for both the presence of relevant features and the absence of
  irrelevant features; or a quick, partial implementation which checks
  only for the presence of the relevant features. */
  ModelParam<bool> itsFull; 

  //! keep track of our VisualCortex
  VisualCortex *itsVCX;

private:
  int R0; // relevance of 0 deg. feature
  int R1; // relevance of 45 deg. feature
  int R2; // relevance of 90 deg. feature
  int R3; // relevance of 135 deg. feature
  int R4; // relevance of 180 deg. feature
  int R5; // relevance of 225 deg. feature
  int R6; // relevance of 270 deg. feature
  int R7;  // relevance of 315 deg. feature
};

// ######################################################################
//! An entropy channel.
/*! CAUTION: this channel will throw a fatal exception if your
  LevelSpec contains more than one level or if that level is different
  from the maplevel or if the deltas are non-zero. */
class EntropyChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  EntropyChannel(ModelManager& mgr);

  //! Destructor
  virtual ~EntropyChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  ModelParam<int> itsQstep; //!< greyscale quantization step

  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  Image<float> itsMap;  //!< our entropy map
};

// ######################################################################
//! A variance channel, computes the variance in an image patch
/*! CAUTION: this channel will throw a fatal exception if your
  LevelSpec contains more than one level or if that level is different
  from the maplevel or if the deltas are non-zero. */
class VarianceChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  VarianceChannel(ModelManager& mgr);

  //! Destructor
  virtual ~VarianceChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  Image<float> itsMap;  //!< our variance map
};

// ######################################################################
//! A temporal correlation channel
/*! Note that in this channel, high 'salience' means low temporal
  correlation, i.e., lots of change. CAUTION: this channel will throw
  a fatal exception if your LevelSpec contains more than one level or
  if that level is different from the maplevel or if the deltas are
  non-zero. */
class TcorrChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  TcorrChannel(ModelManager& mgr);

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~TcorrChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  ModelParam<int> itsFrameLag; //!< Lag (in frames) to use for correlation

  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  ImageCache<float> itsCache; //!< cache of the recent inputs
  Image<float> itsMap;        //!< our temporal correlation map
};

// ######################################################################
//! A spatial correlation channel
/*! Spatial correlation is computed at every location between the
  local patch and other patches at a given radius from the local
  patch. Radius is given in saliency map pixels. Note that in this
  channel, high 'salience' means low spatial correlation, i.e., lots
  of change. CAUTION: this channel will throw a fatal exception if
  your LevelSpec contains more than one level or if that level is
  different from the maplevel or if the deltas are non-zero. */
class ScorrChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  ScorrChannel(ModelManager& mgr);

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~ScorrChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  ModelParam<int> itsRadius; //!< Radius (in saliency map pixels)

  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  Image<float> itsMap;        //!< our temporal correlation map
};

// ######################################################################
//! An information channel.
/*! CAUTION: this channel will throw a fatal exception if your
  LevelSpec contains more than one level or if that level is different
  from the maplevel or if the deltas are non-zero. */
class InformationChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  InformationChannel(ModelManager& mgr);

  //! Destructor
  virtual ~InformationChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  ModelParam<float> itsEps; //!< epsilon on contrast sensitivity threshold

  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  Image<float> itsMap;  //!< our information map
};

// ######################################################################
//! A dummy channel that simply passes its greyscale input to its output
/*! USE WITH CAUTION: only tested for --levelspec=0-0,0-0,0 */
class DummyChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  DummyChannel(ModelManager& mgr);

  //! Destructor
  virtual ~DummyChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! overload
  bool outputAvailable() const;

  //! overload
  const Dims& getMapDims() const;

protected:
  ModelParam<float> itsFactor; //!< output = input * factor if non-zero

  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special submap computation function
  virtual Image<float> centerSurround(const uint cntrlev,
                                      const uint surrlev) const;

  Image<float> itsMap;  //!< our output map
};


// ######################################################################
//! A Contrast channel (based on Parkhurst & Niebur 2003)
/*! CAUTION: this channel will throw a fatal exception if your
  LevelSpec contains more than one level or if that level is different
  from the maplevel or if the deltas are non-zero. */
class PN03contrastChannel : public SingleChannel
{
public:
  //! Construct with standard params.
  PN03contrastChannel(ModelManager& mgr);

  //! Destructor
  virtual ~PN03contrastChannel();

  //! overload
  virtual bool outputAvailable() const;

protected:
  //! special input function
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! special output function
  virtual Image<float> getOutput();

  Image<float> itsMap;  //!< our contrast map
};


// ######################################################################
//! An disparity channel 
class DisparityChannel : public SingleChannel
{
public:

  //! Construct with standard params plus a disparity index for this object.
  /*! @param dIndex is used for various indexing such as with ParamMap
    and Jet interfacing.
    @param ang is the preferred orientation disparity for this channel, in
      degrees (0..180). 
    @param pha is the preferred phase disparity for this channel, in
      degrees (0..360). */
  DisparityChannel(ModelManager& mgr, const uint dIndex, const double ang, 
                   const double pha);

  //! Destructor
  virtual ~DisparityChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the preferred orientation (in degrees) of this channel.
  double angle() const;

  //! Get the preferred phase (in degrees) of this channel.
  double phase() const;

  //! set the raw filtered images from stereo channel to avoid redundancy 
  void setRawFilteredImages(Image<float> ***fImgLE, Image<float> ***fImgRI);

  //! access to the raw filtered images for easy viewing
  void getRawFilteredImages(Image<float> ****fImgLE,Image<float> ****fImgRI);

  //! access to the disparity even before stored in the queue (SingleChannel)
  void getDispMap(ImageSet<float> *dMap);

  //!  normalize the disparity map to [0, 1]
  void normalizeDispMap(ImageSet<float> tDispMap, int nChan);

  //! to get the dimensions of the raw filtered images from stereo channel
  void setNumPhase(int n);
  void setNumTheta(int t);

  //! perform the disparity calculation
  void doInput(const Image< PixRGB<float> >*, const Image<float>* bwimg, 
               const double t, const Image<byte>& clipMask);
  
  void storePyramid(double t);

  //! overload start1()
  void start1();

protected:
  ModelParam<uint>   itsIndex;
  ModelParam<double> itsOrientation;
  ModelParam<double> itsPhase;

private:

  Image<float> ***fImgL;
  Image<float> ***fImgR;
  
  ImageSet<float> dispMap;

  uint depth, nPhase, nTheta;

  const static float period   =  5.0;
};


// ######################################################################
//! A hue similarity channel
/*! Hue is represented as a bivariate Gaussian distibution in
  the C.I.E. color space, which is R/I vs. G/I.*/
class HueChannel : public SingleChannel
{
 public:

  //! Constructor with parameters of the bivariate Gaussian
  /*!@param mgr the ModelManager to assign this channel to
    @param muR the mean of the distribution in the R/I coordinate
    @param muG the mean in the G/I coordinate of the C.I.E. color space
    @param sigR the std in R/I
    @param sigG the std in G/I
    @param rho the correlation between R/I and G/I */
  HueChannel(ModelManager& mgr, 
             float muR, float muG, 
             float sigR, float sigG, float rho);

  //! Constuctor with sample pixel and std
  /*! Simplified version of the constructor that derives
    muR and muG from the pixel value hue, and it assumes
    that sigR=sigG=sig, and rho = 0.*/
  HueChannel(ModelManager& mgr,
             const PixRGB<byte> hue,
             float sig = 0.04);

  //! virtual destructor for inheritance
  virtual ~HueChannel();

  //! Combine all feature maps into a single output map
  /*! This overrides SingleChannel::combineSubMaps.
    Instead of doing center-surround and combining across scales,
    we just return the hue-filtered image at the map
    resolution.*/
  Image<float> combineSubMaps();

  //! always returns 1
  virtual uint numSubmaps() const;

  //! This is just a caching wrapper around computeSubmap()
  virtual Image<float> getSubmap(const uint index);

  //! Always returns "Hue"
  virtual const char* getSubmapName(const uint index) const;


 protected:
  //! special input function to compute the hue distance map
  virtual void doInput(const Image<PixRGB<float> >* colimg,
                       const Image<float>*, const double t,
                       const Image<byte>& clipMask);

  float itsMuR, itsMuG, itsSigR, itsSigG, itsRho;
};

// ######################################################################
//! A channel for detecting skin color.
/*! This implements a HueChannel with special values for face color
  derived from a statistical analysis of ~8000 faces in color images.*/
class SkinHueChannel : public HueChannel
{
 public:

  //! Constructor
  SkinHueChannel(ModelManager& mgr);

  //! virtual destructor, for inheritance
  virtual ~SkinHueChannel();
};

// ######################################################################
// ######################################################################
// #################### COMPLEX CHANNELS
// ######################################################################
// ######################################################################
// ######################################################################

// ######################################################################
//! A generic multi-convolution channel
/*! This channel reads a filter definition file from disk (text
  format) and builds a number of ConvolveChannel subchannels. In the
  current implementation, only 4 subchannels are supported, and their
  inputs will be the red, green, blue and yellow color components of
  the image as given by getRGBY(). Each subchannel will convolve its
  input using the convolution filter definition given in the text
  file. */
class MultiConvolveChannel : public ComplexChannel
{
public:
  //! Construct with standard params
  MultiConvolveChannel(ModelManager& mgr);

  //! Destructor
  virtual ~MultiConvolveChannel();

protected:
  ModelParam<std::string> itsFilterFname; //!< file name for the filters
  ModelParam<float> itsLumThresh; //!< Luminance threshold (see getRGBY())

  //! read filter definition and instantiate subchannels
  void start1();

  //! MultiConvolveChannel requires only color input
  virtual void doInput(const Image<PixRGB<float> >* colimg,
                       const Image<float>*, const double t,
                       const Image<byte>& clipMask);
};
//
//######################################################################
//! A composite color channel class that contains single and double !
// opponent color channels: the double opponent channel is useful for
// bottom-up salience and the single-opponent channel is used for
// recognition and top-down detection
class CompositeColorChannel : public ComplexChannel
{
public:
  //! Construct with standard params. and weight of the double
  //! opponent and single opponent channels
  CompositeColorChannel::CompositeColorChannel(ModelManager& mgr,
                                               const double dw = 1.0, 
                                               const double sw = 1.0);

  //! Destructor
  virtual ~CompositeColorChannel();

  //! lookup the member channel by its index
  ChannelBase& channel(const uint idx) const;

  //! lookup the member channel by its name
  ChannelBase& channel(const char* name) const;  

protected:
  //! Computes the double and single opponent color channels  
  void doInput(const Image<PixRGB<float> >* inp,
               const Image<float>* bwimg, const double t,
               const Image<byte>& clipMask);

  //! member color channels
  double itsDoubleOpponent;
  double itsSingleOpponent;  
};

// ######################################################################
//! A double opponent color channel class that combines r/g, b/y subchannels.
/*! This is an example of an heterogeneous complex channel where the
  various subchannels have distinct visual features. Thus, the JetSpec
  representation for such complex channel will be a series of
  independent single-channel representations, one for each
  subchannel. */
class ColorChannel : public ComplexChannel
{
public:
  //! Construct with standard params.
  ColorChannel(ModelManager& mgr);

  //! Destructor
  virtual ~ColorChannel();

  //! Get access the RedGreenChannel subchannel.
  virtual RedGreenChannel& rg() const;

  //! Get access to the BlueYellowChannel subchannel.
  virtual BlueYellowChannel& by() const;

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

protected:
  ModelParam<float> itsLumThresh; //!< Luminance threshold (see getRGBY())
  ModelParam<ColorComputeType> itsMethod;

  //! ColorChannel requires only color input.
  virtual void doInput(const Image<PixRGB<float> >* colimg,
                       const Image<float>*, const double t,
                       const Image<byte>& clipMask);

  uint itsRGidx, itsBYidx; // subComponent indices for rg and by
};

//
//######################################################################
//! A single opponent color channel class that combines r/g, g/r, b/y
//! and y/b subchannels.
/*! This is an example of an heterogeneous complex channel where the
  various subchannels have distinct visual features. Thus, the JetSpec
  representation for such complex channel will be a series of
  independent single-channel representations, one for each
  subchannel. */
class SOColorChannel : public ComplexChannel
{
public:
  //! Construct with standard params.
  SOColorChannel(ModelManager& mgr);

  //! Destructor
  virtual ~SOColorChannel();

  //! Get access the single opponent RedGreenChannel subchannel.
  virtual SORedGreenChannel& rg() const;

  //! Get access the single opponent GreenRedChannel subchannel.
  virtual SOGreenRedChannel& gr() const;

  //! Get access to the single opponent BlueYellowChannel subchannel.
  virtual SOBlueYellowChannel& by() const;

  //! Get access to the single opponent YellowBlueChannel subchannel.
  virtual SOYellowBlueChannel& yb() const;

  //! Get access the RedChannel subchannel.
  virtual RedChannel& r() const;

  //! Get access the  GreenChannel subchannel.
  virtual GreenChannel& g() const;

  //! Get access to the BlueChannel subchannel.
  virtual BlueChannel& b() const;

  //! Get access to the YellowChannel subchannel.
  virtual YellowChannel& y() const;

protected:
  ModelParam<float> itsLumThresh; //!< Luminance threshold (see getRGBY())

  //! SOColorChannel requires only color input.
  virtual void doInput(const Image<PixRGB<float> >* colimg,
                       const Image<float>*, const double t,
                       const Image<byte>& clipMask);

  //! subComponent indices for rg, gr, by, yb
  uint itsRGidx, itsGRidx, itsBYidx, itsYBidx;

  //! subComponent indices for r, g, b, y
  uint itsRidx, itsGidx, itsBidx, itsYidx;
};

//######################################################################
//! A composite channel containing a set of orientation-tuned gabor channels.
/*! This is an example of a homogeneous complex channel. Thus, the
  JetSpec representation will have an index for the subchannel number,
  and further indices for the various Jet components (which are
  identical for all subchannels). */
class OrientationChannel : public ComplexChannel
{
public:
  //! Construct with standard params
  OrientationChannel(ModelManager& mgr);

  //! Destructor
  virtual ~OrientationChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Wraps ComplexChannel's subChan() and casts the result to GaborChannel.
  virtual GaborChannel& gabor(const uint idx) const;

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

  //! the possible interaction types between orientations
  enum InteractType { NONE = 0, SUB_MEAN = 1, CUSTOM = 2 };

  //! returns the currently set interaction type between the orientation
  InteractType getInteractType();

  //! sets the interaction type between the orientations
  void setInteractType(InteractType type);

  //! returns the coefficients vector for interactions between orientations
  std::vector<float> getInteractCoeffs();

  //! sets the coefficients vector for the interactions between orientations
  void setInteractCoeffs(std::vector<float>& coeffs);

  //! compute interactions, then call ComplexChannel::getOutput()
  virtual Image<float> getOutput();

protected:
  //! trigger the interaction computation between the gabor filters
  void doInteractions();

  ModelParam<uint> itsNumOrients;  //!< number of Gabor subchannels

  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! parse the string for the interaction type between orientations
  virtual void parseInteractString(const std::string& value);

  //! OrientationChannel requires only luminance input.
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  //! type of interaction mode between orientations
  ModelParam<std::string> itsInteractString;

private:
  InteractType itsInteractType;
  std::vector<float> itsInteractCoeffs;
  bool itsDidInteractions;
};


// ######################################################################
//! A composite channel containing a set of color-band channels
/*! This is an example of a homogeneous complex channel. Thus, the
  JetSpec representation will have an index for the subchannel number,
  and further indices for the various Jet components (which are
  identical for all subchannels). */
class MultiColorBandChannel : public ComplexChannel
{
public:
  //! Construct with standard params
  MultiColorBandChannel(ModelManager& mgr);

  //! Destructor
  virtual ~MultiColorBandChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Wraps ComplexChannel's subChan() and casts the result to ColorBandChannel
  virtual ColorBandChannel& band(const uint idx) const;

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  ModelParam<uint> itsNumBands; //!< number of color bands
  ModelParam<uint> itsSatBands; //!< number of saturation bands
  ModelParam<float> itsHueSigma;   //!< width of each hue band
  ModelParam<float> itsSatSigma;   //!< width of each saturation band

  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! ColorBandChannel requires only the color image input
  virtual void doInput(const Image<PixRGB<float> > *colimg,
                       const Image<float>*, const double t,
                       const Image<byte>& clipMask);

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

};

// ######################################################################
//! A composite channel containing a set of direction channels
class MotionChannel : public ComplexChannel
{
public:
  //! Construct
  MotionChannel(ModelManager& mgr);

  //! destructor
  virtual ~MotionChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! returns a specific DirectionChannel
  virtual DirectionChannel& dirChan(const uint idx) const;

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  //! type of pyramid to use in our DirectionChannel subcomponents
  ModelParam<PyramidType> itsPyrType;

  //! number of DirectionChannel subcomponents
  ModelParam<uint> itsNumDirs;

  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! MotionChannel only required luminosity input
  virtual void doInput(const Image< PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;
};

// ######################################################################
//! A composite channel containing a set of Disparity channels
class StereoChannel : public ComplexChannel
{
public:
  //! Construct
  StereoChannel(ModelManager& mgr);
  
  //! destructor
  virtual ~StereoChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Wraps ComplexChannel's subChan() and casts the result to DisparityChannel
  virtual DisparityChannel& dispChan(const uint idO, 
                                          const uint idP) const;

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

  //! set the second image to complete stereo data  
  void setSecondImage(const Image<float>* bwimg);

  //! access method for easy viewing
  void getRawFilteredImages(Image<float> ****fImgLE,Image<float> ****fImgRI);

protected:
  //! type of pyramid to use in our DisparityChannel subcomponents
  ModelParam<PyramidType> itsPyrType;
 
  //! number of orientations of disparity subchannels
  ModelParam<uint> itsNumTheta;  

  //! number of phases of each orientations of disparity subchannels
  ModelParam<uint> itsNumPhase;  

  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! StereoChannel only required luminosity input AT THIS POINT
  virtual void doInput(const Image< PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  //! overload start1()
  virtual void start1();

private:

  void applyGabor();
  void calcDispResponse();  
  void selectMax();
  void createPhasedGaborPyramids();

  // depth of the feature map pyramid
  uint itsDepth;
  
  // captured color images
  Image< PixRGB<byte> > rgbImgL;
  Image< PixRGB<byte> > rgbImgR;

  // luminance values of the captured images
  Image<float> imgL;
  Image<float> imgR;
  
  // image pyramid for both images
  ImageSet<float> pyrL; 
  ImageSet<float> pyrR;
  
  // TUNED PARAMETERS
  const static uint  nPhase   = 4;
  const static uint  nTheta   = 4;
  const static float dPhase   = 360.0/nPhase;
  const static float dTheta   = 180.0/nTheta;
  const static float stddev   =  1.4;
  const static float period   =  4.2;

  Image<float> ***fImgL;
  Image<float> ***fImgR;

};

//
//######################################################################
//! An L junction channel that detects L junctions at different
//orientations
class LJunctionChannel :  public ComplexChannel
{
public:
  //! Construct with standard params.
  LJunctionChannel(ModelManager& mgr, VisualCortex* vcx);

  //! Destructor
  virtual ~LJunctionChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! Computes the L junctions at different orientations
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  ModelParam<uint> itsNumOrients;  //!< number of junction subchannels 

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  VisualCortex *itsVCX;
};

//
//######################################################################
//! A T junction channel that detects T junctions at different
//orientations
class TJunctionChannel :  public ComplexChannel
{
public:
  //! Construct with standard params.
  TJunctionChannel(ModelManager& mgr, VisualCortex* vcx);

  //! Destructor
  virtual ~TJunctionChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! Computes the T junctions at different orientations
  virtual void doInput(const Image<PixRGB<float> >*,
                                 const Image<float>* bwimg, const double t,
                                 const Image<byte>& clipMask);

  ModelParam<uint> itsNumOrients;  //!< number of T junction subchannels 

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  VisualCortex *itsVCX;
};

//
//######################################################################
//! A X junction channel that detects crosses at different
//orientations
class XJunctionChannel :  public ComplexChannel
{
public:
  //! Construct with standard params.
  XJunctionChannel(ModelManager& mgr, VisualCortex *vcx);

  //! Destructor
  virtual ~XJunctionChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! Computes the cross junctions at different orientations
  virtual void doInput(const Image<PixRGB<float> >*,
                                 const Image<float>* bwimg, const double t,
                                 const Image<byte>& clipMask);

  ModelParam<uint> itsNumOrients;  //!< number of junction subchannels 

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  VisualCortex *itsVCX;
};

//
//######################################################################
//! An end point detector channel that detects line ends at different
//orientations
class EndPointChannel :  public ComplexChannel
{
public:
  //! Construct with standard params.
  EndPointChannel(ModelManager& mgr, VisualCortex* vcx);

  //! Destructor
  virtual ~EndPointChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Overload so that we can reconfigure when our params get changed
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  //! (re-)build our subchannels
  virtual void buildSubChans();

  //! Computes the end stops at different orientations
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);
  
  ModelParam<uint> itsNumOrients;  //!< number of junction subchannels 

  //! keep track of what options to export in case we rebuild our subchans
  int itsExportOpts;

  VisualCortex *itsVCX;
};

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !CHANNELS_H_DEFINED
