/*!@file ChannelsBeo.C Definitions for the Beo version of the channels */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ChannelsBeo.C 3078 2004-01-29 03:57:24Z itti $
//

#include "ChannelsBeo.H"
#include "ModelOptionDefs.H"
#include "Beowulf.H"

// ######################################################################
// ######################################################################
// #################### SINGLE CHANNELS
// ######################################################################
// ######################################################################

// ######################################################################
IntensityChannelBeo::
IntensityChannelBeo(ModelManager& mgr, const SharedPtr<Beowulf>& beow) :
  IntensityChannel(mgr), SingleChannelBeo(this, beow)
{ }

IntensityChannelBeo::~IntensityChannelBeo()
{ }

bool IntensityChannelBeo::outputAvailable() const
{
  const_cast<IntensityChannelBeo*>(this)->waitForOutput();
  return IntensityChannel::outputAvailable();
}

void IntensityChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void IntensityChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
RedGreenChannelBeo::
RedGreenChannelBeo(ModelManager& mgr, const SharedPtr<Beowulf>& beow) :
  RedGreenChannel(mgr), SingleChannelBeo(this, beow)
{ }

RedGreenChannelBeo::~RedGreenChannelBeo()
{ }

bool RedGreenChannelBeo::outputAvailable() const
{
  const_cast<RedGreenChannelBeo*>(this)->waitForOutput();
  return RedGreenChannel::outputAvailable();
}

void RedGreenChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void RedGreenChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
BlueYellowChannelBeo::
BlueYellowChannelBeo(ModelManager& mgr, const SharedPtr<Beowulf>& beow) :
  BlueYellowChannel(mgr), SingleChannelBeo(this, beow)
{ }

BlueYellowChannelBeo::~BlueYellowChannelBeo()
{ }

bool BlueYellowChannelBeo::outputAvailable() const
{
  const_cast<BlueYellowChannelBeo*>(this)->waitForOutput();
  return BlueYellowChannel::outputAvailable();
}

void BlueYellowChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void BlueYellowChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
GaborChannelBeo::
GaborChannelBeo(ModelManager& mgr, const uint oriIndex, const double ang,
                const SharedPtr<Beowulf>& beow) :
  GaborChannel(mgr, oriIndex, ang), SingleChannelBeo(this, beow)
{ }

GaborChannelBeo::~GaborChannelBeo()
{ }

bool GaborChannelBeo::outputAvailable() const
{
  const_cast<GaborChannelBeo*>(this)->waitForOutput();
  return GaborChannel::outputAvailable();
}

void GaborChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void GaborChannelBeo::start1()
{ GaborChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
FlickerChannelBeo::
FlickerChannelBeo(ModelManager& mgr, const SharedPtr<Beowulf>& beow) :
  FlickerChannel(mgr), SingleChannelBeo(this, beow)
{ }

FlickerChannelBeo::~FlickerChannelBeo()
{ }

bool FlickerChannelBeo::outputAvailable() const
{
  FlickerChannelBeo *flick = const_cast<FlickerChannelBeo*>(this);
  flick->waitForOutput();

  // FlickerChannel has a secondary output cache...
  if (!itsOutput.initialized())
    flick->itsOutput = flick->SingleChannel::getOutput();

  return FlickerChannel::outputAvailable();
}

void FlickerChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{
  sendInput(bwimg, t, clipMask);
  // FlickerChannel needs to keep track of the previous input, to
  // return a blank image based on that size as long as the output is
  // not available:
  itsPrevInput = *bwimg;
  // Note: actually, FlickerChannel does more complicated stuff to
  // compute itsPrevInput. There is a possible source of discrepancy
  // here, if people were to try to get itsPrevInput back out of
  // FlickerChannelBeo.

  // also invalidate our local output cache
  itsOutput.freeMem();
}

void FlickerChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
ColorBandChannelBeo::
ColorBandChannelBeo(ModelManager& mgr, const uint bandidx,
                    const SharedPtr<Beowulf>& beow) :
  ColorBandChannel(mgr, bandidx), SingleChannelBeo(this, beow)
{ }

ColorBandChannelBeo::~ColorBandChannelBeo()
{ }

bool ColorBandChannelBeo::outputAvailable() const
{
  const_cast<ColorBandChannelBeo*>(this)->waitForOutput();
  return ColorBandChannel::outputAvailable();
}

void ColorBandChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void ColorBandChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
DirectionChannelBeo::
DirectionChannelBeo(ModelManager& mgr, const uint dirIndex,
                    const double direction,  const PyramidType type,
                    const SharedPtr<Beowulf>& beow) :
  DirectionChannel(mgr, dirIndex, direction, type),
  SingleChannelBeo(this, beow)
{ }

DirectionChannelBeo::~DirectionChannelBeo()
{ }

bool DirectionChannelBeo::outputAvailable() const
{
  const_cast<DirectionChannelBeo*>(this)->waitForOutput();
  return DirectionChannel::outputAvailable();
}

void DirectionChannelBeo::
doInput(const Image<PixRGB<float> >*, const Image<float>* bwimg,
        const double t, const Image<byte>& clipMask)
{ sendInput(bwimg, t, clipMask); }

void DirectionChannelBeo::start1()
{ SingleChannel::start1(); SingleChannelBeo::requestBeowulfNode(); }

// ######################################################################
// ######################################################################
// #################### COMPLEX CHANNELS
// ######################################################################
// ######################################################################

// ######################################################################
ColorChannelBeo::ColorChannelBeo(ModelManager& mgr,
                                 const SharedPtr<Beowulf>& beow) :
  ColorChannel(mgr)
{
  // replace ColorChannel's subchannels by their beo version:
  LINFO("Switching to Beo subchannels");
  removeAllSubComponents();
  itsRGidx = addSubChan(makeShared(new RedGreenChannelBeo(mgr, beow)));
  itsBYidx = addSubChan(makeShared(new BlueYellowChannelBeo(mgr, beow)));
}

ColorChannelBeo::~ColorChannelBeo()
{  }

// ######################################################################
OrientationChannelBeo::OrientationChannelBeo(ModelManager& mgr,
                                             const SharedPtr<Beowulf>& beow) :
  OrientationChannel(mgr), itsBeo(beow)
{
  // replace OrientationChannel's normal channels by our Beo ones:
  LINFO("Switching to Beo subchannels");
  buildSubChans();
}

OrientationChannelBeo::~OrientationChannelBeo()
{  }

void OrientationChannelBeo::buildSubChans()
{
  // see GaborChannel::buildSubchans() for comments...
  removeAllSubComponents(); setHomogeneous(true);
  LINFO("Using %d Beo orientations spanning [0..180]deg",
        itsNumOrients.getVal());
  for (uint ori = 0; ori < itsNumOrients.getVal(); ++ori) {
    SharedPtr<GaborChannelBeo>
      ch(new GaborChannelBeo(*manager, ori, 180.0 * double(ori) /
                             double(itsNumOrients.getVal()), itsBeo));
    addSubChan(ch);
  }
  if (itsExportOpts != OPTEXP_NONE)
    OrientationChannel::exportOptions(itsExportOpts, true);
}

// ######################################################################
MultiColorBandChannelBeo::
MultiColorBandChannelBeo(ModelManager& mgr, const SharedPtr<Beowulf>& beow) :
  MultiColorBandChannel(mgr), itsBeo(beow)
{
  // replace normal channels by our Beo ones:
  LINFO("Switching to Beo subchannels");
  buildSubChans();
}

MultiColorBandChannelBeo::~MultiColorBandChannelBeo()
{  }

void MultiColorBandChannelBeo::buildSubChans()
{
  // see MultiColorBandChannel::buildSubchans() for comments...
  removeAllSubComponents(); setHomogeneous(true);
  LINFO("Using %d Beo color bands", itsNumBands.getVal());
  for (uint i = 0; i < itsNumBands.getVal(); i ++) {
    SharedPtr<ColorBandChannelBeo>
      ch(new ColorBandChannelBeo(*manager, i, itsBeo));
    addSubChan(ch);
  }
  if (itsExportOpts != OPTEXP_NONE)
    ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
MotionChannelBeo::MotionChannelBeo(ModelManager& mgr,
                                   const SharedPtr<Beowulf>& beow) :
  MotionChannel(mgr), itsBeo(beow)
{
  // replace normal channels by our Beo ones:
  LINFO("Switching to Beo subchannels");
  buildSubChans();
}

MotionChannelBeo::~MotionChannelBeo()
{  }

void MotionChannelBeo::buildSubChans()
{
  // see MotionChannel::buildSubchans() for comments...
  removeAllSubComponents(); setHomogeneous(true);
  LINFO("Using %d Beo directions spanning [0..360]deg", itsNumDirs.getVal());
  for (uint i = 0; i < itsNumDirs.getVal(); i ++) {
    SharedPtr<DirectionChannelBeo>
      ch(new DirectionChannelBeo(*manager, i, 360.0 * double(i) / 
                                 double(itsNumDirs.getVal()),
                                 itsPyrType.getVal(), itsBeo));
    addSubChan(ch);
  }
  if (itsExportOpts != OPTEXP_NONE)
    ModelComponent::exportOptions(itsExportOpts, true);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
