/*!@file ComplexChannel.C Channel class that pools across multiple
  subchannels. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ComplexChannel.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "ComplexChannel.H"
#include "ModelManager.H"
#include "fancynorm.H"
#include "log.H"
#include "ParamMap.H"
#include "SharedPtr.H"
#include "Jet.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"

#include <algorithm>

// ######################################################################
// ######################################################################
// ComplexChannel member definitions:
// ######################################################################
// ######################################################################


// ######################################################################
ComplexChannel::ComplexChannel(ModelManager& mgr, const std::string& descrName,
                               const std::string& tagName,
                               const VisualFeature vs) :
  ChannelBase(mgr, descrName, tagName, vs),
  itsNormType(this, "MaxNormType", VCXNORM_FANCY),
  itsSaveOutputMap(this, "ComplexChannelSaveOutputMap", false),
  itsUseOlderVersion(this, "UseOlderVersion", true) 
{  }

// ######################################################################
ComplexChannel::~ComplexChannel()
{  }

// ######################################################################
void ComplexChannel::start2()
{
  // in the new version, leave the output range of our subchannels
  // open rather than in some cases forcing it to a given range of values:
  if (itsUseOlderVersion.getVal() == false)
    for (uint i = 0; i < numChans(); ++i)
      subChan(i).setOutputRange(0.0f, 0.0f);

  // in the older version, we used to set the map range as we would
  // also apply spatial competition for salience to the output map,
  // only if using the MAXNORM type of competition, and otherwise we
  // would not touch the range:
  if (itsUseOlderVersion.getVal() && itsNormType.getVal() != VCXNORM_MAXNORM)
    setOutputRange(0.0f, 0.0f);
}

// ######################################################################
void ComplexChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    {
      manager->requestOption(this, itsNormType);
      manager->requestOption(this, itsUseOlderVersion);
    }
  if (whatToExport & OPTEXP_SAVE)
    manager->requestOption(this, itsSaveOutputMap);

  // the base class will propagate this call to our subcomponents:
  ChannelBase::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ComplexChannel::reset(const bool recurse)
{
  // reset some stuff for ComplexChannel
  itsOutput.freeMem();

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
uint ComplexChannel::numChans() const
{ return numSubComp(); }

// ######################################################################
ChannelBase& ComplexChannel::subChan(const uint idx) const
{ return dynamic_cast<ChannelBase&>(*(subComponent(idx).get())); }

// ######################################################################
ChannelBase& ComplexChannel::subChanFromSubmapNum(const uint oldIdx,
                                                  uint& newIdx) const
{
  ASSERT(oldIdx < numSubmaps());
  uint subchan = 0;
  lookupSubmap(oldIdx, subchan, newIdx);

  return subChan(subchan);
}

// ######################################################################
void ComplexChannel::readFrom(const ParamMap& pmap)
{
  ChannelBase::readFrom(pmap);

  for (uint i = 0; i < numChans(); ++i)
    {
      const char* tagname = subChan(i).tagName().c_str();
      if (pmap.hasParam(tagname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(tagname);
          subChan(i).readFrom(*submap);
        }
    }
}

// ######################################################################
void ComplexChannel::writeTo(ParamMap& pmap) const
{
  ChannelBase::writeTo(pmap);

  for (uint i = 0; i < numChans(); ++i)
    {
      SharedPtr<ParamMap> submap(new ParamMap);
      subChan(i).writeTo(*submap);
      pmap.putSubmap(subChan(i).tagName().c_str(), submap);
    }
}

// ######################################################################
void ComplexChannel::setListener(SharedPtr<ChannelListener> listener)
{
  for (uint i = 0; i < numChans(); ++i)
    subChan(i).setListener(listener);
}

// ######################################################################
void ComplexChannel::cleanCoeffs(const double eps)
{
  for (uint i = 0; i < numChans(); ++i)
    subChan(i).cleanCoeffs(eps);
}

// ######################################################################
void ComplexChannel::clampCoeffs(const double cmin, const double cmax)
{
  for (uint i = 0; i < numChans(); ++i)
    subChan(i).clampCoeffs(cmin, cmax);
}

// ######################################################################
double ComplexChannel::absSumCoeffs() const
{
  double sum = 0.0;
  for (uint i = 0; i < numChans(); ++i)
    sum += subChan(i).absSumCoeffs() * subChan(i).getTotalWeight();
  
  return sum;
}

// ######################################################################
void ComplexChannel::normalizeCoeffs(const double div)
{
  for (uint i = 0; i < numChans(); ++i)
    subChan(i).normalizeCoeffs(div);
}
// ######################################################################
void ComplexChannel::learn(const Image<byte>& dmap, const double eta,
                           const bool softmask,
                           const int inthresh, const int outthresh)
{
  killCaches();

  for (uint i = 0; i < numChans(); ++i)
    subChan(i).learn(dmap, eta, softmask, inthresh, outthresh);
}

// ######################################################################
bool ComplexChannel::outputAvailable() const
{
  for (uint i = 0; i < numChans(); ++i)
    if (subChan(i).outputAvailable() == false) return false;
  return true;
}

// ######################################################################
const Dims& ComplexChannel::getMapDims() const
{ return subChan(0).getMapDims(); }

// ######################################################################
const Dims& ComplexChannel::getInputDims() const
{ return subChan(0).getInputDims(); }

// ######################################################################
void ComplexChannel::fillJet(const Point2D& loc, Jet<float>& j,
                             std::vector<int>& v) const
{
  if (isHomogeneous())
    {
      v.push_back(0); // add one index for the subchannels
      int vidx = v.size() - 1;
      for (uint i = 0; i < numChans(); ++i)
        {
          v[vidx] = int(i);
          subChan(i).fillJet(loc, j, v);
        }
      v.pop_back(); // restore v for parent to use again
    }
  else
    for (uint i = 0; i < numChans(); ++i)
      subChan(i).fillJet(loc, j, v);
}

// ######################################################################
void ComplexChannel::fillJetNI(const Point2D& loc, Jet<float>& j,
                               std::vector<int>& v) const
{
  if (isHomogeneous())
    {
      v.push_back(0); // add one index for the subchannels
      int vidx = v.size() - 1;
      for (uint i = 0; i < numChans(); ++i)
        {
          v[vidx] = int(i);
          subChan(i).fillJetNI(loc, j, v);
        }
      v.pop_back(); // restore v for parent to use again
    }
  else
    for (uint i = 0; i < numChans(); ++i)
      subChan(i).fillJetNI(loc, j, v);
}

// ######################################################################
void ComplexChannel::fillJetSpec(JetSpec& js) const
{
  if (numChans()) {
    // are we a homogeneous complex channel? If so, add a range for
    // number of subchannels and initialize rest from first subchannel
    if (isHomogeneous()) {
      // create ranges for the number of subchannels we hold:
      js.addIndexRange(visualFeature(), RAW, 0, numChans() - 1);
      //js.addIndexRange(visualFeature(), RAW_CS, 0, numChans() - 1);
      //js.addIndexRange(visualFeature(), NORM_CS, 0, numChans() - 1);
      
      // use the first subchannel to add the range specs for us:
      subChan(0).fillJetSpec(js);
    } else {
      // We are heterogeneous; each subchannel holds a different
      // VisualFeature; initialize all subchannels in turn, and we do
      // not have to do anything for us proper:
      for (uint i = 0; i < numChans(); i ++)
        subChan(i).fillJetSpec(js);
    }
  }
}

// ######################################################################
uint ComplexChannel::numSubmaps() const
{
  uint sum = 0;
  for (uint i = 0; i < numChans(); ++i) sum += subChan(i).numSubmaps();
  return sum;
}

// ######################################################################
void ComplexChannel::lookupSubmap(const uint idx, uint& subchan,
                                  uint& subidx) const
{
  ASSERT(idx < numSubmaps());
  uint offset = 0;
  for (subchan = 0; subchan < numChans(); ++subchan)
    {
      subidx = idx - offset;
      const uint nsub = subChan(subchan).numSubmaps();
      if (subidx < nsub)
        {
          // OK, we have found the right subchan+submap combination:
          return;
        }
      else
        offset += nsub;
    }
  LFATAL("invalid submap index: %d", idx);
}

// ######################################################################
Image<float> ComplexChannel::getSubmap(const uint idx) const
{
  uint subchan = 0, subidx = 0;
  lookupSubmap(idx, subchan, subidx);
  return subChan(subchan).getSubmap(subidx);
}

// ######################################################################
const char* ComplexChannel::getSubmapName(const uint idx) const
{
  uint subchan = 0, subidx = 0;
  lookupSubmap(idx, subchan, subidx);
  return subChan(subchan).getSubmapName(subidx);
}

// ######################################################################
void ComplexChannel::getCoeff(std::vector<double>& coeff)
{
  for(uint i = 0; i < numChans(); i++)
    subChan(i).getCoeff(coeff);
}

// ######################################################################
void ComplexChannel::setGroupingCoeff(std::vector<double>& coeff, uint& idx)
{
  for(uint i = 0; i < numChans(); i++)
    subChan(i).setGroupingCoeff(coeff, idx);
}

// ######################################################################
Image<float> ComplexChannel::perceptualGrouping()
{
  Image<float> myPercept(getMapDims(), ZEROS);
  for (uint i = 0; i < numChans(); ++i)
    {
      const float w = float(subChan(i).getTotalWeight());
      if (w != 0.0f)
        {
          Image<float> subChanOut = subChan(i).perceptualGrouping();
          if (w != 1.0f) subChanOut *= w;
          myPercept += subChanOut;
        }
    }
  return myPercept;
}

// ######################################################################
void ComplexChannel::biasFeatures(const double* mean,
                                  const double* sigma, int& idx)
{
  for(uint i = 0; i < numChans(); i++)
    subChan(i).biasFeatures(mean, sigma, idx);
}

// ######################################################################
void ComplexChannel::getFeatures(const Point2D& locn,
                                 std::vector<double>& mean) const
{
  for(uint i = 0; i < numChans(); i++)
    subChan(i).getFeatures(locn, mean);
}

// ######################################################################
void ComplexChannel::getFeaturesBatch(std::vector<Point2D*> *locn,
                                      std::vector<std::vector<float> > *mean,
                                      int *count) const
{
  for(uint i = 0; i < numChans(); i++)
    subChan(i).getFeaturesBatch(locn, mean, count);
}




// ######################################################################
void ComplexChannel::findWeight()
{
  double max = 0.0, sum = 0.0;
  for(uint i = 0; i < numChans(); i++)
    {
      // first find the subchannel's weights
      subChan(i).findWeight(); 
      double wt = subChan(i).getTotalWeight();
      max = std::max(max, wt);
      sum += wt;
    }
  // initialize the channel's weight with the max subchannel weight
  setTotalWeight (max);  

  // normalize subchannel weights so that they add to 1
  for(uint i = 0; i < numChans(); i++)
    {
      if (sum != 0.0) 
        subChan(i).setTotalWeight (subChan(i).getTotalWeight()/sum);
    }

}

// ######################################################################
bool ComplexChannel::reco(const ParamMap& pmap, double& mismatch)
{
  double mm = 0.0;
  for (uint i = 0; i < numChans(); ++i)
    {
      const char* tagname = subChan(i).tagName().c_str();
      if (pmap.hasParam(tagname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(tagname);
          subChan(i).reco(*submap, mm);
        }
    }
  mismatch += mm;
  return true;
}

// ######################################################################
void ComplexChannel::findLikelihood(const ParamMap& pmap,
                                    double& LogLikelihood)
{
  double LogProb = 0.0;
  for (uint i = 0; i < numChans(); ++i)
    {
      const char* tagname = subChan(i).tagName().c_str();
      if (pmap.hasParam(tagname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(tagname);
          subChan(i).findLikelihood(*submap, LogProb);
        }
    }
  LogLikelihood += LogProb;  // channels are independent
  return;
}

// ######################################################################
uint ComplexChannel::addSubChan(const SharedPtr<ChannelBase> ch)
{
  // update our homogeneity status:
  if (ch->visualFeature() != visualFeature()) setHomogeneous(false);

  // add the subchannel:
  return addSubComponent(ch);
}

// ######################################################################
void ComplexChannel::killCaches()
{
  ChannelBase::killCaches(); // call our base class's implementation
  itsOutput.freeMem();
}

// ######################################################################
Image<float> ComplexChannel::getOutput()
{
  if (!itsOutput.initialized()) itsOutput = combineOutputs();
  return itsOutput;
}

// ######################################################################
Image<float> ComplexChannel::combineOutputs()
{
  Image<float> output(getMapDims(), ZEROS);
  for (uint i = 0; i < numChans(); ++i)
    {
      const float w = float(subChan(i).getTotalWeight());
      if (w != 0.0f)
        {
          Image<float> subChanOut = subChan(i).getOutput();
          if (w != 1.0f) subChanOut *= w;
          output += subChanOut;

          if (manager->debugMode())
            LDEBUG("%s: %s weight %f", tagName().c_str(),
                   subChan(i).tagName().c_str(), w);
        }
    }

  // apply max-normalization on the output as needed:
  if (itsNormType.getVal() == VCXNORM_LANDMARK)
    {
      float goodness = pow(goodness_map(output), 0.05);
      LINFO("GOODNESS = %f", goodness);
      output *= goodness;
    }
  else
    {
      LDEBUG("%s: Normalizing output: %s(%f .. %f)", tagName().c_str(),
             maxNormTypeName(itsNormType.getVal()), itsOutputRangeMin.getVal(),
             itsOutputRangeMax.getVal());
      output = maxNormalize(output, itsOutputRangeMin.getVal(),
                            itsOutputRangeMax.getVal(), itsNormType.getVal());
    }

  // print some debug info if in debug mode:
  if (manager->debugMode())
    {
      float mi, ma; output.getMinMax(mi, ma);
      LDEBUG("%s: final range [%f .. %f]", tagName().c_str(), mi, ma);
    }

  LINFO("Computed %s Conspicuity Map", descriptiveName().c_str());

  return output;
}

// ######################################################################
void ComplexChannel::drawResults(Image< PixRGB<byte> >& traj)
{ for (uint i = 0; i < numChans(); i++) subChan(i).drawResults(traj); }

// ######################################################################
void ComplexChannel::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // save our own map:
  if (itsSaveOutputMap.getVal())
    ofs->writeFloat(getOutput(), PNM, -1, true, "CO%s-", tagName().c_str());

  // now do it for our subchannels:
  for (uint i = 0; i < numChans(); i++) subChan(i).saveResults(ofs);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
