/*!@file ComplexChannel.H Channel class that pools across multiple
  subchannels. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ComplexChannel.H 4228 2005-02-23 17:19:25Z vidhya $
//

#ifndef COMPLEXCHANNEL_H_DEFINED
#define COMPLEXCHANNEL_H_DEFINED

#include "ChannelBase.H"
#include "Image.H"
#include "fancynorm.H" // for MaxNormType

// ######################################################################
//! A an abstract channel class that pools across several subchannels.
class ComplexChannel : public ChannelBase
{
public:
  //! Constructor.
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage
      @param vs The VisualFeature implemented by the channel */
  ComplexChannel(ModelManager& mgr, const std::string& descrName,
                 const std::string& tagName, const VisualFeature vs);

  //! Virtual destructor.
  virtual ~ComplexChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset ComplexChannel and propagate the reset signal to the sub-components
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! Get the number of subchannels.
  virtual uint numChans() const;

  //! Look up one of the subchannels by its integer index.
  virtual ChannelBase& subChan(const uint idx) const;

  //! Look up one of the subchannels by a submap number
  /*!@param oldIdx - the submap index (in the ComplexChannel); 
     @param newIdx - return value: the number of the submap within 
     the subchannel*/
  virtual ChannelBase& subChanFromSubmapNum(const uint oldIdx,
                                            uint& newIdx) const;

  //! Add a subchannel
  /*! This is just a wrapper to ModelComponent::addSubComponent(),
    with some checking that the subcomponent indeed is a ChannelBase
    derivative, and some adjustment of our homogeneity (if the newly
    added subchannel has a different VisualFeature from ours, then be
    become heterogeneous). Returns the subChan index. */
  virtual uint addSubChan(const SharedPtr<ChannelBase> ch);

  virtual void readFrom(const ParamMap& pmap);

  virtual void writeTo(ParamMap& pmap) const;

  virtual void setListener(SharedPtr<ChannelListener> listener);

  virtual void cleanCoeffs(const double eps);

  virtual void clampCoeffs(const double cmin, const double cmax);

  virtual double absSumCoeffs() const;

  virtual void normalizeCoeffs(const double div);

  virtual void learn(const Image<byte>& dmap, const double eta,
                     const bool softmask = false,
                     const int inthresh = 20, const int outthresh = 255);

  virtual bool outputAvailable() const;

  virtual const Dims& getMapDims() const;

  virtual const Dims& getInputDims() const;

  virtual void fillJet(const Point2D& loc, Jet<float>& j,
                       std::vector<int>& v) const;

  virtual void fillJetNI(const Point2D& loc, Jet<float>& j,
                         std::vector<int>& v) const;

  virtual void fillJetSpec(JetSpec &js) const;

  virtual uint numSubmaps() const;

  virtual Image<float> getSubmap(const uint index) const;

  virtual const char* getSubmapName(const uint index) const;

  virtual void getCoeff(std::vector<double>& coeff);

  virtual void setGroupingCoeff(std::vector<double>& coeff, uint& idx);

  virtual Image<float> perceptualGrouping();

  virtual void biasFeatures(const double* mean, const double* sigma, int& idx);

  virtual void getFeatures(const Point2D& locn,
                           std::vector<double>& mean) const;

  virtual void getFeaturesBatch(std::vector<Point2D*> *locn,
                                std::vector<std::vector<float> > *mean,
                                int *count) const;
  virtual void findWeight();
  
  virtual bool reco(const ParamMap& pmap, double& mismatch);

  virtual void findLikelihood(const ParamMap& pmap, double& likelihood);
  
  //! Form output by combining output of subchannels.
  /*! This function just provides a caching wrapper around
      combineOutputs(), where the real work is done. ComplexChannel
      provides a default implementation of combineOutputs() that sums the
      outputs of the subchannels and then optionally maxNormalize()'s the
      result. */
  virtual Image<float> getOutput();

  //! Draws something channel-specific in the trajectory image.
  /*! By default, we do nothing but propagate the call to our subchannels */
  virtual void drawResults(Image< PixRGB<byte> >& traj);

  //! Save our various maps using an OutputFrameSeries
  /*! By default, we save our output map (if itsSaveOutputMap is true)
    and propagate the call to our subchannels. Our output
    (conspicuity) map will be prefixed by "CO<tagname>-" */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

protected:
  ModelParam<MaxNormType> itsNormType; //!< Normalization; see fancynorm.H
  ModelParam<bool> itsSaveOutputMap;  //!< Save our output map?
  ModelParam<bool> itsUseOlderVersion; /*! use old version? see
                                         Visualcortex.H */
  //! Clears the output cache.
  virtual void killCaches();

  //! Combines the outputs of subchannels.
  /*! The default implementation just sums the outputs of the subchannels
      and then maxNormalize()'s if requested by itsNormType. */
  virtual Image<float> combineOutputs();

  //! get started
  virtual void start2();

private:
  ComplexChannel(const ComplexChannel&); // not allowed
  ComplexChannel& operator=(const ComplexChannel&); // not allowed

  void lookupSubmap(const uint idx, uint& subchan, uint& subidx) const;

  Image<float>               itsOutput;
};


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !COMPLEXCHANNEL_H_DEFINED
