/*!@file FrameSeries.C a series of frames */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: FrameSeries.C 3107 2004-02-10 10:13:24Z itti $
//

#include "FrameSeries.H"
#include "log.H"
#include <cstdio>
#include "Pixels.H"
#include "Image.H"
#include "ModelManager.H"
#include "Image_ShapeOps.H"   // for rescale()
#include "Image_ColorOps.H"   // for luminance()

// maximum number of displays we cando before giving up
#define MAXDISP 20

// ######################################################################
// #################### FrameSeries
// ######################################################################
FrameSeries::FrameSeries(ModelManager& mgr, const std::string& prefixName,
                         const std::string& descrName,
                         const std::string& tagName) :
  ModelComponent(mgr, prefixName+" "+descrName, prefixName+tagName),
  itsFrameRange(this, prefixName+"FrameRange", FrameRange(0, 0, 0.0F)),
  itsDims(this, prefixName+"FrameDims", Dims(0, 0)),
  itsPreserveAspect(this,prefixName+"PreserveAspect", false),
  itsShowFrames(this, prefixName+"ShowFrames", false),
  itsRasterFileFormat(this, prefixName+"RasterFileFormat",
                      prefixName.compare("Input")? Auto : PNM),
  itsStem(""), itsCurrentFrame(-1), itsOldTime(0.0), itsDidDisplay(0)
{  }

// ######################################################################
void FrameSeries::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsFrameRange);
    manager->requestOption(this, itsDims);
    manager->requestOption(this, itsPreserveAspect);
    manager->requestOption(this, itsShowFrames);
    manager->requestOption(this, itsRasterFileFormat);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void FrameSeries::start2()
{
  itsCurrentFrame = itsFrameRange.getVal().first - 1;
  itsOldTime = 0.0;
}

// ######################################################################
FrameSeries::~FrameSeries()
{ }

// ######################################################################
void FrameSeries::reset(const bool recurse)
{
  // reset some stuff for FrameSeries
  itsCurrentFrame = itsFrameRange.getVal().first - 1;
  itsOldTime = 0.0;
  itsDidDisplay = 0;

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void FrameSeries::setFileStem(const std::string& stem)
{
  if (started())
    LFATAL("Cannot change file stem while started -- ABORT");
  itsStem = stem;
}

// ######################################################################
std::string FrameSeries::getFileStem() const
{ return itsStem; }

// ######################################################################
bool FrameSeries::isMultiframe() const
{ return itsFrameRange.getVal().isMultiframe(); }

// ######################################################################
double FrameSeries::delay() const
{ return itsFrameRange.getVal().delay * 0.001; }

// ######################################################################
int FrameSeries::frame() const
{ return itsCurrentFrame; }

// ######################################################################
int FrameSeries::minframe() const
{ return itsFrameRange.getVal().first; }

// ######################################################################
int FrameSeries::maxframe() const
{ return itsFrameRange.getVal().last; }

// ######################################################################
int FrameSeries::numFrames() const
{ return (maxframe() - minframe() + 1); }

// ######################################################################
bool FrameSeries::finished() const
{ return itsCurrentFrame > itsFrameRange.getVal().last; }

// ######################################################################
FrameSeries::State FrameSeries::update(double stime, double sdelay)
{
  // reset our itsDidDisplay:
  itsDidDisplay = 0;

  // check whether we need to move on to the next frame:
  if (itsCurrentFrame == itsFrameRange.getVal().first - 1 ||
      (itsFrameRange.getVal().isMultiframe() &&
       stime + 1.5 * sdelay - itsOldTime > itsFrameRange.getVal().delay*0.001))
    {
      ++itsCurrentFrame;
      itsOldTime = itsFrameRange.getVal().delay * 0.001 *
        (itsCurrentFrame - itsFrameRange.getVal().first);
      if (isMultiframe()) {
        LINFO("%s advancing to frame %d at %.02fms",
              descriptiveName().c_str(), itsCurrentFrame,
              (stime + sdelay) * 1000.0);
        manager->textLog("%s: frame %d at %.02fms",
                         descriptiveName().c_str(), itsCurrentFrame,
                         (stime + sdelay) * 1000.0);
      }
      if (finished()) return COMPLETE;
      return NEXT;
    }
  return SAME;
}

// ######################################################################
FrameSeries::State FrameSeries::updateNext()
{
  ++itsCurrentFrame;
  if (finished()) return COMPLETE;
  else return NEXT;
}

// ######################################################################
bool FrameSeries::didDisplayFrames() const
{ return itsDidDisplay != 0; }

// ######################################################################
std::string FrameSeries::computeFileName(const int framenum,
                                         const char *otherstem,
                                         va_list ap) const
{
  // first split itsStem into path and file components:
  uint idx = itsStem.rfind('/'); std::string path, file;
  if (idx == itsStem.npos)  // no / in the stem
    file = itsStem;
  else
    {
      path = itsStem.substr(0, idx) + '/';
      file = itsStem.substr(idx+1, itsStem.npos);
    }

  // if we are not multiframe, ignore frame number but append "_itsStem"
  // to filename instead:
  if (itsFrameRange.getVal().isMultiframe() == false)
    {
      // do we want to use an alternate stem and args?
      if (otherstem)
        {
          char fname[2048];
          vsnprintf(fname, 2047, otherstem, ap);
          return path + std::string(fname) + file;
        }
      else
        return itsStem;
    }
  else  // append frame number to file stem
    {
      int fnum = framenum;
      if (fnum == -1)
        {
          // if we are an OutputFrameSeries, when update() is called,
          // by the time we get a NEXT result, our counter has already
          // been incremented. But we want to save the results for the
          // current frame just as we get this NEXT outcome. So we
          // need to use itsCurrFrame - 1 for the results filename:
          if (tagName().compare(0, 6, "Output") == 0)
            fnum = itsCurrentFrame - 1;
          else           // otherwise, just use current frame:
            fnum = itsCurrentFrame;
        }
      // format frame number as a 6-digit string:
      char num[10]; sprintf(num, "%06d", fnum);
      
      // check that frame number is valid:
      if (fnum < itsFrameRange.getVal().first ||
          fnum > itsFrameRange.getVal().last)
        LFATAL("Frame %d not in range [%d .. %d] -- ABORT",
               fnum, itsFrameRange.getVal().first,
               itsFrameRange.getVal().last);
      
      // do we want to use an alternate stem and args?
      if (otherstem)
        {
          char fname[2048];
          vsnprintf(fname, 2047, otherstem, ap);
          strcat(fname, num);
          return path + std::string(fname);
        }
      else
        return itsStem + std::string(num);
    }
}

// ######################################################################
template <class T>
Image<T> FrameSeries::resizeImage(const Image<T>& input) const
{
  if (itsDims.getVal().isEmpty()) return input;
  if (itsPreserveAspect.getVal())
    {
      Image<T> res(itsDims.getVal(), ZEROS);
      T bg = T(); bg += 64;
      Rectangle r(0, 0, res.getHeight()-1, res.getWidth()-1);
      res.embed(input, r, bg, true);
      return res;
    }
  return rescale(input, itsDims.getVal());
}

// ######################################################################
bool FrameSeries::okToDisplay()
{
  if (manager->testMode() == false && itsShowFrames.getVal()) 
    {
      ++ itsDidDisplay;
      if (itsDidDisplay > MAXDISP)
        LERROR("**** TOO MANY WINDOWS! NOT DISPLAYING IMAGE...");
      else
        return true;
    }
  return false;
  // itsDidDisplay is reset to zero at each update()
}

// ######################################################################
// #################### InputFrameSeries
// ######################################################################

InputFrameSeries::InputFrameSeries(ModelManager& mgr,
                                   const std::string& descrName,
                                   const std::string& tagName) :
  FrameSeries(mgr, "Input", descrName, tagName)
{  }

// ######################################################################
InputFrameSeries::~InputFrameSeries()
{  }

// ######################################################################
void InputFrameSeries::start2()
{
  FrameSeries::start2();

  // get the fully cooked input frame dims (after possible resizing
  // and framing):
  Dims indims = getImageDims();

  // set the FOA radius if someone wants it but it is not set:
  int foaRadius = 0;
  if (manager->getModelParamVal("FOAradius", foaRadius, true, false) &&
      foaRadius == 0) {
    foaRadius = std::min(indims.w(), indims.h()) / 12;
    manager->setModelParamVal("FOAradius", foaRadius, true, true);
  }
  LINFO("Using FOA radius = %d pixels", foaRadius);
  manager->textLog("FOAradius = %d", foaRadius);

  // set the fovea radius if someone wants it but it is not set:
  int foveaRadius = 0;
  if (manager->getModelParamVal("FoveaRadius", foveaRadius, true, false) &&
      foveaRadius == 0) {
    foveaRadius = std::min(indims.w(), indims.h()) / 12;
    manager->setModelParamVal("FoveaRadius", foveaRadius, true, true);
  }
  LINFO("Using fovea radius = %d pixels", foveaRadius);
  manager->textLog("FoveaRadius = %d", foveaRadius);

  // set the input dims if someone wants them but they are not set:
  Dims idim(0, 0);
  if (manager->getModelParamVal("InputFrameDims", idim, true, false) &&
      idim.isEmpty()) {
    manager->setModelParamVal("InputFrameDims", indims, true, true);
  }
  LINFO("Using input dims of (%d, %d) pixels", indims.w(), indims.h());
  manager->textLog("InputDims = %d %d", indims.w(), indims.h());

  // set the initial eye position at center if it is 0,0:
  Point2D iep(0, 0);
  if (manager->getModelParamVal("SCeyeInitialPosition", iep, true, false) &&
      iep.i == 0 && iep.j == 0) {
    manager->setModelParamVal("SCeyeInitialPosition",
                              Point2D(indims.w() / 2, indims.h() / 2),
                                      true, true);
    LINFO("Using initial eye position at (%d, %d)",
          indims.w()/2, indims.h()/2);
    manager->textLog("IEP = %d %d", indims.w()/2, indims.h()/2);
  }

  // set the initial head position at center if it is 0,0:
  Point2D ihp(0, 0);
  if (manager->getModelParamVal("SCheadInitialPosition", ihp, true, false) &&
      ihp.i == 0 && ihp.j == 0) {
    manager->setModelParamVal("SCheadInitialPosition",
                              Point2D(indims.w() / 2, indims.h() / 2),
                                      true, true);
    LINFO("Using initial head position at (%d, %d)",
          indims.w()/2, indims.h()/2);
    manager->textLog("IHP = %d %d", indims.w()/2, indims.h()/2);
  }
}

// ######################################################################
Dims InputFrameSeries::getImageDims(const RasterFileFormat ft,
                                    const int framenum,
                                    const Raster::ColorMode cmode,
                                    const char *otherstem, ...) const
{
  // since we may be called before start(), see whether
  // itsCurrentFrame is available:
  int fn = framenum;
  if (fn == -1)
    {
      if (started()) fn = itsCurrentFrame;
      else fn = itsFrameRange.getVal().first;
    }

  // if we are doing resizing, our dims are the resized dims:
  if (itsDims.getVal().w() != 0 && itsDims.getVal().h() != 0)
    return itsDims.getVal();

  // Otherwise, we'll use the dims of the frames on disk:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(fn, otherstem, a)); va_end(a);
  return Raster::getImageDims(itsFrameRange.getVal().isMultiframe() &&
                              ft == Auto ? itsRasterFileFormat.getVal():ft,
                              cmode, "%s", fname.c_str());
}

// ######################################################################
Dims InputFrameSeries::getImageDims(const RasterFileFormat ft,
                                    const int framenum,
                                    const Raster::ColorMode cmode) const
{ return getImageDims(ft, framenum, cmode, NULL); }

// ######################################################################
Image< PixRGB<byte> >
InputFrameSeries::readRGB(const RasterFileFormat ft, const int framenum,
                          const char *otherstem, ...)
{
  // figure out the file name to use:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(framenum, otherstem, a)); va_end(a);

  // find out file format:
  RasterFileFormat ff = ft;
  if (itsFrameRange.getVal().isMultiframe() && ft == Auto)
    ff = itsRasterFileFormat.getVal();

  // load the image:
  Image< PixRGB<byte> > ima =
    resizeImage(Raster::ReadRGB(ff, "%s", fname.c_str()));
  manager->textLog("ReadRGB = %s", fname.c_str());

  // do we want to show the image?
  if (okToDisplay()) Raster::Display(fname.c_str(), ff, Raster::MODE_RGB);

  return ima;
}

// ######################################################################
Image< PixRGB<byte> >
InputFrameSeries::readRGB(const RasterFileFormat ft, const int framenum)
{ return readRGB(ft, framenum, NULL); }

// ######################################################################
Image<byte>
InputFrameSeries::readGray(const RasterFileFormat ft, const int framenum,
                           const char *otherstem, ...)
{
  // figure out the file name to use:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(framenum, otherstem, a)); va_end(a);

  // find out file format:
  RasterFileFormat ff = ft;
  if (itsFrameRange.getVal().isMultiframe() && ft == Auto)
    ff = itsRasterFileFormat.getVal();

  // load the image:
  Image<byte> ima = resizeImage(Raster::ReadGray(ff,"%s", fname.c_str()));
  manager->textLog("ReadGray = %s", fname.c_str());

  // do we want to show the image?
  if (okToDisplay())
    Raster::Display(fname.c_str(), ff, Raster::MODE_GREYSCALE);

  return ima;
}

// ######################################################################
Image<byte>
InputFrameSeries::readGray(const RasterFileFormat ft, const int framenum)
{ return readGray(ft, framenum, NULL); }



// ######################################################################
// #################### OutputFrameSeries
// ######################################################################
OutputFrameSeries::OutputFrameSeries(ModelManager& mgr,
                                     const std::string& descrName,
                                     const std::string& tagName) :
  FrameSeries(mgr, "Output", descrName, tagName)
{  }

// ######################################################################
OutputFrameSeries::~OutputFrameSeries()
{  }

// ######################################################################
void OutputFrameSeries::writeRGB(const Image< PixRGB<byte> >& image,
                                 const RasterFileFormat ft, const int framenum,
                                 const char *otherstem, ...)
{
  // figure out the file name to use:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(framenum, otherstem, a)); va_end(a);

  // find out file format:
  RasterFileFormat ff = ft;
  if (itsFrameRange.getVal().isMultiframe() && ft == Auto)
    ff = itsRasterFileFormat.getVal();

  // resize the image as appropriate:
  Image< PixRGB<byte> > ima = resizeImage(image);

  // write the image:
  Raster::WriteRGB(ima, ff, "%s", fname.c_str());

  // do we want to show the image?
  if (okToDisplay())
    Raster::Display(fname.c_str(), ff, Raster::MODE_RGB);
}

// ######################################################################
void OutputFrameSeries::writeRGB(const Image< PixRGB<byte> >& image,
                                 const RasterFileFormat ft,
                                 const int framenum)
{ writeRGB(image, ft, framenum, 0); }

// ######################################################################
void OutputFrameSeries::writeGray(const Image<byte>& image,
                                  const RasterFileFormat ft,
                                  const int framenum,
                                  const char *otherstem, ...)
{
  // figure out the file name to use:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(framenum, otherstem, a)); va_end(a);
  
  // find out file format:
  RasterFileFormat ff = ft;
  if (itsFrameRange.getVal().isMultiframe() && ft == Auto)
    ff = itsRasterFileFormat.getVal();

  // resize the image as appropriate:
  Image<byte> ima = resizeImage(image);

  // write the image:
  Raster::WriteGray(ima, ff, "%s", fname.c_str());
  manager->textLog("WriteGray = %s", fname.c_str());

  // do we want to show the image?  do it only if we are not in test mode
  if (okToDisplay())
    Raster::Display(fname.c_str(), ff, Raster::MODE_GREYSCALE);
}

// ######################################################################
void OutputFrameSeries::writeGray(const Image<byte>& image,
                                  const RasterFileFormat ft,
                                  const int framenum)
{ writeGray(image, ft, framenum, 0); }

// ######################################################################
void OutputFrameSeries::writeFloat(const Image<float>& image,
                                   const RasterFileFormat ft,
                                   const int framenum,
                                   const bool normalize,
                                   const char *otherstem, ...)
{
  // figure out the file name to use:
  va_list a; va_start(a, otherstem);
  std::string fname(computeFileName(framenum, otherstem, a)); va_end(a);
  
  // find out file format:
  RasterFileFormat ff = ft;
  if (itsFrameRange.getVal().isMultiframe() && ft == Auto)
    ff = itsRasterFileFormat.getVal();

  // resize the image as appropriate:
  Image<float> ima = resizeImage(image);

  // write the image:
  Raster::WriteFloat(ima, ff, normalize, "%s", fname.c_str());
  manager->textLog("WriteFloat = %s", fname.c_str());
  
  // do we want to show the image?  do it only if we are not in test mode
  if (okToDisplay())
    Raster::Display(fname.c_str(), ff, Raster::MODE_GREYSCALE);
}

// ######################################################################
void OutputFrameSeries::writeFloat(const Image<float>& image,
                                   const RasterFileFormat ft,
                                   const int framenum,
                                   const bool normalize)
{ writeFloat(image, ft, framenum, normalize, 0); }

// Include the explicit instantiations of our template functions
#define INST_CLASS FrameSeries::
// MA #include "inst/FrameSeries.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
