/*!@file FrameSeries.H a series of frames */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: FrameSeries.H 3107 2004-02-10 10:13:24Z itti $
//

#ifndef FRAMESERIES_H_DEFINED
#define FRAMESERIES_H_DEFINED

#include "saliency.H"
#include "ModelComponent.H"
#include "Raster.H"
#include <stdarg.h>
#include "FrameRange.H" // for FrameRange
#include "Point2D.H"
#include "Pixels.H"
#include "Image.H"
#include "ModelOptionDefs.H"

// ######################################################################
//! FrameSeries manages a series of input or output frames
/*! It keeps track of a range of frames ([first..last]) and an
    interframe delay in seconds (see FrameRange.H). Its key behavior
    is the update() function, which takes the current simulation time
    in seconds and the simulation step in s, and returns a status code
    indicating whether the frame series (a) was unchanged, (b) will
    require that the next frame be loaded just before the simulation
    time is advanced by the current simulation step, or (c) is
    finished. InputFrameSeries and OutputFrameSeries derive from
    FrameSeries and implement a number of helper functions to
    facilitate reading/writing frames in a standardized manner, with
    possible on-the-fly rescaling and visualization. NOTE: if our
    ModelManager is in test mode (manager.testMode() returns true,
    typically because --test-mode was given on the command line), we
    will turn off all displays even though they may be on here. */
class FrameSeries : public ModelComponent
{
public:
  /*! @name Constructors, destructors and basic setup */
  //@{

  //! Constructor
  /*! @param prefixName should be "Input" or "Output" and will be
    automatically provided by the constructors of InputFrameSeries and
    OutputFrameSeries */
  FrameSeries(ModelManager& mgr,
              const std::string& prefixName,
              const std::string& descrName = "Frame Series",
              const std::string& tagName = "FrameSeries");

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);
  
  //! Destructor
  virtual ~FrameSeries();

  //! resets this FrameSeries to ots state after construction
  virtual void reset(const bool recurse = true);

  //! Set the file stem - for more info see ModelComponent
  void setFileStem(const std::string& stem);

  //@}

  /*! @name Access functions */
  //@{

  //! Get the file stem
  std::string getFileStem() const;

  //! The possible states returned by update()
  enum State { SAME, NEXT, COMPLETE };

  //! Update internal state and return corresponding FrameSeries::State
  /*! @param stime current time in seconds
    @param sdelay time step in seconds */
  State update(double stime, double sdelay);

  //! Update by moving to the next frame and not worrying about time counts
  State updateNext();

  //! Returns true if the series has multiple frames
  bool isMultiframe() const;

  //! Returns the inter-frame delay in seconds
  double delay() const;

  //! Returns the current frame
  int frame() const;

  //! Returns the first frame in the series
  int minframe() const;
  
  //! Returns the last valid frame in the series
  int maxframe() const;

  //! Returns the number of frames in the series
  int numFrames() const;

  //! True if reached beyond last frame in series
  bool finished() const;

  //! Did we display any frames since last update?
  /*! This is reset to false on each call to update(). It is set to
    true each time a frame is displayed (if itsShowFrames is turned
    on). You can check it in order to decide whether you should pause
    and wait for the user after each series of displays at a given
    time step). */
  bool didDisplayFrames() const;

  //! resize an image to our dims
  template <class T> Image<T> resizeImage(const Image<T>& input) const;
  //@}

protected:
  //! range of frames and inter-frame delay in SECONDS
  ModelParam<FrameRange> itsFrameRange;
  
  //! dims to which frames should be resized on read/write, or 0x0 for none
  ModelParam<Dims> itsDims;

  //! if true, aspect ratio of original image will be preserved when resizing
  ModelParam<bool> itsPreserveAspect;
  
  //! show the frames using xv
  ModelParam<bool> itsShowFrames;
  
  //! default file format; will be used if Auto is specified at read/write
  ModelParam<RasterFileFormat> itsRasterFileFormat;
  
  std::string itsStem;        //!< filename stem
  int itsCurrentFrame;        //!< current frame number
  double itsOldTime;          //!< time of last update()
  int itsDidDisplay;          //!< did we show some frames since last update?

  //! compute filename corresponding to args and return as string
  std::string computeFileName(const int framenum = -1,
                              const char *otherstem = 0, va_list ap = 0) const;

  //! have we displayed too many windows, or can we still display more?
  bool okToDisplay();

  //! post-commandline-parsing initialization (see ModelComponent.H)
  virtual void start2();
};

// ######################################################################
//! A FrameSeries with additional facilities for reading input frames
/*! NOTE: when start() is called, we will set the "FOAradius",
    "FoveaRadius", "InputFrameDims", "SCinitialEyePosition" and
    "SCinitialHeadPosition" model parameters from input image dims, if
    someone wants them and they have not been set. */
class InputFrameSeries : public FrameSeries
{
public:
  //! Constructor
  InputFrameSeries(ModelManager& mgr,
                   const std::string& descrName = "Frame Series",
                   const std::string& tagName = "FrameSeries");

  //! Destructor
  virtual ~InputFrameSeries();

  //! Get dims of a frame on disk, without fully reading it
  /*! It is okay to call this before the model is started().  If input
    resizing is being done, this will return the resized dims.  All in
    all, this will return the size of whatever you will get when
    calling readRGB() or readGray(). */
  Dims getImageDims(const RasterFileFormat ft, const int framenum,
                    const Raster::ColorMode cmode,
                    const char *otherstem, ...) const;

  //! Get dims of a frame on disk, without fully reading it
  /*! It is okay to call this before the model is started(). If input
    resizing is being done, this will return the resized dims. All in
    all, this will return the size of whatever you will get when
    calling readRGB() or readGray(). */
  Dims getImageDims(const RasterFileFormat ft = Auto, const int framenum = -1,
                    const Raster::ColorMode cmode = Raster::MODE_RGB) const;

  //! Read a frame from disk
  /*! @param ft RasterFileFormat as supported by Raster 
      @param framenum frame number, or use -1 for the current frame.
      @param otherstem prefix stem that may contain printf-like fields
             and will be used with any additional arguments. If
             multiframe, the filename will be of the form
             <cooked_otherstem><framenum>, otherwise it will be
             <cooked_otherstem><defaultstem> */
  Image< PixRGB<byte> > readRGB(const RasterFileFormat ft, const int framenum,
                                const char *otherstem, ...);

  //! Read a frame from disk, using default filename stem
  /*! The default is to read using the stem set by setFileStem(), and
    the current frame number (if multiframe) */
  Image< PixRGB<byte> > readRGB(const RasterFileFormat ft = Auto,
                                const int framenum = -1);

  //! Read a greyscale frame
  Image<byte> readGray(const RasterFileFormat ft, const int framenum,
                       const char *otherstem, ...);

  //! Read a greyscale frame
  Image<byte> readGray(const RasterFileFormat ft = Auto,
                       const int framenum = -1);

protected:
  //! post-commandline-parsing initialization (see ModelComponent.H)
  /*! Will peek the first frame (without fully loading it) and set
    "FOAradius", "FoveaRadius", "InputFrameDims"
    "SCinitialEyePosition" and "SCinitialHeadPosition" model
    parameters from input image dims if someone wants them and they
    have not been set */
  virtual void start2();
};

// ######################################################################
//! A FrameSeries with additional facilities for writing output frames
class OutputFrameSeries : public FrameSeries
{
public:
  //! Constructor
  OutputFrameSeries(ModelManager& mgr,
                   const std::string& descrName = "Frame Series",
                    const std::string& tagName = "FrameSeries");

  //! Destructor
  virtual ~OutputFrameSeries();

  //! Write a frame to disk
  void writeRGB(const Image< PixRGB<byte> >& image, const RasterFileFormat ft,
                const int framenum, const char *otherstem, ...);

  //! Write a frame to disk
  void writeRGB(const Image< PixRGB<byte> >& image,
                const RasterFileFormat ft = Auto,
                const int framenum = -1);

  //! Write a frame to disk
  void writeGray(const Image<byte>& image, const RasterFileFormat ft,
                 const int framenum, const char *otherstem, ...);

  //! Write a frame to disk
  void writeGray(const Image<byte>& image, const RasterFileFormat ft = Auto,
                 const int framenum = -1);

  //! Write a frame to disk
  void writeFloat(const Image<float>& image, const RasterFileFormat ft,
                  const int framenum, const bool normalize,
                  const char *otherstem, ...);

  //! Write a frame to disk
  void writeFloat(const Image<float>& image, const RasterFileFormat ft = Auto,
                  const int framenum = -1, const bool normalize = true);
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
