/*!@file GistEstimator.C extract estimated gist
         using available features of the image                          */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Christian Siagian <siagian@usc.edu>
// $Id: GistEstimator.C 4198 2005-02-15 02:35:14Z siagian $
//

// ######################################################################
/*! Extract gist of image                                               */

#include "GistEstimator.H"
#include "gistParams.H"

#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "Raster.H"
#include "XWindow.H"
#include "XWinManaged.H"

#include "saliency.H"
#include "ChannelBase.H"
#include "SingleChannel.H"
#include "ComplexChannel.H"
#include "Channels.H"
#include "VisualCortex.H"
#include "StdBrain.H"
#include "Jet.H"

#include "Image.H"
#include "Pyramid_Ops.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "Image_FilterOps.H"
#include "Image_ColorOps.H"
#include "Image_Transforms.H"
#include "ImageSet.H"
#include "PyrBuilder.H"
#include "Point2D.H"
#include "Dims.H"

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <stdio.h>
#include <iostream>
#include <math.h>

// ######################################################################
GistEstimator::GistEstimator(ModelManager& mgr, const std::string& descrName,
                               const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName)
{
  tempV = new double[NUM_GIST_FEAT];
  gistV = new double[NUM_GIST_FEAT_SIZE];
}

// ######################################################################
GistEstimator::~GistEstimator()
{
  delete [] tempV;
  delete [] gistV;
}

// ######################################################################
void GistEstimator::exportOptions(const int whatToExport, const bool recurse)
{
}

// ######################################################################
void GistEstimator::setVC(const SharedPtr<VisualCortex>& inVC)
{ 
  vc = inVC;
}

// ######################################################################
// void GistEstimator::start1()
// {
// }

// ######################################################################
void GistEstimator::compute()
{
  getFeatureVector();
}

// ######################################################################
void GistEstimator::getGist(double* retV)
{
  for(int i = 0; i < NUM_GIST_FEAT_SIZE; i++)
    retV[i] = gistV[i];
}

// ######################################################################
void GistEstimator::getFeatureVector()
{
  int drawImages = 0;
  int drawIntens = 0;
  int printRes   = 1;
  int count;

//   printf("want to see the individual features?\n"); 
//   char spC = getchar();getchar();
//   if(spC == 'y')
//     {
//       drawImages = 1;
//     }
 

  // get the appropriate sub-channels and features

  // get the orientation channel
  SharedPtr<OrientationChannel> oc; oc.dynCastFrom(vc->channelPtr("orientation"));
  uint nc_oc = 4;
  GaborChannel& oc_gc000 = oc->gabor(0);
  GaborChannel& oc_gc045 = oc->gabor(1);
  GaborChannel& oc_gc090 = oc->gabor(2);
  GaborChannel& oc_gc135 = oc->gabor(3);

  // get the color channel
  SharedPtr<ColorChannel> cc; cc.dynCastFrom(vc->channelPtr("color"));
  RedGreenChannel& cc_rg = cc->rg();
  BlueYellowChannel& cc_by = cc->by();

  // get the intensity channel
  SharedPtr<IntensityChannel> ic; ic.dynCastFrom(vc->channelPtr("intensity"));

  // show the images from which we extracted the image
  XWinManaged *fWin;
  Image<float> tImg = oc_gc000.getImage(0);
  int w = tImg.getWidth();  
  int h = tImg.getHeight();
  if(drawImages || drawIntens)
    {    
      fWin = new XWinManaged(w, h, 0, 0, "Image"); //wList.add(*fWin);
    } 

  int scale = 1; float min,max; count = 0;
  // orientation channel 0 degree
  for(int i = 0; i < NUM_GIST_LEV; i++)
    {      
      if(printRes)
	printf("Sum oc000_%d: \n",i);  
      getSubSum(tempV, oc_gc000.getImage(i),printRes);
      memcpy(&gistV[i*nc_oc * NUM_GIST_FEAT], 
	     tempV, NUM_GIST_FEAT*sizeof(double));

      if(drawImages)
	{
	  printf("oc_gc000.getImage(%d)\n",i);
	  tImg = zoomXY(oc_gc000.getImage(i),scale,-1);
	  tImg.getMinMax(min,max); //max = 200;
	  tImg.drawGrid(w/4,h/4,1,1,max);
	  fWin->drawImage(tImg, 0, 0);
	  Raster::waitForKey();
	}
      scale *= 2;
    }

  scale = 1;
  // orientation channel 45 degree
  for(int i = 0; i < NUM_GIST_LEV; i++)
    {
      if(printRes)
	printf("Sum oc045_%d: \n",i);  
      getSubSum(tempV, oc_gc045.getImage(i),printRes);
      memcpy(&gistV[(i*nc_oc + 1)*NUM_GIST_FEAT], 
	     tempV, NUM_GIST_FEAT*sizeof(double));

      if(drawImages)
	{
	  printf("oc_gc045.getImage(%d)\n",i);
	  tImg = zoomXY(oc_gc045.getImage(i),scale,-1);
	  tImg.getMinMax(min,max); //max = 200;
	  tImg.drawGrid(w/4,h/4,1,1,max);
	  fWin->drawImage(tImg, 0, 0);
	  Raster::waitForKey();
	}
      scale *= 2;
    }

  scale = 1;
  // orientation channel 90 degree
  for(int i = 0; i < NUM_GIST_LEV; i++)
    {
      if(printRes)
	printf("Sum oc090_%d: \n",i);
      getSubSum(tempV, oc_gc090.getImage(i),printRes);
      memcpy(&gistV[(i*nc_oc + 2)*NUM_GIST_FEAT], 
	     tempV, NUM_GIST_FEAT*sizeof(double));

      if(drawImages)
	{
	  printf("oc_gc090.getImage(%d)\n",i);
	  tImg = zoomXY(oc_gc090.getImage(i),scale,-1);
	  tImg.getMinMax(min,max); //max = 200;
	  tImg.drawGrid(w/4,h/4,1,1,max);
	  fWin->drawImage(tImg, 0, 0);
	  Raster::waitForKey();
	}
      scale *= 2;
    }

  scale = 1;
  // orientation channel 135 degree
  for(int i = 0; i < NUM_GIST_LEV; i++)
    {
      if(printRes)
	printf("Sum oc135_%d: \n",i);  
      getSubSum(tempV, oc_gc135.getImage(i),printRes);
      memcpy(&gistV[(i*nc_oc + 3)*NUM_GIST_FEAT], 
	     tempV, NUM_GIST_FEAT*sizeof(double));

      if(drawImages)
	{
	  printf("oc_gc135.getImage(%d)\n",i);
	  tImg = zoomXY(oc_gc135.getImage(i),scale,-1);
	  tImg.getMinMax(min,max); //max = 200;
	  tImg.drawGrid(w/4,h/4,1,1,max);
	  fWin->drawImage(tImg, 0, 0);
	  Raster::waitForKey();
	}
      scale *= 2;
    }

  // color channel RG color opponency
  scale = 4; count = 16;
  for(int i = NUM_GIST_CTR_S; i < NUM_GIST_CTR_S+NUM_GIST_CTR; i++)
    {
      for(int j = NUM_GIST_SURR_DELTA_S; 
	  j < NUM_GIST_SURR_DELTA_S + NUM_GIST_SURR_DELTA; j++)
	{
	  if(printRes)
	    printf("Sum cc_rg_%d_%d: \n",i,i+j);
	  getSubSum(tempV, cc_rg.centerSurround(i,i+j),printRes);
	  //printf("dd: %d * 21 = %d\n", count, count * NUM_GIST_FEAT);
 	  memcpy(&gistV[count * NUM_GIST_FEAT], 
		 tempV, NUM_GIST_FEAT*sizeof(double));
	  count++;
	
	  if(drawImages)
	    {
	      printf("cc_rg.centerSurround(%d,%d)\n",i,i+j);
	      tImg = zoomXY(cc_rg.centerSurround(i,i+j), scale, -1);
	      tImg.getMinMax(min,max);
	      tImg.drawGrid(w/4,h/4,1,1,max);
	      fWin->drawImage(tImg, 0, 0);
	      Raster::waitForKey();
	    }
	}
      scale *= 2;
    }

  scale = 4;
  // color channel BY color opponency
  for(int i = NUM_GIST_CTR_S; i < NUM_GIST_CTR_S+NUM_GIST_CTR; i++)
    {
      for(int j = NUM_GIST_SURR_DELTA_S; 
	  j < NUM_GIST_SURR_DELTA_S + NUM_GIST_SURR_DELTA; j++)
	{
	  if(printRes)
	    printf("Sum cc_by_%d_%d: \n",i,i+j);  
	  getSubSum(tempV, cc_by.centerSurround(i,i+j),printRes);
	  //printf("dd: %d * 21 =%d\n", count, count * NUM_GIST_FEAT);
	  memcpy(&gistV[count * NUM_GIST_FEAT], 
		 tempV, NUM_GIST_FEAT*sizeof(double));
	  count++;

	  if(drawImages)
	    {
	      printf("cc_by.centerSurround(%d,%d)\n",i,i+j);
	      tImg = zoomXY(cc_by.centerSurround(i,i+j), scale, -1);
	      tImg.getMinMax(min,max);
	      tImg.drawGrid(w/4,h/4,1,1,max);
	      fWin->drawImage(tImg, 0, 0);
	      Raster::waitForKey();
	    }
	}
      scale *= 2;
    }
      
  // intensity channel
  scale = 4;
  for(int i = NUM_GIST_CTR_S; i < NUM_GIST_CTR_S+NUM_GIST_CTR; i++)
    {
      for(int j = NUM_GIST_SURR_DELTA_S; 
	  j < NUM_GIST_SURR_DELTA_S + NUM_GIST_SURR_DELTA; j++)
	{
	  if(printRes)
	    printf("Sum ic_%d_%d: \n",i,i+j);  
	  getSubSum(tempV, ic->centerSurround(i,i+j),printRes);
	  memcpy(&gistV[count * NUM_GIST_FEAT], 
		 tempV, NUM_GIST_FEAT*sizeof(double));
	  count++;

	  if(drawImages || drawIntens)
	    {
	      printf("ic->centerSurround(%d,%d)\n",i,i+j);
	      tImg = zoomXY(ic->centerSurround(i,i+j), scale, -1);
	      tImg.getMinMax(min,max);
	      tImg.drawGrid(w/4,h/4,1,1,max);
	      fWin->drawImage(tImg, 0, 0);
	      Raster::waitForKey();
	    }
	}
      scale *= 2;
    }
}

// ######################################################################
// get subsum of an image
void GistEstimator::getSubSum(double* res, Image<float> img, int printRes)
{
  int useMean = 1; 
  // first level
  res[0] = mean(img); //sum(img);
  if(printRes)
    printf("lev1   : %14.4f\n", res[0]);  

  // second level
  // 1 2 
  // 3 4
  int w = img.getWidth();  int h = img.getHeight();
  Image<float> img1   = copy(img, Rectangle(0    , 0    , h/2,  w/2));
  Image<float> img2   = copy(img, Rectangle(0    , w/2+1, h/2,  w-1));
  Image<float> img3   = copy(img, Rectangle(h/2+1, 0    , h-1,  w/2));
  Image<float> img4   = copy(img, Rectangle(h/2+1, w/2+1, h-1,  w-1));
  if(useMean)
    {
      res[1] = mean(img1); //sum(img1);
      res[2] = mean(img2); //sum(img2);
      res[3] = mean(img3); //sum(img3);
      res[4] = mean(img4); //sum(img4);
    }
  else
    {
      res[1] = img1.getStdev();//mean(img1); //sum(img1);
      res[2] = img2.getStdev();//mean(img2); //sum(img2);
      res[3] = img3.getStdev();//mean(img3); //sum(img3);
      res[4] = img4.getStdev();//mean(img4); //sum(img4);
    }
  if(printRes)
    printf("lev2   : %14.4f, %14.4f, %14.4f, %14.4f\n", 
	   res[1], res[2], res[3], res[4]);

  // third level
  // 1 2 
  // 3 4
  int wX = img1.getWidth();  int hX = img1.getHeight();
  Image<float> imgX_1 = copy(img1, Rectangle(0     , 0     , hX/2,  wX/2));
  Image<float> imgX_2 = copy(img1, Rectangle(0     , wX/2+1, hX/2,  wX-1));
  Image<float> imgX_3 = copy(img1, Rectangle(hX/2+1, 0     , hX-1,  wX/2));
  Image<float> imgX_4 = copy(img1, Rectangle(hX/2+1, wX/2+1, hX-1,  wX-1));
  if(useMean)
    {
      res[5] = mean(imgX_1); //sum(imgX_1);
      res[6] = mean(imgX_2); //sum(imgX_2);
      res[7] = mean(imgX_3); //sum(imgX_3);
      res[8] = mean(imgX_4); //sum(imgX_4);
    }
  else
    {
      res[5] = imgX_1.getStdev();//mean(imgX_1); //sum(imgX_1);
      res[6] = imgX_2.getStdev();//mean(imgX_2); //sum(imgX_2);
      res[7] = imgX_3.getStdev();//mean(imgX_3); //sum(imgX_3);
      res[8] = imgX_4.getStdev();//mean(imgX_4); //sum(imgX_4);
    }
  if(printRes)
    printf("lev3   : %14.4f, %14.4f, %14.4f, %14.4f\n", 
	   res[5], res[6], res[7], res[8]);

  wX = img2.getWidth();  hX = img2.getHeight();
  imgX_1 = copy(img2, Rectangle(0     , 0     , hX/2,  wX/2));
  imgX_2 = copy(img2, Rectangle(0     , wX/2+1, hX/2,  wX-1));
  imgX_3 = copy(img2, Rectangle(hX/2+1, 0     , hX-1,  wX/2));
  imgX_4 = copy(img2, Rectangle(hX/2+1, wX/2+1, hX-1,  wX-1));
  if(useMean)
    {
      res[ 9] = mean(imgX_1); //sum(imgX_1);
      res[10] = mean(imgX_2); //sum(imgX_2);
      res[11] = mean(imgX_3); //sum(imgX_3);
      res[12] = mean(imgX_4); //sum(imgX_4);
    }
  else
    {
      res[ 9] = imgX_1.getStdev();//mean(imgX_1); //sum(imgX_1);
      res[10] = imgX_2.getStdev();//mean(imgX_2); //sum(imgX_2);
      res[11] = imgX_3.getStdev();//mean(imgX_3); //sum(imgX_3);
      res[12] = imgX_4.getStdev();//mean(imgX_4); //sum(imgX_4);
    }
  if(printRes)
    printf("       : %14.4f, %14.4f, %14.4f, %14.4f\n", 
	   res[9], res[10], res[11], res[12]);

  wX = img3.getWidth();  hX = img3.getHeight();
  imgX_1 = copy(img3, Rectangle(0     , 0     , hX/2,  wX/2));
  imgX_2 = copy(img3, Rectangle(0     , wX/2+1, hX/2,  wX-1));
  imgX_3 = copy(img3, Rectangle(hX/2+1, 0     , hX-1,  wX/2));
  imgX_4 = copy(img3, Rectangle(hX/2+1, wX/2+1, hX-1,  wX-1));
  if(useMean)
    {
      res[13] = mean(imgX_1); //sum(imgX_1);
      res[14] = mean(imgX_2); //sum(imgX_2);
      res[15] = mean(imgX_3); //sum(imgX_3);
      res[16] = mean(imgX_4); //sum(imgX_4);
    }
  else
    {
      res[13] = imgX_1.getStdev();//mean(imgX_1); //sum(imgX_1);
      res[14] = imgX_2.getStdev();//mean(imgX_2); //sum(imgX_2);
      res[15] = imgX_3.getStdev();//mean(imgX_3); //sum(imgX_3);
      res[16] = imgX_4.getStdev();//mean(imgX_4); //sum(imgX_4);
    }
  if(printRes)
    printf("       : %14.4f, %14.4f, %14.4f, %14.4f\n", 
	   res[13], res[14], res[15], res[16]);

  wX = img4.getWidth();  hX = img4.getHeight();
  imgX_1 = copy(img4, Rectangle(0     , 0     , hX/2,  wX/2));
  imgX_2 = copy(img4, Rectangle(0     , wX/2+1, hX/2,  wX-1));
  imgX_3 = copy(img4, Rectangle(hX/2+1, 0     , hX-1,  wX/2));
  imgX_4 = copy(img4, Rectangle(hX/2+1, wX/2+1, hX-1,  wX-1));
  if(useMean)
    {
      res[17] = mean(imgX_1); //sum(imgX_1);
      res[18] = mean(imgX_2); //sum(imgX_2);
      res[19] = mean(imgX_3); //sum(imgX_3);
      res[20] = mean(imgX_4); //sum(imgX_4);
    }
  else
    {
      res[17] = imgX_1.getStdev();//mean(imgX_1); //sum(imgX_1);
      res[18] = imgX_2.getStdev();//mean(imgX_2); //sum(imgX_2);
      res[19] = imgX_3.getStdev();//mean(imgX_3); //sum(imgX_3);
      res[20] = imgX_4.getStdev();//mean(imgX_4); //sum(imgX_4);
    }
  if(printRes)
    printf("       : %14.4f, %14.4f, %14.4f, %14.4f\n",
	   res[17], res[18], res[19], res[20]);
}

// ######################################################################
// get histogram to visualize the data
Image<float> GistEstimator::getHistogram(double* arr, float max, int sqSize)
{
  // square size
  int s = sqSize;
  Image<float> img(NUM_GIST_COL * s, NUM_GIST_FEAT * s, ZEROS);
  for(int i = 0; i < NUM_GIST_COL; i++)
  //for(int i = 0; i < 16; i++)
    for(int j = 0; j < NUM_GIST_FEAT; j++)
      {
	float val = arr[i*NUM_GIST_FEAT+j];
	//float val = log(arr[i*NUM_GIST_FEAT+j]+1);
	//val = val * val;
	img.drawPatch(Point2D(i*s+s/2,j*s+s/2),s/2,val);
        //printf("val[%d]: %f\n",j,val); 
      }
  float min;
  if(max == -1.0)
    {
      img.getMinMax(min,max);
      printf("min: %f, max: %f\n",min,max);
    }
  // draw the delineation
  img.drawLine(Point2D(0,  1*s), Point2D(NUM_GIST_COL*s,  1*s), max, 2);
  img.drawLine(Point2D(0,  5*s), Point2D(NUM_GIST_COL*s,  5*s), max, 2);
  img.drawLine(Point2D(0,  9*s), Point2D(NUM_GIST_COL*s,  9*s), max, 1);
  img.drawLine(Point2D(0, 13*s), Point2D(NUM_GIST_COL*s, 13*s), max, 1);
  img.drawLine(Point2D(0, 17*s), Point2D(NUM_GIST_COL*s, 17*s), max, 1);

  img.drawLine(Point2D( 4*s, 0), Point2D( 4*s, NUM_GIST_FEAT*s), max, 1);
  img.drawLine(Point2D( 8*s, 0), Point2D( 8*s, NUM_GIST_FEAT*s), max, 1);
  img.drawLine(Point2D(12*s, 0), Point2D(12*s, NUM_GIST_FEAT*s), max, 1);
  img.drawLine(Point2D(16*s, 0), Point2D(16*s, NUM_GIST_FEAT*s), max, 1);
  img.drawLine(Point2D(22*s, 0), Point2D(22*s, NUM_GIST_FEAT*s), max, 1);
  img.drawLine(Point2D(28*s, 0), Point2D(28*s, NUM_GIST_FEAT*s), max, 1);

  return img;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
