/*!@file Image.H An image template class with many image processing
  functions */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Image.H 4220 2005-02-18 04:24:50Z mundhenk $
//

#ifndef IMAGE_H_DEFINED
#define IMAGE_H_DEFINED

// Uncomment this to allow range checking in Image iterators:
//#define MEM_DEBUG

#include "saliency.H"
#include "Promotions.H"
#include "ArrayData.H"
#include "Point2D.H"
#include "Rectangle.H"
#include "Pixels.H"
#include <vector>

#ifdef MEM_DEBUG
#include "CheckedIterator.H"
#endif

//! Generic image template class
/*! This is a generic image template class. In principle, it can handle
    grayscale as well as color or multispectral images.  Not all methods
    will work with both types, though (e.g., some methods need luminance()
    to be defined in the base type [will not work on canonical scalar
    types], and some other methods perform arithmetic comparisons on the
    base type [will not work with color pixels]).

    Note also that many other image manipulation functions are defined as
    non-member functions in Image_ColorOps.H, Image_FilterOps.H,
    Image_IO.H, Image_MathOps.H, Image_Omni.H, Image_ShapeOps.H, and
    Image_Transforms.H.
*/

template <class T>
class Image
{
public:

  // ############################################################
  /*! @name Constructors, destructors, assignment */
  //@{

  //! Construct from C array
  /*! Build from C array; an internal copy of the C array will be
      allocated, so the C array can (and should) be freed without affecting
      the Image. */
  inline Image(const T* inarray, int width, int height);

  //! Construct from C array
  /*! Build from C array; an internal copy of the C array will be
      allocated, so the C array can (and should) be freed without affecting
      the Image. */
  inline Image(const T* inarray, const Dims& dims);

  //! Allocates memory for given size, and optionally zero-clear that memory.
  inline Image(int width, int height, InitPolicy init);

  //! Constructor that only allocates memory for given size
  inline explicit Image(const Dims& dims, InitPolicy init);

  //! Construct an empty (0-by-0) image (useful for arrays of Images).
  inline Image();

  //! Copy constructor
  /*! e.g.:
      \code
      Image<byte> im(other);
      // or
      Image<byte> im = other; // with other also of type Image<byte>
      \endcode
  */
  inline Image(const Image<T>& A);

  //! Conversion copy constructor
  /*! e.g.:
      \code
      Image<byte> im(other);
      // or
      Image<byte> im = other; // with other of type Image<float>
      \endcode
  */
  template <class T2> inline Image(const Image<T2>& A);

  //! Assigment operator.
  /*! e.g.:
      \code
      Image<byte> im1, im2; im2 = im1;
      \endcode
  */
  inline Image<T>& operator=(const Image<T>& A);

  //! Conversion assigment operator.
  /*! e.g.:
      \code
      Image<byte> im1; std::vector im2; im1 = im2;
      \endcode

  */


  inline Image<T>& operator=(std::vector<std::vector<T> >& A);
  inline Image<T>& operator=(std::vector<std::vector<T>* >& A);
  inline Image<T>& operator=(std::vector<std::vector<T*> >& A);
  //! Conversion assigment operator.
  /*! e.g.:
      \code
      Image<byte> im1; Image<float> im2; im2 = im1;
      \endcode

  */
  template <class T2> inline Image<T>& operator=(const Image<T2>& A);

  //! Destructor
  inline ~Image();

  //! Free memory and switch to uninitialized state.
  /*! Note that it is \b NOT necessary to call this function to ensure
      proper cleanup, that will be done in the destructor by
      default. Rather, freeMem() is offered just as a performance
      optimization, to allow you to release a potentially large chunk of
      memory when you are finished using it. */
  inline void freeMem();

  //@}

  // ############################################################
  /*! @name Memory management functions */
  //@{

  //! Swap the contents of two images
  inline void swap(Image<T>& other);

  //! Use existing memory.
  /*! This is potentially dangerous and should really be avoided. The only
      case where this really is useful is to attach an image to an existing
      memory segment that is shared or into which data is streaming via
      DMA. */
  inline void attach(T* array, const int width, const int height);

  //! Detach previously attach()'ed image.
  /*! The main purpose of detach() is to make sure that this Image object
      does not continue to point at attach()'ed memory after that memory
      has been freed. Nevertheless, this function is not strictly necessary
      to ensure correct memory handling, since the Image destructor will
      only try to free memory if that memory is owned (i.e., not
      attach()'ed). All that it does is release any association with a
      previously attach()'ed memory block, by setting the current image to
      a new empty (zero-by-zero) ArrayData object. We assume that attached
      memory will be destroyed later. */
  inline void detach();

  //! Free mem and realloc new array (array contents are lost).
  /*! Use rescale() instead if you want to preserve image contents. */
  inline void resize(const Dims& dims, const bool clear = false);

  //! Free mem and realloc new array (array contents are lost).
  /*! Use rescale() instead if you want to preserve image contents. */
  inline void resize(const int width, const int height,
                     const bool clear = false);

  //@}

  // ############################################################
  /*! @name Iterators

      There are const and non-const versions of iterators, which are
      returned by begin()/end() and beginw()/endw(), respectively. The "w"
      in beginw()/endw() is a mnemonic for "write" or "writeable". Beware
      that the non-const versions of these functions will invoke the
      copy-on-write mechanism (see ArrayHandle), so a deep copy of all the
      image data will be made if the current image object is not the unique
      owner of its image data. Thus, begin()/end() should always be used
      unless write access is specifically needed. (It is for this reason
      that we don't overload the const/non-const functions with the same
      names (even though this is possible in C++, as done in the STL
      containers)).

      Debugging iterators (CheckedIterator) will be used if the macro
      MEM_DEBUG is #define'd. A checked iterator will check that it is
      within its proper bounds every time that it is dereferenced with
      either operator*() or operator->(). Note that the existence of the
      checked iterators means you cannot rely on an image iterator being a
      raw pointer. Instead, getArrayPtr() is available in case you
      unconditionally need a raw pointer to the image data.
   */
  //@{

#ifndef MEM_DEBUG
  // standard iterators are just pointers to the data:
  //! Read/write iterator.
  typedef T* iterator;
  //! Read-only iterator.
  typedef const T* const_iterator;
#else
  // range-checked iterators can detect attempts to access past image bounds:
  //! Read/write iterator.
  typedef CheckedIterator<T> iterator;
  //! Read-only iterator.
  typedef CheckedIterator<const T> const_iterator;
#endif

  //! Returns a read-only iterator to the beginning of the image data
  inline const_iterator begin() const;

  //! Returns a read-only iterator to one-past-the-end of the image data
  inline const_iterator end() const;

  //! Returns a read-write iterator to the beginning of the image data
  inline iterator beginw();

  //! Returns a read-write iterator to one-past-the-end of the image data
  inline iterator endw();

  //@}

  // ############################################################
  /*! @name Access functions*/
  //@{

  //! Check whether image is non-empty (i.e., non-zero height and width).
  inline bool initialized() const;

  //! Get image size (width * height)
  inline int getSize() const;

  //! Get image width
  inline int getWidth() const;

  //! Get image height
  inline int getHeight() const;

  //! Get image width+height in Dims struct
  inline const Dims& getDims() const;

  //! Check if *this is the same size as the other thing
  /*! The other thing can be any type that exposes getHeight() and
    getWidth() */
  template <class C>
  inline bool isSameSize(const C& other) const;

  //! Check if the image is 1D, i.e., width == 1 or height == 1
  inline bool is1D() const;

  //! Check wether all pixels have finite values
  /*! This relies on isFinite() being defined for your pixel type
    T. In saliency.H we define it for the canonical types, and you can
    use these canonical definitions to define it for more complex
    types (see, for example, the definition in Pixels.H). */
  bool isFinite() const;

  //! Get min and max values
  inline void getMinMax(T& mini, T& maxi) const;

  //! Get min, max and average values
  void getMinMaxAvg(T& mini, T& maxi, T& avg) const;

  //! find point of max activity and also what that max activity is
  void findMax(Point2D& p, T& maxval) const;

  //! find point of min activity and also what that min activity is
  void findMin(Point2D& p, T& minval) const;

  //! compute mean pixel value over image (promoted to double)
  double getMean() const;

  //! compute standard deviation of pixel values over image
  double getStdev() const;

  //! Access image elements through C array index interface
  inline T& operator[](const int index);

  //! Access image elements through C array index interface
  inline const T& operator[](const int index) const;

  //! Get pixel value at index in image
  inline const T& getVal(const int index) const;

  //! Get pixel value at (x, y) in image
  inline const T& getVal(const int x, const int y) const;

  //! Get pixel value at specified coordinates in image
  inline const T& getVal(const Point2D& p) const;

  //! Get value at (x,y), put in val rather that return it
  template <class T2> inline void getVal(const int x, const int y,
                                         T2& val) const;

  //! Get pixel value at (x, y) in image with bilinear interpolation
  inline const T getValInterp(const float x, const float y) const;

  //! Get a sample which is the max value within a circular aperture
  /*! This only works for monochromatic images.
      @param center coordinates of the center of the aperture
      @param radius radius of the aperture, in pixels */
  T getSample(const Point2D& center, const int radius);

  //! Return true if point p is a local maximum
  /*! We just check that the value at p is >= the values of its 4 neighbors */
  bool isLocalMax(const Point2D& p) const;

  //! Set value in image at index
  template <class T2> inline void setVal(const int index, const T2& value);

  //! Set value in image at Point2D
  template <class T2> inline void setVal(const Point2D& p, const T2& value);

  //! Set value in image at (x, y)
  template <class T2> inline void setVal(const int x, const int y,
                                         const T2& value);

  //! Set values at all locations where mask is non-zero
  void setValMask(const Image<byte>& mask, const T val);

  //! Replace all occurences of a value 'from' by another value 'to'
  void replaceVal(const T& from, const T& to);

  //! Returns read-only (const) pointer to internal image array
  inline const T* getArrayPtr() const;

  //! Returns read/write (non-const) pointer to internal image array
  inline T* getArrayPtr();

  //! Test whether point falls inside array boundaries
  inline bool coordsOk(const Point2D& P) const;

  //! Test whether point falls inside array boundaries
  inline bool coordsOk(const int i, const int j) const;

  //! Test whether rectangle fits in image
  inline bool rectangleOk(const Rectangle& rect) const;

  //@}

  // ############################################################
  /*! @name Operator overloads and basic image manipulations

      Note that many other image manipulation functions are available in
      Image_ColorOps.H, Image_FilterOps.H, Image_IO.H, Image_MathOps.H,
      Image_Omni.H, Image_ShapeOps.H, and Image_Transforms.H.
  */
  //@{

  //! Add constant to image, clamp result as necessary
  Image<T>& operator+=(const T& val);

  //! Subtract constant from image, clamp result as necessary
  Image<T>& operator-=(const T& val);

  //! Multiply image by constant, clamp result as necessary
  Image<T>& operator*=(const T& val);

  //! Divide image by constant, clamp result as necessary
  Image<T>& operator/=(const T& val);


  //! Add image to image, clamp result as necessary
  template <class T2> inline
  Image<T>& operator+=(const Image<T2>& A);

  //! Subtract image from image, clamp result as necessary
  template <class T2> inline
  Image<T>& operator-=(const Image<T2>& A);

  //! Multiply image by image, point-by-point, clamp result as necessary
  template <class T2> inline
  Image<T>& operator*=(const Image<T2>& A);

  //! Divide image by image, point-by-point, clamp result as necessary
  template <class T2> inline
  Image<T>& operator/=(const Image<T2>& A);


  //! Add scalar to each point in *this and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator+(const T2& val) const;

  //! Subtract scalar from each point in *this and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator-(const T2& val) const;

  //! Multiply scalar each point in *this by scalar and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator*(const T2& val) const;

  //! Divide each point in *this by scalar and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator/(const T2& val) const;


  //! Point-wise add img to image and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator+(const
                                                    Image<T2>& img) const;

  //! Point-wise subtract img from *this and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator-(const
                                                    Image<T2>& img) const;

  //! Point-wise multiply *this by img and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator*(const
                                                    Image<T2>& img) const;

  //! Point-wise divide *this by img and return result
  template <class T2> inline
  Image<typename promote_trait<T,T2>::TP> operator/(const
                                                    Image<T2>& img) const;


  //! Scale b to my size and add it to me (using bilinear interpolation)
  void addScaled(const Image<T>& b);

  //! Use pyramid to downsize b and add to me
  void addDownsized(const Image<T>& b);

  //! this += b * coeff
  void addWeighted(const Image<T>& b, const float coeff);

  //! Take square of all elements
  void square();

  //! Normalize values between nmin and nmax
  void normalize(const T nmin, const T nmax);

  //! Normalize image by dividing by total activity
  void activityNormalize(const T nmax = 255.0);

  //! Normalize between mi and ma and multiply by variance
  void varianceNormalize(const T mi, const T ma);

  //! Normalize my values by those in the lum image
  void luminanceNormalize(const Image<T>& lum, const T thresh);

  //! Saturate values < cmin to cmin and > cmax to cmax
  inline void clamp(const T cmin, const T cmax);

  //! Saturate values < 0
  void rectify();

  //! Put all values >= 0 into pos and the negated of all vals <= 0 into neg
  void splitPosNeg(Image<T>& pos, Image<T>& neg);

  //! Cut values < thresh and replace them by val
  void lowThresh(const T thresh, const T val = T());

  //! Cut values whose abs is < thresh and replace them by val
  void lowThreshAbs(const T thresh, const T val = T());

  //! Progressive attenuation of borders by "size" pixels
  void attenuateBorders(int size);

  //! set borders of "size" pixels to given value
  /*! (default value is T's default, which should be T's representation
    of zero).
   */
  void setBorders(const int size, const T value = T());

  //! Paste image into *this at given position
  /*! Image must fit within *this and no resizing/interpolation is
    done.  See embed() for a more sophisticated version with
    interpolation and resizing. */
  void paste(const Image<T>& img, const Point2D& pos);

  //! Paste a gabor into an image area. Account for overlap
  /*! this is a simple extention to paste, but it tests for gabor
    overlap and attempts rectify similar to solution for MakeSnake
    by taking the pixel value furthest from the background value
    @param background This is the color of the image background
  */

  void pasteGabor(const Image<T>& gabor,
                  const Point2D& pos,const T background);

  //! Embed image into *this with position & size given in rectangle
  void embed(const Image<T>& img, const Rectangle& r,
             const T background, const bool keep_aspect = true);

  //! Wilson non-linearity
  void wilsonNormalize(const float H, const float eps);

  //! Pass image through sigmoid: f(x) = x^g / (s + x^h)
  void sigmoid(const float g, const float h, const float s);

  //! Pass image through sigmoid: f(x) = x^2 / (s + x^2)
  void sigmoid(const float s);

  //! clear contents (or set to given value)
  inline void clear(const T& val = T());

  //! convert a grey image to color
  void getColor(const Image<T>& A, Image<PixRGB<T> > &B);
  //! convert a color image to grey, this is a float to keep math clean
  void getGrey(const Image<PixRGB<float> >& A, Image<T>& B);
  //! return a double nested vector that represents data (see also =operator)
  void getVector(std::vector<std::vector<T> > *V) const;
  //! return a double nested vector that represents data (see also =operator)
  void getVector(std::vector<std::vector<T*> > *V) const;

  //! return a single row of an image as a vector
  void getVectorRow(std::vector<T> *V, int row) const;
  //! return a single row of an image as a vector
  void getVectorRow(std::vector<T*> *V, int row) const;

  //! return a single column of an image as a vector
  void getVectorColumn(std::vector<T> *V, int column) const;
  //! return a single column of an image as a vector
  void getVectorColumn(std::vector<T*> *V, int column) const;

  //! retain the num largest values and set the other entries to zero
  void retainMax(const int num);

  //@}

  // ############################################################
  /*! @name Drawing functions */
  //@{

  //! Tells how many pixels are zero
  int emptyArea() const;

  //! Draws a disk with color "value" in image
  void drawDisk(const Point2D& center, const int radius, const T value);

  //! Draws a disk in image and check if it hits a target
  /*! @param mask input binary target mask to draw into
    @param center coordinates of disk center
    @param radius radisu of disk, in pixels
    @param value pixel value used to draw the disk
    @param targetval value with which the targets are represented
    @param floodval value to flood target with if target hit
    @return number of targets hit and flooded */
  int drawDiskCheckTarget(Image<T>& mask, const Point2D& center,
                            const int radius, const T value,
                            const T targetval, const T floodval);

  //! Draw a line from p1 to p2 of color col and pixel radius rad
  void drawLine(const Point2D& p1, const Point2D& p2, const T col,
                const int rad = 1);

  //! Draw a rectangle of color col and pixel radius rad
  void drawRect(const Rectangle& rect, const T col, const int rad = 1);

  //! Draw a rectangle of color col and pixel radius rad, no size assertions
  void drawRectEZ(const Rectangle& rect, const T col, const int rad = 1);

  //! Draw a cross of color col, half-size siz and pixel radius rad
  void drawCross(const Point2D& p, const T col, const int siz = 3,
                 const int rad = 1);

  //! Draw a square patch of color col and size siz
  void drawPatch(const Point2D& p, const int siz, const T col);

  //! Draw a circle of color col and radius radius, with pixel radius rad
  void drawCircle(const Point2D& p, const int radius, const T col,
                  const int rad = 1);

  //! Draw an arrow from p1 to p2 with pixel radius rad
  void drawArrow(const Point2D& p1, const Point2D& p2, const T col,
                 const int rad = 1);

  //! Draw a grid
  void drawGrid(const int spacingX, const int spacingY,
                const int thickX, const int thickY, const T col);

  //! Write some text; only ASCII chars between 32 and 126 are supported
  void writeText(const Point2D& top_left, const char* text,
                 const T col = T(), const T bgcol = T(255));

  //! draw a dot which is a point with 4 connected points
  /*! this is between the setVal method and drawCircle method for
  illustrating points
  */
  void drawPoint(int X,int Y, T pix);

  //@}

  // ############################################################
  /*! @name Miscellaneous functions */
  //@{

  //! Floods object >= thresh with val from seed and returns area
  int flood(Image<T>& dest, const Point2D& seed,
              const T thresh, const T val) const;

  //! How many particles >= thresh ?
  int countParticles(const T thresh) const;

  //! Counts how many pixels are > thresh (in absolute value if absol true)
  int countThresh(const T thresh, const bool absol = true) const;

  //! Add speckle noise to array; thresh = 1 for 100% noise
  void speckleNoise(const float thresh, const int size,
                    const T noise_color, bool random_color=false);

  //! Add background noise to image
  /*! Range should be max value for any image
   of the type to which we add noise */
  void addBGnoise(const float range);

  //! Add background noise to image
  /*! Range should be max value for any image
   of the type to which we add noise
   similar to addBGnoise, but eliminates BNG.. constant*/
  void addBGnoise2(const float range);


  //@}

  // ############################################################
  /*! @name Functions for testing/debugging only */
  //@{

  //! For testing/debugging only.
  /*! See if we are pointing to the same ArrayData<T> as is the other
      Image<T>. */
  bool hasSameData(const Image<T>& b) const;

  //! For testing/debugging only.
  /*! Returns the current reference count. */
  long refCount() const throw();

  //@}

private:
  // ############################################################
  // ##### Data:
  // ############################################################
  ArrayHandle<T> itsHdl;
  inline const ArrayData<T>& impl() const;
  inline ArrayData<T>& uniq();
};

// ######################################################################
// ######################################################################
// ######################################################################
// INLINE FUNCTIONS:
// ######################################################################
// ######################################################################
// ######################################################################

// ######################################################################
// ##### Constructors & Destructors:
// ######################################################################

// ######################################################################
template <class T> inline
Image<T>::Image(const T* inarray, int width, int height) :
  itsHdl(new ArrayData<T>(Dims(width, height), inarray))
{}

// ######################################################################
template <class T> inline
Image<T>::Image(const T* inarray, const Dims& dims) :
  itsHdl(new ArrayData<T>(dims, inarray))
{}

// ######################################################################
template <class T> inline
Image<T>::Image(int width, int height, InitPolicy init) :
  itsHdl(new ArrayData<T>(Dims(width, height), init))
{}

// ######################################################################
template <class T> inline
Image<T>::Image(const Dims& dims, InitPolicy init) :
  itsHdl(new ArrayData<T>(dims, init))
{}

// ######################################################################
template <class T> inline
Image<T>::Image() :
  itsHdl(new ArrayData<T>())
{}

// ######################################################################
template <class T> inline
Image<T>::Image(const Image<T>& A) :
  itsHdl(A.itsHdl)
{}

// ######################################################################
template <class T> template <class T2> inline
Image<T>::Image(const Image<T2>& A) :
  itsHdl(new ArrayData<T>(A.getDims(), NO_INIT))
{
  typename Image<T2>::const_iterator sptr = A.begin();
  typename Image<T>::iterator aptr = beginw();
  typename Image<T>::iterator stop = endw();
  while (aptr != stop) *aptr++ = clamped_convert<T>(*sptr++);
}

// ######################################################################
template <class T> inline
void Image<T>::swap(Image<T>& other)
{
  itsHdl.swap(other.itsHdl);
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator=(const Image<T>& A)
{
  Image<T> A_copy( A );
  this->swap(A_copy);
  return *this;
}

// ######################################################################



// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator=(std::vector<std::vector<T> >& A)
{
  if(((signed)A.size() > this->getHeight()) ||
     ((signed)A[0].size() > this->getWidth()))
    this->resize((signed)A[0].size(),(signed)A.size());
  typename Image<T>::iterator aptr = this->beginw();
  typename std::vector<std::vector<T> >::iterator vptr = A.begin();
  typename std::vector<std::vector<T> >::iterator vstp = A.end();
  while(vptr != vstp)
  {
    typename std::vector<T>::iterator vvptr = vptr->begin();
    typename std::vector<T>::iterator vvstp = vptr->end();
    while(vvptr != vvstp)
      *aptr++ = *vvptr++;
    ++vptr;
  }
  return *this;
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator=(std::vector<std::vector<T>* >& A)
{
  if(((signed)A.size() > this->getHeight()) ||
     ((signed)A[0]->size() > this->getWidth()))
    this->resize((signed)A[0]->size(),(signed)A.size());
  typename Image<T>::iterator aptr = this->beginw();
  for(int ii = 0; ii < (signed)A.size(); ii++)
  {
    //std::vector<T>* temp = A[ii];
    typename std::vector<T>::iterator vvptr = A[ii]->begin();
    typename std::vector<T>::iterator vvstp = A[ii]->end();
    while(vvptr != vvstp)
      *aptr++ = *vvptr++;
  }
  return *this;
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator=(std::vector<std::vector<T*> >& A)
{
  if(((signed)A.size() > this->getHeight()) ||
     ((signed)A[0].size() > this->getWidth()))
    this->resize((signed)A[0].size(),(signed)A.size());
  typename Image<T>::iterator aptr = this->beginw();
  typename std::vector<std::vector<T*> >::iterator vptr = A.begin();
  typename std::vector<std::vector<T*> >::iterator vstp = A.end();
  while(vptr != vstp)
  {
    typename std::vector<T*>::iterator vvptr = vptr->begin();
    typename std::vector<T*>::iterator vvstp = vptr->end();
    while(vvptr != vvstp)
      *aptr++ = **vvptr++;
    ++vptr;
  }
  return *this;
}


// ######################################################################
template <class T> template <class T2> inline
Image<T>& Image<T>::operator=(const Image<T2>& A)
{
  // make a new image block if necessary
  if (this->getDims() != A.getDims() || itsHdl.isShared())
    *this = Image<T>(A.getDims(), NO_INIT);

  typename Image<T2>::const_iterator sptr = A.begin();
  typename Image<T>::iterator aptr = this->beginw();
  typename Image<T>::iterator stop = this->endw();
  while (aptr != stop) *aptr++ = clamped_convert<T>(*sptr++);
  return *this;
}

// ######################################################################
template <class T> inline
void Image<T>::attach(T* array, const int w, const int h)
{
  ArrayHandle<T> attached(new ArrayData<T>(Dims(w, h), array, WRITE_THRU));
  itsHdl.swap(attached);
}

// ######################################################################
template <class T> inline
void Image<T>::detach()
{
  ArrayHandle<T> emptyHdl;
  itsHdl.swap(emptyHdl);
}

// ######################################################################
template <class T> inline
Image<T>::~Image()
{ /* memory deallocation is handled by ArrayData's destructor */ }

// ######################################################################
// ##### Memory management functions:
// ######################################################################

template <class T> inline
bool Image<T>::initialized() const
{ return getWidth() > 0 && getHeight() > 0; }

// ######################################################################
template <class T> inline
void Image<T>::resize(const Dims& dims, const bool do_clear)
{
  // This algorithm is slightly more convoluted than may appear necessary
  // at first, in order to be most efficient in the context of
  // copy-on-write. If the requested size is different than the current
  // size, it's very simple: we just make a new image of the right size,
  // clearing it if so requested. Otherwise, if we're keeping the same
  // size, we don't necessarily have to make a new data block; instead we
  // just delegate to clear(), which does the right thing depending on
  // whether we currently have a shared or unshared data block.

  if (dims != getDims())
    {
      ArrayHandle<T> resized(new ArrayData<T>(dims, do_clear ? ZEROS:NO_INIT));
      itsHdl.swap(resized);
    }
  else // we're keeping the same size, so just clear() if necessary
    {
      if (do_clear) clear( T() );
    }
}

// ######################################################################
template <class T> inline
void Image<T>::resize(const int width, const int height, const bool clear)
{
  resize(Dims(width, height), clear);
}

// ######################################################################
template <class T> inline
void Image<T>::freeMem()
{
  Image<T> empty;
  this->swap(empty);
}

// ######################################################################
// ##### Iterators:
// ######################################################################

#ifndef MEM_DEBUG

template<class T> inline
typename Image<T>::const_iterator Image<T>::begin() const
{ return impl().data(); }

template<class T> inline
typename Image<T>::const_iterator Image<T>::end() const
{ return impl().end(); }

template<class T> inline
typename Image<T>::iterator Image<T>::beginw()
{ return uniq().dataw(); }

template<class T> inline
typename Image<T>::iterator Image<T>::endw()
{ return uniq().endw(); }

#else

template <class T> inline
typename Image<T>::const_iterator Image<T>::begin() const
{ return const_iterator(impl().data(), impl().end()); }

template <class T> inline
typename Image<T>::const_iterator Image<T>::end() const
{ return const_iterator(impl().end(), impl().end()); }

template <class T> inline
typename Image<T>::iterator Image<T>::beginw()
{ return iterator(uniq().dataw(), uniq().endw()); }

template <class T> inline
typename Image<T>::iterator Image<T>::endw()
{ return iterator(uniq().endw(), uniq().endw()); }

#endif

// ######################################################################
// ##### Access functions:
// ######################################################################

template <class T> inline
int Image<T>::getSize() const
{ return getDims().sz(); }

// ######################################################################
template <class T> inline
int Image<T>::getWidth() const
{ return itsHdl.get().w(); }

// ######################################################################
template <class T> inline
int Image<T>::getHeight() const
{ return itsHdl.get().h(); }

// ######################################################################
template <class T> inline
const Dims& Image<T>::getDims() const
{ return itsHdl.get().dims(); }

// ######################################################################
template <class T> template <class C> inline
bool Image<T>::isSameSize(const C& other) const
{ return getWidth() == other.getWidth() && getHeight() == other.getHeight(); }

// ######################################################################
template <class T> inline
bool Image<T>::is1D() const
{ return (getWidth() == 1) || (getHeight() == 1); }

// ######################################################################
template<class T> inline
void Image<T>::getMinMax(T& xmini, T& xmaxi) const
{
  ASSERT(initialized());
  typename Image<T>::const_iterator aptr = begin(), stop = end();
  xmini = *aptr; xmaxi = *aptr++;
  while (aptr != stop)
    {
      if (*aptr < xmini) xmini = *aptr;
      else if (*aptr > xmaxi) xmaxi = *aptr;
      ++aptr;
    }
}

// ######################################################################
template <class T> inline
T& Image<T>::operator[](int index)
{
  ASSERT(index >= 0 && index < this->getSize());
  return beginw()[index];
}

// ######################################################################
template <class T> inline
const T& Image<T>::operator[](int index) const
{
  ASSERT(index >= 0 && index < this->getSize());
  return begin()[index];
}

// ######################################################################
template <class T> inline
const T& Image<T>::getVal(const int index) const
{
  ASSERT(index >= 0 && index < this->getSize());
  return begin()[index];
}

// ######################################################################
template <class T> inline
const T& Image<T>::getVal(const int x, const int y) const
{
  ASSERT(coordsOk(x, y));
  return(begin()[x + y * getWidth()]);
}

// ######################################################################
template <class T> inline
const T& Image<T>::getVal(const Point2D& p) const
{
  ASSERT(coordsOk(p));
  return(begin()[p.i + p.j * getWidth()]);
}

// ######################################################################
template <class T> template <class T2> inline
void Image<T>::getVal(const int x, const int y, T2& val) const
{
  ASSERT(coordsOk(x, y));
  val = begin()[x + y * getWidth()];
}

// ######################################################################
template <class T> inline
const T Image<T>::getValInterp(const float x, const float y) const
{
  typename Image<T>::const_iterator const sptr = begin();
  const int wid = getWidth(), hei = getHeight();

  const int y0 = int(y);
  const int y1 = std::min(y0 + 1, hei - 1);
  const float fy = y - float(y0);

  const int wy0 = wid * y0;
  const int wy1 = wid * y1;

  const int x0 = int(x);
  const int x1 = std::min(x0 + 1, wid - 1);
  const float fx = x - float(x0);

  ASSERT(coordsOk(x0, y0));
  typename promote_trait<T, float>::TP const
    d00 = sptr[x0 + wy0], d10 = sptr[x1 + wy0],
    d01 = sptr[x0 + wy1], d11 = sptr[x1 + wy1],
    dx0 = d00 + (d10 - d00) * fx,
    dx1 = d01 + (d11 - d01) * fx;
  return T( dx0 + (dx1 - dx0) * fy );  // no need to clamp
}

// ######################################################################
template <class T> template <class T2> inline
void Image<T>::setVal(const int index, const T2& value)
{
  ASSERT(index >= 0 && index < this->getSize());
  beginw()[index] = clamped_convert<T>(value);
}

// ######################################################################
template <class T> template <class T2> inline
void Image<T>::setVal(const int x, const int y, const T2& value)
{
  ASSERT(coordsOk(x, y));
  beginw()[x + y * getWidth()] = clamped_convert<T>(value);
}

// ######################################################################
template <class T> template <class T2> inline
void Image<T>::setVal(const Point2D& p, const T2& value)
{
  ASSERT(coordsOk(p));
  beginw()[p.i + p.j * getWidth()] = clamped_convert<T>(value);
}

// ######################################################################
template <class T> inline
const T* Image<T>::getArrayPtr() const
{
  return impl().data();
}

// ######################################################################
template <class T> inline
T* Image<T>::getArrayPtr()
{
  return uniq().dataw();
}

// ######################################################################
template <class T> inline
bool Image<T>::coordsOk(const Point2D& P) const
{
  return (P.i >= 0 && P.j >= 0 && P.i < getWidth() && P.j < getHeight());
}

// ######################################################################
template <class T> inline
bool Image<T>::coordsOk(const int i, const int j) const
{
  return (i >= 0 && j >= 0 && i < getWidth() && j < getHeight());
}

// ######################################################################
template <class T> inline
bool Image<T>::rectangleOk(const Rectangle& rect) const
{
  return (rect.left() < getWidth() && rect.right() < getWidth() &&
          rect.top() < getHeight() && rect.bottom() < getHeight() &&
          rect.left() >= 0 && rect.right() >= 0 &&
          rect.top() >= 0 && rect.bottom() >= 0);
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator+=(const T& val)
{
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr)
    *itr = clamped_convert<T>( (*itr) + val );
  return *this;
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator-=(const T& val)
{
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr)
    *itr = clamped_convert<T>( (*itr) - val );
  return *this;
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator*=(const T& val)
{
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr)
    *itr = clamped_convert<T>( (*itr) * val );
  return *this;
}

// ######################################################################
template <class T> inline
Image<T>& Image<T>::operator/=(const T& val)
{
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr)
    *itr = clamped_convert<T>( (*itr) / val );
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
Image<T>& Image<T>::operator+=(const Image<T2>& A)
{
  ASSERT(isSameSize(A));
  typename Image<T2>::const_iterator itr2 = A.begin();
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr, ++itr2)
    *itr = clamped_convert<T>( (*itr) + (*itr2) );
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
Image<T>& Image<T>::operator-=(const Image<T2>& A)
{
  ASSERT(isSameSize(A));
  typename Image<T2>::const_iterator itr2 = A.begin();
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr, ++itr2)
    *itr = clamped_convert<T>( (*itr) - (*itr2) );
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
Image<T>& Image<T>::operator*=(const Image<T2>& A)
{
  ASSERT(isSameSize(A));
  typename Image<T2>::const_iterator itr2 = A.begin();
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr, ++itr2)
    *itr = clamped_convert<T>( (*itr) * (*itr2) );
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
Image<T>& Image<T>::operator/=(const Image<T2>& A)
{
  ASSERT(isSameSize(A));
  typename Image<T2>::const_iterator itr2 = A.begin();
  for (typename Image<T>::iterator itr = beginw(), stop = endw();
       itr != stop; ++itr, ++itr2)
    *itr = clamped_convert<T>( (*itr) / (*itr2) );
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator+(const T2& val) const
{
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++ritr)
    *ritr = (*itr) + val;
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator-(const T2& val) const
{
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++ritr)
    *ritr = (*itr) - val;
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator*(const T2& val) const
{
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++ritr)
    *ritr = (*itr) * val;
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator/(const T2& val) const
{
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++ritr)
    *ritr = (*itr) / val;
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator+(const Image<T2>& img) const
{
  ASSERT(isSameSize(img));
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  typename Image<T2>::const_iterator itr2 = img.begin();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++itr2, ++ritr)
    *ritr = (*itr) + (*itr2);
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator-(const Image<T2>& img) const
{
  ASSERT(isSameSize(img));
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  typename Image<T2>::const_iterator itr2 = img.begin();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++itr2, ++ritr)
    *ritr = (*itr) - (*itr2);
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator*(const Image<T2>& img) const
{
  ASSERT(isSameSize(img));
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  typename Image<T2>::const_iterator itr2 = img.begin();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++itr2, ++ritr)
    *ritr = (*itr) * (*itr2);
  return result;
}

// ######################################################################
template <class T> template <class T2> inline
Image<typename promote_trait<T,T2>::TP> Image<T>::operator/(const Image<T2>& img) const
{
  ASSERT(isSameSize(img));
  typedef typename promote_trait<T,T2>::TP TPRO;
  Image<TPRO> result(getWidth(), getHeight(), NO_INIT);
  typename Image<TPRO>::iterator ritr = result.beginw();
  typename Image<T2>::const_iterator itr2 = img.begin();
  for (typename Image<T>::const_iterator itr = begin(), stop = end();
       itr != stop; ++itr, ++itr2, ++ritr)
    *ritr = (*itr) / (*itr2);
  return result;
}

// ######################################################################
template <class T> inline
void Image<T>::clamp(const T cmin, const T cmax)
{
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();
  while (aptr != stop)
    {
      if (*aptr < cmin) *aptr = cmin;
      else if (*aptr > cmax) *aptr = cmax;
      ++aptr;
    }
}

// ######################################################################
template <class T> inline
void Image<T>::clear(const T& val)
{
  // Check if we have a shared implementation... if not, then for efficiency
  // we should release our copy before doing the iterative loop, which would
  // otherwise unnecessarily make a unique copy of the shared data for us,
  // which we would then ceremoniously ignore as we clear it to a new value.
  if (itsHdl.isShared())
    *this = Image<T>(getDims(), NO_INIT);

  for (iterator itr = beginw(), stop = endw(); itr != stop; ++itr)
    *itr = val;
}

// ######################################################################

template <class T> inline
void getGrey(const Image<PixRGB<float> >& A,Image<T>& B)
{
  B.resize(A.getWidth(),A.getHeight());
  typename Image<T>::iterator BWptr  = B.beginw();
  typename Image<PixRGB<float> >::const_iterator COLptr = A.begin();
  float h,s,v;
  while(BWptr != B.endw())
  {
    PixRGB<float> pix = *COLptr;
    pix.getHSV(h,s,v);
    *BWptr = (T)(v/255.0F);
    ++BWptr; ++COLptr;
  }
}

// ######################################################################

template <class T> inline
void getColor(const Image<T>& A, Image<PixRGB<T> >& B)
{
  B.resize(A.getWidth(),A.getHeight()); 
  typename Image<T>::const_iterator BWptr     = A.begin();
  typename Image<PixRGB<T> >::iterator COLptr = B.beginw();
  PixRGB<double> pix;

  while(BWptr != A.end())
  {
    pix.setHSV(0,0,*BWptr*255.0F);
    *COLptr = pix;
    ++BWptr; ++COLptr;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVector(std::vector<std::vector<T> > *V) const
{
  ASSERT(V->size() >= (unsigned int)this->getHeight());
  ASSERT(V->at(0).size() >= (unsigned int)this->getWidth());
  typename Image<T>::const_iterator aptr = this->begin();
  typename std::vector<std::vector<T> >::iterator vptr = V->begin();
  typename std::vector<std::vector<T> >::iterator vstp = V->end();
  while(vptr != vstp)
  {
    typename std::vector<T>::iterator vvptr = vptr->begin();
    typename std::vector<T>::iterator vvstp = vptr->end();
    while(vvptr != vvstp)
      *vvptr++ = *aptr++;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVector(std::vector<std::vector<T*> > *V) const
{
  ASSERT(V->size() >= (unsigned int)this->getHeight());
  ASSERT(V->at(0).size() >= (unsigned int)this->getWidth());
  typename Image<T>::const_iterator aptr = this->begin();
  typename std::vector<std::vector<T*> >::iterator vptr = V->begin();
  typename std::vector<std::vector<T*> >::iterator vstp = V->end();
  while(vptr != vstp)
  {
    typename std::vector<T*>::iterator vvptr = vptr->begin();
    typename std::vector<T*>::iterator vvstp = vptr->end();
    while(vvptr != vvstp)
      **vvptr++ = *aptr++;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVectorRow(std::vector<T> *V, int row) const
{
  ASSERT(V->size() >= (unsigned int)this->getWidth());
  typename Image<T>::const_iterator aptr =
    this->begin() + (row * getWidth());
  typename std::vector<T>::iterator vptr = V->begin();
  typename std::vector<T>::iterator vstp = V->end();
  while(vptr != vstp)
  {
      *vptr++ = *aptr++;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVectorRow(std::vector<T*> *V, int row) const
{
  ASSERT(V->size() >= (unsigned int)this->getWidth());
  typename Image<T>::const_iterator aptr =
    this->begin() + (row * getWidth());

  typename std::vector<T*>::iterator vptr = V->begin();
  typename std::vector<T*>::iterator vstp = V->end();
  while(vptr != vstp)
  {
      **vptr++ = *aptr++;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVectorColumn(std::vector<T> *V, int column) const
{
  ASSERT(V->size() >= (unsigned int)this->getHeight());
  typename std::vector<T>::iterator vptr = V->begin();
  typename std::vector<T>::iterator vstp = V->end();
  int current = 0;
  while(vptr != vstp)
  {
    *vptr++ = this->begin()[column + current*getWidth()];
    current++;
  }
}

// ######################################################################
template <class T> inline
void Image<T>::getVectorColumn(std::vector<T*> *V, int column) const
{
  ASSERT(V->size() >= (unsigned int)this->getHeight());
  typename std::vector<T*>::iterator vptr = V->begin();
  typename std::vector<T*>::iterator vstp = V->end();
  int current = 0;
  while(vptr != vstp)
  {
    **vptr++ = this->begin()[column + current*getWidth()];
    current++;
  }
}

// ######################################################################
// ##### Functions for testing/debugging only:
// ######################################################################

template <class T> inline
bool Image<T>::hasSameData(const Image<T>& b) const
{ return itsHdl.hasSameData(b.itsHdl); }

// ######################################################################
template <class T> inline
long Image<T>::refCount() const throw() { return itsHdl.refCount(); }


// ######################################################################
// ##### Private methods:
// ######################################################################

template <class T> inline
const ArrayData<T>& Image<T>::impl() const
{ return itsHdl.get(); }

// ######################################################################
template <class T> inline
ArrayData<T>& Image<T>::uniq()
{ return itsHdl.uniq(); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_templates.h"
#include "Image_MonoOnly_templates.h"

#endif
