/*!@file ImageSet.H */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ImageSet.H 3121 2004-02-21 21:10:00Z itti $
//

#ifndef IMAGESET_H_DEFINED
#define IMAGESET_H_DEFINED

#include "saliency.H"
#include "ArrayData.H"
#include <deque>

class Dims;
class Point2D;
template <class T> class Image;

// ######################################################################
//! This class implements a set of images, often used as a dyadic pyramid.
template <class T>
class ImageSet
{
public:
  //! Construct with a given number of empty images.
  ImageSet(const uint n = 0);

  //! Construct with a given number of images of the given size.
  ImageSet(const uint n, const Dims& dims, InitPolicy init = ZEROS);

  // default dtor, copy-ctor and assignment OK

  //! Swap contents with another ImageSet
  void swap(ImageSet<T>& that);

  //! Keep the same number of images, but make all images empty (0-by-0 size).
  /*! This returns the ImageSet to a state where isEmpty() is true. */
  void clear();

  //! Reset to a new number of images, all images will be empty (0-by-0 size).
  void reset(uint newDepth = 0);

  //! Return number of images in image set.
  inline uint size() const;

  //! A synonym for size() (to be backward-compatible with Pyramid syntax).
  inline uint getDepth() const;

  //! Get image from a given level.
  const Image<T>& getImage(const uint lev) const;

  //! Get image from a given level (shorthand for getImage()).
  inline const Image<T>& operator[](const uint lev) const;

  //! Get mutable image from a given level.
  Image<T>& getImageMut(const uint lev);

  //! Get mutable image from a given level (shorthand for getImageMut()).
  inline Image<T>& operator[](const uint lev);

  //! Get value of a pixel at a given level and coordinates
  /*! @param pix The pixel value returned
    @param p Coordinates given at scale 0 (base pyramid level);
    bilinear interpolation will be used to compute pixel value at
    given level.
    @param lev The pyramid level to get the pixel value from. */
  T getPixel(const Point2D& p, const uint lev) const;

  //! Get value of a pixel at a given level and coordinates, no interpolation
  /*! @param pix The pixel value returned
    @param p Coordinates given at scale 0 (base pyramid level);
    This version does not use interpolation.
    @param lev The pyramid level to get the pixel value from. */
  T getPixelNI(const Point2D& p, const uint lev) const;

  //! push an image to the back of the set
  void push_back(Image<T>& img);

  //! push an image to the front of the set
  void push_front(Image<T>& img);

  //! return the last image in the set and pop it off
  Image<T> pop_back();

  //! return the first image in the set and pop it off
  Image<T> pop_front();

  //! return the last image in the set (without popping it off)
  Image<T> front() const;

  //! return the first image in the set (without popping it off)
  Image<T> back() const;

  //! Return true if the pyramid has no non-empty images (width*height > 0).
  inline bool isEmpty() const;

  //! Return true if the pyramid has any non-empty images (width*height > 0).
  bool isNonEmpty() const;


private:
  std::deque< Image<T> > itsLevels;
};



// ######################################################################
// ######################################################################
// ##### INLINED FUNCTIONS:
// ######################################################################
// ######################################################################

// ######################################################################
template <class T> inline
uint ImageSet<T>::getDepth() const { return itsLevels.size(); }

// ######################################################################
template <class T> inline
uint ImageSet<T>::size() const { return itsLevels.size(); }

// ######################################################################
template <class T> inline
const Image<T>& ImageSet<T>::operator[](const uint lev) const
{ return getImage(lev); }

// ######################################################################
template <class T> inline
Image<T>& ImageSet<T>::operator[](const uint lev) { return getImageMut(lev); }

// ######################################################################
template <class T> inline
bool ImageSet<T>::isEmpty() const { return !isNonEmpty(); }

#include "ImageSet_templates.h"

#endif // !IMAGESET_H_DEFINED

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
