/*!@file ImageSet.C Implementation for image pyramid template class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ImageSet.C 4290 2005-03-12 02:03:27Z mundhenk $
//

#include "ImageSet.H"

#include "Image.H"
#include "log.H"
#include "Point2D.H"

#include <algorithm>
#include <cmath>

// ######################################################################
template <class T>
ImageSet<T>::ImageSet(const uint n)
  : itsLevels(n)
{}

// ######################################################################
template <class T>
ImageSet<T>::ImageSet(const uint n, const Dims& dims, InitPolicy init)
  : itsLevels(n, Image<T>(dims,init))
  //itsLevels(std::max(n, uint(1)), Image<T>(dims, init))
{}

// ######################################################################
template <class T>
void ImageSet<T>::swap(ImageSet<T>& that)
{
  this->itsLevels.swap(that.itsLevels);
}

// ######################################################################
template <class T>
void ImageSet<T>::clear()
{
  // Reset to an empty vector of given size by swapping with a temp vector
  std::deque< Image<T> >(itsLevels.size()).swap(itsLevels);
}

// ######################################################################
template <class T>
void ImageSet<T>::reset(uint newDepth)
{
  // Reset to an empty vector of given size by swapping with a temp vector
  std::deque< Image<T> >(newDepth).swap(itsLevels);
}

template <class T>
const Image<T>& ImageSet<T>::getImage(const uint lev) const
{
  ASSERT(lev < itsLevels.size());
  return itsLevels[lev];
}

// ######################################################################
template <class T>
Image<T>& ImageSet<T>::getImageMut(const uint lev)
{
  ASSERT(lev < itsLevels.size());
  return itsLevels[lev];
}

// ######################################################################
template <class T>
T ImageSet<T>::getPixel(const Point2D& p, const uint lev) const
{
  ASSERT(lev < itsLevels.size());
  ASSERT(itsLevels[0].coordsOk(p));

  float fac = float(1 << lev);
  float ii = float(p.i) / fac, jj = float(p.j) / fac;
  int ww = itsLevels[lev].getWidth() - 1, hh = itsLevels[lev].getHeight() - 1;
  if (ii > ww) ii = ww; if (jj > hh) jj = hh;

  return itsLevels[lev].getValInterp(ii, jj);
}

// ######################################################################
template <class T>
T ImageSet<T>::getPixelNI(const Point2D& p, const uint lev) const
{
  ASSERT(lev < itsLevels.size());
  ASSERT(itsLevels[0].coordsOk(p));

  int ii = p.i >> lev, jj = p.j >> lev;
  int ww = itsLevels[lev].getWidth() - 1, hh = itsLevels[lev].getHeight() - 1;
  if (ii > ww) ii = ww; if (jj > hh) jj = hh;

  return itsLevels[lev].getVal(ii, jj);
}

// ######################################################################
template <class T>
void ImageSet<T>::push_back(Image<T>& img)
{
  itsLevels.push_back(img);
}

// ######################################################################
template <class T>
void ImageSet<T>::push_front(Image<T>& img)
{
  itsLevels.push_front(img);
}

// ######################################################################
template <class T>
Image<T> ImageSet<T>::pop_back()
{
  ASSERT(isNonEmpty());
  Image<T> tmp = itsLevels.back();
  itsLevels.pop_back();
  return tmp;
}

// ######################################################################
template <class T>
Image<T> ImageSet<T>::pop_front()
{
  ASSERT(isNonEmpty());
  Image<T> tmp = itsLevels.front();
  itsLevels.pop_front();
  return tmp;
}

// ######################################################################
template <class T>
Image<T> ImageSet<T>::back() const
{
  ASSERT(isNonEmpty());
  return itsLevels.back();
}

// ######################################################################
template <class T>
Image<T> ImageSet<T>::front() const
{
  ASSERT(isNonEmpty());
  return itsLevels.front();
}

// ######################################################################
template <class T>
bool ImageSet<T>::isNonEmpty() const
{
  for (uint i = 0; i < getDepth(); ++i)
    if (itsLevels[i].initialized()) return true;

  return false;
}


#define INSTANTIATE(T) template class ImageSet< T >;

#ifdef INST_BYTE
INSTANTIATE(byte); INSTANTIATE(PixRGB<byte>);
INSTANTIATE(PixH2SV<byte>);
#endif
#ifdef INST_INT16
INSTANTIATE(int16); INSTANTIATE(PixRGB<int16>);
INSTANTIATE(PixH2SV<int16>);
#endif
#ifdef INST_INT32
INSTANTIATE(int32); INSTANTIATE(PixRGB<int32>);
INSTANTIATE(PixH2SV<int32>);
#endif
#ifdef INST_FLOAT
INSTANTIATE(float); INSTANTIATE(PixRGB<float>);
INSTANTIATE(PixH2SV<float>);
#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
