/*!@file Image_ColorOps.C Color operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_ColorOps.C 4239 2005-02-28 04:06:59Z mundhenk $
//

#ifndef IMAGE_COLOROPS_C_DEFINED
#define IMAGE_COLOROPS_C_DEFINED

#include "Image_ColorOps.H"
#include "MathFunctions.H"
#include "Image.H"
#include "Image_FilterOps.H"
#include "Image_ShapeOps.H"
#include "Image_Transforms.H"
#include "Image_MathOps.H"
#include "log.H"

#include <algorithm>
#include <cmath>

// ######################################################################
Image<PixRGB<float> > normalizeRGPolar(const Image<float>& src,
                                       const float max,
                                       const float min)
{
  float aMin = fabs(min); //abs value of negative component
  float maxVal = std::max(aMin, max); // assigned the value of which is larger
                                      // fabs(min) or max

  Image<PixRGB<float> > result(src.getDims(), NO_INIT);

  Image<PixRGB<float> >::iterator dptr = result.beginw();
  Image<PixRGB<float> >::iterator stop = result.endw();
  Image<float>::const_iterator sptr = src.begin();
  
  while (dptr != stop)
    {
      float src_val = *sptr++;
      if(src_val < 0) //is negative value, set red pixel
        {
          *dptr++ = PixRGB<float>(src_val*(255.0F/maxVal)*(-1.0F),
                                  0.0F,
                                  0.0F);
        }
      else //is positive value, set green pixel
        {
          *dptr++ = PixRGB<float>(0.0F,
                                  src_val*(255.0F/maxVal),
                                  0.0F);
        }
    }

  return result;
}

// ######################################################################
Image<PixRGB<float> > overlayStain(const Image<float>& top,
                                   const Image<float>& bottom,
                                   const float trans, const char channel)
{
  // 0 to 100 percent only:
  ASSERT((trans >= 0.0F) && (trans <= 100.0F));
  // images must be same size:
  ASSERT(top.isSameSize(bottom));

  // result is the size of the input images:
  Image<PixRGB<float> > result(top.getDims(), NO_INIT);

  Image<PixRGB<float> >::iterator
    dptr = result.beginw(),
    stop = result.endw();

  Image<float>::const_iterator
    tptr = top.begin(),
    bptr = bottom.begin();
  
  float tfac = trans * 0.01F;

  while (dptr != stop)
    {
      float top_val = *tptr++;

      PixRGB<float> pix;

      if(channel == 'r'){pix.setRed(top_val);pix.setGreen(0);pix.setBlue(0);}
      if(channel == 'g'){pix.setRed(0);pix.setGreen(top_val);pix.setBlue(0);}
      if(channel == 'b'){pix.setRed(0);pix.setGreen(0);pix.setBlue(top_val);}

      *dptr++ = pix - ( pix - (*bptr++) ) * tfac;
    }

  return result;
}

// ######################################################################
Image<PixRGB<byte> > fromRGB555(const byte* data, const Dims& dims)
{
  Image< PixRGB<byte> > dest(dims, NO_INIT);
  Image<PixRGB<byte> >::iterator aptr = dest.beginw();
  Image<PixRGB<byte> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: 1 ignored bit, 5-bit R, 5-bit G, 5-bit B
      int rgb = (int(data[0]) << 8) | data[1];
      byte r = byte((rgb & 0x7C00) >> 7);
      byte g = byte((rgb & 0x03E0) >> 2);
      byte b = byte((rgb & 0x001F) << 3);
      (*aptr++).set(r, g, b);
      data += 2;
    }

  return dest;
}

// ######################################################################
Image<PixRGB<byte> > fromRGB555bs(const byte* data, const Dims& dims)
{
  Image< PixRGB<byte> > dest(dims, NO_INIT);
  Image<PixRGB<byte> >::iterator aptr = dest.beginw();
  Image<PixRGB<byte> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: 1 ignored bit, 5-bit R, 5-bit G, 5-bit B
      int rgb = (int(data[1]) << 8) | data[0];
      byte r = byte((rgb & 0x7C00) >> 7);
      byte g = byte((rgb & 0x03E0) >> 2);
      byte b = byte((rgb & 0x001F) << 3);
      (*aptr++).set(r, g, b);
      data += 2;
    }

  return dest;
}

// ######################################################################
Image<PixRGB<byte> > fromRGB565(const byte* data, const Dims& dims)
{
  Image< PixRGB<byte> > dest(dims, NO_INIT);
  Image<PixRGB<byte> >::iterator aptr = dest.beginw();
  Image<PixRGB<byte> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: 5-bit R, 6-bit G, 5-bit B
      int rgb = (int(data[0]) << 8) | data[1];
      byte r = byte((rgb & 0xF800) >> 8);
      byte g = byte((rgb & 0x07E0) >> 3);
      byte b = byte((rgb & 0x001F) << 3);
      (*aptr++).set(r, g, b);
      data += 2;
    }

  return dest;
}

// ######################################################################
Image<PixRGB<byte> > fromRGB565bs(const byte* data, const Dims& dims)
{
  Image< PixRGB<byte> > dest(dims, NO_INIT);
  Image<PixRGB<byte> >::iterator aptr = dest.beginw();
  Image<PixRGB<byte> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: 5-bit R, 6-bit G, 5-bit B
      int rgb = (int(data[1]) << 8) | data[0];
      byte r = byte((rgb & 0xF800) >> 8);
      byte g = byte((rgb & 0x07E0) >> 3);
      byte b = byte((rgb & 0x001F) << 3);
      (*aptr++).set(r, g, b);
      data += 2;
    }

  return dest;
}

// ######################################################################
void toYUV422(const Image<PixRGB<byte> > img, byte* y, byte *u, byte *v,
              const int ystride, const int ustride, const int vstride)
{
  byte *src = (byte*)(img.getArrayPtr());
  int w = img.getWidth(), h = img.getHeight();

  for (int j = 0; j < h; j += 2)
    {
      // NOTE: the formulas here is an Pixels.H, but we don't use the
      // functions from there just for higher speed here.
      
      // NOTE: this code will probably not work with odd image or sizes

      for (int i = 0; i < w; i += 2)
        {
          // fully convert first RGB pixel:
          float yf, uf, vf;
          float r = float(*src++), g = float(*src++), b = float(*src++);
          yf =  0.299f   * r + 0.587f   * g + 0.114f   * b;
          uf = -0.16874f * r - 0.33126f * g + 0.500f   * b + 128.0f;
          vf =  0.500f   * r - 0.41869f * g - 0.08131f * b + 128.0f;
          if (uf > 255.0f) uf = 255.0f; if (vf > 255.0f) vf = 255.0f;
          *y++ = byte(yf); *u++ = byte(uf); *v++ = byte(vf);
          
          // only get luminance for second RGB pixel:
          r = float(*src++), g = float(*src++), b = float(*src++);
          yf =  0.299f   * r + 0.587f   * g + 0.114f   * b;
          *y++ = byte(yf);
        }
      y += ystride; u += ustride; v += vstride;
      for (int i = 0; i < w; i += 2)
        {
          // only get luminance for third RGB pixel:
          float r = float(*src++), g = float(*src++), b = float(*src++);
          float yf =  0.299f   * r + 0.587f   * g + 0.114f   * b;
          *y++ = byte(yf);
          
          // only get luminance for fourth RGB pixel:
          r = float(*src++), g = float(*src++), b = float(*src++);
          yf =  0.299f   * r + 0.587f   * g + 0.114f   * b;
          *y++ = byte(yf);
        }
      y += ystride;
    }
}

// ######################################################################
Image< PixRGB<byte> > contrastModulate(const Image< PixRGB<byte> >& img,
                                       const Image<float>& mask,
                                       float baseContrast,
                                       byte baseBright)
{
  ASSERT(img.isSameSize(mask));
  ASSERT((baseContrast >= 0.0) && (baseContrast <= 1.0));
  
  Image<float> contrast(mask);
  contrast.clamp(0.01,1.0);
  contrast.normalize(0.0,1.0-baseContrast);
  contrast += baseContrast;
  Image<float> base(contrast * (-baseBright));
  base += baseBright;

  return img * contrast + base;
}

// ######################################################################
void RGBtoCIE(const PixRGB<byte>& rgbColor, 
              float& cr, float& cg, float& intens)
{
  PixRGB<float> fpix(rgbColor);
  intens = (fpix.red() + fpix.green() + fpix.blue());
  if (intens > 0.0f)
    {
      cr = fpix.red() / intens;
      cg = fpix.green() / intens;
    }
  else
    {
      cr = 0.0f;
      cg = 0.0f;
    }
}

// ######################################################################
Image<float> hueDistance(const Image< PixRGB<byte> >& img, 
                         float muR, float muG,
                         float sigR, float sigG, float rho)
{
  Image<float> result(img.getDims(), NO_INIT);
  Image< PixRGB<byte> >::const_iterator iptr, stop = img.end();
  Image<float>::iterator rptr = result.beginw();

  float sR2 = 1/sigR/sigR;
  float sG2 = 1/sigG/sigG;
  float sRG = 2*rho/sigG/sigR;

  for (iptr = img.begin(); iptr != stop; ++iptr, ++rptr)
    {
      float cr, cg, intens;
      RGBtoCIE(*iptr,cr,cg,intens);
      *rptr = exp(-0.5f*((cr-muR)*(cr-muR)*sR2 + (cg-muG)*(cg-muG)*sG2
                         - (cr-muR)*(cg-muG)*sRG));
    }
  return result;
}


// Include the explicit instantiations
// MA #include "inst/Image_ColorOps.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_COLOROPS_C_DEFINED
