/*!@file Image_ColorOps.H Color operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_ColorOps.H 4035 2004-11-12 19:35:59Z rjpeters $
//

// The fromXXX() code here is somewhat inspired from the
// grab_gray_image.c example code provided with the libdc1394
// distribution by Gord Peters, and from the Coriander source code.

#ifndef IMAGE_COLOROPS_H_DEFINED
#define IMAGE_COLOROPS_H_DEFINED

#include "saliency.H"
#include "Promotions.H"

class Dims;
template <class T> class Image;
template <class T> class PixRGB;

//! Set the red, green and blue components from 3 monochromatic images
template <class T>
Image<PixRGB<T> > makeRGB(const Image<T>& red,
                          const Image<T>& green,
                          const Image<T>& blue);

//! Get the red, green and blue components as 3 monochromatic images
template <class T>
void getComponents(const Image<PixRGB<T> >& src,
                   Image<T>& red, Image<T>& green, Image<T>& blue);

/*! This method will normalize an image based upon the maximum value
  provided to it. This is either max or the absolute value of min.
  It then creates an image where all negative values are assigned a red
  pixel value while all postive numbers are assigned a green pixel value
  NOTE: please use a PixRGB type with this class otherwise this method
  makes no sense
  @param max The maximum pixel value
  @param min The minimum pixel value
  @param image The image of values (positive and negative) to be processed
*/
Image<PixRGB<float> > normalizeRGPolar(const Image<float>& src,
                                       const float max,
                                       const float min);

//! Create a color image by staining a monochrome image.
/*! Each pixel in the result is given by the stain color multiplied by the
    corresponding pixel in the monochrome source image. This is entirely
    equivalent to, but more efficient than, creating a color image from the
    monochrome image, then multiplying the color image by the stain
    color.*/
template <class T>
Image<PixRGB<T> > stain(const Image<T>& src, PixRGB<float> color);

//! Overlay an image over another, however, stain the top image
/*! this method works almost exactly like overlay except that the bottom
  image is stained (colored) red green or blue depending on whether
  char channel is set to r,g, or b
*/
Image<PixRGB<float> > overlayStain(const Image<float>& top,
                                   const Image<float>& bottom,
                                   const float trans, const char channel);

//! Get min and max of the components
template <class T>
void getMinMaxC(const Image<PixRGB<T> >& src, T& mi, T& ma);

//! Normalize values between nmin and nmax
template <class T, class T2>
inline void normalizeC(Image<PixRGB<T> >& src, const T2 nmin, const T2 nmax)
{
  ASSERT(src.initialized());
  T mi, ma; getMinMaxC(src, mi, ma);
  const float mif = float(mi);
  const float maf = float(ma);
  const float scale = maf - mif;
  if (scale > -1.0e-20 && scale < 1.0e-20) return;  // no need to rescale...
  const float nscale = (float(nmax) - float(nmin)) / scale;

  typename Image<PixRGB<T> >::iterator aptr = src.beginw();
  typename Image<PixRGB<T> >::iterator stop = src.endw();

  while (aptr != stop)
    {
      aptr->setRed  (nmin+(T2)(((float(aptr->red()))-mif)*nscale));
      aptr->setGreen(nmin+(T2)(((float(aptr->green()))-mif)*nscale));
      aptr->setBlue (nmin+(T2)(((float(aptr->blue()))-mif)*nscale));
      ++aptr;
    }
}

//! Compute luminance of a color image
template <class T>
Image<T> luminance(const Image<PixRGB<T> >& src);

//! Compute luminance of a greyscale image (no-op)
/*! This no-op function is provided so that luminance() can be applied
  to template images where it is not known whether the template
  argument is scalar or PixRGB. */
template <class T>
Image<T> luminance(const Image<T>& src);

//! Convert grayscale image to RGB
template <class T>
Image< PixRGB<T> > toRGB(const Image<T>& src);

//! Convert RGB image to RGB (no-op)
/*! This no-op function is provided so that toRGB() can be applied to
  template images where it is not known whether the template argument
  is scalar or PixRGB. */
template <class T>
Image< PixRGB<T> > toRGB(const Image< PixRGB<T> >& src);

//! Warp image in 3D onto z-map
/*! @param ima Input image
  @param zmap Height map of same size as ima
  @param pitch 3D pitch angle, in degrees
  @param yaw 3D yaw angle, in degrees
  @param dims The result will have these dimensions, unless the dimensions are
  empty (i.e. 0-by-0), in which case the result will be the best size computed
  from ima's size and the pitch and yaw angles.
*/
template <class T>
Image<PixRGB<T> > warp3D(const Image<PixRGB<T> >& ima,
                         const Image<float>& zmap,
                         const float pitch, const float yaw,
                         const float zMax, Dims& dims);

//! Compute color information measure based in infoFFT
template <class T>
Image<float> infoMeasure(const Image<PixRGB<T> >& src,
                         const float eps, const int size);

//! Get YIQ color components
template <class T>
void getYIQ(const Image<PixRGB<T> >& src,
            Image<T>& y, Image<T>& i, Image<T>& q);

//! Get YUV color components from an RGB image
template <class T>
void getYUV(const Image<PixRGB<T> >& src,
            Image<T>& y, Image<T>& u, Image<T>& v);

//! Build an RGB image from YUV color components
template <class T>
Image<PixRGB<T> > setYUV(const Image<T>& y, const Image<T>& u,
                         const Image<T>& v);

//! Build an RGB image from an RGB C array
/*! This will operate a deep copy of the data; this function is
  somewhat redundant with the Image constructor and is provided for
  completeness with respect to the fromXXX methods only. Note that
  data should have 3*dims.w()*dims.h() elements, and it is the
  responsibility of the caller to free that data later. */
template <class T>
Image<PixRGB<T> > fromRGB(const T* data, const Dims& dims);

//! Build an RGB image from a alpha+RGB C array
template <class T>
Image<PixRGB<T> > fromARGB(const T* data, const Dims& dims);

//! Build an RGB image from a BGR C array
template <class T>
Image<PixRGB<T> > fromBGR(const T* data, const Dims& dims);

//! Build an RGB image from a BGR+alpha C array
template <class T>
Image<PixRGB<T> > fromBGRA(const T* data, const Dims& dims);

//! Build an RGB image from an RGB555 C array
Image<PixRGB<byte> > fromRGB555(const byte* data, const Dims& dims);

//! Build an RGB image from an RGB555 C array, byte-swapping in the process
Image<PixRGB<byte> > fromRGB555bs(const byte* data, const Dims& dims);

//! Build an RGB image from an RGB565 C array
Image<PixRGB<byte> > fromRGB565(const byte* data, const Dims& dims);

//! Build an RGB image from an RGB565 C array, byte-swapping in the process
Image<PixRGB<byte> > fromRGB565bs(const byte* data, const Dims& dims);

//! Build an RGB image from a YUV444 C array
template <class T>
Image<PixRGB<T> > fromYUV444(const T* data, const Dims& dims);

//! Build an RGB image from a YUV444 C array, byte-swapping in the process
template <class T>
Image<PixRGB<T> > fromYUV444bs(const T* data, const Dims& dims);

//! Build an RGB image from a YUV422 C array
template <class T>
Image<PixRGB<T> > fromYUV422(const T* data, const Dims& dims);

//! Fill pre-allocated YUV422 C arrays with values from an image
void toYUV422(const Image<PixRGB<byte> > img, byte* y, byte* u, byte* v,
              const int ystride = 0, const int ustride = 0,
              const int vstride = 0);

//! Build an RGB image from a YUV422 C array, byte-swapping in the process
template <class T>
Image<PixRGB<T> > fromYUV422bs(const T* data, const Dims& dims);

//! Build an RGB image from a YUV411 C array
template <class T>
Image<PixRGB<T> > fromYUV411(const T* data, const Dims& dims);

//! Build an RGB image from a YUV411 C array, byte-swapping in the process
template <class T>
Image<PixRGB<T> > fromYUV411bs(const T* data, const Dims& dims);

//! Build an RGB image from a YUV422-planar C array
template <class T>
Image<PixRGB<T> > fromYUV444P(const T* data, const Dims& dims);

//! Build an RGB image from a YUV422-planar C array
template <class T>
Image<PixRGB<T> > fromYUV422P(const T* data, const Dims& dims);

//! Build an RGB image from a YUV411-planar C array
template <class T>
Image<PixRGB<T> > fromYUV411P(const T* data, const Dims& dims);

//! Build an RGB image from a YUV420-planar C array
template <class T>
Image<PixRGB<T> > fromYUV420P(const T* data, const Dims& dims);

//! Build an RGB image from a YUV410-planar C array
template <class T>
Image<PixRGB<T> > fromYUV410P(const T* data, const Dims& dims);

//! Build an RGB image from a grayscale C array of size dims.w()*dims.h()
template <class T>
Image<PixRGB<T> > fromMono(const T* data, const Dims& dims);

//! Normalize my values by those in the lum image
template <class T>
Image<PixRGB<T> > luminanceNormalize(const Image<PixRGB<T> >& src,
                                     const T thresh);

//! Compute R/G and B/Y channels from a luminanceNormalized color image
/*! This version does automatic promotion on the results, so that
  negative values can be represented even if the input is of type
  PixRGB<byte> */
template <class T>
void getRGBY(const Image<PixRGB<T> >& src,
             Image<typename promote_trait<T, float>::TP>& rg,
             Image<typename promote_trait<T, float>::TP>& by, const T thresh);

//! Compute R/G and B/Y channels in a simpler way
/*! Params are as in getRGBY. The computation for the RG and BY maps is:
  RG = (R-G)/max(R,G,B); BY = (B-min(R,G))/max(R,G,B).*/
template <class T>
void getRGBYsimple(const Image<PixRGB<T> >& src,
                   Image<typename promote_trait<T, float>::TP>& rg,
                   Image<typename promote_trait<T, float>::TP>& by,
                   const T thresh);

//! Compute R, G, B, Y channels from a luminanceNormalized color image
/*! This version keeps the four channels separate so that correct
  results can be obtained without requiring promotion even if the
  input is of type PixRGB<byte> */
template <class T>
void getRGBY(const Image<PixRGB<T> >& src, Image<T>& r, Image<T>& g,
             Image<T>& b, Image<T>& y, const T thresh);

//! contrast-modulate an RGB image with a mask
Image< PixRGB<byte> > contrastModulate(const Image< PixRGB<byte> >& img,
                                       const Image<float>& mask,
                                       float baseContrast = 0.05,
                                       byte baseBright = 255);

//! Compute peak signal-to-noise ratio between two color images
/*! For color images, pSNR computed from the average
  mean-squared-error from the red, green and blue channels */
template <class T>
double pSNRcolor(const Image< PixRGB<T> >& img1,
                 const Image< PixRGB<T> >& img2);

//! Compute weighted peak signal-to-noise ratio between two color images
/*! This is like the other pSNRcolor() except that it uses the weighted
  version of distance() internally */
template <class T>
double pSNRcolor(const Image< PixRGB<T> >& img1,
                 const Image< PixRGB<T> >& img2,
                 const Image<float>& weight);

//! Normalize each component of the RGB image separately
/*! Normalize the red component with min.red() and max.red() etc.*/
template <class T>
Image< PixRGB<T> > normalizeRGB(const Image< PixRGB<T> >& img,
                                PixRGB<T> min,
                                PixRGB<T> max);

//! Return the pixel-wise maximum of the r, g and b component
template <class T>
Image<T> maxRGB(const Image< PixRGB<T> >& img);

//! Stain a black and white image with a color
/*!@param src the black and white image to be stained
  @param min This value will be converted to 0 (black)
  @param max This value will be converted to color
  @param color The color used for staining */
template <class T>
Image< PixRGB<T> > colorStain (const Image<T>& src,
                               const T& min, const T& max,
                               const PixRGB<T>& color);

//! convert RGB color values into C.I.E. coordinates
/*!@param rgbColor the color to convert from
  @param cr returns the R/I value
  @param cg returns the G/I value
  @param intens returns the intensity */
void RGBtoCIE(const PixRGB<byte>& rgbColor,
              float& cr, float& cg, float& intens);

//! distance between hue of image pixels and a given hue
/*!@param img The color image to compare
  @param muR the R/I coordinate of the target hue in the C.I.E. color space
  @param muG the G/I coordinate in C.I.E.
  @param sigR the standard deviation of a 2D Gaussian in R/I direction
  @param sigG the standard deviation of a 2D Gaussian in G/I direction
  @param rho the correlation between the R/I and G/I directions
  @return a float image coding for how close the hue in img is
  to the target hue - 1 for right on the mark, 0 for very different */
Image<float> hueDistance(const Image< PixRGB<byte> >& img,
                         float muR, float muG,
                         float sigR, float sigG, float rho);

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_ColorOps_templates.h"

#endif // !IMAGE_COLOROPS_H_DEFINED
