/*!@file Image_ColorOps.C Color operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_ColorOps.C 4239 2005-02-28 04:06:59Z mundhenk $
//

#ifndef IMAGE_COLOROPS_C_DEFINED
#define IMAGE_COLOROPS_C_DEFINED

#include "Image_ColorOps.H"
#include "MathFunctions.H"
#include "Image.H"
#include "Image_FilterOps.H"
#include "Image_ShapeOps.H"
#include "Image_Transforms.H"
#include "Image_MathOps.H"
#include "log.H"

#include <algorithm>
#include <cmath>

// ######################################################################
template <class T>
Image<PixRGB<T> > makeRGB(const Image<T>& red,
                          const Image<T>& green,
                          const Image<T>& blue)
{
  ASSERT(red.isSameSize(green) && red.isSameSize(blue));

  Image<PixRGB<T> > result(red.getDims(), NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = result.beginw();
  typename Image<PixRGB<T> >::iterator stop = result.endw();

  typename Image<T>::const_iterator rptr = red.begin();
  typename Image<T>::const_iterator gptr = green.begin();
  typename Image<T>::const_iterator bptr = blue.begin();

  while (aptr != stop)
    {
      *aptr++ = PixRGB<T>(*rptr++, *gptr++, *bptr++);
    }

  return result;
}

// ######################################################################
template <class T>
void getComponents(const Image<PixRGB<T> >& src,
                   Image<T>& red, Image<T>& green, Image<T>& blue)
{
  ASSERT(src.initialized());

  red = Image<T>(src.getDims(), NO_INIT);
  green = Image<T>(src.getDims(), NO_INIT);
  blue = Image<T>(src.getDims(), NO_INIT);

  typename Image<T>::iterator rptr = red.beginw();
  typename Image<T>::iterator gptr = green.beginw();
  typename Image<T>::iterator bptr = blue.beginw();

  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image<PixRGB<T> >::const_iterator stop = src.end();

  while (aptr != stop)
    {
      *rptr++ = aptr->red();
      *gptr++ = aptr->green();
      *bptr++ = aptr->blue();
      ++aptr;
    }
}

// ######################################################################
template <class T>
Image<PixRGB<T> > stain(const Image<T>& src, PixRGB<float> color)
{
  Image<PixRGB<T> > result(src.getDims(), NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<PixRGB<T> >::iterator dptr = result.beginw(),
    stop = result.endw();

  while (dptr != stop)
    {
      *dptr = color * (*sptr);
      ++sptr;
      ++dptr;
    }

  return result;
}

// ######################################################################
template <class T>
void getMinMaxC(const Image<PixRGB<T> >& src, T& mi, T& ma)
{
  ASSERT(src.initialized());
  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image<PixRGB<T> >::const_iterator stop = src.end();

  mi = aptr->red(); ma = aptr->red();

  while (aptr != stop)
    {
      T x = aptr->red();
      if (x < mi) mi = x; else if (x > ma) ma = x;

      x = aptr->green();
      if (x < mi) mi = x; else if (x > ma) ma = x;

      x = aptr->blue();
      if (x < mi) mi = x; else if (x > ma) ma = x;

      ++aptr;
    }
}

// ######################################################################
template <class T>
Image<T> luminance(const Image<PixRGB<T> >& src)
{
  ASSERT(src.initialized());

  Image<T> result(src.getDims(), NO_INIT);

  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();

  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();

  while (dptr != stop)
    *dptr++ = (*aptr++).luminance();

  return result;
}

// ######################################################################
template <class T>
Image<T> luminance(const Image<T>& src)
{ return src; }

// ######################################################################
template <class T>
Image< PixRGB<T> > toRGB(const Image<T>& src)
{
  ASSERT(src.initialized());

  Image< PixRGB<T> > result(src.getDims(), NO_INIT);
  typename Image<T>::const_iterator aptr = src.begin();
  typename Image< PixRGB<T> >::iterator dptr = result.beginw();
  typename Image< PixRGB<T> >::iterator stop = result.endw();

  while (dptr != stop) *dptr++ = PixRGB<T>(*aptr++);

  return result;
}

// ######################################################################
template <class T>
Image< PixRGB<T> > toRGB(const Image< PixRGB<T> >& src)
{ return src; }

// ######################################################################
template <class T>
Image<PixRGB<T> > warp3D(const Image<PixRGB<T> >& ima,
                         const Image<float>& zmap,
                         const float pitch, const float yaw,
                         const float zMax, Dims& dims)
{
  ASSERT(ima.isSameSize(zmap));

  float cp = float(cos(double(pitch) * M_PI / 180.0));
  float sp = float(sin(double(pitch) * M_PI / 180.0));
  float cy = float(cos(double(yaw) * M_PI / 180.0));
  float sy = float(sin(double(yaw) * M_PI / 180.0));
  int iw = ima.getWidth(), ih = ima.getHeight();

  // compute coords of bounding box to determine 3D image size
  int xmin = 0, zmin = 0, xmax = 0, zmax = 0, xx, zz;
  // (0, 0, 0) is implicitly done: yields (0,0)
  // (0, 0, zMax)
  zz = int(-cp * zMax);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (0,h,0)
  xx = int(sy * ih);
  xmin = std::min(xmin, xx);
  xmax = std::max(xmax, xx);
  zz = int(- cy * sp * ih);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (0, h, zMax)
  zz = int(- cy * sp * ih - cp * zMax);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (w, 0, 0)
  xx = int(cy * iw);
  xmin = std::min(xmin, xx);
  xmax = std::max(xmax, xx);
  zz = int(sy * sp * iw);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (w, 0, zMax)
  zz = int(sy * sp * iw - cp * zMax);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (w, h, 0)
  xx = int(cy * iw + sy * ih);
  xmin = std::min(xmin, xx);
  xmax = std::max(xmax, xx);
  zz = int(sy * sp * iw - cy * sp * ih);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);
  // (w, h, zMax)
  zz = int(sy * sp * iw - cy * sp * ih - cp * zMax);
  zmin = std::min(zmin, zz);
  zmax = std::max(zmax, zz);

  int nw, nh;
  if (dims.isEmpty())  // need to compute image size
    {
      nw = ((xmax - xmin) * 12) / 10;
      nh = ((zmax - zmin) * 16) / 10;  // height is oversized: place_image_3D
      //LDEBUG("image Width=%d, Height=%d", nw, nh);
      dims = Dims(nw, nh);
    }
  else  // image size known; just place the 3D in it
    {
      nw = dims.w(); nh = dims.h();
    }
  Image<PixRGB<T> > result(nw, nh, ZEROS);  // set background to black

  int xoff = -xmin + nw / 10;  // center image
  int zoff = -zmin + nh / 10;

  // store previously drawn points, for gap filling
  int *prev_zz = new int[nw]; memset(prev_zz, 0, nw * sizeof(int));
  int *prev_xx = new int[nw]; memset(prev_xx, 0, nw * sizeof(int));
  int ppzz = 0, ppxx = 0;
  for (int j = 0; j < ih; j ++)
    for (int i = 0; i < iw; i ++)
      {
        float x = float(i);
        float y = float(j);
        float z = - zmap.getVal(i, j);   // z-axis pointing downwards

        // rotate and project
        xx = xoff + int(cy * x + sy * y);
        zz = zoff + int(sy * sp * x - cy * sp * y + cp * z);
        xx = CLAMP(xx, 0, nw-1);
        zz = CLAMP(zz, 0, nh-1);

        // check if some gap-filling is needed due to high slopes
        PixRGB<T> pix, col;
        bool drawn = false;
        ima.getVal(i, j, col);  // color of current pixel
        if (j > 0 && zz > prev_zz[i] + 1)  // we are much lower than (i, j-1)
          {
            result.getVal(prev_xx[i], prev_zz[i], pix);  // save this pixel
            result.drawLine(Point2D(xx, zz),
                            Point2D(prev_xx[i], prev_zz[i]),
                            col);
            result.setVal(prev_xx[i], prev_zz[i], pix);
            drawn = true;
          }
        if (i > 0 && j > 0 && zz > ppzz + 1) // much lower than (i-1, j-1)
          {
            result.getVal(ppxx, ppzz, pix);  // save this pixel
            result.drawLine(Point2D(xx, zz),
                            Point2D(ppxx, ppzz),
                            col);
            result.setVal(ppxx, ppzz, pix);
            drawn = true;
          }
        if (i > 0 && (zz > prev_zz[i-1] + 1||   // much lower than (i-1, j)
                      zz < prev_zz[i-1] - 1))    // much higher than (i-1, j)
          {
            result.getVal(prev_xx[i-1], prev_zz[i-1], pix);  // save this pixel
            result.drawLine(Point2D(xx, zz),
                            Point2D(prev_xx[i-1], prev_zz[i-1]),
                            col);
            result.setVal(prev_xx[i-1], prev_zz[i-1], pix);
            drawn = true;
          }
        if (drawn == false)     // no line to draw, only one point
          result.setVal(xx, zz, col);

        ppzz = prev_zz[i]; ppxx = prev_xx[i];
        prev_zz[i] = zz; prev_xx[i] = xx;
      }
  delete [] prev_zz; delete [] prev_xx;

  return result;
}

// ######################################################################
template <class T>
Image<float> infoMeasure(const Image<PixRGB<T> >& src,
                         const float eps, const int size)
{
  ASSERT(src.initialized());

  Image< PixRGB<float> > float_copy(src);

  Image<float> y, i, q;
  getYIQ(float_copy, y, i, q);

  i = ::decXY(::lowPass3(i));
  q = ::decXY(::lowPass3(q));

  // eps is given between 0 and 1; modify it according to image size
  // we want that eps=0.01 corresponds to a cosine with amplitude 127.5*0.01
  float epsy = eps * ((float)(size*size) * 127.5);
  float epsiq = epsy / 4.0;

  Image<float> result(src.getWidth() / size, src.getHeight() / size, NO_INIT);

  for (int offy = 0; offy < src.getHeight(); offy += size)
    for (int offx = 0; offx < src.getWidth(); offx += size)
      {
        Image<float> ydct = dct(y, offx, offy, size);
        Image<float> idct = dct(i, offx/2, offy/2, size/2);
        Image<float> qdct = dct(q, offx/2, offy/2, size/2);

        result.setVal(offx/size, offy/size,
                      (float(ydct.countThresh(epsy)) +
                       4.0 * float(idct.countThresh(epsiq)) +
                       4.0 * float(qdct.countThresh(epsiq)))
                      / (3.0 * float(size*size)));
      }

  return result;
}

// ######################################################################
template <class T> inline
void getYIQ(const Image<PixRGB<T> >& src,
            Image<T>& y, Image<T>& i, Image<T>& q)
{
  ASSERT(src.initialized());
  y = Image<T>(src.getDims(), NO_INIT);
  i = Image<T>(src.getDims(), NO_INIT);
  q = Image<T>(src.getDims(), NO_INIT);

  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image<PixRGB<T> >::const_iterator stop = src.end();

  typename Image<T>::iterator yptr = y.beginw();
  typename Image<T>::iterator iptr = i.beginw();
  typename Image<T>::iterator qptr = q.beginw();

  while (aptr != stop)
    {
      T yy, ii, qq;
      (*aptr++).getYIQ(yy, ii, qq);
      *yptr++ = yy;
      *iptr++ = ii;
      *qptr++ = qq;
    }
}

// ######################################################################
template <class T> inline
void getYUV(const Image<PixRGB<T> >& src,
            Image<T>& y, Image<T>& u, Image<T>& v)
{
  ASSERT(src.initialized());
  y = Image<T>(src.getDims(), NO_INIT);
  u = Image<T>(src.getDims(), NO_INIT);
  v = Image<T>(src.getDims(), NO_INIT);

  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image<PixRGB<T> >::const_iterator stop = src.end();

  typename Image<T>::iterator yptr = y.beginw();
  typename Image<T>::iterator uptr = u.beginw();
  typename Image<T>::iterator vptr = v.beginw();

  T yy, uu, vv;
  while (aptr != stop)
    {
      (*aptr++).getYUV(yy, uu, vv);
      *yptr++ = yy; *uptr++ = uu; *vptr++ = vv;
    }
}

// ######################################################################
template <class T> inline
Image< PixRGB<T> > setYUV(const Image<T>& y, const Image<T>& u,
                          const Image<T>& v)
{
  ASSERT(y.isSameSize(u) && y.isSameSize(v));
  Image< PixRGB<T> > dest(y.getDims(), NO_INIT);

  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  typename Image<T>::const_iterator yptr = y.begin();
  typename Image<T>::const_iterator uptr = u.begin();
  typename Image<T>::const_iterator vptr = v.begin();

  T yy, uu, vv;
  while (aptr != stop)
    {
      yy = *yptr++; uu = *uptr++; vv = *vptr++;
      (*aptr++).setYUV(yy, uu, vv);
    }
  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromRGB(const T* data, const Dims& dims)
{
  // dirty hack: assumes that we know the internal representation of PixRGB!
  Image< PixRGB<T> > dest(dims, NO_INIT);
  memcpy(dest.getArrayPtr(), data, dims.w() * dims.h() * 3);
  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromARGB(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: alpha, r, g, b
      (*aptr++).set(data[1], data[2], data[3]);
      data += 4;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromBGR(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: b, g, r
      (*aptr++).set(data[2], data[1], data[0]);
      data += 3;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromBGRA(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: b, g, r, alpha
      (*aptr++).set(data[2], data[1], data[0]);
      data += 4;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV444(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: u0, y0, v0, u1, y1, v1
      (*aptr++).setYUV(data[1], data[0], data[2]);
      (*aptr++).setYUV(data[4], data[3], data[5]);
      data += 6;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV444bs(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: y0, u0, u1, v0, v1, y1
      (*aptr++).setYUV(data[0], data[1], data[3]);
      (*aptr++).setYUV(data[5], data[2], data[4]);
      data += 6;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV422(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: u, y0, v, y1
      // Here we bypass the setYUV() function of Pixels.H, for max speed
      
      (*aptr++).setYUV(data[1], data[0], data[2]);
      (*aptr++).setYUV(data[3], data[0], data[2]);
      data += 4;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV422bs(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as:  y0, u, y1, v
      (*aptr++).setYUV(data[0], data[1], data[3]);
      (*aptr++).setYUV(data[2], data[1], data[3]);
      data += 4;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV411(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: u, y0, y1, v, y2, y3
      (*aptr++).setYUV(data[1], data[0], data[3]);
      (*aptr++).setYUV(data[2], data[0], data[3]);
      (*aptr++).setYUV(data[4], data[0], data[3]);
      (*aptr++).setYUV(data[5], data[0], data[3]);
      data += 6;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV411bs(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  while(aptr != stop)
    {
      // data stored as: y0, u, v, y1, y3, y2
      (*aptr++).setYUV(data[0], data[1], data[2]);
      (*aptr++).setYUV(data[3], data[1], data[2]);
      (*aptr++).setYUV(data[5], data[1], data[2]);
      (*aptr++).setYUV(data[4], data[1], data[2]);
      data += 6;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV444P(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  int w = dims.w(), h = dims.h();
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  const T *yptr = data;
  const T *uptr = yptr + w * h;
  const T *vptr = uptr + w * h;

  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i ++)
      {
        // we have a 1 luminance pixels per chroma pair
        (*aptr++).setYUV(*yptr++, *uptr++, *vptr++);
      }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV422P(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  int w = dims.w(), h = dims.h();
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  const T *yptr = data;
  const T *uptr = yptr + w * h;
  const T *vptr = uptr + (w/2) * h;

  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i += 2)
      {
        // we have a 2 luminance pixels per chroma pair
        (*aptr++).setYUV(*yptr++, *uptr, *vptr);
        (*aptr++).setYUV(*yptr++, *uptr++, *vptr++);
      }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV411P(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  int w = dims.w(), h = dims.h();
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  const T *yptr = data;
  const T *uptr = yptr + w * h;
  const T *vptr = uptr + (w/4) * h;

  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i += 4)
      {
        // we have a 4 luminance pixels per chroma pair
        (*aptr++).setYUV(*yptr++, *uptr, *vptr);
        (*aptr++).setYUV(*yptr++, *uptr, *vptr);
        (*aptr++).setYUV(*yptr++, *uptr, *vptr);
        (*aptr++).setYUV(*yptr++, *uptr++, *vptr++);
      }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV420P(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  int w = dims.w(), h = dims.h();
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  const T *yptr = data;
  const T *uptr = yptr + w * h;
  const T *vptr = uptr + (w/2) * (h/2);

  for (int j = 0; j < h; j += 2)
    {
      for (int i = 0; i < w; i += 2)
        {
          T u = *uptr++, v = *vptr++;
          // we have a 2x2 luminance block per chroma pair
          (*aptr).setYUV(*yptr, u, v);
          aptr[1].setYUV(yptr[1], u, v);
          aptr[w].setYUV(yptr[w], u, v);
          aptr[w+1].setYUV(yptr[w + 1], u, v);
          aptr += 2; yptr += 2;
        }
      aptr += w; yptr += w;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromYUV410P(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  int w = dims.w(), h = dims.h(); int w2 = w * 2, w3 = w * 3;
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  const T *yptr = data;
  const T *uptr = yptr + w * h;
  const T *vptr = uptr + (w/4) * (h/4);

  for (int j = 0; j < h; j += 4)
    {
      for (int i = 0; i < w; i += 4)
        {
          T u = *uptr++, v = *vptr++;

          // we have a 4x4 luminance block per chroma pair
          (*aptr).setYUV(*yptr, u, v);
          aptr[1].setYUV(yptr[1], u, v);
          aptr[2].setYUV(yptr[2], u, v);
          aptr[3].setYUV(yptr[3], u, v);

          aptr[w].setYUV(yptr[w], u, v);
          aptr[w+1].setYUV(yptr[w+1], u, v);
          aptr[w+2].setYUV(yptr[w+2], u, v);
          aptr[w+3].setYUV(yptr[w+3], u, v);

          aptr[w2].setYUV(yptr[w2], u, v);
          aptr[w2+1].setYUV(yptr[w2+1], u, v);
          aptr[w2+2].setYUV(yptr[w2+2], u, v);
          aptr[w2+3].setYUV(yptr[w2+3], u, v);

          aptr[w3].setYUV(yptr[w3], u, v);
          aptr[w3+1].setYUV(yptr[w3+1], u, v);
          aptr[w3+2].setYUV(yptr[w3+2], u, v);
          aptr[w3+3].setYUV(yptr[w3+3], u, v);

          aptr += 4; yptr += 4;
        }
      aptr += w3; yptr += w3;
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > fromMono(const T* data, const Dims& dims)
{
  Image< PixRGB<T> > dest(dims, NO_INIT);
  typename Image<PixRGB<T> >::iterator aptr = dest.beginw();
  typename Image<PixRGB<T> >::iterator stop = dest.endw();

  T m;
  while(aptr != stop)
    {
      m = (*data++); (*aptr++).set(m, m, m);
    }

  return dest;
}

// ######################################################################
template <class T>
Image<PixRGB<T> > luminanceNormalize(const Image<PixRGB<T> >& src,
                                     const T thresh)
{
  ASSERT(src.initialized());

  Image<PixRGB<T> > result(src.getDims(), NO_INIT);

  typename Image<PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image<PixRGB<T> >::const_iterator stop = src.end();

  typename Image<PixRGB<T> >::iterator dptr = result.beginw();
  PixRGB<T> zero = T(0);

  while (aptr != stop)
    {
      float v = float(aptr->luminance());
      if (v >= thresh)
        {
          const float fac = 255.0 / (v * 3.0);
          dptr->setRed  ( T(float(aptr->red()) * fac) );
          dptr->setGreen( T(float(aptr->green()) * fac) );
          dptr->setBlue ( T(float(aptr->blue()) * fac) );
        }
      else
        *dptr = zero;

      ++aptr; ++dptr;
    }

  return result;
}

// ######################################################################
template <class T>
void getRGBY(const Image<PixRGB<T> >& src,
             Image<typename promote_trait<T, float>::TP>& rg,
             Image<typename promote_trait<T, float>::TP>& by, const T thresh)
{
  // red = [r - (g+b)/2]        [.] = clamp between 0 and 255
  // green = [g - (r+b)/2]
  // blue = [b - (r+g)/2]
  // yellow = [2*((r+g)/2 - |r-g| - b)]

  ASSERT(src.initialized());
  typedef typename promote_trait<T, float>::TP TF;
  rg = Image<TF>(src.getDims(), NO_INIT);
  by = Image<TF>(src.getDims(), NO_INIT);

  typename Image< PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image< PixRGB<T> >::const_iterator stop = src.end();

  typename Image<TF>::iterator rgptr = rg.beginw(), byptr = by.beginw();
  TF zero = TF(), threshf = TF(thresh);

  while (aptr != stop)
    {
      TF r = aptr->red(), g = aptr->green(), b = aptr->blue();

      // first do the luminanceNormalization:
      TF lum = (r + g + b) / 3.0f;
      if (lum < threshf)  // too dark... no response from anybody
        { *rgptr++ = zero; *byptr++ = zero; ++aptr; }
      else
        {
          // normalize chroma by luminance:
          TF fac = 255.0f / (3.0f * lum);
          r *= fac; g *= fac; b *= fac;

          // now compute color opponencies:
          // yellow gets a factor 2 to compensate for its previous attenuation
          // by luminanceNormalize():
          TF red = r - 0.5f * (g + b), green = g - 0.5f * (r + b);
          TF blue = b - 0.5f * (r + g), yellow = -2.0f * (blue + fabs(r-g));

          if (red < 0.0f) red = 0.0f;
          else if (red > 255.0f) red = 255.0f;
          if (green < 0.0f) green = 0.0f;
          else if (green > 255.0f) green = 255.0f;
          if (blue < 0.0f) blue = 0.0f;
          else if (blue > 255.0f) blue = 255.0f;
          if (yellow < 0.0f) yellow=0.0f;
          else if (yellow > 255.0f) yellow=255.0f;

          *rgptr++ = red - green; *byptr++ = blue - yellow;
          ++aptr;
        }
    }
}

// ######################################################################
template <class T>
void getRGBYsimple(const Image<PixRGB<T> >& src,
                   Image<typename promote_trait<T, float>::TP>& rg,
                   Image<typename promote_trait<T, float>::TP>& by, 
                   const T thresh)
{
  // I = max(R,G,B)
  // RG = (R-G)/I
  // BY = (B-min(R,G))/I
  // Dirk Walther, September 2004
  ASSERT(src.initialized());
  typedef typename promote_trait<T, float>::TP TF;
  rg = Image<TF>(src.getDims(), NO_INIT);
  by = Image<TF>(src.getDims(), NO_INIT);

  typename Image< PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image< PixRGB<T> >::const_iterator stop = src.end();

  typename Image<TF>::iterator rgptr = rg.beginw(), byptr = by.beginw();
  TF zero = TF(), threshf = TF(thresh);

  while (aptr != stop)
    {
      TF r = aptr->red(), g = aptr->green(), b = aptr->blue();

      // first do in = max(R,G,B):
      TF in = r;
      if (g > in) in = g;
      if (b > in) in = b;
      if (in < threshf)  // too dark... no response from anybody
        { *rgptr++ = zero; *byptr++ = zero; ++aptr; }
      else
        {
          TF y = (r < g) ? r : g;
          *rgptr++ = (r - g) / in;
          *byptr++ = (b - y) / in;
          ++aptr;
        }
    }
}

// ######################################################################
template <class T>
void getRGBY(const Image<PixRGB<T> >& src, Image<T>& rr, Image<T>& gg,
             Image<T>& bb, Image<T>& yy, const T thresh)
{
  // this is essentially the same code as the other getRGBY, just
  // duplicated for efficiency
  ASSERT(src.initialized());
  rr = Image<T>(src.getDims(), NO_INIT);
  gg = Image<T>(src.getDims(), NO_INIT);
  bb = Image<T>(src.getDims(), NO_INIT);
  yy = Image<T>(src.getDims(), NO_INIT);

  typename Image< PixRGB<T> >::const_iterator aptr = src.begin();
  typename Image< PixRGB<T> >::const_iterator stop = src.end();
  typename Image<T>::iterator rptr = rr.beginw(), gptr = gg.beginw(),
    bptr = bb.beginw(), yptr = yy.beginw();

  T zero = T(); float threshf = float(thresh);

  while (aptr != stop)
    {
      float r = aptr->red(), g = aptr->green(), b = aptr->blue();

      // first do the luminanceNormalization:
      float lum = (r + g + b) / 3.0f;
      if (lum < threshf)  // too dark... no response from anybody
        { *rptr++=zero; *gptr++=zero; *bptr++=zero; *yptr++=zero; ++aptr; }
      else
        {
          // normalize chroma by luminance:
          float fac = 255.0f / (3.0f * lum);
          r *= fac; g *= fac; b *= fac;

          // now compute color opponencies:
          // yellow gets a factor 2 to compensate for its previous attenuation
          // by luminanceNormalize():
          float red = r - 0.5f * (g + b), green = g - 0.5f * (r + b);
          float blue = b - 0.5f * (r + g), yellow = -2.0f * (blue + fabs(r-g));

          if (red < 0.0f) red = 0.0f;
          else if (red > 255.0f) red = 255.0f;
          if (green < 0.0f) green = 0.0f;
          else if (green > 255.0f) green = 255.0f;
          if (blue < 0.0f) blue = 0.0f;
          else if (blue > 255.0f) blue = 255.0f;
          if (yellow < 0.0f) yellow=0.0f;
          else if (yellow > 255.0f) yellow=255.0f;

          *rptr++ = T(red);     // no clamping necessary
          *gptr++ = T(green);
          *bptr++ = T(blue);
          *yptr++ = T(yellow);

          ++aptr;
        }
    }
}

// ######################################################################
template <class T>
double pSNRcolor(const Image< PixRGB<T> >& img1,
                 const Image< PixRGB<T> >& img2)
{
  ASSERT(img1.isSameSize(img2));
  Image<T> r1, g1, b1, r2, g2, b2; int siz = img1.getSize();
  getComponents(img1, r1, g1, b1); getComponents(img2, r2, g2, b2);
  double mser = distance(r1, r2); mser = mser * mser / double(siz);
  double mseg = distance(g1, g2); mseg = mseg * mseg / double(siz);
  double mseb = distance(b1, b2); mseb = mseb * mseb / double(siz);

  double mse = (mser + mseg + mseb) / 3.0;

  return 10.0 * log10(255.0 * 255.0 / mse);
}

// ######################################################################
template <class T>
double pSNRcolor(const Image< PixRGB<T> >& img1,
                 const Image< PixRGB<T> >& img2,
                 const Image<float>& weight)
{
  ASSERT(img1.isSameSize(img2));
  Image<T> r1, g1, b1, r2, g2, b2; int siz = img1.getSize();
  getComponents(img1, r1, g1, b1); getComponents(img2, r2, g2, b2);
  double mser = distance(r1, r2, weight); mser = mser * mser / double(siz);
  double mseg = distance(g1, g2, weight); mseg = mseg * mseg / double(siz);
  double mseb = distance(b1, b2, weight); mseb = mseb * mseb / double(siz);

  double mse = (mser + mseg + mseb) / 3.0;

  return 10.0 * log10(255.0 * 255.0 / mse);
}

// ######################################################################
template <class T>
Image< PixRGB<T> > normalizeRGB(const Image< PixRGB<T> >& img, 
                                PixRGB<T> min,
                                PixRGB<T> max)
{
  Image<T> r,g,b;
  getComponents(img, r, g, b);
  r.normalize(min.red(),   max.red());
  g.normalize(min.green(), max.green());
  b.normalize(min.blue(),  max.blue());
  return makeRGB(r, g, b);
}

// ######################################################################
template <class T>
Image<T> maxRGB(const Image< PixRGB<T> >& img)
{
  Image<T> r,g,b;
  getComponents(img, r, g, b);
  return takeMax(r, takeMax(g, b));
}

// ######################################################################
template <class T>
Image< PixRGB<T> > colorStain (const Image<T>& src, 
                               const T& min, const T& max,
                               const PixRGB<T>& color)
{
  float fmin = (float)min;
  float fdiff = (float)(max - min);
  ASSERT (fdiff != 0.0F);

  Image < PixRGB<T> > result(src.getDims(), NO_INIT);
  typename Image<T>::const_iterator sptr, stop = src.end();
  //typename Image< PixRGB<T> >::const_iterator stop = src.end();
  typename Image< PixRGB<T> >::iterator rptr = result.beginw();

  for (sptr = src.begin(); sptr != stop; ++sptr, ++rptr)
    {
      *rptr = color;
      *rptr *= ((float)(*sptr) - fmin) / fdiff;
    }
  return result;
}


// Include the explicit instantiations
// MA #include "inst/Image_ColorOps.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_COLOROPS_C_DEFINED
