/*!@file Image_FilterOps.C Filtering operations on Image */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_FilterOps.C 4290 2005-03-12 02:03:27Z mundhenk $
//

#include "Image_FilterOps.H"
#include "MathFunctions.H"
#include "Image.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"
#include "log.H"
#include "Promotions.H"
#include "Raster.H"

#include <algorithm>
#include <cmath>
#include <limits>
#include <string>
#include <iostream>

using namespace std;

namespace
{
  Image<PixRGB<float> > lowPass3x_rgb(const Image<PixRGB<float> >& src)
  {
    const int w = src.getWidth(), h = src.getHeight();
    if (w < 2) return src; // nothing to smooth
    Image<PixRGB<float> > result(w, h, NO_INIT);
    Image<PixRGB<float> >::const_iterator sptr = src.begin();
    Image<PixRGB<float> >::iterator dptr = result.beginw();

    // *** unfolded version (all particular cases treated) for max speed.
    // do not use access overload since can access the C array directly.
    // the bet is: float computations are faster than int (true for most
    // float copros), and int2float converts fast->do not have to divide
    // by the coeff sum since we directly use float coeffs.
    // notations: in () is the position of dest ptr, and ^ is src ptr
    // ########## horizontal pass
    for (int j = 0; j < h; j ++)
      {
        // leftmost point  [ (2^) 1 ] / 3
        dptr->r = sptr[0].r * (2.0F / 3.0F) + sptr[1].r * (1.0F / 3.0F);
        dptr->g = sptr[0].g * (2.0F / 3.0F) + sptr[1].g * (1.0F / 3.0F);
        dptr->b = sptr[0].b * (2.0F / 3.0F) + sptr[1].b * (1.0F / 3.0F);
        ++dptr;

        // rest of the line except last point  [ 1^ (2) 1 ] / 4
        for (int i = 0; i < w - 2; i ++)
          {
            dptr->r = (sptr[0].r + sptr[2].r) * 0.25F + sptr[1].r * 0.5F;
            dptr->g = (sptr[0].g + sptr[2].g) * 0.25F + sptr[1].g * 0.5F;
            dptr->b = (sptr[0].b + sptr[2].b) * 0.25F + sptr[1].b * 0.5F;
            ++dptr;
            ++sptr;
          }

        // last point [ 1^ (2) ] / 3
        dptr->r = sptr[0].r * (1.0F / 3.0f) + sptr[1].r * (2.0F / 3.0F);
        dptr->g = sptr[0].g * (1.0F / 3.0f) + sptr[1].g * (2.0F / 3.0F);
        dptr->b = sptr[0].b * (1.0F / 3.0f) + sptr[1].b * (2.0F / 3.0F);
        ++dptr;
        sptr += 2;  // sptr back to same position as dptr
      }
    return result;
  }

  Image<PixRGB<float> > lowPass3y_rgb(const Image<PixRGB<float> >& src)
  {
    const int w = src.getWidth(), h = src.getHeight();
    if (h < 2) return src; // nothing to smooth
    Image<PixRGB<float> > result(w, h, NO_INIT);
    Image<PixRGB<float> >::const_iterator sptr = src.begin();
    Image<PixRGB<float> >::iterator dptr = result.beginw();
    const int w2 = w * 2; // speedup

    // ########## vertical pass  (even though we scan horiz for speedup)
    // topmost points  ( [ (2^) 1 ] / 3 )^T
    for (int i = 0; i < w; i ++)
      {
        //*dptr = sptr[0] * (2.0F / 3.0F) + sptr[w] * (1.0F / 3.0F);
        dptr->r = sptr[0].r * (2.0F / 3.0F) + sptr[w].r * (1.0F / 3.0F);
        dptr->g = sptr[0].g * (2.0F / 3.0F) + sptr[w].g * (1.0F / 3.0F);
        dptr->b = sptr[0].b * (2.0F / 3.0F) + sptr[w].b * (1.0F / 3.0F);
        ++dptr;
        ++sptr;
      }
    sptr -= w;  // go back to top-left

    // rest of the column except last point ( [ 1^ (2) 1 ] / 4.0 )^T
    for (int j = 0; j < h - 2; j ++)
      for (int i = 0; i < w; i ++)
        {
          //*dptr = (sptr[0] + sptr[w2]) * 0.25F + sptr[w] * 0.5F;
          dptr->r = (sptr[0].r + sptr[w2].r) * 0.25F + sptr[w].r * 0.5F;
          dptr->g = (sptr[0].g + sptr[w2].g) * 0.25F + sptr[w].g * 0.5F;
          dptr->b = (sptr[0].b + sptr[w2].b) * 0.25F + sptr[w].b * 0.5F;
          ++dptr;
          ++sptr;
        }

    // last points ( [ 1^ (2) ] / 3 )^T
    for (int i = 0; i < w; i ++)
      {
        //*dptr = sptr[0] * (1.0F / 3.0F) + sptr[w] * (2.0F / 3.0F);
        dptr->r = sptr[0].r * (1.0F / 3.0F) + sptr[w].r * (2.0F / 3.0F);
        dptr->g = sptr[0].g * (1.0F / 3.0F) + sptr[w].g * (2.0F / 3.0F);
        dptr->b = sptr[0].b * (1.0F / 3.0F) + sptr[w].b * (2.0F / 3.0F);
        ++dptr;
        ++sptr;
      }

    // finished. here, sptr is one line before the end of array.
    return result;
  }
}

// ######################################################################
template <>
Image<promote_trait<PixRGB<float>, float>::TP>
lowPass3x(const Image<PixRGB<float> >& src)
{
  return lowPass3x_rgb(src);
}

// ######################################################################
template <>
Image<promote_trait<PixRGB<float>, float>::TP>
lowPass3y(const Image<PixRGB<float> >& src)
{
  return lowPass3y_rgb(src);
}

// ######################################################################
// Produces a Gabor kernel with optionally unequal major+minor axis lengths.
Image<float> gaborFilter3(const float major_stddev, const float minor_stddev,
                          const float period, const float phase,
                          const float theta, int size)
{
  const float max_stddev =
    major_stddev > minor_stddev ? major_stddev : minor_stddev;

  // figure the proper size for the result
  if (size == -1) size = int(ceil(max_stddev * sqrt(-2.0F * log(exp(-5.0F)))));
  else size = size/2;
  //const int size = int(ceil(max_stddev * 2));

  Image<float> result(2 * size + 1, 2 * size + 1, NO_INIT);

  // change the angles in degree to the those in radians:
  const float psi = M_PI / 180.0F * phase;
  const float rtDeg = M_PI / 180.0F * theta;

  // calculate constants:
  const float omega = (2.0F * M_PI) / period;
  const float co = cos(rtDeg), si = sin(rtDeg);
  const float major_sigq = 2.0F * major_stddev * major_stddev;
  const float minor_sigq = 2.0F * minor_stddev * minor_stddev;
  Image<float>::iterator aptr = result.beginw();

  // compute gabor:
  for (int y = -size; y <= size; ++y)
    for (int x = -size; x <= size; ++x)
      {
        const float major = x*co + y*si;
        const float minor = x*si - y*co;
        *aptr++ = float(cos(omega * major + psi)
                        * exp(-(major*major) / major_sigq)
                      * exp(-(minor*minor) / minor_sigq));
      }

  // make the result have (mean == 0.0)
  return (result - mean(result));
}

// ######################################################################
namespace
{
  // Helper function for binomialKernel():
  double binomial(int n, int k)
  {
    /*
      binomial coefficient refresher:

           k:    0   1   2   3   4   5   6   7   8
         N:  +------------------------------------
         0   |   1
         1   |   1   1
         2   |   1   2   1
         3   |   1   3   3   1
         4   |   1   4   6   4   1
         5   |   1   5  10  10   5   1
         6   |   1   6  15  20  15   6   1
         7   |   1   7  21  35  35  21   7   1
         8   |   1   8  28  56  70  56  28   8   1

      n-choose-k := n! / k! (n-k)!

                 == n*(n-1)*...(n-(k-1)) / k*(k-1)*...*1

      gamma(n+1) == n! for integral n

      lgamma(x) := ln(gamma(x))

      By doing the factorial multiplications as additions in the
      log-domain, we avoid numerical overflow in intermediate results.
    */
    return round(exp(lgamma(n+1) - lgamma(k+1) - lgamma(n-k+1)));
  }
}

Image<float> binomialKernel(const int sz)
{
  ASSERT(sz > 0);

  Image<float> result(sz, 1, NO_INIT);

  const int N = sz-1;

  const double div = double(1 << N);

  // compute a series of N-choose-k, where N == (sz-1), with k running
  // from 0 to N, inclusive
  for (int k = 0; k <= N; ++k)
    {
      result.setVal(k, 0, float(binomial(N, k) / div));
    }

  return result;
}

// ######################################################################
Image<byte> ones(Dims d)
{
  Image<byte> result(d,NO_INIT);
  result.clear(255);
  return result;
}

// ######################################################################
Image<byte> ones(int w, int h)
{
  return ones(Dims(w,h));
}

// ######################################################################
Image<byte> ones(int w)
{
  return ones(Dims(w,w));
}

// ######################################################################
Image< promote_trait< float, float >::TP >
lowPass9x(const Image<float>& src)
{
  //cout << "lowPass9x(const Image<float>&)" << endl;
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef promote_trait<float, float>::TP TF;
  const Image<TF> source = src;
  if (w < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  Image<TF>::const_iterator sptr = source.begin();
  Image<TF>::iterator dptr = result.beginw();

  if (w < 9)  // use inefficient implementation for small images
    {
      float kernel[9] = { 1.0F / 256.0F, 8.0F / 256.0F, 28.0F / 256.0F,
                          56.0F / 256.0F, 70.0F / 256.0F, 56.0F / 256.0F,
                          28.0F / 256.0F, 8.0F / 256.0F, 1.0F / 256.0F };
       return xFilterClean(src, kernel, 9);
    }

  // boundary conditions: truncated filter
  for (int j = 0; j < h; j ++)
    {
      // leftmost points
      *dptr++ = sptr[0] * (70.0F / 163.0F) +
        sptr[1] * (56.0F / 163.0F) +
        sptr[2] * (28.0F / 163.0F) +
        sptr[3] * ( 8.0F / 163.0F) +
        sptr[4] * ( 1.0F / 163.0F);
      *dptr++ = (sptr[0] + sptr[2]) * (56.0F / 219.0F) +
        sptr[1] * (70.0F / 219.0F) +
        sptr[3] * (28.0F / 219.0F) +
        sptr[4] * ( 8.0F / 219.0F) +
        sptr[5] * ( 1.0F / 219.0F);
      *dptr++ = (sptr[0] + sptr[4]) * (28.0F / 247.0F) +
        (sptr[1] + sptr[3]) * (56.0F / 247.0F) +
        sptr[2] * (70.0F / 247.0F) +
        sptr[5] * ( 8.0F / 247.0F) +
        sptr[6] * ( 1.0F / 247.0F);
      *dptr++ = (sptr[0] + sptr[6]) * ( 8.0F / 255.0F) +
        (sptr[1] + sptr[5]) * (28.0F / 255.0F) +
        (sptr[2] + sptr[4]) * (56.0F / 255.0F) +
        sptr[3] * (70.0F / 255.0F) +
        sptr[7] * ( 1.0F / 255.0F);

      // far from the borders
      for (int i = 0; i < w - 8; i ++)
        {
          *dptr++ = (sptr[0] + sptr[8]) * ( 1.0F / 256.0F) +
            (sptr[1] + sptr[7]) * ( 8.0F / 256.0F) +
            (sptr[2] + sptr[6]) * (28.0F / 256.0F) +
            (sptr[3] + sptr[5]) * (56.0F / 256.0F) +
            sptr[4] * (70.0F / 256.0F);
          sptr ++;
        }

      // rightmost points
      *dptr++ = sptr[0] * ( 1.0F / 255.0F) +
        (sptr[1] + sptr[7]) * ( 8.0F / 255.0F) +
        (sptr[2] + sptr[6]) * (28.0F / 255.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 255.0F) +
        sptr[4] * (70.0F / 255.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 247.0F) +
        sptr[1] * ( 8.0F / 247.0F) +
        (sptr[2] + sptr[6]) * (28.0F / 247.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 247.0F) +
        sptr[4] * (70.0F / 247.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 219.0F) +
        sptr[1] * ( 8.0F / 219.0F) +
        sptr[2] * (28.0F / 219.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 219.0F) +
        sptr[4] * (70.0F / 219.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 163.0F) +
        sptr[1] * ( 8.0F / 163.0F) +
        sptr[2] * (28.0F / 163.0F) +
        sptr[3] * (56.0F / 163.0F) +
        sptr[4] * (70.0F / 163.0F);
      sptr += 5;  // sptr back to same as dptr (start of next line)
    }
  return result;
}

// ######################################################################
Image< promote_trait< float, float >::TP >
lowPass9y(const Image<float>& src)
{
  //cout << "lowPass9y(const Image<float>&)" << endl;

  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef promote_trait<float, float>::TP TF;
  const Image<TF> source = src;
  if (h < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  Image<TF>::const_iterator sptr = source.begin();
  Image<TF>::iterator dptr = result.beginw();

  if (h < 9)  // use inefficient implementation for small images
    {
      float kernel[9] = { 1.0F / 256.0F, 8.0F / 256.0F, 28.0F / 256.0F,
                          56.0F / 256.0F, 70.0F / 256.0F, 56.0F / 256.0F,
                          28.0F / 256.0F, 8.0F / 256.0F, 1.0F / 256.0F };
      return yFilterClean(src, kernel, 9);
    }

  // *** vertical pass ***
  const int w2 = w + w, w3 = w2 + w, w4 = w3 + w, w5 = w4 + w, w6 = w5 + w,
    w7 = w6 + w,  w8 = w7 + w;  // index computation speedup
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * (70.0F / 163.0F) +
        sptr[ w] * (56.0F / 163.0F) +
        sptr[w2] * (28.0F / 163.0F) +
        sptr[w3] * ( 8.0F / 163.0F) +
        sptr[w4] * ( 1.0F / 163.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w2]) * (56.0F / 219.0F) +
        sptr[ w] * (70.0F / 219.0F) +
        sptr[w3] * (28.0F / 219.0F) +
        sptr[w4] * ( 8.0F / 219.0F) +
        sptr[w5] * ( 1.0F / 219.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w4]) * (28.0F / 247.0F) +
        (sptr[ w] + sptr[w3]) * (56.0F / 247.0F) +
        sptr[w2] * (70.0F / 247.0F) +
        sptr[w5] * ( 8.0F / 247.0F) +
        sptr[w6] * ( 1.0F / 247.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w6]) * ( 8.0F / 255.0F) +
        (sptr[ w] + sptr[w5]) * (28.0F / 255.0F) +
        (sptr[w2] + sptr[w4]) * (56.0F / 255.0F) +
        sptr[w3] * (70.0F / 255.0F) +
        sptr[w7] * ( 1.0F / 255.0F);
      sptr ++;
    }
  sptr -= w;   // back to top-left
  for (int j = 0; j < h - 8; j ++)
    for (int i = 0; i < w; i ++)
      {
        *dptr++ = (sptr[ 0] + sptr[w8]) * ( 1.0F / 256.0F) +
          (sptr[ w] + sptr[w7]) * ( 8.0F / 256.0F) +
          (sptr[w2] + sptr[w6]) * (28.0F / 256.0F) +
          (sptr[w3] + sptr[w5]) * (56.0F / 256.0F) +
          sptr[w4]  * (70.0F / 256.0F);
        sptr ++;
      }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 255.0F) +
        (sptr[ w] + sptr[w7]) * ( 8.0F / 255.0F) +
        (sptr[w2] + sptr[w6]) * (28.0F / 255.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 255.0F) +
        sptr[w4] * (70.0F / 255.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 247.0F) +
        sptr[ w] * ( 8.0F / 247.0F) +
        (sptr[w2] + sptr[w6]) * (28.0F / 247.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 247.0F) +
        sptr[w4] * (70.0F / 247.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 219.0F) +
        sptr[ w] * ( 8.0F / 219.0F) +
        sptr[w2] * (28.0F / 219.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 219.0F) +
        sptr[w4] * (70.0F / 219.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 163.0F) +
        sptr[ w] * ( 8.0F / 163.0F) +
        sptr[w2] * (28.0F / 163.0F) +
        sptr[w3] * (56.0F / 163.0F) +
        sptr[w4] * (70.0F / 163.0F);
      sptr ++;
    }
  return result;
}

// Include the explicit instantiations
// MA #include "inst/Image_FilterOps.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
