/*!@file Image_FilterOps.H Filtering operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_FilterOps.H 4290 2005-03-12 02:03:27Z mundhenk $
//


#ifndef IMAGE_FILTEROPS_H_DEFINED
#define IMAGE_FILTEROPS_H_DEFINED

#include "saliency.H"
#include "Promotions.H"
#include <vector>

template <class T> class Image;
class Dims;
class Point2D;

//! Low-pass filter, coeff * [0.25 0.5 0.25], applied separably in X and Y
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass3(const Image<T_or_RGB>& src,
         const bool go_x = true, const bool go_y = true);

//! Low-pass filter, coeff * [0.25 0.5 0.25], applied in X
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass3x(const Image<T_or_RGB>& src);

//! Low-pass filter, coeff * [0.25 0.5 0.25], applied in Y
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass3y(const Image<T_or_RGB>& src);

//! Low-pass filter, Anderson's 5x5 separable
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass5(const Image<T_or_RGB>& src,
         const bool go_x = true, const bool go_y = true);

//! Low-pass filter, Anderson's 5x5 separable, applied in X
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass5x(const Image<T_or_RGB>& src);

//! Low-pass filter, Anderson's 5x5 separable, applied in Y
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass5y(const Image<T_or_RGB>& src);

//! Low-pass filter, Anderson's 5x5 separable, applied in X, and X decimated by factor
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass5xDecX(const Image<T_or_RGB>& src, const int factor = 2);

//! Low-pass filter, Anderson's 5x5 separable, applied in Y, and Y decimated by factor
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass5yDecY(const Image<T_or_RGB>& src, const int factor = 2);

//! Low-pass filter, 9x9 separable
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass9(const Image<T_or_RGB>& src,
         const bool go_x = true, const bool go_y = true);

Image< promote_trait< float, float >::TP >
lowPass9x(const Image<float>& src);

//! Low-pass filter, 9x9 separable, applied in X
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass9x(const Image<T_or_RGB>& src);

Image< promote_trait< float, float >::TP >
lowPass9y(const Image<float>& src);

//! Low-pass filter, 9x9 separable, applied in Y
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass9y(const Image<T_or_RGB>& src);

//! Low-pass filter, NxN separable
/*! Just forwards to lowPass3/lowPass5/lowPass9, but allows the filter
    width to be selected at runtime. */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPass(const int N, const Image<T_or_RGB>& src,
        const bool go_x = true, const bool go_y = true);

//! Low-pass filter, NxN separable, applied in X
/*! Just forwards to lowPass3x/lowPass5x/lowPass9x, but allows the filter
    width to be selected at runtime. */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPassX(const int N, const Image<T_or_RGB>& src);

//! Low-pass filter, NxN separable, applied in Y
/*! Just forwards to lowPass3y/lowPass5y/lowPass9y, but allows the filter
    width to be selected at runtime. */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
lowPassY(const int N, const Image<T_or_RGB>& src);

//! Brute force, super inefficient 2D convolution
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolve(const Image<T_or_RGB>& src, const Image<float>& filter);

//! Brute force, super inefficient 2D convolution, with C filter of Nx*Ny
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolve(const Image<T_or_RGB>& src, const float* filter,
         const int Nx, const int Ny);

//! Brute force, super inefficient 2D convolution (truncated filter boundary)
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolveClean(const Image<T_or_RGB>& src, const Image<float>& filter);

//! Brute force, super inefficient 2D convolution, with C filter of Nx*Ny
/*! (truncated filter boundary condition) */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolveClean(const Image<T_or_RGB>& src, const float* filter,
              const int Nx, const int Ny);

//! Brute force, super inefficient 2D convolution (truncated filter boundary)
/*! check for zero pixels and skip */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolveCleanZero(const Image<T_or_RGB>& src, const Image<float>& filter);


//! Brute force, super inefficient 2D convolution, with C filter of Nx*Ny
/*! (truncated filter boundary condition) */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convolveCleanZero(const Image<T_or_RGB>& src, const float* filter,
              const int Nx, const int Ny);

//! This is a somewhat-optimized 2-D convolution.
/*! It is optimized in the sense that there are special-case inner loops
  for (1) cases where the filter doesn't completely overlap with the
  local image patch because we're at the image boundary (in this case
  we need a specialized inner loop to handle these boundary
  conditions), and (2) cases where the filter and local image patch
  overlap completely (in which case we can use a more efficient inner
  loop). */
template <class T>
Image<typename promote_trait<T, float>::TP>
optConvolve(const Image<T>& src, const Image<float>& filter);

//! Brute force, super inefficient 2D convolution (truncated filter boundary)
/*! With normalization by the local image energy, a la HMAX */
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveHmax(const Image<T>& src, const Image<float>& filter);

//! Spatial pooling taking the max value over a neighborhood
/*! Compute max over a rectangle of size (si, sj) starting at positions
  that increment in steps (sti, stj) */
template <class T>
Image<typename promote_trait<T, float>::TP>
spatialPoolMax(const Image<T>& src, const int si, const int sj,
               const int sti, const int stj);

//! Feature/spatial pooling for the S2 layer of Hmax
template <class T>
float featurePoolHmax(const Image<T>& img1, const Image<T>& img2,
                      const Image<T>& img3, const Image<T>& img4,
                      const int si, const int sj, const float s2t);

//! Separable filter, any size (size=0 for no filter), zero-fill boundary
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
sepFilter(const Image<T_or_RGB>& src, const Image<float>& hFilter,
          const Image<float>& vFilter);

template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
sepFilter(const Image<T_or_RGB>& src, const float* hFilter,
          const float* vFilter, const int hfsize, const int vfsize);

//! Separable filter, any size (size=0 for no filter), clean boundary
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
sepFiltClean(const Image<T_or_RGB>& src, const Image<float>& hFilter,
             const Image<float>& vFilter);

template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
xFilterClean(const Image<T_or_RGB>& src, const float* hFilter,
             const int hfsize);

template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
yFilterClean(const Image<T_or_RGB>& src, const float* vFilter,
             const int vfsize);

template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
sepFiltClean(const Image<T_or_RGB>& src, const float* hFilter,
             const float* vFilter, const int hfsize, const int vfsize);


//! Oriented filter along theta (deg), spatial freq k
/*! This filter works on the difference between the image and itself
  low-passed. It modulates this difference by a complex sinusoid, and
  then low-passes the real and imaginary parts, and finally takes the
  modulus. This method is by Greenspan et al. (1994).  CAUTION: *you*
  must provide as input (i.e., in "this") the difference between an
  image and itself low-pass filtered.  CAUTION: results are possibly
  bad with the integer versions. Use float.  */
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
orientedFilter(const Image<T_or_RGB>& src, const float k, const float theta,
               const float intensity = 1.0);

//! Compute center-surround difference, taking the abs if absol==true
/*! To the (hires) center will be substracted a (lowres)
  surround. Image sizes must be int multiples.  CAUTION: values will
  be clamped to the range of your template type.
  @param absol if true, take the absolute value of the difference (otherwise,
  negative values are clamped to zero). */
template <class T_or_RGB>
Image<T_or_RGB> centerSurround(const Image<T_or_RGB>& center,
                               const Image<T_or_RGB>& surround,
                               const bool absol = false);

//! Compute center-surround difference, without clamping or rectifying
/*! To the (hires) center will be substracted a (lowres)
  surround. Image sizes must be int multiples. This is implemented in
  a different function from centerSurround() as it returns two images,
  for the positive and negative values of the difference */
template <class T_or_RGB>
void centerSurround(const Image<T_or_RGB>& center,
                    const Image<T_or_RGB>& surround,
                    Image<T_or_RGB>& pos, Image<T_or_RGB>& neg);

//! compute double-opponent response
/*! (cplus - cminus) [-] (splus - sminus) where [-] is subtraction of two
  images of possibly different sizes, followed by absolute value. The result
  will have the size of the larger (center) image. */
template <class T>
Image<typename promote_trait<T, float>::TP>
doubleOpp(const Image<T>& cplus, const Image<T>& cminus,
          const Image<T>& splus, const Image<T>& sminus);

//! Compute average orientation and strength using steerable filters
template <class T>
void avgOrient(const Image<T>& src,
               Image<float>& orient, Image<float>& strength);

//! Create a DoG kernel; if halfsize is 0, compute it automatically
template <class T>
Image<T> dogFilter(const float stddev, const float theta,
                   const int halfsize = 0);

//! Create a dog kernel for HMAX
template <class T>
Image<T> dogFilterHmax(const float stddev, const float theta,
                       const int cBegin, const int cEnd);

//! Create a Gabor kernel
template <class T>
Image<T> gaborFilter(const float stddev, const float period,
                     const float phase, const float theta,
                     const T bg = T(0), const T ampl = T(1));

//! Create a Gabor kernel with default 1*sigma^2 instead of 2*sigma^2
/*! this is essentailly used in some psycho metric experiments (for instance
, Polat and Sagi, 1994) instead
of 2*sigma^2 so it is included here, if your weird you may make sigMod
something other than 1.0 (or 2.0). As always feel free to be loopy */
template <class T>
Image<T> gaborFilter2(const float stddev, const float period,
                      const float phase, const float theta,
                      const float sigMod = 1.0F,
                      const float amplitude = 1.0F);

//! Produces a Gabor kernel with optionally unequal major+minor axis lengths.
/*!@param size determines the filter size (in pixels) - in the default case 
  (-1), the size is computed from the stddev*/
Image<float> gaborFilter3(const float major_stddev, const float minor_stddev,
                          const float period, const float phase, 
                          const float theta, int size = -1);

//! Create a 2D gaussian, compatible with gaborFilter2 for mask purposes
template <class T>
Image<T> gaussian2D(const float stddev, const float sigMod = 1.0F,
                    const float amplitude = 255.0F);

//! Create a 2D Gaussian blob centered at given coords
template <class T>
Image<T> gaussianBlob(const Dims& dims, const Point2D& center,
                      const float sigmaX, const float sigmaY);

//! Create a 1D binomial distribution kernel (shape of sz-by-1).
/*! These kernels are normalized to sum to 1, and are used (implicitly) in
    the various lowPass*() functions. */
Image<float> binomialKernel(const int sz);

//! Create a sinusoidal grating
template <class T>
Image<T> grating(const int width, const int height,
                 const float period, const float phase,
                 const float theta);

//! Generate 1D gaussian
/*! @param coeff a coefficient so that the peak of the Gaussian has the value
  coeff. If coeff is given as 0.0, then it will be computed internally as
  1/(sigma*sqrt(2*pi)), i.e., the normal Gaussian coefficient.
  @param sigma the Gaussian standard deviation, in pixels
  @param maxhw maximum allower filter half-width. If maxhw is 0 there 
  is no limit.
  @param threshperc cut the spatial extent of the kernel so that only
  values greater than threshperc times the max remain in the
  kernel. This allows us to automatically compute the appropriate
  kernel size to use for a given sigma and approximation tolerance
  given by threshpec. */
template <class T>
Image<T> gaussian(const float coeff, const float sigma, const int maxhw,
                  const float threshperc = 1.0);

//! Build a long-range excitatory filter
template <class T>
Image<T> longRangeExcFilter(const float factor, const float orient);

//! convolve with a separable 2D gaussian
template <class T_or_RGB>
Image<typename promote_trait<T_or_RGB, float>::TP>
convGauss(const Image<T_or_RGB>& src, const float sigmaX, const float sigmaY,
          const float threshperc);

//! Generate a fixation mask, i.e., a Gaussian blob
template <class T>
Image<T> fixationMask(const Dims& dims, const Point2D& fixation,
                      const float pixperdeg, const T maxval,
                      const float sigma = 0.5f);

//! Generate a fixation mask, adding it to existing mask
template <class T>
Image<T> fixationMask(const Image<T>& mask, const Point2D& fixation,
                      const float pixperdeg, const T maxval,
                      const float sigma = 0.5f);

//! Return an image of Dimensions d filled with 1s
Image<byte> ones(Dims d);

//! Return an image of width w and height h filled with 1s
Image<byte> ones(int w, int h);

//! Return a square image of width w filled with 1s
Image<byte> ones(int w);

//! Dilate the img with structuring element se with a given origin
/*! @param origin is the origin of the structuring element. This 
  parameter is optional - the default is the center of the SE*/
template <class T>
Image<T> dilateImg(const Image<T>& img, const Image<T>& se, 
                Point2D origin = Point2D(-1,-1));

//! Erode the img with structuring element se with a given origin
/*! @param origin is the origin of the structuring element. This 
  parameter is optional - the default is the center of the SE*/
template <class T>
Image<T> erodeImg(const Image<T>& img, const Image<T>& se, 
               Point2D origin = Point2D(-1,-1));

//! Open (erode and the close) img with struct element se with a given origin
/*! @param origin is the origin of the structuring element. This 
  parameter is optional - the default is the center of the SE*/
template <class T>
Image<T> openImg(const Image<T>& img, const Image<T>& se, 
              Point2D origin = Point2D(-1,-1));

//! Close (dilate and erode) img with struct element se with a given origin
/*! @param origin is the origin of the structuring element. This 
  parameter is optional - the default is the center of the SE*/
template <class T>
Image<T> closeImg(const Image<T>& img, const Image<T>& se, 
               Point2D origin = Point2D(-1,-1));

//! Divide image by the local image energy, then subtract overall mean.
template <class T>
Image<T> energyNorm(const Image<T>& img);

//! Compute response of a junction detector filter
/*! In the full implementation here, the junction filter responds only
if the relevant features are present and the irrelevant features are
absent.
  @param i0 image filtered by a 0deg (horisontal) Gabor filter
  @param i45 image filtered by a 45deg Gabor filter
  @param i90 image filtered by a 90deg (vertical) Gabor filter
  @param i135 image filtered by a 135deg Gabor filter
  @param r boolean array of which features are considered relevant to
  the junction. The first element is for the horizontal (i0) feature
  at (x+dx, y), the second for the 45deg (i45) feature at (x+dx,
  y-dy), and so on, going counterclockwise.
  @param dx horizontal distance from current pixel at which the
  presence or absence of a given feature should be checked for.
  @param dy vertical distance from current pixel at which the
  presence or absence of a given feature should be checked for. */
template <class T>
Image<T> junctionFilterFull(const Image<T>& i0, const Image<T>& i45,
                            const Image<T>& i90, const Image<T>& i135,
                            const bool r[8], const int dx = 6,
                            const int dy = 6);

//! Compute response of a junction detector filter, partial implementation
/*! In the partial implementation here, the junction filter responds
  when the relevant features are present, without consideration of
  what the values of the irrelevant features might be.
  @param i0 image filtered by a 0deg (horisontal) Gabor filter
  @param i45 image filtered by a 45deg Gabor filter
  @param i90 image filtered by a 90deg (vertical) Gabor filter
  @param i135 image filtered by a 135deg Gabor filter
  @param r boolean array of which features are considered relevant to
  the junction. The first element is for the horizontal (i0) feature
  at (x+dx, y), the second for the 45deg (i45) feature at (x+dx,
  y-dy), and so on, going counterclockwise.
  @param dx horizontal distance from current pixel at which the
  presence or absence of a given feature should be checked for.
  @param dy vertical distance from current pixel at which the
  presence or absence of a given feature should be checked for. */
template <class T>
Image<T> junctionFilterPartial(const Image<T>& i0, const Image<T>& i45,
                               const Image<T>& i90, const Image<T>& i135,
                               const bool r[8], const int dx = 6,
                               const int dy = 6);

//! this method will apply a filter at a location and return a value
/*! This method is useful if your need to filter an image in a 
  non-linear method or only need to extract filtered patches.
  Other than that, you should use the other methods above.
*/
template <class T>
PixRGB<T> filterAtLocation(Image<PixRGB<T> > *image,
                           Image<T> *kernel, Point2D *location);
//! this method will apply a filter at a location and return a value
/*! This method is useful if your need to filter an image in a 
  non-linear method or only need to extract filtered patches.
  Other than that, you should use the other methods above.
  This method differs from filterAtLocation in that it accepts 
  a batch of locations.
*/

template <class T>
void filterAtLocationBatch(const Image<PixRGB<T> > *image,
                           const Image<T> *kernel,
                           const std::vector<Point2D*> *location,
                           std::vector<PixRGB<T> > *results,
                           const unsigned int size);

//! This method applies filter on an image at multiple scales at select locals
/*! When called this method will run each of the supplied filters on the 
    image at the scales ranging from full scale to size/2^n. To keep the
    workings straight, supply a name for each filter and then you can 
    be sure how the results allign in the output vector. The method
    for each filter is basic convolution, but it limited to each location
    rather than the whole image.
    @param image this is the input raw image
    @param location This is a list of the locations at which to apply filters
    @param locations This is how many locations are in the list
    @param filterNames a list of what each filter is called for labels
    @param filters these are several kernels for filter use
    @param scales how man scales down do you want to filter each image
    @param quarterSize scale image by sqrt(.5)(quarter-ish) or 1/2
    @param normalize normalize over the filter outputs
    @param results what you get back from all this
    @param resultsLables User supplied descriptors for each filter
*/
template <class T_or_RGB>
void multiScaleBatchFilter(const Image<T_or_RGB> *image,
                           const std::vector<Point2D*> *location,
                           const unsigned int *locations,
                           const std::vector<Image<float> > *filters,
                           const unsigned int scales,
                           const bool quarterSize,
                           const bool normalize,
                           std::vector<std::vector<T_or_RGB> > *results);

//! this will find junctions after calling multiScaleBatchFilter
/*! This is optional and may be called to find junctions if
  multiScaleBatchFilter was called using gabor filters at 
  0,45,90 and 135 degrees. Call this after calling multiScaleBatchFilter
  if wanted.
*/
template <class T_or_RGB>
void multiScaleJunctionFilter(const unsigned int size,
                              const unsigned int scales,
                              const std::vector<std::vector<T_or_RGB> >
                              *input,
                              std::vector<std::vector<T_or_RGB> > *results);

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_FilterOps_templates.h"

#endif // !IMAGE_FILTEROPS_H_DEFINED
