/*!@file Image_FilterOps.C Filtering operations on Image */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_FilterOps.C 4290 2005-03-12 02:03:27Z mundhenk $
//

#include "Image_FilterOps.H"
#include "MathFunctions.H"
#include "Image.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"
#include "log.H"
#include "Promotions.H"
#include "Raster.H"

#include <algorithm>
#include <cmath>
#include <limits>
#include <string>
#include <iostream>

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass3x(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (w < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  // *** unfolded version (all particular cases treated) for max speed.
  // do not use access overload since can access the C array directly.
  // the bet is: float computations are faster than int (true for most
  // float copros), and int2float converts fast->do not have to divide
  // by the coeff sum since we directly use float coeffs.
  // notations: in () is the position of dest ptr, and ^ is src ptr
  // ########## horizontal pass
  for (int j = 0; j < h; j ++)
    {
      // leftmost point  [ (2^) 1 ] / 3
      *dptr++ = sptr[0] * (2.0F / 3.0F) + sptr[1] * (1.0F / 3.0F);

      // rest of the line except last point  [ 1^ (2) 1 ] / 4
      for (int i = 0; i < w - 2; i ++)
        {
          *dptr++ = (sptr[0] + sptr[2]) * 0.25F + sptr[1] * 0.5F;
          sptr++;
        }

      // last point [ 1^ (2) ] / 3
      *dptr++ = sptr[0] * (1.0F / 3.0F) + sptr[1] * (2.0F / 3.0F);
      sptr += 2;  // sptr back to same position as dptr
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass3y(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (h < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();
  const int w2 = w * 2; // speedup

  // ########## vertical pass  (even though we scan horiz for speedup)
  // topmost points  ( [ (2^) 1 ] / 3 )^T
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[0] * (2.0F / 3.0F) + sptr[w] * (1.0F / 3.0F);
      sptr++;
    }
  sptr -= w;  // go back to top-left

  // rest of the column except last point ( [ 1^ (2) 1 ] / 4.0 )^T
  for (int j = 0; j < h - 2; j ++)
    for (int i = 0; i < w; i ++)
      {
        *dptr++ = (sptr[0] + sptr[w2]) * 0.25F + sptr[w] * 0.5F;
        sptr++;
      }

  // last points ( [ 1^ (2) ] / 3 )^T
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[0] * (1.0F / 3.0F) + sptr[w] * (2.0F / 3.0F);
      sptr++;
    }

  // finished. here, sptr is one line before the end of array.
  return result;
}

// ######################################################################
template <class T> // Anderson's separable kernel: 1/16 * [1 4 6 4 1]
Image<typename promote_trait<T, float>::TP>
lowPass5x(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (w < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  if (w == 2) //////////////////////////////////////////////////
    for (int j = 0; j < h; j ++)
      {
        // leftmost point  [ (6^) 4 ] / 10
        *dptr++ = sptr[0] * (6.0F / 10.0F) + sptr[1] * (4.0F / 10.0F);

        // rightmost point  [ 4^ (6) ] / 10
        *dptr++ = sptr[0] * (4.0F / 10.0F) + sptr[1] * (6.0F / 10.0F);

        sptr += 2;  // sptr back to same position as dptr
      }
  else if (w == 3) //////////////////////////////////////////////////
    for (int j = 0; j < h; j ++)
      {
        // leftmost point  [ (6^) 4 1 ] / 11
        *dptr++ = sptr[0] * (6.0F / 11.0F) +
          sptr[1] * (4.0F / 11.0F) +
          sptr[2] * (1.0F / 11.0F);

        // middle point    [ 4^ (6) 4 ] / 14
        *dptr++ = (sptr[0] + sptr[2]) * (4.0F / 14.0F) +
          sptr[1] * (6.0F / 14.0F);

        // rightmost point  [ 1^ 4 (6) ] / 11
        *dptr++ = sptr[0] * (1.0F / 11.0F) +
          sptr[1] * (4.0F / 11.0F) +
          sptr[2] * (6.0F / 11.0F);

        sptr += 3;  // sptr back to same position as dptr
      }
  else  ////////////////////////////// general case for width() >= 4
        // *** unfolded version (all particular cases treated) for
        // max speed.  do not use access overloading since can
        // access the C array directly.  the bet is: float
        // computations are faster than int (true for most float
        // copros), and int2float convert is fast -> do not have to
        // divide by the coeff sum since we directly use float
        // coeffs.
        // notations: in () is the position of dest ptr, and ^ is src ptr
        // ########## horizontal pass
    for (int j = 0; j < h; j ++)
      {
        // leftmost point  [ (6^) 4 1 ] / 11
        *dptr++ = sptr[0] * (6.0F / 11.0F) +
          sptr[1] * (4.0F / 11.0F) +
          sptr[2] * (1.0F / 11.0F);

        // second point    [ 4^ (6) 4 1 ] / 15
        *dptr++ = (sptr[0] + sptr[2]) * (4.0F / 15.0F) +
          sptr[1] * (6.0F / 15.0F) +
          sptr[3] * (1.0F / 15.0F);

        // rest of the line except last 2 points  [ 1^ 4 (6) 4 1 ] / 16.0
        for (int i = 0; i < w - 4; i ++)
          {
            *dptr++ = (sptr[0] + sptr[4]) * (1.0F / 16.0F) +
              (sptr[1] + sptr[3]) * (4.0F / 16.0F) +
              sptr[2]  * (6.0F / 16.0F);
            sptr++;
          }

        // before last point [ 1^ 4 (6) 4 ] / 15
        *dptr++ = sptr[0] * (1.0F / 15.0F) +
          (sptr[1] + sptr[3]) * (4.0F / 15.0F) +
          sptr[2] * (6.0F / 15.0F);
        sptr++;

        // last point [ 1^ 4 (6) ] / 11
        *dptr++ = sptr[0] * (1.0F / 11.0F) +
          sptr[1] * (4.0F / 11.0F) +
          sptr[2] * (6.0F / 11.0F);
        sptr += 3;  // sptr back to same position as dptr
      }
  return result;
}

// ######################################################################
template <class T> // Anderson's separable kernel: 1/16 * [1 4 6 4 1]
Image<typename promote_trait<T, float>::TP>
lowPass5y(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (h < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  // ########## vertical pass  (even though we scan horiz for speedup)
  const int w2 = w * 2, w3 = w * 3, w4 = w * 4; // speedup

  if (h == 2) //////////////////////////////////////////////////
    {
      // topmost points  ( [ (6^) 4 ] / 10 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[0] * (6.0F / 10.0F) +
            sptr[w] * (4.0F / 10.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // bottommost points  ( [ 4^ (6) ] / 10 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[0] * (4.0F / 10.0F) +
            sptr[w] * (6.0F / 10.0F);
          sptr++;
        }
    }
  else if (h == 3) //////////////////////////////////////////////////
    {
      // topmost points  ( [ (6^) 4 1 ] / 11 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (6.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (1.0F / 11.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // middle points  ( [ 4^ (6) 4 ] / 14 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = (sptr[ 0] + sptr[w2]) * (4.0F / 14.0F) +
            sptr[ w] * (6.0F / 14.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // bottommost points  ( [ 1^ 4 (6) ] / 11 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (1.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (6.0F / 11.0F);
          sptr++;
        }
    }
  else  ///////////////////////////////// general case for height >= 4
    {
      // topmost points  ( [ (6^) 4 1 ] / 11 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (6.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (1.0F / 11.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // second topmost points  ( [ 4^ (6) 4 1 ] / 15 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = (sptr[ 0] + sptr[w2]) * (4.0F / 15.0F) +
            sptr[ w] * (6.0F / 15.0F) +
            sptr[w3] * (1.0F / 15.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // rest of the column except last 2 points ( [ 1^ 4 (6) 4 1 ] / 16 )T
      for (int j = 0; j < h - 4; j ++)
        for (int i = 0; i < w; i ++)
          {
            *dptr++ = (sptr[ 0] + sptr[w4]) * (1.0F / 16.0F) +
              (sptr[ w] + sptr[w3]) * (4.0F / 16.0F) +
              sptr[w2]  * (6.0F / 16.0F);
            sptr++;
          }

      // before last points ( [ 1^ 4 (6) 4 ] / 15 )T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (1.0F / 15.0F) +
            (sptr[ w] + sptr[w3]) * (4.0F / 15.0F) +
            sptr[w2] * (6.0F / 15.0F);
          sptr++;
        }

      // last points ( [ 1^ 4 (6) ] / 11 )T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (1.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (6.0F / 11.0F);
          sptr++;
        }
      // finished. here, sptr is two lines before the end of array.
    }
  return result;
}

// ######################################################################
template <class T> // Anderson's separable kernel: 1/16 * [1 4 6 4 1]
Image<typename promote_trait<T, float>::TP>
lowPass5xDecX(const Image<T>& src, const int factor)
{
  ASSERT(factor > 0);
  const int w = src.getWidth(), h = src.getHeight();
  int w2 = w / factor;
  if (w2 == 0) w2 = 1; // do not go smaller than 1 pixel

  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (w < 2) return source; // nothing to smooth

  Image<TF> result(w2, h, NO_INIT);
  if (w2 == 0) return result;

  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();


  if (w == 2) //////////////////////////////////////////////////
    for (int j = 0; j < h; j ++)
      {
        // leftmost point  [ (6^) 4 ] / 10
        *dptr++ = sptr[0] * (6.0F / 10.0F) + sptr[1] * (4.0F / 10.0F);

        // rightmost point  [ 4^ (6) ] / 10
        if (factor == 1)
          *dptr++ = sptr[0] * (4.0F / 10.0F) + sptr[1] * (6.0F / 10.0F);

        sptr += 2;  // sptr back to same position as dptr
      }
  else if (w == 3) //////////////////////////////////////////////////
    for (int j = 0; j < h; j ++)
      {
        // need the left most point in any case
        // leftmost point  [ (6^) 4 1 ] / 11
        *dptr++ = sptr[0] * (6.0F / 11.0F) +
          sptr[1] * (4.0F / 11.0F) +
          sptr[2] * (1.0F / 11.0F);

        // need the middle point only if we subsample by 1
        if (factor == 1)
          {
            // middle point    [ 4^ (6) 4 ] / 14
            *dptr++ = (sptr[0] + sptr[2]) * (4.0F / 14.0F) +
              sptr[1] * (6.0F / 14.0F);

            // rightmost point  [ 1^ 4 (6) ] / 11
            *dptr++ = sptr[0] * (1.0F / 11.0F) +
              sptr[1] * (4.0F / 11.0F) +
              sptr[2] * (6.0F / 11.0F);
          }

        sptr += 3;  // sptr back to same position as dptr
      }
  else  ////////////////////////////// general case for width() >= 4
        // *** unfolded version (all particular cases treated) for
        // max speed.  do not use access overloading since can
        // access the C array directly.  the bet is: float
        // computations are faster than int (true for most float
        // copros), and int2float convert is fast -> do not have to
        // divide by the coeff sum since we directly use float
        // coeffs.
        // notations: in () is the position of dest ptr, and ^ is src ptr
        // ########## horizontal pass
    for (int j = 0; j < h; j ++)
      {
        int i1 = 0, i2 = 0;
        typename Image<TF>::const_iterator sptr2 = sptr;


        // leftmost point  [ (6^) 4 1 ] / 11
        *dptr++ = sptr2[0] * (6.0F / 11.0F) +
          sptr2[1] * (4.0F / 11.0F) +
          sptr2[2] * (1.0F / 11.0F);
        ++i2;
        i1 += factor;

        // need special case for second point?
        if (factor == 1)
          {
            // second point    [ 4^ (6) 4 1 ] / 15
            *dptr++ = (sptr2[0] + sptr2[2]) * (4.0F / 15.0F) +
              sptr2[1] * (6.0F / 15.0F) +
              sptr2[3] * (1.0F / 15.0F);
            ++i2;
            i1 += factor;
          }
        else
          {
            sptr2 += (factor-2);
          }

        // rest of the line except last 2 points  [ 1^ 4 (6) 4 1 ] / 16.0
        while ((i1 < (w-2)) && (i2 < w2))
          {
            *dptr++ = (sptr2[0] + sptr2[4]) * (1.0F / 16.0F) +
              (sptr2[1] + sptr2[3]) * (4.0F / 16.0F) +
              sptr2[2]  * (6.0F / 16.0F);
            i1 += factor; sptr2 += factor;
            ++i2;
          }

        // need special case for second to last point?
        if ((i2 < w2) && (i1 == (w-2)))
          {
            sptr2 = sptr + w - 4;
            // before last point [ 1^ 4 (6) 4 ] / 15
            *dptr++ = sptr2[0] * (1.0F / 15.0F) +
              (sptr2[1] + sptr2[3]) * (4.0F / 15.0F) +
              sptr2[2] * (6.0F / 15.0F);
            i1 += factor;
            ++i2;
          }

        // need special case for last point?
        if ((i2 < w2) && (i1 == (w-1)))
          {
            sptr2 = sptr + w - 3;
            // last point [ 1^ 4 (6) ] / 11
            *dptr++ = sptr2[0] * (1.0F / 11.0F) +
              sptr2[1] * (4.0F / 11.0F) +
              sptr2[2] * (6.0F / 11.0F);
            ++i2;
          }
        sptr += w;
      }
  return result;
}

// ######################################################################
template <class T> // Anderson's separable kernel: 1/16 * [1 4 6 4 1]
Image<typename promote_trait<T, float>::TP>
lowPass5yDecY(const Image<T>& src, const int factor)
{
  ASSERT(factor > 0);
  const int w = src.getWidth(), h = src.getHeight();
  int h2 = h / factor;
  if (h2 == 0) h2 = 1; // do not go smaller than 1 pixel

  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (h < 2) return source; // nothing to smooth

  Image<TF> result(w, h2, NO_INIT);
  if (h2 == 0) return result;

  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  // ########## vertical pass  (even though we scan horiz for speedup)
  const int w2 = w * 2, w3 = w * 3, w4 = w * 4; // speedup

  if (h == 2) //////////////////////////////////////////////////
    {
      // topmost points  ( [ (6^) 4 ] / 10 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[0] * (6.0F / 10.0F) + sptr[w] * (4.0F / 10.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // bottommost points  ( [ 4^ (6) ] / 10 )^T
      if (factor == 1)
        {
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = sptr[0] * (4.0F / 10.0F) +
                sptr[w] * (6.0F / 10.0F);
              sptr++;
            }
        }
    }
  else if (h == 3) //////////////////////////////////////////////////
    {
      // needed in any case
      // topmost points  ( [ (6^) 4 1 ] / 11 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (6.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (1.0F / 11.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left

      // need the middle point only if we subsample by 1
      if (factor == 1)
        {
          // middle points  ( [ 4^ (6) 4 ] / 14 )^T
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = (sptr[ 0] + sptr[w2]) * (4.0F / 14.0F) +
                sptr[ w] * (6.0F / 14.0F);
              sptr++;
            }
          sptr -= w;  // go back to top-left

          // bottommost points  ( [ 1^ 4 (6) ] / 11 )^T
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = sptr[ 0] * (1.0F / 11.0F) +
                sptr[ w] * (4.0F / 11.0F) +
                sptr[w2] * (6.0F / 11.0F);
              sptr++;
            }
        }
    }
  else  ///////////////////////////////// general case for height >= 4
    {
      int i1 = 0, i2 = 0;
      const int skip = (factor - 1) * w;

      // topmost points  ( [ (6^) 4 1 ] / 11 )^T
      for (int i = 0; i < w; i ++)
        {
          *dptr++ = sptr[ 0] * (6.0F / 11.0F) +
            sptr[ w] * (4.0F / 11.0F) +
            sptr[w2] * (1.0F / 11.0F);
          sptr++;
        }
      sptr -= w;  // go back to top-left
      ++i2;
      i1 += factor;

      // second point only needed if we subsample by 1
      if (factor == 1)
        {
          // second topmost points  ( [ 4^ (6) 4 1 ] / 15 )^T
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = (sptr[ 0] + sptr[w2]) * (4.0F / 15.0F) +
                sptr[ w] * (6.0F / 15.0F) +
                sptr[w3] * (1.0F / 15.0F);
              sptr++;
            }
          sptr -= w;  // go back to top-left
          ++i2;
          i1 += factor;
        }
      else
        {
          sptr += (skip - w);
        }

      // rest of the column except last 2 points ( [ 1^ 4 (6) 4 1 ] / 16 )T
      while((i1 < (h-2)) && (i2 < h2))
        {
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = (sptr[ 0] + sptr[w4]) * (1.0F / 16.0F) +
                (sptr[ w] + sptr[w3]) * (4.0F / 16.0F) +
                sptr[w2]  * (6.0F / 16.0F);
              sptr++;
            }
          sptr += skip;
          i1 += factor;
          ++ i2;
        }

      // need special case for second to last point?
      if ((i2 < h2) && (i1 == (h-2)))
        {
          sptr = source.end() - w4;
          // before last points ( [ 1^ 4 (6) 4 ] / 15 )T
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = sptr[ 0] * (1.0F / 15.0F) +
                (sptr[ w] + sptr[w3]) * (4.0F / 15.0F) +
                sptr[w2] * (6.0F / 15.0F);
              sptr++;
            }
          i1 += factor;
          ++i2;
        }

      // need special case for last point?
      if ((i2 < h2) && (i1 == (h-1)))
        {
          sptr = source.end() - w3;
          // last points ( [ 1^ 4 (6) ] / 11 )T
          for (int i = 0; i < w; i ++)
            {
              *dptr++ = sptr[ 0] * (1.0F / 11.0F) +
                sptr[ w] * (4.0F / 11.0F) +
                sptr[w2] * (6.0F / 11.0F);
              sptr++;
            }
        }
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass9x(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (w < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  if (w < 9)  // use inefficient implementation for small images
    {
      float kernel[9] = { 1.0F / 256.0F, 8.0F / 256.0F, 28.0F / 256.0F,
                          56.0F / 256.0F, 70.0F / 256.0F, 56.0F / 256.0F,
                          28.0F / 256.0F, 8.0F / 256.0F, 1.0F / 256.0F };
       return xFilterClean(src, kernel, 9);
    }

  // boundary conditions: truncated filter
  for (int j = 0; j < h; j ++)
    {
      // leftmost points
      *dptr++ = sptr[0] * (70.0F / 163.0F) +
        sptr[1] * (56.0F / 163.0F) +
        sptr[2] * (28.0F / 163.0F) +
        sptr[3] * ( 8.0F / 163.0F) +
        sptr[4] * ( 1.0F / 163.0F);
      *dptr++ = (sptr[0] + sptr[2]) * (56.0F / 219.0F) +
        sptr[1] * (70.0F / 219.0F) +
        sptr[3] * (28.0F / 219.0F) +
        sptr[4] * ( 8.0F / 219.0F) +
        sptr[5] * ( 1.0F / 219.0F);
      *dptr++ = (sptr[0] + sptr[4]) * (28.0F / 247.0F) +
        (sptr[1] + sptr[3]) * (56.0F / 247.0F) +
        sptr[2] * (70.0F / 247.0F) +
        sptr[5] * ( 8.0F / 247.0F) +
        sptr[6] * ( 1.0F / 247.0F);
      *dptr++ = (sptr[0] + sptr[6]) * ( 8.0F / 255.0F) +
        (sptr[1] + sptr[5]) * (28.0F / 255.0F) +
        (sptr[2] + sptr[4]) * (56.0F / 255.0F) +
        sptr[3] * (70.0F / 255.0F) +
        sptr[7] * ( 1.0F / 255.0F);

      // far from the borders
      for (int i = 0; i < w - 8; i ++)
        {
          *dptr++ = (sptr[0] + sptr[8]) * ( 1.0F / 256.0F) +
            (sptr[1] + sptr[7]) * ( 8.0F / 256.0F) +
            (sptr[2] + sptr[6]) * (28.0F / 256.0F) +
            (sptr[3] + sptr[5]) * (56.0F / 256.0F) +
            sptr[4] * (70.0F / 256.0F);
          sptr ++;
        }

      // rightmost points
      *dptr++ = sptr[0] * ( 1.0F / 255.0F) +
        (sptr[1] + sptr[7]) * ( 8.0F / 255.0F) +
        (sptr[2] + sptr[6]) * (28.0F / 255.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 255.0F) +
        sptr[4] * (70.0F / 255.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 247.0F) +
        sptr[1] * ( 8.0F / 247.0F) +
        (sptr[2] + sptr[6]) * (28.0F / 247.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 247.0F) +
        sptr[4] * (70.0F / 247.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 219.0F) +
        sptr[1] * ( 8.0F / 219.0F) +
        sptr[2] * (28.0F / 219.0F) +
        (sptr[3] + sptr[5]) * (56.0F / 219.0F) +
        sptr[4] * (70.0F / 219.0F);
      sptr ++;
      *dptr++ = sptr[0] * ( 1.0F / 163.0F) +
        sptr[1] * ( 8.0F / 163.0F) +
        sptr[2] * (28.0F / 163.0F) +
        sptr[3] * (56.0F / 163.0F) +
        sptr[4] * (70.0F / 163.0F);
      sptr += 5;  // sptr back to same as dptr (start of next line)
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass9y(const Image<T>& src)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (h < 2) return source; // nothing to smooth
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  if (h < 9)  // use inefficient implementation for small images
    {
      float kernel[9] = { 1.0F / 256.0F, 8.0F / 256.0F, 28.0F / 256.0F,
                          56.0F / 256.0F, 70.0F / 256.0F, 56.0F / 256.0F,
                          28.0F / 256.0F, 8.0F / 256.0F, 1.0F / 256.0F };
      return yFilterClean(src, kernel, 9);
    }

  // *** vertical pass ***
  const int w2 = w + w, w3 = w2 + w, w4 = w3 + w, w5 = w4 + w, w6 = w5 + w,
    w7 = w6 + w,  w8 = w7 + w;  // index computation speedup
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * (70.0F / 163.0F) +
        sptr[ w] * (56.0F / 163.0F) +
        sptr[w2] * (28.0F / 163.0F) +
        sptr[w3] * ( 8.0F / 163.0F) +
        sptr[w4] * ( 1.0F / 163.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w2]) * (56.0F / 219.0F) +
        sptr[ w] * (70.0F / 219.0F) +
        sptr[w3] * (28.0F / 219.0F) +
        sptr[w4] * ( 8.0F / 219.0F) +
        sptr[w5] * ( 1.0F / 219.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w4]) * (28.0F / 247.0F) +
        (sptr[ w] + sptr[w3]) * (56.0F / 247.0F) +
        sptr[w2] * (70.0F / 247.0F) +
        sptr[w5] * ( 8.0F / 247.0F) +
        sptr[w6] * ( 1.0F / 247.0F);
      sptr ++;
    }
  sptr -= w; // back to top-left
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = (sptr[ 0] + sptr[w6]) * ( 8.0F / 255.0F) +
        (sptr[ w] + sptr[w5]) * (28.0F / 255.0F) +
        (sptr[w2] + sptr[w4]) * (56.0F / 255.0F) +
        sptr[w3] * (70.0F / 255.0F) +
        sptr[w7] * ( 1.0F / 255.0F);
      sptr ++;
    }
  sptr -= w;   // back to top-left
  for (int j = 0; j < h - 8; j ++)
    for (int i = 0; i < w; i ++)
      {
        *dptr++ = (sptr[ 0] + sptr[w8]) * ( 1.0F / 256.0F) +
          (sptr[ w] + sptr[w7]) * ( 8.0F / 256.0F) +
          (sptr[w2] + sptr[w6]) * (28.0F / 256.0F) +
          (sptr[w3] + sptr[w5]) * (56.0F / 256.0F) +
          sptr[w4]  * (70.0F / 256.0F);
        sptr ++;
      }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 255.0F) +
        (sptr[ w] + sptr[w7]) * ( 8.0F / 255.0F) +
        (sptr[w2] + sptr[w6]) * (28.0F / 255.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 255.0F) +
        sptr[w4] * (70.0F / 255.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 247.0F) +
        sptr[ w] * ( 8.0F / 247.0F) +
        (sptr[w2] + sptr[w6]) * (28.0F / 247.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 247.0F) +
        sptr[w4] * (70.0F / 247.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 219.0F) +
        sptr[ w] * ( 8.0F / 219.0F) +
        sptr[w2] * (28.0F / 219.0F) +
        (sptr[w3] + sptr[w5]) * (56.0F / 219.0F) +
        sptr[w4] * (70.0F / 219.0F);
      sptr ++;
    }
  for (int i = 0; i < w; i ++)
    {
      *dptr++ = sptr[ 0] * ( 1.0F / 163.0F) +
        sptr[ w] * ( 8.0F / 163.0F) +
        sptr[w2] * (28.0F / 163.0F) +
        sptr[w3] * (56.0F / 163.0F) +
        sptr[w4] * (70.0F / 163.0F);
      sptr ++;
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass3(const Image<T>& src, const bool go_x, const bool go_y)
{
  Image<typename promote_trait<T, float>::TP> result = src;
  if (go_x) result = lowPass3x(result);
  if (go_y) result = lowPass3y(result);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass5(const Image<T>& src, const bool go_x, const bool go_y)
{
  Image<typename promote_trait<T, float>::TP> result = src;
  if (go_x) result = lowPass5x(result);
  if (go_y) result = lowPass5y(result);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass9(const Image<T>& src, const bool go_x, const bool go_y)
{
  Image<typename promote_trait<T, float>::TP> result = src;
  if (go_x) result = lowPass9x(result);
  if (go_y) result = lowPass9y(result);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPass(const int N, const Image<T>& src,
        const bool go_x = true, const bool go_y = true)
{
  Image<typename promote_trait<T, float>::TP> result = src;
  if (go_x) result = lowPassX(N, result);
  if (go_y) result = lowPassY(N, result);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPassX(const int N, const Image<T>& src)
{
  switch (N)
    {
    case 3: return lowPass3x(src);
    case 5: return lowPass5x(src);
    case 9: return lowPass9x(src);
    default:
      {
        // Only allow odd-sized filters, since even-sized kernels will
        // spatially shift the output relative to the input
        if (N % 2 == 1)
          {
            const Image<float> kern = binomialKernel(N);
            return xFilterClean(src, kern.getArrayPtr(), N);
          }
      }
    }
  LFATAL("Low-pass filter size %d is not available", N);
  return Image<typename promote_trait<T, float>::TP>(); // "can't happen"
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
lowPassY(const int N, const Image<T>& src)
{
  switch (N)
    {
    case 3: return lowPass3y(src);
    case 5: return lowPass5y(src);
    case 9: return lowPass9y(src);
    default:
      {
        // Only allow odd-sized filters, since even-sized kernels will
        // spatially shift the output relative to the input
        if (N % 2 == 1)
          {
            const Image<float> kern = binomialKernel(N);
            return yFilterClean(src, kern.getArrayPtr(), N);
          }
      }
    }
  LFATAL("Low-pass filter size %d is not available", N);
  return Image<typename promote_trait<T, float>::TP>(); // "can't happen"
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolve(const Image<T>& src, const Image<float>& filter)
{
  return convolve(src, filter.getArrayPtr(),
                  filter.getWidth(), filter.getHeight());
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolve(const Image<T>& src, const float* filter, const int Nx, const int Ny)
{
  ASSERT(src.initialized()); //ASSERT((Nx & 1)  && (Ny & 1));
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  const int kkk = Nx * Ny - 1;
  const int Nx2 = (Nx - 1) / 2;
  const int Ny2 = (Ny - 1) / 2;

  // very inefficient implementation; one has to be crazy to use non
  // separable filters anyway...
  for (int j = 0; j < h; ++j) // LOOP rows of src
    {
      const int j_offset = j-Ny2;

      for (int i = 0; i < w; ++i) // LOOP columns of src
        {
          TF sum = TF();
          const int i_offset = i-Nx2;
          const int kj_bgn = std::max(- j_offset, 0);
          const int kj_end = std::min(h - j_offset, Ny);

          for (int kj = kj_bgn; kj < kj_end; ++kj) // LOOP rows of filter
            {
              const int kjj = kj + j_offset; // row of src
              const int src_offset = w * kjj + i_offset;
              const int filt_offset = kkk - Nx*kj;

              const int ki_bgn = std::max(-i_offset, 0);
              const int ki_end = std::min(w-i_offset, Nx);

              for (int ki = ki_bgn; ki < ki_end; ++ki) // LOOP cols of filt
                sum += sptr[ki + src_offset] * filter[filt_offset - ki];
            }
          *dptr++ = sum;
        }
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveClean(const Image<T>& src, const Image<float>& filter)
{
  return convolveClean(src, filter.getArrayPtr(),
                       filter.getWidth(), filter.getHeight());
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveClean(const Image<T>& src, const float* filter,
              const int Nx, const int Ny)
{
  ASSERT(src.initialized()); //ASSERT((Nx & 1) && (Ny & 1));
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  int kkk = Nx * Ny - 1;
  int Nx2 = (Nx - 1) / 2, Ny2 = (Ny - 1) / 2;
  // very inefficient implementation; one has to be crazy to use non
  // separable filters anyway...
  for (int j = 0; j < h; ++j)
    for (int i = 0; i < w; ++i)
      {
        TF sum = TF(); float sumw = 0.0F;
        for (int kj = 0; kj < Ny; ++kj)
          {
            int kjj = kj + j - Ny2;
            if (kjj >= 0 && kjj < h)
              for (int ki = 0; ki < Nx; ++ki)
                {
                  int kii = ki + i - Nx2;
                  if (kii >= 0 && kii < w)
                    {
                      float fil = filter[kkk - ki - Nx*kj];
                      sum += sptr[kii + w * kjj] * fil;
                      sumw += fil;
                    }
                }
          }
        *dptr++ = sum;
        // if (sumw)
//           *dptr++ = sum / sumw;
//         else
//           *dptr++ = sum;
      }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveCleanZero(const Image<T>& src, const Image<float>& filter)
{
  return convolveCleanZero(src, filter.getArrayPtr(),
                       filter.getWidth(), filter.getHeight());
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveCleanZero(const Image<T>& src, const float* filter,
              const int Nx, const int Ny)
{
  ASSERT(src.initialized()); ASSERT((Nx & 1) && (Ny & 1));
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  int kkk = Nx * Ny - 1;
  int Nx2 = (Nx - 1) / 2, Ny2 = (Ny - 1) / 2;

  TF zero = T(0);

  // very inefficient implementation; one has to be crazy to use non
  // separable filters anyway...
  for (int j = 0; j < h; ++j)
    for (int i = 0; i < w; ++i)
      {
        if((src.getVal(i,j) > 0.001F) || (src.getVal(i,j) < -0.001F))
        {
          TF sum = TF(); float sumw = 0.0F;
          for (int kj = 0; kj < Ny; ++kj)
          {
            int kjj = kj + j - Ny2;
            if (kjj >= 0 && kjj < h)
              for (int ki = 0; ki < Nx; ++ki)
              {
                int kii = ki + i - Nx2;
                if (kii >= 0 && kii < w)
                {
                  float fil = filter[kkk - ki - Nx*kj];
                  sum += sptr[kii + w * kjj] * fil;
                  sumw += fil;
                }
              }
          }
          *dptr++ = sum / sumw;
        }
        else
          *dptr++ = zero;
      }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
optConvolve(const Image<T>& src, const Image<float>& f)
{
  ASSERT(src.initialized());

  const int src_w = src.getWidth();
  const int src_h = src.getHeight();

  Image<float>::const_iterator filter = f.begin();
  const int fil_w = f.getWidth();
  const int fil_h = f.getHeight();

  ASSERT((fil_w & 1) && (fil_h & 1));

  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result(src_w, src_h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  const int fil_end = fil_w * fil_h - 1;
  const int Nx2 = (fil_w - 1) / 2;
  const int Ny2 = (fil_h - 1) / 2;

  const int srow_skip = src_w-fil_w;

  for (int dst_y = 0; dst_y < src_h; ++dst_y)
    {
      // Determine if we're safely inside the image in the y-direction:
      const bool y_clean = dst_y >= Ny2 && dst_y <  (src_h - Nx2);

      for (int dst_x = 0; dst_x < src_w; ++dst_x, ++dptr)
        {
          // Determine if we're safely inside the image in the x-direction:
          const bool x_clean = dst_x >= Nx2 && dst_x <  (src_w - Nx2);

          // Here is where we pick whether we can use the optimized inner
          // loop (in cases where the filter and image patch overlap
          // completely) or whether we must use the inner loop that can
          // handle boundary conditions.

          if (x_clean && y_clean)
            {
              float dst_val = 0.0f;

              Image<float>::const_iterator fptr = filter+fil_end;

              Image<float>::const_iterator srow_ptr =
                sptr + src_w*(dst_y-Nx2) + dst_x - Nx2;

              for (int f_y = 0; f_y < fil_h; ++f_y)
                {
                  for (int f_x = 0; f_x < fil_w; ++f_x)
                    dst_val += (*srow_ptr++) * (*fptr--);

                  srow_ptr += srow_skip;
                }
              *dptr = dst_val;
              continue;
            }
          else
            {
              // OK, we're at an image boundary, so what do we do to make
              // up for the missing pixels? The approach here is to
              // compute the average value of the non-missing pixels, and
              // proceed as if the missing pixels had that value. This
              // minimizes the introduction of edge artifacts e.g. when
              // convolving with an oriented filter.

              float dst_val = 0.0f;
              float src_sum = 0.0f;
              int src_cnt = 0;
              float fil_sum_skipped = 0.0f;

              for (int f_y = 0; f_y < fil_h; ++f_y)
                {
                  const int src_y = f_y + dst_y - Ny2;
                  if (src_y >= 0 && src_y < src_h)
                    {
                      for (int f_x = 0; f_x < fil_w; ++f_x)
                        {
                          const float fil = filter[fil_end - f_x - fil_w*f_y];

                          const int src_x = f_x + dst_x - Nx2;
                          if (src_x >= 0 && src_x < src_w)
                            {
                              const float src_val = sptr[src_x + src_w * src_y];
                              dst_val += src_val * fil;
                              src_sum += src_val;
                              ++src_cnt;
                            }
                          else
                            {
                              fil_sum_skipped += fil;
                            }
                        }
                    }
                  else
                    {
                      for (int f_x = 0; f_x < fil_w; ++f_x)
                        fil_sum_skipped += filter[fil_end - f_x - fil_w*f_y];
                    }
                }
              const float src_avg = src_sum / src_cnt;
              *dptr = dst_val + (fil_sum_skipped * src_avg);
            }
        }
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convolveHmax(const Image<T>& src, const Image<float>& filter)
{
  ASSERT(src.initialized());
  const int Nx = filter.getWidth(), Ny = filter.getHeight();
  ASSERT((Nx & 1) && (Ny & 1));

  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();
  typename Image<float>::const_iterator const filt = filter.begin();

  int kkk = Nx * Ny - 1;
  int Nx2 = (Nx - 1) / 2, Ny2 = (Ny - 1) / 2;
  // very inefficient implementation; one has to be crazy to use non
  // separable filters anyway...
  for (int j = 0; j < h; ++j)
    for (int i = 0; i < w; ++i)
      {
        TF sum = TF(), sumw = TF();
        for (int kj = 0; kj < Ny; ++kj)
          {
            int kjj = kj + j - Ny2;
            if (kjj >= 0 && kjj < h)
              for (int ki = 0; ki < Nx; ++ki)
                {
                  int kii = ki + i - Nx2;
                  if (kii >= 0 && kii < w)
                    {
                      float fil = filt[kkk - ki - Nx*kj];
                      TF img = sptr[kii + w * kjj];
                      sum += img * fil; sumw += img * img;
                    }
                }
          }
        if (sumw > 0.0F) sum /= std::sqrt(sumw);
        *dptr++ = sum;
      }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
spatialPoolMax(const Image<T>& src, const int si, const int sj,
               const int sti, const int stj)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result((int)ceil((float)w / (float)si),
                  (int)ceil((float)h / (float)sj),
                  NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  // very inefficient implementation, ha ha
  for (int j = 0; j < h; j += sj)
    {
      int jmax = std::min(j + stj, h);
      for (int i = 0; i < w; i += si)
        {
          TF val = std::numeric_limits<TF>::min();
          int imax = std::min(i + sti, w);

          for (int jj = j; jj < jmax; jj ++)
            for (int ii = i; ii < imax; ii ++)
              if (sptr[jj * w + ii] > val) val = sptr[jj * w + ii];
          *dptr++ = val;
        }
    }
  return result;
}

// ######################################################################
template<class T>
float featurePoolHmax(const Image<T>& img1, const Image<T>& img2,
                      const Image<T>& img3, const Image<T>& img4,
                      const int si, const int sj, const float s2t)
{
  ASSERT(img1.isSameSize(img2)&&img1.isSameSize(img3)&&img1.isSameSize(img4));

  float res = std::numeric_limits<float>::min();
  int w = img1.getWidth(), h = img1.getHeight();
  typename Image<T>::const_iterator i1 = img1.begin() + si + w * sj;
  typename Image<T>::const_iterator i2 = img2.begin() + si;
  typename Image<T>::const_iterator i3 = img3.begin() + w * sj;
  typename Image<T>::const_iterator i4 = img4.begin();

  for (int y = sj; y < h; y++)
    {
      for (int x = si; x < w; x++)
        {
          float s2r = (*i1 - s2t) * (*i1 - s2t) + (*i2 - s2t) * (*i2 - s2t) +
            (*i3 - s2t) * (*i3 - s2t) + (*i4 - s2t) * (*i4 - s2t);
          if (s2r > res) res = s2r;
          i1++; i2++; i3++; i4++;
        }
      i1 += si; i2 += si; i3 += si; i4 += si;
    }
  return res;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
sepFilter(const Image<T>& src, const Image<float>& hFilter,
          const Image<float>& vFilter)
{
  ASSERT(hFilter.is1D());
  ASSERT(vFilter.is1D());
  return sepFilter(src, hFilter.getArrayPtr(), vFilter.getArrayPtr(),
                   hFilter.getSize(), vFilter.getSize());
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
sepFilter(const Image<T>& src, const float* hFilt, const float* vFilt,
          const int hfs, const int vfs)
{
  ASSERT(hfs & 1 || hfs == 0); ASSERT(vfs & 1 || vfs == 0);
  ASSERT(src.getWidth() >= hfs); ASSERT(src.getHeight() >= vfs);

  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  Image<TF> result = source;
  const int hfs2 = (hfs - 1) / 2, vfs2 = (vfs - 1) / 2;

  if (hfs)  // non-zero filter for horizontal pass
    {
      Image<TF> tmp(w, h, NO_INIT);

      // flip the filter to accelerate convolution:
      float *hFilter = new float[hfs];
      for (int x = 0; x < hfs; x ++) hFilter[hfs - 1 - x] = hFilt[x];
      typename Image<TF>::const_iterator sptr = result.begin();
      typename Image<TF>::iterator dptr = tmp.beginw();

      // *** horizontal pass ***
      for (int jj = 0; jj < h; jj ++)
        {
          // leftmost points:
          for (int j = hfs2; j < hfs - 1; j ++)
            {
              TF val = TF();
              int fp = hfs - 1 - j;
              for (int k = 0; k <= j; k ++) val += sptr[k] * hFilter[fp + k];
              *dptr++ = val;
            }

          // bulk (far from the borders):
          for (int i = 0; i < w - hfs + 1; i ++)
            {
              TF val = TF();
              for (int k = 0; k < hfs; k ++) val += sptr[k] * hFilter[k];
              *dptr++ = val; sptr++;
            }

          // rightmost points:
          for (int j = hfs - 2; j >= hfs2; j --)
            {
              TF val = TF();
              for (int k = 0; k <= j; k ++) val += sptr[k] * hFilter[k];
              *dptr++ = val; sptr++;
            }
          sptr += hfs2;  // sptr back to same as dptr (start of next line)
        }
      delete [] hFilter;
      result.swap(tmp);
    }

  if (vfs)  // non-null vertical filter
    {
      // flip & normalize the filter to accelerate convolution:
      float *vFilter = new float[vfs];
      for (int x = 0; x < vfs; x ++) vFilter[vfs-1-x] = vFilt[x];

      Image<TF> tmp(w, h, NO_INIT);

      // *** vertical pass ***
      typename Image<TF>::const_iterator sptr = result.begin();
      typename Image<TF>::iterator dptr = tmp.beginw();
      int ww[vfs];
      for (int i = 0; i < vfs; i ++) ww[i] = w * i; // speedup precompute

      // top points:
      for (int j = vfs2; j < vfs - 1; j ++)
        {
          int fp = vfs - 1 - j;
          for (int i = 0; i < w; i ++)  // scan all points of given horiz
            {
              TF val = TF();
              for (int k = 0; k <= j; k ++) val += sptr[ww[k]]*vFilter[fp + k];
              *dptr++ = val; sptr++;
            }
          sptr -= w;   // back to top-left corner
        }

      // bulk (far from edges):
      for (int j = 0; j < h - vfs + 1; j ++)
        for (int i = 0; i < w; i ++)
          {
            TF val = TF();
            for (int k = 0; k < vfs; k ++) val += sptr[ww[k]] * vFilter[k];
            *dptr++ = val; sptr ++;
          }

      // bottommost points:
      for (int j = vfs - 2; j >= vfs2; j --)
        for (int i = 0; i < w; i ++)
          {
            TF val = TF();
            for (int k = 0; k <= j; k ++) val += sptr[ww[k]] * vFilter[k];
            *dptr++ = val; sptr ++;
          }
      delete [] vFilter;
      result.swap(tmp);
    }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
sepFiltClean(const Image<T>& src, const Image<float>& hFilter,
             const Image<float>& vFilter)
{
  ASSERT(hFilter.is1D());
  ASSERT(vFilter.is1D());
  return sepFiltClean(src, hFilter.getArrayPtr(), vFilter.getArrayPtr(),
                      hFilter.getSize(), vFilter.getSize());
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
sepFiltClean(const Image<T>& src, const float* hFilter, const float* vFilter,
             const int hfs, const int vfs)
{
  Image<typename promote_trait<T, float>::TP> result = src;

  if (hfs > 0) result = xFilterClean(result, hFilter, hfs);
  if (vfs > 0) result = yFilterClean(result, vFilter, vfs);

  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
xFilterClean(const Image<T>& src, const float* hFilt, const int hfs)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (hfs == 0) return source;  // no filter
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  ASSERT(hfs & 1);
  const int hfs2 = (hfs - 1) / 2;

  // flip the filter to accelerate convolution:
  float *hFilter = new float[hfs]; float sumh = 0.0F;
  for (int x = 0; x < hfs; x ++)
    { sumh += hFilt[x]; hFilter[hfs-1-x] = hFilt[x]; }

  // *** horizontal pass ***
  if (w < hfs)  // special function for very small images
    for (int j = 0; j < h; j ++)
      for (int i = 0; i < w; i ++)
        {
          TF val = TF(); float sum = 0.0F;
          for (int k = 0; k < hfs; k ++)
            if (i + k - hfs2 >= 0 && i + k - hfs2 < w)
              {
                val += sptr[k - hfs2] * hFilter[k];
                sum += hFilter[k];
              }
          *dptr++ = val * sumh / sum; sptr ++;
        }
  else  // function for reasonably large images
    for (int jj = 0; jj < h; jj ++)
      {
        // leftmost points:
        for (int j = hfs2; j < hfs - 1; j ++)
          {
            TF val = TF(); float sum = 0.0F; int fp = hfs - 1 - j;
            for (int k = 0; k <= j; k ++)
              {
                val += sptr[k] * hFilter[fp + k];
                sum += hFilter[fp + k];
              }
            *dptr++ = val * sumh / sum;
          }
        // bulk (far from the borders):
        for (int i = 0; i < w - hfs + 1; i ++)
          {
            TF val = TF();
            for (int k = 0; k < hfs; k ++) val += sptr[k] * hFilter[k];
            *dptr++ = val; sptr++;
          }
        // rightmost points:
        for (int j = hfs - 2; j >= hfs2; j --)
          {
            TF val = TF(); float sum = 0.0F;
            for (int k = 0; k <= j; k ++)
              {
                val += sptr[k] * hFilter[k];
                sum += hFilter[k];
              }
            *dptr++ = val * sumh / sum;
            sptr++;
          }
        sptr += hfs2;  // sptr back to same as dptr (start of next line)
      }
  delete [] hFilter;
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
yFilterClean(const Image<T>& src, const float* vFilt, const int vfs)
{
  const int w = src.getWidth(), h = src.getHeight();
  // promote the source image to float if necessary, so that we do the
  // promotion only once for all, rather than many times as we access
  // the pixels of the image; if no promotion is necessary, "source"
  // will just point to the original data of "src" through the
  // copy-on-write/ref-counting behavior of Image:
  typedef typename promote_trait<T, float>::TP TF;
  const Image<TF> source = src;
  if (vfs == 0) return source;  // no filter
  Image<TF> result(w, h, NO_INIT);
  typename Image<TF>::const_iterator sptr = source.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  ASSERT(vfs & 1);
  const int vfs2 = (vfs - 1) / 2;

  // flip the filter to accelerate convolution:
  float *vFilter = new float[vfs]; float sumv = 0.0;
  for (int x = 0; x < vfs; x ++)
    { sumv += vFilt[x]; vFilter[vfs-1-x] = vFilt[x]; }

  int ww[vfs];
  for (int i = 0; i < vfs; i ++) ww[i] = w * i; // speedup precompute

  if (h < vfs)  // special function for very small images
    for (int j = 0; j < h; j ++)
      for (int i = 0; i < w; i ++)
        {
          TF val = TF(); float sum = 0.0;
          for (int k = 0; k < vfs; k ++)
            if (j + k - vfs2 >= 0 && j + k - vfs2 < h)
              {
                val += sptr[w * (k - vfs2)] * vFilter[k];
                sum += vFilter[k];
              }
          *dptr++ = val * sumv / sum; sptr ++;
        }
  else  // function for reasonably large images
    {
      // top points:
      for (int j = vfs2; j < vfs - 1; j ++)
        {
          int fp = vfs - 1 - j;
          for (int i = 0; i < w; i ++)  // scan all points of given horiz
            {
              TF val = TF(); float sum = 0.0;
              for (int k = 0; k <= j; k ++)
                {
                  val += sptr[ww[k]] * vFilter[fp + k];
                  sum += vFilter[fp + k];
                }
              *dptr++ = val * sumv / sum;
              sptr++;
            }
          sptr -= w;   // back to top-left corner
        }
      // bulk (far from edges):
      for (int j = 0; j < h - vfs + 1; j ++)
        for (int i = 0; i < w; i ++)
          {
            TF val = TF();
            for (int k = 0; k < vfs; k ++) val += sptr[ww[k]] * vFilter[k];
            *dptr++ = val; sptr ++;
          }
      // bottommost points:
      for (int j = vfs - 2; j >= vfs2; j --)
        for (int i = 0; i < w; i ++)
          {
            TF val = TF(); float sum = 0.0;
            for (int k = 0; k <= j; k ++)
              {
                val += sptr[ww[k]] * vFilter[k];
                sum += vFilter[k];
              }
            *dptr++ = val * sumv / sum;
            sptr ++;
          }
    }
  delete [] vFilter;
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
orientedFilter(const Image<T>& src, const float k,
               const float theta, const float intensity)
{
  double kx = double(k) * cos((theta + 90.0) * M_PI / 180.0);
  double ky = double(k) * sin((theta + 90.0) * M_PI / 180.0);

  typedef typename promote_trait<T, float>::TP TF;
  Image<TF> re(src.getDims(), NO_INIT), im(src.getDims(), NO_INIT);
  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<TF>::iterator reptr = re.beginw(), imptr = im.beginw();

  // (x,y) = (0,0) at center of image:
  int w2l = src.getWidth() / 2, w2r = src.getWidth() - w2l;
  int h2l = src.getHeight() / 2, h2r = src.getHeight() - h2l;

  for (int j = -h2l; j < h2r; j ++)
    for (int i = -w2l; i < w2r; i ++)
      {
        double arg = kx * double(i) + ky * double(j);
        TF val = (*sptr++) * intensity;
        *reptr++ = val * float(cos(arg));
        *imptr++ = val * float(sin(arg));
      }

  re = ::lowPass9(re);
  im = ::lowPass9(im);

  return quadEnergy(re, im);
}

// ######################################################################
template <class T>
Image<T> centerSurround(const Image<T>& center,
                        const Image<T>& surround,
                        const bool absol)
{
  const int lw = center.getWidth(), lh = center.getHeight();
  const int sw = surround.getWidth(), sh = surround.getHeight();

  if (sw > lw || sh > lh) LFATAL("center must be larger than surround");

  int scalex = lw / sw, remx = lw - 1 - (lw % sw);
  int scaley = lh / sh, remy = lh - 1 - (lh % sh);

  // result has the size of the larger image:
  Image<T> result(center.getDims(), NO_INIT);

  // scan large image and subtract corresponding pixel from small image:
  int ci = 0, cj = 0;
  typename Image<T>::const_iterator lptr = center.begin();
  typename Image<T>::const_iterator sptr = surround.begin();
  typename Image<T>::iterator dptr = result.beginw();
  T zero = T();

  if (absol == true)  // compute abs(hires - lowres):
    {
      for (int j = 0; j < lh; ++j)
        {
          for (int i = 0; i < lw; ++i)
            {
              if (*lptr > *sptr)
                *dptr++ = *lptr++ - *sptr;
              else
                *dptr++ = *sptr - *lptr++;

              if ((++ci) == scalex && i != remx) { ci = 0; ++sptr; }
            }
          if (ci) { ci = 0; ++sptr; }  // in case the reduction is not round
          if ((++cj) == scaley && j != remy) cj = 0; else sptr -= sw;
        }
    }
  else  // compute hires - lowres, clamped to 0:
    {
      for (int j = 0; j < lh; ++j)
        {
          for (int i = 0; i < lw; ++i)
            {
              if (*lptr > *sptr)
                *dptr++ = *lptr++ - *sptr;
              else
                { *dptr++ = zero; lptr++; }

              if ((++ci) == scalex && i != remx) { ci = 0; ++sptr; }
            }
          if (ci) { ci = 0; ++sptr; } // in case the reduction is not round
          if ((++cj) == scaley && j != remy) cj = 0; else sptr -= sw;
        }
    }

  // attenuate borders:
  result.attenuateBorders(result.getDims().max() / 20);

  return result;
}

// ######################################################################
template <class T>
void centerSurround(const Image<T>& center, const Image<T>& surround,
                    Image<T>& pos, Image<T>& neg)
{
  const int lw = center.getWidth(), lh = center.getHeight();
  const int sw = surround.getWidth(), sh = surround.getHeight();

  if (sw > lw || sh > lh) LFATAL("center must be larger than surround");

  int scalex = lw / sw, remx = lw - 1 - (lw % sw);
  int scaley = lh / sh, remy = lh - 1 - (lh % sh);

  // result has the size of the larger image:
  pos.resize(center.getDims(), NO_INIT);
  neg.resize(center.getDims(), NO_INIT);

  // scan large image and subtract corresponding pixel from small image:
  int ci = 0, cj = 0;
  typename Image<T>::const_iterator lptr = center.begin();
  typename Image<T>::const_iterator sptr = surround.begin();
  typename Image<T>::iterator pptr = pos.beginw(), nptr = neg.beginw();
  T zero = T();

  for (int j = 0; j < lh; ++j)
    {
      for (int i = 0; i < lw; ++i)
        {
          if (*lptr > *sptr) { *pptr++ = *lptr++ - *sptr; *nptr++ = zero; }
          else { *pptr++ = zero; *nptr++ = *sptr - *lptr++; }

          if ((++ci) == scalex && i != remx) { ci = 0; ++sptr; }
        }
      if (ci) { ci = 0; ++sptr; }  // in case the reduction is not round
      if ((++cj) == scaley && j != remy) cj = 0; else sptr -= sw;
    }

  // attenuate borders:
  pos.attenuateBorders(pos.getDims().max() / 20);
  neg.attenuateBorders(neg.getDims().max() / 20);
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
doubleOpp(const Image<T>& cplus, const Image<T>& cminus,
          const Image<T>& splus, const Image<T>& sminus)
{
  ASSERT(cplus.isSameSize(cminus)); ASSERT(splus.isSameSize(sminus));

  typedef typename promote_trait<T, float>::TP TF;

  // compute difference between both center arrays
  Image<TF> cdiff = cplus; cdiff -= cminus;

  // compute difference between both surround arrays
  Image<TF> sdiff = splus; sdiff -= sminus;

  // compute center-surround cdiff - sdiff = (cp - cm) [-] (sp - sm)
  return centerSurround(cdiff, sdiff, true);  // take abs
}

// ######################################################################
template <class T>
void avgOrient(const Image<T>& src,
               Image<float>& orient, Image<float>& strength)
{
  float Gf1[9] = { 0.0094, 0.1148, 0.3964, -0.0601, -0.9213,  -0.0601,
                   0.3964, 0.1148, 0.0094 };
  float Gf2[9] = { 0.0008, 0.0176, 0.1660, 0.6383, 1.0, 0.6383, 0.1660,
                   0.0176, 0.0008 };
  float Gf3[9] = { -0.0028, -0.0480, -0.3020, -0.5806, 0.0, 0.5806, 0.3020,
                   0.0480, 0.0028 };
  float Hf1[9] = { -0.0098, -0.0618, 0.0998, 0.7551, 0.0, -0.7551, -0.0998,
                   0.0618, 0.0098 };
  float Hf2[9] = { 0.0008, 0.0176, 0.1660, 0.6383, 1.0, 0.6383, 0.1660,
                   0.0176, 0.0008 };
  float Hf3[9] = { -0.0020, -0.0354, -0.2225, -0.4277, 0.0, 0.4277, 0.2225,
                   0.0354, 0.0020 };
  float Hf4[9] = { 0.0048, 0.0566, 0.1695, -0.1889, -0.7349, -0.1889, 0.1695,
                   0.0566, 0.0048 };

  Image<float> fima = src;

  // apply Nine-Tap filters to create the separable bases of G2 and H2
  Image<float> G2a = sepFilter(fima, Gf1, Gf2, 9, 9);
  Image<float> G2b = sepFilter(fima, Gf3, Gf3, 9, 9);
  Image<float> G2c = sepFilter(fima, Gf2, Gf1, 9, 9);
  Image<float> H2a = sepFilter(fima, Hf1, Hf2, 9, 9);
  Image<float> H2b = sepFilter(fima, Hf4, Hf3, 9, 9);
  Image<float> H2c = sepFilter(fima, Hf3, Hf4, 9, 9);
  Image<float> H2d = sepFilter(fima, Hf2, Hf1, 9, 9);

  // combine the basis-filtered images:
  orient = Image<float>(src.getDims(), NO_INIT);
  strength = Image<float>(src.getDims(), NO_INIT);

  const int siz = src.getSize();

  typename Image<float>::const_iterator
    pG2a = G2a.begin(),
    pG2b = G2b.begin(),
    pG2c = G2c.begin(),
    pH2a = H2a.begin(),
    pH2b = H2b.begin(),
    pH2c = H2c.begin(),
    pH2d = H2d.begin();

  Image<float>::iterator porient = orient.beginw();
  Image<float>::iterator pstrength = strength.beginw();

  for (int k = 0; k < siz; ++k)
    {
      const float c2 = 0.5 * (SQR(pG2a[k]) - SQR(pG2c[k]))
        + 0.46875 * (SQR(pH2a[k]) - SQR(pH2d[k]))
        + 0.28125 * (SQR(pH2b[k]) - SQR(pH2c[k]))
        + 0.1875 * (pH2a[k] * pH2c[k] - pH2b[k] * pH2d[k]);
      const float c3 =
        - pG2a[k] * pG2b[k]
        - pG2b[k] * pG2c[k]
        - 0.9375 * (pH2c[k] * pH2d[k] + pH2a[k] * pH2b[k])
        - 1.6875 * pH2b[k] * pH2c[k]
        - 0.1875 * pH2a[k] * pH2d[k];
      porient[k] = (atan2(c3, c2) + M_PI)/ 2.0;  // between 0 and PI
      pstrength[k] = sqrt(SQR(c2) + SQR(c3));
    }
}

// ######################################################################
template <class T>
Image<T> dogFilter(const float stddev, const float theta, const int halfsize)
{
  // resize the data buffer
  int size = halfsize;
  if (size <= 0) size = int(ceil(stddev * sqrt(7.4F)));
  Image<T> dest(2 * size + 1, 2 * size + 1, NO_INIT);

  // change the angles in degree to the those in radian : rotation degree
  float rtDeg = M_PI / 180.0F * theta;

  // calculate constants
  float sigq = stddev * stddev;
  typename Image<T>::iterator aptr = dest.beginw();

  // for DOG operation : to give orientation, it uses omit y-directional
  // component
  for (int x = -size; x <= size; x ++)
    for (int y = -size; y <= size; y ++) {
      float rtX = x * cos(rtDeg) + y * sin(rtDeg);
      float rtY = -x * sin(rtDeg) + y * cos(rtDeg);
      *aptr++ = T((((rtX * rtX)) / sigq - 1.0F)/sigq *
                  exp(-(rtX * rtX + rtY * rtY) / (2.0F * sigq)));
    }
  return dest;
}

// ######################################################################
template <class T>
Image<T> dogFilterHmax(const float stddev, const float theta,
                       const int cBegin, const int cEnd)
{
  const Image<T> srcFilt = dogFilter<T>(stddev, theta);
  const Rectangle cropRegion(cBegin, cBegin, cEnd-1, cEnd-1);
  const Image<T> cropped = crop(srcFilt, cropRegion, false);
  const Image<T> diff = cropped - T(mean(cropped));
  return diff / T(sum(squared(diff)));
}

// ######################################################################
template <class T>
Image<T> gaborFilter(const float stddev, const float period,
                     const float phase, const float theta,
                     const T bg, const T ampl)
{
  // figure the proper size for the result
  int size = int(ceil(stddev * sqrt(-2.0F * log(exp(-5.0F)))));

  Image<T> result(2 * size + 1, 2 * size + 1, NO_INIT);

  // change the angles in degree to the those in radians:
  float psi = M_PI / 180.0F * phase;
  float rtDeg = M_PI / 180.0F * theta;

  // calculate constants:
  float omega = (2.0F * M_PI) / period;
  float co = cos(rtDeg), si = sin(rtDeg);
  float sigq = 2.0F * stddev * stddev;
  typename Image<T>::iterator aptr = result.beginw();
  float a = float(ampl), b = float(bg);

  // compute gabor:
  for (int y = -size; y <= size; y ++)
    for (int x = -size; x <= size; x ++)
      *aptr++ = T(a * cos(omega * (x * co + y * si) + psi) *
                  exp(-(x * x + y * y) / sigq) + b);
  return result;
}

// ######################################################################
template <class T>
Image<T> gaborFilter2(const float stddev, const float period,
                      const float phase, const float theta,
                      const float sigMod = 1.0F,
                      const float amplitude = 128.0F)
{
  // figure the proper size for the result
  int size = int(ceil(stddev * sqrt(-2.0F * log(exp(-5.0F)))));

  Image<T> result(2 * size + 1, 2 * size + 1, NO_INIT);

  // change the angles in degree to the those in radians:
  float psi = M_PI / 180.0F * phase;
  float rtDeg = M_PI / 180.0F * theta;

  // calculate constants:
  float omega = (2.0F * M_PI) / period;
  float co = cos(rtDeg), si = sin(rtDeg);
  float sigq = sigMod * stddev * stddev;
  typename Image<T>::iterator aptr = result.beginw();

  // compute gabor:
  for (int y = -size; y <= size; y ++)
    for (int x = -size; x <= size; x ++)
      *aptr++ = T(amplitude*(cos(omega * (x * co + y * si) + psi) *
                  exp(-(x * x + y * y) / sigq)));
  // who's your daddy?
  return result;
}

// ######################################################################

template <class T>
Image<T> gaussian2D(const float stddev, const float sigMod = 1.0F,
                    const float amplitude = 255.0F)
{
  // figure the proper size for the result
  int size = int(ceil(stddev * sqrt(-2.0F * log(exp(-5.0F)))));
  //float norm = 40.010587F;
  Image<T> result(2 * size + 1, 2 * size + 1, NO_INIT);
  T resultX,resultY;

  typename Image<T>::iterator aptr = result.beginw();
  for (int y = -size; y <= size; y++)
  {
    resultY = T(/*((1.0F/(stddev*sqrt(2.0F*M_PI)))*/1.0F*
                 exp(-((y*y)/(stddev*stddev*sigMod))));
    for (int x = -size; x <= size; x++)
    {
      resultX = T(/*((1.0F/(stddev*sqrt(2.0F*M_PI)))*/1.0F*
                 exp(-((x*x)/(stddev*stddev*sigMod))));
      *aptr++ = T((resultX*resultY)*amplitude);
    }
  }
  return result;
}

// ######################################################################
template <class T>
Image<T> gaussianBlob(const Dims& dims, const Point2D& center,
                      const float sigmaX, const float sigmaY)
{
  Image<T> ret(dims, NO_INIT);
  int w = dims.w(), h = dims.h();

  float fac = 0.5f / (M_PI * sigmaX * sigmaY);
  float vx = -0.5f / (sigmaX * sigmaX);
  float vy = -0.5f / (sigmaY * sigmaY);
  for (int jj = 0; jj < h; jj ++)
    {
      float vydy2 = float(jj - center.j); vydy2 *= vydy2 * vy;
      for (int ii = 0; ii < w; ii ++)
        {
          float dx2 = float(ii - center.i); dx2 *= dx2;

          ret.setVal(ii, jj,
                     clamped_convert<T>(fac * expf(vx * dx2 + vydy2)));
        }
    }
  return ret;
}

// ######################################################################
template <class T>
Image<T> grating(const int width, const int height,
                 const float period, const float phase, const float theta)
{
  Image<T> result(width, height, NO_INIT);

  // change the angles in degree to the those in radian
  float psi = M_PI / 180.0F * phase;
  float rtDeg = M_PI / 180.0F * theta;

  // calculate constants
  float omega = (2.0F * M_PI) / period;
  float co = cos(rtDeg), si = sin(rtDeg);
  typename Image<T>::iterator aptr = result.beginw();

  // for gabor operation
  for (int y = 0; y < height; y ++)
    for (int x = 0; x < width; x ++)
      *aptr++ = T(cos(omega * (x * co + y * si) + psi));
  return result;
}

// ######################################################################
template <class T>
Image<T> gaussian(const float coeff, const float sigma,
                  const int maxhw, const float threshperc)
{
  // determine size: keep only values larger that threshperc*max (here max=1)
  int hw = (int)(sigma * sqrt(-2.0F * log(threshperc / 100.0F)));

  // if kernel turns out to be too large, cut it off:
  if (maxhw > 0 && hw > maxhw) hw = maxhw;

  // allocate image for result:
  Image<T> result(2 * hw + 1, 1, NO_INIT);

  // if coeff is given as 0, compute it from sigma:
  float c = coeff;
  if (coeff == 0.0f) c = 1.0f / (sigma * sqrtf(2.0f * float(M_PI)));

  // build both halves simultaneously
  result.setVal(hw, T(c));
  const float sig22 = - 0.5F / (sigma * sigma);
  for (int i = 1; i <= hw; ++i)
    {
      T val = T(c * exp(float(i * i) * sig22));
      result.setVal(hw + i, val);
      result.setVal(hw - i, val);
    }
  return result;
}

// ######################################################################
template <class T>
Image<T> longRangeExcFilter(const float factor, const float orient)
{
  int siz = 9;         // kernel size
  float peak = 3.0F;   // [3] excitation is max 'peak' pixels away from center
  float sigmaR = 1.0F; // [1] radial sigma, in pixels
  float sigmaT = 5.0F; // [15] angular sigma, in degrees

  Image<T> result(siz, siz, NO_INIT); int mid = (siz - 1) / 2;
  float o = orient * M_PI / 180.0F, st = sigmaT * M_PI / 180.0F;
  st = 2.0F * st * st; float sr = 2.0F * sigmaR * sigmaR;

  typename Image<T>::iterator aptr = result.beginw();
  for (int j = 0; j < siz; j ++)
    for (int i = 0; i < siz; i ++)
      {
        float x = (float)(i - mid), y = (float)(j - mid);
        float r = sqrt(x * x + y * y) - peak, t = atan2(y, x);
        float odiff = t - o;
        while (odiff < -M_PI_2) odiff += M_PI;
        while (odiff > M_PI_2) odiff -= M_PI;
        *aptr++ = T(factor * exp(- r*r/sr - odiff*odiff/st));
      }
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T, float>::TP>
convGauss(const Image<T>& src, const float sigmaX, const float sigmaY,
          const float threshperc)
{
  Image<typename promote_trait<T, float>::TP> result = src;
  Image <float> kernel;
  if (sigmaX > 0.0)
    {
      kernel = gaussian<float>(0.0F, sigmaX, result.getWidth(),threshperc);
      result = xFilterClean(result, kernel.getArrayPtr(), kernel.getSize());
    }
  if (sigmaY > 0.0)
    {
      kernel = gaussian<float>(0.0F, sigmaY, src.getHeight(), threshperc);
      result = yFilterClean(result, kernel.getArrayPtr(), kernel.getSize());
    }
  return result;
}

// ######################################################################
template <class T>
Image<T> fixationMask(const Dims& dims, const Point2D& fixation,
                      const float pixperdeg, const T maxval, const float sigma)
{
  Image<T> img(dims, ZEROS);
  return fixationMask(img, fixation, pixperdeg, maxval, sigma);
}

// ######################################################################
template <class T>
Image<T> fixationMask(const Image<T>& mask, const Point2D& fixation,
                      const float pixperdeg, const T maxval, const float sigma)
{
  Image<T> img = mask;
  typename Image<T>::iterator src = img.beginw();
  float mvf = float(maxval); int w = img.getWidth(), h = img.getHeight();
  float sigfac = 0.5f / (sigma * sigma);
  int fi = fixation.i, fj = fixation.j;
  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i ++)
      {
        float ang = sqrt((i-fi) * (i-fi) + (j-fj) * (j-fj)) / pixperdeg;
        float val = exp(- ang * ang * sigfac);
        *src++ += T(mvf * val);  // range 0..maxval
      }
  return img;
}

// ######################################################################
template <class T>
Image<T> dilateImg(const Image<T>& img, const Image<T>& se, Point2D origin)
{
  ASSERT(img.initialized() && se.initialized());

  int iw = img.getWidth(); int ih = img.getHeight();
  int sw =  se.getWidth(); int sh =  se.getHeight();

  if (origin == Point2D(-1,-1))
    {
      origin.i = (sw - 1) / 2;
      origin.j = (sh - 1) / 2;
    }
  ASSERT((origin.i < sw) && (origin.j < sh));

  Image<T> result(iw,ih,ZEROS);

  typename Image<T>::const_iterator iptr, sptr1, sptr2;
  typename Image<T>::iterator rptr1, rptr2;
  iptr = img.begin();

  for (int iy = 0; iy < ih; ++iy)
    {
      // initialize height
      int rYstart = iy - origin.j;
      int rYend = rYstart + sh;
      int sYstart = 0;
      if (rYstart < 0)
        {
          sYstart -= rYstart;
          rYstart = 0;
        }
      if (rYend > ih) rYend = ih;

      for (int ix = 0; ix < iw; ++ix)
        {
          // do we have to do anything?
          if (*iptr != T())
            {
              //initialize pointers and width
              int rXstart = ix - origin.i;
              int rXend = rXstart + sw;
              int sXstart = 0;
              if (rXstart < 0)
                {
                  sXstart -= rXstart;
                  rXstart = 0;
                }
              if (rXend > iw) rXend = iw;

              rptr1 = result.beginw() + rYstart * iw + rXstart;
              sptr1 = se.begin() + sYstart * sw + sXstart;

              for (int ry = rYstart; ry < rYend; ++ry)
                {
                  rptr2 = rptr1; sptr2 = sptr1;
                  for (int rx = rXstart; rx < rXend; ++rx)
                    {
                      *rptr2 = std::max(*rptr2,*sptr2);
                      ++rptr2; ++sptr2;
                    }
                  rptr1 += iw; sptr1 += sw;
                }
            }  // end: we have to do something
          ++iptr;
        }  // end: for ix
    }  // end: for iy
  return result;
}

// ######################################################################
template <class T>
Image<T> erodeImg(const Image<T>& img, const Image<T>& se, Point2D origin)
{
  ASSERT(img.initialized() && se.initialized());

  int iw = img.getWidth(); int ih = img.getHeight();
  int sw =  se.getWidth(); int sh =  se.getHeight();

  if (origin == Point2D(-1,-1))
    {
      origin.i = (sw - 1) / 2;
      origin.j = (sh - 1) / 2;
    }
  ASSERT((origin.i < sw) && (origin.j < sh));

  Image<T> result(iw,ih,ZEROS);

  typename Image<T>::const_iterator iptr1, iptr2, sptr1, sptr2;
  typename Image<T>::iterator rptr = result.beginw();
  T se_orig_val = se.getVal(origin);

  //loop over result X
  for (int ry = 0; ry < ih; ++ry)
    {
      // initialize height
      int iYstart = ry - origin.j;
      int iYend = iYstart + sh;
      int sYstart = 0;
      if (iYstart < 0)
        {
          sYstart -= iYstart;
          iYstart = 0;
        }
      if (iYend > ih) iYend = ih;


      // loop over result Y
      for (int rx = 0; rx < iw; ++rx)
        {
          int iXstart = rx - origin.i;
          int iXend = iXstart + sw;
          int sXstart = 0;
          if (iXstart < 0)
            {
              sXstart -= iXstart;
              iXstart = 0;
            }
          if (iXend > iw) iXend = iw;

          bool flag = true;   // reset the flag
          iptr1 = img.begin() + iYstart * iw + iXstart;
          sptr1 = se.begin() + sYstart * sw + sXstart;

          //loop over the image covered by the structuring element
          for (int iy = iYstart; iy < iYend; ++iy)
            {

              iptr2 = iptr1; sptr2 = sptr1;
              for (int ix = iXstart; ix < iXend; ++ ix)
                {
                  if ((*sptr2 != 0) && (*iptr2 == 0))
                    {
                      flag = false;
                      break;
                    } // end: if (*iptr2 == 0)
                  ++iptr2; ++sptr2;
                } // end: for ix

              if (!flag) break;
              iptr1 += iw; sptr1 += sw;

            } // end: for iy

          // should we set the pixel?
          if (flag) *rptr = std::max(*rptr, se_orig_val);

          ++rptr;

        } // end: for rx
    } // end: for ry
  return result;
}

// ######################################################################
template <class T>
Image<T> openImg(const Image<T>& img, const Image<T>& se, Point2D origin)
{
  return dilateImg(erodeImg(img,se,origin),se,origin);
}

// ######################################################################
template <class T>
Image<T> closeImg(const Image<T>& img, const Image<T>& se, Point2D origin)
{
  return erodeImg(dilateImg(img,se,origin),se,origin);
}

// ######################################################################
template <class T>
Image<T> energyNorm(const Image<T>& img)
{
  // Need to bump everything above zero here so that we don't have
  // divide-by-zero problems when we later divide by the energy. FIXME
  // should be a better way???
  Image<float> result = img + 1.0f;

  // Compute the local image energy
  const Image<float> energy = sqrt(lowPass9(result * result));

  // Normalize by the local image energy
  result = result / energy; // watch divbyzero here!!!

  // Set to zero-mean
  result -= mean(result);

  return result;
}

// ######################################################################
template <class T>
Image<T> junctionFilterFull(const Image<T>& i0, const Image<T>& i45,
                            const Image<T>& i90, const Image<T>& i135,
                            const bool r[8], const int dx = 6,
                            const int dy = 6)
{
  ASSERT(i0.isSameSize(i45) && i0.isSameSize(i90) && i0.isSameSize(i135));
  Image<T> result(i0.getDims(), ZEROS);
  const int w = i0.getWidth(), h = i0.getHeight();
  /* example L junction: upright L(x,y) = |(x, y-1) and (NOT
     /(x+1,y-1)) and -(x+1, y) and (NOT \(x+1,y+1)) and (NOT |(x,
     y+1)) and (NOT /(x-1,y+1)) and (NOT -(x-1, y)) and (NOT
     \(x-1,y-1))

     To find NOT of an irrelevant feature, just consider how strong it
     is in comparison to the weakest relevant feature, i.e. d =
     min(all Rel responses)/2 - Irr. response

     d is high <=> the absence of this Irr. feature
     d is low <=> the presence of this Irr. feature  */

  // pixel offsets for (x+dx, y), (x+dx, y-dy), (x, y-dy), etc. going
  // counterclockwise:
  const int o0 = dx, o1 = dx - dy*w, o2 = -dy*w, o3 = -dx - dy*w,
    o4 = -dx, o5 = -dx + dy*w, o6 = dy*w, o7 = dx + dy*w;

  // get pointers to the image pixel data:
  typename Image<T>::const_iterator
    ip0  = i0.begin()  + dx + dy*w, ip45  = i45.begin()  + dx + dy*w,
    ip90 = i90.begin() + dx + dy*w, ip135 = i135.begin() + dx + dy*w;
  typename Image<T>::iterator rp = result.beginw() + dx + dy*w;

  // loop over the bulk of the images (i.e., all except a border of
  // width dx and height dy):
  for (int j = dy*2; j < h; j ++) {
    for (int i = dx*2; i < w; i ++) {

      // get the various feature values:
      T r0 = ip0[o0], r1 = ip45[o1], r2 = ip90[o2], r3 = ip135[o3],
        r4 = ip0[o4], r5 = ip45[o5], r6 = ip90[o6], r7 = ip135[o7];

      // for this full implementation, let's start by finding out the
      // minimum relevant response, so that we can then check the
      // irrelevant responses against that baseline:
      T minRel = 255;
      if (r[0] && r0 < minRel) minRel = r0;
      if (r[1] && r1 < minRel) minRel = r1;
      if (r[2] && r2 < minRel) minRel = r2;
      if (r[3] && r3 < minRel) minRel = r3;
      if (r[4] && r4 < minRel) minRel = r4;
      if (r[5] && r5 < minRel) minRel = r5;
      if (r[6] && r6 < minRel) minRel = r6;
      if (r[7] && r7 < minRel) minRel = r7;
      minRel /= 2;

      // now let's compute the response of the junction:
      float resp = 1.0f;

      if (r[0]) resp *= r0; else resp *= (minRel - r0);
      if (resp < 0.0f) resp = 0.0f;

      if (resp != 0.0f) {
        if (r[1]) resp *= r1; else resp *= (minRel - r1);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[2]) resp *= r2; else resp *= (minRel - r2);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[3]) resp *= r3; else resp *= (minRel - r3);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[4]) resp *= r4; else resp *= (minRel - r4);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[5]) resp *= r5; else resp *= (minRel - r5);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[6]) resp *= r6; else resp *= (minRel - r6);
        if (resp < 0.0f) resp = 0.0f;
      }

      if (resp != 0.0f) {
        if (r[7]) resp *= r7; else resp *= (minRel - r7);
        if (resp < 0.0f) resp = 0.0f;
      }

      // store the response after normalizing it by the number of features:
      *rp = T( pow(resp, 1.0/8.0) );

      // move on to the next pixel:
      ++ip0; ++ip45; ++ip90; ++ip135; ++rp;
    }
    // skip border and go to next row of pixels:
    ip0 += dx*2; ip45 += dx*2; ip90 += dx*2; ip135 += dx*2; rp += dx*2;
  }

  return result;
}

// ######################################################################
template <class T>
Image<T> junctionFilterPartial(const Image<T>& i0, const Image<T>& i45,
                               const Image<T>& i90, const Image<T>& i135,
                               const bool r[8], const int dx = 6,
                               const int dy = 6)
{
  // the preamble is identical to junctionFilterFull:
  ASSERT(i0.isSameSize(i45) && i0.isSameSize(i90) && i0.isSameSize(i135));
  Image<T> result(i0.getDims(), ZEROS);
  const int w = i0.getWidth(), h = i0.getHeight();
  const int o0 = dx, o1 = dx - dy*w, o2 = -dy*w, o3 = -dx - dy*w,
    o4 = -dx, o5 = -dx + dy*w, o6 = dy*w, o7 = dx + dy*w;
  typename Image<T>::const_iterator
    ip0 =  i0.begin()  + dx + dy*w, ip45 =  i45.begin()  + dx + dy*w,
    ip90 = i90.begin() + dx + dy*w, ip135 = i135.begin() + dx + dy*w;
  typename Image<T>::iterator rp = result.beginw() + dx + dy*w;

  // compute the number of relevant features (for normalization):
  int nr = 0; for (int i = 0; i < 8; i ++) if (r[i]) nr ++;

  // loop over the bulk of the images:
  for (int j = dy*2; j < h; j ++) {
    for (int i = dx*2; i < w; i ++) {

      // get the various feature values:
      T r0 = ip0[o0], r1 = ip45[o1], r2 = ip90[o2], r3 = ip135[o3],
        r4 = ip0[o4], r5 = ip45[o5], r6 = ip90[o6], r7 = ip135[o7];

      // let's compute the response of the junction:
      float resp = 1.0f;

      if (r[0]) resp *= r0;
      if (r[1]) resp *= r1;
      if (r[2]) resp *= r2;
      if (r[3]) resp *= r3;
      if (r[4]) resp *= r4;
      if (r[5]) resp *= r5;
      if (r[6]) resp *= r6;
      if (r[7]) resp *= r7;

      // store the response after normalizing it by the number of
      // relevant features:
      *rp = T( pow(resp, 1.0/nr) );

      // move on to the next pixel:
      ++ip0; ++ip45; ++ip90; ++ip135; ++rp;
    }
    // skip border and go to next row of pixels:
    ip0 += dx*2; ip45 += dx*2; ip90 += dx*2; ip135 += dx*2; rp += dx*2;
  }

  return result;
}

// ######################################################################

template <class T> inline
PixRGB<T> filterAtLocation(Image<PixRGB<T> > *image,
                           Image<T> *kernel, Point2D *location)
{
  ASSERT((kernel->getWidth() % 2 == 1) && (kernel->getHeight() % 2 == 1));
  int w        = image->getWidth();
  int h        = image->getHeight();
  int kw       = kernel->getWidth();
  int kh       = kernel->getHeight();
  int kx       = (kw-1)/2;
  int ky       = (kh-1)/2;
  int kstartx  = 0;
  int kstarty  = 0;
  
  // do some extra work to account for overlapping on edge
  // e.g. bounds checking

  int startx = location->i - kx;
  if(startx < 0){ kstartx = abs(startx); startx = 0;}
  int endx   = location->i + kx;
  if(endx >= w) { endx = w - 1; }

  int starty = location->j - ky;
  if(starty < 0){ kstarty = abs(starty); starty = 0;}
  int endy   = location->j + ky;
  if(endy >= h) { endy = h - 1; }

  PixRGB<T> result;
  float sum = 0;
  //LINFO("C");
  //LINFO("%d %d %d %d",startx,endx,starty,endy);
  //LINFO("%d %d",kstartx,kstarty); 
  typename Image<T>::iterator ker;
  typename Image<PixRGB<T> >::iterator img; 
  for(int iy = starty; iy <= endy; iy++)
  { 
    // start at kernel offset for any edges
    ker = kernel->beginw() +
      (kw*(kstarty+iy-starty)) + kstartx;
    //LINFO("%d", (kw*(kstarty+iy-starty))+ kstartx);
    //LINFO("%d", w*iy + startx);
    //LINFO("%d", w*iy + endx);
    // start in image offset for any edges
    for(img = image->beginw() + w*iy + startx; 
        img != image->beginw() + w*iy + endx; ++img, ++ker)
    {  
      result.r += img->r * (*ker);
      result.g += img->g * (*ker);
      result.b += img->b * (*ker);
      //LINFO("C2");
      sum += *ker;
      //LINFO("C3");
    }
  }
  // normalize results
  if(sum != 0)
  {
    result.r = (T)(result.r/sum);
    result.g = (T)(result.g/sum);
    result.b = (T)(result.b/sum);
  }
  else
  {
    result.r = 0;
    result.g = 0;
    result.b = 0;
  }
  return result;
}
 
// ######################################################################

template <class T> inline
void filterAtLocationBatch(const Image<PixRGB<T> > *image,
                           const Image<T> *kernel, 
                           const std::vector<Point2D*> *location,
                           std::vector<PixRGB<T> > *results,
                           const unsigned int size)
{
  ASSERT((kernel->getWidth() % 2 == 1) && (kernel->getHeight() % 2 == 1));
  ASSERT(results->size() >= size);
  int w        = image->getWidth();
  int h        = image->getHeight();
  int kw       = kernel->getWidth();
  int kh       = kernel->getHeight();
  int kx       = (kw-1)/2;
  int ky       = (kh-1)/2;
  int kstartx  = 0;
  int kstarty  = 0;

  typename std::vector<PixRGB<T> >::iterator resultsItr = results->begin();

  std::vector<Point2D*>::const_iterator locationItr = location->begin();

  // do some extra work to account for overlapping on edge
  // e.g. bounds checking

  int startx,endx,starty,endy;

  typename Image<T>::const_iterator ker;
  typename Image<PixRGB<T> >::const_iterator img; 
  //bool useSlow;
  bool useSlow;


  for(unsigned int i = 0; i < size; i++, ++resultsItr, ++locationItr)
  {
    useSlow = false;
    startx = (*locationItr)->i - kx;
    if(startx < 0){ kstartx = abs(startx); startx = 0; useSlow = true; }
    endx   = (*locationItr)->i + kx;
    if(endx >= w) { endx = w - 1; useSlow = true; }

    starty = (*locationItr)->j - ky;
    if(starty < 0){ kstarty = abs(starty); starty = 0; useSlow = true; }
    endy   = (*locationItr)->j + ky;
    if(endy >= h) { endy = h - 1; useSlow = true; }

    float sum = 0;
    // use slow if the kernal runs off the image 
    // otherwise use a slightly faster iterator incr.
    resultsItr->r = 0;
    resultsItr->g = 0;
    resultsItr->b = 0;

    if(useSlow == true)
    { 
      //LINFO("SLOW");
      for(int iy = starty; iy <= endy; iy++)
      { 
        // start at kernel offset for any edges
        ker = kernel->begin() +
          (kw*(kstarty+iy-starty)) + kstartx;
        // start in image offset for any edges
        for(img = image->begin() + w*iy + startx; 
            img != image->begin() + w*iy + endx; ++img, ++ker)
        {  
          resultsItr->r += img->r * (*ker);
          resultsItr->g += img->g * (*ker);
          resultsItr->b += img->b * (*ker);
          sum += *ker;
        }
      }
      // normalize results
      if(sum != 0)
      {
        resultsItr->r = (T)(resultsItr->r/sum);
        resultsItr->g = (T)(resultsItr->g/sum);
        resultsItr->b = (T)(resultsItr->b/sum);
      }
      else
      {
        resultsItr->r = 0;
        resultsItr->g = 0;
        resultsItr->b = 0;
      }
    }
    else
    {
      //LINFO("FAST");
      ker = kernel->begin();
      for(int iy = starty; iy <= endy; iy++)
      {
        for(img = image->begin() + w*iy + startx; 
            img != image->begin() + w*iy + endx; ++img, ++ker)
        {
          resultsItr->r += img->r * (*ker);
          resultsItr->g += img->g * (*ker);
          resultsItr->b += img->b * (*ker);
          sum += *ker;
        }
      }
      if(sum != 0)
      {
        resultsItr->r = (T)(resultsItr->r/sum);
        resultsItr->g = (T)(resultsItr->g/sum);
        resultsItr->b = (T)(resultsItr->b/sum);
      }
      else
      {
        resultsItr->r = 0;
        resultsItr->g = 0;
        resultsItr->b = 0;
      }
    }
  }
}
 
// ######################################################################
template <class T_or_RGB>
void multiScaleBatchFilter(const Image<T_or_RGB> *image,
                           const std::vector<Point2D*> *location,
                           const unsigned int *locations,
                           const std::vector<Image<float> > *filters,
                           const unsigned int scales,
                           const bool quarterSize,
                           const bool normalize,
                           std::vector<std::vector<T_or_RGB> > *results)
{

  // set up a scan line map
  T_or_RGB zero = T_or_RGB(0);
  //PixH2SV<T_or_RGB> zero = PixH2SV<T_or_RGB>(0);

  // this is how much to reduce each image by in fact for img pyramid
  const double rooty = sqrt(.5);

  // An operable copt of the image locations
  std::vector<Point2D> currentLocation1;
  std::vector<Point2D> currentLocation2;
  std::vector<Point2D> *currentLocation;

  // A working copy of the current image
  Image<T_or_RGB> currentImage1 = *image; 
  Image<T_or_RGB> currentImage2;  
  Image<T_or_RGB> *currentImage;

  // Copy all pertinent locations over
  //LINFO("COPY LOCATIONS");

  std::vector<Point2D*>::const_iterator ilocation = location->begin();

  for(unsigned int i = 0; i < *locations; i++, ++ilocation)
  {
    currentLocation1.push_back(**ilocation);
  }
  if(quarterSize)
  { 
    Point2D pp;
    currentImage2 = rescale(currentImage1,
                            (int)floor(currentImage1.getWidth()*rooty),
                            (int)floor(currentImage1.getHeight()*rooty));
    currentLocation2.resize(currentLocation1.size(),pp);
    std::vector<Point2D>::iterator icur1 =  currentLocation1.begin();
    std::vector<Point2D>::iterator icur2 =  currentLocation2.begin();
    while(icur1 != currentLocation1.end())
    {
      icur2->i = (int)floor(icur1->i * rooty);
      icur2->j = (int)floor(icur1->j * rooty);
      ++icur1; ++icur2;
    }
  }

  int startx,starty;
  unsigned int endx,endy;

  Image<float>::const_iterator ker;
  typename Image<T_or_RGB>::iterator img; 

  typename std::vector<std::vector<T_or_RGB> >::iterator iresults =
    results->begin();

  int foo = 0;
  for(std::vector<Image<float> >::const_iterator
        ifilters = filters->begin();
        ifilters != filters->end(); ++ifilters, foo++)
  { 
    Raster::Visu(*ifilters,PNM,true,false,"filter.%d",foo);
  }

  //LINFO("START");
  // iterate over each scale
  for(unsigned int sc = 0; sc < scales; sc++)
  {
    if((!quarterSize) || (sc%2 == 0) || (sc == 0))
    {
      currentLocation = &currentLocation1;
      currentImage    = &currentImage1;
    }
    else
    {
      currentLocation = &currentLocation2;
      currentImage    = &currentImage2;
    }
    
    const unsigned int w        = currentImage->getWidth();
    const unsigned int h        = currentImage->getHeight(); 
    unsigned int kstartx        = 0;
    unsigned int kstarty        = 0;

    for(std::vector<Image<float> >::const_iterator
          ifilters = filters->begin();
          ifilters != filters->end(); ++ifilters, ++iresults)
    { 
      // iterate over each location

      const unsigned int kw       = ifilters->getWidth();
      const unsigned int kh       = ifilters->getHeight();
      const unsigned int kx       = (kw-1)/2;
      const unsigned int ky       = (kh-1)/2;

      typename std::vector<T_or_RGB>::iterator iiresults =
        iresults->begin(); 
      std::vector<Point2D>::iterator icurrentLocation
        = currentLocation->begin();
      for(unsigned int l = 0; l < *locations; l++,
          ++icurrentLocation, ++iiresults)
      {  
        bool useSlow = false;
        startx = icurrentLocation->i - kx;
        if(startx < 0)
          { kstartx = abs(startx); startx = 0; useSlow = true; }
        endx   = icurrentLocation->i + kx;
        if(endx >= w) { endx = w - 1; useSlow = true; }

        starty = icurrentLocation->j - ky; 
        if(starty < 0)
          { kstarty = abs(starty); starty = 0; useSlow = true; }
        endy   = icurrentLocation->j + ky; 
        if(endy >= h) { endy = h - 1; useSlow = true; }
        
        float sum = 0;
        // use slow if the kernal runs off the image 
        // otherwise use a slightly faster iterator incr.

        for(unsigned int iy = starty; iy <= endy; iy++)
        { 
          // start at kernel offset for any edges
          ker = ifilters->begin() +
            (kw*(kstarty+iy-starty)) + kstartx;
            // start in image offset for any edges
          for(img  = currentImage->beginw() + w*iy + startx; 
              img != currentImage->beginw() + w*iy + endx; ++img, ++ker)
          { 
            *iiresults += (*img) * (*ker);
            sum += fabs(*ker);
          }
        }
        // normalize results
        if(normalize)
          if(sum != 0)
          {
            *iiresults = *iiresults/sum;
          }
          else
          {
            *iiresults = zero;
          }
        kstarty = 0;
        kstartx = 0;
      }
    }
   
    // reduce the image using decimation and 
    // adjust the coordinate selection accordingly
    *currentImage    = decXY(*currentImage);
    // note rooty = sqrt(.5);

    for(std::vector<Point2D>::iterator icurrentLocation
          = currentLocation->begin();
        icurrentLocation != currentLocation->end();
        ++icurrentLocation)
    {
      *icurrentLocation = *icurrentLocation/2;
    }
  }
}

// ######################################################################
// call this optionally after multiScaleBatchFilter 
template <class T_or_RGB>
void multiScaleJunctionFilter(const unsigned int size,
                              const unsigned int scales, 
                              const std::vector<std::vector<T_or_RGB> > 
                              *input,
                              std::vector<std::vector<T_or_RGB> > *results)
{
  // assume 4 angels at the moment
  // const unsigned int offset     = 4 * scales;
  // total size is offset * H1,H2,S,V
  //const unsigned int totalSize  = offset * 4;
  typename std::vector<std::vector<T_or_RGB > >::const_iterator 
    iinput1a = input->begin();
  typename std::vector<std::vector<T_or_RGB > >::const_iterator 
    iinput2a = input->begin()+1;
  typename std::vector<std::vector<T_or_RGB > >::const_iterator 
    iinput1b = input->begin()+2;
  typename std::vector<std::vector<T_or_RGB > >::const_iterator 
    iinput2b = input->begin()+3;
  typename std::vector<std::vector<T_or_RGB > >::iterator 
    iresults = results->begin();

  for(unsigned int s = 0; s < scales; s++)
  {
    typename std::vector<T_or_RGB >::const_iterator
      iiinput1a = iinput1a->begin();
    typename std::vector<T_or_RGB >::const_iterator
      iiinput2a = iinput2a->begin();
    typename std::vector<T_or_RGB >::const_iterator
      iiinput1b = iinput1b->begin();
    typename std::vector<T_or_RGB >::const_iterator
      iiinput2b = iinput2b->begin();
    typename std::vector<T_or_RGB >::iterator iiresults = iresults->begin();

    T_or_RGB max1a = T_or_RGB(0); T_or_RGB max2a = T_or_RGB(0); 
    T_or_RGB max1b = T_or_RGB(0); T_or_RGB max2b = T_or_RGB(0);

    for(unsigned int i = 0; i < size; i++,++iiinput1a)
    {
      max1a = max1a.takeMax(*iiinput1a);
    }
    for(unsigned int i = 0; i < size; i++,++iiinput2a)
    {
      max2a =max2a.takeMax(*iiinput2a);
    }
    for(unsigned int i = 0; i < size; i++,++iiinput1b)
    {
      max1b = max1b.takeMax(*iiinput1b);
    }
    for(unsigned int i = 0; i < size; i++,++iiinput2b)
    {
      max2b = max2b.takeMax(*iiinput2b);
    }

    iiinput1a = iinput1a->begin();
    iiinput2a = iinput2a->begin();
    iiinput1b = iinput1b->begin();
    iiinput2b = iinput2b->begin();

    T_or_RGB norm1 = max1a + max1b;

    T_or_RGB norm2 = max2a + max2b;

    for(unsigned int i = 0; i < size; i++,
          ++iiinput1a,++iiinput2a,
          ++iiinput1b,++iiinput2b,
          ++iiresults)
    {
      const T_or_RGB c1mix  = abs((*iiinput1a)-(*iiinput1b))/(norm1);
      const T_or_RGB c2mix  = abs((*iiinput2a)-(*iiinput2b))/(norm2); 
      // compute crossyness at location (strength of junction)
      const T_or_RGB c1Gmix = (((*iiinput1a)+(*iiinput1b))/(norm1));
      const T_or_RGB c2Gmix = (((*iiinput2a)+(*iiinput2b))/(norm2));
      // Indicates general intensity of the output
      const T_or_RGB iOutA  = (c1Gmix+c2Gmix)/2;
      // Finish computing crossyness by subtracting lineyness
      // indicates the crispness of lines
      const T_or_RGB iOutB  = abs(c1mix-c2mix)/2; 
      //LINFO("BETA %f", *iOutB);
      // indicates the crispness of junctions
      *iiresults            = iOutA - iOutB;
      // Include the explicit instantiations
    }
    iinput1a = iinput1a + 4;
    iinput2a = iinput2a + 4;
    iinput1b = iinput1b + 4;
    iinput2b = iinput2b + 4;
    ++iresults;
  }
}

// Include the explicit instantiations
// MA #include "inst/Image_FilterOps.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
