/*!@file Image_IO.C I/O operations for Image's
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_IO.C 4290 2005-03-12 02:03:27Z mundhenk $
//

#ifndef IMAGE_IO_C_DEFINED
#define IMAGE_IO_C_DEFINED

#include "Image_IO.H"

#include "Image.H"

#include <iomanip>
#include <iostream>
#include <vector>

namespace
{
  // The point of these little functions is to convert numbers to a type
  // that will print as a number; basically the only problem is with byte,
  // which is likely a typedef for char, so if it is printed as-is, then we
  // get the characters themselves rather than their numeric
  // representation... so by overloading we can make printable(byte) return
  // an int.

  // The default printable() is just a template that returns the val unchanged.
  template <class T> T printable(T val) { return val; }

  // Overloads of printable() can specialize the behavior for specific
  // types, such as byte here:
  int printable(byte val) { return int(val); }

  template <class T>
  std::ostream& operator<<(std::ostream& os, const PixRGB<T>& pix)
  {
    os << '(' << printable(pix.red())
       << ',' << printable(pix.green())
       << ',' << printable(pix.blue())
       << ')';
    return os;
  }

  template <class T>
  std::ostream& operator<<(std::ostream& os, const PixH2SV<T>& pix)
  {
    T r,g,b;
    PixH2SV<T> pax = pix;
    pax.getRGBP(&r,&g,&b);
    os << '(' << printable(r)
       << ',' << printable(g)
       << ',' << printable(b)
       << ')';
    return os;
  }

}

// ######################################################################
void writeBinImage(const Image<byte>& img, std::ostream& os)
{
  std::vector<byte> mask;
  for (uint i = 0; i < 8; ++i)
    mask.push_back(1 << i);

  Image<byte>::const_iterator iptr = img.begin();
  int w = img.getWidth(), h = img.getHeight();

  os << img.getWidth() << ' ' << img.getHeight() << '\n';

  byte count = 0;
  unsigned char currByte;
  for (int y = 0; y < h; ++y)
    {
      currByte = 0;
      for (int x = 0; x < w; ++x)
        {
          if (*iptr > 0) currByte |= mask[count];
          ++count;
          ++iptr;
          if ((count == 8) || (x == (w - 1)))
            {
              os << (int)currByte << " ";
              count = 0;
              currByte = 0;
            }
        }
      os << '\n';
    }
}

// ######################################################################
Image<byte> readBinImage(std::istream& is)
{
  std::vector<unsigned char> mask;
  for (uint i = 0; i < 8; ++i)
    mask.push_back(1 << i);

  byte count = 0;
  byte currByte = 0;
  int tmp;
  int w, h;
  is >> w;
  is >> h;

  Image<byte> img(w,h,NO_INIT);
  Image<byte>::iterator iptr = img.beginw();

  for (int y = 0; y < h; ++y)
    {
      count = 0;
      for (int x = 0; x < w; ++x)
        {
          if (count == 0)
            {
              is >> tmp; 
              currByte = (byte)tmp;
            }
          if ((currByte & mask[count]) > 0) *iptr = byte(255);
          else *iptr = byte(0);
          ++count;
          if (count == 8) count = 0;
          ++iptr;
        }
    }
  return img;
}

// ######################################################################
// Get the explicit instantiations:
//#include "inst/Image_IO.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_IO_C_DEFINED
