/*!@file Image_MathOps.C Mathematical operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_MathOps.C 4232 2005-02-24 07:49:57Z itti $

#ifndef IMAGE_MATHOPS_C_DEFINED
#define IMAGE_MATHOPS_C_DEFINED

#include "Image_MathOps.H"
#include "MathFunctions.H"
#include "Image_ColorOps.H"
#include "Image.H"
#include "Range.H"

#include <algorithm>
#include <cmath>
#include <numeric> // for std::accumulate(), etc.

#ifdef USE_MMX
#include "mmx-sse.H"
#endif

// ######################################################################
// Optimized implementations using SSE2/SSE for some canonical datatypes:
#ifdef USE_SSE
double sum(const Image<double> &a)
{
  double d;
  sse_sum((const double *)(a.begin()), &d, a.getSize());
  return d;
}
#endif
#ifdef USE_SSE2
double sum(const Image<float> &a)
{
  double d;
  sse2_sum((const float *)(a.begin()), &d, a.getSize());
  return d;
}

double sum(const Image<int> &a)
{
  double d;
  sse2_sum((const int *)(a.begin()), &d, a.getSize());
  return d;
}

double sum(const Image<byte> &a)
{
  double d;
  sse2_sum((const byte *)(a.begin()), &d, a.getSize());
  return d;
}
#endif


// ######################################################################
Image<byte> normalizeWithScale(const Image<float>& src)
{
  // First, remember the original range of src
  assert(false);  // MA
  const Range<float> r /*= rangeOf(src)*/;

  // Now convert to byte and normalize to [0,255]
  Image<float> nsrc = src;
  nsrc.normalize(0.0f, 255.0f);
  Image<byte> bytesrc = nsrc;

  // Allow 12 pixels width per character; number of characters should be
  // at least 12, so min width of 144

  const int width = std::max(src.getWidth(), 144);

  Image<byte> result(width, src.getHeight() + 50, ZEROS);
  result.paste(bytesrc, Point2D(0,0));

  // Write text showing the min+max values from the original src

  char buf[256];

  snprintf(buf, 256, "min: %f", r.min());
  result.writeText(Point2D(0, src.getHeight()), buf);

  snprintf(buf, 256, "max: %f", r.max());
  result.writeText(Point2D(0, src.getHeight()+25), buf);

  return result;
}

// ######################################################################
// Optimized absDiff implementations using SSE2 for some canonical datatypes:
#ifdef USE_SSE2
Image<byte> absDiff(const Image<byte>& b, const Image<byte> &c)
{
  Image<byte> result(b.getDims(), NO_INIT);
  sse2_absDiff((const byte *)(b.begin()),(const byte *)(c.begin()),
              (byte *)(result.beginw()), b.getSize());
  return result;
}

Image<float> absDiff(const Image<float>& b, const Image<float> &c)
{
  Image<float> result(b.getDims(), NO_INIT);
  sse2_absDiff((const float *)(b.begin()),(float *)(c.begin()),(float *)
              (result.beginw()), b.getSize());
  return result;
}

Image<int> absDiff(const Image<int>& b, const Image<int> &c)
{
  Image<int> result(b.getDims(), NO_INIT);
  sse2_absDiff((const int *)(b.begin()),(const int *)(c.begin()),
              (int *)(result.beginw()), b.getSize());
  return result;
}
#endif
#ifdef USE_SSE
Image<double> absDiff(const Image<double>& b, const Image<double> &c)
{
  Image<double> result(b.getDims(), NO_INIT);
  sse_absDiff((const double *)(b.begin()),(double *)(c.begin()),
              (double *)(result.beginw()), b.getSize());
  return result;
}
#endif


#ifdef USE_SSE

// ######################################################################

Image<float> clampedDiff(const Image<float>& b, const Image<float>& c)
{
  ASSERT(c.isSameSize(b));

  Image<float> result(b.getDims(), NO_INIT);

  const float *bptr= b.begin();
  const float *cptr= c.begin();

  sse_clampedDiff(bptr, cptr,
                  (float *) (result.beginw()), b.getSize());

  return result;
}


// ######################################################################

Image<byte> clampedDiff(const Image<byte>& b, const Image<byte>& c)
{
  ASSERT(c.isSameSize(b));

  Image<byte> result(b.getDims(), NO_INIT);

  const byte *aptr = (const_cast <byte *> (b.begin()));
  const byte *bptr = (const_cast <byte *> (c.begin()));
  byte *dptr= (const_cast <byte *> (result.beginw()));

  sse_clampedDiff( aptr, bptr, dptr, b.getSize());

  return result;
}

// ######################################################################

Image<int> clampedDiff(const Image<int>& b, const Image<int>& c)
{
  ASSERT(c.isSameSize(b));

  Image<int> result(b.getDims(), NO_INIT);

  const int *bptr = (const_cast <int *> (b.begin()));
  const int *cptr = (const_cast <int *> (c.begin()));
  int *dptr = (const_cast <int *> (result.beginw()));

  sse_clampedDiff(bptr, cptr, dptr, b.getSize());

  return result;
}


#endif

// ######################################################################
double corrcoef(const Image<float>& img1, const Image<float>& img2)
{
  ASSERT(img1.isSameSize(img2));

  Image<float>::const_iterator p1 = img1.begin();
  Image<float>::const_iterator p2 = img2.begin();
  const Image<float>::const_iterator stop1 = img1.end();

  double sum1 = 0.0;
  double sum2 = 0.0;

  double sumsq1 = 0.0;
  double sumsq2 = 0.0;

  double sumprod = 0.0;

  while (p1 != stop1)
    {
      sum1    += (*p1);
      sum2    += (*p2);
      sumsq1  += (*p1) * (*p1);
      sumsq2  += (*p2) * (*p2);
      sumprod += (*p1) * (*p2);

      ++p1;
      ++p2;
    }

  const int n = img1.getSize();

  const double avg1 = sum1 / n;
  const double avg2 = sum2 / n;

  const double rsq =
    ((sumsq1 - n*avg1*avg1) != 0.0 && (sumsq2 - n*avg2*avg2) != 0.0)
    ? (SQR(sumprod - n*avg1*avg2)
       /
       ( (sumsq1 - n*avg1*avg1) * (sumsq2 - n*avg2*avg2) ))
    : 0.0;

  ASSERT(rsq >= 0.0);
  ASSERT(rsq <= 1.0);

  return rsq;
}

// ######################################################################
Image<float> matrixMult(const Image<float>& A, const Image<float>& B)
{
  ASSERT(A.initialized());
  ASSERT(B.initialized());
  const int wa = A.getWidth(), ha = A.getHeight();
  const int wb = B.getWidth(), hb = B.getHeight();
  ASSERT(wa == hb);

  Image<float> C(wb, ha, ZEROS);
  Image<float>::iterator pc = C.beginw();

  for (int yc = 0; yc < ha; ++yc)
    for (int xc = 0; xc < wb; ++xc, ++pc)
      for (int xa = 0; xa < wa; ++xa)
        *pc += (A.getVal(xa,yc) * B.getVal(xc,xa));

  return C;
}

// ######################################################################
Image<float> matrixMult(const Image<float>& A, const Image<float>& B,
                        unsigned int *beginAX, unsigned int *endAX, 
                        unsigned int *beginBX, unsigned int *endBX,
                        unsigned int *beginAY, unsigned int *endAY)
{  
  ASSERT(A.initialized());
  ASSERT(B.initialized());
  ASSERT(*beginAX <= *endAX);
  ASSERT(*beginBX <= *endBX);  
  ASSERT(*beginAY <= *endAY); 
  const int hb = B.getHeight();
  ASSERT(hb == (signed)((*endAX - *beginAX)+1)); 
  Image<float> C((signed)(*endBX - *beginBX),
                 (signed)(*endAY - *beginAY), ZEROS);
  Image<float>::iterator pc = C.beginw();
  unsigned int ya; 

  for (unsigned int yc = *beginAY; yc < *endAY; yc++)
    for (unsigned int xc = *beginBX; xc < *endBX; xc++, ++pc)
    {
      ya = 0;
      for (unsigned int xa = *beginAX; xa < *endAX; xa++ , ya++)
        *pc += (A.getVal(xa,yc) * B.getVal(xc,ya));
    }           
  return C;
}

// ######################################################################
Image<float> matrixMult(const Image<float>& A, const Image<float>& B,
                        unsigned long *beginAX, unsigned long *endAX, 
                        unsigned long *beginBX, unsigned long *endBX,
                        unsigned long *beginAY, unsigned long *endAY)
{  
  ASSERT(A.initialized());
  ASSERT(B.initialized());
  ASSERT(*beginAX <= *endAX);
  ASSERT(*beginBX <= *endBX);  
  ASSERT(*beginAY <= *endAY); 
  const long hb = B.getHeight();
  ASSERT(hb == (signed)((*endAX - *beginAX)+1)); 
  Image<float> C((signed)(*endBX - *beginBX),
                 (signed)(*endAY - *beginAY), ZEROS);
  Image<float>::iterator pc = C.beginw();
  unsigned long ya; 

  for (unsigned long yc = *beginAY; yc < *endAY; yc++)
    for (unsigned long xc = *beginBX; xc < *endBX; xc++, ++pc)
    {
      ya = 0;
      for (unsigned long xa = *beginAX; xa < *endAX; xa++ , ya++)
        *pc += (A.getVal(xa,yc) * B.getVal(xc,ya));
    }           
  return C;
}

// ######################################################################
Image<float> eye(const int size)
{
  Image<float> result(size,size,ZEROS);
  Image<float>::iterator rptr = result.beginw();
  for (int i = 0; i < size; ++i)
    {
      *rptr = 1;
      rptr += (size+1);
    }

  return result;
}

// ######################################################################
Image<float> logSig(const Image<float>& ima, float o, float b)
{ 
  const int w = ima.getWidth(), h = ima.getHeight();
  Image<float> result(w,h,ZEROS);
  Image<float>::const_iterator ptr;
  Image<float>::iterator res = result.beginw();
  for(ptr = ima.begin(); ptr != ima.end(); ++ptr,++res)
    //*res = 1/(1+pow(2.718281,(-1*(((*ptr)-o)*b))));
    *res = 1/(1+pow(2.718281,-(b*(*ptr-o))));
  return result;
}

// ######################################################################

int32 findMonteMap(Image<float>& ima,
                   std::vector<Point2D>* coords,
                   int decimation, float bias = 1.0F)
{  
  const unsigned int siz = (unsigned)ima.getSize(); ASSERT(siz > 0);
  std::vector<Point2D>::iterator itr;
  ASSERT(siz <= coords->size());
  Image<float>::iterator a = ima.beginw();

  float minVal     = *a;
  float maxVal     = 0;

  //First find upper and lower bounds on image values
  //apply an exponent bias if needed
  if(bias != 1.0F)
    for(a = ima.beginw(); a != ima.endw(); ++a)
    {
      *a = *a * *a;
      //*a = pow(*a,bias); 
    }

  for(a = ima.beginw(); a != ima.endw(); ++a)
  { 
    if(*a > maxVal)
      maxVal = *a;
    if(*a < minVal)
      minVal = *a;
  }

  a                 = ima.beginw();
  itr               = coords->begin();
  int32 counter     = 0;
  float salVal      = 0.0F;

  // determin which pixels are selected based upon there 
  // relation to min/max values
  for (unsigned int i = 0; i < siz; i ++)
  {  
    // higher order bits (better)
    salVal = ((minVal)+((maxVal)*rand()/(RAND_MAX+1.0)));
    // lower order bits (faster)
    //salVal = (*minVal)+(rand() % (int)*maxVal);
    if(*a > salVal) 
    {
      *itr = Point2D(i % ima.getWidth(), i/ima.getWidth()); 
      ++itr;
      counter++;
    }
    ++a;
  }

  return counter;
}

// ######################################################################
int32 makeSparceMap(std::vector<Point2D>* coords, std::vector<Point2D*>* cmap,
                    std::vector<Point2D>* Oldcoords,
                    std::vector<bool>* keep, int inPoints, int outPoints)
{
  ASSERT(coords->size() >= cmap->size());
  std::vector<Point2D*>::iterator icmap;
  std::vector<Point2D>::iterator icoords;
  std::vector<Point2D>::iterator iOldcoords;
  std::vector<bool>::iterator ikeep;
  int interval = inPoints/(outPoints+1);
  int mod = 0;
  int count = 0;
  icoords = coords->begin();
  iOldcoords = Oldcoords->begin();
  icmap = cmap->begin();
  ikeep = keep->begin();
  for(int i = 0; i < inPoints; i++)
  {
    if(mod == 0)
    {
      if(*ikeep == false)
        *icmap = &*icoords;
      else
        *icmap = &*iOldcoords;
      ++icmap;
      ++ikeep;
      ++iOldcoords;
      count++;
    }
    else
    {
      if(mod == interval)
        mod = -1;
    }
    mod++;
    ++icoords;  
  }
  return count;
}

// ######################################################################

int32 makeCoordArray(std::vector<Point2D>* coords, unsigned int initX, unsigned int initY, unsigned int sizeX, unsigned int sizeY)
{
  coords->resize(initX*initY);
  std::vector<Point2D>::iterator icoords = coords->begin();
  float scaleX = sizeX / initX;
  float scaleY = sizeY / initY;
  for(unsigned int i = 0; i < initX; i++)
  {
    for(unsigned int j = 0; j < initY; j++)
    {
      icoords->i = (int)(i*scaleX);
      icoords->j = (int)(j*scaleY);
    }
  }
  return initX*initY;
}

// Include the explicit instantiations
// MA #include "inst/Image_MathOps.I"

// additional explicit instantiations:
/*template double sum(const Image<double>& a);
template double mean(const Image<double>& a);
template Image<double> abs(const Image<double>& a);
template Image<byte> thresholdedMix(const Image<float>& mask,
                                    const float& thresh,
                                    const Image<byte>& lower,
                                    const Image<byte>& higher);
template Image<double> toPower(const Image<double>& a, double pow);*/

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_MATHOPS_C_DEFINED
