/*!@file Image_MathOps.H Mathematical operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_MathOps.H 4147 2005-01-27 07:26:48Z mundhenk $
//

#ifndef IMAGE_MATHOPS_H_DEFINED
#define IMAGE_MATHOPS_H_DEFINED
#include <vector>
#include "saliency.H"
#include "Promotions.H"
#include "Point2D.H"

class Rectangle;
template <class T> class Image;
template <class T> class Range;

//! Equality: true if the images are the same size and all pixels are equal.
template <class T_or_RGB>
bool operator==(const Image<T_or_RGB>& img1, const Image<T_or_RGB>& img2);

//! Returns the sum of all input pixels.
template <class T>
double sum(const Image<T>& a);

//! Get the average of all input pixels.
template <class T>
double mean(const Image<T>& a);

//! Get the min/max range of the input pixels.
/*! This provides similar functionality to Image<T>::getMinMax(). */
template <class T>
Range<T> rangeOf(const Image<T>& img);

//! Linearly remap pixels from the source to dest dynamic ranges.
/*! This is similar to, but more general than, Image<T>::normalize(). In
    effect, Image<T>::normalize() always assumes that the source range is
    the image's own min/max range. In contrast, remapRange() allows one to
    find the global min/max across a range of images, and then use that as
    the source range while normalizing each the images using that range. */
template <class T>
Image<T> remapRange(const Image<T>& img,
                    const Range<T>& from, const Range<T>& to);

//! Converts a float image to a byte image, normalizing to [0.0-255.0].
/*! Two lines of text are appended to the output indicating the original
    min and max values of the float image. */
Image<byte> normalizeWithScale(const Image<float>& src);

//! Square each element of the input image.
template <class T>
Image<T> squared(const Image<T>& a);

//! Raise each element of the input image to the given power.
template <class T>
Image<typename promote_trait<T,float>::TP>
toPower(const Image<T>& a, double pow);

//! Take absolute value of every pixel.
template <class T>
Image<T> abs(const Image<T>& a);

//! Compute exp(pixel / (2 * sigma^2)) for every pixel.
template <class T>
Image<typename promote_trait<T,float>::TP>
hmaxActivation(const Image<T>& a, const float sigma);

//! Returns abs(b - c).
template <class T>
Image<T> absDiff(const Image<T>& b, const Image<T>& c);

//! Returns abs(b - c) for each color channel
template <class T>
Image< PixRGB<T> > absDiff(const Image< PixRGB<T> >& b,
                           const Image< PixRGB<T> >& c);

//! Returns (b - c), clamping to zero all negative values.
template <class T>
Image<T> clampedDiff(const Image<T>& b, const Image<T>& c);

//! Returns (b - c) for each color channel, clamping to zero all neg. values
template <class T>
Image< PixRGB<T> > clampedDiff(const Image< PixRGB<T> >& b,
                               const Image< PixRGB<T> >& c);

//! Returns (val - a).
template <class T>
Image<T> binaryReverse(const Image<T>& a, const T val);

//! Returns (b + c) / 2, clean & prevents overflows.
template <class T>
Image<T> average(const Image<T>& b, const Image<T>& c);

//! Returns (a*w + b) / 2, the weighted average between images  
template <class T>
Image<T> averageWeighted(Image<T>& a, const Image<T>& b, T *aWeight);

//! Returns max(a, b), element by element.
template <class T>
Image<T> takeMax(const Image<T>& a, const Image<T>& b);

//! Retuns max(a, b) for each color component separatly
template <class T>
Image< PixRGB<T> > takeMax(const Image< PixRGB<T> >& a, 
                           const Image< PixRGB<T> >& b);

//! Returns min(a, b), element by element.
template <class T>
Image<T> takeMin(const Image<T>& a, const Image<T>& b);

//! Compute energy for quadrature pair (sqrt(img1^2 + img2^2)).
template <class T_or_RGB>
Image<T_or_RGB> quadEnergy(const Image<T_or_RGB>& img1,
                           const Image<T_or_RGB>& img2);

//! RMS error between arr1 and arr2.
template <class T>
float RMSerr(const Image<T>& arr1, const Image<T>& arr2);

//! Create transparency overlay of two images
/*! by laying top over bottom with transparency of top set to
 trans. A new image will be returned (use 0 to 100 for percent
 transparency). NOTE: top and bottom must be the same size!
 this function only works for grey scale images */
template <class T>
Image<T> overlay(const Image<T>& top, const Image<T>& bottom,
                 const float trans);

//! sum pixels along the X and Y axes (sumX, sumY) and in total (return value)
/*!@param img is a w x h size image
  @param sumX will return a vector of length w with the sum over the rows of img
  @param sumY will return a vector of length h with the sum over the columns*/
template <class T>
float sumXY(const Image<T>& img, std::vector<float>& sumX,
            std::vector<float>& sumY);

//! Returns Euclidean distance between two images
/*! Returns sqrt(sum_over_pixels((img1[i] - img2[i])^2)) */
template <class T>
double distance(const Image<T> &img1, const Image<T> &img2);

//! Returns weighted Euclidean distance between two images
/*! Returns sqrt(sum_over_pixels(weight[i] * (img1[i] - img2[i])^2)) */
template <class T>
double distance(const Image<T> &img1, const Image<T> &img2,
                const Image<float> &weight);

//! Returns the correlation coefficient r^2 of the two images.
double corrcoef(const Image<float>& img1, const Image<float>& img2);

//! Returns the cross-correlation coefficient r between two patches
/*! Note that the computations done here differ from what is done in
  corrcoeff, in that a signed correlation coefficient is returned
  (between -1 and 1), and we add a convention that if both patches
  have zero variance then the correlation is 1. If only one patch has
  zero variance then the correlation is zero. Current implementation
  requires that the patch fits within both image bounds, and assumes a
  small patch size so that it will not attempt to use a faster
  FFT-based implementation which can save time with larger patches.
  @param img1 first image
  @param topleft1 coordinates of top-left corner of patch in img1
  @param patchdims dimensions of patch
  @param img2 second image
  @param topleft2 coordinates of top-left corner of patch in img2
  @param eps epsilon used to determine whether patch variance is zero */
template <class T>
double corrpatch(const Image<T>& img1, const Point2D& topleft1,
                 const Dims& patchdims, const Image<T>& img2,
                 const Point2D& topleft2, const double eps = 1.0e-20);

//! Returns peak signal-to-noise ratio
/*! pSNR is computed as 10.log_10(255^2/sigma^2) where
  sigma^2=/N.sum(x1_i - x2_i)^2 and is a measure of distortion between
  two images */
template <class T>
double pSNR(const Image<T> &img1, const Image<T> &img2);

//! Returns weighted peak signal-to-noise ratio
/*! This is like the other pSNR() except that it uses the weighted
  version of distance() internally */
template <class T>
double pSNR(const Image<T> &img1, const Image<T> &img2,
            const Image<float>& weight);

//! Take square root of all elements
template <class T>
Image<typename promote_trait<T,float>::TP> sqrt(const Image<T>& img);

//! Take inverse of all elements > eps in abs, 0 otherwise
template <class T>
Image<typename promote_trait<T,float>::TP> inverse(const Image<T>& img,
                                                   const T eps);

//! Take exponential of all pixels
template <class T>
Image<typename promote_trait<T,float>::TP> exp(const Image<T>& img);

//! Take negative exponential of all pixels, that is exp(-x)
template <class T>
Image<typename promote_trait<T,float>::TP> negexp(const Image<T>& img);


//! determine the first and last non-zero values and the centroid
/*!@param vect a vector of numbers
  @param centroid returns the centroid of vect
  @param first returns the position of the first non-zero value in vect
  @param last returns the position of the last non-zero value in vect
  @return true if we actually found all these values, false if vect
  only contains zeros*/
template <class T>
bool getCentroidFirstLast(std::vector<T> vect, float& centroid,
                          int& first, int& last);

//! determine the centroid of all the points in the image
/*! The centroid is rounded to integer pixel positions.
  When there are no values > 0 in the img, (-1,-1) is returned
  @param boundingBox the bounding box of the object(s) contained in
  the image is returned
  @param cenX returns the float value of the centroid along the x axis
  @param cenY returns the float value of the centroid along the y axis
  @return the centroid rounded to integer coordinates*/
template <class T>
Point2D centroid(const Image<T>& img, Rectangle& boundingBox,
                 float& cenX, float& cenY);

//! determine the center of mass of all the points in the image
/*! The centroid is rounded to integer pixel positions.
  When there are no values > 0 in the img, (-1,-1) is returned */
template <class T>
Point2D centroid(const Image<T>& img);

//! Apply a squashing function to the pixel values
/*! The squashing function is a 4th-order polynomial with sigmoidal
  shape. Denoting by x a pixel value in the original image and by
  y=f(x) the transformed pixel value that will be put in the result
  image, the contraints on the polynomial are: 1) f(oldmin)=newmin, 2)
  f(oldmax)=newmax, 3) f(oldmid)=newmid, 4) f'(oldmin)=0, and 5)
  f'(oldmax)=0. So, typically, the polynomial will have horizontal
  slope at both ends of the input range, and will remap the old
  min/max to the new min/max while also remapping a user-chosen
  mid-point oldmid to newmid. Playing with this mid-point allows the
  user to give more emphasis to higher or lower values while
  remapping. This is the full form, but two simplified forms are also
  available that will compute some of the arguments here from the
  input image. Because the main use of this function is when oldmin is
  the true minimum value of the input image, and oldmax its true
  maximum, here we will not waste time checking for input values
  outside this range, and their remapped values are unclear as the
  polynomial is unconstrained outside the input range. Internal
  remapping is done using floats, and results are converted back (with
  possible clamping) to the same type as the original image, so that
  this function may be efficiently used to do contrast adjustments in
  Image<byte> data. */
template<class T>
Image<T> squash(const Image<T>& src,
                const T oldmin, const T newmin,
                const T oldmid, const T newmid,
                const T oldmax, const T newmax);

//! Apply a squashing function to the pixel values
/*! In this specialization of the general form of squash(), the old
  min/max are computed using getMinMax(). */
template<class T>
Image<T> squash(const Image<T>& src, const T newmin,
                const T oldmid, const T newmid, const T newmax);

//! Apply a squashing function to the pixel values
/*! In this specialization of the general form of squash(), the old
  min/max are computed using getMinMax(), and the new min/max will be
  kept equal to the old ones, so that there is no overall change in
  range (unless your mid-point is outside that range). */
template<class T>
Image<T> squash(const Image<T>& src, const T oldmid, const T newmid);

//! Mix between two images based on comparing mask values to a threshold
/*! On every pixel, if mask >= thresh, the returned value is taken
  from higher, otherwise it is taken from lower. All three of lower,
  highr and mask must have same dims. */
template <class T, class T_or_RGB>
Image<T_or_RGB> thresholdedMix(const Image<T>& mask, const T& thresh,
                               const Image<T_or_RGB>& lower,
                               const Image<T_or_RGB>& higher);

//! Matrix Multiplication: C = A * B
/*! The images A and B are interpreted as matrices. For multiplying
  a (m x n) with a (n x p) matrix, this is the generic matrix multiplication
  with complexity (m * n * p).*/
Image<float> matrixMult(const Image<float>& A, const Image<float>& B);

//! Matrix Multiplication: C = A * B
/*! The images A and B are interpreted as matrices. For multiplying
  a (m x n) with a (n x p) matrix, this is the generic matrix multiplication
  with complexity (m * n * p). This version will allow you to multiply
  values in a matrix that lie between index endA/B and beginA/B. The
  destination matrix will be the same size as:
  (endAX - beginAX) x (endAX - beginAX)
  per usual the width of A (indexed between beginAX and endAX)
  and height of B MUST be equal. If not then
  perhaps use transpose...
  @param A input matrix 1
  @param B input matrix 2
  @param beginAX starting (width) index in matrix A
  @param endAX ending (width) index in matrix A
  @param beginBY starting (height) index in matrix B
  @param endBY ending (height) index in matrix B
  @param beginAY starting (height) index in matrix A
  @param endAY ending (height) index in matrix A
*/
Image<float> matrixMult(const Image<float>& A, const Image<float>& B,
                        unsigned int *beginAX, unsigned int *endAX, 
                        unsigned int *beginBY, unsigned int *endBY,
                        unsigned int *beginAY, unsigned int *endAY);
//! Matrix Multiplication: C = A * B
/*! The images A and B are interpreted as matrices. For multiplying
  a (m x n) with a (n x p) matrix, this is the generic matrix multiplication
  with complexity (m * n * p). This version will allow you to multiply
  values in a matrix that lie between index endA/B and beginA/B. The
  destination matrix will be the same size as:
  (endAX - beginAX) x (endAX - beginAX)
  per usual the width of A (indexed between beginAX and endAX)
  and height of B MUST be equal. If not then
  perhaps use transpose...
  This method is for large matrices
  @param A input matrix 1
  @param B input matrix 2
  @param beginAX starting (width) index in matrix A
  @param endAX ending (width) index in matrix A
  @param beginBY starting (height) index in matrix B
  @param endBY ending (height) index in matrix B
  @param beginAY starting (height) index in matrix A
  @param endAY ending (height) index in matrix A
*/
Image<float> matrixMult(const Image<float>& A, const Image<float>& B,
                        unsigned long *beginAX, unsigned long *endAX, 
                        unsigned long *beginBY, unsigned long *endBY,
                        unsigned long *beginAY, unsigned long *endAY);

//! Matrix Multiplication: C = A * B
/*! The images A and B are interpreted as matrices. For multiplying
  a (m x n) with a (n x p) matrix, this is the generic matrix multiplication
  with complexity (m * n * p). This version will allow you to multiply
  values in a matrix that lie between index endA/B and beginA/B. The
  destination matrix will be the same size as:
  (endAX - beginAX) x (endAX - beginAX)
  per usual the width of A (indexed between beginAX and endAX)
  and height of B MUST be equal. If not then
  perhaps use transpose...
  This method is for large matrices
  @param A input matrix 1
  @param B input matrix 2
  @param beginAX starting (width) index in matrix A
  @param endAX ending (width) index in matrix A
  @param beginBY starting (height) index in matrix B
  @param endBY ending (height) index in matrix B
  @param beginAY starting (height) index in matrix A
  @param endAY ending (height) index in matrix A
*/
template <class T>
std::vector<std::vector<T> > matrixMult(std::vector<std::vector<T> > *A,
                                        std::vector<std::vector<T> > *B,
                                        unsigned long *beginAX,
                                        unsigned long *endAX, 
                                        unsigned long *beginBX,
                                        unsigned long *endBX,
                                        unsigned long *beginAY,
                                        unsigned long *endAY);

//! transpose matrix M
template <class T_or_RGB>
Image<T_or_RGB> transpose(const Image<T_or_RGB>& M);

//! return the identity matrix of dimensions (size x size)
Image<float> eye(const int size);

//! take the logistic sigmoid 1/(1+e^x(...)) over the image
/*! This is a standard logistic sigmoid with offset o and slope b
  @param ima the input image
  @param o Offset for this sigmoid
  @param b Slope for this sigmoid
 */
Image<float> logSig(const Image<float>& ima, float o, float b);

//! randomly scramble image pixels
template <class T_or_RGB>
Image<T_or_RGB> scramble(const Image<T_or_RGB>& ima);

//! take in an image and return it's statistically most relevent points
/*! 
  Input an image to find a monte carlo like map. This will in essence cut
  out about half the pixels in a simple quick fashion by applying  the formula:
  select if -> pow(pixelValue,bias) > minVal+(maxVal*rand()/RAND_MAX+1.0)
  The bias is used to scew the distribution using an exponent
  if no bias is given, this method will work more efficently by skipping
  the exponent computation over each pixel.
  @param ima This is the input probability map
  @param coords this is a pointer to a vector which will hold selected coords
  @param bais this scews the probability map by pow(pixVal,bias)
  
*/
int32 findMonteMap(Image<float>& ima,
                   std::vector<Point2D>* coords,
                   int decimation, float bias);

//! Take in a vector and decimate it according how many points you want back
/*!
  In essence this will decimate a list of coordinates attempting to create a
  list of the size in outPoints. Thus, if you have list of 23890 points and 
  want a list of 300 points, this will take every nth point and create a
  list of about this size.
  @param coords This is the list of input coordinates
  @param cmap This is the final list of coordinates after decimation
  @param inPoints this is the input list size
  @param outPoints this is how many points you would like out
*/

int32 makeSparceMap(std::vector<Point2D>* coords, std::vector<Point2D*>* cmap,
                    std::vector<Point2D>* Oldcoords,
                    std::vector<bool>* keep, int inPoints, int outPoints);


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_MathOps_templates.h"

#endif // !IMAGE_MATHOPS_H_DEFINED
