/*!@file Image_MathOps.C Mathematical operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_MathOps.C 4232 2005-02-24 07:49:57Z itti $

#ifndef IMAGE_MATHOPS_C_DEFINED
#define IMAGE_MATHOPS_C_DEFINED

#include "Image_MathOps.H"
#include "MathFunctions.H"
#include "Image_ColorOps.H"
#include "Image.H"
#include "Range.H"

#include <algorithm>
#include <cmath>
#include <numeric> // for std::accumulate(), etc.

#ifdef USE_MMX
#include "mmx-sse.H"
#endif

namespace
{
  template <class T>
  struct ToPow
  {
    double pp;
    ToPow(double p) : pp(p) {}
    T operator()(T x) { return clamped_convert<T>(pow(x, pp)); }
  };

  template <class T>
  struct EpsInv
  {
    double ee;
    EpsInv(T e) : ee(e) {}
    T operator()(T x) {
      if (ABS<T>(x) > ee) return clamped_convert<T>(1.0F / x);
      else return T();
    }
  };
}

// ######################################################################
template <class T>
bool operator==(const Image<T>& img1, const Image<T>& img2)
{
  if (!img1.isSameSize(img2)) return false;

  return std::equal(img1.begin(), img1.end(),
                    img2.begin());
}

// ######################################################################
template <class T>
double sum(const Image<T>& a)
{
  return std::accumulate(a.begin(), a.end(), 0.0);
}


// ######################################################################
template <class T>
double mean(const Image<T>& a)
{
  ASSERT(a.getSize() > 0);
  return sum(a) / double(a.getSize());
}

// ######################################################################
template <class T>
Range<T> rangeOf(const Image<T>& img)
{
  return std::for_each(img.begin(), img.end(), Range<T>());
}

// ######################################################################
template <class T>
Image<T> remapRange(const Image<T>& img,
                    const Range<T>& from, const Range<T>& to)
{
  Image<T> result(img.getDims(), NO_INIT);

  typename Image<T>::const_iterator sptr = img.begin();
  typename Image<T>::iterator dptr = result.beginw(), stop = result.endw();

  double scale = double(to.range()) / double(from.range());

  while (dptr != stop)
    {
      *dptr = to.min() + T((*sptr - from.min()) * scale);
      ++sptr; ++dptr;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> squared(const Image<T>& a)
{
  Image<T> result(a.getDims(), NO_INIT);
  std::transform(a.begin(), a.end(), result.beginw(), SQR<T>);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP>
toPower(const Image<T>& a, double pow)
{
  Image<typename promote_trait<T,float>::TP> result(a.getDims(), NO_INIT);
  std::transform(a.begin(), a.end(), result.beginw(), ToPow<T>(pow));
  return result;
}

// ######################################################################
template <class T>
Image<T> abs(const Image<T>& a)
{
  Image<T> result(a.getDims(), NO_INIT);
  std::transform(a.begin(), a.end(), result.beginw(), ABS<T>);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP>
hmaxActivation(const Image<T>& a, const float sigma)
{
  typedef typename promote_trait<T, float>::TP TF;
  Image<TF> result(a.getDims(), NO_INIT);
  float sq = 0.5F / (sigma * sigma);
  int siz = a.getSize();
  typename Image<T>::const_iterator sptr = a.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  for (int i = 0; i < siz; i ++)
    *dptr++ = clamped_convert<TF>(exp((*sptr++) * sq));

  return result;
}

// ######################################################################
template <class T>
Image<T> absDiff(const Image<T>& b, const Image<T>& c)
{
  ASSERT(c.isSameSize(b));

  Image<T> result(b.getDims(), NO_INIT);

  typename Image<T>::const_iterator bptr = b.begin();
  typename Image<T>::const_iterator cptr = c.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();

  while (dptr != stop)
    {
      if (*bptr < *cptr) *dptr++ = *cptr++ - *bptr++;
      else *dptr++ = *bptr++ - *cptr++;
    }

  return result;
}

// ######################################################################
template <class T>
Image< PixRGB<T> > absDiff(const Image< PixRGB<T> >& b,
                           const Image< PixRGB<T> >& c)
{
  ASSERT(c.isSameSize(b));

  Image< PixRGB<T> > result(b.getDims(), NO_INIT);

  typename Image< PixRGB<T> >::const_iterator bptr = b.begin();
  typename Image< PixRGB<T> >::const_iterator cptr = c.begin();
  typename Image< PixRGB<T> >::iterator dptr = result.beginw();
  typename Image< PixRGB<T> >::iterator stop = result.endw();

  while (dptr != stop)
    {
      if (bptr->red() < cptr->red()) dptr->setRed(cptr->red() - bptr->red());
      else dptr->setRed(bptr->red() - cptr->red());

      if (bptr->blue() < cptr->blue()) dptr->setBlue(cptr->blue() - bptr->blue());
      else dptr->setBlue(bptr->blue() - cptr->blue());

      if (bptr->green() < cptr->green()) dptr->setGreen(cptr->green() - bptr->green());
      else dptr->setGreen(bptr->green() - cptr->green());

      ++bptr; ++cptr; ++dptr;
    }

  return result;
}


// ######################################################################
template <class T>
Image<T> clampedDiff(const Image<T>& b, const Image<T>& c)
{
  ASSERT(c.isSameSize(b));

  Image<T> result(b.getDims(), NO_INIT);

  T zero = T();

  typename Image<T>::const_iterator bptr = b.begin();
  typename Image<T>::const_iterator cptr = c.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();

  while (dptr != stop)
    {
      if (*bptr < *cptr) *dptr = zero; else *dptr = *bptr - *cptr;
      bptr ++; cptr ++; dptr ++;
    }

  return result;
}

// ######################################################################
template <class T>
Image< PixRGB<T> > clampedDiff(const Image< PixRGB<T> >& b,
                               const Image< PixRGB<T> >& c)
{
  ASSERT(c.isSameSize(b));

  Image< PixRGB<T> > result(b.getDims(), NO_INIT);

  T zero = T();

  typename Image< PixRGB<T> >::const_iterator bptr = b.begin();
  typename Image< PixRGB<T> >::const_iterator cptr = c.begin();
  typename Image< PixRGB<T> >::iterator dptr = result.beginw();
  typename Image< PixRGB<T> >::iterator stop = result.endw();

  while (dptr != stop)
    {
      if (bptr->red() < cptr->red()) dptr->setRed(zero);
      else dptr->setRed(bptr->red() - cptr->red());

      if (bptr->blue() < cptr->blue()) dptr->setBlue(zero);
      else dptr->setBlue(bptr->blue() - cptr->blue());

      if (bptr->green() < cptr->green()) dptr->setGreen(zero);
      else dptr->setGreen(bptr->green() - cptr->green());

      ++bptr; ++cptr; ++dptr;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> binaryReverse(const Image<T>& a, const T val)
{
  Image<T> result(a.getDims(), NO_INIT);

  typename Image<T>::const_iterator aptr = a.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();
  while (dptr != stop)
    {
      *dptr++ = val - *aptr++;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> average(const Image<T>& b, const Image<T>& c)
{
  ASSERT(b.isSameSize(c));
  Image<T> result(b.getDims(), NO_INIT);
  typename Image<T>::iterator aptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();
  typename Image<T>::const_iterator bptr = b.begin();
  typename Image<T>::const_iterator cptr = c.begin();

  while (aptr != stop)
    *aptr++ = (T)( (*bptr++ * 0.5F) + (*cptr++ * 0.5F) );

  return result;
}

// ######################################################################

template <class T>
Image<T> averageWeighted(Image<T>& a, const Image<T>& b, T *aWeight)
{
  ASSERT(a.isSameSize(b));
  typename Image<T>::iterator aptr;
  typename Image<T>::const_iterator bptr = b.begin();
  for(aptr = a.beginw();  aptr != a.endw(); ++aptr, ++bptr)
    *aptr = (((*aptr)*(*aWeight))+(*bptr))/(1+(*aWeight));
  return a;
}

// ######################################################################
template <class T>
Image<T> takeMax(const Image<T>& a, const Image<T>& b)
{
  ASSERT(a.isSameSize(b));

  Image<T> result(a.getDims(), NO_INIT);

  typename Image<T>::const_iterator aptr = a.begin();
  typename Image<T>::const_iterator bptr = b.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();

  while (dptr != stop)
    {
      *dptr++ = std::max(*aptr++, *bptr++);
    }

  return result;
}

// ######################################################################
template <class T>
Image< PixRGB<T> > takeMax(const Image< PixRGB<T> >& a, 
                           const Image< PixRGB<T> >& b)
{
  Image<T> ra, rb, ga, gb, ba, bb;
  getComponents(a, ra, ga, ba);
  getComponents(b, rb, gb, bb);

  return makeRGB(takeMax(ra,rb), takeMax(ga,gb), takeMax(ba,bb));
}

// ######################################################################
template <class T>
Image<T> takeMin(const Image<T>& a, const Image<T>& b)
{
  ASSERT(a.isSameSize(b));

  Image<T> result(a.getDims(), NO_INIT);

  typename Image<T>::const_iterator aptr = a.begin();
  typename Image<T>::const_iterator bptr = b.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();

  while (dptr != stop)
    {
      *dptr++ = std::min(*aptr++, *bptr++);
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> quadEnergy(const Image<T>& img1, const Image<T>& img2)
{
  ASSERT(img1.isSameSize(img2));

  Image<T> result(img1.getDims(), NO_INIT);

  typename Image<T>::const_iterator s1ptr = img1.begin();
  typename Image<T>::const_iterator s2ptr = img2.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();
  typedef typename promote_trait<T, float>::TP TF;
  while (dptr != stop)
    {
      TF s1 = *s1ptr++, s2 = *s2ptr++;
      *dptr++ = T(sqrt(s1 * s1 + s2 * s2));
    }

  return result;
}

// ######################################################################
template <class T>
float RMSerr(const Image<T>& arr1, const Image<T>& arr2)
{
  ASSERT(arr1.initialized() && arr2.initialized());
  ASSERT(arr1.isSameSize(arr2));

  typename Image<T>::const_iterator ap1 = arr1.begin();
  typename Image<T>::const_iterator ap2 = arr2.begin();
  typename Image<T>::const_iterator stop2 = arr2.end();

  float err = 0.0f;
  while (ap2 != stop2)
    { float v = *ap1++ - *ap2++; err += v * v;  }
  return sqrt(err / float(arr1.getSize()));
}

// ######################################################################
template <class T>
Image<T> overlay(const Image<T>& top, const Image<T>& bottom,
                 const float trans)
{
  // 0 to 100 percent only:
  ASSERT((trans >= 0.0F) && (trans <= 100.0F));
  // images must be same size:
  ASSERT(top.isSameSize(bottom));

  // result is the same size as the input images:
  Image<T> result(top.getDims(), NO_INIT);

  typename Image<T>::const_iterator tptr = top.begin();
  typename Image<T>::const_iterator bptr = bottom.begin();
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::iterator stop = result.endw();
  float tfac = trans * 0.01F;
  while (dptr != stop)
    {
      *dptr = T((*tptr) - ( (*tptr) - (*bptr) ) * tfac);
      ++dptr; ++bptr; ++tptr;
    }

  return result;
}

// ######################################################################
template <class T>
float sumXY(const Image<T>& img, std::vector<float>& sumX,
            std::vector<float>& sumY)
{
  int w = img.getWidth(), h = img.getHeight();
  sumX.resize(w,0.0F);
  sumY.resize(h,0.0F);
  std::vector<float>::iterator sX, sY = sumY.begin();
  float sum = 0.0F;
  typename Image<T>::const_iterator ptr = img.begin();
  for (int y = 0; y < h; y++)
    {
      sX = sumX.begin();
      for (int x = 0; x < w; x++)
        {
          //sumX[x] += (float)*ptr;
          //sumY[y] += (float)*ptr;
          *sX += (float)*ptr;
          *sY += (float)*ptr;
          sum += (float)*ptr++;
          ++sX;
        }
      ++sY;
    }

  return sum;
}

// ######################################################################
template <class T>
double distance(const Image<T> &img1, const Image<T> &img2)
{
  ASSERT(img1.isSameSize(img2));
  double d = 0.0;
  typename Image<T>::const_iterator ptr1 = img1.begin(),
    ptr2 = img2.begin(), e = img1.end();
  while(ptr1 != e) {
    d += (*ptr1 - *ptr2) * (*ptr1 - *ptr2);
    ptr1 ++; ptr2 ++;
  }
  return sqrt(d);
}

// ######################################################################
template <class T>
double distance(const Image<T> &img1, const Image<T> &img2,
                const Image<float> &weight)
{
  ASSERT(img1.isSameSize(img2) && img1.isSameSize(weight));

  double d = 0.0;
  typename Image<T>::const_iterator ptr1 = img1.begin(),
    ptr2 = img2.begin(), e = img1.end();
  Image<float>::const_iterator w = weight.begin();

  while(ptr1 != e) {
    d += *w * (*ptr1 - *ptr2) * (*ptr1 - *ptr2);
    ptr1 ++; ptr2 ++; w ++;
  }
  return sqrt(d);
}

// ######################################################################
template <class T>
double corrpatch(const Image<T>& img1, const Point2D& topleft1,
                 const Dims& patchdims, const Image<T>& img2,
                 const Point2D& topleft2, const double eps)
{
  ASSERT(img1.rectangleOk(Rectangle(topleft1, patchdims)));
  ASSERT(img2.rectangleOk(Rectangle(topleft2, patchdims)));

  const int pw = patchdims.w(), ph = patchdims.h();
  const int i1w = img1.getWidth(), i2w = img2.getWidth();

  typename Image<T>::const_iterator p1 =
    img1.begin() + topleft1.j * i1w + topleft1.i;
  typename Image<T>::const_iterator p2 =
    img2.begin() + topleft2.j * i2w + topleft2.i;
  const int stride1 = i1w - pw, stride2 = i2w - pw;

  double sum1 = 0.0, sum2 = 0.0, sumsq1 = 0.0, sumsq2 = 0.0, sumprod = 0.0;

  for (int j = 0; j < ph; j ++)
    {
      for (int i = 0; i < pw; i ++)
        {
          const double val1 = double(*p1++), val2 = (*p2++);
          
          sum1    += val1;
          sum2    += val2;
          sumsq1  += val1 * val1;
          sumsq2  += val2 * val2;
          sumprod += val1 * val2;
        }

      // switch to next row in inputs:
      p1 += stride1; p2 += stride2;
    }

  const double n = double(pw * ph);
  const double avg1 = sum1 / n, avg2 = sum2 / n;

  const double denom1 = sumsq1 - n * avg1 * avg1;
  const double denom2 = sumsq2 - n * avg2 * avg2;

  if (fabs(denom1) < eps || fabs(denom2) < eps)
    {
      if (fabs(denom1) < eps && fabs(denom2) < eps) return 1.0;
      return 0.0;
    }

  return (sumprod - n * avg1 * avg2) / sqrt(denom1 * denom2);
}

// ######################################################################
template <class T>
double pSNR(const Image<T> &img1, const Image<T> &img2)
{
  double sigma2 = distance(img1, img2);
  sigma2 = sigma2 * sigma2 / double(img1.getSize());
  return 10.0 * log10(255.0*255.0 / sigma2);
}

// ######################################################################
template <class T>
double pSNR(const Image<T> &img1, const Image<T> &img2,
            const Image<float>& weight)
{
  double sigma2 = distance(img1, img2, weight);
  sigma2 = sigma2 * sigma2 / double(img1.getSize());
  return 10.0 * log10(255.0*255.0 / sigma2);
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP> sqrt(const Image<T>& a)
{
  Image<typename promote_trait<T,float>::TP> result(a.getDims(), NO_INIT);
  typename Image<T>::const_iterator src = a.begin(), stop = a.end();
  typename Image<typename promote_trait<T,float>::TP>::iterator
    dest = result.beginw();
  while(src != stop) *dest++ = sqrt(*src++);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP> inverse(const Image<T>& a,
                                                   const T eps)
{
  Image<typename promote_trait<T,float>::TP> result(a.getDims(), NO_INIT);
  std::transform(a.begin(), a.end(), result.beginw(), EpsInv<T>(eps));
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP> exp(const Image<T>& a)
{
  Image<typename promote_trait<T,float>::TP> result(a.getDims(), NO_INIT);
  typename Image<T>::const_iterator src = a.begin(), stop = a.end();
  typename Image<typename promote_trait<T,float>::TP>::iterator
    dest = result.beginw();
  while(src != stop) *dest++ = exp(*src++);
  return result;
}

// ######################################################################
template <class T>
Image<typename promote_trait<T,float>::TP> negexp(const Image<T>& a)
{
  Image<typename promote_trait<T,float>::TP> result(a.getDims(), NO_INIT);
  typename Image<T>::const_iterator src = a.begin(), stop = a.end();
  typename Image<typename promote_trait<T,float>::TP>::iterator
    dest = result.beginw();
  while(src != stop) *dest++ = exp( - (*src++));
  return result;
}

// ######################################################################
template <class T>
bool getCentroidFirstLast(std::vector<T> vect, float& centroid,
                          int& first, int& last)
{
  double sum = 0.0;
  double mi = 0.0;
  bool before = true;
  const T zero = T();

  typename std::vector<T>::iterator vit = vect.begin();

  for (uint i = 0; i < vect.size(); ++i, ++vit)
    {
      //LINFO("vect[%d] = %g",i,vect[i]);
      mi  += (*vit * i);
      sum += *vit;
      if (before) first = i;
      if (*vit != zero)
        {
          before = false;
          last = i;
        }
    }

  // there was nobody at home?
  if (before)
    {
      centroid = -1.0F; first = -1; last = -1;
      return false;
    }

  if (sum == 0.0)
    LFATAL("The sum of non-zero numbers is zero - don't know how "
           "to compute the center of mass in this case.");

  centroid = mi / sum;
  return true;
}

// ######################################################################
template <class T>
Point2D centroid(const Image<T>& img, Rectangle& boundingBox,
                 float& cenX, float& cenY)
{
  std::vector<float> sumx, sumy;
  sumXY(img, sumx, sumy);
  int firstX, lastX, firstY, lastY;
  cenX = 0.0F; cenY = 0.0F;

  bool success = getCentroidFirstLast(sumx, cenX, firstX, lastX);
  success |= getCentroidFirstLast(sumy, cenY, firstY, lastY);

  if (!success)
    {
      boundingBox = Rectangle();
      return Point2D(-1,-1);
    }

  boundingBox.modify(firstY, firstX, lastY, lastX);
  return Point2D(int(cenX + 0.5F), int(cenY + 0.5F));
}

// ######################################################################
template <class T>
Point2D centroid(const Image<T>& img)
{
  Rectangle boundingBox;
  float cenX, cenY;
  return centroid(img, boundingBox, cenX, cenY);
}

// ######################################################################
template<class T>
Image<T> squash(const Image<T>& ima,
                const T oldmin, const T newmin,
                const T oldmid, const T newmid,
                const T oldmax, const T newmax)
{
  // if the input is blank, let's return blank:
  if (oldmin == oldmax) return ima;

  // let's compute the remapping coeffs and do some checking that we
  // actually can remap:
  float a = float(oldmin), b = float(oldmid), c = float(oldmax);
  float d = float(newmin), e = float(newmid), f = float(newmax);

  // here are the polynomial coefficients, straight out of
  // Mathematica. They are not optimized for computation speed, but we
  // don't really care here as we only need to compute them once. They
  // all need to be divided by the denominator below:
  float q0 = c*c*(b*(b-c)*(b-c)*(-4.0F*a*a + 3.0F*a*b + 2.0F*a*c - b*c)*d +
                  a*a*(a-c)*(a-c)*(a-c)*e) + a*a*(a-b)*(a-b)*b*
    (a*(b-2.0F*c) + c*(-3.0F*b + 4.0F*c))*f;
  float q1 = 2.0F*a*c*(c*c*c*c*(e-d) - 3.0F*b*b*b*b*(d-f) +
                       4.0F*b*b*b*c*(d-f) + 2.0F*a*a*a*c*(e-f) + a*a*a*a*(f-e)+
                       2.0F*a*(c*c*c*(d-e) + 2.0F*b*b*b*(d-f) +
                               3.0F*b*b*c*(f-d)));
  float q2 = a*a*a*a*a*(e-f) + a*a*a*a*c*(e-f) + 8.0F*a*a*a*c*c*(f-e) +
    (a+c)*(c*c*c*c*(d-e) + 3.0F*b*b*b*b*(d-f) + 4.0F*b*b*b*c*(f-d)) -
    4.0F*a*a*(2.0F*c*c*c*(d-e) + b*b*b*(d-f) + 3.0F*b*c*c*(f-d));
  float q3 = 2.0F*(c*c*c*c*(e-d) + 2.0F*a*a*(b-c)*(b-c)*(d-f) +
                   2.0F*b*b*c*c*(d-f) + 2.0F*a*a*a*c*(e-f) +
                   b*b*b*b*(f-d) + a*a*a*a*(f-e) +
                   2.0F*a*c*(c*c*(d-e) + b*b*(d-f) + 2.0F*b*c*(f-d)));
  float q4 = -3.0F*a*b*b*d + 2.0F*b*b*b*d + 6.0F*a*b*c*d - 3.0F*b*b*c*d -
    3.0F*a*c*c*d + c*c*c*d + a*a*a*e - 3.0F*a*a*c*e + 3.0F*a*c*c*e -
    c*c*c*e - (a-b)*(a-b)*(a + 2.0F*b - 3.0F*c)*f;

  float denom = (a-b)*(a-b)*(a-c)*(a-c)*(a-c)*(b-c)*(b-c);
  if (fabs(denom) < 1.0e-5F)
    { LERROR("Zero denominator - cannot remap - RETURNING INPUT"); return ima;}
  q0 /= denom; q1 /= denom; q2 /= denom; q3 /= denom; q4 /= denom;

  Image<T> result(ima.getDims(), NO_INIT);
  typename Image<T>::const_iterator src = ima.begin(), stop = ima.end();
  typename Image<T>::iterator dest = result.beginw();
  while(src != stop) {
    float val = float(*src++);
    // here we optimize the computation for speed:
    val = q0 + val*(q1 + val*(q2 + val*(q3 + val*q4)));
    *dest++ = clamped_convert<T>(val);
  }
  return result;
}

// ######################################################################
template<class T>
Image<T> squash(const Image<T>& src, const T newmin,
                const T oldmid, const T newmid, const T newmax)
{
  // get oldmin and oldmax from the source image:
  T oldmin, oldmax; src.getMinMax(oldmin, oldmax);

  // call the general squash:
  return squash(src, oldmin, newmin, oldmid, newmid, oldmax, newmax);
}

// ######################################################################
template<class T>
Image<T> squash(const Image<T>& src, const T oldmid, const T newmid)
{
  // get oldmin and oldmax from the source image:
  T oldmin, oldmax; src.getMinMax(oldmin, oldmax);

  // call the general squash:
  return squash(src, oldmin, oldmin, oldmid, newmid, oldmax, oldmax);
}

// ######################################################################
template <class T, class TT>
Image<TT> thresholdedMix(const Image<T>& mask, const T& thresh,
                         const Image<TT>& lower, const Image<TT>& higher)
{
  ASSERT(mask.isSameSize(lower) && mask.isSameSize(higher));

  Image<TT> result(mask.getDims(), NO_INIT);
  typename Image<T>::const_iterator msrc = mask.begin(), mstop = mask.end();
  typename Image<TT>::const_iterator
    lsrc = lower.begin(), hsrc = higher.begin();
  typename Image<TT>::iterator dest = result.beginw();

  while(msrc != mstop) {
    if (*msrc >= thresh) *dest = *hsrc; else *dest = *lsrc;
    msrc ++; lsrc++; hsrc++; dest ++;
  }

  return result;
}

// ######################################################################
// Work in progress, do not use yet! //
template <class T>
std::vector<std::vector<T> > matrixMult(std::vector<std::vector<T> > *A,
                                        std::vector<std::vector<T> > *B,
                                        unsigned long *beginAX,
                                        unsigned long *endAX, 
                                        unsigned long *beginBX,
                                        unsigned long *endBX,
                                        unsigned long *beginAY,
                                        unsigned long *endAY)
{
  ASSERT(*beginAX <= *endAX);
  ASSERT(*beginBX <= *endBX);  
  ASSERT(*beginAY <= *endAY); 
  typename std::vector<T> temp((signed)(*endBX - *beginBX), ZEROS);
  typename std::vector<std::vector<T> > C((signed)(*endAY - *beginAY),temp);
  typename std::vector<std::vector<T> >::iterator iA;
  typename std::vector<std::vector<T> >::iterator iB;
  typename std::vector<std::vector<T> >::iterator iC = C.begin();
  typename std::vector<T>::iterator iiB;
  typename std::vector<T>::iterator iiC;
  //Image<float> C((signed)(*endBX - *beginBX),
  //               (signed)(*endAY - *beginAY), ZEROS);
  unsigned long ya; 
  for (unsigned long yc = *beginAY; yc < *endAY; yc++, ++iC)
    {
      iB = B->begin() + *beginBX;
      iiC = iC->begin();
      for (unsigned long xc = *beginBX; xc < *endBX; ++iB, ++iiC)
        {
          ya = 0;
          iA = A->begin() + *beginAX;
          iiB = iB->begin();
          for (unsigned long xa = *beginAX; xa < *endAX; ++iA , ++iiB)
            *iiC += iA->at(yc) * (*iiB);
        }    
    }       
  return C;
}

// ######################################################################
template <class T_or_RGB>
Image<T_or_RGB> transpose(const Image<T_or_RGB>& M)
{
  const int w = M.getWidth(), h = M.getHeight();
  Image<T_or_RGB> result(h,w,NO_INIT);
  typename Image<T_or_RGB>::const_iterator mptr, mptr2 = M.begin();
  typename Image<T_or_RGB>::iterator rptr = result.beginw();

  for (int y = 0; y < w; ++y)
    {
      mptr = mptr2;
      for (int x = 0; x < h; ++x)
        {
          *rptr = *mptr;
          ++rptr; mptr += w;
        }
      ++mptr2;
    }

  return result;
}
    
// ######################################################################
template <class T>
Image<T> scramble(const Image<T>& ima)
{
  const int siz = ima.getSize(); ASSERT(siz > 0);
  Image<T> result = ima; typename Image<T>::iterator a = result.beginw();

  // this is very similar to randShuffle() in saliency.C:
  for (int i = 0; i < siz; i ++)
    {
      T tmp = a[i];
      int idx = i + (lrand48() % (siz - i));
      a[i] = a[idx]; a[idx] = tmp;
    }

  return result;
}

// Include the explicit instantiations
// MA #include "inst/Image_MathOps.I"

// additional explicit instantiations:
// MA 
/*template double sum(const Image<double>& a);
template double mean(const Image<double>& a);
template Image<double> abs(const Image<double>& a);
template Image<byte> thresholdedMix(const Image<float>& mask,
                                    const float& thresh,
                                    const Image<byte>& lower,
                                    const Image<byte>& higher);
template Image<double> toPower(const Image<double>& a, double pow);*/

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_MATHOPS_C_DEFINED
