/*!@file Image_MonoOnly.C Operations for monochrome Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_MonoOnly.C 4188 2005-02-13 07:01:15Z itti $
//

#include "saliency.H"
#include "MathFunctions.H"
#include "Image.H"
#include "Image_ShapeOps.H"
#include "log.H"

#include <algorithm> // for std::count(), etc.
#include <cmath>
#include <numeric> // for std::accumulate(), etc.
#include <list>    // used in retainMax()

// background noise level: as coeff of map full dynamic range:
#define BGNOISELEVEL 0.00001

// ######################################################################
// ##### Access functions:
// ######################################################################

// see .H file for inlined functions.

// ######################################################################
template <class T>
T Image<T>::getSample(const Point2D& center, const int radius)
{
  ASSERT(initialized());
  T val = getVal(center); // a bit wasteful but less bogus that numeric_limits
  const int r2 = radius * radius, ci = center.i, cj = center.j;
  for (int y = -radius; y <= radius; y ++)
    {
      int bound = int(sqrtf(float(r2 - y*y)));
      int sj = y + cj;
      for (int x = -bound; x <= bound; x ++)
        {
          int si = x + ci;
          if (coordsOk(si, sj))
            { T v = getVal(si, sj); if (v > val) val = v; }
        }
    }
  return val;
}

// ######################################################################
template<class T>
void Image<T>::getMinMaxAvg(T& xmini, T& xmaxi, T& xavg) const
{
  ASSERT(initialized());
  typename Image<T>::const_iterator aptr = begin(), stop = end();
  double avg = 0.0;
  xmini = *aptr; xmaxi = *aptr++;
  while (aptr != stop)
    {
      if (*aptr < xmini) xmini = *aptr;
      else if (*aptr > xmaxi) xmaxi = *aptr;
      avg += (double)(*aptr);
      aptr++;
    }
  xavg = clamped_convert<T>(avg / double(getSize()));
}

// ######################################################################
template<class T>
bool Image<T>::isFinite() const
{
  ASSERT(initialized());
  typename Image<T>::const_iterator aptr = begin(), stop = end();
  while (aptr != stop)
    {
      if (::isFinite(*aptr) == false) return false;
      aptr++;
    }
  return true;
}

// ######################################################################
template<class T>
void Image<T>::findMax(Point2D& p, T& maxval) const
{
  ASSERT(initialized());
  typename Image<T>::const_iterator aptr = begin();
  const int w = getWidth(), h = getHeight();
  p.i = 0; p.j = 0; maxval = *aptr;
  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i ++)
      {
        if (*aptr > maxval) { maxval = *aptr; p.i = i; p.j = j; }
        aptr++;
      }
}

// ######################################################################
template<class T>
void Image<T>::findMin(Point2D& p, T& minval) const
{
  ASSERT(initialized());
  typename Image<T>::const_iterator aptr = begin();
  const int w = getWidth(), h = getHeight();
  p.i = 0; p.j = 0; minval = *aptr;
  for (int j = 0; j < h; j ++)
    for (int i = 0; i < w; i ++)
      {
        if (*aptr < minval) { minval = *aptr; p.i = i; p.j = j; }
        aptr++;
      }
}

// ######################################################################
template<class T>
double Image<T>::getMean() const
{
  ASSERT(initialized());
  double sum = std::accumulate(this->begin(), this->end(), 0.0);
  return sum / double(getSize());
}

// ######################################################################
template<class T>
double Image<T>::getStdev() const
{
  ASSERT(initialized());

  typename Image<T>::const_iterator aptr = begin(), stop = end();

  double sum = 0.0;
  double mean = getMean();

  while (aptr != stop)
    {
      double val = ((double)(*aptr++)) - mean;
      sum += val * val;
    }
  return sqrt(sum / ((double)(getSize() - 1)));
}

// ######################################################################
template <class T>
void Image<T>::addDownsized(const Image<T>& b)
{
  (*this) += downSize(b, this->getDims());
}

// ######################################################################
template <class T>
void Image<T>::normalize(const T nmin, const T nmax)
{
  ASSERT(initialized()); T mi, ma; getMinMax(mi, ma);
  float scale = (float)ma - (float)mi;
  if (fabs(scale) < 1.0e-10F) scale = 1.0; // image is uniform
  float nscale = ((float)nmax - (float)nmin) / scale;

  typename Image<T>::iterator aptr = this->beginw();
  typename Image<T>::iterator stop = this->endw();

  while (aptr != stop)
    {
      *aptr = nmin + (T)( (float(*aptr) - float(mi)) * nscale );
      ++aptr;
    }
}

// ######################################################################
template <class T>
bool Image<T>::isLocalMax(const Point2D& p) const
{
  typename Image<T>::const_iterator sptr = this->begin();
  const int w = getWidth(), h = getHeight(); const int i = p.i, j = p.j;
  sptr += i + w * j; T val = *sptr;

  if (coordsOk(p) == false) return false;     // outside the image
  if (i > 0 && sptr[-1] > val) return false;  // (i-1,j) is higher
  if (i < w-1 && sptr[1] > val) return false; // (i+1,j) is higher
  if (j > 0 && sptr[-w] > val) return false;  // (i,j-1) is higher
  if (j < h-1 && sptr[w] > val) return false; // (i,j+1) is higher

  // no neighbor was higher, so we must be the highest:
  return true;
}

// ######################################################################
template <class T>
void Image<T>::activityNormalize(const T nmax)
{
  typename Image<T>::iterator sptr = this->beginw(), stop = this->endw();

  double sum = 0.0;
  while (sptr != stop)
    sum += fabs((double)(*sptr++));
  if (sum < 1.0e-10) return; // not enough energy -> forget it!
  T zero = T();

  sptr = this->beginw();
  double coeff = nmax / sum;
  while (sptr != stop)
    {
      double xx = double(*sptr) * coeff;
      if (xx > 0.0) *sptr++ = (T)xx;
      else          *sptr++ = zero;
    }
}

// ######################################################################
template <class T>
void Image<T>::varianceNormalize(const T mi, const T ma)
{
  ASSERT(initialized()); normalize(mi, ma); double sdev = getStdev();
  (*this) *= T(sdev * sdev);
}

// ######################################################################
template <class T>
void Image<T>::luminanceNormalize(const Image<T>& lum, const T thresh)
{
  ASSERT(this->isSameSize(lum));

  T zero = T();
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();
  typename Image<T>::const_iterator lptr = lum.begin();
  while (aptr != stop)
    {
      T v = *lptr++;
      if (v >= thresh) *aptr++ = (T)( (*aptr) / v * (255.0 / 3.0) );
      else *aptr++ = zero;
    }
}

// ######################################################################
template <class T>
void Image<T>::rectify()
{
  ASSERT(initialized());

  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();
  T zero = T();
  while (aptr != stop)
    {
      if (*aptr < zero) *aptr = zero;
      ++aptr;
    }
}

// ######################################################################
template <class T>
void Image<T>::splitPosNeg(Image<T>& pos, Image<T>& neg)
{
  ASSERT(initialized());
  pos.resize(getDims(), NO_INIT); neg.resize(getDims(), NO_INIT);
  typename Image<T>::const_iterator sptr = this->begin(), stop = this->end();
  typename Image<T>::iterator pptr = pos.beginw(), nptr = neg.beginw();

  T zero = T();
  while (sptr != stop)
    if (*sptr <= zero) { *nptr++ = - (*sptr++); *pptr++ = zero; }
    else { *nptr++ = zero; *pptr++ = *sptr++; }
}

// ######################################################################
template <class T>
void Image<T>::lowThresh(const T thresh, const T val)
{
  ASSERT(initialized());
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();

  while (aptr != stop)
    {
      if (*aptr < thresh) *aptr = val;
      ++aptr;
    }
}

// ######################################################################
template <class T>
void Image<T>::lowThreshAbs(const T thresh, const T val)
{
  ASSERT(initialized());
  if (thresh <= 0.0)
    {
      if (thresh < 0.0)
        LERROR("With thresh = %f < 0.0, this is a no-op -- IGNORED",
               double(thresh));
      // it's still a no-op with thresh = 0 but we won't report this
      // as 0.0 is a convenient default when no threshold should be
      // applied...
      return;
    }

  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();

  while (aptr != stop)
    {
      if (std::abs(*aptr) < thresh) *aptr = val;
      ++aptr;
    }
}

// ######################################################################
template <class T>
void Image<T>::wilsonNormalize(const float H, const float eps)
{
  ASSERT(initialized());
  // if this = S = Image*oriented_filter, F(S)=(S^2+KS^(3-eps))/(K+S^2)
  // with K = 1/H(1-eps)    CAUTION: IMAGE MUST BE >=0 EVERYWHERE
  float K = float( 1.0F / (H * (1.0F - eps)) );
  float epsd = float(3.0F - eps);

  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();

  while (aptr != stop)
    {
      float S = float(*aptr);
      float S2 = S * S;
      T val = clamped_convert<T>( (S2 + K * powf(S, epsd)) / (K + S2) );
      *aptr++ = val;
    }
}

// ######################################################################
template <class T>
void Image<T>::sigmoid(const float g, const float h, const float s)
{
  ASSERT(initialized());  // Sigmoidal normalization: f(x)=x^g/(s+x^h)
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();

  while (aptr != stop) {
    T val = clamped_convert<T>(pow(*aptr, g) / (s + pow(*aptr, h)));
    *aptr++ = val;
  }
}

// ######################################################################
template <class T>
void Image<T>::sigmoid(const float s)
{
  ASSERT(initialized());  // Sigmoidal normalization: f(x)=x^2/(s+x^2)
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();

  while (aptr != stop)
    {
      float val = float(*aptr); val *= val;
      T v = clamped_convert<T>(val / (s + val));
      *aptr ++ = v;
    }
}

// ######################################################################
template <class T>
void Image<T>::retainMax(const int num)
{
  std::list<T> vals; std::list<int> idx; int siz = getSize();
  typename Image<T>::const_iterator aptr = this->begin();
  T zero = T();

  for (int i = 0; i < siz; i ++)
    {
      T j = *aptr++;
      typename std::list<T>::iterator valptr = vals.begin();
      std::list<int>::iterator idxptr = idx.begin();
      while (true)
         {
           if (valptr == vals.end())
             { vals.push_back(j); idx.push_back(i); break; }
           if (j > *valptr)
             { vals.insert(valptr,j); idx.insert(idxptr, i); break; }
           valptr++; idxptr ++;
         }
       if (int(vals.size()) > num)
         { setVal(idx.back(), zero); vals.pop_back(); idx.pop_back(); }
    }
}

// ######################################################################
template <class T>
int Image<T>::emptyArea() const
{
  return std::count(this->begin(), this->end(), T());
}

// ######################################################################
template <class T>
int Image<T>::drawDiskCheckTarget(Image<T>& mask,
                                    const Point2D& center,
                                    const int radius,
                                    const T value,
                                    const T targetvalue,
                                    const T floodvalue)
{
  ASSERT(initialized()); ASSERT(floodvalue < targetvalue);
  int nbhit = 0; Point2D hitpt;
  for (int y = -radius; y <= radius; y ++)
    {
      int bound = (int)(sqrt( (double)( SQR(radius) - SQR(y) ) ));
      for (int x = -bound; x <= bound; x ++)
        if (coordsOk(x + center.i, y + center.j))
          {
            setVal(x + center.i, y + center.j, value);
            if (mask.getVal(x + center.i, y + center.j) == targetvalue)
              {
                nbhit ++;
                hitpt.i = x + center.i, hitpt.j = y + center.j;
                // flood the hit target:
                Image<T> tmp = mask;
                tmp.flood(mask, hitpt, targetvalue, floodvalue);
              }
          }
    }
  return nbhit;
}

#include <vector>

// ######################################################################
template <class T>
int Image<T>::flood(Image<T>& dest, const Point2D& seed,
                    const T thresh, const T val) const
{
  ASSERT(initialized()); ASSERT(coordsOk(seed));

  // clear destination image if necessary
  if (!dest.initialized()) dest.resize(getDims(), true);

  if (this->getVal(seed) < thresh) return -1; // starting point not in object

  const T zero = T();

  const int size = this->getSize();
  const int w = this->getWidth();

  // Allocate space for the recursion stack
  std::vector<int> stk;

  const int STACKSIZE_ = 256 * 256 + 20;
  stk.reserve(STACKSIZE_);

  stk.push_back(seed.i + w * seed.j);

  Image<T> input = *this;  // copy input because we will modify it
  typename Image<T>::iterator const inptr = input.beginw();

  int nbpix = 0;

  while (! stk.empty())
    {
      int index = stk.back(); stk.pop_back();
      if (inptr[index] != zero)  // we have not already visited this point
        {
          // we know that the current point IS in our object:
          dest.setVal(index, val);  // flood output
          inptr[index] = zero;      // clear this point in input
          ++nbpix;                 // one more flooded pixel

          // explore recursively the 8-connected neighbors:
          if ((index+1 < size) && (inptr[index+1] >= thresh))
            stk.push_back( index+1 );

          if ((index+w < size) && (inptr[index+w] >= thresh))
            stk.push_back( index+w );

          if ((index+w+1 < size) && (inptr[index+w+1] >= thresh))
            stk.push_back( index+w+1 );

          if ((index+w-1 < size) && (inptr[index+w-1] >= thresh))
            stk.push_back( index+w-1 );

          if ((index-1 >= 0) && (inptr[index-1] >= thresh))
            stk.push_back( index-1 );

          if ((index-w >= 0) && (inptr[index-w] >= thresh))
            stk.push_back( index-w );

          if ((index-w-1 >= 0) && (inptr[index-w-1] >= thresh))
            stk.push_back( index-w-1 );

          if ((index-w+1 >= 0) && (inptr[index-w+1] >= thresh))
            stk.push_back( index-w+1 );
        }
    }
  return nbpix;
}

// ######################################################################
template <class T>
int Image<T>::countParticles(const T thresh) const
{
  Point2D p;
  int nbpart = 0;
  Image<T> tmp = *this;
  T fill = thresh;
  const T zero = T();
  if (fill >= zero + (T)1.0) fill -= (T)1.0;  // fill with val below thresh
  else LFATAL("Cannot count particles with such low threshold");

  for (p.i = 0; p.i < getWidth(); ++p.i)
    for (p.j = 0; p.j < getHeight(); ++p.j)
      if (tmp.getVal(p) >= thresh)  // found one particle
        {
          ++nbpart;
          Image<T> tmp2 = tmp;
          tmp2.flood(tmp, p, thresh, fill);  // discard it
        }
  return nbpart;
}

// ######################################################################
template <class T>
int Image<T>::countThresh(const T thresh, const bool absol) const
{
  ASSERT(initialized());
  int nb = 0;
  typename Image<T>::const_iterator ap = this->begin(), stop = this->end();

  if (absol)
    {
      float threshd = (float)thresh;
      while (ap != stop)
        {
          if (fabs(float(*ap)) > threshd) ++nb;
          ++ap;
        }
    }
  else
    {
      while (ap != stop)
        if (*ap++ > thresh) ++nb;
    }

  return nb;
}

// ######################################################################
template <class T>
void Image<T>::addBGnoise(const float range)
{
  ASSERT(initialized());

  // do not put noise very close to image borders:
  int siz = std::min(getWidth(), getHeight()) / 10;

  typename Image<T>::iterator sptr = beginw() + siz + siz * getWidth();

  float r = range * BGNOISELEVEL;

  for (int j = siz; j < getHeight()-siz; ++j)
    {
      for (int i = siz; i < getWidth()-siz; ++i)
        *sptr++ += clamped_convert<T>(r * drand48());
      sptr += siz + siz;
    }
}

// ######################################################################
template <class T>
void Image<T>::addBGnoise2(const float range)
{
  ASSERT(initialized());

  // do not put noise very close to image borders:
  int siz = std::min(getWidth(), getHeight()) / 10;

  typename Image<T>::iterator sptr = beginw() + siz + siz * getWidth();

  float r = range;

  for (int j = siz; j < getHeight()-siz; ++j)
    {
      for (int i = siz; i < getWidth()-siz; ++i)
        *sptr++ += clamped_convert<T>(r * drand48());
      sptr += siz + siz;
    }
}

// template instantiations:

// MA template void Image<double>::rectify();

#define INSTANTIATE(T)                                               \
template T Image<T>::getSample(const Point2D&, const int);           \
template void Image<T>::getMinMaxAvg(T&, T&, T&) const;              \
template bool Image<T>::isFinite() const;                            \
template bool Image<T>::isLocalMax(const Point2D&) const;            \
template void Image<T>::findMax(Point2D&, T&) const;                 \
template void Image<T>::findMin(Point2D&, T&) const;                 \
template double Image<T>::getMean() const;                           \
template double Image<T>::getStdev() const;                          \
template void Image<T>::addDownsized(const Image<T>&);               \
template void Image<T>::normalize(T, T);                             \
template void Image<T>::activityNormalize(T);                        \
template void Image<T>::varianceNormalize(T, T);                     \
template void Image<T>::luminanceNormalize(const Image<T>&, T);      \
template void Image<T>::rectify();                                   \
template void Image<T>::splitPosNeg(Image<T>&, Image<T>&);           \
template void Image<T>::lowThresh(T, T);                             \
template void Image<T>::lowThreshAbs(T, T);                          \
template void Image<T>::wilsonNormalize(float, float);               \
template void Image<T>::sigmoid(float, float, float);                \
template void Image<T>::sigmoid(float);                              \
template void Image<T>::retainMax(const int);                        \
template int Image<T>::emptyArea() const;                            \
template int Image<T>::drawDiskCheckTarget(Image<T>&,                \
                                           const Point2D&,           \
                                           int, T, T, T);            \
template int Image<T>::flood(Image<T>&, const Point2D&, T, T) const; \
template int Image<T>::countParticles(T) const;                      \
template int Image<T>::countThresh(T, const bool) const;             \
template void Image<T>::addBGnoise(float);                           \
template void Image<T>::addBGnoise2(float);                          \

#ifdef INST_BYTE
INSTANTIATE(byte)
#endif
#ifdef INST_INT16
INSTANTIATE(int16)
#endif
#ifdef INST_INT32
INSTANTIATE(int32)
#endif
#ifdef INST_FLOAT
INSTANTIATE(float)
#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
