/*!@file Image_ShapeOps.H Shape operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_ShapeOps.H 4194 2005-02-14 21:44:08Z mundhenk $
//

#ifndef IMAGE_SHAPEOPS_H_DEFINED
#define IMAGE_SHAPEOPS_H_DEFINED

#include "saliency.H"
#include "Pixels.H"

class Rectangle;

class Dims;
class Point2D;

template <class T> class Image;

//! Compute quick and dirty scaling down of array.
template <class T_or_RGB>
Image<T_or_RGB> quickLocalAvg(const Image<T_or_RGB>& array, const int scale);

//! Compute quick and dirty scaling down of array.
template <class T_or_RGB>
Image<T_or_RGB> quickLocalMax(const Image<T_or_RGB>& array, const int scale);

//! Compute quick and dirty scaling down of array.
template <class T_or_RGB>
Image<T_or_RGB> quickLocalMin(const Image<T_or_RGB>& array, const int scale);

//! Dumb interpolation.
template <class T_or_RGB>
Image<T_or_RGB> quickInterpolate(const Image<T_or_RGB>& src,
                                 const int sfactor);

//! Scale to new width & height using bilinear interpolation.
template <class T_or_RGB>
Image<T_or_RGB> rescale(const Image<T_or_RGB>& src, const Dims& dims);

//! Scale to new width & height using bilinear interpolation.
template <class T_or_RGB>
Image<T_or_RGB> rescale(const Image<T_or_RGB>& src, int width, int height);

//! downscale an image using fancy widgets like anti-aliasing, resampling
/*! Adapted from PhotoPNMtools by Boris Van Schooten boris@13thmonkey.org
 */
template <class T>
Image<PixRGB<T> > downscaleFancy(const Image<PixRGB<T> >& src,
                                 int width, int height, int weighting_slope,
                                 bool no_weight_black);

//! Downsize using alternating lowpass/decimate operations.
template <class T>
Image<T> downSize(const Image<T>& src, const Dims& dims);

//! Downsize using alternating lowpass/decimate operations.
template <class T>
Image<T> downSize(const Image<T>& src, const int width, const int height);

//! Like downSize(), except image proportions don't have to match exactly.
/*! This operation applies alternating lowpass/decimate as long as the
    result is larger than the desired size, and then finally does a
    bilinear interpolation to the final size. If the desired size is
    actually larger than the input size, then the image is simply upscaled
    with a bilinear interpolation. The advantage of using this operation
    over just using rescale() directly is that rescale() may introduce
    aliasing if an image is being downsized by several ocatves; successive
    lowpass/decimate operations ensure that total image energy is
    preserved. */
Image<float> downSizeClean(const Image<float>& src, const Dims& new_dims);

//! Juxtapose two arrays of same height into one.
template <class T_or_RGB>
Image<T_or_RGB> concatX(const Image<T_or_RGB>& left,
                        const Image<T_or_RGB>& right);

//! Juxtapose two arrays of same width into one.
template <class T_or_RGB>
Image<T_or_RGB> concatY(const Image<T_or_RGB>& top,
                        const Image<T_or_RGB>& bottom);

//! Concatenate all image arrays, Nx arrays per line.
/*! Also reshape each array to (destX, destY) if these are != -1. */
template <class T_or_RGB>
Image<T_or_RGB> concatArray(const Image<T_or_RGB> arr[],
                            const int nbarr, const int Nx,
                            const int destX = -1, const int destY = -1);

//! Decimate in X and Y (take one every 'factor' pixels).
template <class T_or_RGB>
Image<T_or_RGB> decXY(const Image<T_or_RGB>& src, const int factor = 2);

//! Decimate in X (take one every 'factor' pixels).
template <class T_or_RGB>
Image<T_or_RGB> decX(const Image<T_or_RGB>& src, const int factor = 2);

//! Decimate in Y (take one every 'factor' pixels).
template <class T_or_RGB>
Image<T_or_RGB> decY(const Image<T_or_RGB>& src, const int factor = 2);

//! Interpolate in X and Y (zero-pad if dupli == false, or duplicate).
template <class T_or_RGB>
Image<T_or_RGB> intXY(const Image<T_or_RGB>& src, const bool dupli);

//! Interpolate in X (zero-pad if dupli == false or duplicate).
template <class T_or_RGB>
Image<T_or_RGB> intX(const Image<T_or_RGB>& src, const bool dupli);

//! Interpolate in Y (zero-pad if dupli == false or duplicate).
template <class T_or_RGB>
Image<T_or_RGB> intY(const Image<T_or_RGB>& src, const bool dupli);

//! Zoom up an image by duplicating pixels.
/*! Separate scaling factors can be given for the x and y dimensions.

    @param src Input image.
    @param xzoom Zoom factor for x dimension (default 2).
    @param yzoom Zoom factor for y dimension. If negative (the default),
    then yzoom is taken to be the same as xzoom.
*/
template <class T_or_RGB>
Image<T_or_RGB> zoomXY(const Image<T_or_RGB>& src,
                       int xzoom = 2, int yzoom = -1);

//! Extract a subimage from the input image.
/*! The upper left corner of the subimage is given by a Point2D, and the
    size of the subimage is given by the Dims. If zerofill is false, then
    the rectangle given by Point2D+Dims must be wholly contained within the
    bounds of the input image. If zerofill is true, then any output region
    that lies outside the bounds of the input will be zero-filled (where
    zero means the default value for type T). */
template <class T_or_RGB>
Image<T_or_RGB> crop(const Image<T_or_RGB>& src,
                     const Point2D& upperLeft, const Dims& size,
                     const bool zerofill = false);

//! Overload of crop() using a Rectangle instead of a Point2D+Dims.
template <class T_or_RGB>
Image<T_or_RGB> crop(const Image<T_or_RGB>& src,
                     const Rectangle& rect, const bool zerofill = false);

//! Shift an image by (int dx, int dy), with horizontal wraparound
template <class T_or_RGB>
Image<T_or_RGB> shift(const Image<T_or_RGB>& srcImg,
                      const int dx, const int dy);

//! Shift an image by (float dx, float dy)
template <class T_or_RGB>
Image<T_or_RGB> shiftImage(const Image<T_or_RGB>& srcImg,
                           const float dx, const float dy);

//! Shift an image by (int dx, int dy), without wraparound
template <class T_or_RGB>
Image<T_or_RGB> shiftClean(const Image<T_or_RGB>& srcImg,
                           const int dx, const int dy,
                           const T_or_RGB bgval = T_or_RGB());

//! Rotate an image about (x,y) by ang(in Radians), without interpolation
template <class T_or_RGB>
Image<T_or_RGB> rotate(const Image<T_or_RGB>& srcImg,
                       const int x, const int y,
                       const float ang);

//! Returns an image consisting of topImg above bottomImg
template <class T_or_RGB>
Image<T_or_RGB> combineImagesVertically(const Image<T_or_RGB>& topImg, 
                                        const Image<T_or_RGB>& bottomImg);
;
// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_ShapeOps_templates.h"

#endif // !IMAGE_SHAPEOPS_H_DEFINED
