/*!@file Image_ShapeOps.C Shape operations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_ShapeOps.C 4194 2005-02-14 21:44:08Z mundhenk $
//

#ifndef IMAGE_SHAPEOPS_C_DEFINED
#define IMAGE_SHAPEOPS_C_DEFINED

#include "Image_ShapeOps.H"

#include "Image.H"
#include "Image_FilterOps.H"
#include "log.H"
#include "safecopy.H"

#include <algorithm>
#include <cmath>
#include <limits>

#define ADD_RGB(buf,Weight,bWeight,in) do { \
	buf->r += in->r*(Weight); \
	buf->g += in->g*(Weight); \
	buf->b += in->b*(Weight); \
	*bWeight += Weight; } while (0);

// ######################################################################
template <class T>
Image<T> quickLocalAvg(const Image<T>& array, const int scale)
{
  ASSERT(array.initialized());
  int lw = array.getWidth(), lh = array.getHeight();
  int sw = std::max(1, lw / scale), sh = std::max(1, lh / scale);
  int scalex = lw / sw, scaley = lh / sh; // in case the image is very small
  int remx = lw - 1 - (lw % sw), remy = lh - 1 - (lh % sh);

  typedef typename promote_trait<T, float>::TP TF;
  Image<TF> result(sw, sh, ZEROS);

  int i, j, ci = 0, cj = 0; float fac = 1.0f / float(scale * scale);
  typename Image<T>::const_iterator lptr = array.begin();
  typename Image<TF>::iterator dptr = result.beginw();

  for (j = 0; j < lh; ++j)
    {
      for (i = 0; i < lw; ++i)
        {
          *dptr += (*lptr++);
          if ((++ci) == scalex && i != remx) { ci = 0; ++dptr; }
        }
      if (ci) { ci = 0; ++dptr; } // in case the reduction is not round
      if ( (++cj) == scaley && j != remy) cj = 0; else dptr -= sw;
    }

  Image<T> ret = result * fac;  // normalize by pixel area
  return ret;
}

// ######################################################################
template <class T>
Image<T> quickLocalMax(const Image<T>& array, const int scale)
{
  ASSERT(array.initialized());
  int lw = array.getWidth(), lh = array.getHeight();
  int sw = std::max(1, lw / scale), sh = std::max(1, lh / scale);
  int scalex = lw / sw, scaley = lh / sh;  // in case the image is very small
  int remx = lw - 1 - (lw % sw), remy = lh - 1 - (lh % sh);
  Image<T> result(sw, sh, ZEROS); // clear dest
  int i, j, ci = 0, cj = 0;
  typename Image<T>::const_iterator lptr = array.begin();
  typename Image<T>::iterator dptr = result.beginw();

  for (j = 0; j < lh; ++j)
    {
      for (i = 0; i < lw; ++i)
        {
          if (*lptr > *dptr) *dptr = *lptr;
          ++lptr;
          if ((++ci) == scalex && i != remx) { ci = 0; ++dptr; }
        }
      if (ci) { ci = 0; ++dptr; } // in case the reduction is not round
      if ((++cj) == scaley && j != remy) cj = 0; else dptr -= sw;
    }
  return result;
}

// ######################################################################
template <class T>
Image<T> quickLocalMin(const Image<T>& array, const int scale)
{
  ASSERT(array.initialized());
  int lw = array.getWidth(), lh = array.getHeight();
  int sw = std::max(1, lw / scale), sh = std::max(1, lh / scale);
  int scalex = lw / sw, scaley = lh / sh;  // in case the image is very small
  int remx = lw - 1 - (lw % sw), remy = lh - 1 - (lh % sh);
  Image<T> result(sw, sh, ZEROS); // clear dest
  int i, j, ci = 0, cj = 0;
  typename Image<T>::const_iterator lptr = array.begin();
  typename Image<T>::iterator dptr = result.beginw();

  for (j = 0; j < lh; ++j)
    {
      for (i = 0; i < lw; ++i)
        {
          if (*lptr < *dptr) *dptr = *lptr;
          ++lptr;
          if ((++ci) == scalex && i != remx) { ci = 0; ++dptr; }
        }
      if (ci) { ci = 0; ++dptr; } // in case the reduction is not round
      if ((++cj) == scaley && j != remy) cj = 0; else dptr -= sw;
    }
  return result;
}

// ######################################################################
template <class T>
Image<T> quickInterpolate(const Image<T>& src, const int sfactor)
{
  ASSERT(src.initialized()); ASSERT(sfactor >= 2);
  const int sw = src.getWidth();
  const int sh = src.getHeight();
  int lw = sw * sfactor, lh = sh * sfactor;
  Image<T> result(lw, lh, NO_INIT);
  int i, j, ci = 0, cj = 0;
  typename Image<T>::iterator lptr = result.beginw();
  typename Image<T>::const_iterator sptr = src.begin();

  for (j = 0; j < lh; ++j)
    {
      for (i = 0; i < lw; ++i)
        {
          *lptr++ = *sptr;
          if ((++ci) == sfactor) {  ci = 0; ++sptr; }
        }
      if ((++cj) == sfactor) cj = 0; else sptr -= sw;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> rescale(const Image<T>& src, const Dims& dims)
{
  return rescale(src, dims.w(), dims.h());
}

// ######################################################################
template <class T>
Image<T> rescale(const Image<T>& src, const int new_w, const int new_h)
{
  ASSERT(src.initialized()); ASSERT(new_w > 0 && new_h > 0);

  const int orig_w = src.getWidth();
  const int orig_h = src.getHeight();

  // check if same size already
  if (new_w == orig_w && new_h == orig_h) return src;

  const float sw = float(orig_w) / float(new_w);
  const float sh = float(orig_h) / float(new_h);

  Image<T> result(new_w, new_h, NO_INIT);
  typename Image<T>::iterator dptr = result.beginw();
  typename Image<T>::const_iterator const sptr = src.begin();

  // code inspired from one of the Graphics Gems book:
  /*
    (1) (x,y) are the original coords corresponding to scaled coords (i,j)
    (2) (x0,y0) are the greatest lower bound integral coords from (x,y)
    (3) (x1,y1) are the least upper bound integral coords from (x,y)
    (4) d00, d10, d01, d11 are the values of the original image at the corners
        of the rect (x0,y0),(x1,y1)
    (5) the value in the scaled image is computed from bilinear interpolation
        among d00,d10,d01,d11
  */
  for (int j = 0; j < new_h; ++j)
    {
      const float y = std::max(0.0f, (j+0.5f) * sh - 0.5f);

      const int y0 = int(y);
      const int y1 = std::min(y0 + 1, orig_h - 1);

      const float fy = y - float(y0);

      const int wy0 = orig_w * y0;
      const int wy1 = orig_w * y1;

      for (int i = 0; i < new_w; ++i)
        {
          const float x = std::max(0.0f, (i+0.5f) * sw - 0.5f);

          const int x0 = int(x);
          const int x1 = std::min(x0 + 1, orig_w - 1);

          const float fx = x - float(x0);

          typename promote_trait<T, float>::TP const
            d00 = sptr[x0 + wy0], d10 = sptr[x1 + wy0],
            d01 = sptr[x0 + wy1], d11 = sptr[x1 + wy1],
            dx0 = d00 + (d10 - d00) * fx,
            dx1 = d01 + (d11 - d01) * fx;

          *dptr++ = T( dx0 + (dx1 - dx0) * fy );  // no need to clamp
        }
    }
  return result;
}

// ######################################################################

// rescale an image using fancy widgets like anti-aliasing, resampling
template <class T>
Image<PixRGB<T> > downscaleFancy(const Image<PixRGB<T> >& src,
                                 int width, int height, int weighting_slope,
                                 bool no_weight_black)

{
  PixRGB<T> pix;
  pix.r = 0; pix.g = 0; pix.b = 0;
  Image<PixRGB<T> > buffer;
  Image<PixRGB<T> > out;
  Image<T>          bufferWeight;

  buffer.resize(width,height,true);
  out.resize(width,height,true);
  bufferWeight.resize(width,height,true);

  T scalex = (T)width  / (T)src.getWidth();
  T scaley = (T)height / (T)src.getHeight();
  T dx, dy, weight, xweight, yweight, xfrac, yfrac;
  typename Image<PixRGB<T> >::iterator bufelem;
  typename Image<T>::iterator bufw;
  dx = (T)0.0;
  dy = (T)0.0;

  for (int sy = (int)0; sy < (int)src.getHeight();
       sy++, dy += scaley)
  {
    // outer loop for Y axis
    yfrac = dy - (T)floor(dy);
    switch (weighting_slope) 
    {
    case 5: yweight = yfrac < 0.5 ? 0 : 1;
      break;
    case 4: yweight = (T)(0.5 + 0.5*tanh(15*(yfrac - 0.5)));
      break;
    case 3: yweight = (T)(0.5 + 0.5*tanh(8*(yfrac - 0.5)));
      break;
    case 2: yweight = (T)(0.5 + 0.5*tanh(5*(yfrac - 0.5)));
      break;
    case 1: yweight = (T)(0.5 - 0.5*cos(yfrac*M_PI));
      break;
    case 0: yweight = yfrac;
      break;
    default : fprintf(stderr,"illegal weighting slope\n");
      yweight = yfrac;
    } 
    // inner loop for X axis
    dx = (T)0;
    for (int sx = (int)0; sx < (int)src.getWidth();
         sx++, dx += scalex)
    {
      //LINFO("X %d",sx);
      xfrac = dx - (T)floor(dx);
      switch (weighting_slope)
      {
      case 5: xweight = xfrac < 0.5 ? 0 : 1;
        break;
      case 4: xweight = (T)(0.5 + 0.5*tanh(15*(xfrac - 0.5)));
        break;
      case 3: xweight = (T)(0.5 + 0.5*tanh(8*(xfrac - 0.5)));
        break;
      case 2: xweight = (T)(0.5 + 0.5*tanh(5*(xfrac - 0.5)));
        break;
      case 1: xweight = (T)(0.5 - 0.5*cos(xfrac*M_PI));
        /*almost same as tanh(4*x)*/
        break;
      case 0: xweight = xfrac;
        break;
      default : LINFO("illegal weighting slope");
        xweight = xfrac;
      } 
      //LINFO("XWEIGHT %f",xweight);
      int floordx = (int)floor((T)dx);
      int floordy = (int)floor((T)dy);
      const PixRGB<T> *in_sy_sx = &src.getVal(sx,sy);

      if (no_weight_black)
        if (in_sy_sx->r == 0 && in_sy_sx->g == 0 && in_sy_sx->b == 0)
          continue;

      bufelem = buffer.beginw() 
        + buffer.getWidth()*floordy + floordx;

      bufw    = bufferWeight.beginw() 
        + bufferWeight.getWidth()*floordy + floordx;

      ADD_RGB(bufelem, ((T)1.0-xweight)*((T)1.0-yweight), bufw, in_sy_sx);

      if (dx < width - 1)
      {
        bufelem++; bufw++;
        ADD_RGB(bufelem, xweight*((T)1.0-yweight), bufw, in_sy_sx);
      }

      if (dy < height - 1) 
      {
        bufelem = buffer.beginw() 
          + buffer.getWidth()*(floordy+1) + floordx;
        bufw    = bufferWeight.beginw() 
          + bufferWeight.getWidth()*(floordy+1) + floordx;

        ADD_RGB(bufelem, ((T)1.0-xweight)*yweight, bufw, in_sy_sx);
        if (dx < width - 1) {
          bufelem++;
          bufw++;
          //bufelem = &(buf[BUFIDX(floordx+1,floordy+1)]);
          ADD_RGB(bufelem, xweight*yweight, bufw, in_sy_sx);
        }
      }
    }	
    if (floorf(dy + scaley) > floorf(dy)) 
    {
      /* line finished -> write to out */
      int dsty = (int)floor(dy);
      for (int dstx = 0; dstx < width; dstx++) 
      {
        weight = bufferWeight.getVal(dstx,dsty);
        if(weight != 0.0) 
        {
          pix = buffer.getVal(dstx,dsty);
          pix = pix/weight;
          out.setVal(dstx,dsty,pix);
        }
        PixRGB<T> zero;
        zero.r = 0; zero.g = 0; zero.b = 0;
        buffer.setVal(dstx,dsty,zero);
        bufferWeight.setVal(dstx,dsty,0);
      }
    }
  }
  return out;
}
    

// ######################################################################
template <class T>
Image<T> downSize(const Image<T>& src, const Dims& dims)
{
  return downSize(src, dims.w(), dims.h());
}

// ######################################################################
template <class T>
Image<T> downSize(const Image<T>& src, const int new_w, const int new_h)
{
  if (src.getWidth() == new_w && src.getHeight() == new_h) return src;

  ASSERT(src.getWidth() / new_w > 1 && src.getHeight() / new_h > 1);

  const int wdepth = int(0.5+log(double(src.getWidth() / new_w)) / M_LN2);
  const int hdepth = int(0.5+log(double(src.getHeight() / new_h)) / M_LN2);

  if (wdepth != hdepth)
    LFATAL("arrays must have same proportions");

  Image<T> result = src;

  for (int i = 0; i < wdepth; ++i)
    {
      result = decX(lowPassX(9, result));
      result = decY(lowPassY(9, result));
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> concatX(const Image<T>& left, const Image<T>& right)
{
  ASSERT(left.getHeight() == right.getHeight());
  const int h = left.getHeight();
  const int wl = left.getWidth();
  const int wr = right.getWidth();
  Image<T> result(wl + wr, left.getHeight(), NO_INIT);
  typename Image<T>::const_iterator lptr = left.begin();
  typename Image<T>::const_iterator rptr = right.begin();
  typename Image<T>::iterator dptr = result.beginw();
  for (int j = 0; j < h; ++j)
    {
      for (int i1 = 0; i1 < wl; i1 ++) *dptr++ = *lptr++;
      for (int i2 = 0; i2 < wr; i2 ++) *dptr++ = *rptr++;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> concatY(const Image<T>& top, const Image<T>& bottom)
{
  ASSERT(top.getWidth() == bottom.getWidth());
  const int w = top.getWidth();
  const int ht = top.getHeight();
  const int hb = bottom.getHeight();
  Image<T> result(w, ht + hb, NO_INIT);
  typename Image<T>::const_iterator tptr = top.begin();
  typename Image<T>::const_iterator bptr = bottom.begin();
  typename Image<T>::iterator dptr = result.beginw();
  for (int i = 0; i < ht * w; ++i) *dptr++ = *tptr++;
  for (int i2 = 0; i2 < hb * w; i2 ++) *dptr++ = *bptr++;
  return result;
}

// ######################################################################
template <class T>
Image<T> concatArray(const Image<T> arr[], const int nbarr, const int Nx,
                     const int destX, const int destY)
{
  int width, height;
  // check if we are going to reshape:
  if (destX <= 0 || destY <= 0)
    {
      width = arr[0].getWidth(); height = arr[0].getHeight();
    }
  else { width = destX; height = destY; }
  int tw = width * Nx;
  int th = height * (int)ceil(((double)nbarr) / ((double)Nx));
  Image<T> result(tw, th, NO_INIT);
  T zeros[width]; for (int i = 0; i < width; ++i) zeros[i] = T();

  int arrnb = 0;
  for (int j = 0; j < th; j += height)
    for (int i = 0; i < tw; i += width)
      {
        if (destX <= 0 || destY <= 0)  // no reshaping
          {
            if (arrnb < nbarr)  // copy image
              {
                ASSERT(arr[arrnb].getWidth() == width &&
                       arr[arrnb].getHeight() == height);
                for (int jj = 0; jj < height; jj ++)
                  safecopy(result.beginw() + (i + (j + jj) * tw),
                           arr[arrnb].begin() + (jj * width),
                           width);
              }
            else                // put blanks at the end
              for (int jj = 0; jj < height; jj ++)
                safecopy(result.beginw() + (i + (j + jj) * tw),
                         zeros,
                         width);
          }
        else  // reshape first
          {
            if (arrnb < nbarr)  // copy image
              {
                Image<T> tmp = rescale(arr[arrnb], width, height);
                for (int jj = 0; jj < height; jj ++)
                  safecopy(result.beginw() + (i + (j + jj) * tw),
                           tmp.begin() + (jj * width),
                           width);
              }
            else                // put blanks at the end
              for (int jj = 0; jj < height; jj ++)
                safecopy(result.beginw() + (i + (j + jj) * tw),
                         zeros, width);
          }
        arrnb ++;
      }

  return result;
}

// ######################################################################
template <class T>
Image<T> decXY(const Image<T>& src, const int factor)
{
  // do not go smaller than 1x1:
  if (src.getWidth() <= 1 && src.getHeight() <= 1) return src;

  if (src.getWidth() == 1)  // only thinout vertic
    {
      int h2 = src.getHeight() / factor;

      Image<T> result(1, h2, NO_INIT);

      for (int j = 0; j < h2; ++j)
        result.setVal(j, src.getVal(j * factor));

      return result;
    }
  if (src.getHeight() == 1)
    {
      int w2 = src.getWidth() / factor;

      Image<T> result(w2, 1, NO_INIT);

      for (int i = 0; i < w2; ++i)
        result.setVal(i, src.getVal(i * factor));

      return result;
    }

  const int w2 = src.getWidth() / factor;
  const int h2 = src.getHeight() / factor;

  Image<T> result(w2, h2, NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();
  const int skip = src.getWidth() % factor + src.getWidth() * (factor - 1);

  for (int j = 0; j < h2; ++j)
    {
      for (int i = 0; i < w2; ++i)
        {
          *dptr++ = *sptr;   // copy one pixel
          sptr += factor;    // skip some pixels
        }
      sptr += skip;          // skip to start of next line
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> decX(const Image<T>& src, const int factor)
{
  ASSERT(factor > 0);

  if (src.getWidth() <= 1) return src;  // do not go smaller than 1 pixel wide

  const int h = src.getHeight();
  int w2 = src.getWidth() / factor;
  int skip = src.getWidth() % factor;
  if (w2 == 0) 
    {
      w2 = 1;
      skip = src.getWidth() - factor;
    }

  Image<T> result(w2, h, NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();

  for (int j = 0; j < h; ++j)
    {
      for (int i = 0; i < w2; ++i)
        {
          *dptr++ = *sptr;   // copy one point
          sptr += factor;    // skip a few points
        }
      sptr += skip;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> decY(const Image<T>& src, const int factor)
{
  ASSERT(factor > 0);

  if (src.getHeight() <= 1) return src;  // do not go smaller than 1 pixel high

  const int w = src.getWidth();
  int h2 = src.getHeight() / factor;
  if (h2 == 0) h2 = 1;

  Image<T> result(w, h2, NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();
  int skip = w * factor;

  for (int j = 0; j < h2; ++j)
    {
      safecopy(dptr, sptr, w);
      dptr += w; sptr += skip;
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> intXY(const Image<T>& src, const bool dupli)
{
  ASSERT(src.initialized());   // works only if there is an image

  const int w = src.getWidth();
  const int h = src.getHeight();

  const int w2 = w * 2;
  const int h2 = h * 2;

  Image<T> result(w2, h2, NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();
  if (dupli)
    for (int j = 0; j < h; ++j)
      {
        for (int i = 0; i < w; ++i)
          {
            *dptr++ = *sptr;     // copy one point
            *dptr++ = *sptr++;   // again
          }
        // duplicate line we just wrote:
        safecopy(dptr, dptr - w2, w2);
        dptr += w2;
      }
  else
    {
      T zeros[w2];
      for (int i = 0; i < w2; ++i) zeros[i] = T();  // needed??
      for (int j = 0; j < h; ++j)
        {
          for (int i = 0; i < w; ++i)
            {
              *dptr++ = *sptr++;        // copy one point
              *dptr++ = zeros[0];   // put a blank
            }
          safecopy(dptr, zeros, w2);
          dptr += w2;
        }
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> intX(const Image<T>& src, const bool dupli)
{
  ASSERT(src.initialized());

  const int w2 = src.getWidth() * 2;

  Image<T> result(w2, src.getHeight(), NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();
  T zero = T();

  const int siz = src.getSize();

  if (dupli)
    for (int i = 0; i < siz; ++i)
      {
        *dptr++ = *sptr;     // copy one point
        *dptr++ = *sptr++;   // again
      }
  else
    for (int i = 0; i < siz; ++i)
      {
        *dptr++ = *sptr++;  // copy one point
        *dptr++ = zero;     // put a zero
      }

  return result;
}

// ######################################################################
template <class T>
Image<T> intY(const Image<T>& src, const bool dupli)
{
  ASSERT(src.initialized());

  const int w = src.getWidth();
  const int h = src.getHeight();
  const int h2 = h * 2;

  Image<T> result(w, h2, NO_INIT);

  typename Image<T>::const_iterator sptr = src.begin();
  typename Image<T>::iterator dptr = result.beginw();

  if (dupli)
    for (int j = 0; j < h; ++j)
      {
        safecopy(dptr, sptr, w);
        dptr += w;
        safecopy(dptr, sptr, w);
        dptr += w;
        sptr += w;
      }
  else
    {
      T zeros[w];
      for (int i = 0; i < w; ++i) zeros[i] = T(); // needed??
      for (int j = 0; j < h; ++j)
        {
          safecopy(dptr, sptr, w);
          dptr += w;
          safecopy(dptr, zeros, w);
          dptr += w;
          sptr += w;
        }
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> zoomXY(const Image<T>& src, int xzoom, int yzoom)
{
  if (yzoom < 0) yzoom = xzoom;

  if (xzoom == 0 || yzoom == 0) return Image<T>();
  if (xzoom == 1 && yzoom == 1) return src;

  Image<T> result(src.getWidth()*xzoom, src.getHeight()*yzoom, NO_INIT);

  typename Image<T>::iterator dptr = result.beginw();

  for (int y = 0; y < src.getHeight(); ++y)
    {
      for (int yz = 0; yz < yzoom; ++yz)
        {
          typename Image<T>::const_iterator sptr =
            src.begin() + y*src.getWidth();
          for (int x = 0; x < src.getWidth(); ++x)
            {
              for (int xz = 0; xz < xzoom; ++xz)
                {
                  *dptr = *sptr;
                  ++dptr;
                }
              ++sptr;
            }
        }
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> crop(const Image<T>& src, const Point2D& pt, const Dims& dims,
              const bool zerofill)
{
  Image<T> result(dims, NO_INIT);

  typename Image<T>::iterator dptr = result.beginw();

  if (!zerofill)
    {
      ASSERT(src.coordsOk(pt));
      ASSERT(src.coordsOk(pt.i + dims.w() - 1, pt.j + dims.h() - 1));

      typename Image<T>::const_iterator sptr =
        src.begin()
        + pt.j * src.getWidth()   // Skip downward over starting rows
        + pt.i;                   // Skip rightward over starting columns

      const int src_catchup = src.getWidth() - result.getWidth();

      for (int y = 0; y < dims.h(); ++y)
        {
          for (int x = 0; x < dims.w(); ++x)
            {
              *dptr++ = *sptr++;
            }
          sptr += src_catchup;
        }
    }
  else
    {
      const T zero = T();

      const int ymax = pt.j + dims.h();
      const int xmax = pt.i + dims.w();

      for (int y = pt.j; y < ymax; ++y)
        for (int x = pt.i; x < xmax; ++x)
          {
            if (src.coordsOk(x, y)) *dptr = src.getVal(x, y);
            else                    *dptr = zero;
            ++dptr;
          }
    }

  return result;
}

// ######################################################################
template <class T>
Image<T> crop(const Image<T>& src, const Rectangle& rect, const bool zerofill)
{
  return crop(src,
              Point2D(rect.left(), rect.top()),
              Dims(rect.width(), rect.height()),
              zerofill);
}

// ######################################################################
template <class T>
Image<T> shift(const Image<T>& srcImg, const int dx, const int dy)
{
  // make sure the source image is valid
  ASSERT(srcImg.initialized());

  // create and clear the return image
  Dims dim(srcImg.getDims());
  Image<T> retImg(srcImg.getDims(), ZEROS);

  // create the source and destination iterators
  typename Image<T>::const_iterator srcBegPtr = srcImg.begin();
  typename Image<T>::const_iterator srcEndPtr = srcImg.end();
  typename Image<T>::iterator retBegPtr = retImg.beginw();
  typename Image<T>::iterator retEndPtr = retImg.endw();

  // adjust the pointers according to the direction of image shift
  int shiftAmt = dy * dim.w() + dx;
  if(shiftAmt < 0)
    srcBegPtr -= shiftAmt;
  else
    retBegPtr += shiftAmt;

  // determine width of the actual data to be transferred
  //  int newW = dim.w() - abs(dx);
  //  int quad = dx*dy;

  // transfer from source to destination image
  //  for(int i=0; (srcBegPtr != srcEndPtr) && (retBegPtr != retEndPtr);
  //      ++i, ++srcBegPtr, ++retBegPtr) {
    //    if(((quad < 0) && (i%dim.w() < abs(dx))) ||
    //       ((quad > 0) && (i%dim.w() >= newW)))
    //      *retBegPtr = T(0);
    //    else
  //      *retBegPtr = *srcEndPtr;
  //  } // for block

  // transfer from source to destination image
  while((srcBegPtr != srcEndPtr) && (retBegPtr != retEndPtr))
    *retBegPtr++ = *srcBegPtr++;

  return retImg;
}

// ######################################################################
template <class T>
Image<T> shiftImage(const Image<T>& srcImg, const float dx, const float dy)
{
  // make sure the source image is valid
  ASSERT(srcImg.initialized());

  // create and clear the return image
  Dims dim(srcImg.getDims());
  int w = dim.w(), h = dim.h();
  Image<T> retImg(dim, ZEROS);

  // prepare a couple of variable for the x direction  
  int xt = (int)floor(dx);
  float xfrac = dx - xt;
  int startx = std::max(0,xt);
  int endx = std::min(0,xt) + w;
  if (fabs(xfrac) < 1.0e-10) xfrac = 0.0;
  else endx--; 

  // prepare a couple of variable for the y direction  
  int yt = (int)floor(dy);
  float yfrac = dy - yt;
  int starty = std::max(0,yt);
  int endy = std::min(0,yt) + h;
  if (fabs(yfrac) < 1.0e-10) yfrac = 0.0;
  else endy--; 
  
  // dispatch to faster shiftClean() if displacements are roughly integer:
  if (fabs(xfrac) < 1.0e-10 && fabs(yfrac) < 1.0e-10)
    return shiftClean(srcImg, xt, yt);

  if (xfrac > 0.0)
  {
    xfrac = 1.0 - xfrac;
    xt++;
  }
  
  if (yfrac > 0.0)
  {
    yfrac = 1.0 - yfrac;
    yt++;
  }

  // prepare the coefficients
  float tl = (1.0 - xfrac) * (1.0 - yfrac);
  float tr = xfrac * (1.0 - yfrac);
  float bl = (1.0 - xfrac) * yfrac;
  float br = xfrac * yfrac;

  // prepare the pointers
  typename Image<T>::const_iterator src, src2 = srcImg.begin();
  typename Image<T>::iterator ret, ret2 = retImg.beginw();
  if (xt > 0) ret2 += xt;
  if (xt < 0) src2 -= xt;
  if (yt > 0) ret2 += yt * w;
  if (yt < 0) src2 -= yt * w;

  // now loop over the images
  for (int y = starty; y < endy; ++y)
    {
      src = src2; ret = ret2;
      for (int x = startx; x < endx; ++x)
        {
          (*ret) = (T)((*src) * tl);
          if (tr > 0.0) (*ret) += (T)((*(src + 1)) * tr);
          if (bl > 0.0) (*ret) += (T)((*(src + w)) * bl);
          if (br > 0.0) (*ret) += (T)((*(src + w + 1)) * br);
          ++src; ++ret;
        }
      src2 += w; ret2 += w;
    }
  return retImg;
}

// ######################################################################
template <class T>
Image<T> shiftClean(const Image<T>& srcImg, const int dx, const int dy,
                    const T bgval)
{
  // make sure the source image is valid
  ASSERT(srcImg.initialized());

  // create and clear the return image
  int w = srcImg.getWidth(), h = srcImg.getHeight();
  Image<T> retImg(w, h, NO_INIT); retImg.clear(bgval);

  // create the source and destination iterators
  typename Image<T>::const_iterator src = srcImg.begin();
  typename Image<T>::iterator dst = retImg.beginw();

  // find range of pixels to copy:
  int startx = std::max(0, -dx), endx = std::min(w - 1, w - 1 - dx);
  if (startx >= w || endx < 0) return retImg; // empty result
  int starty = std::max(0, -dy), endy = std::min(h - 1, h - 1 - dy);
  if (starty >= h || endy < 0) return retImg; // empty result

  int dstx = std::max(0, std::min(w - 1, dx));
  int dsty = std::max(0, std::min(h - 1, dy));

  src += startx + starty * w;
  dst += dstx + dsty * w;

  int skip = w - endx + startx - 1;

  // do the copy:
  for (int j = starty; j <= endy; j ++)
    {
      for (int i = startx; i <= endx; i ++) *dst++ = *src++;

      // ready for next row of pixels:
      src += skip; dst += skip;
    }

  return retImg;
}

// ######################################################################
template <class T>
Image<T> rotate(const Image<T>& srcImg, const int x, const int y,
                const float ang)
{
  // make sure the source image is valid
  ASSERT(srcImg.initialized());

  // create and clear the return image
  int w = srcImg.getWidth(), h = srcImg.getHeight();
  Image<T> retImg(w, h, ZEROS);

  // create temporary working image, put srcImg in the middle 
  // put some padding (values are 0) around the image 
  // so we won't need to check out of bounds when rotating
  int pad = int(ceil(sqrt(w * w + h * h))); 
  int wR  = 2 * pad + w;
  int hR  = 2 * pad + h;
  Image<T> tempImg(wR, hR, ZEROS);
  tempImg.paste(srcImg, Point2D(pad,pad));

  typename Image<T>::iterator rBPtr = retImg.beginw();
  float cAng = cos(ang), sAng = sin(ang);

  // fill in the return image with the appropriate values 
  float xR = 0.0; float yR = 0.0;
  int dx = pad + x, dy = pad + y; 
  for(int j = -y; j < h-y; j++)
    for(int i = -x; i < w-x; i++)
      {  
        xR = dx + i*cAng + j*sAng;
        yR = dy - i*sAng + j*cAng;
        *rBPtr++ = T(tempImg.getValInterp(xR,yR)); 
      }

  return retImg;
}

// ######################################################################
template <class T>
Image<T> combineImagesVertically(const Image<T>& topImg, 
                                 const Image<T>& bottomImg)
{
  int height = topImg.getHeight() + bottomImg.getHeight();
  int width;
  if(topImg.getWidth() > bottomImg.getWidth())
    width = topImg.getWidth();
  else
    width = bottomImg.getWidth();
  Image<T> result = Image<T>(width, height, ZEROS);

  for(int y=0; y < topImg.getHeight(); y++)
    {
      for(int x=0; x < topImg.getWidth(); x++)
        {
          result.setVal(x,y, topImg.getVal(x,y));
        }
    }
  for(int y=0; y < bottomImg.getHeight(); y++)
    {
      for(int x=0; x < bottomImg.getWidth(); x++)
        {
          result.setVal(x, y + topImg.getHeight(), bottomImg.getVal(x,y));
        }
    }

  return result;
}

// Include the explicit instantiations (color instantiations are now
// requested by using "T_or_RGB" for the template formal parameter name in
// the declarations in the .H file).
// MA #include "inst/Image_ShapeOps.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_SHAPEOPS_C_DEFINED
