/*!@file Image_Transforms.C Transformations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_Transforms.C 3375 2004-06-18 15:11:39Z itti $
//

#ifndef IMAGE_TRANSFORMS_C_DEFINED
#define IMAGE_TRANSFORMS_C_DEFINED

#include "Image_Transforms.H"

#include "Image.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "Image_FilterOps.H"
#include "ImageSet.H"
#include "log.H"
#include "Rectangle.H"

#include <algorithm>
#include <vector>
#include <cmath>

#define SEGOBJ_START 0.9
#define SEGOBJ_STEP  0.05
#define SEGOBJ_MINA  0.005
#define SEGOBJ_EXPL  1.05

// ######################################################################
double learningCoeff(const Image<float>& featureMap,
                     const Image<byte>& targetDMap,
                     const bool softmask,
                     const int in_thresh, const int out_thresh)
{
  ASSERT(featureMap.initialized()); ASSERT(targetDMap.initialized());
  ASSERT(featureMap.isSameSize(targetDMap));

  const bool havein = (in_thresh >= 0 && in_thresh <= 255);
  const bool haveou = (out_thresh >= 0 && out_thresh <= 255);

  if (softmask)
    {
      // remember that in the dmap zero means in the target and 255
      // outside. In the soft version, we just take the product of the
      // feature map by the mask (or 255-mask) and look for the avg
      // resulting value:
      float av = mean(featureMap) / float(featureMap.getSize());
      float oav = sum(featureMap * targetDMap) / (255.0f * sum(targetDMap));
      Image<float> mask = binaryReverse(targetDMap, byte(255));
      float iav = sum(featureMap * mask) / (255.0f * sum(mask));

      if (!havein && !haveou)
        { LFATAL("At least one threshold must be valid -- ABORT"); return 0.0;}
      else if (!havein && haveou)  // bg only: strong is bad
        return double((av - oav) / av);
      else if (havein && !haveou)  // target only: strong good
        return double((iav - av) / av);
      else // target+bg: higher inside than outside is good (>0), opposite bad
        return double((iav - oav) / av);
    }
  else
    {
      // Hard mask approach: examine local maxs in/out of targets and
      // compute learning weight convention: in the given distance
      // map, distance 0 is inside targets; is considered target
      // everything for which map value <= in_thresh; is considered
      // non-target everything for which map value >= out_thresh; the
      // transition between target and non-target is not used for
      // computation

      Image<byte>::const_iterator dmapptr = targetDMap.begin();
      Image<float>::const_iterator sptr = featureMap.begin(),
        stop = featureMap.end();
      
      float glob_min = *sptr; float glob_max = *sptr;
      float in_max = 0.0f; float out_max = 0.0f;
      
      while (sptr != stop)
        {
          const int mapval = *dmapptr++;
          const float val = *sptr++;
          if (mapval <= in_thresh && val > in_max) in_max = val;
          if (mapval >= out_thresh && val > out_max) out_max = val;
          if (val < glob_min) glob_min = val;
          if (val > glob_max) glob_max = val;
        }
      
      if (!havein && !haveou)
        { LFATAL("At least one threshold must be valid -- ABORT"); return 0.0;}
      else if (!havein && haveou)  // bg only: strong is bad
        return -double(in_max) / (double(glob_max) - double(glob_min));
      else if (havein && !haveou)  // target only: strong good
        return double(in_max) / (double(glob_max) - double(glob_min));
      else // target+bg: higher inside than outside is good (>0), opposite bad
        return (double(in_max) - double(out_max)) /
          (double(glob_max) - double(glob_min));
    }
}


// Include the explicit instantiations
// MA #include "inst/Image_Transforms.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !IMAGE_TRANSFORMS_C_DEFINED
