/*!@file Image_Transforms.H Transformations on Image
 */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: Image_Transforms.H 3372 2004-06-17 13:37:35Z itti $
//

#ifndef IMAGE_TRANSFORMS_H_DEFINED
#define IMAGE_TRANSFORMS_H_DEFINED

#include "saliency.H"
#include "Promotions.H"

class Point2D;
class Dims;
class Rectangle;
template <class T> class Image;
template <class T> class ImageSet;

//! Segment object using an adaptive thresholding & region growing method.
/*! Returns an image that masks the object (see Itti et al, J Neuroimaging,
    2001) */
template <class T>
Image<byte> segmentObject(const Image<T>& src, const Point2D& center);

//! Segment landmark using an adaptive thresholding & region growing method.
/*! Returns an image that masks the landmark
  @param activity is a measure of mean activity in the landmark
  @param standout is a measure of how much the landmark stands out from
  the rest of the image
  @param area is the percentage area of landmark
*/
template <class T>
int segmentLandmark(const Image<T>& src,
                    const Point2D& seed, Image<byte>& target,
                    double& activity, double& standout,
                    float& area_percentage);

//! Compute 2D 4-neighbor contours of binary objects in image
template <class T>
Image<byte> contour2D(const Image<T>& src);

//! Compute 2D 8-neighbor contours of binary objects in image
template <class T>
Image<byte> contour2D8(const Image<T>& src);

//! 3/4-chamfer distance transform; object = non 0
template <class T>
Image<T> chamfer34(const Image<T>& src, const T dmax = 255);

//! Computes 2D discrete cosine transform
/*! of square image patch size*size at offset (offx, offy) */
template <class T>
Image<typename promote_trait<T, float>::TP> dct(const Image<T>& src,
                                                const int offx,
                                                const int offy,
                                                const int size);

//! FFT-based information measure
template <class T>
float infoFFT(const Image<T>& src, const float eps, const Rectangle& rect);

//! FFT-based information measure
template <class T>
Image<float> infoFFT(const Image<T>& src, const float eps);

//! Examine local maxes in/out of targets and compute a learning weight
/*! Valid ranges for in_thresh and out_thresh are 0..255, plus: use -1
  for in_thresh to learn from the background only, or use -1 for
  out_thresh to learn from the target only. */
double learningCoeff(const Image<float>& featureMap,
                     const Image<byte>& targetDMap,
                     const bool softmask = false,
                     const int in_thresh = 20,
                     const int out_thresh = 255);

//! scale up an image using blocks
template <class T_or_RGB>
Image<T_or_RGB> scaleBlock(const Image<T_or_RGB>& src, const Dims newDims);

//! Floods object >= thresh with val from seed and returns area
/*! a clean version of Image::flood -
  problems solved: correct boundary checking, allow for thresh = 0.
  @param dest if dest is not initialized, it will be initialized to ZEROS.
  ATTENTION: If you want to have the old behavior of dest being initialized
  to src, you have to this explicitely before calling floodClean*/
template <class T>
int floodClean(const Image<T>& src, Image<T>& dest, const Point2D& seed,
               const T thresh, const T val);


//! same as floodClean, only finds the bounding box of the region in addition
template <class T>
int floodCleanBB(const Image<T>& src, Image<T>& dest, const Point2D& seed,
                 const T thresh, const T val, Rectangle& bbox);

//! same as segmentObject, only with floodClean and different parameters
template <class T>
Image<byte> segmentObjectClean(const Image<T>& src,
                               const Point2D& seed);

//! degrades a binary image mask s.t. it decays away from the FOA
template <class T>
Image<float> distDegrade(const Image<T>& src, const Point2D& foa,
                         const float area);

//! set all values >= thresh to val in src and return the result
template <class T>
Image<T> highThresh(const Image<T>& src, const T thresh, const T val);

//! Generate a foveation blur
/*! Mask should be 255 inside the object and 0 outside; the pyramid
  should be a Gaussian pyramid and will be used to create the
  eccentricity-dependent blur. The deeper the pyramid, the more severe
  the blur */
template <class T_or_RGB>
Image<T_or_RGB> foveate(const Image<byte>& mask,
                        const ImageSet<T_or_RGB>& pyr);

//! replace all values "from" with "to" and all other values with "other"
template <class T_or_RGB>
Image<T_or_RGB> replaceVals(const Image<T_or_RGB>& src, 
                            const T_or_RGB from, 
                            const T_or_RGB to,
                            const T_or_RGB other);

//! replace all values "from" with "to" and leave all other values in place
template <class T_or_RGB>
Image<T_or_RGB> replaceVals(const Image<T_or_RGB>& src, 
                            const T_or_RGB from, 
                            const T_or_RGB to);

//! generate a weighted blur
/*! The higher the modulator value (0..255), the more blurring at that
  location */
template <class T_or_RGB>
Image<T_or_RGB> weightedBlur(const Image<byte>& modulator,
                             const ImageSet<T_or_RGB>& pyr);

//! make a binary version of the source image
/*!@param src source image
  @param threshold the threshold used for binarization
  @param lowVal value for all points <= threshold
  @param highVal value for all points > threshold*/
template <class T>
Image<byte> makeBinary(const Image<T>& src, 
                       const T& threshold,
                       const byte lowVal = byte(0), 
                       const byte highVal = byte(255));

//! paste foreground into background image at (location)
/*!@param transparent pixels with this value are not replaced in background*/
template <class T_or_RGB>
void pasteImage(Image<T_or_RGB>& background, 
                const Image<T_or_RGB>& foreground,
                const T_or_RGB& transparent, 
                const Point2D location,
                float opacity = 1.0F);

//! return a Copy a window of image passed in
template <class T_or_RGB>
Image<T_or_RGB> copy(const Image<T_or_RGB>& src, const Rectangle& r);

// ######################################################################
// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Image_Transforms_templates.h"

#endif // !IMAGE_TRANSFORMS_H_DEFINED
