/*!@file Image.C An image template class with many image processing
  functions */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Image.C 4290 2005-03-12 02:03:27Z mundhenk $
//

#include "Image.H"
#include "MathFunctions.H"
#include "Image_ShapeOps.H" // this isn't strictly needed in Image.C,
                            // but it IS needed and Image_MonoOnly.C
                            // so we need to include it here for the
                            // dependency calculator -- this will go
                            // away when we find a better solution for
                            // the double instantiation of Image
                            // functions that we do in Image.C and
                            // Image_MonoOnly.C

#include "log.H"
#include "safecopy.H"
#include "saliency.H"
#include "Pixels.H"

#include <algorithm> // for std::fill(), etc.
#include <cmath>
#include <functional> // for std::plus(), etc.

// load up font definition for text writing
#include "font.H"

// ######################################################################
// ##### Constructors & Destructors:
// ######################################################################

// see .H file for inlined functions.

// ######################################################################
// ##### Memory management functions:
// ######################################################################

// see .H file for inlined functions.


// ######################################################################
// ##### Operator overloads and basic image manipulations:
// ######################################################################

// ######################################################################
template <class T>
void Image<T>::setValMask(const Image<byte>& mask, const T val)
{
  ASSERT(initialized()); ASSERT(isSameSize(mask));
  Image<byte>::const_iterator mptr = mask.begin();
  typename Image<T>::iterator dptr = this->beginw(), endptr = this->endw();
  while (dptr != endptr) {
    if (*mptr) *dptr = val;
    mptr ++; dptr ++;
  }
}

// ######################################################################
template <class T>
void Image<T>::replaceVal(const T& from, const T& to)
{
  ASSERT(initialized());
  typename Image<T>::iterator aptr = this->beginw(), endptr = this->endw();
  while (aptr != endptr)
    {
      if (*aptr == from) *aptr = to;
      aptr ++;
    }
}

// ######################################################################
// from GraphicsGems.h:
// linear interpolation from l (when a=0) to h (when a=1)
// (equal to (a*h)+((1-a)*l)
#define LERP(a,l,h)     ((l)+(((h)-(l))*(a)))

template <class T>
void Image<T>::addScaled(const Image<T>& b)
{
  ASSERT(initialized()); ASSERT(b.initialized());

  // check if images have same size:
  if (this->isSameSize(b)) { (*this) += b; return; }

  // ok, we need to interpolate:
  const int new_w = getWidth(), new_h = getHeight();
  const int orig_w = b.getWidth(), orig_h = b.getHeight();

  typename Image<T>::iterator aptr = this->beginw();
  typename Image<T>::const_iterator sptr = b.begin();

  const float sw = float(orig_w) / float(new_w);
  const float sh = float(orig_h) / float(new_h);

  for (int j = 0; j < new_h; ++j)
    {
      const float y = std::max(0.0f, (j+0.5f) * sh - 0.5f);

      const int y0 = int(y);
      const int y1 = std::min(y0 + 1, orig_h - 1);

      const float fy = y - float(y0);

      const int wy0 = orig_w * y0;
      const int wy1 = orig_w * y1;

      for (int i = 0; i < new_w; ++i)
        {
          const float x = std::max(0.0f, (i+0.5f) * sw - 0.5f);

          const int x0 = int(x);
          const int x1 = std::min(x0 + 1, orig_w - 1);

          const float fx = x - float(x0);

          typename promote_trait<T, float>::TP const
            d00 = sptr[x0 + wy0], d10 = sptr[x1 + wy0],
            d01 = sptr[x0 + wy1], d11 = sptr[x1 + wy1],
            dx0 = d00 + (d10 - d00) * fx,
            dx1 = d01 + (d11 - d01) * fx;

          *aptr++ = clamped_convert<T>( *aptr + dx0 + (dx1 - dx0) * fy );
        }
    }
}

// ######################################################################
template <class T>
void Image<T>::addWeighted(const Image<T>& b, const float coeff)
{
  ASSERT(b.initialized());
  if (initialized())   // this += b * coeff
    {
      ASSERT(this->isSameSize(b));
      typename Image<T>::iterator aptr = beginw();
      typename Image<T>::const_iterator bptr = b.begin();
      typename Image<T>::const_iterator stop = b.end();

      while (bptr != stop)
        *aptr++ += (T)(*bptr++ * coeff);
    }
  else   // this = b * coeff
    {
      *this = Image<T>(b.getDims(), NO_INIT);
      typename Image<T>::iterator aptr = this->beginw();
      typename Image<T>::const_iterator bptr = b.begin();
      typename Image<T>::const_iterator stop = b.end();

      while (bptr != stop)
        *aptr++ = (T)(*bptr++ * coeff);
    }
}


// ######################################################################
template <class T>
void Image<T>::square()
{
  ASSERT(initialized());
  typename Image<T>::iterator aptr = this->beginw(), stop = this->endw();
  while (aptr != stop) *aptr *= *aptr++;
}

// ######################################################################
template <class T>
void Image<T>::attenuateBorders(int size)
{
  ASSERT(initialized());

  Dims dims = getDims();

  if (size * 2 > dims.w()) size = dims.w() / 2;
  if (size * 2 > dims.h()) size = dims.h() / 2;
  if (size < 1) return;  // forget it

  float increment = 1.0 / (float)(size + 1);
  // top lines:
  float coeff = increment;
  typename Image<T>::iterator aptr = beginw();
  for (int y = 0; y < size; y ++)
    {
      for (int x = 0; x < dims.w(); x ++)
        {
          *aptr = (T)( (*aptr) * coeff );
          ++aptr;
        }
      coeff += increment;
    }
  // normal lines: start again from beginning to attenuate corners twice:
  aptr = beginw();
  for (int y = 0; y < dims.h(); y ++)
    {
      coeff = increment;
      for (int x = 0; x < size; x ++)
        {
          *(aptr + dims.w() - 1 - x * 2) =
            (T)(*(aptr + dims.w() - 1 - x * 2) * coeff);

          *aptr = (T)( (*aptr) * coeff );
          ++aptr;
          coeff += increment;
        }
      aptr += dims.w() - size;
    }
  // bottom lines
  aptr = beginw() + (dims.h() - size) * dims.w();
  coeff = increment * (float)size;
  for (int y = dims.h() - size; y < dims.h(); y ++)
    {
      for (int x = 0; x < dims.w(); ++x)
        {
          *aptr = (T)( (*aptr) * coeff );
          ++aptr;
        }
      coeff -= increment;
    }
}

// ######################################################################
template <class T>
void Image<T>::setBorders(const int size, const T value)
{
  ASSERT(initialized()); int siz = size;
  if (size * 2 > getWidth()) siz = getWidth() / 2;
  if (size * 2 > getHeight()) siz = getHeight() / 2;
  if (siz < 1) return;  // forget it

  typename Image<T>::iterator const data = beginw();

  // do the first line explicitly:
  for (int i = 0; i < getWidth(); ++i)
    data[i] = value;

  // then copy to next lines:
  for (int y = 1; y < siz; ++y)
    safecopy(data + (y * getWidth()),
             data,
             getWidth());

  // then do the vertical borders explicitly:
  for (int y = siz; y < getHeight() - siz; ++y)
    {
      typename Image<T>::iterator aptr = data + (y * getWidth());
      for (int x = 0; x < siz; ++x)
        {
          aptr[x] = value;
          aptr[getWidth() - 1 - x] = value;
        }
    }

  // finally do the bottom lines as a copy of the top lines
  safecopy(data + (getHeight() - siz) * getWidth(),
           data,
           siz * getWidth());
}

// ######################################################################
template <class T>
void Image<T>::paste(const Image<T>& img, const Point2D& pos)
{
  int w = getWidth(), h = getHeight(), iw = img.getWidth(), ih=img.getHeight();
  ASSERT(pos.i + iw <= w && pos.j + ih <= h);

  typename Image<T>::const_iterator sptr = img.begin();
  typename Image<T>::iterator dptr = beginw() + pos.i + pos.j * w;
  for (int j = 0; j < ih; j ++)
    {
      for (int i = 0; i < iw; i ++) *dptr++ = *sptr++;
      dptr += w - iw;
    }
}

// ######################################################################
template <class T>
void Image<T>::pasteGabor(const Image<T>& gabor,
                          const Point2D& pos, const T background)
{
  int w = getWidth(), h = getHeight();
  int iw = gabor.getWidth(), ih=gabor.getHeight();
  ASSERT(pos.i + iw <= w && pos.j + ih <= h);
  T dist, idist;

  typename Image<T>::const_iterator sptr = gabor.begin();
  typename Image<T>::iterator dptr = beginw() + pos.i + pos.j * w;
  for (int j = 0; j < ih; j ++)
    {
      for (int i = 0; i < iw; i ++)
        {
          dist = *dptr - background;
          if(dist < 0.0F) dist = dist * -1;
          if(*sptr < 0.0F) idist = *sptr * -1;
          else idist = *sptr;
          //take the value furthest from grey (background)
          if(dist < idist)
          {
            *dptr = *sptr+background;
          }
          *dptr++;
          *sptr++;
        }
      dptr += w - iw;
    }
}

// ######################################################################
template <class T>
void Image<T>::embed(const Image<T>& img, const Rectangle& r,
                     const T background, const bool keep_aspect)
{
  ASSERT(initialized() && img.initialized()); ASSERT(rectangleOk(r));
  int ww = r.width(), hh = r.height();
  int iw = img.getWidth(), ih = img.getHeight(); int myw = getWidth();
  float sw = float(iw) / float(ww), sh = float(ih) / float(hh);

  // take larger of scale factors if keep aspect ratio:
  if (keep_aspect)
    {
      if (sw > sh) sh = sw;
      if (sh > sw) sw = sh;
    }

  // determine bounds of actual embedded image, within rectangle:
  const int imin = (ww - int(float(iw) / sw + 0.4999F)) / 2;
  const int imax = ww - imin;
  const int jmin = (hh - int(float(ih) / sh + 0.4999F)) / 2;
  const int jmax = hh - jmin;

  typename Image<T>::iterator aptr = beginw();
  typename Image<T>::const_iterator sptr = img.begin();
  aptr += r.top() * myw + r.left();

  // top empty lines:
  for (int j = 0; j < jmin; j ++)
    {
      for (int i = 0; i < ww; i ++) *aptr++ = background;
      aptr += myw - ww;
    }

  // bulk of embedded image:
  for (int j = jmin; j < jmax; j ++)
    {
      for (int i = 0; i < imin; i ++) *aptr++ = background;

      const float y = std::max(0.0F, (float(j - jmin) + 0.5F) * sh - 0.5F);
      const int y0 = int(y);
      const float fy = y - float(y0);
      const int y1 = std::min(y0 + 1, ih - 1);
      const int wy0 = iw * y0, wy1 = iw * y1;

      for (int i = imin; i < imax; i ++)
        {
          const float x = std::max(0.0F, (float(i - imin) + 0.5F) * sw - 0.5F);
          const int x0 = int(x);
          const float fx = x - float(x0);
          const int x1 = std::min(x0 + 1, iw - 1);

          typename promote_trait<T, float>::TP
            d00 = sptr[x0 + wy0], d10 = sptr[x1 + wy0],
            d01 = sptr[x0 + wy1], d11 = sptr[x1 + wy1],
            dx0 = d00 + (d10 - d00) * fx,
            dx1 = d01 + (d11 - d01) * fx;

          *aptr++ = T( dx0 + (dx1 - dx0) * fy );  // no need to clamp
        }
      for (int i = imax; i < ww; i ++) *aptr++ = background;
      aptr += myw - ww;
    }

  // bottom empty lines:
  for (int j = jmax; j < hh; j ++)
    {
      for (int i = 0; i < ww; i ++) *aptr++ = background;
      aptr += myw - ww;
    }
}


// ######################################################################
// ##### Drawing functions:
// ######################################################################

// ######################################################################
template <class T>
void Image<T>::drawDisk(const Point2D& center,
                        const int radius, const T value)
{
  ASSERT(initialized());
  if (radius == 1)
    { if (coordsOk(center)) setVal(center.i + getWidth() * center.j, value); }
  else
    for (int y = -radius; y <= radius; y ++)
      {
        int bound = (int)(sqrt( (double)( SQR(radius) - SQR(y) ) ));
        for (int x = -bound; x <= bound; x ++)
          if (coordsOk(x + center.i, y + center.j))
            setVal(x + center.i, y + center.j, value);
      }
}

// ######################################################################
template <class T>
void Image<T>::drawLine(const Point2D& p1, const Point2D& p2,
                        const T col, const int rad)
{
  ASSERT(initialized());         // from Graphics Gems / Paul Heckbert
  int dx = p2.i - p1.i, ax = ABS(dx) << 1, sx = SGN(dx);
  int dy = p2.j - p1.j, ay = ABS(dy) << 1, sy = SGN(dy);
  int x = p1.i, y = p1.j;
  if (ax > ay)
    {
      int d = ay - (ax >> 1);
      for (;;)
        {
          Point2D p(x, y);
          drawDisk(p, rad, col);
          if (x == p2.i) return;
          if (d >= 0) { y += sy; d -= ax; }
          x += sx; d += ay;
        }
    }
  else
    {
      int d = ax - (ay >> 1);
      for (;;)
        {
          Point2D p(x, y);
          drawDisk(p, rad, col);
          if (y == p2.j) return;
          if (d >= 0) { x += sx; d -= ay; }
          y += sy; d += ax;
        }
    }
}

// ######################################################################
template <class T>
void Image<T>::drawRect(const Rectangle &r, const T color, const int rad)
{
  ASSERT(initialized()); ASSERT(rectangleOk(r));
  Point2D p1(r.left(), r.top()); Point2D p2(r.right(), r.top());
  Point2D p3(r.left(), r.bottom()); Point2D p4(r.right(), r.bottom());

  drawLine(p1, p2, color, rad); drawLine(p1, p3, color, rad);
  drawLine(p3, p4, color, rad); drawLine(p2, p4, color, rad);
}

// ######################################################################
template <class T>
void Image<T>::drawRectEZ(const Rectangle &r, const T color, const int rad)
{
  ASSERT(initialized()); 
  if(rectangleOk(r))
  {
    Point2D p1(r.left(), r.top()); Point2D p2(r.right(), r.top());
    Point2D p3(r.left(), r.bottom()); Point2D p4(r.right(), r.bottom());

    drawLine(p1, p2, color, rad); drawLine(p1, p3, color, rad);
    drawLine(p3, p4, color, rad); drawLine(p2, p4, color, rad);
  }
}

// ######################################################################
template <class T>
void Image<T>::drawCross(const Point2D& p, const T col, const int siz,
                         const int rad)
{
  ASSERT(initialized());
  Point2D p1, p2;

  p1.i = CLAMP(p.i - siz, 0, getWidth() - 1);
  p1.j = CLAMP(p.j, 0, getHeight() - 1);
  p2.i = CLAMP(p.i + siz, 0, getWidth() - 1);
  p2.j = CLAMP(p.j, 0, getHeight() - 1);

  drawLine(p1, p2, col, rad);

  p1.i = CLAMP(p.i, 0, getWidth() - 1);
  p1.j = CLAMP(p.j - siz, 0, getHeight() - 1);
  p2.i = CLAMP(p.i, 0, getWidth() - 1);
  p2.j = CLAMP(p.j + siz, 0, getHeight() - 1);

  drawLine(p1, p2, col, rad);
}

// ######################################################################
template <class T>
void Image<T>::drawPatch(const Point2D& p, const int siz, const T col)
{
  ASSERT(initialized());
  for (int i = -siz; i <= siz; ++i)
    for (int j = -siz; j <= siz; ++j)
      {
        int xx = CLAMP(p.i + i, 0, getWidth() - 1);
        int yy = CLAMP(p.j + j, 0, getHeight() - 1);
        setVal(xx, yy, col);
      }
}

// ######################################################################
template <class T>
void Image<T>::drawCircle(const Point2D& p, const int radius,
                          const T col, const int rad)
{
  ASSERT(initialized());
  if (radius == 1) { if (coordsOk(p)) setVal(p.i, p.j, col); return; }
  Point2D pp(p); pp.j -= radius; drawDisk(pp, rad, col);
  pp.j += radius + radius; drawDisk(pp, rad, col);
  int bound1 = radius, bound2 = radius;

  for (int y = 1; y <= radius; y ++)
    {
      bound2 = bound1;
      bound1 = (int)(sqrt( (double)( SQR(radius) - SQR(y) ) ));
      for (int i = bound1; i <= bound2; i ++)
        {
          pp.j = p.j - y;
          pp.i = p.i - i; drawDisk(pp, rad, col);
          pp.i = p.i + i; drawDisk(pp, rad, col);
          pp.j = p.j + y; drawDisk(pp, rad, col);
          pp.i = p.i - i; drawDisk(pp, rad, col);
        }
    }
}

// ######################################################################
template <class T>
void Image<T>::drawArrow(const Point2D& p1, const Point2D& p2,
                         const T col, const int rad)
{
  // first draw a line connecting p1 to p2:
  drawLine(p1, p2, col, rad);
  // second, draw an arrow:
  Point2D pm; pm.i = (p1.i + p2.i) / 2; pm.j = (p1.j + p2.j) / 2;
  float norm = float(std::max(getWidth(), getHeight())) / 30.0F; // arrow size
  float angle = atan2((float)(p2.j - p1.j), (float)(p2.i - p1.i));
  float arrow_angle = 20.0F * M_PI / 180.0F;
  Point2D pp;
  pp.i = pm.i - (int)(norm * cos(angle + arrow_angle));
  pp.j = pm.j - (int)(norm * sin(angle + arrow_angle));
  drawLine(pm, pp, col, rad);
  pp.i = pm.i - (int)(norm * cos(angle - arrow_angle));
  pp.j = pm.j - (int)(norm * sin(angle - arrow_angle));
  drawLine(pm, pp, col, rad);
}

// ######################################################################
template <class T>
void Image<T>::drawGrid(const int spacingX, const int spacingY,
                        const int thickX, const int thickY,
                        const T col)
{
  ASSERT(initialized());
  for (int i = thickX / 2; i < getWidth() - thickX / 2; i += spacingX)
    for (int j = 0; j < getHeight(); j ++)
      for (int t = -thickX / 2; t <= thickX / 2; t ++)
        setVal(i + t, j, col);

  for (int j = thickY / 2; j < getHeight() - thickY / 2; j += spacingY)
    for (int i = 0; i < getWidth(); i ++)
      for (int t = -thickY / 2; t <= thickY / 2; t ++)
        setVal(i, t + j, col);
}

// ######################################################################
template <class T>
void Image<T>::writeText(const Point2D& top_left, const char* text,
                         const T col, const T bgcol)
{
  Point2D p = top_left; // copy for modif
  int ww = getWidth(), hh = getHeight();
  for (int i = 0; i < (int)(strlen(text)); i ++)
    {
      int letter = int(text[i]) - 32;
      if (letter < 0 || letter > 94) letter = 0; // use space for unknown chars
      for (int y = 0; y < FONT_HEIGHT; y ++)
        for (int x = 0; x < FONT_WIDTH; x ++)
          if (p.i + x >= 0 && p.i + x < ww && p.j + y >= 0 && p.j + y < hh)
            {
              if (font[letter][y * FONT_WIDTH + x])
                setVal(p.i + x, p.j + y, bgcol);
              else
                setVal(p.i + x, p.j + y, col);
            }
      p.i += FONT_WIDTH;
    }
}

// ######################################################################
template <class T>
void Image<T>::drawPoint(int X,int Y,T pix)
{
  this->setVal(X,Y,pix);
  if(X > 0) this->setVal((X-1),Y,pix);
  if(Y > 0) this->setVal(X,(Y-1),pix);
  if(X < (this->getWidth()-1)) this->setVal((X+1),Y,pix);
  if(Y < (this->getHeight()-1)) this->setVal(X,(Y+1),pix);
}

// ######################################################################
// ##### Misc functions:
// ######################################################################

// ######################################################################
template <class T>
void Image<T>::speckleNoise(const float thresh, const int size,
                            const T noise_color, bool random_color)
{
  ASSERT(initialized()); ASSERT(size >= 1);

  if (random_color)
    {
      for (int j = 0; j <= getHeight() - size; j += size)
        for (int i = 0; i <= getWidth() - size; i += size)
          if (drand48() < thresh) {
            T rcol = clamped_convert<T>(noise_color * drand48());
            for (int k = j; k < j + size; k ++)
              for (int l = i; l < i + size; l ++)
                setVal(l, k, rcol);
          }
    }
  else
    {
      for (int j = 0; j <= getHeight() - size; j += size)
        for (int i = 0; i <= getWidth() - size; i += size)
          if (drand48() < thresh)
            for (int k = j; k < j + size; k ++)
              for (int l = i; l < i + size; l ++)
                setVal(l, k, noise_color);
    }
}


#ifdef INST_BYTE
template class Image<byte>;
template class Image< PixRGB<byte> >;
template class Image< PixH2SV<byte> >;
#endif
#ifdef INST_INT16
template class Image<int16>;
template class Image< PixRGB<int16> >;
template class Image< PixH2SV<int16> >;
#endif
#ifdef INST_INT32
template class Image<int32>;
template class Image< PixRGB<int32> >;
template class Image< PixH2SV<int32> >;
#endif
#ifdef INST_FLOAT
template class Image<float>;
template class Image< PixRGB<float> >;
template class Image< PixH2SV<float> >;
#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
