/*!@file InferoTemporal.C Object recognition module */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: InferoTemporal.C 4070 2004-12-02 02:56:16Z jbonaiuto $
//

#include "InferoTemporal.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "Raster.H"
#include "Image_ShapeOps.H"
#include "Image_ColorOps.H"
#include "ScaleSpace.H"
#include "Timer.H"
#include <cstring>
#include <iostream>
#include <sstream>
using std::vector;
using std::string;
using std::cin;
// ######################################################################
InferoTemporal::InferoTemporal(ModelManager& mgr, const std::string& descrName,
                               const std::string& tagName) :
  ModelComponent::ModelComponent(mgr, descrName, tagName),
  itsInitialScaleUp(this, "InitialScaleUp", true),
  itsUseAttention(this, "AttentionObjRecog", true),
  itsObjectDatabaseFile(this, "ObjectDatabaseFileName", ""),
  itsTrainObjectDB(this, "TrainObjectDB", true),
  itsPromptUserTrainDB(this, "PromptUserTrainDB", false),
  itsMatchObjects(this, "MatchObjects", true),
  itsRecogMinMatch(this, "RecognitionMinMatch", 3),
  itsRecogMinMatchPercent(this, "RecognitionMinMatchPercent", 0.75f),
  itsOutputFile(this, "OutputFileName", ""),
  itsSortObjects(this, "SortObjects", false),
  itsSortObjectsBy(this, "SortObjectsBy", "features"),
  itsMatchingAlg(this, "MatchingAlgorithm", "Simple"),
  itsSortKeys(this, "SortKeys", false)
{
}

// ######################################################################
void InferoTemporal::start1()
{
  const char* dbfilename = itsObjectDatabaseFile.getValString().c_str();
  //Load database
  if(strlen(dbfilename) > 0)
    objectDB = VisualObjectDB(dbfilename);
}

// ######################################################################
InferoTemporal::~InferoTemporal()
{
}

// ######################################################################
void InferoTemporal::exportOptions(const int whatToExport, const bool recurse)
{
   if (whatToExport & OPTEXP_CORE) {
     manager->requestOption(this, itsInitialScaleUp);
     manager->requestOption(this, itsUseAttention);
     manager->requestOption(this, itsObjectDatabaseFile);
     manager->requestOption(this, itsTrainObjectDB);
     manager->requestOption(this, itsPromptUserTrainDB);
     manager->requestOption(this, itsMatchObjects);
     manager->requestOption(this, itsRecogMinMatch);
     manager->requestOption(this, itsRecogMinMatchPercent);
     manager->requestOption(this, itsOutputFile);
     manager->requestOption(this, itsSortObjects);
     manager->requestOption(this, itsSortObjectsBy);
     manager->requestOption(this, itsMatchingAlg);
     manager->requestOption(this, itsSortKeys);
   }

   ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void InferoTemporal::freeMem()
{
}

// ######################################################################
void InferoTemporal::input(Image<PixRGB<byte> >& objImg, Point2D& location,
                           std::vector<double>& feature, int attnShiftNumber)
{
  Timer stopwatch;

  obj = VisualObject();
  obj.setImage(objImg);
  obj.setFeature(feature);

  if(computeVisualObject(luminance(objImg)))
    {
      if(itsMatchObjects.getVal())
        {
          LINFO("trying to recognize object");
          int minMatches = itsRecogMinMatch.getVal();
          float minMatchPer = itsRecogMinMatchPercent.getVal();
          char* matchingAlg = (char*)itsMatchingAlg.getValString().c_str();
          bool sortKeys = itsSortKeys.getVal();
          bool sortObjects = itsSortObjects.getVal();
          char* sortObjectsBy = (char*)itsSortObjectsBy.getVal().c_str();

          VisualObjectMatch match  = objectDB.matchObject(obj, location,
                                                          minMatches, 
                                                          minMatchPer,
                                                          matchingAlg, 
                                                          sortKeys, 
                                                          sortObjects,
                                                          sortObjectsBy); 
                                                              
          //If an object was identified
          if(match.getMatchCount() > 0)
            {
              std::stringstream s;
              s << attnShiftNumber;
              char* out = (char*)itsOutputFile.getValString().c_str();
              char* filename = strcat(strcat(out,"_"), (char*)s.str().c_str());
              match.writeOutput(filename);

              double secs = stopwatch.getSecs();
              LINFO("********** RECOGNIZED OBJECT *************");
              LINFO("OBJECT: %s; SCORE:%f; COMPARISONS: %i; TIME: %f",
                    match.getObjectName(), match.getScore(),  
                    match.getNumberKeyCompare(), secs);
              manager->textLog("************* RECOGNIZED OBJECT ************");
              manager->textLog("OBJECT: %s; COMPARISONS: %i; TIME: %f",
                               match.getObjectName(), 
                               match.getNumberKeyCompare(), secs);
            }
          else
            {
              LINFO("could not identify object!");
              if(itsTrainObjectDB.getVal())
                {
                  if(itsPromptUserTrainDB.getVal())
                    {
                      std::string s;
                      LINFO("add object to database? (y/n)");
                      std::getline(cin, s, '\n');
                      if((s == "Y") || (s == "y"))
                        //train database with new object
                        objectDB.train(obj);
                    }
                }
            }
        }
      else if(itsTrainObjectDB.getVal())
        {
          if(itsPromptUserTrainDB.getVal())
            {
              std::string s;
              LINFO("add object to database? (y/n)");
              std::getline(cin, s, '\n');
              if((s == "Y") || (s == "y"))
                //train database with new object
                objectDB.train(obj);
            }
        }
    }
}

// ######################################################################
bool InferoTemporal::getUseAttention() const
{
  return itsUseAttention.getVal();
}

// ######################################################################
void InferoTemporal::stop1()
{
  // write database
  objectDB.save(itsObjectDatabaseFile.getValString().c_str());
}

// ######################################################################
bool InferoTemporal::computeVisualObject(const Image<float>& bwimg)
{
  int keyPointNumber = 0;
  float currentScale = 1.0;
  Image<float> temp(bwimg);
  if(itsInitialScaleUp.getVal())
    {
      temp = rescale(temp, temp.getWidth()*2, temp.getHeight()*2);
      currentScale = 0.5;
    }
  
  // Loop over Scale Spaces
  ScaleSpace ss;

  // as long as minimum image size requirements are met:
  // get keypoints, scale down, rinse, repeat
  while ( temp.getWidth() > 24 && temp.getHeight() > 24)
    {
      // debug
      LINFO("Octave Size %i x %i", temp.getWidth(), temp.getHeight());

      // initialize scale space
      LINFO("  Computing Dogs");
      ss.initScaleSpace( currentScale, temp );

      // get keys in this scale space
      LINFO("  Retrieving keys");
      keyPointNumber += ss.findKeypoints(obj);
      		
      // resize
      temp = downSize(temp, temp.getWidth()/2, temp.getHeight()/2);
   
      currentScale *= 2.0;
    }
  if(keyPointNumber >= 3)
    return true;
  else
    return false;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
