/*!@file InferoTemporalConfigurator.C Class to select a InferoTemporal at
  runtime */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>
//

#include "InferoTemporalConfigurator.H"
#include "InferoTemporal.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"

#include <sstream>

// ######################################################################
InferoTemporalConfigurator::
InferoTemporalConfigurator(ModelManager& mgr,
                         const std::string& descrName,
                         const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsITtype(this, "InferoTemporalType", "None"),
  itsIT(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
InferoTemporalConfigurator::~InferoTemporalConfigurator()
{  }

// ######################################################################
void InferoTemporalConfigurator::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
  {
    manager->requestOption(this, itsITtype);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<InferoTemporal> InferoTemporalConfigurator::getIT() const
{ return itsIT; }

// ######################################################################
bool InferoTemporalConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsITtype.getName()) == 0) {
    // some info message:
    LINFO("Configuring IT of type %s", value.c_str());

    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsIT.get()) { 
      LINFO("Resetting InferoTemporal...");
      removeSubComponent(itsIT); itsIT.reset(NULL);
    }

    // instantiate a IT of the appropriate type:

    // #################### No inferior temporal cortex:
    if (value.compare("None") == 0)
      itsIT.reset(NULL);

    // #################### Standard inferior temporal cortex:
    else if (value.compare("Std") == 0)
      itsIT.reset(new InferoTemporal(*manager, "Inferior Temporal Cortex", 
				     "InferoTemporal"));

    // #################### Custom-defined:
    else 
      itsIT.reset(new InferoTemporal(*manager, "Inferior Temporal Cortex", 
				     "InferoTemporal"));
  }
  //itsIT.reset(new InferoTemporal(*manager, "Inferior Temporal Cortex",
	//		  	"InferoTemporal"));

  // add our babies as a subcomponents of us so that they will
  // become linked to the manager through us (hopefully we are
  // registered with the manager), which in turn will allow them to
  // export their command-line options and get configured. 
  if (itsIT.get()) {
    addSubComponent(itsIT);
    // tell the controller to export its options:
    itsIT->exportOptions(itsWhatToExport, true);
  }

  return ret;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
