/*!@file KDTree.C k-d tree implementation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>
// $Id: KDTree.C 3842 2004-10-14 21:28:52Z jbonaiuto $
//
#include "KDTree.H"
#include "log.H"
#include <math.h>

// #######################################################################
// #######################################################################
// ############# BestEntry implementation
// #######################################################################
// #######################################################################

// #######################################################################
BestEntry::BestEntry()
{
}

// #######################################################################
BestEntry::BestEntry(const Keypoint& neighbor, int distSq)
{
  this->neighbor = neighbor;
  this->distSq = distSq;
  this->dist = sqrt(distSq);
}

// #######################################################################
BestEntry::BestEntry(const Keypoint& neighbor, double dist)
{
  this->neighbor = neighbor;
  this->dist = dist;
  this->distSq = (int)(dist * dist);
}

// #######################################################################
const double BestEntry::getDistance() const
{
  return dist;
}

// #######################################################################
void BestEntry::setDistance(double value)
{
  dist = value;
  distSq = (int)(value * value);
}

// #######################################################################
const int BestEntry::getDistanceSq() const
{
  return distSq;
}

// #######################################################################
void BestEntry::setDistanceSq(int value)
{
  distSq = value;
  dist = sqrt(value);
}

// #######################################################################
const Keypoint& BestEntry::getNeighbor() const
{
  return neighbor;
}

// #######################################################################
bool BestEntry::equals(const BestEntry& be2) const
{
  Keypoint k2 = be2.getNeighbor();
  bool equal = neighbor.equals(k2);
  return equal;
}

// #######################################################################
int BestEntry::compareTo(const BestEntry& be2) const
{
  if(distSq < be2.getDistanceSq())
    return -1;
  else if(distSq > be2.getDistanceSq())
    return 1;
  else
    return 0;
}

// #######################################################################
// #######################################################################
// ############# HyperRectangle implementation
// #######################################################################
// #######################################################################

// #######################################################################
HyperRectangle::HyperRectangle()
{
}

// #######################################################################
HyperRectangle::HyperRectangle(int dim)
{
  this->dim = dim;
  leftTop = std::vector<int>(dim);
  rightBottom = std::vector<int>(dim);
}

// #######################################################################
HyperRectangle HyperRectangle::clone() const
{
  HyperRectangle rec(dim);

  for (int n = 0 ; n < dim ; ++n) {
    rec.leftTop[n] = leftTop[n];
    rec.rightBottom[n] = rightBottom[n];
  }
  
  return (rec);
}

// #######################################################################
HyperRectangle HyperRectangle::createUniverseRectangle(int dim)
{
  HyperRectangle rec(dim);

  for (int n = 0 ; n < dim ; ++n) 
    {
      rec.leftTop[n] = -1000000000;
      rec.rightBottom[n] = 1000000000;
    }
  
  return (rec);
}

// #######################################################################
HyperRectangle HyperRectangle::splitAt(int splitDim, int splitVal)
{
  HyperRectangle r2;
  //if(splitDim >= (int)leftTop.size())
  //  leftTop.resize(splitDim+1,0);
  //if(splitDim >= (int)rightBottom.size())
  //  rightBottom.resize(splitDim+1,0);
  if (leftTop[splitDim] < splitVal && rightBottom[splitDim] >= splitVal)
    {
      r2 = this->clone ();
      rightBottom[splitDim] = splitVal;
      r2.leftTop[splitDim] = splitVal;
    }
  return (r2);
}

// #######################################################################
bool HyperRectangle::isIn(const Keypoint& target) const
{
  if (target.getDimensionCount() == dim)
    {
      for (int n = 0 ; n < dim ; ++n) {
	int targD = target.getDimensionElement(n);

	if (targD < leftTop[n] || targD >= rightBottom[n])
	  return (false);
      }

      return (true);
    }
  else
    return false;
}

// #######################################################################
bool HyperRectangle::isInReach(const Keypoint& target, 
			       double distRad) const
{
  return (distance(target) < distRad);
}

// #######################################################################
double HyperRectangle::distance(const Keypoint& target) const
{
  int closestPointN;
  int distance = 0;

  // first compute the closest point within hr to the target. if
  // this point is within reach of target, then there is an
  // intersection between the hypersphere around target with radius
  // 'dist' and this hyperrectangle.
  //if(leftTop.size() < (uint)dim)
  //  leftTop.resize(dim,0);
  //if(rightBottom.size() < (uint)dim)
  //  rightBottom.resize(dim, 0);
  for (int n = 0 ; n < dim ; ++n) {
    int tI = target.getDimensionElement(n);
    int hrMin = leftTop[n];
    int hrMax = rightBottom[n];

    closestPointN = 0;
    if (tI <= hrMin) {
      closestPointN = hrMin;
    } else if (tI > hrMin && tI < hrMax) {
      closestPointN = tI;
    } else if (tI >= hrMax) {
      closestPointN = hrMax;
    }
    
    int dimDist = tI - closestPointN;
    distance += dimDist * dimDist;
  }

  return sqrt((double) distance);
}

// #######################################################################
// #######################################################################
// ############# KDTree implementation
// #######################################################################
// #######################################################################

// #######################################################################
KDTree::KDTree()
{
  initialized  = false;
}

// #######################################################################
const bool KDTree::isInitialized() const
{
  return initialized;
}

// #######################################################################
Keypoint KDTree::nearestNeighbor(const Keypoint& target, double& resDist) const
{
  HyperRectangle hr = 
    HyperRectangle::createUniverseRectangle(target.getDimensionCount());

  Keypoint nearest = nearestNeighborI(target, hr, 10000000000000.0, 
					   resDist);
  resDist = sqrt(resDist);

  return (nearest);
}

// #######################################################################
SortedLimitedList<BestEntry> KDTree::nearestNeighborList(const Keypoint& target, 
							 double& resDist, 
							 int q) const
{
  HyperRectangle hr = 
    HyperRectangle::createUniverseRectangle(target.getDimensionCount());

  SortedLimitedList<BestEntry> best(q);

  Keypoint nearest = nearestNeighborListI(best, q, target, hr,
					  10000000000000.0, resDist);
  resDist = sqrt(resDist);

  return (best);
}

// #######################################################################
SortedLimitedList<BestEntry> KDTree::nearestNeighborListBBF(const Keypoint& target, 
							    int q, 
							    int searchSteps) const
{
  HyperRectangle hr =
    HyperRectangle::createUniverseRectangle(target.getDimensionCount());

  SortedLimitedList<BestEntry> best(q);
  SortedLimitedList<HREntry> searchHr(searchSteps);

  int dummyDist;
  Keypoint nearest = nearestNeighborListBBFI<BestEntry, HREntry>(best, q, 
								 target, hr,
								 1000000000,
								 dummyDist, 
								 searchHr, 
								 searchSteps);

  return (best);
}

// ######################################################################
SharedPtr<KDTree> KDTree::createKDTree(const std::vector<Keypoint>& exset)
{
  SharedPtr<KDTree> cur = makeShared(new KDTree());
  cur->initialized = false;
  if (exset.size() > 0)
    {
      cur->initialized = true;
      cur->dr = KDTree::goodCandidate(exset, cur->splitDim);

      std::vector<Keypoint> leftElems;
      std::vector<Keypoint> rightElems;

      // split the exemplar set into left/right elements relative to the
      // splitting dimension
      double bound = cur->dr.getDimensionElement(cur->splitDim);
      for(uint i=0; i<exset.size(); i++)
	{
	  Keypoint dom = exset[i];
	  // ignore the current element
	  if (dom.equals(cur->dr))
	    continue;
	  
	  if (dom.getDimensionElement(cur->splitDim) <= bound) 
	    leftElems.push_back(dom);
	  else 
	    rightElems.push_back(dom);
		       
	}

      // recurse
      cur->left = KDTree::createKDTree(leftElems);
      cur->right = KDTree::createKDTree(rightElems);
    }
  return (cur);
}

// ######################################################################
Keypoint KDTree::nearestNeighborI(const Keypoint& target, 
				  const HyperRectangle& hr, double maxDistSq, 
				  double& resDistSq) const
{
  resDistSq = 10000000000000.0;

  Keypoint pivot = dr;

  HyperRectangle leftHr = hr;
  HyperRectangle rightHr = leftHr.splitAt(splitDim,
					  pivot.getDimensionElement(splitDim));
  
  HyperRectangle nearerHr, furtherHr;
  SharedPtr<KDTree> nearerKd = makeShared(new KDTree());
  SharedPtr<KDTree> furtherKd = makeShared(new KDTree());

  // step 5-7
  if (target.getDimensionElement(splitDim) <=
      pivot.getDimensionElement(splitDim))
    {
      nearerKd = SharedPtr<KDTree>(left);
      nearerHr = leftHr;
      furtherKd = SharedPtr<KDTree>(right);
      furtherHr = rightHr;
    } 
  else 
    {
      nearerKd = SharedPtr<KDTree>(right);
      nearerHr = rightHr;
      furtherKd = SharedPtr<KDTree>(left);
      furtherHr = leftHr;
    }

  // step 8
  Keypoint nearest;
  double distSq;
  if (!nearerKd->isInitialized()) 
    distSq = 10000000000000.0;
  else 
    nearest = nearerKd->nearestNeighborI(target, nearerHr, maxDistSq, distSq);
		
  // step 9
  if(distSq < maxDistSq)
    maxDistSq = distSq;

  // step 10
  if (furtherHr.isInReach(target, sqrt(maxDistSq))) 
    {
      double ptDistSq = pivot.distSquared(target);
      if (ptDistSq < distSq) {
	// steps 10.1.1 to 10.1.3
	nearest = pivot;
	distSq = ptDistSq;
	maxDistSq = distSq;
      }

      // step 10.2
      double tempDistSq;
      Keypoint tempNearest;
      if (!furtherKd->isInitialized()) 
	tempDistSq = 10000000000000.0;
      else 
	tempNearest = furtherKd->nearestNeighborI(target, furtherHr, maxDistSq,
						  tempDistSq);
			
      // step 10.3
      if (tempDistSq < distSq) 
	{
	  nearest = tempNearest;
	  distSq = tempDistSq;
	}
    }

  resDistSq = distSq;
  return (nearest);
}

// ######################################################################
template <class T>
Keypoint KDTree::nearestNeighborListI(SortedLimitedList<T>& best, 
				      int q, const Keypoint& target, 
				      const HyperRectangle& hr, 
				      double maxDistSq, 
				      double& resDistSq) const
{
  resDistSq = 10000000000000.0;

  Keypoint pivot = dr;

  BestEntry be(dr, target.distSquared(dr));
  best.add(be);

  HyperRectangle leftHr = hr;
  HyperRectangle rightHr = leftHr.splitAt(splitDim,
					  pivot.getDimensionElement(splitDim));

  HyperRectangle nearerHr, furtherHr;
  SharedPtr<KDTree> nearerKd = makeShared(new KDTree());
  SharedPtr<KDTree> furtherKd = makeShared(new KDTree());

  // step 5-7
  if (target.getDimensionElement(splitDim) <= 
      pivot.getDimensionElement(splitDim))
    {
      nearerKd = SharedPtr<KDTree>(left);
      nearerHr = leftHr;
      furtherKd = SharedPtr<KDTree>(right);
      furtherHr = rightHr;
    } 
  else 
    {
      nearerKd = SharedPtr<KDTree>(right);
      nearerHr = rightHr;
      furtherKd = SharedPtr<KDTree>(left);
      furtherHr = leftHr;
    }
  
  // step 8
  Keypoint nearest;
  double distSq;

  // No child, bottom reached!
  if(!nearerKd->isInitialized()) 
    distSq = 10000000000000.0;
  else 
    nearest = nearerKd->nearestNeighborListI(best, q, target, nearerHr,
					     maxDistSq, distSq);
		
  // step 9
  if (best.getSize() >= q)
    maxDistSq = best.getValue(q - 1).getDistance();
  else
    maxDistSq = 10000000000000.0;

  // step 10
  if (furtherHr.isInReach(target, sqrt(maxDistSq))) 
    {
      double ptDistSq = pivot.distSquared(target);
      if (ptDistSq < distSq) {
	// steps 10.1.1 to 10.1.3
	nearest = pivot;
	distSq = ptDistSq;
	
	maxDistSq = distSq;
      }

      // step 10.2
      double tempDistSq;
      Keypoint tempNearest;
      if (!furtherKd->isInitialized()) 
	tempDistSq = 10000000000000.0;
      else 
	tempNearest = furtherKd->nearestNeighborListI(best, q, target,
						      furtherHr, maxDistSq, 
						      tempDistSq);
    
      // step 10.3
      if (tempDistSq < distSq) 
	{
	  nearest = tempNearest;
	  distSq = tempDistSq;
	}
    }

  resDistSq = distSq;
  return (nearest);
}

// ######################################################################
template<class T1,class T2>
Keypoint KDTree::nearestNeighborListBBFI(SortedLimitedList<T1>& best, 
					 int q, const Keypoint& target, 
					 const HyperRectangle& hr, 
					 int maxDistSq, int& resDistSq, 
					 SortedLimitedList<T2>& searchHr,
					 int& searchSteps) const
{
  resDistSq = 1000000000;

  Keypoint pivot = dr;
  BestEntry be(dr, target.distSquared(dr));
  best.add(be);

  HyperRectangle leftHr = hr;
  HyperRectangle rightHr = leftHr.splitAt(splitDim,
					  pivot.getDimensionElement(splitDim));
  HyperRectangle nearerHr, furtherHr;
  SharedPtr<KDTree> nearerKd = makeShared(new KDTree());
  SharedPtr<KDTree> furtherKd = makeShared(new KDTree());

  // step 5-7
  if (target.getDimensionElement(splitDim) <=
      pivot.getDimensionElement(splitDim))
    {
      nearerKd = SharedPtr<KDTree>(left);
      nearerHr = leftHr;
      furtherKd = SharedPtr<KDTree>(right);
      furtherHr = rightHr;
    } 
  else 
    {
      nearerKd = SharedPtr<KDTree>(right);
      nearerHr = rightHr;
      furtherKd = SharedPtr<KDTree>(left);
      furtherHr = leftHr;
    }

  // step 8
  Keypoint nearest;
  int distSq;

  HREntry hre(furtherHr, furtherKd, pivot, 
	      furtherHr.distance(target));
  searchHr.add(hre);
  
  // No child, bottom reached!
  if (nearerKd->isInitialized())
    nearest = nearerKd->nearestNeighborListBBFI(best, q, target, nearerHr,
						maxDistSq, distSq, searchHr, 
						searchSteps);
  else
    distSq = 1000000000;

  // step 9
  if (best.getSize() >= q) 
    maxDistSq = best.getValue(q - 1).getDistanceSq();
  else
    maxDistSq = 1000000000;

  if (searchHr.getSize() > 0) 
    {
      HREntry hre = (HREntry)searchHr.getValue(0);
      searchHr.remove(0);

      furtherHr = hre.getHR();
      furtherKd = hre.getTree();
      pivot = hre.getPivot();
    }

  // step 10
  searchSteps -= 1;
  if (searchSteps > 0 &&
      furtherHr.isInReach (target, sqrt(maxDistSq)))
    {
      int ptDistSq = pivot.distSquared(target);
      if (ptDistSq < distSq) 
	{
	  // steps 10.1.1 to 10.1.3
	  nearest = pivot;
	  distSq = ptDistSq;
	  
	  maxDistSq = distSq;
	}

      // step 10.2
      int tempDistSq;
      Keypoint tempNearest;
      if (!furtherKd->isInitialized()) 
	tempDistSq = 1000000000;
      else 
	tempNearest = furtherKd->nearestNeighborListBBFI(best, q, target, 
							 furtherHr, maxDistSq, 
							 tempDistSq, searchHr,
							 searchSteps);
      // step 10.3
      if (tempDistSq < distSq) 
	{
	  nearest = tempNearest;
	  distSq = tempDistSq;
	}
    }
  
  resDistSq = distSq;
  return (nearest);
}

// ######################################################################
bool KDTree::operator ==(const KDTree& tree) const
{
  Keypoint k2 = tree.dr;
  if(!this->dr.equals(k2))
    return false;
  if(!(*(this->left) == *(tree.left)))
    return false;
  if(!(*(this->right) == *(tree.right)))
    return false;
  return true;
}

// ######################################################################
Keypoint KDTree::goodCandidate(std::vector<Keypoint> exset,
			       int& splitDim)
{
  Keypoint exemplar;
  if(exset.size() > 0)
    {
      Keypoint first = exset[0];
      int dim = first.getDimensionCount();

      // initialize temporary hr search min/max values
      std::vector<double> minHr(dim);
      std::vector<double> maxHr(dim);
      for (int k = 0 ; k < dim ; k++) 
	{
	  minHr[k] =  10000000000000.0;
	  maxHr[k] = -10000000000000.0;
	}

      for(uint i=0; i<exset.size(); i++)
	{
	  Keypoint dom = exset[i];
	  for (int k = 0; k < dim; k++) 
	    {
	      double dimE = dom.getDimensionElement(k);
	  
	      if (dimE < minHr[k])
		minHr[k] = dimE;
	      if (dimE > maxHr[k])
		maxHr[k] = dimE;
	    }
	}
      
      // find the maximum range dimension
      std::vector<double> diffHr(dim);
      int maxDiffDim = 0;
      double maxDiff = 0.0;
      for (int k = 0; k < dim; k++) 
	{
	  diffHr[k] = maxHr[k] - minHr[k];
	  if (diffHr[k] > maxDiff) 
	    {
	      maxDiff = diffHr[k];
	      maxDiffDim = k;
	    }
	}
      
      // the splitting dimension is maxDiffDim
      // now find a exemplar as close to the arithmetic middle as possible
      double middle = (maxDiff / 2.0) + minHr[maxDiffDim];
      double exemMinDiff = 10000000000000.0;
      
      for(uint i=0; i<exset.size(); i++)
	{
	  Keypoint dom = exset[i];
	  double curDiff = fabs(dom.getDimensionElement(maxDiffDim) - 
			       middle);
	  if (curDiff < exemMinDiff) 
	    {
	      exemMinDiff = curDiff;
	      exemplar = dom;
	    }
	}
      
      // return the values
      splitDim = maxDiffDim;
    }
  return (exemplar);
}

// #######################################################################
// #######################################################################
// ############# HREntry implementation
// #######################################################################
// #######################################################################

// #######################################################################
HREntry::HREntry()
{
}

// #######################################################################
HREntry::HREntry(const HyperRectangle& rect, const SharedPtr<KDTree>& tree, 
		 const Keypoint& pivot, double dist)
{
  this->rect = rect;
  this->tree = tree;
  this->pivot = pivot;
  this->dist = dist;
}

// #######################################################################
const double HREntry::getDistance() const
{
  return dist;
}

// #######################################################################
const HyperRectangle& HREntry::getHR() const
{
  return rect;
}

// #######################################################################
const Keypoint& HREntry::getPivot() const
{
  return pivot;
}

// #######################################################################
const SharedPtr<KDTree>& HREntry::getTree() const
{
  return tree;
}

// #######################################################################
bool HREntry::equals(const HREntry& obj) const
{
  if(dist != obj.getDistance())
    return false;
  Keypoint p2 = obj.getPivot();
  if(!pivot.equals(p2))
    return false;
  return true;
}

// #######################################################################
int HREntry::compareTo(const HREntry& obj) const
{
  if(dist < obj.getDistance())
    return -1;
  else if(dist > obj.getDistance())
    return 1;
  else
    return 0;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
