/*!@file KDTree.H k-d tree implementation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>
// $Id: KDTree.H 3994 2004-11-05 04:26:16Z jbonaiuto $
//
#ifndef KDTREE_H_DEFINED
#define KDTREE_H_DEFINED

#include "Keypoint.H"
#include "SharedPtr.H"
#include <vector>

//! The nearest entry in the tree to the target entry
class BestEntry
{
public:
  //! Constructor
  BestEntry();

  //! Constructor
  BestEntry(const Keypoint& neighbor, int distSq);
  
  //! Constructor
  BestEntry(const Keypoint& neighbor, double dist);
  
  //! Get the distance between the neighbor point and the target point
  const double getDistance() const;

  //! Set the distance between the neighbor point and the target point
  void setDistance(double value);

  //! Get the distance squared between the neighbor point and the target point
  const int getDistanceSq() const;

  //! Set the distance squared between the neighbor point and the target point
  void setDistanceSq(int value);

  //! Get the neighbor point
  const Keypoint& getNeighbor() const;

  //! Returns whether or not this entry is equal to the specified entry
  bool equals(const BestEntry& be2) const;

  //! Compares this entry to the specified entry
  int compareTo(const BestEntry& be2) const;

private:
  // Distance between this neigbor point and the target point
  double dist;
  int distSq;
  // The neighbor
  Keypoint neighbor;
};

//! Maintains a sorted list with a limited number of elements
template<class T>
class SortedLimitedList
{
public:
  //! Constructor
  inline SortedLimitedList();

  //! Constructor
  inline SortedLimitedList(int maxElements);

  //! Add an entry to the list
  inline int add(const T& entry, int order=0);

  //! Add an entry to the list without yet sorting
  inline void addUnsorted(const T& entry);

  //! Get an entry from the list
  inline const T& getValue(int index) const;
  
  //! Get the size of the list
  inline int getSize() const;
  
  //! Get the index of the entry
  inline int indexOf(const T& entry) const;

  //! Sort the list 0=ascending order, 1=descending order
  inline void sort(int n, int start=0, const int order=0);
  
  //! Remove the object at the specified index from the list
  inline void remove(int index);
  
  //! Change the max number of elements
  inline void setMaxElements(int value);

  //! Convert to vector
  inline const std::vector<T>& toVector() const;

private:
  int max;
  std::vector<T> list;
  void partition(int n, int& pivot_index, int start, int order);
};

// #######################################################################
// #######################################################################
// ############# SortedLimitedList implementation
// #######################################################################
// #######################################################################

// #######################################################################
template <class T> inline
SortedLimitedList<T>::SortedLimitedList()
{
  max = 0;
}

// #######################################################################
template <class T> inline
SortedLimitedList<T>::SortedLimitedList(int maxElements)
{
  max = maxElements;
}

// #######################################################################
template <class T> inline
int SortedLimitedList<T>::add(const T& entry, int order)
{
  list.push_back(entry);
  sort(list.size(), order);
  if(max > 0 && list.size() > (uint)max)
    list.pop_back();
  return indexOf(entry);
}

// #######################################################################
template <class T> inline
void SortedLimitedList<T>::addUnsorted(const T& entry)
{
  list.push_back(entry);
}

// #######################################################################
template <class T> inline
int SortedLimitedList<T>::indexOf(const T& entry) const
{
  for(uint i=0; i<list.size(); i++)
    {
      bool equal = list[i].equals(entry);
      if(equal)
	return i;
    }
  return -1;
}

// #######################################################################
template <class T> inline
const T& SortedLimitedList<T>::getValue(int index) const
{
  return list[index];
}

// #######################################################################
template<class T> inline
int SortedLimitedList<T>::getSize() const
{
  return (int)list.size();
}

// #######################################################################
template<class T> inline
void SortedLimitedList<T>::setMaxElements(int value)
{
  max = value;
  while(list.size() > (uint)max)
    list.pop_back();
}

// #######################################################################
template<class T> inline
void SortedLimitedList<T>::sort(int n, int start, const int order)
{
  int pivot_index;
  int n1;
  int n2;

  if(n > 1)
    {
      partition(n, pivot_index,start,order);
      n1 = pivot_index;
      n2 = n - n1 - 1;
      sort(n1);
      sort(n2, pivot_index+1);
    }
}

// ######################################################################
template<class T> inline
void SortedLimitedList<T>::partition(int n, int& pivot_index, int start,
                                     int order)
{
  T pivot = list[start];
  int l_index = start+1;
  int r_index = n-1;

  while(l_index < r_index)
    {
      while(l_index < n && 
	    ((order == 0 && list[l_index].compareTo(pivot) <= 0) ||
             (order == 1 && list[l_index].compareTo(pivot) > 0)))
	l_index++;
      while((order == 0 && list[r_index].compareTo(pivot) > 0) ||
            (order == 1 && list[r_index].compareTo(pivot) <= 0))
	r_index--;
      if(l_index < r_index)
	{
	  T l = T(list[l_index]);
	  list[l_index] = T(list[r_index]);
	  list[r_index] = l;
	}
    }

  pivot_index = r_index;
  list[start] = T(list[pivot_index]);
  list[pivot_index] = T(pivot);
}

// #######################################################################
template<class T> inline
void SortedLimitedList<T>::remove(int index)
{
  list.erase(list.begin()+index);
}

// #######################################################################
template<class T> inline
const std::vector<T>& SortedLimitedList<T>::toVector() const
{
  return list;
}

// #######################################################################

//! Represents a rectangle in n-dimensional space
class HyperRectangle
{
public:
  //! Constructor
  HyperRectangle();

  //! Constructor
  HyperRectangle(int dim);
  
  //! Clones this HyperRectangle
  HyperRectangle clone() const;

  //! Creates an all-encompassing rectangle in the dimension specified
  static HyperRectangle createUniverseRectangle(int dim);

  //! Splits the rectangle at the specified dimension and value
  HyperRectangle splitAt(int splitDim, int splitVal);
  
  //! Returns true if the target point is within this HR
  bool isIn(const Keypoint& target) const;
  
  //! Return true if any part of this HR is reachable from target
  /*! Must be reachable by no more than 'distRad', false otherwise.
   */
  bool isInReach(const Keypoint& target, double distRad) const;

  //! Return the dist from nearest point within the HR to the target point.
  double distance(const Keypoint& target) const;

private:
  std::vector<int> leftTop;
  std::vector<int> rightBottom;
  int dim;
};


//! A simple K-d tree implementation
class KDTree
{
public:
  //! Constructor
  KDTree();

  //! Whether or not the tree has any values
  const bool isInitialized() const;

  //! Find the nearest neighbour to hyperspace point 'target' in kd-tree. 
  /*! After return 'resDist' contains the absolute distance from the
      target point. The nearest neighbour is returned.
  */
  Keypoint nearestNeighbor(const Keypoint& target, double& resDist) const;
  
  //! Find a list of the nearest-neighbors to the target in kd-tree of size q
  SortedLimitedList<BestEntry> nearestNeighborList(const Keypoint& target, 
						   double& resDist,
						   int q) const;

  //! Limited Best-Bin-First k-d-tree nearest neighbour search.
  /*! Find the approximate nearest neighbour to the hyperspace point 'target'
      within the kd-tree using 'searchSteps' tail recursions at most (each
      recursion deciding about one neighbours' fitness).
      After return 'resDist' contains the absolute distance of the
      approximate nearest neighbour from the target point. The approximate
      nearest neighbour is returned.
  */
  SortedLimitedList<BestEntry> nearestNeighborListBBF(const Keypoint& target, 
						      int q, 
						      int searchSteps) const;

  //! Build a kd-tree from a list of elements. 
  /*! All elements must have the same dimensionality.
  */
  static SharedPtr<KDTree> createKDTree(const std::vector<Keypoint>& exset);

  //! Find wether or not this tree is equal to the specified tree
  bool operator ==(const KDTree& tree) const;

private:
  bool initialized;
  //The current element
  Keypoint dr;
  // The splitting dimension for subtrees.
  int splitDim;

  // The left and the right kd-subtree.
  SharedPtr<KDTree> left;
  SharedPtr<KDTree> right;

  // Internal recursive algorithm for the kd-tree nearest neighbour search.
  Keypoint nearestNeighborI(const Keypoint& target, const HyperRectangle& hr,
			    double maxDistSq, double& resDistSq) const;
  
  // Internal recursive algorithm for the kd-tree nearest neighbor list search
  template<class T>
  Keypoint nearestNeighborListI(SortedLimitedList<T>& best, int q,
				const Keypoint& target, 
				const HyperRectangle& hr, double maxDistSq, 
				double& resDistSq) const;

  // Internal recursive algorithm for the limited best-bin-first kd-tree
  // nearest neighbor search
  template<class T1,class T2>
  Keypoint nearestNeighborListBBFI(SortedLimitedList<T1>& best, 
				   int q, const Keypoint& target, 
				   const HyperRectangle& hr, int maxDistSq,
				   int& resDistSq, 
				   SortedLimitedList<T2>& searchHr,
				   int& searchSteps) const;

  static Keypoint goodCandidate(std::vector<Keypoint> exset, int& splitDim);
  
};

//! Represents a HyperRectangle entry in a list
class HREntry
{
public:
  //! Constructor
  HREntry();

  //! Constructor
  HREntry(const HyperRectangle& rect, const SharedPtr<KDTree>& tree, 
	  const Keypoint& pivot, double dist);
  
  //! Get distance
  const double getDistance() const;

  //! Get the HyperRectangle
  const HyperRectangle& getHR() const;

  //! Get the pivot
  const Keypoint& getPivot() const;

  //! Get the tree
  const SharedPtr<KDTree>& getTree() const;

  //! Find if this HREntry equals the specified HREntry
  bool equals(const HREntry& obj) const;

  //! Compare this HREntry to the specified HREntry
  int compareTo(const HREntry& obj) const;

private:
  double dist;
  HyperRectangle rect;
  Keypoint pivot;
  SharedPtr<KDTree> tree;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
