/*!@file Keypoint.C Keypoint for SIFT obj recognition */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>

#include "Keypoint.H"
#include "KDTree.H"
#include "log.H"
#include <math.h>

// #######################################################################
// #######################################################################
// ########## Keypoint implementation
// #######################################################################
// #######################################################################

// #######################################################################
Keypoint::Keypoint(int fvLength)
{
  FVBytes = std::vector<unsigned char>(fvLength);
}

// #######################################################################
Keypoint::Keypoint(int fvLength, float x, float y, float s, float o, 
                   double mag)
{
  FVBytes = std::vector<unsigned char>(fvLength);
  col_x = x;
  row_y = y;
  scale = s;
  orientation = o;
  magnitude = mag;
}

// #######################################################################
Keypoint::~Keypoint()
{}

// #######################################################################
float Keypoint::getX() const
{
  return col_x;
}

// #######################################################################
void Keypoint::setX(float x)
{
  col_x = x;
}

// #######################################################################
float Keypoint::getY() const
{
  return row_y;
}

// #######################################################################
void Keypoint::setY(float y)
{
  row_y = y;
}

// #######################################################################
float Keypoint::getScale() const
{
  return scale;
}

// #######################################################################
void Keypoint::setScale(float scale)
{
  this->scale = scale;
}

// #######################################################################
float Keypoint::getOrientation() const
{
  return orientation;
}

// #######################################################################
void Keypoint::setOrientation(float orientation)
{
  this->orientation = orientation;
}

// ########################################################################
double Keypoint::getMagnitude() const
{
  return magnitude;
}

// ########################################################################
bool Keypoint::simpleMatch(std::vector<Keypoint>& klist, 
                           Keypoint& match, int& numberComparisons) const
{
  bool matched = true;
  int matchIndex = -1;
  int distsq1 = 100000000, distsq2 = 100000000;
  int distSq;

  /* Find the two closest matches, and put their squared distances in
     distsq1 and distsq2.
  */
  for(uint i=0; i<klist.size(); i++)
    {
      Keypoint k = klist[i];
      distSq = distSquared(k);
      numberComparisons++;

      if (distSq < distsq1) 
	{
	  distsq2 = distsq1;
	  distsq1 = distSq;
	  match = klist[i];
          matchIndex = i;
	} 
      else if (distSq < distsq2) 
	{
	  distsq2 = distSq;
	}
    }
  
  /* Check whether closest distance is less than 0.6 of second. */
  if (10 * 10 * distsq1 >= 6 * 6 * distsq2)
    matched = false;
  if(matched && matchIndex >= 0)
    klist.erase(klist.begin()+matchIndex);
  return matched;
}

// ########################################################################
bool Keypoint::kdTreeMatch(SharedPtr<KDTree>& tree, Keypoint& match, 
                           bool useBBF, int& numberComparisons) const
{
  bool matched = true;
  double dist;
  BestEntry matchEntry;
  BestEntry secondMatchEntry;
  if(!useBBF)
    {
      SortedLimitedList<BestEntry> closest = tree->nearestNeighborList(*this,
                                                                       dist,2);
      if(closest.getSize() == 2)
        {
          matchEntry = (BestEntry)closest.getValue(0);
          secondMatchEntry = (BestEntry)closest.getValue(1);
        }
      else
        return false;
    }
  else
    {
      SortedLimitedList<BestEntry> closest=tree->nearestNeighborListBBF(*this, 
                                                                        2, 40);
      if(closest.getSize() == 2)
        {
          matchEntry = (BestEntry)closest.getValue(0);
          secondMatchEntry = (BestEntry)closest.getValue(1);
        }
      else
        return false;
    }
  int distSq1 = matchEntry.getDistanceSq();
  int distSq2 = secondMatchEntry.getDistanceSq();
  if(10 * 10 * distSq1 >= 6 * 6 * distSq2)
    matched = false;
  match = matchEntry.getNeighbor();

  return matched;
}

// ########################################################################
const int Keypoint::getDimensionElement(int index) const
{
  if((uint)index < FVBytes.size())
    return (int)FVBytes[index];
  else
    return -1;
}

// ########################################################################
int Keypoint::getDimensionCount() const
{
  return (int)FVBytes.size();
}

// ########################################################################
int Keypoint::distSquared(const Keypoint& k) const
{
  int dif, distsq = 0;
  
  for(uint i=0; i<FVBytes.size(); i++)
    {
      dif = (int)FVBytes[i] - k.getDimensionElement(i);
      distsq += dif * dif;
    }
  return distsq;
}

// #######################################################################
void Keypoint::save(FILE *fp) const
{
  fprintf(fp, "%f %f %f %f \n", row_y, col_x, scale,
	  orientation);
  for (uint i=0; i<FVBytes.size(); i++)
    fprintf(fp, "%3d ", FVBytes[i]);
  fprintf(fp, "\r\n");
}

// #######################################################################
int Keypoint::maxDistSquared() const
{
  return (255*255)*FVBytes.size();
}

// #######################################################################
void Keypoint::load(FILE *fp)
{
  int val;
  if (fscanf(fp, "%f %f %f %f", &(row_y), &(col_x), &(scale),
	     &(orientation)) != 4)
    LINFO("Invalid keypoint file format.");

  for (uint j = 0; j < FVBytes.size(); j++)
    {
      if (fscanf(fp, "%d", &val) != 1 || val < 0 || val > 255)
	LINFO("Invalid keypoint file value.");
      FVBytes[j] = (unsigned char) val;
    }
}

// ######################################################################
void Keypoint::computeFeature(float x, float y, float s,
                              Image<float>& orie, Image<float>& grad,
                              int dimX, int dimY)
{
  FeatureVector fv(FVBytes.size());
  
  int xi = (int) (x + 0.5f);
  int yi = (int) (y + 0.5f);
  float sinAngle = sin( orientation );
  float cosAngle = cos( orientation );
  
  // check this scale
  float scaleCompute = 3.0 * s;
  int radius = 8;
  
  for (int rx=-radius; rx<radius; rx++ )
    {
      float xf = (rx+radius)/4.0f;
      
      for (int ry=-radius; ry<radius; ry++ )
        {
	  float yf = (ry+radius)/4.0f;
	  
	  // ouside circle
	  if ( pow(rx,2.0) + pow(ry,2.0) >= radius*radius )
	    continue;
	  if ( ( xi+rx<0) || ( xi+rx>= dimX ) || (yi+ry<0) || 
	       ( yi+ry >= dimY ) )
	    continue;
	  
	  // rotate the points
	  float newX = rx * cosAngle - ry * sinAngle;
	  float newY = rx * sinAngle + ry * cosAngle;
	  float xRel = newX + xi - x;
	  float yRel = newY + yi - y;
	  float gaussFactor = exp ( -(xRel * xRel + yRel * yRel) / 
				    ( scaleCompute * scaleCompute ) );
	  float weightedMagnitude = gaussFactor * grad.getVal( xi+rx, yi+ry );

	  //if (weightedMagnitude < 0.0001)
	  //	continue;

	  float gradAng = orie.getVal( xi+rx, yi+ry ) - orientation;
	  if (gradAng>M_PI)
	    gradAng -= 2* M_PI;
	  if (gradAng<-M_PI)
	    gradAng += 2* M_PI;
	  float orient = (gradAng+M_PI) * 8 / (2 * M_PI);
	  
	  // will be placed in 2 x 2 x 2 cases :
	  //  [ xi xi+1 ] x [ yi yi+1 ] x [ angle, angle+1 ]
	  fv.addValue( xf, yf, orient, weightedMagnitude );
        }
    }

  // normalize and threshold
  fv.normalize();
  fv.threshold( 0.2f );
  
  // convert to array of unsigned chars
  for (uint n = 0 ; n < FVBytes.size() ; n++)
    {
      int intval = (int) ( fv.getValue(n) * 512.0);
      if (intval > 255)
	intval = 255;
      else if(intval < 0)
	intval = 0;
      FVBytes[n] = (unsigned char) intval;
    }
}

// #######################################################################
bool Keypoint::equals(const Keypoint& key2) const
{
  bool equal = true;
  if(getDimensionCount() == key2.getDimensionCount())
    {
      for(int i=0; i<getDimensionCount(); i++)
	{
	  if(getDimensionElement(i) != key2.getDimensionElement(i))
	    {
	      equal = false;
	      break;
	    }
	}
    }
  return equal;
}

// ########################################################################
int Keypoint::compareTo(const Keypoint& key2) const
{
  if(magnitude < key2.getMagnitude())
    return -1;
  else if(magnitude > key2.getMagnitude())
    return 1;
  else
    return 0;
}

// ########################################################################
// ########################################################################
// ########### FeatureVector member definitions
// ########################################################################
// ########################################################################

// ########################################################################
void FeatureVector::addValue( float x, float y, float orientation, float value)
{
  // indexes
  //  0 <= x < 4 : [0,4[
  //  0 <= y < 4 : [0,4[
  //  0 <= o < 8 : [0,8[

  // will be lineary added in 8 cases

  int xi = (int)x;
  int yi = (int)y;
  int oi = (int)orientation;

  float magx = x-xi;
  float magy = y-yi;
  float mago = orientation-oi;

  for (int i=0; i<=1; i++)
    {
      if ((i+xi)>3)
	continue;
      
      float mag;
      
      if (i==0)
	mag = magx * value;
      else
	mag = (1.0f - magx) * value;
      
      for (int j=0; j<=1; j++)
        {
	  if ((j+yi)>3)
	    continue;
	  
	  if (j==0)
	    mag *= magy;
	  else
	    mag *= (1.0f - magy);
	  
	  for (int k=0; k<=1; k++)
            {
	      int oriCur = oi + k;
	      if (oriCur > 7)
		oriCur = 0;
	      
	      if (k==0)
		mag *= mago;
	      else
		mag *= (1.0f - mago);
	      
	      FVFloat[ 4*8*(i+xi) + 8*(j+yi) + oriCur ] += mag;
            }
        }
    }
}

// ########################################################################
void FeatureVector::normalize()
{
  uint i;
  float sq = 0.0f;
  for (i=0; i<FVFloat.size(); i++)
    sq += FVFloat[i] * FVFloat[i];
  sq = sqrt(sq);
  for (i=0; i<FVFloat.size(); i++)
    {
      if(sq == 0)
	continue;
      FVFloat[i] /= sq;
    }
}

// ########################################################################
void FeatureVector::threshold( float limit )
{
    bool changed = false;
    for (uint i=0; i<FVFloat.size(); i++)
    {
        if (FVFloat[i] > limit)
        {
            FVFloat[i] = limit;
            changed = true;
        }
    }
    if (changed)
        normalize();
}

// ########################################################################
float FeatureVector::getValue(int index) const
{
  if((uint)index < FVFloat.size())
    return FVFloat[index];
  else
    return 0.0f;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
