/*!@file Keypoint.H Keypoint for SIFT obj recognition */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>

#ifndef KEYPOINT_DEFINED
#define KEYPOINT_DEFINED

#include "Image.H"
#include "SharedPtr.H"
#include "log.H"
#include <iostream>
#include <vector>

// ########################################################################
//! The keypoint base class
// ########################################################################
/*! The sift keypoint feature descriptor
    3D position (x, y, scale)
    Mean orientation
*/
class KDTree;
class Keypoint
{
public :

  //! Constructor
  Keypoint(int fvLength=0);

  //! Constructor
  Keypoint(int fvLength, float x, float y, float s, float o, double mag);
  
  //! Destructor
  virtual ~Keypoint();

  //! Construct from file pointer
  virtual void load(FILE *fp);

  //! x getter
  virtual float getX() const;

  //! x setter
  virtual void setX(float x);

  //! y getter
  virtual float getY() const;

  //! y setter
  virtual void setY(float y);

  //! scale getter
  virtual float getScale() const;

  //! scale setter
  virtual void setScale(float scale);

  //! orientation getter
  virtual float getOrientation() const;

  //! orientation setter
  virtual void setOrientation(float orientation);

  //! magnitude getter
  virtual double getMagnitude() const;

  //! checks for a match in the list using the simple algorithm
  virtual bool simpleMatch(std::vector<Keypoint>& klist, 
                           Keypoint& match, int& numberComparisons) const;

  //! checks for a match in the kd tree
  virtual bool kdTreeMatch(SharedPtr<KDTree>& tree, Keypoint& match,
                           bool useBBF, int& numberComparisons) const;

  //! save to file
  virtual void save(FILE *fp) const;

  //! Return squared distance between two keypoint descriptors.
  virtual int distSquared(const Keypoint& k) const;

  //! Return the max dist sq between two feature vectors
  virtual int maxDistSquared() const;

  //! Return the length of this keypoint's feature vector
  virtual int getDimensionCount() const;

  //! Return the value of the feature vector at the given index
  virtual const int getDimensionElement(int index) const;

  //! Compute the feature vector for this keypoint
  virtual void computeFeature(float x, float y, float s,
                              Image<float>& orie,
                              Image<float>& grad,
                              int dimX, int dimY);

  //! Returns whether or not the two keypoints are equal
  virtual bool equals(const Keypoint& key2) const;

  //! Compares the magnitude of this keypoint to the given keypoint
  int compareTo(const Keypoint& key2) const;

protected:
  float	col_x;				// col, ie x coord
  float row_y;				// row, ie y coord
  float	scale;				// scale
  float	orientation;			// orientation
  std::vector<unsigned char> FVBytes;   // feature vector
  double magnitude;                     // magnitude
};

// ########################################################################
//! Feature Vector
// ########################################################################
/*! Stores temporary keypoint features
    The final feature vector is formed with 'unsigned char'. This one is just 
    float
*/
class FeatureVector
{
public :
  //! Constructor
  FeatureVector(int length)
  {
    FVFloat = std::vector<float>(length);
  }
  
  //! Destructor
  ~FeatureVector() {}

  //! get a value from the feature vector
  float getValue(int index) const;

  //! add value
  /*! linearly add a value. Each couple (x,y) will influence 4 positions in 
      the vector
  */
  void addValue( float x, float y, float orientation, float value );
	
  //! normalize Vector
  void normalize();
	
  //! threshold Vector and renormalize
  void threshold(float limit);
  
private:
  std::vector<float> FVFloat;	// vector of features
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
