/*!@file MathFunctions.C Miscellaneous math functions */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: MathFunctions.C 4206 2005-02-15 20:48:16Z itti $
//

#include "MathFunctions.H"
#include "log.H"
#include <cmath>
#include <sys/types.h> // for getpid()
#include <unistd.h>    // for getpid()

// ######################################################################
bool isFinite(const byte& arg) { return true; }
bool isFinite(const int16& arg) { return true; }
bool isFinite(const int32& arg) { return true; }

#ifdef HAVE_ISFINITE
bool isFinite(const float& arg) { return std::isfinite(arg); }
bool isFinite(const double& arg) { return std::isfinite(arg); }
#else
// Use built-in functions, e.g. on Mac OSX
bool isFinite(const float& arg) { return __isfinitef(arg); }
bool isFinite(const double& arg) { return __isfinited(arg); }
#endif

// #####################################################################
double ran2 (int& idum)
{
  const int IM1 = 2147483563, IM2 = 2147483399; 
  const double AM = (1.0/IM1); 
  const int IMM1 = IM1-1; 
  const int IA1 = 40014, IA2 = 40692, IQ1 = 53668, IQ2 = 52774; 
  const int IR1 = 12211, IR2 = 3791, NTAB = 32; 
  const int NDIV = 1+IMM1/NTAB; 
  const double EPS = 3.0e-16, RNMX = 1.0-EPS; 

  int j, k; 
  static int idum2 = 123456789, iy = 0; 
  static int iv[NTAB]; 
  double temp; 

  if (idum <= 0) { 
    idum = (idum == 0 ? 1 : -idum); 
    idum2=idum; 
    for (j=NTAB+7;j>=0;j--) { 
      k=idum/IQ1; 
      idum=IA1*(idum-k*IQ1)-k*IR1; 
      if (idum < 0) idum += IM1; 
      if (j < NTAB) iv[j] = idum; 
    } 
    iy=iv[0]; 
  } 
  k=idum/IQ1; 
  idum=IA1*(idum-k*IQ1)-k*IR1; 
  
  if (idum < 0) idum += IM1; 
  k=idum2/IQ2; 
  idum2=IA2*(idum2-k*IQ2)-k*IR2; 
  if (idum2 < 0) idum2 += IM2; 
  j=iy/NDIV; 
  iy=iv[j]-idum2; 
  iv[j] = idum; 
  if (iy < 1) iy += IMM1; 
  if ((temp=AM*iy) > RNMX) return RNMX; 
  else return temp; 
}

// ######################################################################
int getIdum(const bool useRandom)
{
  if (useRandom)
    return -(time(NULL) + 17 * getpid());
  else
    return -123456;
}

// ######################################################################
double gasdev (int& idum) {
  static int iset = 0; 
  static double gset;
  
  double fac, rsq, v1, v2; 
  if (idum < 0) iset = 0; 
  if (iset == 0) { 
    do {
      v1 = 2.0*ran2(idum)-1.0;
      v2 = 2.0*ran2(idum)-1.0;
      rsq = v1*v1 + v2*v2; 
    } while (rsq >= 1.0 || rsq == 0.0); 

    fac = std::sqrt(-2.0*std::log(rsq)/rsq); 
    gset = v1*fac; iset = 1; 
    return v2*fac; 
  } else { 
    iset = 0; 
    return gset; 
  } 
}

// ######################################################################
double expdev(int& idum)
{
  double dum;
  do dum = ran2(idum); while (dum == 0.0);
  return -log(dum);
}

// ######################################################################
// Code for computation of psi function (derivative of log gamma)
// ripped from GSL, the GNU scientific library
//
/* Copyright (C) 1996, 1997, 1998, 1999, 2000 Gerard Jungman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* data for a Chebyshev series over a given interval */
struct cheb_series_struct {
  double * c;   /* coefficients                */
  int order;    /* order of expansion          */
  double a;     /* lower interval point        */
  double b;     /* upper interval point        */
  int order_sp; /* effective single precision order */
};
typedef struct cheb_series_struct cheb_series;

/* Chebyshev fits from SLATEC code for psi(x) */
static double psics_data[23] = {
  -.038057080835217922,
   .491415393029387130, 
  -.056815747821244730,
   .008357821225914313,
  -.001333232857994342,
   .000220313287069308,
  -.000037040238178456,
   .000006283793654854,
  -.000001071263908506,
   .000000183128394654,
  -.000000031353509361,
   .000000005372808776,
  -.000000000921168141,
   .000000000157981265,
  -.000000000027098646,
   .000000000004648722,
  -.000000000000797527,
   .000000000000136827,
  -.000000000000023475,
   .000000000000004027,
  -.000000000000000691,
   .000000000000000118,
  -.000000000000000020
};
static cheb_series psi_cs = { psics_data, 22, -1, 1, 17 };

static double apsics_data[16] = {    
  -.0204749044678185,
  -.0101801271534859,
   .0000559718725387,
  -.0000012917176570,
   .0000000572858606,
  -.0000000038213539,
   .0000000003397434,
  -.0000000000374838,
   .0000000000048990,
  -.0000000000007344,
   .0000000000001233,
  -.0000000000000228,
   .0000000000000045,
  -.0000000000000009,
   .0000000000000002,
  -.0000000000000000 
};    
static cheb_series apsi_cs = { apsics_data, 15, -1, 1, 9 };

// Ripped from GSL, the GNU scientific library, function
// cheb_eval_e(); modified to eliminate the error computation and just
// return the result:
double cheb_eval(const cheb_series* cs, const double x)
{
  int j; double d  = 0.0, dd = 0.0;
  double y  = (2.0*x - cs->a - cs->b) / (cs->b - cs->a);
  double y2 = 2.0 * y;

  for (j = cs->order; j >= 1; j--)
    {
      double temp = d;
      d = y2*d - dd + cs->c[j];
      dd = temp;
    }

  return y*d - dd + 0.5 * cs->c[0];
}

// Ripped from GSL, the GNU scientific library, function psi_x();
// modified to eliminate the error computation and just return the
// result:
double psi(double x)
{
  if (x <= 0.0) LFATAL("Only defined for x > 0.0 -- ABORT");
  if (x >= 2.0)
    {
      const double t = 8.0 / (x*x) - 1.0;
      const double cheb = cheb_eval(&apsi_cs, t);
      return log(x) - 0.5/x + cheb;
    }
  else if (x < 1.0)
    {
      const double t1 = 1.0/x;
      const double cheb = cheb_eval(&psi_cs, 2.0*x-1.0);
      return -t1 + cheb;
    }
  else
    {
      const double v = x - 1.0;
      return cheb_eval(&psi_cs, 2.0*v-1.0);
    }
}

// ripped from GSL (see above).
/* coefficients for gamma=7, kmax=8  Lanczos method */
static double lanczos_7_c[9] = {
  0.99999999999980993227684700473478,
  676.520368121885098567009190444019,
 -1259.13921672240287047156078755283,
  771.3234287776530788486528258894,
 -176.61502916214059906584551354,
  12.507343278686904814458936853,
 -0.13857109526572011689554707,
  9.984369578019570859563e-6,
  1.50563273514931155834e-7
};

#define LogRootTwoPi_  0.9189385332046727418

/* Lanczos method for real x > 0;
 * gamma=7, truncated at 1/(z+8) 
 * [J. SIAM Numer. Anal, Ser. B, 1 (1964) 86]
 */
double lngamma(double x)
{
  double Ag;
  double term1, term2;

  x -= 1.0; /* Lanczos writes z! instead of Gamma(z) */

  Ag = lanczos_7_c[0];
  for (int k = 1; k <= 8; k++) { Ag += lanczos_7_c[k] / (x+k); }

  /* (x+0.5)*log(x+7.5) - (x+7.5) + LogRootTwoPi_ + log(Ag(x)) */
  term1 = (x + 0.5) * log((x + 7.5) / M_E);
  term2 = LogRootTwoPi_ + log(Ag);
  return term1 + (term2 - 7.0);
}

// ripped from GSL (see above). Compute probability of getting a value
// k from a Poisson distribution with mean mu:
double poisson(unsigned int k, double mu)
{
  double p;
  double lf = lngamma(k+1); 

  p = exp(log(mu) * k - lf - mu);
  return p;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
