/*!@file MathFunctions.H Miscellaneous math functions */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: MathFunctions.H 4206 2005-02-15 20:48:16Z itti $
//

#ifndef MATHFUNCTIONS_H_DEFINED
#define MATHFUNCTIONS_H_DEFINED

#include "saliency.H"
#include <stdlib.h>  // for lrand48()
#include <algorithm> // for std::swap()

//! Miscellaneous math functions

// ######################################################################
/*! @name Basic numerical operations */
//@{

//! True if \a x is odd.
template <class T>
inline bool ODD(const T& x);

//! Returns the absolute value of \a x.
template <class T>
inline T ABS(const T& x);

//! Returns -1 if \a x is negative, or 1 if \a x is positve.
template <class T>
inline int SGN(const T& x);

//! Returns the square of \a x.
template <class T>
inline T SQR(const T& x);

//! Returns the number closest to \a val and between \a bound1 and \a bound2.
/*! The bounds may be specified in either order (i.e., either (lo, hi) or
  (hi, lo). */
template <class T, class T2>
inline T CLAMP(const T& val, const T2& bound1, const T2& bound2);

//@}

// ######################################################################
/*! @name Numerical range and finiteness testing */
//@{

//! Function to check for finite value
bool isFinite(const byte& arg);

//! Function to check for finite value
bool isFinite(const int16& arg);

//! Function to check for finite value
bool isFinite(const int32& arg);

//! Function to check for finite value
bool isFinite(const float& arg);

//! Function to check for finite value
bool isFinite(const double& arg);

//@}

// ######################################################################
/*! @name Random number generation */
//@{

//! Generate a random number between 0 and n (inclusive)
inline int randomUpTo(const int n);

//! Randomly shuffle given array
template <class T> inline
void randShuffle(T* array, const uint size);

//! Random number generator from Numerical Recipes in C book
/*! Long period >2x10^18 random number generator of L Ecuyer with
    Bays-Durham shuffle. Return a uniform deviate in the interval
    (0,1).  Call with int variable as argument set to a negative
    value, and don't change this variable between calls unless you
    want to reseed the generator. You can get ain initial seed using
    the function getIdum() */
double ran2(int& idum);

//! Get a random seed for ran2
/*! If useRandom is true, a random negative number will be return,
  otherwise a negative number that is always the same will be
  returned. */
int getIdum(const bool useRandom = true);

//! Randomly distributed Gaussian deviate with zero mean and unit variance
double gasdev(int& idum);

//! Randomly distributed Exponential deviate
double expdev(int& idum);

//@}

// ######################################################################
/*! @name Special math functions with iterative implementations */
//@{

//! Compute probability of drawing a value k from a Poisson of mean mu
double poisson(unsigned int k, double mu);

//! Compute log of Gamma
double lngamma(double x);

//! Compute the psi (derivarive of log of gamma) function (requires x > 0.0)
double psi(double x);


//@}

// ######################################################################
// ######################################################################
// ######################################################################
// ##### Implementation of inlined functions
// ######################################################################
// ######################################################################
// ######################################################################
template <class T> inline bool ODD(const T& x)
{ return (x & 1); }

template <class T> inline T ABS(const T& x)
{ return (x > 0) ? x : -x; }

template <class T> inline int SGN(const T& x)
{ return (x < 0) ? -1 : 1; }

template <class T> inline T SQR(const T& x)
{ return (x * x); }

template <class T, class T2>
inline T CLAMP(const T& val, const T2& bound1, const T2& bound2)
{
  if (bound1 < bound2)
    {
      if (val < bound1) return bound1;
      if (val > bound2) return bound2;
      return val;
    }
  else
    {
      if (val < bound2) return bound2;
      if (val > bound1) return bound1;
    }
  return val;
}

inline int randomUpTo(const int n)
{ return lrand48() % (n + 1); }

template <class T> inline void randShuffle(T* array, const uint size)
{
  for (uint i = 0; i < size; i ++)
    {
      uint idx = i + (lrand48() % (size - i));
      std::swap(array[i], array[idx]);
    }
}

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
