// -*- c-basic-offset: 4 -*-
/** @file Matrix3.c
 *
 *  @author Alexandre Jenny <alexandre.jenny@le-geo.com>
 *
 *  $Id: Matrix3.C 3586 2004-09-17 01:19:27Z jbonaiuto $
 *
 *  This is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "Matrix3.H"
#include <math.h>

Matrix3::Matrix3()
{
}

Matrix3::Matrix3(const Matrix3& ot)
{
  (*this) = ot; // call copy operator
}

/** Set the identity matrix */
void Matrix3::setIdentity()
{
  m[0][0] = 1; m[0][1] = 0;  m[0][2] = 0;
  m[1][0] = 0; m[1][1] = 1;  m[1][2] = 0;
  m[2][0] = 0; m[2][1] = 0;  m[2][2] = 1;
}

/** Set the matrice to rotation using yaw, pitch, roll angle */
void Matrix3::setRotation( double Yaw, double Pitch, double Roll )
{
  double	SR	= sin(Roll),
    SP	= sin(Pitch),
    SY	= sin(Yaw),
    CR	= cos(Roll),
    CP	= cos(Pitch),
    CY	= cos(Yaw);
  
  m[0][0]	= CP * CY;
  m[0][1]	= CP * SY;
  m[0][2]	= SP;
  
  m[1][0]	= SR * SP * CY - CR * SY;
  m[1][1]	= SR * SP * SY + CR * CY;
  m[1][2]	= - SR * CP;
  
  m[2][0]	= -( CR * SP * CY + SR * SY );
  m[2][1]	= CY * SR - CR * SP * SY;
  m[2][2]	= CR * CP;
}

/** set the matrice to rotation around X */
void Matrix3::setRotationX( double a )
{
  m[0][0] = 1.0;		m[0][1] = 0.0; 		m[0][2] = 0.0;
  m[1][0] = 0.0; 		m[1][1] = cos(a); 	m[1][2] = sin(a);
  m[2][0] = 0.0;		m[2][1] =-m[1][2];	m[2][2] = m[1][1];
}

void Matrix3::setRotationY( double a )
{
  m[0][0] = cos(a); 	m[0][1] = 0.0;		m[0][2] =-sin(a);
  m[1][0] = 0.0; 		m[1][1] = 1.0;		m[1][2] = 0.0;
  m[2][0] =-m[0][2];	m[2][1] = 0.0;		m[2][2] = m[0][0];
}
	
void Matrix3::setRotationZ( double a )
{
  m[0][0] = cos(a); 	m[0][1] = sin(a); 	m[0][2] = 0.0;
  m[1][0] =-m[0][1];	m[1][1] = m[0][0]; 	m[1][2] = 0.0;
  m[2][0] = 0.0;		m[2][1] = 0.0;		m[2][2] = 1.0;
}

/** copy operator */
Matrix3& Matrix3::operator= (const Matrix3& ot)
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      m[i][j] = ot.m[i][j];
  return *this;
}

/** multiplication with another matrix */
Matrix3 Matrix3::operator*(const Matrix3& ot) const
{
  Matrix3	Result;
  Result.m[0][0] = m[0][0] * ot.m[0][0] + m[0][1] * ot.m[1][0] + m[0][2] * ot.m[2][0];
  Result.m[0][1] = m[0][0] * ot.m[0][1] + m[0][1] * ot.m[1][1] + m[0][2] * ot.m[2][1];
  Result.m[0][2] = m[0][0] * ot.m[0][2] + m[0][1] * ot.m[1][2] + m[0][2] * ot.m[2][2];
  
  Result.m[1][0] = m[1][0] * ot.m[0][0] + m[1][1] * ot.m[1][0] + m[1][2] * ot.m[2][0];
  Result.m[1][1] = m[1][0] * ot.m[0][1] + m[1][1] * ot.m[1][1] + m[1][2] * ot.m[2][1];
  Result.m[1][2] = m[1][0] * ot.m[0][2] + m[1][1] * ot.m[1][2] + m[1][2] * ot.m[2][2];
  
  Result.m[2][0] = m[2][0] * ot.m[0][0] + m[2][1] * ot.m[1][0] + m[2][2] * ot.m[2][0];
  Result.m[2][1] = m[2][0] * ot.m[0][1] + m[2][1] * ot.m[1][1] + m[2][2] * ot.m[2][1];
  Result.m[2][2] = m[2][0] * ot.m[0][2] + m[2][1] * ot.m[1][2] + m[2][2] * ot.m[2][2];
  return Result;
}

/** operator *= */
void Matrix3::operator*=(Matrix3 ot)
{
  Matrix3 Result;
  Result.m[0][0] = m[0][0] * ot.m[0][0] + m[0][1] * ot.m[1][0] + m[0][2] * ot.m[2][0];
  Result.m[0][1] = m[0][0] * ot.m[0][1] + m[0][1] * ot.m[1][1] + m[0][2] * ot.m[2][1];
  Result.m[0][2] = m[0][0] * ot.m[0][2] + m[0][1] * ot.m[1][2] + m[0][2] * ot.m[2][2];
  
  Result.m[1][0] = m[1][0] * ot.m[0][0] + m[1][1] * ot.m[1][0] + m[1][2] * ot.m[2][0];
  Result.m[1][1] = m[1][0] * ot.m[0][1] + m[1][1] * ot.m[1][1] + m[1][2] * ot.m[2][1];
  Result.m[1][2] = m[1][0] * ot.m[0][2] + m[1][1] * ot.m[1][2] + m[1][2] * ot.m[2][2];
  
  Result.m[2][0] = m[2][0] * ot.m[0][0] + m[2][1] * ot.m[1][0] + m[2][2] * ot.m[2][0];
  Result.m[2][1] = m[2][0] * ot.m[0][1] + m[2][1] * ot.m[1][1] + m[2][2] * ot.m[2][1];
  Result.m[2][2] = m[2][0] * ot.m[0][2] + m[2][1] * ot.m[1][2] + m[2][2] * ot.m[2][2];
  *this = Result;
}


/** return inverse if it exists, otherwise identity */
Matrix3 Matrix3::inverse() const
{
  Matrix3 Result;
  double Det = determinant();
  
  if (Det == 0.0f)
    {
      Result.setIdentity();
      return Result;
    }
  
  double	invDet = 1.0f / Det;
  
  Result.m[0][0] =  invDet * ( m[1][1] * m[2][2] - m[2][1] * m[1][2] );
  Result.m[0][1] = -invDet * ( m[0][1] * m[2][2] - m[2][1] * m[0][2] );
  Result.m[0][2] =  invDet * ( m[0][1] * m[1][2] - m[1][1] * m[0][2] );
  
  Result.m[1][0] = -invDet * ( m[1][0] * m[2][2] - m[2][0] * m[1][2] );
  Result.m[1][1] =  invDet * ( m[0][0] * m[2][2] - m[2][0] * m[0][2] );
  Result.m[1][2] = -invDet * ( m[0][0] * m[1][2] - m[1][0] * m[0][2] );
  
  Result.m[2][0] =  invDet * ( m[1][0] * m[2][1] - m[2][0] * m[1][1] );
  Result.m[2][1] = -invDet * ( m[0][0] * m[2][1] - m[2][0] * m[0][1] );
  Result.m[2][2] =  invDet * ( m[0][0] * m[1][1] - m[1][0] * m[0][1] );
  return Result;
}
