// -*- c-basic-offset: 4 -*-
/** @file Matrix3.h
 *
 *  @author Alexandre Jenny <alexandre.jenny@le-geo.com>
 *
 *  $Id: Matrix3.H 3586 2004-09-17 01:19:27Z jbonaiuto $
 *
 *  This is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _MATRIX3_H_
#define _MATRIX3_H_

#include "Vector3.H"

/** general : Matrix3 is a class for handling 3x3 Matrix manipulation.
 *
 * We do not use 4x4 matrix for view point changement as the calculus could be inefficent
 * (some of the coefficients are null, m14 = m24 = m34 = 0 et m44 = 1.0 always).
 */

class Matrix3
{
public:
    /** we define the Matrix3 as 3 colums of 3 rows */
    double m[3][3];
	
public:
    /** default constructor : initialise to zero */
    Matrix3();
	
    /** copy constructor */
    Matrix3(const Matrix3& ot);

    /** Set the identity matrix */
    void setIdentity();

    /** Set the matrice to rotation using yaw, pitch, roll angle */
    void setRotation( double Yaw, double Pitch, double Roll );
    
    /** set the matrice to rotation around X */
    void setRotationX( double a );

    void setRotationY( double a );
	
    void setRotationZ( double a );

    /** copy operator */
    Matrix3& operator= (const Matrix3& ot);
	
    /** multiplication with another matrix */
    Matrix3 operator*(const Matrix3& ot) const;
    
    /** operator *= */
    void operator*=(Matrix3 ot);
    
    /** comparison */
    inline bool operator==(Matrix3& ot) const
    {
	for(int i=0; i<3; i++)
	    for(int j=0; j<3; j++)
		if(m[i][j] != ot.m[i][j])
		    return false;
	return true;
    }
    
    /** comparison */
    inline bool operator!=(Matrix3& ot) const
    {
	return !(*this == ot);
    }
    
    /** retrieves transpose */
    inline Matrix3 transpose()
    {
	Matrix3	Result;
	Result.m[0][0] = m[0][0];
	Result.m[0][1] = m[1][0];
	Result.m[0][2] = m[2][0];
	Result.m[1][0] = m[0][1];
	Result.m[1][1] = m[1][1];
	Result.m[1][2] = m[2][1];
	Result.m[2][0] = m[0][2];
	Result.m[2][1] = m[1][2];
	Result.m[2][2] = m[2][2];
	return Result;
    }
    
    /** get the determinant */
    inline double determinant() const
    {
	double result = 0.0;
	result  = m[0][0] * ( m[1][1] * m[2][2] - m[2][1] * m[1][2] );
	result -= m[1][0] * ( m[0][1] * m[2][2] - m[2][1] * m[0][2] );
	result += m[2][0] * ( m[0][1] * m[1][2] - m[1][1] * m[0][2] );
	return result;
    }
    
    /** transforms a vector */
    inline Vector3 transformVector(const Vector3 &V) const
    {
	Vector3 Result;
	Result.x = V.x * m[0][0] + V.y * m[1][0] + V.z * m[2][0];
	Result.y = V.x * m[0][1] + V.y * m[1][1] + V.z * m[2][1];
	Result.z = V.x * m[0][2] + V.y * m[1][2] + V.z * m[2][2];
	return Result;
    }
    
    /** return inverse if it exists, otherwise identity */
    Matrix3 inverse() const;
};
#endif
