/*!@file ModelComponent.C base class for parameterized model components */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelComponent.C 3043 2004-01-14 02:07:35Z vidhya $
//

#include "ModelComponent.H"
#include "ModelManager.H"
#include "ParamMap.H"
#include "log.H"
#include <iostream>

// ######################################################################
ModelComponent::ModelComponent(ModelManager& mgr,
                               const std::string& descrName,
                               const std::string& tagName) :
  manager(&mgr), dname(descrName), tname(tagName), ModelComponentStarted(false)
{
  if (manager && manager->debugMode()) CLDEBUG(">>>> Constructed <<<<");
}

// ######################################################################
ModelComponent::~ModelComponent()
{
  if (manager && manager->debugMode()) CLDEBUG(">>>> Destroying...");

  // stop ourselves if necessary:
  if (ModelComponentStarted) stop();

  // let our manager know that we are gone, in case we had some
  // command-line options or were registered with it:
  if (manager) manager->destroyComponent(this);

  if (manager && manager->debugMode()) CLDEBUG("<<<< Destroyed");
}

// ######################################################################
void ModelComponent::start()
{
  if (ModelComponentStarted)
    CLFATAL("ModelComponent '%s' already started -- ABORT", dname.c_str());
  if (manager && manager->debugMode()) CLDEBUG(">>>> Starting...");

  // do our startup that is before our subs have been started:
  this->start1();

  // start our subcomponents in order of addition:
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  while(citr != subcomp.end()) { (*citr)->start(); citr ++; }

  // do our startup that is after our subs have been started:
  this->start2();

  // we are started:
  ModelComponentStarted = true;
  if (manager && manager->debugMode()) CLDEBUG("<<<< Started");
}

// ######################################################################
bool ModelComponent::started() const
{ return ModelComponentStarted; }

// ######################################################################
void ModelComponent::stop()
{
  if (ModelComponentStarted == false)
    CLFATAL("ModelComponent '%s' not started -- ABORT", dname.c_str());
  if (manager && manager->debugMode()) CLDEBUG(">>>> Stopping...");

  // do our shutdown that is before our subs have been stopped:
  this->stop1();

  // stop our subcomponents in reverse order of addition:
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.end();
  while(citr != subcomp.begin()) { -- citr;  (*citr)->stop(); }

  // do our shutdown that is after our subs have been stopped:
  this->stop2();

  // we are stopped:
  ModelComponentStarted = false;
  if (manager && manager->debugMode()) CLDEBUG("<<<< Stopped");
}

// ######################################################################
void ModelComponent::destroyManager()
{
  //CLINFO("destroyManager");

  manager = NULL;

  // also let all our subcomponents know about this sad situation:
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  while(citr != subcomp.end()) { (*citr)->destroyManager(); citr ++; }
}

// ######################################################################
void ModelComponent::printout(std::ostream& s, const std::string& prefix) const
{
  // add our tag name to the prefix:
  std::string newprefix;
  if (prefix.size()) newprefix = prefix + "." + tname;
  else newprefix = tname;

  // first, call printout() on all our ModelParam's, in the order in which
  // they registered with us:
  std::list<ModelParamBase *>::const_iterator itr = params.begin();
  while(itr != params.end()) { (*itr)->printout(s, newprefix); itr ++; }

  // then call printout() on all our subcomponents, in the order in
  // which thery were added:
  std::vector< SharedPtr<ModelComponent> >::const_iterator
    citr = subcomp.begin();
  while(citr != subcomp.end()) { (*citr)->printout(s, newprefix); citr ++; }
}

// ######################################################################
void ModelComponent::start1()
{ }

// ######################################################################
void ModelComponent::start2()
{ }

// ######################################################################
void ModelComponent::stop1()
{ }

// ######################################################################
void ModelComponent::stop2()
{ }

// ######################################################################
std::string ModelComponent::descriptiveName() const
{ return dname; }

// ######################################################################
void ModelComponent::setDescriptiveName(const std::string& name)
{ dname = name; }

// ######################################################################
std::string ModelComponent::tagName() const
{ return tname; }

// ######################################################################
void ModelComponent::setTagName(const std::string& name)
{ tname = name; }

// ######################################################################
uint ModelComponent::addSubComponent(const SharedPtr<ModelComponent>& subc)
{ subcomp.push_back(subc); return subcomp.size() - 1; }

// ######################################################################
void ModelComponent::removeSubComponent(const SharedPtr<ModelComponent>& subc)
{
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  bool gotit = false;
  while(citr != subcomp.end()) {
    if ((*citr).get() == subc.get()) { subcomp.erase(citr); gotit=true; break;}
    citr ++;
  }
  if (gotit == false)
    CLERROR("Request to erase unknown subcomponent -- IGNORED");
}

// ######################################################################
void ModelComponent::removeSubComponent(const uint idx)
{
  if (idx >= subcomp.size())
    CLERROR("Attempt to remove subcomp whith index beyond range -- IGNORED");
  else {
    std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
    citr += idx; subcomp.erase(citr);
  }
}

// ######################################################################
void ModelComponent::removeSubComponent(const std::string& tagname)
{
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  while(citr != subcomp.end()) {
    if (tagname.compare((*citr)->tagName()) == 0)
      { subcomp.erase(citr); return; }
    citr ++;
  }
  CLFATAL("Cannot find subComponent '%s' -- ABORT", tagname.c_str());
}

// ######################################################################
void ModelComponent::removeAllSubComponents()
{ while(subcomp.size()) subcomp.pop_back(); }

// ######################################################################
SharedPtr<ModelComponent> ModelComponent::subComponent(const uint idx) const
{
  if (idx >= subcomp.size())
    CLFATAL("%s: request for subcomponent %d but I have only %d -- FATAL",
           dname.c_str(), idx, subcomp.size());
  return subcomp[idx];
}

// ######################################################################
SharedPtr<ModelComponent>
ModelComponent::subComponent(const std::string& tagname) const
{
  std::vector< SharedPtr<ModelComponent> >::const_iterator
    itr = subcomp.begin();
  while(itr != subcomp.end()) {
    if (tagname.compare((*itr)->tagName()) == 0) return *itr;
    itr ++;
  }
  CLFATAL("Cannot find subComponent '%s' -- ABORT", tagname.c_str());
  return SharedPtr<ModelComponent>();  // keep g++ happy
}

// ######################################################################
uint ModelComponent::numSubComp() const
{ return subcomp.size(); }

// ######################################################################
bool ModelComponent::hasSubComponent(const std::string& tagname,
                                     const bool recurse) const
{
  std::vector< SharedPtr<ModelComponent> >::const_iterator
    itr = subcomp.begin();
  while(itr != subcomp.end()) {
    // is the one we are looking for one of our subcomponents?
    if (tagname.compare((*itr)->tagName()) == 0) return true;

    // otherwise, if recursive, see whether it's a subcomp of the current:
    if (recurse && (*itr)->hasSubComponent(tagname, true)) return true;
    itr ++;
  }
  // nowhere to be found
  return false;
}

// ######################################################################
bool ModelComponent::hasSubComponent(const SharedPtr<ModelComponent>& c,
                                     const bool recurse) const
{
  std::vector< SharedPtr<ModelComponent> >::const_iterator
    itr = subcomp.begin();
  while(itr != subcomp.end()) {
    // is the one we are looking for one of our subcomponents?
    if (c.get() == itr->get()) return true;

    // otherwise, if recursive, see whether it's a subcomp of the current:
    if (recurse && (*itr)->hasSubComponent(c, true)) return true;
    itr ++;
  }
  // nowhere to be found
  return false;
}

// ######################################################################
void ModelComponent::exportOptions(const int whatToExport, const bool recurse)

{
  // in the base class implementation, we do nothing for us proper...
  if (manager && manager->debugMode()) CLDEBUG(">>>> Exporting Options...");

  // if recursive, let's propagate to the subcomps:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::iterator citr=subcomp.begin();
      while(citr != subcomp.end()) {
        (*citr)->exportOptions(whatToExport, recurse);
        citr ++;
      }
    }
  if (manager && manager->debugMode()) CLDEBUG("<<<< Exported Options");
}

// ######################################################################
bool ModelComponent::setModelParamString(const std::string& name,
                                         const std::string& value,
                                         const bool recurse, const bool warn)
{
  if (ModelComponentStarted)
    CLFATAL("Cannot change ModelParam values while started -- ABORT");

  bool gotit = false;

  // first let's see if we ourselves have a ModelParam by this name:
  std::list<ModelParamBase *>::iterator itr = params.begin();
  while (itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0)
      { (*itr)->setValString(value); gotit = true; }
    itr ++;
  }

  // if recursive, let's also do the subcomp
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::const_iterator
        citr = subcomp.begin();
      while(citr != subcomp.end()) {
        gotit |= (*citr)->setModelParamString(name, value, recurse, false);
        citr ++;
      }
    }
  if (warn && gotit == false)
    CLERROR("No parameter named '%s' -- IGNORING", name.c_str());
  return gotit;
}

// ######################################################################
std::string ModelComponent::getModelParamString(const std::string& name,
                                                const bool recurse,
                                                const bool warn) const
{
  // first let's see if we ourselves have a ModelParam by this name:
  std::list<ModelParamBase *>::const_iterator itr = params.begin();
  while (itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0) return (*itr)->getValString();
    itr ++;
  }

  // if recursive, let's also do the subcomps:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::const_iterator
        citr = subcomp.begin();
      while(citr != subcomp.end()) {
        std::string str = (*citr)->getModelParamString(name, recurse, false);
        if (str.compare("<unknown>")) return str; // found it!
        citr ++;
      }
    }
  if (warn)
    CLERROR("No parameter named '%s' -- IGNORING", name.c_str());
  return std::string("<unknown>");
}

// ######################################################################
bool ModelComponent::doRequestOption(const std::string& name,
                                     const bool useMyVal,
                                     const bool recurse,
                                     const bool warn)
{
  if (ModelComponentStarted)
    CLFATAL("Cannot request an option while started -- ABORT");
  bool gotit = false;

  // do we have this ModelParam?
  std::list<ModelParamBase *>::iterator itr = findParam(name);
  if (itr != params.end())  // ok, let's send the request to our manager
    { manager->requestOption(this, **itr, useMyVal); gotit = true; }

  // if recurse is true, let's also see if our subcomps have it:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::iterator citr=subcomp.begin();
      while(citr != subcomp.end()) {
        gotit |= (*citr)->doRequestOption(name, useMyVal, recurse, false);
        citr ++;
      }
  }

  if (warn && gotit == false)
    CLERROR("No ModelParam named '%s' -- IGNORED", name.c_str());
  return gotit;
}

// ######################################################################
void ModelComponent::readParamsFrom(const ParamMap& pmap, const bool noerr)
{
  if (ModelComponentStarted)
    CLFATAL("Cannot read ModelParam values while started -- ABORT");

  // if we want to ignore missing params, check first whether we have one:
  if (noerr && pmap.hasParam(tname.c_str()) == false) return;

  // we should have a submap for all our stuff (will trigger error otherwise):
  SharedPtr<ParamMap> submap = pmap.getSubmap(tname.c_str());

  // ok, do our own params:
  std::list<ModelParamBase *>::iterator itr = params.begin();
  while(itr != params.end()) { (*itr)->readFrom(*submap, noerr); itr ++; }

  // then go over our subcomponents, and recurse:
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  while(citr != subcomp.end())
    { (*citr)->readParamsFrom(*submap, noerr); citr ++; }
}

// ######################################################################
void ModelComponent::writeParamsTo(ParamMap& pmap) const
{
  //  if (ModelComponentStarted)
  //  CLERROR("It is not safe to write ModelParam values while started");

  // first, create a submap for all our stuff:
  SharedPtr<ParamMap> submap(new ParamMap);

  // then, do our own params:
  std::list<ModelParamBase *>::const_iterator itr = params.begin();
  while(itr != params.end()) { (*itr)->writeTo(*submap); itr ++; }

  // then go over our subcomponents, and recurse:
  std::vector< SharedPtr<ModelComponent> >::const_iterator
    citr = subcomp.begin();
  while(citr != subcomp.end()) { (*citr)->writeParamsTo(*submap); citr ++; }
  
  // finally add this new submap to our incoming pmap:
  pmap.putSubmap(tname.c_str(), submap);
}

// ######################################################################
void ModelComponent::registerParam(ModelParamBase* mp)
{ params.push_back(mp); }

// ######################################################################
void ModelComponent::unregisterParam(const ModelParamBase* mp)
{
  std::list<ModelParamBase *>::iterator itr = params.begin();
  bool gotit = false;
  while(itr != params.end()) {
    if (*itr == mp) { params.erase(itr); gotit = true; break; }
    itr ++;
  }
  if (gotit == false) CLERROR("Request to unregister unknown parameter");
}

// ######################################################################
std::list<ModelParamBase *>::const_iterator
ModelComponent::findParam(const std::string& name) const
{
  std::list<ModelParamBase *>::const_iterator itr = params.begin();
  while(itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0) return itr;
    itr ++;
  }
  return itr;
}

// ######################################################################
std::list<ModelParamBase *>::iterator
ModelComponent::findParam(const std::string& name)
{
  std::list<ModelParamBase *>::iterator itr = params.begin();
  while(itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0) return itr;
    itr ++;
  }
  return itr;
}

// ######################################################################
void ModelComponent::reset(const bool recurse)

{
  // in the base class implementation, we do nothing for us proper...

  // if recursive, let's propagate to the subcomps:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::iterator citr=subcomp.begin();
      while(citr != subcomp.end()) {
        (*citr)->reset(recurse);
        citr ++;
      }
    }
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
