/*!@file ModelComponent.H base class for parameterized model components */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelComponent.H 2893 2003-10-08 20:36:13Z itti $
//

#ifndef MODELCOMPONENT_H_DEFINED
#define MODELCOMPONENT_H_DEFINED

#include <list>
#include <vector>
#include <string>
#include <iosfwd>
#include "ModelParam.H"
#include "SharedPtr.H"
#include "log.H"
class ParamMap;
class ModelManager;

//! Base class for parameterized model components
/*! The goal is to provide a unified interface by which tunable
  parameters can be set, read, or exported as command-line options or
  in parameter maps or files, so that executables using a variable
  collection of model components can easily be built and configured
  using command-line options and config files. To this end,
  ModelComponent provides a set of facilities that allow it to hold a
  number of ModelParam<T> data members. A ModelComponent also holds a
  list of sub-components that are ModelComponent as well, and many of
  the ModelComponent functions will first apply to the component
  itself, then recursively to its subcomponents. The ModelParam<T> are
  the tunable parameters that are persistent and may be associated
  with a command-line option; see ModelParam.H for details. A
  ModelComponent is attached to a ModelManager that will handle option
  parsing and ModelParam<T> load/save. A ModelComponent may request
  its master to associate some of its ModelParam<T> data members with
  command-line options. A global list of all possible command-line
  option definitions for our entire code base is provided in
  ModelOptionDefs.C. Command-line parsing will set the value of ALL
  ModelParam<T> objects in all ModelComponent objects and their
  subcomponents that have requested the option to the value specified
  on the command line. So this mechanism should be used sparingly, for
  parameters that may be used across a variety of models and a variety
  of ModelComponent derivatives.

  There is one trick to be aware of with respect to the basic
  operation and setup of ModelComponent. For speed reasons, we want
  our ModelParam parameters to be true data members, so that we can
  access them as efficiently as normal data members (rather than
  having to first look them up by name). On the other hand, the
  ModelComponent must also be able to produce a list of all its
  ModelParam members, so that they can be saved to a ParamMap. To
  resolve this, ModelComponent both holds each ModelParam as a regular
  data member, and a list of pointers to all its ModelParam
  members. This list is automatically maintained when the ModelParam
  objects are constructed and destroyed, so that the user does not
  have to worry about it. */ 
class ModelComponent
{
public:
  // ######################################################################
  /*! @name Constructors and Destructors */
  //@{

  //! Constructor
  /*! Typically, in the constructor you should perform all
    initializations that are independent of the values of your
    ModelParam data members. Since the command-line usually has not
    been parsed yet at the time you construct your various
    ModelComponent objects, however, the start() function below is
    provided to allow you to perform additional initializations after
    the command-line has been parsed and the ModelParam data members
    have been configured.
    @param mgr our model manager; see ModelManager.H
    @param descrName a human-readable descriptive name
    @param tagname a computer-oriented unique name (used for storage
           in ParamMap) */
  ModelComponent(ModelManager& mgr, const std::string& descrName,
                 const std::string& tagName);

  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~ModelComponent();

  //! Start the component
  /*! This will start the component by triggering the following
    sequence of actions:
        - Check if we are already started and emit a fatal error if so
        - Call start1() that may be overloaded by derived classes
        - Call start() on all our subcomponents
        - Call start2() that may be overloaded by derived classes
        - Switch us to "started" state, in which modifying our ModelParam
          values externally is forbidden

    Derived classes should overload start1() and start2() as a way to
    get the component started, once the command-line options and
    config files have been parsed and all the ModelParam objects
    have been configured. Default no-op implementations are provided
    for start1() and start2(), so if you don't need to do anything to
    start your component, you don't need to implement start1() and
    start2(). Because this is a recursive call on our sub-components,
    typically you only need to call it once on your ModelManager to
    get your entire hierarchy of ModelComponent objects started. */
  void start();

  //! Did we already get a call to start()?
  bool started() const;

  //! Stop the component
  /*! This will stop the component by triggering the following
    sequence of actions:
        - Check if we are already started and emit a fatal error if not
        - Call stop1() that may be overloaded by derived classes
        - Call stop() on all our subcomponents
        - Call stop2() that may be overloaded by derived classes
        - Switch us out of "started" state, so that modifying our ModelParam
          values externally is allowed

    Derived classes should overload stop1() and stop2() as a way to
    get the component stopped, once the command-line options and
    config files have been parsed and all the ModelParam and objects
    have been configured. Default no-op implementations are provided
    for stop1() and stop2(), so if you don't need to do anything to
    stop your component, you don't need to implement stop1() and
    stop2(). Because this is a recursive call on our sub-component,
    typically you only need to call it once on your ModelManager to
    get the entire hierarchy of ModelComponent objects stopped. */
  void stop();

  //! Signify that our ModelManager just got destroyed
  /*! This should not be called manually, but will be called from
    within the ModelManager destructor. It will ensure that we do not
    try to unregister with our ModelManager if the manager is
    destroyed before us. */
  void destroyManager();
  //@}

  // ######################################################################
  /*! @name Name functions */
  //@{

  //! Get a human-readable descriptive name for this component.
  /*! This is NOT intended to be a name for programmatic use (such as for
      indexing into associative arrays). It is not necessarily unique on a
      per-object basis; it's possible that all objects of the same class
      return the same descriptiveName(). */
  std::string descriptiveName() const;

  //! Change this component's descriptive name.
  void setDescriptiveName(const std::string& name);

  //! Get a programmer-usable tag name for this object.
  /*! This is intended to be unique among component objects, so that this
      string could be used for associate-array indexing, for example. */
  std::string tagName() const;

  //! Change this component's tag name.
  void setTagName(const std::string& name);
  //@}

  // ######################################################################
  /*! @name Sub-Component management functions */
  //@{

  //! Add a subcomponent to our internal list; return its index
  uint addSubComponent(const SharedPtr<ModelComponent>& subc);

  //! Remove a subcomponent from our internal list, by address
  /*! This will only remove the component from our list of registered
    subcomponents, but will not attempt to destroy it. */
  void removeSubComponent(const SharedPtr<ModelComponent>& subc);

  //! Remove a subcomponent from our internal list, by index
  /*! This will only remove the component from our list of registered
    subcomponents, but will not attempt to destroy it. */
  void removeSubComponent(const uint idx);

  //! Remove a subcomponent from our internal list, by tagName
  /*! This will only remove the component from our list of registered
    subcomponents, but will not attempt to destroy it. */
  void removeSubComponent(const std::string& tagname);

  //! Remove all our subcomponents
  void removeAllSubComponents();

  //! Access a sub-component by index
  SharedPtr<ModelComponent> subComponent(const uint idx) const;

  //! Access a sub-component by tagname
  SharedPtr<ModelComponent> subComponent(const std::string& tagname) const;

  //! Get current number of subcomponents
  uint numSubComp() const;

  //! Return true if we have a subcomponent by that tagname
  /*! @param recurse if true, will look not only in our subcomponents
    but also recursively in their subcomponents (depth-first search)*/
  bool hasSubComponent(const std::string& tagname,
                       const bool recurse = false) const;

  //! Return true if we have the pointee as subcomponent
  /*! @param recurse if true, will look not only in our subcomponents
    but also recursively in their subcomponents (depth-first search)*/
  bool hasSubComponent(const SharedPtr<ModelComponent>& c,
                       const bool recurse = false) const;

  //! Show our ModelParam internals and those of our subcomponents
  void printout(std::ostream& s, const std::string& prefix = "") const;

  //! Reset a model component and propagate the request to the sub-components
  /*! For each ModelComponent reset should set it into the state that
    it was just after calling the constructor. It will depend on
    the component what exactly that involves. PROGRAMMER NOTE: when
    you overload this in one of your derived classes, be sure to call
    the base function somewhere in your overload; the base function
    just implements the recursion through the subcomponents*/
  virtual void reset(const bool recurse = true);

  //@}

  // ######################################################################
  /*! @name Access to tunable ModelParam's */
  //@{
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! This should be called once on the model manager to decide which
    sets of ModelParam paremeters should become command-line options;
    the call will then recurse through all ModelComponent objects that
    have been registered with the manager. This is provided so that
    external meta-knowledge about what a collection of ModelComponent
    objects forming a computational model can be used to decide on
    which command-line options are relevent: for instance, if you
    build a model that has no OutputFrameSeries, then you will not be
    able to call saveResults() on your Brain or other objects, so you
    would not want to export command-line options that relate to which
    results should be saved. See ModelOptionDefs.H for the various
    classes of options that you may export. PROGRAMMER NOTE: when you
    overload this in one of your derived classes, be sure to call the
    base function somewhere in your overload; the base function just
    implements the recursion through the subcomponents, and your
    overload should implement a series of calls to
    ModelManager::requestOption() depending on the contents of
    the whatToExport argument. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Parse a parameter value given as string and set the param
  /*! Should not be needed in normal operation, inefficient text-based
      interface. In normal operation, the internals of the
      ModelComponent derivative will directly access the ModelParam,
      not external people. If you want to provide external access to
      your ModelParam values, you should implement a wrapper for them
      in your ModelComponent derivative (e.g., implement a function
      setSize() that will set the value of your Size ModelParam).
      Returns true if the ModelParam was found and set. Throws a fatal
      error if the ModelComponent was started, to avoid external
      people messing around with our ModelParam values during
      operation of the component. PROGRAMMER NOTE: This is called by
      the ModelManager to update us whenever one of our ModelParam
      members for which we have requested a command-line option is
      being changed, and it is also called on each ModelParam when we
      do a readParamsFrom(); so, you may want to overload this call if
      you need to do any internal reconfiguration as some of your
      parameters change; see OrientationChannel in Channels.C for an
      example; also see SaccadeControllerConfigurator in
      SaccadeControllers.H for how to use this to select subcomponents
      at runtime.

      @param name the ModelParam tagname
      @param value the parameter value, to be parsed
      @param recurse will propagate this call to all our subcomponents and
             their subs as well
      @param warn will produce a warning message if parameter not found */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

  //! Set a parameter value
  /*! This is slow, as it will search for the ModelParam by name. If
      you need efficient access to ModelParam values, you should
      define a wrapper method in your ModelComponent derivative. In
      normal operation, indeed, only the methods of the ModelComponent
      derivative should directly access the ModelParam, not external
      people. Returns true if the ModelParam was found and set. Throws
      a fatal error if the ModelComponent was started, to avoid
      external people messing around with our ModelParam values during
      operation of the component. 

      @param name the ModelParam tagname
      @param val the parameter value
      @param recurse will propagate this call to all our subcomponents and
      their subs as well
      @param warn will produce a warning message if parameter not found */
  template <class T> inline
  bool setModelParamVal(const std::string& name, const T& val,
                        const bool recurse = false, const bool warn = true);

  //! Get a parameter value
  /*! Because the parameter may exist both in us and in several of our
      subcomponents, we will stop and return the value of the first
      instance encountered, searching first in us, then recursively in
      our subcomponents (depth-first search, if recurse is true), in
      the order in which they were added. If the parameter is not
      found in us or any of our subcomponents, the string "<unknown>"
      is returned (and a warning is printed out, if warn is true). It
      is ok to call this on a started ModelComponent. */
  std::string getModelParamString(const std::string& name,
                                  const bool recurse = false,
                                  const bool warn = true) const;

  //! Get a parameter value
  /*! This is slow, as it will search for the ModelParam by name. If
      you need efficient access to ModelParam values, you should
      define a wrapper method in your ModelComponent derivative. In
      normal operation, indeed, only the methods of the ModelComponent
      derivative should directly access the ModelParam, not external
      people. Because the parameter may exist both in us and in
      several of our subcomponents, we will stop and return the value
      of the first instance encountered, searching first in us, then
      recursively in our subcomponents (depth-first search, if recurse
      is true), in the order in which they were added. If the
      parameter is not found in us or any of our subcomponents, false
      is returned and val is unchanged (and a warning is issued if
      warn is true). It is ok to call this on a started
      ModelComponent. */
  template <class T> inline
  bool getModelParamVal(const std::string& name, T& val,
                        const bool recurse = false,
                        const bool warn = true) const;

  //! External request that one of our ModelParam becomes an option
  /*! This is to allow external callers to selectively decide that
    some of our internal ModelParam parameters should be exported as
    command-line options. If recurse is true, will look for the
    ModelParam in all our subcomponents, and will link all matching
    ModelParam in us and our subcomponents to the option (so that if a
    command-line value is provided, it will affect all matching
    ModelParam objects in us and all our subcomponents). Will return
    false (and generate a warning message if warn is true) if we (or
    our subcomponents if recurse is true) don't have a ModelParam of
    the specified name. Will generate a fatal error if we have the
    ModelParam but no known command-line option exists for that name
    in ModelOptionDefs.C. Returns true if the ModelParam was found and
    added as a command-line option. Throws a fatal error is we are
    started. */
  bool doRequestOption(const std::string& name, const bool useMyVal = false,
                       const bool recurse = true, const bool warn = true);
  //@}

  // ######################################################################
  /*! @name ParamMap-based configuration functions */
  //@{

  //! Read params from the ParamMap
  /*! Throws a fatal error is we are started.
      @param noerr will not generate an error message if some of our
             parameters do not exist in ParamMap */
  void readParamsFrom(const ParamMap& pmap, const bool noerr = true);

  //! Write params to the ParamMap
  /*! Prints out a warning message if we are started. */
  void writeParamsTo(ParamMap& pmap) const;

  //@}

protected:
  ModelManager *manager;   //<! our manager

  //! This is called from within start() before the subcomponents start
  virtual void start1();

  //! This is called from within start() after the subcomponents have started
  virtual void start2();

  //! This is called from within stop() before the subcomponents stop
  virtual void stop1();

  //! This is called from within stop() after the subcomponents have stopped
  virtual void stop2();

private:
  std::string dname;                                // descriptive name
  std::string tname;                                // tag name
  bool ModelComponentStarted;                       // did we start()?
  std::list<ModelParamBase *> params;               // our model parameters
  std::vector< SharedPtr<ModelComponent> > subcomp; // our sub-components

  // our parameters will register with us upon construction
  void registerParam(ModelParamBase* mp);

  // our parameters will un-register with us upon destruction
  void unregisterParam(const ModelParamBase* mp);

  friend class ModelParamBase; friend class ModelManager;

  // find a ModelParam by name, or return params.end(); const version
  std::list<ModelParamBase *>::const_iterator
  findParam(const std::string& name) const;

  // find a ModelParam by name, or return params.end(); non-const version
  std::list<ModelParamBase *>::iterator
  findParam(const std::string& name);
};

// ######################################################################
// #################### INLINED METHODS:
// ######################################################################

// yes, this is long but if we don't inline it, we'll have to
// implement it in ModelComponent.C for all known ModelParam<T> types

template <class T> inline
bool ModelComponent::setModelParamVal(const std::string& name, const T& val,
                                      const bool recurse, const bool warn)
{
  if (ModelComponentStarted)
    LFATAL("Cannot change ModelParam values while started -- ABORT");

  bool gotit = false;
  
  // first let's see if we ourselves have a ModelParam by this name:
  std::list<ModelParamBase *>::iterator itr = params.begin();
  while (itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0) {
      try {
        (dynamic_cast<ModelParam<T>&>(**itr)).setVal(val);
      } catch (...) {
        LFATAL("Wrong type used with parameter '%s' -- ABORT", name.c_str());
      }

      gotit = true;
    }
    itr ++;
  }

  // if recursive, let's also do the subcomps:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::const_iterator
        citr = subcomp.begin();
      while(citr != subcomp.end()) {
        gotit |= (*citr)->setModelParamVal(name, val, recurse, false);
        citr ++;
      }
    }
  if (warn && gotit == false)
    LERROR("No parameter named '%s' -- IGNORING", name.c_str());
  return gotit;
}

// ######################################################################
template <class T> inline
bool ModelComponent::getModelParamVal(const std::string& name, T& val,
                                      const bool recurse,
                                      const bool warn) const
{
  // first let's see if we ourselves have a ModelParam by this name:
  std::list<ModelParamBase *>::const_iterator itr = params.begin();
  while (itr != params.end()) {
    if (name.compare((*itr)->getName()) == 0) {  // got it!
      try {
        val = (dynamic_cast<const ModelParam<T>&>(**itr)).getVal();
      } catch (...) {
        LFATAL("Wrong type used with parameter '%s' -- ABORT", name.c_str());
      }
      return true;
    }
    itr ++;
  }

  // if recursive, let's also do the subcomps:
  if (recurse)
    {
      std::vector< SharedPtr<ModelComponent> >::const_iterator
        citr = subcomp.begin();
      while(citr != subcomp.end()) {
        if ((*citr)->getModelParamVal(name, val, recurse, false)) return true;
        citr ++;
      }
    }
  if (warn)
    LERROR("No parameter named '%s' -- IGNORING", name.c_str());
  return false;
}

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
