/*!@file ModelManager.C manages a model as collection of ModelComponents */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelManager.C 4218 2005-02-17 20:50:20Z itti $
//

#include "fpe.H"
#include "log.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "ParamMap.H"
#include "StringConversions.H"
#include <iostream>
#include <string>
#include <fstream>

#include "ArrayData.H"     // for ModelOption<Dims>
#include "Point2D.H"       // for ModelOption<Point2D>

#include <unistd.h>        // for isatty()
#include <sys/ioctl.h>     // for TIOCGWINSZ
#include <stdarg.h>        // for ... arguments

// ######################################################################
ModelManager::ModelManager(const std::string& descrName,
                           const std::string& tagName,
                           const bool loadCfgOpt, const bool saveCfgOpt,
                           const bool autoLoadCfg) :
  ModelComponent(*this, descrName, tagName),
  itsPrintHelpMsg(this, "ShowHelpMessage", true),
  itsDebugMode(this, "DebugMode", false),
  itsUsingFPE(this, "UsingFPE", fpIsEnabled()),
  itsTestMode(this, "TestMode", false),
  itsTextLogFile(this, "TextLogFile", ""),
  itsTextLogStream(NULL)
{
  // figure out how many possible options exist:
  const ModelOptionDef *optr = AllModelOptions; // defined in ModelOptionDefs.C
  numopt = 0; while(optr->typ != MODOPT_NONE) { numopt ++; optr ++; }

  // create array for option values:
  opt = new ModelOption[numopt];

  // initialize all our options' default values:
  for (int i = 0; i < numopt; i ++) opt[i].val = AllModelOptions[i].defval;

  // request --help, --debug, --use-fpe and --test-mode
  requestOption(this, itsPrintHelpMsg);
  requestOption(this, itsDebugMode);
  requestOption(this, itsTestMode);
  requestOption(this, itsUsingFPE, true);
  requestOption(this, itsTextLogFile);

  // do we want an option to load a config file?
  if (loadCfgOpt) {
    LoadConfigFile.init(this, "LoadConfigFile", "");
    requestOption(this, LoadConfigFile);
  }

  // do we want an option to save a config file?
  if (saveCfgOpt) {
    SaveConfigFile.init(this, "SaveConfigFile", "");
    requestOption(this, SaveConfigFile);
  }

  // keep track of autoload:
  autoLoadConfig = autoLoadCfg;

  // keep track of whether exportOptions() and saveConfig() get called:
  didExport = false; didSave = false;
}

// ######################################################################
ModelManager::~ModelManager()
{
  // NOTE: one could be tempted to call saveConfig() here for
  // automatic persistence of our ModelParam values. But that would be
  // dangerous as some of our ModelComponents may already have been
  // destroyed by the time we get here. So, saveConfig() should be
  // called manually, at a point in the code where all ModelComponents
  // are still alive.

  // let our components know that we are gone:
  destroyManager();

  // delete our command-line options:
  delete [] opt;
}

// ######################################################################
void ModelManager::registerComponent(const SharedPtr<ModelComponent> component)
{ addSubComponent(component); }

// ######################################################################
void ModelManager::unregisterComponent(const SharedPtr<ModelComponent> c)
{
  // delete this component from our list of option customers:
  for (int i = 0; i < numopt; i ++) {
    std::list<ModelComponent *>::iterator itr = opt[i].comp.begin();
    while(itr != opt[i].comp.end())
      if (c.get() == *itr) itr = opt[i].comp.erase(itr); else itr ++;
  }

  // forget about this component:
  removeSubComponent(c);
}

// ######################################################################
void ModelManager::destroyComponent(const ModelComponent* c)
{
  //LINFO("destroyComponent on %s", c->tagName().c_str());

  // delete this component from our list of option customers:
  for (int i = 0; i < numopt; i ++) {
    std::list<ModelComponent *>::iterator itr = opt[i].comp.begin();
    while(itr != opt[i].comp.end())
      if (c == *itr) itr = opt[i].comp.erase(itr); else itr ++;
  }

  // unregister this component if it was registered:
  std::vector< SharedPtr<ModelComponent> >::iterator citr = subcomp.begin();
  while(citr != subcomp.end())
    if ((*citr).get() == c) citr = subcomp.erase(citr); else citr ++;
}

// ######################################################################
void ModelManager::requestOption(ModelComponent *c, ModelParamBase& p,
                                 const bool useMyVal)
{
  // find the option based on the name of the model param:
  int num = findOption(p.getName());

  // ok, found it. See if we already have this component for this option:
  bool gotit = false;
  std::list<ModelComponent *>::iterator itr = opt[num].comp.begin();
  while(itr != opt[num].comp.end()) {
    if ((*itr) == c) { gotit = true; break; }
    itr ++;
  }

  // Add this component to our client list for that option:
  if (gotit == false) opt[num].comp.push_back(c);

  // Do we want to use its val as our new default? Otherwise, reset
  // its val to our default:
  if (useMyVal) setOptionValString(p.getName(), p.getValString());
  else c->setModelParamString(p.getName(), opt[num].val, false, true);
}

// ######################################################################
void ModelManager::unRequestOption(ModelComponent *c, ModelParamBase& p)
{
  // find the option based on the name of the model param:
  int num = findOption(p.getName());

  // ok, found it. See if we already have this component for this
  // option, and if so, let's remove it from the list of clients for
  // the option:
  std::list<ModelComponent *>::iterator itr = opt[num].comp.begin();
  while(itr != opt[num].comp.end())
    if (c == *itr) itr = opt[num].comp.erase(itr); else itr ++;
}

// ######################################################################
void ModelManager::requestOptionAlias(const std::string name)
{
  // find the option based on the name:
  int num = findOption(name);

  // ok, found it. See if we already have requested this alias:
  bool gotit = false;
  std::list<ModelComponent *>::iterator itr = opt[num].comp.begin();
  while(itr != opt[num].comp.end()) {
    if ((*itr) == this) { gotit = true; break; }
    itr ++;
  }

  // Add this component to our client list for that option:
  if (gotit == false) opt[num].comp.push_back(this);
}

// ######################################################################
bool ModelManager::setOptionValString(const std::string& name,
                                      const std::string& val)
{
  // find the option by name:
  int num = findOption(name);

  // keep the new value so that we can return it via getOptionValString:
  opt[num].val = val;

  // tell the new value to all our clients:
  bool ret = true;
  std::list<ModelComponent *>::iterator itr = opt[num].comp.begin();
  while(itr != opt[num].comp.end()) {
    ret &= (*itr)->setModelParamString(AllModelOptions[num].name, val);
    itr ++;
  }

  // will return true unless we had components and setting the option
  // failed for them:
  return ret;
}

// ######################################################################
std::string ModelManager::getOptionValString(const std::string& name)
{
  // find the option by name:
  int num = findOption(name);

  // return the value:
  return opt[num].val;
}

// ######################################################################
bool ModelManager::parseCommandLine(const int argc, const char **argv,
                                    const char *usage, const int minarg,
                                    const int maxarg)
{
  // did anybody export the options yet? otherwise let's export them
  // all as our default behavior:
  if (didExport == false)
    { exportOptions(OPTEXP_ALL, true); didExport = true; }

  // do a few resets:
  extraargs.clear();

  // let's get our application name:
  std::string procname(argv[0]);
  uint ii = procname.rfind('/'); // eliminate path and keep only the filename:
  if (ii < procname.size()) procname = procname.substr(ii + 1);

  // if we wanted to automatically load a config from ~/.execname,
  // let's do that now:
  if (autoLoadConfig)
    {
      std::string fname = std::string(getenv("HOME")) + "/." + procname;
      FILE *tryit = fopen(fname.c_str(), "r");
      if (tryit) {   // config file exists; let's load it:
        fclose(tryit);
        LINFO("Autoloading configuration from '%s'", fname.c_str());
        ParamMap pmap; pmap.load(fname.c_str());
        readParamsFrom(pmap);
      }
    }

  // do the core of the parsing:
  bool ret = parseCommandLineCore(argc, argv);
  if (ret == false) return false;

  // let's finally check that the number of leftover args is within
  // what we want:
  if (int(extraargs.size()) < minarg ||
      (maxarg >= 0 && int(extraargs.size()) > maxarg))
    {
      if (maxarg == -1)
        LERROR("Incorrect number of (non-opt) arg: %d [%d..Inf]",
               extraargs.size(), minarg);
      else
        LERROR("Incorrect number of (non-opt) arg: %d [%d..%d]",
               extraargs.size(), minarg, maxarg);
      printUsage(procname.c_str(), usage);
    }

  // all went well
  return true;
}

// ######################################################################
bool ModelManager::parseCommandLineCore(const int argc, const char **argv)
{
  // loop over all the options and trigger the appropriate parsing.
  // There are several special cases in here. First, the
  // ModelParam<OptionFlag> objects take no option arguments and are
  // processed as special cases. Second, for the loadConfigFile, we
  // want to load it up as soon as the option is encountered, so that
  // subsequent options are applied to the configured model (indeed,
  // loading the config file will overwrite all ModelParam values). We
  // keep track in loadcfg of the name of the last config file loaded
  // (in case we encounter several --load-config-from=X options):
  std::string loadcfg;
  if (LoadConfigFile.initialized()) loadcfg = LoadConfigFile.getVal();
  int nextarg = 1;  // position of next argument for short options

  // Let's loop over the args and parse each one:
  for (int i = 1; i < argc; i ++)
    {
      // if it's not an option, let's add it as an extra arg:
      if (argv[i][0] != '-') {
        if (i < nextarg)
          continue; // already processed as a short opt arg
        else {
          extraargs.push_back(std::string(argv[i]));
          nextarg = i + 1;
          continue;
        }
      }

      // is it a single '-'?
      if (strlen(argv[i]) < 2) printHelp("Unknown option: -");

      // is it a short option (or a collection of short options)?
      if (argv[i][1] != '-') {
        // loop over all short options specified here:
        for (int j = 1; j < int(strlen(argv[i])); j ++) {
          // let's look it up:
          int o;
          for (o = 0; o < numopt; o ++)
            if (AllModelOptions[o].shortoptname == argv[i][j]) break;

          // if not found, let's give an error:
          if (o == numopt) printHelp("Unknown option: ", argv[i]);

          // if it's special option number 0, let's print the help:
          if (o == 0) printHelp();

          // did anybody request this option?
          if (opt[o].comp.empty()) printHelp("Unimplemented option: ",argv[i]);

          // if it's a MODOPT_ALIAS, let's recursively parse its value
          // and we'll be done with that option:
          if (AllModelOptions[o].typ == MODOPT_ALIAS)
            { parseCommandLineCore(AllModelOptions[o].defval); continue; }

          // if it's a MODOPT_FLAG, let's turn it on:
          if (AllModelOptions[o].typ == MODOPT_FLAG)
            setOptionValString(AllModelOptions[o].name, "true");
          else {
            // otherwise, we need an argument for the option:
            while (nextarg < argc && argv[nextarg][0] == '-') nextarg ++;
            if (nextarg >= argc)
              printHelp("Missing argument for option: ", argv[i]);

            // ok, let's set the value:
            if (setOptionValString(AllModelOptions[o].name,
                                   argv[nextarg]) == false)
              printHelp("Invalid value for option: ", argv[i]);

            // ready for the next short option in our current arg
            nextarg ++;
          }
        }
      } else {
        // it's a long option

        // is it just '--'?
        if (strlen(argv[i]) < 3) printHelp("Unknown option: --");

        // do we have an '=' symbol in the string?
        const char *eq = strchr(argv[i], '=');
        if (eq == NULL) {
          // is it a MODOPT_FLAG or MODOPT_ALIAS? let's look it up:
          int o; const char *cval = "true";
          for (o = 0; o < numopt; o ++)
            if (strcmp(AllModelOptions[o].longoptname,
                       &(argv[i][2]))==0) break;

          // if it's special option number 0, let's print the help:
          if (o == 0) printHelp();
          // if not found, since there is no '=', it could still be --noXXX
          else if (o == numopt) {
            // check that we have 'no' as prefix, plus something else:
            if (argv[i][2] != 'n' || argv[i][3] != 'o' || strlen(argv[i]) == 4)
              printHelp("Unknown option: ", argv[i]);

            // let's look up that --noXXX:
            for (o = 0; o < numopt; o ++)
              if (strcmp(AllModelOptions[o].longoptname,
                         &(argv[i][4])) == 0) break;

            // if not found, let's give up:
            if (o == numopt) printHelp("Unknown option: ", argv[i]);

            // we'll continue, knowing that we want to set the value to false:
            cval = "false";
          }

          // did anybody request this option?
          if (opt[o].comp.empty())
            printHelp("Unimplemented option: ", argv[i]);

          // if it's a MODOPT_ALIAS, let's recursively parse its value
          // and we'll be done for that option:
          if (AllModelOptions[o].typ == MODOPT_ALIAS)
            { parseCommandLineCore(AllModelOptions[o].defval); continue; }

          // if it's a MODOPT_FLAG, let's turn it on/off, otherwise trouble:
          if (AllModelOptions[o].typ == MODOPT_FLAG)
            setOptionValString(AllModelOptions[o].name, cval);
          else
            printHelp("Missing required argument for option: ", argv[i]);
        } else {
          // ok, we have an '='; let's isolate the option name:
          char oname[eq - argv[i] - 1];
          strncpy(oname, &(argv[i][2]), eq - argv[i] - 2);
          oname[eq - argv[i] - 2] = '\0';

          // ok. let's look it up:
          int o; for (o = 0; o < numopt; o ++)
            if (strcmp(AllModelOptions[o].longoptname, oname) == 0) break;

          // if not found, it's a bogus option name:
          if (o == numopt) printHelp("Unknown option: --", oname);

          // did anybody request this option?
          if (opt[o].comp.empty())
            printHelp("Unimplemented option: --", oname);

          // if it's a MODOPT_ALIAS, user is goofing off:
          if (AllModelOptions[o].typ == MODOPT_ALIAS)
            printHelp("No argument permitted for option alias --", oname);

          // is there anything after the '='?
          if (eq - argv[i] == int(strlen(argv[i])) - 1)
            printHelp("Missing value for option: --", oname);

          // all right, let's set the value:
          if (setOptionValString(AllModelOptions[o].name, &(eq[1])) == false)
            printHelp("Invalid value for option: --", oname);
        }
      }

      // if processing this option just changed our loadConfigFile
      // parameter, that means that we just parsed an option that
      // specified that value. Let's load up the file immediately:
      if (LoadConfigFile.initialized() &&
          loadcfg.compare(LoadConfigFile.getVal()))
        { loadConfig(); loadcfg = LoadConfigFile.getVal(); }
    }

  // all went well:
  return true;
}

// ######################################################################
bool ModelManager::parseCommandLineCore(const char *args)
{
  // let's just split our single string into an argv-like array and
  // call the other version of parseCommandLineCore:
  char *a = new char[strlen(args) + 1]; strcpy(a, args);

  const int maxnum = 100; // max number of options in the alias
  typedef const char* const_char_ptr;
  const char **argv = new const_char_ptr[maxnum];
  argv[0] = NULL;  // this usually is the executable name and will be ignored

  int argc = 1;     // current argv number
  char *ptr = a, *stop = a + strlen(a); // pointer to next option and to end
  while(ptr < stop)
    {
      argv[argc++] = ptr;  // store address of option
      if (argc >= maxnum)
        LFATAL("Wooooo, your alias is too long! -- ABORT on '%s'", args);
      while(!isspace(*ptr) && ptr != stop) ptr++; // find next white space
      *ptr++ = '\0';  // blank it out and get ready for next one
    }

  // ok, let's parse them all!
  bool ret = parseCommandLineCore(argc, argv);

  // free our local memory and return:
  delete [] argv; delete [] a;
  return ret;
}

// ######################################################################
void ModelManager::exportOptions(const int whatToExport, const bool recurse)
{
  didExport = true;
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
uint ModelManager::numExtraArgs() const
{ return extraargs.size(); }

// ######################################################################
std::string ModelManager::getExtraArg(const uint num) const
{
  if (num >= extraargs.size()) {
    LERROR("Invalid arg number %d (0..%d) -- IGNORED", num, extraargs.size());
    return std::string("");
  }
  return extraargs[num];
}

// ######################################################################
void ModelManager::printUsage(const char *procname, const char *usage)
{
  std::cerr<<"USAGE: "<<procname<<" "<<usage<<std::endl;
  std::cerr<<"Try '"<<procname<<" --help' for additional information."<<
    std::endl;
  exitProgram(1);
}

// ######################################################################
void ModelManager::printHelp(const char *errormsg, const char *err2)
{
  // if we have an error message, let's start with that:
  if (errormsg && err2)
    std::cerr<<"ERROR: "<<errormsg<<err2<<std::endl<<std::endl;
  else if (errormsg)
    std::cerr<<"ERROR: "<<errormsg<<std::endl<<std::endl;

  // figure out number of columns in our terminal:
  unsigned int ncols = 80;  // default width
#ifdef TIOCGWINSZ
  if (isatty(STDIN_FILENO)) {
    struct winsize ws;
    if (!ioctl(STDIN_FILENO, TIOCGWINSZ, &ws) && ws.ws_col > 0)
      ncols = ws.ws_col;
  }
#endif

  // we'll pile up the relevant help messages for each category; so
  // here we need a vector (one entry per category) of vectors (one
  // entry per help message) of strings (each help message):
  std::vector< std::vector<std::string> > bycateg(NBMODELOPTIONCATEGS);

  // Go over all possible options and add some help to the correct
  // category if we have at least one ModelComponent that has
  // requested the option:
  for (int i = 0; i < numopt; i ++) {
    int c = int(AllModelOptions[i].categ);
    const ModelOptionDef *mo = &(AllModelOptions[i]);
    std::string help = "";

    if (opt[i].comp.empty() == false) // this option has some clients!
      {
        switch(AllModelOptions[i].typ)
          {
          case MODOPT_ALIAS:
            if (mo->shortoptname != '\0')
              help += std::string("-") + mo->shortoptname + std::string(", ");
            help += std::string("--") + mo->longoptname +
              std::string("\n") + mo->descr +
              std::string(". EQUIVALENT TO: ") + mo->defval;
            break;
          case MODOPT_FLAG:
            {
              if (mo->shortoptname != '\0')
                help += std::string("-") + mo->shortoptname+std::string(", ");
              help += std::string("--[no]") + mo->longoptname;
              bool on; convertFromString(opt[i].val, on);
              if (on) help += " [yes]"; else help += " [no]";
              help += std::string("\n") + mo->descr + std::string(" [") +
                mo->name + std::string("]");
            }
            break;
          default:
            {
              if (mo->shortoptname != '\0')
                help += std::string("-") + mo->shortoptname+std::string(", ");
              help += std::string("--") + mo->longoptname + std::string("=") +
                mo->validvals + std::string(" [") + opt[i].val +
                std::string("]\n") + mo->descr + std::string(" [") +
                mo->name + std::string("]");
            }
          }
        bycateg[c].push_back(help);
      }
  }

  // now loop over all categories and print their options if any:
  std::cerr<<"\nCOMMAND-LINE OPTION SYNTAX:\n";
  for (int c = 0; c < NBMODELOPTIONCATEGS; c ++)
    if (bycateg[c].empty() == false)
      {
        // let's print out the category heading:
        std::cerr<<std::endl<<AllModelOptionCategs[c]<<':'
                 <<std::endl<<std::endl;

        // now loop over all options in that category:
        for (uint i = 0; i < bycateg[c].size(); i ++) {
          // let's format the help message to terminal width:
          std::string h = bycateg[c][i];
          // we have 2 lines, separated by a \n; let's split them:
          unsigned int pos = h.find('\n');
          std::string h1 = h.substr(0, pos), h2 = h.substr(pos + 1);
          // let's print this out:
          std::cerr<<formatHelpString(h1, 2, 4, ncols)<<std::endl;
          std::cerr<<formatHelpString(h2, 6, 6, ncols)<<std::endl<<std::endl;
        }
      }

  // exit
  exitProgram(0);
}

// ######################################################################
void ModelManager::exitProgram(int return_code)
{ exit(return_code); }

// ######################################################################
std::string ModelManager::formatHelpString(const std::string& h,
                                           const unsigned int spc1,
                                           const unsigned int spc2,
                                           const unsigned int ncols) const
{
  // let's start with a bunch of spaces:
  std::string ret = std::string(spc1, ' '); unsigned int pos = 0;
  unsigned int spc = spc1;

  // get a chunk of the incoming string:
  while(pos < h.length())
    {
      // does the rest of h fit?
      if (h.length() - pos + spc < ncols) { ret += h.substr(pos); break; }
      
      // here is the max we could fit in the current line of output:
      unsigned int pos2 = pos + ncols - spc - 1;
      
      // do we have an LF?
      unsigned int poslf = h.find('\n', pos);
      if (poslf != h.npos && poslf < pos2) pos2 = poslf;
      
      unsigned int pos3 = h.find_last_of(" |=+-,:*@#\n", pos2);
      if (pos3 != h.npos && pos3 >= pos)
        {
          if (h[pos3] != '\n')
            ret += h.substr(pos, pos3 - pos + 1)+"\n"+std::string(spc2, ' ');
          else
            ret += h.substr(pos, pos3 - pos + 1) + std::string(spc2, ' ');
          
          pos = pos3 + 1; spc = spc2;
          if (h[pos] == '\n') pos++; // skip any LF
        }
      else
        {
          // cannot cut -- abort this mess
          ret += h.substr(pos); break;
        }
    }
  return ret;
}

// ######################################################################
bool ModelManager::loadConfig()
{
  if (LoadConfigFile.initialized() &&
      LoadConfigFile.getVal().empty() == false) {
    LINFO("Loading configuration from '%s'", LoadConfigFile.getVal().c_str());
    ParamMap pmap; pmap.load(LoadConfigFile.getVal().c_str());
    readParamsFrom(pmap);
  }
  return true;  //FIXME: ParamMap should return false if file not found, etc
}

// ######################################################################
bool ModelManager::loadConfig(const std::string& fname)
{
  LINFO("Loading configuration from '%s'", fname.c_str());
  ParamMap pmap; pmap.load(fname.c_str());
  readParamsFrom(pmap);
  return true;  //FIXME: ParamMap should return false if file not found, etc
}

// ######################################################################
bool ModelManager::saveConfig() const
{
  if (SaveConfigFile.initialized() &&
      SaveConfigFile.getVal().empty() == false)
    return saveConfig(SaveConfigFile.getVal());
  else
    return true; //FIXME: ParamMap should return false if cannot write file
}

// ######################################################################
bool ModelManager::saveConfig(const std::string& fname) const
{
  LINFO("Saving configuration to '%s'", fname.c_str());
  ParamMap pmap; writeParamsTo(pmap);
  pmap.format(fname.c_str());

  return true; //FIXME: ParamMap should return false if cannot write file
}

// ######################################################################
int ModelManager::findOption(const std::string& name) const
{
  int num = -1; const char* nam = name.c_str();
  for (int i = 0; i < numopt; i ++)
    if (strcmp(nam, AllModelOptions[i].name) == 0) { num = i; break; }
  if (num == -1) LFATAL("Unknown option named '%s' -- ABORT", nam);
  return num;
}

// ######################################################################
void ModelManager::start1()
{
  // report a bunch of settings:
  LINFO("Debug: %s, FPE: %s, TestMode: %s",
        itsDebugMode.getVal() ? "ON" : "OFF",
        itsUsingFPE.getVal() ? "ON" : "OFF",
        itsTestMode.getVal() ? "ON" : "OFF");

  // did we want a debug printout?
  if (itsDebugMode.getVal()) {
    std::cerr<<"==================== MODEL PARAMETERS ====================\n";
    printout(std::cerr);
    std::cerr<<"==========================================================\n";
  }

  // open a text log if so requested by command-line:
  if (itsTextLogFile.getVal().empty() == false)
    {
      itsTextLogStream = new std::ofstream(itsTextLogFile.getVal().c_str());
      if (itsTextLogStream->is_open() == false)
        LFATAL("Cannot open '%s' for output -- ABORT.",
               itsTextLogFile.getVal().c_str());
      LINFO("Using textlog file: %s", itsTextLogFile.getVal().c_str());
    }

  // if saveConfig() has not been called yet, let's do it now just
  // before we start all our components:
  if (didSave == false) { didSave = true; saveConfig(); }
}

// ######################################################################
bool ModelManager::setModelParamString(const std::string& name,
                                       const std::string& value,
                                       const bool recurse,
                                       const bool warn)
{
  // keep track of which values changed:
  bool prevFPE = itsUsingFPE.getVal();

  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // debug mode on/off?
  if (name.compare(itsDebugMode.getName()) == 0) {
    if (itsDebugMode.getVal()) MYLOGVERB = LOG_DEBUG;
    else MYLOGVERB = LOG_INFO;
  }

  // floating-point exceptions on/off?
  if (name.compare(itsUsingFPE.getName()) == 0 &&
      itsUsingFPE.getVal() != prevFPE) {  // param did change
    if (itsUsingFPE.getVal()) fpExceptionsOn();
    else fpExceptionsOff();
  }

  // test mode on/off?
  if (name.compare(itsTestMode.getName()) == 0) {
    if (itsTestMode.getVal()) {
      srand48(0);  // will produce deterministic pseudo-random sequences
      srand(0);
      setOptionValString("UseRandom", "false");
    } else {
      srand48(time(0) + getpid() * 123);
      srand(time(0) + getpid() * 345);
      setOptionValString("UseRandom", "true");
    }
  }

  return ret;
}

// ######################################################################
void ModelManager::stop2()
{
  // close textlog file if open:
  if (itsTextLogStream)
    {
      if (itsTextLogStream->is_open()) itsTextLogStream->close();
      delete itsTextLogStream;
    }
  itsTextLogStream = NULL;
}

// ######################################################################
bool ModelManager::debugMode() const
{ return itsDebugMode.getVal(); }

// ######################################################################
bool ModelManager::usingFPE() const
{ return itsUsingFPE.getVal(); }

// ######################################################################
bool ModelManager::testMode() const
{ return itsTestMode.getVal(); }

// ######################################################################
void ModelManager::textLog(const std::string& message)
{
  if (itsTextLogStream && itsTextLogStream->is_open())
    {
      (*itsTextLogStream) << message << std::endl;
      itsTextLogStream->flush();  // don't lose messages if ctrl-C
    }
}

// ######################################################################
void ModelManager::textLog(const char *message, ...)
{
  if (itsTextLogStream && itsTextLogStream->is_open())
    {
      va_list a; va_start(a, message);
      char txt[5000]; vsnprintf(txt, 5000, message, a); va_end(a);
      (*itsTextLogStream) << txt << std::endl;
      itsTextLogStream->flush();  // don't lose messages if ctrl-C
    }
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
