/*!@file ModelManager.H manages a model as collection of ModelComponents */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelManager.H 3738 2004-09-28 09:56:21Z walther $
//

#ifndef MODELMANAGER_H_DEFINED
#define MODELMANAGER_H_DEFINED

#include "ModelComponent.H"
class ParamMap;

//! Manages a collection of ModelComponent objects forming a model
/*! ModelManager should be implemented once in each executable, and
  will facilitate the management of a collection of ModelComponent
  objects, load/save of tunable ModelParam<T> parameters that may be
  in those objects, and generation/parsing of command-line
  options. ModelComponent objects must manually register with the
  ModelManager to fall under its management.  Typically, when a
  ModelComponent has subcomponents, only the top-level component
  should be registered with the ModelManager, as all ModelParam<T>
  access functions will recurse through the ModelComponent's
  subcomponents. See test-model.C and test-audioGrab.C for simple
  examples of how to write executables that are managed by
  ModelManager. */
class ModelManager : public ModelComponent
{
public:
  // ######################################################################
  /*! @name Constructors and Destructors */
  //@{

  //! Default constructor
  /*! @param descrName descriptine name
    @param tagName tag name (see ModelComponent.H)
    @param loadCfgOpt create a --load-config-from=X option if true
    @param saveCfgOpt create a --save-config-to=X option if true
    @param autoLoadCfg try to load config from ~/.execname if true */
  ModelManager(const std::string& descrName = "The Model",
               const std::string& tagName = "model",
               const bool loadCfgOpt = true, const bool saveCfgOpt = true,
               const bool autoLoadCfg = true);

  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~ModelManager();

  //! Returns true if we are running in debug mode
  /*! Debug mode can be invoked via the --debug command-line
    option. Programs can use this function to do additional debug
    outputs if the manager is in debug mode. In addition, we will
    generate a debug printout of all model parameters just before
    start() and we will force the log level to LOG_DEBUG (see log.H)
    if debug mode is on. */
  bool debugMode() const;

  //! Returns true if we are using floating-point exceptions
  /*! Floating point exceptions are on by default, but may be turned
    off via the command-line option --nouse-fpe. */
  bool usingFPE() const;

  //! Returns true if we are in test mode as per --test-mode
  /*! Various components will adjust their behavior in various ways
    when in test mode. The goal is to provide an exactly reproducible,
    batch-oriented execution of each program, to be used in the test
    suite. At the manager level, we turn off --use-random and
    initialize random numbers to a fixed seed (in case some components
    still use random numbers while --use-random is off), while we
    initialize the random numbers to a time and PID-dependent seed
    otherwise. Do a grep for the name of this function on other
    components to see if they use it (e.g., FrameSeries will turn off
    all displays when in test mode). */
  bool testMode() const;

  //@}

  // ######################################################################
  /*! @name ModelComponent management functions */
  //@{

  //! Register a component with the manager
  /*! This is only a different name for ModelComponent's
    addSubComponent(), just to make ModelManager feel special. */
  void registerComponent(const SharedPtr<ModelComponent> component);

  //! Unregister a component with the manager
  /*! This is only a different name for ModelComponent's
    removeSubComponent(), just to make ModelManager feel special. If
    you have a component that has requested options from the manager
    and is about to be destroyed, be sure to call this first so that
    the manager will not try to set the option values on your
    component. */
  void unregisterComponent(const SharedPtr<ModelComponent> component);

  //! ModelComponent objects will call this upon destruction
  /*! This should not be called manually; ModelComponent objects will
    call this upon destruction. This will remove the corresponding
    ModelComponent from our list of command-line option customers and
    from our list of registered components (both as applicable). */
  void destroyComponent(const ModelComponent *c);
  //@}

  // ######################################################################
  /*! @name Command-line options */
  //@{

  //! ModelComponent objects may request a command-line option here
  /*! @param c the ModelComponent object requesting the option
    @param p the ModelParam data member of c corresponding to the
    option
    @param useMyVal if true, we will do a setOptionValString() using
    the current value of the parameter, so that this value will be
    propagated to all ModelComponent objects that have requested this
    option and will become the new command-line default value. If
    false, the ModelParam p will be updated to the current value of
    the option. So, use useMyVal=true if your component wishes to push
    its current parameter value as the default for everyone using this
    option, and use false if your component wishes to use the current
    default instead. */
  void requestOption(ModelComponent *c, ModelParamBase& p,
                     const bool useMyVal = false);

  //! ModelComponents may request the removal from the command-line options
  /*! @param c the ModelComponent object requesting the option
    @param p the ModelParam data member of c corresponding to the option */
  void unRequestOption(ModelComponent *c, ModelParamBase& p);

  //! Users may request model option aliases
  /*! @param name the name of the option alias, from ModelOptionDefs.C */
  void requestOptionAlias(const std::string name);

  //! Set an option value
  /*! The ModelParam objects in all ModelComponent objects that have
    requested this option will be updated with the new value. Will
    throw a fatal error if the model has been started (see
    ModelComponent::start()). */
  bool setOptionValString(const std::string& name, const std::string& val);

  //! Get an option value
  /*! The value returned here is the value currently held by the
    ModelManager. This value is updated in three cases: 1) when a new
    component requests the option using requestOption() with useMyVal
    true; 2) when setOptionValString() is called, and 3) when the
    command-line arguments are parsed by ParseCommandLine, if a value
    for this option has been specified in the command line. If
    ModelComponent objects modify their local ModelParam values for
    their local parameter with this name, the value returned here will
    not be affected. */
  std::string getOptionValString(const std::string& name);

  //! Parse all command-line options and set our ModelComponent params
  /*! This function will create a set of command-line options from our
    current collection of ModelComponents. It will then parse the
    command-line arguments and set the various option values
    accordingly, using a call to setOptionValString(). Any left-over
    args (i.e., not formatted as options) will be available through
    getExtraArg(), and if the number of such extra args is not within
    the specified minarg/maxarg bounds, a usage message and optionlist
    will be printed out (also printed out if a '--help' option is
    encountered). To decide which options should be exported by your
    various ModelComponent objects, typically you will call
    exportOptions() on the manager, which will propagate the call to
    all registered components. Here we have an additional default
    behavior that if exportOptions() has not been called on us, we
    will call it with an order to export all options, just before we
    start parsing the command-line. So, if you want to export all
    options, you don't have to do anything; if you want to only export
    special classes of options, then you will need to call
    exportOptions() manually some time before you parse the
    command-line. Returns true upon fully successful parse.

    @param argc the number of command-line args 
    @param argv the command-line args 
    @param usage a textual description of the expected usage format,
      describing only what the non-option arguments should be (e.g.,
      "<input.ppm> <output.ppm>")
    @param minarg the minimum number of non-optional arguments
    @param maxarg the maximum number of non-optional args, or -1 if
      unbounded */
  bool parseCommandLine(const int argc, const char **argv,
                        const char *usage, const int minarg, const int maxarg);

  //! Get the number of available extra args
  uint numExtraArgs() const;

  //! Get an extra arg, by number
  std::string getExtraArg(const uint num) const;

  //! Print short usage message
  /*! This is automatically called during parseCommandLine() if necessary */
  void printUsage(const char *procname, const char *usage);

  //! Print a categorized help message depending on requested options
  /*! This is automatically called during parseCommandLine() if necessary */
  void printHelp(const char *errormsg = NULL, const char *err2 = NULL);


  //! Overload of the ModelComponent function
  /*! This just behaves like the ModelComponent exportOptions(), plus
    we keep track of whether this was called or not, so that we may
    call it automatically from within parseCommandLine(), if
    exportOptions() has not been called before. */
  void exportOptions(const int whatToExport, const bool recurse = true);


  //@}

  // ######################################################################
  /*! @name Logging to a text file (--textlog=XX option) */
  //@{

  //! Log a message to our textlog file defined via --textlog=XX option
  /*! This version is for C++ formatted message. If the option was not
    specified in the command line, this is a no-op. */
  void textLog(const std::string& message);

  //! Log a message to our textlog file defined via --textlog=XX option
  /*! This version if for printf()-like message syntax. If the option
    was not specified in the command line, this is a no-op. */
  void textLog(const char *message, ...);

  //@}

  // ######################################################################
  /*! @name Persistent ModelParam<T> management */
  //@{

  //! Load all our ModelParam<T> values from file
  /*! This function is automatically called during parseCommandLine()
    if --load-config-from=XX option is encountered or autoload was
    specified in our constructor (thus ensuring that command-line
    options encountered after the option specifying which config file
    to load are applied after the config file has been loaded). It may
    also be called manually. This function is a no-op if no filename
    is in our loadConfigFile internal ModelParam (this parameter can
    be set by the --load-config-from=X command-line option).
    Otherwise, load the file (which should be a ParamMap) and
    configure all our parameters. Returns true on success. */
  bool loadConfig();

  //! Explicitly load all our ModelParams from file
  /*! This function will always load the params from the given file,
    even is --load-config-from=X was not specified. */
  bool loadConfig(const std::string& fname);

  //! Save all our ModelParams to file
  /*! This function is a no-op if no filename is in our saveConfigFile
    internal ModelParam (this parameter can be set by the
    --save-config-to=X command-line option). Otherwise, save all our
    params to the file as a ParamMap. Returns true on success. You
    should typically call this either some time before you start()
    your model, or after you have stop()'ed it.  NOTE: There is a
    default behavior which is to call this as the first thing in
    start(), i.e., just before all our registered components will be
    started, unless you have manually called it yourself earlier, in
    which case it is not called again. */
  bool saveConfig() const;

  //! Explicitly save all our ModelParams to file
  /*! This function will always save the params to the given file,
    even is --save-config-to=X was not specified. */
  bool saveConfig(const std::string& fname) const;

  //@}

  //! Intercept people changing our ModelParam's
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  bool setModelParamString(const std::string& name,
                           const std::string& value,
                           const bool recurse = false,
                           const bool warn = true);

protected:
  ModelParam<bool> itsPrintHelpMsg;  // print help message?
  ModelParam<bool> itsDebugMode;     // use debug mode?
  ModelParam<bool> itsUsingFPE;      // use floating-point exceptions?
  ModelParam<bool> itsTestMode;      // use test mode?
  ModelParam<std::string> itsTextLogFile; // Name of file to save text logs to
  ModelParam<std::string> LoadConfigFile; // Name of a config file to load up
  ModelParam<std::string> SaveConfigFile; // Name of a config file to save to
  bool autoLoadConfig; // Load ~/.execname just before parsing command-line
  std::vector<std::string> extraargs;  // non-option command-line args

  struct ModelOption {
    std::string val;  // current value in string representation
    std::list<ModelComponent *> comp; // the components that requested it
  };
  ModelOption *opt;  // array of options
  int numopt;        // total number of known options
  bool didExport;    // was exportOptions() already called?
  bool didSave;      // was saveConfig() already called?

  // find an option by name and return its index:
  int findOption(const std::string& name) const;

  // format a string for help message
  std::string formatHelpString(const std::string& h, const unsigned int spc1,
                               const unsigned int spc2,
                               const unsigned int ncols) const;

  // stream for the textlog file (--textlog=XXX option):
  /* we declare as pointer just so that we don't have to #include
     <fstream>, which would slow donw lots of compilations. */
  std::ofstream *itsTextLogStream;

  // start and saveConfig first if not saved yet
  void start1();

  // stop the model
  void stop2();

  // this is the core of parseCommandLine() and may be called
  // recursively to handle command-line aliases. Note: beware that the
  // first element of argv will be ignored (as it usually is the
  // program name).
  bool parseCommandLineCore(const int argc, const char **argv);
  bool parseCommandLineCore(const char *args);

  //! virtual function for exiting
  virtual void exitProgram(int return_code = 0);

};

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
