/*!@file ModelOptionDefs.C Global list of all known ModelOption instances */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelOptionDefs.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "ModelOptionDefs.H"

//! Descriptions of the option categories:
const char* AllModelOptionCategs[NBMODELOPTIONCATEGS] = {
  "General Simulation Options",
  "Input Reading/Formatting Options",
  "Output Writing/Formatting Options",
  "What to Display/Save Options",
  "Audio-Related Options",
  "Retina-Related ooptions",
  "Channel-Related Options",
  "General Brain/Attention-Related Options",
  "Eye/Head Saccade Controller Options",
  "Video FrameGrabber-Related Options",
  "Beowulf-Related Options",
  "Special-Hardware-Related Options",
  "PsychoDisplay-Related Options",
  "MBARI Result Viewer Related Options",
  "ImageDisplay Related Options",
  "Saliency Map Related Options",
  "Task-Relevance Map Related Options",
  "Attention Guidance Map Related Options",
  "Winner-Take-All Related Options",
  "Visual Buffer Related Options",
  "I-Room Related Options",
  "Vision Cluster Feature Related Options",
  "Inferior Temporal Cortex Options",
  "Gist Estimator Options",
  // caution: make sure this one remains the last one:
  "Option Aliases and Shortcuts (may not always work)"
};

// Format here is:
//
// { MODOPT_TYPE, "name", MOC_CATEG,
//   "description of what option does",
//   "long option name", 'short option name', "valid values", "default value" }
//

// alternatively, for MODOPT_ALIAS option types, format is:
//
// { MODOPT_ALIAS, "", MOC_ALIAS,
//   "description of what alias does",
//   "long option name", 'short option name', "", "list of options" }
//

// NOTE: do not change the default value of any existing option unless
// you really know what you are doing!  Many components will determine
// their default behavior from that default value, so you may break
// lots of executables if you change it.

//! All possible model options
const ModelOptionDef AllModelOptions[] = {

  // #################### General model options:
  // Used by: ModelManager
  { MODOPT_FLAG, "ShowHelpMessage", MOC_GENERAL, // LEAVE THIS AS FIRST IN LIST
    "Show help message and option syntax",
    "help", 'h', "", "true" },
  // Used by: ModelManager
  { MODOPT_FLAG, "DebugMode", MOC_GENERAL,
    "Use debug mode, which, in particular, will increase the verbosity of "
    "the log messages and printout all model parameter values just before "
    "the model starts",
    "debug", 'd', "", "false" },
  // Used by: ModelManager
  { MODOPT_FLAG, "UsingFPE", MOC_GENERAL,
    "Use floating-point exceptions, which will abort execution if overflow, "
    "underflow, invalid or division by zero are encountered",
    "use-fpe", '\0', "", "true" },
  // Used by: ModelManager
  { MODOPT_FLAG, "TestMode", MOC_GENERAL,
    "Use test mode, which, in particular, will turn off randomness (like "
    "--nouse-random), reset random numbers to a reproducible pseudo "
    "sequence, and turn off most displays. This is mostly useful for "
    "execution of the test suite, or any situation where you need "
    "deterministic, reproducible behavior",
    "test-mode", 'Z', "", "false" },
  // Used by: ModelManager
  { MODOPT_STRING, "LoadConfigFile", MOC_GENERAL,
    "Load configuration from file",
    "load-config-from", '\0', "<file.pmap>", "" },
  // Used by: ModelManager
  { MODOPT_STRING, "SaveConfigFile", MOC_GENERAL,
    "Save configuration to file",
    "save-config-to", '\0', "<file.pmap>", "" },
  // Used by: ModelManager
  { MODOPT_STRING, "TextLogFile", MOC_GENERAL,
    "Save text log messages to file",
    "textlog", '\0', "<file.log>", "" },
  // Used by: Brain, FrictionSaccadeController, various
  { MODOPT_DOUBLE, "SimulationTimeStep", MOC_GENERAL,
    "Simulation time step (in seconds) to use for various difference "
    "equations in the model",
    "timestep", '\0', "<double>", "0.0001" },

  // #################### Beowulf options
  // Used by: Beowulf
  { MODOPT_STRING, "BeowulfSlaveNames", MOC_BEOWULF,
    "Comma-separated list of names of slaves to use, in the hostname:port "
    "format (default port will be used if not specified), or single absolute "
    "path to a text file that contains the node names, one per line.",
    "beowulf-slaves", '\0', "<host:port,host:port,...|/path/to/nodelist.txt>",
    "/etc/nodes" },
  // Used by: Beowulf
  { MODOPT_FLAG, "BeowulfMaster", MOC_BEOWULF,
    "Flags this as the master node",
    "beowulf-master", '\0', "", "false" },
  // Used by: TCPcommunicator
  { MODOPT_INADDR, "TCPcommunicatorIPaddr", MOC_BEOWULF,
    "Our IP address to use for Beowulf communications (useful if we have "
    "several ethernet interfaces), or 0.0.0.0 to determine it automatically "
    "from our hostname",
    "ip-addr", '\0', "<int>.<int>.<int>.<int>", "0.0.0.0" },
  // Used by: TCPcommunicator
  { MODOPT_INT, "TCPcommunicatorInQlen", MOC_BEOWULF,
    "Queue length for incoming messages, or zero for unlimited",
    "tcp-inqlen", '\0', "<int>", "100" },
  // Used by: TCPcommunicator
  { MODOPT_INT, "TCPcommunicatorOuQlen", MOC_BEOWULF,
    "Queue length for outgoing messages, or zero for unlimited",
    "tcp-outqlen", '\0', "<int>", "100" },
  // Used by: TCPcommunicator
  { MODOPT_FLAG, "TCPcommunicatorInDropLast", MOC_BEOWULF,
    "Message dropping policy when incoming queue is full; if true, the "
    "most recent (last received) message will be dropped, otherwise, the "
    "least recent will be dropped",
    "tcp-indroplast", '\0', "", "false" },
  // Used by: TCPcommunicator
  { MODOPT_FLAG, "TCPcommunicatorOuDropLast", MOC_BEOWULF,
    "Message dropping policy when outgoing queue is full; if true, the "
    "most recent (last queued) message will be dropped, otherwise, the "
    "least recent will be dropped",
    "tcp-outdroplast", '\0', "", "false" },
  // Used by: TCPcommunicator
  { MODOPT_FLAG, "TCPcommunicatorDisableShm", MOC_BEOWULF,
    "Disable the use of shared memory for faster transfers between two "
    "Beowulf nodes running on the same physical machine",
    "tcp-disable-shm", '\0', "", "false" },
  // Used by: TCPcommunicator
  { MODOPT_SHORT, "SockServPort", MOC_BEOWULF,
    "Port on which to listen for incoming Beowulf connections, "
    "or 0 to determine it from /etc/services",
    "ip-port", '\0', "<int>", "0" },
  // Used by: Beowulf
  { MODOPT_INT, "BeowulfSelfQlen", MOC_BEOWULF,
    "Queue length for self-addressed messages, or zero for unlimited",
    "tcp-selfqlen", '\0', "<int>", "100" },
  // Used by: Beowulf
  { MODOPT_FLAG, "BeowulfSelfDropLast", MOC_BEOWULF,
    "Message dropping policy when self-addressed message queue is full; "
    "if true, the most recent (last received) message will be dropped, "
    "otherwise, the least recent will be dropped",
    "tcp-selfdroplast", '\0', "", "false" },
  // Used by: SingleChannelBeoServer
  { MODOPT_FLAG, "SingleChannelBeoServerQuickMode", MOC_BEOWULF,
    "Use quick mode where only the output map is sent back for each"
    "channel; this will break any program that attempts to access other"
    "channel results, such as the pyramid queue, submaps or clipPyramid",
    "scbserver-quickmode", '\0', "", "true" },

  // #################### Video framegrabbing options
  // Used by: FrameGrabberConfigurator
  { MODOPT_STRING, "FrameGrabberType", MOC_FRAMEGRAB,
    "Framegrabber type name",
    "fg-type", '\0', "<None|V4L|1394>", "None" },
  // Used by: FrameGrabber
  { MODOPT_STRING, "FrameGrabberDevice", MOC_FRAMEGRAB,
    "Framegrabber device name",
    "framegrabber-dev", '\0', "<file>", "/dev/video0" },
  // Used by: FrameGrabber
  { MODOPT_INT, "FrameGrabberChannel", MOC_FRAMEGRAB,
    "Framegrabber input channel (typically, 0=TV, 1=Composite)",
    "framegrabber-chan", '\0', "<int>", "1" },
  // Used by: FrameGrabber
  { MODOPT_INT, "FrameGrabberSubChan", MOC_FRAMEGRAB,
    "Framegrabber input sub-channel",
    "framegrabber-subchan", '\0', "<int>", "0" },
  // Used by: FrameGrabber
  { MODOPT_DIMS, "FrameGrabberDims", MOC_FRAMEGRAB,
    "Framegrabber grab image dimensions",
    "framegrabber-dims", '\0', "<width>x<height>", "320x240" },
  // Used by: FrameGrabber
  { MODOPT_GRABMODE, "FrameGrabberMode", MOC_FRAMEGRAB,
    "Framegrabber grab mode",
    "framegrabber-mode", '\0', "<GREY|RAW|RGB555|RGB565|RGB24|RGB32|YUYV|"
    "UYVY|YUV444|YUV422|YUV411|YUV420|YUV410|YUV444P|YUV422P|YUV411P|YUV420P|"
    "YUV410P>", "YUV422" },
  // Used by: FrameGrabber
  { MODOPT_FLAG, "FrameGrabberByteSwap", MOC_FRAMEGRAB,
    "Framegrabber use byte-swapping on grabbed data",
    "framegrabber-bswap", '\0', "", "false" },
  // Used by: FrameGrabber
  { MODOPT_FLOAT, "FrameGrabberFPS", MOC_FRAMEGRAB,
    "Framegrabber grabbing frames per seconds",
    "framegrabber-fps", '\0', "<float>", "30.0" },
  // Used by: FrameGrabber
  { MODOPT_FLOAT, "FrameGrabberNbuf", MOC_FRAMEGRAB,
    "Framegrabber number fo grab buffers",
    "framegrabber-nbuf", '\0', "<int>", "2" },
  { MODOPT_INT, "FrameGrabberBrightness", MOC_FRAMEGRAB,
    "Brightness of the grabbed picture",
    "framegrabber-brightness", '\0', "<int>", "32768" },
  { MODOPT_INT, "FrameGrabberColour", MOC_FRAMEGRAB,
    "Brightness of the grabbed picture",
    "framegrabber-colour", '\0', "<int>", "32768" },
  { MODOPT_INT, "FrameGrabberHue", MOC_FRAMEGRAB,
    "Hue of the grabbed picture",
    "framegrabber-hue", '\0', "<int>", "32768" },
  { MODOPT_INT, "FrameGrabberSaturation", MOC_FRAMEGRAB,
    "Saturation of the grabbed picture",
    "framegrabber-saturation", '\0', "<int>", "90" },
  { MODOPT_INT, "FrameGrabberContrast", MOC_FRAMEGRAB,
    "Contrast of the grabbed picture",
    "framegrabber-contrast", '\0', "<int>", "32768" },
  { MODOPT_INT, "FrameGrabberWhiteness", MOC_FRAMEGRAB,
    "Whiteness of the grabbed picture",
    "framegrabber-whiteness", '\0', "<int>", "32768" },
  { MODOPT_INT, "FrameGrabberExposure", MOC_FRAMEGRAB,
    "Exposure of the frame grabber",
    "framegrabber-exposure", '\0', "<int>", "511" },
  { MODOPT_INT, "FrameGrabberSharpness", MOC_FRAMEGRAB,
    "Sharpness of the grabbed picture",
    "framegrabber-sharpness", '\0', "<int>", "80" },
  { MODOPT_INT, "FrameGrabberWhiteBalBU", MOC_FRAMEGRAB,
    "B/U white balance of the grabbed picture",
    "framegrabber-whitebalBU", '\0', "<int>", "95" },
  { MODOPT_INT, "FrameGrabberWhiteBalRV", MOC_FRAMEGRAB,
    "R/V white balance of the grabbed picture",
    "framegrabber-whitebalRV", '\0', "<int>", "87" },
  { MODOPT_INT, "FrameGrabberGamma", MOC_FRAMEGRAB,
    "Gamma correction of the frame grabber",
    "framegrabber-gamma", '\0', "<int>", "1" },
  { MODOPT_INT, "FrameGrabberShutter", MOC_FRAMEGRAB,
    "Shutter speed of the frame grabber",
    "framegrabber-shutter", '\0', "<int>", "6" },
  { MODOPT_INT, "FrameGrabberGain", MOC_FRAMEGRAB,
    "Gain of the frame grabber",
    "framegrabber-gain", '\0', "<int>", "87" },

  // #################### Special hardware options
  // Used by: VCC4
  { MODOPT_STRING, "VCC4serialDevice", MOC_SPECHARD,
    "the serial device for the VCC4 camera",
    "vcc4-serial-device", '\0', "<dev>", "/dev/usb/tts/0" },
  // Used by: VCC4
  { MODOPT_INT, "VCC4unitNo", MOC_SPECHARD,
    "VCC4 unit number",
    "vcc4-unit-no", '\0', "<0..9>", "0" },
  // Used by: VCC4
  { MODOPT_FLOAT, "VCC4pulseRatio", MOC_SPECHARD,
    "VCC4 pulse ratio, or 0.0 to query it from the VCC4 unit",
    "vcc4-pulse-ratio", '\0', "<float>", "0.0" },
  // Used by: VCC4
  { MODOPT_FLOAT, "VCC4maxAngle", MOC_SPECHARD,
    "VCC4 max angle, or 0.0 to query it from the VCC4 unit",
    "vcc4-max-angle", '\0', "<float>", "0.0" },

  // #################### Audio recording options
  // Used by: AudioMixer
  { MODOPT_STRING, "AudioMixerDevice", MOC_AUDIO,
    "Audio mixer device name",
    "audio-mix-dev", '\0', "<file>", "/dev/mixer" },
  // Used by: AudioMixer
  { MODOPT_FLAG, "AudioMixerLineIn", MOC_AUDIO,
    "Audio mixer Line input",
    "audio-mix-linein", '\0', "", "true" },
  // Used by: AudioMixer
  { MODOPT_FLAG, "AudioMixerCdIn", MOC_AUDIO,
    "Audio mixer CD input",
    "audio-mix-cdin", '\0', "", "false" },
  // Used by: AudioMixer
  { MODOPT_FLAG, "AudioMixerMicIn", MOC_AUDIO,
    "Audio mixer Mic input",
    "audio-mix-micin", '\0', "", "false" },

  // Used by: AudioGrabber
  { MODOPT_STRING, "AudioGrabberDevice", MOC_AUDIO,
    "Audio grabber device name",
    "audio-grab-dev", '\0', "<file>", "/dev/dsp" },
  // Used by: AudioGrabber
  { MODOPT_UINT, "AudioGrabberBits", MOC_AUDIO,
    "Audio grabber bits per sample",
    "audio-grab-bits", '\0', "<8|16>", "8" },
  // Used by: AudioGrabber
  { MODOPT_UINT, "AudioGrabberFreq", MOC_AUDIO,
    "Audio grabber sampling frequency (Hz)",
    "audio-grab-freq", '\0', "<int>", "44100" },
  // Used by: AudioGrabber
  { MODOPT_UINT, "AudioGrabberBufSamples", MOC_AUDIO,
    "Audio grabber buffer length (in samples)",
    "audio-grab-bufsamp", '\0', "<int>", "256" },
  // Used by: AudioGrabber
  { MODOPT_BOOL, "AudioGrabberStereo", MOC_AUDIO,
    "Audio grabber stereo recording",
    "audio-grab-stereo", '\0', "<true|false>", "true" },

  // #################### FrameSeries handling options:
  // Used by: InputFrameSeries
  { MODOPT_FRAMERANGE, "InputFrameRange", MOC_INPUT,
    "Input frame range and delay in ms",
    "input-frames", 'M', "<first>-<last>@<delay>", "0-0@0.0" },
  // Used by: InputFrameSeries, SaccadeControllers
  { MODOPT_DIMS, "InputFrameDims", MOC_INPUT,
    "Rescale input frames to fixed dims, or 0x0 for no rescaling",
    "rescale-input", '\0', "<width>x<height>", "0x0" },
  // Used by: InputFrameSeries
  { MODOPT_FLAG, "InputPreserveAspect", MOC_INPUT,
    "Preserve input frame aspect ratio if rescaling to fixed dims",
    "preserve-input-aspect", '\0', "", "false" },
  // Used by: InputFrameSeries
  { MODOPT_FLAG, "InputShowFrames", MOC_INPUT,
    "Show input frames",
    "display-input-frames", '\0', "", "false" },
  // Used by: InputFrameSeries
  { MODOPT_RASTERFFORMAT, "InputRasterFileFormat", MOC_INPUT,
    "Input file format",
    "input-format", '\0', "<PNM|PNG|RAWIMAGE|Auto>", "PNM" },

  // Used by: OutputFrameSeries
  { MODOPT_FRAMERANGE, "OutputFrameRange", MOC_OUTPUT,
    "Output frame range and delay in ms",
    "output-frames", 'R', "<first>-<last>@<delay>", "0-0@0.0" },
  // Used by: OutputFrameSeries
  { MODOPT_DIMS, "OutputFrameDims", MOC_OUTPUT,
    "Rescale output frames to fixed dims, or 0x0 for no rescaling",
    "rescale-output", '\0', "<width>x<height>", "0x0" },
  // Used by: OutputFrameSeries
  { MODOPT_FLAG, "OutputPreserveAspect", MOC_OUTPUT,
    "Preserve output frame aspect ratio if rescaling to fixed dims",
    "preserve-output-aspect", '\0', "", "false" },
  // Used by: OutputFrameSeries
  { MODOPT_FLAG, "OutputShowFrames", MOC_OUTPUT,
    "Show output frames",
    "display-output-frames", '\0', "", "true" },
  // Used by: OutputFrameSeries
  { MODOPT_RASTERFFORMAT, "OutputRasterFileFormat", MOC_OUTPUT,
    "Output file format",
    "output-format", '\0', "<PNM|PNG|RAWIMAGE|Auto>", "PNM" },
  // Used by: OutputFrameSeries
  { MODOPT_STRING, "OutputFramingImageName", MOC_OUTPUT,
    "Filename of an image to used as a background into which output "
    "frames will be embedded before processing. This image must be larger "
    "than the output frames to be processed.",
    "output-framing", '\0', "<imagefile>", "" },
  // Used by: OutputFrameSeries
  { MODOPT_POINT2D, "OutputFramingImagePos", MOC_OUTPUT,
    "Position of the top-left corner of the output frames once embedded "
    "into the larger background image specified by --output-framing, if any.",
    "output-framing-pos", '\0', "<i,j>", "0,0" },

  // Used by: OutputMPEGStream
  { MODOPT_STRING, "OutputMPEGStreamCodec", MOC_OUTPUT,
    "Type of video output codec to use (use value 'List' to see list of "
    "available codecs on your system)",
    "output-codec", '\0', "<List|name>", "mpeg1video" },
  // Used by: InputMPEGStream
  { MODOPT_STRING, "InputMPEGStreamCodec", MOC_INPUT,
    "Type of video input codec to use (use value 'List' to see list of "
    "available codecs on your system, or 'Auto' to automatically determine "
    "which codec to use)",
    "input-codec", '\0', "<Auto|List|name>", "Auto" },

  // #################### Retina options:
  // Used by: RetinaConfigurator
  { MODOPT_STRING, "RetinaType", MOC_RETINA,
    "Type of Retina to use. 'None' for a simple pass-through of input "
    "frames, 'Std' for the standard retina that can foveate, shift inputs "
    "to eye position, embed inputs within a larger framing image, foveate "
    "inputs, etc.",
    "retina-type", '\0', "<None|Std>", "Std" },
  // Used by: RetinaStd
  { MODOPT_STRING, "InputFramingImageName", MOC_RETINA,
    "Filename of an image to used as a background into which input "
    "frames will be embedded before processing. This image must be larger "
    "than the input frames to be processed.",
    "input-framing", '\0', "<imagefile>", "" },
  // Used by: RetinaStd
  { MODOPT_POINT2D, "InputFramingImagePos", MOC_RETINA,
    "Position of the top-left corner of the input frames once embedded "
    "into the larger background image specified by --input-framing, if any.",
    "input-framing-pos", '\0', "<i,j>", "0,0" },
  // Used by: Retina
  { MODOPT_FLAG, "FoveateInput", MOC_RETINA,
    "Foveate input frames according to current eye position",
    "foveate-input", '\0', "", "false" },
  // Used by: Retina
  { MODOPT_INT, "FoveateInputDepth", MOC_RETINA,
    "Depth of pyramid to use to foveate input frames",
    "foveate-input-depth", '\0', "<int>", "5" },
  // Used by: Retina
  { MODOPT_FLAG, "ShiftInputToEye", MOC_RETINA,
    "Shift input frames so that they are centered at current eye position",
    "shift-input", '\0', "", "false" },
  // Used by: Retina
  { MODOPT_RGB, "ShiftInputToEyeBGcol", MOC_RETINA,
    "Background color to use when shifting inputs using --shift-input",
    "shift-input-bgcol", '\0', "<r,g,b>", "64,64,64" },
  // Used by: Retina
  { MODOPT_FLAG, "InputFOV", MOC_RETINA,
    "If non-empty, centrally crop the input images to the given "
    "field-of-view dimensions",
    "input-fov", '\0', "<w>x<h>", "0x0" },
  // Used by: Retina
  { MODOPT_FLAG, "RetinaSaveInput", MOC_RETINA,
    "Save retinal input images, with prefix RETIN-",
    "save-retina-input", '\0', "", "false" },
  // Used by: Retina
  { MODOPT_FLAG, "RetinaSaveOutput", MOC_RETINA,
    "Save retina output (possibly including foveation, shifting, "
    "embedding, etc), with prefix RETOUT-",
    "save-retina-output", '\0', "", "false" },
  // Used by: RetinaStd
  { MODOPT_FLAG, "RetinaStdSavePyramid", MOC_RETINA,
    "Save pyramid used for retinal foveation, with prefix RET<level>-",
    "save-retina-pyr", '\0', "", "false" },

  // #################### Channel options:
  // Used by: ChannelBase (and derivatives)
  { MODOPT_FLOAT, "ChannelOutputRangeMax", MOC_CHANNEL,
    "Max of the channel's output range",
    "chanout-max", '\0', "<float>", "10.0" },
  // Used by: ChannelBase (and derivatives)
  { MODOPT_FLOAT, "ChannelOutputRangeMin", MOC_CHANNEL,
    "Min of the channel's output range",
    "chanout-min", '\0', "<float>", "0.0" },
  // Used by: VisualCortex (and derivatives)
  { MODOPT_FLOAT, "VisualCortexOutputFactor", MOC_CHANNEL,
    "Factor applied to outputs of VisualCortex to scale them to Amps of "
    "synaptic input currents to saliency map",
    "vcx-outfac", '\0', "<float>", "1.0e-9" },
  // Used by: DirectFeedChannel
  { MODOPT_UINT, "MapLevel", MOC_CHANNEL,
    "The level of the saliency map in Channels that don't need LevelSpec",
    "maplevel", '\0', "<uint>", "4" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_LEVELSPEC, "LevelSpec", MOC_CHANNEL,
    "LevelSpec to use in channels",
    "levelspec", '\0', "<cmin,cmax,smin,smax,sml>", "2,4,3,4,4" },
  // Used by: SingleChannel, ComplexChannel (and derivatives), VisualCortex
  { MODOPT_MAXNORM, "MaxNormType", MOC_CHANNEL,
    "Type of MaxNormalization to use",
    "maxnorm-type", '\0', "<None|Maxnorm|Fancy|FancyFast|FancyOne|"
    "FancyLandmark|Landmark|FancyWeak|Ignore|Surprise>", "Fancy" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_INT, "SingleChannelQueueLen", MOC_CHANNEL,
    "Queue length for channels",
    "qlen", '\0', "<int>", "1" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_DOUBLE, "SingleChannelTimeDecay", MOC_CHANNEL,
    "Time decay for channel queues",
    "qtime-decay", '\0', "<float>", "20.0" },
  // Used by: OrientationChannel
  { MODOPT_UINT, "NumOrientations", MOC_CHANNEL,
    "Number of oriented channels",
    "num-orient", '\0', "<int>", "4" },
  // Used by: OrientationChannel
  { MODOPT_STRING, "OriInteraction", MOC_CHANNEL,
    "This describes the way in which the different orientations "
    "within the OrientationChannel interact with each other. "
    "'None' is for no interaction (default); for 'SubtractMean', "
    "the mean of all orientation pyramids is subtracted from each "
    "orientation pyramid. You can determine your own interaction by "
    "specifying a vector with interaction coefficients: 'c1,...,cn'"
    " where n is the number of orientations. c1 is the coeffiecient "
    "for an orientation itself, c2 for the one with the next higher "
    "angle and so on. For instance, for n=4, 'None' is the same as "
    "'1,0,0,0', and 'SubtractMean' is the same as "
    "'0.75,-0.25,-0.25,-0.25'",
    "ori-interaction", '\0', "<None|SubtractMean|c1,...,cn>",
    "None" },
  // Used by: GaborChannel
  { MODOPT_ORICOMPTYPE, "OrientComputeType", MOC_CHANNEL,
    "Type of computation used to compute orientations",
    "oricomp-type", '\0', "<Steerable|Gabor|GaborEnergyNorm>",
    "Steerable" },
  // Used by: GaborChannel
  { MODOPT_DOUBLE, "GaborChannelIntensity", MOC_CHANNEL,
    "Intensity coefficient for Gabor channel",
    "gabor-intens", '\0', "<double>", "10.0" },
  // Used by: DirectionChannel
  { MODOPT_FLAG, "DirectionChannelTakeSqrt", MOC_CHANNEL,
    "Take square root of our Reichardt output if true",
    "direction-sqrt", '\0', "", "false" },
  // Used by: DirectionChannel
  { MODOPT_FLOAT, "DirectionChannelLowThresh", MOC_CHANNEL,
    "Low threshold to allpy to eliminate small motion responses",
    "direction-lowthresh", '\0', "<float>", "3.0" },
  // Used by: MultiColorBandChannel
  { MODOPT_UINT, "NumColorBands", MOC_CHANNEL,
    "Number of hue-band channels",
    "num-colorband", '\0', "<int>", "3" },
  // Used by: MultiColorBandChannel
  { MODOPT_UINT, "NumSatBands", MOC_CHANNEL,
    "Number of saturation-band channels",
    "num-satband", '\0', "<int>", "3" },
  // Used by: MultiColorBandChannel
  { MODOPT_UINT, "HueBandWidth", MOC_CHANNEL,
    "Width of each hue band",
    "sigma-hueband", '\0', "<float>", "30" },
  // Used by: MultiColorBandChannel
  { MODOPT_UINT, "SatBandWidth", MOC_CHANNEL,
    "Width of each saturation band",
    "sigma-satband", '\0', "<float>", "0.3" },
  // Used by: MotionChannel
  { MODOPT_UINT, "NumDirections", MOC_CHANNEL,
    "Number of direction-selective motion channels",
    "num-directions", '\0', "<int>", "4" },
  // Used by: StereoChannel
  { MODOPT_UINT, "NumTheta", MOC_CHANNEL,
    "Number of oriented disparity channels",
    "num-theta", '\0', "<int>", "1" },
  // Used by: SingleChannel (and derivatives)
  // Used by: LJunctionChannel
  { MODOPT_UINT, "NumLOrients", MOC_CHANNEL,
    "Number of oriented l-junction channels",
    "num-l-orients", '\0', "<int>", "4" },
   // Used by: TJunctionChannel
  { MODOPT_UINT, "NumTOrients", MOC_CHANNEL,
    "Number of oriented t-junction channels",
    "num-t-orients", '\0', "<int>", "4" },
   // Used by: XJunctionChannel
  { MODOPT_UINT, "NumXOrients", MOC_CHANNEL,
    "Number of oriented x-junction channels",
    "num-x-orients", '\0', "<int>", "1" },
   // Used by: EndPointChannel
  { MODOPT_UINT, "NumEOrients", MOC_CHANNEL,
    "Number of oriented end-point channels",
    "num-e-orients", '\0', "<int>", "4" },
  // Used by: LJunctionChannel
  { MODOPT_FLAG, "LFullImplementation", MOC_CHANNEL,
    "Use full implementation of the l-junction channel",
    "use-full-l", '\0', "", "false" },
  // Used by: TJunctionChannel
  { MODOPT_FLAG, "TFullImplementation", MOC_CHANNEL,
    "Use full implementation of the t-junction channel",
    "use-full-t", '\0', "", "false" },
  // Used by: XJunctionChannel
  { MODOPT_FLAG, "XFullImplementation", MOC_CHANNEL,
    "Use full implementation of the x-junction channel",
    "use-full-x", '\0', "", "false" },
  // Used by: EndPointChannel
  { MODOPT_FLAG, "EFullImplementation", MOC_CHANNEL,
    "Use full implementation of the end-point channel",
    "use-full-e", '\0', "", "false" },
   // Used by: SingleChannel (and derivatives)
  { MODOPT_FLAG, "SingleChannelUseSplitCS", MOC_CHANNEL,
    "Use split positive/negative center-surround computations",
    "use-split-cs", '\0', "", "false" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_FLAG, "SingleChannelSaveRawMaps", MOC_CHANNEL,
    "Save raw input maps (pyramid levels) from all single channels",
    "save-raw-maps", '\0', "", "false" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_FLAG, "SingleChannelSaveFeatureMaps", MOC_CHANNEL,
    "Save center-surround feature maps from all single channels",
    "save-feature-maps", '\0', "", "false" },
  // Used by: SingleChannel (and derivatives)
  { MODOPT_FLAG, "SingleChannelSaveOutputMap", MOC_CHANNEL,
    "Save combined center-surround output maps from all single channels",
    "save-featurecomb-maps", '\0', "", "false" },
  // Used by: ComplexChannel (and derivatives)
  { MODOPT_FLAG, "ComplexChannelSaveOutputMap", MOC_CHANNEL,
    "Save conspicuity maps from all complex channels",
    "save-conspic-maps", '\0', "", "false" },
  // Used by: ColorChannel
  { MODOPT_COLCOMPTYPE, "ColorComputeType", MOC_CHANNEL,
    "Type of computation used to compute RG and BY color opponencies",
    "color-comp-type", '\0', "<Standard|Simple>",
    "Standard" },

  // Used by: VisualCortex (and derivatives)
  { MODOPT_FLAG, "VisualCortexSaveOutput", MOC_CHANNEL,
    "Save output of visual cortex (=input to the saliency map)",
    "save-vcx-output", '\0', "", "false" },

  // Used by: SingleChannelSurprise
  { MODOPT_UINT, "SingleChannelSurpriseSQlen", MOC_CHANNEL,
    "Number of feed-through surprise models",
    "surprise-qlen", '\0', "<uint>", "5" },
  // Used by: SingleChannelSurprise
  { MODOPT_DOUBLE, "SingleChannelSurpriseUpdFac", MOC_CHANNEL,
    "Local surprise update factor",
    "surprise-updfac", '\0', "<double>", "0.7" },
  // Used by: SingleChannelSurprise
  { MODOPT_DOUBLE, "SingleChannelSurpriseNeighUpdFac", MOC_CHANNEL,
    "Neighborhood surprise update factor, or 0.0 to use the same value as "
    "the local surprise update factor",
    "surprise-neighupdfac", '\0', "<double>", "0.7" },
  // Used by: SingleChannelSurprise
  { MODOPT_POINT2D, "SingleChannelSurpriseProbe", MOC_CHANNEL,
    "Location of a virtual electrode probe; various model and surprise "
    "readings from this location will be printed out if not (-1, -1)",
    "surprise-probe", '\0', "<x>,<y>", "-1,-1" },
  // Used by: SingleChannelSurprise
  { MODOPT_DOUBLE, "SingleChannelSurpriseSLfac", MOC_CHANNEL,
    "Factor for contribution of local temporal surprise to total surprise",
    "surprise-slfac", '\0', "<double>", "1.0" },
  // Used by: SingleChannelSurprise
  { MODOPT_DOUBLE, "SingleChannelSurpriseSSfac", MOC_CHANNEL,
    "Factor for contribution of spatial surprise to total surprise",
    "surprise-ssfac", '\0', "<double>", "0.1" },

  // Used by: DummyChannel
  { MODOPT_DOUBLE, "DummyChannelFactor", MOC_CHANNEL,
    "Factor to multiply the input by, or 0.0 for nothing",
    "dummychannel-factor", '\0', "<float>", "0.0" },

  // Used by: TcorrChannel
  { MODOPT_INT, "TcorrChannelFrameLag", MOC_CHANNEL,
    "Frame lag with which to compute the temporal correlations",
    "tcorrchannel-framelag", '\0', "<int>", "1" },

  // Used by: ScorrChannel
  { MODOPT_INT, "ScorrChannelRadius", MOC_CHANNEL,
    "Radius (in saliency map pixels) to compute the spatial correlations",
    "scorrchannel-radius", '\0', "<int>", "1" },

  // Used by: VisualCortexEyeMvt
  { MODOPT_STRING, "VCEMeyeFnames", MOC_CHANNEL,
    "Comma-separated list of human eye movement file names",
    "vcem-eyefnames", '\0', "<name1,name2,...>", "" },
  // Used by: VisualCortexEyeMvt
  { MODOPT_FLOAT, "VCEMsigma", MOC_CHANNEL,
    "Sigma of gaussian blob plotted at each human eye position, in pixels "
    "at the scale of the saliency map, or 0.0 to plot a single pixel",
    "vcem-sigma", '\0', "<float>", "3.0" },
  // Used by: VisualCortexEyeMvt
  { MODOPT_FLOAT, "VCEMforgetFac", MOC_CHANNEL,
    "Forgetting factor to be applied at every evolve of VisualCortexEyeMvt",
    "vcem-forgetfac", '\0', "<float>", "0.9995" },
  // Used by: VisualCortexEyeMvt
  { MODOPT_INT, "VCEMdelay", MOC_CHANNEL,
    "Human-to-human delay, in eye tracker samples",
    "vcem-delay", '\0', "<int>", "24" },
  // Used by: VisualCortexEyeMvt
  { MODOPT_FLAG, "VCEMuseMax", MOC_CHANNEL,
    "Use max to comabine across humans, otherwise sum",
    "vcem-usemax", '\0', "", "false" },
  // Used by: VisualCortexEyeMvt
  { MODOPT_FLAG, "VCEMsaccadeOnly", MOC_CHANNEL,
    "Only plot endpoint locations of saccades instead of every sample",
    "vcem-saconly", '\0', "", "false" },


  // #################### SimulationViewer options:
  // Used by: SimulationViewerConfigurator, StdBrain, etc
  { MODOPT_STRING, "SimulationViewerType", MOC_DISPLAY,
    "Type of SimulationViewer to use. 'Std' is the standard 2D "
    "viewer. 'Compress' is a simple multi-foveated blurring viewer, in "
    "which a selection of most salient locations will be represented "
    "crisply, while the rest of the image will be increasingly blurred as "
    "we move away from those hot spots. 'EyeMvt' is a viewer for comparison "
    "between saliency maps and human eye movements nd 'EyeMvt' also plus "
    "it adds a concept of visual memory buffer. 'EyeSim' simulates an "
    "eye-tracker recording from the model.",
    "sv-type", '\0', "<None|Std|Compress|EyeMvt|EyeMvt2|EyeSim>", "Std" },
   // Used by: SimulationViewer
  { MODOPT_FLAG, "SVsaveTraj", MOC_DISPLAY,
    "Save attention/eye/head trajectories",
    "save-trajectory", 'T', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVsaveXcombo", MOC_DISPLAY,
    "Show combination trajec (left) + salmap (right)",
    "save-x-combo", 'X', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVsaveYcombo", MOC_DISPLAY,
    "Show combination trajec (top) + salmap (bottom)",
    "save-y-combo", 'Y', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVsaveTRMXcombo", MOC_DISPLAY,
    "Show combination trajec (left) + salmap (middle)+ "
    " task-relevance-map (right)",
    "save-trm-x-combo", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVsaveTRMYcombo", MOC_DISPLAY,
    "Show combination trajec (top) + salmap (middle) + "
    " task-relevance-map (bottom) ",
    "save-trm-y-combo", '\0', "", "false" },
   // Used by: SimulationViewer
  { MODOPT_FLAG, "SVwarp3D", MOC_DISPLAY,
    "Show color image warped onto 3D salmap",
    "warp-salmap-3d", '3', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVmegaCombo", MOC_DISPLAY,
    "Show trajectory, saliency and channels as a combo",
    "mega-combo", 'K', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVmotionSixPack", MOC_DISPLAY,
    "show the outputs of the four DirectionChannels, of the "
    "MotionChannel and the original image (aka. >Sixpack<)",
    "show-motion-sixpack", '6', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_DIMS, "SVcropFOA", MOC_DISPLAY,
    "Crop image to <w>x<h> around FOA location",
    "crop-foa", 'v', "<w>x<h>", "0x0" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVxwindow", MOC_DISPLAY,
    "Show results in an XWindow instead of saving to disk",
    "show-xwindow", 'x', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVfoveateTraj", MOC_DISPLAY,
    "foveate trajectory (cumulative if --display-additive)",
    "foveate-traj", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayFOA", MOC_DISPLAY,
    "Display focus-of-attention as a circle",
    "display-foa", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayPatch", MOC_DISPLAY,
    "Display small filled square at attended location",
    "display-patch", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayFOALinks", MOC_DISPLAY,
    "Display attentional trajectory using red arrows",
    "display-traj", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayEye", MOC_DISPLAY,
    "Display small hollow square at eye position",
    "display-eye", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayEyeLinks", MOC_DISPLAY,
    "Display eye trajectory using red lines",
    "display-eye-traj", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayHead", MOC_DISPLAY,
    "Display larger hollow square at head position",
    "display-head", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayHeadLinks", MOC_DISPLAY,
    "Display head trajectory using red lines",
    "display-head-traj", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayAdditive", MOC_DISPLAY,
    "Display things additively",
    "display-additive", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayTime", MOC_DISPLAY,
    "Display internal simulation time",
    "display-time", '\0', "", "true" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayHighlights", MOC_DISPLAY,
    "Display highlight at focus-of-attention",
    "display-highlight", '\0', "", "false" },
  // Used by: SimulationViewer
  { MODOPT_FLAG, "SVdisplayBoring", MOC_DISPLAY,
    "Display attention shifts to boring targets in green",
    "display-boring", '\0', "", "true" },
  // Used by: SimulationViewerStd
  { MODOPT_FLAG, "SVuseLargerDrawings", MOC_DISPLAY,
    "Use larger drawings for FOA, FOV, and head markers than default",
    "display-larger-markers", '\0', "", "false" },
  
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveXcombo", MOC_DISPLAY,
    "Show combination multifoveated image (left) + foveation mask (right)",
    "save-compress-xcombo", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveYcombo", MOC_DISPLAY,
    "Show combination multifoveated image (top) + foveation mask (bottom)",
    "save-compress-ycombo", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveXcomboOrig", MOC_DISPLAY,
    "Show combination of original (left) and multifoveated image (right)",
    "save-compress-xcomboorig", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveYcomboOrig", MOC_DISPLAY,
    "Show combination of original (top) and multifoveated image (bottom)",
    "save-compress-ycomboorig", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveMegaCombo", MOC_DISPLAY,
    "Show combination of original (top-left), multifoveated image "
    "(top-right), raw saliency map (bottom-left) and blurring mask "
    "(bottom-right)",
    "save-compress-megacombo", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPsaveEyeCombo", MOC_DISPLAY,
    "Show combination of original (left), foveated image + human eye "
    "(middle) and raw saliency map (right)",
    "save-compress-eyecombo", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_INT, "SVCOMPnumFoveas", MOC_DISPLAY,
    "Number of foveas to use for multi-foveated image blurring. A value of "
    "zero is valid and means that the blur will be continuously computed "
    "from the saliency map, rather than from a discrete number of foveas",
    "num-foveas", '\0', "<int>", "3" },
  // Used by: SimulationViewerCompress
  { MODOPT_INT, "SVCOMPcacheSize", MOC_DISPLAY,
    "Number of frames over which the object masks are averaged to "
    "determine current foveation mask",
    "maskcache-size", '\0', "<int>", "5" },
  // Used by: SimulationViewerCompress
  { MODOPT_INT, "SVCOMPiframePeriod", MOC_DISPLAY,
    "Period of occurence of I-Frames when compressing with MPEG-1. If this "
    "is not 1, the foveation will be allowed to change only on every "
    "I-Frame, but will remain stable on P-frames and B-frames",
    "iframe-period", '\0', "<int>", "1" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPdisplayHumanEye", MOC_DISPLAY,
    "Display human eye positions",
    "display-human-eye", '\0', "<true|false>", "true" },
  // Used by: SimulationViewerCompress
  { MODOPT_FLAG, "SVCOMPuseTRMmax", MOC_DISPLAY,
    "Use TRM-max trick",
    "use-trmmax", '\0', "", "false" },
  // Used by: SimulationViewerCompress
  { MODOPT_STRING, "SVCOMPfoveaSCtype", MOC_DISPLAY,
    "Type of SC to use for the foveas of the compression viewer",
    "compress-sc-type", '\0', "<type>", "Friction" },
  // Used by: SimulationViewerCompress
  { MODOPT_INT, "SVCOMPMultiRetinaDepth", MOC_DISPLAY,
    "Depth of pyramid used for multi-foveation",
    "compress-multiretina-depth", '\0', "<int>", "5" },

  // Used by: SimulationViewerEyeMvt
  { MODOPT_FLAG, "SVEMsaveMegaCombo", MOC_DISPLAY,
    "Show combination of original (top-left), max-saliency-weighted image "
    "(top-right), raw saliency map (bottom-left) and max-saliency mask "
    "(bottom-right)",
    "save-eyemvt-megacombo", '\0', "", "false" },
  // Used by: SimulationViewerEyeMvt
  { MODOPT_INT, "SVEMdelayCacheSize", MOC_DISPLAY,
    "Number of eye movement samples to skip between current time and "
    "time at which we start computing the max saliency over a period of "
    "time, or 0 for no delay",
    "delaycache-size", '\0', "<int>", "0" },
  // Used by: SimulationViewerEyeMvt
  { MODOPT_INT, "SVEMmaxCacheSize", MOC_DISPLAY,
    "Number of eye movement samples over which to compute the max saliency, "
    "or 0 for no cache",
    "maxcache-size", '\0', "<int>", "0" },
  // Used by: SimulationViewerEyeMvt
  { MODOPT_STRING, "SVEMoutFname", MOC_DISPLAY,
    "File name for output data, required",
    "svem-out-fname", '\0', "<file>", "" },
  // Used by: SimulationViewerEyeMvt
  { MODOPT_BOOL, "SVEMsampleAtStart", MOC_DISPLAY,
    "Sample at start of saccade (otherwise at end of saccade)",
    "svem-sampleatstart", '\0', "<true|false>", "true" },

  // Used by: SimulationViewerEyeMvt2
  { MODOPT_DIMS, "SVEMbufDims", MOC_DISPLAY,
    "Dimensions of internal visual buffer",
    "svem-bufdims", '\0', "<w>x<h>", "" },
  // Used by: SimulationViewerEyeMvt2
  { MODOPT_FLAG, "SVEMuseIOR", MOC_DISPLAY,
    "Attempt to guess IOR from human eye movements",
    "svem-use-ior", '\0', "", "true" },

  // Used by: SimulationViewerEyeSim
  { MODOPT_STRING, "SVeyeSimFname", MOC_DISPLAY,
    "File name for output eye movement data (required)",
    "eyesim-fname", '\0', "<file>", "" },
  // Used by: SimulationViewerEyeSim
  { MODOPT_FLOAT, "SVeyeSimPeriod", MOC_DISPLAY,
    "Eye movements sampling period",
    "eyesim-period", '\0', "<float>", "0.0041666666" }, // 240Hz
  // Used by: SimulationViewerEyeSim
  { MODOPT_INT, "SVeyeSimTrash", MOC_DISPLAY,
    "Number of initial eye movement samples to trash (e.g., "
    "corresponding to initial fixation before actual stimulus)",
    "eyesim-trash", '\0', "<int>", "0" },

  // #################### PsychoDisplay options:
  // Used by: PsychoDisplay
  { MODOPT_STRING, "PsychoDisplayFileName", MOC_PSYCHODISP,
    "Filename for experimental data collection trace, or empty to not "
    "save any data",
    "psycho-data", '\0', "<filename>", "psychodata.psy" },
  // Used by: PsychoDisplay
  { MODOPT_DIMS, "PsychoDisplayDims", MOC_PSYCHODISP,
    "PsychoDisplay screen dimensions",
    "psycho-dims", '\0', "<width>x<height>", "640x480" },
  // Used by: PsychoDisplay
  { MODOPT_INT, "PsychoDisplayPriority", MOC_PSYCHODISP,
    "Priority to run at in SCHED_FIFO mode (need to run as root), or 0 for "
    "normal process scheduling and normal priority",
    "psycho-priority", '\0', "<int>", "80" },
  // Used by: PsychoDisplay
  { MODOPT_FLOAT, "PsychoDisplayRefreshUsec", MOC_PSYCHODISP,
    "Desired refresh delay in microseconds. All screen refresh opeations "
    "will enforce that at least 90% of this delay has occurred, but it "
    "could be more depending on the refresh rate of the video mode used",
    "psycho-refresh", '\0', "<float>", "33333.3333" },

  // Used by: DiskDataStream
  { MODOPT_STRING, "DiskDataStreamFileBase", MOC_PSYCHODISP,
    "Filename stem for even-numbered files streamed to disk. If no stem"
    "is given for odd-numbered files, then this stem will be used for"
    "all files",
    "datastream-even-filestem", '\0', "<filename>", "img" },
  // Used by: DiskDataStream
  { MODOPT_STRING, "DiskDataStreamFileBaseOdd", MOC_PSYCHODISP,
    "Filename stem for odd-numbered files streamed to disk. If not given,"
    "then odd-numbered files will have the same stem as the even-numbered"
    "files",
    "datastream-odd-filestem", '\0', "<filename>", "" },
  // Used by: DiskDataStream
  { MODOPT_STRING, "DiskDataStreamFileExt", MOC_PSYCHODISP,
    "Filename extension for files streamed to disk. A '.' must be included"
    "here if desired.",
    "datastream-ext", '\0', "<.extension>", "" },
  // Used by: DiskDataStream
  { MODOPT_BOOL, "DiskDataStreamUseMmap", MOC_PSYCHODISP,
    "Whether to use mmap() instead of write() to stream data to disk. The"
    "default is to use mmap(), which may be significantly faster, but may"
    "be somewhat less portable",
    "datastream-use-mmap", '\0', "<true|false>", "true" },

  // #################### InferoTemporal options:
  // Used by: Brain
  { MODOPT_STRING, "InferoTemporalType", MOC_BRAIN,
    "Type of InferoTemporal to use. 'None' or 'Std'",
    "it-type", '\0', "<None|Std>", "None" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "InitialScaleUp", MOC_ITC,
    "Whether or not to scale up the image before keypoint extraction.",
    "init-scaleup", '\0', "", "yes" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "AttentionObjRecog", MOC_ITC,
    "0=don't use attention for object recognition, 1=mask attended region "
    "before object recognition.",
    "attn-objrecog", '\0', "", "yes" },

  // Used by: InferoTemporal
  { MODOPT_STRING, "ObjectDatabaseFileName", MOC_ITC,
    "Filename for the object database",
    "object-db", '\0', "<filename>", "" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "TrainObjectDB", MOC_ITC,
    "0=don't train the database on new objects, 1=train database if the "
    "object could not be recognized.",
    "train-db", '\0', "", "yes" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "PromptUserTrainDB", MOC_ITC,
    "Whether or not to prompt the user before training the database on a "
    "new object.",
    "prompt-user-train-db", '\0', "", "no" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "MatchObjects", MOC_ITC,
    "0=don't attempt to match new objects, 1=attempt to match new objects.",
    "match-objects", '\0', "", "yes"},

  // Used by: InferoTemporal
  { MODOPT_STRING, "MatchingAlgorithm", MOC_ITC,
    "The algorithm to use for object matching",
    "matching-alg", '\0', "<Simple|KdTree|KdTreeBBF>", "Simple"},

  // Used by: InferoTemporal
  { MODOPT_INT, "RecognitionMinMatch", MOC_ITC,
    "The minimum number of keypoint matches for recognizing an object. If "
    "greater than 0, search will be terminated once an object with the "
    "minimum number of keypoint matches is found. If equal to 0, the "
    "percentage of keypoints specified by recog-minmatch-percent must be "
    "matched to declare an object match",
    "recog-minmatch", '\0', "<int>", "3" },

  // Used by: InferoTemporal
  { MODOPT_FLOAT, "RecognitionMinMatchPercent", MOC_ITC,
    "The minimum percentage of keypoints in the target object that must be "
    "successfully matched in order to declare an object match.",
    "recog-minmatch-percent", '\0', "<float>", "0.75" },

  // Used by: InferoTemporal
  { MODOPT_FLAG, "SortObjects", MOC_ITC,
    "Whether or not to sort the objects in the database before matching. "
    "similarity before matching.",
    "sort-objects", '\0', "", "no"},

  // Used by: InferoTemporal
  { MODOPT_STRING, "SortObjectsBy", MOC_ITC,
    "The metric to use to sort the objects in the database - 'features' or "
    "'location'. Only has effect when --sort-objects=yes. ",
    "sort-objects-by", '\0', "<features|location>", "features"},

  // Used by: InferoTemporal
  { MODOPT_FLAG, "SortKeys", MOC_ITC,
    "Whether or not to sort the keys by scale space extrema magnitude "
    "before matching.",
    "sort-keys", '\0', "", "no"},

  // Used by: InferoTemporal
  { MODOPT_STRING, "OutputFileName", MOC_ITC,
    "The name of the object recognition output file",
    "objrecog-output", '\0', "<filename>", ""},

  // #################### VisualCortex options:
  // Used by: VisualCortex, Brain, SingleChannel (and derivatives)
  { MODOPT_STRING, "VisualCortexType", MOC_BRAIN,
    "Type of VisualCortex to use. There are two ways to setup a "
    "VisualCortex: either by using one of the standard names defined "
    "below, which correspond to standard bundled collections of "
    "underlying visual processing channels, or by explicitly listing "
    "the visual channels to use. Pre-bundled collections of channels "
    "include:\n"
    "  None: no VisualCortex at all\n"
    "  Std: use all standard channels with unit weights\n"
    "  BeoStd: use all Beowulf channels with unit weights\n"
    "  SurpStd: use all standard Surprise channels\n"
    "  Entrop: entropy model, computing pixel entropy in 16x16 image patches\n"
    "  EyeMvt: fake visual cortex built from human eye movement traces\n" 
    "  PN03contrast: Parkhurst & Niebur'03 contrast model\n"
    "  Variance: local variance in 16x16 image patches\n"
    "  Tcorr: temporal correlation in image patches across frames\n"
    "  Scorr: spatial correlation between image patches in a frame\n"
    "  Info: DCT-based local information measure as in Itti et al, PAMI 1998\n"
    "You may also configure which channels to use in your VisualCortex by "
    "specifying a series of letters from:\n"
    "  C: double-opponent color center-surround\n"
    "  I: intensity center-surround\n"
    "  O: orientation contrast\n"
    "  F: flicker center-surround\n"
    "  M: motion energy center-surround\n"
    "  D: dummy channel\n"
    "  G: multi-color band channel\n"
    "  K: skin hue detector\n"
    "  S: composite single-opponent color center-surround\n"
    "  L: L-junction detector\n"
    "  T: T-junction detector\n"
    "  X: X-junction detector\n"
    "  E: end-stop detector\n"
    "with possible optional weights (given just after the channel letter, "
    "with a ':' in between). Finally, in this string a single letter 'B' "
    "will instruct to use Beowulf channels rather than normal channels "
    "(for this to work, you will need to be running a beochannel-server "
    "process on each of your Beowulf nodes), while a single letter 'U' "
    "will use the Surprise channels.\n"
    "EXAMPLE: 'IO:5.0M' will use intensity (weight 1.0), orientation "
    "(weight 5.0) and motion (weight 1.0) channels",
    "vc-type", '\0',
    "<None|Std|...|CIOFM...BU>",
    "Std" },
  // Used by: VisualCortex, Brain, SingleChannel (and derivatives)
  { MODOPT_FLAG, "UseRandom", MOC_BRAIN,
    "Add small amounts of random noise to various stages of model",
    "use-random", '\0', "", "true" },
  // Used by: VisualCortex, Brain, SingleChannel (and derivatives)
  { MODOPT_STRING, "BiasFeatures", MOC_BRAIN,
    "Bias the features to detect the target",
    "bias-features", '\0', "<ParamMap>", "" },
  // Used by: VisualCortex, ComplexChannel, SingleChannel (and derivatives)
  { MODOPT_FLAG, "UseOlderVersion", MOC_BRAIN,
    "Use the older version where we normalize responses within all feature "
    "types",
    "use-older-version", '\0', "", "true" },
  // Used by: VisualCortex and derivatives
  { MODOPT_FLOAT, "VCXSurpriseThresh", MOC_BRAIN,
    "Threshold for nonlinearity applied to Visual Cortex surprise output",
    "surprise-thresh", '\0', "<float>", "20.0" },
  // Used by: VisualCortex and derivatives
  { MODOPT_FLOAT, "VCXSurpriseSemisat", MOC_BRAIN,
    "Semisaturation constant for nonlinearity applied to Visual Cortex "
    "surprise output",
    "surprise-semisat", '\0', "<float>", "20.0" },
  // Used by: VisualCortex and derivatives
  { MODOPT_FLOAT, "VCXSurpriseExp", MOC_BRAIN,
    "Exponent for nonlinearity applied to Visual Cortex surprise output",
    "surprise-exp", '\0', "<float>", "4.0" },

  // #################### ShapeEstimator options:
  // Used by: ShapeEstimator
  { MODOPT_SEMODE, "ShapeEstimatorMode", MOC_BRAIN,
    "Shape estimator mode",
    "shape-estim-mode", '\0', "<None|FeatureMap|ConspicuityMap|SaliencyMap>",
    "None" },
  { MODOPT_SESMOOTHMETHOD, "ShapeEstimatorSmoothMethod", MOC_BRAIN,
    "Shape estimator smooth method",
    "shape-estim-smoothmethod", '\0', "<None|Gaussian|Chamfer>",
    "Gaussian" },

  // #################### Brain options:
  // Used by: Brain
  { MODOPT_IORTYPE, "IORtype", MOC_BRAIN,
    "Type of IOR to use, or infer from ShapeEstimatorMode if Auto",
    "ior-type", '\0', "<None|Auto|Disc|ShapeEst|ShapeEstSM|ShapeEstCM>",
    "Auto" },
  // Used by: Brain
  { MODOPT_FLAG, "BrainSaveObjMask", MOC_DISPLAY,
    "Save object mask",
    "save-objmask", '\0', "", "false" },
  // Used by: Brain
  { MODOPT_FLAG, "BlankBlink", MOC_BRAIN,
    "Blank-out visual input during eye blinks",
    "blank-blink", '\0', "", "false" },
  // Used by: Brain
  { MODOPT_INT, "BrainTooManyShifts", MOC_BRAIN,
    "Exit after this number of covert attention shifts",
    "too-many-shifts", '\0', "<int>", "0" },
  // Used by: Brain
  { MODOPT_INT, "BrainTooMuchTime", MOC_BRAIN,
    "Exit after the maximum specified time in seconds (unlimited if zero)",
    "too-much-time", '\0', "<float>", "0.0" },
  // Used by: Brain
  { MODOPT_FLAG, "BrainSaveWinnerFeatures", MOC_BRAIN,
    "Save low-level feature values at each covertly attended location into "
    "the text file specified by the --textlog=XXX option.",
    "save-winner-features", '\0', "", "false" },

  // #################### Saliency Map options:
  // Used by: SaliencyMapConfigurator
  { MODOPT_STRING, "SaliencyMapType", MOC_SM,
    "Type of Saliency Map to use. 'None' for no saliency map, 'Std' for the "
    "standard saliency map using LeakyIntegrator neurons (takes a lot "
    "of CPU cycles to evolve), 'Trivial' for just a non-evolving copy of the "
    "inputs, 'Fast' for a faster implementation that"
    "just takes a weighted average between current and new saliency "
    "map at each time step.",
    "sm-type", '\0', "<None|Std|Trivial|Fast>", "Std" },
  // Used by: SaliencyMap and derivatives
  { MODOPT_FLAG, "SMsaveResults", MOC_SM,
    "Save saliency map",
    "save-salmap", '\0', "", "false" },
  // Used by: SaliencyMap and derivatives
  { MODOPT_FLAG, "SMsaveCumResults", MOC_SM,
    "Save cumulative saliency map so far",
    "save-cumsalmap", '\0', "", "false" },
  // Used by: SaliencyMap and derivatives
  { MODOPT_FLOAT, "SMfactor", MOC_SM,
    "multiply salmap by <factor> instead of normalize, if not 0.0",
    "salmap-factor", 'S', "<factor>", "0.0" },
  // Used by: SaliencyMap and derivatives
  { MODOPT_FLAG, "SMuseSacSuppress", MOC_SM,
    "Use saccadic suppression in the saliency map",
    "salmap-sacsupp", '\0', "", "true" },
  // Used by: SaliencyMap and derivatives
  { MODOPT_FLAG, "SMuseBlinkSuppress", MOC_SM,
    "Use blink suppression in the saliency map",
    "salmap-blinksupp", '\0', "", "true" },
  // Used by: SaliencyMapFast
  { MODOPT_FLOAT, "SMfastInputCoeff", MOC_SM,
    "Coefficient by which new inputs merge with old ones at each time step",
    "salmap-inputcoeff", '\0', "<0..1>", "0.1" },
  // Used by: SaliencyMapFast
  { MODOPT_FLOAT, "SMfastItoVcoeff", MOC_SM,
    "Coefficient used to convert from synaptic current inputs (typically "
    "in the nA range) to voltage outputs (typically in the mV range)",
    "salmap-itovcoeff", '\0', "<float>", "5000000.0" },
  // Used by: SaliencyMapStd
  { MODOPT_FLOAT, "SMginhDecay", MOC_SM,
    "Coefficient by which inhibition-of-return dies off at each time step",
    "salmap-iordecay", '\0', "<0..1>", "0.9999" },

  // #################### Task-Relevance Map options:
  // Used by: TaskRelevanceMapConfigurator
  { MODOPT_STRING, "TaskRelevanceMapType", MOC_TRM,
    "Type of Task-Relevance Map to use. 'None' for no map, 'Std' for the "
    "standard TRM that needs an agent to compute the relevance of "
    "objects, 'KillStatic' for a TRM that progressively decreases "
    "the relevance of static objects, and 'KillN' for a TRM that kills "
    "the last N objects that have been passed to it.",
    "trm-type", '\0', "<None|Std|KillStatic|KillN>", "Std" },
  // Used by: TaskRelevanceMap and derivatives
  { MODOPT_FLAG, "TRMsaveResults", MOC_TRM,
    "Save task-relevance map",
    "save-trm", '\0', "", "false" },
  // Used by: TaskRelevanceMap and derivatives
  { MODOPT_FLOAT, "TRMfactor", MOC_TRM,
    "multiply TRM by <factor> instead of normalize, if not 0.0",
    "trm-factor", '\0', "<factor>", "0.0" },
  // Used by: Scene understanding model
  { MODOPT_FLAG, "LearnTRM", MOC_TRM,
    "Learn TRM over a number of scenes", "learn-trm", '\0', "", "false" },
  // Used by: Scene understanding model
  { MODOPT_FLAG, "LearnUpdateTRM", MOC_TRM,
    "Learn and update TRM over a number of scenes",
    "learn-update-trm", '\0', "", "false" },
  // Used by: Brain for the Scene understanding model
  { MODOPT_STRING, "BiasTRM", MOC_TRM,
    "bias the TRM with the given image", "bias-trm", '\0', "<file>", "" },
  // Used by: TaskRelevanceMapKillStatic
  { MODOPT_FLOAT, "TRMKillStaticThresh", MOC_TRM,
    "Minimum difference between current luminance and previously cumulated "
    "luminance for a point to be considered non-static",
    "kill-static-thresh", '\0', "<float>", "20.0" },
  // Used by: TaskRelevanceMapKillStatic
  { MODOPT_FLOAT, "TRMKillStaticCoeff", MOC_TRM,
    "Killing achieved by a purely static location. "
    "A value of 1.0 here means that all salience of a static location will "
    "be killed; a value of 0 means no killing",
    "kill-static-coeff", '\0', "<0.0 .. 1.0>", "0.95" },
  // Used by: TaskRelevanceMapKillN
  { MODOPT_INT, "TRMkillN", MOC_TRM,
    "Number of objects to kill.",
    "killn", '\0', "<int>", "3" },

  // #################### Attention Guidance Map options:
  // Used by: AttentionGuidanceMapConfigurator
  { MODOPT_STRING, "AttentionGuidanceMapType", MOC_AGM,
    "Type of Attention Guidance Map to use. 'None' for no map (in which "
    "case Brain will use the bottom-up saliency map for attention "
    "guidance), 'Std' for the standard AGM that just computes the pointwise "
    "product between bottom-up saliency map and top-down task-relevance "
    "map.",
    "agm-type", '\0', "<None|Std>", "Std" },
  // Used by: AttentionGuidanceMap and derivatives
  { MODOPT_FLAG, "AGMsaveResults", MOC_AGM,
    "Save attention guidance map",
    "save-agm", '\0', "", "false" },
  // Used by: AttentionGuidanceMap and derivatives
  { MODOPT_FLOAT, "AGMfactor", MOC_AGM,
    "multiply AGM by <factor> instead of normalize, if not 0.0",
    "agm-factor", '\0', "<factor>", "0.0" },

  // #################### Winner-Take-All options:
  // Used by: WinnerTakeAllConfigurator
  { MODOPT_STRING, "WinnerTakeAllType", MOC_WTA,
    "Type of Winner-Take-All to use. 'None' for no winner-take-all, 'Std' "
    "for the standard winner-take-all using LeakyIntFire neurons "
    "(takes a lot of CPU cycles to evolve), 'Fast' for a faster "
    "implementation that just computes the location of the max at "
    "every time step, 'Dummy' is a dummy one that does nothing and never "
    "returns a winner, 'Greedy' is one that returns, out of a number of "
    "possible targets above a threshold, the one closest to current eye "
    "position.",
    "wta-type", '\0', "<None|Std|Fast|Dummy|Greedy>", "Std" },
  // Used by: WinnerTakeAll and derivatives
  { MODOPT_FLAG, "WTAsaveResults", MOC_WTA,
    "Save winner-take-all membrane potential values",
    "save-wta", '\0', "", "false" },
  // Used by: WinnerTakeAll and derivatives
  { MODOPT_FLOAT, "WTAfactor", MOC_WTA,
    "multiply WTA values by <factor> instead of normalize, if not 0.0",
    "wta-factor", '\0', "<factor>", "0.0" },
  // Used by: WinnerTakeAll and derivatives
  { MODOPT_FLAG, "WTAuseSacSuppress", MOC_WTA,
    "Use saccadic suppression in the winner-take-all",
    "wta-sacsupp", '\0', "", "true" },
  // Used by: WinnerTakeAll and derivatives
  { MODOPT_FLAG, "WTAuseBlinkSuppress", MOC_WTA,
    "Use blink suppression in the winner-take-all",
    "wta-blinksupp", '\0', "", "true" },
  // Used by: WinnerTakeAllGreedy
  { MODOPT_FLOAT, "WinnerTakeAllGreedyThreshFac", MOC_WTA,
    "Threshold, as a factor of the max, over which possible local "
    "maxima will be considered",
    "wta-greedyfac", '\0', "<0.0 .. 1.0>", "0.5" },

  // #################### SaccadeController options:
  // Used by: SaccadeControllerConfigurator
  { MODOPT_STRING, "SaccadeControllerType", MOC_SAC,
    "Saccade Controller name (pick a name and then use --help to see "
    "options specific to the chosen controller)",
    "sc-type", '\0',
    "<None|Trivial|Fixed|Friction|Threshold|Threshfric|Eyehead|"
    "Monkey|Monkey2|EyeTrack|VisBuf|ModelW>", "None" },
  // Used by: SaccadeControllers
  { MODOPT_POINT2D, "SCeyeInitialPosition", MOC_SAC,
    "Initial eye position, or 0,0 to start at image center",
    "initial-eyepos", '\0', "<x,y>", "0,0" },
  // Used by: SaccadeControllers
  { MODOPT_FLAG, "SCeyeStartAtIP", MOC_SAC,
    "Start at initial eye position (otherwise at first focus of attention)",
    "start-at-iep", '\0', "", "false" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCeyeSpringK", MOC_SAC,
    "Eye spring stiffness to use in mass/spring simulations",
    "eye-spring-k", '\0', "<float>", "50000.0" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCeyeFrictionMu", MOC_SAC,
    "Eye friction coefficient to use in mass/spring simulations",
    "eye-friction-mu", '\0', "<float>", "500.0" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCeyeThreshMinOvert", MOC_SAC,
    "Minimum overt distance for a saccade (as factor of fovea radius)",
    "eye-minovert-fac", '\0', "<float>", "2.0" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCeyeThreshMaxCovert", MOC_SAC,
    "Maximum covert distance for a saccade (as factor of fovea radius)",
    "eye-maxcovert-fac", '\0', "<float>", "1.5" },
  // Used by: SaccadeControllers
  { MODOPT_INT, "SCeyeThreshMinNum", MOC_SAC,
    "Minimum required number of covert shifts within max covert distance",
    "eye-mincovertnum", '\0', "<int>", "4" },
  // Used by: SaccadeControllers
  { MODOPT_FLAG, "SCeyeThreshSalWeigh", MOC_SAC,
    "Weight covert positions by saliency when computing average distance",
    "eye-saliencyweight", '\0', "<true|false>", "true" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCeyeMaxIdleSecs", MOC_SAC,
    "Maximum idle time (in s) before eyes return to initial eye position",
    "eye-maxidle-time", '\0', "<float>", "0.75" },

  // Used by: SaccadeControllers
  { MODOPT_POINT2D, "SCheadInitialPosition", MOC_SAC,
    "Initial head position, or 0,0 to start at image center",
    "initial-headpos", '\0', "<x,y>", "0,0" },
  // Used by: SaccadeControllers
  { MODOPT_FLAG, "SCheadStartAtIP", MOC_SAC,
    "Start at initial head position (otherwise at first focus of attention)",
    "start-at-ihp", '\0', "", "false" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCheadSpringK", MOC_SAC,
    "Head spring stiffness to use in mass/spring simulations",
    "head-spring-k", '\0', "<float>", "10000.0" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCheadFrictionMu", MOC_SAC,
    "Head friction coefficient to use in mass/spring simulations",
    "head-friction-mu", '\0', "<float>", "5000.0" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCheadThreshMinOvert", MOC_SAC,
    "Minimum overt distance for a head saccade (as factor of fovea radius)",
    "head-minovert-fac", '\0', "<float>", "2.0" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCheadThreshMaxCovert", MOC_SAC,
    "Maximum covert distance for a head saccade (as factor of fovea radius)",
    "head-maxcovert-fac", '\0', "<float>", "1.5" },
  // Used by: SaccadeControllers
  { MODOPT_INT, "SCheadThreshMinNum", MOC_SAC,
    "Minimum required number of covert shifts within max covert distance",
    "head-mincovertnum", '\0', "<int>", "4" },
  // Used by: SaccadeControllers
  { MODOPT_FLAG, "SCheadThreshSalWeigh", MOC_SAC,
    "Weight covert positions by saliency when computing average distance",
    "head-saliencyweight", '\0', "<true|false>", "true" },
  // Used by: SaccadeControllers
  { MODOPT_FLOAT, "SCheadMaxIdleSecs", MOC_SAC,
    "Maximum idle time (in s) before head return to initial eye position",
    "head-maxidle-time", '\0', "<float>", "0.75" },

  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCeyeBlinkWaitTime", MOC_SAC,
    "Typical wait time between eye blinks (in s)",
    "blink-wait", '\0', "<float>", "1.8" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "SCeyeBlinkDuration", MOC_SAC,
    "Eye blink duration (in s)",
    "blink-duration", '\0', "<float>", "0.190" },

  // Used by: SimulationViewerCompress, SaccadeControllerEyeTracker
  { MODOPT_STRING, "EyeTrackFname", MOC_SAC,
    "File name for eye movement data, or empty if no data",
    "eye-fname", '\0', "<file>", "" },
  // Used by: SimulationViewerCompress, SaccadeControllerEyeTracker
  { MODOPT_FLOAT, "EyeTrackPeriod", MOC_SAC,
    "Eye movements sampling period",
    "eye-period", '\0', "<float>", "0.0041666666" }, // 240Hz
  // Used by: SimulationViewerCompress, SaccadeControllerEyeTracker
  { MODOPT_INT, "EyeTrackTrash", MOC_SAC,
    "Number of initial eye movement samples to trash (e.g., "
    "corresponding to initial fixation before actual stimulus)",
    "eye-trash", '\0', "<int>", "0" },

  // #################### GistEstimator options:
  // Used by: SaccadeControllerConfigurator
  { MODOPT_STRING, "GistEstimatorType", MOC_GE,
    "Type of gist estimator to use",
    "ge-type", '\0',
    "<None|Std>", "None" },

  // #################### Other:
  // Used by: SimulationViewer, InputFrameSeries
  { MODOPT_INT, "FOAradius", MOC_BRAIN,
    "Radius of the Focus Of Attention, or 0 to set it from input image dims",
    "foa-radius", '\0', "<pixels>", "0" },
  // Used by: SimulationViewer, InputFrameSeries, SaccadeControllers
  { MODOPT_INT, "FoveaRadius", MOC_BRAIN,
    "Radius of the fovea, or 0 to set it from input image dims",
    "fovea-radius", '\0', "<pixels>", "0" },
  // Used by: SimulationViewer
  { MODOPT_INT, "HeadMarkerRadius", MOC_BRAIN,
    "Radius of the head position marker",
    "head-radius", '\0', "<pixels>", "50" },
  // Used by: SimulationViewer
  { MODOPT_INT, "MultiRetinaDepth", MOC_BRAIN,
    "Depth of pyramid used for foveation",
    "multiretina-depth", '\0', "<int>", "5" },
  // Used by: SaccadeControllers
  { MODOPT_DOUBLE, "PixelsPerDegree", MOC_BRAIN,
    "Pixels per degree of visual angle",
    "pixperdeg", '\0', "<float>", "5.333333333" }, // 120deg for 640pix

  // #################### MbariResultViewer options:
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVsaveEvents", MOC_MBARIRV,
    "Save the event structure to a text file",
    "mbari-save-events", '\0', "fileName", "" },
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVloadEvents", MOC_MBARIRV,
    "Load the event structure from a text file "
    "instead of computing it from the frames",
    "mbari-load-events", '\0', "fileName", "" },
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVsaveProperties", MOC_MBARIRV,
    "Save the event property vector to a text file",
    "mbari-save-properties", '\0', "fileName", "" },
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVloadProperties", MOC_MBARIRV,
    "Load the event property vector from a text file",
    "mbari-load-properties", '\0', "fileName", "" },
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVsavePositions", MOC_MBARIRV,
    "Save the positions of events to a text file",
    "mbari-save-positions", '\0', "fileName", "" },
  // Used by: MbariResultViewer
  { MODOPT_BODRAWMODE, "MRVmarkInteresting", MOC_MBARIRV,
    "Way to mark interesting events in output frames of MBARI programs",
    "mbari-mark-interesting", '\0', "<None|Shape|Outline|BoundingBox>",
    "BoundingBox" },
  // Used by: MbariResultViewer
  { MODOPT_FLOAT, "MRVopacity", MOC_MBARIRV,
    "Opacity of shape or outline markings of events",
    "mbari-opacity", '\0', "<0.0 ... 1.0>", "1.0" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVmarkCandidate", MOC_MBARIRV,
    "Mark candidates for interesting events in output frames of MBARI programs",
    "mbari-mark-candidate", '\0', "", "true" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVmarkPrediction", MOC_MBARIRV,
    "Mark the Kalman Filter's prediction for the location of an object "
    "in output frames of MBARI programs",
    "mbari-mark-prediction", '\0', "", "false" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVmarkFOE", MOC_MBARIRV,
    "Mark the focus of expansion in the output frames of MBARI programs",
    "mbari-mark-foe", '\0', "", "false" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVsaveResults", MOC_MBARIRV,
    "Save intermediate results in MBARI programs to disc",
    "mbari-save-results", '\0', "", "false" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVdisplayResults", MOC_MBARIRV,
    "Display intermediate results in MBARI programs",
    "mbari-display-results", '\0', "", "false" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVsaveOutput", MOC_MBARIRV,
    "Save output frames in MBARI programs",
    "mbari-save-output", '\0', "", "true" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVdisplayOutput", MOC_MBARIRV,
    "Display output frames in MBARI programs",
    "mbari-display-output", '\0', "", "false" },
  // Used by: MbariResultViewer
  { MODOPT_FLAG, "MRVshowEventLabels", MOC_MBARIRV,
    "Write event labels into the output frames",
    "mbari-label-events", '\0', "", "true" },
  // Used by: MbariResultViewer
  { MODOPT_DIMS, "MRVrescaleDisplay", MOC_MBARIRV,
    "Rescale displays to <width>x<height>, or 0x0 for no rescaling",
    "mbari-rescale-display", '\0', "<width>x<height>", "0x0" },
  // Used by: MbariResultViewer
  { MODOPT_STRING, "MRVsaveEventNums", MOC_MBARIRV,
    "Save video clips showing specific events",
    "mbari-save-event-clip", '\0', "ev1,ev1,...,evN; or: all", "" },
  // Used by: MbariResultViewer
  { MODOPT_INT, "MRVsizeAvgCache", MOC_MBARIRV,
    "The number of frames used to compute the running average",
    "mbari-cache-size", '\0', "<int>", "10" },

  // #################### ImageDisplay options:
  // Used by: ImageDisplay
  { MODOPT_IMAGEDISPLAYTYPE, "ImageDisplayType", MOC_IMAGEDISP,
    "The type of display to be used for displaying images",
    "image-display", '\0', "<None|XV|XWin|XWinManaged>", "None" },

  // #################### VisualBuffer options:
  // Used by: VisualBuffer
  { MODOPT_FLAG, "VBignoreBoring", MOC_VB,
    "Ignore boring shifts of attention if true",
    "vb-ignore-boring", '\0', "", "true" },
  // Used by: VisualBuffer
  { MODOPT_FLAG, "VBobjectBased", MOC_VB,
    "Use object-based saliency imprinting if true",
    "vb-object-based", '\0', "", "true" },
  // Used by: VisualBuffer
  { MODOPT_DIMS, "VBdims", MOC_VB,
    "Use given dims for visual buffer (at saliency map scale), or use same "
    "dims as the saliency map if set to (0, 0)",
    "vb-dims", '\0', "<w>x<h>", "0x0" },
  // Used by: VisualBuffer
  { MODOPT_DOUBLE, "VBtimePeriod", MOC_VB,
    "Apply internal competitive interactions within the buffer at "
    "intervals given by this parameter (in s).",
    "vb-period", '\0', "<double>", "0.05" },
  // Used by: VisualBuffer
  { MODOPT_FLOAT, "VBfactor", MOC_VB,
    "Multiply buffer by <factor> instead of normalize, if not 0.0",
    "vb-factor", '\0', "<factor>", "0.0" },
  // Used by: VisualBuffer
  { MODOPT_FLOAT, "VBdecayFactor", MOC_VB,
    "Multiply buffer by <factor> at each time period, if not 1.0",
    "vb-decay", '\0', "<factor>", "1.0" },
  // Used by: VisualBuffer
  { MODOPT_MAXNORM, "VBmaxNormType", MOC_VB,
    "Type of MaxNormalization to use",
    "vb-maxnorm-type", '\0', "<see --maxnorm-type>", "Fancy" },

  // #################### I-Room options
  // Used by: I-Room
  { MODOPT_FLAG, "I-Room-CameraNode", MOC_IROOM,
    "Flags this I-Room node as containing a camera",
    "i-room-camera", '\0', "", "false" },

  // #################### Cluster feature options
  { MODOPT_STRING, "NPconfig", MOC_FCV,
    "Config file (readConfig.H) to open for NPclassify",
    "fcv-np-config",'\0',"<file>","NPclassify.conf"},
  { MODOPT_STRING, "Polyconfig", MOC_FCV,
    "Config file (readConfig.H) to open for Kernel parameters",
    "fcv-poly-config",'\0',"<file>","polySet.conf"},
  { MODOPT_STRING, "localconfig", MOC_FCV,
    "Config file (readConfig.H) for FCV and fzvision",
    "fcv-local-config",'\0',"<file>","feature.conf"},
  { MODOPT_STRING, "icamatrix", MOC_FCV,
    "ICA unmixing matrix files tab-delim with names [name].[number]",
    "fcv-ica-matrix",'\0',"<file>","../ICA/ICAunmix"},

  // #################### Option Aliases and Shortcuts
  // These can be requested using ModelManager::requestOptionAlias().
  // See in ModelOptionAliases.H some macros that will request groups
  // of aliases for various models (e.g., all aliases applicable to
  // attention models, etc)
  { MODOPT_ALIAS, "ALIASmovie", MOC_ALIAS,
    "Process in movie (multiframe) mode",
    "movie", '\0', "",
    "--nodisplay-traj --nodisplay-additive --nouse-fpe --salmap-factor=50000 "
    "--agm-factor=50000 --nouse-random --nodisplay-foa --display-patch" },

  { MODOPT_ALIAS, "ALIASmoviefov", MOC_ALIAS,
    "Process in foveated movie (multiframe) mode with a ThreshFric "
    "SaccadeController",
    "moviefov", '\0', "",
    "--movie --foveate-input --foveate-input-depth=6 --sc-type=Threshfric "
    "--ior-type=None --trm-type=KillStatic" },

  { MODOPT_ALIAS, "ALIASmovieeyehead", MOC_ALIAS,
    "Process in foveated movie (multiframe) mode with a Monkey2 "
    "Eye/Head SaccadeController and displays of head position",
    "movieeyehead", '\0', "",
    "--moviefov --sc-type=Monkey2 --display-head" },

  { MODOPT_ALIAS, "ALIASmovieanim", MOC_ALIAS,
    "Process in foveated movie (multiframe) mode with a human-derived "
    "ModelW Eye/Head SaccadeController, a WinnerTakeAllGreedy, no IOR, "
    "and displays of head position",
    "movieanim", '\0', "",
    "--movie --foveate-input --foveate-input-depth=4 --sc-type=ModelW "
    "--ior-type=None --display-head --shape-estim-mode=None "
    "--wta-type=Greedy --start-at-iep --start-at-ihp" },

  { MODOPT_ALIAS, "ALIASsurprise", MOC_ALIAS,
    "Use standard surprise mode",
    "surprise", '\0', "",
    "--vc-type=UCIOFM --maxnorm-type=Surprise "// surprise channels and maxnorm
    "--nouse-older-version --chanout-max=0.0 " // no weighing
    "--gabor-intens=20.0 --direction-sqrt "    // 0..255 range every channel
    "--vcx-outfac=1.0e-8 --nouse-random" },

  // ########## Make sure this record is last (end marker): ##########
  { MODOPT_NONE, "", MOC_GENERAL, "", "", '\0', ""}
};

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
