/*!@file ModelOptionDefs.H Specification for declaration of all valid
  model options */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelOptionDefs.H 3874 2004-10-29 06:16:54Z siagian $
//

#ifndef MODELOPTIONDEFS_H_DEFINED
#define MODELOPTIONDEFS_H_DEFINED


// ######################################################################
//! Known model option types
/*! If you add new types here, be sure to update ModelParam.C,
  StringConversions.[CH] and ModelManager.C (function
  ModelManager::parseCommandLine()) */
enum ModelOptionType {
  MODOPT_NONE, MODOPT_BOOL, MODOPT_INT, MODOPT_UINT, MODOPT_FLOAT,
  MODOPT_DOUBLE, MODOPT_STRING, MODOPT_DIMS, MODOPT_POINT2D, MODOPT_FLAG,
  MODOPT_GRABMODE, MODOPT_FRAMERANGE, MODOPT_LEVELSPEC, MODOPT_RECTANGLE,
  MODOPT_RASTERFFORMAT, MODOPT_RGB, MODOPT_PTYPE, MODOPT_MAXNORM,
  MODOPT_SEMODE, MODOPT_SESMOOTHMETHOD, MODOPT_INADDR, MODOPT_SHORT,
  MODOPT_ORICOMPTYPE, MODOPT_IORTYPE, MODOPT_BODRAWMODE, 
  MODOPT_IMAGEDISPLAYTYPE, MODOPT_ALIAS, MODOPT_COLCOMPTYPE
};

// ######################################################################
//! Known model option categories; used for printing --help message
enum ModelOptionCateg {
  MOC_GENERAL    = 0, //!< General simulation options
  MOC_INPUT      = 1, //!< Input reading/formatting options
  MOC_OUTPUT     = 2, //!< Output writing/formatting options
  MOC_DISPLAY    = 3, //!< What to display/save options
  MOC_AUDIO      = 4, //!< Audio-related options
  MOC_RETINA     = 5, //!< Retina-related options
  MOC_CHANNEL    = 6, //!< Channel-related options
  MOC_BRAIN      = 7, //!< Brain-related options
  MOC_SAC        = 8, //!< SaccadeController-related options
  MOC_FRAMEGRAB  = 9, //!< FrameGrabber-related options
  MOC_BEOWULF    =10, //!< Beowulf-related options
  MOC_SPECHARD   =11, //!< Special-hardware options (e.g., pan/tilt cam, etc)
  MOC_PSYCHODISP =12, //!< PsychoDisplay options
  MOC_MBARIRV    =13, //!< MbariResultViewer options
  MOC_IMAGEDISP  =14, //!< ImageDisplay options
  MOC_SM         =15, //!< SaliencyMap options
  MOC_TRM        =16, //!< TaskRelevanceMap options
  MOC_AGM        =17, //!< AttentionGuidanceMap options
  MOC_WTA        =18, //!< WinnerTakeAll options
  MOC_VB         =19, //!< VisualBuffer options
  MOC_IROOM      =20, //!< I-Room related options
  MOC_FCV        =21, //!< Feature Cluster Options
  MOC_ITC	 =22, //!< Inferior Temporal Cortex Options
  MOC_GE	 =23, //!< GistEstimator Options
  // please keep this one last; if you add a category, put it above
  // and change the number of this last one as well as the number of
  // categories below:
  MOC_ALIAS      =24  //!< Option aliases
};

//! number of option categories
#define NBMODELOPTIONCATEGS 25

//! text description of all option categories
extern const char *AllModelOptionCategs[NBMODELOPTIONCATEGS];

// ######################################################################
//! Specification of a known model option
/*! This struct is used to build a list of all known model options in
  ModelOptions.C */
struct ModelOptionDef
{
  ModelOptionType typ;     //!< Type of the option
  const char *name;        //!< Option name
  const ModelOptionCateg categ; //!< Option category
  const char *descr;       //!< Description of what option does
  const char *longoptname; //!< Long option name (without leading "--")
  char shortoptname;       //!< Short option name (without leading "-")
  const char *validvals;   //!< Textual description of valid values, or NULL
  const char *defval;      //!< Default value in text representation
};

//! Our static array of all known model options, defined in ModelOptionDefs.C
extern const ModelOptionDef AllModelOptions[];

// ######################################################################
//! What to export when calling ModelComponent::exportOptions()

//! Export no option
#define OPTEXP_NONE       0

//! Export all options
#define OPTEXP_ALL        0xffff

/*! Export our core configuration options that do not require any other
    model component to be present in the model */
#define OPTEXP_CORE      (1 << 0)

/*! Export our options that relate to saving results, and require that
  an OutputFrameSeries be available (and used!) in the model */
#define OPTEXP_SAVE      (1 << 1)

/*! Export options that relate to the MbariResultViewer */
#define OPTEXP_MRV       (1 << 2)

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
