/*!@file ModelParam.C A tunable ModelComponent parameter base class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelParam.C 3905 2004-11-02 03:01:02Z rjpeters $
//

#include "ModelParam.H"
#include "ParamMap.H"
#include "ModelComponent.H"
#include "log.H"
#include "StringConversions.H"
#include <iostream>
#include "saliency.H"            // for byte
#include "ArrayData.H"           // for Dims
#include "Point2D.H"             // for Point2D
#include "FrameGrabberMode.H"    // for FrameGrabberMode
#include "FrameRange.H"          // for FrameRange
#include "LevelSpec.H"           // for LevelSpec
#include "Rectangle.H"           // for Rectangle
#include "RasterFileFormat.H"    // for RasterFileFormat
#include "Pixels.H"              // for PixRGB<byte>
#include "VisualFeatures.H"      // for VisualFeature
#include "PyramidTypes.H"        // for PyramidType
#include "fancynorm.H"           // for MaxNormType
#include "IORtypes.H"            // for IORtype
#include "ShapeEstimatorModes.H" // for ShapeEstimatorMode
#include "OrientComputeTypes.H"  // for OrientComputeType
#include <netinet/in.h>          // for in_addr
#include "BitObjectDrawModes.H"  // for BitObjectDrawMode
#include <netinet/in.h>          // for in_addr
#include "ImageDisplayTypes.H"   // for ImageDisplayType
#include "ColorComputeTypes.H"   // for ColorComputeType

// Instantiations of ModelParam<T>:
#define INSTANTIATE(T) template class ModelParam< T >;
#include "InstantiateAll.H"

// ######################################################################
// ############# ModelParamBase implementation
// ######################################################################
ModelParamBase::ModelParamBase() :
  master(NULL), name("")
{  }

// ######################################################################
ModelParamBase::ModelParamBase(ModelComponent *mast, const std::string& nam) :
  master(mast), name(nam)
{
  // register ourselves with our master:
  master->registerParam(this);
}

// ######################################################################
void ModelParamBase::init(ModelComponent *mast, const std::string& nam)
{
  if (master) LFATAL("Already initialized -- ABORT");
  master = mast; name = nam;

  // register ourselves with our master:
  master->registerParam(this);
}

// ######################################################################
bool ModelParamBase::initialized() const
{ return (master != NULL); }

// ######################################################################
ModelParamBase::~ModelParamBase()
{
  // un-register ourselves with our master:
  if (master) { master->unregisterParam(this); master = NULL; }
}

// ######################################################################
std::string ModelParamBase::getName() const
{ return name; }

// ######################################################################
void ModelParamBase::printout(std::ostream& s, const std::string& prefix) const
{
  if (master)
    s<<prefix<<": "<<name<<" = "<<this->getValString()<<std::endl;
  else
    s<<prefix<<": "<<name<<" (uninitialized)"<<std::endl;
}

// ######################################################################
void ModelParamBase::writeTo(ParamMap& pmap) const
{
  if (master) // be quiet if not initialized
    { pmap.putStringParam(name.c_str(), this->getValString().c_str()); }
}

// ######################################################################
void ModelParamBase::readFrom(const ParamMap& pmap, const bool noerr)
{
  if (master) // be quiet if not initialized
    {
      std::string str = this->getValString(); const char* res = str.c_str();
      ParamMap::ReturnCode rc = pmap.queryStringParam(name.c_str(), res);
      if (rc == ParamMap::CHANGED)
        // NOTE: we could do a this->setValString(std::string(res));
        // but instead go through
        // ModelComponent::setModelParamString() as some components
        // may want to overload that call and do additional stuff when
        // their parameter values change:
        master->setModelParamString(name, std::string(res), false, true);
      if (noerr == false && rc == ParamMap::MISSING)
        LERROR("Cannot find parameter %s in ParamMap", name.c_str());
    }
}

// ######################################################################
// ############# ModelParam<T> implementation
// ######################################################################

template <class T>
ModelParam<T>::ModelParam() : ModelParamBase()
{  }

// ######################################################################
template <class T>
ModelParam<T>::ModelParam(ModelComponent *mast, const std::string& nam,
                          const T& initval) :
  ModelParamBase(mast, nam), val(initval)
{  }

// ######################################################################
template <class T>
void ModelParam<T>::init(ModelComponent *mast, const std::string& nam,
                         const T& initval)
{ ModelParamBase::init(mast, nam); val = initval; }

// ######################################################################
template <class T>
void ModelParam<T>::setVal(const T& v)
{ val = v; }

// ######################################################################
template <class T>
T ModelParam<T>::getVal() const
{ return val; }

// ######################################################################
template <class T>
std::string ModelParam<T>::getValString() const
{ std::string str; convertToString(val, str); return str; }

// ######################################################################
template <class T>
bool ModelParam<T>::setValString(const std::string& textval)
{ return convertFromString(textval, val); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
