/*!@file ModelParam.H A tunable ModelComponent parameter base class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ModelParam.H 2493 2003-04-29 08:42:03Z itti $
//

#ifndef MODELPARAM_H_DEFINED
#define MODELPARAM_H_DEFINED

#include <string>
#include <iosfwd>
class ModelComponent;
class ParamMap;
class ModelManager;

// ######################################################################
//! ModelParamBase is a persistent ModelComponent parameter base class
/*! Various model parameter types can be derived from this base, see
  the ModelParam<T> template class definition. ModelParamBase is
  persistent in the sense that it can be saved/loaded into a ParamMap,
  which itself provides disk save/load capabilities. ModelParamBase is
  intended to be used as a data member in a ModelComponent (or
  derived) object. ModelComponent provides an interface by which
  ModelParamBase derivatives may also be tuned via command-line
  options. */
class ModelParamBase
{
public:
  // ######################################################################
  /*! @name Constructors and Destructors */
  //@{
 
  //! Uninitialized constructor
  /*! The ModelParamBase created using this constructor will be
    unusable until it has been configured using init(). This
    constructor is provided for the creation of ModelParamBase
    arrays. */
  ModelParamBase();

  //! Initialized constructor
  /*! @param mast the ModelComponent which we are attached to
      @param nam the name of the ModelParamBase, that will be used to
         store its value in a ParamMap and fetch command-line option
         configuration data. */
  ModelParamBase(ModelComponent *mast, const std::string& nam);

  //! Initialization of an unitialized ModelParamBase
  /*! This will produce a fatal error if the ModelParamBase has already
      been initialized. Should be called before other uses on all
      ModelParamBase objects created with the default constructor. 
      @param mast the ModelComponent that we are attached to
      @param nam the name of the ModelParamBase that will be used to store
      its value in a ParamMap and fetch command-line option
      configuration data.*/
  virtual void init(ModelComponent *mast, const std::string& nam);

  //! Return true if we have been initialized
  bool initialized() const;

  //! Destructor
  virtual ~ModelParamBase();

  //@}

  // ######################################################################
  /*! @name Basic access functions */
  //@{

  //! get the ModelParamBase's name
  std::string getName() const;

  //! Set the parameter value from a textual representation
  /*! Returns true on successful parsing. Must be implemented by
    derived classes. Typically, a derived class of type T will also
    implement a method "void setVal(const T& v)" for more efficient
    direct assignment of the parameter value. See the class definition
    for ModelParam<T> */
  virtual bool setValString(const std::string& textval) = 0;

  //! Get current value as a string (must be implemented by derived classes)
  /*! Typically, a derived class of type T will also implement a
    method "T getVal() const" for more efficient direct access to the
    parameter value. See the class definition for ModelParam<T> */
  virtual std::string getValString() const = 0;

  //@}

  // ######################################################################
  /*! @name Input/Output functions */
  //@{

  //! Print out our name and contents, mostly for debugging
  virtual void printout(std::ostream& s, const std::string& prefix = "") const;

  //! Write parameter value to ParamMap
  void writeTo(ParamMap& pmap) const;

  //! Get parameter value from ParamMap
  /*! @param noerr will not generate an error message if the parameter
    does not exist in ParamMap. */
  void readFrom(const ParamMap& pmap, const bool noerr = true);

  //@}

protected:
  // yes, there is no parameter value data proper and no way to get
  // the parameter value directly! This is only the base class.
  // Derived classes will implement those additional things.

  ModelComponent *master;    //!< keep track of our master
  std::string name;          //!< the parameter name as used in ParamMap

private:
  ModelParamBase(const ModelParamBase& m);      //!< forbid copy-contruction
  ModelParamBase& operator=(const ModelParamBase& m); //!< forbid copy
};

// ######################################################################
//! ModelParam<T> definitions
/*! ModelParam<T> is a template class that derives from ModelParamBase
  and implements a tunable parameter of type T. To the ModelParamBase
  interface, ModelParam<T> adds getVal() and setVal() for direct
  access to the parameter value. */
template <class T> class ModelParam : public ModelParamBase
{
public:
  // ######################################################################
  /*! @name Constructors and Destructors */
  //@{

  //! Uninitilized constructor
  /*! You will need to call init() before any other method */
  ModelParam();

  //! Full constructor
  /*! Same as ModelParamBase plus an initial parameter value */
  ModelParam(ModelComponent *mast, const std::string& nam, const T& initval);

  //! Initialization
  /*! Same as ModelParamBase plus an initial parameter value */
  void init(ModelComponent *mast, const std::string& nam, const T& initval);

  //@}
  
  // ######################################################################
  /*! @name Access functions */
  //@{

  //! Set the value
  void setVal(const T& v);

  //! Get the value
  T getVal() const;

  //! Set the value from a string representation
  bool setValString(const std::string& textval);

  //! Get the value as a string
  std::string getValString() const;

  //@}

private:
  T val;     // the parameter value
  ModelParam(const ModelParam<T>& m);      //!< forbid copy-contruction
  ModelParam& operator=(const ModelParam<T>& m); //!< forbid copy
};

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
