/*!@file ParamMap.C I/O operations for parameter files */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: ParamMap.C 3104 2004-02-10 00:36:12Z zhanshi $
//

#ifndef PARAMMAP_C_DEFINED
#define PARAMMAP_C_DEFINED

#include "ParamMap.H"

#include "log.H"
#include "saliency.H"

#include <fstream>
#include <iomanip>
#include <iostream>
#include <limits>
#include <map>
#include <sstream>
#include <string>
#include <vector>

// ######################################################################
namespace
{
  void escapeWrite(std::ostream& ostrm, const std::string& s)
  {
    for (uint i = 0; i < s.size(); ++i)
      {
        switch (s[i])
          {
          // Need to escape whitespace in order to allow interior
          // whitespace within attribute values:
          case ' '   : ostrm << '\\' << ' '; break;
          case '\t'  : ostrm << '\\' << '\t'; break;
          case '\n'  : ostrm << '\\' << '\n'; break;

          // Must escape a backslash itself in order to not confuse it with
          // "backslash-as-escape-character"
          case '\\'  : ostrm << '\\' << '\\'; break;

          // Anything else can be written literally
          default    : ostrm << s[i]; break;
          }
      }
  }

  std::string escapeRead(std::istream& istrm)
  {
    // Read into a std::vector<char> rather than a std::string since it has
    // amortized-constant push_back(), which is not necessarily the case
    // for std::string.
    std::vector<char> buf;

    bool finished = false;

    // Skip over any unescaped whitespace that precedes the text of interest:
    istrm >> std::ws;

    while (!finished)
      {
        const int c = char(istrm.get());

        switch(c)
          {
          // (1) In case of unescaped whitespace or EOF, we're done:
          case EOF:
          case ' ':
          case '\t':
          case '\n':
            finished = true;
            break;

          // (2) In case of a backslash, take the subsequent character literally:
          case '\\':
            if (istrm.peek() == EOF)
              {
                // We'd better not find EOF following a backslash
                LERROR("Incomplete escape sequence before EOF!");
                throw ParamMap::Error();
              }
            buf.push_back(char(istrm.get()));
            break;

          // (3) Otherwise, take the current character literally:
          default:
            buf.push_back(char(c));
            break;
          }
      }

    // Must add a null-terminator
    buf.push_back('\0');

    // Gobble up any trailing un-escaped whitespace from the input stream
    while( istrm.peek() == '\t' || istrm.peek() == ' ' )
      istrm.get();

    return std::string(&buf[0]);
  }

  struct Param
  {
    Param() : str(), pmap() {}

    Param(double v) : str(), pmap()
    {
      std::ostringstream oss;
      oss << std::setprecision(25) << v;
      str = oss.str();
    }

    Param(int v) : str(), pmap()
    {
      std::ostringstream oss; oss << v; str = oss.str();
    }

    Param(const std::string& s) : str(s), pmap() {}

    Param(const SharedPtr<ParamMap>& p) : str(), pmap(p) {}

    void put(std::ostream& ostrm, int indentlev = 0) const
    {
      if (pmap.get() != 0)
        {
          ostrm << "{\n";
          pmap->format(ostrm, indentlev+1);
          for (int i = 0; i < indentlev; ++i) ostrm << '\t';
          ostrm << '}';
        }
      else
        {
          escapeWrite(ostrm, str);
        }
    }

    const char* getString() const { return str.c_str(); }

    double getDouble() const
    {
      std::istringstream iss(str);
      double val;
      iss >> val;
      return val;
    }

    int getInt() const
    {
      std::istringstream iss(str);
      int val;
      iss >> val;
      return val;
    }

    SharedPtr<ParamMap> getMap() const
    {
      if (pmap.get() == 0)
        {
          LERROR("No such parameter submap.");
          throw ParamMap::Error();
        }
      return pmap;
    }

    bool isLeaf() const
    {
      if (pmap.get() == 0)
        return true;
      else
        return false;
    }

    std::string str;
    mutable SharedPtr<ParamMap> pmap;
  };
}

// ######################################################################
struct ParamMap::Impl
{
  typedef std::map<std::string, Param> MapType;

  MapType itsParams;
};

// ######################################################################
struct ParamMap::key_iterator::IterRep
{
  ParamMap::Impl::MapType::const_iterator iter;
};

// ######################################################################
ParamMap::key_iterator::key_iterator() :
  rep(new IterRep)
{}

// ######################################################################
ParamMap::key_iterator::~key_iterator()
{
  delete rep;
}

// ######################################################################
ParamMap::key_iterator::key_iterator(const key_iterator& other) :
  rep(new IterRep)
{
  rep->iter = other.rep->iter;
}

// ######################################################################
ParamMap::key_iterator&
ParamMap::key_iterator::operator=(const key_iterator& other)
{
  rep->iter = other.rep->iter; return *this;
}

// ######################################################################
const char* ParamMap::key_iterator::operator*() const
{
  return (*(rep->iter)).first.c_str();
}

// ######################################################################
ParamMap::key_iterator&
ParamMap::key_iterator::operator++()
{
  ++(rep->iter); return *this;
}

// ######################################################################
bool ParamMap::key_iterator::operator==(const key_iterator& other) const
{
  return rep->iter == other.rep->iter;
}

// ######################################################################
bool ParamMap::key_iterator::operator!=(const key_iterator& other) const
{
  return rep->iter != other.rep->iter;
}

// ######################################################################
SharedPtr<ParamMap> ParamMap::loadPmapFile(const char* fname)
{
  SharedPtr<ParamMap> result(new ParamMap);
  result->load(fname);
  return result;
}

// ######################################################################
SharedPtr<ParamMap> ParamMap::loadConfFile(const char* fname)
{
  SharedPtr<ParamMap> pmap(new ParamMap);

  std::ifstream inFile(fname);

  int counter = 0;
  bool inComment = false;
  std::string instring;
  std::string paramName;
  while (inFile >> instring)
    {
      if (instring == "#")
        {
          // the comment code "#" toggles our in-comment state
          inComment = !inComment;
        }
      else if (!inComment) //real line found
        {
          if (paramName.length() == 0)
            {
              paramName.swap(instring);
            }
          else
            {
              pmap->putStringParam(paramName.c_str(), instring.c_str());
              LINFO("%s[%d]: %s = %s",
                    fname, counter, paramName.c_str(), instring.c_str());
              paramName.clear();
              ++counter;
            }
        }
    }

  return pmap;
}

// ######################################################################
ParamMap::ParamMap() :
  rep(new Impl)
{}

// ######################################################################
ParamMap::ParamMap(const char* fname) :
  rep(new Impl)
{
  load(fname);
}

// ######################################################################
ParamMap::~ParamMap()
{
  delete rep;
}

// ######################################################################
ParamMap::key_iterator
ParamMap::keys_begin() const
{
  key_iterator result;
  result.rep->iter = rep->itsParams.begin();
  return result;
}

// ######################################################################
ParamMap::key_iterator
ParamMap::keys_end() const
{
  key_iterator result;
  result.rep->iter = rep->itsParams.end();
  return result;
}

// ######################################################################
void ParamMap::load(std::istream& istrm)
{
  while (istrm.peek() != EOF)
    {
      if (istrm.peek() == '#')
        {
          istrm.ignore(std::numeric_limits<int>::max(), '\n');
          istrm >> std::ws;
          continue;
        }

      const std::string attribname = escapeRead(istrm);

      if (istrm.peek() != '{')
        {
          std::string attribval;
          while( istrm.peek() == '\t' || istrm.peek() == ' ' )
              istrm.get();
          if( istrm.peek() == '\n' )
              attribval = "";
          else
            attribval = escapeRead(istrm);

          rep->itsParams.insert(Impl::MapType::value_type(attribname,
                                                          attribval));
        }
      else
        {
          std::string attribval;

          int bracelevel = 0;
          // First time through loop we pick up the opening brace '{'
          do
            {
              int c = istrm.get();

              if (c == EOF)
                {
                  LERROR("Unexpected EOF before closing brace '}'.");
                  throw Error();
                }
              switch (c)
                {
                case '\\':
                  {
                    int cc = istrm.get();
                    if (cc == EOF)
                      {
                        LERROR("Unexpected EOF found after \\.");
                        throw Error();
                      }
                    attribval.push_back(char(cc));
                  }
                  break;
                case '{':
                  // Don't add the opening brace to the value string
                  if (bracelevel++ > 0)
                    attribval.push_back(char(c));
                  break;
                case '}':
                  // Don't add the closing brace to the value string
                  if (--bracelevel > 0)
                    attribval.push_back(char(c));
                  break;
                default:
                  attribval.push_back(char(c));
                  break;
                }
            }
          while (bracelevel > 0);

          SharedPtr<ParamMap> submap(new ParamMap);

          std::istringstream iss(attribval);
          submap->load(iss);

          rep->itsParams.insert(Impl::MapType::value_type(attribname, submap));
        }

      istrm >> std::ws;
    }
}

// ######################################################################
void ParamMap::load(const char* fname)
{
  std::ifstream ifs(fname);

  if (!ifs.is_open())
    {
      LERROR("Couldn't open file '%s' for reading.", fname);
      throw Error();
    }

  load(ifs);

  ifs.close();
}

// ######################################################################
void ParamMap::format(std::ostream& ostrm, int indentlev) const
{
  for (Impl::MapType::const_iterator
         itr = rep->itsParams.begin(), stop = rep->itsParams.end();
       itr != stop;
       ++itr)
    {
      for (int i = 0; i < indentlev; ++i) ostrm << '\t';
      ostrm << (*itr).first << "  ";
      (*itr).second.put(ostrm, indentlev);
      ostrm << '\n';
    }
}

// ######################################################################
void ParamMap::format(const char* fname) const
{
  std::ofstream ofs(fname);

  if (!ofs.is_open())
    {
      LERROR("Couldn't open file '%s' for writing.", fname);
      throw Error();
    }

  format(ofs);

  ofs.close();
}

// ######################################################################
bool ParamMap::hasParam(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  return (itr != rep->itsParams.end());
}

// ######################################################################
bool ParamMap::isLeaf(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if(itr->second.isLeaf())
    return true;
  else
    return false;
}

// ######################################################################
SharedPtr<ParamMap> ParamMap::getSubmap(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr == rep->itsParams.end())
    {
      LERROR("No parameter named '%s'.", paramname);
      throw Error();
    }
  return (*itr).second.getMap();
}

// ######################################################################
const char* ParamMap::getStringParam(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr == rep->itsParams.end())
    {
      LERROR("No parameter named '%s'.", paramname);
      throw Error();
    }
  return (*itr).second.getString();
}

// ######################################################################
double ParamMap::getDoubleParam(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr == rep->itsParams.end())
    {
      LERROR("No parameter named '%s'.", paramname);
      throw Error();
    }
  return (*itr).second.getDouble();
}

// ######################################################################
int ParamMap::getIntParam(const char* paramname) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr == rep->itsParams.end())
    {
      LERROR("No parameter named '%s'.", paramname);
      throw Error();
    }
  return (*itr).second.getInt();
}

// ######################################################################
const char* ParamMap::getStringParam(const char* paramname,
                                     const char* defval) const
{
  const char* result = defval;
  queryStringParam(paramname, result);
  return result;
}

// ######################################################################
double ParamMap::getDoubleParam(const char* paramname,
                                const double defval) const
{
  double result = defval;
  queryDoubleParam(paramname, result);
  return result;
}

// ######################################################################
int ParamMap::getIntParam(const char* paramname,
                          const int defval) const
{
  int result = defval;
  queryIntParam(paramname, result);
  return result;
}

// ######################################################################
SharedPtr<ParamMap> ParamMap::lookupSubmap(const char* paramname)
{
  if (hasParam(paramname))
    return getSubmap(paramname);

  // else...
  SharedPtr<ParamMap> submap(new ParamMap);
  putSubmap(paramname, submap);
  return submap;
}

// ######################################################################
ParamMap::ReturnCode
ParamMap::queryStringParam(const char* paramname, const char*& result) const
{
  ASSERT(result != 0);

  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr != rep->itsParams.end())
    {
      const char* new_val = (*itr).second.getString();
      if (strcmp(result, new_val) != 0)
        {
          result = new_val;
          return CHANGED;
        }
      else
        return UNCHANGED;
    }

  // else...
  LINFO("Parameter '%s' not found; using default value '%s'.",
        paramname, result);
  return MISSING;
}

// ######################################################################
ParamMap::ReturnCode
ParamMap::queryDoubleParam(const char* paramname, double& result) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr != rep->itsParams.end())
    {
      double new_val = (*itr).second.getDouble();
      if (new_val != result)
        {
          result = new_val;
          return CHANGED;
        }
      else
        return UNCHANGED;
    }

  // else...
  LINFO("Parameter '%s' not found; using default value '%f'.",
        paramname, result);
  return MISSING;
}

// ######################################################################
ParamMap::ReturnCode
ParamMap::queryIntParam(const char* paramname, int& result) const
{
  Impl::MapType::const_iterator itr = rep->itsParams.find(paramname);
  if (itr != rep->itsParams.end())
    {
      int new_val = (*itr).second.getInt();
      if (new_val != result)
        {
          result = new_val;
          return CHANGED;
        }
      else
        return UNCHANGED;
    }

  // else...
  LINFO("Parameter '%s' not found; using default value '%d'.",
        paramname, result);
  return MISSING;
}

// ######################################################################
void ParamMap::putSubmap(const char* paramname, const SharedPtr<ParamMap>& val)
{
  rep->itsParams.insert(Impl::MapType::value_type(paramname, val));
}

// ######################################################################
void ParamMap::putStringParam(const char* paramname, const char* val)
{
  rep->itsParams.insert(Impl::MapType::value_type(paramname,
                                                  std::string(val)));
}

// ######################################################################
void ParamMap::putDoubleParam(const char* paramname, double val)
{
  rep->itsParams.insert(Impl::MapType::value_type(paramname, val));
}

// ######################################################################
void ParamMap::putIntParam(const char* paramname, int val)
{
  rep->itsParams.insert(Impl::MapType::value_type(paramname, val));
}

// ######################################################################
void ParamMap::replaceStringParam(const char* paramname, const char* val)
{
   rep->itsParams.erase(rep->itsParams.find(paramname));
   rep->itsParams.insert(Impl::MapType::value_type(paramname, 
                                                   std::string(val)));   
}

// ######################################################################
void ParamMap::replaceDoubleParam(const char* paramname, double val)
{
   rep->itsParams.erase(rep->itsParams.find(paramname));
   rep->itsParams.insert(Impl::MapType::value_type(paramname, val));
}

// ######################################################################
void ParamMap::replaceIntParam(const char* paramname, int val)
{
   rep->itsParams.erase(rep->itsParams.find(paramname));
   rep->itsParams.insert(Impl::MapType::value_type(paramname, val));   
}

// ######################################################################
void ParamMap::clear()
{
   rep->itsParams.clear();
}

// ######################################################################
void ParamMap::print(const char* name) const
{
  std::cout << "--- begin " << name << " ---\n";
  for (Impl::MapType::const_iterator
         itr = rep->itsParams.begin(), stop = rep->itsParams.end();
       itr != stop;
       ++itr)
    {
      std::cout << "name: " << (*itr).first <<"****" <<std::endl;
      std::cout << "val: " << (*itr).second.str <<"****"<< std::endl;
    }
  std::cout << "--- end " << name << " ---\n";
}

// ######################################################################
void ParamMap::getAllSubmaps (const char* prefix,
                              std::vector <SharedPtr<ParamMap> >& submaps,
                              int& min_ctr,  int& max_ctr, 
                              int& min_surr, int& max_surr)
{
  Impl::MapType::const_iterator itr = rep->itsParams.begin();
  Impl::MapType::const_iterator itr_end = rep->itsParams.end();

  for ( ; itr != itr_end; ++itr)
    {
      const char* name = itr->first.c_str();
      // does this submap match the prefix?
      if(itr->first.find(prefix) == 0)
        {
          // match found
          SharedPtr<ParamMap> submap = getSubmap(name);
          submaps.push_back(submap);

          // find the min and max ctr, surr of the "coeff" submaps
          // (this is used by ObjectBase)
          if(strcmp(prefix, "coeff") == 0)
            {
              int c = -1, s = -1; sscanf(name, "coeff(%d,%d)", &c, &s);
              if(min_ctr == -1 || min_ctr >= c && min_surr > s)
                {
                  min_ctr = c; 
                  min_surr = s;
                }
              else if(max_ctr == -1 || (c >= max_ctr) && (s > max_surr))
                {
                  max_ctr = c;
                  max_surr = s;
                }
            }
        }
      else
        {
          // match not found
          if (itr->second.isLeaf() == true)
            continue;  // goto the next submap
          else
            {
              // not a leaf, hence recurse
              SharedPtr<ParamMap> submap = getSubmap(name);
              submap->getAllSubmaps(prefix, submaps, 
                                    min_ctr, max_ctr, min_surr, max_surr);
           }
        }
    }
  return;
}
// ######################################################################
void ParamMap::setAllSubmaps (const char* prefix, int& idx,
                              std::vector <SharedPtr<ParamMap> >& submaps)
{
  Impl::MapType::const_iterator itr = rep->itsParams.begin();
  Impl::MapType::const_iterator itr_end = rep->itsParams.end();
  for ( ; itr != itr_end; ++itr)
    {
      const char* name = itr->first.c_str();
      // check prefix
      if(itr->first.find(prefix) == 0)
        {
          // match found
          //LDEBUG("match found --- %s\n",name);
          rep->itsParams.erase(itr->first.c_str());
          putSubmap(name, submaps[idx]);
          idx++;
        }
      else
        {
          // match not found, recurse
          //LDEBUG("match not found --- %s\n",name);
          if (itr->second.isLeaf() == true)
            continue;
          else
            {
              // not a leaf, hence recurse
               SharedPtr<ParamMap> submap = getSubmap(name);
               submap->setAllSubmaps(prefix, idx, submaps);
            }
        }

    }
  return;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !PARAMMAP_C_DEFINED
