/*!@file PbmParser.C Parse pbm/pgm/ppm image files. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: PbmParser.C 3353 2004-06-15 08:40:39Z itti $
//

#include "PbmParser.H"

#include "Image.H"
#include "Image_ColorOps.H"
#include "log.H"

#include <fstream>
#include <limits>
#include <string>

// ######################################################################
struct PbmParser::Rep
{
  Rep(const char* fname) :
    strm(fname), mode(-1), w(-1), h(-1), maxGrey(1), comments("")
  {}

  std::ifstream strm;
  int mode;
  int w, h;
  int maxGrey;
  std::string comments;
};

// ######################################################################
PbmParser::PbmParser(const char* fname) :
  rep(new Rep(fname))
{
  if (!rep->strm.is_open())
    LFATAL("Couldn't open file '%s' for reading.", fname);

  int c = rep->strm.get();
  if (c != 'P')
    LFATAL("Missing magic number in pbm file '%s'.", fname);

  rep->strm >> rep->mode;

  rep->strm >> std::ws;
  // copy and concatenate optional comment line(s) starting with '#'
  // into comments string

  std::string temp;

  while ( rep->strm.peek() == '#' )
    {
      //get the full line
      std::getline( rep->strm, temp, '\n');
      rep->comments += temp;

      //This is what we did before we were interested in the comments
      //rep->strm.ignore(std::numeric_limits<int>::max(), '\n');
    }

  rep->strm >> rep->w >> rep->h;

  if (rep->mode != 1 && rep->mode != 4)
    {
      rep->strm >> rep->maxGrey;
    }

  // read one more character of whitespace from the stream after maxGrey
  c = rep->strm.get();
  if ( !isspace(c) )
    LFATAL("Missing whitespace after maxGrey in pbm file '%s'.", fname);
LDEBUG("PBM Reading RGB Image: %s", fname);
}

// ######################################################################
PbmParser::~PbmParser()
{
  delete rep;
}

// ######################################################################
bool PbmParser::isBitmap() const
{ return (rep->mode == 1 || rep->mode == 4); }

// ######################################################################
bool PbmParser::isGray() const
{ return (rep->mode == 2 || rep->mode == 5); }

// ######################################################################
bool PbmParser::isRGB() const
{ return (rep->mode == 3 || rep->mode == 6); }

// ######################################################################
Dims PbmParser::getDims() const
{ return Dims(rep->w, rep->h); }

// ######################################################################
std::string PbmParser::getComments() const
{ return rep->comments; }

// ######################################################################
Image<PixRGB<byte> > PbmParser::getRGB()
{
  if      (isBitmap())  LFATAL("Pbm mode '%d' not supported", rep->mode);
  else if (isGray())    return toRGB(parseGray());
  /* else... */         return parseRGB();
}

// ######################################################################
Image<byte> PbmParser::getGray()
{
  if      (isBitmap())  LFATAL("Pbm mode '%d' not supported", rep->mode);
  else if (isGray())    return parseGray();
  /* else... */         return luminance(parseRGB());
}

// ######################################################################
Image<float> PbmParser::getFloat()
{
  if      (isBitmap())  LFATAL("Pbm mode '%d' not supported", rep->mode);
  else if (isGray())    return Image<float>(parseGray());
  /* else... */         return Image<float>(luminance(parseRGB()));
}

// ######################################################################
void PbmParser::readBytes(byte* space, int byte_count)
{
  ASSERT(rep->mode == 2 || rep->mode == 3 || rep->mode == 5 || rep->mode == 6);

  if (rep->mode == 2 || rep->mode == 3)
    {
      // Parse ascii:
      const double scale = 255.0/double(rep->maxGrey);

      int position = 0;
      int val = 0;

      while (rep->strm.peek() != EOF && position < byte_count)
        {
          rep->strm >> val;
          space[position++] = static_cast<byte>(val * scale);
        }
    }
  else
    {
      // Parse raw:
      rep->strm.read(reinterpret_cast<char*>(space), byte_count);
    }
}

// ######################################################################
Image<byte> PbmParser::parseGray()
{
  ASSERT(rep->mode == 2 || rep->mode == 5);

  Image<byte> img(rep->w, rep->h, NO_INIT);
  readBytes(img.getArrayPtr(), img.getSize());
  return img;
}

// ######################################################################
Image<PixRGB<byte> > PbmParser::parseRGB()
{
  ASSERT(rep->mode == 3 || rep->mode == 6);

  Image<PixRGB<byte> > img(rep->w, rep->h, NO_INIT);
  readBytes(reinterpret_cast<byte*>(img.getArrayPtr()), img.getSize()*3);
  return img;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
