/*!@file PfmParser.C Parse pfm image files. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <Itti@usc.edu>
// $Id: PfmParser.C 3353 2004-06-15 08:40:39Z itti $
//

#include "PfmParser.H"

#include "Image.H"
#include "Image_ColorOps.H"
#include "log.H"

#include <fstream>
#include <limits>
#include <string>

// ######################################################################
struct PfmParser::Rep
{
  Rep(const char* fname) :
    strm(fname), mode('\0'), w(-1), h(-1), maxGrey(1), comments("")
  {}

  std::ifstream strm;
  char mode;
  int w, h;
  float maxGrey;
  std::string comments;
};

// ######################################################################
PfmParser::PfmParser(const char* fname) :
  rep(new Rep(fname))
{
  if (!rep->strm.is_open())
    LFATAL("Couldn't open file '%s' for reading.", fname);

  int c = rep->strm.get();
  if (c != 'P')
    LFATAL("Missing magic number in pbm file '%s'.", fname);

  rep->strm >> rep->mode;

  rep->strm >> std::ws;
  // copy and concatenate optional comment line(s) starting with '#'
  // into comments string

  std::string temp;

  while ( rep->strm.peek() == '#' )
    {
      //get the full line
      std::getline( rep->strm, temp, '\n');
      rep->comments += temp;

      //This is what we did before we were interested in the comments
      //rep->strm.ignore(std::numeric_limits<int>::max(), '\n');
    }

  rep->strm >> rep->w >> rep->h;

  if (rep->mode != 1 && rep->mode != 4)
    {
      rep->strm >> rep->maxGrey;
    }

  // read one more character of whitespace from the stream after maxGrey
  c = rep->strm.get();
  if ( !isspace(c) )
    LFATAL("Missing whitespace after maxGrey in pfm file '%s'.", fname);
  LDEBUG("PFM Reading Image: %s", fname);
}

// ######################################################################
PfmParser::~PfmParser()
{
  delete rep;
}

// ######################################################################
bool PfmParser::isBitmap() const
{ return false; }

// ######################################################################
bool PfmParser::isGray() const
{ return (rep->mode == 'F'); }

// ######################################################################
bool PfmParser::isRGB() const
{ return false; }

// ######################################################################
Dims PfmParser::getDims() const
{ return Dims(rep->w, rep->h); }

// ######################################################################
std::string PfmParser::getComments() const
{ return rep->comments; }

// ######################################################################
Image<PixRGB<byte> > PfmParser::getRGB()
{ return toRGB(Image<byte>(parseFloat())); }

// ######################################################################
Image<byte> PfmParser::getGray()
{ return Image<byte>(parseFloat()); }

// ######################################################################
Image<float> PfmParser::getFloat()
{ return parseFloat(); }

// ######################################################################
void PfmParser::readFloats(float* space, int count)
{
  ASSERT(rep->mode == 'F');
  rep->strm.read(reinterpret_cast<char *>(space), count * sizeof(float));
}

// ######################################################################
Image<float> PfmParser::parseFloat()
{
  ASSERT(rep->mode == 'F');

  Image<float> img(rep->w, rep->h, NO_INIT);
  readFloats(img.getArrayPtr(), img.getSize());
  return img;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
