/*!@file PfmParser.H Parse pfm image files. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: PfmParser.H 3353 2004-06-15 08:40:39Z itti $
//

#ifndef PFMPARSER_H_DEFINED
#define PFMPARSER_H_DEFINED

#include "saliency.H"
#include "ArrayData.H"

#include <string>

template <class T> class Image;
template <class T> class PixRGB;

// ######################################################################
//! This class handles reading/parsing of PFM image files.
/*! PFM is an unofficial format that looks very much like PGM except
  that raw pixel values are floats. Caution, this is not portable
  across endianness. The PNM magic for this unofficial format is
  PF. */
class PfmParser
{
public:
  //! Initialize with a PGM or PPM file (either is OK).
  PfmParser(const char* fname);

  //! Destructor.
  ~PfmParser();

  //! Query whether the file is a black+white bitmap.
  bool isBitmap() const;

  //! Query whether the file is grayscale.
  bool isGray() const;

  //! Query whether the file is RGB.
  bool isRGB() const;

  //! Get the dims of the image without fully reading it
  Dims getDims() const;

  //! Get the comments from the image without fully reading it
  std::string getComments() const;

  //! Get an RGB image from the file.
  /*! This works regardless of the actual mode of the underlying file
    (i.e., if the file was grayscale, this will convert that to RGB). */
  Image<PixRGB<byte> > getRGB();

  //! Get a grayscale image from the file.
  /*! This works regardless of the actual mode of the underlying file
    (i.e., if the file was RGB, getGray() the returns luminance of that). */
  Image<byte> getGray();

  //! Get a float grayscale image from the file.
  /*! This works regardless of the actual mode of the underlying file
    (i.e., if the file was RGB, getFloat() the returns luminance of that). */
  Image<float> getFloat();

private:
  //! Read the actual bytes out of the file.
  /*! This works for any supported mode or file type. */
  void readFloats(float* space, int count);

  //! Parse a float grayscale image
  Image<float> parseFloat();

  PfmParser(const PfmParser&);
  PfmParser& operator=(const PfmParser&);

  struct Rep;
  Rep* const rep;
};

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !PBMPARSER_H_DEFINED
