/*!@file Pixels.H Basic pixel types */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Pixels.H 4290 2005-03-12 02:03:27Z mundhenk $
//

#ifndef PIXEL_H_DEFINED
#define PIXEL_H_DEFINED

#define DEFHTYPE 2

#include "saliency.H"
#include "TypeTraits.H"
#include "log.H"
#include "Promotions.H"
#include <cmath>

template <class T> class PixRGB;

template <class T> class PixH2SV;

//! Specialize TypeTraits to indicate that PixRGB is a "trivial" type.
/*! That is, it is a type that can be safely initialized with memset(), and
    safely copied with memcpy(). In practical terms, this means we're
    relying on PixRGB not to do anything significant in its copy
    constructor or destructor. */
template <class T> struct TypeTraits<PixRGB<T> >
{
  enum { isTrivial = TypeTraits<T>::isTrivial };
};

template <class T> struct TypeTraits<PixH2SV<T> >
{
  enum { isTrivial = TypeTraits<T>::isTrivial };
};

//! Generic pixel class for color RGB pixels
/*! These definitions are for color pixels. The pixels initialize to
  black through the default constructor. The following other methods
  are required if those pixels are to be used for methods available for
  color images (see Image.H):

  - constructor(red, green, blue) - construct from an RGB triplet
  - constructor(lum) - construct gray pixel(lum, lum, lum)
  - luminance() - return luminance value
  - red(), green(), blue() - return respective component values
  - setRed(), setGreen(), setBlue() - set respective component values

  WARNING: you should NOT define comparison operators for color pixels,
  otherwise you will get the impression that many Image methods work
  while they will actually be bogus... */

template <class T> class PixRGB {
public:
  //! Construct an RGB pixel, black
  inline PixRGB();

  //! Construct an RGB pixel with given values red, green, blue
  inline PixRGB(const T redval, const T greenval, const T blueval);

  //! Construct from given values, clamping & converting as needed
  template <class T2, class T3, class T4> inline
  PixRGB(const T2 redval, const T3 greenval, const T4 blueval);

  //! Construct an RGB pixel of luminance lum
  inline PixRGB(const T lum);

  //! Construct an RGB pixel of luminance lum, clamping & converting as needed
  template <class T2> inline PixRGB(const T2 lum);

  //! Construct an RGB pixel from another
  inline PixRGB(const PixRGB<T>& A);

  //! Contructor from another PixRGB with possibly different base type
  template <class T2> inline PixRGB(const PixRGB<T2>& A);

  //! Construct an RGB pixel equal to another
  inline PixRGB<T>& operator=(const PixRGB<T>& A);

  //! Construct an RGB pixel equal to another of possibly different base type
  template <class T2> inline PixRGB<T>& operator=(const PixRGB<T2>& A);  

  //! Construct an RGB pixel equal to another
  inline PixRGB<T>& operator=(const PixH2SV<T>& A);

  //! Construct an RGB pixel equal to another of possibly different base type
  template <class T2> inline PixRGB<T>& operator=(const PixH2SV<T2>& A);

  //! Compute luminance as (r + g + b) / 3
  inline T luminance() const;

  //! Access red channel
  inline T red() const;

  //! Access green channel
  inline T green() const;

  //! Access blue channel
  inline T blue() const;

  //! Set red channel value, clamping & converting as necessary
  template <class T2> inline void setRed(const T2 redval);

  //! Set green channel value, clamping & converting as necessary
  template <class T2> inline void setGreen(const T2 greenval);

  //! Set blue channel value, clamping & converting as necessary
  template <class T2> inline void setBlue(const T2 blueval);

  //! Set all channel values, clamping & converting as necessary
  template <class T2, class T3, class T4> inline void
  set(const T2 redval, const T3 greenval, const T4 blueval);

  //! PixRGB equality operator
  inline bool operator==(const PixRGB<T>& other) const;

  //! PixRGB inequality operator
  inline bool operator!=(const PixRGB<T>& other) const;

  //! Add a value to all components, clamp result as necessary
  template <class T2> inline
  PixRGB<T>& operator+=(const T2 val);

  //! Add component-by-component, clamp result as necessary
  template <class T2> inline
  PixRGB<T>& operator+=(const PixRGB<T2>& A);

  //! Add component-by-component, clamp result as necessary
  template <class T2> inline
  PixRGB<T>& operator+=(const PixH2SV<T2>& A);
  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator-=(const T2 val);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator-=(const PixRGB<T2>& A); 

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator-=(const PixH2SV<T2>& A);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator*=(const T2 val);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator*=(const PixRGB<T2>& A);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator*=(const PixH2SV<T2>& A);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator/=(const T2 val);

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator/=(const PixRGB<T2>& A); 

  //! PixRGB operator
  template <class T2> inline
  PixRGB<T>& operator/=(const PixH2SV<T2>& A);

  //! Add component-by-component, promote & clamp as necessary
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator+(const PixRGB<T2>& B) const;

  //! Add component-by-component, promote & clamp as necessary
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator+(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator+(const T2 val) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator-(const PixRGB<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator-(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator-(const T2 val) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator*(const PixRGB<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator*(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator*(const T2 val) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator/(const PixRGB<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator/(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixRGB<typename promote_trait<T,T2>::TP> operator/(const T2 val) const;

  //! This nonsense comparison operator will abort the program when called.
  /*! It is implemented only to keep the compiler happy by allowing it
    to instantiate some methods that use comparisons but should never
    be used on color pixels. For instance, this is necessary for
    centerSurround() if we want to be able to use color pyramids. */
  inline bool operator>(const PixRGB<T>& B) const;

  //! This nonsense comparison operator will abort the program when called.
  /*! It is implemented only to keep the compiler happy by allowing it
    to instantiate some methods that use comparisons but should never
    be used on color pixels. For instance, this is necessary for
    centerSurround() if we want to be able to use color pyramids. */
  inline bool operator<(const PixRGB<T>& B) const;

  //! get pixel that has the max elements of both bixels 
  inline PixRGB<T> takeMax(const PixRGB<T>& B);

  // ##### color conversion functions:

  //! Color conversion for YIQ
  inline void getYIQ(T& y, T& i, T& q) const;

  //! Color conversion for YUV
  inline void getYUV(T& y, T& u, T& v) const;

  //! Color conversion for HSV
  inline void getHSV(T& h, T& s, T& v) const; 

  //! Color conversion for HSV, old algo
  inline void getHSVOld(T& h, T& s, T& v) const;

  //! return HSV but decompose the modulo value of hue into h1 and h2
  /*! Note: this uses standard p1 radial decomposition where
    Lp = (|x|^p + |y|^p)^(1/p) thus, the shape is diamond and
    not circular. This maintains linear scaling. Also, h1 and h2
    are normalized from 0 to 1
  */

  inline void getHSV(T& h1, T& h2, T& s, T& v) const;

  //! Color conversion for YUV
  inline void setYUV(const T y, const T u, const T v);

  //! Color conversion for HSV
  inline void setHSV(const T h, const T s, const T v);

  //! Color conversion for HSV
  inline void setHSV(const T h1, const T h2, const T s, const T v);

public:
  // Keep the data members public for run-time efficiency, so that we
  // can access data members directly without having to rely on the
  // compiler optimizing away inline function calls. Example:
  // lowPass3x() and lowPass3y() were spending roughly half their CPU
  // time in PixRGB::operator+() and PixRGB::operator*(). When these
  // calls were replaced with a lowpass implementation that just
  // accesses PixRGB's data members directly, the total CPU time spent
  // in lowPass3() dropped by more than 50%. This is significant,
  // since lowPass3() is often where most of the CPU time is spent
  // overall in any given salience computation.
  T r, g, b;
};

// ########## free functions for PixRGB:

//! take square root of r, g, and b:
template <class T> inline
PixRGB<T> sqrt(const PixRGB<T> &x);

//! Check that all three of r, g, and b have finite values
template <class T> inline
bool isFinite(const PixRGB<T> &x);



// ######################################################################
// ##### INLINE FUNCTIONS FOR PixRGB:
// ######################################################################

template <class T> inline
PixRGB<T>::PixRGB() : r(), g(), b()
{ }

// ######################################################################
template <class T> inline
PixRGB<T>::PixRGB(const T redval, const T greenval, const T blueval)
{ r = redval; g = greenval; b = blueval; }

// ######################################################################
template <class T> template <class T2, class T3, class T4> inline
PixRGB<T>::PixRGB(const T2 redval, const T3 greenval, const T4 blueval)
{
  r = clamped_convert<T>(redval);
  g = clamped_convert<T>(greenval);
  b = clamped_convert<T>(blueval);
}

// ######################################################################
template <class T> inline
PixRGB<T>::PixRGB(const T lum)
{ r = lum; g = lum; b = lum; }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>::PixRGB(const T2 lum)
{ r = clamped_convert<T>(lum); g = r; b = r; }

// ######################################################################
template <class T> inline
PixRGB<T>::PixRGB(const PixRGB<T>& A)
{ r = A.r; g = A.g; b = A.b; }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>::PixRGB(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(A.red());
  g = clamped_convert<T>(A.green());
  b = clamped_convert<T>(A.blue());
}

// ######################################################################
template <class T> inline
PixRGB<T>& PixRGB<T>::operator=(const PixRGB<T>& A)
{ r = A.r; g = A.g; b = A.b; return *this; }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator=(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(A.red());
  g = clamped_convert<T>(A.green());
  b = clamped_convert<T>(A.blue());
  return *this;
}

// ######################################################################
template <class T> inline
PixRGB<T>& PixRGB<T>::operator=(const PixH2SV<T>& A)
{
  return A.getRGB();
}

// ######################################################################

template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator=(const PixH2SV<T2>& A)
{
  PixH2SV<T> B = A;
  return B.getRGB();
}

// ######################################################################
template <class T> inline
T PixRGB<T>::luminance() const
{ return clamped_convert<T>( (float(r) + float(g) + float(b)) / 3.0f ); }

// ######################################################################
template <class T> inline
T PixRGB<T>::red() const
{ return r; }

// ######################################################################
template <class T> inline
T PixRGB<T>::green() const
{ return g; }

// ######################################################################
template <class T> inline
T PixRGB<T>::blue() const
{ return b; }

// ######################################################################
template <class T> template <class T2> inline
void PixRGB<T>::setRed(const T2 redval)
{ r = clamped_convert<T>(redval); }

// ######################################################################
template <class T> template <class T2> inline
void PixRGB<T>::setGreen(const T2 greenval)
{ g = clamped_convert<T>(greenval); }

// ######################################################################
template <class T> template <class T2> inline
void PixRGB<T>::setBlue(const T2 blueval)
{ b = clamped_convert<T>(blueval); }

// ######################################################################
template <class T> template <class T2, class T3, class T4> inline
void PixRGB<T>::set(const T2 redval, const T3 greenval, const T4 blueval)
{
  r = clamped_convert<T>(redval);
  g = clamped_convert<T>(greenval);
  b = clamped_convert<T>(blueval);
}

// ######################################################################
template <class T> inline
bool PixRGB<T>::operator==(const PixRGB<T>& x) const
{ return ( r == x.r && g == x.g && b == x.b ); }

// ######################################################################
template <class T> inline
bool PixRGB<T>::operator!=(const PixRGB<T>& x) const
{ return ( r != x.r || g != x.g || b != x.b ); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator+=(const T2 val)
{
  r = clamped_convert<T>(r + val);
  g = clamped_convert<T>(g + val);
  b = clamped_convert<T>(b + val);
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator+=(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(r + A.red());
  g = clamped_convert<T>(g + A.green());
  b = clamped_convert<T>(b + A.blue());
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator+=(const PixH2SV<T2>& A)
{
  PixH2SV<T> B = A;
  PixRGB<T>  C = B;
  C = C + *this;
  return C;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator-=(const T2 val)
{
  r = clamped_convert<T>(r - val);
  g = clamped_convert<T>(g - val);
  b = clamped_convert<T>(b - val);
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator-=(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(r - A.red());
  g = clamped_convert<T>(g - A.green());
  b = clamped_convert<T>(b - A.blue());
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator-=(const PixH2SV<T2>& A)
{
  PixH2SV<T> B = A;
  PixRGB<T>  C = B;
  C = *this - C;
  return C;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator*=(const T2 val)
{
  r = clamped_convert<T>(r * val);
  g = clamped_convert<T>(g * val);
  b = clamped_convert<T>(b * val);
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator*=(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(r * A.red());
  g = clamped_convert<T>(g * A.green());
  b = clamped_convert<T>(b * A.blue());
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator*=(const PixH2SV<T2>& A)
{
  PixH2SV<T> B = A;
  PixRGB<T>  C = B;
  C = C * (*this);
  return C;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator/=(const T2 val)
{
  r = clamped_convert<T>(r / val);
  g = clamped_convert<T>(g / val);
  b = clamped_convert<T>(b / val);
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator/=(const PixRGB<T2>& A)
{
  r = clamped_convert<T>(r / A.red());
  g = clamped_convert<T>(g / A.green());
  b = clamped_convert<T>(b / A.blue());
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<T>& PixRGB<T>::operator/=(const PixH2SV<T2>& A)
{
  PixH2SV<T> B = A;
  PixRGB<T>  C = B;
  C = *this/C;
  return C;
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator+(const PixRGB<T2>& B) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r+B.red(), g+B.green(), b+B.blue()); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator+(const PixH2SV<T2>& B) const
{
  PixH2SV<T> C = B;
  PixRGB<T>  D = C;
  return PixRGB<typename promote_trait<T,T2>::TP>(r+D.red(), g+D.green(), b+D.blue()); 
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator+(const T2 val) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r + val, g + val, b + val); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator-(const PixRGB<T2>& B) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r-B.red(), g-B.green(), b-B.blue()); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator-(const PixH2SV<T2>& B) const
{
  PixH2SV<T> C = B;
  PixRGB<T>  D = C;
  return PixRGB<typename promote_trait<T,T2>::TP>(r-D.red(), g-D.green(), b-D.blue()); 
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator-(const T2 val) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r - val, g - val, b - val); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator*(const PixRGB<T2>& B) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r*B.red(), g*B.green(), b*B.blue()); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator*(const PixH2SV<T2>& B) const
{ 
  PixH2SV<T> C = B;
  PixRGB<T>  D = C;
  return PixRGB<typename promote_trait<T,T2>::TP>(r*D.red(), g*D.green(), b*D.blue()); 
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP> PixRGB<T>::operator*(const T2 val)
  const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r * val, g * val, b * val); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP>
PixRGB<T>::operator/(const PixRGB<T2>& B) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r/B.red(),g/B.green(),b/B.blue()); }

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP>
PixRGB<T>::operator/(const PixH2SV<T2>& B) const
{
  PixH2SV<T> C = B;
  PixRGB<T>  D = C;
  return PixRGB<typename promote_trait<T,T2>::TP>(r/D.red(), g/D.green(), b/D.blue()); 
}

// ######################################################################
template <class T> template <class T2> inline
PixRGB<typename promote_trait<T,T2>::TP>
PixRGB<T>::operator/(const T2 val) const
{ return PixRGB<typename promote_trait<T,T2>::TP>(r / val, g / val, b / val); }

// ######################################################################
template <class T> inline
bool PixRGB<T>::operator>(const PixRGB<T>& B) const
{
  LFATAL("Operator > does not make sense on color pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
bool PixRGB<T>::operator<(const PixRGB<T>& B) const
{
  LFATAL("Operator < does not make sense on color pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
PixRGB<T> PixRGB<T>::takeMax(const PixRGB<T>& B)
{
  if(B.r > r) r = B.r;
  if(B.g > g) g = B.g;
  if(B.b > b) b = B.b;
  return *this;
}

// ######################################################################
template <class T> inline
void PixRGB<T>::getYIQ(T& y, T& i, T& q) const
{
  // coeffs from Foley & Van Dam
  y = T(0.299f * r + 0.587f * g + 0.114f * b);
  i = T(0.596f * r - 0.275f * g - 0.321f * b);
  q = T(0.212f * r - 0.523f * g + 0.311f * b);
}

// ######################################################################
template <class T> inline
void PixRGB<T>::getYUV(T& y, T& u, T& v) const
{
  float yf, uf, vf;
  yf =  0.299f   * r + 0.587f   * g + 0.114f   * b; // coeffs from PVRG-JPEG
  uf = -0.16874f * r - 0.33126f * g + 0.500f   * b + 128.0f;
  vf =  0.500f   * r - 0.41869f * g - 0.08131f * b + 128.0f;
  if (uf > 255.0f) uf = 255.0f; if (vf > 255.0f) vf = 255.0f;
  y = (T)yf; u = (T)uf; v = (T)vf;
}

// ######################################################################
template <class T> inline
void PixRGB<T>::getHSVOld(T& h, T& s, T& v) const
{
  T min = (T)0.0F, delta;
  if((r >= g) && (r >= b)) v = r;
  if((g >= r) && (g >= b)) v = g;
  if((b >= g) && (b >= r)) v = b;
  if((r <= g) && (r <= b)) min = r;
  if((g <= r) && (g <= b)) min = g;
  if((b <= g) && (b <= r)) min = b;
  delta = v - min;
  if(delta == (T)0.0F) delta = (T)1.0F;
  if (v){ s = delta/v;}
  else { s = 0; h = -1; return; }
  if(r == v){ h = (g - b) / delta;}
  else if(g == v){ h = 2 + (b - r) / delta;}
  else { h = 4 + (r - g) / delta;}
  h *= (T)60; if(h < (T)0) h += (T)360;
}

// ######################################################################
template <class T> inline
void PixRGB<T>::getHSV(T& h, T& s, T& v) const
{
  // this method is a bit verbose, but it has been unrolled a little 
  // and inlined so make sure it runs efficently
  T min = 0.0F, delta;
  if((r >= g) && (r >= b))      // SET value = red
  {    
    v = r;
    if((g <= r) && (g <= b))      min = g;
    else if((b <= g) && (b <= r)) min = b;
    delta = v - min;            // delta is max - min 
    if(delta == 0.0F) delta = 1.0F; 
    if  (v){ s = delta/v;}      // SET sat = (v-min)/v
    else   { s = 0; h = -1; return; }
    h =     (g - b) / delta;    // SET hue = green - blue / delta
    h *= 60; if(h < 0) h += 360;
    return;
  }
  else if((g >= r) && (g >= b)) // SET value = green
  {    
    v = g;
    if((r <= g) && (r <= b))      min = r;
    else if((b <= g) && (b <= r)) min = b;
    delta = v - min;           // delta is max - min 
    if(delta == 0.0F) delta = 1.0F;
    if  (v){ s = delta/v;}     // SET sat = (v-min)/v
    else   { s = 0; h = -1; return; }
    h = 2 + (b - r) / delta;   // SET hue = 2 + blue - red  / delta 
    h *= 60; if(h < 0) h += 360;
    return;
  }
  else if((b >= g) && (b >= r)) // SET value = blue
  {
    v = b;
    if((r <= g) && (r <= b))      min = r;
    else if((g <= r) && (g <= b)) min = g; 
    delta = v - min;          // delta is max - min 
    if(delta == 0.0F) delta = 1.0F;
    if  (v){ s = delta/v;}    // SET sat = (v-min)/v
    else   { s = 0; h = -1; return; }
    h = 4 + (r - g) / delta;  // SET hue = 4 + red - green / delta 
    h *= 60; if(h < 0) h += 360;
    return;
  }
}

// ######################################################################
// return HSV with modulo decompisition over hue
template <class T> inline
void PixRGB<T>::getHSV(T& h1, T& h2, T& s, T& v) const
{
  T h;
  getHSV(h,s,v);
  // demoduloize hue into two components
  if(DEFHTYPE == 1)
  {
    if((h >= 0) && (h <= 90))
      { h1 = .5 + h/180; h2 = 1 - h/180;}
    else if((h > 90) && (h <= 180))
      { h1 = 1 - (h - 90)/180; h2 = 1 - h/180;}
    else if((h > 180) && (h <= 270))
      { h1 = 1 - (h - 90)/180; h2 = (h - 180)/180;}
    else if((h > 270) && (h <= 360))
      { h1 = (h - 270)/180; h2 = (h - 180)/180;}
    else // undefined color, set to 0
      { h1 = .5; h2 = 1;}
  }
  else
  {
    if((h >= 0) && (h <= 60))
      { h1 = 0.66667 + h/180;      h2 = 1 - h/120; }
    else if((h > 60) && (h <= 120))
      { h1 = 1 - (h - 60)/180; h2 = 1 - h/120; }
    else if((h > 120) && (h <= 240))
      { h1 = 1 - (h - 60)/180; h2 = (h - 120)/240; }
    else if((h > 240) && (h <= 360))
      { h1 = (h - 240)/180;    h2 = (h - 120)/240; }
    else
      { h1 = 0.66667; h2 = 1; }
  }
}

// ######################################################################
template <class T> inline
void PixRGB<T>::setYUV(const T y, const T u, const T v)
{
  float yf = y;
  float uf = u - 128.0f;
  float vf = v - 128.0f;
  float rf = yf                 + 1.402f   * vf; // coeffs from PVRG-JPEG
  float gf = yf - 0.34414f * uf - 0.71414f * vf;
  float bf = yf + 1.772f   * uf;
  if (rf < 0.0f) rf = 0.0f; if (rf > 255.0f) rf = 255.0f;
  if (gf < 0.0f) gf = 0.0f; if (gf > 255.0f) gf = 255.0f;
  if (bf < 0.0f) bf = 0.0f; if (bf > 255.0f) bf = 255.0f;
  r = (T)rf; g = (T)gf; b = (T)bf;
}

// ######################################################################
template <class T> inline
void PixRGB<T>::setHSV(const T h, const T s, const T v)
{
  float vf = v / 256.0f;
  float sf = s;
  if( vf == 0.0F )
    { r = 0; g = 0; b = 0; }
  else if( sf == 0.0F )
    { r = v; g = v; b = v; }
  else
    {
      float hf = h / 60.0F;
      int i = (int) floor( hf );
      float f = hf - i;
      float p = v * ( 1 - sf );
      float q = v * ( 1 - sf * f );
      float t = v * ( 1 - sf * ( 1 - f ) );
      switch( i )
        {
        case 0:
          r = v; g = t; b = p; break;
        case 1:
          r = q; g = v; b = p; break;
        case 2:
          r = p; g = v; b = t; break;
        case 3:
          r = p; g = q; b = v; break;
        case 4:
          r = t; g = p; b = v; break;
        case 5:
          r = v; g = p; b = q; break;
        }
      r *= 256.0f; g *= 256.0f; b *= 256.0f;
    }
}

// ######################################################################
// Recompose hue from hue1 and hue2 and set color

template <class T> inline
void PixRGB<T>::setHSV(const T h1,const T h2, const T s, const T v)
{
  T h;
  if(DEFHTYPE == 1)
  {
    if((h1 > 0.5) && (h2 > 0.5))
      { h = 180*(h1-0.5);}
    else if((h1 > 0.5) && (h2 <= 0.5))
      { h = 90 + 180*(1-h1); }
    else if((h1 <= 0.5) && (h2 <= 0.5))
      { h = 90 + 180*(1-h1); }
    else if((h1 <= 0.5) && (h2 > 0.5))
      { h = 270 + 180*h1; }
  }
  else
  {
    if(h1 >= 0.66667)
      { h = 120 * (1 - h2) + 0 ;}
    else if(h1 <= 0.5)
      { h = 240 * h2 + 120; }
    else if((h1 > 0.5) && (h1 < 0.66667) && (h2 > h1))  
      { h = 180 * h1 + 240; }
    else if((h1 > 0.5) && (h1 < 0.66667) && (h2 < h1))
      { h = 180 * (1 - h1) + 60; }
  }
  setHSV(h,s,v);
}

// ########## free functions:

template <class T> inline
PixRGB<T> sqrt(const PixRGB<T> &x)
{
  return PixRGB<T>(clamped_convert<T>(std::sqrt((float)(x.red()))),
                   clamped_convert<T>(std::sqrt((float)(x.green()))),
                   clamped_convert<T>(std::sqrt((float)(x.blue()))));
}

template <class T>
bool isFinite(const PixRGB<T> &x)
{
  // we use the isFinite() functions defined in saliency.H
  return (isFinite(x.red()) && isFinite(x.green()) &&
          isFinite(x.blue()));
}


//***********************************************************************
//***********************************************************************
//***** H2SV Pixel Class
//***********************************************************************
//***********************************************************************

//! This is a Pixel class for H2SV (non-polar-coord HSV), should work similar to PixRGB
/*! Here we have 4 color components. All should be normalized between 0 to 1. The components are:
  H1 - This is a Yellow / Blue opponent color component. is 1 for yellow and 0 for blue
  H2 - This is a Red / Green opponent color component. is 1 for red and 0 fro green
  S  - This is the basic color saturation like HSV. Is 1 for total color saturation, 0 for none.
  V  - This is the "Value" or intensity the same as HSV. Is 1 for total color and 0 for black.
  
  There are oporators in this class to convert and opperate between H2SV and RGB easily.

  Also note: Htype defines the way H is converted to H1 and H2. It is an unsigned short
  1 - H1 is semetric with H2 and peaks at 90, H2 peaks at 0
  2 - RG/BY oponent colors with H1 and H2 defines slightly different to reflect this
  3 - HSV standard, H2 is left to 0 and H1 ranges from 0 to 1 as 0 to 360
*/

template <class T> class PixH2SV {
public:
  //! Construct an H2SV pixel, black
  inline PixH2SV();

  //! Construct an H2SV pixel with given values red, green, blue
  inline PixH2SV(const T H1, const T H2, const T S, const T V);

  //! Construct from given values, clamping & converting as needed
  template <class T2, class T3, class T4, class T5> inline
  PixH2SV(const T2 H1, const T3 H2, const T4 S, const T5 V);
  
  //! Construct an H2SV pixel, unsaturated red, at value
  inline PixH2SV(const T lum); 
  
  template <class T2> inline PixH2SV(const PixRGB<T2> pix);
  //! Construct an H2SV pixel, unsaturated red, at value
  template <class T2> inline PixH2SV(const T2 lum);

  //! Construct an H2SV pixel from another
  inline PixH2SV(const PixH2SV<T>& A);

  //! Contructor from another PixH2SV with possibly different base type
  template <class T2> inline PixH2SV(const PixH2SV<T2>& A);

  //! Construct an H2SV pixel equal to another
  inline PixH2SV<T>& operator=(const PixH2SV<T>& A); 
  
  //! Construct an H2SV pixel equal an RGB pixel
  inline PixH2SV<T>& operator=(const PixRGB<T>& A);

  //! Construct an H2SV pixel equal an RGB pixel
  template <class T2> inline PixH2SV<T>& operator=(const PixRGB<T2>& A);

  //! get RGB values for H2SV
  inline PixRGB<T> getRGB();
  
  //! get RGB values directly
  inline void getRGBP(T *r, T *g, T *b);

  //! set H2SV values from RGB
  inline void setH2SV(const PixRGB<T>& A);

  //! Construct an H2SV pixel equal to another of possibly different base type
  template <class T2> inline PixH2SV<T>& operator=(const PixH2SV<T2>& A);

  //! Access hue 1 channel
  inline T H1() const;  

  //! Access hue 2 channel
  inline T H2() const;

  //! Access saturation channel
  inline T S() const;

  //! Access intensity value channel
  inline T V() const;

  //! Set H1 channel value, clamping & converting as necessary
  template <class T2> inline void setH1(const T2 H1);

  //! Set H2 channel value, clamping & converting as necessary
  template <class T2> inline void setH2(const T2 H2);

  //! Set S channel value, clamping & converting as necessary
  template <class T2> inline void setS(const T2 S);

  //! Set V channel value, clamping & converting as necessary
  template <class T2> inline void setV(const T2 S);


  //! Set all channel values, clamping & converting as necessary
  template <class T2, class T3, class T4, class T5> inline void
  set(const T2 H1, const T3 H2, const T4 S, const T5 V);

  //! PixH2SV equality operator
  inline bool operator==(const PixH2SV<T>& other) const;

  //! PixH2SV inequality operator
  inline bool operator!=(const PixH2SV<T>& other) const;

  //! Add a value to all components, clamp result as necessary
  template <class T2> inline
  PixH2SV<T>& operator+=(const T2 val);

  //! Add component-by-component, clamp result as necessary
  template <class T2> inline
  PixH2SV<T>& operator+=(const PixH2SV<T2>& A);

  //! Add component-by-component, clamp result as necessary
  template <class T2> inline
  PixH2SV<T>& operator+=(const PixRGB<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator-=(const T2 val);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator-=(const PixH2SV<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator-=(const PixRGB<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator*=(const T2 val);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator*=(const PixH2SV<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator*=(const PixRGB<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator/=(const T2 val);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator/=(const PixH2SV<T2>& A);

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<T>& operator/=(const PixRGB<T2>& A);

  //! Add component-by-component, promote & clamp as necessary
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator+(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator+(const PixRGB<T2>& B) const;

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator+(const T2 val) const;


  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator-(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator-(const PixRGB<T2>& B) const;

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator-(const T2 val) const;


  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator*(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator*(const PixRGB<T2>& B) const;

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator*(const T2 val) const;



  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator/(const PixH2SV<T2>& B) const;

  //! PixRGB operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator/(const PixRGB<T2>& B) const;

  //! PixH2SV operator
  template <class T2> inline
  PixH2SV<typename promote_trait<T,T2>::TP> operator/(const T2 val) const;

  //! This nonsense comparison operator will abort the program when called.
  /*! It is implemented only to keep the compiler happy by allowing it
    to instantiate some methods that use comparisons but should never
    be used on color pixels. For instance, this is necessary for
    centerSurround() if we want to be able to use color pyramids. */
  inline bool operator>(const PixH2SV<T>& B) const;

  //! This nonsense comparison operator will abort the program when called.
  /*! It is implemented only to keep the compiler happy by allowing it
    to instantiate some methods that use comparisons but should never
    be used on color pixels. For instance, this is necessary for
    centerSurround() if we want to be able to use color pyramids. */
  inline bool operator<(const PixH2SV<T>& B) const;
  //! This nonsense comparison operator will abort the program when called.
  inline bool operator<<(const PixH2SV<T>& B) const;
  //! This nonsense comparison operator will abort the program when called.
  inline bool operator>>(const PixH2SV<T>& B) const;

  //! get pixel that has the max elements of both bixels 
  inline PixH2SV<T> takeMax(const PixH2SV<T>& B);
  // ##### color conversion functions:

  //! Color conversion for YIQ
  inline void getYIQ(T& y, T& i, T& q) const;

  //! Color conversion for YUV
  inline void getYUV(T& y, T& u, T& v) const;

  //! Color conversion for YUV
  inline void setYUV(const T y, const T u, const T v);

public:
  // Keep the data members public for run-time efficiency, so that we
  // can access data members directly without having to rely on the
  // compiler optimizing away inline function calls. Example:
  // lowPass3x() and lowPass3y() were spending roughly half their CPU
  // time in PixH2SV::operator+() and PixH2SV::operator*(). When these
  // calls were replaced with a lowpass implementation that just
  // accesses PixH2SV's data members directly, the total CPU time spent
  // in lowPass3() dropped by more than 50%. This is significant,
  // since lowPass3() is often where most of the CPU time is spent
  // overall in any given salience computation.
  T h1, h2, s, v;
  unsigned short Htype;
};
// ########## free functions for PixRGB:

//! take square root of r, g, and b:
template <class T> inline
PixH2SV<T> sqrt(const PixH2SV<T> &x);

//! find absolute value of all elements
template <class T> inline
PixH2SV<T> abs(const PixH2SV<T> &x);

//! Check that all three of r, g, and b have finite values
template <class T> inline
bool isFinite(const PixH2SV<T> &x);



// ######################################################################
// ##### INLINE FUNCTIONS FOR PixH2SV:
// ######################################################################

template <class T> inline
PixH2SV<T>::PixH2SV() : h1(), h2(), s(), v()
{ 
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> inline
PixH2SV<T>::PixH2SV(const T H1, const T H2, const T S, const T V)
{ h1 = H1; h2 = H2; s = S; v = V; Htype = DEFHTYPE; }

// ######################################################################
template <class T> template <class T2, class T3, class T4, class T5> inline
PixH2SV<T>::PixH2SV(const T2 H1, const T3 H2, const T4 S, const T5 V)
{
  h1 = clamped_convert<T>(H1);
  h2 = clamped_convert<T>(H2);
  s  = clamped_convert<T>(S);
  v  = clamped_convert<T>(V);
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> inline
PixH2SV<T>::PixH2SV(const T lum)
{
  v  = lum;
  s  = clamped_convert<T>(0); 
  h1 = clamped_convert<T>(0.5);
  h2 = clamped_convert<T>(1.0);
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>::PixH2SV(const PixRGB<T2> pix)
{ 
  PixRGB<T> temp = pix;
  *this = pix;
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>::PixH2SV(const T2 lum)
{ 
  v  = clamped_convert<T>(lum);
  s  = clamped_convert<T>(1); 
  h1 = clamped_convert<T>(0.5);
  h2 = clamped_convert<T>(1.0); 
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> inline
PixH2SV<T>::PixH2SV(const PixH2SV<T>& A)
{ h1 = A.h1; h2 = A.h2; s = A.s; v = A.v; Htype = DEFHTYPE;}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>::PixH2SV(const PixH2SV<T2>& A)
{
  h1 = clamped_convert<T>(A.h1);
  h2 = clamped_convert<T>(A.h2);
  s  = clamped_convert<T>(A.s );
  v  = clamped_convert<T>(A.v );
  Htype = DEFHTYPE;
}

// ######################################################################
template <class T> inline
PixH2SV<T>& PixH2SV<T>::operator=(const PixH2SV<T>& A)
{ h1 = A.h1; h2 = A.h2; s = A.s; v = A.v; return *this; }

// ######################################################################
template <class T> inline
PixH2SV<T>& PixH2SV<T>::operator=(const PixRGB<T>& A)
{
  setH2SV(A);
  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator=(const PixRGB<T2>& A)
{
  PixRGB<T> B = A; setH2SV(B);
  return *this;
}

// ######################################################################
template <class T> inline
void PixH2SV<T>::setH2SV(const PixRGB<T>& A)
{
  T min = 0, delta, h;
  Htype = DEFHTYPE;
  if((A.r >= A.g) && (A.r >= A.b))      // SET value = red
  {    
    v = A.r;
    if((A.g <= A.r)      && (A.g <= A.b))      min = A.g;
    else if((A.b <= A.g) && (A.b <= A.r))      min = A.b;
    delta = v - min;            // delta is max - min 
    if(delta == 0) delta = 1; 
    if(v != 0)
    { 
      s = delta/v;     // SET sat = (v-min)/v
      h =     (A.g - A.b) / delta;    // SET hue = green - blue / delta
      h *= 60; if(h < 0) h += 360;
    }
    else { s = 0; h = 0;}
  }
  else if((A.g >= A.r) && (A.g >= A.b)) // SET value = green
  {    
    v = A.g;
    if((A.r <= A.g)      && (A.r <= A.b)) min = A.r;
    else if((A.b <= A.g) && (A.b <= A.r)) min = A.b;
    delta = v - min;           // delta is max - min 
    if(delta == 0) delta = 1;
    if(v != 0)
    {
      s = delta/v;     // SET sat = (v-min)/v
      h = 2 + (A.b - A.r) / delta;   // SET hue = 2 + blue - red  / delta 
      h *= 60; if(h < 0) h += 360;
    }
    else { s = 0; h = 0;}
  }
  else if((A.b >= A.g) && (A.b >= A.r)) // SET value = blue
  {
    v = A.b;
    if((A.r <= A.g)      && (A.r <= A.b)) min = A.r;
    else if((A.g <= A.r) && (A.g <= A.b)) min = A.g; 
    delta = v - min;          // delta is max - min 
    if(delta == 0) delta = 1;
    if(v != 0)
    { 
      s = delta/v;
      h = 4 + (A.r - A.g) / delta;  // SET hue = 4 + red - green / delta 
      h *= 60; if(h < 0) h += 360;
    }    // SET sat = (v-min)/v
    else { s = 0; h = 0;}
  }
  
  v = v/255;

  if(Htype == 1) // std H2SV
  {
    // demoduloize hue into two components
    if((h >= 0) && (h <= 90))
      { h1 = .5 + h/180;       h2 = 1 - h/180; }
    else if((h > 90) && (h <= 180))
      { h1 = 1 - (h - 90)/180; h2 = 1 - h/180; }
    else if((h > 180) && (h <= 270))
      { h1 = 1 - (h - 90)/180; h2 = (h - 180)/180; }
    else if((h > 270) && (h <= 360))
      { h1 = (h - 270)/180;    h2 = (h - 180)/180; } 
    else // undefined color, set to 0
      { h1 = .5; h2 = 1; }
  }
  else if(Htype == 2) // RG/BY style H2SV
  {
    if((h >= 0) && (h <= 60))
      { h1 = 0.66667 + h/180;      h2 = 1 - h/120; }
    else if((h > 60) && (h <= 120))
      { h1 = 1 - (h - 60)/180; h2 = 1 - h/120; }
    else if((h > 120) && (h <= 240))
      { h1 = 1 - (h - 60)/180; h2 = (h - 120)/240; }
    else if((h > 240) && (h <= 360))
      { h1 = (h - 240)/180;    h2 = (h - 120)/240; }
    else
      { h1 = 0.66667; h2 = 1; }
  }
  else if(Htype == 3) // HSV traditional 
  {
    h2 = 0;
    h1 = h/360;
  }
  else
  {
    LFATAL("The type %d is not supported by H2SV pixels -- ABORT",Htype);
  }

}

// ######################################################################
template <class T> inline
void PixH2SV<T>::getRGBP(T *r, T *g, T *b)
{
  PixRGB<T> pixr;
  pixr = getRGB();
  *r = pixr.r;
  *g = pixr.g;
  *b = pixr.b;
}

// ######################################################################

template <class T> inline
PixRGB<T> PixH2SV<T>::getRGB()
{
  T h;
  PixRGB<T> A;
  if(Htype == 1)
  {
    // re-modulo-ize the hue component
    if((h1 > 0.5) && (h2 > 0.5))
      { h = 180*(h1-0.5);}
    else if((h1 > 0.5) && (h2 <= 0.5))
      { h = 90 + 180*(1-h1); }
    else if((h1 <= 0.5) && (h2 <= 0.5))
      { h = 90 + 180*(1-h1); }
    else if((h1 <= 0.5) && (h2 > 0.5))
      { h = 270 + 180*h1; }
  }
  else if(Htype == 2)
  {
    if(h1 >= 0.66667)
      { h = 120 * (1 - h2) + 0 ;}
    else if(h1 <= 0.5)
      { h = 240 * h2 + 120; }
    else if((h1 > 0.5) && (h1 < 0.66667) && (h2 > h1))  
      { h = 180 * h1 + 240; }
    else if((h1 > 0.5) && (h1 < 0.66667) && (h2 < h1))
      { h = 180 * (1 - h1) + 60; }
  }
  else if(Htype == 3)
  {
    h = h1*360;
  }
  else
  {
    LFATAL("The type %d is not supported by H2SV pixels -- ABORT",Htype);
  }

  T vf = v; // Assume normalized H2SV between 0 to 1 for all channels
  T sf = s; // Assume normalized H2SV between 0 to 1 for all channels
  if( vf == 0 )
    { A.r = 0; A.g = 0; A.b = 0; }
  else if( sf == 0 )
    { A.r = v; A.g = v; A.b = v; }
  else
  {
    T hf  = h / 60;
    int i = (int) floor( hf );
    T f   = hf - i;
    switch( i )
    {
    case 0:
    {
      T t   = v * ( 1 - sf * ( 1 - f ) ); 
      T p   = v * ( 1 - sf );
      A.r = v; A.g = t; A.b = p; break;
    }
    case 1:
    {
      T q   = v * ( 1 - sf * f );
      T p   = v * ( 1 - sf );
      A.r = q; A.g = v; A.b = p; break;
    }
    case 2:
    {
      T p   = v * ( 1 - sf );
      T t   = v * ( 1 - sf * ( 1 - f ) );
      A.r = p; A.g = v; A.b = t; break;
    }
    case 3:
    {
      T p   = v * ( 1 - sf );
      T q   = v * ( 1 - sf * f );
      A.r = p; A.g = q; A.b = v; break;
    }
    case 4:
    {
      T t   = v * ( 1 - sf * ( 1 - f ) );
      T p   = v * ( 1 - sf );
      A.r = t; A.g = p; A.b = v; break;
    }
    case 5:
    {
      T p   = v * ( 1 - sf );
      T q   = v * ( 1 - sf * f );
      A.r = v; A.g = p; A.b = q; break;
    }
    }
    A.r *= 255; A.g *= 255; A.b *= 255;
  }
 
}


// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator=(const PixH2SV<T2>& A)
{
  h1 = clamped_convert<T>(A.h1);
  h2 = clamped_convert<T>(A.h2);
  s  = clamped_convert<T>(A.s );
  v  = clamped_convert<T>(A.v ); 
  Htype = DEFHTYPE;
  return *this;
}

// ######################################################################
template <class T> inline
T PixH2SV<T>::H1() const
{ return h1; }

// ######################################################################
template <class T> inline
T PixH2SV<T>::H2() const
{ return h2; }

// ######################################################################
template <class T> inline
T PixH2SV<T>::S() const
{ return s; }

// ######################################################################
template <class T> inline
T PixH2SV<T>::V() const
{ return v; }

// ######################################################################
template <class T> template <class T2> inline
void PixH2SV<T>::setH1(const T2 H1)
{ h1 = clamped_convert<T>(H1); }

// ######################################################################
template <class T> template <class T2> inline
void PixH2SV<T>::setH2(const T2 H2)
{ h2 = clamped_convert<T>(H2); }

// ######################################################################
template <class T> template <class T2> inline
void PixH2SV<T>::setS(const T2 S)
{ s = clamped_convert<T>(S); }

// ######################################################################
template <class T> template <class T2> inline
void PixH2SV<T>::setV(const T2 V)
{ v = clamped_convert<T>(V); }

// ######################################################################
template <class T> template <class T2, class T3, class T4, class T5> inline
void PixH2SV<T>::set(const T2 H1, const T3 H2, const T4 S, const T5 V)
{ 
  Htype = DEFHTYPE;
  h1 = clamped_convert<T>(H1);
  h2 = clamped_convert<T>(H2);
  s  = clamped_convert<T>(S);
  v  = clamped_convert<T>(V);
}

// ######################################################################
template <class T> inline
bool PixH2SV<T>::operator==(const PixH2SV<T>& x) const
{ return ( h1 == x.h1 && h2 == x.h2 && s == x.s && v == x.v); }

// ######################################################################
template <class T> inline
bool PixH2SV<T>::operator!=(const PixH2SV<T>& x) const
{ return ( h1 != x.h1 || h2 != x.h2 || s != x.s || v != x.v ); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator+=(const T2 val)
{ 
  Htype = DEFHTYPE;
  h1 = clamped_convert<T>(h1 + val);
  h2 = clamped_convert<T>(h2 + val);
  s  = clamped_convert<T>(s  + val);
  v  = clamped_convert<T>(v  + val);

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator+=(const PixH2SV<T2>& A)
{
  Htype = DEFHTYPE;
  h1 = clamped_convert<T>(h1 + A.h1);
  h2 = clamped_convert<T>(h2 + A.h2);
  s  = clamped_convert<T>(s  + A.s );
  v  = clamped_convert<T>(v  + A.v );

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator+=(const PixRGB<T2>& A)
{
  PixRGB<T>  B = A;
  PixH2SV<T> C = B;
  h1 = h1 + C.h1;
  h2 = h2 + C.h2; 
  s  = s  + C.s;
  v  = v  + C.v;

  return *this; 
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator-=(const T2 val)
{
  h1 = clamped_convert<T>(h1 - val);
  h2 = clamped_convert<T>(h2 - val);
  s  = clamped_convert<T>(s  - val);
  v  = clamped_convert<T>(v  - val);

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator-=(const PixH2SV<T2>& A)
{
  h1 = clamped_convert<T>(h1 - A.h1);
  h2 = clamped_convert<T>(h2 - A.h2);
  s  = clamped_convert<T>(s  - A.s );
  v  = clamped_convert<T>(v  - A.v );

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator-=(const PixRGB<T2>& A)
{
  PixRGB<T>  B = A;
  PixH2SV<T> C = B;
  h1 = h1 - C.h1;
  h2 = h2 - C.h2; 
  s  = s  - C.s;
  v  = v  - C.v;

  return *this; 
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator*=(const T2 val)
{
  h1 = clamped_convert<T>(h1 * val);
  h2 = clamped_convert<T>(h2 * val);
  s  = clamped_convert<T>(s  * val);
  v  = clamped_convert<T>(v  * val);
    

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator*=(const PixH2SV<T2>& A)
{
  h1 = clamped_convert<T>(h1 * A.h1);
  h2 = clamped_convert<T>(h2 * A.h2);
  s  = clamped_convert<T>(s  * A.s );
  v  = clamped_convert<T>(v  * A.v );

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator*=(const PixRGB<T2>& A)
{
  PixRGB<T>  B = A;
  PixH2SV<T> C = B;
  h1 = h1 * C.h1;
  h2 = h2 * C.h2; 
  s  = s  * C.s;
  v  = v  * C.v;

  return *this; 
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator/=(const T2 val)
{
  h1 = clamped_convert<T>(h1 / val);
  h2 = clamped_convert<T>(h2 / val);
  s  = clamped_convert<T>(s  / val);
  v  = clamped_convert<T>(v  / val);

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator/=(const PixH2SV<T2>& A)
{
  h1 = clamped_convert<T>(h1 / A.h1);
  h2 = clamped_convert<T>(h2 / A.h2);
  s  = clamped_convert<T>(s  / A.s );
  v  = clamped_convert<T>(v  / A.v );

  return *this;
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<T>& PixH2SV<T>::operator/=(const PixRGB<T2>& A)
{
  PixRGB<T>  B = A;
  PixH2SV<T> C = B;
  h1 = h1 / C.h1;
  h2 = h2 / C.h2; 
  s  = s  / C.s;
  v  = v  / C.v;

  return *this; 
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator+(const PixH2SV<T2>& B) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1+B.h1, h2+B.h2, s+B.s, v+B.v ); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator+(const PixRGB<T2>& B) const
{
  PixRGB<T> C = B;
  PixH2SV<T> D = C;
  return PixH2SV<typename promote_trait<T,T2>::TP>(h1+D.h1, h2+D.h2, s+D.s, v+D.v );
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator+(const T2 val) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1 + val, h2 + val, s + val, v + val ); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator-(const PixH2SV<T2>& B) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1-B.h1, h2-B.h2, s-B.s, v-B.v); }


// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator-(const PixRGB<T2>& B) const
{
  PixRGB<T> C = B;
  PixH2SV<T> D = C;
  return PixH2SV<typename promote_trait<T,T2>::TP>(h1-D.h1, h2-D.h2, s-D.s, v-D.v );
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator-(const T2 val) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1 - val, h2 - val, s - val, v - val); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator*(const PixH2SV<T2>& B) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1*B.h1, h2*B.h2, s*B.s, v*B.v); }


// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator*(const PixRGB<T2>& B) const
{
  PixRGB<T> C = B;
  PixH2SV<T> D = C;
  return PixH2SV<typename promote_trait<T,T2>::TP>(h1*D.h1, h2*D.h2, s*D.s, v*D.v );
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP> PixH2SV<T>::operator*(const T2 val) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1 * val, h2 * val, s * val, v * val); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP>
PixH2SV<T>::operator/(const PixH2SV<T2>& B) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1/B.h1, h2/B.h2, s/B.s, v/B.v); }

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP>
PixH2SV<T>::operator/(const PixRGB<T2>& B) const
{
  PixRGB<T> C = B;
  PixH2SV<T> D = C;
  return PixH2SV<typename promote_trait<T,T2>::TP>(h1/D.h1, h2/D.h2, s/D.s, v/D.v );
}

// ######################################################################
template <class T> template <class T2> inline
PixH2SV<typename promote_trait<T,T2>::TP>
PixH2SV<T>::operator/(const T2 val) const
{ return PixH2SV<typename promote_trait<T,T2>::TP>(h1 / val, h2 / val, s / val, v / val ); }

// ######################################################################
template <class T> inline
bool PixH2SV<T>::operator>(const PixH2SV<T>& B) const
{
  LFATAL("Operator > does not make sense on H2SV pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
bool PixH2SV<T>::operator<(const PixH2SV<T>& B) const
{
  LFATAL("Operator < does not make sense on H2SV pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline 
bool PixH2SV<T>::operator<<(const PixH2SV<T>& B) const
{
  LFATAL("Operator << does not make sense on H2SV pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline 
bool PixH2SV<T>::operator>>(const PixH2SV<T>& B) const
{
  LFATAL("Operator >> does not make sense on H2SV pixels -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
void PixH2SV<T>::getYIQ(T& y, T& i, T& q) const
{
  LFATAL("This operation not yet supported, convert to RGB first -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
void PixH2SV<T>::getYUV(T& y, T& u, T& v) const
{
  LFATAL("This operation not yet supported, convert to RGB first -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
void PixH2SV<T>::setYUV(const T y, const T u, const T v)
{
  LFATAL("This operation not yet supported, convert to RGB first -- ABORT");
  return false;
}

// ######################################################################
template <class T> inline
PixH2SV<T> PixH2SV<T>::takeMax(const PixH2SV<T>& B)
{
  if(B.h1 > h1) h1 = B.h1;
  if(B.h2 > h2) h2 = B.h2;
  if(B.s  > s ) s  = B.s;
  if(B.v  > v ) v  = B.v;
  return *this;
}

// ########## free functions:

template <class T> inline
PixH2SV<T> sqrt(const PixH2SV<T> &x)
{
  return PixH2SV<T>(clamped_convert<T>(std::sqrt((float)(x.h1))),
                    clamped_convert<T>(std::sqrt((float)(x.h2))),
                    clamped_convert<T>(std::sqrt((float)(x.s ))),
                    clamped_convert<T>(std::sqrt((float)(x.v )))
                    );
}

// ######################################################################
template <class T> inline
PixH2SV<T> abs(const PixH2SV<T> &x)
{
  return PixH2SV<T>(clamped_convert<T>(std::fabs((float)(x.h1))),
                    clamped_convert<T>(std::fabs((float)(x.h2))),
                    clamped_convert<T>(std::fabs((float)(x.s ))),
                    clamped_convert<T>(std::fabs((float)(x.v )))
                    );
}

// ######################################################################
template <class T>
bool isFinite(const PixH2SV<T> &x)
{
  // we use the isFinite() functions defined in saliency.H
  return (isFinite(x.h1()) && isFinite(x.h2()) &&
          isFinite(x.s() ) && isFinate(x.v() ) 
          );
}

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
