/*!@file Point2D.H A basic 2D point class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Point2D.H 2905 2003-10-15 23:31:38Z itti $
//

#ifndef POINT2D_H_DEFINED
#define POINT2D_H_DEFINED

#include "Dims.H"
#include <cmath>

//! This is a basic class to encode 2D integer coordinates
/*! This is a completely public class whose goal is just to provide a
 shorthand notation for 2D integer coordinates.
 All methods are inlined, so there is no .C file, just a .H file.
*/

class Point2D
{
public:
  //! The default constructor initializes the coordinates to (0,0)
  inline Point2D();

  //! Initialize the Point2D from horizontal & vertical coordinates
  inline Point2D(const int ii, const int jj);

  //! += operator
  inline Point2D& operator+=(const Point2D &p);
  //! -= operator
  inline Point2D& operator-=(const Point2D &p);
  //! *= operator
  inline Point2D& operator*=(const Point2D &p);
  //! /= operator
  inline Point2D& operator/=(const Point2D &p);

  //! + operator
  inline Point2D operator+(const Point2D &p) const;
  //! - operator
  inline Point2D operator-(const Point2D &p) const;
  //! * operator
  inline Point2D operator*(const Point2D &p) const;
  //! / operator
  inline Point2D operator/(const Point2D &p) const;

  //! += operator
  inline Point2D& operator+=(const int val);
  //! -= operator
  inline Point2D& operator-=(const int val);
  //! *= operator
  inline Point2D& operator*=(const int val);
  //! /= operator
  inline Point2D& operator/=(const int val);

  //! + operator
  inline Point2D operator+(const int val) const;
  //! - operator
  inline Point2D operator-(const int val) const;
  //! * operator
  inline Point2D operator*(const int val) const;
  //! / operator
  inline Point2D operator/(const int val) const;

  //! test whether i and j are both positive 
  inline bool isValid() const;

  //! the square of the euclidean distance
  inline float squdist(const Point2D& p) const;

  //! the euclidean distance from p
  inline float distance(const Point2D& p) const;

  //! clamp the coords such that the point fits into image of dimensions dims
  /*! returns the difference between the corrected values and the old ones */
  inline Point2D fitToDims(const Dims dims);

  //! clamp the coords such that the point fits into image of dimensions dims
  /*! fatsre than fitToDims() and no return value. */
  inline void clampToDims(const Dims dims);

  //! 2D coordinates
  int i, j;
};

//! == operator
inline bool operator==(const Point2D& p1, const Point2D& p2);

//! != operator
inline bool operator!=(const Point2D& p1, const Point2D& p2);





// ######################################################################
inline Point2D::Point2D()
{ i = 0; j = 0; }

// ######################################################################
inline Point2D::Point2D(const int ii, const int jj)
{ i = ii; j = jj; }

// ######################################################################
inline Point2D& Point2D::operator+=(const Point2D &p)
{ i += p.i; j += p.j; return *this; }

// ######################################################################
inline Point2D& Point2D::operator-=(const Point2D &p)
{ i -= p.i; j -= p.j; return *this; }

// ######################################################################
inline Point2D& Point2D::operator*=(const Point2D &p)
{ i *= p.i; j *= p.j; return *this; }

// ######################################################################
inline Point2D& Point2D::operator/=(const Point2D &p)
{ i /= p.i; j /= p.j; return *this; }

// ######################################################################
inline Point2D Point2D::operator+(const Point2D &p) const
{ return Point2D(i + p.i, j + p.j); }

// ######################################################################
inline Point2D Point2D::operator-(const Point2D &p) const
{ return Point2D(i - p.i, j - p.j); }

// ######################################################################
inline Point2D Point2D::operator*(const Point2D &p) const
{ return Point2D(i * p.i, j * p.j); }

// ######################################################################
inline Point2D Point2D::operator/(const Point2D &p) const
{ return Point2D(i / p.i, j / p.j); }

// ######################################################################
inline Point2D& Point2D::operator+=(const int val)
{ i += val; j += val; return *this; }

// ######################################################################
inline Point2D& Point2D::operator-=(const int val)
{ i -= val; j -= val; return *this; }

// ######################################################################
inline Point2D& Point2D::operator*=(const int val)
{ i *= val; j *= val; return *this; }

// ######################################################################
inline Point2D& Point2D::operator/=(const int val)
{ i /= val; j /= val; return *this; }

// ######################################################################
inline Point2D Point2D::operator+(const int val) const
{ return Point2D(i + val, j + val); }

// ######################################################################
inline Point2D Point2D::operator-(const int val) const
{ return Point2D(i - val, j - val); }

// ######################################################################
inline Point2D Point2D::operator*(const int val) const
{ return Point2D(i * val, j * val); }

// ######################################################################
inline Point2D Point2D::operator/(const int val) const
{ return Point2D(i / val, j / val); }

// ######################################################################
inline bool operator==(const Point2D& p1, const Point2D& p2)
{ return p1.i == p2.i && p1.j == p2.j; }

// ######################################################################
inline bool operator!=(const Point2D& p1, const Point2D& p2)
{ return p1.i != p2.i || p1.j != p2.j; }

// ######################################################################
inline bool Point2D::isValid() const
{ return ((i >= 0) && (j >= 0)); }

// ######################################################################
inline float Point2D::squdist(const Point2D& p) const
{ 
  int d1 = p.i - i, d2 = p.j - j;
  return (d1 * d1 + d2 * d2);
}

// ######################################################################
inline float Point2D::distance(const Point2D& p) const
{ return sqrt(squdist(p)); }

// ######################################################################
inline Point2D Point2D::fitToDims(const Dims dims)
{
  int ri = 0, rj = 0;
  if (i < 0) { ri = -i; i = 0; }
  if (j < 0) { rj = -j; j = 0; }
  if (i >= dims.w()) { ri = dims.w() - i - 1; i = dims.w() - 1; }
  if (j >= dims.h()) { rj = dims.h() - j - 1; j = dims.h() - 1; }
  return Point2D(ri, rj);
}

// ######################################################################
inline void Point2D::clampToDims(const Dims dims)
{
  if (i < 0) i = 0; else if (i >= dims.w()) i = dims.w()-1;
  if (j < 0) j = 0; else if (j >= dims.h()) j = dims.h()-1;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif
