/*!@file PyrBuilder.C Classes for building dyadic pyramids */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: PyrBuilder.C 3420 2004-07-13 21:52:04Z bosa $
//

#include "PyrBuilder.H"

#include "Image.H"
#include "ImageSet.H"
#include "Pyramid_Ops.H"
#include "Image_FilterOps.H"
#include "Image_ColorOps.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"

// ######################################################################
// ##### TemplateMatchPyrBuilder Functions:
// ######################################################################
TemplateMatchPyrBuilder::TemplateMatchPyrBuilder(const Image<float>& templ) :
  PyrBuilder<float>(), itsFilt(templ)
{ }

// ######################################################################
TemplateMatchPyrBuilder* TemplateMatchPyrBuilder::clone() const
{ return new TemplateMatchPyrBuilder(*this); }

// ######################################################################
ImageSet<float> TemplateMatchPyrBuilder::build(const Image<float>& image,
                                               const int depth)
{
  ImageSet<float> result(depth);
  result[0] = templateMatch(image);
  Image<float> prev_scale = image;

  for (int lev = 1; lev < depth; ++lev)
    {
      Image<float> cur_scale = prev_scale;

      cur_scale = decX(lowPass5x(cur_scale));
      cur_scale = decY(lowPass5y(cur_scale));

      // Save the unconvolved image at this scale to be used as the
      // starting point for the next scale
      prev_scale = cur_scale;

      result[lev] = templateMatch(cur_scale);
    }

  return result;
}

// ######################################################################
Image<float> TemplateMatchPyrBuilder::templateMatch(const Image<float>& img)
{
  Image<float> conv = convolve(img, itsFilt);
  Image<float> conv2 = convolve(binaryReverse(img, 255.0f), itsFilt);
  return takeMax(conv, conv2);
}


// ######################################################################
// ##### Instantiations
// ######################################################################
template class GaborPyrBuilder<float>;
template class RGBConvolvePyrBuilder<float>;

#define INSTANTIATE(T) \
template class PyrBuilder< T >; \
template class GenericPyrBuilder< T >; \
template class GaussianPyrBuilder< T >; \
template class ConvolvePyrBuilder< T >; \
template class LaplacianPyrBuilder< T >; \
template class OrientedPyrBuilder< T >; \
template class ReichardtPyrBuilder< T >; \

#ifdef INST_BYTE
INSTANTIATE(byte);
INSTANTIATE(PixRGB<byte>);
#endif
#ifdef INST_INT16
INSTANTIATE(int16);
INSTANTIATE(PixRGB<int16>);
#endif
#ifdef INST_INT32
INSTANTIATE(int32);
INSTANTIATE(PixRGB<int32>);
#endif
#ifdef INST_FLOAT
INSTANTIATE(float);
INSTANTIATE(PixRGB<float>);
#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
