/*!@file PyrBuilder.C Classes for building dyadic pyramids */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: PyrBuilder.C 3420 2004-07-13 21:52:04Z bosa $
//

#include "PyrBuilder.H"

#include "Image.H"
#include "ImageSet.H"
#include "Pyramid_Ops.H"
#include "Image_FilterOps.H"
#include "Image_ColorOps.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"

// ######################################################################
// ##### PyrBuilder functions:
// ######################################################################

// ######################################################################
template <class T>
PyrBuilder<T>::PyrBuilder()
{ }

// ######################################################################
template <class T>
PyrBuilder<T>::~PyrBuilder()
{ }

// ######################################################################
template <class T>
void PyrBuilder<T>::reset()
{ }

// ######################################################################
// ##### GaussianPyrBuilder Functions:
// ######################################################################

template <class T>
GaussianPyrBuilder<T>::GaussianPyrBuilder(int filter_size) :
  PyrBuilder<T>(), itsFiltSize(filter_size)
{ }

template <class T>
ImageSet<T> GaussianPyrBuilder<T>::build(const Image<T>& img, const int depth)
{ return buildPyrGaussian(img, depth, itsFiltSize); }

template <class T> inline
GaussianPyrBuilder<T>* GaussianPyrBuilder<T>::clone() const
{ return new GaussianPyrBuilder<T>(*this); }

// ######################################################################
// ##### ConvolvePyrBuilder Functions:
// ######################################################################

template <class T>
ConvolvePyrBuilder<T>::ConvolvePyrBuilder(const Image<float>& filt) :
  PyrBuilder<T>(), itsFilt(filt)
{ }

template <class T>
ImageSet<T> ConvolvePyrBuilder<T>::build(const Image<T>& img, const int depth)
{ return buildPyrConvolve(img, depth, itsFilt); }

template <class T>
ConvolvePyrBuilder<T>* ConvolvePyrBuilder<T>::clone() const
{ return new ConvolvePyrBuilder<T>(*this); }

// ######################################################################
// ##### RGBConvolvePyrBuilder Functions:
// ######################################################################

template <class T>
RGBConvolvePyrBuilder<T>::RGBConvolvePyrBuilder(const Image<float>& rfilt,
                                                const Image<float>& gfilt,
                                                const Image<float>& bfilt) :
  PyrBuilder<T>(), itsRFilt(rfilt), itsGFilt(rfilt), itsBFilt(rfilt)
{ }

template <class T>
ImageSet<T> RGBConvolvePyrBuilder<T>::build(const Image< PixRGB<T> >& img,
                                            const int depth)
{ 
  Image<T> r, g, b; getComponents(img, r, g, b);
  ImageSet<T> rset = buildPyrConvolve(r, depth, itsRFilt);
  ImageSet<T> gset = buildPyrConvolve(g, depth, itsGFilt);
  ImageSet<T> bset = buildPyrConvolve(b, depth, itsBFilt);
  ImageSet<T> result(depth);
  
  for (int lev = 0; lev < depth; ++lev)
    result[lev] = (rset[lev] + gset[lev] + bset[lev]) / 3;
  
  return result;
}

template <class T>
ImageSet<T> RGBConvolvePyrBuilder<T>::build(const Image<T>& img,
                                            const int depth)
{
  LFATAL("Cannot process greyscale images -- ABORT");
  return ImageSet<T>();
}

template <class T>
ImageSet< PixRGB<T> > RGBConvolvePyrBuilder<T>::build2(const Image< PixRGB<T> >& img,
                                                       const int depth)
{
  Image<T> r, g, b; getComponents(img, r, g, b);
  ImageSet<T> rset = buildPyrConvolve(r, depth, itsRFilt);
  ImageSet<T> gset = buildPyrConvolve(g, depth, itsGFilt);
  ImageSet<T> bset = buildPyrConvolve(b, depth, itsBFilt);
  ImageSet< PixRGB<T> > result(depth);
  
  for (int lev = 0; lev < depth; ++lev)
    result[lev] = makeRGB (rset[lev], gset[lev], bset[lev]);
  
  return result;
}

template <class T>
RGBConvolvePyrBuilder<T>* RGBConvolvePyrBuilder<T>::clone() const
{ return new RGBConvolvePyrBuilder<T>(*this); }

// ######################################################################
// ##### LaplacianPyrBuilder Functions:
// ######################################################################

template <class T>
LaplacianPyrBuilder<T>::LaplacianPyrBuilder(const int filter_size) :
  PyrBuilder<T>(), itsFiltSize(filter_size)
{ }

template <class T>
ImageSet<T> LaplacianPyrBuilder<T>::build(const Image<T>& img, const int depth)
{ return buildPyrLaplacian(img, depth, itsFiltSize); }

template <class T> inline
LaplacianPyrBuilder<T>* LaplacianPyrBuilder<T>::clone() const
{ return new LaplacianPyrBuilder<T>(*this); }

// ######################################################################
// ##### OrientedPyrBuilder Functions:
// ######################################################################

template <class T>
OrientedPyrBuilder<T>::OrientedPyrBuilder(const int filter_size,
                                          const float theta,
                                          const float intens) :
  PyrBuilder<T>(), itsFiltSize(filter_size), itsAngle(theta),
  itsGaborIntens(intens)
{  }

template <class T>
ImageSet<T> OrientedPyrBuilder<T>::build(const Image<T>& img, const int depth)
{ return buildPyrOriented(img, depth, itsFiltSize, itsAngle, itsGaborIntens); }

template <class T>
OrientedPyrBuilder<T>* OrientedPyrBuilder<T>::clone() const
{ return new OrientedPyrBuilder<T>(*this); }


// ######################################################################
// ##### GenericPyrBuilder Functions:
// ######################################################################

template <class T>
GenericPyrBuilder<T>::GenericPyrBuilder(const PyramidType typ,
                                        const float gabor_theta,
                                        const float intens) :
  PyrBuilder<T>(), itsPtype(typ), itsGaborAngle(gabor_theta),
  itsGaborIntens(intens)
{ }

template <class T>
ImageSet<T> GenericPyrBuilder<T>::build(const Image<T>& image, const int depth)
{
  return buildPyrGeneric(image, depth, itsPtype,
                         itsGaborAngle, itsGaborIntens);
}

template <class T>
GenericPyrBuilder<T>* GenericPyrBuilder<T>::clone() const
{ return new GenericPyrBuilder<T>(*this); }

// ######################################################################
// ##### ReichardtPyrBuilder Functions:
// ######################################################################
template <class T>
ReichardtPyrBuilder<T>::ReichardtPyrBuilder(const float dx,
                                            const float dy,
                                            const PyramidType typ,
                                            const float gabor_theta,
                                            const float intens) :
  PyrBuilder<T>(), itsDX(dx), itsDY(dy), itsPtype(typ),
  itsGaborAngle(gabor_theta), itsGaborIntens(intens)
{ }

template <class T>
ImageSet<T> ReichardtPyrBuilder<T>::build(const Image<T>& image,
                                          const int depth)
{
  // create a pyramid with the input image
  ImageSet<T> upyr = buildPyrGeneric(image, depth, itsPtype,
                                     itsGaborAngle, itsGaborIntens);
  // create an empty pyramid
  ImageSet<T> spyr(depth);

  // fill the empty pyramid with the shifted version
  for (int i = 0; i < depth; i++)
    spyr[i] = shiftImage(upyr[i], itsDX, itsDY);

  // store both pyramids in the deques
  unshifted.push_back(upyr);
  shifted.push_back(spyr);

  ImageSet<T> result(depth);

  // so, it's our first time? Pretend the pyramid before this was
  // the same as the current one ...
  if (unshifted.size() == 1)
    {
      unshifted.push_back(upyr);
      shifted.push_back(spyr);
    }

  // need to pop off old pyramid?
  if (unshifted.size() == 3)
    {
      unshifted.pop_front();
      shifted.pop_front();
    }

  // compute the Reichardt maps
  for (int i = 0; i < depth; i++)
    {
      result[i] =
        (unshifted.back()[i] * shifted.front()[i]) -
        (unshifted.front()[i] * shifted.back()[i]);
    }

  return result;
}

// ######################################################################
template <class T>
ReichardtPyrBuilder<T>* ReichardtPyrBuilder<T>::clone() const
{ return new ReichardtPyrBuilder<T>(*this); }

// ######################################################################
template <class T>
void ReichardtPyrBuilder<T>::reset()
{
  shifted.clear();
  unshifted.clear();
}

// ######################################################################
// ##### GaborPyrBuilder Functions:
// ######################################################################
template <class T>
GaborPyrBuilder<T>::GaborPyrBuilder(double angle,
                                    double filter_period,
                                    double elongation,
                                    int size,
                                    int buildFlags) :
  PyrBuilder<T>(),
  itsAngle(angle),
  itsPeriod(filter_period),
  itsElongation(elongation),
  itsSize(size),
  itsBuildFlags(buildFlags)
{}

// ######################################################################
template <class T>
GaborPyrBuilder<T>* GaborPyrBuilder<T>::clone() const
{
  return new GaborPyrBuilder<T>(*this);
}

// ######################################################################
template <class T>
ImageSet<T> GaborPyrBuilder<T>::build(const Image<T>& img, const int depth)
{
  return buildPyrGabor(img, depth, itsAngle, itsPeriod, itsElongation, 
                       itsSize, itsBuildFlags);
}

// ######################################################################
template <class T>
ImageSet<T> GaborPyrBuilder<T>::input(const ImageSet<T>& pyr)
{
  return buildPyrGabor(pyr, itsAngle, itsPeriod, itsElongation, 
                       itsSize, itsBuildFlags);
}


// ######################################################################
// ##### Instantiations
// ######################################################################
//template class GaborPyrBuilder<float>;
//template class RGBConvolvePyrBuilder<float>;

#define INSTANTIATE(T) \
template class PyrBuilder< T >; \
template class GenericPyrBuilder< T >; \
template class GaussianPyrBuilder< T >; \
template class ConvolvePyrBuilder< T >; \
template class LaplacianPyrBuilder< T >; \
template class OrientedPyrBuilder< T >; \
template class ReichardtPyrBuilder< T >; \

#ifdef INST_BYTE
INSTANTIATE(byte);
INSTANTIATE(PixRGB<byte>);
#endif
#ifdef INST_INT16
INSTANTIATE(int16);
INSTANTIATE(PixRGB<int16>);
#endif
#ifdef INST_INT32
INSTANTIATE(int32);
INSTANTIATE(PixRGB<int32>);
#endif
#ifdef INST_FLOAT
INSTANTIATE(float);
INSTANTIATE(PixRGB<float>);
#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
