/*!@file Pyramid_Ops.C Free functions operating on pyramid data structures */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Pyramid_Ops.C 3155 2004-03-04 18:45:51Z walther $
//

#include "Pyramid_Ops.H"

#include "Image.H"
#include "Image_FilterOps.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"
#include "Image_ColorOps.H"
#include "ImageSet.H"
#include "ImageSet_Ops.H"

// ######################################################################
ImageSet<float> buildPyrGabor(const ImageSet<float>& gaussianPyr, float angle,
                              float filter_period, float elongation,
                              int size, int flags)
{
  const double major_stddev = filter_period / 3.0;
  const double minor_stddev = major_stddev * elongation;

  // We have to add 90 to the angle here when constructing the gabor
  // filter. That's because the angle used to build the gabor filter
  // actually specifies the direction along which the grating
  // varies. This direction is orthogonal to the the direction of the
  // contours that the grating will detect.
  const double theta = angle + 90.0f;

  // In concept, we want to filter with four phases of the filter: odd
  // on+off, and even on+off (that would be xox, oxo, xo, and ox). But
  // since the on version just produces the negation of the off version,
  // we can get the summed output of both by just taking the absolute
  // value of the outputs of one of them. So we only need to convolve
  // with the filter in two phases, then take the absolute value (i.e.,
  // we're doing |xox| and |xo|).

  Image<float> g0 = gaborFilter3(major_stddev, minor_stddev,
                                 filter_period, 0.0f, theta, size);
  Image<float> g90 = gaborFilter3(major_stddev, minor_stddev,
                                  filter_period, 90.0f, theta, size);

  LDEBUG("angle = %.2f, period = %.2f pix, size = %dx%d pix",
         angle, filter_period, g0.getWidth(), g0.getHeight());

  Image<float> f0, f90;
  ImageSet<float> result(gaussianPyr.getDepth());

  for (uint i = 0; i < gaussianPyr.getDepth(); ++i)
    {
      if (flags & DO_ENERGY_NORM)
        {
          Image<float> temp = energyNorm(gaussianPyr[i]);
          f0 = optConvolve(temp, g0);
          f90 = optConvolve(temp, g90);
        }
      else
        {
          f0 = optConvolve(gaussianPyr[i], g0);
          f90 = optConvolve(gaussianPyr[i], g90);
        }

      if (!(flags & NO_ABS))
        {
          f0 = abs(f0);
          f90 = abs(f90);
        }

      // this really doesn't do much - take it out for now
      // DW 04/25/03
      //if (flags & DO_CLAMPED_DIFF)
      //{
      //  result[i] =
      //    clampedDiff(f0, gaussianPyr[i]) + clampedDiff(f90, gaussianPyr[i]);
      //}
      //else
      //{
      result[i] = f0 + f90;
      //}
    }

  return result;
}

// ######################################################################
ImageSet<float> buildPyrGabor(const Image<float>& img, int depth, float angle,
                              float filter_period, float elongation,
                              int size, int flags)
{
  ImageSet<float> pyr;

  //LINFO("flags: %d", flags);
  if (flags & DO_LAPLACIAN)
    {
      pyr = buildPyrLaplacian(img, depth, 9);
    }
  else
    {
      pyr = buildPyrGaussian(img, depth, 9);
    }

  return buildPyrGabor(pyr, angle, filter_period, elongation, size, flags);
}

// Include the explicit instantiations
// MA #include "inst/Pyramid_Ops.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
