/*!@file Pyramid_Ops.H Free functions operating on pyramid data structures */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Pyramid_Ops.H 3067 2004-01-25 22:57:30Z itti $
//

#ifndef PYRAMID_OPS_H_DEFINED
#define PYRAMID_OPS_H_DEFINED

#include "PyramidTypes.H"

template <class T> class Image;
template <class T> class ImageSet;

// ############################################################
// ##### Pyramid builder functions:
// ############################################################

//! Uses a Gaussian filter of size \a filterSize.
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrGaussian(const Image<T_or_RGB>& image,
                                    int depth, int filterSize);

//! Uses an arbitrary filter given as argument
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrConvolve(const Image<T_or_RGB>& image,
                                    int depth, const Image<float>& filter);

//! Uses a Laplacian filter of size \a filterSize.
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrLaplacian(const Image<T_or_RGB>& image,
                                     int depth, int filterSize);

//! Uses an oriented Laplacian filter with a given size/orientation/intensity.
/*! This is a fast approximation to gabor convolution. */
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrOriented(const Image<T_or_RGB>& image,
                                    int depth, int filterSize,
                                    float theta, float intens = 10.0f);

//! Just fills the deepest level with the local image average at that scale.
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrLocalAvg(const Image<T_or_RGB>& image, int depth);

//! Just fills the deepest level with the local image maximum at that scale.
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrLocalMax(const Image<T_or_RGB>& image, int depth);

//! Build a pyramid of type given by the PyramidType.
template <class T_or_RGB>
ImageSet<T_or_RGB> buildPyrGeneric(const Image<T_or_RGB>& image,
                                   int depth, const PyramidType typ,
                                   const float gabor_theta = 0.0f,
                                   const float gabor_intens = 10.0f);

enum GaborFlags
  {
    DO_ENERGY_NORM = 1,
    DO_LAPLACIAN = 2,
    DO_CLAMPED_DIFF = 4,
    NO_ABS = 8
  };

//! Build a Gabor Pyramid with a given angle, filter period and elongation
/*!@param size determines the filter size (in pixels) - in the default case
  (-1), the size is computed from the filter_preiod and elongation*/
ImageSet<float> buildPyrGabor(const ImageSet<float>& gaussianPyr, float angle,
                              float filter_period, float elongation = 1.0,
                              int size = -1, int flags = 0);

//! Build a Gabor Pyramid with a given angle, filter period and elongation
/*!@param size determines the filter size (in pixels) - in the default case
  (-1), the size is computed from the filter_preiod and elongation*/
ImageSet<float> buildPyrGabor(const Image<float>& img, int depth, float angle,
                              float filter_period, float elongation = 1.0,
                              int size = -1, int flags = 0);

// ############################################################
// ##### Processing Functions:
// ############################################################

//! Check whether the pyramid is dyadic.
/*! A dyadic pyramid is one in which each level is one half the width and
    one half the height of the preceding level. */
template <class T>
bool isDyadic(const ImageSet<T>& pyr);

//! Compute center-surround between images at lev1 and lev2
template <class T>
Image<T> centerSurround(const ImageSet<T>& pyr,
                        const int lev1, const int lev2,
                        const bool absol = false,
                        const ImageSet<float>* clipPyr = 0);

//! Compute center-surround between images, splitting positive and negative
template <class T>
void centerSurround(const ImageSet<T>& pyr,
                    const int lev1, const int lev2,
                    Image<T>& pos, Image<T>& neg,
                    const ImageSet<float>* clipPyr = 0);

//! Compute single-opponent center-surround between images at lev1 and lev2
/*! This is essentially cpyr[lev1] - spyr[lev2] */
template <class T>
Image<T> centerSurroundSingleOpponent(const ImageSet<T>& cpyr,
                                      const ImageSet<T>& spyr,
                                      const int lev1, const int lev2,
                                      const bool absol = false,
                                      const ImageSet<float>* clipPyr = 0);

//! Compute single-opp center-surr between imgs, splitting pos and neg
template <class T>
void centerSurroundSingleOpponent(const ImageSet<T>& cpyr,
                                  const ImageSet<T>& spyr,
                                  const int lev1, const int lev2,
                                  Image<T>& pos, Image<T>& neg,
                                  const ImageSet<float>* clipPyr = 0);

//! Compute center-surround between difference images at lev1 and lev2
/*! Difference images are obtained at each level by taking the difference
  between the image from *this minus the image from other_pyr */
template <class T>
Image<T> centerSurroundDiff(const ImageSet<T>& pyr1,
                            const ImageSet<T>& pyr2,
                            const int lev1, const int lev2,
                            const bool absol = false,
                            const ImageSet<float>* clipPyr = 0);

//! Compute center-surround differnces, splitting pos & neg values
template <class T>
void centerSurroundDiff(const ImageSet<T>& pyr1,
                        const ImageSet<T>& pyr2,
                        const int lev1, const int lev2,
                        Image<T>& pos, Image<T>& neg,
                        const ImageSet<float>* clipPyr = 0);

//! Compute single-opp center-surround btw difference images at lev1 and lev2
/*! Difference images are obtained at each level by taking the difference
  between the image from *this minus the image from other_pyr */
template <class T>
Image<T> centerSurroundDiffSingleOpponent(const ImageSet<T>& cpyr1,
                                          const ImageSet<T>& cpyr2,
                                          const ImageSet<T>& spyr1,
                                          const ImageSet<T>& spyr2,
                                          const int lev1, const int lev2,
                                          const bool absol = false,
                                          const ImageSet<float>* clipPyr = 0);

//! Compute single-opp center-surround differnces, splitting pos & neg values
template <class T>
void centerSurroundDiffSingleOpponent(const ImageSet<T>& cpyr1,
                                      const ImageSet<T>& cpyr2,
                                      const ImageSet<T>& spyr1,
                                      const ImageSet<T>& spyr2,
                                      const int lev1, const int lev2,
                                      Image<T>& pos, Image<T>& neg,
                                      const ImageSet<float>* clipPyr = 0);

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#include "Pyramid_Ops_templates.h"

#endif // !PYRAMID_OPS_H_DEFINED
