/*!@file Pyramid_Ops.C Free functions operating on pyramid data structures */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Pyramid_Ops.C 3155 2004-03-04 18:45:51Z walther $
//

#include "Pyramid_Ops.H"

#include "Image.H"
#include "Image_FilterOps.H"
#include "Image_ShapeOps.H"
#include "Image_MathOps.H"
#include "Image_ColorOps.H"
#include "ImageSet.H"
#include "ImageSet_Ops.H"

// ######################################################################
// ##### Pyramid builder functions:
// ######################################################################

// ######################################################################
template <class T>
ImageSet<T> buildPyrGaussian(const Image<T>& image, int depth, int filterSize)
{
  ASSERT(image.initialized());

  ImageSet<T> result(depth);

  result[0] = image;
  for (int lev = 1; lev < depth; ++lev)
    {
      Image<T> a = result.getImage(lev-1);
      if (filterSize == 5)
        {
          a = lowPass5yDecY(lowPass5xDecX(a,2),2);
        }
      else
        {
          a = decX(lowPassX(filterSize, a));
          a = decY(lowPassY(filterSize, a));
        }
      result[lev] = a;
    }

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrConvolve(const Image<T>& image, int depth,
                             const Image<float>& filter)
{
  ASSERT(image.initialized());

  ImageSet<T> result(depth);

  Image<T> base;
  for (int lev = 0; lev < depth; ++lev)
    {
      if (lev == 0) base = image; else base = decXY(lowPass5(base));
      result[lev] = convolveClean(base, filter);
    }

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrLaplacian(const Image<T>& image, int depth, int filterSize)
{
  ASSERT(image.initialized());

  ImageSet<T> result(depth);

  Image<T> lpf = lowPass(filterSize, image);
  result[0] = image - lpf;

  for (int lev = 1; lev < depth; ++lev)
    {
      Image<T> dec = decXY(lpf);
      lpf = lowPass(filterSize, dec);
      result[lev] = dec - lpf;
    }

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrOriented(const Image<T>& image, int depth, int filterSize,
                             float theta, float intens)
{
  ASSERT(image.initialized());

  ImageSet<T> result(depth);

  int attenuation_width = -1;
  float spatial_freq = -1.0;

  switch (filterSize)
    {
    case 5: attenuation_width = 3; spatial_freq = M_PI_2; break;
    case 9: attenuation_width = 5; spatial_freq = 2.6; break;
    default:
      LFATAL("Filter size %d is not supported", filterSize);
    }

  // let's promote the intermediary results to float
  typedef typename promote_trait<T, float>::TP TF;

  // compute laplacian as image - LPF(image), then compute oriented
  // filter:
  Image<TF> lpfima = lowPass(filterSize, image);
  result[0] = orientedFilter(image - lpfima, spatial_freq, theta, intens);

  // attenuate borders that are overestimated due to filter trunctation:
  result[0].attenuateBorders(attenuation_width);

  // iterate procedure over subsequent levels:
  for (int lev = 1; lev < depth; ++lev)
    {
      Image<TF> dec = decXY(lpfima);
      lpfima = lowPass(filterSize, dec);
      result[lev] = orientedFilter(dec-lpfima, spatial_freq, theta, intens);
      result[lev].attenuateBorders(attenuation_width);
    }

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrLocalAvg(const Image<T>& image, int depth)
{
  ASSERT(image.initialized());

  ASSERT(depth >= 1);

  ImageSet<T> result(depth);

  // only deepest level of pyr is filled with local avg:
  const int scale = int(pow(2.0, depth - 1));
  result[depth - 1] = quickLocalAvg(image, scale);

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrLocalMax(const Image<T>& image, int depth)
{
  ASSERT(image.initialized());

  ASSERT(depth >= 1);

  ImageSet<T> result(depth);

  // only deepest level of pyr is filled with local max:
  const int scale = int(pow(2.0, depth - 1));
  result[depth - 1] = quickLocalMax(image, scale);

  return result;
}

// ######################################################################
template <class T>
ImageSet<T> buildPyrGeneric(const Image<T>& image, int depth,
                            const PyramidType typ, const float gabor_theta,
                            const float intens)
{
  switch(typ)
    {
    case Gaussian3: return buildPyrGaussian(image, depth, 3);
    case Gaussian5: return buildPyrGaussian(image, depth, 5);
    case Gaussian9: return buildPyrGaussian(image, depth, 9);

    case Laplacian5: return buildPyrLaplacian(image, depth, 5);
    case Laplacian9: return buildPyrLaplacian(image, depth, 9);

    case Oriented5: return buildPyrOriented(image, depth, 5, gabor_theta, intens);
    case Oriented9: return buildPyrOriented(image, depth, 9, gabor_theta, intens);

    case QuickLocalAvg: return buildPyrLocalAvg(image, depth);
    case QuickLocalMax: return buildPyrLocalMax(image, depth);

    default:
      LFATAL("Attempt to create Pyramid of unknown type %d", typ);
    }

  ASSERT(false);
  return ImageSet<T>(); // "can't happen", but placate compiler
}

// ######################################################################
// ##### Processing Functions:
// ######################################################################

// ######################################################################
template <class T>
bool isDyadic(const ImageSet<T>& pyr)
{
  if (pyr.getDepth() == 0) return false;

  for (uint i = 1; i < pyr.getDepth(); ++i)
    {
      const Dims prevdims = pyr[i-1].getDims();
      const Dims curdims = pyr[i].getDims();

      // make sure we don't go below 1
      const int pw2 = std::max(prevdims.w()/2,1);
      const int ph2 = std::max(prevdims.h()/2,1);

      if (curdims.w() != pw2) return false;
      if (curdims.h() != ph2) return false;
    }

  return true;
}

// ######################################################################
template <class T>
Image<T> centerSurround(const ImageSet<T>& pyr,
                        const int lev1, const int lev2,
                        const bool absol,
                        const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < pyr.getDepth() && uint(lev2) < pyr.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == pyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == pyr[smallLev].getDims());

      return
        ::centerSurround(Image<T>(pyr[largeLev] * (*clipPyr)[largeLev]),
                         Image<T>(pyr[smallLev] * (*clipPyr)[smallLev]),
                         absol);
    }
  else
    return ::centerSurround(pyr[largeLev], pyr[smallLev], absol);
}

// ######################################################################
template <class T>
void centerSurround(const ImageSet<T>& pyr,
                    const int lev1, const int lev2,
                    Image<T>& pos, Image<T>& neg,
                    const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < pyr.getDepth() && uint(lev2) < pyr.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == pyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == pyr[smallLev].getDims());

      ::centerSurround(Image<T>(pyr[largeLev] * (*clipPyr)[largeLev]),
                       Image<T>(pyr[smallLev] * (*clipPyr)[smallLev]),
                       pos, neg);
    }
  else
    ::centerSurround(pyr[largeLev], pyr[smallLev], pos, neg);
}

// ######################################################################
template <class T>
Image<T> centerSurroundSingleOpponent(const ImageSet<T>& cpyr,
                                      const ImageSet<T>& spyr,
                                      const int lev1, const int lev2,
                                      const bool absol,
                                      const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < cpyr.getDepth() && uint(lev2) < cpyr.getDepth());
  ASSERT(uint(lev1) < spyr.getDepth() && uint(lev2) < spyr.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == cpyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == cpyr[smallLev].getDims());
      ASSERT((*clipPyr)[largeLev].getDims() == spyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == spyr[smallLev].getDims());

      return
        ::centerSurround(Image<T>(cpyr[largeLev] * (*clipPyr)[largeLev]),
                         Image<T>(spyr[smallLev] * (*clipPyr)[smallLev]),
                         absol);
    }
  else
    return ::centerSurround(cpyr[largeLev], spyr[smallLev], absol);
}

// ######################################################################
template <class T>
void centerSurroundSingleOpponent(const ImageSet<T>& cpyr,
                                  const ImageSet<T>& spyr,
                                  const int lev1, const int lev2,
                                  Image<T>& pos, Image<T>& neg,
                                  const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < cpyr.getDepth() && uint(lev2) < cpyr.getDepth());
  ASSERT(uint(lev1) < spyr.getDepth() && uint(lev2) < spyr.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == cpyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == cpyr[smallLev].getDims());
      ASSERT((*clipPyr)[largeLev].getDims() == spyr[largeLev].getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == spyr[smallLev].getDims());

      ::centerSurround(Image<T>(cpyr[largeLev] * (*clipPyr)[largeLev]),
                       Image<T>(spyr[smallLev] * (*clipPyr)[smallLev]),
                       pos, neg);
    }
  else
    ::centerSurround(cpyr[largeLev], spyr[smallLev], pos, neg);
}

// ######################################################################
template <class T>
Image<T> centerSurroundDiff(const ImageSet<T>& pyr1,
                            const ImageSet<T>& pyr2,
                            const int lev1, const int lev2,
                            const bool absol,
                            const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < pyr1.getDepth() && uint(lev2) < pyr1.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  ASSERT(pyr1[largeLev].getDims() == pyr2[largeLev].getDims());
  ASSERT(pyr1[smallLev].getDims() == pyr2[smallLev].getDims());

  // compute differences between the two pyramids:
  const Image<T> limg = pyr1[largeLev] - pyr2[largeLev];
  const Image<T> simg = pyr1[smallLev] - pyr2[smallLev];

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == limg.getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == simg.getDims());

      return ::centerSurround(Image<T>(limg * (*clipPyr)[largeLev]),
                              Image<T>(simg * (*clipPyr)[smallLev]),
                              absol);
    }
  else
    return ::centerSurround(limg, simg, absol);
}

// ######################################################################
template <class T>
void centerSurroundDiff(const ImageSet<T>& pyr1,
                        const ImageSet<T>& pyr2,
                        const int lev1, const int lev2,
                        Image<T>& pos, Image<T>& neg,
                        const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < pyr1.getDepth() && uint(lev2) < pyr1.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  ASSERT(pyr1[largeLev].getDims() == pyr2[largeLev].getDims());
  ASSERT(pyr1[smallLev].getDims() == pyr2[smallLev].getDims());

  // compute differences between the two pyramids:
  const Image<T> limg = pyr1[largeLev] - pyr2[largeLev];
  const Image<T> simg = pyr1[smallLev] - pyr2[smallLev];

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == limg.getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == simg.getDims());

      ::centerSurround(Image<T>(limg * (*clipPyr)[largeLev]),
                       Image<T>(simg * (*clipPyr)[smallLev]),
                       pos, neg);
    }
  else
    ::centerSurround(limg, simg, pos, neg);
}

// ######################################################################
template <class T>
Image<T> centerSurroundDiffSingleOpponent(const ImageSet<T>& cpyr1,
                                          const ImageSet<T>& cpyr2,
                                          const ImageSet<T>& spyr1,
                                          const ImageSet<T>& spyr2,
                                          const int lev1, const int lev2,
                                          const bool absol,
                                          const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < cpyr1.getDepth() && uint(lev2) < cpyr1.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  ASSERT(cpyr1[largeLev].getDims() == cpyr2[largeLev].getDims());
  ASSERT(cpyr1[smallLev].getDims() == cpyr2[smallLev].getDims());
  ASSERT(spyr1[largeLev].getDims() == spyr2[largeLev].getDims());
  ASSERT(spyr1[smallLev].getDims() == spyr2[smallLev].getDims());

  // compute differences between the two pyramids:
  const Image<T> limg = cpyr1[largeLev] - cpyr2[largeLev];
  const Image<T> simg = spyr1[smallLev] - spyr2[smallLev];

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == limg.getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == simg.getDims());

      return ::centerSurround(Image<T>(limg * (*clipPyr)[largeLev]),
                              Image<T>(simg * (*clipPyr)[smallLev]),
                              absol);
    }
  else
    return ::centerSurround(limg, simg, absol);
}

// ######################################################################
template <class T>
void centerSurroundDiffSingleOpponent(const ImageSet<T>& cpyr1,
                                      const ImageSet<T>& cpyr2,
                                      const ImageSet<T>& spyr1,
                                      const ImageSet<T>& spyr2,
                                      const int lev1, const int lev2,
                                      Image<T>& pos, Image<T>& neg,
                                      const ImageSet<float>* clipPyr)
{
  ASSERT(lev1 >= 0 && lev2 >= 0);
  ASSERT(uint(lev1) < cpyr1.getDepth() && uint(lev2) < cpyr1.getDepth());

  const int largeLev = std::min(lev1, lev2);
  const int smallLev = std::max(lev1, lev2);

  ASSERT(cpyr1[largeLev].getDims() == cpyr2[largeLev].getDims());
  ASSERT(cpyr1[smallLev].getDims() == cpyr2[smallLev].getDims());
  ASSERT(spyr1[largeLev].getDims() == spyr2[largeLev].getDims());
  ASSERT(spyr1[smallLev].getDims() == spyr2[smallLev].getDims());

  // compute differences between the two pyramids:
  const Image<T> limg = cpyr1[largeLev] - cpyr2[largeLev];
  const Image<T> simg = spyr1[smallLev] - spyr2[smallLev];

  if (clipPyr != 0 && clipPyr->isNonEmpty())
    {
      ASSERT((*clipPyr)[largeLev].getDims() == limg.getDims());
      ASSERT((*clipPyr)[smallLev].getDims() == simg.getDims());

      ::centerSurround(Image<T>(limg * (*clipPyr)[largeLev]),
                       Image<T>(simg * (*clipPyr)[smallLev]),
                       pos, neg);
    }
  else
    ::centerSurround(limg, simg, pos, neg);
}

// Include the explicit instantiations
// MA #include "inst/Pyramid_Ops.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
