/*!@file Raster.C Read/write/display raster images */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Raster.C 4117 2005-01-14 07:14:37Z mundhenk $
//

#include "Raster.H"

#include "Image.H"
#include "Image_ColorOps.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "log.H"
#include "Pixels.H"
#include "PbmParser.H"
#include "PfmParser.H"
// MA #include "PngParser.H"
// MA #include "PngWriter.H"

#include <cstring> // for strcat(), strcasecmp()
#include <fstream>
#include <iostream>
#include <unistd.h> // for unlink()

// ######################################################################
// ########## Raster helper functions:
// ######################################################################
namespace
{
  const int FNAMELEN = 1024;

  // ######################################################################
  RasterFileFormat AddFileExtension(char* file, const RasterFileFormat ftype,
                                    const Raster::ColorMode mode)
  {
    // if an automatic file type was requested, let's first see
    // whether we can determine the file type from the file
    // extension. Otherwise, we don't touch the filename and just
    // return Auto:
    if (ftype == Auto)
      {
        if (!strcasecmp(&file[strlen(file) - 4], ".ppm") ||
            !strcasecmp(&file[strlen(file) - 4], ".pgm")) return PNM;
        if (!strcasecmp(&file[strlen(file) - 4], ".png")) return PNG;
        if (!strcasecmp(&file[strlen(file) - 4], ".pfm")) return PFM;
        return Auto;
      }

    // if a filetype was explicitly specified, let's make sure the
    // filename (which may have come in without any extension) has the
    // correct extension:
    switch(ftype)
      {
      case PNM:
        if (mode == Raster::MODE_GREYSCALE)
          {
            if (strcasecmp(&file[strlen(file) - 4], ".pgm"))
              strcat(file, ".pgm");
          }
        else if (mode == Raster::MODE_RGB)
          {
            if (strcasecmp(&file[strlen(file) - 4], ".ppm"))
              strcat(file, ".ppm");
          }
        else
          {
            LFATAL("unknown color mode %d", int(mode));
          }
        break;
      case PNG:
        if (strcasecmp(&file[strlen(file) - 4], ".png"))
          strcat(file, ".png");
        break;
      case PFM:
        if (strcasecmp(&file[strlen(file) - 4], ".pfm"))
          strcat(file, ".pfm");
        break;
      case RAWIMAGE:
        // do not add any extension
        break;
      default:
        LFATAL("RasterFileFormat not supported");
      }
    return ftype;
  }

  // ######################################################################
  void waitDelete(const char* fname)
  {
    std::cout<<".......... press <RETURN> to continue .........."<<std::endl;
    getchar(); unlink(fname);
  }

  // ######################################################################
  void writeGrayA(const Image<byte>& image, RasterFileFormat ftype,
                  const char* fname, va_list ap)
  {
    ASSERT(image.initialized()); char fil[FNAMELEN];
    vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_GREYSCALE);
    LDEBUG("Writing Gray Image: %s", fil);
    switch(ftype)
      {
      case PNM:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open PNM file '%s' for writing.", fil);
          ofs << "P5\n" << image.getWidth() << ' ' << image.getHeight()
              << "\n255\n";
          ofs.write(reinterpret_cast<const char*>(image.getArrayPtr()),
                    image.getSize());
          break;
        }
      case PFM:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open PFM file '%s' for writing.", fil);
          Image<float> tmp = image;
          ofs << "PF\n" << tmp.getWidth() << ' ' << tmp.getHeight()
              << "\n1.0\n";
          ofs.write(reinterpret_cast<const char*>(tmp.getArrayPtr()),
                    tmp.getSize() * sizeof(float));
          break;
        }
      case PNG:
        {
          assert(false);
          // MA Png::writeGray(fil, image);
          break;
        }
      case RAWIMAGE:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open RAW file '%s' for writing.", fil);
          ofs.write(reinterpret_cast<const char*>(image.getArrayPtr()),
                    image.getSize());
          break;
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
  }

  // ######################################################################
  void writeFloatA(const Image<float>& image, RasterFileFormat ftype,
                   const char* fname, va_list ap)
  {
    ASSERT(image.initialized()); char fil[FNAMELEN];
    vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_GREYSCALE);
    LDEBUG("Writing Float Image: %s", fil);
    switch(ftype)
      {
      case PFM:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open PFM file '%s' for writing.", fil);
          Image<float> tmp = image;
          ofs << "PF\n" << tmp.getWidth() << ' ' << tmp.getHeight()
              << "\n1.0\n";
          ofs.write(reinterpret_cast<const char*>(tmp.getArrayPtr()),
                    tmp.getSize() * sizeof(float));
          break;
        }
      case RAWIMAGE:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open RAW file '%s' for writing.", fil);
          ofs.write(reinterpret_cast<const char*>(image.getArrayPtr()),
                    image.getSize() * sizeof(float));
          break;
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
  }

  // ######################################################################
  Image<byte> readGrayA(RasterFileFormat ftype, const char* fname, va_list ap)
  {
    char fil[FNAMELEN]; vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_GREYSCALE);
    LDEBUG("Reading Gray Image: %s", fil);

    switch(ftype)
      {
      case PNM:
        {
          PbmParser pp(fil);
          return pp.getGray();
        }
      case PNG:
        {
          assert(false);
          return Image<byte>();
          //PngParser pp(fil);
          //return pp.getGray();
        }
      case PFM:
        {
          PfmParser pp(fil);
          return pp.getGray();
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
    /* can't happen: */ return Image<byte>();
  }

  // ######################################################################
  Image<float> readFloatA(RasterFileFormat ftype,
                          const char* fname, va_list ap)
  {
    char fil[FNAMELEN]; vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_GREYSCALE);
    LDEBUG("Reading Float Image: %s", fil);

    switch(ftype)
      {
      case PNM:
        {
          PbmParser pp(fil);
          return pp.getFloat();
        }
      case PNG:
        assert(false);
        return Image<float>();
        {
          //PngParser pp(fil);
          //return pp.getFloat();
        }
      case PFM:
        {
          PfmParser pp(fil);
          return pp.getFloat();
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
    /* can't happen: */ return Image<byte>();
  }

  // ######################################################################
  void writeRGBA(const Image< PixRGB<byte> >& image, RasterFileFormat ftype,
                 const char* fname, va_list ap)
  {
    ASSERT(image.initialized()); char fil[FNAMELEN];
    vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_RGB);
    LDEBUG("Writing RGB Image: %s", fil);

    switch(ftype)
      {
      case PNM:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open PNM file '%s' for writing.", fil);
          ofs << "P6\n" << image.getWidth() << ' ' << image.getHeight()
              << "\n255\n";
          ofs.write(reinterpret_cast<const char*>(image.getArrayPtr()),
                    3 * image.getSize());
          break;
        }
      case PNG:
        {
          assert(false);
          // MA Png::writeRGB(fil, image);
          break;
        }
      case RAWIMAGE:
        {
          std::ofstream ofs(fil, std::ios::binary);
          if (!ofs.is_open())
            LFATAL("Couldn't open RAW file '%s' for writing.", fil);
          ofs.write(reinterpret_cast<const char*>(image.getArrayPtr()),
                    3 * image.getSize());
          break;
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
  }

  // ######################################################################
  Image<PixRGB<byte> > readRGBA(RasterFileFormat ftype,
                                const char* fname, va_list ap)
  {
    char fil[FNAMELEN]; vsnprintf(fil, FNAMELEN, fname, ap);
    ftype = AddFileExtension(fil, ftype, Raster::MODE_RGB);
    LDEBUG("Reading RGB Image: %s", fil);

    switch(ftype)
      {
      case PNM:
        {
          PbmParser pp(fil);
          return pp.getRGB();
        }
      case PNG:
        assert(false);
        return Image<PixRGB<byte> >();
        {
          //PngParser pp(fil);
          //return pp.getRGB();
        }
      case PFM:
        {
          PfmParser pp(fil);
          return pp.getGray();
        }
      default:
        LFATAL("Unsupported file type: %d", int(ftype));
      }
    /* can't happen: */ return Image<PixRGB<byte> >();
  }

} // end anonymous namespace


// ######################################################################
// ########## Metadata stuff:
// ######################################################################
Raster::MetaData::MetaData() :
  accel(0), steer(0), lat(0), lon(0), ev(0), nv(0)
{ }

Raster::MetaData::MetaData( const MetaData& m ) :
  accel( m.accel ), steer( m.steer ), lat( m.lat ), lon( m.lon),
  ev( m.ev ), nv( m.nv )
{ }

Raster::MetaData::MetaData( std::string s ) :
  accel(0), steer(0), lat(0), lon(0), ev(0), nv(0)
{
  parseMetaData( s );
}

void Raster::MetaData::parseMetaData( std::string s )
{
  std::string temp = "";
  int space;
  int index;
  while( ( space = s.find( " ", 0 ) ) != int(std::string::npos) )
    {
      temp = s.substr( 0, space );
      s = s.substr( space, s.length() - space );

      index = temp.find( " ", 0 );

      if( temp.find( "ACCEL:", 0 ) != std::string::npos )
        accel = atof( temp.c_str() );
      if( temp.find("STEER:", 0) != std::string::npos )
        steer = atof( temp.c_str() );
      if( temp.find("LAT:", 0 ) != std::string::npos )
        lat = atof( temp.c_str() );
      if( temp.find("LON:", 0 ) != std::string::npos )
        lon = atof( temp.c_str() );
      if( temp.find("EV:", 0) != std::string::npos )
        ev = atof( temp.c_str() );
      if( temp.find("NV:", 0) != std::string::npos )
        nv = atof( temp.c_str() );
    }
}

float Raster::MetaData::getAccel() const
{ return accel; }

float Raster::MetaData::getSteer() const
{ return steer; }

float Raster::MetaData::getLat() const
{ return lat; }

float Raster::MetaData::getLon() const
{ return lon; }

float Raster::MetaData::getEV() const
{ return ev; }

float Raster::MetaData::getNV() const
{ return nv; }


// ######################################################################
// ########## Raster functions:
// ######################################################################
void Raster::WriteGray(const Image<byte>& image, const RasterFileFormat ftype,
                       const char* fname, ...)
{
  va_list a; va_start(a, fname); writeGrayA(image, ftype, fname, a); va_end(a);
}

// ######################################################################
void Raster::WriteFloat(const Image<float>& image,
                        const RasterFileFormat ftype,
                        const bool normalize,
                        const char* fname, ...)
{
  if (ftype == PFM) { // directly write the float image
    va_list a; va_start(a, fname);
    writeFloatA(image, ftype, fname, a); va_end(a);
  } else {            // first convert to byte then write gray
    Image<byte> ima;
    if (normalize)
      {
        Image<float> tmp(image);
        tmp.normalize(0.0, 255.0);
        ima = Image<byte>(tmp);
      }
    else
      ima = Image<byte>(image);
    va_list a; va_start(a, fname); writeGrayA(ima, ftype, fname, a); va_end(a);
  }
}

// ######################################################################
void Raster::WriteFloatNorm(const Image<float>& image,
                            const RasterFileFormat ftype,
                            NormalizeMode nmode,
                            const char* fname, ...)
{
  va_list a;
  va_start(a, fname);
  switch (nmode)
    {
    case NO_NORM:
      writeGrayA(Image<byte>(image), ftype, fname, a);
      break;
    case STD_NORM:
      {
        Image<float> tmp(image);
        tmp.normalize(0.0, 255.0);
        writeGrayA(Image<byte>(tmp), ftype, fname, a);
      }
      break;
    case NORM_WITH_SCALE:
      writeGrayA(normalizeWithScale(image), ftype, fname, a);
      break;
    default:
      LFATAL("unknown Raster::NormalizeMode '%d'", nmode);
    }
  va_end(a);
}

// ######################################################################
Dims Raster::getImageDims(const RasterFileFormat ftype,
                          const Raster::ColorMode cmode,
                          const char* fname, ...)
{
  va_list a; va_start(a, fname);
  char fil[FNAMELEN]; vsnprintf(fil, FNAMELEN, fname, a); va_end(a);
  RasterFileFormat newftype = AddFileExtension(fil, ftype, cmode);

  switch(newftype)
    {
    case PNM:
      {
        PbmParser pp(fil);
        return pp.getDims();
      }
    case PNG:
      assert(false);
      return Dims();
      {
        //PngParser pp(fil);
        //return pp.getDims();
      }
    default:
      LFATAL("Unsupported file type: %d", int(ftype));
    }
  /* can't happen: */ return Dims(-1, -1);
}


// ######################################################################
Raster::MetaData Raster::getImageMeta(const RasterFileFormat ftype,
                                      const Raster::ColorMode cmode,
                                      const char* fname, ...)
{
  va_list a; va_start(a, fname);
  char fil[FNAMELEN]; vsnprintf(fil, FNAMELEN, fname, a); va_end(a);
  RasterFileFormat newftype = AddFileExtension(fil, ftype, cmode);
  std::string s;


  switch(newftype)
    {
    case PNM:
      {
        PbmParser pp(fil);
        s = pp.getComments();
        Raster::MetaData mData( s );
        return mData;
      }
    case PNG:
      {
        LINFO("PNG MetaData not currently supported");
        MetaData mData("");
        return mData;
      }
    default:
      LFATAL("Unsupported file type: %d", int(ftype));
    }
  MetaData mData("");
  /* can't happen: */ return mData;
}


// ######################################################################
Image<byte> Raster::ReadGray(const RasterFileFormat ftype,
                             const char* fname, ...)
{
  va_list a; va_start(a, fname);
  Image<byte> result = readGrayA(ftype, fname, a);
  va_end(a);
  return result;
}

// ######################################################################
Image<float> Raster::ReadFloat(const RasterFileFormat ftype,
                               const char* fname, ...)
{
  va_list a; va_start(a, fname);
  Image<float> result = readFloatA(ftype, fname, a);
  va_end(a);
  return result;
}

// ######################################################################
void Raster::WriteRGB(const Image< PixRGB<byte> >& image,
                      const RasterFileFormat ftype,
                      const char* fname, ...)
{
  va_list a; va_start(a, fname); writeRGBA(image, ftype, fname, a); va_end(a);
}

// ######################################################################
Image<PixRGB<byte> > Raster::ReadRGB(const RasterFileFormat ftype,
                                     const char* fname, ...)
{
  va_list a; va_start(a, fname);
  Image<PixRGB<byte> > result = readRGBA(ftype, fname, a);
  va_end(a);
  return result;
}

namespace
{
  bool USING_XV = true;
}

// ######################################################################
void Raster::NoXV()
{
  USING_XV = false;
}

// ######################################################################
void Raster::Display(const char* file)
{
  if (USING_XV)
    {
      char command[FNAMELEN];
      snprintf(command, FNAMELEN, "xv \"%s\" &", file);
      system(command);
    }
}

// ######################################################################
void Raster::Display(const char* file, const RasterFileFormat ft,
                     const Raster::ColorMode cmode)
{
  char fil[FNAMELEN]; snprintf(fil, FNAMELEN, file);
  AddFileExtension(fil, ft, cmode);
  Display(fil);
}

// ######################################################################
void Raster::waitForKey()
{
  char tmp[100];
  printf("<<<<< Press <Return> to continue >>>>>\n");
  fgets(tmp, 100, stdin);
}

namespace
{
  // ######################################################################
  // This is the shared implementation for Raster::Visu(Image) and
  // Raster::VisuArray(Image)
  template <class T>
  void doVisu(const Image<T>& image, const RasterFileFormat ftype,
              const bool normalize, const bool waitdel,
              const char* fname, va_list ap)
  {
    char fil[FNAMELEN];
    vsnprintf(fil, FNAMELEN, fname, ap);
    AddFileExtension(fil, ftype, Raster::MODE_GREYSCALE);

    Image<T> tt = image;
    T mi, ma; tt.getMinMax(mi, ma);

    if (normalize)
      {
        tt.normalize(T(0), T(255));
        LINFO("%s [%f..%f] -- Normalized", fil, float(mi), float(ma));
      }
    else
      LINFO("%s [%f..%f]", fil, float(mi), float(ma));
    Image<byte> tt2 = tt; // won't actually copy data if T is byte
    Raster::WriteGray(tt2, ftype, fil);
    Raster::Display(fil);
    if (waitdel) waitDelete(fil);
  }

  // ######################################################################
  // This is the shared implementation for Raster::Visu(Image<PixRGB>) and
  // Raster::VisuArray(Image<PixRGB>)
  template <class T>
  void doVisuRGB(const Image<PixRGB<T> >& image, const RasterFileFormat ftype,
                 const bool normalize, const bool waitdel,
                 const char* fname, va_list ap)
  {
    char fil[FNAMELEN];
    vsnprintf(fil, FNAMELEN, fname, ap);
    AddFileExtension(fil, ftype, Raster::MODE_RGB);

    Image<PixRGB<T> > tt = image;
    T mi, ma; getMinMaxC(tt, mi, ma);

    if (normalize)
      {
        normalizeC(tt, T(0), T(255));
        LINFO("%s [%f..%f] -- Normalized", fil, float(mi), float(ma));
      }
    else
      LINFO("%s [%f..%f]", fil, float(mi), float(ma));
    Image< PixRGB<byte> > tt2 = tt; // won't actually copy data if T is byte
    Raster::WriteRGB(tt2, ftype, fil);
    Raster::Display(fil);
    if (waitdel) waitDelete(fil);
  }
}

// ######################################################################
template <class T>
void Raster::Visu(const Image<T>& image, const RasterFileFormat ft,
                  const bool normalize, const bool wdel,
                  const char* fname,...)
{
  va_list a; va_start(a, fname);
  doVisu(image, ft, normalize, wdel, fname, a);
  va_end(a);
}

// ######################################################################
template <class T>
void Raster::Visu(const Image< PixRGB<T> >& image, const RasterFileFormat ft,
                  const bool normalize, const bool wdel,
                  const char* fname, ...)
{
  va_list a; va_start(a, fname);
  doVisuRGB(image, ft, normalize, wdel, fname, a);
  va_end(a);
}

// ######################################################################
template <class T>
void Raster::VisuArray(const Image<T>* image, const int nbima,
                       const int Nx, const RasterFileFormat ftype,
                       const bool normalize, const bool waitdel,
                       const char* fname, ...)
{
  Image<T> tt = concatArray(image, nbima, Nx);

  va_list a; va_start(a, fname);
  doVisu(tt, ftype, normalize, waitdel, fname, a);
  va_end(a);
}

// ######################################################################
template <class T>
void Raster::VisuArray(const Image< PixRGB<T> >* image, const int nbima,
                       const int Nx, const RasterFileFormat ftype,
                       const bool normalize, const bool waitdel,
                       const char* fname, ...)
{
  Image< PixRGB<T> > tt = concatArray(image, nbima, Nx);

  va_list a; va_start(a, fname);
  doVisuRGB(tt, ftype, normalize, waitdel, fname, a);
  va_end(a);
}


// Include the explicit instantiations, and make sure that they go in the
// Raster:: namespace

#define INST_NAMESPACE Raster
// MA #include "inst/Raster.I"

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
