/*!@file Raster.H Writes a raster image to disk / reads it / displays it */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Raster.H 3359 2004-06-16 06:08:00Z itti $
//

#ifndef RASTER_H_DEFINED
#define RASTER_H_DEFINED

#include "saliency.H"
#include <stdarg.h>
#include "RasterFileFormat.H"
#include "ArrayData.H"

//Ouch, sorry, but has to be done at this point
#include <string>

template <class T> class Image;
template <class T> class PixRGB;
class MetaData;

// ######################################################################
//! For reading, writing and displaying images.
/*! Raster works with PGM files (for grayscale images) and PPM files (for
    RGB images). Selection between PGM and PPM given the PNM
    RasterFileFormat is automatic based on whether the image is greyscale
    or color. In addition, it offers PNG read capability (grayscale, RGB,
    or color-palette, at any bit-depth), and write capability (8-bit
    grayscale and RGB). Similarly, it offers write capability for RAWIMAGE
    data (straight pixel data with no header information), but no such read
    capability. Both the raw and ASCII variants of PGM/PPM files are
    supported. Display of images uses the program "xv" to output images to
    the screen. The filename argument is passed to the various methods of
    this class in the same way as in printf(). */
namespace Raster
{
  //! Raster color modes
  enum ColorMode { MODE_GREYSCALE, MODE_RGB };

  //! Normalization modes when saving float images
  enum NormalizeMode
    {
      NO_NORM = 0,        //!< no normalization
      STD_NORM = 1,       //!< norm to [0,255]
      NORM_WITH_SCALE = 2 //!< norm to [0,255]; text shows original min+max
    };

  //! Metadata information (typically obtained from GPS)
  class MetaData
  {
  public:
    //! Constructor
    MetaData();
    
    //! Constructor
    MetaData( const MetaData& m );

    //! Constructor
    MetaData( std::string s );

    //! Parse from string
    void parseMetaData( std::string s );

    //! Access function
    float getAccel() const;

    //! Access function
    float getSteer() const;

    //! Access function
    float getLat() const;

    //! Access function
    float getLon() const;

    //! Access function
    float getEV() const;

    //! Access function
    float getNV() const;

    //protected:
    float accel; //!< Acceleration
    float steer; //!< Steering
    float lat;   //!< Latitude
    float lon;   //!< Longitude
    float ev;    //!< Velocity in east direction
    float nv;    //!< Velocity in north direction
  };

  // ######################################################################
  // Read/Write in various formats, for byte-based images:

  //! Write a grayscale byte image as greyscale - pgm format
  void WriteGray(const Image<byte>& image, const RasterFileFormat ftype,
                 const char *fname, ...);

  //! Write a float image, normalize if necessary - pgm format
  void WriteFloat(const Image<float>& image, const RasterFileFormat ftype,
                  const bool normalize,
                  const char* fname, ...);

  //! Write a float image with optional normalization style.
  void WriteFloatNorm(const Image<float>& image,
                     const RasterFileFormat ftype,
                     NormalizeMode nmode,
                     const char* fname, ...);

  //! Write an RGB byte image - ppm format
  void WriteRGB(const Image< PixRGB<byte> >& image,
                const RasterFileFormat ftype,
                const char *fname, ...);

  //! Get image dimensions without fully reading the image
  Dims getImageDims(const RasterFileFormat ftype,
                    const Raster::ColorMode cmode,
                    const char *fname, ...);

  //! Get image Meta Data without fully reading the image (PPM only)
  MetaData getImageMeta(const RasterFileFormat ftype,
                    const Raster::ColorMode cmode,
                    const char *fname, ...);

  //! Read a greyscale byte image from a greyscale pgm file
  Image<byte> ReadGray(const RasterFileFormat ftype, const char *fname, ...);

  //! Read a greyscale float image from a greyscale file
  Image<float> ReadFloat(const RasterFileFormat ftype, const char *fname, ...);

  //! Read an RGB byte image from a ppm file
  Image<PixRGB<byte> > ReadRGB(const RasterFileFormat ftype,
                               const char *fname, ...);

  //! Turn off the use of xv
  /*! Turn off xv, so that calls to Display() do nothing, and calls to
      Visu() just save the image to disk without displaying it onscreen. */
  void NoXV();

  //! calls xv to display an image that is on disk
  void Display(const char* file);

  //! calls xv to display an image that is on disk, guesses file extension
  void Display(const char* file, const RasterFileFormat ft,
               const Raster::ColorMode cmode);

  //! wait for the [return] key
  void waitForKey();

  // save image and display it:

  //! Save image of type Image<T> to disk then display it using xv
  template <class T>
  void Visu(const Image<T>& image, const RasterFileFormat ft,
            const bool normalize, const bool wdel,
            const char *fname, ...);

  //! Save image of type Image< PixRGB<T> > to disk then display it using xv
  template <class T>
  void Visu(const Image< PixRGB<T> >& image, const RasterFileFormat ft,
            const bool normalize, const bool wdel,
            const char *fname, ...);

  //! Save an array of images of type Image<T>
  /*! Save array of images and display them together with same
      normalization put Nx images in every line of the array, and as many
      lines as necessary */
  template <class T>
  void VisuArray(const Image<T>* image, const int nbima, const int Nx,
                 const RasterFileFormat ftype, const bool normalize,
                 const bool waitdel, const char *fname, ...);

  //! Save an array of images of type Image< PixRGB<T> >
  /*! save array of images and display them together with same
      normalization put Nx images in every line of the array, and as many
      lines as necessary */
  template <class T>
  void VisuArray(const Image< PixRGB<T> >* image, const int nbima,
                 const int Nx, const RasterFileFormat ftype,
                 const bool normalize, const bool waitdel,
                 const char *fname, ...);
};

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif
