/*!@file Rectangle.H A basic rectangle class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Rectangle.H 3849 2004-10-14 22:53:40Z walther $
//

#ifndef RECTANGLE_H_DEFINED
#define RECTANGLE_H_DEFINED

#include "saliency.H"
#include "ArrayData.H"
#include "Point2D.H"

//! A basic rectangle class
/*! This is a shorthand class to represent a 2D rectangle, as defined
by integer top, left, bottom and right coordinates.  The goal of this
object is not to have to check if left < right, etc...  for each
function using a Rectangle. The object checks for that at its
creation, and is always internally coherent.  Image convention: top <
bottom, left < right. This class is 100% inlined, so there is no .C
file. */

class Rectangle {
 public:
  // ############################################################
  // ##### Constructors & Destructors:
  // ############################################################

  //! Build from 4 coordinates (top, left, bottom, right)
  inline Rectangle(int tt, int ll, int bb, int rr);

  //! Build from 4 coordinates and adapt to fit into dims
  inline Rectangle(int tt, int ll, int bb, int rr, Dims dims);

  //! Build from a top-left corner and some dims
  inline Rectangle(const Point2D& topleft, const Dims& dims);

  //! Copy constructor
  inline Rectangle(const Rectangle& rect);

  //! Uninitialized constructor (useful for arrays).
  /*! Use modify() to change Rectangle contents later */
  inline Rectangle();
  
  // ############################################################
  // ##### Access functions:
  // ############################################################

  //! modify the rectangle so that it fits within an image of dimensions dims
  inline void fitToDims(const Dims dims);

  //! modify the rectangle so that it fits within the rectangle rect
  inline void fitToRect(const Rectangle& rect);

  //! Change coordinates and rebuild and fit into dims
  inline void modify(int tt, int ll, int bb, int rr, 
                     Dims dims = Dims(0,0));  

  //! Access internal components
  inline int top() const; 
  //! Access internal components
  inline int bottom() const; 
  //! Access internal components
  inline int left() const; 
  //! Access internal components
  inline int right() const;

  //! Access some metric info
  inline int width() const;  
  //! Access some metric info
  inline int height() const;

  //! whether the rectangle is valid
  inline bool isValid() const;
  
private:
  bool valid;       //!< True if Rectangle contains something coherent
  int t, b, l, r; //!< top, bottom, left, right
};


// ######################################################################
// INLINE FUNCTIONS:
// ######################################################################
// This is a 100% inlined class.  No .C file for this class.

// ######################################################################
inline Rectangle::Rectangle(int tt, int ll, int bb, int rr)
{ modify(tt, ll, bb, rr); }

// ######################################################################
inline Rectangle::Rectangle(int tt, int ll, int bb, int rr, Dims dims)
{ modify(tt, ll, bb, rr, dims); }

// ######################################################################
inline Rectangle::Rectangle(const Point2D& topleft, const Dims& dims)
{ modify(topleft.j, topleft.i, topleft.j+dims.h()-1, topleft.i+dims.w()-1); }


// ######################################################################
inline Rectangle::Rectangle(const Rectangle& rect)
  : valid(rect.valid), t(rect.t), b(rect.b), l(rect.l), r(rect.r)
{}

// ######################################################################
inline Rectangle::Rectangle()
  : valid(false), t(-1), b(-1), l(-1), r(-1)
{}

// ######################################################################
inline void Rectangle::fitToDims(const Dims dims)
{
  ASSERT(dims.isNonEmpty());
  fitToRect(Rectangle(0, 0, dims.h()-1, dims.w()-1));
}

// ######################################################################
inline void Rectangle::fitToRect(const Rectangle& rect)
{
  if (l < rect.left()) l = rect.left();
  if (l > rect.right()) l = rect.right();
  if (r < rect.left()) r = rect.left();
  if (r > rect.right()) r = rect.right();
  if (t < rect.top()) t = rect.top();
  if (t > rect.bottom()) t = rect.bottom();
  if (b < rect.top()) b = rect.top();
  if (b > rect.bottom()) b = rect.bottom();
}

// ######################################################################
inline void Rectangle::modify(int tt, int ll, int bb, int rr, Dims dims)
{
  if (bb < tt) { int tmp = tt; tt = bb; bb = tmp; }
  if (ll > rr) { int tmp = ll; ll = rr; rr = tmp; }
  t = tt; b = bb; l = ll; r = rr; 
  if (dims.isNonEmpty()) fitToDims(dims);
  valid = true;
} 

// ######################################################################
inline int Rectangle::top() const
{ ASSERT(valid); return t; }

inline int Rectangle::bottom() const
{ ASSERT(valid); return b; }

inline int Rectangle::left() const
{ ASSERT(valid); return l; }

inline int Rectangle::right() const
{ ASSERT(valid); return r; }

inline int Rectangle::width() const
{ ASSERT(valid); return r - l + 1; } // *** CAUTION: BOUNDARIES ARE INCLUDED

inline int Rectangle::height() const
{ ASSERT(valid); return b - t + 1; } // *** CAUTION: BOUNDARIES ARE INCLUDED

inline bool Rectangle::isValid() const
{ return valid; }

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
