/*!@file Retina.C a human retina */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Retina.C 4154 2005-02-01 20:29:21Z itti $
//

#include "Retina.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Raster.H"
#include "Image_ShapeOps.H"
#include "SaccadeController.H"
#include "Brain.H"
#include "Image_Transforms.H"
#include "Pyramid_Ops.H"
#include "FrameSeries.H"

// ######################################################################
// ######################################################################
// ########## Retina implementation
// ######################################################################
// ######################################################################
Retina::Retina(ModelManager& mgr, const std::string& descrName,
               const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSaveInput(this, "RetinaSaveInput", false), // see ModelOptionDefs.C
  itsSaveOutput(this, "RetinaSaveOutput", false),
  itsBrain(NULL), itsRawInput(), itsOutputCache()
{ }

// ######################################################################
void Retina::setBrain(Brain* brain)
{ itsBrain = brain; }

// ######################################################################
Retina::~Retina()
{ }

// ######################################################################
void Retina::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsSaveInput);
    manager->requestOption(this, itsSaveOutput);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void Retina::input(const Image< PixRGB<byte> >& in)
{ itsOutputCache.freeMem(); itsRawInput = in; }

// ######################################################################
bool Retina::hasInput() const
{ return itsRawInput.initialized(); }

// ######################################################################
Image< PixRGB<byte> > Retina::getRawInput() const
{ return itsRawInput; }

// ######################################################################
Image< PixRGB<byte> > Retina::getOutput()
{
  // already cached? otherwise recompute:
  if (itsOutputCache.initialized() == false) itsOutputCache = doGetOutput();

  // now it's cached until the next input():
  return itsOutputCache;
}

// ######################################################################
Rectangle Retina::getRawInputRectangle(const int border) const
{
  Rectangle r(border, border,
              itsRawInput.getHeight() - 1 - border,
              itsRawInput.getWidth() - 1 - border);
  r.fitToDims(itsOutputCache.getDims());
  return r;
}

// ######################################################################
Point2D Retina::rawToRetinal(const Point2D& rawpos) const
{ return rawpos; }

// ######################################################################
Point2D Retina::retinalToRaw(const Point2D& retpos) const
{ return retpos; }

// ######################################################################
void Retina::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // save input?
  if (itsSaveInput.getVal() && itsRawInput.initialized())
    ofs->writeRGB(itsRawInput, PNM, -1, "RETIN-");

  // save output?
  if (itsSaveOutput.getVal() && itsRawInput.initialized())
    ofs->writeRGB(getOutput(), PNM, -1, "RETOUT-");
}

// ######################################################################
// ######################################################################
// ########## RetinaConfigurator implementation
// ######################################################################
// ######################################################################
RetinaConfigurator::
RetinaConfigurator(ModelManager& mgr,
                          const std::string& descrName,
                          const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsType(this, "RetinaType", "Std"),
  itsRET(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
RetinaConfigurator::~RetinaConfigurator()
{  }

// ######################################################################
void RetinaConfigurator::exportOptions(const int whatToExport,
                                              const bool recurse)
{
  // keep a copy of whatToExport so that we can let our component
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsType);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<Retina> RetinaConfigurator::getRET() const
{ return itsRET; }

// ######################################################################
bool RetinaConfigurator::setModelParamString(const std::string& name,
                                             const std::string& value,
                                             const bool recurse,
                                             const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsType.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsRET.get()) { removeSubComponent(itsRET); itsRET.reset(NULL); }

    // instantiate a Retina of the appropriate type:
    if (value.compare("None") == 0)          // none
      itsRET.reset(NULL);
    else if (value.compare("Std") == 0)      // standard
      itsRET.reset(new RetinaStd(*manager));
    else
      LFATAL("Unknown Retina type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsRET.get()) {
      addSubComponent(itsRET);

      // tell the controller to export its options:
      itsRET->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected RET of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ########## RetinaStd implementation
// ######################################################################
// ######################################################################
RetinaStd::RetinaStd(ModelManager& mgr, const std::string& descrName,
                     const std::string& tagName) :
  Retina(mgr, descrName, tagName),
  itsFramingImageName(this, "InputFramingImageName", ""),
  itsFramingImagePos(this, "InputFramingImagePos", Point2D(0, 0)),
  itsFoveateInput(this, "FoveateInput", false),  // see ModelOptionDefs.C
  itsFoveateInputDepth(this, "FoveateInputDepth", 5), // see ModelOptionDefs.C
  itsShiftInput(this, "ShiftInputToEye", false), // see ModelOptionDefs.C
  itsShiftInputBGcol(this, "ShiftInputToEyeBGcol", false), // idem
  itsInputFOV(this, "InputFOV", Dims(0, 0)),  // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", -1),  // see ModelOptionDefs.C
  itsSavePyr(this, "RetinaStdSavePyramid", false),
  itsFramingImage(), itsRetinalShift(0, 0)
{ }

// ######################################################################
void RetinaStd::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsFramingImageName);
    manager->requestOption(this, itsFramingImagePos);
    manager->requestOption(this, itsFoveateInput);
    manager->requestOption(this, itsFoveateInputDepth);
    manager->requestOption(this, itsShiftInput);
    manager->requestOption(this, itsShiftInputBGcol);
    manager->requestOption(this, itsInputFOV);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsSavePyr);
  }

  // the base class will propagate this call to our subcomponents:
  Retina::exportOptions(whatToExport, recurse);
}

// ######################################################################
void RetinaStd::start2()
{
  // if doing framing, read the framing image:
  if (itsFramingImageName.getVal().empty() == false)
    {
      itsFramingImage =
        Raster::ReadRGB(Auto, "%s", itsFramingImageName.getVal().c_str());
      LINFO("Using %dx%d framing image %s", itsFramingImage.getWidth(),
            itsFramingImage.getHeight(), itsFramingImageName.getVal().c_str());
    }
}

// ######################################################################
RetinaStd::~RetinaStd()
{ }

// ######################################################################
Image< PixRGB<byte> > RetinaStd::doGetOutput()
{
  // start with our latest input:
  Image< PixRGB<byte> > ret;

  // embed our latest raw input within a larger framing image if any:
  if (itsFramingImage.initialized())
    {
      ret = itsFramingImage;
      ret.paste(itsRawInput, itsFramingImagePos.getVal());
    }
  else  // if not doing framing, let's simply start with the raw input:
    ret = itsRawInput;

  // do we have a valid eye position?
  Point2D eye(-1, -1);
  if (itsBrain->hasModule("SaccadeController"))
    eye = itsBrain->getSC()->getPreviousDecision(0).p;

  // do we want to shift the input so that it is centered at the
  // current eye position?
  if (itsShiftInput.getVal())
    {
      if (itsBrain == NULL)
        LFATAL("I need a Brain to shift input");

      // do we have a valid eye position yet?
      if (eye.i != -1 || eye.j != -1)
        {
          LINFO("Shifting input to eye position (%d, %d)", eye.i, eye.j);
          
          itsRetinalShift.i += ret.getWidth() / 2 - eye.i;
          itsRetinalShift.j += ret.getHeight() / 2 - eye.j;
          
          ret = shiftClean(ret, itsRetinalShift.i, itsRetinalShift.j,
                           itsShiftInputBGcol.getVal());
        }

      // in addition, do we want to crop the shifted input to a
      // central field of view?
      Dims fovd(itsInputFOV.getVal());
      if (fovd.isEmpty() == false)
        {
          Point2D fovoff( (ret.getWidth() - fovd.w()) / 2,
                          (ret.getHeight() - fovd.h()) / 2);
          if (eye.i != -1 || eye.j != -1)
            {
              itsRetinalShift.i -= fovoff.i;
              itsRetinalShift.j -= fovoff.j;
            }
          ret = crop(ret, fovoff, fovd);
        }
    }

  // if doing foveation, create color pyramid for itsMultiRetina, and
  // foveate the input image, storing the result into itsRetina;
  // otherwise, itsMultiRetina remains unitialized and itsRetina is a
  // straight copy of the input frame:
  if (itsFoveateInput.getVal())
    {
      LINFO("Initializing multiretina (depth %d)",
            itsFoveateInputDepth.getVal());
      itsMultiRetina =
        buildPyrGaussian(ret, itsFoveateInputDepth.getVal(), 9);
      
      // create a mask with a disk at current overt fixation:
      Image<byte> mask(ret.getDims(), ZEROS);
      
      // do we have a valid eye position yet?
      if (eye.i != -1 && eye.j != -1)
        {
          mask.drawDisk(eye, itsFoveaRadius.getVal(), 255);
          LINFO("Foveating input at (%d, %d)", eye.i, eye.j);
        }
      else
        LINFO("Foveating input with uniform medium blur");
      
      // use the mask to foveate; if we don't have a SaccadeController but
      // have nevertheless requested itsFoveateInput, the behavior of
      // foveate() is to apply a uniform medium blur to the entire
      // image. If we do have a SaccadeController, this will only happen
      // as long as no overt fixation has been made yet:
      ret = foveate(mask, itsMultiRetina);
    }
  
  // ready to return (base class will do some caching on that):
  return ret;
}

// ######################################################################
Rectangle RetinaStd::getRawInputRectangle(const int border) const
{
  Point2D fpos = itsFramingImagePos.getVal();
  Rectangle r(itsRetinalShift.j + fpos.j + border,
              itsRetinalShift.i + fpos.i + border,
              itsRetinalShift.j + fpos.j + itsRawInput.getHeight()-1 - border,
              itsRetinalShift.i + fpos.i + itsRawInput.getWidth()-1 - border);
  r.fitToDims(itsOutputCache.getDims());
  return r;
}

// ######################################################################
Point2D RetinaStd::rawToRetinal(const Point2D& rawpos) const
{
  Point2D ret(rawpos);

  // if we are doing input framing, shift by the offset:
  ret.i += itsFramingImagePos.getVal().i;
  ret.j += itsFramingImagePos.getVal().j;
                
  // add any possible input shift:
  ret.i += itsRetinalShift.i;
  ret.j += itsRetinalShift.j;

  return ret;
}

// ######################################################################
Point2D RetinaStd::retinalToRaw(const Point2D& retpos) const
{
  Point2D ret(retpos);

  // remove any possible input shift:
  ret.i -= itsRetinalShift.i;
  ret.j -= itsRetinalShift.j;

  // if we are doing input framing, unshift by the offset:
  ret.i -= itsFramingImagePos.getVal().i;
  ret.j -= itsFramingImagePos.getVal().j;
                
  return ret;
}

// ######################################################################
void RetinaStd::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // have our base class save its stuff:
  Retina::saveResults(ofs);

  // save pyramid? (will force computation of output if not done yet)
  if (itsSavePyr.getVal() && itsRawInput.initialized() &&
      getOutput().initialized())
    for (uint i = 0; i < itsMultiRetina.size(); i ++)
      ofs->writeRGB(itsMultiRetina[i], PNM, -1, "RET%d-", i);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
