/*!@file Retina.H simulation of a human retina */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: Retina.H 4154 2005-02-01 20:29:21Z itti $
//

#ifndef RETINA_H_DEFINED
#define RETIBA_H_DEFINED

#include "saliency.H"
#include "ModelComponent.H"
#include "Point2D.H"
#include "Image.H"
#include "ImageSet.H"
class ModelManager;
class Brain;
class OutputFrameSeries;

// ######################################################################
//! The retina base class
/*! The Retina is the entry point to the Brain and VisualCortex, and
  this is a base class that mostly specifies the interface. Every
  input image to be processed first passes through the Retina (which
  is a subcomponent of Brain). The Retina may transform the input in
  various ways, e.g., apply a foveation filter, embed the input into a
  larger background framing image, or shift the image to current eye
  position. */

class Retina : public ModelComponent
{
public:
  //! Constructor
  /*! See ModelComponent.H for details */
  Retina(ModelManager& mgr, const std::string& descrName = "Retina",
         const std::string& tagName = "Retina");

  //! Set our brain
  /*! Some viewings may require a Brain. Since in the base class we
    don't have a Brain data member, the base class implementation of
    this function is a no-op. Derived classes should overload this
    accordingly if they need a Brain. */
  virtual void setBrain(Brain* brain);

  //! Destructor
  virtual ~Retina();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  /*! @name Access functions */
  //@{
  
  //! Input a new frame
  /*! The base implementation just keeps a copy of the input image in
  itsRawInput.  
  @param in the raw input image (e.g., from an InputFrameSeries). */
  virtual void input(const Image< PixRGB<byte> >& in);

  //! Have we received an input yet?
  /*! The base implementation just checks whether itsRawInput is
    initialized */
  virtual bool hasInput() const;

  //! Get latest raw input frame
  /*! The base implementation just returns itsRawInput */
  virtual Image< PixRGB<byte> > getRawInput() const;

  //! Get the output of the retina
  /*! The retina may apply a number of transforms to the input such as
    to yield the output, depending on ModelParam parameter
    values. This includes embedding the input within a larger
    background image, shifting it to current eye position, etc.
    Derived classes should not overload this getOutput() method, as it
    is a caching wrapper around doGetOutput() which should be
    overloaded. */
  virtual Image< PixRGB<byte> > getOutput();

  //! Get a rectangle delineating the raw input image area
  /*! As the raw input may be embedded into a background, shifted,
    cropped by a field of view, etc., this function allows users to
    recover the boundaries of the original frames. Note that the
    rectangle will be possibly cropped to fit within the dims of the
    Retina's output. A border may be specified that will be used to
    shrink (if positive) or enlarge (if negative) the returned
    rectangle (and will be applied before the rectangle is cropped to
    dims). The default implementation in the base class is to just
    return a rectangle at (0,0) with dims of the raw input. */
  virtual Rectangle getRawInputRectangle(const int border = 0) const;

  //! Translate a raw eye position from stimulus coordinates to retinal
  /*! The raw eye position should be given in the coordinates of the
    raw input frame (before shifting, embedding, etc). The returned
    position is in retinal coordinates, taking into account possible
    shifting, framing, fov'ing, etc. Default implementation in the
    base class is a no-op. */
  virtual Point2D rawToRetinal(const Point2D& rawpos) const;

  //! Translate a retinal eye position back into raw stimulus coordinates
  /*! The returned raw eye position is in the coordinates of the raw
    input frame (before shifting, embedding, etc). The given retinal
    position is in retinal coordinates, taking into account possible
    shifting, framing, fov'ing, etc. Default implementation in the
    base class is a no-op. */
  virtual Point2D retinalToRaw(const Point2D& retpos) const;

  //@}

  //! Save our internals using an OutputFrameSeries
  /*! Depending on our ModelParam settings, we can save our raw input
   frames (with name prefix "RETIN-"), and/or our foveated output images
   (prefix "RETOUT-") */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

protected:
  ModelParam<bool> itsSaveInput;    //!< Save our raw input?
  ModelParam<bool> itsSaveOutput;   //!< Save our output image?

  //! Compute a new output
  /*! Derived classes should overload this, which is called by
    getOutput() */
  virtual Image< PixRGB<byte> > doGetOutput() = 0;

  Brain *itsBrain;                      //!< keep track of our Brain
  Image< PixRGB<byte> > itsRawInput;    //!< last raw input image received
  Image< PixRGB<byte> > itsOutputCache; //!< last output computed
};


// ######################################################################
//! Retina configurator
// ######################################################################
/*! This will export the --retina-type=XX command-line option and will
  instantiate a Retina of the desired type as the option gets assigned a
  value. As this happens, new options may become available in the
  command-line. To see them, use --help AFTER you have chosen the type
  to use. The current Retina may be retrieved using getRET(). */
class RetinaConfigurator : public ModelComponent
{
public:
  //! Constructor
  RetinaConfigurator(ModelManager& mgr,
                     const std::string& descrName = "Retina Configurator",
                     const std::string& tagName = "RetinaConfigurator");

  //! destructor
  virtual ~RetinaConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our baby once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen Retina
  /*! You should call this during start() of the ModelComponent that
    needs the Retina. CAUTION: this may be a SharedPtr(NULL) if no
    Retina type has been chosen. */
  SharedPtr<Retina> getRET() const;

protected:
  ModelParam<std::string> itsType; //!< type of retina

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    retina of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<Retina> itsRET; // the retina
  int itsWhatToExport;
};

// ######################################################################
//! Standard Retina implementation
// ######################################################################
/*! Can do foveation, shifting, framing. */
class RetinaStd : public Retina
{
public:
  //! Constructor
  /*! See ModelComponent.H for details */
  RetinaStd(ModelManager& mgr, const std::string& descrName = "RetinaStd",
            const std::string& tagName = "RetinaStd");
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~RetinaStd();

  /*! @name Access functions */
  //@{

  // standard access functions are inherited fro the base class

  //! Get a rectangle delineating the raw input image area
  virtual Rectangle getRawInputRectangle(const int border = 0) const;
  
  //! Frame an image
  /*! This will embed the given image into a larger 'framing image'
    that may be specified via command-line options. The goal is to add
    a standard border around every incoming image. The image will be
    embedded with its top-left corner at the position specified via
    the command-line. The offset is available via getFramingOffset()
    at any time after start().  This function is provided in the
    public section so that users can frame various images themselves,
    but note that it is also called internally by getOutput() as soon
    as the relevant command-line options have non-empty values. */
  template <class T>
  Image<T> frameImage(const Image<T>& img, const Point2D& pos) const;

  template <class T>
  Image< PixRGB<T> > frameImage(const Image< PixRGB<T> >& img,
                                const Point2D& pos) const;

  //! Return current framing offset
  /*! This returns the coordinates of the top-left corner of the
    frames read from disk, once they have been 'framed' into a larger
    static background image using frameImage(). */
  Point2D getFramingOffset() const;

  //! Translate a raw eye position from stimulus coordinates to retinal
  virtual Point2D rawToRetinal(const Point2D& rawpos) const;

  //! Translate a retinal eye position back into raw stimulus coordinates
  virtual Point2D retinalToRaw(const Point2D& retpos) const;
  //@}

  //! Save our internals using an OutputFrameSeries
  /*! Depending on our ModelParam settings, we can save our pyramid
    frames (with name prefix "RET<level>-"), in addition to what the
    base class can save. */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

protected:
  //! Image file to use for input framing
  ModelParam<std::string> itsFramingImageName;

  //! Offset to apply for input framing
  ModelParam<Point2D> itsFramingImagePos;

  //! foveate the input?
  ModelParam<bool> itsFoveateInput;

  //! Pyramid depth to use when foveating the input if itsFoveateInput true
  ModelParam<uint> itsFoveateInputDepth;

  //! shift input to eye position?
  ModelParam<bool> itsShiftInput;

  //! background color to use when shifting input to eye position
  ModelParam< PixRGB<byte> > itsShiftInputBGcol;

  //! crop input to a given FOV?
  ModelParam<Dims> itsInputFOV;

  //! radius of the Fovea, in pixels
  ModelParam<int> itsFoveaRadius;

  //! Save our internal pyramid?
  ModelParam<bool> itsSavePyr;

  //! Compute a new output
  virtual Image< PixRGB<byte> > doGetOutput();

  //! get started
  virtual void start2();

private:
  Image< PixRGB<byte> > itsFramingImage; // image used for framing
  Point2D itsRetinalShift;
  ImageSet< PixRGB<byte> > itsMultiRetina;
};


#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
