/*!@file SaccadeController.C Base class for saccade generation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaccadeController.C 3271 2004-05-11 08:49:17Z itti $
//

#include "SaccadeController.H"
#include "log.H"

// ######################################################################
SaccadeController::SaccadeController(ModelManager& mgr,
                                     const std::string& descrName,
                                     const std::string& tagName,
                                     const int percept_qlen,
                                     const int decision_qlen,
                                     const int decision_qlen_head) :
  ModelComponent(mgr, descrName, tagName),
  itsBrain(NULL),
  pqlen(percept_qlen),
  dqlen(decision_qlen),
  dqlenh(decision_qlen_head),
  inSaccade(false),
  inBlink(false),
  inSaccadeHead(false)
{
  if (pqlen < 2 || dqlen < 2)
    LFATAL("Queue lengths must be at least 2 for proper operation.");
}

// ######################################################################
void SaccadeController::setBrain(Brain *brain)
{ itsBrain = brain; }

// ######################################################################
SaccadeController::~SaccadeController()
{  }

// ######################################################################
void SaccadeController::reset(const bool recurse)
{
  // reset some stuff for SaccadeController
  percept.clear();
  decision.clear();
  decisionh.clear();
  inSaccade = false;
  inBlink = false;
  inSaccadeHead = false;

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void SaccadeController::setPercept(const WTAwinner& fix)
{
  // push new percept to front of percept queue:
  percept.push_front(fix);

  // truncate percept queue if necessary:
  while(percept.size() > pqlen) percept.pop_back();

  // do some processing:
  computeWhenNewPercept();
}

// ######################################################################
void SaccadeController::resetPos(const Point2DT& p)
{
  // push new percept to front of percept queue:
  WTAwinner win; win.p = p.p; win.t = p.t;
  percept.push_front(win);

  // truncate percept queue if necessary:
  while(percept.size() > pqlen) percept.pop_back();

  // do some processing:
  computeWhenResetPos();
}

// ######################################################################
Point2D SaccadeController::getDecision(const double t)
{
  // enforce that passed time larger than that of last percept:
  if (percept.size() && t < percept.front().t)
  LERROR("Last percept was for t=%.3fs, asking time %.3fs must be larger.",
         percept.front().t, t);

  // do some computation and push a new decision into decision queue,
  // if the new decision is different from the last one pushed into
  // the queue:
  Point2D last(-1, -1);
  if (decision.size()) last = decision.front().p;
  Point2D loc = computeWhenNewDecision(t);
  bool nochange = (loc.i == last.i && loc.j == last.j);
  if (loc.i != -1 && nochange == false)
    decision.push_front(Point2DT(loc, t));

  // truncate decision queue as necessary:
  while(decision.size() > dqlen) decision.pop_back();

  // return current decision or (-1, -1) if no change since last time:
  if (nochange) return Point2D(-1, -1); else return loc;
}

// ######################################################################
Point2D SaccadeController::getDecisionHead(const double t)
{
  // enforce that passed time larger than that of last percept:
  if (percept.size() && t < percept.front().t)
  LERROR("Last percept was for t=%.3fs, asking time %.3fs must be larger.",
         percept.front().t, t);

  // do some computation and push a new decision into decision queue,
  // if the new decision is different from the last one pushed into
  // the queue:
  Point2D last(-1, -1);
  if (decision.size()) last = decisionh.front().p;
  Point2D loc = computeWhenNewDecisionHead(t);
  bool nochange = (loc.i == last.i && loc.j == last.j);
  if (loc.i != -1 && nochange == false)
    decisionh.push_front(Point2DT(loc, t));

  // truncate decision queue as necessary:
  while(decisionh.size() > dqlenh) decisionh.pop_back();

  // return current decision or (-1, -1) if no change since last time:
  if (nochange) return Point2D(-1, -1); else return loc;
}

// ######################################################################
void SaccadeController::dumpQueues() const
{
  char ptxt[percept.size() * 40], dtxt[decision.size() * 30];
  char tmp[40]; ptxt[0] = '\0'; dtxt[0] = '\0';
  for (unsigned int i = 0; i < percept.size(); i ++) {
    WTAwinner win = percept[i];
    sprintf(tmp, "[(%d, %d) %.2fms %.2fmV %.2fmV] ", win.p.i, win.p.j,
            win.t * 1000.0, win.sv * 1000.0, win.av * 1000.0);
    strcat(ptxt, tmp);
  }
  LINFO("PERCEPT: %s", ptxt);
  for (unsigned int i = 0; i < decision.size(); i ++) {
    Point2DT pt = decision[i];
    sprintf(tmp, "[(%d, %d) %.2fms] ", pt.p.i, pt.p.j, pt.t * 1000.0);
    strcat(dtxt, tmp);
  }
  LINFO("DECISION: %s", dtxt);
}

// ######################################################################
Point2D SaccadeController::computeWhenNewDecisionHead(const double t)
{ return getPreviousDecisionHead(0).p; }

// ######################################################################
WTAwinner SaccadeController::getPreviousPercept(const unsigned int
                                                       index) const
{
  if (percept.size() > index) return percept[index];
  else return WTAwinner(-1, -1, 0.0, 0.0, 0.0, false);
}

// ######################################################################
bool SaccadeController::havePercepts() const
{ return !(percept.empty()); }

// ######################################################################
int SaccadeController::getPqlen() const
{ return pqlen; }

// ######################################################################
void SaccadeController::resetPqlen(const int len)
{ percept.clear(); pqlen = len; }

// ######################################################################
void SaccadeController::killPercepts()
{ while(percept.size()) percept.pop_front(); }

// ######################################################################
Point2DT SaccadeController::getPreviousDecision(const unsigned int
                                                       index) const
{
  if (decision.size() > index) return decision[index];
  else return Point2DT(-1, -1, 0.0);
}

// ######################################################################
Point2DT SaccadeController::getPreviousDecisionHead(const unsigned int
                                                           index) const
{
  if (decisionh.size() > index) return decisionh[index];
  else return Point2DT(-1, -1, 0.0);
}

// ######################################################################
bool SaccadeController::haveDecisions() const
{ return !(decision.empty()); }

// ######################################################################
int SaccadeController::getDqlen() const
{ return dqlen; }

// ######################################################################
void SaccadeController::resetDqlen(const int len)
{ decision.clear(); dqlen = len; }

// ######################################################################
void SaccadeController::killDecisions()
{ while(decision.size()) decision.pop_front(); }

// ######################################################################
bool SaccadeController::haveDecisionsHead() const
{ return !(decisionh.empty()); }

// ######################################################################
int SaccadeController::getDqlenHead() const
{ return dqlenh; }

// ######################################################################
void SaccadeController::resetDqlenHead(const int len)
{ decisionh.clear(); dqlenh = len; }

// ######################################################################
void SaccadeController::killDecisionsHead()
{ while(decisionh.size()) decisionh.pop_front(); }

// ######################################################################
bool SaccadeController::isInSaccade() const
{ return inSaccade; }

// ######################################################################
bool SaccadeController::isInSaccadeHead() const
{ return inSaccadeHead; }

// ######################################################################
void SaccadeController::beginSaccade()
{ inSaccade = true; }

// ######################################################################
void SaccadeController::endSaccade()
{ inSaccade = false; }

// ######################################################################
bool SaccadeController::isInBlink() const
{ return inBlink; }

// ######################################################################
void SaccadeController::beginBlink()
{ inBlink = true; }

// ######################################################################
void SaccadeController::endBlink()
{ inBlink = false; }

// ######################################################################
void SaccadeController::beginSaccadeHead()
{ inSaccadeHead = true; }

// ######################################################################
void SaccadeController::endSaccadeHead()
{ inSaccadeHead = false; }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
