/*!@file SaccadeController.H Base class for saccade generation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaccadeController.H 3108 2004-02-10 22:07:18Z itti $
//

#ifndef SACCADECONTROLLER_H_DEFINED
#define SACCADECONTROLLER_H_DEFINED

#include "Point2DT.H"
#include "WTAwinner.H"
#include "ModelComponent.H"
#include <deque>
class Brain;

//! A base class to implement saccade controllers
/*! This base class handles the basic queuing of percepts and
fixations. Typically, new percepts (e.g., winner-take-all winners over
the saliency map) should be passed to the saccade controller using
setPercept(). During normal evolution of a simulation, evolve() should
be called at every time step, in a manner similar to that of Brain,
WTA and SM. In order to obtain coordinates of saccade targets computed
by this class, getDecision() should be called. It will return (-1, -1)
if no new saccade target is decided (either because it cannot be
decided, or because it is the same as the last valid returned
one). So, saccading to and fixating a new target will yield one
initial valid set of returned coordinates, followed by subsequent
returns of (-1, -1) as long as fixation is to be held
stationary. Derived classes should overload computeWhenNewPercept(),
called each time a new percept has arrived, and
computeWhenNewDecision(), called each time a new decision is
requested. In addition, derived classes should usually overload
evolve(), typically called at each time step during a simulation. For
example, the Brain class needs a SaccadeController derivative object
that implements evolve(), though that implementation may be a
no-op. The controller may include a head movement component; this
works just like the eye movement component. If no head movement is
being simulated, just use a decision queue length of zero for the
head. Callers can check whether a head is simulated using
haveDecisionsHead(). See SaccadeControllers.H for examples of derived
classes which actually implement saccadic control strategies. */

class SaccadeController : public ModelComponent
{
public:
  //! Constructor
  /*! @param percept_qlen length of the queue of accumulated percepts
    @param decision_qlen length of the queue of generated decisions
    @param decision_qlen_head length of the queue of generated decisions
    for the head. If your controller has no head component, just use 0 here. */
  SaccadeController(ModelManager& mgr, const std::string& descrName,
                    const std::string& tagName,
                    const int percept_qlen, const int decision_qlen,
                    const int decision_qlen_head = 0);

  //! Set a hook to our Brain
  virtual void setBrain(Brain *brain);
  
  //! Destructor (virtual to ensure that derived classes correctly destroyed)
  virtual ~SaccadeController();
  
  //! Reset SaccadeController
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! Receive a new percept (e.g., winner from saliency map)
  /*! This method should be invoked by external callers each time a
    new percept is available. Do not overload this method, overload
    computeWhenNewPercept() instead if you need to perform some
    computation for each new percept. */
  virtual void setPercept(const WTAwinner& fix);

  //! Reset to an externally-given position
  /*! This will force the current position to the position given as
    argument, disregarding any internal physical model. So this may be
    used to externally drive the SaccadeController to given
    locations. A percept will be queued up, which will be built from
    the passed position/time plus default values for the WTAwinner
    fields. Do not overload this method, overload computeWhenReset()
    instead. */
  virtual void resetPos(const Point2DT& p);

  //! Get the fixation decided by the controller
  /*! @param t should be the current time (in seconds) and should be
    greater than or equal to the time of the last percept passed to
    setPercept(). The coordinates of the decided fixation are returned
    (or (-1, -1) of no decision has been made). If a decision is made
    but is the same as the previous one, the behavior here is to
    return (-1, -1), and the caller has the responsibility of holding
    fixation until a new, valid decision can be reached. Do not
    overload this method, overload computeWhenNewDecision() instead if
    you need to do some computation before returning a fixation. */
  virtual Point2D getDecision(const double t);

  //! Same as getDecision() but for the head position
  virtual Point2D getDecisionHead(const double t);

  //! Get a previous percept
  /*! This method allows one to access a previous percept passed
    through setPercept(), or (-1, -1) if no valid previous percept
    exists.
    @param index indicates which previous percept; a value of zero
    will return the last percept that was passed to setPercept(), and
    increasing values will return increasingly older percepts. If
    index is more then the queue length, (-1, -1) will be returned. */
  WTAwinner getPreviousPercept(const unsigned int index = 0) const;

  //! Do we already have any percepts queued up?
  bool havePercepts() const;

  //! Returns length of percept queue
  /*! Note that the queue may not be full to that length yet */
  int getPqlen() const;

  //! Reset length of percept queue (contents will be lost)
  void resetPqlen(const int len);

  //! Kill all previous percepts (empty the percept queue)
  void killPercepts();

  //! Get a previous decision
  /*! This method allows one to access a previous valid decision. It
    will not attempt to determine whether a new decision can be made
    now, like getDecision() does. It will simply return the result of
    what was previously decided (or (-1, -1) if no valid previous
    decision exists).
    @param index indicates which previous decision; a value of zero
    will return the last decision that was made and returned by
    getDecision(), and increasing values will return increasingly
    older decisions. if index is more than the eueue length, (-1, -1)
    will be returned. */
  Point2DT getPreviousDecision(const unsigned int index = 0) const;

  //! Same as getPreviousDecision but for the head
  Point2DT getPreviousDecisionHead(const unsigned int index = 0) const;

  //! Do we already have any decisions queued up?
  bool haveDecisions() const;

  //! Returns length of decision queue
  /*! Note that the queue may not be full to that length yet */
  int getDqlen() const;

  //! Reset length of decision queue (contents will be lost)
  void resetDqlen(const int len);

  //! Kill all previous decisions (empty the decision queue)
  void killDecisions();

  //! Do we already have any head decisions queued up?
  bool haveDecisionsHead() const;

  //! Returns length of head decision queue
  /*! Note that the queue may not be full to that length yet */
  int getDqlenHead() const;

  //! Reset length of head decision queue (contents will be lost)
  void resetDqlenHead(const int len);

  //! Kill all previous decisions (empty the decision queue)
  void killDecisionsHead();

  //! evolve one time step (called, e.g., by Brain::evolve())
  /*! This method should be invoked by external callers to allow the
    internals of the controller to evolve as time passes by but no new
    percept is available. */
  virtual void evolve(const double t) = 0;

  //! Are we executing a saccade?
  bool isInSaccade() const;

  //! Are we executing a head "saccade"?
  bool isInSaccadeHead() const;

  //! Are we executing an eye blink?
  bool isInBlink() const;

  //! Show contents of the queues, for debugging
  void dumpQueues() const;

protected:
  //! Call this when you start a saccade (will make isInSaccade() true)
  void beginSaccade();

  //! Call this when you end a saccade (will make isInSaccade() false)
  void endSaccade();

  //! Call this when you start a blink (will make isInBlink() true)
  void beginBlink();

  //! Call this when you end a blink (will make isInBlink() false)
  void endBlink();

   //! Call this when you start a saccade (will make isInSaccade() true)
  void beginSaccadeHead();

  //! Call this when you end a saccade (will make isInSaccade() false)
  void endSaccadeHead();

 //! This method is called each time a new percept has arrived
  /*! The latest percept is at the front of the percept queue. Derived
    classes must provide some implementation for this method, as it is
    called from within setPercept(). */
  virtual void computeWhenNewPercept() = 0;

  //! This method is called each time a reset() occurs
  /*! The reset location/time is at the front of teh percept
    queue. Derived classes must provide some implementation for this
    method, as it is called from within reset(). */
  virtual void computeWhenResetPos() = 0;

  //! This method is called each time a new decision is requested
  /*! The latest percept is at the front of the percept queue, and the
    latest previous decision at the front of the decision queue. The
    method should return the decision to be taken given the time
    passed as argument. This method should return the current decision
    for fixation. The wrapping getDecision() call will take care of
    eliminating decisions that are the same as the previous ones.
    Derived classes must provide some implementation for this method,
    as it is called from within getDecision(). */
  virtual Point2D computeWhenNewDecision(const double t) = 0;

  //! Same as computeWhenNewDecision but for the head
  /*! A default no-op implementation is provided in the base class, so
    if your derived class has no head, you don't need to implement
    this. */
  virtual Point2D computeWhenNewDecisionHead(const double t);

  Brain *itsBrain;               //!< hook to our Brain or NULL

private:
  std::deque<WTAwinner> percept; // queue of accumulated percepts; front=latest
  std::deque<Point2DT> decision; // queue of accumulated decisions;front=latest
  std::deque<Point2DT> decisionh; // queue of accumulated head decisions
  unsigned int pqlen, dqlen;     // size of percept & decisioon queues
  unsigned int dqlenh;           // size of decision queue for head movements
  bool inSaccade;                // we are executing a saccade
  bool inBlink;                  // we are executing a blink
  bool inSaccadeHead;            // we are executing a saccade
};

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
#endif
