/*!@file SaccadeControllers.C Derived classes for saccade generation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaccadeControllers.C 4212 2005-02-17 03:51:20Z itti $
//

#include "SaccadeControllers.H"
#include "MathFunctions.H"
#ifdef HAVE_VALUES_H
#  include <values.h> // for MAXDOUBLE
#endif
#include "log.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Retina.H"
#include "Brain.H"

// ######################################################################
// ######################################################################
// ######################################################################
SaccadeControllerConfigurator::
SaccadeControllerConfigurator(ModelManager& mgr,
                              const std::string& descrName,
                              const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSacCtrlType(this, "SaccadeControllerType", "None"),
  itsSC(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
SaccadeControllerConfigurator::~SaccadeControllerConfigurator()
{  }
// ######################################################################

void SaccadeControllerConfigurator::exportOptions(const int whatToExport,
                                                  const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsSacCtrlType);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<SaccadeController> SaccadeControllerConfigurator::getSC() const
{ return itsSC; }

// ######################################################################
bool SaccadeControllerConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsSacCtrlType.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsSC.get()) { removeSubComponent(itsSC); itsSC.reset(NULL); }

    // instantiate a controller of the appropriate type:
    if (value.compare("None") == 0)
      itsSC.reset(NULL);
    else if (value.compare("Trivial") == 0)
      itsSC.reset(new TrivialSaccadeController(*manager));
    else if (value.compare("Fixed") == 0)
      itsSC.reset(new FixedSaccadeController(*manager));
    else if (value.compare("Friction") == 0)
      itsSC.reset(new FrictionSaccadeController(*manager));
    else if (value.compare("Threshold") == 0)
      itsSC.reset(new ThresholdSaccadeController(*manager));
    else if (value.compare("Threshfric") == 0)
      itsSC.reset(new ThresholdFrictionSaccadeController(*manager));
    else if (value.compare("Eyehead") == 0)
      itsSC.reset(new SimpleEyeHeadSaccadeController(*manager));
    else if (value.compare("Monkey") == 0)
      itsSC.reset(new MonkeySaccadeController(*manager));
    else if (value.compare("Monkey2") == 0)
      itsSC.reset(new Monkey2SaccadeController(*manager));
    else if (value.compare("ModelW") == 0)
      itsSC.reset(new ModelWrapperSaccadeController(*manager));
    else if (value.compare("EyeTrack") == 0)
      itsSC.reset(new EyeTrackerSaccadeController(*manager));
    else if (value.compare("VisBuf") == 0)
      itsSC.reset(new VisBufSaccadeController(*manager));
    else
      LFATAL("Unknown SaccadeController type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsSC.get()) {
      addSubComponent(itsSC);

      // tell the controller to export its options:
      itsSC->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected SC of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ######################################################################
TrivialSaccadeController::TrivialSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Trivial Saccade Controller",
                    "TrivialSaccadeController",
                    2, 2)  // queue length 2 for current and previous
{  }

// ######################################################################
TrivialSaccadeController::~TrivialSaccadeController()
{  }

// ######################################################################
void TrivialSaccadeController::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // let's first do our exports:

  // (we have none)

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void TrivialSaccadeController::evolve(const double t)
{  }

// ######################################################################
void TrivialSaccadeController::computeWhenNewPercept()
{  }

// ######################################################################
void TrivialSaccadeController::computeWhenResetPos()
{  }

// ######################################################################
Point2D TrivialSaccadeController::computeWhenNewDecision(const double t)
{
  // just return last percept received (or (-1,-1) if no percept yet):
  return getPreviousPercept(0).p;
}

// ######################################################################
// ######################################################################
// ######################################################################
FixedSaccadeController::FixedSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Fixed Saccade Controller", "FixedSaccadeController",
                    2, 2),  // queue length 2 for current and previous
  itsIEP(this, "SCeyeInitialPosition", Point2D(0, 0)) // see ModelOptionDefs.C
{  }

// ######################################################################
FixedSaccadeController::~FixedSaccadeController()
{  }

// ######################################################################
void FixedSaccadeController::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsIEP);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void FixedSaccadeController::evolve(const double t)
{  }

// ######################################################################
void FixedSaccadeController::saccade(const bool insac)
{
  if (isInSaccade() == false && insac == true) beginSaccade();
  else if (isInSaccade() == true && insac == false) endSaccade();
}

// ######################################################################
void FixedSaccadeController::blink(const bool inbli)
{
  if (isInBlink() == false && inbli == true) beginBlink();
  else if (isInBlink() == true && inbli == false) endBlink();
}

// ######################################################################
void FixedSaccadeController::computeWhenNewPercept()
{  }

// ######################################################################
void FixedSaccadeController::computeWhenResetPos()
{ itsIEP.setVal(getPreviousPercept(0).p); }

// ######################################################################
Point2D FixedSaccadeController::computeWhenNewDecision(const double t)
{ return itsIEP.getVal(); }

// ######################################################################
// ######################################################################
// ######################################################################
FrictionSaccadeController::FrictionSaccadeController(ModelManager& mgr,
                                                     const std::string pfx) :
  SaccadeController(mgr, pfx+" Friction Saccade Controller",
                    pfx+"FrictionSaccadeController",
                    2, 2),  // queue length 2 for current and previous
  itsSpringK(this, pfx+"SpringK", 50000.0),  // see ModelOptionDefs.C
  itsFrictionMu(this, pfx+"FrictionMu", 500.0),  // see ModelOptionDefs.C
  itsStartAtIEP(this, pfx+"StartAtIP", false),  // see ModelOptionDefs.C
  itsIEP(this, pfx+"InitialPosition", Point2D(0, 0)), // see ModelOptionDefs.C
  itsDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  itsTimeStep(this, "SimulationTimeStep", 0.0001), // see ModelOptionDefs.C
  x(0.0), y(0.0), oldx(0.0), oldy(0.0), oldt(-1.0)
{  }

// ######################################################################
void FrictionSaccadeController::start1()
{
  // start simulation and initially fixate at specified initial eye position?
  if (itsStartAtIEP.getVal())
    {
      WTAwinner win;  // create a fake percept at the IEP
      win.p = itsIEP.getVal();
      setPercept(win);
    }
}

// ######################################################################
FrictionSaccadeController::~FrictionSaccadeController()
{  }

// ######################################################################
void FrictionSaccadeController::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsDims);
    manager->requestOption(this, itsIEP);
    manager->requestOption(this, itsStartAtIEP);
    manager->requestOption(this, itsSpringK);
    manager->requestOption(this, itsFrictionMu);
    manager->requestOption(this, itsTimeStep);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void FrictionSaccadeController::evolve(const double t)
{
  if (oldt < 0.0) return;  // not simulating yet
  if (havePercepts() == false) { oldt = t; return; } // no percept yet

  // get FoA coordinates:
  WTAwinner last = getPreviousPercept(0);   // get latest percept
  double foax = last.p.i + 0.49999;
  double foay = last.p.j + 0.49999;

  // For stability of our difference equations, we need to run them at
  // the given simulation time step. Here, we however support for us
  // being called at much coarser time intervals. If we are called
  // more than 1.5 simulation time steps after last time we were
  // called, we'll loop over several smaller time steps as close to
  // itsTimeStep as possible:
  double dt, ts = itsTimeStep.getVal(), stop = t - ts * 0.01;
  if (t - oldt < 1.5 * ts) dt = t - oldt;       // do it in one step
  else dt = (t - oldt) / int((t - oldt) / ts);  // several steps

  for (double tt = oldt; tt < stop; tt += dt)
    {
      // compute force:
      double k = itsSpringK.getVal(), mu = itsFrictionMu.getVal();
      double fx = k * (foax - x);
      double fy = k * (foay - y);

      // update position:
      double newx = 2.0 * x - oldx + dt * dt * fx - dt * mu * (x - oldx);
      double newy = 2.0 * y - oldy + dt * dt * fy - dt * mu * (y - oldy);
      //LINFO("curr=(%.2f, %.2f) foa=(%.2f, %.2f) new=(%.2f, %.2f) "
      //"f=(%.2f, %.2f)", x, y, foax, foay, newx, newy, fx, fy);

      // keep track of the time step and current/old position:
      oldt = tt; oldx = x; oldy = y; x = newx; y = newy;
    }

  // finally turn off inSaccade if it had been set:
  if (isInSaccade()) endSaccade();
}

// ######################################################################
void FrictionSaccadeController::computeWhenNewPercept()
{
  // if this is our first percept, let's start the mass/spring now and here:
  if (oldt < 0.0)
    {
      WTAwinner last = getPreviousPercept(0);
      x = oldx = last.p.i + 0.49999;
      y = oldy = last.p.j + 0.49999;
      oldt = last.t;
    }
}

// ######################################################################
void FrictionSaccadeController::computeWhenResetPos()
{
  // get desired new fixation location:
  Point2D newloc = getPreviousPercept(0).p;

  // just shift overt attention coordinates to the new location,
  // disregarding physics of friction and the spring:
  x = newloc.i + 0.49999;
  y = newloc.j + 0.49999;
  oldx = x; oldy = y; oldt = getPreviousPercept(0).t;

  // we are doing a saccade:
  beginSaccade();
}

// ######################################################################
Point2D FrictionSaccadeController::computeWhenNewDecision(const double t)
{
  if (oldt < 0.0 || havePercepts() == false)
    return Point2D(-1, -1); // not started yet

  // round current coordinates to pixels:
  Point2D fi = Point2D(int(x), int(y));

  // possibly force fixation to inside image:
  int width = itsDims.getVal().w(), height = itsDims.getVal().h();
  if (fi.i < 0) fi.i = 0; else if (fi.i >= width) fi.i = width - 1;
  if (fi.j < 0) fi.j = 0; else if (fi.j >= height) fi.j = height - 1;

  // return our current rounded & clamped coordinates:
  return fi;
}

// ######################################################################
// ######################################################################
// ######################################################################
ThresholdSaccadeController::ThresholdSaccadeController(ModelManager& mgr,
                                                       const std::string pfx) :
  SaccadeController(mgr, pfx+" Threshold Saccade Controller",
                    pfx+"ThresholdSaccadeController", 2, 2),
  itsOdist(this, pfx+"ThreshMinOvert", 1.0F), // see ModelOptionDefs.C
  itsCdist(this, pfx+"ThreshMaxCovert", 0.5F),  // see ModelOptionDefs.C
  itsMinNum(this, pfx+"ThreshMinNum", 4),  // see ModelOptionDefs.C
  itsSalWeight(this, pfx+"ThreshSalWeigh", true),  // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", 30)  // see ModelOptionDefs.C
{  }

// ######################################################################
void ThresholdSaccadeController::start1()
{
  resetPqlen(itsMinNum.getVal());
  odistsq = float(itsFoveaRadius.getVal()) * itsOdist.getVal();
  odistsq *= odistsq;
  cdistsq = float(itsFoveaRadius.getVal()) * itsCdist.getVal();
  cdistsq *= cdistsq;
  didresetpos = false;
}

// ######################################################################
ThresholdSaccadeController::~ThresholdSaccadeController()
{  }

// ######################################################################
void ThresholdSaccadeController::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsOdist);
    manager->requestOption(this, itsCdist);
    manager->requestOption(this, itsMinNum);
    manager->requestOption(this, itsSalWeight);
    manager->requestOption(this, itsFoveaRadius);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ThresholdSaccadeController::evolve(const double t)
{ if (isInSaccade()) endSaccade(); }

// ######################################################################
void ThresholdSaccadeController::reset(const bool recurse)
{
  // reset some stuff for ThresholdSaccadeController
  didresetpos = false;

  // call reset in SaccadeController
  SaccadeController::reset(recurse);
}

// ######################################################################
void ThresholdSaccadeController::checkPercepts(bool& areclose, Point2D& avgp)
{
  // compute average location of queued percepts:
  float ii = 0.0f, jj = 0.0f, w = 0.0f; int idx = 0;
  while(1) {
    WTAwinner win = getPreviousPercept(idx);
    if (win.isValid() == false) break; // reached end of percept queue
    float weight = itsSalWeight.getVal() ? win.sv : 1.0f;
    ii += float(win.p.i) * weight;
    jj += float(win.p.j) * weight;
    w += weight; ++ idx;
  }
  if (w) { avgp.i = int(ii / w + 0.4999f); avgp.j = int(jj / w + 0.4999f); }
  else { avgp.i = -1; avgp.j = -1; }

  // test whether all of our percepts are within cdistsq of the average:
  idx = 0; areclose = true;
  while(1) {
    WTAwinner win = getPreviousPercept(idx);
    if (win.isValid() == false) break; // reached end of percept queue
    if (avgp.squdist(win.p) > cdistsq) { areclose = false; break; }
    ++ idx;
  }
}

// ######################################################################
void ThresholdSaccadeController::computeWhenNewPercept()
{  }

// ######################################################################
void ThresholdSaccadeController::computeWhenResetPos()
{
  // get desired new fixation location:
  resetloc = getPreviousPercept(0).p;
  didresetpos = true;

  // we are doing a saccade:
  beginSaccade();
}

// ######################################################################
Point2D ThresholdSaccadeController::computeWhenNewDecision(const double t)
{
  // if we just did a reset, return the reset location:
  if (didresetpos) { didresetpos = false; return resetloc; }

  // if no percept yet, return invalid position:
  if (havePercepts() == false) return Point2D(-1, -1);

  // are our queued-up percepts close to each other and get their average:
  bool areclose; Point2D avgp;
  checkPercepts(areclose, avgp);

  // if the covert shifts are nicely clustered, check that the average
  // if sufficiently far from our current overt location (if we have
  // made an overt shift already; otherwise make our first overt
  // shift if areclose is true):
  if (areclose && haveDecisions() == false)
    { beginSaccade(); return avgp; } // first overt shift
  if (areclose && avgp.squdist(getPreviousDecision(0).p) >= odistsq)
    { beginSaccade(); return avgp; } // do overt shift to avg covert location

  // either the covert shifts are not close to each other, or they are
  // too close to our current overt fixation. Don't move overt
  // attention yet and return whatever previous decision we have made
  // (if any):
  return getPreviousDecision(0).p;
}

// ######################################################################
// ######################################################################
// ######################################################################
ThresholdFrictionSaccadeController::
ThresholdFrictionSaccadeController(ModelManager& mgr, const std::string pfx) :
  SaccadeController(mgr, pfx+" Threshold-Friction Saccade Controller",
                    pfx+"ThresholdFrictionSaccadeController",
                    2, 2),  // queue length 2 for current and previous
  itsMaxIdle(this, pfx+"MaxIdleSecs", 0.75), // see ModelOptionDefs.C
  itsIEP(this, pfx+"InitialPosition", Point2D(0, 0)), // see ModelOptionDefs.C
  tsc(new ThresholdSaccadeController(mgr, pfx)),
  fsc(new FrictionSaccadeController(mgr, pfx))
{
  addSubComponent(tsc);
  addSubComponent(fsc);
}

// ######################################################################
ThresholdFrictionSaccadeController::~ThresholdFrictionSaccadeController()
{  }

// ######################################################################
void ThresholdFrictionSaccadeController::exportOptions(const int whatToExport,
                                                       const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsMaxIdle);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ThresholdFrictionSaccadeController::evolve(const double t)
{
  // evolve the friction part of our business:
  fsc->evolve(t);

  // evolve the threshold part of our business:
  tsc->evolve(t);

  // end saccade if any:
  if (isInSaccade()) endSaccade();
}

// ######################################################################
void ThresholdFrictionSaccadeController::computeWhenNewPercept()
{
  // get the percept we just received:
  WTAwinner win = getPreviousPercept(0);

  // Feed that new percept directly to the ThresholdSaccadeController:
  tsc->setPercept(win);

  // feed the average covert location as new percept to the friction
  // controller, only if the queued-up percepts are close to each
  // other (so that the average represents a fairly stable covert
  // fixation), and if that new percept is different from the last one
  // we gave to the FrictionController:
  bool areclose; Point2D avgp; tsc->checkPercepts(areclose, avgp);
  if (areclose) {
    Point2D flast = fsc->getPreviousPercept(0).p;
    if (avgp.i != -1 && (avgp.i != flast.i || avgp.j != flast.j)) {
      win.p = avgp;  // replace covert location by average covert location
      fsc->setPercept(win);
    }
  }
}

// ######################################################################
void ThresholdFrictionSaccadeController::computeWhenResetPos()
{
  // get the percept we just received:
  WTAwinner win = getPreviousPercept(0);
  Point2DT newloc(win.p, win.t);

  // reset friction:
  fsc->resetPos(newloc);

  // reset threhold:
  tsc->resetPos(newloc);

  // next time computeWhenNewDecision() is called, tsc will recommend
  // a move to the new location. Thus we are in saccade:
  beginSaccade();
}

// ######################################################################
Point2D ThresholdFrictionSaccadeController::computeWhenNewDecision(const
                                                                   double t)
{
  // what does the FrictionSaccadeController recommend?
  Point2D pf = fsc->getDecision(t);

  // what does the ThresholdSaccadeController recommend?
  Point2D pt = tsc->getDecision(t);


  // if we have a saccade recommendation from the threshold
  // controller, override the slow drift recommended by the friction
  // controller. Otherwise, if the friction controller is not
  // recommending a move either, see if we have been idle for too
  // long, in which case let's home to the center of the image:
  if (pt.i != -1 ||          // threshold controller has a new move, -- OR --
      (pt.i == -1 &&         // no new threshold move
       pf.i == -1 &&         // no new friction move
       haveDecisions() &&    // but we have moved in the past
       t - getPreviousDecision(0).t >
       itsMaxIdle.getVal())) // and it's been a long time
    {
      // if homing, let's get the coords:
      if (pt.i == -1) pt = itsIEP.getVal();

      // reset the mass/spring simulation of the friction controller
      // to new overt fixation:
      fsc->resetPos(Point2DT(pt, t));

      // we are saccading:
      beginSaccade();

      // do overt shift:
      return pt;
    }

  // otherwise, follow the recommendation of the friction controller,
  // if any (damped drifting of overt attention in between two
  // saccades). This will not yield chaotic damped movements, since
  // the friction controller is only fed with stable percepts
  // (queued-up percepts that are close to each other):
  if (pf.i != -1) return pf;

  // otherwise, return whatever our current position:
  return getPreviousDecision(0).p;
}

// ######################################################################
// ######################################################################
// ######################################################################
SimpleEyeHeadSaccadeController::
SimpleEyeHeadSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Simple Eye/Head Saccade Controller",
                    "SimpleEyeHeadSaccadeController", 2, 2, 2),
  eyesc(new ThresholdFrictionSaccadeController(mgr, "SCeye")),
  headsc(new FrictionSaccadeController(mgr, "SChead"))
{
  addSubComponent(eyesc);
  addSubComponent(headsc);
}

// ######################################################################
SimpleEyeHeadSaccadeController::~SimpleEyeHeadSaccadeController()
{  }

// ######################################################################
void SimpleEyeHeadSaccadeController::exportOptions(const int whatToExport,
                                                   const bool recurse)
{
  // let's first do our exports:

  // (we have none)

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimpleEyeHeadSaccadeController::evolve(const double t)
{
  // evolve the eye/thresholdfriction part of our business:
  eyesc->evolve(t);

  // evolve the head/friction part of our business:
  headsc->evolve(t);
}

// ######################################################################
void SimpleEyeHeadSaccadeController::computeWhenNewPercept()
{
  WTAwinner win = getPreviousPercept(0);

  // do the thresholdfriction part of our business:
  eyesc->setPercept(win);

  // the head will be filled-in with the decision of the eyes
}

// ######################################################################
void SimpleEyeHeadSaccadeController::computeWhenResetPos()
{
  LFATAL("Unimplemented");
}

// ######################################################################
Point2D SimpleEyeHeadSaccadeController::computeWhenNewDecision(const
                                                               double t)
{
  // figure out where the eyes are:
  Point2D eye = eyesc->getDecision(t);

  // feed the current eye position as new percept to the head
  // controller, if it is different from what the head controller
  // already has as its last percept:
  Point2D headlast = headsc->getPreviousPercept(0).p;
  if (eye.i != -1 && (eye.i != headlast.i || eye.j != headlast.j)) {
    WTAwinner win(eye, t, 0.0, 0.0, false);
    headsc->setPercept(win);
  }

  // we only return the eye movements here, so it's just whatever the
  // ThresholdFrictionController decides:
  return eye;
}

// ######################################################################
Point2D SimpleEyeHeadSaccadeController::computeWhenNewDecisionHead(const
                                                               double t)
{
  // get current head position:
  Point2D head = headsc->getDecision(t);   // run decision on headsc

  if (head.i != -1)
    LINFO("### head at (%d, %d) target (%d, %d) ###", head.i, head.j,
          headsc->getPreviousPercept(0).p.i,
          headsc->getPreviousPercept(0).p.j);

  return head;
}

// ######################################################################
// ######################################################################
// ######################################################################
MonkeyEyeHeadController::MonkeyEyeHeadController(ModelManager& mgr,
                                                 const std::string pfx) :
  SaccadeController(mgr, pfx+" Monkey Eye or Head Controller",
                    pfx+"MonkeyEyeHeadController", 2, 2),
  itsPPD(this, "PixelsPerDegree", 5.33333333), // see ModelOptionDefs.C
  itsDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  fsc(new FrictionSaccadeController(mgr, pfx))
{
  oldt = 0;
  addSubComponent(fsc);
  if (pfx.compare("SCeye") == 0) eyemodel = true; else eyemodel = false;
}

// ######################################################################
MonkeyEyeHeadController::~MonkeyEyeHeadController()
{  }

// ######################################################################
void MonkeyEyeHeadController::exportOptions(const int whatToExport,
                                            const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsDims);
    manager->requestOption(this, itsPPD);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void MonkeyEyeHeadController::saccade(const Point2DT& target)
{
  // our starting point is whatever fsc is currently at:
  Point2D currfsc = fsc->getPreviousDecision(0).p;
  pix2deg(currfsc, currx, curry);

  // initialize simulation:
  startt = oldt = target.t;
  double ampX, ampY; pix2deg(target.p, ampX, ampY);
  ampX -= currx; ampY -= curry;  // amplitude of displacement from current

  double amplitude = sqrt(ampX * ampX + ampY * ampY);
  theta = atan(ampY / ampX); if (ampX < 0.0) theta += M_PI;

  // If amplitude too small, don't even bother starting the saccade:
  if (amplitude < 0.0001) return;

  if (eyemodel)
    {
      // set us up for a realistic eye saccade:
      vmax = 473.0 * (1.0 - exp(-amplitude / 7.8)); // degrees/ms
      duration = 2.0 * amplitude / vmax;
      accel = 2.0 * vmax / duration; // degrees/s^2
    }
  else
    {
      // set us up for a realistic head saccade:
      vmax = (3.9 * amplitude + 16.0); // degrees/ms
      duration = 2.0 * amplitude / vmax;
      accel = 2.0 * vmax / duration; // degrees/ms^2
    }

  // we are in saccade:
  beginSaccade();

  LINFO("%s: Ampl=%.2fdeg vMax=%.2f dur=%.2fms theta=%.2fdeg start=%.2fms",
        eyemodel?"Eye ":"Head", amplitude, vmax, duration * 1000.0,
        theta * 180.0 / M_PI, startt * 1000.0);
}

// ######################################################################
void MonkeyEyeHeadController::pix2deg(const Point2D& pixloc,
                                      double& xdeg, double& ydeg) const
{
  double width = double(itsDims.getVal().w());
  double height = double(itsDims.getVal().h());
  double ppd = itsPPD.getVal();

  xdeg = atan((2.0 * pixloc.i / width - 1.0) *
              tan(width / ppd * M_PI / 360.0)) * 180.0 / M_PI;
  ydeg = atan((1.0 - 2.0 * pixloc.j / height) *
              tan(height / ppd * M_PI / 360.0)) * 180.0 / M_PI;
}

// ######################################################################
void MonkeyEyeHeadController::deg2pix(const double xdeg, const double ydeg,
                                      Point2D& pixloc) const
{
  double width = double(itsDims.getVal().w());
  double height = double(itsDims.getVal().h());
  double ppd = itsPPD.getVal();

  pixloc.i = int(0.5 * width * (1.0 + tan(xdeg * M_PI / 180.0) /
                                tan(width / ppd * M_PI / 360.0)));
  pixloc.j = int(0.5 * height * (1.0 - tan(ydeg * M_PI / 180.0) /
                                 tan(height / ppd * M_PI / 360.0)));
}

// ######################################################################
void MonkeyEyeHeadController::computeWhenNewPercept()
{
  // get our new percept:
  WTAwinner win = getPreviousPercept(0);

  // just pass it on to fsc; that's our new spring anchor point:
  fsc->setPercept(win);
}

// ######################################################################
void MonkeyEyeHeadController::evolve(const double t)
{
  // are we executing a realistic saccade?
  if (isInSaccade())
    {
      // do the realistic simulation:
      double deltaT = t - oldt, elapsedT = t - startt;

      // Get the average velocity during this interval, that's the
      // velocity at mid-point of the interval:
      double v;
      if (elapsedT > 0.5 * duration)
        v = vmax - accel * (elapsedT + 0.5 * deltaT - 0.5 * duration);
      else
        v = accel * (elapsedT + 0.5 * deltaT);

      // update current position:
      currx += v * cos(theta) * deltaT;
      curry += v * sin(theta) * deltaT;

      // drive the friction controller with our current coordinates:
      Point2D currpt; deg2pix(currx, curry, currpt);
      fsc->resetPos(Point2DT(currpt, t));

      // if reached the end, finish the saccade:
      if (elapsedT >= duration) {
        LINFO("=== %s: end of saccade at (%d, %d) at %.2fms ===",
              eyemodel?"Eye ":"Head", currpt.i, currpt.j, t * 1000.0);
        endSaccade();
      }
    }
  else
    // let the frictioncontroller evolve:
    fsc->evolve(t);

  // get ready for next time step:
  oldt = t;
}

// ######################################################################
void MonkeyEyeHeadController::computeWhenResetPos()
{
  // get our reset target:
  WTAwinner win = getPreviousPercept(0);
  Point2DT newloc(win.p, win.t);

  // feed it to fsc:
  fsc->resetPos(newloc);

  // abort any saccade we may be executing:
  if (isInSaccade()) endSaccade();
}

// ######################################################################
Point2D MonkeyEyeHeadController::computeWhenNewDecision(const double t)
{
  // since our realistic simulation directly drives the
  // FrictionController, it is always up to date, so return its
  // current position, or our current if fsc has nothing new for us:
  Point2D p = fsc->getDecision(t);
  if (p.i != -1) return p; else return getPreviousDecision(0).p;
}

// ######################################################################
// ######################################################################
// ######################################################################
Monkey2SaccadeController::
Monkey2SaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Monkey Saccade Controller v2",
                    "Monkey2SaccadeController", 2, 2, 2),
  itsBlinkWait(this, "SCeyeBlinkWaitTime", 1.8), // see ModelOptionDefs.C
  itsBlinkDur(this, "SCeyeBlinkDuration", 0.15), // see ModelOptionDefs.C
  itsOdist(this, "SCeyeThreshMinOvert", 1.0F), // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", 30),  // see ModelOptionDefs.C
  tsc(new ThresholdSaccadeController(mgr, "SCeye")),
  eyesc(new MonkeyEyeHeadController(mgr, "SCeye")),
  headsc(new MonkeyEyeHeadController(mgr, "SChead"))
{
  addSubComponent(tsc);
  addSubComponent(eyesc);
  addSubComponent(headsc);

  lastsbt = 0.0; blinkt = 0.0;

  // we need to start at some initial eye/head position for this
  // controller to work:
  manager->setOptionValString("SCeyeStartAtIP", "true");
  manager->setOptionValString("SCheadStartAtIP", "true");
}

// ######################################################################
Monkey2SaccadeController::~Monkey2SaccadeController()
{  }

// ######################################################################
void Monkey2SaccadeController::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsBlinkWait);
    manager->requestOption(this, itsBlinkDur);
    manager->requestOption(this, itsOdist);
    manager->requestOption(this, itsFoveaRadius);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void Monkey2SaccadeController::evolve(const double t)
{
  // evolve the thresholdfriction part of our business:
  tsc->evolve(t);

  // are we doing an eye saccade?
  bool esac = eyesc->isInSaccade();
  bool hsac = headsc->isInSaccade();

  // evolve the eye part of our business:
  eyesc->evolve(t);

  // did we just end our eye saccade?
  if (esac == true && eyesc->isInSaccade() == false)
    { endSaccade(); lastsbt = t; }

  // evolve the head part of our business:
  headsc->evolve(t);

  // did we just end our head saccade?
  if (hsac == true && headsc->isInSaccade() == false)
    endSaccadeHead();

  // are we blinking but time is up?
  if (isInBlink() == true && t - blinkt > itsBlinkDur.getVal())
    {
      LINFO("===== Ending eye blink at t=%.2fms =====", t * 1000.0);
      endBlink(); lastsbt = t;
    }

  // do we want to blink?
  if (isInBlink() == false && isInSaccade() == false &&
      t - lastsbt > itsBlinkWait.getVal())
    {
      // check whether our percepts are nicely clustered (no action going on)
      bool areclose; Point2D avgp;
      tsc->checkPercepts(areclose, avgp);
      if (areclose)
        {
          // start a blink with very small probability, or with a higher
          // probability if the percepts are clustered, or for sure if it
          // has been a long time even though the percepts may not be
          // clustered:
          double r = drand48();
          if (r > 0.99995 ||
              (r > 0.9999 && areclose == true) ||
              (r > 0.999 && t - lastsbt > itsBlinkWait.getVal() * 2.0) ||
              (r > 0.9 && t - lastsbt > itsBlinkWait.getVal() * 3.0))
            {
              LINFO("===== Starting blink at t=%.2fms (last at %.2fms)=====",
                    t * 1000.0, lastsbt * 1000.0);
              beginBlink(); blinkt = t;
            }
        }
    }
}

// ######################################################################
void Monkey2SaccadeController::computeWhenNewPercept()
{
  // get the percept just receoved:
  WTAwinner win = getPreviousPercept(0);

  // do the thresholdfriction part of our business:
  tsc->setPercept(win);

  // get our current eye and head positions:
  Point2D curreye = eyesc->getPreviousDecision(0).p;
  Point2D currhead = headsc->getPreviousDecision(0).p;

  // if tsc recommends a saccade, let's execute it:
  Point2D target = tsc->getDecision(win.t);
  if (target.i != -1)
    {
      win.p = target;  // use target location as percept location

      float sacthresh = float(itsFoveaRadius.getVal()) * itsOdist.getVal();
      // if the move recommended by tsc is large enough, and we are
      // not blinking, initiate a realistic saccade:
      if (target.distance(curreye) >= sacthresh && isInBlink() == false)
        {
          LINFO("=== Initiating saccade to (%d, %d) at %.2fms ===",
                target.i, target.j, win.t * 1000.0);

          // find the eye and head contributions:
          double eyecx, eyecy, headcx, headcy, tpx, tpy;
          eyesc->pix2deg(curreye, eyecx, eyecy);
          headsc->pix2deg(currhead, headcx, headcy);
          eyesc->pix2deg(target, tpx, tpy);

          // get horizontal head displacement:
          double hampx = headAmplitude(eyecx, headcx, tpx);

          // get vertical head displacement:
          double hampy = headAmplitude(eyecy, headcy, tpy);

          // now, we'll send the eyes directly to the target and we'll
          // send the head to an angular displacement of (hampx, hampy):
          eyesc->saccade(Point2DT(target, win.t));

          Point2D htarget;
          headsc->deg2pix(hampx + headcx, hampy + headcy, htarget);
          headsc->saccade(Point2DT(htarget, win.t));

          // we are in saccade:
          beginSaccade();
          beginSaccadeHead();
        }
      else
        {
          LINFO("=== Smooth pursuit to (%d, %d) at %.2fms ===",
                target.i, target.j, win.t * 1000.0);

          // we have a recommendation from tsc, but it's too small for
          // a realistic saccade. Just use it as friction targets:
          win.p = target; eyesc->setPercept(win);

          // the head follows the eye:
          win.p = eyesc->getPreviousDecision(0).p;
          headsc->setPercept(win);
        }
    }
  else
    {
      // we have no recommendation from tsc

      // set friction targets: the eye follows the moving average if
      // the recent percepts are nicely clustered, otherwise the eyes
      // just don't move:
      bool areclose; Point2D avgp;
      tsc->checkPercepts(areclose, avgp);
      if (areclose) {
        LINFO("=== Smooth pursuit to moving avg (%d, %d) at %.2fms ===",
              avgp.i, avgp.j, win.t * 1000.0);
        win.p = avgp;
        eyesc->setPercept(win);
      }

      // the head follows the eye:
      win.p = eyesc->getPreviousDecision(0).p;
      headsc->setPercept(win);
    }
}

// ######################################################################
void Monkey2SaccadeController::computeWhenResetPos()
{
  LFATAL("Unimplemented");
}

// ######################################################################
Point2D Monkey2SaccadeController::computeWhenNewDecision(const
                                                         double t)
{
  // get current eye position:
  Point2D eye = eyesc->getDecision(t);  // run decision on eyesc

  // feed the current eye position as new percept to the head
  // controller, if it is differemt from what the head controller
  // already has as its last percept:
  Point2D headlast = headsc->getPreviousPercept(0).p;
  if (eye.i != -1 && (eye.i != headlast.i || eye.j != headlast.j)) {
    WTAwinner win(eye, t, 0.0, 0.0, false);
    headsc->setPercept(win);
  }

  // we only return eye movements here, so it's whatever eyesc says:
  return eye;
}

// ######################################################################
Point2D Monkey2SaccadeController::computeWhenNewDecisionHead(const
                                                             double t)
{
  // get current head position:
  Point2D head = headsc->getDecision(t);   // run decision on headsc

  if (head.i != -1)
    LINFO("### head at (%d, %d) target (%d, %d) ###", head.i, head.j,
          headsc->getPreviousPercept(0).p.i,
          headsc->getPreviousPercept(0).p.j);

  return head;
}

// ######################################################################
double Monkey2SaccadeController::headAmplitude(const double curreye,
                                               const double currhead,
                                               const double target)
{
  // This implementation follows Freedman, Biol Cybern, 2001.

  // initial eye position with respect to the head. Sign is positive
  // if the eyes begin deviated in the direction of the subsequent
  // movement, negative otherwise:
  double iep = (curreye - currhead) * SGN(target - curreye);

  // compute desired gaze displacement:
  double hgazedd = target - curreye;

  // compute head displacement dead zone max boundary (note: we here
  // deviate from Freedman's formulation by clamping out negative
  // values):
  double hhdz = 0.56 * (20.0 - 0.5 * iep); if (hhdz < 0.0) hhdz = 0.0;

  // if desired gaze displacement in dead zone, do not move the head,
  // otherwise compute slope hslh of tradeoff between dead zone
  // boundary and actual gaze displacement amplitude, and derive head
  // displacement:
  double hheaddd = 0.0, hslh = 0.0;

  if (fabs(hgazedd) >= hhdz)
    {
      // I think Freedman forgot to take the absolute value of iep, so
      // we do that here. Also we'll make sure the slope is smaller
      // than 1, as otherwise Freeman's equation for hheaddd yields
      // weird results. Finally, let's also make sure both terms of
      // the tradeoff have the sign of hgazedd:
      hslh = 0.65 * (fabs(iep) / 35.0 + 1.1); if (hslh > 1.0) hslh = 1.0;
      hheaddd = hgazedd * hslh + hhdz * (1.0 - hslh) * SGN(hgazedd);
    }

  LINFO("curreye=%f currhead=%f target=%f iep=%f hgazedd=%f hhdz=%f "
        "hslh=%f hheaddd=%f", curreye, currhead, target, iep, hgazedd, hhdz,
        hslh, hheaddd);
  return hheaddd;
}

// ######################################################################
// ######################################################################
// ######################################################################
MonkeySaccadeController::
MonkeySaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Monkey Saccade Controller",
                    "MonkeySaccadeController", 2, 2),
  itsBlinkWait(this, "SCeyeBlinkWaitTime", 1.8), // see ModelOptionDefs.C
  itsBlinkDur(this, "SCeyeBlinkDuration", 0.15), // see ModelOptionDefs.C
  itsOdist(this, "SCeyeThreshMinOvert", 1.0F), // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", 30),  // see ModelOptionDefs.C
  itsPPD(this, "PixelsPerDegree", 5.33333333), // see ModelOptionDefs.C
  itsDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  headtsc(new ThresholdSaccadeController(mgr, "SChead")),
  tfsc(new ThresholdFrictionSaccadeController(mgr, "SCeye"))
{
  addSubComponent(tfsc);
  addSubComponent(headtsc);

  blinkqlen = 5;
  memset((void *)&eye, 0, sizeof(eye));
  memset((void *)&head, 0, sizeof(head));
  lastsbt = 0.0; blinkt = 0.0;
}

// ######################################################################
MonkeySaccadeController::~MonkeySaccadeController()
{  }

// ######################################################################
void MonkeySaccadeController::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsBlinkWait);
    manager->requestOption(this, itsBlinkDur);
    manager->requestOption(this, itsOdist);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsPPD);
    manager->requestOption(this, itsDims);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void MonkeySaccadeController::evolve(const double t)
{
  // evolve the ThresholdFriction part of our business:
  tfsc->evolve(t);

  // evolve the head threshold controller
  headtsc->evolve(t);

  // are we blinking but time is up?
  if (isInBlink() == true && t - blinkt > itsBlinkDur.getVal())
    {
      LINFO("===== Ending eye blink at t=%.2fms =====", t * 1000.0);
      endBlink(); lastsbt = t;
    }

  // do we want to blink? we can blink while in a saccade .. no ?
  if (isInBlink() == false && t - lastsbt > itsBlinkWait.getVal())
    {
      // start a blink with very small probability, or for sure if it
      // has been a long time even though the percepts may not be
      // clustered:
      double r = drand48();
      if (r > 0.999 ||
          (r > 0.95 && t - lastsbt > itsBlinkWait.getVal() * 2.0) ||
          (r > 0.8 && t - lastsbt > itsBlinkWait.getVal() * 3.0) )
        {
          LINFO("===== Starting blink at t=%.2fms (last at %.2fms)=====",
                t * 1000.0, lastsbt * 1000.0);
          beginBlink(); blinkt = t;
        }
    }

  head_evolve(t);
  eye_evolve(t);
}

// ######################################################################
void MonkeySaccadeController::computeWhenNewPercept()
{
  // get the percept we just received:
  WTAwinner win = getPreviousPercept(0);

  // Feed that new percept directly to the ThresholdFrictionSaccadeController:
  tfsc->setPercept(win);
}

// ######################################################################
void MonkeySaccadeController::computeWhenResetPos()
{
  // get the percept we just received:
  WTAwinner win = getPreviousPercept(0);
  Point2DT newloc(win.p, win.t);

  // reset ThresholdFriction:
  tfsc->resetPos(newloc);
}

// ######################################################################
Point2D MonkeySaccadeController::computeWhenNewDecision(const double t)
{
  int width = itsDims.getVal().w(), height = itsDims.getVal().h();
  double ppd = itsPPD.getVal();

  WTAwinner win = getPreviousPercept(0);
  if (havePercepts() == false)
    return Point2D(-1, -1);  // no percept yet

  // If we are currently in an eye saccade return the current coordinate
  if (inSaccade(&eye) == true || inSaccade(&head)==true ) {
    Point2DT ee = getEye();

    // force the ThresholdFriction controller to where the eyes are, so that
    // they can start smooth pursuit when our current saccade ends:
    tfsc->resetPos(ee);

    // return current eye position
    if(inSaccade(&eye)==true) return ee.p;
  }

  // what does the ThresholdFrictionSaccadeController recommend?
  Point2D ptf = tfsc->getDecision(t);
  Point2D cg = getEye().p;

  if ( ptf.i != -1)
    {
      float sacthresh = float(itsFoveaRadius.getVal()) * itsOdist.getVal();

      if(sqrt((ptf.i-cg.i)*(ptf.i-cg.i)+(ptf.j-cg.j)*(ptf.j-cg.j))>=sacthresh)
        {
          // compute angular positions corresponding to target point:
          double betaX = atan(2.0 * (ptf.i - width/2.0) / width *
                              tan((width / ppd) * M_PI / 360.0)) *
            180.0 / M_PI;
          double betaY = atan(2.0 * (ptf.j - height/2.0) / height *
                              tan((height / ppd) * M_PI / 360.0)) *
            180.0 / M_PI;

          // setup the realistic saccade controller to that target:
          setContrib(&eye, &head, betaX, betaY);
          setSaccadeTarget(&eye, t, 1);
          setSaccadeTarget(&head, t, 2);

          // initiate saccade in the realistic controller:
          LINFO("***** saccading to (%d, %d) = (%.2f, %.2f)deg *****",
                ptf.i, ptf.j, betaX, betaY);

          // tell brain we are in saccade:
          beginSaccade();
        }
      else
        { // attribute entire movement to head
            win.p = ptf; win.t = t;
            tfsc->setPercept(win);
            double betaX = atan(2.0 * (ptf.i - width/2.0) / width *
                                tan((width / ppd) * M_PI / 360.0)) *
              180.0 / M_PI;
            double betaY = atan(2.0 * (ptf.j - height/2.0) / height *
                                tan((height / ppd) * M_PI / 360.0)) *
              180.0 / M_PI;

            setAlign(&eye,&head,betaX,betaY);
            setSaccadeTarget(&head, t, 2);
          }
    }

  if(inSaccade(&head)==false)
    {
      // move head towards the eye
      double betaX = atan(2.0 * (cg.i - width/2.0) / width *
                          tan((width / ppd) * M_PI / 360.0)) * 180.0 / M_PI;
      double betaY = atan(2.0 * (cg.j - height/2.0) / height *
                          tan((height / ppd) * M_PI / 360.0)) * 180.0 / M_PI;
      setAlign(&eye,&head,betaX,betaY);
      setSaccadeTarget(&head, t, 2);
    }
  // return current gaze position:
  return getEye().p;
}

// ######################################################################
void MonkeySaccadeController::formulaEye(SomeInfo *inf)
{
  // timeleft = duration = (23 + 3.1*amplitude); // duration in ms
  inf->vMax = 473*(1.0 - exp(-inf->amplitude/7.8)); // degrees/ms
  inf->duration = 2*inf->amplitude/inf->vMax;
  inf->accel = 2*inf->vMax/(inf->duration); // degrees/s^2
  inf->latency =  0.160;
}

// ######################################################################
void MonkeySaccadeController::formulaHead(SomeInfo *inf)
{
  inf->vMax = (3.9*inf->amplitude + 16); // degrees/ms
  inf->timeleft = inf->duration = 2*inf->amplitude/inf->vMax;
  inf->accel = 2*inf->vMax/(inf->duration); // degrees/ms^2
  inf->latency =0.22;
}

// ######################################################################
double MonkeySaccadeController::getVelocity(SomeInfo *info, double t)
{
  if(t>info->duration/2)
    return info->vMax - (info->accel)*(t-info->duration/2);

  return info->accel*t;
}

// ######################################################################
double MonkeySaccadeController::getVelocityX(SomeInfo *info, const double t)
{
  double v = getVelocity(info, t);
  return v*cos(info->theta);
}

double MonkeySaccadeController::getVelocityY(SomeInfo *info, const double t)
{
  double v = getVelocity(info, t);
  return v*sin(info->theta);
}

// ######################################################################
void MonkeySaccadeController::setContrib(SomeInfo *e, SomeInfo *h,
                                         double tpX, double tpY)
{
  LINFO("eye(%f,%f)->(%f,%f)   head(%f,%f)->(%f,%f)  tp-> %f %f",
        e->cpX,e->cpY,e->tpX,e->tpY,h->cpX,h->cpY,h->tpX,h->tpY,tpX,tpY);

  double ampx = tpX - (e->cpX+h->cpX); // total amplitue X component
  double ampy = tpY - (e->cpY+h->cpY); //   "    "       Y component
  double cpx = e->cpX, cpy = e->cpY;

  if(ampx>0) e->sgnx = 1.0; else e->sgnx = -1.0;
  if(ampy>0) e->sgny = 1.0; else e->sgny = -1.0;
  if(ampx>0) h->sgnx = 1.0; else h->sgnx = -1.0;
  if(ampy>0) h->sgny = 1.0; else h->sgny = -1.0;

  if( (cpx > 0 && ampx < 0) || (cpx < 0 && ampx > 0) ) cpx = -cpx;
  else cpx = -1.0*cpx;
  if( (cpy > 0 && ampy < 0) || (cpy < 0 && ampy > 0) ) cpy = -cpy;
  else cpy = -1.0*cpy;

  ampx = fabs(ampx);
  ampy = fabs(ampy);

  // Horizontal Eye Displacement
  float thresholdE = -cpx/2 + 20;
  if(ampx < thresholdE)
    e->ampX = ampx;
  else {
    float slope = (-cpx/35.0 + 1)*0.35;
    e->ampX = ampx*slope + thresholdE*(1-slope);
  }

  h->ampX = ampx - e->ampX;

  // Vertical Eye Displacement
  thresholdE = -cpy/2 + 20;
  if(ampy< thresholdE)
    e->ampY = ampy;
  else {
    float slope = (-cpy/35.0 + 1)*0.35;
    e->ampY = ampy*slope + thresholdE*(1-slope);
  }

  h->ampY = ampy - e->ampY;

  e->ampX = ampx;
  e->ampY = ampy;


  LINFO("total=(%.2f, %.2f)deg eye=(%.2f, %.2f)deg head=(%.2f, %.2f) %f %f %f %f",
        ampx, ampy, e->ampX, e->ampY, h->ampX,h->ampY,e->sgnx,e->sgny,h->sgnx,
        h->sgny);
}

// ######################################################################
void MonkeySaccadeController::setAlign(SomeInfo *e, SomeInfo *h,
                                         double tpX, double tpY)
{
  double ampx = tpX - (h->cpX); // total amplitue X component
  double ampy = tpY - (h->cpY); //   "    "       Y component

  if(ampx>0) h->sgnx = 1.0; else h->sgnx = -1.0;
  if(ampy>0) h->sgny = 1.0; else h->sgny = -1.0;

  ampx = fabs(ampx);
  ampy = fabs(ampy);

  //e->ampX = e->ampY = 0.0;
  h->ampX = ampx; h->ampY = ampy;

  //  LINFO("total=(%.2f, %.2f)deg eye=(%.2f, %.2f)deg head=(%.2f, %.2f) %f %f %f %f",
  //        ampx, ampy, e->ampX, e->ampY, h->ampX,h->ampY,e->sgnx,e->sgny,h->sgnx,
  //        h->sgny);
}


// ######################################################################
void MonkeySaccadeController::setSaccadeTarget(SomeInfo *info, const double t,
                                               const int opt)
{
  info->isinsaccade = true;
  info->coveredx = info->coveredy = 0.0;

  info->tpX += info->sgnx*info->ampX; info->tpY += info->sgny*info->ampY;

  info->startt = info->oldt = t;

  info->amplitude = sqrt(info->ampX*info->ampX + info->ampY*info->ampY);
  info->theta = fabs(atan(info->ampY/info->ampX));

  if(info->ampX<0.00001 && info->ampY<0.00001)
    {
      info->isinsaccade = false;
      return;
    }

  if (opt==1) formulaEye(info);
  else if (opt==2) formulaHead(info);
  else LINFO("Unknow Option %d",opt);

  info->timeleft = info->duration;

  char *nam[3] = {"","Eye","Head"};
  LINFO("%s: Ampl=%.2fdeg vMax=%.2f dur=%.2fs theta=%.2fdeg dist=%.2f",
        nam[opt], info->amplitude, info->vMax, info->duration,
        info->theta*180/M_PI, info->vMax/2*info->duration);
}

void MonkeySaccadeController::beginVOR(const double t)
{
  invor = true;
}

void MonkeySaccadeController::endVOR(const double t)
{
  invor = false;
}

bool MonkeySaccadeController::inVOR()
{
  return invor;
}

// ######################################################################
void MonkeySaccadeController::eye_evolve(const double t)
{
  double deltaT = (t - eye.oldt);
  double elapsedT = (t - eye.startt);
  eye.oldt = t;

  eye.timeleft -= deltaT;
  double vx=0, vy=0;

  if (eye.isinsaccade == false) return;

  if(eye.timeleft < deltaT) {
    eye.isinsaccade = false;
    //    if(head.isinsaccade==false) endSaccade();
    LINFO("XXXXXXXXXXXXXXX EYE saccade ended !XXXXXXXXXXX");
    endSaccade();
  }  else {
    vx = (getVelocityX(&eye, elapsedT) +
          getVelocityX(&eye, elapsedT+deltaT))/2.0;
    vy = (getVelocityY(&eye, elapsedT) +
          getVelocityY(&eye, elapsedT+deltaT))/2.0;
    eye.coveredx += vx*deltaT;   eye.coveredy += vy*deltaT;
    eye.cpX += eye.sgnx*vx*deltaT;   eye.cpY += eye.sgny*vy*deltaT;

    //    LINFO("vx=%.2f vy=%.2f cpX=(%.2f, %.2f)deg TP=(%.2f,%.2f)deg",
    //          vx, vy, eye.cpX+head.cpX, eye.cpY+head.cpY, eye.tpX, eye.tpY);
  }

  return;
}

// ######################################################################
void MonkeySaccadeController::head_evolve(const double t)
{
  if (inSaccade(&head) == false) return;

  double deltaT = (t - head.oldt);  double elapsedT = (t - head.startt);
  head.oldt = t;

  head.timeleft -= deltaT;


  // Get the average velocity during this interval
  double vx = (getVelocityX(&head, elapsedT) +
               getVelocityX(&head, elapsedT+deltaT))/2.0;
  double vy = (getVelocityY(&head, elapsedT) +
               getVelocityY(&head, elapsedT+deltaT))/2.0;

  head.coveredx = vx*deltaT;   head.coveredy = vy*deltaT;
  head.cpX += head.sgnx*vx*deltaT;  head.cpY += head.sgny*vy*deltaT;
  eye.cpX  -= head.sgnx*vx*deltaT;  eye.cpY  -= head.sgny*vy*deltaT;

  if (eye.isinsaccade==false) {
    eye.tpX = eye.cpX; eye.tpY = eye.cpY;
  }

  //  LINFO("cpX=%0.2f cpY=%0.2f TP=(%0.2f,%0.2f)deg vx=(%0.2f,%0.2f)deg",
  //       head.cpX, head.cpY, head.tpX, head.tpY,vx,vy);

  if (head.timeleft < deltaT) {
    head.isinsaccade = false;
    LINFO("XXXXXXXXXXX end head saccade: cp=(%.2f, %.2f)", head.cpX, head.cpY);
    return;
  }
}

// ######################################################################
Point2DT MonkeySaccadeController::getEye() const
{
  int width = itsDims.getVal().w(), height = itsDims.getVal().h();
  double ppd = itsPPD.getVal();

  Point2DT pt;
  pt.p.i = int(tan((eye.cpX + head.cpX) * M_PI / 180.0) /
               tan((width / ppd) * M_PI / 360.0) * width / 2.0 + width / 2.0);
  pt.p.j = int(tan((eye.cpY + head.cpY) * M_PI / 180.0) /
               tan((height / ppd) * M_PI / 360.0) * height / 2.0 + height/2.0);
  pt.t = eye.oldt;
  return pt;
}

// ######################################################################
Point2DT MonkeySaccadeController::getJustEye() const
{
  int width = itsDims.getVal().w(), height = itsDims.getVal().h();
  double ppd = itsPPD.getVal();

  Point2DT pt;
  pt.p.i = int(tan((eye.cpX) * M_PI / 180.0) /
               tan((width / ppd) * M_PI / 360.0) * width / 2.0 + width / 2.0);
  pt.p.j = int(tan((eye.cpY) * M_PI / 180.0) /
               tan((height / ppd) * M_PI / 360.0) * height / 2.0 + height/2.0);
  pt.t = eye.oldt;
  return pt;
}

// ######################################################################
Point2DT MonkeySaccadeController::getHead() const
{
  int width = itsDims.getVal().w(), height = itsDims.getVal().h();
  double ppd = itsPPD.getVal();

  // c = tan(B)/tan(A/2)*w/2
  Point2DT pt;
  pt.p.i = int(tan(head.cpX * M_PI / 180.0) /
               tan((width / ppd) * M_PI / 360.0) * width / 2.0 + width / 2.0);
  pt.p.j = int(tan(head.cpY * M_PI / 180.0) /
               tan((height / ppd) * M_PI / 360.0) * height / 2.0 + height/2.0);
  pt.t = head.oldt;
  return pt;
}

// ######################################################################
void MonkeySaccadeController::writePos()
{
  Point2DT ee = getEye();
  Point2DT hh = getHead();
  LINFO("eye = (%d, %d)  head = (%d, %d)", ee.p.i, ee.p.j, hh.p.i, hh.p.j);
}

// ######################################################################
bool MonkeySaccadeController::inSaccade(SomeInfo *info)
{ return info->isinsaccade; }

// ######################################################################
// ######################################################################
// ######################################################################
EyeTrackerSaccadeController::EyeTrackerSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "EyeTracker Saccade Controller",
                    "EyeTrackerSaccadeController",
                    2, 2),  // queue length 2 for current and previous
  itsDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  itsEyeFname(this, "EyeTrackFname", ""),
  itsEyePeriod(this, "EyeTrackPeriod", 0.0041666666), // 240Hz
  itsEyeTrash(this, "EyeTrackTrash", 0),
  itsEyeSample(0), itsEyeFile(NULL), itsCurrEye(-1, -1),
  itsNextSaccadeTarget(-1, -1), itsNextSaccadeAmpl(0.0),
  itsNextFixationDuration(0.0)
{  }

// ######################################################################
EyeTrackerSaccadeController::~EyeTrackerSaccadeController()
{  }

// ######################################################################
void EyeTrackerSaccadeController::exportOptions(const int whatToExport,
                                                const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsDims);
    manager->requestOption(this, itsEyeFname);
    manager->requestOption(this, itsEyePeriod);
    manager->requestOption(this, itsEyeTrash);
  }

  // the base class will propagate this call to our subcomponents:
  SaccadeController::exportOptions(whatToExport, recurse);
}

// ######################################################################
void EyeTrackerSaccadeController::start1()
{
  // abort if no eye movement file:
  if (itsEyeFname.getVal().empty()) LFATAL("I need an eye movement file!");

  // open eye position file:
  itsEyeFile = fopen(itsEyeFname.getVal().c_str(), "r");
  if (itsEyeFile == NULL)
    PLFATAL("Cannot read '%s'", itsEyeFname.getVal().c_str());

  // trash some initial fixation samples if necessary:
  char trash[1024];
  for (int i = 0; i < itsEyeTrash.getVal(); i ++)
    fgets(trash, 1024, itsEyeFile);

  // ready for our first actually-used sample:
  itsEyeSample = 0;
}

// ######################################################################
void EyeTrackerSaccadeController::stop1()
{
  if (itsEyeFile) { fclose(itsEyeFile); itsEyeFile = NULL; }
}

// ######################################################################
void EyeTrackerSaccadeController::
getExtraEyeData(Point2D& nextSacTarget, double& nextSacAmpl,
                double& nextFixDuration)
{
  nextSacTarget = itsNextSaccadeTarget;
  nextSacAmpl = itsNextSaccadeAmpl;
  nextFixDuration = itsNextFixationDuration;

  // reset the extra data:
  itsNextSaccadeTarget.i = -1; itsNextSaccadeTarget.j = -1;
  itsNextSaccadeAmpl = 0.0; itsNextFixationDuration = 0.0;
}

// ######################################################################
void EyeTrackerSaccadeController::evolve(const double t)
{
  int w = itsDims.getVal().w(), h = itsDims.getVal().h();
  SharedPtr<Retina> retina = itsBrain->getRET();
  if (retina.get() == NULL) LFATAL("I need a Retina!");

  // loop over all fixations (if any) that happened since last evolve:
  while(double(itsEyeSample + 1) * itsEyePeriod.getVal() < t)
    {
      // current time:
      double currt = double(itsEyeSample) * itsEyePeriod.getVal();

      // get the eye movement data:
      double ex, ey, dx, dy, ampl; int status, durs;
      char line[512];
      if (fgets(line, 512, itsEyeFile) == NULL)
        CLDEBUG("No eye fixation available at t=%f ?", currt);
      else {
        bool gotextra = false;
        int n = sscanf(line, "%lf %lf %d %lf %lf %lf %d\n",
                       &ex, &ey, &status, &dx, &dy, &ampl, &durs);
        if (n == 7) gotextra = true;
        else if (n != 2) LERROR("Bogus eye file?");
        else status = 0;

#ifndef HAVE_ISNAN
        // built-in __isnand e.g. on Mac OSX
#  define isnan __isnand
#endif

        if (isnan(ex) == false && isnan(ey) == false)
          {
            int ei = int(ex + 0.49999), ej = int(ey + 0.49999);
            if (ei >= 0 && ei < w && ej >= 0 && ej < h)
              {
                CLINFO("Human eye %d (%d, %d)[%d] at %.1fms",
                       itsEyeSample + itsEyeTrash.getVal(),
                       ei, ej, status, itsEyeSample *
                       itsEyePeriod.getVal() * 1000.0);

                // convert from raw to retinal coordinates, taking
                // into account possible retinal shifting, embedding,
                // cropping to an FOV, etc:
                itsCurrEye = retina->rawToRetinal(Point2D(ei, ej));

                // if we have extra data from segmented
                // saccades/blinks, let's process it now:
                if (gotextra) {
                  // store extra data for later, if non-zero:
                  if (ampl > 0.0)
                    {
                      itsNextSaccadeTarget =
                        retina->rawToRetinal(Point2D(int(dx + 0.49999),
                                                     int(dy + 0.49999)));
                      CLINFO("**** Saccade to (%d, %d) ****",
                             itsNextSaccadeTarget.i, itsNextSaccadeTarget.j);
                      itsNextSaccadeAmpl = ampl;
                      itsNextFixationDuration =
                        double(durs) * itsEyePeriod.getVal();
                    }

                  // trigger/end a saccade/blink depending on status:
                  if (status & 1) // saccade
                    { if (isInSaccade() == false) beginSaccade(); }
                  else
                    { if (isInSaccade()) endSaccade(); }

                  // how about blinks?
                  if (status & 2)
                    { if (isInBlink() == false) beginBlink(); }
                  else
                    { if (isInBlink()) endBlink(); }
                }
              }
          }
      }
      // ready for next eye movement sample:
      itsEyeSample ++;

      // kill our decision queue, so that we will return our current eye
      // position as an eye movement event even though we may not have
      // moved. The reason for that is to keep files that have one line
      // per eye movement sample in sync:
      killDecisions();
    }
}

// ######################################################################
void EyeTrackerSaccadeController::computeWhenNewPercept()
{
  // we ignore percepts given to us by Brain, since eye position is
  // not driven by Brain but by the human data read from file.
}

// ######################################################################
void EyeTrackerSaccadeController::computeWhenResetPos()
{  }

// ######################################################################
Point2D EyeTrackerSaccadeController::computeWhenNewDecision(const double t)
{
  // just return our current eye position:
  return itsCurrEye;
}

// ######################################################################
double EyeTrackerSaccadeController::getEyePeriod() const
{ return itsEyePeriod.getVal(); }


// ######################################################################
// ######################################################################
// ######################################################################
VisBufSaccadeController::VisBufSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "VisBuf Saccade Controller",
                    "VisBufSaccadeController",
                    2, 2),  // queue length 2 for current and previous
  itsDims(this, "VisBufDims", Dims(0, 0)) // see ModelOptionDefs.C
{  }

// ######################################################################
VisBufSaccadeController::~VisBufSaccadeController()
{  }

// ######################################################################
void VisBufSaccadeController::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsDims);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void VisBufSaccadeController::start1()
{
}

// ######################################################################
void VisBufSaccadeController::stop1()
{
}

// ######################################################################
void VisBufSaccadeController::evolve(const double t)
{
}

// ######################################################################
void VisBufSaccadeController::computeWhenNewPercept()
{
}

// ######################################################################
void VisBufSaccadeController::computeWhenResetPos()
{  }

// ######################################################################
Point2D VisBufSaccadeController::computeWhenNewDecision(const double t)
{
  return Point2D(0,0);
}

// ######################################################################
// ######################################################################
// ######################################################################
ModelWrapperSaccadeController::
ModelWrapperSaccadeController(ModelManager& mgr) :
  SaccadeController(mgr, "Model Wrapper Saccade Controller",
                    "ModelWrapperSaccadeController", 2, 2, 2),
  itsOdist(this, "SCeyeThreshMinOvert", 1.0F), // see ModelOptionDefs.C
  itsFoveaRadius(this, "FoveaRadius", 30),  // see ModelOptionDefs.C
  tsc(new ThresholdSaccadeController(mgr, "SCeye")),
  eyesc(new ModelEyeHeadController(mgr, "SCeye")),
  headsc(new ModelEyeHeadController(mgr, "SChead")),
  itsBlinkTime(0.0), itsBlinkDuration(0.0), idum(-123)
{
  addSubComponent(tsc);
  addSubComponent(eyesc);
  addSubComponent(headsc);

  // we need to start at some initial eye/head position for this
  // controller to work:
  manager->setOptionValString("SCeyeStartAtIP", "true");
  manager->setOptionValString("SCheadStartAtIP", "true");
}

// ######################################################################
ModelWrapperSaccadeController::~ModelWrapperSaccadeController()
{  }

// ######################################################################
void ModelWrapperSaccadeController::start1()
{
  // start our base object:
  SaccadeController::start1();

  // init random numbers:
  bool usernd; manager->getModelParamVal("UseRandom", usernd);
  idum = getIdum(usernd);

  // schedule our first blink:
  pickNextBlink(0.0);
}

// ######################################################################
void ModelWrapperSaccadeController::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsOdist);
    manager->requestOption(this, itsFoveaRadius);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ModelWrapperSaccadeController::evolve(const double t)
{
  // evolve the threshold part of our business:
  tsc->evolve(t);

  // are we doing an eye saccade?
  bool esac = eyesc->isInSaccade();
  bool hsac = headsc->isInSaccade();

  // evolve the eye part of our business:
  eyesc->evolve(t);

  // did we just end our eye saccade?
  if (esac == true && eyesc->isInSaccade() == false) endSaccade();

  // evolve the head part of our business:
  headsc->evolve(t);

  // did we just end our head saccade?
  if (hsac == true && headsc->isInSaccade() == false) endSaccadeHead();

  // are we blinking but time is up?
  if (isInBlink() == true && t > itsBlinkTime + itsBlinkDuration)
    {
      LINFO("===== Ending eye blink at t=%.2fms =====", t * 1000.0);
      endBlink();

      // let's pick our next blink:
      pickNextBlink(t); // will set itsBlinkTime and itsBlinkDuration
    }

  // time to execute our scheduled next blink?
  if (isInBlink() == false && t >= itsBlinkTime)
    {
      // check whether percepts are nicely clustered (no action going
      // on), i.e., we are fixating on a rather stable location. If
      // not, wait until that becomes the case:
      bool areclose; Point2D avgp; tsc->checkPercepts(areclose, avgp);
      if (areclose)
        {
          // ok, let's start a blink:
          LINFO("===== Starting eye blink at t=%.2fms (duration %.2fms) =====",
                t * 1000.0, itsBlinkDuration * 1000.0);
          beginBlink(); itsBlinkTime = t;
        }
    }
}

// ######################################################################
void ModelWrapperSaccadeController::pickNextBlink(const double t)
{
  const double BLINKDURMEAN = 0.190; // mean blink duration
  const double BLINKDURSTD = 0.040;  // stdev blink duration
  const double BLINKCOEFF = 0.12;    // exponential factor blink start

  // pick the time and duration for our next blink. From fitting a
  // collection of human blinks, we know that P(blink at t|blink ended
  // at t0)=0.5 exp(-0.12(t-t0)), and duration is Gaussian distributed
  // with mean 190ms and stdev 40ms:
  while(true)
    {
      itsBlinkTime = t + expdev(idum) / BLINKCOEFF;
      itsBlinkDuration = BLINKDURMEAN + BLINKDURSTD * gasdev(idum);
      
      // accept if values drawn were not too extreme:
      if (itsBlinkTime > t +  0.100 &&
          itsBlinkTime < t + 15.000 &&
          itsBlinkDuration > BLINKDURMEAN - 3.0 * BLINKDURSTD &&
          itsBlinkDuration < BLINKDURMEAN + 3.0 * BLINKDURSTD) break;
    }
  
  LINFO("===== Scheduled eye blink for t=%.2fms (duration %.2fms) =====",
        itsBlinkTime * 1000.0, itsBlinkDuration * 1000.0);
}

// ######################################################################
void ModelWrapperSaccadeController::computeWhenNewPercept()
{
  // get the percept just received:
  WTAwinner win = getPreviousPercept(0);

  // do the threshold part of our business:
  tsc->setPercept(win);

  // get our current eye and head positions:
  Point2D curreye = eyesc->getPreviousDecision(0).p;
  Point2D currhead = headsc->getPreviousDecision(0).p;

  // if tsc recommends a saccade, let's execute it:
  Point2D target = tsc->getDecision(win.t);
  if (target.i != -1)
    {
      win.p = target;  // use target location as percept location

      float sacthresh = float(itsFoveaRadius.getVal()) * itsOdist.getVal();
      // if the move recommended by tsc is large enough, initiate a
      // realistic saccade:
      if (target.distance(curreye) >= sacthresh)
        {
          LINFO("=== Initiating saccade to (%d, %d) at %.2fms ===",
                target.i, target.j, win.t * 1000.0);

          // find the eye and head contributions:
          double eyecx, eyecy, headcx, headcy, tpx, tpy;
          eyesc->pix2deg(curreye, eyecx, eyecy);
          headsc->pix2deg(currhead, headcx, headcy);
          eyesc->pix2deg(target, tpx, tpy);

          // get horizontal head displacement:
          double hampx = headAmplitude(eyecx, headcx, tpx);

          // get vertical head displacement:
          double hampy = headAmplitude(eyecy, headcy, tpy);

          LINFO("Eye=(%.2f,%.2f)deg, Head=(%.2f,%.2f)deg, "
                "Target=(%.2f,%.2f)deg, HeadContrib=(%.2f,%.2f)deg",
                eyecx, eyecy, headcx, headcy, tpx, tpy, hampx, hampy);

          // now, we'll send the eyes directly to the target and we'll
          // send the head to an angular displacement of (hampx, hampy):
          eyesc->saccade(Point2DT(target, win.t));

          Point2D htarget;
          headsc->deg2pix(hampx + headcx, hampy + headcy, htarget);
          headsc->saccade(Point2DT(htarget, win.t));

          // we are in saccade:
          beginSaccade();
          beginSaccadeHead();
        }
      else
        {
          LINFO("=== Smooth pursuit to (%d, %d) at %.2fms ===",
                target.i, target.j, win.t * 1000.0);

          // we have a recommendation from tsc, but it's too small for
          // a realistic saccade. Just use it as friction targets:
          win.p = target; eyesc->setPercept(win);

          // the head follows the eye:
          win.p = eyesc->getPreviousDecision(0).p;
          headsc->setPercept(win);
        }
    }
  else
    {
      // we have no recommendation from tsc

      // set friction targets: the eye follows the moving average if
      // the recent percepts are nicely clustered, otherwise the eyes
      // just don't move:
      bool areclose; Point2D avgp;
      tsc->checkPercepts(areclose, avgp);
      if (areclose) {
        LINFO("=== Smooth pursuit to moving avg (%d, %d) at %.2fms ===",
              avgp.i, avgp.j, win.t * 1000.0);
        win.p = avgp;
        eyesc->setPercept(win);
      }

      // the head follows the eye:
      win.p = eyesc->getPreviousDecision(0).p;
      headsc->setPercept(win);

      // added
      currhead = win.p;
    }
}

// ######################################################################
void ModelWrapperSaccadeController::computeWhenResetPos()
{
  LFATAL("Unimplemented");
}

// ######################################################################
Point2D ModelWrapperSaccadeController::computeWhenNewDecision(const
                                                         double t)
{
  // get current eye position:
  Point2D eye = eyesc->getDecision(t);  // run decision on eyesc

  // feed the current eye position as new percept to the head
  // controller, if it is differemt from what the head controller
  // already has as its last percept:
  Point2D headlast = headsc->getPreviousPercept(0).p;
  if (eye.i != -1 && (eye.i != headlast.i || eye.j != headlast.j)) {
    WTAwinner win(eye, t, 0.0, 0.0, false);
    headsc->setPercept(win);
  }

  // we only return eye movements here, so it's whatever eyesc says:
  return eye;
}

// ######################################################################
Point2D ModelWrapperSaccadeController::computeWhenNewDecisionHead(const
                                                                  double t)
{
  // get current head position:
  return headsc->getDecision(t);   // run decision on headsc
}

// ######################################################################
double ModelWrapperSaccadeController::headAmplitude(const double curreye,
                                                    const double currhead,
                                                    const double target)
{
  double iep = (curreye - currhead) * SGN(target - curreye);
  double amp = fabs(target - curreye);
  double e = 0.0; // eye contribution; head contribution will be amp - e

  if (iep > 0)
    { if (amp < 25) e = amp; else e = 25 + 0.18*(amp - 25); }
  else
    {
      if (iep >= -15)
        { if (amp < 35) e = amp; else e = 35 + 0.22*(amp - 35); }
      else if (iep >= -25)
        { if (amp < 45) e = amp; else e = 45 + 0.28*(amp - 45); }
      else if (iep >= -35)
        { if (amp < 55) e = amp; else e = 55 + 0.40*(amp - 55); }
      else
        {
          LINFO("IEP is very off: %fdeg -- HACKING", iep);
          e = amp;
        }
    }
  return (amp - e) * SGN(target - curreye);
}

// ######################################################################
// ######################################################################
// ######################################################################
ModelEyeHeadController::ModelEyeHeadController(ModelManager& mgr,
                                               const std::string pfx) :
  SaccadeController(mgr, pfx+" Monkey Eye or Head Controller",
                    pfx+"ModelEyeHeadController", 2, 2),
  itsPPD(this, "PixelsPerDegree", 5.33333333), // see ModelOptionDefs.C
  itsDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  fsc(new FrictionSaccadeController(mgr, pfx)),
  eyemodel(false), duration(0.0), ampX(0.0), ampY(0.0), startt(0.0),
  currX(0.0), currY(0.0), prevX(0.0), prevY(0.0)
{
  addSubComponent(fsc);
  if (pfx.compare("SCeye") == 0) eyemodel = true;
}

// ######################################################################
ModelEyeHeadController::~ModelEyeHeadController()
{  }

// ######################################################################
void ModelEyeHeadController::exportOptions(const int whatToExport,
                                            const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsDims);
    manager->requestOption(this, itsPPD);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ModelEyeHeadController::saccade(const Point2DT& target)
{
  // our starting point is whatever fsc is currently at:
  Point2D currfsc = fsc->getPreviousDecision(0).p;
  pix2deg(currfsc, currX, currY);  // currX,currY = current position in deg
  prevX = currX; prevY = currY;    // save it, will be required later

  // initialize simulation:
  startt = target.t;
  pix2deg(target.p, ampX, ampY);
  ampX -= currX; ampY -= currY;  // ampl in deg of displacement from current
  double amplitude = sqrt(ampX * ampX + ampY * ampY);

  // If amplitude too small, don't even bother starting the saccade:
  if (amplitude < 0.5) return;

  // set us up for a realistic eye or head saccade:
  if (eyemodel) duration = 0.00164 * amplitude + 0.073;
  else          duration = 0.01000 * amplitude + 0.350;

  // we are in saccade:
  beginSaccade();

  LINFO("%s: Ampl=%.2fdeg dur=%.2fms start=%.2fms",
        eyemodel?"Eye ":"Head", amplitude, duration * 1000.0,
        startt * 1000.0);
}

// ######################################################################
void ModelEyeHeadController::pix2deg(const Point2D& pixloc,
                                      double& xdeg, double& ydeg) const
{
  double width = double(itsDims.getVal().w());
  double height = double(itsDims.getVal().h());
  double ppd = itsPPD.getVal();

  xdeg = atan((2.0 * pixloc.i / width - 1.0) *
              tan(width / ppd * M_PI / 360.0)) * 180.0 / M_PI;
  ydeg = atan((1.0 - 2.0 * pixloc.j / height) *
              tan(height / ppd * M_PI / 360.0)) * 180.0 / M_PI;
}

// ######################################################################
void ModelEyeHeadController::deg2pix(const double xdeg, const double ydeg,
                                      Point2D& pixloc) const
{
  double width = double(itsDims.getVal().w());
  double height = double(itsDims.getVal().h());
  double ppd = itsPPD.getVal();

  pixloc.i = int(0.5 * width * (1.0 + tan(xdeg * M_PI / 180.0) /
                                tan(width / ppd * M_PI / 360.0)));
  pixloc.j = int(0.5 * height * (1.0 - tan(ydeg * M_PI / 180.0) /
                                 tan(height / ppd * M_PI / 360.0)));
}

// ######################################################################
void ModelEyeHeadController::computeWhenNewPercept()
{
  // get our new percept:
  WTAwinner win = getPreviousPercept(0);

  // just pass it on to fsc; that's our new spring anchor point:
  fsc->setPercept(win);
}

// our damped harmonig is exp(-k(x-1)) [sin(pi/2*x)]^a; we have (x-1)
// in the exponential to account for a normalizing constant, so that
// the function is equal to exactly 1 when x = 1:
#define DAMPED_HARMONIC(a,k,x) (exp(-k*(x-1.0f)) * pow(sin(M_PI_2*x),a))

// ######################################################################
void ModelEyeHeadController::evolve(const double t)
{
  // are we executing a realistic saccade?
  if (isInSaccade())
    {
      // do the realistic simulation:
      double elapsedT = t - startt, a, k;
      static const double a_eye  = 3.36, k_eye  = -0.001767;
      static const double a_head = 1.70, k_head =  0.011500;
      if (eyemodel) { a = a_eye; k = k_eye; } else { a = a_head, k = k_head; }

      // update current position:
      if (elapsedT < duration)
        {
          // Saccade is in progress; use the damped harmonic to
          // compute current position:
          double x = elapsedT / duration;
          currX = prevX + ampX * DAMPED_HARMONIC(a, k, x);
          currY = prevY + ampY * DAMPED_HARMONIC(a, k, x);
        }
      else
        {
          // we are at the end of saccade, possibly overshooting due to
          // discrete time steps:
          currX = prevX + ampX;
          currY = prevY + ampY;
        }

      // drive the friction controller with our current coordinates:
      Point2D currpt; deg2pix(currX, currY, currpt);
      fsc->resetPos(Point2DT(currpt, t));

      // end the saccade if our time is up:
      if (elapsedT >= duration)
        {
          LINFO("=== %s: end of saccade at (%d, %d), t=%.2fms ===",
                eyemodel?"Eye ":"Head", currpt.i, currpt.j, t * 1000.0);
          endSaccade();
        }
    }
  else
    // let the frictioncontroller evolve:
    fsc->evolve(t);
}

// ######################################################################
void ModelEyeHeadController::computeWhenResetPos()
{
  // get our reset target:
  WTAwinner win = getPreviousPercept(0);
  Point2DT newloc(win.p, win.t);

  // feed it to fsc:
  fsc->resetPos(newloc);

  pix2deg(win.p, currX, currY);
  // abort any saccade we may be executing:
  if (isInSaccade()) endSaccade();
}

// ######################################################################
Point2D ModelEyeHeadController::computeWhenNewDecision(const double t)
{
  // since our realistic simulation directly drives the
  // FrictionController, it is always up to date, so return its
  // current position, or our current if fsc has nothing new for us:
  Point2D p = fsc->getDecision(t);
  if (p.i != -1) return p; else return getPreviousDecision(0).p;
}




// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
