/*!@file SaccadeControllers.H Derived classes for saccade generation */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaccadeControllers.H 4211 2005-02-17 01:44:48Z itti $
//

#ifndef SACCADECONTROLLERS_H_DEFINED
#define SACCADECONTROLLERS_H_DEFINED

#include "SaccadeController.H"
#include "ArrayData.H" // for class Dims
#include <fstream>
#include <string>
#include <deque>

// ######################################################################
//! SaccadeController configurator
/*! This will export the --sc-type=XX command-line option and will
  instantiate a SaccadeController of the desired type as the option
  gets assigned a value. As this happens, new options may become
  available in the command-line. To see them, use --help AFTER you
  have chosen the name of the controller to use. The current
  SaccadeController may be retrieved using getSC(). */
class SaccadeControllerConfigurator : public ModelComponent
{
public:
  //! Constructor
  SaccadeControllerConfigurator(ModelManager& mgr,
                                const std::string& descrName =
                                "Saccade Controller Configurator",
                                const std::string& tagName =
                                "SaccadeControllerConfigurator");

  //! destructor
  virtual ~SaccadeControllerConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SaccadeController once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen SaccadeController
  /*! You should call this during start() of the ModelComponent that
    needs the SaccadeController. CAUTION: this may be a
    SharedPtr(NULL) if no controller has been chosen. */
  SharedPtr<SaccadeController> getSC() const;

  //! Intercept people changing our itsSacCtrlName ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

protected:
  ModelParam<std::string> itsSacCtrlType; //!< Type of the controller

private:
  SharedPtr<SaccadeController> itsSC; // the controller
  int itsWhatToExport;
};

// ######################################################################
//! Trivial saccade controller
/*! TrivialSaccadeController is essentially a no-op (direct mapping
from percept to decision). */

class TrivialSaccadeController : public SaccadeController
{
public:
  //! Constructor
  TrivialSaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~TrivialSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);
};

// ######################################################################
//! Fixed saccade controller
/*! FixedSaccadeController holds fixation at a fixed point, ignoring
its percepts. Useful to render the difference between covert and overt
attention evident in the model. Also useful for creating foveated
displays with fixed viewpoint. Like with all saccade controllers, the
fixation point may externally be changed using resetPos(). Also here,
the controller may be externally set as in saccade or blinking using
saccade() and blink(). */

class FixedSaccadeController : public SaccadeController
{
public:
  //! Constructor
  FixedSaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~FixedSaccadeController();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

  //! externally decide whether we are in a saccade
  void saccade(const bool insac);

  //! externally decide whether we are in blink
  void blink(const bool inbli);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  ModelParam<Point2D> itsIEP; //!< Initial (and permanent) eye position
};

// ######################################################################
//! Friction-damped mass-spring saccade controller
/*! FrictionSaccadeController acts like a mass (at saccade center)
connected to the covert focus of attention through a spring of zero
rest length, and sliding on a plane in the presence of friction. */

class FrictionSaccadeController : public SaccadeController
{
public:
  //! Constructor
  /*! @param paramprefix As this controller may be used either for the
    eye or for the head (as part of bigger controllers that include
    both eye and head), this should be either "SCeye" or "SChead" and
    will determine which model options we use for our internal
    parameters (eye or head) */
  FrictionSaccadeController(ModelManager& mgr,
                            const std::string paramprefix = "SCeye");

  //! Destructor
  virtual ~FrictionSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

protected:
  ModelParam<double> itsSpringK;    //!< spring stiffness
  ModelParam<double> itsFrictionMu; //!< friction coeff
  ModelParam<bool> itsStartAtIEP;   //!< start at initial eye position?
  ModelParam<Point2D> itsIEP;       //!< Initial (and permanent) eye position
  ModelParam<Dims> itsDims;         //!< Image size so that we stay inside it
  ModelParam<double> itsTimeStep;   //!< Simulation time step

  //! get us started:
  virtual void start1();

private:
  double x, y, oldx, oldy, oldt;
};

// ######################################################################
//! Threshold-limited controller that eliminates small or unreliable shifts
/*! ThresholdSaccadeController will issue an overt control whenever
  the received covert shifts are large and go to a same target for
  several successive covert shifts. It is hence to be used with IOR
  turned off, so that the covert system will keep attending to a same
  location for a while. Once a minimum number of covert shifts have
  been made that are sufficiently far from the current overt location
  and that are sufficiently close to each other, an overt shift will
  be issued, to the average location of the accumulated covert
  locations. */

class ThresholdSaccadeController : public SaccadeController
{
public:
  //! Constructor
  /*! @param minovertdist minimum distance required to warrant an overt
  shift (in pixels at the original image's scale).
  @param maxcovertdist maximum distance to consider that two covert
  shifts point to the "same" location.
  @param minfix minimum number of covert shifts within maxcovertdist
  of each other and farther than minovertdist from the curent overt
  fixation, such as to warrant an overt shift. 
  @param sweigh if true, average covert location is a weighted average
  with saliency values at the covert locations used as weights. Otherwise,
  the weight of each covert location is unity. */

  /*! @param paramprefix As this controller may be used either for the
    eye or for the head (as part of bigger controllers that include
    both eye and head), this should be either "SCeye" or "SChead" and
    will determine which model options we use for our internal
    parameters (eye or head) */
  ThresholdSaccadeController(ModelManager& mgr,
                            const std::string paramprefix = "SCeye");

  //! Destructor
  virtual ~ThresholdSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset ThresholdSaccadeController
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

  //! Check queued-up percepts and return average
  /*! check whether queued-up percepts are close to each other, and
    also return average position of percept queue */
  virtual void checkPercepts(bool& areclose, Point2D& avgp);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  //! Get us started
  virtual void start1();

  ModelParam<float> itsOdist;    //!< min overt distance
  ModelParam<float> itsCdist;    //!< max covert distance
  ModelParam<int> itsMinNum;     //!< min number of close covert shifts
  ModelParam<bool> itsSalWeight; //!< weight average covert pos by saliency
  ModelParam<int> itsFoveaRadius;//!< radius of the fovea

  float odistsq;  // square of min overt distance
  float cdistsq;  // square of max covert distance
  bool sweight;   // use saliency as weight for average covert location
  bool didresetpos;  // just did a resetPos
  Point2D resetloc; // location of reset
};

// ######################################################################
//! Threshold-limited friction-damped controller
/*! ThresholdFrictionSaccadeController combines the features of
  ThresholdSaccadeController and FrictionSaccadeController.

  The ThresholdSaccadeController determines when a new saccade is
  warranted. The FrictionSaccadeController is fed with the average
  covert location over the duration of the percept queue of the
  Threshold controller. It then is responsible for inducing slow
  drifts of overt attention towards that average, in between two
  saccades. In addition, this controller will return to the center of
  the image if it does not move in some time. */
class ThresholdFrictionSaccadeController : public SaccadeController
{
public:
  //! Constructor
  /*! See the two base classes for parameters.
   @param maxidle max time (in seconds) before we naturally return to the
   center of the image. This will happen if the covert shifts are so
   chaotic that we never decide to follow them overtly. */
  /*! @param paramprefix As this controller may be used either for the
    eye or for the head (as part of bigger controllers that include
    both eye and head), this should be either "SCeye" or "SChead" and
    will determine which model options we use for our internal
    parameters (eye or head) */
  ThresholdFrictionSaccadeController(ModelManager& mgr,
                                     const std::string paramprefix = "SCeye");

  //! Destructor
  virtual ~ThresholdFrictionSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  ModelParam<double> itsMaxIdle; //!< max time (in s) before we return to IEP
  ModelParam<Point2D> itsIEP;  //!< homing position if idle

private:
  SharedPtr<ThresholdSaccadeController> tsc;
  SharedPtr<FrictionSaccadeController> fsc;
};


// ######################################################################
//! Simple eye/head controller
/*! This fairly trivial controller uses a
  ThresholdFrictionSaccadeController to move the eyes, plus a slow
  FrictionSaccadeController to move the head.  This is a fairly bogus
  implementation just used here for testing purposes. The eyes
  basically do all the work, and the head part here just slowly
  follows the eyes around. The fixations returned are the same as for
  the ThresholdFrictionSaccadeController. The getHead() function
  returns coordinates of where the head is pointed to. */
class SimpleEyeHeadSaccadeController : public SaccadeController
{
public:
  //! Constructor. See the base classes for parameters.
  SimpleEyeHeadSaccadeController(ModelManager& mgr);
  
  //! Destructor
  virtual ~SimpleEyeHeadSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

  //! Get the current direction of the head
  Point2DT getHead() const;

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  //! This method is called each time a new head decision is requested
  virtual Point2D computeWhenNewDecisionHead(const double t);

private:
  SharedPtr<ThresholdFrictionSaccadeController> eyesc;  // for the eyes
  SharedPtr<FrictionSaccadeController> headsc;          // for the head
};

// ######################################################################
//! Realistic Monkey Eye/Head controller
/*! This is part of an attempt at a simplified approach to the
  MonkeySaccadeCopntroller, which turned out quite complex because of
  the interdependency between head and eye movements. Here, we fake it
  by having independent controllers for the eyes and the head, such
  that we don't need to even deal with gaze stabilization during head
  movements and VOR. This controller will do either the head or the
  eye. For a controller that does both, see
  Monkey2SaccadeController. */
class MonkeyEyeHeadController : public SaccadeController
{
public:
  //! Constructor
  /*! Parameters are similar to the FrictionSaccadeController, plus
    @param iseye use eye model if true, otherwise use head model.
    @param pixperdeg used to calibrate the size (in degrees of
    visual angle) of the incoming images */
  /*! @param paramprefix As this controller may be used either for the
    eye or for the head (as part of bigger controllers that include
    both eye and head), this should be either "SCeye" or "SChead" and
    will determine which model options we use for our internal
    parameters (eye or head) */
  MonkeyEyeHeadController(ModelManager& mgr,
                          const std::string paramprefix = "SCeye");

  //! Destructor
  MonkeyEyeHeadController::~MonkeyEyeHeadController();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Initiate a realistic saccade
  void saccade(const Point2DT& target);

  //! Helper function converting from pixels to degrees
  void pix2deg(const Point2D& pixloc, double& xdeg, double& ydeg) const;

  //! Helper function converting from degrees to pixels
  void deg2pix(const double xdeg, const double ydeg, Point2D& pixloc) const;

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  ModelParam<double> itsPPD;  //!< pixels per degree
  ModelParam<Dims> itsDims;   //!< Image size

private:
  SharedPtr<FrictionSaccadeController> fsc;
  bool eyemodel;   // use eye formulas if true, otherwise use head
  double duration; // duration of realistic saccade
  double oldt;    // time of last call to evolve()
  double startt;  // time of start of realistic saccade
  double currx, curry; // current eye position
  double vmax;   // max velocity
  double accel;  // acceleration
  double theta; // angle of saccade
};

// ######################################################################
//! Realistic rhesus monkey saccade controller
/*! This is a hack based on Nitin's MonkeySaccadeController but
  decoupling eye and head (so that the head minds its own business
  without influencing the eye). Not as realistic in the details, but
  the end result is the same */
class Monkey2SaccadeController : public SaccadeController
{
public:
  //! Constructor
  Monkey2SaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~Monkey2SaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  //! This method is called each time a new head decision is requested
  virtual Point2D computeWhenNewDecisionHead(const double t);

  ModelParam<double> itsBlinkWait; //!< wait time before a blink
  ModelParam<double> itsBlinkDur;   //!< blink duration
  ModelParam<float> itsOdist;
  ModelParam<int> itsFoveaRadius;

private:
  SharedPtr<ThresholdSaccadeController> tsc;
  SharedPtr<MonkeyEyeHeadController> eyesc;
  SharedPtr<MonkeyEyeHeadController> headsc;
  double lastsbt; // time of end of last saccade/blink
  double blinkt;  // time of start of current blink

  // compute amplitude of head displacement. All angles in degrees
  double headAmplitude(const double curreye, const double currhead,
                       const double target);
};

// ######################################################################
//! Realistic rhesus monkey saccade controller
/*! Nitin's masterpiece! */
class MonkeySaccadeController : public SaccadeController
{
public:
  //! Constructor
  MonkeySaccadeController(ModelManager& mgr);
  
  //! Destructor
  virtual ~MonkeySaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

  //! Get the current position of the eyes (same as returned decisions)
  Point2DT getEye() const;

  //! Get the current direction of the head
  Point2DT getHead() const;

  //! Get the current direction of the eye
  Point2DT getJustEye() const;

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  ModelParam<double> itsBlinkWait; //!< wait time before a blink
  ModelParam<double> itsBlinkDur;   //!< blink duration
  ModelParam<float> itsOdist;
  ModelParam<int> itsFoveaRadius;
  ModelParam<double> itsPPD;  //!< pixels per degree
  ModelParam<Dims> itsDims;   //!< Image size

  SharedPtr<ThresholdSaccadeController> headtsc;
  SharedPtr<ThresholdFrictionSaccadeController> tfsc;

private:
  class SomeInfo {
  public:
    double cpX, cpY; // current position
    double tpX, tpY; // target position of the saccade
    double ampX, ampY, amplitude, theta; // amplitude, angle of the saccade
    double ampX2, ampY2;
    double coveredx, coveredy;
    double latency, latency2;// latency of the saccade
    double startt, oldt; // start time
    double duration, timeleft; // duration of the saccade
    double vMax, accel; // maximum x-y velocities of the saccade
    double sgnx, sgny;  // -1 for contralateral, 1 for ipsilateral eye movment
    bool isinsaccade;  // bool flag indicating if we are in a saccade
    int type;
  };
  
  class BlinkInfo {
  public:
    std::deque<Point2DT> history; // history of the previous blinks
    double tlast; // time since last blink
    double tleft; // time left for this blink
    double startt, oldt;
  };

  typedef void (*Formula)(SomeInfo *info);
  void setContrib(SomeInfo *e, SomeInfo *h, double tpX, double tpY);
  void setAlign(SomeInfo *e, SomeInfo *h, double tpX, double tpY);
  void formulaEye(SomeInfo *inf);
  void formulaHead(SomeInfo *inf);
  
  double getVelocity(SomeInfo *info, double t);
  double getVelocityX(SomeInfo *info, const double t);
  double getVelocityY(SomeInfo *info, const double t);
  void setSaccadeTarget(SomeInfo *info, const double t, const int opt);
  
  void eye_evolve(const double t);
  void head_evolve(const double t);
  void blink_evolve(const double t);
  bool inSaccade(SomeInfo *info);
  void writePos();
  void beginVOR(const double t);
  void endVOR(const double t);
  bool inVOR();

  SomeInfo eye, head;
  BlinkInfo blink;
  unsigned int blinkqlen;
  bool invor;
  
  double lastsbt; // time of end of last saccade/blink
  double blinkt;  // time of start of current blink
};

// ######################################################################
//! EyeTracker saccade controller
/*! EyeTrackerSaccadeController is driven by a text file that
  describes human eye position over time, as recorded from an
  eye-tracker apparatus. The text file format should be: fixed
  sampling frequency (default 240Hz), one line per sample:

  ex ey status dx dy sx sy ampl dur

  where (ex, ey) are the coordinates of eye fixation (point of
  regard); status is 0 during fixation/smooth pursuit, 1 during
  saccades, 2 during blinks, 3 during blinks-during-saccades; (dx, dy)
  coordinates of saccade target location (valid only at start of a
  saccade, and the following measures are too); (sx, sy) stdev of
  those target coords over up to 0.25s after end of saccade (might be
  shorted if next saccade starts sooner then 0.25s after current one);
  ampl is the amplitude of the saccade, in pixels; dur is the duration
  of the saccade (in eye-tracker samples). */
class EyeTrackerSaccadeController : public SaccadeController
{
public:
  //! Constructor
  EyeTrackerSaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~EyeTrackerSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

  //! Get all the extra eye movement data
  /*! Get all the extra eye movement data from the data file.
    @param nextSacTarget coordinates of next saccade target. This is
    updated each time a saccade starts; so typically, the caller should
    get this info as soon as the saccade controller has entered a
    saccade.
    @param nextSacAmpl amplitude of the saccade, in pixels.
    @param nextFixDuration duration of the next fixation period, in s. */
  void getExtraEyeData(Point2D& nextSacTarget, double& nextSacAmpl,
                       double& nextFixDuration);

  //! get the eye movement sampling period, in seconds
  double getEyePeriod() const;

protected:
  ModelParam<Dims> itsDims;         //!< Image size so that we stay inside it

  //! Name of a calibrated eye movement file
  ModelParam<std::string> itsEyeFname;

  //! Period (in s) at which eye movement samples should be read in
  ModelParam<float> itsEyePeriod;

  //! Number of initial eye movement samples to trash
  ModelParam<int> itsEyeTrash;

  //! Get started
  virtual void start1();

  //! get stopped
  virtual void stop1();

  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

private:
  int itsEyeSample; // current eye sample number
  FILE *itsEyeFile;  // input eye data file handle

  Point2D itsCurrEye;
  Point2D itsNextSaccadeTarget;
  double itsNextSaccadeAmpl;
  double itsNextFixationDuration;
};

// ######################################################################
//! VisBuf saccade controller
/*! VisBufSaccadeController uses a world-centered VisualBuffer to
  accumulate interesting locations (each time a covert attention shift
  occurs and a percept is given to the controller), with internal
  competitive dynamics within the buffer. */
class VisBufSaccadeController : public SaccadeController
{
public:
  //! Constructor
  VisBufSaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~VisBufSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  ModelParam<Dims> itsDims;    //!< dimensions of the buffer

  //! Get started
  virtual void start1();

  //! get stopped
  virtual void stop1();

  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

private:

};

// ######################################################################
//! Realistic Eye/Head controller
/*! This is part of an attempt at a simplified approach to the
  MonkeySaccadeCopntroller, which turned out quite complex because of
  the interdependency between head and eye movements. Here, we fake it
  by having independent controllers for the eyes and the head, such
  that we don't need to even deal with gaze stabilization during head
  movements and VOR. This controller will do either the head or the
  eye. For a controller that does both, see
  Monkey2SaccadeController. */
class ModelEyeHeadController : public SaccadeController
{
public:
  //! Constructor
  /*! Parameters are similar to the FrictionSaccadeController, plus
    @param iseye use eye model if true, otherwise use head model.
    @param pixperdeg used to calibrate the size (in degrees of
    visual angle) of the incoming images */
  /*! @param paramprefix As this controller may be used either for the
    eye or for the head (as part of bigger controllers that include
    both eye and head), this should be either "SCeye" or "SChead" and
    will determine which model options we use for our internal
    parameters (eye or head) */
  ModelEyeHeadController(ModelManager& mgr,
                          const std::string paramprefix = "SCeye");

  //! Destructor
  ModelEyeHeadController::~ModelEyeHeadController();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Initiate a realistic saccade
  void saccade(const Point2DT& target);

  //! Helper function converting from pixels to degrees
  void pix2deg(const Point2D& pixloc, double& xdeg, double& ydeg) const;

  //! Helper function converting from degrees to pixels
  void deg2pix(const double xdeg, const double ydeg, Point2D& pixloc) const;

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  ModelParam<double> itsPPD;  //!< pixels per degree
  ModelParam<Dims> itsDims;   //!< Image size

private:
  SharedPtr<FrictionSaccadeController> fsc;
  bool eyemodel;   // use eye formulas if true, otherwise use head
  double duration; // duration of realistic saccade
  double ampX, ampY; // amplitude (in deg) along the X and Y axes
  double oldt;    // time of last call to evolve()
  double startt;  // time of start of realistic saccade
  double currX, currY; // current position
  double prevX, prevY; // position at start of previous saccade/smooth pursuit
};

// ######################################################################
//! Realistic saccade controller based on human eye/head recordings
/*! This is a hack based on Nitin's MonkeySaccadeController but
  decoupling eye and head (so that the head minds its own business
  without influencing the eye). Not as realistic in the details, but
  the end result is the same */
class ModelWrapperSaccadeController: public SaccadeController
{
public:
  //! Constructor
  ModelWrapperSaccadeController(ModelManager& mgr);

  //! Destructor
  virtual ~ModelWrapperSaccadeController();
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! evolve one time step (called, e.g., by Brain::evolve())
  virtual void evolve(const double t);

protected:
  //! This method is called each time a new percept has arrived
  virtual void computeWhenNewPercept();

  //! This method is called each time a reset is made
  virtual void computeWhenResetPos();

  //! This method is called each time a new decision is requested
  virtual Point2D computeWhenNewDecision(const double t);

  //! This method is called each time a new head decision is requested
  virtual Point2D computeWhenNewDecisionHead(const double t);

  //! get started
  virtual void start1();

  ModelParam<float> itsOdist;
  ModelParam<int> itsFoveaRadius;

private:
  SharedPtr<ThresholdSaccadeController> tsc;
  SharedPtr<ModelEyeHeadController> eyesc;
  SharedPtr<ModelEyeHeadController> headsc;
  double lastsbt; // time of end of last saccade/blink
  double itsBlinkTime;     // time (in s) of start of current blink
  double itsBlinkDuration; // duration of current blink (s)
  int idum;  // for random number generation

  // compute amplitude of head displacement. All angles in degrees
  double headAmplitude(const double curreye, const double currhead,
                       const double target);

  // decide on the next blink
  void pickNextBlink(const double t);
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
