/*!@file SaliencyMap.C Implementation for saliency map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaliencyMap.C 4291 2005-03-13 00:24:48Z itti $
//

// MA
#include <iostream>
using namespace std;

#include "saliency.H"
#include "MathFunctions.H"
#include "Image.H"
#include "SaliencyMap.H"
#include "log.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "FrameSeries.H"
#include <cmath>

using namespace iNVT;

// ######################################################################
// ######################################################################
// ########## SaliencyMap implementation
// ######################################################################
// ######################################################################

SaliencyMap::SaliencyMap(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSaveResults(this, "SMsaveResults", false), // see ModelOptionDefs.C
  itsSaveCumResults(this, "SMsaveCumResults", false), // see ModelOptionDefs.C
  itsSalmapFactor(this, "SMfactor", 50000.0),
  itsUseSaccadicSuppression(this, "SMuseSacSuppress", true),
  itsUseBlinkSuppression(this, "SMuseBlinkSuppress", true)
{ }

// ######################################################################
SaliencyMap::~SaliencyMap()
{ }

// ######################################################################
void SaliencyMap::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsSalmapFactor);
    manager->requestOption(this, itsUseSaccadicSuppression);
    manager->requestOption(this, itsUseBlinkSuppression);
  }

  if (whatToExport & OPTEXP_SAVE) {
    manager->requestOption(this, itsSaveResults);
    manager->requestOption(this, itsSaveCumResults);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
double SaliencyMap::w(const double& d, const double& ampl, const double& sdev)
{
// MA
if (sdev == 0)
{
  LDEBUG("!!! ERROR !!!");
}

  double d_dev = d / sdev; d_dev = -0.5 * d_dev * d_dev; 
  if (d_dev < -20.0) return 0.0;
  return ampl * exp(d_dev);
}

// ######################################################################
double SaliencyMap::w(const Point2D& p1, const Point2D& p2,
                      const double& pampl, const double& mampl,
                      const double& psdev, const double& msdev)
{
  double dx = double(p1.i - p2.i);
  double dy = double(p1.j - p2.j);
  double d = sqrt(SQR(dx) + SQR(dy));
  return w(d, pampl, psdev) - w(d, mampl, msdev);
}

// ######################################################################
void SaliencyMap::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  if (itsSaveResults.getVal())
    ofs->writeFloat(getV(true), PNM, -1, true, "SM");

  if (itsSaveCumResults.getVal())
    {
      Image<float> current = getV(false); // do not normalize
      if (itsCumMap.initialized()) itsCumMap += current;
      else itsCumMap = current;
      ofs->writeFloat(itsCumMap, PFM, -1, true, "CUMSM");
    }
}


// ######################################################################
// ######################################################################
// ########## SaliencyMapConfigurator implementation
// ######################################################################
// ######################################################################
SaliencyMapConfigurator::SaliencyMapConfigurator(ModelManager& mgr,
                                                 const std::string& descrName,
                                                 const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSMtype(this, "SaliencyMapType", "None"),
  itsSM(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
SaliencyMapConfigurator::~SaliencyMapConfigurator()
{  }

// ######################################################################
void SaliencyMapConfigurator::exportOptions(const int whatToExport,
                                            const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsSMtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<SaliencyMap> SaliencyMapConfigurator::getSM() const
{ return itsSM; }

// ######################################################################
bool SaliencyMapConfigurator::setModelParamString(const std::string& name,
                                                  const std::string& value,
                                                  const bool recurse,
                                                  const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsSMtype.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsSM.get()) { removeSubComponent(itsSM); itsSM.reset(NULL); }

    // instantiate a SM of the appropriate type:
    if (value.compare("None") == 0)          // no SM
      itsSM.reset(NULL);
    else if (value.compare("Std") == 0)      // standard
      itsSM.reset(new SaliencyMapStd(*manager));
    else if (value.compare("Trivial") == 0)     // trivial
      itsSM.reset(new SaliencyMapTrivial(*manager));
    else if (value.compare("Fast") == 0)     // fast
      itsSM.reset(new SaliencyMapFast(*manager));
    else
      LFATAL("Unknown SM type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsSM.get()) {
      addSubComponent(itsSM);

      // tell the controller to export its options:
      itsSM->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected SM of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ########## SaliencyMapStd implementation
// ######################################################################
// ######################################################################

// ######################################################################
SaliencyMapStd::SaliencyMapStd(ModelManager& mgr, const std::string& descrName,
                               const std::string& tagName) :
  SaliencyMap(mgr, descrName, tagName), Image<LeakyIntegrator>(),
  itsGinhDecay(this, "SMginhDecay", 0.9999F)
{  }

// ######################################################################
void SaliencyMapStd::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsGinhDecay);
  }

  // the base class will propagate this call to our subcomponents:
  SaliencyMap::exportOptions(whatToExport, recurse);
}

// ######################################################################
SaliencyMapStd::~SaliencyMapStd()
{ }

// ######################################################################
void SaliencyMapStd::input(const Image<float>& current)
{
  if (initialized() == false) {
    resize(current.getDims(), true);
    Image<LeakyIntegrator>::iterator nptr = beginw(), stop = endw();
    float decay = itsGinhDecay.getVal();
    while(nptr != stop) (nptr++)->setGinhDecay(decay);
  }

  Image<LeakyIntegrator>::iterator nptr = beginw();
  Image<float>::const_iterator cptr = current.begin();
  Image<float>::const_iterator stop = current.end();

  while (cptr != stop) (nptr++)->input(*cptr++);
}

// ######################################################################
Image<float> SaliencyMapStd::getV(const bool normalize) const
{
  Image<float> ret(getDims(), NO_INIT);
  Image<float>::iterator dest = ret.beginw();
  Image<LeakyIntegrator>::const_iterator src = begin();
  Image<LeakyIntegrator>::const_iterator stop = end();

  while(src != stop) *dest++ = (src++)->getV();

  if (normalize)
    {
      if (itsSalmapFactor.getVal() > 0.0F) ret *= itsSalmapFactor.getVal();
      else ret.normalize(0.0F, 255.0F);
    }

  return ret;
}

// ######################################################################
void SaliencyMapStd::depress(const Point2D& winner,
                             const double& pampl, const double& mampl,
                             const double& psdev, const double& msdev)
{
// MA
LDEBUG("psdev = %f, msdev = %f", (float)psdev, (float)msdev);

  float g; Point2D p; int ww = getWidth(), hh = getHeight();
  Image<LeakyIntegrator>::iterator src = beginw();

  for (p.j = 0; p.j < hh; p.j ++)
    for (p.i = 0; p.i < ww; p.i ++)
      {
        g = float(w(winner, p, pampl, mampl, psdev, msdev));
        if (p.i == winner.i && p.j == winner.j)
          LDEBUG("Peak IOR conductance: %fmS", g * 1000.0F);

        // open inhibitory conductance:
        (src++)->addGinh(g);
      }
}

// ######################################################################
void SaliencyMapStd::depress(const Point2D& winner, const double& ampl,
                             const Image<byte>& objectMask)
{
  Point2D p; int ww = getWidth(), hh = getHeight();
  Image<LeakyIntegrator>::iterator src = beginw();

  for (p.j = 0; p.j < hh; p.j ++)
    for (p.i = 0; p.i < ww; p.i ++)
      {
        float g = 0.1F / 255.F * ampl * float(objectMask.getVal(p));
        if (p.i == winner.i && p.j == winner.j)
          LDEBUG("Peak IOR conductance: %fmS", g * 1000.0F);
        
        // open inhibitory conductance:
        
        (src++)->addGinh(g);
      }
}

// ######################################################################
void SaliencyMapStd::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for saccadic suppression
  // of the Saliency Map, as there is evidence that some intra-saccadic visual
  // input may still influence eye movements.
  float inG;
  if (on) inG = 1.0e-6F;  // on inhibitory conductance, in Siemens
  else inG = 0.0F;        // turning off

  // NOTE: this will erase any previous conductance pattern that may
  // have been present in the neurons, e.g., due to recent IORs. So it
  // is assumed here that IOR and saccadic suppression share the same
  // inhibitory mechanism.
  Image<LeakyIntegrator>::iterator src = beginw();
  Image<LeakyIntegrator>::iterator stop = endw();

  while(src != stop) (src++)->setGinh(inG);
  
  LINFO("------- SaliencyMapStd saccadic suppression %s -------",
        on ? "on":"off");
}

// ######################################################################
void SaliencyMapStd::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for blink suppression of
  // the Saliency Map, as the visual input should also be turned off
  // in Brain
  float inG;
  if (on) inG = 1.0e-6F;  // on inhibitory conductance, in Siemens
  else inG = 0.0F;        // turning off

  // NOTE: this will erase any previous conductance pattern that may
  // have been present in the neurons, e.g., due to recent IORs. So it
  // is assumed here that IOR and saccadic suppression share the same
  // inhibitory mechanism.
  Image<LeakyIntegrator>::iterator src = beginw();
  Image<LeakyIntegrator>::iterator stop = endw();

  while(src != stop) (src++)->setGinh(inG);
  
  LINFO("------- SaliencyMapStd blink suppression %s -------",
        on ? "on":"off");
}

// ######################################################################
void SaliencyMapStd::evolve(const double t)
{
  ASSERT(initialized());

  Image<LeakyIntegrator>::iterator itr = beginw();
  Image<LeakyIntegrator>::iterator stop = endw();

  while (itr != stop) (itr++)->integrate(t);
}

// ######################################################################
float SaliencyMapStd::getV(const Point2D& p) const
{ return getVal(p).getV(); }

// ######################################################################
void SaliencyMapStd::reset()
{ freeMem(); }

// ######################################################################
int SaliencyMapStd::getWidth() const
{ return Image<LeakyIntegrator>::getWidth(); }

// ######################################################################
int SaliencyMapStd::getHeight() const
{ return Image<LeakyIntegrator>::getHeight(); }


// ######################################################################
// ########## SaliencyMapTrivial implementation
// ######################################################################
// ######################################################################

// ######################################################################
SaliencyMapTrivial::SaliencyMapTrivial(ModelManager& mgr,
                                       const std::string& descrName,
                                       const std::string& tagName) :
  SaliencyMap(mgr, descrName, tagName), Image<float>()
{  }

// ######################################################################
SaliencyMapTrivial::~SaliencyMapTrivial()
{ }

// ######################################################################
void SaliencyMapTrivial::input(const Image<float>& current)
{
  *(static_cast< Image<float> *>(this)) = current;
}

// ######################################################################
Image<float> SaliencyMapTrivial::getV(const bool normalize) const
{
  if (normalize == false) return *this;

  Image<float> ret(*this);
  if (itsSalmapFactor.getVal() > 0.0F) ret *= itsSalmapFactor.getVal();
  else ret.normalize(0.0F, 255.0F);
  return ret;
}

// ######################################################################
void SaliencyMapTrivial::depress(const Point2D& winner,
                                 const double& pampl, const double& mampl,
                                 const double& psdev, const double& msdev)
{
  float g; Point2D p; int ww = getWidth(), hh = getHeight();
  Image<float>::iterator src = beginw();

  for (p.j = 0; p.j < hh; p.j ++)
    for (p.i = 0; p.i < ww; p.i ++)
      {
        g = float(w(winner, p, pampl, mampl, psdev, msdev));
        if (p.i == winner.i && p.j == winner.j)
          LDEBUG("Peak IOR conductance: %fmS", g * 1000.0F);

        // open inhibitory conductance:
        *src -= g; if (*src < 0.0F) *src = 0.0F;
        ++ src;
      }
}

// ######################################################################
void SaliencyMapTrivial::depress(const Point2D& winner, const double& ampl,
                                 const Image<byte>& objectMask)
{
  Image<float> inhib = objectMask * (ampl / 255.0F);
  *this -= inhib;  // subtract inhibition for current saliency values
  rectify();       // remove possible negative values
}

// ######################################################################
void SaliencyMapTrivial::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for saccadic suppression
  // of the Saliency Map, as there is evidence that some intra-saccadic visual
  // input may still influence eye movements.
  if (on) *this -= 1.0e-6F;
  LINFO("------- SaliencyMapTrivial saccadic suppression %s -------",
        on ? "on":"off");
}

// ######################################################################
void SaliencyMapTrivial::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for blink suppression of
  // the Saliency Map, as the visual input should also be turned off
  // in Brain
  if (on) *this -= 1.0e-6F;
  LINFO("------- SaliencyMapTrivial blink suppression %s -------",
        on ? "on":"off");

}

// ######################################################################
void SaliencyMapTrivial::evolve(const double t)
{ }

// ######################################################################
float SaliencyMapTrivial::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
void SaliencyMapTrivial::reset()
{ freeMem(); }

// ######################################################################
int SaliencyMapTrivial::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int SaliencyMapTrivial::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
// ######################################################################
// ########## SaliencyMapFast implementation
// ######################################################################
// ######################################################################

// ######################################################################
SaliencyMapFast::SaliencyMapFast(ModelManager& mgr,
                                 const std::string& descrName,
                                 const std::string& tagName) :
  SaliencyMap(mgr, descrName, tagName), Image<float>(),
  itsTimeStep(this, "SimulationTimeStep", 0.0001), // see ModelOptionDefs.C
  itsInputCoeff(this, "SMfastInputCoeff", 0.1), // ModelOptionDefs.C
  itsItoVcoeff(this, "SMfastItoVcoeff", 5000000.0f), // idem
  itsInputCopy(), itsT(0.0)
{  }

// ######################################################################
void SaliencyMapFast::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsTimeStep);
    manager->requestOption(this, itsInputCoeff);
    manager->requestOption(this, itsItoVcoeff);
  }

  // the base class will propagate this call to our subcomponents:
  SaliencyMap::exportOptions(whatToExport, recurse);
}

// ######################################################################
SaliencyMapFast::~SaliencyMapFast()
{ }

// ######################################################################
void SaliencyMapFast::input(const Image<float>& current)
{
  if (initialized() == false) resize(current.getDims(), true);
  itsInputCopy = current * itsItoVcoeff.getVal();
}

// ######################################################################
Image<float> SaliencyMapFast::getV(const bool normalize) const
{
  if (normalize == false) return *this;

  Image<float> ret(*this);
  if (itsSalmapFactor.getVal() > 0.0F) ret *= itsSalmapFactor.getVal();
  else ret.normalize(0.0F, 255.0F);
  return ret;
}

// ######################################################################
void SaliencyMapFast::depress(const Point2D& winner,
                              const double& pampl, const double& mampl,
                              const double& psdev, const double& msdev)
{
cout << "SaliencyMapFast::depress" << endl;
  float g; Point2D p; int ww = getWidth(), hh = getHeight();
  Image<float>::iterator src = beginw();

  for (p.j = 0; p.j < hh; p.j ++)
    for (p.i = 0; p.i < ww; p.i ++)
      {
        g = float(w(winner, p, pampl, mampl, psdev, msdev));
        if (p.i == winner.i && p.j == winner.j)
          LDEBUG("Peak IOR conductance: %fmS", g * 1000.0F);

        // open inhibitory conductance:
        *src -= g; if (*src < 0.0F) *src = 0.0F;
        ++ src;
      }
}

// ######################################################################
void SaliencyMapFast::depress(const Point2D& winner, const double& ampl,
                 const Image<byte>& objectMask)
{
  Image<float> inhib = objectMask * (ampl / 255.0F);
  *this -= inhib;  // subtract inhibition for current saliency values
  rectify();       // remove possible negative values
}

// ######################################################################
void SaliencyMapFast::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for saccadic suppression
  // of the Saliency Map, as there is evidence that some intra-saccadic visual
  // input may still influence eye movements.
  if (on) *this -= 1.0e-6F;
  LINFO("------- SaliencyMapFast saccadic suppression %s -------",
        on ? "on":"off");
}

// ######################################################################
void SaliencyMapFast::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;

  // we setup a weak inhibitory conductance for blink suppression of
  // the Saliency Map, as the visual input should also be turned off
  // in Brain
  if (on) *this -= 1.0e-6F;
  LINFO("------- SaliencyMapFast blink suppression %s -------",
        on ? "on":"off");

}

// ######################################################################
void SaliencyMapFast::evolve(const double t)
{
  ASSERT(initialized());
  double nsteps = (t - itsT) / itsTimeStep.getVal();
  float c1 = pow(itsInputCoeff.getVal(), nsteps), c2 = 1.0F - c1;
  *this *= c1; addWeighted(itsInputCopy, c2);
  itsT = t;
}

// ######################################################################
float SaliencyMapFast::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
void SaliencyMapFast::reset()
{ freeMem(); itsT = 0.0; }

// ######################################################################
int SaliencyMapFast::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int SaliencyMapFast::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
