/*!@file SaliencyMap.H Class declarations for saliency map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SaliencyMap.H 4217 2005-02-17 18:13:54Z beobot $
//

#ifndef SALIENCYMAP_H_DEFINED
#define SALIENCYMAP_H_DEFINED

#include "Image.H"
#include "LeakyIntegrator.H"
#include "ModelComponent.H"
class ModelManager;
class OutputFrameSeries;

namespace iNVT
{
// ######################################################################
//! The saliency map base class
// ######################################################################
/*! This is a 2D salience map. It is just a base class with virtual
  function definitions. Various saliency maps are available below,
  which provide implementations for these virtual functions. The
  saliency map receives input through the input() function, and its
  current state (membrane potential of neurons) can be retrieved using
  getV(). Various inplementations will typically derive from class
  Image and inherit all the standard methods of class Image.  The
  user has the responsibility of calling evolve() often to ensure that
  possible difference equations in the map's neurons run well. */
class SaliencyMap : public ModelComponent
{
 public:
  // ######################################################################
  //! @name Constructor, destructor, and reset
  //@{

  //! Ininitialized constructor
  /*! The SM will be resized and initialized the first time input() is
    called */
  SaliencyMap(ModelManager& mgr, const std::string& descrName = "Saliency Map",
              const std::string& tagName = "SaliencyMap");
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. A base
    implementation is provided that will only export the base class
    options (see the ModelParam definitions in the protected section
    below). */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);
  //! Destructor
  virtual ~SaliencyMap();

  //! Reset to initial state just after construction
  virtual void reset() = 0;

  //@}

  // ######################################################################
  //! @name Input, evolve and output
  //@{

  //! Set new input currents for all neurons
  /*! This should initialize and resize the saliency map if the
    saliency map is currently uninitialized (e.g., just after
    construction or reset()). */
  virtual void input(const Image<float>& in) = 0;

  //! Get width of saliency map
  virtual int getWidth() const = 0;

  //! Get height of saliency map
  virtual int getHeight() const = 0;

  //! Return all our membrane potential voltages as an Image<float>
  /*! The curent values are possibly normalized to 0..255. The
    normalization will use itsSalmapFactor if it is non-null */
  virtual Image<float> getV(const bool normalize = false) const = 0;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const = 0;

  //! Integrate inputs until time t (in s) and update membrane potentials
  virtual void evolve(const double t) = 0;

  //@}

  // ######################################################################
  //! @name Inhibition of return
  //@{

  //! difference-of-Gaussians inhibition of return
  /*! The positive Gaussian is defined by its amplitute pampl and
    width psdev; the negative Gaussian by mampl and msdev.  Both
    centered at winner.  */
  virtual void depress(const Point2D& winner,
                       const double& pampl, const double& mampl,
                       const double& psdev, const double& msdev) = 0;

  //! object-based inhibition of return
  /*! The object mask defines where in the image the inhibition will
    be applied. The winner coordinates and ampl are used to calibrate
    the inhibition strength */
  virtual void depress(const Point2D& winner, const double& ampl,
                       const Image<byte>& objectMask) = 0;

  //@}

  // ######################################################################
  //! @name Saccadic and blink suppression
  //@{

  //! Turn saccadic suppression on/off
  virtual void saccadicSuppression(const bool on) = 0;

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on) = 0;

  //@}

  // ######################################################################
  //! @name Results
  //@{

  //! Save our various results
  /*! A default implementation is provided, which just calls
      getV() and saves that map with a file name prefix of "SM".
      @param ofs will figure out file name/format and save our
      results. */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

  // ######################################################################
protected:
  //! Save our internals when saveResults() is called?
  ModelParam<bool> itsSaveResults;

  //! Save our cumulative internals when saveResults() is called?
  ModelParam<bool> itsSaveCumResults;

  //! Multiply SM by a factor rather than normalize when doing getV()?
  ModelParam<float> itsSalmapFactor;

  //! use saccadic suppression?
  ModelParam<bool> itsUseSaccadicSuppression;

  //! use blink suppression?
  ModelParam<bool> itsUseBlinkSuppression;

  //! Helper function to compute inhibition of return
  double w(const double& d, const double& ampl, const double& sdev);

  //! Helper function to compute inhibition of return
  double w(const Point2D& p1, const Point2D& p2, const double& pampl,
           const double& mampl, const double& psdev, const double& msdev);

  //! Cumulative map so far. Cumulates each time saveResults() is called
  /*! This is computed only if itsSaveCumResults() is true */
  Image<float> itsCumMap;

private:
  // forbid assignment and copy-construction:
  SaliencyMap& operator=(const SaliencyMap& sm);
  SaliencyMap(const SaliencyMap& sm);
};
}

// ######################################################################
//! SaliencyMap configurator
// ######################################################################
/*! This will export the --sm-type=XX command-line option and will
  instantiate an SM of the desired type as the option gets assigned a
  value. As this happens, new options may become available in the
  command-line. To see them, use --help AFTER you have chosen the type
  to use. The current SM may be retrieved using getSM(). */
class SaliencyMapConfigurator : public ModelComponent
{
public:
  //! Constructor
  SaliencyMapConfigurator(ModelManager& mgr,
                          const std::string& descrName =
                          "Saliency Map Configurator",
                          const std::string& tagName =
                          "SaliencyMapConfigurator");

  //! destructor
  virtual ~SaliencyMapConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SM once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen SM
  /*! You should call this during start() of the ModelComponent that
    needs the SM. CAUTION: this may be a SharedPtr(NULL) if no
    SM type has been chosen. */
  SharedPtr<iNVT::SaliencyMap> getSM() const;

protected:
  ModelParam<std::string> itsSMtype; //!< type of saliency map

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<iNVT::SaliencyMap> itsSM; // the saliency map
  int itsWhatToExport;
};


// ######################################################################
//! The standard saliency map: a 2D array of LeakyIntegrator neurons
// ######################################################################
/*! This is a 2D salience map. It is an Image<LeakyIntegrator> and
  inherits all the standard methods of class Image. To avoid
  confusion, we here add explicit input() and getV() methods (rather
  than providing conversion functions between LeakyIntegrator and
  float, which could make the process more transparent but also
  possibly more confusing). The user has the responsibility of calling
  evolve() often to ensure that the difference equations in the
  LeakyIntegrator neurons run well. */
class SaliencyMapStd : public iNVT::SaliencyMap, public Image<LeakyIntegrator>
{
 public:
  //! Ininitialized constructor
  /*! The SM will be resized and initialized the first time input() is
    called */
  SaliencyMapStd(ModelManager& mgr, const std::string& descrName =
                 "Saliency Map Std",
                 const std::string& tagName =
                 "SaliencyMapStd");
  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~SaliencyMapStd();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  virtual void input(const Image<float>& in);

  //! Get width of saliency map
  virtual int getWidth() const;

  //! Get height of saliency map
  virtual int getHeight() const;

  //! Return all our membrane potential voltages as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const; // get output at a given location

  //! Integrate inputs until time t (in s) and update membrane potential
  virtual void evolve(const double t);

  //! difference-of-Gaussians inhibition of return
  /*! The positive Gaussian is defined by its amplitute pampl and
    width psdev; the negative Gaussian by mampl and msdev.  Both
    centered at winner.  */
  virtual void depress(const Point2D& winner,
                       const double& pampl, const double& mampl,
                       const double& psdev, const double& msdev);

  //! object-based inhibition of return
  /*! The object mask defines where in the image the inhibition will
    be applied. The winner coordinates and ampl are used to calibrate
    the inhibition strength */
  virtual void depress(const Point2D& winner, const double& ampl,
                       const Image<byte>& objectMask);

  //! Turn saccadic suppression on/off
  /*! The behavior here is that turning saccadic suppression on will
    open the inhibitory conductances of the SM neurons just like
    inhibition of return does. This function should be called
    repeatedly during suppression as, otherwise, the inhibitory
    conductances have a natural tendency to fade away (as is needed
    for IOR to pregressively fade away). When this function is called
    to turn saccadic suppression off, the inhibitory conductances are
    violently reset to zero.  So, for full saccadic suppression, call
    SM.saccadicSuppression(true) at every time step during the
    suppression time interval, then call SM.saccadicSuppression(false)
    once at the end of the suppression period. Calling
    SM.saccadicSuppression(true) less often (or just once at the start of
    a saccade) will yield milder suppression. */
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);

protected:
  ModelParam<float> itsGinhDecay; //!< decay for IOR inhibition
};

// ######################################################################
//! The trivial saliency map: a 2D array of floats
// ######################################################################
/*! This is a 2D salience map. It is an Image<float> and inherits all
  the standard methods of class Image. To avoid confusion, we here add
  explicit input() and getV() methods. evolve() does nothing in this
  version. */
class SaliencyMapTrivial : public iNVT::SaliencyMap, public Image<float>
{
 public:
  //! Ininitialized constructor
  /*! The SM will be resized and initialized the first time input() is
    called */
  SaliencyMapTrivial(ModelManager& mgr, const std::string& descrName =
                     "Saliency Map Trivial",
                     const std::string& tagName =
                     "SaliencyMapTrivial");

  //! Destructor
  virtual ~SaliencyMapTrivial();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  virtual void input(const Image<float>& in);

  //! Return all our membrane potential voltages as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const; // get output at a given location

  //! Get width of saliency map
  virtual int getWidth() const;

  //! Get height of saliency map
  virtual int getHeight() const;

  //! Integrate inputs until time t (in s) and update membrane potential
  virtual void evolve(const double t);

  //! difference-of-Gaussians inhibition of return
  /*! The positive Gaussian is defined by its amplitute pampl and
    width psdev; the negative Gaussian by mampl and msdev.  Both
    centered at winner.  */
  virtual void depress(const Point2D& winner,
                       const double& pampl, const double& mampl,
                       const double& psdev, const double& msdev);

  //! object-based inhibition of return
  /*! The object mask defines where in the image the inhibition will
    be applied. The winner coordinates and ampl are used to calibrate
    the inhibition strength */
  virtual void depress(const Point2D& winner, const double& ampl,
                       const Image<byte>& objectMask);

  //! Turn saccadic suppression on/off
  /*! The behavior here is that turning saccadic suppression on will
    open the inhibitory conductances of the SM neurons just like
    inhibition of return does. This function should be called
    repeatedly during suppression as, otherwise, the inhibitory
    conductances have a natural tendency to fade away (as is needed
    for IOR to pregressively fade away). When this function is called
    to turn saccadic suppression off, the inhibitory conductances are
    violently reset to zero.  So, for full saccadic suppression, call
    SM.saccadicSuppression(true) at every time step during the
    suppression time interval, then call SM.saccadicSuppression(false)
    once at the end of the suppression period. Calling
    SM.saccadicSuppression(true) less often (or just once at the start of
    a saccade) will yield milder suppression. */
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);
};

// ######################################################################
//! The fast saliency map: a 2D array of floats
// ######################################################################
/*! This is a 2D salience map. It is an Image<float> and inherits all
  the standard methods of class Image. To avoid confusion, we here add
  explicit input() and getV() methods. The user has the responsibility
  of calling evolve() often to ensure that the saliency map evolves
  smoothly over time. */
class SaliencyMapFast : public iNVT::SaliencyMap, public Image<float>
{
 public:
  //! Ininitialized constructor
  /*! The SM will be resized and initialized the first time input() is
    called */
  SaliencyMapFast(ModelManager& mgr, const std::string& descrName =
                  "Saliency Map Fast",
                  const std::string& tagName =
                  "SaliencyMapFast");

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);
  //! Destructor
  virtual ~SaliencyMapFast();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  virtual void input(const Image<float>& in);

  //! Return all our membrane potential voltages as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const; // get output at a given location

  //! Get width of saliency map
  virtual int getWidth() const;

  //! Get height of saliency map
  virtual int getHeight() const;

  //! Integrate inputs until time t (in s) and update membrane potential
  virtual void evolve(const double t);

  //! difference-of-Gaussians inhibition of return
  /*! The positive Gaussian is defined by its amplitute pampl and
    width psdev; the negative Gaussian by mampl and msdev.  Both
    centered at winner.  */
  virtual void depress(const Point2D& winner,
                       const double& pampl, const double& mampl,
                       const double& psdev, const double& msdev);

  //! object-based inhibition of return
  /*! The object mask defines where in the image the inhibition will
    be applied. The winner coordinates and ampl are used to calibrate
    the inhibition strength */
  virtual void depress(const Point2D& winner, const double& ampl,
                       const Image<byte>& objectMask);

  //! Turn saccadic suppression on/off
  /*! The behavior here is that turning saccadic suppression on will
    open the inhibitory conductances of the SM neurons just like
    inhibition of return does. This function should be called
    repeatedly during suppression as, otherwise, the inhibitory
    conductances have a natural tendency to fade away (as is needed
    for IOR to pregressively fade away). When this function is called
    to turn saccadic suppression off, the inhibitory conductances are
    violently reset to zero.  So, for full saccadic suppression, call
    SM.saccadicSuppression(true) at every time step during the
    suppression time interval, then call SM.saccadicSuppression(false)
    once at the end of the suppression period. Calling
    SM.saccadicSuppression(true) less often (or just once at the start of
    a saccade) will yield milder suppression. */
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);

protected:
  //! Simulation time step, in seconds
  ModelParam<double> itsTimeStep;

  //! Coeff by which new inputs replace the current SM at each time step (0..1)
  ModelParam<float> itsInputCoeff;

  //! Coeff to go from current inputs to voltage outputs
  ModelParam<float> itsItoVcoeff;

private:
  Image<float> itsInputCopy;  // copy of inputs
  double itsT;                // keep track of time
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
