/*!@file ScaleSpace.C Keypoint computation for SIFT obj recognition */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>

#include "ScaleSpace.H"
#include "ArrayData.H"
#include "Image_FilterOps.H"
#include "Matrix3.H"
#include <math.h>

// #######################################################################
// ScaleSpace member definitions
// #######################################################################

// 
// Initialize this Scale Space
// Create gaussian blurred and dogs pictures
//
void ScaleSpace::initScaleSpace(float currentscale, const Image<float> & image)
{
    // Reset
    reset();

    // init local variables
    dimX = image.getWidth();
    dimY = image.getHeight();
    octSize = currentscale;
    int s;
    blur = ImageSet<float>(SCALES + 3);
    dogs = ImageSet<float>(SCALES + 2);
    float k = pow( 2.0f, 1.0f / SCALES );
    float sigmaW = SIGMA; 
    blur[0] = image;    

    // compute gaussian blurred maps
    for (s=1; s<SCALES+3; s++)
      {
	// get previous image in pyramid
	blur[s] = Image<float>(blur[s-1]);
    
	// calculate sigma -- is this the right sigma to use?
	float sigmaCalc = sigmaW * sqrt( k*k - 1.0);
      
	// convolve with gaussian filter
	blur[s] = convGauss(blur[s],sigmaCalc,sigmaCalc,1.0f);
	
	// update sigma
	sigmaW *= k;
      }

    // computes difference of gaussian
    for ( s=0; s<SCALES+2; s++ )
      {
        dogs[s] = blur[s+1] - blur[s];
      }

    // prepare analysed image
    analysed = Image<float>(dimX, dimY, ZEROS );
}

//
// Computes gradient and orientation maps
//
void gradientOrientationImages(const Image<float>& in, Image<float>& gradient,
			       Image<float>& orientation)
{
  // initialize images
  gradient = Image<float>(in.getWidth(), in.getHeight(), ZEROS);
  orientation = Image<float>(in.getWidth(), in.getHeight(), ZEROS);
  
  // for each pixel
  for(int x=1; x<in.getWidth()-1; x++)
    {
      for(int y=1; y<in.getHeight()-1; y++)
	{
	  
	  // set gradient value
	  gradient.setVal(x, y, sqrt( pow( in.getVal(x+1,y) - in.getVal(x-1,y),
					   2.0)
				      + pow( in.getVal(x,y+1) - 
					     in.getVal(x,y-1), 2.0)));

	  // set the orientation value
	  orientation.setVal(x, y, atan2( in.getVal(x,y+1) - in.getVal(x,y-1), 
					  in.getVal(x+1,y) - 
					  in.getVal(x-1,y)));
	}
    }
}

//
// Main entry for keypoints retrieval
//   look for keypoints
//
int ScaleSpace::findKeypoints(VisualObject & obj)
{
  int keyNumber = 0;
  // loop through scale spaces
  for (int sc=1; sc<=SCALES; sc++)
    {
      // precalculate for this scale every orientation and gradient
      gradientOrientationImages( blur[sc], grad, orie );
      
      // for every pixels in image except for edges
      for ( int x=EXT; x<dimX-EXT; x++ )
        {
	  for ( int y=EXT; y<dimY-EXT; y++ )
	    {
	      // check for maximum or minimum of difference-of-gaussian
	      // function in scale space
              double magnitude = 0.0;
	      if (!checkForMinMax(x, y, magnitude, dogs[sc-1], dogs[sc], 
                                  dogs[sc+1]))
		  continue;

	      // accurate localization and pruning of keypoint
	      keyNumber += accurateLocalizationAndPruning(x, y, sc, magnitude, 
                                                          obj);
            }
        }
    }
  return keyNumber;
}

//
// AccurateLocalizationAndPruning
//   This is really the heart of the SIFT algorithm
//     - 3D interpolation to find the real position of the point in (x, y, s) 
//       space
//     - handling of already done position
//     - feature vector calculation
//     - returns number of keypoints found
int ScaleSpace::accurateLocalizationAndPruning(int x, int y, int s, double mag,
                                               VisualObject & obj)
{
  Vector3 dDog, dpos;
  float new_heigth, dx2, dy2, dxdy;
  int MaxIter = 6;
  bool moved;
  int new_x = x;
  int new_y = y;
  
  // 1. Interactive 3D interpolation (x direction, y direction, and scale 
  // direction) Iterate to find good extremum position until stabilization or 
  // too many moves
  while ( (MaxIter--) >= 0 )
    {
      // 3d fit of this extremum
      dpos = fit3D( new_x, new_y, s, dDog, dx2, dy2, dxdy );
      
      // calculate new peak height
      new_heigth = dogs[s].getVal(new_x, new_y) + 0.5 * dpos.Dot( dDog );
      
      // pruning small peak
      if ( fabs(new_heigth)< PEAK_THRESH )
	return 0;
      
      // pruning big edge response
      if ( fabs(dx2*dy2-dxdy*dxdy) == 0)
	return 0;
      
      // calculate edge test
      float egdetest = pow(dx2+dy2,2.0) / fabs(dx2*dy2-dxdy*dxdy);
      if ( egdetest > pow(R_EDGE +1,2.0)/R_EDGE )
	return 0;
	
      // some overflow cases -> leave
      if ( fabs(dpos.x) > 1.5f )
	return 0;
      if ( fabs(dpos.y) > 1.5f )
	return 0;
      if ( fabs(dpos.z) > 1.5f )
	return 0;

      // make the moves
      moved = false;
      if ( dpos.x > 0.5 )
        {
	  if ( new_x < dimX - EXT )
            {
	      new_x++;
	      moved = true;
            }
        }
      if ( dpos.x < -0.5 )
        {
	  if (new_x > EXT)
            {
	      new_x--;
	      moved = true;
            }
        }
      if ( dpos.y > 0.5 )
        {	
	  if ( new_y < dimY - EXT )
            {
	      new_y++;
	      moved = true;
            }
        }
      if ( dpos.y < -0.5 )
        {
	  if (new_y > EXT )
            {
	      new_y--;
	      moved = true;
            }
        }
      
      // if stable, leave
      if (!moved)
	break;
    }
  
  // 2. if already analyzed then quit
  if (analysed.getVal(new_x, new_y) != 0.0 )
    return 0;
  analysed.setVal(new_x, new_y, 1.0);
  
  // 3. create feature vector
  float new_s = SIGMA * pow (2.0, (s + dpos.z) / SCALES);
  return obj.createKeys(new_x + dpos.x, new_y + dpos.y, new_s, dimX, dimY, 
                        octSize, mag, grad, orie);
}

// 
// Just Check in the local 3x3 space, if it's a max or min
//
bool ScaleSpace::checkForMinMax(int x, int y, double& magnitude, 
                                const Image<float> & im0, 
                                const Image<float> & im1, 
                                const Image<float> & im2 )
{
  //get value
  double val = im1.getVal(x, y);
  magnitude = fabs(val);

  // peak threshold
  if ( fabs(val) < PEAK_THRESH )
    return false;

  // verify for max or min
  if (val<im1.getVal(x-1, y))
    {
      // check for minimal
      if (val>im1.getVal(x-1, y-1))  return false;
      if (val>im1.getVal(x-1, y+1))	return false;
      if (val>im1.getVal(x  , y-1))	return false;
      if (val>im1.getVal(x  , y+1))	return false;
      if (val>im1.getVal(x+1, y-1))	return false;
      if (val>im1.getVal(x+1, y  ))	return false;		
      if (val>im1.getVal(x+1, y+1))	return false;

      // check for minimal level -1
      if (val>im0.getVal(x-1, y-1))	return false;
      if (val>im0.getVal(x-1, y  ))	return false;
      if (val>im0.getVal(x-1, y+1))	return false;
      if (val>im0.getVal(x  , y-1))	return false;
      if (val>im0.getVal(x  , y  ))	return false;
      if (val>im0.getVal(x  , y+1))	return false;
      if (val>im0.getVal(x+1, y-1))	return false;
      if (val>im0.getVal(x+1, y  ))	return false;
      if (val>im0.getVal(x+1, y+1))	return false;

      // check for minimal level +1
      if (val>im2.getVal(x-1, y-1))	return false;
      if (val>im2.getVal(x-1, y  ))	return false;
      if (val>im2.getVal(x-1, y+1))	return false;
      if (val>im2.getVal(x  , y-1))	return false;
      if (val>im2.getVal(x  , y  ))	return false;
      if (val>im2.getVal(x  , y+1))	return false;
      if (val>im2.getVal(x+1, y-1))	return false;
      if (val>im2.getVal(x+1, y  ))	return false;
      if (val>im2.getVal(x+1, y+1))  return false;
      return true;
    }
  else
    {
      // check for maximum
      if (val<im1.getVal(x-1, y-1))	return false;
      if (val<im1.getVal(x-1, y+1))	return false;
      if (val<im1.getVal(x  , y-1))	return false;
      if (val<im1.getVal(x  , y+1))	return false;
      if (val<im1.getVal(x+1, y-1))	return false;
      if (val<im1.getVal(x+1, y  ))	return false;
      if (val<im1.getVal(x+1, y+1))	return false;

      // check for minimal level -1
      if (val<im0.getVal(x-1, y-1))	return false;
      if (val<im0.getVal(x-1, y  ))	return false;
      if (val<im0.getVal(x-1, y+1))	return false;
      if (val<im0.getVal(x  , y-1))	return false;
      if (val<im0.getVal(x  , y  ))	return false;
      if (val<im0.getVal(x  , y+1))	return false;
      if (val<im0.getVal(x+1, y-1))	return false;
      if (val<im0.getVal(x+1, y  ))	return false;
      if (val<im0.getVal(x+1, y+1))	return false;

      // check for minimal level +1
      if (val<im2.getVal(x-1, y-1))	return false;
      if (val<im2.getVal(x-1, y  ))	return false;
      if (val<im2.getVal(x-1, y+1))	return false;
      if (val<im2.getVal(x  , y-1))	return false;
      if (val<im2.getVal(x  , y  ))	return false;
      if (val<im2.getVal(x  , y+1))	return false;
      if (val<im2.getVal(x+1, y-1))	return false;
      if (val<im2.getVal(x+1, y  ))	return false;
      if (val<im2.getVal(x+1, y+1))	return false;
      return true;
    }
}

//
// Fit in 3D
//   in the (X, Y, S) space, fit the peak, to find it's best extrenum
//   return solution and dDog as first derivative of Dog
//
Vector3 ScaleSpace::fit3D( int x, int y, int s, Vector3 & dDog, float& dX2, 
			   float& dY2, float& dXdY)
{
  Image<float>& Below = dogs[s - 1];
  Image<float>& Cur = dogs[s];
  Image<float>& Above = dogs[s + 1];

  // standard approximations of first derivatives d(DOG)/d(?)
  float dX = 0.5 * ( Cur.getVal(x+1, y) - Cur.getVal(x-1, y) );
  float dY = 0.5 * ( Cur.getVal(x, y+1) - Cur.getVal(x, y-1) );
  float dS = 0.5 * ( Above.getVal(x, y) - Below.getVal(x, y) );
  
  // standard approximations of seconds derivatives d²(DOG)/d(?²)
  dX2	= Cur.getVal(x-1, y) - 2.0 * Cur.getVal(x, y) + Cur.getVal(x+1, y);
  dY2 = Cur.getVal(x, y-1) - 2.0 * Cur.getVal(x, y) + Cur.getVal(x, y+1);
  float dS2	= Below.getVal(x, y) - 2.0 * Cur.getVal(x, y) + 
    Above.getVal(x, y);
	
  // standard approximation of crossed derivatives d²(DOG)/d?d?
  dXdY  = 0.25 * ( Cur.getVal(x+1, y+1) - Cur.getVal(x-1, y+1)) - 
    0.25 * ( Cur.getVal(x+1, y-1) - Cur.getVal(x-1, y-1) );
  float dSdX	= 0.25 * ( Above.getVal(x+1, y) - Above.getVal(x-1, y)) - 
    0.25 * ( Below.getVal(x+1, y) - Below.getVal(x-1, y) );
  float dSdY	= 0.25 * ( Above.getVal(x, y+1) - Above.getVal(x, y-1)) - 
    0.25 * ( Below.getVal(x, y+1) - Below.getVal(x, y-1) );
		
  // matrix to solve is
  // dX         [ dX²  dXdY dXdS ]
  // dY = [ V ] [ dXdY dY²  dYdS ]
  // dS         [ dXdS dYdS dS²  ]
  Matrix3 mat;
  Vector3 solution;
  dDog.Set( -dX, -dY, -dS );
  mat.m[0][0] = dX2;
  mat.m[1][1] = dY2;
  mat.m[2][2] = dS2;
  mat.m[1][0] = mat.m[0][1] = dXdY;
  mat.m[2][0] = mat.m[0][2] = dSdX;
  mat.m[2][1] = mat.m[1][2] = dSdY;
  solution = mat.inverse().transformVector( dDog );
  
  return solution;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
