/*!@file ScaleSpace.H Keypoint computation for SIFT obj recognition */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: James Bonaiuto <bonaiuto@usc.edu>

#ifdef MSVC
#pragma once
#endif

#ifndef _SCALE_SPACE_H
#define _SCALE_SPACE_H

#include "ImageSet.H"
#include "Image.H"
#include "VisualObject.H"
#include "Vector3.H"

const int   SCALES     	= 3;	    // number of stages in an octave
const float SIGMA	= 1.6f;     // sigma
const int   EXT		= 10;	    // edge width around picture where no 
                                    // keypoint can be located
const int   R_EDGE      = 10;	    // edge response threshold
const float PEAK_THRESH = 0.001f;   // peak below that doesn't count
const float PI          = M_PI;	    // PI

//! ScaleSpace
/*! It handles a scale space with all the pictures associated.
    Implements what is needed for sift detection in one level of the full 
    SpaceScale
*/
class ScaleSpace
{
public:
  //! Constructor
  ScaleSpace() {  }

  //! Destructor
  ~ScaleSpace(void) { reset(); }
  
  //! reset
  void reset()
  {
    blur.clear();
    dogs.clear();
  }
  
  //! Init this ScaleSpace provided the scale dimension and the picture
  void initScaleSpace(float currentscale, const Image<float> & image);
	
  //! Find the keypoints in the previously provided picture
  /*! Returns the number of keypoints found
   */
  int findKeypoints(VisualObject& obj);

private :
  int dimX, dimY;	// dimension of picture
  float octSize;	// dimension of octave ( the current scale )
  ImageSet<float> blur;	// gaussian blurred pictures for this scale space
  ImageSet<float> dogs;	// difference of gaussian picture for this scale space

  Image<float>  analysed;// map which tells if a pixel was already analysed
  Image<float>  grad;	 // gradient maps
  Image<float>  orie;	 // orientation maps
	
  //! return true, if it's a max or min in the local 3x3 matrix
  bool checkForMinMax(int x, int y, double& magnitude, 
                      const Image<float> & im0, const Image<float> & im1, 
                      const Image<float> & im2 );
  
  //! fits the paraboloid in (X, Y, S) space : 3D Hessian matrix inversion
  Vector3 fit3D(int x, int y, int s, Vector3 & dDog, float& dX2, float& dY2, 
                float& dXdY);

  //! localization and pruning of keypoint
  int accurateLocalizationAndPruning(int x, int y, int s, double mag,
                                     VisualObject & obj); 
};


#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
