/*!@file ShapeEstimator.C Estimate the shape/size of an attended object */

// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: ShapeEstimator.C 3992 2004-11-05 04:23:21Z jbonaiuto $
//

#include "saliency.H"
#include "log.H"
#include "ModelManager.H"
#include "ShapeEstimator.H"
#include "Point2D.H"
#include "ChannelBase.H"
#include "Image_FilterOps.H"
#include "Image_Transforms.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "ImageSet.H"
#include "SingleChannel.H"
#include "ComplexChannel.H"
#include "ModelOptionDefs.H"

#define SIGMA 20
#define CLAMP 0.4

// ######################################################################
ShapeEstimator::ShapeEstimator(ModelManager& mgr, const std::string& descrName,
                               const std::string& tagName,
                               const SharedPtr<VisualCortex> vcx) :
  ModelComponent(mgr, descrName, tagName),
  itsMode(this, "ShapeEstimatorMode", SEMfeatureMap),
  itsSmMethod(this, "ShapeEstimatorSmoothMethod", SESMgaussian),
  itsInitialized(false),
  itsVCX(vcx),
  itsSubmapNum(-1),
  itsObjMaskComputed(false)
{
  //manager->requestOption(this, itsMode);
}


// ######################################################################
void ShapeEstimator::setVC(const SharedPtr<VisualCortex>& vc)
{ 
  itsVCX = vc; 
  reset(false);
}


// ######################################################################
void ShapeEstimator::exportOptions(const int whatToExport, const bool recurse)
{
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this,itsMode);
    manager->requestOption(this,itsSmMethod);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void ShapeEstimator::reset(const bool recurse)
{
  // reset some stuff for ShapeEstimator
  itsInitialized = false;
  itsObjMaskComputed = false;
  itsChannel.reset();
  itsSubmapNum = -1;
  itsWinningMap.freeMem();
  winMapNormalized.freeMem();
  itsObjectMask.freeMem();
  itsIORMask.freeMem();
  itsCumMask.freeMem();

  // call reset in ModelComponent
  ModelComponent::reset(recurse);
}

// ######################################################################
void ShapeEstimator::compute(const Point2D& winner)
{
  ASSERT(itsVCX.get() != NULL);

  itsInitialized = true; 
  itsObjMaskComputed = false;
  itsWinner = winner;

  // kill old results:
  itsWinningMap.freeMem(); itsObjectMask.freeMem(); 
  winMapNormalized.freeMem(); itsIORMask.freeMem(); 
  itsSmoothMask.freeMem(); itsWinningLabel.clear();

  // nothing to do?
  if (itsMode.getVal() == SEMnone) 
    return;

  // in case we use saliency map, just store the SM as the winning map
  if (itsMode.getVal() == SEMsaliencyMap)
    {
      itsChannel.reset();
      itsSubmapNum = -1;
      itsWinningMap = itsVCX->getOutput();
      itsWinningLabel = "SM";
      return;
    }

  // loop over the channels
  float mx = -1.0F;
  for (uint i = 0; i < itsVCX->numChannels(); ++i)
    {
      SharedPtr<ChannelBase> chan = itsVCX->channelPtr(i);
      Image<float> output = chan->getOutput();
      output *= chan->getTotalWeight();

      float curr_val = output.getVal(downScaleCoords(winner,output.getDims()));
      LDEBUG("i = %i; channel = %s; val = %g",i,chan->tagName().data(),curr_val); 

      // do we have a new max?
      if (curr_val >= mx) { mx = curr_val; itsChannel = chan; }
    }

  // we did not find a max here? -> resort back to saliency map
  if (mx == -1.0F)
    {
      itsChannel.reset();
      itsSubmapNum = -1;
      itsWinningMap = itsVCX->getOutput();
      itsWinningLabel = "SM";

      return;
    }

  // in case we use conspicuity map, store the channel's output
  // as the winning map
  if (itsMode.getVal() == SEMconspicuityMap)
    {
      itsSubmapNum = -1;
      itsWinningMap = itsChannel->getOutput();
      itsWinningLabel = itsChannel->tagName();
      return;
    }

  // loop over the submaps of the winning channel
  mx = -1.0F;
  for (uint i = 0; i < itsChannel->numSubmaps(); ++i)
    {
      Image<float> submap = itsChannel->getSubmap(i);

      float curr_val = submap.getVal(downScaleCoords(winner,submap.getDims()));

      // do we have a new max?
      if (curr_val >= mx) { mx = curr_val; itsSubmapNum = i; }
    }

  // we did not find a max here? -> resort back to conspicuity map
  if (mx == -1.0F)
    {
      itsSubmapNum = -1;
      itsWinningMap = itsChannel->getOutput();
      itsWinningLabel = itsChannel->tagName();
      return;
    }

  // in case we use feature map, store the result ...
  itsWinningMap = itsChannel->getSubmap(itsSubmapNum);
  itsWinningLabel = itsChannel->getSubmapName(itsSubmapNum);

}

// ######################################################################
Image<byte>& ShapeEstimator::getObjectMask()
{
  ASSERT(itsInitialized);
  if (itsObjMaskComputed) return itsObjectMask; // cached
  if (itsMode.getVal() == SEMnone) return itsObjectMask; // should be empty

  winMapNormalized = itsWinningMap;
  winMapNormalized.normalize(0.0F, 1.0F);

  Point2D wp = downScaleCoords(itsWinner, itsWinningMap.getDims());

  // the winner usually falls at the junction of four pixels in the
  // saliency/conspicuity/feature map - have a look at all of the pixels
  std::vector<Point2D> surr(4);
  surr[0] = Point2D(0,0); surr[1] = Point2D(-1,0);
  surr[2] = Point2D(0,1); surr[3] = Point2D(-1,1);

  bool haveValue = false;
  for (uint i = 0; i < surr.size(); ++i)
    {
      Point2D wp2 = wp + surr[i];
      wp2.fitToDims(itsWinningMap.getDims());
      if (itsWinningMap.getVal(wp2) > 0.0F)
        {
          wp = wp2;
          haveValue = true;
          break;
        }
    }

  if (haveValue)
    itsObjectMask = segmentObjectClean(winMapNormalized, wp);
  else
    itsObjectMask.freeMem();

  itsObjMaskComputed = true;
  return itsObjectMask;
}

// ######################################################################
Point2D ShapeEstimator::downScaleCoords(Point2D winner, Dims targetDims)
{
  ASSERT(itsVCX.get() != NULL);
  Dims sDims = itsVCX->getInputDims();
  int i = (winner.i * targetDims.w()) / sDims.w();
  int j = (winner.j * targetDims.h()) / sDims.h();
  return Point2D(i,j);
}

// ######################################################################
int ShapeEstimator::getObjectArea()
{
  ASSERT(itsInitialized);

  // make sure our Object Mask is up to date
  getObjectMask();

  // object mask not initialized? -> return 0
  if (!itsObjectMask.initialized()) return 0;

  double scale = itsVCX->getInputDims().sz() / itsObjectMask.getSize();

  return int(sum(itsObjectMask)/255.0 * scale);
}

// #####################################################################
Image<PixRGB<byte> > ShapeEstimator::getObjectImage(Image<PixRGB<byte> >& scene)
{
  if(!itsSmoothMask.initialized()) getSmoothMask();

  int xOffset = 0;
  int yOffset = 0;
  int w = itsSmoothMask.getWidth();
  int h = itsSmoothMask.getHeight();
  float threshold = 1.0;

  //Find x
  for(int i=0; i<itsSmoothMask.getWidth(); i++)
    {
      bool found = false;
      for(int j=0; j<itsSmoothMask.getHeight(); j++)
        {
          if(itsSmoothMask.getVal(i,j) >= threshold)
            {
              xOffset = i;
              found = true;
              break;
            }
        }
      if(found)
        break;
    }

  //Find y
  for(int i=0; i<itsSmoothMask.getHeight(); i++)
    {
      bool found = false;
      for(int j=0; j<itsSmoothMask.getWidth(); j++)
        {
          if(itsSmoothMask.getVal(j,i) >= threshold)
            {
              yOffset = i;
              found = true;
              break;
            }
        }
      if(found)
        break;
    }

  //Find width
  for(int i=itsSmoothMask.getWidth()-1; i>=0; i--)
    {
      bool found = false;
      for(int j=0; j<itsSmoothMask.getHeight(); j++)
        {
          if(itsSmoothMask.getVal(i,j) >= threshold)
            {
              w = i;
              found = true;
              break;
            }
        }
      if(found)
        break;
    }
  w -= xOffset;

  //Find height
  for(int i=itsSmoothMask.getHeight()-1; i>=0; i--)
    {
      bool found = false;
      for(int j=0; j<itsSmoothMask.getWidth(); j++)
        {
          if(itsSmoothMask.getVal(j,i) >= threshold)
            {
              h = i;
              found = true;
              break;
            }
        }
      if(found)
        break;
    }
  h -= yOffset;

  Rectangle r(Point2D(xOffset, yOffset), Dims(w, h));
  Image<PixRGB<byte> > croppedObjImg = crop(scene, r);
  return croppedObjImg;
}

// ######################################################################
Image<byte>& ShapeEstimator::getIORMask()
{
  ASSERT(itsInitialized);
  if (itsIORMask.initialized()) return itsIORMask; // cached
  if (itsMode.getVal() == SEMnone) return itsIORMask; // should be empty

  getObjectMask();
  if (itsObjectMask.initialized())
    itsIORMask = (Image<byte>) (itsObjectMask * winMapNormalized);
  else
    itsIORMask.freeMem();

  return itsIORMask;
}

// ######################################################################
Image<float>& ShapeEstimator::getSmoothMask()
{
  ASSERT(itsInitialized);
  if (itsSmoothMask.initialized()) return itsSmoothMask; // cached
  if (itsMode.getVal() == SEMnone) return itsSmoothMask; // should be empty

  getObjectMask();
  Dims dims = itsVCX->getInputDims();
  if (itsObjectMask.initialized())
    {
      switch(itsSmMethod.getVal())
        {
        case SESMgaussian:
          {
            Image<float> temp = 
              (Image<float>)scaleBlock(itsObjectMask, dims);
            itsSmoothMask = convGauss<float>(temp,SIGMA,SIGMA,5);
            itsSmoothMask.normalize(0.0F,3.0F);
            itsSmoothMask.clamp(0.0F,1.0F);
            break;
          }
          
        case SESMchamfer:
          {
            if (!structEl.initialized())
              {
                //create the structuring element for the opening
                const int ss = 8;
                structEl = Image<byte>(ss+ss,ss+ss,ZEROS);
                structEl.drawDisk(Point2D(ss,ss),ss,(byte)1);
              }

            const byte cutoff = 100;
            Image<byte> temp = scaleBlock(itsObjectMask, dims);
            temp = chamfer34(openImg(temp,structEl),cutoff);
            itsSmoothMask = binaryReverse((Image<float>)temp,255.0F);
            itsSmoothMask.normalize(0.0F,1.0F);
            break;
          }

        case SESMnone:
          {
            itsSmoothMask = scaleBlock(itsObjectMask, dims)/255.0;
            break;
          }
          
        default: LFATAL("Unknown Smoothing Method");
        }


      // update the cumulative mask
      if (itsCumMask.initialized())
        itsCumMask = takeMax(itsCumMask,itsSmoothMask);
      else
        itsCumMask = itsSmoothMask;
    }
  else
    {
      // no object mask available -> invalidate itsSmoothMask
      itsSmoothMask.freeMem();
    }
  
  return itsSmoothMask;
}

// ######################################################################
Image<float> ShapeEstimator::getCumSmoothMask() const
{
  if (itsCumMask.initialized() == false) return itsCumMask;

  Image<float> result(itsCumMask);
  result.clamp(CLAMP, 1.0F);
  return result;
}

// ######################################################################
Image<float> ShapeEstimator::getNegCumSmoothMask() const
{
  if (itsCumMask.initialized() == false) return itsCumMask;

  Image<float> result(itsCumMask * (-1.0F) + 1.0F);
  result.clamp(1.0F-CLAMP, 1.0F);
  return result;
}

// ######################################################################
void ShapeEstimator::resetCumMask()
{ itsCumMask.freeMem(); }

// ######################################################################
Point2D ShapeEstimator::getWinner() const
{ ASSERT (itsInitialized); return itsWinner; }

// ######################################################################
ChannelBase& ShapeEstimator::getChannel() const
{ ASSERT (itsInitialized); return *itsChannel; }

// ######################################################################
SingleChannel& ShapeEstimator::getSingleChannel() const
{
  ASSERT(itsInitialized);
  if (itsMode.getVal() != SEMfeatureMap) return (SingleChannel&)*itsChannel;
  ChannelBase* result = itsChannel.get();

  uint oldIdx = itsSubmapNum;

  while (result->numChans() >= 0)
    {
      uint newIdx;
      ComplexChannel* tmp = (ComplexChannel*)result;
      result = &(tmp->subChanFromSubmapNum(oldIdx, newIdx));
      oldIdx = newIdx;
    }

  return (SingleChannel&)*result;
}

// ######################################################################
const ImageSet<float>& ShapeEstimator::getPyramid(const int idx) const
{
  ASSERT(itsMode.getVal() == SEMfeatureMap);
  return getSingleChannel().pyramid(idx);
}

// ######################################################################
int ShapeEstimator::getSubmapNum() const
{ ASSERT(itsInitialized); return itsSubmapNum; }

// ######################################################################
Image<float>& ShapeEstimator::getWinningMap()
{ ASSERT(itsInitialized); return itsWinningMap; }

// ######################################################################
std::string ShapeEstimator::getWinningLabel() const
{ ASSERT(itsInitialized); return itsWinningLabel; } 

// ######################################################################
bool ShapeEstimator::isInitialized() const
{ return itsInitialized; }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
