/*!@file ShapeEstimator.H Estimate the shape/size of an attended object */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Dirk Walther <walther>
// $Id: ShapeEstimator.H 3826 2004-10-08 06:04:59Z jbonaiuto $
//

#ifndef SHAPEESTIMATOR_H_DEFINED
#define SHAPEESTIMATOR_H_DEFINED

#include <string>
#include "Image.H"
#include "VisualCortex.H"
#include "SharedPtr.H"
#include "ModelComponent.H"
#include "ShapeEstimatorModes.H"
#include "Rectangle.H"

class VisualCortex;
class Point2D;
class ChannelBase;
class SingleChannel;
template <class T> class ImageSet;


//! everything realted to the shape estimation procedure
/*! class to take care of all things related to the shape estimation
  procedure published in Walther et al, BMCV 02 */
class ShapeEstimator : public ModelComponent
{
public:

  //! Constructor
  /*! @param vcx this is the hook for the class to obtain its information
    on channels, submaps etc. */
  ShapeEstimator(ModelManager& mgr,
                 const std::string& descrName = "Shape Estimator",
                 const std::string& tagName = "shapeestimator",
                 const SharedPtr<VisualCortex> vcx =
                 SharedPtr<VisualCortex>(NULL));

  //! set the VisualCortex if it was not given at construction
  void setVC(const SharedPtr<VisualCortex>& vc);

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset ShapeEstimator
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! re-compute for a new winner
  void compute(const Point2D& winner);

  //! return the winner from the last compute call
  Point2D getWinner() const;

  //! returns a pointer to the winning channel
  /*! for <tt>mode = saliencyMap</tt> the pointer will be empty*/
  ChannelBase& getChannel() const;

  //! returns pointer to the winning SingleChannel (only for mode = featureMap)
  SingleChannel& getSingleChannel() const;

  //! returns pyramid of the winning SingleChannel (only for mode = featureMap)
  /*!@param idx index of the pyramid in the queue (default: 0)*/
  const ImageSet<float>& getPyramid(const int idx = 0) const;

  //! returns the number (the index) of the winning submap in the winning channel
  /*! makes only sense for <tt>mode = featureMap</tt>*/
  int getSubmapNum() const;

  //! returns the winning map (in saliency map coordinates)
  /*! for <tt>mode == saliencyMap</tt>, a copy of the saliency map;
    for <tt>mode == conspicuityMap</tt>, the output of the winning channel;
    for <tt>mode == featureMap</tt>, winning submap of the winning channel*/
  Image<float>& getWinningMap();

  //! the label of the winning map
  std::string getWinningLabel() const;

  //! whether compute has been run at least once
  bool isInitialized() const;

  //! return a binary mask of the attended object {0,255} in saliency map coordinates
  Image<byte>& getObjectMask();

  //! returns the area of the object Mask, scaled up to image coordinates
  int getObjectArea();

  //! returns the object image
  Image<PixRGB<byte> > getObjectImage(Image<PixRGB<byte> >& scene);

  //! return a graded mask for IOR [0,255] in saliency map coordinates
  Image<byte>& getIORMask();

  //! return an upscaled and smoothed mask [0.0,1.0] in image coordinates
  Image<float>& getSmoothMask();

  //! return the cumulative smoothed mask [0.0,1.0] in image coordinates
  /*! IMPORTANT: for this to work, getSmoothMask has to be called
    at least once after each new compute call */
  Image<float> getCumSmoothMask() const;

  //! return the negative cumulative smoothed mask [0.0,1.0] in image coordinates
  /*! IMPORTANT: for this to work, getSmoothMask has to be called
    at least once after each new compute call */
  Image<float> getNegCumSmoothMask() const;

  //! reset the cumulative mask
  void resetCumMask();

protected:

  //! Determines what information is used as a source for the shape estimation
  /*! possible values:<ul><li>"None" - shape estimator is not used;
    <li>"FeatureMap" - the shape is extracted from the winning Feature Map;
    <li>"ConspicuityMap" - the shape is extracted from the winning Conspicuity Map;
    <li>"SaliencyMap" - the shape is extracted from the Saliency Map</ul>*/
  ModelParam<ShapeEstimatorMode> itsMode;

  //! Determines the Smoothing Method used to extract a map in image coordinates
  /*! possible values:<ul><li>"None" - no smoothing, the result is just scaled up in blocks;
    <li>"Gaussian" - smoothing is done by convolving with a large 2D Gaussian kernel;
    <li>"Chamfer" - smoothing is done with Opening and Chamfering (takes about half 
    the time compared to "Gaussian"</ul>*/
  ModelParam<ShapeEstimatorSmoothMethod> itsSmMethod;

private:
  // compute the winner in some other coordinates
  Point2D downScaleCoords(Point2D winner, Dims targetDims);

  // true if compute() has been called at least once; initially false
  bool itsInitialized;

  // a hook to the VisualCortex from which the ShapeEstimator
  // sucks all the info it needs
  SharedPtr<VisualCortex> itsVCX;

  // the currently attended position
  Point2D itsWinner;

  // a pointer to the winning channel
  // for mode = SEMfeatureMap this is a SingleChannel
  // for mode = SEMconspicuityMap this may be a CompexChannel
  // for mode = SEMsaliencyMap this is not initialized
  SharedPtr<ChannelBase> itsChannel;

  // the index of the winning map in itsChannel
  // for mode = SEMsaliencyMap, this is -1
  int itsSubmapNum;

  // a copy of the winning map in saliency map coordinates
  Image<float> itsWinningMap;

  // the winning map, normalized to (0.0,1.0)
  Image<float> winMapNormalized;

  // a string containing info on where the winning map comes from
  std::string itsWinningLabel;

  // a binary mask {0,255}  in saliency map coordinates
  // specifiying the extend of the object
  Image<byte> itsObjectMask;

  // true if the object mask has been computed yet
  bool itsObjMaskComputed;

  // a byte map [0,255] - 0 everywhere outside the object,
  // (winMapNormalized * 255) everywhere inside the object
  // this is used for IOR in a graded manner - the stronger
  // the winning property is in a certain position, the stronger
  // this position is inhibited 
  Image<byte> itsIORMask;

  // a float map [0.0,1.0] in image coordinates
  // created from itsObjectMask by scaling itsObjectMask
  // up and smoothing out the edges with some specified
  // smoothing method
  Image<float> itsSmoothMask;

  // a float map [0.0,1.0] in image coordinates that
  // cumulates all the itsSmoothMasks using max
  // until resetCumMask is called
  Image<float> itsCumMask;

  // the structuring element for eroding and dilating
  // for the chamfer smoothing method
  Image<byte> structEl;
};

#endif


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
