/*!@file SharedPtr.H a shared pointer class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: SharedPtr.H 3840 2004-10-14 17:51:36Z itti $
//

#ifndef SHAREDPTR_H_DEFINED
#define SHAREDPTR_H_DEFINED

//! A smart pointer with reference counted copy semantics.
/*! The object pointed to is deleted when the last SharedPtr pointing to it is
    destroyed or reset. Take note of the WARNING given in the documentation
    for the constructor that takes a raw pointer as its argument.

    Borrowed and modified from boost.org smart_ptr.hpp
    Original copyright notice:

     (C) Copyright Greg Colvin and Beman Dawes 1998, 1999. Permission
     to copy, use, modify, sell and distribute this software is granted
     provided this copyright notice appears in all copies. This
     software is provided "as is" without express or implied warranty,
     and with no claim as to its suitability for any purpose.
 */
template<class T>
class SharedPtr
{
public:
  //! Construct from a raw pointer. WARNING: do this only ONCE per pointer!
  /*! Bad Things (TM) will happen if this constructor is used multiple
      times to make multiple SharedPtr's from a single raw pointer; that's
      because each SharedPtr will start off thinking that it is the unique
      owner, and so they will each start their own ref count. If multiple
      SharedPtr's are needed for the same object (after all, that's the whole
      point of a shared ptr), only the first one should be constructed from
      the raw pointer, and subsequent ones should be constructed from the
      first using the copy constructor or assignment operator. */
  explicit inline SharedPtr(T* p =0);

  //! Copy another SharedPtr. The ref count is incremented by one.
  inline SharedPtr(const SharedPtr& r) throw();

  //! Destructor decrements the ref count by one.
  /*! If we were the last pointer to the pointee, then the pointee and the
      reference count are both delete'd. */
  inline ~SharedPtr();

  //! Assign from another SharedPtr. The ref count is incremented by one.
  inline SharedPtr& operator=(const SharedPtr& r);

  //! Copy construct from a SharedPtr of a different type.
  /*! The other type TT must be related to T by an appropriate inheritance
      relationship or by a change in const-qualification. */
  template<class TT>
  inline SharedPtr(const SharedPtr<TT>& r) throw();

  //! Assign from a SharedPtr of a different type.
  /*! The other type TT must be related to T by an appropriate inheritance
      relationship or by a change in const-qualification. */
  template<class TT>
  inline SharedPtr& operator=(const SharedPtr<TT>& r);

  //! Assign from a SharedPtr of a different type, with dynamic casting.
  /*! This is the SharedPtr equivalent of doing a dynamic_cast between
    two raw pointers. Doing 'SharedPtr<T> ptr;
    ptr.dynCastFrom(other);' with 'other' a SharedPtr<TT> will
    internally do a dynamic_cast from TT* to T* and will assign the
    result to ptr. If the dynamic_cast succeeds, ptr and other will
    share the pointee as if they were two normal SharedPtr onto it
    (even though ptr and other are SharedPtr of different types). If
    the dynamic_cast fails then ptr will point to NULL. Typical usage
    of this is to recover an expected derived class from a base class:
    for example,
    <PRE>

    // get our visual cortex from Brain. Brain provides a getVC() function
    // which returns a SharedPtr onto the VisualCortex base class no matter
    // which derivation of that we are actually running right now:
    SharedPtr<VisualCortex> vcx = brain->getVC();

    // do something special if our VisualCortex actually happens to be of
    // type VisualCortexEyeMvt which is derived from VisualCortex:
    SharedPtr<VisualCortexEyeMvt> vcem;
    vcem.dynCastFrom(vcx);

    // vcem points to NULL unless the pointee of vcx could be
    // dynamic_cast'ed to vcem's type.
    if (vcem.get()) {
        // yes, indeed we are running a VisualCortexEyeMvt
        ...
    }
    </PRE>

    Note: make sure you always use virtual destructors in your
    pointees, especially when you use dynCastFrom. Indeed, whichever
    SharedPtr is the last one to run out of scope will destroy the
    pointee. */
  template<class TT>
  inline SharedPtr& dynCastFrom(const SharedPtr<TT>& r);

  //! Make the SharedPtr point to a different (optionally null) pointee.
  inline void reset(T* p=0);

  //! Get a reference to the pointee.
  inline T& operator*() const throw() { return *px; }

  //! Get the pointee for accessing its members.
  inline T* operator->() const throw() { return px; }

  //! Get the pointee.
  inline T* get() const throw() { return px; }

  //! Query how many SharedPtr's are sharing the pointee.
  inline long useCount() const throw() { return *pn; }

  //! Query whether the SharedPtr is the unique owner of its pointee.
  inline bool unique() const throw() { return *pn == 1; }

  //! Swap the pointees of two SharedPtr's.
  inline void swap(SharedPtr<T>& that) throw();

private:
  T*     px;     // contained pointer
  long*  pn;     // ptr to reference counter

  template<class TT> friend class SharedPtr;
};



// ######################################################################
// ######################################################################
// ##### INLINED FREE FUNCTIONS:
// ######################################################################
// ######################################################################

//! Test whether two SharedPtr's point to the same object.
template<class T, class U>
inline bool operator==(const SharedPtr<T>& a, const SharedPtr<U>& b)
{
  return a.get() == b.get();
}

//! Test whether two SharedPtr's point to different objects.
template<class T, class U>
inline bool operator!=(const SharedPtr<T>& a, const SharedPtr<U>& b)
{
  return a.get() != b.get();
}

//! A convenience function for making a SharedPtr out of a raw pointer.
template <class T>
SharedPtr<T> makeShared(T* t) { return SharedPtr<T>(t); }



// ######################################################################
// ######################################################################
// ##### INLINED MEMBER FUNCTIONS:
// ######################################################################
// ######################################################################

// ######################################################################
template <class T> inline
SharedPtr<T>::SharedPtr(T* p) :
  px(p), pn(0)
{
  // prevent leak if new throws:
  try { pn = new long(1); }
  catch (...) { delete p; throw; }
}

// ######################################################################
template <class T> inline
SharedPtr<T>::SharedPtr(const SharedPtr<T>& r) throw() :
  px(r.px), pn(r.pn)
{
  ++(*pn);
}

// ######################################################################
template <class T> inline
SharedPtr<T>::~SharedPtr() { if (--*pn == 0) { delete px; delete pn; } }

// ######################################################################
template <class T> inline
SharedPtr<T>& SharedPtr<T>::operator=(const SharedPtr<T>& r)
{
  SharedPtr(r).swap(*this);
  return *this;
}

// ######################################################################
template <class T>
template<class TT> inline
SharedPtr<T>::SharedPtr(const SharedPtr<TT>& r) throw() :
  px(r.px), pn(r.pn)
{
  ++(*pn);
}

// ######################################################################
template <class T>
template<class TT> inline
SharedPtr<T>& SharedPtr<T>::operator=(const SharedPtr<TT>& r)
{
  SharedPtr(r).swap(*this);
  return *this;
}

// ######################################################################
template <class T>
template<class TT> inline 
SharedPtr<T>& SharedPtr<T>::dynCastFrom(const SharedPtr<TT>& r)
{
  // if we are already initialized to something (and we always are),
  // simulate a destroy:
  if (--*pn == 0) { delete px; delete pn; }

  // share the ref count without increasing it:
  pn = r.pn;

  // increase it to account for our existence:
  ++(*pn);

  // get a pointer to the contents and dynamic_cast it to our type:
  px = dynamic_cast<T*>(r.px);

  return *this;
}

// ######################################################################
template <class T> inline
void SharedPtr<T>::reset(T* p)
{
  SharedPtr(p).swap(*this);
}

// ######################################################################
template <class T> inline
void SharedPtr<T>::swap(SharedPtr<T>& that) throw()
{
  T* that_px = that.px; that.px = this->px; this->px = that_px;
  long* that_pn = that.pn; that.pn = this->pn; this->pn = that_pn;
}

#endif // !SHAREDPTR_H_DEFINED

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
