/*!@file SimulationViewer.C visualize various model simulations */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewer.C 4182 2005-02-12 23:18:35Z itti $
//

#include "SimulationViewer.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "SimulationViewerStd.H"
#include "SimulationViewerCompress.H"
#include "SimulationViewerEyeMvt.H"
#include "SimulationViewerEyeMvt2.H"
#include "SimulationViewerEyeSim.H"

// ######################################################################
SimulationViewer::SimulationViewer(ModelManager& mgr,
                                   const std::string& descrName,
                                   const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName)
{ }

// ######################################################################
SimulationViewer::~SimulationViewer()
{ }

// ######################################################################
void SimulationViewer::setBrain(Brain* brain)
{ }

// ######################################################################
// ######################################################################
// ######################################################################
SimulationViewerConfigurator::
SimulationViewerConfigurator(ModelManager& mgr,
                         const std::string& descrName,
                         const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSVtype(this, "SimulationViewerType", "None"),
  itsSV(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
SimulationViewerConfigurator::~SimulationViewerConfigurator()
{  }

// ######################################################################
void SimulationViewerConfigurator::exportOptions(const int whatToExport,
                                                 const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsSVtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<SimulationViewer> SimulationViewerConfigurator::getSV() const
{ return itsSV; }

// ######################################################################
bool SimulationViewerConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsSVtype.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsSV.get()) { removeSubComponent(itsSV); itsSV.reset(NULL); }

    // instantiate a SV of the appropriate type:
    if (value.compare("None") == 0)          // no SimulationViewer
      itsSV.reset(NULL);
    else if (value.compare("Std") == 0)      // standard
      itsSV.reset(new SimulationViewerStd(*manager));
    else if (value.compare("Compress") == 0) // compression
      itsSV.reset(new SimulationViewerCompress(*manager));
    else if (value.compare("EyeMvt") == 0) // eye movements
      itsSV.reset(new SimulationViewerEyeMvt(*manager));
    else if (value.compare("EyeMvt2") == 0) // eye movements v 2.0
      itsSV.reset(new SimulationViewerEyeMvt2(*manager));
    else if (value.compare("EyeSim") == 0) // eye movements simulation
      itsSV.reset(new SimulationViewerEyeSim(*manager));
    else
      LFATAL("Unknown SimulationViewer type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsSV.get()) {
      addSubComponent(itsSV);

      // tell the controller to export its options:
      itsSV->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected SV of type %s", value.c_str());
  }
  return ret;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
