/*!@file SimulationViewer.H visualize various model simulations */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewer.H 2493 2003-04-29 08:42:03Z itti $
//

#ifndef SIMULATIONVIEWER_H_DEFINED
#define SIMULATIONVIEWER_H_DEFINED

#include "ModelComponent.H"
class OutputFrameSeries;
class VisualCortex;
template <class T> class PixRGB;
template <class T> class Image;
class Point2DT;
class WTAwinner;
class Brain;

#include "Pixels.H"
#include "Image.H" // needed by getTraj()

//! This class provides an abstract interface for a simulation viewer
/*! Its typical usage is with a model that generates shifts of
  attention, eye and head movements. Derivatives from this class will
  handle visualizing those events. See SimulationViewerStd.H for an
  example of such derivative. The base class described here just
  specifies the interface. */
class SimulationViewer : public ModelComponent {
public:
  // ######################################################################
  /*! @name Constructors and destructors */
  //@{

  //! Constructor. See ModelComponent.H.
    /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage */
  SimulationViewer(ModelManager& mgr,
                   const std::string& descrName = "Simulation Viewer",
                   const std::string& tagName = "SimulationViewer");

  //! Destructor
  virtual ~SimulationViewer();

  //! Set our brain
  /*! Some viewings may require a Brain. Since in the base class we
    don't have a Brain data member, the base class implementation of
    this function is a no-op. Derived class should overload this
    accordingly if they need a Brain. */
  virtual void setBrain(Brain* brain);

  //@}

  // ######################################################################
  /*! @name Events that will trigger new visualizations */
  //@{

  //! Ready to load up a new input frame
  /*! This is called by Brain just before a new input frame is being
    loaded into Brain. Thus, this gives a last opportunity for
    visualizations on the current input, just before it gets replaced
    by the new one. */
  virtual void newInputReady() = 0;

  //! Got a new input frame at the specified (current) time
  /*! This is called by Brain after the frame is loaded into the
    Brain, VisualCortex and others, and after all the low-level
    convoultions have been made.
      @param input the new input image
      @param t the current time, in seconds */
  virtual void newInputFrame(const Image< PixRGB<byte> >& input,
                             const double t) = 0;

  //! Update with availability of new channel drawings
  /*! @param vcx The VisualCortex pointer allows us to access the
    channels if we need to. */
  virtual void channelUpdate(SharedPtr<VisualCortex>& vcx) = 0;

  //! Called every simulation time step (time is in seconds)
  virtual void evolve(const double t) = 0;

  //! Update with new covert attention position
  /*! @param win the coordinates and other info about this new covert
      shift of attention
      @param foaMask an image that is all-zero except that it is 255
      inside a region that corresponds to the shape and size of the
      current focus of attention. If such image is not available, an
      uninitialized image should be passed here, otherwise it should
      have same dims as the input frames. */
  virtual void attentionShift(const WTAwinner& win,
                              const Image<byte>& foaMask) = 0;

  //! Update with new eye position
  /*! @param curreye the coordinates and time of this new overt
      shift of attention (=eye movement)
      @param foveaMask an image that is all-zero except that it is 255
      inside a region that corresponds to the shape and size of the
      current fovea. If such image is not available, an uninitialized
      image should be passed here, otherwise it should have same dims as 
      the input frames. */
  virtual void eyeMovement(const Point2DT& curreye,
                           const Image<byte>& foveaMask) = 0;

  //! Update with new head position
  /*! @param currhead the coordinates and time of the new head position
      @param headMask an image that is all-zero except that it is 255
      inside a region that corresponds to the shape and size of the
      current head marker. If such image is not available, an uninitialized
      image should be passed here, otherwise it should have same dims as 
      the input frames. */
  virtual void headMovement(const Point2DT& currhead,
                            const Image<byte>& headMask) = 0;

  //! Update eye saccade status
  /*! inSaccade should be true when a saccade starts and false when it ends */
  virtual void eyeSaccade(const bool inSaccade) = 0;

  //! Update head saccade status
  /*! inSaccade should be true when a saccade starts and false when it ends */
  virtual void headSaccade(const bool inSaccade) = 0;

  //! Update eye blink status
  /*! inBlink should be true when a blink starts and false when it ends */
  virtual void eyeBlink(const bool inBlink) = 0;

  //@}

  // ######################################################################
  /*! @name Output of our current internal drawings */
  //@{

  //! Get the attention/eye/head trajectory image
  /*! This is called internally by saveResults(), which should
    normally be used. It is provided here for those models which do
    not use FrameSeries and hence cannot call saveResults(). */
  virtual Image< PixRGB<byte> > getTraj(const double t) = 0;

  //! Save our various results
  /*! @param ofs this will figure out file names and formats and save our
      results */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs) = 0;

  //@}
};

// ######################################################################
// ######################################################################
// ######################################################################
//! SimulationViewer configurator
/*! This will export the --sv-type=XX command-line option and will
  instantiate a SimulationViewer of the desired type as the option gets
  assigned a value. As this happens, new options may become available
  in the command-line. To see them, use --help AFTER you have chosen
  the type to use. The current SimulationViewer may be retrieved using
  getSV(). */
class SimulationViewerConfigurator : public ModelComponent
{
public:
  //! Constructor
  SimulationViewerConfigurator(ModelManager& mgr,
                           const std::string& descrName =
                           "Simulation Viewer Configurator",
                           const std::string& tagName =
                           "SimulationViewerConfigurator");

  //! destructor
  virtual ~SimulationViewerConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SimulationViewer once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen SimulationViewer
  /*! You should call this during start() of the ModelComponent that
    needs the SimulationViewer. CAUTION: this may be a
    SharedPtr(NULL) if no controller has been chosen. */
  SharedPtr<SimulationViewer> getSV() const;

protected:
  ModelParam<std::string> itsSVtype; //!< type of viewer

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<SimulationViewer> itsSV; // the viewer
  int itsWhatToExport;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
