/*!@file SimulationViewerCompress.C multi-foveated saliency-based
  compression */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerCompress.C 4188 2005-02-13 07:01:15Z itti $
//

#include "SimulationViewerCompress.H"
#include "MathFunctions.H"
#include "ModelManager.H"
#include "Image_ShapeOps.H"   // for rescale()
#include "Pyramid_Ops.H"      // for buildPyrGaussian()
#include "Image_Transforms.H" // for foveate()
#include "Image_DrawOps.H"    // for weightedBlur()
#include "Image_MathOps.H"    // for binaryReverse(), thresholdedMix()
#include "Image_FilterOps.H"  // for lowPass9()
#include "Image_ColorOps.H"

#include "Brain.H"
#include "Retina.H"
#include "VisualCortex.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"
#include "SaliencyMap.H"
#include "TaskRelevanceMap.H"
#include "AttentionGuidanceMap.H"
#include "SaccadeControllers.H"
#include "ShapeEstimator.H"
#include "Channels.H"
#include "VisualCortexEyeMvt.H"

//! Number of random samples to take
#define NRND 1

// ######################################################################
SimulationViewerCompress::
SimulationViewerCompress(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName) :
  SimulationViewer(mgr, descrName, tagName),
  itsNumFoveas(this, "SVCOMPnumFoveas", 3),
  itsSaveTraj(this, "SVsaveTraj", false),
  itsSaveXcombo(this, "SVCOMPsaveXcombo", false),
  itsSaveYcombo(this, "SVCOMPsaveYcombo", false),
  itsSaveXcomboOrig(this, "SVCOMPsaveXcomboOrig", false),
  itsSaveYcomboOrig(this, "SVCOMPsaveYcomboOrig", false),
  itsSaveMegaCombo(this, "SVCOMPsaveMegaCombo", false),
  itsSaveEyeCombo(this, "SVCOMPsaveEyeCombo", false),
  itsDisplayPatch(this, "SVdisplayPatch", true),
  itsDisplayFOA(this, "SVdisplayFOA", true),
  itsDisplayEye(this, "SVCOMPdisplayHumanEye", true),
  itsColorNormal(this, "SVcolorNormal", PixRGB<byte>(255, 255, 0)),
  itsColorEye(this, "SVcolorHumanEye", PixRGB<byte>(128, 255, 255)),
  itsFOAradius(this, "FOAradius", 30),
  itsFoveaRadius(this, "FoveaRadius", 30),
  itsHeadRadius(this, "HeadMarkerRadius", 50),
  itsMultiRetinaDepth(this, "SVCOMPMultiRetinaDepth", 5),
  itsTimeStep(this, "SimulationTimeStep", 0.0001),
  itsCacheSize(this, "SVCOMPcacheSize", 5),
  itsUseTRMmax(this, "SVCOMPuseTRMmax", false),
  itsFoveaSCtype(this, "SVCOMPfoveaSCtype", "Friction"),
  itsOutFname(this, "SVEMoutFname", ""),
  itsEyeTrash(this, "EyeTrackTrash", 0),
  itsIFramePeriod(this, "SVCOMPiframePeriod", 8),
  itsWhatToExport(OPTEXP_NONE),
  itsBrain(NULL),
  itsMultiTraj(),
  itsSC(), itsTime(0.0), itsInputTime(0.0), itsFrame(-1), itsMask(),
  itsIgnoreSC(), itsCurrentMask(), itsOutFile(NULL), itsBlurMask(),
  itsEyeData()
{
  // build an initial set of SCCs:
  buildSCC();

  // disable IOR:
  LINFO("NOTE: disabling IOR");
  manager->setOptionValString("IORtype", "None");
}

// ######################################################################
SimulationViewerCompress::~SimulationViewerCompress()
{ }

// ######################################################################
bool SimulationViewerCompress::setModelParamString(const std::string& name,
                                                   const std::string& value,
                                                   const bool recurse,
                                                   const bool warn)
{
  std::string oldNumFov = itsNumFoveas.getValString();
  std::string oldType = itsFoveaSCtype.getValString();

  // accept the parameter change:
  bool ret = SimulationViewer::setModelParamString(name, value, recurse, warn);

  // change number of foveas?
  if (name.compare(itsNumFoveas.getName()) == 0 &&
      value.compare(oldNumFov) != 0)
    buildSCC(); // rebuild SCCs

  // change type of controller for foveas?
  if (name.compare(itsFoveaSCtype.getName()) == 0 &&
      value.compare(oldType) != 0)
    buildSCC(); // rebuild SCCs

  return ret; 
}

// ######################################################################
void SimulationViewerCompress::buildSCC()
{
  // drop any old SCCs:
  removeAllSubComponents();

  LINFO("Using %d SaccadeControllers of type %s",
        itsNumFoveas.getVal(), itsFoveaSCtype.getVal().c_str());

  // build an array of SCCs and export their options:
  for (int i = 0; i < itsNumFoveas.getVal(); i ++)
    {
      SharedPtr<SaccadeControllerConfigurator>
        scc(new SaccadeControllerConfigurator(*manager));
      // we need to change the tag name so that we won't get confused
      // among our various SCCs:
      char num[10]; sprintf(num, "%d", i);
      scc->setTagName(scc->tagName() + num);
      scc->setDescriptiveName(scc->descriptiveName() + " " + std::string(num));
      scc->exportOptions(itsWhatToExport);
      addSubComponent(scc);

      // let's change its SC type to not what the command-line says,
      // but what we say:
      scc->setModelParamString("SaccadeControllerType",
                               itsFoveaSCtype.getVal(), true, false);
    }
}

// ######################################################################
void SimulationViewerCompress::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // keep track of what to export:
  itsWhatToExport = whatToExport;

  // let's first do our exports. NOTE: here we consider as OPTEXP_CORE
  // the options that are involved in doing a getTraj(), i.e., which
  // are bout our core output, and we consider as OPTEXP_SAVE options
  // that would save other results:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsNumFoveas);
    manager->requestOption(this, itsFOAradius);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsHeadRadius);
    manager->requestOption(this, itsMultiRetinaDepth);
    manager->requestOption(this, itsCacheSize);
    manager->requestOption(this, itsIFramePeriod);
    manager->requestOption(this, itsUseTRMmax);
    manager->requestOption(this, itsFoveaSCtype);
    manager->requestOption(this, itsOutFname);
    manager->requestOption(this, itsEyeTrash);

    manager->requestOption(this, itsSaveTraj);
    manager->requestOption(this, itsSaveXcombo);
    manager->requestOption(this, itsSaveYcombo);
    manager->requestOption(this, itsSaveXcomboOrig);
    manager->requestOption(this, itsSaveYcomboOrig);
    manager->requestOption(this, itsSaveMegaCombo);
    manager->requestOption(this, itsSaveEyeCombo);
    manager->requestOption(this, itsDisplayPatch);
    manager->requestOption(this, itsDisplayFOA);
    manager->requestOption(this, itsDisplayEye);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimulationViewerCompress::setBrain(Brain* brain)
{ itsBrain = brain; }

// ######################################################################
void SimulationViewerCompress::start1()
{
  itsSC.clear(); itsEyeData.clear();
  itsMask.setMaxSize(itsCacheSize.getVal());

  // setup shortcuts to our configured SCs:
  for (uint i = 0; i < numSubComp(); i ++) {
    SharedPtr<SaccadeControllerConfigurator> scc;
    scc.dynCastFrom(subComponent(i));
    itsSC.push_back(scc->getSC());
    itsIgnoreSC.push_back(false);
  }
  if (itsBrain == NULL) LFATAL("I need a brain!");

  // open output file if any:
  if (itsOutFname.getVal().empty() == false) {
    itsOutFile = fopen(itsOutFname.getVal().c_str(), "w");
    if (itsOutFile == NULL)
      PLFATAL("Cannot write '%s'", itsOutFname.getVal().c_str());

    // fill in a number of trashed samples, so that we have the same
    // number of output samples as we had of eye tracker samples:
    for (int i = 0; i < itsEyeTrash.getVal(); i ++)
      fprintf(itsOutFile, "-\n");
  }
}

// ######################################################################
void SimulationViewerCompress::stop1()
{
  if (itsOutFile) { fclose(itsOutFile); itsOutFile = NULL; }
}

// ######################################################################
void SimulationViewerCompress::newInputReady()
{  }

// ######################################################################
void SimulationViewerCompress::newInputFrame(const
                                             Image< PixRGB<byte> >& input,
                                             const double t)
{
  itsInputTime = t; // keep track of time of last input
  itsTime = t; // update our time
  itsFrame ++; // keep track of frame number, to decide whether I-frame

  //  get a foveation pyramid ready:
  itsMultiTraj = buildPyrGaussian(input, itsMultiRetinaDepth.getVal(), 9);
}

// ######################################################################
void SimulationViewerCompress::channelUpdate(SharedPtr<VisualCortex>& vcx)
{  }

// ######################################################################
void SimulationViewerCompress::evolve(const double t)
{
  // update our time:
  itsTime = t;
  
  // we don't evolve our SCs here. Indeed, the timing of this viewer
  // is a bit hacked: we want to wait until the end of the input
  // frame, so that the saliency map and other Brain internals have
  // been evolved for teh duration of the current frame. Then we want
  // to pick the salient locations and figure out their features, so
  // that we can assign our SCs. Only then do we want to evolve our
  // SCs. So we evolve the SCs in getTraj() rather than here. If we
  // were to assign the SCs in newInputFrame(), the salmap would not
  // have been evolved for that new frame yet, and hence would
  // highlight the salient objects of the previous frame; if those
  // objects just moved, then their features will not match well, and
  // tracking will not work well.
}

// ######################################################################
void SimulationViewerCompress::attentionShift(const WTAwinner& win,
                                              const Image<byte>& foaMask)
{  }

// ######################################################################
void SimulationViewerCompress::eyeMovement(const Point2DT& win,
                                           const Image<byte>& foveaMask)
{
  // all the analysis will be done in getTraj(), so that we don't have
  // to recompute the whole blur mask at every eye movement
  // sample. Hence, here we just queue up the eye positions received:
  if (win.p.i != -1 && win.p.j != -1) itsEyeData.push_back(win.p);
}

// ######################################################################
void SimulationViewerCompress::headMovement(const Point2DT& win,
                                    const Image<byte>& headMask)
{  }

// ######################################################################
void SimulationViewerCompress::eyeSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerCompress::headSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerCompress::eyeBlink(const bool inBlink)
{  }

// ######################################################################
Image< PixRGB<byte> > SimulationViewerCompress::getTraj(const double t)
{
  itsTime = t; // update our time
  Dims dims = itsMultiTraj[0].getDims(); // input image dims

  // let's get the current normalized (values in 0..255) saliency map:
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  if (itsBrain->getAGM().get() == NULL) LFATAL("I need an AGM!");
  if (itsBrain->getTRM().get() == NULL) LFATAL("I need a TRM!");
  if (itsBrain->getRET().get() == NULL) LFATAL("I need a Retina!");

  Image<float> sm;
  if (itsUseTRMmax.getVal()) sm = itsBrain->getSM()->getV(true);// raw saliency
  else sm = itsBrain->getAGM()->getV(true);  // already weighted by relevance

  // find the top itsNumFoveas salient locations; to this end, we will
  // find the max in sm, then use the FOAradius to draw a disk at the
  // top location, then loop as many times as we have foveas:
  Image<float> smf = rescale(sm, dims);

  // in this function, we merge two behaviors: if we have
  // subcomponents (i.e., SaccadeControllers), then we will work in a
  // mode where we have a bunch of foveas moving around. Otherwise, we
  // will use the saliency map as a continuous modulator of blur:
  Image<byte> msk;
  if (itsSC.size()) msk = getMaskSC(smf); else msk = getMaskSM(smf);

  // add this mask to our sliding average:
  if (itsCacheSize.getVal() > 0)  // using a sliding average cache
    {
      itsMask.push_back(msk);
      
      // are we on an I-frame? If so, update our itsCurrentMask, and use
      // it. Otherwise, use the current contents of itsCurrentMask:
      if (itsFrame % itsIFramePeriod.getVal() == 0)
        itsCurrentMask = itsMask.mean();
    }
  else
    itsCurrentMask = msk;  // just using instantaneous mask

  // update blur mask using TRM if needed, otherwise it's just itsCurrentMask:
  if (itsUseTRMmax.getVal() && itsBlurMask.initialized()) {
    // if a location is rapidly changing (high TRM value), we use the
    // value of itsCurrentMask for our blur; otherwise, we take the
    // min between itsCurrentMask and our accumulated itsBlurMask:
    Image<byte> minMask = takeMin(itsBlurMask, itsCurrentMask);
    Image<float> trm = rescale(itsBrain->getTRM()->getV(false), dims);
    itsBlurMask = thresholdedMix(trm, 0.99F, minMask, itsCurrentMask);
  } else itsBlurMask = itsCurrentMask;

  // we use the mean of our mask sliding average for a weighted blur:
  Image< PixRGB<byte> > traj;
  if (itsSaveEyeCombo.getVal())
    traj = itsMultiTraj[0]; // just use the retina
  else
    traj = weightedBlur(itsBlurMask, itsMultiTraj); // weighted blur

  // draw a patch at center of each fovea?
  if (itsDisplayPatch.getVal())
    {
      // select a drawing color & size:
      PixRGB<byte> col(itsColorNormal.getVal()); int psiz = 4 + 2*itsSC.size();
      PixRGB<byte> bk(0, 0, 0);

      // draw a patch at current position of each fovea:
      for (uint i = 0; i < itsSC.size(); i ++)
        if (itsIgnoreSC[i] == false) {
          traj.drawPatch(itsSC[i]->getPreviousDecision(0).p, psiz-i*2+2, bk);
          traj.drawPatch(itsSC[i]->getPreviousDecision(0).p, psiz-i*2, col);
        }
    }

  // draw FOA outlines?
  if (itsDisplayFOA.getVal())
    {
      // select a drawing color & size:
      PixRGB<byte> col(itsColorNormal.getVal()); int thick = 3;
      
      Image<byte> om(itsCurrentMask);
      om = binaryReverse(om, byte(255));
      om.lowThresh(220, 0); // get the objects
      om = contour2D8(om);  // compute binary 8-connected contour image
      int w = traj.getWidth(), h = traj.getHeight();
      Point2D ppp;
      for (ppp.j = 0; ppp.j < h; ppp.j ++)
        for (ppp.i = 0; ppp.i < w; ppp.i ++)
          if (om.getVal(ppp.i, ppp.j))  // got a contour point -> draw here
            traj.drawDisk(ppp, thick, col);  // small disk for each point
    }

  // prepare a full-size color version of the SM for our various markings:
  Image< PixRGB<byte> > colorsm = toRGB(Image<byte>(rescale(sm, dims)));

  // get the raw, unfoveated input image and paste it into an image
  // with our dims:
  Image< PixRGB<byte> > rawinp(dims, NO_INIT); rawinp.clear(PixRGB<byte>(64));
  Image< PixRGB<byte> > rawinp2 = itsBrain->getRawInput();
  Point2D rawinpoff((rawinp.getWidth() - rawinp2.getWidth())/2,
                    (rawinp.getHeight() - rawinp2.getHeight())/2);
  rawinp.paste(rawinp2, rawinpoff);

  // do we want to compare to human eye movement data?
  if (itsOutFile)
    {
      // compute average blur for this frame:
      byte mi, ma, avg; itsCurrentMask.getMinMaxAvg(mi, ma, avg);
      
      // get the raw SM:
      Image<float> rawsm = itsBrain->getSM()->getV(false);

      // get the map level to scale things down:
      LevelSpec lspec; itsBrain->getModelParamVal("LevelSpec", lspec);
      int sml = lspec.mapLevel();
      
      // sample in a disk of radius itsFoveaRadius:
      int rad = itsFoveaRadius.getVal();
      int radsm = rad >> sml;
      
      // let's get the raw saliency map and a vector of all our
      // conspicuity maps and of their min/max/avg:
      std::vector< Image<float> > cmap;
      std::vector<float> cmi, cma, cav;
      SharedPtr<VisualCortex> vc = itsBrain->getVC();
      
      // find out a window to use for our random values, which is
      // important in case we are doing input shifting with
      // --shift-input and apply a field of view with
      // --input-fov. In these cases, we want to take random samples
      // only withing the actual display area:
      const int border = itsFoveaRadius.getVal()*2;  // eliminate borders
      Rectangle r = itsBrain->getRET()->getRawInputRectangle(border);
      //colorsm.drawRect(r, PixRGB<byte>(0, 255, 0), 2);
      //traj.drawRect(r, PixRGB<byte>(0, 255, 0), 2);
      
      // get a version of the rectangle scaled to SM dims:
      Rectangle rsm(r.top() >> sml, r.left() >> sml,
                    r.bottom() >> sml, r.right() >> sml);
      rsm.fitToDims(rawsm.getDims());

      // let's get the raw saliency map and a vector of all our
      // conspicuity maps and of their min/max/avg:
      Image<float> cropsm = crop(rawsm, rsm);
      float rawsmmi, rawsmma, rawsmav;
      cropsm.getMinMaxAvg(rawsmmi, rawsmma, rawsmav);

      // let's get started with our trivial model, just looking at
      // intensity center-surround contrast:
      if (vc->hasChannel("intensity")) {
        SharedPtr<IntensityChannel> ichan;
        ichan.dynCastFrom(vc->channelPtr("intensity"));
        if (ichan.get() == NULL) LFATAL("I need an intensity channel");
        Image<float> ics =
          downSize(ichan->centerSurround(2, 5), rawsm.getDims());
        Image<float> cropmap = crop(ics, rsm);
        float ccmi, ccma, ccav;
        cropmap.getMinMaxAvg(ccmi, ccma, ccav);
        cmi.push_back(ccmi); cma.push_back(ccma); cav.push_back(ccav);
        cmap.push_back(cropmap);
      }
      
      // now for all the channel conspicuity maps:
      for (uint ii = 0; ii < vc->numChannels(); ii ++) {
        Image<float> cm = vc->channel(ii).getOutput();
        Image<float> cropmap = crop(cm, rsm);
        float ccmi, ccma, ccav;
        cropmap.getMinMaxAvg(ccmi, ccma, ccav);
        cmi.push_back(ccmi); cma.push_back(ccma); cav.push_back(ccav);
        cmap.push_back(cropmap);
      }

      // loop over all fixations that happened during current frame:
      while(itsEyeData.size()) {
        // pick a random location to get blur there:
        Point2D rnd(randomUpTo(itsCurrentMask.getWidth() - 1),
                    randomUpTo(itsCurrentMask.getHeight() - 1));
        
        // get next eye fixation:
        Point2D eye = itsEyeData.front(); itsEyeData.pop_front();
        eye.clampToDims(itsCurrentMask.getDims());

        // also scale down eye coords to sm level:
        Point2D eyesm(eye.i >> sml, eye.j >> sml);
        // finally shift eyesm to reflect our crops of the sm and cmaps:
        eyesm.i -= rsm.left(); eyesm.j -= rsm.top();
        eyesm.clampToDims(cropsm.getDims());

        // do we want to draw it?
        if (itsDisplayEye.getVal())
          {
            // select a drawing color & size:
            PixRGB<byte> col(itsColorEye.getVal()); int psiz = 5;
            PixRGB<byte> bk(0, 0, 0);
            traj.drawPatch(eye, psiz + 2, bk);
            traj.drawPatch(eye, psiz, col);
            colorsm.drawPatch(eye, psiz + 2, bk);
            colorsm.drawPatch(eye, psiz, col);

            Point2D rieye = itsBrain->getRET()->retinalToRaw(eye);
            rieye.i += rawinpoff.i;
            rieye.j += rawinpoff.j;
            rawinp.drawPatch(rieye, psiz + 2, bk);
            rawinp.drawPatch(rieye, psiz, col);
          }
        
        // save blur at eye, average, rnd, raw saliency and vector of
        // features:
        float sateye = this->getSample(cropsm, eyesm, radsm);
        float satrnd = this->getRandomSample(cropsm, radsm, NRND);
        
        fprintf(itsOutFile, "%d %d %d    %g %g %g %g", 
                itsCurrentMask.getVal(eye),    // blur at eye
                avg,                           // average blur over frame
                itsCurrentMask.getVal(rnd),    // blur at random location
                sateye,                        // raw saliency value
                rawsmma,                       // max raw saliency
                rawsmav,                       // avg raw saliency
                satrnd);                       // raw saliency at rnd loc
        //LINFO("BLUR: eye=%d avg=%d rnd=%d eye at (%d, %d)",
        //      int(itsCurrentMask.getVal(eye)), int(avg),
        //      int(itsCurrentMask.getVal(rnd)), eye.i, eye.j);
        LINFO("RAWSM: eye=%.2g max=%.2g avg=%.2g rnd=%.2g",
              sateye, rawsmma, rawsmav, satrnd);
        
        for (uint ii = 0; ii < cmap.size(); ii ++) {
          float cateye = this->getSample(cmap[ii], eyesm, radsm);
          float catrnd = this->getRandomSample(cmap[ii], radsm, NRND);

          fprintf(itsOutFile, "    %g %g %g %g",  // for each feature:
                  cateye,    // value at eye
                  cma[ii],   // max value over map
                  cav[ii],   // average value over map
                  catrnd);   // value at rnd loc
          //LINFO("CMAP(%s):  eye=%.2g max=%.2g avg=%.2g rnd=%.2g",
          //      vc->channel(ii).tagName().c_str(), cateye,
          //      cma[ii], cav[ii], catrnd);
        }
        fprintf(itsOutFile, "\n");
      }
    }
  
  // do we want a mega combo instead of the plain blurred image?
  if (itsSaveMegaCombo.getVal())
    {
      Image< PixRGB<byte> > ret =
        concatX(colGreyCombo(itsMultiTraj[0], rescale(sm, dims), false),
                colGreyCombo(traj, itsCurrentMask, false));
      ret.drawGrid(dims.w()-1, dims.h()-1, 3, 3, PixRGB<byte>(128, 128, 128));
      return ret;
    }

  // do we want an eye combo?
  if (itsSaveEyeCombo.getVal())
    {
      Image< PixRGB<byte> > ret =
        concatX(concatX(rawinp, traj), colorsm);
      ret.drawLine(Point2D(dims.w()-1, 0), Point2D(dims.w()-1, dims.h()-1), 
                   PixRGB<byte>(255,255,0), 3);
      ret.drawLine(Point2D(dims.w()*2-1, 0), Point2D(dims.w()*2-1,dims.h()-1), 
                   PixRGB<byte>(255,255,0), 3);

      // make sure the size is reasonable...
      while(ret.getWidth() > 1024) ret = decXY(lowPass3(ret));

      return ret;
    }
  
  // do we want a combo instead of the plain blurred image?
  if (itsSaveXcombo.getVal() || itsSaveYcombo.getVal())
    return colGreyCombo(traj, itsCurrentMask, itsSaveXcombo.getVal());

  // do we want to save a combo of original + blurred?
  if (itsSaveXcomboOrig.getVal()) return concatX(itsMultiTraj[0], traj);
  if (itsSaveYcomboOrig.getVal()) return concatY(itsMultiTraj[0], traj);

  // otherwise return the blurred image:
  return traj;
}

// ######################################################################
Image<byte> SimulationViewerCompress::getMaskSM(const Image<float>& smf)
{
  // let's start by smoothing the interpolated salmap a bit:
  Image<float> maskf = lowPass9(smf);
  
  // let's squash the SM a bit. We downplay values below average and
  // give more range to those above average (with possible saturation):
  float mi, ma, av; maskf.getMinMaxAvg(mi, ma, av);
  maskf = squash(maskf, mi, mi, 0.5F*(av-mi), 0.55F*(av-mi), ma, ma);
  
  // make a blurring mask:
  Image<byte> mask = binaryReverse(maskf, 255.0F);  // will clamp to 0..255

  return mask;
}

// ######################################################################
Image<byte> SimulationViewerCompress::getMaskSC(const Image<float>& smf)
{
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  SharedPtr<VisualCortex> vc = itsBrain->getVC();
  if (vc.get() == NULL) LFATAL("I need a VisualCortex!");
  SharedPtr<ShapeEstimator> se = itsBrain->getSE();
  if (se.get() == NULL) LFATAL("I need a ShapeEstimator!");

  // violently reset the se:
  se->reset();

  // find the top salient locations; to this end, we will find the max
  // in sm, then use the FOAradius to draw a disk at the top location,
  // then loop. We will extract more locations than we have foveas, so
  // that we are robust to slight changes in saliency ordering:
  std::vector<Point2DT> topsal;  // will store saliency in 't' field
  Image<float> smff = smf;  // get a copy we can modify
  for (uint i = 0; i < itsSC.size() + 4; i ++) {
    // find max:
    Point2D p; float sal; smff.findMax(p, sal);
    
    // store coords & saliency:
    topsal.push_back(Point2DT(p, double(sal)));
    
    // get object shape at that location, or revert to a disk of no object:
    se->compute(p);
    Image<byte> objmask = se->getSmoothMask() * 255.0F;
    if (objmask.initialized() == false) {
      objmask.resize(smff.getDims(), true);
      objmask.drawDisk(p, itsFOAradius.getVal(), 255);
    }

    // inhibit the sm by the object shape:
    smff.setValMask(objmask, 0.0F);
  }

  // if this is our first time (itsFeatures is empty), just assign an
  // SC to each of the top salient locations:
  if (itsFeatures.empty())
    for (uint i = 0; i < itsSC.size(); i ++)
      {
        // feed the SC:
        itsSC[i]->setPercept(WTAwinner(topsal[i].p, itsTime, topsal[i].t,
                                       topsal[i].t, false));
        
        // keep track of the features each SC is tracking:
        std::vector<double> f; vc->getFeatures(topsal[i].p, f);
        itsFeatures.push_back(f);
        
        LINFO("Initializing SC[%d] to (%d,%d)", i,
              topsal[i].p.i, topsal[i].p.j);
      }
  else
    {
      // let's get an idea of which features are important for
      // differentiating between our current foveas, and of what
      // their range is:
      std::vector<double> minf, maxf; uint nf = itsFeatures[0].size();
      for (uint i = 0; i < nf; i ++)
        { minf.push_back(1.0e50); maxf.push_back(-1.0e50); }
      
      for (uint fov = 0; fov < itsSC.size(); fov ++)
        for (uint i = 0; i < nf; i ++)
          {
            if (itsFeatures[fov][i] < minf[i]) minf[i] = itsFeatures[fov][i];
            if (itsFeatures[fov][i] > maxf[i]) maxf[i] = itsFeatures[fov][i];
          }
      
      // solve correspondence problem: Score each salient location with
      // respect to each SC, based on distance and feature similarity in a
      // neighborhood:
      Image<float> score(topsal.size(), itsSC.size(), NO_INIT);
      for (uint i = 0; i < topsal.size(); i ++)
        for (uint j = 0; j < itsSC.size(); j ++)
          {
            // how well does salient location 'i' score with fovea
            // 'j'? First let's look at distance; what counts then is
            // the distance to the target (=percept) of the SC, not to
            // where the SC currently is:
            Point2D pi = topsal[i].p;
            Point2D pj = itsSC[j]->getPreviousPercept(0).p;
            float dist = pi.distance(pj);
            
            // a distance of up to twice our FOA radius yields no penalty:
            if (dist < 2.0F * float(itsFOAradius.getVal())) dist = 0.0F;
            
            // normalize the distance to 0..1 range:
            dist /= sqrt(smf.getWidth() * smf.getWidth() +
                         smf.getHeight() * smf.getHeight());
            
            // now look at feature similarity; to this end, we explore
            // a neighborhood of the salient location and see whether
            // we can find the features that the SC is interested in:
            std::vector<double> scf = itsFeatures[j];
            double fdist = 1.0e50;
            int ci = pi.i; if (ci & 1) ci --;
            int cj = pi.j; if (cj & 1) cj --;
            for (int jj = cj - 10; jj <= cj + 10; jj += 2)
              for (int ii = ci - 10; ii <= ci + 10; ii += 2)
                {
                  Point2D p(ii, jj);
                  if (smf.coordsOk(p))
                    {
                      // get a vector of features:
                      std::vector<double> ff; vc->getFeatures(p, ff);
                      
                      // compute feature distance: get feature
                      // difference and normalize by feature range
                      // if range not too small:
                      double d = 0.0; int numf = 0;
                      for (uint k = 0; k < nf; k ++)
                        if (maxf[k] > minf[k] + 1.0)
                          {
                            double dd = (ff[k]-scf[k]) / (maxf[k]-minf[k]);
                            
                            // accumulate compound feature distance:
                            d += dd * dd; numf ++;
                          }
                      
                      // compute normalized weighted feature distance:
                      if (numf) d = sqrt(d / double(numf));
                      
                      // if distance better than what we had, update:
                      if (d < fdist) fdist = d;
                    }
                }
            
            // point distance and feature distance both contribute
            // to score. In addition, we add here a penalty for
            // changing the ordering of the SCs. Best possible score
            // is zero and all scores are negative, growing larger
            // negatively as they get worse:
            float sco = -
              float(dist * 100.0) -
              float(fdist * 0.5) -
              10.0F * fabs(float(i)-float(j)) +
              100.0F * topsal[i].t / topsal[0].t;
            
            score.setVal(i, j, sco);
            LINFO("[topsal(%d)(%d,%d), SC(%d)(%d,%d)]: dist=%f fdist=%f "
                  "score=%f", i, pi.i, pi.j, j, pj.i, pj.j,
                  dist, fdist, sco);
          }
      
      // find the best score and assign the corresponding salient
      // location to the corresponding SC; then kill that SC in the
      // score map and loop until all SCs have been assigned:
      for (uint i = 0; i < itsSC.size(); i ++)
        {
          Point2D best; float val; score.findMax(best, val);
          int bi = best.i;  // salient location
          int bj = best.j;  // saccade controller
 
          // set new percept to the winning SC:
          itsSC[bj]->setPercept(WTAwinner(topsal[bi].p, itsTime,
                                          topsal[bi].t, topsal[bi].t,
                                          false));
          // also assign new feature vector to that SC; even though
          // the best score may have been achieved for a neighbor of
          // the topsal location, we use as feature vector the one
          // from the topsal location:
          std::vector<double> fvec; vc->getFeatures(topsal[bi].p, fvec);
          itsFeatures[bj] = fvec;
          
          // done with this SC; let's make sure we will not pick it up again:
          for (int k = 0; k < score.getWidth(); k ++)
            score.setVal(k, bj, -1.0e30F);
          
          // also make sure we will not pick up that salient loc again:
          for (int k = 0; k < score.getHeight(); k ++)
            score.setVal(bi, k, -1.0e30F);
          
          LINFO("Assigned topsal[%d](%d,%d) to SC[%d](%d,%d)", bi,
                topsal[bi].p.i, topsal[bi].p.j, bj,
                itsSC[bj]->getPreviousPercept(1).p.i,
                itsSC[bj]->getPreviousPercept(1).p.j);
        }
    }

  // evolve our SCs in batch. See comments in evolve() for why we do that here:
  double t = itsInputTime;
  while(t < itsTime) {
    for (uint i = 0; i < itsSC.size(); i ++) itsSC[i]->evolve(t);
    t += itsTimeStep.getVal();
  }

  // run getDecision() on our SaccadeControllers:
  for (uint i = 0; i < itsSC.size(); i ++) itsSC[i]->getDecision(itsTime);

  // create a mask with the object shapes at each SC:
  se->reset();  // violently reset the SE
  Image<float> maskf(smf.getDims(), ZEROS);
  for (uint i = 0; i < itsSC.size(); i ++)
    {
      // if we terminate this iteration early we will ignore this SC:
      itsIgnoreSC[i] = true;

      // get the current fixation for this SC:
      Point2DT p = itsSC[i]->getPreviousDecision(0);
      if (p.p.i < 0) continue;

      // if salience was very low, don't bother using this SC:
      if (smf.getVal(itsSC[i]->getPreviousPercept(0).p) < topsal[0].t * 0.05) {
        LINFO("Ignoring SC[%d] because salience too low", i);
        continue;
      }

      // otherwise segment the object and mark it:
      se->compute(p.p);
      Image<float> objf = se->getSmoothMask();
      if (objf.initialized()) maskf = takeMax(maskf, objf);
      else maskf.drawDisk(p.p, itsFOAradius.getVal(), 1.0F);

      // ok, we won't ignore this SC:
      itsIgnoreSC[i] = false;
    }
  // binarize the object mask:
  maskf.lowThresh(0.5F, 0.0F);

  // create a distance map from the mask (this code similar to that in
  // foveate() of Image_Transforms.C):
  int maxdist = std::max(smf.getWidth(), smf.getHeight()) * 2;
  float scalefac = float(maxdist) / 255.0f;
  maskf = chamfer34(maskf, float(maxdist)) / scalefac;

  // if modulator does not contain any point at zero (inside object),
  // that means that the mask was empty, which is the case at the
  // beginning of a simulation. Set it to some intermediary value to
  // provide a uniform medium blur; otherwise, squash it:
  float mi, ma, av; maskf.getMinMaxAvg(mi, ma, av);
  if (mi > 0.0F) maskf /= 3.0F;
  else maskf = squash(maskf, mi, mi, 0.5F*(av-mi), 0.6F*(av-mi), ma, ma);

  // return byte version of the mask:
  Image<byte> mask = maskf;  // will clamp as necessary
  return mask;
}

// ######################################################################
void SimulationViewerCompress::saveResults(const
                                           SharedPtr<OutputFrameSeries>& ofs)
{
  // save results?
  if (itsSaveTraj.getVal() || itsSaveXcombo.getVal() ||
      itsSaveYcombo.getVal() || itsSaveXcomboOrig.getVal() ||
      itsSaveYcomboOrig.getVal() || itsSaveMegaCombo.getVal() ||
      itsSaveEyeCombo.getVal())
    ofs->writeRGB(getTraj(itsTime), Auto, -1, "T");
}

// ######################################################################
float SimulationViewerCompress::getSample(const Image<float>& smap,
                                          const Point2D& p,
                                          const int radius) const
{
  // ### disk version:
  Image<float> fov(smap.getDims(), ZEROS);
  fov.drawDisk(p, radius, 1.0F);
  fov *= smap;  // max smap by the disk
  float junk, salience;
  fov.getMinMax(junk, salience);
  return salience;

  // ### point version:
  //  return smap.getVal(p);
}

// ######################################################################
float SimulationViewerCompress::
getRandomSample(const Image<float>& smap, const int radius, const int n) const
{
  float rndval = 0.0f;
  for (int i = 0; i < n; i ++)
    {
      Point2D rndsm(randomUpTo(smap.getWidth() - 1),
                    randomUpTo(smap.getHeight() - 1));
      rndval += getSample(smap, rndsm, radius);
    }
  return rndval / n;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
