/*!@file SimulationViewerCompress.H multi-foveated saliency-based
  compression */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerCompress.H 3116 2004-02-15 22:40:13Z itti $
//

#ifndef SIMULATIONVIEWERCOMPRESS_H_DEFINED
#define SIMULATIONVIEWERCOMPRESS_H_DEFINED

#include "SimulationViewer.H"
#include "ImageSet.H"
#include "ImageCache.H"
class SaccadeController;

//! Do multi-foveated saliency-based image compression
/*! This viewer will consider the top N salient locations and apply
  blurring to the image that increases in strength (sigma) as we get
  farther away from any of the top N most salient locations. The
  resulting trajectory image hence will be crisp at the top N
  locations and increasingly blurred away from those. Compressing the
  output of this viewer, e.g., using MPEG, yields smaller file size
  while (hopefully) preserving high quality at the important image
  locations. This viewer relies on a collection of N
  SaccadeControllers to track the N hotspots; you can pick which type
  of controller to use and configure their parameters as usual, via
  the command line. Note that in programs like ezvision that use an
  StdBrain, the StdBrain also contains a SaccadeController; but here
  we will explicitly drop it, as we won;t be using Brain's controller,
  since we will decide on the N hotspots here based onthe raw saliency
  map. We will also drop Brain's ShapeEstimator and force Brain's IOR
  type to none, so that we have a clean saliency map here to work
  with. */
class SimulationViewerCompress : public SimulationViewer {
public:
  // ######################################################################
  /*! @name Constructors and destructors */
  //@{

  //! Constructor. See ModelComponent.H.
  SimulationViewerCompress(ModelManager& mgr,
                           const std::string& descrName =
                           "Multi-Foveated Compression Simulation Viewer",
                           const std::string& tagName =
                           "SimulationViewerCompress");

  //! Destructor
  virtual ~SimulationViewerCompress();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SaccadeController once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Set our brain
  virtual void setBrain(Brain* brain);

  //@}

  // ######################################################################
  /*! @name Events that will trigger new visualizations */
  //@{

  //! Ready to switch to next input frame
  virtual void newInputReady();

  //! Got a new input frame at the specified (current) time
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void newInputFrame(const Image< PixRGB<byte> >& input,
                             const double t);

  //! Update with availability of new channel drawings
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void channelUpdate(SharedPtr<VisualCortex>& vcx);

  //! Called every simulation time step (time is in seconds)
  /*! In SimulationViewerCompress, we evolve the positions of our
    internal foveas. */
  virtual void evolve(const double t);

  //! Update with new covert attention position
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void attentionShift(const WTAwinner& win,
                              const Image<byte>& foaMask);

  //! Update with new eye position
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void eyeMovement(const Point2DT& curreye,
                           const Image<byte>& foveaMask);

  //! Update with new head position
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void headMovement(const Point2DT& currhead,
                            const Image<byte>& headMask);

  //! Update eye saccade status
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void eyeSaccade(const bool inSaccade);

  //! Update head saccade status
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void headSaccade(const bool inSaccade);

  //! Update eye blink status
  /*! In SimulationViewerCompress, this is a no-op. */
  virtual void eyeBlink(const bool inBlink);

  //@}

  // ######################################################################
  /*! @name Output of our current internal drawings */
  //@{

  //! Get the attention/eye/head trajectory image
  virtual Image< PixRGB<byte> > getTraj(const double t);

  //! Save our various results
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

protected:
  ModelParam<int> itsNumFoveas;        //!< number of foveas
  ModelParam<bool> itsSaveTraj;        //!< save trajectory?
  ModelParam<bool> itsSaveXcombo;      //!< save trajectory+mask?
  ModelParam<bool> itsSaveYcombo;      //!< save trajectory+mask?
  ModelParam<bool> itsSaveXcomboOrig;  //!< save original+trajectory?
  ModelParam<bool> itsSaveYcomboOrig;  //!< save original+trajectory?
  ModelParam<bool> itsSaveMegaCombo;   //!< save mega combo?
  ModelParam<bool> itsSaveEyeCombo;    //!< save eye combo?
  ModelParam<bool> itsDisplayPatch;    //!< draw patches
  ModelParam<bool> itsDisplayFOA;      //!< draw objetc outlines
  ModelParam<bool> itsDisplayEye;      //!< draw human eye movements
  ModelParam< PixRGB<byte> > itsColorNormal; //!< patch color
  ModelParam< PixRGB<byte> > itsColorEye;  //!< patch color for human eye mvts
  ModelParam<int> itsFOAradius;        //!< radius of FOA
  ModelParam<int> itsFoveaRadius;      //!< radius of Fovea
  ModelParam<int> itsHeadRadius;       //!< head radius
  ModelParam<int> itsMultiRetinaDepth; //!< depth of blur pyramid
  ModelParam<double> itsTimeStep;      //!< simulation time step
  ModelParam<int> itsCacheSize;        //!< size of our mask cache
  ModelParam<bool> itsUseTRMmax;       //!< use TRM to take max in cache
  ModelParam<std::string> itsFoveaSCtype; //!< type of SC for foveas
  ModelParam<std::string> itsOutFname;     //!< Name of output file
  ModelParam<int> itsEyeTrash;         //!< how many initial eye samples trash

  //! This parameter is the period (in frames) for foveation mask to change
  /*! The idea here is to see whether keeping the foveation mask
    stable for predicted frames and changing it only for intra-coded
    frames may improve the compression ratio when using MPEG-1
    compression. Have a look at
    http://www.disctronics.co.uk/technology/video/video_mpeg.htm if
    you are not familiar with those various types of frames. In
    particular (quoted from that web site):

    "I-frames (Intra coded frames) use DCT encoding only to compress a
    single frame without reference to any other frame in the sequence.
    [...]
    P-frames (Predicted frames) are coded as differences from the last
    I or P frame. The new P-frame is first predicted by taking the
    last I or P frame and 'predicting' the values of each new
    pixel. P-frames use Motion Prediction and DCT encoding. As a
    result P-frames will give a compression ratio better than I-frames
    but depending on the amount of motion present. The differences
    between the predicted and actual values are encoded. [...]
    B-frames (Bidirectional frames) are coded as differences from the
    last or next I or P frame. B-frames use prediction as for P-frames
    but for each block either the previous I or P frame is used or the
    next I or P frame. [...]"

    So, in our case, changing the foveation mask on a P or B frame may
    yield lots of prediction errors, that can be reduced if we force
    the mask to be only allowed to change on I frames. Here we assume
    that the first frame is an I-frame. */
  ModelParam<int> itsIFramePeriod;

  //! Get started and disable any SC the Brain may want to use
  virtual void start1();

  //! get stopped
  virtual void stop1();

  //! Intercept people changing our number of foveas
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);
private:
  int itsWhatToExport;                   // keep track of what to export
  Brain *itsBrain;                       // hook to brain or null
  ImageSet< PixRGB<byte> > itsMultiTraj; // used to foveate traj
  std::vector< SharedPtr<SaccadeController> > itsSC; // our SCs
  double itsTime;                        // keep track of internal time
  double itsInputTime;                   // time of last input
  int itsFrame;                          // keep track of frame number

  void buildSCC(); // build our SaccadeControllerConfigurators

  float getSample(const Image<float>& smap, const Point2D& p,
                  const int radius) const;
  float getRandomSample(const Image<float>& smap,
                        const int radius, const int n) const;

  Image<byte> getMaskSC(const Image<float>& smf); // create mask from SCs
  Image<byte> getMaskSM(const Image<float>& smf); // create mask from SM

  // keep track of features being tracked by each SC:
  std::vector< std::vector<double> > itsFeatures;

  // use a sliding image cache for our foveation masks:
  ImageCacheAvg<byte> itsMask;

  // ignore SCs that have too low salience:
  std::vector<bool> itsIgnoreSC;

  Image<byte> itsCurrentMask;  // last I-frame mask computed

  // eye movement comparison stuff:
  FILE *itsOutFile;

  Image<byte> itsBlurMask;  // final current blur mask to use
  std::deque<Point2D> itsEyeData;  // queued-up eye movement data over 1 frame
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
