/*!@file SimulationViewerEyeMvt.C comparison between saliency and
  human eye movements */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerEyeMvt.C 4125 2005-01-17 02:22:19Z itti $
//

#include "SimulationViewerEyeMvt.H"
#include "ModelManager.H"
#include "Image_ShapeOps.H"    // for rescale()
#include "Image_ColorOps.H"    // for contrastModulate()
#include "Image_DrawOps.H"     // for colGreyCombo()
#include "Image_Transforms.H"  // for segmentObjectClean(), contour2D8()
#include "Image_FilterOps.H"
#include "Image_MathOps.H"     // for scramble()
#include "SaccadeControllers.H"
#include "AttentionGuidanceMap.H"

#include "Brain.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"
#include "VisualCortex.H"
#include "ChannelBase.H"
#include "VisualCortexEyeMvt.H"

// number of random samples to take:
#define NBRND 100

// ######################################################################
SimulationViewerEyeMvt::
SimulationViewerEyeMvt(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName) :
  SimulationViewer(mgr, descrName, tagName),
  itsSaveMegaCombo(this, "SVEMsaveMegaCombo", false),
  itsColorEye(this, "SVcolorHumanEye", PixRGB<byte>(128, 255, 255)),
  itsFoveaRadius(this, "FoveaRadius", 30),
  itsDelayCacheSize(this, "SVEMdelayCacheSize", 0),
  itsMaxCacheSize(this, "SVEMmaxCacheSize", 120),
  itsSampleAtStart(this, "SVEMsampleAtStart", true),
  itsOutFname(this, "SVEMoutFname", ""),
  itsBrain(NULL), itsTime(0.0), itsInput(), itsTraj(),
  itsOutFile(NULL), itsLastSample(-1, -1), itsLastRadius(0),
  itsLastAmplitude(0.0), itsLastDuration(0.0),
  itsDelayCache(), itsMaxCache(), itsSC(NULL)
{
  // disable IOR:
  LINFO("NOTE: disabling IOR and SE, selecting FixedSaccadeController");
  manager->setOptionValString("IORtype", "None");
  manager->setOptionValString("ShapeEstimatorMode", "None");
  manager->setOptionValString("SaccadeControllerType", "EyeTrack");
}

// ######################################################################
SimulationViewerEyeMvt::~SimulationViewerEyeMvt()
{ }

// ######################################################################
void SimulationViewerEyeMvt::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports. NOTE: here we consider as OPTEXP_CORE
  // the options that are involved in doing a getTraj(), i.e., which
  // are bout our core output, and we consider as OPTEXP_SAVE options
  // that would save other results:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsDelayCacheSize);
    manager->requestOption(this, itsMaxCacheSize);
    manager->requestOption(this, itsOutFname);
    manager->requestOption(this, itsSaveMegaCombo);
    manager->requestOption(this, itsSampleAtStart);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimulationViewerEyeMvt::setBrain(Brain* brain)
{ itsBrain = brain; }

// ######################################################################
void SimulationViewerEyeMvt::start1()
{
  if (itsDelayCacheSize.getVal())
    itsDelayCache.setMaxSize(itsDelayCacheSize.getVal());
  if (itsMaxCacheSize.getVal())
    itsMaxCache.setMaxSize(itsMaxCacheSize.getVal());

  // abort if no output file:
  if (itsOutFname.getVal().empty()) LFATAL("I need an output file!");

  // open output file:
  itsOutFile = fopen(itsOutFname.getVal().c_str(), "w");
  if (itsOutFile == NULL)
    PLFATAL("Cannot write '%s'", itsOutFname.getVal().c_str());

  // make sure we have the right type of SC and get a pointer to it:
  itsSC.dynCastFrom(itsBrain->getSC());
  if (itsSC.get() == NULL) LFATAL("I need an EyeTrackerSaccadeController");
}

// ######################################################################
void SimulationViewerEyeMvt::stop1()
{
  if (itsOutFile) { fclose(itsOutFile); itsOutFile = NULL; }
}

// ######################################################################
void SimulationViewerEyeMvt::newInputReady()
{  }

// ######################################################################
void SimulationViewerEyeMvt::newInputFrame(const
                                           Image< PixRGB<byte> >& input,
                                           const double t)
{
  itsTime = t; // update our time
  itsInput = input;
  itsTraj = input; // keep the frame for later displays
}

// ######################################################################
void SimulationViewerEyeMvt::channelUpdate(SharedPtr<VisualCortex>& vcx)
{  }

// ######################################################################
void SimulationViewerEyeMvt::evolve(const double t)
{ }

// ######################################################################
void SimulationViewerEyeMvt::attentionShift(const WTAwinner& win,
                                            const Image<byte>& foaMask)
{
  // a new object has been selected by covert attention
}

// ######################################################################
void SimulationViewerEyeMvt::eyeMovement(const Point2DT& win,
                                         const Image<byte>& foveaMask)
{
  Dims dims = itsTraj.getDims();

  // let's get the current saliency map, in mV:
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  if (itsBrain->getAGM().get() == NULL) LFATAL("I need an AGM!");
  Image<float> sm = itsBrain->getAGM()->getV(false);

  // if we have a delay cache, we push the SM into it and push
  // what comes out of it into our max-computation cache;
  // otherwise, we directly push the SM into the max-computation
  // cache:
  Image<float> sm2(sm);
  if (itsDelayCacheSize.getVal())
    {
      // let's push the SM into our delay sliding cache:
      itsDelayCache.push_back(sm);
      
      // transfer oldest delay image to our max computation cache:
      if (int(itsDelayCache.size()) >= itsDelayCacheSize.getVal())
        sm2 = itsDelayCache.front();
    }

  // if we have a max cache, let's push the (possibly delayed) sm into it:
  if (itsMaxCacheSize.getVal()) itsMaxCache.push_back(sm2);
      
  // select a drawing color & size:
  PixRGB<byte> col(itsColorEye.getVal()); int psiz = 5;
  if (itsSC->isInSaccade()) { col.setGreen(0); col.setRed(255); }
  if (itsSC->isInBlink()) { col.setBlue(0); }

  PixRGB<byte> bk(0, 0, 0);
  itsTraj.drawPatch(win.p, psiz + 2, bk);
  itsTraj.drawPatch(win.p, psiz, col);
  
  // get the eye movement data:
  Point2D nextTarg; double ampl, durs;
  itsSC->getExtraEyeData(nextTarg, ampl, durs);

  // are we starting a saccade?
  if (ampl > 0.0)
    {
      // check that saccade target is okay:
      if (itsTraj.coordsOk(nextTarg) == false)
        {
          LERROR("Hum, saccade target at (%d, %d)? -- CLAMPING",
                 nextTarg.i, nextTarg.j);
          nextTarg.clampToDims(dims);
        }

      // store these values for later drawings:
      itsLastSample = nextTarg;
      itsLastRadius = itsFoveaRadius.getVal();
      itsLastAmplitude = ampl;
      itsLastDuration = durs;
    }

  if (itsLastAmplitude > 0.0 &&
      (itsSampleAtStart.getVal() || itsSC->isInSaccade() == false))
    {
      // draw in the traj:
      itsTraj.drawCircle(itsLastSample, itsLastRadius,
                         PixRGB<byte>(255, 255, 0), 2);

      // save amplitude and last fixation duration:
      fprintf(itsOutFile, "%g %g   ", itsLastAmplitude, itsLastDuration);

      // get the map level to scale things down:
      LevelSpec lspec; itsBrain->getModelParamVal("LevelSpec", lspec);
      int sml = lspec.mapLevel();

      // get ready to take samples in the saliency map and in all
      // conspicuity maps, at the scale of the saliency map:
      int rad = itsLastRadius >> sml;
      Point2D p(itsLastSample.i >> sml, itsLastSample.j >> sml);
      SharedPtr<VisualCortex> vc = itsBrain->getVC();
      Image<float> smap;
      if (itsMaxCache.size())
        smap = itsMaxCache.front(); // used to be: itsMaxCache.getMax();
      else
        smap = sm;

      // just in case the image has weird dims, make sure the sampling
      // location is in it:
      p.clampToDims(smap.getDims());

      // get the min/max/avg and saliency samples:
      float mi, ma, av; smap.getMinMaxAvg(mi, ma, av);

      // get a sample around the saccade target location:
      float val = smap.getSample(p, rad);

      // save that:
      fprintf(itsOutFile, " %g %g %g %g  ", val, mi, ma, av);
      
     // now get a bunch of random samples:
      for (int k = 0; k < NBRND; k ++)
        {
            Point2D randp(int(smap.getWidth() * drand48()),
                          int(smap.getHeight() * drand48()));
            fprintf(itsOutFile, " %g %d %d",
                    smap.getSample(randp, rad), randp.i, randp.j);
            // fprintf(itsOutFile, " %g",
            //   smap.getSample(Point2D(int(smap.getWidth() * drand48()),
            //                  int(smap.getHeight() * drand48())), rad));
        }

      // if we have a VisualCortexEyeMvt let's also store some stats
      // on human-to-human distances:
      SharedPtr<VisualCortexEyeMvt> vcem;
      vcem.dynCastFrom(vc);
      if (vcem.get()) // yes, it's a VisualCortexEyeMvt
        for (uint i = 0; i < vcem->numHumans(); i ++)
          {
            Point2D p = vcem->getEyePos(i);
            fprintf(itsOutFile, " %g", itsLastSample.distance(p));
          }

      // done:
      fprintf(itsOutFile, "\n"); fflush(itsOutFile);
      itsLastAmplitude = 0.0;
    }
}

// ######################################################################
void SimulationViewerEyeMvt::headMovement(const Point2DT& win,
                                    const Image<byte>& headMask)
{  }

// ######################################################################
void SimulationViewerEyeMvt::eyeSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerEyeMvt::headSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerEyeMvt::eyeBlink(const bool inBlink)
{  }

// ######################################################################
Image< PixRGB<byte> > SimulationViewerEyeMvt::getTraj(const double t)
{
  itsTime = t; // update our time
  Dims dims = itsTraj.getDims();

  // let's get the current saliency map (normalized):
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  if (itsBrain->getAGM().get() == NULL) LFATAL("I need an AGM!");
  Image<float> sm = rescale(itsBrain->getAGM()->getV(true), dims);

  // compute max saliency over our min/max cache:
  Image<float> maxsm;
  if (itsMaxCache.size()) maxsm = rescale(itsMaxCache.getMax(), dims);
  else maxsm = rescale(sm, dims);

  // get color version of SM:
  Image< PixRGB<byte> > smc = toRGB(Image<byte>(sm));

  // draw circle around saliency sample if we just took one:
  if (itsLastRadius && itsLastAmplitude == 0.0) {
      maxsm.drawCircle(itsLastSample, itsLastRadius, 255.0F, 2);
      smc.drawCircle(itsLastSample, itsLastRadius, PixRGB<byte>(255, 255, 0), 2);
      itsLastRadius = 0;
  }

  // return mega combo; we have two formats: if we have a max-cache,
  // it will be a 4-image format, otherwise a 3-image format:
  Image< PixRGB<byte> > ret;
  if (itsMaxCacheSize.getVal())
    {
      ret = concatX(colGreyCombo(itsInput, sm, false),
                    colGreyCombo(itsTraj, maxsm, false));
      ret.drawGrid(dims.w()-1, dims.h()-1, 3, 3, PixRGB<byte>(128, 128, 128));
    }
  else
    {
      /****** old 3-image format:
      ret = concatX(itsInput, concatX(itsTraj, smc));
      ret.drawLine(Point2D(dims.w()-1, 0), Point2D(dims.w()-1, dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
      ret.drawLine(Point2D(dims.w(), 0), Point2D(dims.w(), dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
      ret.drawLine(Point2D(dims.w()*2-1, 0), Point2D(dims.w()*2-1, dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
      ret.drawLine(Point2D(dims.w()*2, 0), Point2D(dims.w()*2, dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
      ********/

      // new 2-image format (no original):
      ret = concatX(itsTraj, smc);
      ret.drawLine(Point2D(dims.w()-1, 0), Point2D(dims.w()-1, dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
      ret.drawLine(Point2D(dims.w(), 0), Point2D(dims.w(), dims.h()-1),
                   PixRGB<byte>(255, 255, 0), 1);
    }
  // make sure image is not unreasonably large:
  while(ret.getWidth() > 1024) ret = decXY(lowPass3(ret));
  return ret;
}

// ######################################################################
void SimulationViewerEyeMvt::saveResults(const
                                         SharedPtr<OutputFrameSeries>& ofs)
{
  // save results?
  if (itsSaveMegaCombo.getVal())
    ofs->writeRGB(getTraj(itsTime), Auto, -1, "T");
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
