/*!@file SimulationViewerEyeMvt.H comparison between saliency and
  human eye movements */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerEyeMvt.H 3321 2004-05-28 00:12:32Z itti $
//

#ifndef SIMULATIONVIEWEREYEMVT_H_DEFINED
#define SIMULATIONVIEWEREYEMVT_H_DEFINED

#include "SimulationViewer.H"
#include "ImageSet.H"
#include "ImageCache.H"
class EyeTrackerSaccadeController;

//! Measure salience at human eye positions
class SimulationViewerEyeMvt : public SimulationViewer {
public:
  // ######################################################################
  /*! @name Constructors and destructors */
  //@{

  //! Constructor. See ModelComponent.H.
  SimulationViewerEyeMvt(ModelManager& mgr,
                         const std::string& descrName =
                         "EyeMvt Simulation Viewer",
                         const std::string& tagName =
                         "SimulationViewerEyeMvt");

  //! Destructor
  virtual ~SimulationViewerEyeMvt();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SaccadeController once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Set our brain
  virtual void setBrain(Brain* brain);

  //@}

  // ######################################################################
  /*! @name Events that will trigger new visualizations */
  //@{

  //! Ready to switch to next input frame
  virtual void newInputReady();

  //! Got a new input frame at the specified (current) time
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void newInputFrame(const Image< PixRGB<byte> >& input,
                             const double t);

  //! Update with availability of new channel drawings
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void channelUpdate(SharedPtr<VisualCortex>& vcx);

  //! Called every simulation time step (time is in seconds)
  /*! In SimulationViewerEyeMvt, we evolve the positions of our
    internal foveas. */
  virtual void evolve(const double t);

  //! Update with new covert attention position
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void attentionShift(const WTAwinner& win,
                              const Image<byte>& foaMask);

  //! Update with new eye position
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void eyeMovement(const Point2DT& curreye,
                           const Image<byte>& foveaMask);

  //! Update with new head position
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void headMovement(const Point2DT& currhead,
                            const Image<byte>& headMask);

  //! Update eye saccade status
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void eyeSaccade(const bool inSaccade);

  //! Update head saccade status
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void headSaccade(const bool inSaccade);

  //! Update eye blink status
  /*! In SimulationViewerEyeMvt, this is a no-op. */
  virtual void eyeBlink(const bool inBlink);

  //@}

  // ######################################################################
  /*! @name Output of our current internal drawings */
  //@{

  //! Get the attention/eye/head trajectory image
  virtual Image< PixRGB<byte> > getTraj(const double t);

  //! Save our various results
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

protected:
  ModelParam<bool> itsSaveMegaCombo;   //!< save mega combo?
  ModelParam< PixRGB<byte> > itsColorEye;  //!< patch color for human eye mvts
  ModelParam<int> itsFoveaRadius;      //!< radius of Fovea
  ModelParam<int> itsDelayCacheSize;   //!< size of our delay cache
  ModelParam<int> itsMaxCacheSize;     //!< size of our max cache
  ModelParam<bool> itsSampleAtStart;   //!< take samples at start or end of sac

  //! Name of a calibrated eye movement file
  ModelParam<std::string> itsOutFname;

  //! Get started
  virtual void start1();

  //! get stopped
  virtual void stop1();

private:
  Brain *itsBrain;                       // hook to brain or null
  double itsTime;                        // keep track of internal time
  Image< PixRGB<byte> > itsInput;        // copy of input
  Image< PixRGB<byte> > itsTraj;         // trajectory
  FILE *itsOutFile;
  Point2D itsLastSample;
  int itsLastRadius;
  double itsLastAmplitude;
  double itsLastDuration;

  // use a sliding image cache for salience computations:
  ImageCacheMinMax<float> itsDelayCache;
  ImageCacheMinMax<float> itsMaxCache;

  SharedPtr<EyeTrackerSaccadeController> itsSC;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
