/*!@file SimulationViewerEyeMvt2.C comparison between saliency and
  human eye movements */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerEyeMvt2.C 3486 2004-08-05 18:53:01Z itti $
//

#include "SimulationViewerEyeMvt2.H"
#include "ModelManager.H"
#include "Image_MathOps.H"     // for takeMax()
#include "Image_ColorOps.H"    // for contrastModulate()
#include "Image_DrawOps.H"     // for colGreyCombo()
#include "Image_Transforms.H"  // for segmentObjectClean(), contour2D8()
#include "Image_ShapeOps.H"
#include "Image_ColorOps.H"
#include "SaccadeControllers.H"
#include "VisualBuffer.H"
#include "Brain.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"
#include "AttentionGuidanceMap.H"
#include "SaliencyMap.H"
#include "TaskRelevanceMap.H"
#include "VisualCortex.H"
#include "ChannelBase.H"
#include "Retina.H"

// ######################################################################
SimulationViewerEyeMvt2::
SimulationViewerEyeMvt2(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName) :
  SimulationViewer(mgr, descrName, tagName),
  itsSaveMegaCombo(this, "SVEMsaveMegaCombo", false),
  itsColorEye(this, "SVcolorHumanEye", PixRGB<byte>(128, 255, 255)),
  itsFOAradius(this, "FOAradius", 30),
  itsFoveaRadius(this, "FoveaRadius", 30),
  itsOutFname(this, "SVEMoutFname", ""),
  itsLevelSpec(this, "LevelSpec", LevelSpec()),
  itsUseIOR(this, "SVEMuseIOR", true),
  itsBrain(NULL), itsTime(0.0), itsTraj(), itsRawTraj(),
  itsOutFile(NULL), itsLastSample(-1, -1), itsLastRadius(0),
  itsSC(NULL), itsBuffer(new VisualBuffer(mgr)), itsIORmask(),
  itsDidAttentionShift(false)
{
  LINFO("NOTE: Selecting SaccadeController of type EyeTrack");
  manager->setOptionValString("SaccadeControllerType", "EyeTrack");

  // appropriate our visual buffer:
  addSubComponent(itsBuffer);
}

// ######################################################################
SimulationViewerEyeMvt2::~SimulationViewerEyeMvt2()
{ }

// ######################################################################
void SimulationViewerEyeMvt2::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports. NOTE: here we consider as OPTEXP_CORE
  // the options that are involved in doing a getTraj(), i.e., which
  // are bout our core output, and we consider as OPTEXP_SAVE options
  // that would save other results:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsFOAradius);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsOutFname);
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsUseIOR);
    manager->requestOption(this, itsSaveMegaCombo);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimulationViewerEyeMvt2::setBrain(Brain* brain)
{ itsBrain = brain; itsBuffer->setBrain(brain); }

// ######################################################################
void SimulationViewerEyeMvt2::start1()
{
  // abort if no output file:
  if (itsOutFname.getVal().empty()) LFATAL("I need an output file!");

  // open output file:
  itsOutFile = fopen(itsOutFname.getVal().c_str(), "w");
  if (itsOutFile == NULL)
    PLFATAL("Cannot write '%s'", itsOutFname.getVal().c_str());

  // make sure we have the right type of SC and get a pointer to it:
  itsSC.dynCastFrom(itsBrain->getSC());
  if (itsSC.get() == NULL) LFATAL("I need an EyeTrackerSaccadeController");

  // make sure we have a retina:
  if (itsBrain->getRET().get() == NULL) LFATAL("I need a Retina!");
}

// ######################################################################
void SimulationViewerEyeMvt2::stop1()
{
  if (itsOutFile) { fclose(itsOutFile); itsOutFile = NULL; }
}

// ######################################################################
void SimulationViewerEyeMvt2::newInputReady()
{  }

// ######################################################################
void SimulationViewerEyeMvt2::newInputFrame(const
                                           Image< PixRGB<byte> >& input,
                                           const double t)
{
  itsTime = t; // update our time
  itsTraj = input; // keep the frame for later displays
  itsRawTraj = itsBrain->getRET()->getRawInput(); // also for later displays

  // if our buffer is not object based, pass it the current saliency
  // map now (the winner location does not matter):
  if (itsBuffer->isObjectBased() == false)
    itsBuffer->input(WTAwinner(Point2D(0, 0), t, 0.0, 0.0, false),
                     itsBrain->getSM()->getV(), Image<byte>());

  // let's force an initialization of our buffer if first frame:
  if (itsBuffer->getBuffer(false).initialized() == false)
    itsBuffer->input(WTAwinner(),
                     Image<float>(itsBrain->getSM()->getV().getDims(), ZEROS),
                     Image<byte>());
}

// ######################################################################
void SimulationViewerEyeMvt2::channelUpdate(SharedPtr<VisualCortex>& vcx)
{  }

// ######################################################################
void SimulationViewerEyeMvt2::evolve(const double t)
{
  // evolve our visual buffer:
  itsBuffer->evolve(t);
}

// ######################################################################
void SimulationViewerEyeMvt2::attentionShift(const WTAwinner& win,
                                             const Image<byte>& foaMask)
{
  // pass inputs on to our visual buffer, if it is object-based:
  if (itsBuffer->isObjectBased()) {
    itsBuffer->input(win, itsBrain->getAGM()->getV(), foaMask);

    // do some drawings:
    itsTraj.drawCircle(win.p, itsFOAradius.getVal(),
                       PixRGB<byte>(0, 127 + (win.boring?0:128), 0), 2);
    itsDidAttentionShift = true;
    
    // if we are being used jointly with a TRM of type KillN, let's pass
    // the object mask on to it:
    SharedPtr<TaskRelevanceMapKillN> trm; trm.dynCastFrom(itsBrain->getTRM());
    if (trm.get()) {
      Image<float> tmask;
      int w = itsBrain->getSM()->getWidth();
      int h = itsBrain->getSM()->getHeight();
      
      if (foaMask.initialized()) {
        // let's just use the foaMask:
        tmask = foaMask * (1.0F / 255.0F);
        tmask = binaryReverse(rescale(tmask, w, h), 1.0F);
      } else {
        // let's draw a disc:
        tmask.resize(itsTraj.getDims(), false); tmask.clear(1);
        tmask.drawDisk(win.p, itsFOAradius.getVal(), 0);
        tmask = rescale(tmask, w, h);
      }
      // let's feed that to the TRM:
      LINFO("Feeding object mask to TRM of type KillN");
      trm->input(tmask);
    }
  }
}

// ######################################################################
void SimulationViewerEyeMvt2::eyeMovement(const Point2DT& win,
                                          const Image<byte>& foveaMask)
{
  // get a snapshot of our buffer:
  Image<float> buf = itsBuffer->getBuffer(false);
  Dims dims = buf.getDims();
  int maplevel = itsLevelSpec.getVal().mapLevel();

  // select a drawing color & size:
  PixRGB<byte> col(itsColorEye.getVal()); int psiz = 5;
  if (itsSC->isInSaccade()) { col.setGreen(0); col.setRed(255); }
  if (itsSC->isInBlink()) { col.setBlue(0); }
  PixRGB<byte> bk(0, 0, 0);
  itsTraj.drawPatch(win.p, psiz + 2, bk);
  itsTraj.drawPatch(win.p, psiz, col);
  Point2D rawp = itsBrain->getRET()->retinalToRaw(win.p);
  itsRawTraj.drawPatch(rawp, psiz + 2, bk);
  itsRawTraj.drawPatch(rawp, psiz, col);

  // get the eye movement data:
  Point2D nextTarg; double ampl, durs;
  itsSC->getExtraEyeData(nextTarg, ampl, durs);

  // get buffer-centered coords of our current eye position:
  Point2D curr = itsBuffer->retinalToBuffer(win.p);

  // are we starting a saccade?
  if (ampl > 0.0)
    {
      // the coords we have from the SC are retinotopic; transform
      // into buffer-centered:
      itsLastSample = itsBuffer->retinalToBuffer(nextTarg);
      itsLastRadius = (itsFoveaRadius.getVal() +
                       (1 << maplevel) - 1) >> maplevel;
      
      // check that saccade target is okay:
      if (buf.coordsOk(itsLastSample) == false)
        {
          LERROR("Hum, saccade target at (%d, %d)? -- CLAMPING",
                 itsLastSample.i, itsLastSample.j);
          itsLastSample.clampToDims(dims);
        }
      float mi, ma, av; buf.getMinMaxAvg(mi, ma, av);

      // mask the buffer by a disk of radius FOAradius+stdev around
      // the saccade target location and compute the max inside:
      Image<float> fov(dims, ZEROS);
      itsTraj.drawCircle(nextTarg, itsFoveaRadius.getVal(),
                         PixRGB<byte>(255, 255, 0), 2);
      float salience = buf.getSample(itsLastSample, itsLastRadius);

      if (itsUseIOR.getVal())
        {
          // the human may or may not have turned on IOR, we don't really
          // know. Here let's just estimate what would happen if we
          // trigger IOR at our current location. If it improves, we'll
          // assume it was done. Improvement means that our measured
          // salience at the target location does not change, but the
          // average (and possibly max) over the entire image will go down
          // since we have eliminated our potentially very salient current
          // location. We first cut off anything less salient than 1/4 the
          // max salience, to make sure the segmentation will not spread
          // too far.
          Image<float> buf2(buf);
          buf2.lowThresh(ma * 0.25F, 0.0F);
          if (buf2.getVal(curr) > ma * 0.05F)
            {
              itsIORmask = segmentObjectClean(buf2, curr);
              Image<float> iormaskf(itsIORmask);
              iormaskf.normalize(0.0F, 1.0F);
              buf2 = buf * binaryReverse(iormaskf, 1.0F);
              float iorsal = buf2.getSample(itsLastSample, itsLastRadius);
              
              if (fabs(iorsal - salience) < ma * 1.0e-4F)
                {
                  LINFO("### Triggering IOR in Visual Buffer ###");
                  
                  // update min/max/avg/salience computation
                  itsBuffer->inhibit(iormaskf);
                  buf = itsBuffer->getBuffer(false);
                  buf.getMinMaxAvg(mi, ma, av);
                  salience = buf.getSample(itsLastSample, itsLastRadius);
                }
            }
        }
      
      // get salience we would obtain with a random
      // saccade of random amplitude:
      Point2D p(int(dims.w() * drand48()),
                int(dims.h() * drand48()));
      itsTraj.drawCircle(itsBuffer->bufferToRetinal(p),
                         itsFoveaRadius.getVal(),
                         PixRGB<byte>(255, 255, 255), 2);

      float rndsal2 = buf.getSample(p, itsLastRadius);
      float rnddist = p.distance(itsLastSample) * (1<<maplevel);

      // get location that our buffer would predict is the best target
      // for a saccade right now:
      p = itsBuffer->findMostInterestingTargetLocMax(curr);
      itsTraj.drawPatch(itsBuffer->bufferToRetinal(p), psiz + 2, bk);
      itsTraj.drawPatch(itsBuffer->bufferToRetinal(p), psiz,
                        PixRGB<byte>(0, 0, 255));
      float dist = p.distance(itsLastSample) * (1<<maplevel);

      // save vbuf contents at eye:
      fprintf(itsOutFile, "%g %g %g %g  %g %g %g  %g %g\n", salience,
              mi, ma, av, ampl, durs, rndsal2, dist, rnddist);
      fflush(itsOutFile);
    }
}

// ######################################################################
void SimulationViewerEyeMvt2::headMovement(const Point2DT& win,
                                    const Image<byte>& headMask)
{  }

// ######################################################################
void SimulationViewerEyeMvt2::eyeSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerEyeMvt2::headSaccade(const bool inSaccade)
{  }

// ######################################################################
void SimulationViewerEyeMvt2::eyeBlink(const bool inBlink)
{  }

// ######################################################################
Image< PixRGB<byte> > SimulationViewerEyeMvt2::getTraj(const double t)
{
  PixRGB<byte> bgcol(64, 128, 0); // background color
  PixRGB<byte> gridcol(255, 255, 255); // grid color

  itsTime = t; // update our time
  Dims tdims = itsTraj.getDims();
  int maplevel = itsLevelSpec.getVal().mapLevel();
  Image<float> sm = itsBrain->getAGM()->getV(true);
  Image< PixRGB<byte> > smc = toRGB(Image<byte>(rescale(sm, tdims)));

  // get the saliency mask (will show a blob depicting the area that
  // gets transferred from saliency map to buffer at each attention
  // shift):
  Image<byte> mask;
  if (itsBuffer->isObjectBased() && itsDidAttentionShift)
    mask = rescale(itsBuffer->getSaliencyMask(), tdims);
  else
    mask.resize(tdims, true);
  itsDidAttentionShift = false;

  // get the buffer:
  Image<float> buf = itsBuffer->getBuffer(true);
  Dims bdims(buf.getWidth() << maplevel, buf.getHeight() << maplevel);

  // get a color, full-scale version of the buffer:
  Image< PixRGB<byte> > bufc = toRGB(Image<byte>(rescale(buf, bdims)));

  // draw circle around saliency sample if we just took one:
  if (itsLastRadius != 0)
    {
      Point2D p(itsLastSample); p.i <<= maplevel; p.j <<= maplevel;
      int r(itsLastRadius); r <<= maplevel;
      bufc.drawCircle(p, r, PixRGB<byte>(255, 255, 0), 2);
      smc.drawCircle(itsBuffer->bufferToRetinal(p), r, 
                     PixRGB<byte>(255, 255, 0), 2);
      itsLastRadius = 0;
    }

  // draw IOR contours if any:
  if (itsIORmask.initialized())
    {
      Image<byte> contou = contour2D8(rescale(itsIORmask, bdims));
      
      Point2D ppp;
      for (ppp.j = 0; ppp.j < bdims.h(); ppp.j ++)
        for (ppp.i = 0; ppp.i < bdims.w(); ppp.i ++)
          if (contou.getVal(ppp.i, ppp.j))
            bufc.drawDisk(ppp, 2, PixRGB<byte>(255, 255, 0));
      itsIORmask.freeMem();
    }

  // return mega combo; we have 4 panels of the dims of the blown-up
  // buffer (bdims), and we will paste images of various dims in those
  // panels:
  Image< PixRGB<byte> > ret;

  // start with the unshifted raw input:
  Image< PixRGB<byte> > rawinp(bdims, NO_INIT); rawinp.clear(bgcol);
  Point2D rawinpoff((rawinp.getWidth() - itsRawTraj.getWidth())/2,
                    (rawinp.getHeight() - itsRawTraj.getHeight())/2);
  rawinp.paste(itsRawTraj, rawinpoff);

  // now the foveal/shifted image:
  Image< PixRGB<byte> > rtraj(bdims, NO_INIT); rtraj.clear(bgcol);
  Point2D rtrajoff((rtraj.getWidth() - itsTraj.getWidth())/2,
                   (rtraj.getHeight() - itsTraj.getHeight())/2);
  rtraj.paste(itsTraj, rtrajoff);

  // now the saliency map:
  Image< PixRGB<byte> > smc2(bdims, NO_INIT); smc2.clear(bgcol);
  smc2.paste(smc, rtrajoff);

  // ready for action:
  ret = concatY(concatX(rawinp, rtraj), concatX(bufc, smc2));

  // draw separating borders:
  ret.drawLine(Point2D(0, bdims.h()), Point2D(bdims.w()*2-1, bdims.h()),
               gridcol, 1);
  ret.drawLine(Point2D(bdims.w(), 0), Point2D(bdims.w(), bdims.h()*2-1),
               gridcol, 1);

  return ret;
}

// ######################################################################
void SimulationViewerEyeMvt2::saveResults(const
                                         SharedPtr<OutputFrameSeries>& ofs)
{
  // save results?
  if (itsSaveMegaCombo.getVal())
    ofs->writeRGB(getTraj(itsTime), Auto, -1, "T");
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
