/*!@file SimulationViewerEyeSim.C simulate an eye-tracker recording */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerEyeSim.C 4219 2005-02-17 22:04:00Z itti $
//

#include "SimulationViewerEyeSim.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Brain.H"
#include "SaccadeController.H"

// ######################################################################
SimulationViewerEyeSim::
SimulationViewerEyeSim(ModelManager& mgr, const std::string& descrName,
                         const std::string& tagName) :
  SimulationViewerStd(mgr, descrName, tagName),
  itsEyeFname(this, "SVeyeSimFname", ""),
  itsEyePeriod(this, "SVeyeSimPeriod", 0.0041666666), // 240Hz
  itsEyeTrash(this, "SVeyeSimTrash", 0),
  itsOutFile(NULL), itsSC(NULL), itsLastN(0)
{  }

// ######################################################################
SimulationViewerEyeSim::~SimulationViewerEyeSim()
{ }

// ######################################################################
void SimulationViewerEyeSim::exportOptions(const int whatToExport,
                                           const bool recurse)
{
  // let's first do our exports. NOTE: here we consider as OPTEXP_CORE
  // the options that are involved in doing a getTraj(), i.e., which
  // are bout our core output, and we consider as OPTEXP_SAVE options
  // that would save other results:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsEyeFname);
    manager->requestOption(this, itsEyePeriod);
    manager->requestOption(this, itsEyeTrash);
  }

  // the base class will propagate this call to our subcomponents:
  SimulationViewerStd::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimulationViewerEyeSim::start1()
{
  // get our base class started:
  SimulationViewerStd::start1();

  // abort if no output file:
  if (itsEyeFname.getVal().empty()) LFATAL("I need an output file!");

  // open output file:
  itsOutFile = fopen(itsEyeFname.getVal().c_str(), "w");
  if (itsOutFile == NULL)
    PLFATAL("Cannot write '%s'", itsEyeFname.getVal().c_str());

  // write our initial trash, here we just assume it's a long blink:
  for (int i = 0; i < itsEyeTrash.getVal(); i ++)
    fprintf(itsOutFile, "NaN NaN\n");

  // get our SC from brain:
  itsSC = itsBrain->getSC();
  if (itsSC.get() == NULL) LFATAL("I need a SaccadeController");
}

// ######################################################################
void SimulationViewerEyeSim::stop1()
{
  if (itsOutFile)
    {
      // write a few extra samples at end of simulation:
      Point2D eye = itsSC->getPreviousDecision(0).p;
      for (int i = 0; i < 50; i ++)
        fprintf(itsOutFile, "%d.0 %d.0\n", eye.i, eye.j);

      // close down:
      fclose(itsOutFile);
      itsOutFile = NULL;
    }
}

// ######################################################################
void SimulationViewerEyeSim::evolve(const double t)
{
  // check whether the time has arrived to write out our next sample:
  int64 nstep = int64(t / itsEyePeriod.getVal());

  while(nstep > itsLastN)
    {
      // get current eye position and blink status:
      if (itsSC->isInBlink())
        fprintf(itsOutFile, "NaN NaN\n");
      else
        {
          Point2D eye = itsSC->getPreviousDecision(0).p;
          fprintf(itsOutFile, "%d.0 %d.0\n", eye.i, eye.j);
        }

      // ready for next sample:
      ++ itsLastN;
    }

  // do a normal evolve on SimulationViewerStd:
  SimulationViewerStd::evolve(t);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
