/*!@file SimulationViewerStd.C visualize various model simulations */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerStd.C 4219 2005-02-17 22:04:00Z itti $
//

#include "SimulationViewerStd.H"
#include "ModelManager.H"
#include "Image_Transforms.H"  // for contour2D()
#include "Image_ColorOps.H"    // for normalizeC()
#include "VisualCortex.H"
#include "SaliencyMap.H"
#include "TaskRelevanceMap.H"
#include "ImageSet.H"
#include "Pyramid_Ops.H"
#include "Image_MathOps.H"   // for takeMax()
#include "Image_ShapeOps.H"  // for crop()
#include "Image_FilterOps.H"
#include "Image_DrawOps.H"
#include "Brain.H"
#include "FrameSeries.H"
#include "XWindow.H"
#include "Channels.H"
#include "ModelOptionDefs.H"

#include <unistd.h> // for usleep

// ######################################################################
SimulationViewerStd::SimulationViewerStd(ModelManager& mgr,
                                         const std::string& descrName,
                                         const std::string& tagName) :
  SimulationViewer(mgr, descrName, tagName),
  itsSaveTraj(this, "SVsaveTraj", false),
  itsSaveXcombo(this, "SVsaveXcombo", false),
  itsSaveYcombo(this, "SVsaveYcombo", false),
  itsSaveTRMXcombo(this, "SVsaveTRMXcombo", false),
  itsSaveTRMYcombo(this, "SVsaveTRMYcombo", false),
  itsWarp3D(this, "SVwarp3D", false),
  itsMegaCombo(this, "SVmegaCombo", false),
  itsMotionSixPack(this, "SVmotionSixPack", false),
  itsCropFOA(this, "SVcropFOA", Dims(0, 0)),
  itsXwindow(this, "SVxwindow", false),
  itsFoveateTraj(this, "SVfoveateTraj", false),
  itsDisplayFOA(this, "SVdisplayFOA", true),
  itsDisplayPatch(this, "SVdisplayPatch", true),
  itsDisplayFOALinks(this, "SVdisplayFOALinks", true),
  itsDisplayEye(this, "SVdisplayEye", true),
  itsDisplayEyeLinks(this, "SVdisplayEyeLinks", false),
  itsDisplayHead(this, "SVdisplayHead", false),
  itsDisplayHeadLinks(this, "SVdisplayHeadLinks", false),
  itsDisplayTime(this, "SVdisplayTime", true),
  itsDisplayAdditive(this, "SVdisplayAdditive", true),
  itsDisplayHighlights(this, "SVdisplayHighlights", false),
  itsDisplayBoring(this, "SVdisplayBoring", true),
  itsDisplayShapeEstimator(this, "SVdisplayShapeEstimator", false),
  itsColorNormal(this, "SVcolorNormal", PixRGB<byte>(255, 255, 0)),
  itsColorBoring(this, "SVcolorBoring", PixRGB<byte>(127, 0, 0)),
  itsColorBlink(this, "SVcolorBlink", PixRGB<byte>(0, 128, 255)),
  itsColorSaccade(this, "SVcolorSaccade", PixRGB<byte>(255, 128, 255)),
  itsColorLink(this, "SVcolorLink", PixRGB<byte>(255, 0, 0)),
  itsHighlightMax(this, "SVhighlightMax", 320),
  itsShapeEstimatorBaseContrast(this, "SVShapeEstimatorBaseContrast", 0.1F),
  itsShapeEstimatorBaseBright(this, "SVShapeEstimatorBaseBright", 128),
  itsWarp3Dpitch(this, "SVwarp3DInitialPitch", -25.0F),
  itsWarp3Dyaw(this, "SVwarp3DInitialYaw", -15.0F),
  itsWarp3DpitchRate(this, "SVwarp3DpitchRate", 0.0F),
  itsWarp3DyawRate(this, "SVwarp3DyawRate", 15.0F),
  itsWarp3DpitchMax(this, "SVwarp3DpitchMax", 0.0F),
  itsWarp3DyawMax(this, "SVwarp3DyawMax", 20.0F),
  itsFOAradius(this, "FOAradius", 30),
  itsFoveaRadius(this, "FoveaRadius", 30),
  itsHeadRadius(this, "HeadMarkerRadius", 50),
  itsMultiRetinaDepth(this, "MultiRetinaDepth", 5),
  itsFOApsiz(this, "SVfoapsiz", 3),
  itsFOAthick(this, "SVfoathick", 3),
  itsFOAlinkThick(this, "SVfoalinkthick", 2),
  itsFOVpsiz(this, "SVfovpsiz", 3),
  itsFOVpthick(this, "SVfovpthick", 1),
  itsFOVthick(this, "SVfovthick", 2),
  itsFOVlinkThick(this, "SVfovlinkthick", 1),
  itsHEDpsiz(this, "SVhedpsiz", 9),
  itsHEDpthick(this, "SVhedpthick", 1),
  itsHEDthick(this, "SVhedthick", 1),
  itsHEDlinkThick(this, "SVhedlinkthick", 1),
  itsUseLargerDrawings(this, "SVuseLargerDrawings", false),
  itsBrain(NULL),
  itsInput(),
  itsTraj(),
  itsCurrFOA(-1, -1, 0.0, 0.0, 0.0, false),
  itsCurrFOAmask(),
  itsCumFOAmask(),
  itsCurrEye(-1, -1, 0.0),
  itsCurrFoveaMask(),
  itsCumFoveaMask(),
  itsCurrHead(-1, -1, 0.0),
  itsCurrHeadMask(),
  itsCumHeadMask(),
  itsPrevFOA(-1, -1, 0.0, 0.0, 0.0, false),
  itsPrevEye(-1, -1, 0.0),
  itsPrevHead(-1, -1, 0.0),
  itsCurrTime(0.0),
  itsEyeSaccade(false),
  itsHeadSaccade(false),
  itsEyeBlink(false),
  itsMultiTraj(),
  itsDims3D(),
  itsPitch3D(-1.0e10F),
  itsYaw3D(-1.0e10F),
  itsWin(NULL),
  itsHasNewInput(false)
{  }

// ######################################################################
SimulationViewerStd::~SimulationViewerStd()
{ }

// ######################################################################
void SimulationViewerStd::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // let's first do our exports. NOTE: here we consider as OPTEXP_CORE
  // the options that are involved in doing a getTraj(), i.e., which
  // are bout our core output, and we consider as OPTEXP_SAVE options
  // that would save other results:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsSaveTraj);
    manager->requestOption(this, itsSaveXcombo);
    manager->requestOption(this, itsSaveYcombo);
    manager->requestOption(this, itsSaveTRMXcombo);
    manager->requestOption(this, itsSaveTRMYcombo);
    manager->requestOption(this, itsWarp3D);
    manager->requestOption(this, itsMegaCombo);
    manager->requestOption(this, itsMotionSixPack);
    manager->requestOption(this, itsCropFOA);
    manager->requestOption(this, itsXwindow);
    manager->requestOption(this, itsFoveateTraj);
    
    manager->requestOption(this, itsDisplayFOA);
    manager->requestOption(this, itsDisplayPatch);
    manager->requestOption(this, itsDisplayFOALinks);
    manager->requestOption(this, itsDisplayEye);
    manager->requestOption(this, itsDisplayEyeLinks);
    manager->requestOption(this, itsDisplayHead);
    manager->requestOption(this, itsDisplayHeadLinks);
    manager->requestOption(this, itsDisplayTime);
    manager->requestOption(this, itsDisplayAdditive);
    manager->requestOption(this, itsDisplayHighlights);
    manager->requestOption(this, itsDisplayBoring);

    // don't request options for the various colors, that would be too
    // extreme in terms of the number of command-line options
    
    manager->requestOption(this, itsFOAradius);
    manager->requestOption(this, itsFoveaRadius);
    manager->requestOption(this, itsHeadRadius);
    manager->requestOption(this, itsMultiRetinaDepth);

    manager->requestOption(this, itsUseLargerDrawings);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SimulationViewerStd::reset(const bool recurse)
{
  // reset some stuff for SimulationViewerStd
  itsInput.freeMem();
  itsTraj.freeMem();
  itsCurrFOA = WTAwinner(-1, -1, 0.0, 0.0, 0.0, false);
  itsCurrFOAmask.freeMem();
  itsCumFOAmask.freeMem();
  itsCurrEye = Point2DT(-1, -1, 0.0);
  itsCurrFoveaMask.freeMem();
  itsCumFoveaMask.freeMem();
  itsCurrHead = Point2DT(-1, -1, 0.0);
  itsCurrHeadMask.freeMem();
  itsCumHeadMask.freeMem();
  itsPrevFOA = WTAwinner(-1, -1, 0.0, 0.0, 0.0, false);
  itsPrevEye = Point2DT(-1, -1, 0.0);
  itsPrevHead = Point2DT(-1, -1, 0.0);
  itsCurrTime = 0.0;
  itsEyeSaccade = false;
  itsHeadSaccade = false;
  itsEyeBlink = false;
  itsMultiTraj.reset();
  itsDims3D = Dims();
  itsPitch3D = -1.0e10F;
  itsYaw3D = -1.0e10F;
  itsWin.reset();
  itsHasNewInput = false;

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void SimulationViewerStd::start1()
{
  // do we wnat larger drawings?
  if (itsUseLargerDrawings.getVal())
    {
      itsFOApsiz.setVal(5);
      itsFOVpsiz.setVal(5);
      itsFOVpthick.setVal(2);
      itsHEDpsiz.setVal(12);
      itsHEDpthick.setVal(2);
    }
}

// ######################################################################
void SimulationViewerStd::setBrain(Brain* brain)
{ itsBrain = brain; }

// ######################################################################
void SimulationViewerStd::newInputReady()
{  }

// ######################################################################
void SimulationViewerStd::newInputFrame(const Image< PixRGB<byte> >& input,
                                     const double t)
{
  // keep a copy of the image
  itsInput = input;

  // for now, we don't care about the time...

  // erase old trajectory and replace it by fresh new input:
  itsTraj = itsInput;

  // if foveating the traj, get a foveation pyramid ready:
  if (itsFoveateTraj.getVal())
    itsMultiTraj = buildPyrGaussian(input, itsMultiRetinaDepth.getVal(), 9);

  // we have a new input; this will force redrawing various things on
  // the trajectory in case people request it before a new shift of
  // attention or other event occurrs:
  itsHasNewInput = true;
}

// ######################################################################
void SimulationViewerStd::channelUpdate(SharedPtr<VisualCortex>& vcx)
{
  if (itsTraj.initialized() == false || vcx.get() == NULL) return;
  vcx->drawResults(itsTraj);
}

// ######################################################################
void SimulationViewerStd::evolve(const double t)
{
  
  // do the 3D rotation:
  if (itsWarp3D.getVal()) {
    if (itsPitch3D == -1.0e10F) itsPitch3D = itsWarp3Dpitch.getVal();
    if (itsYaw3D == -1.0e10F) itsYaw3D = itsWarp3Dyaw.getVal();

    itsPitch3D += itsWarp3DpitchRate.getVal() * (t - itsCurrTime);
    itsYaw3D += itsWarp3DyawRate.getVal() * (t - itsCurrTime);
    
    if (itsYaw3D >= itsWarp3DyawMax.getVal() ||
        itsYaw3D <= -itsWarp3DyawMax.getVal())
      itsWarp3DyawRate.setVal(- itsWarp3DyawRate.getVal());
    if (itsPitch3D >= itsWarp3DpitchMax.getVal() ||
        itsPitch3D <= -itsWarp3DpitchMax.getVal())
      itsWarp3DpitchRate.setVal(- itsWarp3DpitchRate.getVal());
  }

  // update our internal time:
  itsCurrTime = t;
}

// ######################################################################
void SimulationViewerStd::attentionShift(const WTAwinner& win,
                                         const Image<byte>& foaMask)
{
  // check dims of mask:
  if (foaMask.initialized() && foaMask.isSameSize(itsInput) == false)
    LFATAL("Dimensions of FOAmask must match those of input -- ABORT");

  // update our internals:
  itsPrevFOA = itsCurrFOA; itsCurrFOA = win;
  if (foaMask.initialized())
    itsCurrFOAmask = foaMask;  // keep a copy of the FOA mask
  else
    {
      // draw a disk at current foa position:
      itsCurrFOAmask.resize(itsInput.getDims(), true);
      if (win.isValid())
        itsCurrFOAmask.drawDisk(win.p, itsFOAradius.getVal(), 255);
    }

  // update cumulative FOA mask:
  if (itsDisplayAdditive.getVal() &&            // cumulative display?
      itsCumFOAmask.initialized() &&            // not first frame?
      itsCumFOAmask.isSameSize(itsCurrFOAmask)) // not changing frame dims?
    itsCumFOAmask = takeMax(itsCumFOAmask, itsCurrFOAmask);
  else
    itsCumFOAmask = itsCurrFOAmask;

  // forget it if we don't have a traj yet:
  if (itsTraj.initialized() == false) return;

  // draw the FOA:
  if (itsDisplayFOA.getVal() || itsDisplayPatch.getVal()) drawFOA();

  // link the last two FOAs:
  if (itsDisplayFOALinks.getVal()) linkFOAs();
}

// ######################################################################
void SimulationViewerStd::eyeMovement(const Point2DT& win,
                                   const Image<byte>& foveaMask)
{
  // check dims of mask:
  if (foveaMask.initialized() && foveaMask.isSameSize(itsInput) == false)
    LFATAL("Dimensions of FoveaMask must match those of input -- ABORT");

  // update our internals:
  itsPrevEye = itsCurrEye; itsCurrEye = win;
  if (foveaMask.initialized())
    itsCurrFoveaMask = foveaMask;  // keep a copy fo the fovea mask
  else
    {
      // draw a disk at current eye position:
      itsCurrFoveaMask.resize(itsInput.getDims(), true);
      if (win.isValid())
        itsCurrFoveaMask.drawDisk(win.p, itsFoveaRadius.getVal(), 255);
    }
  
  // update cumulative fovea mask:
  if (itsDisplayAdditive.getVal() &&
      itsCumFoveaMask.initialized() &&
      itsCumFoveaMask.isSameSize(itsCurrFoveaMask))
    itsCumFoveaMask = takeMax(itsCumFoveaMask, itsCurrFoveaMask);
  else
    itsCumFoveaMask = itsCurrFoveaMask;

  // forget it if we don't have a traj yet:
  if (itsTraj.initialized() == false) return;

  // draw the eye position:
  if (itsDisplayEye.getVal()) drawEye();

  // link the last two eye positions:
  if (itsDisplayEyeLinks.getVal()) linkEyes();
}

// ######################################################################
void SimulationViewerStd::headMovement(const Point2DT& win,
                                    const Image<byte>& headMask)
{
  // check dims of mask:
  if (headMask.initialized() && headMask.isSameSize(itsInput) == false)
    LFATAL("Dimensions of HeadMask must match those of input -- ABORT");

  // update our internals:
  itsPrevHead = itsCurrHead; itsCurrHead = win;
  if (headMask.initialized())
    itsCurrHeadMask = headMask;  // keep a copy of the mask
  else
    {
      // draw a disk at current head position:
      itsCurrHeadMask.resize(itsInput.getDims(), true);
      if (win.isValid())
        itsCurrHeadMask.drawDisk(win.p, itsHeadRadius.getVal(), 255);
    }

  // update cumulative head mask:
  if (itsDisplayAdditive.getVal() &&
      itsCumHeadMask.initialized() &&
      itsCumHeadMask.isSameSize(itsCurrHeadMask))
    itsCumHeadMask = takeMax(itsCumHeadMask, itsCurrHeadMask);
  else
    itsCumHeadMask = itsCurrHeadMask;

  // forget it if we don't have a traj yet:
  if (itsTraj.initialized() == false) return;

  // draw the head position:
  if (itsDisplayHead.getVal()) drawHead();

  // link the last two head positions:
  if (itsDisplayHeadLinks.getVal()) linkHeads();
}

// ######################################################################
void SimulationViewerStd::eyeSaccade(const bool inSaccade)
{ itsEyeSaccade = inSaccade; }

// ######################################################################
void SimulationViewerStd::headSaccade(const bool inSaccade)
{ itsHeadSaccade = inSaccade; }

// ######################################################################
void SimulationViewerStd::eyeBlink(const bool inBlink)
{ itsEyeBlink = inBlink; }

// ######################################################################
Image< PixRGB<byte> > SimulationViewerStd::getTraj(const double t)
{
  itsCurrTime = t; bool redraw = false;
  if (itsTraj.initialized() == false) return itsTraj;

  // ##### if not doing additive displays, clear traj and redraw only
  // ##### current foa/eye/head data:
  if (itsDisplayAdditive.getVal() == false)
    { itsTraj = itsInput; redraw = true; }

  // ##### Do a bunch of drawings that will completely destroy and
  // ##### rebuild the trajectory image:

  // show ShapeEstimator masking (at the location(s) of covert attention):
  if (itsDisplayShapeEstimator.getVal())
    {
      itsTraj = contrastModulate(itsInput, itsCumFOAmask,
                                 itsShapeEstimatorBaseContrast.getVal(),
                                 itsShapeEstimatorBaseBright.getVal());
      redraw = true;
    }
  
  // display FOA highlights using mask of eye positions:
  if (itsDisplayHighlights.getVal())
    {
      itsTraj = highlightRegions(itsInput, itsCumFoveaMask,
                                 itsHighlightMax.getVal());
      redraw = true;
    }

  // show foveated masking, at the location of overt attention:
  if (itsFoveateTraj.getVal())
    {
      itsTraj = foveate(itsCumFoveaMask, itsMultiTraj);
      redraw = true;
    }

  // ##### re-do a bunch of drawings if we destroyed the traj
  // ##### (foveation, highlights, etc). We only care about current
  // ##### attention/eye/head position and not about any previous ones
  // ##### or links:
  if (redraw || itsHasNewInput)
    {
      if (itsDisplayFOA.getVal() || itsDisplayPatch.getVal()) drawFOA();
      if (itsDisplayEye.getVal()) drawEye();
      if (itsDisplayHead.getVal()) drawHead();
      itsHasNewInput = false;
    }

  // ##### let's get ready to return the traj, but we may reshape it
  // ##### just before that, or return something that involves the
  // ##### traj plus some other maps in some combined display:
  Image< PixRGB<byte> > ret;

  // if we want a megacombo, let's return it now:
  if (itsMegaCombo.getVal()) ret = drawMegaCombo();
  
  // if we want a motion sixpack, let's return it now:
  else if (itsMotionSixPack.getVal()) ret = drawMotionSixPack();
  
  // if cropping let's crop now:
  // resizing and other normalizations are taken care of by OutputFrameSeries  
  else if (itsCropFOA.getVal().isNonEmpty())
    {
      Dims crop_dims = itsCropFOA.getVal();
      Rectangle crect(itsCurrFOA.p.j - crop_dims.h() / 2,
                      itsCurrFOA.p.i - crop_dims.w() / 2,
                      itsCurrFOA.p.j + crop_dims.h() / 2 - 1,
                      itsCurrFOA.p.i + crop_dims.w() / 2 - 1);
      ret = crop(itsTraj, crect, true);
    }

  // return a 3D traj warped onto saliency map?
  else if (itsWarp3D.getVal() && itsBrain) {
    ret = itsTraj;

    ret.drawGrid(ret.getWidth() / 6, ret.getHeight() / 6, 3, 3,
                 itsColorNormal.getVal());
    ret.drawRect(Rectangle(1, 1, ret.getHeight()-2, ret.getWidth()-2),
                           itsColorNormal.getVal(), 3);
    ret = warp3Dmap(ret, itsBrain->getSM()->getV(true),
                    itsPitch3D, itsYaw3D, itsDims3D);
  }

  // display combo traj + salmap?
  else if ((itsSaveXcombo.getVal() || itsSaveYcombo.getVal()) && itsBrain)
    ret = colGreyCombo(itsTraj, itsBrain->getSM()->getV(true),
                       itsSaveXcombo.getVal());

  // display combo traj + salmap + highlighted trm?
  else if ((itsSaveTRMXcombo.getVal() || itsSaveTRMYcombo.getVal()) 
           && itsBrain)
    {
      ret = colGreyCombo(itsTraj, itsBrain->getSM()->getV(true), 
                         itsSaveTRMXcombo.getVal());
      Image<float> trm = itsBrain->getTRM()->getV(false); 
      trm.normalize(0.0f, 1.0f);
      Image<PixRGB<float> > highlight_trm = toRGB(trm);
      highlight_trm = (Image<PixRGB<float> >)itsInput *
        rescale(trm, itsInput.getDims());
      ret = colColCombo(ret, (Image<PixRGB<byte> >)highlight_trm, 
                        itsSaveTRMXcombo.getVal());
    }

  // otherwise, use the traj so far:
  else ret = itsTraj;

  // ##### do a bunch of last-minute drawings:
  if (itsDisplayTime.getVal()) drawTime(ret);

  // ##### return final transformed traj:
  return ret;
}

// ######################################################################
void SimulationViewerStd::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // update the trajectory:
  Image< PixRGB<byte> > res = getTraj(itsCurrTime);

  // if using Xwindow, display the results instead of saving:
  if (itsXwindow.getVal())
    {
      // although we are not using ofs to save, we nevertheless use
      // the settings of ofs to resize the image according to
      // what the user wants:
      Image< PixRGB<byte> > xres = ofs->resizeImage(res);
      
      // open the X window if it does not exist yet:
      if (itsWin.get() == NULL)
        {
          itsWin =
            makeShared(new XWindow(xres.getWidth(), xres.getHeight(), -1, -1,
                                   "iLab Neuromorphic SimulationViewer"));
          // somehow, the first image does not get displayed?
          usleep(10000); itsWin->drawImage(xres); usleep(10000);
        }

      // draw the image in the window:
      itsWin->drawImage(xres);
    }
  else
    {
      // save results?
      if (itsSaveTraj.getVal() || itsSaveXcombo.getVal() ||
          itsSaveYcombo.getVal() || itsSaveTRMXcombo.getVal() ||
          itsSaveTRMYcombo.getVal() || itsWarp3D.getVal() ||
          itsMotionSixPack.getVal() || itsMegaCombo.getVal())
        ofs->writeRGB(res, Auto, -1, "T");
    }
}

// ######################################################################
void SimulationViewerStd::drawFOA()
{
  if (itsCurrFOA.isValid() == false) return;

  // select a drawing color:
  PixRGB<byte> col(itsColorNormal.getVal());
  if (itsCurrFOA.boring) col -= itsColorBoring.getVal();

  // draw patch at current eye position:
  itsTraj.drawPatch(itsCurrFOA.p, itsFOApsiz.getVal(), col);

  // display focus of attention, possibly object-shaped:
  if (itsDisplayFOA.getVal())
    drawMaskOutline(itsTraj, itsCurrFOAmask, col, itsFOAthick.getVal(),
                    itsCurrFOA.p, itsFOAradius.getVal());
}

// ######################################################################
void SimulationViewerStd::linkFOAs()
{
  if (itsCurrFOA.isValid() == false) return;

  PixRGB<byte> col = itsColorLink.getVal();
  if (itsPrevFOA.isValid())
    {
      int d = int(itsPrevFOA.p.distance(itsCurrFOA.p));
      if (d > 0) itsTraj.drawArrow(itsPrevFOA.p, itsCurrFOA.p,
                                   col, itsFOAlinkThick.getVal());
    }
}

// ######################################################################
void SimulationViewerStd::drawEye()
{
  if (itsCurrEye.isValid() == false) return;

  int psiz = itsFOVpsiz.getVal();    // half-size of patch

  // select a drawing color:
  PixRGB<byte> col(itsColorNormal.getVal());
  if (itsEyeSaccade) { col = itsColorSaccade.getVal(); psiz += 2; }
  if (itsEyeBlink) { col = itsColorBlink.getVal(); psiz += 4; }
  if (itsCurrFOA.boring) col -= itsColorBoring.getVal();

  // draw patch (thin hollow square):
  itsTraj.drawRect(Rectangle(itsCurrEye.p.j - psiz, itsCurrEye.p.i - psiz,
                             itsCurrEye.p.j + psiz, itsCurrEye.p.i + psiz,
                             itsTraj.getDims()), col, itsFOVpthick.getVal());

  // draw fovea, possibly mask-shaped:
  if (itsDisplayFOA.getVal())
    drawMaskOutline(itsTraj, itsCurrFoveaMask, col, itsFOVthick.getVal(),
                    itsCurrEye.p, itsFoveaRadius.getVal());
}

// ######################################################################
void SimulationViewerStd::linkEyes()
{
  if (itsCurrEye.isValid() == false) return;

  PixRGB<byte> col = itsColorLink.getVal();

  // draw a line segment:
  if (itsPrevEye.isValid())
    itsTraj.drawLine(itsPrevEye.p, itsCurrEye.p, col,
                     itsFOVlinkThick.getVal());
}

// ######################################################################
void SimulationViewerStd::drawHead()
{
  if (itsCurrHead.isValid() == false) return;

  int psiz = itsHEDpsiz.getVal();   // half-size of patch

  // select a drawing color:
  PixRGB<byte> col(itsColorNormal.getVal());
  if (itsHeadSaccade) { col = itsColorSaccade.getVal(); psiz += 3; }
  if (itsCurrFOA.boring) col -= itsColorBoring.getVal();

  // draw square at current head position:
  itsTraj.drawRect(Rectangle(itsCurrHead.p.j - psiz, itsCurrHead.p.i - psiz,
                             itsCurrHead.p.j + psiz, itsCurrHead.p.i + psiz,
                             itsTraj.getDims()), col, itsHEDpthick.getVal());

  // draw head marker, possibly object-shaped:
  if (itsDisplayFOA.getVal())
    drawMaskOutline(itsTraj, itsCurrHeadMask, col, itsHEDthick.getVal(),
                    itsCurrHead.p, itsHeadRadius.getVal());
}

// ######################################################################
void SimulationViewerStd::linkHeads()
{
  if (itsCurrHead.isValid() == false) return;

  PixRGB<byte> col = itsColorLink.getVal();

  // draw a line segment:
  if (itsPrevHead.isValid())
    itsTraj.drawLine(itsPrevHead.p, itsCurrHead.p, col,
                     itsHEDlinkThick.getVal());
}

// ######################################################################
void SimulationViewerStd::drawTime(Image<PixRGB<byte> >& image)
{
  char txt[20]; sprintf(txt, " %dms ", (int)(itsCurrTime * 1000.0 + 0.4999));
  image.writeText(Point2D(0, 0), txt);
}

// ######################################################################
Image< PixRGB<byte> > SimulationViewerStd::drawMegaCombo()
{
  if (itsBrain == NULL) LFATAL("I need a brain!");
  if (itsBrain->getSM().get() == NULL) LFATAL("I need an SM");
  if (itsBrain->getTRM().get() == NULL) LFATAL("I need a TRM");

  PixRGB<byte> bg(128, 128, 255);
  int XWW = itsTraj.getWidth() * 2, XWH = itsTraj.getHeight() * 2;
  Image< PixRGB<byte> > xwi(XWW/2, XWH/2, NO_INIT);
  Image< PixRGB<byte> > xwin(XWW, XWH, ZEROS);
  
  Rectangle r(0, 0, XWH/2-1, XWW/2-1); xwi.embed(itsTraj, r, bg, true);
  xwin.paste(xwi, Point2D(0, 0));

  // get the normalized SM:
  Image<float> agm = itsBrain->getSM()->getV(true);
  Image< PixRGB<byte> > xwi2(XWW/2, XWH/2, NO_INIT);

  // get the non-normalized TRM (neutral is 1.0):
  Image<float> trm = itsBrain->getTRM()->getV();

  // useful range is 0.0 .. 3.0; let's show values smaller
  // than 1 as red and those larger that 1 as green:
  Image<byte> rr, gg, bb;
  Image<float> trmfac(trm); trmfac.clamp(0.0F, 1.0F);
  trmfac *= -1.0F; trmfac += 1.0F;
  rr = trmfac * 255.0F; // the redder, the lower the relevance
  trmfac = trm; trmfac.clamp(1.0F, 3.0F); trmfac -= 1.0F;
  gg = trmfac * 127.5F; // the greener, the higher the relevance
  bb = agm;  // in blue is the AGM

  // place the agm into our result image:
  Image<PixRGB<byte> > cbtmp = rescale(makeRGB(rr, gg, bb), XWW/2, XWH/2);
  xwi2.embed(cbtmp, r, PixRGB<byte>(192, 192, 192), true);
  xwi2.writeText(Point2D(0,0), " Attention Map ");
  xwin.paste(xwi2, Point2D(XWW/2, 0));

  // now do the conspicuity maps:
  r.modify(0, 0, XWH/4 - 1, XWW/4 - 1);
  xwi2.resize(XWW/4, XWH/4);

  SharedPtr<VisualCortex> vc = itsBrain->getVC();
  if (vc.get() == NULL) LFATAL("I need a visual cortex!");

  if (vc->hasChannel("color")) {
    Image<float> ftmp = vc->channel("color").getOutput();
    ftmp = rescale(ftmp, XWW / 4, XWH / 4); ftmp.normalize(0.0, 255.0);
    Image<byte> btmp = ftmp; btmp.writeText(Point2D(0,0), " Color ");
    xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
    xwin.paste(xwi2, Point2D(0, XWH/2));
  }

  if (vc->hasChannel("intensity")) {
    Image<float> ftmp = vc->channel("intensity").getOutput();
    ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
    Image<byte> btmp = ftmp; btmp.writeText(Point2D(0,0), " Intensity ");
    xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
    xwin.paste(xwi2, Point2D(XWW/4, XWH/2));
  }

  if (vc->hasChannel("orientation")) {
    Image<float> ftmp = vc->channel("orientation").getOutput();
    ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
    Image<byte> btmp = ftmp; btmp.writeText(Point2D(0,0), " Orientation ");
    xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
    xwin.paste(xwi2, Point2D(XWW/2, XWH/2));
  }
  
  if (vc->hasChannel("flicker")) {
    Image<float> ftmp = vc->channel("flicker").getOutput();
    ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
    Image<byte> btmp = ftmp; btmp.writeText(Point2D(0,0), " Flicker ");
    xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
    xwin.paste(xwi2, Point2D(XWW/4+XWW/2, XWH/2));
  }

  if (vc->hasChannel("motion")) {
    SharedPtr<MotionChannel> mc; mc.dynCastFrom(vc->channelPtr("motion"));
    if (mc.get() == NULL) LFATAL("Channel named 'motion' not a MotionChannel");
    if (mc->numChans() == 4) {
      Image<float> ftmp = mc->dirChan(2).getOutput();
      ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
      Image<byte> btmp = ftmp; btmp.writeText(Point2D(0,0), " MotionLeft ");
      xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
      xwin.paste(xwi2, Point2D(0, XWH/4 + XWH/2));
      
      ftmp = mc->dirChan(1).getOutput();
      ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
      btmp = ftmp; btmp.writeText(Point2D(0,0), " MotionDown ");
      xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
      xwin.paste(xwi2, Point2D(XWW/4, XWH/4 + XWH/2));
      
      ftmp = mc->dirChan(3).getOutput();
      ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
      btmp = ftmp; btmp.writeText(Point2D(0,0), " MotionUp ");
      xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
      xwin.paste(xwi2, Point2D(XWW/2, XWH/4 + XWH/2));
      
      ftmp = mc->dirChan(0).getOutput();
      ftmp = rescale(ftmp, XWW/4, XWH/4); ftmp.normalize(0.0, 255.0);
      btmp = ftmp; btmp.writeText(Point2D(0,0), " MotionRight ");
      xwi2.embed(btmp, r, 192, true); xwi2.drawRect(r, 255, 1);
      xwin.paste(xwi2, Point2D(XWW/2 + XWW/4, XWH/4 + XWH/2));
    }
  }

  return xwin;
}

// ######################################################################
Image< PixRGB<byte> > SimulationViewerStd::drawMotionSixPack()
{
  if (itsBrain == NULL) LFATAL("I need a brain!");

  int margin1 = 10, margin2 = 20;

  Dims origDim(itsTraj.getDims());
  Image<byte> img1(luminance(itsTraj));

  SharedPtr<VisualCortex> vc = itsBrain->getVC();
  if (vc.get() == NULL) LFATAL("I need a visual cortex!");
  if (vc->hasChannel("motion") == false) LFATAL("I need a motion channel!");
  SharedPtr<MotionChannel> mc; mc.dynCastFrom(vc->channelPtr("motion"));
  if (mc.get() == NULL) LFATAL("Channel named 'motion' not a MotionChannel");
  if (mc->numChans() != 4)
    LFATAL("I only work with a 4-direction motion channel!");

  // left
  img1 = concatX(scaleBlock((Image<byte>)mc->dirChan(2).getOutput(),
                            origDim), img1);
  img1.drawArrow(Point2D(margin2, margin1), Point2D(margin1, margin1), 255);
  
  // down
  img1 = concatX(scaleBlock((Image<byte>)mc->dirChan(1).getOutput(),
                            origDim), img1);
  img1.drawArrow(Point2D(margin1, margin1), Point2D(margin1, margin2), 255);
  
  Image<byte> img2(scaleBlock(mc->getOutput(), origDim));
  
  // up
  img2 = concatX(scaleBlock((Image<byte>)mc->dirChan(3).getOutput(),
                            origDim), img2);
  img2.drawArrow(Point2D(margin1, margin2), Point2D(margin1, margin1), 255);
  
  // right
  img2 = concatX(scaleBlock((Image<byte>)mc->dirChan(0).getOutput(),
                            origDim), img2);
  img2.drawArrow(Point2D(margin1, margin1), Point2D(margin2, margin1), 255);
  
  img1 = concatY(img2, img1);
  img1.drawGrid(origDim.w(), origDim.h(), 1, 1, 255);
  
  return img1;
}

// ######################################################################
void SimulationViewerStd::drawMaskOutline(Image< PixRGB<byte> >& traj,
                                          const Image<byte> mask,
                                          const PixRGB<byte>& col,
                                          const int thick,
                                          const Point2D& pos,
                                          const int radius)
{
  if (traj.initialized() == false) return; // can't draw...

  // object-shaped drawing
  Image<byte> om(mask);
  om.lowThresh(128);    // cut off fuzzy (interpolated) object boundaries
  om = contour2D8(om);  // compute binary 8-connected contour image
  int w = traj.getWidth(), h = traj.getHeight();
  Point2D ppp;
  for (ppp.j = 0; ppp.j < h; ppp.j ++)
    for (ppp.i = 0; ppp.i < w; ppp.i ++)
      if (om.getVal(ppp.i, ppp.j))  // got a contour point -> draw here
        traj.drawDisk(ppp, thick, col);  // small disk for each point
  // OBSOLETE: traj.drawCircle(pos, radius, col, thick);
}
// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
