/*!@file SimulationViewerStd.H visualize various model simulations */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SimulationViewerStd.H 4219 2005-02-17 22:04:00Z itti $
//

#ifndef SIMULATIONVIEWERSTD_H_DEFINED
#define SIMULATIONVIEWERSTD_H_DEFINED

#include "SimulationViewer.H"
#include "Pixels.H"      // for PixRGB<byte>
#include "Image.H"       // for Image
#include "ArrayData.H"   // for Dims
#include "SharedPtr.H"   // for SharedPtr
#include "Point2DT.H"    // for Point2DT
#include "WTAwinner.H"   // for WTAwinner
#include "ImageSet.H"
class Brain;
class XWindow;

//! This class provides a standard set of visualization tools
/*! Its typical usage is with a model that generates shifts of
  attention, eye and head movements. This class will handle
  visualizing those events. There are three ways of customizing it to
  your specific needs: 1) by changing the values of some of its
  internal ModelParam parameters, either via the command-line or via a
  config file for those that do not have command-line options; 2) by
  overloading (in a derived class) the low-level drawing functions in
  the protected section of the class definition; 3) by overloading the
  trigger functions in the public section of the class definition. */
class SimulationViewerStd : public SimulationViewer {
public:
  // ######################################################################
  /*! @name Constructors and destructors */
  //@{

  //! Constructor. See ModelComponent.H.
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage */
  SimulationViewerStd(ModelManager& mgr,
                      const std::string& descrName = "Simulation Viewer Std",
                      const std::string& tagName = "SimulationViewerStd");
  //! Destructor
  virtual ~SimulationViewerStd();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset SimulationViewerStd
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! Set our brain
  /*! Some drawings, in particular those involving the saliency map
    (like x-combo, warp3d, etc) require a brain */
  virtual void setBrain(Brain* brain);

  //@}

  // ######################################################################
  /*! @name Events that will trigger new drawings */
  /*! Each of these will trigger a sequence of calls to the
    lower-level functions in the protected section of this class
    definition. These sequences will depend on the current settings of
    our ModelParam parameters. */
  //@{

  //! Ready to switch to next input frame
  void newInputReady();

  //! Got a new input frame
  void newInputFrame(const Image< PixRGB<byte> >& input, const double t);

  //! Update with availability of new channel drawings
  void channelUpdate(SharedPtr<VisualCortex>& vcx);

  //! Called every simulation time step (time is in seconds)
  void evolve(const double t);

  //! Update with new covert attention position
  void attentionShift(const WTAwinner& win, const Image<byte>& foaMask);

  //! Update with new eye position
  void eyeMovement(const Point2DT& curreye, const Image<byte>& foveaMask);

  //! Update with new head position
  void headMovement(const Point2DT& currhead, const Image<byte>& headMask);

  //! Update eye saccade status
  void eyeSaccade(const bool inSaccade);

  //! Update head saccade status
  void headSaccade(const bool inSaccade);

  //! Update eye blink status
  void eyeBlink(const bool inBlink);

  //@}

  // ######################################################################
  /*! @name Output of our current internal drawings */
  //@{

  //! Get the attention/eye/head trajectory image
  Image< PixRGB<byte> > getTraj(const double t);

  //! Save our various results
  void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

protected:
  // ######################################################################
  /*! @name Derived classes should overload those to fit their cosmetic needs*/
  //@{

  //! Draw the focus of covert attention
  void drawFOA();

  //! Link previous focus of attention to current one
  void linkFOAs();

  //! Draw current eye position
  void drawEye();

  //! Link previous eye position to current one
  void linkEyes();

  //! Draw current head position
  void drawHead();

  //! Link previous head position to current one
  void linkHeads();

  //! Draw current time onto given image
  void drawTime(Image< PixRGB<byte> >& image);

  //! Draw highlights around a given position
  void drawHighlights();

  //! Draw mega combo for X display
  Image< PixRGB<byte> > drawMegaCombo();

  //! Draw motion sixpack
  Image< PixRGB<byte> > drawMotionSixPack();

  //! Draw outline of a mask (used by drawFOA, drawEye, etc)
  /*! if the mask is initialized, revert to drawing a circle centered
    at "pos" an of radius "radius" */
  void drawMaskOutline(Image< PixRGB<byte> >& traj, const Image<byte> mask,
                       const PixRGB<byte>& col, const int thick,
                       const Point2D& pos, const int radius);
  //@}

protected:
  // ##########what to compute and show parameters:
  ModelParam<bool> itsSaveTraj;
  ModelParam<bool> itsSaveXcombo;
  ModelParam<bool> itsSaveYcombo;
  ModelParam<bool> itsSaveTRMXcombo;
  ModelParam<bool> itsSaveTRMYcombo;
  ModelParam<bool> itsWarp3D;
  ModelParam<bool> itsMegaCombo;
  ModelParam<bool> itsMotionSixPack;
  ModelParam<Dims> itsCropFOA;
  ModelParam<bool> itsXwindow;
  ModelParam<bool> itsFoveateTraj;

  // ########## display selection parameters:
  ModelParam<bool> itsDisplayFOA;
  ModelParam<bool> itsDisplayPatch;
  ModelParam<bool> itsDisplayFOALinks;
  ModelParam<bool> itsDisplayEye;
  ModelParam<bool> itsDisplayEyeLinks;
  ModelParam<bool> itsDisplayHead;
  ModelParam<bool> itsDisplayHeadLinks;
  ModelParam<bool> itsDisplayTime;
  ModelParam<bool> itsDisplayAdditive;
  ModelParam<bool> itsDisplayHighlights;
  ModelParam<bool> itsDisplayBoring;
  ModelParam<bool> itsDisplayShapeEstimator;

  // ########## drawing color selection parameters:
  ModelParam< PixRGB<byte> > itsColorNormal;

  //! This gets subtracted (with clamping) from the current color if boring
  ModelParam< PixRGB<byte> > itsColorBoring;
  ModelParam< PixRGB<byte> > itsColorBlink;
  ModelParam< PixRGB<byte> > itsColorSaccade;
  ModelParam< PixRGB<byte> > itsColorLink;

  // ########## misc default values parameters:
  ModelParam<int> itsHighlightMax;
  ModelParam<float> itsShapeEstimatorBaseContrast;
  ModelParam<int> itsShapeEstimatorBaseBright;

  ModelParam<float> itsWarp3Dpitch;
  ModelParam<float> itsWarp3Dyaw;
  ModelParam<float> itsWarp3DpitchRate;
  ModelParam<float> itsWarp3DyawRate;
  ModelParam<float> itsWarp3DpitchMax;
  ModelParam<float> itsWarp3DyawMax;

  // ########## general model parameters, shared with other components:
  ModelParam<int> itsFOAradius;
  ModelParam<int> itsFoveaRadius;
  ModelParam<int> itsHeadRadius;
  ModelParam<int> itsMultiRetinaDepth;

  // ########## parameters for drawings (not exported as command-line opts):
  ModelParam<int> itsFOApsiz;     //!< half-size of (filled square) FOA patch
  ModelParam<int> itsFOAthick;    //!< line thickness for FOA outline
  ModelParam<int> itsFOAlinkThick;//!< line thickness for links bewteen FOAs
  ModelParam<int> itsFOVpsiz;     //!< half-size of (hollow square) FOV patch
  ModelParam<int> itsFOVpthick;   //!< line thickness for FOV patch
  ModelParam<int> itsFOVthick;    //!< line thickness for FOV outline
  ModelParam<int> itsFOVlinkThick;//!< line thickness for links btw FOVs
  ModelParam<int> itsHEDpsiz;     //!< half-size of (hollow-square) head patch
  ModelParam<int> itsHEDpthick;   //!< line thickness for head patch
  ModelParam<int> itsHEDthick;    //!< line thickness for head outline
  ModelParam<int> itsHEDlinkThick;//!< line thickness for links btw heads

  ModelParam<bool> itsUseLargerDrawings;  //!< use larger drawings?

  //! Get started:
  virtual void start1();

private:
  friend class SimulationViewerEyeSim;
  Brain *itsBrain;                // hook to brain or null
  Image< PixRGB<byte> > itsInput; // clean input image
  Image< PixRGB<byte> > itsTraj;  // all our drawings go here
  WTAwinner itsCurrFOA;           // current attention position
  Image<byte> itsCurrFOAmask;     // current FOA shape
  Image<byte> itsCumFOAmask;      // cumulative FOA shape
  Point2DT itsCurrEye;            // current eye position
  Image<byte> itsCurrFoveaMask;   // current fovea shape
  Image<byte> itsCumFoveaMask;    // cumulative fovea shape
  Point2DT itsCurrHead;           // current head position
  Image<byte> itsCurrHeadMask;    // current head shape
  Image<byte> itsCumHeadMask;     // cumulative head shape

  WTAwinner itsPrevFOA;           // previous attention position
  Point2DT itsPrevEye;            // previous eye position
  Point2DT itsPrevHead;           // previous head position
  double itsCurrTime;             // current time
  bool itsEyeSaccade;             // are we in eye saccade?
  bool itsHeadSaccade;            // are we in head saccade?
  bool itsEyeBlink;               // are we in eye blink?
  ImageSet< PixRGB<byte> > itsMultiTraj; // used to foveate traj
  Dims itsDims3D;                 // dims of 3D drawing
  mutable float itsPitch3D, itsYaw3D; // 3D drawing; mutable for testsuite
  SharedPtr<XWindow> itsWin;      // a window for displays
  bool itsHasNewInput;            // input changed since last getTraj()
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
