/*!@file SingleChannel.C Channel for a single stream of processing. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: SingleChannel.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "SingleChannel.H"
#include "MathFunctions.H"
#include "ModelManager.H"
#include "fancynorm.H"
#include "Image.H"
#include "Image_ShapeOps.H"
#include "Image_Transforms.H" // for learningCoeff()
#include "Image_MathOps.H"    // for binaryReverse()
#include "Image_FilterOps.H"  // for single-opponent's centerSurround()
#include "ImageSet_Ops.H"
#include "Jet.H"
#include "log.H"
#include "ParamMap.H"
#include "Pyramid_Ops.H"
#include "PyrBuilder.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"

#include <cmath>

// ######################################################################
SingleChannel::SingleChannel(ModelManager& mgr, const std::string& descrName,
                             const std::string& tagName,
                             const VisualFeature vs,
                             SharedPtr<PyrBuilder<float> > pbuild) :
  ChannelBase(mgr, descrName, tagName, vs),
  itsTakeAbs(this, "SingleChannelTakeAbs", false),
  itsNormalizeOutput(this, "SingleChannelNormalizeOutput", false),
  itsUseRandom(this, "UseRandom", true),  // see ModelOptionDefs.C
  itsScaleNoiseToMax(this, "SingleChannelScaleNoiseToMax", false),
  itsLowThresh(this, "SingleChannelLowThresh", 0.0F),
  itsRectifyPyramid(this, "SingleChannelRectifyPyramid", false),
  itsUseSplitCS(this, "SingleChannelUseSplitCS", false), // ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()),    // see ModelOptionDefs.C
  itsNormType(this, "MaxNormType", VCXNORM_FANCY), // see ModelOptionDefs.C
  itsQlen(this, "SingleChannelQueueLen", 1),       // see ModelOptionDefs.C
  itsUseOlderVersion(this, "UseOlderVersion", true), // see ModelOptionDefs.C 
  itsTimeDecay(this, "SingleChannelTimeDecay", 20.0), // see ModelOptionDefs.C
  itsSaveRawMaps(this, "SingleChannelSaveRawMaps", false),
  itsSaveFeatureMaps(this, "SingleChannelSaveFeatureMaps", false),
  itsSaveOutputMap(this, "SingleChannelSaveOutputMap", false),
  itsPq(),
  itsCoeff(NULL),
  itsMean(NULL),
  itsSigma(NULL),
  itsListener(),
  itsOutputCache(),
  itsSubmapCache(NULL),
  itsPyrBuilder(pbuild),
  itsClipPyr(),
  itsUseSingleOpponence(false),
  itsPq2()
{  }

// ######################################################################
SingleChannel::~SingleChannel()
{ }

// ######################################################################
void SingleChannel::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsUseRandom);
    manager->requestOption(this, itsNormType);
    manager->requestOption(this, itsQlen);
    manager->requestOption(this, itsUseOlderVersion);
    manager->requestOption(this, itsTimeDecay);
    manager->requestOption(this, itsUseSplitCS);
  }
  if (whatToExport & OPTEXP_SAVE) {
    manager->requestOption(this, itsSaveRawMaps);
    manager->requestOption(this, itsSaveFeatureMaps);
    manager->requestOption(this, itsSaveOutputMap);
  }

  // the base class will propagate this call to our subcomponents:
  ChannelBase::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SingleChannel::start1()
{
  int maxind = itsLevelSpec.getVal().maxIndex();
  itsCoeff = new double[maxind];
  itsMean = new double[maxind]; 
  itsSigma = new double[maxind];
  itsSubmapCache = new Image<float>[maxind];
  for (int i = 0; i < maxind; ++i)
    { itsCoeff[i] = 1.0; itsMean[i] = 0.0; itsSigma[i] = 0.0; }

  // in the older version, we used to set the map range as we would
  // also apply spatial competition for salience to the output map,
  // only if using the MAXNORM type of competition, and otherwise we
  // would not touch the range:
  if (itsUseOlderVersion.getVal() && itsNormType.getVal() != VCXNORM_MAXNORM)
    setOutputRange(0.0f, 0.0f);
}

// ######################################################################
void SingleChannel::stop2()
{
  delete [] itsSubmapCache;
  delete [] itsCoeff;
  delete [] itsMean;  
  delete [] itsSigma;
}

// ######################################################################
void SingleChannel::reset(const bool recurse)
{
  // reset some stuff for SingleChannel
  itsPyrBuilder->reset();
  itsPq.clear(); itsPq2.clear();
  itsClipPyr.clear();
  itsOutputCache.freeMem();

  for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i) 
    { 
      itsCoeff[i] = 1.0; 
      itsMean[i] = 0.0;
      itsSigma[i] = 0.0; 
    }

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void SingleChannel::readFrom(const ParamMap& pmap)
{
  ChannelBase::readFrom(pmap);

  for (uint delta = itsLevelSpec.getVal().delMin();
       delta <= itsLevelSpec.getVal().delMax(); ++delta)
    for (uint ctr = itsLevelSpec.getVal().levMin();
         ctr <= itsLevelSpec.getVal().levMax(); ++ctr)
      {
        char name[256];
        snprintf(name, 256, "coeff(%d,%d)", ctr, ctr + delta);
        if (pmap.hasParam(name))
          {
            if (pmap.isLeaf(name))
              {
                setCoeff(ctr, ctr+delta, 1.0);
                setMean(ctr, ctr+delta, pmap.getDoubleParam(name));
                setSigma(ctr, ctr+delta, 0.0);
              }
            else
              {
                SharedPtr<ParamMap> submap = pmap.getSubmap(name);
                if (submap->hasParam("coeff"))
                  setCoeff(ctr, ctr+delta, submap->getDoubleParam("coeff"));
                setMean(ctr, ctr+delta, submap->getDoubleParam("mean"));
                setSigma(ctr, ctr+delta, submap->getDoubleParam("sigma"));
              }
          }
        else
          {
            // coeff not available in pmap
            setCoeff(ctr, ctr+delta, 1.0);
            setMean(ctr, ctr+delta, 0.0);
            setSigma(ctr, ctr+delta, 0.0);
          }
      }
}

// ######################################################################
void SingleChannel::writeTo(ParamMap& pmap) const
{
  ChannelBase::writeTo(pmap);

  for (uint delta = itsLevelSpec.getVal().delMin();
       delta <= itsLevelSpec.getVal().delMax(); ++delta)
    for (uint ctr = itsLevelSpec.getVal().levMin();
         ctr <= itsLevelSpec.getVal().levMax(); ++ctr)
      {
        char pname[256], name[256];
        snprintf(pname, 256, "coeff(%d,%d)", ctr, ctr+delta);
        SharedPtr<ParamMap> submap(new ParamMap);
        snprintf(name, 256, "coeff");
        submap->putDoubleParam(name, getCoeff(ctr, ctr+delta));
        snprintf(name, 256, "mean");
        submap->putDoubleParam(name, getMean(ctr, ctr+delta));
        snprintf(name, 256, "sigma");
        submap->putDoubleParam(name, getSigma(ctr, ctr+delta));
        pmap.putSubmap(pname, submap);
      }
}

// ######################################################################
void SingleChannel::setListener(SharedPtr<ChannelListener> listener)
{ itsListener = listener; }

// ######################################################################
uint SingleChannel::numChans() const
{ return 0; }

// ######################################################################
void SingleChannel::cleanCoeffs(double eps)
{
  killCaches();

  for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i)
    {
      if (itsCoeff[i] <= eps) itsCoeff[i] = 0.0;
      if (itsMean[i] <= eps) itsMean[i] = 0.0;      
      if (itsSigma[i] <= eps) itsSigma[i] = 0.0;
    }
}

// ######################################################################
void SingleChannel::clampCoeffs(double cmin, double cmax)
{
  killCaches();

  for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i)
    {
      itsCoeff[i] = CLAMP(itsCoeff[i], cmin, cmax);
      itsMean[i] =  CLAMP(itsMean[i], cmin, cmax);
      itsSigma[i] = CLAMP(itsSigma[i], cmin, cmax);
    }
}

// ######################################################################
double SingleChannel::absSumCoeffs() const
{
  double sum = 0.0;
  for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i)
    sum += std::abs(itsCoeff[i]);
  return sum;
}

// ######################################################################
void SingleChannel::normalizeCoeffs(const double div)
{
  killCaches();
  for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i)
    itsCoeff[i] /= div;
}

// ######################################################################
void SingleChannel::setCoeff(const uint cntr, const uint surr,
                             const double val)
{
  killCaches();
  itsCoeff[csToIndex(cntr, surr)] = val;
}
// ######################################################################
void SingleChannel::setMean(const uint cntr, const uint surr,
                             const double val)
{
  killCaches();
  itsMean[csToIndex(cntr, surr)] = val;
}
// ######################################################################

void SingleChannel::setSigma(const uint cntr, const uint surr,
                             const double val)
{
  killCaches();
  itsSigma[csToIndex(cntr, surr)] = val;
}

// ######################################################################
double SingleChannel::getCoeff(const uint cntr, const uint surr) const
{ return itsCoeff[csToIndex(cntr, surr)]; }

// ######################################################################
double SingleChannel::getMean(const uint cntr, const uint surr) const
{ return itsMean[csToIndex(cntr, surr)]; }

// ######################################################################
double SingleChannel::getSigma(const uint cntr, const uint surr) const
{ return itsSigma[csToIndex(cntr, surr)]; }

// ######################################################################
void SingleChannel::learn(const Image<byte>& dmap, const double eta,
                          const bool softmask,
                          const int inthresh, const int outthresh)
{
  killCaches();

  for (uint del = itsLevelSpec.getVal().delMin();
       del <= itsLevelSpec.getVal().delMax(); ++del)
    for (uint lev = itsLevelSpec.getVal().levMin();
         lev <= itsLevelSpec.getVal().levMax(); ++lev)
      {
        uint idx = csToIndex(lev, lev+del);
        Image<float> fmap = getSubmap(idx);
        double oldcoeff = itsCoeff[idx];
        
        itsCoeff[idx] += eta * 
          learningCoeff(fmap, rescale(dmap, fmap.getDims()),
                        softmask, inthresh, outthresh);
        LINFO("%s(%d,%d): %f -> %f", tagName().c_str(), lev, lev+del,
              oldcoeff, itsCoeff[idx]);
      }
}

// ######################################################################
bool SingleChannel::outputAvailable() const
{ return itsPq.empty() == false; }

// ######################################################################
const Dims& SingleChannel::getMapDims() const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");
  return itsPq.front().pyr[itsLevelSpec.getVal().mapLevel()].getDims();
}

// ######################################################################
const Dims& SingleChannel::getInputDims() const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");
  return itsPq.front().pyr[0].getDims();
}

// ######################################################################
const Image<float>& SingleChannel::getImage(const uint lev) const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");
  return itsPq.front().pyr.getImage(lev);
}

// ######################################################################
Image<float> SingleChannel::centerSurround(const uint cntrlev,
                                           const uint surrlev) const
{
  ASSERT(itsLevelSpec.getVal().csOK(cntrlev, surrlev));
  if (outputAvailable() == false) LFATAL("I have no output yet!");

  // in the single-opponent version, we do pyr[cntrlev] - pyr2[surrlev]
  if (itsUseSingleOpponence)
    {
      // do basic center-surround for the front (latest) pyramid in queue:
      const ImageSet<float>& cpyr = itsPq.front().pyr;
      const ImageSet<float>& spyr = itsPq2.front().pyr;
      double t = itsPq.front().t;

      // compute single-opponent center-surround:
      Image<float> cs =
        ::centerSurroundSingleOpponent(cpyr, spyr, cntrlev, surrlev,
                                       itsTakeAbs.getVal(), &itsClipPyr);

      // do additional processing with other pyramids in queue:
      for (uint i = 1; i < itsPq.size(); ++i)
        {
          const ImageSet<float>& cpyr2 = itsPq[i].pyr;
          const ImageSet<float>& spyr2 = itsPq2[i].pyr;
          double t2 = itsPq[i].t;
          
          // compute a decay factor based on how old the second pyramid is
          // compared to the latest one:
          float fac = exp( (t2 - t) * itsTimeDecay.getVal());
          cs +=
            ::centerSurroundDiffSingleOpponent(cpyr, cpyr2, spyr, spyr2,
                                               cntrlev, surrlev,
                                               itsTakeAbs.getVal(),
                                               &itsClipPyr) * fac;
        }
      return cs;
    }
  else
    {
      // standard center-surround uses only one pyramid, in itsPq:

      // do basic center-surround for the front (latest) pyramid in queue:
      const ImageSet<float>& pyr = itsPq.front().pyr;
      double t = itsPq.front().t;
      
      // compute center-surround:
      Image<float> cs = ::centerSurround(pyr, cntrlev, surrlev,
                                         itsTakeAbs.getVal(), &itsClipPyr);
      
      // do additional processing with other pyramids in queue:
      for (uint i = 1; i < itsPq.size(); ++i)
        {
          const ImageSet<float>& pyr2 = itsPq[i].pyr;
          double t2 = itsPq[i].t;
          
          // compute a decay factor based on how old the second pyramid is
          // compared to the latest one:
          float fac = exp( (t2 - t) * itsTimeDecay.getVal());
          cs += ::centerSurroundDiff(pyr, pyr2, cntrlev, surrlev,
                                     itsTakeAbs.getVal(), &itsClipPyr) * fac;
        }
      
      return cs;
    }
}

// ######################################################################
void SingleChannel::centerSurround(const uint cntrlev, const uint surrlev,
                                   Image<float>& pos, Image<float>& neg) const
{
  ASSERT(itsLevelSpec.getVal().csOK(cntrlev, surrlev));
  if (outputAvailable() == false) LFATAL("I have no output yet!");

  if (itsUseSingleOpponence)
    {
      // do basic center-surround for the front (latest) pyramid in queue:
      const ImageSet<float>& cpyr = itsPq.front().pyr;
      const ImageSet<float>& spyr = itsPq2.front().pyr;
      double t = itsPq.front().t;
      
      // compute center-surround:
      ::centerSurroundSingleOpponent(cpyr, spyr, cntrlev, surrlev,
                                     pos, neg, &itsClipPyr);
      
      // do additional processing with other pyramids in queue:
      for (uint i = 1; i < itsPq.size(); i ++)
        {
          const ImageSet<float>& cpyr2 = itsPq[i].pyr;
          const ImageSet<float>& spyr2 = itsPq2[i].pyr;
          double t2 = itsPq[i].t;
          
          // compute a decay factor based on how old the second pyramid is
          // compared to the latest one:
          float fac = exp( (t2 - t) * itsTimeDecay.getVal());
          Image<float> pos2, neg2;
          ::centerSurroundDiffSingleOpponent(cpyr, cpyr2, spyr, spyr2,
                                             cntrlev, surrlev,
                                             pos2, neg2, &itsClipPyr);
          pos += pos2 * fac;
          neg += neg2 * fac;
        }
    }
  else
    {
      // do basic center-surround for the front (latest) pyramid in queue:
      const ImageSet<float>& pyr = itsPq.front().pyr;
      double t = itsPq.front().t;
      
      // compute center-surround:
      ::centerSurround(pyr, cntrlev, surrlev, pos, neg, &itsClipPyr);
      
      // do additional processing with other pyramids in queue:
      for (uint i = 1; i < itsPq.size(); i ++)
        {
          const ImageSet<float>& pyr2 = itsPq[i].pyr;
          double t2 = itsPq[i].t;
          
          // compute a decay factor based on how old the second pyramid is
          // compared to the latest one:
          float fac = exp( (t2 - t) * itsTimeDecay.getVal());
          Image<float> pos2, neg2;
          ::centerSurroundDiff(pyr, pyr2, cntrlev, surrlev,
                               pos2, neg2, &itsClipPyr);
          pos += pos2 * fac;
          neg += neg2 * fac;
        }
    }
}

// ######################################################################
void SingleChannel::fillJet(const Point2D& loc, Jet<float>& j,
                            std::vector<int>& index) const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");

  // A trivial implementation would go as follows for the RAW case
  // (modulo some checking that the jet can accept the indices we
  // have):
  /*
  for (uint i = 0; i < itsLevelSpec.getVal().levMax() +
       itsLevelSpec.getVal().delMax(); i++)
    {
      float pix = itsPq.front().pyr->getPixel(loc, i);
      index.push_back(i);
      j.setVal(pix, visualFeature(), RAW, index);
      index.pop_back();
    }
  */

  // The implementation below is slightly faster by minimizing the
  // number of Jet index computations, using only one getIndexV() call
  // and then raw Jet array accesses. But it is identical in spirit.
  int rmin, rmax;
  if (j.getIndexRange(visualFeature(), RAW, index.size(), rmin, rmax))
    {
      index.push_back(rmin);  // raw features use a single index
      int idx = j.getSpec()->getIndexV(visualFeature(), RAW, index);
      int ns = int(itsLevelSpec.getVal().levMax() +
                   itsLevelSpec.getVal().delMax());
      int start = std::max(rmin, 0), stop = std::min(rmax, ns);

      // We have data for scales [0..ns]; the Jet wants data for
      // scales [rmin..rmax] and we will put zeros in the Jet for all
      // Jet scales that it has but we don't have
      idx -= rmin;  // offset to account for rmin
      for (int i = rmin; i < 0; ++i)
        j.setVal(idx + i, 0.0f);

      for (int i = start; i <= stop; ++i)
        {
          float pix = 0.0;
          if (itsPq.empty() == false)
            pix = itsPq.front().pyr.getPixel(loc, i);
          j.setVal(idx + i, pix);
        }
      for (int i = ns + 1; i <= rmax; i ++) j.setVal(idx + i, 0.0f);
      index.pop_back();
    }

  //FIXME: also do it for RAW_CS and NORM_CS
}

// ######################################################################
void SingleChannel::fillJetNI(const Point2D& loc, Jet<float>& j,
                              std::vector<int>& index) const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");

  int rmin, rmax;
  if (j.getIndexRange(visualFeature(), RAW, index.size(), rmin, rmax))
    {
      index.push_back(rmin);  // raw features use a single index
      int idx = j.getSpec()->getIndexV(visualFeature(), RAW, index);
      int ns = int(itsLevelSpec.getVal().levMax() +
                   itsLevelSpec.getVal().delMax());
      int start = std::max(rmin, 0), stop = std::min(rmax, ns);

      // We have data for scales [0..ns]; the Jet wants data for
      // scales [rmin..rmax] and we will put zeros in the Jet for all
      // Jet scales that it has but we don't have
      idx -= rmin;  // offset to account for rmin
      for (int i = rmin; i < 0; ++i)
        j.setVal(idx + i, 0.0f);

      for (int i = start; i <= stop; ++i)
        {
          float pix = 0.0;
          if (itsPq.empty() == false)
            pix = itsPq.front().pyr.getPixelNI(loc, i);
          j.setVal(idx + i, pix);
        }
      for (int i = ns + 1; i <= rmax; i ++) j.setVal(idx + i, 0.0f);
      index.pop_back();
    }

  //FIXME: also do it for RAW_CS and NORM_CS
}

// ######################################################################
void SingleChannel::fillJetSpec(JetSpec &js) const
{
  js.addIndexRange(visualFeature(), RAW, 0,
                   itsLevelSpec.getVal().levMax() +
                   itsLevelSpec.getVal().delMax());
  /*  FIXME: FUTURE...
  js.addIndexRange(visualFeature(), RAW_CS,
                   itsLevelSpec.getVal().levMin(),
                   itsLevelSpec.getVal().levMax());
  js.addIndexRange(visualFeature(), RAW_CS,
                   itsLevelSpec.getVal().delMin(),
                   itsLevelSpec.getVal().delMax());
  js.addIndexRange(visualFeature(), NORM_CS,
                   itsLevelSpec.getVal().levMin(),
                   itsLevelSpec.getVal().levMax());
  js.addIndexRange(visualFeature(), NORM_CS,
                   itsLevelSpec.getVal().delMin(),
                   itsLevelSpec.getVal().delMax());
  */
}

// ######################################################################
uint SingleChannel::numSubmaps() const
{ return itsLevelSpec.getVal().maxIndex(); }

// ######################################################################
Image<float> SingleChannel::getSubmap(const uint idx) const
{
  // cache the results for later use if we don't already have cached it:
  if (itsSubmapCache[idx].initialized() == false)
    itsSubmapCache[idx] = computeSubmap(idx);

  return itsSubmapCache[idx];
}

// ######################################################################
Image<float> SingleChannel::computeSubmap(const uint idx) const
{
  // default implementation is just to get the raw CS map and
  // post-process it:
  return postProcessMap(getRawCSmap(idx), idx);
}

// ######################################################################
Image<float> SingleChannel::getRawCSmap(const uint idx) const
{
  ASSERT(itsLevelSpec.getVal().indexOK(idx));
  if (outputAvailable() == false) LFATAL("I have no output yet!");

  const float w = float(itsCoeff[idx]);
  
  // if we have a zero weight, just return an empty image:
  if (w == 0.0f) return Image<float>(getMapDims(), ZEROS);
  
  // otherwise, compute center-surround map:
  uint clev = 0, slev = 0;
  itsLevelSpec.getVal().indexToCS(idx, clev, slev);
  
  Image<float> submap;
  if (itsUseSplitCS.getVal())
    {
      // if using split center-surround, compute both the positive
      // and negative submaps, normalize them, sum them and we
      // will below further normalize the sum:
      Image<float> subpos, subneg, subpos2, subneg2;
      
      // positive and negative submaps are computed from a
      // single center-surround difference:
      this->centerSurround(clev, slev, subpos, subneg);
      
      // also compute maps containing max(map)-map:
      float pmin, pmax, nmin, nmax;
      subpos.getMinMax(pmin, pmax); subpos2 = binaryReverse(subpos, pmax);
      subneg.getMinMax(nmin, nmax); subneg2 = binaryReverse(subneg, nmax);
      
      // apply spatial competition for salience, independently to
      // both positive and negative parts of the submap,
      // preserving their original range rather than first
      // normalizing them to [MAXNORMMIN..MAXNORMMAX] as we
      // usually do, so that those maps who contain nothing do not
      // get artificially amplified:
      subpos = maxNormalize(subpos, 0.0f, 0.0f, itsNormType.getVal());
      subneg = maxNormalize(subneg, 0.0f, 0.0f, itsNormType.getVal());
      subpos2 = maxNormalize(subpos2, 0.0f, 0.0f, itsNormType.getVal());
      subneg2 = maxNormalize(subneg2, 0.0f, 0.0f, itsNormType.getVal());
      
      // the raw submap is the sum of the normalized positive and
      // negative sides:
      submap = subpos + subneg + subpos2 + subneg2;
    }
  else
    // submap is computed from a center-surround difference:
    submap = this->centerSurround(clev, slev);

  // print some debug info if in debug mode:
  if (manager->debugMode())
    {
      float mi, ma; submap.getMinMax(mi, ma);
      LDEBUG("%s(%d,%d): raw range [%f .. %f]", tagName().c_str(),
             clev, slev, mi, ma);
    }

  return submap;
}
  
// ######################################################################
Image<float> SingleChannel::postProcessMap(const Image<float>& smap,
                                           const uint idx) const
{
  Image<float> submap(smap);
  uint clev = 0, slev = 0;
  itsLevelSpec.getVal().indexToCS(idx, clev, slev);

  // resize submap to fixed scale if necessary:
  if (submap.getWidth() > getMapDims().w())
    submap = downSize(submap, getMapDims());
  else if (submap.getWidth() < getMapDims().w())
    submap = rescale(submap, getMapDims());
  
  // add noise if wanted:
  if (itsUseRandom.getVal())
    {
      // check if we want to scale noise to maximum of image
      if (itsScaleNoiseToMax.getVal())
        {
          float fMin, fMax; submap.getMinMax(fMin, fMax);
          submap.addBGnoise(fMax);
        }
      else
        submap.addBGnoise(255.0);
    }
  
  // if using the older version: first normalize the submap to a
  // fixed dynamic range and then apply spatial competition for
  // salience to the submap; otherwise, just apply competition:
  if (itsUseOlderVersion.getVal())
    {
      LDEBUG("%s(%d,%d): applying %s(%f .. %f)", tagName().c_str(), clev, slev,
             maxNormTypeName(itsNormType.getVal()), MAXNORMMIN, MAXNORMMAX);
      submap = maxNormalize(submap, MAXNORMMIN, MAXNORMMAX,
                            itsNormType.getVal());
    }
  else
    {
      LDEBUG("%s(%d,%d): applying %s(0.0 .. 0.0)", tagName().c_str(),
             clev, slev, maxNormTypeName(itsNormType.getVal()));
      submap = maxNormalize(submap, 0.0f, 0.0f, itsNormType.getVal());
    }

  // let our listeners know we have a new submap:
  if (itsListener.get() != NULL)
    itsListener->onEachCS(*this, clev, slev-clev, submap);

  // print some debug info if in debug mode:
  if (manager->debugMode())
    {
      float mi, ma; submap.getMinMax(mi, ma);
      LDEBUG("%s(%d,%d): final range [%f .. %f]",
             tagName().c_str(), clev, slev, mi, ma);
    }

  return submap;
}

// ######################################################################
const char* SingleChannel::getSubmapName(const uint idx) const
{
  ASSERT( itsLevelSpec.getVal().indexOK(idx) );

  uint clev = 0, slev = 0;
  itsLevelSpec.getVal().indexToCS(idx, clev, slev);

  static char buf[256];
  snprintf(buf, 256, "%s lev: %d delta: %d",
           descriptiveName().c_str(), clev, slev-clev);

  return buf;
}

// ######################################################################
void SingleChannel::getCoeff(std::vector<double>& coeff)
{
  for (uint idx = 0; idx < numSubmaps(); idx ++)
    {
      // get center and surround scales for this submap index:
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(idx, clev, slev);
      // find the coefficient for this submap as a function of the
      // amount of signal present
      double sum = 0.0;
      Image<float> submap = getSubmap(idx);
      uint w = submap.getWidth(), h = submap.getHeight();
      for (uint i = 0; i < w; i++)
        for (uint j = 0; j < h; j++){
          double salience = submap.getVal(i,j);
          sum += salience*salience*salience;
        }
      sum /= w*h;
      LINFO("%s(%d,%d): -- amount of signal = %lf", tagName().c_str(), 
            clev, slev,sum);
      coeff.push_back(sum);
    }
}

// ######################################################################
void SingleChannel::setGroupingCoeff(std::vector<double>& coeff, uint& idx)
{
  uint num = numSubmaps();
  for (uint i = 0; i < num; i++)
    {
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(i, clev, slev);
      setCoeff(clev, slev, coeff[idx]);
      idx ++;
    }
}

// ######################################################################
Image<float> SingleChannel::perceptualGrouping()
{
  Image<float> myPercept(getMapDims(), ZEROS);

  // compute a weighted sum of raw feature maps at all levels:
  for (uint idx = 0; idx < itsLevelSpec.getVal().maxIndex(); ++idx)
    {
      const float w = float(itsCoeff[idx]);     // weight for that submap
      if (w != 0.0f)
        {
          Image<float> submap = getRawCSmap(idx); // get raw map
          if (w != 1.0f) submap *= w;            // weigh the submap
          // resize submap to fixed scale if necessary:
          if (submap.getWidth() > getMapDims().w())
            submap = downSize(submap, getMapDims());
          else if (submap.getWidth() < getMapDims().w())
            submap = rescale(submap, getMapDims());
          myPercept += submap;                  // add submap to our sum
        }
    }
  return myPercept;
}

// ######################################################################
void SingleChannel::biasFeatures(const double* mean, const double* sigma,
                                 int& idx)
{
  uint num = numSubmaps();
  for (uint i = 0; i < num; i++)
    {
      uint subidx = idx % num;
      ASSERT( itsLevelSpec.getVal().indexOK(subidx) );

      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(subidx, clev, slev);
      setMean(clev, slev, mean[idx]);
      setSigma(clev, slev, sigma[idx]);
      idx ++;
    }
}

// ######################################################################
void SingleChannel::getFeatures(const Point2D& locn,
                                std::vector<double>& mean) const
{
  if (outputAvailable() == false)
    {
      CLDEBUG("I have no output yet -- RETURNING ZEROS");
      for (uint idx = 0; idx < numSubmaps(); idx ++) mean.push_back(0.0);
      return;
    }

  // The coordinates we receive are at the scale of the original
  // image, and we will need to rescale them to the size of the
  // various submaps we read from. The first image in our first
  // pyramid has the dims of the input:
  ImageSet<float> pyr = itsPq.front().pyr;
  float iw = float(pyr[0].getWidth());
  float ih = float(pyr[0].getHeight());

  for (uint idx = 0; idx < numSubmaps(); idx ++)
    {
      // get center and surround scales for this submap index:
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(idx, clev, slev);

      // read center value with bilinear interpolation:
      const float cx = std::max(0.0f, (float(locn.i) + 0.5f) *
                                float(pyr[clev].getWidth()) / iw - 0.5f);
      const float cy = std::max(0.0f, (float(locn.j) + 0.5f) *
                                float(pyr[clev].getHeight()) / ih - 0.5f);
      const float cval = pyr[clev].getValInterp(cx, cy);

      // read surround value with bilinear interpolation:
      const float sx = std::max(0.0f, (float(locn.i) + 0.5f) *
                                float(pyr[slev].getWidth()) / iw - 0.5f);
      const float sy = std::max(0.0f, (float(locn.j) + 0.5f) *
                                float(pyr[slev].getHeight()) / ih - 0.5f);
      float sval;
      if (itsUseSingleOpponence)  // take from surround pyramid:
        sval = itsPq2.front().pyr[slev].getValInterp(sx, sy);
      else
        sval = pyr[slev].getValInterp(sx, sy);
    
      // compute center - surround and take absolute value if our
      // channel wants that:
      float val = cval - sval; if (itsTakeAbs.getVal()) val = fabs(val);

      // store the submap value at the chosen location:
      mean.push_back(double(val));
    }
}
// ######################################################################

void SingleChannel::getFeaturesBatch(std::vector<Point2D*> *locn,
                                     std::vector<std::vector<float> > *mean,
                                     int *count) const
{
  if (outputAvailable() == false)
  {
    CLDEBUG("I have no output yet -- RETURNING ZEROS");
    for (uint idx = 0; idx < numSubmaps(); idx ++)
    {
      std::vector<std::vector<float> >::iterator imean = mean->begin();
      for(int i = 0; i < *count; i++, ++imean)
        imean->push_back(0.0);
    }
    return;
  }

  // The coordinates we receive are at the scale of the original
  // image, and we will need to rescale them to the size of the
  // various submaps we read from. The first image in our first
  // pyramid has the dims of the input:
  const ImageSet<float> *pyr = &(itsPq.front().pyr);
  const ImageSet<float> *pyr2 = &(itsPq2.front().pyr);
  float iw = float(pyr->getImage(0).getWidth());
  float ih = float(pyr->getImage(0).getHeight());   
  uint clev = 0, slev = 0;
  uint sm = numSubmaps();
  for (uint idx = 0; idx < sm; idx++)
  {
    // get center and surround scales for this submap index:
    itsLevelSpec.getVal().indexToCS(idx, clev, slev);
    std::vector<Point2D*>::iterator ilocn = locn->begin();
    std::vector<std::vector<float> >::iterator imean = mean->begin();
    float pyrwic = pyr->getImage(clev).getWidth();
    float pyrhec = pyr->getImage(clev).getHeight();
    float pyrwis = pyr->getImage(slev).getWidth();
    float pyrhes = pyr->getImage(slev).getHeight();
    
    if (itsUseSingleOpponence == false)
      for(int i = 0; i < *count; i++, ++ilocn, ++imean)
      {
        // read center value with bilinear interpolation:
        // This is the same expression (mostly) used in the other 
        // get feature. It's just more compact.
        float val = (pyr->getImage(clev).getValInterp(
                                                      (std::max(0.0f, 
                                                      (((*ilocn)->i) + 0.5f) *
                                                      pyrwic / iw - 0.5f)),
                                                      (std::max(0.0f,
                                                      (((*ilocn)->j) + 0.5f) *
                                                      pyrhec / ih - 0.5f))) -
                     pyr->getImage(slev).getValInterp(
                                                      (std::max(0.0f,
                                                      (((*ilocn)->i) + 0.5f) *
                                                      pyrwis / iw - 0.5f)),
                                                      (std::max(0.0f,
                                                      (((*ilocn)->j) + 0.5f)) *
                                                       pyrhes / ih - 0.5f)));
        
        if (itsTakeAbs.getVal()) val = fabs(val); 
        // store the submap value at the chosen location:
        imean->push_back(float(val));
      }
    else
      for(int i = 0; i < *count; i++, ++ilocn, ++imean)
      {  
        // read center value with bilinear interpolation:
        float val = (pyr->getImage(clev).getValInterp(
                                                      (std::max(0.0f, 
                                                      (((*ilocn)->i) + 0.5f) *
                                                      pyrwic / iw - 0.5f)),
                                                      (std::max(0.0f,
                                                      (((*ilocn)->j) + 0.5f) *
                                                      pyrhec / ih - 0.5f))) -
                    pyr2->getImage(slev).getValInterp(     
                                                      (std::max(0.0f,
                                                      (((*ilocn)->i) + 0.5f) *
                                                      pyrwis / iw - 0.5f)),
                                                      (std::max(0.0f,
                                                      (((*ilocn)->j) + 0.5f)) *
                                                       pyrhes / ih - 0.5f)));

        if (itsTakeAbs.getVal()) val = fabs(val); 
        // store the submap value at the chosen location:
        imean->push_back(float(val)); 
    }
  }
}

// ######################################################################
void SingleChannel::findWeight()
{
  double max = 0.0, sum = 0.0;
  // initialise the submap weights
  for (uint idx = 0; idx < numSubmaps(); idx++)
    {
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(idx, clev, slev);
      double wt = getMean(clev, slev) / (1.0 + getSigma(clev, slev));
      sum += wt;
      max = std::max(max, wt);
    }
  // initialize channel's weight with the max submap weight
  setTotalWeight(max);

  // normalize submap weights so that they add to 1
  for (uint idx = 0; idx < numSubmaps(); idx++)
    {
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(idx, clev, slev);
      double wt = getMean(clev, slev) / (1.0 + getSigma(clev, slev));
      if (sum != 0.0) setCoeff (clev, slev, wt/sum);
    }

}

// ######################################################################
bool SingleChannel::reco(const ParamMap& pmap, double& mismatch)
{
  double mm = 0.0, mean = 0.0, sigma = 0.0;

  for (uint delta = itsLevelSpec.getVal().delMin();
       delta <= itsLevelSpec.getVal().delMax(); ++delta)
    for (uint ctr = itsLevelSpec.getVal().levMin();
         ctr <= itsLevelSpec.getVal().levMax(); ++ctr)
      {
        char name[256];
        snprintf(name, 256, "coeff(%d,%d)", ctr, ctr+delta);
        SharedPtr<ParamMap> submap = pmap.getSubmap(name);
        snprintf(name, 256, "mean");
        mean = submap->getDoubleParam(name);
        snprintf(name, 256, "sigma");
        sigma = submap->getDoubleParam(name);

        // find the euclidean distance between observed data and mean
        mm += pow (getMean(ctr, ctr+delta) - mean, 2);
      }
  mismatch += mm; 
  return true;
}

// ######################################################################
void SingleChannel::findLikelihood(const ParamMap& pmap, 
                                   double& LogLikelihood)
{
  double LogProb= 0.0, mean = 0.0, sigma = 0.0;

  for (uint delta = itsLevelSpec.getVal().delMin();
       delta <= itsLevelSpec.getVal().delMax(); ++delta)
    for (uint ctr = itsLevelSpec.getVal().levMin();
         ctr <= itsLevelSpec.getVal().levMax(); ++ctr)
      {
        char name[256];
        // get the mean and sigma of the object being matched
        snprintf(name, 256, "coeff(%d,%d)", ctr, ctr+delta);
        SharedPtr<ParamMap> submap = pmap.getSubmap(name);
        snprintf(name, 256, "mean");
        mean = submap->getDoubleParam(name);
        snprintf(name, 256, "sigma");
        sigma = submap->getDoubleParam(name);
        
        /* find likelihood of fixation's features assuming a normal
        distri. for the features of this object class. Since we are
        only using the likelihoods to compare and find the best match
        for recognition, we can omit the constant log(2*PI) term
        */
        double fixMean = getMean(ctr, ctr+delta);
        if (sigma != 0.0)
          {
            double z = (fixMean - mean) / sigma;
            LogProb = 0.5*z*z - log(sigma);
          }
        else 
          LogProb = 0;
        LogLikelihood += LogProb; // addition since we assume that all
                                  // spatial scales are independent
    }
}

// ######################################################################
void SingleChannel::killCaches()
{
  ChannelBase::killCaches(); // call our base class's implementation

  itsOutputCache.freeMem();

  // our caches exist only when we are in started() state:
  if (started())
    for (uint i = 0; i < itsLevelSpec.getVal().maxIndex(); ++i)
      itsSubmapCache[i] = Image<float>();
}

// ######################################################################
void SingleChannel::doInput(const Image<PixRGB<float> >*,
                            const Image<float>* bwimg,
                            const double t,
                            const Image<byte>& clipMask)
{
  ASSERT(bwimg != NULL);

  setClipPyramid(clipMask);
  storePyramid(computePyramid(*bwimg), t);
}

// ######################################################################
void SingleChannel::inputPyramid(const ImageSet<float>& pyramid, 
                                 const double t,
                                 const Image<byte>& clipMask)
{
  killCaches();
  setClipPyramid(clipMask);

  storePyramid(pyramid, t);
}

// ######################################################################
ImageSet<float> SingleChannel::computePyramid(const Image<float>& bwimg)
{
  // Compute our pyramid:
  ImageSet<float> py = itsPyrBuilder->
    build(bwimg, itsLevelSpec.getVal().maxDepth());

  // Eliminate small values if and/or rectify the pyramid desired:
  if (itsLowThresh.getVal() > 0.0f)
    {
      if (itsRectifyPyramid.getVal())
        doLowThresh(py, itsLowThresh.getVal());
      else
        doLowThreshAbs(py, itsLowThresh.getVal());
    }
  else if (itsRectifyPyramid.getVal())
    doRectify(py);

  // return pyramid:
  return py;
}

// ######################################################################
void SingleChannel::setClipPyramid(const Image<byte>& clipMask)
{
  // if necessary create pyramid of clipping masks
  if (clipMask.initialized())
    {
      itsClipPyr = buildPyrGaussian(Image<float>(clipMask)/255.0f,
                                    itsLevelSpec.getVal().maxDepth(), 9);
      doLowThresh(itsClipPyr, 1.0f, 0.0f);
    }
  else
    itsClipPyr.clear();
}

// ######################################################################
void SingleChannel::storePyramid(const ImageSet<float>& p,
                                 const double t)
{
  // check that inputs are not coming in too fast, as we may otherwise
  // run into trouble with those channels that care about the
  // inter-frame delay:
  if (itsPq.size() && t - itsPq.front().t < 0.002F) // 500fps is the max
    CLFATAL("Inputs coming in too fast! -- BAILING OUT");

  // load our pyramid into our front pyramid:
  itsPq.push_front(TPyr(p, t));

  // truncate the pyramid queue if necessary:
  while(int(itsPq.size()) > itsQlen.getVal()) itsPq.pop_back();

  // We only want dyadic pyramids here:
  ASSERT(isDyadic(itsPq.front().pyr));
}

// ######################################################################
void SingleChannel::storePyramid2(const ImageSet<float>& p,
                                  const double t)
{
  ASSERT(itsUseSingleOpponence);

  // check that inputs are not coming in too fast, as we may otherwise
  // run into trouble with those channels that care about the
  // inter-frame delay:
  if (itsPq2.size() && t - itsPq2.front().t < 0.002F) // 500fps is the max
    CLFATAL("Inputs coming in too fast! -- BAILING OUT");

  // load our pyramid into our front pyramid:
  itsPq2.push_front(TPyr(p, t));

  // truncate the pyramid queue if necessary:
  while(int(itsPq2.size()) > itsQlen.getVal()) itsPq2.pop_back();

  // We only want dyadic pyramids here:
  ASSERT(isDyadic(itsPq2.front().pyr));
}

// ######################################################################
const ImageSet<float>& SingleChannel::pyramid(const uint index) const
{
  if (outputAvailable() == false) LFATAL("I have no output yet!");
  ASSERT(index < itsPq.size());
  return itsPq[index].pyr;
}

// ######################################################################
double SingleChannel::pyramidTime(const uint index) const
{
  if (itsPq.empty()) return 0.0;
  ASSERT(index < itsPq.size());
  return itsPq[index].t;
}

// ######################################################################
const ImageSet<float>& SingleChannel::clipPyramid() const
{ return itsClipPyr; }

// ######################################################################
LevelSpec SingleChannel::getLevelSpec() const
{ return itsLevelSpec.getVal(); }

// ######################################################################
int SingleChannel::getNormType() const
{ return itsNormType.getVal(); }

// ######################################################################
Image<float> SingleChannel::combineSubMaps()
{
  Image<float> output(getMapDims(), ZEROS);

  // compute max-normalized weighted sum of center-surround at all levels:
  for (uint idx = 0; idx < itsLevelSpec.getVal().maxIndex(); ++idx)
    {
      const float w = float(itsCoeff[idx]);     // weight for that submap
      if (w != 0.0f)
        {
          Image<float> submap = getSubmap(idx); // get the unweighted map
          if (w != 1.0f) submap *= w;           // weigh the submap
          output += submap;                     // add submap to our sum

          if (manager->debugMode())
            {
              uint clev = 0, slev = 0;
              itsLevelSpec.getVal().indexToCS(idx, clev, slev);
              LDEBUG("%s(%d,%d): weight %f", tagName().c_str(), clev, slev, w);
            }
        }
    }

  // apply max-normalization on the output as needed:
  if(itsNormType.getVal() == VCXNORM_LANDMARK)
    {
      float goodness = pow(goodness_map(output), 0.1);
      LINFO("GOODNESS = %f", goodness);
      output *= goodness;
    }
  else if (itsNormalizeOutput.getVal())
    {
      LDEBUG("%s: Normalizing output: %s(%f .. %f)", tagName().c_str(),
             maxNormTypeName(itsNormType.getVal()), itsOutputRangeMin.getVal(),
             itsOutputRangeMax.getVal());

      output = maxNormalize(output, itsOutputRangeMin.getVal(),
                            itsOutputRangeMax.getVal(), itsNormType.getVal());
    }

  // print some debug info if in debug mode:
  if (manager->debugMode())
    {
      float mi, ma; output.getMinMax(mi, ma);
      LDEBUG("%s: final range [%f .. %f]", tagName().c_str(), mi, ma);
    }

  LINFO("Computed %s Conspicuity Map", descriptiveName().c_str());

  return output;
}

// ######################################################################
Image<float> SingleChannel::getOutput()
{
  if (!itsOutputCache.initialized()) itsOutputCache = combineSubMaps();

  return itsOutputCache;
}

// ######################################################################
void SingleChannel::setPyramid(SharedPtr<PyrBuilder<float> > pbuild)
{
  ASSERT(pbuild.get() != 0);
  itsPq.clear();  // forget about our old pyramids
  itsPyrBuilder = pbuild;
}

// ######################################################################
ImageSet<float>& SingleChannel::pyrMut(const uint index)
{
  ASSERT(index < itsPq.size());
  return itsPq[index].pyr;
}

// ######################################################################
uint SingleChannel::csToIndex(const uint centerlev,
                              const uint surroundlev) const
{ return itsLevelSpec.getVal().csToIndex(centerlev, surroundlev); }

// ######################################################################
void SingleChannel::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  if (outputAvailable() == false) return; // nothing to save
  
  // save raw pyramid levels?
  if (itsSaveRawMaps.getVal()) {
    const ImageSet<float>& pyr = itsPq.front().pyr;
    for (uint i = 0; i < pyr.size(); i ++)
      ofs->writeFloat(pyr[i], PNM, -1, true, "SR%s-%d-", tagName().c_str(),i);
  }

  // save center-surround feature submaps?
  if (itsSaveFeatureMaps.getVal())
    for (uint i = 0; i < numSubmaps(); i ++) {
      uint clev = 0, slev = 0;
      itsLevelSpec.getVal().indexToCS(i, clev, slev);
      ofs->writeFloat(getSubmap(i), PNM, -1, true, "SF%s-%d-%d-",
                      tagName().c_str(),clev, slev);
    }

  // save output map?
  if (itsSaveOutputMap.getVal())
    ofs->writeFloat(getOutput(), PNM, -1, true, "SO%s-", tagName().c_str());
}

// ######################################################################
void SingleChannel::useSingleOpponence(bool use)
{
  itsUseSingleOpponence = use;
}

// ######################################################################
void SingleChannel::singleOpponentInput(const ImageSet<float>& centerPyr,
                                        const ImageSet<float>& surroundPyr,
                                        const double t,
                                        const Image<byte>& clipMask)
{
  ASSERT(itsUseSingleOpponence);
  killCaches();
  setClipPyramid(clipMask);
  storePyramid(centerPyr, t);
  storePyramid2(surroundPyr, t);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
