/*!@file SingleChannel.H Channel for a single stream of processing. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Rob Peters <rjpeters@klab.caltech.edu>
// $Id: SingleChannel.H 4228 2005-02-23 17:19:25Z vidhya $
//

#ifndef SINGLECHANNEL_H_DEFINED
#define SINGLECHANNEL_H_DEFINED

#include "ChannelBase.H"
#include "SharedPtr.H"
#include "ModelParam.H"
#include "LevelSpec.H"
#include "ImageSet.H"
#include "PyrBuilder.H"
#include "fancynorm.H" // for MaxNormType
#include <deque>

// ######################################################################
//! SingleChannel represents a single stream of processing.
/*! The processing is implemented by an internal PyrBuilder object of some
    kind. */
class SingleChannel : public ChannelBase
{
public:
  //! Constructor. See ChannelBase.H
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage
      @param vs The VisualFeature implemented by the channel
      @param pyr The type of pyramid that should be used. */
  SingleChannel(ModelManager& mgr, const std::string& descrName,
                const std::string& tagName, const VisualFeature vs,
                SharedPtr<PyrBuilder<float> > pyr);

  //! destructor
  virtual ~SingleChannel();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset SingleChannel
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //! Send an already computed pyramid as input to the channel.
  virtual void inputPyramid(const ImageSet<float>& pyramid, const double t,
                            const Image<byte>& clipMask = Image<byte>());

  //! Computes the channel's pyramid but doesn't store it
  virtual ImageSet<float> computePyramid(const Image<float>& bwimg);

  virtual void readFrom(const ParamMap& pmap);

  virtual void writeTo(ParamMap& pmap) const;

  virtual void setListener(SharedPtr<ChannelListener> listener);

  virtual uint numChans() const;

  virtual void cleanCoeffs(const double eps);

  virtual void clampCoeffs(const double cmin, const double cmax);

  virtual double absSumCoeffs() const;

  virtual void normalizeCoeffs(const double div);

  //! Set the weighting coefficient for the given center/surround level.
  virtual void setCoeff(const uint cntr, const uint surr, const double val);

  //! Set the mean feature response for the given center/surround
  //! level.
  virtual void setMean(const uint cntr, const uint surr, const double val);
  
  //! Set the std. dev of feature response at the given
  //! center/surround level.
  virtual void setSigma(const uint cntr, const uint surr, const double val);

  //! Get the mean/ weighting coefficient for the given center/surround level.
  virtual double getCoeff(const uint cntr, const uint surr) const;

  //! Get the mean feature response for the given center/surround
  //! level.
  virtual double getMean(const uint cntr, const uint surr) const;

  //! Get the std. dev of the feature response for the given
  //! center/surround level.
  virtual double getSigma(const uint cntr, const uint surr) const;

  virtual void learn(const Image<byte>& dmap, const double eta,
                     const bool softmask = false,
                     const int inthresh = 20, const int outthresh = 255);

  virtual bool outputAvailable() const;

  virtual const Dims& getMapDims() const;

  virtual const Dims& getInputDims() const;

  //! Get the image from the given level of the internal pyramid
  virtual const Image<float>& getImage(const uint lev) const;

  //! Get the center/surround image for the given levels
  virtual Image<float> centerSurround(const uint cntr, const uint surr) const;

  //! Get the center/surround, split into positive and negative parts
  virtual void centerSurround(const uint cntr, const uint surr,
                              Image<float>& pos, Image<float>& neg) const;

  virtual void fillJet(const Point2D& loc, Jet<float>& j,
                       std::vector<int>& index) const;

  virtual void fillJetNI(const Point2D& loc, Jet<float>& j,
                         std::vector<int>& index) const;

  virtual void fillJetSpec(JetSpec &js) const;

  virtual uint numSubmaps() const;

  //! This is just a caching wrapper around computeSubmap()
  virtual Image<float> getSubmap(const uint index) const;

  virtual const char* getSubmapName(const uint index) const;

  //! This is for perceptual feature grouping 
  virtual void getCoeff(std::vector<double>& coeff);

  virtual void setGroupingCoeff(std::vector<double>& coeff, uint& idx);

  virtual Image<float> perceptualGrouping();

  virtual void biasFeatures(const double* mean, const double* sigma, int& idx);

  virtual void getFeatures(const Point2D& locn,
                           std::vector<double>& mean) const;

  virtual void getFeaturesBatch(std::vector<Point2D*> *locn,
                                std::vector<std::vector<float> > *mean,
                                int *count) const;
  virtual void findWeight();

  virtual bool reco(const ParamMap& pmap, double& mismatch);

  virtual void findLikelihood(const ParamMap& pmap, double& likelihood);

  //! This is just a caching wrapper around combineSubMaps()
  virtual Image<float> getOutput();

  //! Combine all feature maps into a single output map
  virtual Image<float> combineSubMaps();

  //! Save our various maps using an OutputFrameSeries
  /*! Depending on our ModelParam settings, we can save raw pyramid
    levels (with name prefix "SR<tagname>-<level>-"), center-surround
    feature maps ("SF<tagname>-<ctr>-<surr>-"), and output map
    ("SO<tagname>-") */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //! Access to the underlying pyramid is provided for completeness.
  /*! However, it probably shouldn't need to be used except for testing,
      debugging, etc. Instead, the Channel interface should be used or
      extended to fill the specific need.
      @param index The index in the queue (0 = newest, qlen-1 = oldest). */
  virtual const ImageSet<float>& pyramid(const uint index) const;

  //! Access to the underlying pyramid times is provided for completeness.
  /*! However, it probably shouldn't need to be used except for testing,
      debugging, etc. Instead, the Channel interface should be used or
      extended to fill the specific need.
      @param index The index in the queue (0 = newest, qlen-1 = oldest). */
  virtual double pyramidTime(const uint index) const;

  //! Access to the underlying clipPyramid
  virtual const ImageSet<float>& clipPyramid() const;

  //! get our the LevelSpec
  LevelSpec getLevelSpec() const;

  //! get our NormType:
  int getNormType() const;

  //! should we use single-opponence?
  void useSingleOpponence(bool use);

  //! Single-opponent SingleChannel requires center and surround pyramids
  virtual void singleOpponentInput(const ImageSet<float>& centerPyr,
                                   const ImageSet<float>& surroundPyr,
                                   const double t,
                                   const Image<byte>& clipMask);

protected:
  //! SingleChannel requires only luminance input
  virtual void doInput(const Image<PixRGB<float> >*,
                       const Image<float>* bwimg, const double t,
                       const Image<byte>& clipMask);

  //! Stores p as the channel's pyramid at time t
  virtual void storePyramid(const ImageSet<float>& p, const double t);

  //! Stores p as the channel's surround pyramid at time t
  /*! This will trow a fatal error is the channel has not been
    declared as single-opponent through a call to useSingleOpponence() */
  virtual void storePyramid2(const ImageSet<float>& p, const double t);

  //! make the clipping pyramid from the clip mask
  virtual void setClipPyramid(const Image<byte>& clipMask);

  //! SingleChannel implements this to clear its cached output.
  virtual void killCaches();

  //! Change to use a different pyramid object.
  virtual void setPyramid(SharedPtr<PyrBuilder<float> > pyr);

  //! Mutable access to underlying pyramid
  virtual ImageSet<float>& pyrMut(const uint index);

  //! compute a submap; part of getSubMap()
  virtual Image<float> computeSubmap(const uint idx) const;

  //! get raw CS map; part of getSubmap()
  virtual Image<float> getRawCSmap(const uint idx) const;

  //! rescale and post-process raw CS map; part of default getSubMap()
  virtual Image<float> postProcessMap(const Image<float>& smap,
                                      const uint idx) const;

protected:
  struct TPyr { //!< a time-stamped pyramid
    TPyr(const ImageSet<float>& pyr_, const double t_) : pyr(pyr_), t(t_) {}
    ImageSet<float> pyr; //!< the pyramid
    double t;            //!< the timestamp
  };

  ModelParam<bool> itsTakeAbs;
  ModelParam<bool> itsNormalizeOutput;
  ModelParam<bool> itsUseRandom;
  ModelParam<bool> itsScaleNoiseToMax;
  ModelParam<float> itsLowThresh;
  ModelParam<bool> itsRectifyPyramid;
  ModelParam<bool> itsUseSplitCS;
  ModelParam<LevelSpec> itsLevelSpec;
  ModelParam<MaxNormType> itsNormType;
  ModelParam<int> itsQlen;
  ModelParam<bool> itsUseOlderVersion;  

  //! Time decay for the contribution of differences in the pyramid queue.
  /*! This decay will yield a factor fac = exp( (t2 - t) * decay)
    applied to the difference image between t (current time, in
    seconds) and t2 (time of previous image, in seconds): */
  ModelParam<double> itsTimeDecay;

  //! Save our raw pyramid levels?
  ModelParam<bool> itsSaveRawMaps;

  //! Save our center-surround feature maps?
  ModelParam<bool> itsSaveFeatureMaps;

  //! Save our output map?
  ModelParam<bool> itsSaveOutputMap;

  // get us started (see ModelComponent.H)
  /*! If you overload this, make sure you call SingleChannel::start1()
    at the beginning of your overload */
  virtual void start1();

  // get us stopped (see ModelComponent.H)
  /*! If you overload this, make sure you call SingleChannel::start1()
    at the beginning of your overload */
  virtual void stop2();

  // shortcut to the csToIndex function of itsLevelSpec:
  /*! If you overload this, make sure you call SingleChannel::stop1()
    at the beginning of your overload */
  uint csToIndex(uint centerlev, uint surroundlev) const;

private:
  SingleChannel(const SingleChannel&); // not allowed
  SingleChannel& operator=(const SingleChannel&); // not allowed

  std::deque<TPyr>               itsPq; // temporal queue of pyramids
  double*                        itsCoeff;   // weighting coefficient
  double*                        itsMean;    // mean feature response
  double*                        itsSigma;   // stdev  of feature response
  SharedPtr<ChannelListener>     itsListener;
  Image<float>                   itsOutputCache;
  Image<float>*                  itsSubmapCache;
  SharedPtr<PyrBuilder<float> >  itsPyrBuilder;
  ImageSet<float>                itsClipPyr;

  friend class SingleChannelBeo;   // this one needs to mess with itsPq
  friend class OrientationChannel; // this one too
  friend class RGBConvolveChannel; // this one messes with itsPyrBuilder

  /* the following members are for single-opponence */
  bool itsUseSingleOpponence;          // is this channel single-opponent?
  std::deque<TPyr> itsPq2; // temporal queue of surround pyramids
};


/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif // !SINGLECHANNEL_H_DEFINED
