/*!@file SingleChannelBeo.C wrapper class to run a SingleChannel on a
  distant CPU */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SingleChannelBeo.C 2493 2003-04-29 08:42:03Z itti $
//

#include "SingleChannelBeo.H"
#include "TCPmessage.H"
#include "Beowulf.H"
#include "ParamMap.H"

#include <sstream>

// ######################################################################
SingleChannelBeo::SingleChannelBeo(SingleChannel *chan,
                                   const SharedPtr<Beowulf>& beow) :
  itsChan(chan), itsBeo(beow), itsNode(-2), itsFrame(-1)
{ }

// ######################################################################
void SingleChannelBeo::requestBeowulfNode()
{
  // ask the Beowulf for a node number:
  itsNode = itsBeo->requestNode();
  if (itsNode == -2)
    LFATAL("No more available nodes - You need a bigger Beowulf -- ABORT"); 

  LINFO("Delegating '%s' [%s] to Beowulf node %d",
        itsChan->descriptiveName().c_str(), itsChan->tagName().c_str(),
        itsNode);
}

// ######################################################################
SingleChannelBeo::~SingleChannelBeo()
{  }

// ######################################################################
void SingleChannelBeo::sendInput(const Image<float>* bwimg, const double t,
                                 const Image<byte>& clipMask)
{
  if (bwimg == NULL) LFATAL("Oooops, NULL input image? -- ABORT");
  // if this is the first time we are called, let's start by sending
  // off a configuration message to our processing node, so that it
  // can create a channel for us:
  if (itsFrame == -1) {
    // we have two types of config data: our ModelParam values, and our
    // map weights. We put each into a ParamMap:
    ParamMap param; itsChan->writeParamsTo(param);
    ParamMap weight; itsChan->writeTo(weight);
    
    // let's convert both ParamMaps to string:
    std::stringstream sparam, sweight;
    param.format(sparam); weight.format(sweight);
    
    // let's put our descriptive name, our tag name and both parameter
    // strings into a TCPmessage, and let's put our VisualFeature in
    // the 'frame' field, and let's send it off (NOTE: VisualFeature
    // also is in param, but having it here will make decoding of the
    // message easier).
    int32 vs = static_cast<int32>(itsChan->visualFeature());
    TCPmessage msg(vs, BEO_SCHANCONF);
    msg.addString(itsChan->descriptiveName().c_str());
    msg.addString(itsChan->tagName().c_str());
    msg.addString(sparam.str().c_str());
    msg.addString(sweight.str().c_str());

    itsBeo->send(itsNode, msg);
  }

  // increment our unofficial frame number; we have this counter just
  // because we have a frame field in TCPmessage, so we may as well
  // use it:
  itsFrame ++;

  // send t, bwimg and clipMask to processor node:
  TCPmessage msg(itsFrame, BEO_SCHANINPUT);
  msg.addDouble(t); msg.addImage(*bwimg); msg.addImage(clipMask);
  itsBeo->send(itsNode, msg);

  // anybody needing our results will need to wait until they have
  // come back from the processor node:
  itsChan->killCaches();
}

// ######################################################################
void SingleChannelBeo::waitForOutput()
{
  // nothing to wait for if we have nothing in progress
  if (itsChan->itsOutputCache.initialized()) return;

  // See if we have a new message on our Beowulf:
  TCPmessage rmsg; bool gotnothing = true;
  int32 rframe, raction, rnode = itsNode;  // receive only from our server
  while(gotnothing)
    if (itsBeo->receive(rnode, rmsg, rframe, raction, 50)){// wait up to 50ms
      // message is supposed to contain: pyramid, submaps, possibly clipPyr:
      if (raction != BEO_SCHANOUTPUT && raction != BEO_SCHANALLOUT)
        LERROR("IGNORING message of wrong type from node %d", rnode);
      else {
        if (rframe != itsFrame)
          LERROR("Received results for frame %d while at frame %d??",
                 rframe, itsFrame);
        // the time:
        double *ttt; int32 typ;
        rmsg.getElement((void **)&ttt, typ); ASSERT(typ == TCPMSG_DOUBLE);
        double t = *ttt; delete ttt;

        // the output map (always valid, but possibly blank):
        Image<float> *out;
        rmsg.getElement((void **)&out, typ); ASSERT(typ == TCPMSG_FLOATIMA);
        itsChan->itsOutputCache = *out; delete out;

        // additional results?
        if (raction == BEO_SCHANALLOUT) {
          // the pyramid:
          ImageSet<float> *pyr;
          rmsg.getElement((void **)&pyr, typ); ASSERT(typ==TCPMSG_FLOATIMASET);
          // if the SingleChannel on the remote node had no output
          // available yet, we have an empty pyramid, which we don't
          // want to push into our pyramid queue; this way, we will also
          // have no pyramid available if the remote had none:
          if (pyr->size()) {
            SingleChannel::TPyr tpyr(*pyr, t);
            itsChan->itsPq.push_front(tpyr);  // store our pyramid
          }
          delete pyr;
          
          // the submaps:
          rmsg.getElement((void **)&pyr, typ); ASSERT(typ==TCPMSG_FLOATIMASET);
          if (pyr->size()) {
            uint maxind = itsChan->numSubmaps();
            for (uint i = 0; i < maxind; i ++)
              itsChan->itsSubmapCache[i] = (*pyr)[i];
          }
          delete pyr;
          
          // clip pyramid (if any):
          rmsg.getElement((void **)&pyr, typ); ASSERT(typ==TCPMSG_FLOATIMASET);
          if (pyr->size()) {
            itsChan->itsClipPyr = *pyr;
          }
          delete pyr;
        } else {
          // in QuickMode we just load an empty ImageSet into the
          // channel's pyramid queue, so that outputAvailable() will
          // return true now that we indeed have the output:
          ImageSet<float> empty; SingleChannel::TPyr tpyr(empty, t);
          itsChan->itsPq.push_front(tpyr);  // store empty pyramid
        }
        // truncate the pyramid queue if necessary, just like is
        // done in SingleChannel::doInput:
        while(int(itsChan->itsPq.size()) > itsChan->itsQlen.getVal())
          itsChan->itsPq.pop_back();
        
        // ok, we got our results, let's return:
        gotnothing = false;
      }
    }
  //else
  //   LINFO("%s waiting for results from node %d...",
  //         itsChan->tagName().c_str(), itsNode);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
