/*!@file SingleChannelSurprise.C Channel for a single stream of processing. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SingleChannelSurprise.C 4291 2005-03-13 00:24:48Z itti $
//

#include "SingleChannelSurprise.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"

// ######################################################################
SingleChannelSurprise::SingleChannelSurprise(ModelManager& mgr,
                                             const std::string& descrName,
                                             const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSQlen(this, "SingleChannelSurpriseSQlen", 3),
  itsUpdateFac(this, "SingleChannelSurpriseUpdFac", 0.7),
  itsNeighUpdateFac(this, "SingleChannelSurpriseNeighUpdFac", 0.7),
  itsInitialVal(this, "SingleChannelSurpriseIniVal", 0.0),
  itsInitialVar(this, "SingleChannelSurpriseIniVar", 25.0),
  itsBgVal(this, "SingleChannelSurpriseBgVal", 5.0),
  itsNeighSigma(this, "SingleChannelSurpriseNeighSig", 0.5f),
  itsLocSigma(this, "SingleChannelSurpriseLocSig", 3.0f),
  itsLevelSpec(this, "LevelSpec", LevelSpec()),
  itsProbe(this, "SingleChannelSurpriseProbe", Point2D(-1, -1)),
  itsSLfac(this, "SingleChannelSurpriseSLfac", 1.0),
  itsSSfac(this, "SingleChannelSurpriseSSfac", 0.1),
  itsSmap()
{ }

// ######################################################################
SingleChannelSurprise::~SingleChannelSurprise()
{ }

// ######################################################################
void SingleChannelSurprise::exportOptions(const int whatToExport,
                                          const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsSQlen);
    manager->requestOption(this, itsUpdateFac);
    manager->requestOption(this, itsNeighUpdateFac);
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsProbe);
    manager->requestOption(this, itsSLfac);
    manager->requestOption(this, itsSSfac);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void SingleChannelSurprise::start1()
{
  int maxind = itsLevelSpec.getVal().maxIndex();

  SurpriseMap<SSSMODEL> initial;
  initial.init(itsSQlen.getVal(), itsUpdateFac.getVal(),
               itsNeighUpdateFac.getVal(),
               itsInitialVal.getVal() + itsBgVal.getVal(),
               itsInitialVar.getVal(), itsNeighSigma.getVal(),
               itsLocSigma.getVal(), itsProbe.getVal(),
               itsSLfac.getVal(), itsSSfac.getVal());

  for (int i = 0; i < maxind; i ++) itsSmap.push_back(initial);
}

// ######################################################################
Image<float> SingleChannelSurprise::getSurpriseMap(const uint index,
                                                   const Image<float>& submap)
{
  Image<double> submapd = submap;
  if (itsBgVal.getVal()) submapd += itsBgVal.getVal(); // add bg firing rate

  // create a SurpriseImage from the submap:
  Image<double> var(submap.getDims(), NO_INIT);
  var.clear(itsInitialVar.getVal());
  SurpriseImage<SSSMODEL> sample(itsUpdateFac.getVal(), submapd, var);

  // now inject that into our models and get the surprise:
  return Image<float>(itsSmap[index].surprise(sample));
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
