/*!@file SingleChannelSurprise.H Channel for a single stream of processing. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SingleChannelSurprise.H 4291 2005-03-13 00:24:48Z itti $
//

#ifndef SINGLECHANNELSURPRISE_H_DEFINED
#define SINGLECHANNELSURPRISE_H_DEFINED

#include "ModelComponent.H"
#include "SurpriseMap.H"
#include <vector>
#include "LevelSpec.H"

// ######################################################################
//! SingleChannelSurprise is a Surprise extension to SingleChannel
/*! This class only provides the computation of surprise. Typically
  you would first build a SingleChannel (or derivative), then build a
  SingleChannelSurprise, then adopt the SingleChannelSurprise into the
  SingleChannel as a subcomponent, then register your SingleChannel
  with your VisualCortex, ModelManager or whoever (but do not register
  the adopted SingleChannelSurprise object). */ 
class SingleChannelSurprise : public ModelComponent
{
public:
  //! Constructor
  SingleChannelSurprise(ModelManager& mgr, const std::string& descrName,
                        const std::string& tagName);

  //! destructor
  virtual ~SingleChannelSurprise();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! get a surprise map
  /*! Pass a submap index and a submap, and you'll get back the
    associated surprise */
  virtual Image<float> getSurpriseMap(const uint index,
                                      const Image<float>& submap);

protected:
  ModelParam<uint> itsSQlen;            //!< number of surprise maps
  ModelParam<double> itsUpdateFac;      //!< update factor
  ModelParam<double> itsNeighUpdateFac; //!< update factor for neighbors
  ModelParam<double> itsInitialVal;     //!< Initial values
  ModelParam<double> itsInitialVar;     //!< Initial variance
  ModelParam<double> itsBgVal;          //!< Background values
  ModelParam<float> itsNeighSigma;      //!< Neighborhood sigma
  ModelParam<float> itsLocSigma;        //!< Local neighborhood sigma
  ModelParam<LevelSpec> itsLevelSpec;   //!< our levelspec
  ModelParam<Point2D> itsProbe;         //!< location of a virtusl electrode
  ModelParam<double> itsSLfac;          //!< factor for local temporal surprise
  ModelParam<double> itsSSfac;          //!< factor for spatial surprise
  
  // get us started (see ModelComponent.H)
  virtual void start1();

private:
  SingleChannelSurprise(const SingleChannelSurprise&); // not allowed
  SingleChannelSurprise& operator=(const SingleChannelSurprise&); // not allow

// surprise models to use:
#define SSSMODEL SurpriseModelSP

  std::vector< SurpriseMap<SSSMODEL> > itsSmap;
};

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
