/*!@file StdBrain.C A canonical brain, which a standard set of channels. */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: StdBrain.C 3874 2004-10-29 06:16:54Z siagian $
//

#include "StdBrain.H"
#include "ChannelBase.H"
#include "Retina.H"
#include "VisualCortex.H"
#include "SaccadeControllers.H"
#include "SimulationViewer.H"
#include "ShapeEstimator.H"
#include "ModelManager.H"
#include "VisualCortexConfigurator.H"
#include "SaliencyMap.H"
#include "TaskRelevanceMap.H"
#include "AttentionGuidanceMap.H"
#include "WinnerTakeAll.H"
#include "InferoTemporal.H"
#include "InferoTemporalConfigurator.H"
#include "GistEstimator.H"
#include "GistEstimatorConfigurator.H"

// ######################################################################
StdBrain::StdBrain(ModelManager& mgr, const std::string& descrName,
		   const std::string& tagName) :
  Brain(mgr, descrName, tagName),
  rec(new RetinaConfigurator(mgr)),
  scc(new SaccadeControllerConfigurator(mgr)),
  vcc(new VisualCortexConfigurator(mgr)),
  smc(new SaliencyMapConfigurator(mgr)),
  trmc(new TaskRelevanceMapConfigurator(mgr)),
  agmc(new AttentionGuidanceMapConfigurator(mgr)),
  wtac(new WinnerTakeAllConfigurator(mgr)),
  svc(new SimulationViewerConfigurator(mgr)),
  it(new InferoTemporalConfigurator(mgr)),
  gec(new GistEstimatorConfigurator(mgr))
{
  SharedPtr<ShapeEstimator>
    se(new ShapeEstimator(mgr, "Shape Estimator", "ShapeEstimator"));
  //SharedPtr<ShapeEstimator>
  //  ge(new ShapeEstimator(mgr, "Gist Estimator", "GistEstimator"));

  // we take ownership of our new babies, except for the
  // Configurators, as we'll need to wait until we get started to know
  // what their chosen implementation is going to be:
  addModule(se, "ShapeEstimator", true);
  addSubComponent(rec);  // make sure it can get configured
  addSubComponent(scc);  // make sure it can get configured
  addSubComponent(vcc);  // make sure it can get configured
  addSubComponent(smc);  // make sure it can get configured
  addSubComponent(trmc); // make sure it can get configured
  addSubComponent(agmc); // make sure it can get configured
  addSubComponent(wtac); // make sure it can get configured
  addSubComponent(svc);  // make sure it can get configured
  addSubComponent(it);   // make sure it can get configured
  addSubComponent(gec);  // make sure it can get configured
}

// ######################################################################
StdBrain::~StdBrain()
{  }

// ######################################################################
void StdBrain::start1()
{
  // did we get a VisualCortex?
  if (rec->getRET().get())
    addModule(rec->getRET(), "Retina", false);

  // did we get a SaccadeController?
  if (scc->getSC().get())
    addModule(scc->getSC(), "SaccadeController", false);

  // did we get a VisualCortex?
  if (vcc->getVC().get())
    addModule(vcc->getVC(), "VisualCortex", false);

  // did we get a SaliencyMap?
  if (smc->getSM().get())
    addModule(smc->getSM(), "SaliencyMap", false);

  // did we get a TaskRelevanceMap?
  if (trmc->getTRM().get())
    addModule(trmc->getTRM(), "TaskRelevanceMap", false);

  // did we get an AttentionGuidanceMap?
  if (agmc->getAGM().get())
    addModule(agmc->getAGM(), "AttentionGuidanceMap", false);

  // did we get a WinnerTakeAll?
  if (wtac->getWTA().get())
    addModule(wtac->getWTA(), "WinnerTakeAll", false);

  // did we get a SimulationViewer?
  if (svc->getSV().get())
    addModule(svc->getSV(), "SimulationViewer", false);

  // did we get an Inferior Temporal Cortex?
  if (it->getIT().get())
    addModule(it->getIT(), "InferoTemporal", false);

  // did we get a GistEstimator?
  if (gec->getGE().get())
    addModule(gec->getGE(), "GistEstimator", false);

  // hook VC to SE:
  if (hasModule("ShapeEstimator") && hasModule("VisualCortex"))
    getSE()->setVC(getVC());

  // hook brain and SV:
  if (hasModule("SimulationViewer")) getSV()->setBrain(this);

  // hook brain and SC:
  if (hasModule("SaccadeController")) getSC()->setBrain(this);

  // hook brain and RET:
  if (hasModule("Retina")) getRET()->setBrain(this);

  // hook brain and TRM:
  if (hasModule("TaskRelevanceMap")) getTRM()->setBrain(this);

  // hook VC and GE:
  if (hasModule("GistEstimator") && hasModule("VisualCortex"))    
    getGE()->setVC(getVC());
    
  // normal brain init:
  Brain::start1();
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
