/*!@file StringConversions.C Convert to and from std::string */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: StringConversions.C 3713 2004-09-24 06:22:51Z walther $
//

#include "log.H"
#include "StringConversions.H"
#include <sstream>
#include "saliency.H"          // for byte

#include "ArrayData.H"         // for Dims
#include "Point2D.H"           // for Point2D
#include "FrameRange.H"        // for FrameRange
#include "LevelSpec.H"         // for LevelSpec
#include "Rectangle.H"         // for Rectangle
#include "Pixels.H"            // for PixRGB<byte>
#include <netinet/in.h>        // for in_addr

// instantiations for canonical types:
template void convertToString(const int& val, std::string& str);
template bool convertFromString(const std::string& str, int& val);
template void convertToString(const short int& val, std::string& str);
template bool convertFromString(const std::string& str, short int& val);
template void convertToString(const uint& val, std::string& str);
template bool convertFromString(const std::string& str, uint& val);
template void convertToString(const float& val, std::string& str);
template bool convertFromString(const std::string& str, float& val);
template void convertToString(const double& val, std::string& str);
template bool convertFromString(const std::string& str, double& val);

// instantiations of toStr(T) and fromStr(std::string)
#define INSTANTIATE(T) template std::string toStr(const T& val); \
                       template T fromStr(const std::string& str) ;
#include "InstantiateAll.H"


// ######################################################################
template <class T>
void convertToString(const T& val, std::string& str)
{ std::stringstream s; s << val; s >> str; }

// ######################################################################
template <class T> inline
bool convertFromString(const std::string& str, T& val)
{ std::stringstream s; s << str; s >> val; return true; }

// ######################################################################
void convertToString(const std::string& val, std::string& str)
{ str = val; }

// ######################################################################
bool convertFromString(const std::string& str, std::string& val)
{ val = str; return true; }

// ######################################################################
void convertToString(const Dims& val, std::string& str)
{ std::stringstream s; s<<val.w()<<'x'<<val.h(); s >> str; }

// ######################################################################
bool convertFromString(const std::string& str, Dims& val)
{
  std::stringstream s; int w = -1, h = -1; char c;
  s<<str; s>>w; s>>c; s>>h;
  if (w == -1 || h == -1 || c != 'x')
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }
  val = Dims(w, h); return true;
}

// ######################################################################
void convertToString(const Point2D& val, std::string& str)
{ std::stringstream s; s<<val.i<<','<<val.j; s>>str; }

// ######################################################################
bool convertFromString(const std::string& str, Point2D& val)
{
  std::stringstream s; char c; int i = -1234567889, j = -123456789;
  s<<str; s>>i; s>>c; s>>j;
  if (i == -123456789 || j == -123456789)
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }
  val.i = i; val.j = j; return true;
}

// ######################################################################
void convertToString(const bool& val, std::string& str)
{ if (val) str = "true"; else str = "false"; }

// ######################################################################
bool convertFromString(const std::string& str, bool& val)
{
  char c = str[0];
  if (c == '1' || c == 't' || c == 'T' || c == 'y' || c == 'Y')
    { val = true; return true; }
  else if (c == '0' || c == 'f' || c == 'F' || c == 'n' || c == 'N')
    { val = false; return true; }
  else
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }
}

// ######################################################################
void convertToString(const FrameGrabberMode val, std::string& str)
{ str = frameGrabberModeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, FrameGrabberMode& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBFRAMEGRABBERMODES; i ++)
    if (str.compare(frameGrabberModeName((FrameGrabberMode)i)) == 0)
      { val = (FrameGrabberMode)i; return true; }

  LERROR("Unknown frame grabber mode: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const FrameRange& val, std::string& str)
{ std::stringstream s; s<<val.first<<'-'<<val.last<<'@'<<val.delay; s >> str; }

// ######################################################################
bool convertFromString(const std::string& str, FrameRange& val)
{
  std::stringstream s; int first = -2, last = -2; float delay = -1.0F; char c;
  s<<str; s>>first>>c>>last>>c>>delay;
  if (first == -2 || last == -2 || c != '@' || delay == -1.0F)
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }

  val.first = first; val.last = last; val.delay = delay; return true;
}

// ######################################################################
void convertToString(const LevelSpec& val, std::string& str)
{
  std::stringstream s;
  s<<val.levMin()<<'-'<<val.levMax()<<','<<val.delMin()<<'-'<<
    val.delMax()<<','<<val.mapLevel();
  s>>str;
}

// ######################################################################
bool convertFromString(const std::string& str, LevelSpec& val)
{
  std::stringstream s; int lmi = -2, lma = -2, dmi = -2, dma = -2, ml = -2;
  char c; s<<str; s>>lmi>>c>>lma>>c>>dmi>>c>>dma>>c>>ml;
  if (lmi == -2 || lma == -2 || dmi == -2 || dma == -2 || ml == -2)
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }

  val.init(lmi, lma, dmi, dma, ml); return true;
}

// ######################################################################
void convertToString(const Rectangle& val, std::string& str)
{
  std::stringstream s;
  s<<val.left()<<','<<val.top()<<','<<val.right()<<','<<val.bottom();
  s>>str;
}

// ######################################################################
bool convertFromString(const std::string& str, Rectangle& val)
{
  std::stringstream s; int t = -1, b = -1, l = -1, r = -1;
  char c; s<<str; s>>l>>c>>t>>c>>r>>c>>b;
  if (t == -1 || b == -1 || l == -1 || r == -1)
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }

  val.modify(t, l, b, r); return true;
}

// ######################################################################
void convertToString(const RasterFileFormat val, std::string& str)
{
  // Text descriptions of the various known formats:
  static const char *RasterFileFormatString[] = {
    "PNM", "PNG", "PFM", "RAWIMAGE", "Auto" };
  str = RasterFileFormatString[val];
}

// ######################################################################
bool convertFromString(const std::string& str, RasterFileFormat& val)
{
  if (str.compare("PNM") == 0) { val = PNM; return true; }
  else if (str.compare("PNG") == 0) { val = PNG; return true; }
  else if (str.compare("PFM") == 0) { val = PFM; return true; }
  else if (str.compare("RAWIMAGE") == 0) { val = RAWIMAGE; return true; }
  else if (str.compare("Auto") == 0) { val = Auto; return true; }

  LERROR("Unknown format: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const PixRGB<byte>& val, std::string& str)
{
  std::stringstream s;
  s<<int(val.red())<<','<<int(val.green())<<','<<int(val.blue());
  s>>str;
}

// ######################################################################
bool convertFromString(const std::string& str, PixRGB<byte>& val)
{
  if (str[0] == '#') {
    // #RRGGBB format
    std::string str2 = str.substr(1);
    std::stringstream s; s<<std::hex<<str2; int rgb; s>>rgb;
    val.setRed((rgb & 0xff0000) >> 16);
    val.setGreen((rgb & 0xff00) >> 8);
    val.setBlue(rgb & 0xff);
    return true;
  } else {
    // <byte>,<byte>,<byte> format
    std::stringstream s; int r = -1, g = -1, b = -1;
    char c; s<<str; s>>r>>c>>g>>c>>b;
    if (r == -1 || g == -1 || b == -1)
      { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }
    val.set(r, g, b); return true;
  }
}

// ######################################################################
void convertToString(const VisualFeature val, std::string& str)
{ str = featureName(val); }

// ######################################################################
bool convertFromString(const std::string& str, VisualFeature& val)
{
  // CAUTION: assumes features are numbered and ordered!
  for (int i = 0; i < NBVISUALFEATURES; i ++)
    if (str.compare(featureName((VisualFeature)i)) == 0)
      { val = (VisualFeature)i; return true; }

  LERROR("Unknown visual feature: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const PyramidType val, std::string& str)
{ str = pyramidTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, PyramidType& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBPYRAMIDTYPES; i ++)
    if (str.compare(pyramidTypeName((PyramidType)i)) == 0)
      { val = (PyramidType)i; return true; }

  LERROR("Unknown pyramid type: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const MaxNormType val, std::string& str)
{ str = maxNormTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, MaxNormType& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBMAXNORMTYPES; i ++)
    if (str.compare(maxNormTypeName((MaxNormType)i)) == 0)
      { val = (MaxNormType)i; return true; }

  LERROR("Unknown maxnorm type: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const IORtype val, std::string& str)
{ str = iorTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, IORtype& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBIORTYPES; i ++)
    if (str.compare(iorTypeName((IORtype)i)) == 0)
      { val = (IORtype)i; return true; }

  LERROR("Unknown IOR type: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const ShapeEstimatorMode val, std::string& str)
{ str = shapeEstimatorModeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, ShapeEstimatorMode& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBSHAPEESTIMATORMODES; i ++)
    if (str.compare(shapeEstimatorModeName((ShapeEstimatorMode)i)) == 0)
      { val = (ShapeEstimatorMode)i; return true; }

  LERROR("Unknown ShapeEstimator mode: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const ShapeEstimatorSmoothMethod val, 
                     std::string& str)
{ str = shapeEstimatorSmoothMethodName(val); }

// ######################################################################
bool convertFromString(const std::string& str, 
                       ShapeEstimatorSmoothMethod& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBSHAPEESTIMATORSMOOTHMETHODS; i ++)
    if (str.compare(shapeEstimatorSmoothMethodName
                    ((ShapeEstimatorSmoothMethod)i)) == 0)
      { val = (ShapeEstimatorSmoothMethod)i; return true; }

  LERROR("Unknown ShapeEstimator smooth method: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const OrientComputeType val, std::string& str)
{ str = orientComputeTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, OrientComputeType& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBORIENTCOMPUTETYPES; i ++)
    if (str.compare(orientComputeTypeName((OrientComputeType)i)) == 0)
      { val = (OrientComputeType)i; return true; }

  LERROR("Unknown OrientComputeType: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const in_addr& val, std::string& str)
{
  std::stringstream s; unsigned long int a = val.s_addr;
  s<<(a>>24)<<'.'<<((a>>16) & 0xFF)<<'.'<<((a>>8) & 0xFF)<<'.'<<(a & 0xFF);
  s>>str;
}

// ######################################################################
bool convertFromString(const std::string& str, in_addr& val)
{
  std::stringstream s; unsigned long int n1 = 300, n2 = 300, n3 = 300, n4=300;
  char c; s<<str; s>>n1>>c>>n2>>c>>n3>>c>>n4;
  if (n1 == 300 || n2 == 300 || n3 == 300 || n4 == 300)
    { LERROR("Bogus format: %s -- IGNORED", str.c_str()); return false; }
  val.s_addr = (n1 << 24) | (n2 << 16) | (n3 << 8) | n4;
  return true;
}

// ######################################################################
void convertToString(const BitObjectDrawMode val, std::string& str)
{ str = bitObjectDrawModeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, BitObjectDrawMode& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBBITOBJECTDRAWMODES; i ++)
    if (str.compare(bitObjectDrawModeName((BitObjectDrawMode)i)) == 0)
      { val = (BitObjectDrawMode)i; return true; }

  LERROR("Unknown BitObjectDrawMode: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const ImageDisplayType val, std::string& str)
{ str = imageDisplayTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, ImageDisplayType& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBIMAGEDISPLAYTYPES; i++)
    if (str.compare(imageDisplayTypeName((ImageDisplayType)i)) == 0)
      { val = (ImageDisplayType)i; return true; }

  LERROR("Unknown ImageDisplay type: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
void convertToString(const ColorComputeType val, std::string& str)
{ str = colorComputeTypeName(val); }

// ######################################################################
bool convertFromString(const std::string& str, ColorComputeType& val)
{
  // CAUTION: assumes types are numbered and ordered!
  for (int i = 0; i < NBCOLORCOMPUTETYPES; i++)
    if (str.compare(colorComputeTypeName((ColorComputeType)i)) == 0)
      { val = (ColorComputeType)i; return true; }

  LERROR("Unknown ColorCompute type: %s -- IGNORED", str.c_str());
  return false;
}

// ######################################################################
template <class T>
std::string toStr(const T& val)
{
  std::string str;
  convertToString(val,str);
  return str;
}

// ######################################################################
template <class T>
T fromStr(const std::string& str)
{
  T val;
  convertFromString(str, val);
  return val;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
