/*!@file StringConversions.H Convert to and from std::string */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: StringConversions.H 3713 2004-09-24 06:22:51Z walther $
//

#ifndef STRINGCONVERSIONS_H_DEFINED
#define STRINGCONVERSIONS_H_DEFINED

#include <string>
class Point2D;
class Dims;
class FrameRange; 
class LevelSpec;
class Rectangle;
template <class T> class PixRGB;
struct in_addr;

#include "FrameGrabberMode.H"    // for FrameGrabberMode
#include "RasterFileFormat.H"    // for RasterFileFormat
#include "saliency.H"            // for byte
#include "VisualFeatures.H"      // for VisualFeature
#include "PyramidTypes.H"        // for PyramidType
#include "fancynorm.H"           // for MaxNormType
#include "IORtypes.H"            // for IORtype
#include "ShapeEstimatorModes.H" // for ShapeEstimatorMode and ShapeEstimatorSmoothMethod
#include "OrientComputeTypes.H"  // for OrientComputeType
#include "BitObjectDrawModes.H"  // for BitObjectDrawMode
#include "ImageDisplayTypes.H"   // for ImageDisplayType
#include "ColorComputeTypes.H"   // for ColorComputeType

/*! \defgroup strcvthelp String conversion helpers

   We define a number of string conversion helpers here just so that
   we don't require each type that we may want to instantiate into a
   ModelParam derivative to define operator<< and operator>>. The main
   reason is that this definition requires #including <sstream>
   (especially if the class is all-inline like Point2D), which would
   largely increase the compile time if we did it for all our simple
   types like Dims, Point2D, etc.  But the code below may some day
   migrate into proper definitions of operator<< and operator>> for
   those data types. */
//@{

/*! @name Generic string conversion functions */
//@{

/*! inefficient function to convert anything that supports operator<<
  into an std::string */
template <class T>
void convertToString(const T& val, std::string& str);

/*! inefficient function to convert from an std::string into anything
  that supports operator>> */
template <class T>
bool convertFromString(const std::string& str, T& val);
//@}

/*! @name Specialized implementations of string conversion functions */
//@{

//! std::string specialization
void convertToString(const std::string& val, std::string& str);
//! std::string specialization
bool convertFromString(const std::string& str, std::string& val);


//! Dims specialization: format is "<int>x<int>"
void convertToString(const Dims& val, std::string& str);
//! Dims specialization: format is "<int>x<int>"
bool convertFromString(const std::string& str, Dims& val);


//! Point2D specialization: format is "<int>,<int>"
void convertToString(const Point2D& val, std::string& str);
//! Point2D specialization: format is "<int>,<int>"
bool convertFromString(const std::string& str, Point2D& val);

//! bool specialization: format is "true|false"
void convertToString(const bool& val, std::string& str);
//! bool specialization
/*! Anything with a first letter 't', T', 'y', 'Y', or '1' will
  convert to true; anything with a first letter 'f', F', 'n', 'N', or
  '0' will convert to false */
bool convertFromString(const std::string& str, bool& val);

//! Conversion from FrameGrabberMode to string
/*! Use it, for example, if you got a grabmode as text from the
  command line and need to convert that to a FrameGrabberMode value for the
  constructor. Text strings should match the symbolic names of the
  FrameGrabberMode enum, minus the "GM" prefix. */
void convertToString(const FrameGrabberMode val, std::string& str);
//! Conversion from string to FrameGrabberMode
/*! The names used in the definition of the enum FrameGrabberMode
  (minus the "GM" prefix) will be returned */
bool convertFromString(const std::string& str, FrameGrabberMode& val);

//! FrameRange specialization: format is "<int>-<int>@<float>"
/*! Format corresponds to "first-last@delay" */
void convertToString(const FrameRange& val, std::string& str);
//! FrameRange specialization: format is "<int>-<int>@<float>"
bool convertFromString(const std::string& str, FrameRange& val);

//! LevelSpec specialization: format is "<int>-<int>,<int>-<int>,<int>"
/*! Format corresponds to "levmin-levmax,delmin-delmax,maplevel" */
void convertToString(const LevelSpec& val, std::string& str);
//! LevelSpec specialization: format is "<int>-<int>,<int>-<int>,<int>"
bool convertFromString(const std::string& str, LevelSpec& val);

//! Rectangle specialization: format is "<int>,<int>,<int>,<int>"
/*! Format corresponds to "xtopleft,ytopleft,xbottomright,ybottomright" */
void convertToString(const Rectangle& val, std::string& str);
//! Rectangle specialization: format is "<int>,<int>,<int>,<int>"
bool convertFromString(const std::string& str, Rectangle& val);

//! RasterFileFormat specialization
/*! Use it, for example, if you got a file format as text from the
  command line and need to convert that to a RasterFileFormat value
  for a ModelParam. Text strings should match the symbolic names of
  the RasterFileFormat enum. */
void convertToString(const RasterFileFormat val, std::string& str);
//! RasterFileFormat specialization
/*! The names used in the definition of the enum RasterFileFormat
    will be returned */
bool convertFromString(const std::string& str, RasterFileFormat& val);

//! PixRGB<byte> specialization
/*! Format is "<byte>,<byte>,<byte>" for R, G, B */
void convertToString(const PixRGB<byte>& val, std::string& str);
//! RasterFileFormat specialization
/*! Format  is "<byte>,<byte>,<byte>" for  R, G, B, or  "#RRGGBB" with
  the RR, GG and BB values in hexadecimal. */
bool convertFromString(const std::string& str, PixRGB<byte>& val);

//! VisualFeature specialization
/*! Format is "name" as defined by featureName() in VisualFeatures.H */
void convertToString(const VisualFeature val, std::string& str);
//! VisualFeature specialization
/*! Format is "name" as defined by featureName() in VisualFeatures.H */
bool convertFromString(const std::string& str, VisualFeature& val);

//! PyramidType specialization
/*! Format is "name" as defined by pyramidTypeName() in PyramidTypes.H */
void convertToString(const PyramidType val, std::string& str);
//! PyramidType specialization
/*! Format is "name" as defined by pyramidTypeName() in PyramidTypes.H */
bool convertFromString(const std::string& str, PyramidType& val);

//! MaxNormType specialization
/*! Format is "name" as defined by maxNormTypeName() in MaxNormTypes.H */
void convertToString(const MaxNormType val, std::string& str);
//! MaxNormType specialization
/*! Format is "name" as defined by maxNormTypeName() in MaxNormTypes.H */
bool convertFromString(const std::string& str, MaxNormType& val);

//! IORtype specialization
/*! Format is "name" as defined by iorTypeName() in IORtypes.H */
void convertToString(const IORtype val, std::string& str);
//! IORtype specialization
/*! Format is "name" as defined by iorTypeName() in IORtypes.H */
bool convertFromString(const std::string& str, IORtype& val);

//! ShapeEstimatorMode specialization
/*! Format is "name" as defined in ShapeEstimatorModes.H */
void convertToString(const ShapeEstimatorMode val, std::string& str);
//! ShapeEstimatorMode specialization
/*! Format is "name" as defined in ShapeEstimatorModes.H */
bool convertFromString(const std::string& str, ShapeEstimatorMode& val);

//! ShapeEstimatorSmoothType specialization
/*! Format is "name" as defined in ShapeEstimatorModes.H */
void convertToString(const ShapeEstimatorSmoothMethod val, 
                     std::string& str);
//! ShapeEstimatorSmoothType specialization
/*! Format is "name" as defined in ShapeEstimatorModes.H */
bool convertFromString(const std::string& str,
                       ShapeEstimatorSmoothMethod& val);

//! OrientComputeType specialization
/*! Format is "name" as defined in OrientComputeTypes.H */
void convertToString(const OrientComputeType val, 
                     std::string& str);
//! OrientComputeType specialization
/*! Format is "name" as defined in OrientComputeTypes.H */
bool convertFromString(const std::string& str,
                       OrientComputeType& val);

//! in_addr specialization: format is "<int>.<int>.<int>.<int>"
void convertToString(const in_addr& val, std::string& str);
//! in_addr specialization: format is "<int>.<int>.<int>.<int>"
bool convertFromString(const std::string& str, in_addr& val);

//! BitObjectDrawMode specialization
/*! Format is "name" as defined in BitObjectDrawModes.H */
void convertToString(const BitObjectDrawMode val, 
                     std::string& str);
//! BitObjectDrawMode specialization
/*! Format is "name" as defined in BitObjectDrawModes.H */
bool convertFromString(const std::string& str,
                       BitObjectDrawMode& val);

//! ImageDisplayType specialization
/*! Format is "name" as defined by imageDisplayTypeName() 
  in ImageDisplayTypes.H */
void convertToString(const ImageDisplayType val, std::string& str);
//! ImageDisplayType specialization
/*! Format is "name" as defined by imageDisplayTypeName() 
  in ImageDisplayTypes.H */
bool convertFromString(const std::string& str, ImageDisplayType& val );

//! ColorCOmputeType specialization
/*! Format is "name" as defined by colorComputeTypeName() 
  in ColorComputeTypes.H */
void convertToString(const ColorComputeType val, std::string& str);
//! ColorComputeType specialization
/*! Format is "name" as defined by colorComputeTypeName() 
  in ColorComputeTypes.H */
bool convertFromString(const std::string& str, ColorComputeType& val );

//@}

/*! @name Easier-syntax functions to convert to/from string */
//@{

//! converts val into a string and returns the string
template <class T>
std::string toStr(const T& val);

//! converts str into a T value and returns the T value
template <class T>
T fromStr(const std::string& str);
//@}

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
