/*!@file SurpriseMap.C a surprise map */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SurpriseMap.C 4129 2005-01-18 19:55:04Z itti $
//

#include "SurpriseMap.H"
#include "Image_FilterOps.H"
#include "Image_MathOps.H"

// ######################################################################
template <class T>
SurpriseMap<T>::SurpriseMap() :
  itsModels(), itsQlen(0), itsInitialModel(),
  itsNeighSigma(0.0f), itsLocSigma(0.0f), itsNweights(), itsNWmin(0.0f),
  itsNeighUpdFac(0.7), itsProbe(-1, -1), itsSLfac(1.0), itsSSfac(0.1)
{ }

// ######################################################################
template <class T>
void SurpriseMap<T>::init(const uint qlen, const double updatefac,
                          const double neighupdatefac,
                          const double sampleval, const double samplevar,
                          const float neighsigma, const float locsigma,
                          const Point2D& probe, const double slfac,
                          const double ssfac)
{
  itsQlen = qlen;
  itsModels.clear();
  itsInitialModel.init(updatefac, sampleval, samplevar);
  itsNeighSigma = neighsigma; itsLocSigma = locsigma;
  itsNweights.freeMem();
  itsNeighUpdFac = neighupdatefac;
  itsProbe = probe;
  itsSLfac = slfac; itsSSfac = ssfac;
}

// ######################################################################
template <class T>
SurpriseMap<T>::~SurpriseMap()
{ }

// ######################################################################
template <class T>
void SurpriseMap<T>::reset()
{
  for (uint i = 0; i < itsModels.size(); i ++)
    itsModels[i].reset();
}

// ######################################################################
template <class T>
Image<double> SurpriseMap<T>::surprise(const SurpriseImage<T>& sample)
{
  // is it the first time we are called? if so, we need to setup our
  // size and reset our neighborhood cache:
  if (itsModels.empty())
    {
      // resize and reset our queue of models:
      SurpriseImage<T> models(sample.getDims()); // NOTE: uninitialized models
      models.clear(itsInitialModel); models.reset();
      itsModels.clear();
      for (uint i = 0; i < itsQlen; i ++) itsModels.push_back(models);

      // compute our Difference-of-Gaussians mask of weights:
      int w = sample.getWidth(), h = sample.getHeight();
      float sigma = itsNeighSigma * float(std::max(w, h));
      Dims d(w * 2 + 1, h * 2 + 1); Point2D p(w, h);
      itsNweights = gaussianBlob<float>(d, p, sigma, sigma);
      itsNweights -= gaussianBlob<float>(d, p, itsLocSigma,
                                         itsLocSigma) *
        (itsLocSigma*itsLocSigma / (sigma * sigma) * 1.5f);

      itsNweights.rectify();  // eliminate negative values

      float mi, ma; itsNweights.getMinMax(mi, ma);
      itsNWmin = 0.01f * ma;
    }
  else if (itsModels[0].isSameSize(sample) == false)
    LFATAL("Inconsistent input size! -- ABORT");
  
  // each model feeds into the next one. The first (fastest) model
  // receives the sample from the image as input, and is updated. The
  // updated model then serves as input to the next slower model, and
  // so on. Total surprise is the product from all models:
  SurpriseImage<T> input(sample);
  Image<double> s;
  bool doprobe = input.coordsOk(itsProbe);
  double locmean = 0.0, locvar = 0.0, neimean = 0.0, neivar = 0.0; // for probe

  for (uint i = 0; i < itsModels.size(); i ++)
    {
      // compute the neighborhood models from our current (old) local models:
      SurpriseImage<T> neigh;
      neigh.neighborhoods(input, itsNweights, itsNWmin);
      if (itsNeighUpdFac != 0.0) // use different update fac for neighs?
        neigh.resetUpdFac(itsNeighUpdFac); // higher updfac -> stronger popout

      // show values at a probe location?
      if (doprobe)
        {
          locmean = itsModels[i].getMean().getVal(itsProbe);
          locvar = itsModels[i].getVar().getVal(itsProbe);
          neimean = neigh.getMean().getVal(itsProbe);
          neivar = neigh.getVar().getVal(itsProbe);
        }
      
      // update local models and compute the local temporal surprise:
      Image<double> sl = itsModels[i].surprise(input);

      // compute spatial surprise:
      Image<double> ss = neigh.surprise(input);

      // the total surprise is roughly the pointwise product between
      // spatial and temporal surprises, for the current model, with a
      // small pedestal of spatial surprise to account for tonic
      // responses to static stimuli:
      Image<double> stot(sl.getDims(), ZEROS);
      if (itsSLfac)
        {
          if (itsSLfac != 1.0) stot = sl * itsSLfac; else stot = sl;
        }
      if (itsSSfac)
        {
          if (itsSSfac != 1.0) stot += ss * itsSSfac; else stot += ss;
        }

      // total surprise combines multiplicatively across models with
      // different time scales:
      if (i == 0) s = stot; else s *= stot;

      // save debug output from a probe location:
      if (doprobe)
        {
          LERROR("MODELS: %d   %g %g   %g %g   %g %g", i,
                 input.getMean().getVal(itsProbe),
                 input.getVar().getVal(itsProbe),
                 locmean, locvar, neimean, neivar);
          LERROR("SURPRISE: %d %g %g", i, sl.getVal(itsProbe),
                 ss.getVal(itsProbe));
        }
      
      // the updated models are the input to the next iteration:
      input = itsModels[i];
    }

  // the results here should always be positive but sometimes turn
  // negative due to rounding errors in the surprise computation.
  // Let's clamp just to be sure, otherwise we'll end up with some
  // NaNs at later stages:
  s.rectify();

  // calm down total surprise:
  s = toPower(s, 1.0 / (3.0*double(itsModels.size())));
  
  // return total surprise:
  return s;
}

// ######################################################################
template <class T>
const SurpriseImage<T>& SurpriseMap<T>::getSurpriseImage(const
                                                         uint index) const
{
  ASSERT(index < itsModels.size());
  return itsModels[index];
}


// ######################################################################
// explicit instantiations:
template class SurpriseMap<SurpriseModelSG>;
template class SurpriseMap<SurpriseModelSP>;
template class SurpriseMap<SurpriseModelOD>;

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
