/*!@file SurpriseModel.C a local (single-point) model of surprise */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SurpriseModel.C 4188 2005-02-13 07:01:15Z itti $
//

#include "SurpriseModel.H"
#include "SurpriseImage.H"
#include "log.H"
#include "MathFunctions.H"

// ######################################################################
// SurpriseModel implementation
// ######################################################################
SurpriseModel::SurpriseModel(const double updatefac, const double sampleval,
                             const double samplevar) :
  itsUpdateFac(updatefac), itsInitialVal(sampleval), itsInitialVar(samplevar)
{ }

// ######################################################################
SurpriseModel::~SurpriseModel()
{ }

// ######################################################################
void SurpriseModel::init(const double updfac,
                         const double sampleval,
                         const double samplevar)
{ itsInitialVal = sampleval; itsInitialVar = samplevar; itsUpdateFac = updfac;}

// ######################################################################
void SurpriseModel::resetUpdFac(const double updfac)
{ itsUpdateFac = updfac; }

// ######################################################################
double SurpriseModel::surprise(const SurpriseModel& other)
{ LFATAL("Unimplemented! use derived classes instead -- ABORT"); return 0.0; }

// ######################################################################
void SurpriseModel::combineFrom(const SurpriseImage<SurpriseModel>& models, 
                           const Image<float>& weights)
{ LFATAL("Unimplemented! use derived classes instead -- ABORT"); }

// ######################################################################
void SurpriseModel::combineFrom(const SurpriseImage<SurpriseModel>& models, 
                                const Image<float>& weights,
                                const Point2D& pos, const float wmin)
{ LFATAL("Unimplemented! use derived classes instead -- ABORT"); }




// ######################################################################
// SurpriseModelSG implementation
// ######################################################################
SurpriseModelSG::SurpriseModelSG(const double updatefac,
                                 const double sampleval,
                                 const double samplevar) :
  SurpriseModel(updatefac, sampleval, samplevar),
  itsMean(sampleval),
  itsVar(samplevar)
{ }

// ######################################################################
SurpriseModelSG::~SurpriseModelSG()
{ }

// ######################################################################
void SurpriseModelSG::reset()
{ load(itsInitialVal, itsInitialVar); }

// ######################################################################
void SurpriseModelSG::init(const double updfac,
                           const double sampleval,
                           const double samplevar)
{
  SurpriseModel::init(updfac, sampleval, samplevar);
  reset();
}

// ######################################################################
void SurpriseModelSG::load(const double sampleval,
                           const double samplevar)
{ itsMean = sampleval; itsVar = samplevar; }
 
// ######################################################################
double SurpriseModelSG::surprise(const SurpriseModelSG& sample)
{
  const double mean1 = itsMean, var1 = itsVar;
  const double mean2 = sample.itsMean, var2 = sample.itsVar;

  // let's find out the mean and variance of our model once we update
  // it with the incoming sample (i.e., let's compute the posterior):
  const double fac1 = itsUpdateFac, fac2 = 1.0 - fac1;
  const double newMean = fac1 * mean1 + fac2 * mean2;
  const double newVar = fac1 * (var1 + mean1 * mean1) +
    fac2 * (var2 + mean2 * mean2) - newMean * newMean;

  // surprise is KL(new || old):
  // KL(Gq || Gp) = 1/2 [Sq^2/Sp^2 - 1 - log(Sq^2/Sp^2) + (Mq-Mp)^2/Sp^2]
  // here, new = Gq = G(newMean, newVar) and old = Gp = G(mean1, var1) 
  double x = newVar / var1, mm = newMean - mean1; mm *= mm;
  double s = 0.5 * (x - 1.0 - log(x) + mm/var1);

  // the posterior becomes our new prior:
  itsMean = newMean; itsVar = newVar;

  // surprise should always be positive but rounding may mess that up:
  if (s < 0.0) s = 0.0;

  // return the surprise
  return s;
}

// ######################################################################
void SurpriseModelSG::combineFrom(const SurpriseImage<SurpriseModelSG>&
                                  models, const Image<float>& weights)
{
  ASSERT(models.isSameSize(weights));

  // combined mean is just the weighted sum of all the means; the
  // formula for combined variance is barely more complicated:
  itsMean = 0.0; itsVar = 0.0;
  SurpriseImage<SurpriseModelSG>::const_iterator m = models.begin(),
    stop = models.end();
  Image<float>::const_iterator w = weights.begin();
  double wsum = 0.0;  // find out how much total weight we use
  while(m != stop) 
    {
      const double mmean = m->itsMean;
      const double weight = double(*w++);
      itsMean += weight * mmean;
      itsVar += weight * (m->itsVar + mmean * mmean);
      wsum += weight;
      ++m;
    }
  itsMean /= wsum; itsVar /= wsum;
  itsVar -= itsMean * itsMean;
}

// ######################################################################
void SurpriseModelSG::combineFrom(const SurpriseImage<SurpriseModelSG>&
                                  models, const Image<float>& weights,
                                  const Point2D& pos, const float wmin)
{
  int width = models.getWidth(), height = models.getHeight();
  ASSERT(weights.getWidth() == 2 * width + 1);
  ASSERT(weights.getHeight() == 2 * height + 1);
  ASSERT(models.coordsOk(pos));

  // combined mean is just the weighted sum of all the means; the
  // formula for combined variance is barely more complicated:
  itsMean = 0.0; itsVar = 0.0;
  SurpriseImage<SurpriseModelSG>::const_iterator m = models.begin();
  Image<float>::const_iterator w = weights.begin();

  // define bounds of our scan in the weight mask and go for it:
  w += width - pos.i + (2 * width + 1) * (height - pos.j);
  double wsum = 0.0;  // find out how much total weight we use
  for (int j = 0; j < height; j ++)
    {
      for (int i = 0; i < width; i ++)
        {
          const double weight = double(*w++);
          if (weight > wmin)
            {
              const double mmean = m->itsMean;
              itsMean += weight * mmean;
              itsVar += weight * (m->itsVar + mmean * mmean);
              wsum += weight;
            }
          ++m;
        }
      // skip to next line:
      w += width + 1;
    }
  itsMean /= wsum; itsVar /= wsum;
  itsVar -= itsMean * itsMean;
}

// ######################################################################
double SurpriseModelSG::getMean() const
{ return itsMean; }

// ######################################################################
double SurpriseModelSG::getVar() const
{ return itsVar; }


// ######################################################################
// SurpriseModelSP implementation
// ######################################################################
SurpriseModelSP::SurpriseModelSP(const double updatefac,
                                 const double sampleval,
                                 const double samplevar) :
  SurpriseModel(updatefac, sampleval, samplevar),
  itsN(1),
  itsAlpha(itsN * sampleval / (1.0 - updatefac)),
  itsBeta(itsN / (1.0 - updatefac))
{ }

// ######################################################################
SurpriseModelSP::~SurpriseModelSP()
{ }

// ######################################################################
void SurpriseModelSP::reset()
{ load(itsInitialVal, itsInitialVar); }

// ######################################################################
void SurpriseModelSP::init(const double updfac,
                           const double sampleval,
                           const double samplevar)
{
  SurpriseModel::init(updfac, sampleval, samplevar);
  load(sampleval, samplevar);
}

// ######################################################################
void SurpriseModelSP::load(const double sampleval,
                           const double samplevar)
{
  itsAlpha = itsN * sampleval / (1.0 - itsUpdateFac);
  itsBeta = itsN / (1.0 - itsUpdateFac);
}
 
// ######################################################################
double SurpriseModelSP::surprise(const SurpriseModelSP& sample)
{
  // first, decay alpha and beta:
  itsAlpha *= itsUpdateFac; itsBeta *= itsUpdateFac;

  // let's find out the alpha and beta of our model once we update
  // it with the incoming sample (i.e., let's compute the posterior):
  const double newAlpha = itsAlpha + itsN * sample.itsAlpha / sample.itsBeta;
  const double newBeta = itsBeta + itsN;

  // surprise is KL(new || old):
  // S = a2 log(b1/b2) + log(G(a2)/G(a1)) + b2 a1/b1 - a1 + (a1-a2) psi(a1)
  // with G the Euler Gamma function and psi the digamma.
  // Here, new = Gam(a1, b1) = Gam(newAlpha, newBeta) and
  // old = Gam(a2, b2) = Gam(itsAlpha, itsBeta)
  double s = itsAlpha*log(newBeta/itsBeta) + lgamma(itsAlpha) -
    lgamma(newAlpha) + newAlpha*itsBeta/newBeta - newAlpha +
    (newAlpha - itsAlpha)*psi(newAlpha);

  // the posterior becomes our new prior:
  itsAlpha = newAlpha; itsBeta = newBeta;

  // surprise should always be positive but rounding may mess that up:
  if (s < 0.0) s = 0.0;

  // return the surprise
  return s;
}

// ######################################################################
void SurpriseModelSP::combineFrom(const SurpriseImage<SurpriseModelSP>&
                                  models, const Image<float>& weights)
{
  ASSERT(models.isSameSize(weights));

  // combined alpha is the weighted sum of means, and combined beta is
  // the sum of weights (both possibly scaled by a factor):
  itsAlpha = 0.0; itsBeta = 0.0;
  SurpriseImage<SurpriseModelSP>::const_iterator m = models.begin(),
    stop = models.end();
  double wsum = 0.0;  // find out how much total weight we use
  Image<float>::const_iterator w = weights.begin();
  while(m != stop) 
    {
      const double weight = double(*w++);
      itsAlpha += weight * m->itsAlpha / m->itsBeta;
      itsBeta += weight * m->itsBeta;
      wsum += weight;
      ++m;
    }
  // scale alpha and beta so that if all local models were the same we
  // get the same parameters for our neighborhood model:
  itsBeta /= wsum;
  itsAlpha *= itsBeta / wsum;
}

// ######################################################################
void SurpriseModelSP::combineFrom(const SurpriseImage<SurpriseModelSP>&
                                  models, const Image<float>& weights,
                                  const Point2D& pos, const float wmin)
{
  int width = models.getWidth(), height = models.getHeight();
  ASSERT(weights.getWidth() == 2 * width + 1);
  ASSERT(weights.getHeight() == 2 * height + 1);
  ASSERT(models.coordsOk(pos));

  // combined mean is just the weighted sum of all the means; the
  // formula for combined variance is barely more complicated:
  itsAlpha = 0.0; itsBeta = 0.0;
  SurpriseImage<SurpriseModelSP>::const_iterator m = models.begin();
  Image<float>::const_iterator w = weights.begin();
  double wsum = 0.0;  // find out how much total weight we use

  // define bounds of our scan in the weight mask and go for it:
  w += width - pos.i + (2 * width + 1) * (height - pos.j);
  for (int j = 0; j < height; j ++)
    {
      for (int i = 0; i < width; i ++)
        {
          const double weight = double(*w++);
          if (weight > wmin)
            {
              itsAlpha += weight * m->itsAlpha / m->itsBeta;
              itsBeta += weight * m->itsBeta;
              wsum += weight;
            }
          ++m;
        }
      // skip to next line:
      w += width + 1;
    }
  // scale alpha and beta so that if all local models were the same we
  // get the same parameters for our neighborhood model:
  itsBeta /= wsum;
  itsAlpha *= itsBeta / wsum;
}

// ######################################################################
double SurpriseModelSP::getMean() const
{ return itsAlpha / itsBeta; }

// ######################################################################
double SurpriseModelSP::getVar() const
{ return itsAlpha / (itsBeta * itsBeta); }


// ######################################################################
// SurpriseModelOD implementation
// ######################################################################
SurpriseModelOD::SurpriseModelOD(const double updatefac,
                                 const double sampleval,
                                 const double samplevar) :
  SurpriseModel(updatefac, sampleval, samplevar),
  itsN(1),
  itsLambda(sampleval)
{ }

// ######################################################################
SurpriseModelOD::~SurpriseModelOD()
{ }

// ######################################################################
void SurpriseModelOD::reset()
{ load(itsInitialVal, itsInitialVar); }

// ######################################################################
void SurpriseModelOD::init(const double updfac,
                           const double sampleval,
                           const double samplevar)
{
  SurpriseModel::init(updfac, sampleval, samplevar);
  load(sampleval, samplevar);
}

// ######################################################################
void SurpriseModelOD::load(const double sampleval,
                           const double samplevar)
{
  itsLambda = sampleval;
}
 
// ######################################################################
double SurpriseModelOD::surprise(const SurpriseModelOD& sample)
{
  // let's compute how likely the data is given our current model:
  double ret = poisson(int(sample.itsLambda+0.4999), itsLambda);

  // now update our lambda:
  itsLambda = itsLambda + itsUpdateFac * (sample.itsLambda - itsLambda);

  // return the "surprise":
  ret = 1.0 / (ret + 1.0e-5) - 1.0;
  if (ret < 0.0) ret = 0.0;

  return ret * 0.00001;
}

// ######################################################################
void SurpriseModelOD::combineFrom(const SurpriseImage<SurpriseModelOD>&
                                  models, const Image<float>& weights)
{
  ASSERT(models.isSameSize(weights));

  // combined lambda is the weighted sum of lambdas:
  itsLambda = 0.0;
  SurpriseImage<SurpriseModelOD>::const_iterator m = models.begin(),
    stop = models.end();
  double wsum = 0.0;  // find out how much total weight we use
  Image<float>::const_iterator w = weights.begin();
  while(m != stop) 
    {
      const double weight = double(*w++);
      itsLambda += weight * m->itsLambda;
      wsum += weight;
      ++m;
    }
}

// ######################################################################
void SurpriseModelOD::combineFrom(const SurpriseImage<SurpriseModelOD>&
                                  models, const Image<float>& weights,
                                  const Point2D& pos, const float wmin)
{
  int width = models.getWidth(), height = models.getHeight();
  ASSERT(weights.getWidth() == 2 * width + 1);
  ASSERT(weights.getHeight() == 2 * height + 1);
  ASSERT(models.coordsOk(pos));

  // combined lambda is the weighted sum of lambdas:
  itsLambda = 0.0;
  SurpriseImage<SurpriseModelOD>::const_iterator m = models.begin();
  Image<float>::const_iterator w = weights.begin();
  double wsum = 0.0;  // find out how much total weight we use

  // define bounds of our scan in the weight mask and go for it:
  w += width - pos.i + (2 * width + 1) * (height - pos.j);
  for (int j = 0; j < height; j ++)
    {
      for (int i = 0; i < width; i ++)
        {
          const double weight = double(*w++);
          if (weight > wmin)
            {
              itsLambda += weight * m->itsLambda;
              wsum += weight;
            }
          ++m;
        }
      // skip to next line:
      w += width + 1;
    }
}

// ######################################################################
double SurpriseModelOD::getMean() const
{ return itsLambda; }

// ######################################################################
double SurpriseModelOD::getVar() const
{ return itsLambda; }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
