/*!@file SurpriseModel.H a local (single-point) model of surprise */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: SurpriseModel.H 4188 2005-02-13 07:01:15Z itti $
//

#ifndef SURPRISEMODEL_H_DEFINED
#define SURPRISEMODEL_H_DEFINED

template <class T> class SurpriseImage;  // defined in SurpriseImage.H
template <class T> class Image;          // defined in Image.H
#include "Point2D.H"

// ######################################################################
//! A local (single-point) surprise model
/*! This the base class and it cannot be implemented as some of its
  functions are purely virtual. It defines the basic interface. See
  the derivative classes for actual implementations. */
class SurpriseModel
{
public:
  //! Constructor
  /*! Build a new SurpriseModel object and initialize it as if an
    infinitely long sequence of samples had been previously received
    by the model.
    @param updatefac factor by which the current model is combine with
    a new model on an update. Valid values are in [0..1]
    @param sampleval the value of the samples previously received.
    @param samplevar a variance expressing the intrinsic variability
    in the measurement of those samples. */
  SurpriseModel(const double updatefac = 0.5, const double sampleval = 0.0,
                const double samplevar = 1.0);
  
  // default copy-construction and copy operators OK.
  
  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~SurpriseModel();

  //! Reset to initial state
  virtual void reset() = 0;

  //! Change intial state parameters
  virtual void init(const double updatefac = 0.5, const double sampleval = 0.0,
                    const double samplevar = 1.0);

  //! Reset our update factor
  virtual void resetUpdFac(const double updfac);

  //! Load given sample mean and variance
  virtual void load(const double sampleval = 0.0,
                    const double samplevar = 1.0) = 0;
  
  //! Compute surprise between us and another model
  /*! Surprise is the KL distance between our current model and what
    that model will become once we update it with the sample passed in
    as argument.  For the update, our current model is weighted by
    itsUpdateFac while the new model by (1.0-itsUpdateFac). Programmer
    note: derived classes will not provide a direct overload of this
    (and other) base functions, as instead they will implement
    functions with basically the same syntax but using an argument of
    the derived class. Since an implementation is not provided for
    this base function, you will get an error at link time if you call
    the derived functions with a base-class argument. */
  virtual double surprise(const SurpriseModel& sample);

  //! Initialize us as a weighted combination of the given map of models
  /*! @param models an array of models, one model at each pixel of the image.
    @param weights the weight to be given to each model. Weights should sum 
    to 1.0 and this is not checked for here for computational efficiency. */
  virtual void combineFrom(const SurpriseImage<SurpriseModel>& models, 
                           const Image<float>& weights);

  //! Initialize us as a weighted combination of the given map of models
  /*! @param models an array of models, one model at each pixel of the image.
    @param weights the weight to be given to each model. Weights
    should sum to 1.0 and this is not checked for here for
    computational efficiency. In this version, the weight image should
    have double+1 the width and height of the Surprise image, and the
    weight for the current point will be taken at the center of the
    weight map, etc.
    @param pos indicates the 2D coordinates of the center of the
    neighborhood in the SurpriseImage
    @param minw is the minimum weight below which we just don't include a
    model in the update*/
  virtual void combineFrom(const SurpriseImage<SurpriseModel>& models, 
                           const Image<float>& weights, const Point2D& pos,
                           const float wmin = 0.0f);

  //! get our mean
  virtual double getMean() const = 0;

  //! get our variance
  virtual double getVar() const = 0;

protected:
  double itsUpdateFac;  //!< our update factor
  double itsInitialVal; //!< our initial value
  double itsInitialVar; //!< our initial variance
};

// ######################################################################
//! A single-Gaussian SurpriseModel
/*! This is a very simple SurpriseModel consisting of a single
  Gaussian, which is updated in a sliding-average manner. */
class SurpriseModelSG : public SurpriseModel
{
public:
  //! Constructor. See base class for details.
  SurpriseModelSG(const double updatefac = 0.5, const double sampleval = 0.0,
                  const double samplevar = 1.0);
  
  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~SurpriseModelSG();

  //! Reset to initial state
  virtual void reset();

  //! Change intial state parameters
  virtual void init(const double updatefac = 0.5, const double sampleval = 0.0,
                    const double samplevar = 1.0);

  //! Load given sample mean and variance
  virtual void load(const double sampleval, const double samplevar = 1.0);
  
  //! Compute surprise between us and another model
  virtual double surprise(const SurpriseModelSG& other);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelSG>& models, 
                           const Image<float>& weights);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelSG>& models, 
                           const Image<float>& weights, const Point2D& pos,
                           const float wmin = 0.0f);

  //! get our mean
  virtual double getMean() const;

  //! get our variance
  virtual double getVar() const;

protected:
  double itsMean;  //!< our current Gaussian mean
  double itsVar;   //!< our current Gaussian variance
};

// ######################################################################
//! A single-Poisson/Gamma SurpriseModel
/*! This is a very simple SurpriseModel consisting of a single Gamma
  prior over Poisson, which is updated in a sliding-average manner. */
class SurpriseModelSP : public SurpriseModel
{
public:
  //! Constructor. See base class for details.
  SurpriseModelSP(const double updatefac = 0.5, const double sampleval = 0.0,
                  const double samplevar = 1.0 /* ignored */);
  
  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~SurpriseModelSP();

  //! Reset to initial state
  virtual void reset();

  //! Change intial state parameters
  virtual void init(const double updatefac = 0.5, const double sampleval = 0.0,
                    const double samplevar = 1.0 /* ignored */);

  //! Load given sample mean and variance
  virtual void load(const double sampleval, const double samplevar = 1.0);
  
  //! Compute surprise between us and another model
  virtual double surprise(const SurpriseModelSP& other);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelSP>& models, 
                           const Image<float>& weights);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelSP>& models, 
                           const Image<float>& weights, const Point2D& pos,
                           const float wmin = 0.0f);

  //! get our mean
  virtual double getMean() const;

  //! get our variance
  virtual double getVar() const;

protected:
  int itsN;         //!< our sample N
  double itsAlpha;  //!< our current Gamma alpha
  double itsBeta;   //!< our current Gamma beta
};

// ######################################################################
//! A single-Poisson Outlier-based model
/*! This is not really a true SurpriseModel, but rather an information
  model based on outlier detection using a single adaptive Poisson
  model. */
class SurpriseModelOD : public SurpriseModel
{
public:
  //! Constructor. See base class for details.
  SurpriseModelOD(const double updatefac = 0.5, const double sampleval = 0.0,
                  const double samplevar = 1.0 /* ignored */);
  
  //! Virtual destructor ensures proper destruction of derived classes
  virtual ~SurpriseModelOD();

  //! Reset to initial state
  virtual void reset();

  //! Change intial state parameters
  virtual void init(const double updatefac = 0.5, const double sampleval = 0.0,
                    const double samplevar = 1.0 /* ignored */);

  //! Load given sample mean and variance
  virtual void load(const double sampleval, const double samplevar = 1.0);
  
  //! Compute surprise between us and another model
  virtual double surprise(const SurpriseModelOD& other);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelOD>& models, 
                           const Image<float>& weights);

  //! Initialize us as a weighted combination of the given map of models
  virtual void combineFrom(const SurpriseImage<SurpriseModelOD>& models, 
                           const Image<float>& weights, const Point2D& pos,
                           const float wmin = 0.0f);

  //! get our mean
  virtual double getMean() const;

  //! get our variance
  virtual double getVar() const;

protected:
  int itsN;          //!< our sample N
  double itsLambda;  //!< our current Poisson lambda
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
