/*!@file TaskRelevanceMap.C Implementation for task-relevance map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: TaskRelevanceMap.C 3048 2004-01-15 03:29:57Z itti $
//

#include "saliency.H"
#include "Image.H"
#include "Image_ShapeOps.H"  // for rescale()
#include "Image_MathOps.H"   // for absDiff()
#include "TaskRelevanceMap.H"
#include "log.H"
#include "ChannelBase.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "FrameSeries.H"
#include "ShapeEstimator.H"
#include "SaccadeController.H"
#include "Channels.H"
#include "Brain.H"
#include "Image_Transforms.H"

// ######################################################################
// ######################################################################
// ########## TaskRelevanceMap implementation
// ######################################################################
// ######################################################################

TaskRelevanceMap::TaskRelevanceMap(ModelManager& mgr,
                                   const std::string& descrName,
                                   const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSaveResults(this, "TRMsaveResults", false), // see ModelOptionDefs.C
  itsTRMFactor(this, "TRMfactor", 100.0F),
  itsBrain(NULL)
{ }

// ######################################################################
TaskRelevanceMap::~TaskRelevanceMap()
{ }

// ######################################################################
void TaskRelevanceMap::exportOptions(const int whatToExport,
                                     const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsTRMFactor);

  if (whatToExport & OPTEXP_SAVE)
    manager->requestOption(this, itsSaveResults);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void TaskRelevanceMap::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  if (itsSaveResults.getVal())
    ofs->writeFloat(getV(true), PNM, -1, true, "TRM");
}

// ######################################################################
void TaskRelevanceMap::setBrain(Brain* brain)
{ itsBrain = brain; }

// ######################################################################
// ######################################################################
// ########## TaskRelevanceMapConfigurator implementation
// ######################################################################
// ######################################################################
TaskRelevanceMapConfigurator::
TaskRelevanceMapConfigurator(ModelManager& mgr,
                             const std::string& descrName,
                             const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsTRMtype(this, "TaskRelevanceMapType", "None"),
  itsTRM(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
TaskRelevanceMapConfigurator::~TaskRelevanceMapConfigurator()
{  }

// ######################################################################
void TaskRelevanceMapConfigurator::exportOptions(const int whatToExport,
                                                 const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsTRMtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<TaskRelevanceMap> TaskRelevanceMapConfigurator::getTRM() const
{ return itsTRM; }

// ######################################################################
bool TaskRelevanceMapConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsTRMtype.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsTRM.get()) { removeSubComponent(itsTRM); itsTRM.reset(NULL); }

    // instantiate a SM of the appropriate type:
    if (value.compare("None") == 0)              // no TRM
      itsTRM.reset(NULL);
    else if (value.compare("Std") == 0)          // standard
      itsTRM.reset(new TaskRelevanceMapStd(*manager));
    else if (value.compare("KillStatic") == 0)   // kill-static
      itsTRM.reset(new TaskRelevanceMapKillStatic(*manager));
    else if (value.compare("KillN") == 0)        // kill-n
      itsTRM.reset(new TaskRelevanceMapKillN(*manager));
    else
      LFATAL("Unknown TRM type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsTRM.get()) {
      addSubComponent(itsTRM);

      // tell the controller to export its options:
      itsTRM->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected TRM of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ########## TaskRelevanceMapStd implementation
// ######################################################################
// ######################################################################

// ######################################################################
TaskRelevanceMapStd::
TaskRelevanceMapStd(ModelManager& mgr, const std::string& descrName,
                    const std::string& tagName) :
  TaskRelevanceMap(mgr, descrName, tagName), Image<float>(),
  itsLearnTRM(this, "LearnTRM", false),
  itsLearnUpdateTRM(this, "LearnUpdateTRM", false),
  itsBiasTRM(this, "BiasTRM", "")
{  }

// ######################################################################
void TaskRelevanceMapStd::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsLearnTRM);
    manager->requestOption(this, itsLearnUpdateTRM);
    manager->requestOption(this, itsBiasTRM);
  }

  // the base class will propagate this call to our subcomponents:
  TaskRelevanceMap::exportOptions(recurse);
}

// ######################################################################
TaskRelevanceMapStd::~TaskRelevanceMapStd()
{ }

// ######################################################################
void TaskRelevanceMapStd::reset()
{ freeMem(); }

// ######################################################################
void TaskRelevanceMapStd::input(const Image<float>& current)
{
  Image<float> in(current); swap(in);
  Dims d = getDims();

  // bias TRM if the option is specified
  if (strcmp(itsBiasTRM.getValString().c_str(), "") != 0)
    {
      LINFO("biasing TRM with %s", itsBiasTRM.getValString().c_str());
        Image<float> newtrm = 
          Raster::ReadGray(Auto, itsBiasTRM.getValString().c_str());
        newtrm = rescale(newtrm * (1.0F/255.0F), d);
        swap(newtrm);
    }
}

// ######################################################################
Image<float> TaskRelevanceMapStd::getV(const bool normalize) const
{
  if (normalize == false) return *this;

  // useful range is 0.0..3.0 with 1.0 being baseline ->
  // we remap 1.0 to 100 or whatever itsTRMFactor may be:
  Image<float> ret(*this);
  ret *= itsTRMFactor.getVal(); ret.clamp(0.0F, 255.0F);
  return ret;
}

// ######################################################################
void TaskRelevanceMapStd::evolve(const double t)
{ }

// ######################################################################
float TaskRelevanceMapStd::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
int TaskRelevanceMapStd::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int TaskRelevanceMapStd::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
void TaskRelevanceMapStd::update(const Point2D& fixation,
                                 const float relevance)
{
  // we need a brain and a shape estimator:
  if (itsBrain == NULL) LFATAL("I need a Brain -- ABORT");

  SharedPtr<ShapeEstimator> se = itsBrain->getSE();
  if (se.get() == NULL) LFATAL("I need a ShapeEstimator -- ABORT");

  // get the object shape:
  Point2D fixn;
  fixn.i = fixation.i * getWidth() / itsBrain->getInputDims().w();
  fixn.j = fixation.j * getHeight() / itsBrain->getInputDims().h();
  if ( ! se->getObjectMask().initialized())
    { LERROR("object mask not initialized!"); return; }
  Image<byte> fmask = rescale(se->getObjectMask(), getDims());
  Point2D p;

  //FIXME: this should be an Image function
  // update relevance only within object
  for (p.i = 0; p.i < getWidth(); p.i ++)
    for (p.j = 0; p.j < getHeight(); p.j ++)
      if (fmask.getVal(p) == 255)
        {
          if (relevance == 0.0f) setVal(p, getVal(p) - 1.0f);
          else setVal(p, getVal(p) + relevance); 
        }

  if (itsLearnTRM.getVal() == false)
    {
      // biasing mode / updating mode / none: 
      // set min to 0 else normalizing will clear the entire TRM
      float min, max; getMinMax(min, max);
      if (min > 0.0f) setVal(0, 0, 0.0f);
    }

  // normalize TRM else AGM will explode
  normalize(0.0f, 3.0f);
  LINFO("Evolved TRM with relevance = %f", relevance);
}

// ######################################################################
// ######################################################################
// ########## TaskRelevanceMapKillStatic implementation
// ######################################################################
// ######################################################################

// ######################################################################
TaskRelevanceMapKillStatic::
TaskRelevanceMapKillStatic(ModelManager& mgr, const std::string& descrName,
                           const std::string& tagName) :
  TaskRelevanceMap(mgr, descrName, tagName), Image<float>(),
  itsKillStaticThresh(this, "TRMKillStaticThresh", 20.0F),
  itsKillStaticCoeff(this, "TRMKillStaticCoeff", 0.95F),
  itsStaticBuff(), itsSC(NULL), itsIntensChan(NULL), itsMapLevel(0),
  itsLastTime(0.0)
{  }

// ######################################################################
void TaskRelevanceMapKillStatic::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsKillStaticThresh);
    manager->requestOption(this, itsKillStaticCoeff);
  }

  // the base class will propagate this call to our subcomponents:
  TaskRelevanceMap::exportOptions(recurse);
}

// ######################################################################
TaskRelevanceMapKillStatic::~TaskRelevanceMapKillStatic()
{ }

// ######################################################################
void TaskRelevanceMapKillStatic::start1()
{
  // does our brain have an SC?
  if (itsBrain->getSC().get()) itsSC = itsBrain->getSC();

  // does it have an intensity channel?
  SharedPtr<VisualCortex> vcx = itsBrain->getVC();
  if (vcx.get() == NULL)
    LFATAL("I need a VisualCortex in my Brain!");
  if (vcx->hasChannel("intensity") == false)
    LFATAL("I need an intensity channel in my VisualCortex!");
  itsIntensChan.dynCastFrom(vcx->channelPtr("intensity"));
  LevelSpec ls = itsIntensChan->getLevelSpec();
  itsMapLevel = ls.mapLevel();
}

// ######################################################################
void TaskRelevanceMapKillStatic::reset()
{ freeMem(); itsStaticBuff.freeMem(); itsLastTime = 0.0; }

// ######################################################################
void TaskRelevanceMapKillStatic::input(const Image<float>& current)
{ Image<float> in(current); swap(in); }

// ######################################################################
Image<float> TaskRelevanceMapKillStatic::getV(const bool normalize) const
{
  if (normalize == false) return *this;

  // useful range is 0.0..3.0 with 1.0 being baseline ->
  // we remap 1.0 to 100 or whatever itsTRMFactor may be:
  Image<float> ret(*this);
  ret *= itsTRMFactor.getVal(); ret.clamp(0.0F, 255.0F);
  return ret;
}

// ######################################################################
void TaskRelevanceMapKillStatic::evolve(const double t)
{
  // reset our static-detection buffer if in saccade:
  if (itsSC.get() && itsSC->isInSaccade())
    { itsStaticBuff.freeMem(); return; }

  // did the input to our IntensityChannel just change?
  if (itsIntensChan->pyramidTime(0) < itsLastTime) return;
  itsLastTime = t;

  // let's build a cumulative image of the luminance channel at the
  // resolution of the saliency map:
  Image<float> img = itsIntensChan->pyramid(0).getImage(itsMapLevel);
    
  // is this our first time here?
  if (itsStaticBuff.initialized() == false) { itsStaticBuff = img; return; }

  // otherwise derive TRM from difference between current frame and
  // staticBuf:
  Image<float> diff = absDiff(itsStaticBuff, img);

  // useful range of diff is 0..255; anything giving an output less
  // that itsKillStaticThresh will be considered static; here let's
  // clamp to isolate that:
  diff.clamp(0.0F, itsKillStaticThresh.getVal());

  // the more static, the greater the killing: so, first let's change
  // the range so that the min is at -itsKillStaticThresh and zero is
  // neutral:
  diff -= itsKillStaticThresh.getVal();

  // itsKillStaticCoeff determines how strong the killing should be:
  diff *= (1.0F / itsKillStaticThresh.getVal()) * // normalize to min at -1.0
    itsKillStaticCoeff.getVal();                  // apply coeff

  // mix our new TRM to the old one; we want to have a fairly high
  // mixing coeff for the new one so that we don't penalize recent
  // onset objects too much (i.e., were previously static and killed,
  // but are not anymore):
  Image<float> newTRM = Image<float>(*this) * 0.25F + (diff + 1.0F) * 0.75F; 
  float mi, ma; newTRM.getMinMax(mi, ma);
  LINFO("TRM range = [%.2f .. %.2f] -- 1.0 is baseline", mi, ma);
  swap(newTRM);

  // update our cumulative buffer:
  itsStaticBuff = itsStaticBuff * 0.75F + img * 0.25F;

}

// ######################################################################
float TaskRelevanceMapKillStatic::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
int TaskRelevanceMapKillStatic::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int TaskRelevanceMapKillStatic::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
void TaskRelevanceMapKillStatic::update(const Point2D& fixation,
                                        const float relevance)
{ LFATAL("Not supported -- ABORT"); }


// ######################################################################
// ######################################################################
// ########## TaskRelevanceMapKillN implementation
// ######################################################################
// ######################################################################

// ######################################################################
TaskRelevanceMapKillN::
TaskRelevanceMapKillN(ModelManager& mgr, const std::string& descrName,
                      const std::string& tagName) :
  TaskRelevanceMap(mgr, descrName, tagName), Image<float>(),
  itsN(this, "TRMkillN", 3), itsCache()
{  }

// ######################################################################
void TaskRelevanceMapKillN::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsN);
  }

  // the base class will propagate this call to our subcomponents:
  TaskRelevanceMap::exportOptions(recurse);
}

// ######################################################################
TaskRelevanceMapKillN::~TaskRelevanceMapKillN()
{ }

// ######################################################################
void TaskRelevanceMapKillN::reset()
{ freeMem(); itsCache.clear(); }

// ######################################################################
void TaskRelevanceMapKillN::input(const Image<float>& current)
{
  // is this our first input ever?
  if (initialized() == false) itsCache.setMaxSize(itsN.getVal());

  // is the input uniform?  In this case, just push it in, otherwise
  // let's segment the objects out:
  float mi, ma; current.getMinMax(mi, ma);
  if (fabs(mi - ma) < 1.0e-10)
    itsCache.push_back(current);
  else
    {
      // let's build a distance map out of the object and threhold it:
      Image<float> dmap = chamfer34(binaryReverse(current, 1.0F), 255.0F);
      dmap.clamp(0.0F, 15.0F); dmap /= 15.0F;  // 0 inside obj, 1 outside

      // get this new input mask into our cache:
      itsCache.push_back(dmap);
    }
  
  // our current relevance map is the min over our cache:
  Image<float> mini = itsCache.getMin();
  swap(mini);
}

// ######################################################################
Image<float> TaskRelevanceMapKillN::getV(const bool normalize) const
{
  if (normalize == false) return *this;

  // useful range is 0.0..3.0 with 1.0 being baseline ->
  // we remap 1.0 to 100 or whatever itsTRMFactor may be:
  Image<float> ret(*this);
  ret *= itsTRMFactor.getVal(); ret.clamp(0.0F, 255.0F);
  return ret;
}

// ######################################################################
void TaskRelevanceMapKillN::evolve(const double t)
{ }

// ######################################################################
float TaskRelevanceMapKillN::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
int TaskRelevanceMapKillN::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int TaskRelevanceMapKillN::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
void TaskRelevanceMapKillN::update(const Point2D& fixation,
                                        const float relevance)
{ LFATAL("Not supported -- ABORT"); }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
