/*!@file TaskRelevanceMap.H Class declarations for task-relevance map class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: TaskRelevanceMap.H 2974 2003-11-07 00:04:47Z itti $
//

#ifndef TASKRELEVANCEMAP_H_DEFINED
#define TASKRELEVANCEMAP_H_DEFINED

#include "Image.H"
#include "ModelComponent.H"
#include "ImageCache.H"
class ModelManager;
class OutputFrameSeries;
class Brain;
class SaccadeController;
class IntensityChannel;

// ######################################################################
//! The task-relevance map base class
// ######################################################################
/*! This is a 2D task-relevance map. It is just a base class with
  virtual function definitions. Various implementations are available
  below.  The TRM is a topographic map where a value of zero indicates
  that a location is not relevant to the task at hand, while a value
  of 1.0 indicates neutral relevance and values higher than 1.0
  indicate high relevance. */ 
class TaskRelevanceMap : public ModelComponent
{
 public:
  // ######################################################################
  //! @name Constructor, destructor, and reset
  //@{

  //! Ininitialized constructor
  /*! The map will be resized and initialized the first time input() is
    called */
  TaskRelevanceMap(ModelManager& mgr,
                   const std::string& descrName = "Task-Relevance Map",
                   const std::string& tagName = "TaskRelevanceMap");
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. A base
    implementation is provided that will only export the base class
    options (see the ModelParam definitions in the protected section
    below). */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~TaskRelevanceMap();

  //! Reset to initial state just after construction
  virtual void reset() = 0;

  //! Setup a hook back to our brain
  void setBrain(Brain* brain);

  //@}

  // ######################################################################
  //! @name Input, evolve and output
  //@{

  //! Set new relevance values
  /*! This should initialize and resize the map if it is currently
      uninitialized (e.g., just after construction or reset()). */
  virtual void input(const Image<float>& in) = 0;

  //! Get width of the map
  virtual int getWidth() const = 0;

  //! Get height of the map
  virtual int getHeight() const = 0;

  //! Return all our relevance values as an Image<float>
  /*! The current values are possibly normalized to 0..255. The
    normalization will use itsTRMFactor if it is non-null */
  virtual Image<float> getV(const bool normalize = false) const = 0;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const = 0;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t) = 0;

  //! update TRM with relevance of current fixation
  /*! @param fixation enter the current fixation point
    @param relevance enter the relevance of current fixation */
  virtual void update(const Point2D& fixation, const float relevance) = 0;

  //@}

  // ######################################################################
  //! @name Results
  //@{

  //! Save our various results
  /*! A default implementation is provided, which just calls
      getV() and saves that map with a file name prefix of "SM".
      @param ofs will figure out file name/format and save our
      results. */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

  // ######################################################################
protected:
  //! Save our internals when saveResults() is called?
  ModelParam<bool> itsSaveResults;

  //! Multiply by a factor rather than normalize when doing getV()?
  ModelParam<float> itsTRMFactor;

  //! Our brain, or SharedPtr<NULL> if it has not been set with setBrain()
  Brain* itsBrain;

private:
  // forbid assignment and copy-construction:
  TaskRelevanceMap& operator=(const TaskRelevanceMap& sm);
  TaskRelevanceMap(const TaskRelevanceMap& sm);
};

// ######################################################################
//! TaskRelevanceMap configurator
// ######################################################################
/*! This will export the --trm-type=XX command-line option and will
  instantiate a TRM of the desired type as the option gets assigned a
  value. As this happens, new options may become available in the
  command-line. To see them, use --help AFTER you have chosen the type
  to use. The current TRM may be retrieved using getTRM(). */
class TaskRelevanceMapConfigurator : public ModelComponent
{
public:
  //! Constructor
  TaskRelevanceMapConfigurator(ModelManager& mgr,
                               const std::string& descrName =
                               "Task-Relevance Map Configurator",
                               const std::string& tagName =
                               "TaskRelevanceMapConfigurator");

  //! destructor
  virtual ~TaskRelevanceMapConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our SM once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen TRM
  /*! You should call this during start() of the ModelComponent that
    needs the TRM. CAUTION: this may be a SharedPtr(NULL) if no
    TRM type has been chosen. */
  SharedPtr<TaskRelevanceMap> getTRM() const;

protected:
  ModelParam<std::string> itsTRMtype; //!< type of map

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<TaskRelevanceMap> itsTRM; // the map
  int itsWhatToExport;
};


// ######################################################################
//! The standard task-relevance map
// ######################################################################
/*! This is our current standard TRM implementation. It can be loaded
  all at once using input() or updated for a specific location using
  update(); evolve() is a no-op. It's a pure spatial short-term
  memory. */
class TaskRelevanceMapStd : public TaskRelevanceMap, public Image<float>
{
 public:
  //! Uninitialized constructor
  TaskRelevanceMapStd(ModelManager& mgr, const std::string& descrName =
                      "Task-Relevance Map Std",
                      const std::string& tagName =
                      "TaskRelevanceMapStd");

  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~TaskRelevanceMapStd();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new relevance values
  virtual void input(const Image<float>& in);

  //! Get width of the map
  virtual int getWidth() const;

  //! Get height of the map
  virtual int getHeight() const;

  //! Return all our relevance values as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t);

  //! update TRM with relevance of current fixation
  virtual void update(const Point2D& fixation, const float relevance);

protected:
  /*! Learn the TRM for the given task by inspecting several scenes */
  ModelParam<bool> itsLearnTRM;

   /*! Learn and update the TRM as we inspecting several scenes */
  ModelParam<bool> itsLearnUpdateTRM;

  /*! If an image is specified with this option, then use it to bias the TRM */
  ModelParam<std::string> itsBiasTRM;
};

// ######################################################################
//! A killStatic task-relevance map
// ######################################################################
/*! This is a very simple TRM where image locations that do not change
  much over time see theri relevance progressively decrease. */
class TaskRelevanceMapKillStatic : public TaskRelevanceMap, public Image<float>
{
 public:
  //! Uninitialized constructor
  TaskRelevanceMapKillStatic(ModelManager& mgr, const std::string& descrName =
                      "Task-Relevance Map KillStatic",
                      const std::string& tagName =
                      "TaskRelevanceMapKillStatic");

  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~TaskRelevanceMapKillStatic();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new relevance values
  virtual void input(const Image<float>& in);

  //! Get width of the map
  virtual int getWidth() const;

  //! Get height of the map
  virtual int getHeight() const;

  //! Return all our relevance values as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t);

  //! update TRM with relevance of current fixation
  /*! Not supported, will throw a fatal error if called. */
  virtual void update(const Point2D& fixation, const float relevance);

protected:
  //! get ourselves started
  virtual void start1();

  /*! Anything in the difference between current and cumulated
    previous frames (range 0..255) less than this will be considered
    static */
  ModelParam<float> itsKillStaticThresh;

  /*! Max killing achievable for a purely static location (should be
    in 0.0 .. 1.0 range) */
  ModelParam<float> itsKillStaticCoeff;

  // cumulative buffer of the luminance image, at the resolution of
  // the saliency map. This is a sliding weighted average of the
  // downscaled luminance of the past few input frames, and it is used
  // to determine how different the current input is from that
  // accumulated average, and in turn to set the task relevance map so
  // that more static regions are less relevant.
  Image<float> itsStaticBuff;

  SharedPtr<SaccadeController> itsSC;
  SharedPtr<IntensityChannel> itsIntensChan;
  int itsMapLevel;
  double itsLastTime;
};

// ######################################################################
//! A KillN task-relevance map
// ######################################################################
/*! This is an alternative to IOR, in which we set the relevance of
  the last N attended locations to zero. */
class TaskRelevanceMapKillN : public TaskRelevanceMap, public Image<float>
{
 public:
  //! Uninitialized constructor
  TaskRelevanceMapKillN(ModelManager& mgr, const std::string& descrName =
                      "Task-Relevance Map KillN",
                      const std::string& tagName =
                      "TaskRelevanceMapKillN");

  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Destructor
  virtual ~TaskRelevanceMapKillN();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new relevance values
  /*! The expected input here is an object mask, with zero values
    inside the attended object and one outside, at the scale and
    dims of the saliency map. Typically, this can be obtained from an
    object mask given by a ShapeEstimator. After each input, our
    relevance map is updated as a compound over the last N objects
    (with zero relevance at those objects, and unit relevance outside,
    with fuzzy boundaries). */
  virtual void input(const Image<float>& in);

  //! Get width of the map
  virtual int getWidth() const;

  //! Get height of the map
  virtual int getHeight() const;

  //! Return all our relevance values as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update internal state
  virtual void evolve(const double t);

  //! update TRM with relevance of current fixation
  /*! Not supported, will throw a fatal error if called. */
  virtual void update(const Point2D& fixation, const float relevance);

protected:
  //! Number of inputs to compute over
  ModelParam<int> itsN;

private:
  ImageCacheMinMax<float> itsCache;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
