/*!@file VisualBuffer.C Grab ausio samples from /dev/dsp */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualBuffer.C 3486 2004-08-05 18:53:01Z itti $
//

#include "saliency.H"
#include "log.H"
#include "VisualBuffer.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Image_MathOps.H"
#include "Image_ShapeOps.H"
#include "Image_Transforms.H"
#include "Image_FilterOps.H"
#include "fancynorm.H"
#include "Retina.H"
#include "Brain.H"

// ######################################################################
VisualBuffer::VisualBuffer(ModelManager& mgr, const std::string& descrName,
			   const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsIgnoreBoring(this, "VBignoreBoring", true), // see ModelOptionDefs.C
  itsObjectBased(this, "VBobjectBased", true), // see ModelOptionDefs.C
  itsBufferDims(this, "VBdims", Dims(0, 0)),  // see ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()), // see ModelOptionDefs.C
  itsInputDims(this, "InputFrameDims", Dims(0, 0)), // see ModelOptionDefs.C
  itsFOAradius(this, "FOAradius", 30),  // see ModelOptionDefs.C
  itsTimePeriod(this, "VBtimePeriod", 0.05),  // see ModelOptionDefs.C
  itsNormalizeFactor(this, "VBfactor", 50000.0),
  itsDecayFactor(this, "VBdecayFactor", 1.0F),
  itsNormType(this, "VBmaxNormType", VCXNORM_FANCY), // see ModelOptionDefs.C
  itsBuffer(), itsSMdims(), itsSaliencyMask(), itsTime(0.0),
  itsLastInteractTime(0.0), itsBrain(NULL)
{  }

// ######################################################################
void VisualBuffer::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsIgnoreBoring);
    manager->requestOption(this, itsObjectBased);
    manager->requestOption(this, itsBufferDims);
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsInputDims);
    manager->requestOption(this, itsFOAradius);
    manager->requestOption(this, itsTimePeriod);
    manager->requestOption(this, itsNormalizeFactor);
    manager->requestOption(this, itsDecayFactor);
    manager->requestOption(this, itsNormType);
  }

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void VisualBuffer::start1()
{
  // initialize our buffer if we have valid dims for it:
  Dims d = itsBufferDims.getVal();
  if (d.w() > 0 && d.h() > 0) {
      itsBuffer.resize(d, true);
      LINFO("Using buffer dims of (%d, %d)", d.w(), d.h());
  }
  LINFO("Using internal maxnorm of type %s, decay %f",
        itsNormType.getValString().c_str(), itsDecayFactor.getVal());
}

// ######################################################################
void VisualBuffer::setBrain(Brain *brain)
{ itsBrain = brain; }

// ######################################################################
VisualBuffer::~VisualBuffer()
{  }

// ######################################################################
void VisualBuffer::input(const WTAwinner& win, const Image<float>& sm,
                         const Image<byte>& objmask)
{
  // update our sm dims:
  itsSMdims = sm.getDims();

  // if our buffer size has not been set yet, do it now, and use the
  // size of the sm as our buffer size:
  if (itsBuffer.initialized() == false) {
    itsBuffer.resize(itsSMdims, true);
    LINFO("Using buffer dims of (%d, %d)", itsSMdims.w(), itsSMdims.h());
  }

  // ignore boring attention shifts:
  if (win.boring && itsIgnoreBoring.getVal())
    { LINFO("Ignoring boring attention shift"); return; }

  // Let's build a multiplicative mask that we will apply to the sm
  // and that will determine what saliency information gets
  // transferred into the buffer. We have two modes here, either
  // object-based (only transfer attended object) or map-based
  // (transfer whole map). In addition, we have two sub-cases
  // depending on whether objmask is initialized (then use it to
  // define the object) or not (then use a disk). 

  Image<float> maskedsm; // the sm masked by our transfer mask
  Image<float> mask;     // the saliency mask

  if (itsObjectBased.getVal()) {
    // Object-based model. We start by building a float mask at retinal
    // resolution and coordinates:
    Image<byte> maskb;

    if (objmask.initialized()) {
      // build our mask using the object definition passed to us:
      maskb = objmask;          // get the (fuzzy-boundary) object shape
      maskb.lowThresh(255);     // get tight object boundaries
    } else {
      // build our mask using a disk and distance-based decay:
      maskb.resize(itsInputDims.getVal(), true);  // create empty mask
      maskb.drawDisk(win.p, itsFOAradius.getVal(), 255);
    }

    maskb = chamfer34(maskb);  // introduce a distance-based spatial decay
    mask = maskb;              // convert to float; range is 0..255
    mask = rescale(mask, sm.getDims()); // downsize
    mask = binaryReverse(mask, 255.0F);
    mask = squash(mask, 0.0F, 0.0F, 128.0F, 0.25F, 255.0F, 1.0F); // squash
    maskedsm = sm * mask;
  } else {
    // we are not object-based and want to transfer the whole map:
    maskedsm = sm;
    mask.resize(sm.getDims()); mask.clear(1.0F);
  }

  // now let's take the max between our current buffer and our masked
  // sm, after shifting the masked sm to world-centered coordinates.
  // We start by computing the world coords of the top-left corner of
  // the masked sm, at maplevel:
  Point2D tl = retinalToBuffer(Point2D(0, 0));
  Image<float> newbuf(itsBuffer.getDims(), ZEROS);
  pasteImage(newbuf, maskedsm, 0.0F, tl);

  itsBuffer = takeMax(itsBuffer, newbuf);

  // apply one iteration of our internal dynamics:
  internalDynamics();
  
  // for display purposes, keep a copy of the mask used to transfer saliency:
  itsSaliencyMask = Image<byte>(mask * 255.0F);

  // the internal dynamics of the buffer are taken care of in evolve()
}

// ######################################################################
Image<byte> VisualBuffer::getSaliencyMask() const
{ return itsSaliencyMask; }

// ######################################################################
void VisualBuffer::evolve(const double t)
{
  itsTime = t;
    // apply one iteration of our internal dynamics if the time has come:
  if (itsTime - itsLastInteractTime >= itsTimePeriod.getVal())
    internalDynamics();  // will update itsLastInteractTime
}

// ######################################################################
void VisualBuffer::inhibit(const Point2D& loc)
{
  Image<float> mask(itsBuffer.getDims(), ZEROS);
  mask.drawDisk(loc,
                itsFOAradius.getVal() >> itsLevelSpec.getVal().mapLevel(),
                1.0F);
  inhibit(mask);
}

// ######################################################################
void VisualBuffer::inhibit(const Image<float>& mask)
{
  Image<float> inhib = binaryReverse(mask, 1.0F);
  itsBuffer *= inhib;
}

// ######################################################################
Point2D VisualBuffer::findMostInterestingTarget(const Point2D& p)
{
  // let's start by cutting-off locations below a given fraction of
  // the max activation over the buffer, and binarize the rest:
  Image<float> buf = itsBuffer; float mi, ma; buf.getMinMax(mi, ma);
  Image<byte> bin = makeBinary(buf, ma * 0.25F, 0, 1);

  // now let's find the cluster that is closest to our current eye
  // position. For that, we compute a distance map from a single pixel
  // at current eye position, multiply it by our binary mask, and look
  // for the smallest non-zero value:
  Image<byte> dmap(bin.getDims(), ZEROS);
  dmap.setVal(p, 255); dmap = chamfer34(dmap);
  Image<byte> prod = bin * dmap;
  prod.replaceVal(0, 255); // eliminate zero distances outside our blobs

  Point2D minp; byte minval;
  prod.findMin(minp, minval);

  // minp belongs to our closest cluster. Let's segment that cluster
  // out so that we can compute its center of gravity:
  Image<byte> obj; bin.flood(obj, minp, 1, 1);
  Image<float> objf = itsBuffer * obj;

  return centroid(objf);
}

// ######################################################################
Point2D VisualBuffer::findMostInterestingTargetLocMax(const Point2D& p)
{
  // let's start by getting an idea of the range of values in our
  // buffer after we blur it a bit:
  Image<float> buf = lowPass9(itsBuffer); float mi, ma; buf.getMinMax(mi, ma);
  float thresh = ma * 0.25F;

  // let's go over the image and find the local max that is above a
  // threshold and closest to our current fixation. The code here is
  // similar to what we have in maxNormalizeStd() to find local maxes,
  // except that here we want to enforce a true local max, not a ridge
  // point (hence we use strict inequalities):
  int w = buf.getWidth(), h = buf.getHeight();
  Point2D best(-1, -1); float bestdist(1.0e10);
  for (int j = 1; j < h - 1; j ++)
    for (int i = 1; i < w - 1; i ++)
      {
        int index = i + w * j;
        float val = buf.getVal(index);
        if (val >= thresh &&                      // strong enough activity
            val > buf.getVal(index - w) &&        // local max
            val > buf.getVal(index + w) &&        // local max
            val > buf.getVal(index - 1) &&        // local max
            val > buf.getVal(index + 1) &&        // local max
            p.distance(Point2D(i, j)) < bestdist) // closest to eye
          { best.i = i; best.j = j; bestdist = p.distance(best); }
      }
  return best;
}

// ######################################################################
Image<float> VisualBuffer::getBuffer(const bool normalize) const
{
  if (normalize == false) return itsBuffer;
  Image<float> ret(itsBuffer);
  if (itsNormalizeFactor.getVal() > 0.0F) ret *= itsNormalizeFactor.getVal();
  else ret.normalize(0.0F, 255.0F);
  return ret;
}

// ######################################################################
Point2D VisualBuffer::retinalToBuffer(const Point2D& p) const
{
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  SharedPtr<Retina> retina = itsBrain->getRET();
  if (retina.get() == NULL) LFATAL("I need a retina!");

  // first subtract our current eye position:
  Point2D pp = retina->retinalToRaw(p);

  // then downscale to maplevel (we do like in WTAwinner.H, but no 
  // random jitter):
  int smlevel = itsLevelSpec.getVal().mapLevel();
  pp.i = int(pp.i / double(1 << smlevel) + 0.49);
  pp.j = int(pp.j / double(1 << smlevel) + 0.49);

  // then re-center into the larger visual buffer:
  pp += Point2D((itsBuffer.getWidth() - itsSMdims.w()) / 2,
                (itsBuffer.getHeight() - itsSMdims.h()) / 2);

  return pp;
}

// ######################################################################
Point2D VisualBuffer::bufferToRetinal(const Point2D& p) const
{
  if (itsBrain == NULL) LFATAL("I need a Brain!");
  SharedPtr<Retina> retina = itsBrain->getRET();
  if (retina.get() == NULL) LFATAL("I need a retina!");

  // re-center out of the larger visual buffer:
  Point2D pp = p - Point2D((itsBuffer.getWidth() - itsSMdims.w()) / 2,
                           (itsBuffer.getHeight() - itsSMdims.h()) / 2);

  // scale back up to retinal scale (we do like in WTAwinner.H):
  int smlevel = itsLevelSpec.getVal().mapLevel();
  pp.i <<= smlevel; pp.j <<= smlevel;

  // then add our current eye position:
  pp = retina->rawToRetinal(pp);

  return pp;
}

// ######################################################################
bool VisualBuffer::isObjectBased() const
{ return itsObjectBased.getVal(); }

// ######################################################################
void VisualBuffer::internalDynamics()
{
  float mi, ma; itsBuffer.getMinMax(mi, ma);
  itsBuffer = maxNormalize(itsBuffer, 0.0F, 0.0F, itsNormType.getVal());
  itsBuffer.normalize(mi, ma);
  if (itsDecayFactor.getVal() != 1.0F) itsBuffer *= itsDecayFactor.getVal();
  itsLastInteractTime = itsTime;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
