/*!@file VisualBuffer.H Grab ausio samples from /dev/dsp */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualBuffer.H 3109 2004-02-11 02:34:10Z itti $
//

#ifndef VISUALBUFFER_H_DEFINED
#define VISUALBUFFER_H_DEFINED

#include "saliency.H"
#include "ModelComponent.H"
#include "Image.H"
#include "WTAwinner.H"
#include "LevelSpec.H"
#include "fancynorm.H"
class Brain;

//! An integrative visual buffer
/*! A world-centered Visual Buffer to accumulate interesting locations
  (each time a covert attention shift occurs), with internal
  competitive dynamics within the buffer. */
class VisualBuffer : public ModelComponent
{
public:
  //! Constructor
  VisualBuffer(ModelManager& mgr,
	       const std::string& descrName = "Visual Buffer",
	       const std::string& tagName = "VisualBuffer");

  //! Set a hook to our Brain
  void setBrain(Brain *brain);

  //! Destructor
  ~VisualBuffer();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! New input
  /*! Beware that the inputs here are matched to what Brain naturally
    provides, which is
    @param win the coordinates of an atteention shift at retinal resolution
    @param sm the saliency map at the maplevel resolution
    @param objmask a mask with (if initialized) the shape of the attended
    object at retinal resolution. */
  void input(const WTAwinner& win, const Image<float>& sm,
             const Image<byte>& objmask);

  //! Get saliency mask
  /*! This is mostly for display purposes. The mask is the one that
    was created at the last input() and has the same size/scale as the
    saliency map and is in retinal coordinates */
  Image<byte> getSaliencyMask() const;

  //! Get current buffer
  Image<float> getBuffer(const bool normalize = false) const;

  //! evolve our internal dynamics
  void evolve(const double t);

  //! inhibit a location, given in retinotopic/retinal-scale coords
  /*! A disk of radius FOAradius will be used for inhibition */
  void inhibit(const Point2D& loc);

  //! inhibit a location, using a mask of same dims/scale as the buffer
  /*! @param mask should have same dims as the buffer, and values
    between 0.0 (no inhibition) and 1.0 (full inhibition). */
  void inhibit(const Image<float>& mask);

  //! Decide on most interesting target location given current eye position
  /*! All coordinates are in world-centered/sm-scale. The algorithm
    used here is inspired from the area activation model; we threshold
    out regions lower than a fraction of the maximum activation, then
    aim for the centroid of the region that is closest to current eye
    position. */
  Point2D findMostInterestingTarget(const Point2D& p);

  //! Decide on most interesting target location given current eye position
  /*! All coordinates are in world-centered/sm-scale. The algorithm
    used here will aim for the local max above a threshold that is
    closest to current eye position. */
  Point2D findMostInterestingTargetLocMax(const Point2D& p);

  //! transform coord from retinotopic/retinal-scale to world-centered/sm-scale
  Point2D retinalToBuffer(const Point2D& p) const;

  //! transform coord from world-centered/sm-scale to retinotopic/retinal-scale
  Point2D bufferToRetinal(const Point2D& p) const;

  //! are we object-based?
  bool isObjectBased() const;

protected:
  ModelParam<bool> itsIgnoreBoring;   //!< ignore boring attention shifts
  ModelParam<bool> itsObjectBased;    //!< true if doing object-based
  ModelParam<Dims> itsBufferDims;     //!< dims of our internal buffer
  ModelParam<LevelSpec> itsLevelSpec; //!< levelspec determines buffer scale
  ModelParam<Dims> itsInputDims;      //!< input image dims
  ModelParam<int> itsFOAradius;       //!< radius of FOA
  ModelParam<double> itsTimePeriod;   //!< period at which to apply interaction
  ModelParam<float> itsNormalizeFactor;//!< normalize factor for getBuffer()
  ModelParam<float> itsDecayFactor;   //!< temporal decay factor
  ModelParam<MaxNormType> itsNormType;//!< maxnorm type for our internals

  //! get started (see ModelComponent.H)
  void start1();

private:
  // apply one iteration of our internal dynamics
  void internalDynamics();

  Image<float> itsBuffer;
  Dims itsSMdims;
  Image<byte> itsSaliencyMask;
  double itsTime;
  double itsLastInteractTime;
  Brain *itsBrain;
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
