/*!@file VisualCortex.C Implementation for visual cortex class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortex.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "VisualCortex.H"

#include "Channels.H"
#include "fancynorm.H" // for maxNormalize(), MaxNormType
#include "Image.H"
#include "Image_ColorOps.H"   // for luminance()
#include "Image_MathOps.H"    // for distance()
#include "Image_ShapeOps.H"
#include "Image_Transforms.H" // for chamfer34()
#include "log.H"
#include "ParamMap.H"
#include "Jet.H"
#include "ModelManager.H"
#include "FrameSeries.H"
#include "ModelOptionDefs.H"
#include <algorithm>
#include <cstdio>
#include <cctype>
#include <vector>

// ######################################################################
VisualCortex::VisualCortex(ModelManager& mgr, const std::string& descrName,
                           const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsNormType(this, "MaxNormType", VCXNORM_FANCY), // see ModelOptionDefs.C
  itsUseRandom(this, "UseRandom", true),           // see ModelOptionDefs.C
  itsBiasFeatures(this, "BiasFeatures", ""), // see ModelOptionDefs.C
  itsSaveOutput(this, "VisualCortexSaveOutput", false),// see ModelOptionDefs.C
  itsOutputFactor(this, "VisualCortexOutputFactor", 1.0e-9f), // idem
  itsUseOlderVersion(this, "UseOlderVersion", true), // see ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()),
  itsSurpriseThresh(this, "VCXSurpriseThresh", 20.0f),
  itsSurpriseSemisat(this, "VCXSurpriseSemisat", 25.0f),
  itsSurpriseExp(this, "VCXSurpriseExp", 4.0f)
{  }

// ######################################################################
VisualCortex::VisualCortex(ModelManager& mgr, const std::string& descrName,
                           const std::string& tagName,
                           const double chanIw, const double chanCw,
                           const double chanOw, const double chanFw,
                           const double chanMw) :
  ModelComponent(mgr, descrName, tagName),
  itsNormType(this, "MaxNormType", VCXNORM_FANCY), // see ModelOptionDefs.C
  itsUseRandom(this, "UseRandom", true),           // see ModelOptionDefs.C
  itsBiasFeatures(this, "BiasFeatures", ""), // see ModelOptionDefs.C
  itsSaveOutput(this, "VisualCortexSaveOutput", false),// see ModelOptionDefs.C
  itsOutputFactor(this, "VisualCortexOutputFactor", 1.0e-9f), // idem
  itsUseOlderVersion(this, "UseOlderVersion", true), // see ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()),
  itsSurpriseThresh(this, "VCXSurpriseThresh", 20.0f),
  itsSurpriseSemisat(this, "VCXSurpriseSemisat", 25.0f),
  itsSurpriseExp(this, "VCXSurpriseExp", 4.0f)
{
  addDefaultChannels(chanIw, chanCw, 0.0, chanOw, chanFw, chanMw,
                     0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
}

// ######################################################################
VisualCortex::VisualCortex(ModelManager& mgr, const std::string& descrName,
                           const std::string& tagName,
                           const double chanIw, const double chanCw,
                           const double chanCow,
                           const double chanOw, const double chanFw,
                           const double chanMw, const double chanLw,
                           const double chanTw, const double chanXw,
                           const double chanEw, const double chanGw,
                           const double chanDw, const double chanSHw) :
  ModelComponent(mgr, descrName, tagName),
  itsNormType(this, "MaxNormType", VCXNORM_FANCY), // see ModelOptionDefs.C
  itsUseRandom(this, "UseRandom", true),           // see ModelOptionDefs.C
  itsBiasFeatures(this, "BiasFeatures", ""), // see ModelOptionDefs.C
  itsSaveOutput(this, "VisualCortexSaveOutput", false),// see ModelOptionDefs.C
  itsOutputFactor(this, "VisualCortexOutputFactor", 1.0e-9f), // idem
  itsUseOlderVersion(this, "UseOlderVersion", true), // see ModelOptionDefs.C
  itsLevelSpec(this, "LevelSpec", LevelSpec()),
  itsSurpriseThresh(this, "VCXSurpriseThresh", 20.0f),
  itsSurpriseSemisat(this, "VCXSurpriseSemisat", 25.0f),
  itsSurpriseExp(this, "VCXSurpriseExp", 4.0f)
{
  addDefaultChannels(chanIw, chanCw,  chanCow, chanOw, chanFw, chanMw,
                     chanLw, chanTw, chanXw, chanEw, chanGw, chanDw, chanSHw);
}

// ######################################################################
VisualCortex::~VisualCortex()
{  }

// ######################################################################
void VisualCortex::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsNormType);
    manager->requestOption(this, itsUseRandom);
    manager->requestOption(this, itsBiasFeatures);
    manager->requestOption(this, itsUseOlderVersion);
    manager->requestOption(this, itsOutputFactor);
    manager->requestOption(this, itsLevelSpec);
    manager->requestOption(this, itsSurpriseThresh);
    manager->requestOption(this, itsSurpriseSemisat);
    manager->requestOption(this, itsSurpriseExp);
  }
  if (whatToExport & OPTEXP_SAVE)
    manager->requestOption(this, itsSaveOutput);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void VisualCortex::reset(const bool recurse)
{
  // reset some stuff for VisualCortex
  itsOutput.freeMem();

  // call reset in ModelComponent to propagate to other components
  ModelComponent::reset(recurse);
}

// ######################################################################
void VisualCortex::addDefaultChannels(const double chanIw,
                                      const double chanCw,
                                      const double chanCow,
                                      const double chanOw,
                                      const double chanFw,
                                      const double chanMw,
                                      const double chanLw,
                                      const double chanTw,
                                      const double chanXw,
                                      const double chanEw,
                                      const double chanGw,
                                      const double chanDw,
                                      const double chanSHw)
{
  if (chanCw > 0.0)
    {
      addChannel("color", makeShared(new ColorChannel(*manager)));
      channel("color").setTotalWeight(chanCw);
    }
  if (chanCow > 0.0)
    {
      addChannel("composite-color", makeShared
                 (new CompositeColorChannel(*manager)));
      channel("composite-color").setTotalWeight(chanCow);
    }
   if (chanFw > 0.0)
    {
      addChannel("flicker", makeShared(new FlickerChannel(*manager)));
      channel("flicker").setTotalWeight(chanFw);
    }
  if (chanIw > 0.0)
    {
      addChannel("intensity", makeShared(new IntensityChannel(*manager)));
      channel("intensity").setTotalWeight(chanIw);
    }
  if (chanOw > 0.0)
    {
      addChannel("orientation", makeShared(new OrientationChannel(*manager)));
      channel("orientation").setTotalWeight(chanOw);
    }
  if (chanMw > 0.0)
    {
      addChannel("motion", makeShared(new MotionChannel(*manager)));
      channel("motion").setTotalWeight(chanMw);
    }
  if (chanLw > 0.0)
    {
      addChannel("LJunction",makeShared(new LJunctionChannel(*manager,this)));
      channel("LJunction").setTotalWeight(chanLw);
    }
  if (chanTw > 0.0)
    {
      addChannel("TJunction",makeShared(new TJunctionChannel(*manager,this)));
      channel("TJunction").setTotalWeight(chanTw);
    }
  if (chanXw > 0.0)
    {
      addChannel("XJunction",makeShared(new XJunctionChannel(*manager,this)));
      channel("XJunction").setTotalWeight(chanXw);
    }
  if (chanEw > 0.0)
    {
      addChannel("EndPoint", makeShared(new EndPointChannel(*manager,this)));
      channel("EndPoint").setTotalWeight(chanEw);
    }
  if (chanGw > 0.0)
    {
      addChannel("MultiColorBand", makeShared(new MultiColorBandChannel
                                              (*manager)));
      channel("MultiColorBand").setTotalWeight(chanGw);
    }
  if (chanDw > 0.0)
    {
      addChannel("Dummy", makeShared(new DummyChannel(*manager)));
      channel("Dummy").setTotalWeight(chanDw);
    }
  if (chanSHw > 0.0)
    {
      addChannel("SkinHue", makeShared(new SkinHueChannel(*manager)));
      channel("SkinHue").setTotalWeight(chanSHw);
    }
}

// ######################################################################
uint VisualCortex::addChannel(const SharedPtr<ChannelBase>& ch)
{
  // let's take ownership of this channel and return its index:
  return addSubComponent(ch);
}

// ######################################################################
uint VisualCortex::addChannel(const char* name,
                              const SharedPtr<ChannelBase>& ch)
{
  // let's rename that channel to the name that given here:
  ch->setTagName(name);

  // let's take ownership of this channel and return its index:
  return addSubComponent(ch);
}

// ######################################################################
void VisualCortex::input(const Image< PixRGB<byte> >& inp,
                         const double t,
                         const Image<byte>& clipMask)
{
  // keep track of the input dims as we may need them later:
  itsInputDims = inp.getDims();

  // convert input to float, as the channels want:
  Image< PixRGB<float> > input = inp;

  // bias the feature channels to promote the target
  if (strcmp(itsBiasFeatures.getValString().c_str(), "") != 0)
    {
      LINFO("biasing the visual cortex with %s",
            itsBiasFeatures.getValString().c_str());
      readParamMap(itsBiasFeatures.getValString().c_str());
    }
  // get luminance of input image for non-color channels (motion, ori):
  Image<float> bwima = luminance(input);
  LINFO("Luminance ok.");

  // make a list of all V1, V2 and V4 channels
  std::vector<uint> v1, v2, v4;
  for (uint i = 0; i < numChannels(); ++i)
    {
      if (featureHierarchyLevel(channel(i).visualFeature()) == 1)
        v1.push_back (i);
      else if (featureHierarchyLevel(channel(i).visualFeature()) == 2)
        v2.push_back (i);
      else if (featureHierarchyLevel(channel(i).visualFeature()) == 4)
        v4.push_back (i);
    }
  // process V1 channels
  if (v1.size())
    {
      for (uint i = 0; i < v1.size(); i ++)
        {
          channel(v1[i]).input(input, bwima, t, clipMask);
          LINFO("Input to %s channel ok.", channel(v1[i]).tagName().c_str());
        }
      LINFO ("Input to V1 channels ok.");
    }
  // process V2 channels
  if (v2.size())
    {
      for (uint i = 0; i < v2.size(); i ++)
        {
          LINFO("entering %s channel", channel(v2[i]).tagName().c_str());
          channel(v2[i]).input(input, bwima, t, clipMask);
          LINFO("Input to %s channel ok.", channel(v2[i]).tagName().c_str());
        }
      LINFO ("Input to V2 channels ok.");
    }

    // process V4 channels
  if (v4.size())
    {
      for (uint i = 0; i < v4.size(); i ++)
        {
          LINFO("entering %s channel", channel(v4[i]).tagName().c_str());
          channel(v4[i]).input(input, bwima, t, clipMask);
          LINFO("Input to %s channel ok.", channel(v4[i]).tagName().c_str());
        }
      LINFO ("Input to V4 channels ok.");
    }

  // invalidate any cached results, which will force recomputation next
  // time we want to need to get the output
  itsOutput.freeMem();
}

// ######################################################################
void VisualCortex::drawResults(Image< PixRGB<byte> >& traj)
{ for (uint i = 0; i < numChannels(); ++i) channel(i).drawResults(traj); }

// ######################################################################
void VisualCortex::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // save our own output:
  if (itsSaveOutput.getVal())
    ofs->writeFloat(getOutput(), PNM, -1, true, "VCO");

  // save our channel outputs:
  for (uint i = 0; i < numChannels(); ++i) channel(i).saveResults(ofs);
}

// ######################################################################
void VisualCortex::learn(const Image<byte>& targets, const double& eta,
                         const bool doDistMap, const bool softmask,
                         const int inthresh, const int outthresh)
{
  // create a chamfer distance map from the target image -> target
  // weighting; result has zeros inside the targets and 255 outside,
  // graded values in between:
  Image<byte> dmap;
  if (doDistMap) dmap = chamfer34(targets);
  else dmap = binaryReverse(targets, byte(255));

  for (uint i = 0; i < numChannels(); ++i)
    channel(i).learn(dmap, eta, softmask, inthresh, outthresh);

  // I leave to the user the opportunity to normalize the coeffs or not
  // after each learning. Some normalization should be done at some point
  // to prevent coeff blowout.
}

// ######################################################################
uint VisualCortex::numChannels() const
{ return numSubComp(); }

// ######################################################################
ChannelBase& VisualCortex::channel(const uint idx) const
{ return dynamic_cast<ChannelBase&>(*channelPtr(idx)); }

// ######################################################################
SharedPtr<ChannelBase> VisualCortex::channelPtr(const char *tagname) const
{
  SharedPtr<ChannelBase> ptr;
  ptr.dynCastFrom(subComponent(tagname));
  if (ptr.get() == NULL)
    LFATAL("Ooops, channel '%s' does not derive from ChannelBase!", tagname);
  return ptr;
}

// ######################################################################
SharedPtr<ChannelBase> VisualCortex::channelPtr(const uint idx) const
{
  SharedPtr<ChannelBase> ptr;
  ptr.dynCastFrom(subComponent(idx));
  if (ptr.get() == NULL)
    LFATAL("Ooops, channel '%s' [%d] does not derive from ChannelBase!",
           subComponent(idx)->tagName().c_str(), idx);
  return ptr;
}

// ######################################################################
ChannelBase& VisualCortex::channel(const char* name) const
{ return dynamic_cast<ChannelBase&>(*channelPtr(name)); }

// ######################################################################
bool VisualCortex::hasChannel(const char *name) const
{ return hasSubComponent(name); }

// ######################################################################
Image<float> VisualCortex::getOutput()
{
  // did we already compute the output ??
  if (itsOutput.initialized()) return itsOutput;

  // ... OK, we have to recompute the output:
  for (uint i = 0; i < numChannels(); ++i)
    {
      if (channel(i).isZeroWeight()) continue;
      if (channel(i).outputAvailable() == false) continue;

      Image<float> chanOut = channel(i).getOutput();

      if (itsUseOlderVersion.getVal())
        {
          const float w = float(channel(i).weightPerChan());
          if (w != 1.0f) chanOut *= w;
          LDEBUG("%s weight %f", channel(i).tagName().c_str(), w);
        }
      else
        {
          const float w = float(channel(i).getTotalWeight());
          if (w != 1.0f) chanOut *= w;
          LDEBUG("%s weight %f", channel(i).tagName().c_str(), w);
        }

      // downSizeClean() gracefully does nothing if chanOut is already
      // the right size. (Eventually we might like to have a way for
      // VisualCortex to enforce a particular chanOut size on its own,
      // rather than just taking the size from the first channel in
      // the array.)
      if (itsOutput.initialized() == false)
        itsOutput = chanOut;  // first channel that has an output
      else
        itsOutput += downSizeClean(chanOut, itsOutput.getDims());
    }

  // Ensure that we still return a valid image even if we have no channels
  // that has a valid output in the VisualCortex:
  if (itsOutput.initialized() == false)
    {
      int sml = itsLevelSpec.getVal().mapLevel();
      itsOutput = Image<float>(itsInputDims.w() >> sml,
                               itsInputDims.h() >> sml, ZEROS);
      return itsOutput;
    }

  if (manager->debugMode())
    {
      float mi, ma; itsOutput.getMinMax(mi, ma);
      LINFO("Raw output range is [%f .. %f]", mi, ma);
    }

  // let's apply a last maxNormalization to the map:
  switch(itsNormType.getVal())
    {
    case VCXNORM_LANDMARK:
      itsOutput = maxNormalize(itsOutput, 0.0f, 2.0f, VCXNORM_NONE);
      break;
    case VCXNORM_SURPRISE:
      // get rid of background firing activity:
      itsOutput -= itsSurpriseThresh.getVal();
      itsOutput.rectify();

      // Pass the final surprise through a sigmoid:
      itsOutput.sigmoid(itsSurpriseExp.getVal(), itsSurpriseExp.getVal(),
                        powf(itsSurpriseSemisat.getVal(),
                             itsSurpriseExp.getVal()));
      break;
    default:
      itsOutput = maxNormalize(itsOutput, 0.0f, 2.0f, itsNormType.getVal());
      LDEBUG("%s(%f .. %f)", maxNormTypeName(itsNormType.getVal()),
             0.0f, 2.0f);
      break;
    }

  // output is now typically in the (0.0..8.0) range;
  // typical images are in (0..4) range; we want input current in nA
  itsOutput *= itsOutputFactor.getVal();
  float mi, ma; itsOutput.getMinMax(mi, ma);
  LINFO("Salmap input range is [%f .. %f] nA", mi * 1.0e9F, ma * 1.0e9F);

  // add a tiny background activity to ensure that the entire image
  // will be visited:
  if (itsUseRandom.getVal())
    {
      LINFO("Adding random background noise to output.");
      itsOutput.addBGnoise(1.0e-12F);
      itsOutput += 1.0e-14F;
    }

  LINFO("Computed VisualCortex output.");

  return itsOutput;
}

// ######################################################################
SharedPtr<JetSpec> VisualCortex::getJetSpec() const
{
  SharedPtr<JetSpec> js(new JetSpec());
  for (uint i = 0; i < numChannels(); ++i) channel(i).fillJetSpec(*js);
  return js;
}

// ######################################################################
void VisualCortex::fillJet(const Point2D& loc, Jet<float>& j) const
{
  std::vector<int> index;
  for (uint i = 0; i < numChannels(); ++i) channel(i).fillJet(loc, j, index);
}

// ######################################################################
void VisualCortex::fillJetNI(const Point2D& loc, Jet<float>& j) const
{
  std::vector<int> index;
  for (uint i = 0; i < numChannels(); ++i) channel(i).fillJetNI(loc, j, index);
}

// ######################################################################
void VisualCortex::clampParams(const double& cmin, const double& cmax)
{
  for (uint i = 0; i < numChannels(); ++i)
    {
      channel(i).clampWeight(cmin, cmax);
      channel(i).clampCoeffs(cmin, cmax);
    }
}

// ######################################################################
void VisualCortex::normalizeParams()
{
  double sum = 0.0;
  uint nbmaps = 0;

  for (uint i = 0; i < numChannels(); ++i)
    {
      sum += channel(i).getTotalWeight() * channel(i).absSumCoeffs();
      nbmaps += channel(i).numSubmaps();
    }

  if (sum < 0.1)
    {
      LERROR("Sum of coeffs very small (%f). Not normalized.", sum);
      return;
    }

  LINFO("Coeff normalization: old sum = %f, nbmaps = %d", sum, nbmaps);

  for (uint i = 0; i < numChannels(); ++i)
    channel(i).normalizeCoeffs(sum / float(nbmaps));
}

// ######################################################################
void VisualCortex::cleanParams(const double& epsilon)
{
  for (uint i = 0; i < numChannels(); ++i)
    {
      channel(i).cleanWeight(epsilon);
      channel(i).cleanCoeffs(epsilon);
    }
}

// ######################################################################
bool VisualCortex::readParamMap(const char* fname)
{
  try
    {
      ParamMap toplev;
      toplev.load(fname); // fill the ParamMap with the file's contents
      readFrom(toplev); // read our parameters from the ParamMap
    }
  catch (ParamMap::Error&)
    {
      return true;
    }
  return false;
}

// ######################################################################
void VisualCortex::writeParamMap(const char* fname) const
{
  try
    {
      ParamMap toplev;
      writeTo(toplev); // fill the ParamMap with our parameters
      toplev.format(fname); // write the parameters to the file
    }
  catch (ParamMap::Error&)
    {
      // Do nothing; just ignore parsing errors for now.
    }
}

// ######################################################################
void VisualCortex::readFrom(const ParamMap& pmap)
{
  for (uint i = 0; i < numChannels(); ++i)
    {
      const char* const pname = channel(i).tagName().c_str();
      if (pmap.hasParam(pname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(pname);
          channel(i).readFrom(*submap);
        }
    }
}

// ######################################################################
void VisualCortex::writeTo(ParamMap& pmap) const
{
  for (uint i = 0; i < numChannels(); ++i)
    {
      SharedPtr<ParamMap> submap(new ParamMap);
      channel(i).writeTo(*submap);
      pmap.putSubmap(channel(i).tagName().c_str(), submap);
    }
}

// ######################################################################
const Dims& VisualCortex::getInputDims() const
{ return itsInputDims; }

// ######################################################################
Image<float> VisualCortex::getJetDistMap(const Point2D& center,
                                         const Jet<float>& weights)
{
  Image<float> result(itsInputDims, NO_INIT);
  int w = itsInputDims.w(), h = itsInputDims.h();
  Jet<float> cenJet(weights.getSpec());
  Jet<float> currJet(weights.getSpec());

  // get the jet values for the specified location:
  fillJet(center, cenJet);

  // now loop over the entire image and compute a weighted Jet
  // distance between that central Jet and the Jets extracted from all
  // the other locations in the image:
  for (int x = 0; x < w; x++)
    for (int y = 0; y < h; y++)
      {
        fillJet(Point2D(x,y),currJet);
        result.setVal(x, y, float(distance(cenJet, currJet, weights)));
      }
  return result;
}

// ######################################################################
Image<float> VisualCortex::getRaoJetMap(const Jet<float>& targ,
                                        const int smin, const int smax)
{
  int w = itsInputDims.w() >> smin, h = itsInputDims.h() >> smin;
  Image<float> result(w, h, NO_INIT);
  Jet<float> currJet(targ.getSpec());

  // now loop over the entire image and compute a weighted Jet
  // distance between that central Jet and the Jets extracted from all
  // the other locations in the image:
  for (int x = 0; x < w; x++)
    for (int y = 0; y < h; y++)
      {
        fillJetNI(Point2D(x << smin, y << smin), currJet);
        result.setVal(x, y, float(raodistance(targ, currJet, smin, smax)));
      }
  return result;
}

// ######################################################################
void VisualCortex::getFeatures(const Point2D& locn,
                               std::vector<double>& mean) const
{
  for (uint i = 0; i < numChannels(); i++)
  {
    channel(i).getFeatures(locn, mean);
  }
}

// ######################################################################
void VisualCortex::getFeatures(const Point2D& locn,
                               std::vector<double>& mean,
                               int c_start, int c_end) const
{
  ASSERT(c_start > -1);
  LINFO("%d,%d",c_end,numChannels());
  ASSERT(c_end <=  int(numChannels()));
  for (int i = c_start; i < c_end; i++)
  {
    channel(i).getFeatures(locn, mean);
  }
}

// ######################################################################
void VisualCortex::getFeaturesBatch(std::vector<Point2D*> *locn,
                                    std::vector<std::vector<float> > *mean,
                                    int *count) const
{
  for (uint i = 0; i < numChannels(); i++)
  {
    channel(i).getFeaturesBatch(locn, mean, count);
  }
}

// ######################################################################
void VisualCortex::getCoeff(std::vector<double>& coeff)
{
  for (uint i = 0; i < numChannels(); i++)
    channel(i).getCoeff(coeff);
}

// ######################################################################
void VisualCortex::setGroupingCoeff(std::vector<double>& coeff)
{
  uint idx = 0; // index to the coeff array
  for (uint i = 0; i < numChannels(); i++)
    channel(i).setGroupingCoeff(coeff, idx);
}

// ######################################################################
Image<float> VisualCortex::perceptualGrouping()
{
  // ... OK, we have to recompute the percept:
  Image<float> myPercept;
  for (uint i = 0; i < numChannels(); ++i)
    {
      if (channel(i).isZeroWeight()) continue;
      if (channel(i).outputAvailable() == false) continue;
      Image<float> chanOut = channel(i).perceptualGrouping();
      const float w = float(channel(i).weightPerChan());
      if (w != 1.0f) chanOut *= w;
      LINFO("%s grouping weight %f", channel(i).tagName().c_str(), w);
      if (myPercept.initialized() == false)
        myPercept = chanOut;  // first channel that has an output
      else
        myPercept += downSizeClean(chanOut, myPercept.getDims());
    }
  
  // Ensure that we still return a valid image even if we have no channels
  // that has a valid percept in the VisualCortex:
  if (myPercept.initialized() == false)
    {
      int sml = itsLevelSpec.getVal().mapLevel();
      myPercept = Image<float>(itsInputDims.w() >> sml,
                               itsInputDims.h() >> sml, ZEROS);
    }
  return myPercept;
}  

// ######################################################################
void VisualCortex::biasFeatures(const double* mean, const double* sigma)
{
  int idx = 0;
  for (uint i = 0; i < numChannels(); i++)
    channel(i).biasFeatures(mean, sigma, idx);

  //normalizeParams();
  clampParams(0.0, 255.0);

  // set weights of the channels
  findWeight();
}

// ######################################################################
void VisualCortex::findWeight()
{
  double sum = 0.0;
  for (uint i = 0; i < numChannels(); i++)
    {
      // find the weight of the channel
      channel(i).findWeight();
      // find the sum of the channel's weight (which is currently the
      // maximum submap response) over its theoretical maximum
      sum += (channel(i).getTotalWeight()/255.0);
    }

  // normalize channel weights so that they add to 1
  for (uint i = 0; i < numChannels(); i++)
    {
      double wt = channel(i).getTotalWeight()/255.0;
      if (sum != 0.0) channel(i).setTotalWeight(wt/sum);
    }

}

// ######################################################################
bool VisualCortex::reco(const ParamMap& pmap, double& mismatch)
{
  int numSubmaps = 0;
  for (uint i = 0; i < numChannels(); i++)
    {
      const char* const pname = channel(i).tagName().c_str();
      if (pmap.hasParam(pname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(pname);
          channel(i).reco(*submap, mismatch);
        }
      numSubmaps += channel(i).numSubmaps();
    }

  mismatch = sqrt(mismatch);
  return true;
}
// ######################################################################
void VisualCortex::findLikelihood(const ParamMap& pmap,
                                  double& LogLikelihood)
{
  LogLikelihood = 0.0; // start with a good assumption of match:)
  for (uint i = 0; i < numChannels(); i++)
    {
      const char* const pname = channel(i).tagName().c_str();

      if (pmap.hasParam(pname))
        {
          SharedPtr<ParamMap> submap = pmap.getSubmap(pname);
          channel(i).findLikelihood(*submap, LogLikelihood );
        }
    }
  return;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
