/*!@file VisualCortex.H The early visual processing stages */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortex.H 4228 2005-02-23 17:19:25Z vidhya $
//

#ifndef VISUALCORTEX_H_DEFINED
#define VISUALCORTEX_H_DEFINED

#include "saliency.H"
#include "Image.H"
#include <string>
#include <utility>
#include <vector>
#include "ModelComponent.H"
#include "fancynorm.H" // for MaxNormType
#include "LevelSpec.H"
#include "ScaleSpace.H"

// Forward declarations will suffice instead of #include's here, and will
// be more efficient in compile times and in minimizing dependencies.
class ChannelBase;
class ChannelListener;
class Dims;
class ParamMap;
class Point2D;
class Rectangle;
class JetSpec;
class OutputFrameSeries;

template <class T> class Jet;
template <class T> class SharedPtr;
template <class T> class PixRGB;


// ######################################################################
//! The Visual Cortex Class
/*! In brief, VisualCortex holds a collection of ChannelBase objects, and
    most of VisualCortex's operations are achieved by some kind of
    iteration over that collection. That is, VisualCortex now does little
    work by itself, but delegates its operations to the channels,
    accumulating their results if necessary. */
class VisualCortex : public ModelComponent
{
public:
  // ######################################################################
  /*! @name Constructors/Destructor */
  //@{

  //! Construct with no channel; channels must then be added with addChannel().
  /*! @param mgr our ModelManager (see ModelManager.H)
      @param descrName descriptive name for human useag
      @param tagName name for ParamMap usage */
  VisualCortex(ModelManager& mgr,
               const std::string& descrName = "Visual Cortex",
               const std::string& tagName = "VisualCortex");
  
  //! Construct with default intens/color/orient/flicker/motion channels.
  /*! This just calls addDefaultChannels(). The total channel weights
    are given here for each channel. If any is <= 0.0, the channel
    will NOT be added. */
  VisualCortex(ModelManager& mgr, const std::string& descrName,
               const std::string& tagName, const double chanIw,
               const double chanCw, const double chanOw, const double chanFw,
               const double chanMw);

  //! Construct with default intens/color/orient/flicker/motion/
  //! junction channels.
  /*! This just calls addDefaultChannels(). The total channel weights
    are given here for each channel. If any is <= 0.0, the channel
    will NOT be added. */
  VisualCortex(ModelManager& mgr, const std::string& descrName,
               const std::string& tagName, const double chanIw,
               const double chanCw, const double chanCow, 
               const double chanOw, const double chanFw,
               const double chanMw, const double chanLw, const double chanTw,
               const double chanXw, const double chanEw, const double chanGw, 
               const double chanDw, const double chanSHw);

  //! Virtual destructor for safe inheritance.
  virtual ~VisualCortex();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset VisualCortex and propagate the reset signal to the sub-components
  /*! See the base function in ModelComponent.H for info. */
  virtual void reset(const bool recurse = true);

  //@}

  // ######################################################################
  /*! @name Channel management functions

      These functions offer an interface in which VisualCortex maintains a
      set of channels, each of which can be looked up either by an integer
      index or by an associated name. */
  //@{

  //! Adds a new channel to the VisualCortex.
  /*! The channel's tagName will be used for associative lookup. Returns
      the index at which the channel was inserted; this index could be
      passed back to channel() to retrieve the channel object. */
  virtual uint addChannel(const SharedPtr<ChannelBase>& ch);

  //! Adds a new named channel to the VisualCortex.
  /*! The channel's tagName will be set to \a name. Returns the index at
      which the channel was inserted; this index could be passed back to
      channel() to retrieve the channel object. */
  virtual uint addChannel(const char* name, const SharedPtr<ChannelBase>& ch);

  //! Adds the default channels to the VisualCortex object
  /*! intens/color/orient/flicker/motion/junctions/dummy channels.
    The total channel weights are given here for each channel.
    If any is <= 0.0, the channel will NOT be added. */
  virtual void addDefaultChannels(const double chanIw = 1.0,
                                  const double chanCw = 1.0,
                                  const double chanCow = 1.0,
                                  const double chanOw = 1.0,
                                  const double chanFw = 1.0,
                                  const double chanMw = 1.0,
                                  const double chanLw = 0.0,
                                  const double chanTw = 0.0,
                                  const double chanXw = 0.0,
                                  const double chanEw = 0.0,
                                  const double chanGw = 0.0,
                                  const double chanDw = 0.0,
                                  const double chanSHw = 0.0);

  //! Get the number of contained channels.
  virtual uint numChannels() const;

  //! Look up a channel by its index (\a idx must be less than numChannels()).
  virtual ChannelBase& channel(const uint idx) const;

  //! Look up a channel by its index (\a idx must be less than numChannels()).
  virtual SharedPtr<ChannelBase> channelPtr(const uint idx) const;

  //! Look up a channel by its tagname
  virtual SharedPtr<ChannelBase> channelPtr(const char *tagname) const;

  //! Look up a channel by its name.
  virtual ChannelBase& channel(const char* name) const;

  //! see whether we have a channel of given name
  virtual bool hasChannel(const char *name) const;
  //@}


  // ######################################################################
  /*! @name Processing functions */
  //@{

  //! Present new input (e.g., a new frame from a video stream):
  virtual void input(const Image< PixRGB<byte> >& inp, const double t,
                     const Image<byte>& clipMask = Image<byte>());

  //! Call drawResults() for each contained channel.
  virtual void drawResults(Image< PixRGB<byte> >& traj);

  //! Call saveResults() for each contained channel.
  /*! Our output will be saved with prefix "VCO<tagname>-" */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //! Learn feature map weights:
  /*! CAUTION: targets should correspond to last loaded input! */
  virtual void learn(const Image<byte>& targets, const double& eta,
                     const bool doDistMap = true,
                     const bool softmask = false,
                     const int inthresh = 20, const int outthresh = 255);

  //! get a Jet-distance map
  /*! @param center the location at which the reference Jet will be
    extracted, using the JetSpec that is underling in the weights
    @param weights the weights to be used when computing a weighted
    RMS distance between the Jet extractred at the center location and
    those extracted at everyother location in the image. This distance
    will be stored in the returned image for every pixel. */
  virtual Image<float> getJetDistMap(const Point2D& center,
                                     const Jet<float>& weights);

  //! Get a jet distance map a la Rao et al., Vis Res, 2002
  virtual Image<float> getRaoJetMap(const Jet<float>& targ,
                                    const int smin, const int smax);

  //@}


  // ######################################################################
  /*! @name Access functions */
  //@{

  //! Get the output from visual cortex.
  /*! This will be some kind of normalized sum of the outputs of each of
      the contained channels. */
  virtual Image<float> getOutput();

  //! get the JetSpec corresponding to our channels
  virtual SharedPtr<JetSpec> getJetSpec() const;

  //! Fill the Jet with the relevant components from each contained channel.
  virtual void fillJet(const Point2D& loc, Jet<float>& j) const;

  //! Fill the Jet with the relevant components from each contained channel.
  /*! this version does not use interpolation */
  virtual void fillJetNI(const Point2D& loc, Jet<float>& j) const;

  //@}


  // ######################################################################
  /*! @name Parameter management functions

      These functions generally just loop over the contained channels and
      forward the request to each of them. */
  //@{

  //! Parameter management function
  void clampParams(const double& cmin, const double& cmax);

  //! Normalize parameters to fixed sum
  /*! Good thing to do after learning to prevent blowup. */
  void normalizeParams();

  //! Remove pars <= epsilon for speedup of output computation:
  void cleanParams(const double& epsilon);

  //! Read params from a ParamMap.
  /*! This is used internally by readParamMap(). This is different
    from ModelComponent::readParamsFrom() as it concerns mostly feature
    map weights rather than the various ModelComponent parameters. */
  void readFrom(const ParamMap& pmap);

  //! Write params to a ParamMap.
  /*! This is used internally by writeParamMap(). This is different
    from ModelComponent::writeParamsTo() as it concerns mostly feature
    map weights rather than the various ModelComponent parameters. */
  void writeTo(ParamMap& pmap) const;

  //! Read params from disk file.
  bool readParamMap(const char* fname);

  //! Write params to disk file.
  void writeParamMap(const char* fname) const;

  //! Get the coefficients (gain factors) for perceptual feature grouping
  void getCoeff(std::vector<double>& coeff);

  //! Set the coefficients (gain factors) for perceptual feature grouping
  void setGroupingCoeff(std::vector<double>& coeff);

  //! Compute the percept by grouping features
  Image<float> perceptualGrouping();

  //! Bias features( mean, sigma) of all featuremaps  
  void biasFeatures(const double* mean, const double* sigma);

  //! Get the center-surround feature responses of all feature maps 
  void getFeatures(const Point2D& locn, std::vector<double>& mean) const;

  //! Get the center-surround feature responses of all feature maps
  /*! this will only return values for channles between c_start and
    c_end
  */
  void getFeatures(const Point2D& locn, std::vector<double>& mean,
                   int c_start, int c_end) const;

  //! get a batch of feature points, is more efficent then one at a time

  void getFeaturesBatch(std::vector<Point2D*> *locn,
                        std::vector<std::vector<float> > *mean,
                        int *count) const;

  //! Find the weights of all channels after biasing features
  void findWeight();

  //! Recognise the location to check if it is the same as obj
  bool reco(const ParamMap& pmap, double& mismatch);

  //! what is the probability that the fixation belongs to this object? 
  void findLikelihood(const ParamMap& pmap, double& likelihood);
  //@}

  //! Returns the dimensions of the most recent input image.
  const Dims& getInputDims() const;

protected:
  ModelParam<MaxNormType> itsNormType; //!< maxNormalization to use
  ModelParam<bool> itsUseRandom;       //!< add random noise to output
  ModelParam<std::string> itsBiasFeatures;//!< bias features with target params
  ModelParam<bool> itsSaveOutput;      //!< save our output?
  ModelParam<float> itsOutputFactor;   //!< output range factor

  /* older version: raw CS submaps -> [normalize, spatial competition,
     sum] -> SO -> [spatial competition, sum] -> CO -> [sum, spatial
     competition] -> VCX o/p : 

     Since different feature types may have different range of
     responses, their responses are normalized to a fixed range so as
     to treat them equally. Ideally, normalization should occur at the
     visualCortex when it combines the outputs of the different
     feature types. Instead, the older version performs the
     normalization within each feature type, forcing all scales within
     a feature type to be treated as equally important. This is
     undesirable as it artificially magnifies even those scales which
     contain only noise, and we lose information about the relative
     goodness of the different scales.

     new version: raw CS submaps -> [spatial competition, sum] -> SO
     -> [spatial competition, sum] -> CO -> [normalize, sum, spatial
     competition] -> VCX o/p
  */
  ModelParam<bool> itsUseOlderVersion;

  //! LevelSpec used by our channels, used to compute output dims
  ModelParam<LevelSpec> itsLevelSpec;

  ModelParam<float> itsSurpriseThresh; //!< surprise nonlinerity threshold
  ModelParam<float> itsSurpriseSemisat; //!< surprise nonlinerity semisat
  ModelParam<float> itsSurpriseExp; //!< surprise nonlinerity exponent

private:
  VisualCortex(const VisualCortex&); // not allowed
  VisualCortex& operator=(const VisualCortex&); // not allowed

  Dims itsInputDims;
  Image<float> itsOutput;
};


/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif
