/*!@file VisualCortexBeo.C a VisualCortex with SingleChannelBeo channels */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortexBeo.C 2493 2003-04-29 08:42:03Z itti $
//

#include "VisualCortexBeo.H"
#include "ModelManager.H"
#include "Beowulf.H"
#include "ChannelsBeo.H"

// ######################################################################
VisualCortexBeo::VisualCortexBeo(ModelManager& mgr,
                                 const std::string& descrName,
                                 const std::string& tagName,
                                 const SharedPtr<Beowulf>& beo) :
  VisualCortex(mgr, descrName, tagName),
  itsBeo(beo)
{
  // check that we are Beowulf master
  if (beo->getNodeNumber() != -1)
    LFATAL("I need to be Beowulf master -- ABORT");
}

// ######################################################################
VisualCortexBeo::VisualCortexBeo(ModelManager& mgr,
                                 const std::string& descrName,
                                 const std::string& tagName,
                                 const SharedPtr<Beowulf>& beo,
                                 const double chanIw, const double chanCw,
                                 const double chanOw, const double chanFw,
                                 const double chanMw) :
  VisualCortex(mgr, descrName, tagName),
  itsBeo(beo)
{
  // check that we are Beowulf master
  if (beo->getNodeNumber() != -1)
    LFATAL("I need to be Beowulf master -- ABORT");
  
  // create our default channels:
  addDefaultChannels(chanIw, chanCw, chanOw, chanFw, chanMw);
}

// ######################################################################
VisualCortexBeo::~VisualCortexBeo()
{  }

// ######################################################################
void VisualCortexBeo::addDefaultChannels(const double chanIw,
                                         const double chanCw,
                                         const double chanOw,
                                         const double chanFw,
                                         const double chanMw)
{
  // order matters here! To compile the results, we will sequentially
  // loop over all channels, and wait for each one to give us its
  // results.
  if (chanOw > 0.0)
    {
      addChannel("orientation",
                 makeShared(new OrientationChannelBeo(*manager, itsBeo)));
      channel("orientation").setTotalWeight(chanOw);
    }
  if (chanMw > 0.0)
    {
      addChannel("motion",
                 makeShared(new MotionChannelBeo(*manager, itsBeo)));
      channel("motion").setTotalWeight(chanMw);
    }
  if (chanCw > 0.0)
    {
      addChannel("color",
                 makeShared(new ColorChannelBeo(*manager, itsBeo)));
      channel("color").setTotalWeight(chanCw);
    }
  if (chanFw > 0.0)
    {
      addChannel("flicker",
                 makeShared(new FlickerChannelBeo(*manager, itsBeo)));
      channel("flicker").setTotalWeight(chanFw);
    }
  if (chanIw > 0.0)
    {
      addChannel("intensity",
                 makeShared(new IntensityChannelBeo(*manager, itsBeo)));
      channel("intensity").setTotalWeight(chanIw);
    }
}

// ######################################################################
SharedPtr<Beowulf> VisualCortexBeo::getBeowulf() const
{ return itsBeo; }

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
