/*!@file VisualCortexConfigurator.C Class to select a VisualCortex at
  runtime */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortexConfigurator.C 4228 2005-02-23 17:19:25Z vidhya $
//

#include "VisualCortexConfigurator.H"
#include "VisualCortex.H"
#include "VisualCortexBeo.H"
#include "VisualCortexSurprise.H"
#include "VisualCortexEyeMvt.H"
#include "Beowulf.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Channels.H"

#include <sstream>

// ######################################################################
VisualCortexConfigurator::
VisualCortexConfigurator(ModelManager& mgr,
                         const std::string& descrName,
                         const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsVCtype(this, "VisualCortexType", "None"),
  itsVC(NULL), itsBeo(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
VisualCortexConfigurator::~VisualCortexConfigurator()
{  }

// ######################################################################
void VisualCortexConfigurator::exportOptions(const int whatToExport,
                                             const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsVCtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<VisualCortex> VisualCortexConfigurator::getVC() const
{ return itsVC; }

// ######################################################################
SharedPtr<Beowulf> VisualCortexConfigurator::getBeo() const
{ return itsBeo; }

// ######################################################################
bool VisualCortexConfigurator::
setModelParamString(const std::string& name,
                    const std::string& value,
                    const bool recurse,
                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsVCtype.getName()) == 0) {
    // some info message:
    LINFO("Configuring VC of type %s", value.c_str());

    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsVC.get()) { 
      LINFO("Resetting VisualCortex...");
      removeSubComponent(itsVC); itsVC.reset(NULL);
    }
    if (itsBeo.get()) {
      LINFO("Resetting VisualCortex Beowulf...");
      removeSubComponent(itsBeo); itsBeo.reset(NULL);
    }

    // instantiate a VC of the appropriate type:

    // #################### No visual cortex:
    if (value.compare("None") == 0)
      itsVC.reset(NULL);

    // #################### Standard channels:
    else if (value.compare("Std") == 0)
      itsVC.reset(new VisualCortex(*manager, "Visual Cortex", "VisualCortex",
                                   1.0, 1.0, 1.0, 1.0, 1.0));

    // #################### Human eye movement fake cortex:
    else if (value.compare("EyeMvt") == 0)
      itsVC.reset(new VisualCortexEyeMvt(*manager));

    // #################### Entropy model:
    else if (value.compare("Entrop") == 0) // entropy
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<EntropyChannel> channel(new EntropyChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### Variance model:
    else if (value.compare("Variance") == 0)
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<VarianceChannel> channel(new VarianceChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### Tcorr model:
    else if (value.compare("Tcorr") == 0)
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<TcorrChannel> channel(new TcorrChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### Scorr model:
    else if (value.compare("Scorr") == 0)
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<ScorrChannel> channel(new ScorrChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### Information model:
    else if (value.compare("Info") == 0)
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<InformationChannel>
          channel(new InformationChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### PN03contrast model:
    else if (value.compare("PN03contrast") == 0) 
      {
        itsVC.reset(new VisualCortex(*manager));
        SharedPtr<PN03contrastChannel>
          channel(new PN03contrastChannel(*manager));
        itsVC->addChannel(channel);
        manager->setOptionValString("LevelSpec", "4-4,0-0,4");
      }

    // #################### Standard beo-channels:
    else if (value.compare("BeoStd") == 0)
      {
        itsBeo.reset(new Beowulf(*manager, "Visual Cortex Beowulf",
                                 "VisualCortexBeowulf", true));
        itsVC.reset(new VisualCortexBeo(*manager, "Visual Cortex",
                                        "VisualCortex", itsBeo,
                                        1.0, 1.0, 1.0, 1.0, 1.0));
      }

    // #################### Standard surprise-channels:
    else if (value.compare("SurpStd") == 0)
      itsVC.reset(new
                  VisualCortexSurprise(*manager, "Visual Cortex Surprise",
                                       "VisualCortexSurprise",
                                       1.0, 1.0, 1.0, 1.0, 1.0));

    // #################### Custom-defined:
    else {
      // ok, format here is a series of "X[:w.w]" where X is any of
      // "ICOFM" and the optional w.w may be channel weights. In
      // addition, a single "B" may be used to specify Beowulf mode,
      // and a single 'U' for Surprise mode:
      bool isbeo = false, issurp = false;
      uint i = 0; uint len = value.length();
      double iw = 0.0, cw = 0.0, cow = 0.0, ow = 0.0, fw = 0.0, mw = 0.0, 
        lw = 0.0, tw = 0.0, xw = 0.0, ew = 0.0, gw = 0.0, dw = 0.0, sw = 0.0;

      while (i < len) {
        char c = value[i];  // the channel to implement
        double weight = 1.0;  // default weight value
        if (c == 'B')
          { isbeo = true; i ++; }
        else if (c == 'U')
          { issurp = true; i ++; }
        else
          {
            // do we have a weight specification?
            if (i < len - 1 && value[i+1] == ':') {
              if (i >= len - 2)
                LFATAL("Missing channel weight value -- ABORT");
              uint end = value.find_first_not_of(".0123456789", i+2);
              std::stringstream s; s << value.substr(i+2, end); s >> weight;
              i = end;  // ready for next one
            } else i ++;
          }

        // assign the weight to the channel:
        switch(c) {
        case 'B':
          LINFO("Using Beowulf Channels");
          break;
        case 'U':
          LINFO("Using Surprise Channels");
          break;
        case 'I':
          LINFO("Using an IntensityChannel with weight %f", weight);
          iw = weight; break;
        case 'C':
          LINFO("Using a Double Opponent ColorChannel with weight %f", weight);
          cw = weight; break;
        case 'S':
          LINFO("Using a Composite ColorChannel with weight %f", weight);
          cow = weight; break;
        case 'O':
          LINFO("Using an OrientationChannel with weight %f", weight);
          ow = weight; break;
        case 'F':
          LINFO("Using a FlickerChannel with weight %f", weight);
          fw = weight; break;
        case 'M':
          LINFO("Using a MotionChannel with weight %f", weight);
          mw = weight; break;
        case 'L':
          LINFO("Using an L Junction Channel with weight %f", weight);
          lw = weight; break; 
        case 'T':
          LINFO("Using a T Junction Channel with weight %f", weight);
          tw = weight; break; 
        case 'X':
          LINFO("Using an X Junction Channel with weight %f", weight);
          xw = weight; break; 
        case 'E':
          LINFO("Using an EndPoint Channel with weight %f", weight);
          ew = weight; break; 
        case 'G':
          LINFO("Using Gaussian multi-color band channels with weight %f", 
                weight);
          gw = weight; break; 
        case 'D':
          LINFO("Using a Dummy Channel with weight %f", weight);
          dw = weight; break; 
        case 'K':
          LINFO("Using a SkinHue channel with weight %f",weight);
          sw = weight; break;
        default:
          LFATAL("Unknown channel type: '%c' -- ABORT", c);
        }
      }

      // all right, let's implement this cortex!
      if (isbeo == false)
        { 
          if (issurp == false)   // standard channels
            itsVC.reset(new VisualCortex(*manager, "Visual Cortex",
                                         "VisualCortex",
                                         iw, cw, cow, ow, fw, mw,
                                         lw, tw, xw, ew, gw, dw, sw));
          else
            itsVC.reset(new VisualCortexSurprise(*manager,
                                                 "Visual Cortex Surprise",
                                                 "VisualCortexSurprise",
                                                 iw, cw, cow, ow, fw, mw,
                                                 lw, tw, xw, ew, dw));
        }
      else // standard BeoChannels
        {
          if (issurp) LFATAL("BeoSurprise channels not supported -- ABORT");
          itsBeo.reset(new Beowulf(*manager, "Visual Cortex Beowulf",
                                   "VisualCortexBeowulf", true));
          itsVC.reset(new VisualCortexBeo(*manager, "Visual Cortex",
                                          "VisualCortex", itsBeo,
                                          iw, cw, ow, fw, mw));
        }
    }

    // add our babies as a subcomponents of us so that they will
    // become linked to the manager through us (hopefully we are
    // registered with the manager), which in turn will allow them to
    // export their command-line options and get configured. Note: the
    // Beowulf needs to be start()'ed by the time the VC starts, so we
    // add it first here if we have one:
    if (itsBeo.get()) {
      addSubComponent(itsBeo);
      // tell the controller to export its options:
      itsBeo->exportOptions(itsWhatToExport, true);
    }
    if (itsVC.get()) {
      addSubComponent(itsVC);
      // tell the controller to export its options:
      itsVC->exportOptions(itsWhatToExport, true);
    }
  }
  return ret;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
