/*!@file VisualCortexEyeMvt.C Implementation for the fake
  human-eye-movement visual cortex class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortexEyeMvt.C 4162 2005-02-03 19:16:39Z itti $
//

#include "VisualCortexEyeMvt.H"

#include "Channels.H"
#include "fancynorm.H" // for maxNormalize(), MaxNormType
#include "Image.H"
#include "Image_ColorOps.H"   // for luminance()
#include "Image_MathOps.H"    // for distance()
#include "Image_ShapeOps.H"
#include "Image_FilterOps.H"  // for gaussianBlob()
#include "Image_Transforms.H" // for chamfer34()
#include "log.H"
#include "ModelManager.H"
#include "ModelOptionDefs.H"
#include "Brain.H"
#include "FrameSeries.H"

#include <algorithm>
#include <cstdio>
#include <cctype>
#include <vector>

// ######################################################################
VisualCortexEyeMvt::VisualCortexEyeMvt(ModelManager& mgr,
                                       const std::string& descrName,
                                       const std::string& tagName) :
  VisualCortex(mgr, descrName, tagName),
  itsFnames(this, "VCEMeyeFnames", ""),
  itsSigma(this, "VCEMsigma", 4),
  itsForgetFac(this, "VCEMforgetFac", 0.999f),
  itsDelay(this, "VCEMdelay", 24),
  itsUseMax(this, "VCEMuseMax", false),
  itsSaccadeOnly(this, "VCEMsaccadeOnly", false),
  itsMaps(), itsSC(), itsSML(0)
{
  // instantiate one SC so that it gets configured:
  SharedPtr<EyeTrackerSaccadeController>
    sc(new EyeTrackerSaccadeController(mgr));
  sc->setDescriptiveName("SChuman00"); sc->setTagName("SChuman00");
  addSubComponent(sc);
  itsSC.push_back(sc);
}

// ######################################################################
VisualCortexEyeMvt::VisualCortexEyeMvt(ModelManager& mgr,
                                       const std::string& descrName,
                                       const std::string& tagName,
                                       const double chanIw,
                                       const double chanCw,
                                       const double chanOw,
                                       const double chanFw,
                                       const double chanMw) :
  VisualCortex(mgr, descrName, tagName), // ignore channels
  itsFnames(this, "VCEMeyeFnames", ""),
  itsSigma(this, "VCEMsigma", 4),
  itsForgetFac(this, "VCEMforgetFac", 0.999f),
  itsDelay(this, "VCEMdelay", 24),
  itsUseMax(this, "VCEMuseMax", false),
  itsSaccadeOnly(this, "VCEMsaccadeOnly", false),
  itsMaps(), itsSC(), itsSML(0)
{
  // instantiate one SC so that it gets configured:
  SharedPtr<EyeTrackerSaccadeController>
    sc(new EyeTrackerSaccadeController(mgr));
  sc->setDescriptiveName("SChuman00"); sc->setTagName("SChuman00");
  addSubComponent(sc);
  itsSC.push_back(sc);
}

// ######################################################################
VisualCortexEyeMvt::VisualCortexEyeMvt(ModelManager& mgr,
                                       const std::string& descrName,
                                       const std::string& tagName,
                                       const double chanIw,
                                       const double chanCw,
                                       const double chanCow,
                                       const double chanOw,
                                       const double chanFw,
                                       const double chanMw,
                                       const double chanLw,
                                       const double chanTw,
                                       const double chanXw, 
                                       const double chanEw,
                                       const double chanDw) :
  VisualCortex(mgr, descrName, tagName),  // ignore channels
  itsFnames(this, "VCEMeyeFnames", ""),
  itsSigma(this, "VCEMsigma", 4),
  itsForgetFac(this, "VCEMforgetFac", 0.999f),
  itsDelay(this, "VCEMdelay", 24),
  itsUseMax(this, "VCEMuseMax", false),
  itsSaccadeOnly(this, "VCEMsaccadeOnly", false),
  itsMaps(), itsSC(), itsSML(0)
{
  // instantiate one SC so that it gets configured:
  SharedPtr<EyeTrackerSaccadeController>
    sc(new EyeTrackerSaccadeController(mgr));
  sc->setDescriptiveName("SChuman00"); sc->setTagName("SChuman00");
  addSubComponent(sc);
  itsSC.push_back(sc);
}

// ######################################################################
VisualCortexEyeMvt::~VisualCortexEyeMvt()
{  }

// ######################################################################
void VisualCortexEyeMvt::start1()
{
  // let's parse our filename definition file and instantiate more SCs:
  std::string fn = itsFnames.getVal();
  uint idx = 0; int n = 0;

  // ok, we will need a brain to hook up to our SCs; let's fish for it:
  SharedPtr<Brain> brain; brain.dynCastFrom(manager->subComponent("Brain"));
  Dims dims; manager->getModelParamVal("InputFrameDims", dims, true);

  // let's build our collection of SCs:
  while(1)
    {
      // get one name from our comma-separated list:
      uint idx2 = fn.find(',', idx);
      std::string name; name.assign(fn, idx, idx2 - idx);

      // create a new SC:
      SharedPtr<EyeTrackerSaccadeController> sc;
      if (n == 0)
        sc.dynCastFrom(subComponent(0));  // we already have this one
      else
        {
          sc.reset(new EyeTrackerSaccadeController(*manager));
          char txt[100]; sprintf(txt, "SChuman%02d", n);
          sc->setTagName(txt); sc->setDescriptiveName(txt);

          // add this SC as a subcomponent, also to our direct-access list:
          addSubComponent(sc);
          itsSC.push_back(sc);
          sc->exportOptions(OPTEXP_ALL);
          sc->setModelParamVal("InputFrameDims", dims, true);
        }
      
      // adjust the SC's eye trash by our delay:
      int trash; sc->getModelParamVal("EyeTrackTrash", trash);
      trash += itsDelay.getVal();
      sc->setModelParamVal("EyeTrackTrash", trash);
      LINFO("Setting EyeTrackTrash on %s to %d samples",
            sc->tagName().c_str(), trash);

      // assign that file to that SC:
      LINFO("Assigned SC %d to %s", n, name.c_str());
      sc->setModelParamVal("EyeTrackFname", name);

      // hook SC to brain:
      sc->setBrain(brain.get());

      // ready for the next one:
      if (idx2 >= fn.size()-1) break; else idx = idx2 + 1;
      ++n;
    }

  // setup one map per SC:
  itsMaps.reset(itsSC.size());

  // get saliency map level:
  LevelSpec ls; getModelParamVal("LevelSpec", ls);
  itsSML = ls.mapLevel();
}

// ######################################################################
void VisualCortexEyeMvt::exportOptions(const int whatToExport,
                                       const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsFnames);
    manager->requestOption(this, itsSigma);
    manager->requestOption(this, itsForgetFac);
    manager->requestOption(this, itsDelay);
    manager->requestOption(this, itsUseMax);
    manager->requestOption(this, itsSaccadeOnly);
  }

  // the base class will propagate this call to our subcomponents:
  VisualCortex::exportOptions(whatToExport, recurse);
}

// ######################################################################
void VisualCortexEyeMvt::reset(const bool recurse)
{
  // call reset in ModelComponent to propagate to other components
  VisualCortex::reset(recurse);
}

// ######################################################################
void VisualCortexEyeMvt::input(const Image< PixRGB<byte> >& inp,
                               const double t,
                               const Image<byte>& clipMask)
{
  // compute map dims:
  int w = inp.getWidth() >> itsSML, h = inp.getHeight() >> itsSML;
  
  // start new maps fresh if first time:
  if (itsMaps[0].initialized() == false)
    for (uint i = 0; i < itsMaps.size(); i ++)
      itsMaps[i].resize(w, h, true);
}

// ######################################################################
void VisualCortexEyeMvt::evolve(const double t)
{
  float sig = itsSigma.getVal();

  for (uint idx = 0; idx < itsSC.size(); idx ++)
    {
      // evolve the SC:
      itsSC[idx]->evolve(t);

      // get current eye position and extra eye data if any:
      itsSC[idx]->getDecision(t);
      Point2D nextTarg; double ampl, durs;
      itsSC[idx]->getExtraEyeData(nextTarg, ampl, durs);
      Point2D p = (itsSC[idx]->getPreviousDecision(0)).p;
      
      // stop here if we are only interested in saccades but there is
      // not one now, otherwise use the saccade endpoint for our plots:
      if (itsSaccadeOnly.getVal())
        {
          if (ampl == 0.0) continue;
          p = nextTarg; // we are interested in saccade endpoint location
        }

      // convert eye coords to scale of saliency map:
      p.i = int(float(p.i) / float(1 << itsSML) + 0.4999f);
      p.j = int(float(p.j) / float(1 << itsSML) + 0.4999f);
      
      // do some forgetting:
      if (itsForgetFac.getVal() != 1.0f) itsMaps[idx] *= itsForgetFac.getVal();
      
      // inject new blob if valid coords:
      if (itsMaps[idx].coordsOk(p))
        {
          if (sig > 0.0f)
            {
              // draw a blob at current eye:
              Image<float> blob =
                gaussianBlob<float>(itsMaps[idx].getDims(), p, sig, sig);
              
              // take max between current and old blobs, with some forgetting:
              itsMaps[idx] = takeMax(itsMaps[idx], blob);
            }
          else
            // single spike mode:
            itsMaps[idx].setVal(p, itsMaps[idx].getVal(p) + 1.0f);
        }
    }
}

// ######################################################################
uint VisualCortexEyeMvt::addChannel(const SharedPtr<ChannelBase>& ch)
{
  LFATAL("Cannot add channels to VisualCortexEyeMvt -- ABORT");
  return 0; // keep compiler happy
}

// ######################################################################
uint VisualCortexEyeMvt::addChannel(const char* name,
                                    const SharedPtr<ChannelBase>& ch)
{
  LFATAL("Cannot add channels to VisualCortexEyeMvt -- ABORT");
  return 0; // keep compiler happy
}

// ######################################################################
uint VisualCortexEyeMvt::numChannels() const
{ return 0; }

// ######################################################################
bool VisualCortexEyeMvt::hasChannel(const char *name) const
{ return false; }
 
// ######################################################################
void VisualCortexEyeMvt::drawResults(Image< PixRGB<byte> >& traj)
{  }

// ######################################################################
void VisualCortexEyeMvt::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  // save our own output:
  if (itsSaveOutput.getVal())
    ofs->writeFloat(getOutput(), PNM, -1, true, "VCO");
}

// ######################################################################
Image<float> VisualCortexEyeMvt::getOutput()
{
  // compute output from all human maps:
  Image<float> out = itsMaps[0];
  if (itsUseMax.getVal())
    for (uint idx = 1; idx < itsMaps.size(); idx ++)
      out = takeMax(out, itsMaps[idx]);
  else
    for (uint idx = 1; idx < itsMaps.size(); idx ++)
      out += itsMaps[idx];

  // reset the maps:
  for (uint idx = 0; idx < itsMaps.size(); idx ++)
    itsMaps[idx].clear();

  // output is now typically in the (0.0..0.1) range; we want saliency
  // map input current in nA:
  out *= 50.0f * itsOutputFactor.getVal();
  float mi, ma; out.getMinMax(mi, ma);
  LINFO("Salmap input range is [%f .. %f] nA", mi * 1.0e9F, ma * 1.0e9F);
  LINFO("Computed VisualCortex output.");

  return out;
}

// ######################################################################
uint VisualCortexEyeMvt::numHumans() const
{ return itsSC.size(); }

// ######################################################################
Point2D VisualCortexEyeMvt::getEyePos(const uint human) const
{
  ASSERT(human < itsSC.size());
  return (itsSC[human]->getPreviousDecision(0)).p;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
