/*!@file VisualCortexEyeMvt.H a fake VisualCortex whose saliency map
  is built from human eye movement traces */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualCortexEyeMvt.H 4162 2005-02-03 19:16:39Z itti $
//

#ifndef VISUALCORTEXEYEMVT_H_DEFINED
#define VISUALCORTEXEYEMVT_H_DEFINED

#include "VisualCortex.H"
#include "ImageSet.H"
#include <vector>
#include "SaccadeControllers.H"

// ######################################################################
//! The VisualCortexEyeMvt Class
/*! This is a fake VisualCortex that outputs a saliency map built from
  a number of human eye movement traces. It contains a bunch of
  EyeTrackerSaccadeController objects that will provide eye position
  of the various humans, and builds some sort of heat map from
  those. Contrary to other VisualCortex variants, this requires an
  evolve(). Since Brain usually will not call evolve() on its
  VisualCortex, a special hook has been setup to handle that. */
class VisualCortexEyeMvt : public VisualCortex
{
public:
  //! Construct with no channel; channels are ignored anyways
  VisualCortexEyeMvt(ModelManager& mgr,
                     const std::string& descrName = "Visual Cortex Eye Mvt",
                     const std::string& tagName = "VisualCortexEyeMvt");
  
  //! Construct, ignoring channel definitions
  VisualCortexEyeMvt(ModelManager& mgr, const std::string& descrName,
                     const std::string& tagName, const double chanIw,
                     const double chanCw, const double chanOw,
                     const double chanFw, const double chanMw);

  //! Construct, ignoring even more channel definitions
  VisualCortexEyeMvt(ModelManager& mgr, const std::string& descrName,
                     const std::string& tagName, const double chanIw,
                     const double chanCw, const double chanCow, 
                     const double chanOw, const double chanFw,
                     const double chanMw, const double chanLw,
                     const double chanTw, const double chanXw,
                     const double chanEw, const double chanDw );

  //! Virtual destructor for safe inheritance.
  virtual ~VisualCortexEyeMvt();

  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Reset VisualCortex and propagate the reset signal to the sub-components
  virtual void reset(const bool recurse = true);

  //! Present new input (e.g., a new frame from a video stream):
  virtual void input(const Image< PixRGB<byte> >& inp, const double t,
             const Image<byte>& clipMask = Image<byte>());

  //! Cannot add channels to this VC -- will throw a fatal error if called
  virtual uint addChannel(const SharedPtr<ChannelBase>& ch);

  //! Cannot add channels to this VC -- will throw a fatal error if called
  virtual uint addChannel(const char* name, const SharedPtr<ChannelBase>& ch);

  //! Get the number of contained channels, always 0
  virtual uint numChannels() const;

  //! see whether we have a channel of given name
  virtual bool hasChannel(const char *name) const;

  //! No-op overload to VisualCortex::drawResults()
  virtual void drawResults(Image< PixRGB<byte> >& traj);

  //! Save our output
  /*! Our output will be saved with prefix "VCO<tagname>-" */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //! Get the output from visual cortex.
  virtual Image<float> getOutput();

  //! evolve...
  /*! Let the major hacking begin! VisualCortex usually does not
    evolve, but here we need it to. Brain will call this evolve from
    its own to get us going. */
  virtual void evolve(const double stime);

  //! Get our number of humans
  uint numHumans() const;

  //! Get current eye position of our of our humans
  Point2D getEyePos(const uint human) const;

protected:
  ModelParam<std::string> itsFnames; //!< names of eye movement files
  ModelParam<float> itsSigma; //!< sigma of gaussian blobs in pix at sml
  ModelParam<float> itsForgetFac; //!< forgetting factor
  ModelParam<int> itsDelay; //!< human-to-human delay, in eye samples
  ModelParam<bool> itsUseMax; //!< use max to combine instead of sum
  ModelParam<bool> itsSaccadeOnly; //!< plot only when a saccade

  // get started
  virtual void start1();

private:
  VisualCortexEyeMvt(const VisualCortexEyeMvt&); // not allowed
  VisualCortexEyeMvt& operator=(const VisualCortexEyeMvt&); // not allowed

  ImageSet<float> itsMaps;                      // one map per human
  std::vector< SharedPtr<EyeTrackerSaccadeController> > itsSC; // one per human
  int itsSML; // saliency map level
};

#endif

/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
