/*!@file VisualFeatures.H definitions of known visual features */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2002 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: VisualFeatures.H 3838 2004-10-12 01:26:45Z itti $
//

#ifndef VISUALFEATURES_H_DEFINED
#define VISUALFEATURES_H_DEFINED

//! Various features that can be represented
/*! We assign a specific number to each definition to speedup name printing */
enum VisualFeature
{
  UNKNOWN          = 0,            //!< Unknown feature type
  COMPOSITECOLOR   = 1,            //!< composite Color channel       
  COLOR            = 2,            //!< double opponent Color complex feature
  SOCOLOR          = 3,            //!< single opponent Color complex feature
  RG               = 4,            //!< Red/green double opponency
  BY               = 5,            //!< Blue/yellow double-opponency
  SORG             = 6,            //!< Red/green single opponency
  SOGR             = 7,            //!< Green/Red single opponency
  SOBY             = 8,            //!< Blue/Yellow single opponency
  SOYB             = 9,            //!< Yellow/Blue single opponency
  R                = 10,           //!< Red channel
  G                = 11,           //!< green channel
  B                = 12,           //!< Blue channel
  Y                = 13,           //!< yellow channel 
  INTENS           = 14,           //!< Intensity center-surround
  FLICKER          = 15,           //!< Flicker center-surround
  ORI              = 16,           //!< Orientation contrast
  TEMPLATE         = 17,           //!< Template matching
  COLBAND          = 18,           //!< Color band
  MOTION           = 19,           //!< Motion channel
  CONVOLVE         = 20,           //!< Convolution channel
  LJUNCTION        = 21,           //!< L junction channel
  TJUNCTION        = 22,           //!< T junction channel
  XJUNCTION        = 23,           //!< X junction channel
  ENDPOINT         = 24,           //!< End Point channel
  RGBCONVOLVE      = 25,           //!< Color convolution channel
  ENTROPY          = 26,           //!< Local entropy channel
  DUMMY            = 27,           //!< Dummy channel
  PN03CONTRAST     = 28,           //!< Local contrast (Parkhurst&Niebur 2003)
  DISPARITY        = 29,           //!< Disparity Channel 
  STEREO           = 30,           //!< Stereo channel
  INFORMATION      = 31,           //!< Information channel
  RECOGNITION      = 32,           //!< Recognition channel
  HUE              = 33,           //!< Hue channel
  DIRECTFEED       = 34,           //!< DirectFeed channel
  MULTIDIRECTFEED  = 35,           //!< MultiDirectFeed channel
  VARIANCE         = 36,           //!< Variance channel
  TCORR            = 37,           //!< Tcorr (temporal correlation) channel
  SCORR            = 38            //!< Scorr (spatial correlation) channel
  // if you add any to this list, be sure to update the following below:
  // NBVISUALFEATURES [add 1]
  // featureName() [add a new name for your feature at end of list]
  // featureHierarchyLevel [add a new entry for your feature]
};
//! Number of visual features
#define NBVISUALFEATURES 39

//! Various types of feature representations
/*! We assign a specific number to each definition to speedup name printing */
enum VisualFeatureType
{
  RAW      = 0, //!< Raw pixel data
  RAW_CS   = 1, //!< Center-surround data before maxnormalization
  NORM_CS  = 2  //!< Center-surround data after maxnormalization
};

//!< Number of visual feature types:
#define NBVISUALFEATURETYPES 3

enum VisualFeatureHierarchy
  {
    V1 = 1, //!< V1 channel in visual hierarchy
    V2 = 2, //!< V2 channel in visual hierarchy
    V4 = 4  //!< V4 channel in visual hierarchy
  };

//!< Number of levels in the visual hierarchy
#define NBVISUALFEATURELEVELS 3

//! Access feature name in clear
inline const char *featureName(const VisualFeature f);

//! Access feature type name in clear
inline const char *featureTypeName(const VisualFeatureType t);

//! Access feature level in the visual hierarchy
inline const int featureHierarchyLevel(const  VisualFeature f);

// ######################################################################
// ######################################################################
// ######################################################################
inline const char *featureName(const VisualFeature f)
{
  static const char n[NBVISUALFEATURES][20] = {
    "<Unknown>", "CompositeColor", "Color", "SOColor", 
    "Red-Green", "Blue-Yellow", 
    "SORed-Green", "SOBlue-Yellow", "SOGreen-Red", "SOYellow-Blue", 
    "Red", "Green", "Blue", "Yellow", 
    "Intensity", "Flicker", "Orientation", "Template", "ColorBand", 
    "Motion", "Convolve", 
    "L-Junction", "T-Junction", "X-Junction", "EndPoint", "RGBConvolve",
    "Entropy", "Dummy", "PN03contrast",
    "Disparity", "Stereo", "Information", "Recognition", "Hue",
    "DirectFeed", "MultiDirectFeed", "Variance", "Tcorr", "Scorr"
  };
  return n[int(f)];
}

// ######################################################################
inline const char *featureTypeName(const VisualFeatureType t)
{
  static const char n[NBVISUALFEATURETYPES][27] = {
    "Raw", "Raw Center-Surround", "Normalized Center-Surround" };
  return n[int(t)];
}

// ######################################################################
inline const int featureHierarchyLevel(const  VisualFeature f)
{
  static const int n[NBVISUALFEATURES] = {
    1,    // UNKNOWN         
    1,    // COMPOSITECOLOR  
    1,    // COLOR           
    1,    // SOCOLOR         
    1,    // RG              
    1,    // BY              
    1,    // SORG            
    1,    // SOGR            
    1,    // SOBY            
    1,    // SOYB            
    1,    // R               
    1,    // G               
    1,    // B               
    1,    // Y               
    1,    // INTENS          
    1,    // FLICKER         
    1,    // ORI             
    1,    // TEMPLATE        
    1,    // COLBAND         
    1,    // MOTION          
    1,    // CONVOLVE        
    2,    // LJUNCTION       
    2,    // TJUNCTION       
    2,    // XJUNCTION       
    2,    // ENDPOINT        
    1,    // RGBCONVOLVE     
    1,    // ENTROPY         
    1,    // DUMMY           
    1,    // PN03CONTRAST    
    1,    // DISPARITY       
    1,    // STEREO
    1,    // INFORMATION
    2,    // RECOGNITION
    1,    // Hue
    1,    // DirectFeed
    1,    // MultiDirectFeed
    1,    // Variance
    1,    // Tcorr
    1     // Scorr
  };
  return n[int(f)];
}

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
