//
//	VisualFeaturesAndPositions.cc		This file is a part of the IKAROS project
// 				<Short description of the module>
//
//    Copyright (C) 2005 Mikael Asker
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//    
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//    
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	Created: 050217
//
//	<Additional description of the module>

#include "VisualFeaturesAndPositions.h"

#include <algorithm>

#include "AttentionGuidanceMap.h"
#include "Channels.h"
#include "SimulationViewer.h"
#include "SaliencyMap.h"
#include "TaskRelevanceMap.h"
#include "VisualCortex.h"
#include "WinnerTakeAll.h"

using namespace iNVT;

// log.h :

//! need to define this once for each executable, typically equals exec name
char *MYLOGNAME = "VFP";


const float aNotAFrameRateValue = -1;


/*static*/ VisualFeaturesAndPositions::FeatureValueVector 
VisualFeaturesAndPositions::featureValuesInputToVector(float** theFeatureValuesInput)
{
    const float * someFeatureValues = *theFeatureValuesInput;
    FeatureValueVector aVector(aNumberOfFeatureValues);
    
    for (unsigned i = 0; i < aNumberOfFeatureValues; i++)
    {
        aVector[i] = someFeatureValues[i];
    }
    
    return aVector;
}


Module * 
VisualFeaturesAndPositions::Create(char* theName, Parameter* theParameters)
{
    return new VisualFeaturesAndPositions(theName, theParameters);
}


/*static*/ const float 
VisualFeaturesAndPositions::anEvolveTickLength = 1E-4;   // 0.1 ms

VisualFeaturesAndPositions::VisualFeaturesAndPositions(char* theName, Parameter* theParameters) :
    Module(theName),
    iAmProcessingInRealTime(GetIntValue(theParameters, "real_time", 0)),
    myFrameRate(GetFloatValue(theParameters, "frame_rate", aNotAFrameRateValue)),
    iAmProcessingAStillImage(! iAmProcessingInRealTime && myFrameRate == aNotAFrameRateValue),
    mySaliencyMapSpeed(iAmProcessingAStillImage ? 1 : myFrameRate / /*(3 * 40)*/ aStillFrameRate),
	myFloatToByteScaleFactor(255 / GetIntValue(theParameters, "max", 255)),
    myNoOfInputsX(0),
    myNoOfInputsY(0),
    myResolution(0),
    myReducedResolution(0),
    myFilterBiasInput(0),
    myTRMInput(0),
    myIORInput(0),
    myInputRed(0),
    myInputGreen(0),
    myInputBlue(0),
    myIntensityOutput(0),
    myColorOutput(0),
    myOrientationOutput(0),
    myFlickerOutput(0),
    myMotionOutput(0),
    myVCXOutput(0),
    mySMOutput(0),
    myAGMOutput(0),
    myWTAOutput(0),
    myTrajectoryRed(0),
    myTrajectoryGreen(0),
    myTrajectoryBlue(0),
    myFOAOutput(0),
    myFeatureValuesOutput(0),
    myModelManager("VisualFeaturesAndPositions"),
    myBrain(new SwitchableIORBrain(myModelManager)),
    myTickCounter(0),
    myInitCountLeft(2),   // Two initialization ticks
    myFrameCounter(0),
    mySaliencyMapTime(0.0)
{
    if (iAmProcessingAStillImage)
    {
        myFrameRate = /*100*/ aStillFrameRate;   // Hertz
    }
    
    assert(iAmProcessingInRealTime || myFrameRate != aNotAFrameRateValue);

    // AddInput here declares one input for the module. This is the same
    // name used in the experiment configuration file.

    AddInput("FILTER_BIAS_INPUT");
    AddInput("TRM_INPUT");
    AddInput("IOR_INPUT");

    AddInput("INPUT_RED");
    AddInput("INPUT_GREEN");
    AddInput("INPUT_BLUE");


    // Here we declare an output and tell the simulator that its size will be
    // figured out later. The parameter is not strictly necessary as it is the
    // default value, but it is a good idea to include it for clarity.

    AddOutput("INTENSITY");
    AddOutput("COLOR");
    AddOutput("ORIENTATION");
    AddOutput("FLICKER");
    AddOutput("MOTION");
    
    AddOutput("VCX");
    AddOutput("SM");
    AddOutput("AGM");
    AddOutput("WTA");

    AddOutput("TRAJECTORY_RED");
    AddOutput("TRAJECTORY_GREEN");
    AddOutput("TRAJECTORY_BLUE");

    AddOutput("FOA_OUTPUT");
    AddOutput("FEATURE_VALUES_OUTPUT");


    MYLOGVERB = LOG_ERR;
    //MYLOGVERB = LOG_DEBUG;    
}


void
VisualFeaturesAndPositions::SetSizes()
{
    // Find out the size of the input image

    const int anImageSizeX = GetInputSizeX("INPUT_RED");
    const int anImageSizeY = GetInputSizeY("INPUT_RED");

    if (anImageSizeX == unknown_size || anImageSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the red input size is not set\n");
        return;
    }

    if (GetInputSizeX("INPUT_GREEN") != anImageSizeX || GetInputSizeY("INPUT_GREEN") != anImageSizeY ||
        GetInputSizeX("INPUT_BLUE") != anImageSizeX  || GetInputSizeY("INPUT_BLUE") != anImageSizeY)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the input sizes are different\n");
        return;
    }

    myResolution = max(anImageSizeX, anImageSizeY);

    myReducedResolution = myResolution >> 4;
    if (myResolution & 0xf)
    {
        myReducedResolution++;
    }


    // Find out the size of the filter bias input.

    // We first check that the size is set.
    // If not, we abort with an error message.

    const int aFilterBiasInputSizeX = GetInputSizeX("FILTER_BIAS_INPUT");
    const int aFilterBiasInputSizeY = GetInputSizeY("FILTER_BIAS_INPUT");

    //cout <<  aFilterBiasInputSizeX << ", " << aFilterBiasInputSizeY << endl;

    if (aFilterBiasInputSizeX == unknown_size || aFilterBiasInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the filter bias input size is not set\n");
        return;
    }

    if (aFilterBiasInputSizeX != aNumberOfFeatureValues + 1 || aFilterBiasInputSizeY != 2)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the filter bias input size is wrong\n");
        return;
    }


    // Find out the size of the TRM input

    const int aTRMInputSizeX = GetInputSizeX("TRM_INPUT");
    const int aTRMInputSizeY = GetInputSizeY("TRM_INPUT");

    //cout <<  myReducedResolution << ", " << aTRMInputSizeX << ", " << aTRMInputSizeY << endl;

    if (aTRMInputSizeX == unknown_size || aTRMInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the TRM input size is not set\n");
        return;
    }

    if (aTRMInputSizeX != myReducedResolution || aTRMInputSizeY != myReducedResolution)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the TRM input size is wrong\n");
        return;
    }


    // Find out the size of the IOR input

    const int anIORInputSizeX = GetInputSizeX("IOR_INPUT");
    const int anIORInputSizeY = GetInputSizeY("IOR_INPUT");

    // We first check that the size is set.
    // If not, we abort with an error message.

    if (anIORInputSizeX == unknown_size || anIORInputSizeY == unknown_size)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the IOR input size is not set\n");
        return;
    }

    if (anIORInputSizeX != 1 || anIORInputSizeY != 1)
    {
        this->Notify(msg_fatal_error, "VisualFeaturesAndPositions: the IOR input is not scalar\n");
        return;
    }


    // Set the size of the output vector to be equal to the size 
    // of the input vector.

    SetOutputSize("INTENSITY", myReducedResolution, myReducedResolution);
    SetOutputSize("COLOR", myReducedResolution, myReducedResolution);
    SetOutputSize("ORIENTATION", myReducedResolution, myReducedResolution);
    SetOutputSize("FLICKER", myReducedResolution, myReducedResolution);
    SetOutputSize("MOTION", myReducedResolution, myReducedResolution);
    
    SetOutputSize("VCX", myReducedResolution, myReducedResolution);
    SetOutputSize("SM", myReducedResolution, myReducedResolution);
    SetOutputSize("AGM", myReducedResolution, myReducedResolution);
    SetOutputSize("WTA", myReducedResolution, myReducedResolution);
    
    SetOutputSize("TRAJECTORY_RED", myResolution, myResolution);
    SetOutputSize("TRAJECTORY_GREEN", myResolution, myResolution);
    SetOutputSize("TRAJECTORY_BLUE", myResolution, myResolution);

    SetOutputSize("FOA_OUTPUT", 2);
    SetOutputSize("FEATURE_VALUES_OUTPUT", aNumberOfFeatureValues);
}


void
VisualFeaturesAndPositions::Init()
{
    if (iAmProcessingInRealTime)
    {
        this->Notify(msg_print,
                     "VisualFeaturesAndPositions::Init : processing in real-time.\n");
    }
    else   // Not real-time
    {
        assert(myFrameRate != aNotAFrameRateValue);

        this->Notify(msg_print,
                     "VisualFeaturesAndPositions::Init : myFrameRate = %f Hz\n",
                     myFrameRate);
    }

    myNoOfInputsX = GetInputSizeX("INPUT_RED");
    myNoOfInputsY = GetInputSizeY("INPUT_RED");
       
    // Here we could add any number of consistency checks and do any memory
    // allocations and other initializations that wer need to do.

    // Set the variables that will be associated with the inputs and outputs
    // during the simulation. Each time the Tick method is entered, myInput
    // will point to the input vector, and myOutput will point to
    // the output vector. 

    myInputRed = GetInputMatrix("INPUT_RED");
    myInputGreen = GetInputMatrix("INPUT_GREEN");
    myInputBlue = GetInputMatrix("INPUT_BLUE");

    myFilterBiasInput = GetInputMatrix("FILTER_BIAS_INPUT");
    myTRMInput = GetInputMatrix("TRM_INPUT");
    myIORInput = GetInputMatrix("IOR_INPUT");

    myIntensityOutput = GetOutputMatrix("INTENSITY");
    myColorOutput = GetOutputMatrix("COLOR");
    myOrientationOutput = GetOutputMatrix("ORIENTATION");
    myFlickerOutput = GetOutputMatrix("FLICKER");
    myMotionOutput = GetOutputMatrix("MOTION");

    myVCXOutput = GetOutputMatrix("VCX");
    mySMOutput = GetOutputMatrix("SM");
    myAGMOutput = GetOutputMatrix("AGM");
    myWTAOutput = GetOutputMatrix("WTA");

    myTrajectoryRed = GetOutputMatrix("TRAJECTORY_RED");
    myTrajectoryGreen = GetOutputMatrix("TRAJECTORY_GREEN");
    myTrajectoryBlue = GetOutputMatrix("TRAJECTORY_BLUE");

    myFOAOutput = GetOutputMatrix("FOA_OUTPUT");
    myFeatureValuesOutput = GetOutputMatrix("FEATURE_VALUES_OUTPUT");


    myModelManager.registerComponent(myBrain);

    const char * aName = "TEST";
    const char * anArgv[1] = {aName};

    if (myModelManager.parseCommandLine(0, anArgv, "", 0, 0) == false)
    {
      throw "ModelManager::parseCommandLine failed in VisualFeaturesAndPositions constructor";
    }

    myModelManager.setModelParamVal("FOAradius", min(myNoOfInputsX, myNoOfInputsY) / /*20*/ 12, true, true);
    //myModelManager.setModelParamVal("SimulationTimeStep", 0.00015, true, true);
    myModelManager.start();

    /*SharedPtr<VisualCortex> aVisualCortex = myBrain->getVC();
    const int aNumberOfChannels = aVisualCortex->numChannels();

    for (int aChannelIndex = 0; aChannelIndex < aNumberOfChannels; aChannelIndex++)
    {
        printf("%d : %s\n",
               aChannelIndex,
               aVisualCortex->subComponent(aChannelIndex)->tagName().c_str());
    }*/
}


VisualFeaturesAndPositions::~VisualFeaturesAndPositions()
{
  myModelManager.stop();
}


inline static void
copyPixel(float thePixel, int x, int y, float** theOutput)
{
    theOutput[y][x] = thePixel;
}


static void 
resetOutputImage(int theResolution, float** theOutput)
{
    for (int i = 0; i < theResolution; i++) 
    {
        for (int j = 0; j < theResolution; j++) 
        {
            copyPixel(0.0, i, j, theOutput);
        }
    }
}


static void 
outputImage(const Image<float>& theImage, float** theOutput)
{
    const int32 aMapWidth = theImage.getWidth();
    const int32 aMapHeight = theImage.getHeight();

    for (int i = 0; i < aMapWidth; i++) 
    {
        for (int j = 0; j < aMapHeight; j++) 
        {
            copyPixel(theImage.getVal(i, j), 
                      i, 
                      j,
                      theOutput);
        }
    }
}


inline static void
copyPixel(const PixRGB<byte>& thePixel, 
          int                 x, 
          int                 y, 
          float**             theRedOutput, 
          float**             theGreenOutput, 
          float**             theBlueOutput)
{
    theRedOutput[y][x] = thePixel.red();
    theGreenOutput[y][x] = thePixel.green();
    theBlueOutput[y][x] = thePixel.blue();
}


static void 
resetOutputImage(int     theResolution,
                 float** theRedOutput, 
                 float** theGreenOutput, 
                 float** theBlueOutput)
{
    const PixRGB<byte> aPixel(0, 0, 0);

    for (int i = 0; i < theResolution; i++) 
    {
        for (int j = 0; j < theResolution; j++) 
        {            
            copyPixel(aPixel, 
                      i, 
                      j,
                      theRedOutput, 
                      theGreenOutput, 
                      theBlueOutput);
        }
    }
}


static void 
outputImage(const Image< PixRGB<byte> >&  theImage, 
            float**                       theRedOutput, 
            float**                       theGreenOutput, 
            float**                       theBlueOutput)
{
    const int32 aMapWidth = theImage.getWidth();
    const int32 aMapHeight = theImage.getHeight();
    //assert(aMapWidth <= myNoOfInputsX && aMapHeight <= myNoOfInputsY);

    for (int i = 0; i < aMapWidth; i++) 
    {
        for (int j = 0; j < aMapHeight; j++) 
        {
            const PixRGB<byte> aPixel = theImage.getVal(i, j);
            
            copyPixel(aPixel, 
                      i, 
                      j,
                      theRedOutput, 
                      theGreenOutput, 
                      theBlueOutput);
        }
    }
}


void
VisualFeaturesAndPositions::Tick()
{
    /*this->Notify(msg_print,
                 "VisualFeaturesAndPositions::Tick : Ikaros time = %.1lf, mySaliencyMapTime = %.1lf ms **********\n",
                 mySaliencyMapTime / mySaliencyMapSpeed * 1000,   // milliseconds
                 mySaliencyMapTime * 1000);   // milliseconds*/

    // VisualFeaturesAndPositions in action. Input is a pointer to the current input
    // vector and output points to the output vector that will be 
    // sent along to whatever modules it is connected to.
    
    if (myInitCountLeft > 0)   // An init tick
    {
        resetOutputImage(myReducedResolution, myIntensityOutput);
        resetOutputImage(myReducedResolution, myColorOutput);
        resetOutputImage(myReducedResolution, myOrientationOutput);
        resetOutputImage(myReducedResolution, myFlickerOutput);
        resetOutputImage(myReducedResolution, myMotionOutput);
        resetOutputImage(myReducedResolution, myVCXOutput);
        resetOutputImage(myReducedResolution, mySMOutput);
        resetOutputImage(myReducedResolution, myAGMOutput);
        resetOutputImage(myReducedResolution, myWTAOutput);

        resetOutputImage(myResolution, myTrajectoryRed, myTrajectoryGreen, myTrajectoryBlue);

        myInitCountLeft--;
    }
    else   // Not an init tick
    {
        myFrameCounter++;   // Count this frame

        // EVOLVE
    
        Image<float> aTRMImage(10, 7, ZEROS);
            
        for (int i = 0; i < 10; i++) 
        {
            for (int j = 0; j < 7; j++) 
            {
                aTRMImage.setVal(i, j, myTRMInput[j][i]);
            }
        }

        SharedPtr<TaskRelevanceMap> aTRM = myBrain->getTRM();
        //cout << "TRM size : " << aTRM->getWidth() << " x " << aTRM->getHeight() << endl;
        aTRM->input(aTRMImage);

        if (**myIORInput < 0.5)
        {
            myBrain->stopIOR();
        }
        else   // **myIORInput >= 0.5
        {
            myBrain->startIOR();
        }
    
        double anIkarosStopTime = 0;
        static double aSaliencyMapStopTime = 0;

        if (iAmProcessingInRealTime)
        {
            static bool aStartedFlag = false;
		    static struct timeval aStartTimeStruct;
        
            if (! aStartedFlag)
            {
    		    gettimeofday(&aStartTimeStruct, NULL);
                aStartedFlag = true;
		        anIkarosStopTime = 0;   // Current time in seconds	
		        aSaliencyMapStopTime = 0;

                /*this->Notify(msg_print,
                             "VisualFeaturesAndPositions::Tick : started at %ld.%06ld s.\n", 
                             aStartTimeStruct.tv_sec,
                             aStartTimeStruct.tv_usec);*/
            }
            else
            {
        		struct timeval aCurrentTimeStruct;
	        	gettimeofday(&aCurrentTimeStruct, NULL);

                /*this->Notify(msg_print,
                             "VisualFeaturesAndPositions::Tick : now at %ld.%06ld s.\n", 
                             aCurrentTimeStruct.tv_sec,
                             aCurrentTimeStruct.tv_usec);*/

    	    	const long aSecondDiff = aCurrentTimeStruct.tv_sec - aStartTimeStruct.tv_sec;
	    	    const long aMicroSecondDiff = aCurrentTimeStruct.tv_usec - aStartTimeStruct.tv_usec;

                // Current time in seconds	
		        anIkarosStopTime = aSecondDiff + (double)aMicroSecondDiff / 1000000L;   

               /* this->Notify(msg_print,
                             "VisualFeaturesAndPositions::Tick : frame %u : time %f s, average interval %f s\n", 
                             myFrameCounter,
                             anIkarosStopTime, 
                             anIkarosStopTime / (myFrameCounter - 1));*/

                static double aLastIkarosStopTime = 0;
                const double anIntervalLength = anIkarosStopTime - aLastIkarosStopTime;
                aLastIkarosStopTime = anIkarosStopTime;

                mySaliencyMapSpeed = 1.0 / anIntervalLength / /*(3 * 40)*/ aStillFrameRate;
                aSaliencyMapStopTime += mySaliencyMapSpeed * anIntervalLength;
            }
        }
        else   // Not real-time
        {
            assert(myFrameRate != aNotAFrameRateValue);
            anIkarosStopTime = (myFrameCounter - 1) / myFrameRate;
            aSaliencyMapStopTime = anIkarosStopTime * mySaliencyMapSpeed;
        }

        /*this->Notify(msg_print,
                     "VisualFeaturesAndPositions::Tick : aSaliencyMapStopTime = %f s\n", 
                     aSaliencyMapStopTime);*/

        bool aWinFlag = false;

        while (mySaliencyMapTime < aSaliencyMapStopTime)
        {
            Point2D aWinner;
            int32 aNumberOfTargetsHit = 0;
            float aWinnerVoltage = 0.0;
            myBrain->evolve(mySaliencyMapTime, aWinner, aNumberOfTargetsHit, aWinnerVoltage);

            if (aWinner.i != -1 || aWinner.j != -1)
            {
                this->Notify(msg_print,
                             "%u VisualFeaturesAndPositions::Tick : aWinner = (%u,%u).\n",
                             myTickCounter,
                             aWinner.i,
                             aWinner.j);
                                 
                aWinFlag = true;
                
                (*myFOAOutput)[0] = aWinner.i;
                (*myFOAOutput)[1] = aWinner.j;

                // Get the center-surround feature responses of all feature maps 
                std::vector<double> someFeatureValues;
                myBrain->getVC()->getFeatures(aWinner, someFeatureValues);
                /*this->Notify(msg_print,
                             "VisualFeaturesAndPositions::Tick : someFeatureValues.size() = %u\n",
                             someFeatureValues.size());*/
                assert(someFeatureValues.size() == aNumberOfFeatureValues);
                for (unsigned i = 0; i < aNumberOfFeatureValues; i++)
                {
                    (*myFeatureValuesOutput)[i] = someFeatureValues[i];
                }
            }
        }

        outputImage(myBrain->getSM()->getV(), mySMOutput);
        outputImage(myBrain->getAGM()->getV(), myAGMOutput);
        outputImage(myBrain->getWTA()->getV(), myWTAOutput);

        Image< PixRGB<byte> > aTrajectory = myBrain->getSV()->getTraj(0);
        assert(aTrajectory.getWidth() <= myNoOfInputsX && aTrajectory.getHeight() <= myNoOfInputsY);

        outputImage(myBrain->getSV()->getTraj(anIkarosStopTime), 
                    myTrajectoryRed, 
                    myTrajectoryGreen, 
                    myTrajectoryBlue);

        // INPUT
                    
        static int anInputDoneFlag = false;   // A still image should be input just once

        if (! (iAmProcessingAStillImage && anInputDoneFlag))   // An input tick - only once for stills
        {
            if (myFilterBiasInput[0][aNumberOfFeatureValues] > 0.5)
            {
                //this->Notify(msg_print, "VisualFeaturesAndPositions::Tick : bias features.\n");

                double aMean[aNumberOfFeatureValues];
                double aSigma[aNumberOfFeatureValues];

                for (unsigned i = 0; i < aNumberOfFeatureValues; i++)
                {
                    aMean[i] = myFilterBiasInput[0][i];   // float -> double conversion
                    aSigma[i] = myFilterBiasInput[1][i];
                }

                myBrain->getVC()->biasFeatures(aMean, aSigma);
            }

            Image< PixRGB<byte> > anInputImage(myNoOfInputsX, myNoOfInputsY, ZEROS);

            for (int i = 0; i < myNoOfInputsX; i++) 
            {
                for (int j = 0; j < myNoOfInputsY; j++) 
                {
                    const PixRGB<byte> aPixel(lround(myFloatToByteScaleFactor * myInputRed[j][i]), 
                                              lround(myFloatToByteScaleFactor * myInputGreen[j][i]), 
                                              lround(myFloatToByteScaleFactor * myInputBlue[j][i]));

                    anInputImage.setVal(i, j, aPixel);
                }
            }

            Image<byte> someTargets;
            myBrain->input(anInputImage, someTargets);

            anInputDoneFlag = true;

            outputImage(myBrain->getVC()->channel("intensity").getOutput(), myIntensityOutput);
            outputImage(myBrain->getVC()->channel("color").getOutput(), myColorOutput);
            outputImage(myBrain->getVC()->channel("orientation").getOutput(), myOrientationOutput);
            outputImage(myBrain->getVC()->channel("flicker").getOutput(), myFlickerOutput);    
            outputImage(myBrain->getVC()->channel("motion").getOutput(), myMotionOutput);
            outputImage(myBrain->getVC()->getOutput(), myVCXOutput);
        }

    }   // Not an init tick

    myTickCounter++;
    //this->Notify(msg_print, "VisualFeaturesAndPositions::Tick ends.\n");
}
