/*!@file VisualObject.C Visual Objects to be recognized */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Philip Williams <plw@usc.edu>
// $Id: VisualObject.C 4072 2004-12-02 09:00:34Z jbonaiuto $
//

#include "VisualObject.H"
#include "Histogram.H"
#include "KDTree.H"
#include "Image_ColorOps.H"
#include "Image_ShapeOps.H"
#include "Raster.H"
#include "log.H"
#include <cstring>
#include <math.h>

// ######################################################################
// ######################################################################
// ############ KeypointMatch implementation
// ######################################################################
// ######################################################################

// ######################################################################
KeypointMatch::KeypointMatch()
{
}

// ######################################################################
KeypointMatch::KeypointMatch(const Keypoint& k1, const Keypoint& k2)
{
  key1 = k1;
  key2 = k2; 
  distSq = k1.distSquared(k2);
}

// ######################################################################
const Keypoint& KeypointMatch::getKeypoint1() const
{
  return key1;
}

// ######################################################################
const Keypoint& KeypointMatch::getKeypoint2() const
{
  return key2;
}

// ######################################################################
const int KeypointMatch::getDistanceSq() const
{
  return distSq;
}

// ######################################################################
int KeypointMatch::compareTo(const KeypointMatch& match) const
{
  if(distSq > match.getDistanceSq())
    return 1;
  else if(distSq < match.getDistanceSq())
    return -1;
  else
    return 0;
}

// ######################################################################
bool KeypointMatch::equals(const KeypointMatch& match) const
{
  if(!((key1.equals(match.getKeypoint1()) && 
        key2.equals(match.getKeypoint2())) ||
       (key1.equals(match.getKeypoint2()) && 
        key2.equals(match.getKeypoint1()))))
    return false;
  return true;
}

// ######################################################################
// ######################################################################
// ############ VisualObjectMatch implementation
// ######################################################################
// ######################################################################

// ######################################################################
VisualObjectMatch::VisualObjectMatch()
{
  numKeyCompare = 0;
  success = false;
}

// ######################################################################
const bool VisualObjectMatch::getSuccess() const
{
  return success;
}

// ######################################################################
void VisualObjectMatch::setSuccess(bool s)
{
  success = s;
}

// ######################################################################
void VisualObjectMatch::setOutputVerticalOffset(int value)
{
  outputVertOffset = value;
}

// ######################################################################
void VisualObjectMatch::match(const Keypoint& k1, const Keypoint& k2)
{
  KeypointMatch match(k1, k2);
  matches.add(match); 
}

// ######################################################################
const char* VisualObjectMatch::getObjectName() const
{
  return objectName;
}

// ######################################################################
void VisualObjectMatch::setObjectName(char* name)
{
  objectName = NULL;
  objectName = new char[strlen(name) + 1];
  strcpy(objectName, name);
}

// ######################################################################
const Image<PixRGB<byte> >& VisualObjectMatch::getOutput() const
{
  return output;
}

// ######################################################################
void VisualObjectMatch::setOutput(Image<PixRGB<byte> >& img)
{
  output = img;
}

// ######################################################################
const int VisualObjectMatch::getMatchCount() const
{
  return matches.getSize();
}

// ######################################################################
const float VisualObjectMatch::getScore() const
{
  return score;
}

// ######################################################################
void VisualObjectMatch::calculateScore()
{
  score = 0.0f;
  for(int i=0; i<matches.getSize(); i++)
    {
      //score += 1.0f - ((float)distSq/k1.maxDistSquared());
      score += ((float)matches.getValue(i).getDistanceSq());
      // / k1.maxDistSquared());
    }
  if(matches.getSize() > 0)
    score /= matches.getSize();
}

// ######################################################################
void VisualObjectMatch::writeOutput(char* outputFile)
{
  calculateScore();
  LINFO("Indentified %s in %i keypoint comparisons",objectName,numKeyCompare);
  LINFO("Score: %f", score);

  for(int i=0; i<matches.getSize(); i++)
    {
      Keypoint k1 = matches.getValue(i).getKeypoint1();
      Keypoint k2 = matches.getValue(i).getKeypoint2();

      /* Draw a line on the image from k1 to k2.  Note that we
         must add row count of first image to row position in second so
         that line ends at correct location in second image.
      */
      int p1x = (int)k1.getX();
      int p1y = (int)k1.getY();
      int p2x = (int)k2.getX();
      int p2y = (int)(k2.getY() + outputVertOffset);
      output.drawLine(Point2D(p1x,p1y), Point2D(p2x,p2y), 
                      255.0, 1);
    }
  
  //If an output filename was specified
  if(strlen(outputFile) > 0)
    {
      // create new output filename based on recognized obj label
      char* filename = (char*)(strcat(strcat(strcat(outputFile, "_"), 
                                             objectName), ".ppm"));
      
      // write the output image:
      Raster::WriteRGB(output, Auto, "%s", filename);
    }
}

// ######################################################################
void VisualObjectMatch::incrementNumberKeyCompare(int incr)
{
  numKeyCompare += incr;
}

// ######################################################################
int VisualObjectMatch::getNumberKeyCompare() const
{
  return numKeyCompare;
}

// ######################################################################
// ######################################################################
// ############ VisualObject implementation
// ######################################################################
// ######################################################################

// ######################################################################
VisualObject::VisualObject(int sortOrder)
{
  this->name = NULL;
  this->imageFilename = NULL;
  this->sortOrder = sortOrder;
  this->timesRecognized = 0;
  this->lastRecognized = Point2D(-1,-1);
}

// ######################################################################
VisualObject::VisualObject(const char* n, int sortOrder) 
{
  this->name = NULL;
  this->imageFilename = NULL;
  setName((char*)n);
  this->sortOrder = sortOrder;
  this->timesRecognized = 0;
  this->lastRecognized = Point2D(-1,-1);
}

// ######################################################################
VisualObject::VisualObject(FILE *fp, int sortOrder)
{
  this->sortOrder = sortOrder;
  name = NULL;
  
  char n[200];
  char i[200];

  // Read name
  if(fscanf(fp, "%s", &n[0]) != 1)
    LINFO("Invalid object database file format");
  setName(n);

  // Read image filename
  if(fscanf(fp, "%s", &i[0]) != 1)
    LINFO("Invalid object database file format");
  setImageFilename(i);
  // Read image
  // Check if we have a valid TargetImageFile name; if not, then just give
  // up and don't do anything (otherwise we'll crash in ReadRGB when trying
  // to read a non-existent file)
  if (strlen(imageFilename) >  0)
    // load the image:
    image = Raster::ReadRGB(Auto, "%s", imageFilename);

  // Read number of times recognized
  if(fscanf(fp, "%d", &timesRecognized) != 1)
    LINFO("Invalid object database file format");


  // Read location last recognized
  int x=-1,y=-1;
  if(fscanf(fp, "%d %d", &x, &y) != 2)
    LINFO("Invalid object database file format");
  lastRecognized = Point2D(x,y);

  // Read the feature vector
  int featureLen;
  float val;
  if(fscanf(fp, "%d", &featureLen) != 1)
    LINFO("Invalid object database file format");
  std::vector<double> f;
  for(int j=0; j<featureLen; j++)
    {
      if(fscanf(fp, "%f", &val) != 1)
        LINFO("Invalid object database file format");
      f.push_back((double)val);
    }
  setFeature(f);
  
  // Read the keypoints
  int num, len;
  if (fscanf(fp, "%d %d", &num, &len) != 2)
    LINFO("Invalid keypoint file beginning.");
  for (int i = 0; i < num; i++)
    {
      Keypoint k(len);
      k.load(fp);
      keys.push_back(k);
      sortedKeys.addUnsorted(k);
    }
  sortedKeys.sort(sortedKeys.getSize(), 0, sortOrder);
}

// ######################################################################
VisualObject::VisualObject(const VisualObject& vo)
{
  if(vo.name != NULL)
    setName(vo.name);
  if(vo.imageFilename != NULL)
    setImageFilename(vo.imageFilename);
  if(vo.image.initialized())
    image = vo.image;
  keys = vo.keys;
  feature = vo.feature;
  timesRecognized = vo.timesRecognized;
  sortedKeys = vo.sortedKeys;
  lastRecognized = vo.lastRecognized;
}

// ######################################################################
VisualObject::~VisualObject()
{
}

// ######################################################################
VisualObject& VisualObject::operator=(const VisualObject& vo)
{
  if(vo.name != NULL)
    setName(vo.name);
  if(vo.imageFilename != NULL)
    setImageFilename(vo.imageFilename);
  if(vo.image.initialized())
    image = vo.image;
  keys = vo.keys;
  feature = vo.feature;
  timesRecognized = vo.timesRecognized;
  sortedKeys = vo.sortedKeys;
  lastRecognized = vo.lastRecognized;
  return *this;
}

// ######################################################################
const char* VisualObject::getName( void ) const
{
  return name;
}

// ######################################################################
void VisualObject::setName(char* n)
{
  name = NULL;
  name = new char[strlen(n) + 1];
  strcpy(name, n);
}

// ######################################################################
const char* VisualObject::getImageFilename() const
{
  return imageFilename;
}

// ######################################################################
void VisualObject::setImageFilename(char* imgFilename)
{
  imageFilename = NULL;
  imageFilename = new char[strlen(imgFilename) + 1];
  strcpy(imageFilename, imgFilename);
}

// ######################################################################
const Image<PixRGB<byte> >& VisualObject::getImage() const
{
  return image;
}

// ######################################################################
void VisualObject::setImage(Image<PixRGB<byte> >& img)
{
  image = img;
}

// ######################################################################
void VisualObject::setFeature(std::vector<double>& f)
{
  feature = f;
  while(feature.size() > 42)
    feature.pop_back();
}

// ######################################################################
double VisualObject::getFeatureDistSq(const VisualObject& obj) const
{
  double distSq = 0.0;
  if(feature.size() == obj.getFeature().size())
    {
      for(uint k=0; k<obj.getFeature().size(); k++)
        {
          distSq += obj.getFeature()[k] - this->feature[k];
        }
      distSq *= distSq;
    }
  else
    distSq = 100000000000000.0;
  return distSq;
}

// ######################################################################
const std::vector<double>& VisualObject::getFeature() const
{
  return this->feature;
}

// ######################################################################
void VisualObject::incTimesRecognized()
{
  timesRecognized++;
}

// ######################################################################
const int VisualObject::getTimesRecognized() const
{
  return timesRecognized;
}

// ######################################################################
const Point2D& VisualObject::getLastRecognized() const
{
  return lastRecognized;
}

// ######################################################################
void VisualObject::setLastRecognized(Point2D p)
{
  lastRecognized = Point2D(p.i,p.j);
}

// ######################################################################
void VisualObject::save(FILE *fp) const
{
  fprintf(fp, "%s\n", name);
  fprintf(fp, "%s\n", imageFilename);
  fprintf(fp, "%d\n", timesRecognized);
  fprintf(fp, "%d %d\n", lastRecognized.i, lastRecognized.j);

  int featureSize = feature.size();
  fprintf(fp, "%d\n", featureSize);
  for(int i=0; i<featureSize; i++)
    fprintf(fp, "%3f ", feature[i]);
       
  if(keys.size() > 0)
    fprintf(fp, "%d %d\n", keys.size(), keys[0].getDimensionCount());
  else
    fprintf(fp, "0 0\n");
  
  for (uint i=0; i<keys.size(); i++)
    ((Keypoint)keys[i]).save(fp);
  
  // write image
  if(image.initialized())
    Raster::WriteRGB(image, Auto, "%s", imageFilename);
}

// ######################################################################
const VisualObjectMatch VisualObject::match(VisualObject& obj, 
                                            char* matchingAlg, 
                                            int recogMinMatch, 
                                            float recogMinMatchPercent,
                                            bool sortKeys)
{
  if(!image.initialized())
    getImage();
  VisualObjectMatch objMatch;
  // Create a new image that joins the two images vertically.
  Image<PixRGB<byte> > combinedImage = combineImagesVertically(this->image, 
                                                               obj.getImage());
  objMatch.setOutput(combinedImage);
  objMatch.setObjectName(this->name);
  
  Keypoint k, match, secondMatch;
  // Match the keys in object to their best matches in scene keypoints
  objMatch.setOutputVerticalOffset(this->image.getHeight());
  if(!sortKeys)
    obj.match(this->keys, objMatch, matchingAlg, recogMinMatch, 
              recogMinMatchPercent, sortKeys);
  else
    {
      obj.match(sortedKeys.toVector(), objMatch, matchingAlg, 
                recogMinMatch, recogMinMatchPercent, sortKeys);
    }
  objMatch.calculateScore();

  return objMatch;
}

// ######################################################################
void VisualObject::match(const std::vector<Keypoint>& objKeys, 
                         VisualObjectMatch& objMatch, char* matchingAlg, 
                         int recogMinMatch, float recogMinMatchPercent, 
                         bool sortKeys)
{
  Keypoint k, match;
  SharedPtr<KDTree> objTree;
  if(strcmp(matchingAlg, "KdTree") == 0 || 
     strcmp(matchingAlg, "KdTreeBBF") == 0)
    objTree = KDTree::createKDTree(objKeys);
  std::vector<Keypoint> myKeys = keys;
  std::vector<Keypoint> otherKeys = objKeys;
  if(sortKeys)
    myKeys = sortedKeys.toVector();
  
  for (uint i = 0; i < myKeys.size(); i++)
    {
      k = myKeys[i];
      int comparisons = 0;
      if(strcmp(matchingAlg, "Simple") == 0)
        {
          if (k.simpleMatch(otherKeys, match, comparisons)) 
            objMatch.match(match, k);
        }
      else if(strcmp(matchingAlg, "KdTree") == 0 ||
              strcmp(matchingAlg, "KdTreeBBF") == 0)
        {
          bool useBBF = (strcmp(matchingAlg, "KdTreeBBF") == 0);
          if (k.kdTreeMatch(objTree, match, useBBF, comparisons))
            objMatch.match(match, k);
        }
      objMatch.incrementNumberKeyCompare(comparisons);
          
      if((recogMinMatch > 0 && objMatch.getMatchCount() >= recogMinMatch) ||
         (recogMinMatch == 0 && objMatch.getMatchCount() >= 
          (myKeys.size() * recogMinMatchPercent)))
        {
          objMatch.setSuccess(true);
          break;
        }
    }
}

// ######################################################################
int VisualObject::createKeys(float x, float y, float s, int dimX, int dimY,
                             float octSize, double magnitude, 
                             Image<float>& grad, Image<float>& orie)
{
  int keyNumber = 0;
  // int position
  int xi = (int) floor(x + 0.5f);
  int yi = (int) floor(y + 0.5f);
  
  // orientation histogram
  Histogram OV(ORIENTARRAY);
  
  // radius calculus
  float sig = 1.5 * s;
  int radius = (int) (3.0 * sig);

  // 1. Calculate orientation vector 
  for (int ind_x = xi - radius; ind_x < xi + radius; ind_x++ )
    {
      for (int ind_y = yi - radius; ind_y < yi + radius; ind_y++ )
        {
	  // check limits
	  if (( ind_x < 0 ) || ( ind_y < 0) || (ind_x >= dimX) || 
	      (ind_y >= dimY) )
	    continue;

	  // get gradient
	  float gradVal = grad.getVal( ind_x, ind_y );
 	  float distSq  = pow( ind_x - x, 2.0 ) + pow( ind_y - y, 2.0 );
			
	  // if outside the circle then continue
	  if ( distSq >= radius * radius )
	    continue;
	  
	  // compute the gaussian weight for this histogram entry
	  float gaussianWeight = exp ( -distSq / (2.0 * sig * sig) );
	  
	  // add this orientation to the histogram
          // [-pi ; pi] -> [0 ; 2pi]
          float angle = orie.getVal( ind_x, ind_y) + M_PI;
  
          // [0 ; 2pi] -> [0 ; 36]
          angle = ( angle * ORIENTARRAY ) / (2.0 * M_PI);

          int index = (int) angle;
	
          // if exactly 2.pi => 0
          if ( index == ORIENTARRAY )
            index = 0;
  
	  OV.addValue( index, gaussianWeight * gradVal );
        }
    }
  
  // smooth 3 times
  int smoothIterations = 3;
  for(int i=0; i<smoothIterations; i++)
    OV.smooth();
   
  // find the max in the histogram
  float maxPeakValue = OV.findMax();
    
  // 2. Create feature vector of with orientation peaks
  for (int bin=0; bin<ORIENTARRAY; bin++ )
    {
      // get value to the left of current value
      int left = ORIENTARRAY - 1;
      if (bin > 0)
	left = bin - 1;
      
      // get value to the right of current value
      int right = 0;
      if (bin < (ORIENTARRAY - 1) )
	right = bin + 1;
      
      // only consider local peaks
      if ( OV.getValue(left) >= OV.getValue(bin) )
	continue;
      if ( OV.getValue(right) >= OV.getValue(bin) )
	continue;
      if ( OV.getValue(bin) < (0.8 * maxPeakValue) )
	continue;
      
      // interpolates the values to get the orientation
      //  with f(x) = ax² + bx + c
      //   f(-1) = x0
      //   f( 0) = x1
      //   f(+1) = x2
      //  => a = (x0+x2)/2 - x1
      //     b = (x2-x0)/2
      //     c = x1
      // f'(x) = 0 => x = -b/2a
      float x0 = OV.getValue(left);
      float x1 = OV.getValue(bin);
      float x2 = OV.getValue(right);
      float a  = 0.5f*(x0+x2) - x1;
      float b  = 0.5f*(x2-x0);
      float realangle = bin-b/(2*a);
      
      // [0:36] to [0:2*pi]
      realangle *= 2 * M_PI / ORIENTARRAY;
      // [0:2pi] to [-pi:pi]
      realangle -= M_PI;
     
      // Create keypoint
      Keypoint newkey(128, x * octSize, y * octSize, 
                      s * octSize, realangle, magnitude);
      image.drawCircle(Point2D((int)(x*octSize),(int)(y*octSize)), 1, 0.0);
      // compute feature vector for this keypoint
      newkey.computeFeature( x, y, s, orie, grad, dimX, dimY );
      
      // add to list of keys
      keys.push_back(newkey);
      sortedKeys.addUnsorted(newkey);
      keyNumber++;
    }
  sortedKeys.sort(sortedKeys.getSize(), 0, sortOrder);
  return keyNumber;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
