/*!@file VisualObject.H Header file for visual objects to be recognized */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Philip Williams <plw@usc.edu>
// $Id: VisualObject.H 4017 2004-11-10 06:11:39Z jbonaiuto $
//

#ifndef VISUALOBJECT_H_DEFINED
#define VISUALOBJECT_H_DEFINED

#include "Image.H"
#include "Point2D.H"
#include "KDTree.H"
#include "Keypoint.H"
#include <vector>

const int   ORIENTARRAY	= 36;	    // size of orientation array

//! KeypointMatch
class KeypointMatch
{
public:
  //! Constructor
  KeypointMatch();

  //! Constructor
  KeypointMatch(const Keypoint& k1, const Keypoint& k2);

  //! keypoint 1 getter
  const Keypoint& getKeypoint1() const;

  //! keypoint 2 getter
  const Keypoint& getKeypoint2() const;

  //! distance squared getter
  const int getDistanceSq() const;

  //! compare to another match
  int compareTo(const KeypointMatch& match) const;

  //! return if this match equals the specified match
  bool equals(const KeypointMatch& match) const;
  
private:
  Keypoint key1;
  Keypoint key2;
  int distSq;
};

//! VisualObjectMatch
class VisualObjectMatch
{
public:
  //! constructor
  VisualObjectMatch();

  //! get whether or not successful
  const bool getSuccess() const;

  //! set successful
  void setSuccess(bool s);

  //! get object name
  const char* getObjectName() const;

  //! set object name
  void setObjectName(char* name);

  //! get output
  const Image<PixRGB<byte> >& getOutput() const;

  //! set output
  void setOutput(Image<PixRGB<byte> >& img);

  //! get count
  const int getMatchCount() const;

  //! get score
  const float getScore() const;

  //! set score
  void calculateScore();

  //! update output with a match
  void match(const Keypoint& k1, const Keypoint& k2);
 
  //! write output
  void writeOutput(char* outputFile);

  //! Set the vertical offset of the output (height of the top image)
  void setOutputVerticalOffset(int value);

  //! Increment the number of keypoint comparisons made
  void incrementNumberKeyCompare(int incr);

  //! Get the number of keypoint comparisons made
  int getNumberKeyCompare() const;

private:
  char* objectName;
  Image<PixRGB<byte> > output;
  float score;
  int outputVertOffset;
  SortedLimitedList<KeypointMatch> matches;
  int numKeyCompare;
  bool success;
};

//! describes name and attributes of a visual object
class VisualObject 
{
public:
  //! Constructor
  VisualObject::VisualObject(int sortOrder=0);

  //! Constructor
  VisualObject(const char* n, int sortOrder=0);

  //! Construct by reading file
  VisualObject(FILE *fp, int sortOrder=0);

  //! Copy constructor
  VisualObject(const VisualObject& vo);

  //! Destructor
  ~VisualObject();
  
  //! Assignment constructor
  VisualObject& operator=(const VisualObject& vo);

  //! name getter
  const char* getName() const;

  //! name setter
  void setName(char* n);
  
  //! image filename getter
  const char* getImageFilename() const;

  //! image filename setter
  void setImageFilename(char* imgFilename);

  //! save object to file
  void save(FILE* fp) const;

  //! image getter
  const Image<PixRGB<byte> >& getImage() const;

  //! image setter
  void setImage(Image<PixRGB<byte> >& img);

  //! features setter
  void setFeature(std::vector<double>& f);

  //! get the distance squared between the closest two feature vectors from 
  /*! this object and the given object
   */
  double getFeatureDistSq(const VisualObject& obj) const;

  //! get feature (needed to add features to matched object
  const std::vector<double>& getFeature() const;

  //! find match using the given object and algorithm
  const VisualObjectMatch match(VisualObject& obj, char* matchingAlg,
                                int recogMinMatch, float recogMinMatchPercent,
                                bool sortKeys);

  //! find match using the given keypoints and algorithm
  void match(const std::vector<Keypoint>& objKeys, 
             VisualObjectMatch& objMatch,
             char* matchingAlg, int recogMinMatch, float recogMinMatchPercent,
             bool sortKeys);

  //! Creates object keys. Returns the number created
  int createKeys(float x, float y, float s, int dimX, int dimY, 
                 float octSize, double magnitude, Image<float>& grad, 
                 Image<float>& orie);

  //! Increments the number of times this object has been recognized
  void incTimesRecognized();

  //! Returns the number of times this object has been recognized
  const int getTimesRecognized() const;

  //! Returns the last location object was recognized at
  const Point2D& getLastRecognized() const;

  //! Sets the last location object was recognized at
  void setLastRecognized(Point2D p);

private:
  char* name;                                   // name of this object
  std::vector<Keypoint> keys;                   // keypoints
  SortedLimitedList<Keypoint> sortedKeys;       // keys sorted by magnitude
  int sortOrder;                                // 0=ascending,1=descending
  char* imageFilename;                          // object img filename
  Image<PixRGB<byte> > image;                   // object image
  std::vector<double> feature;                  // object attentive features
  int timesRecognized;                          // number of times recognized
  Point2D lastRecognized;                       // last point recognized at
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
