/*!@file VisualObjectDB.C Visual object database */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: 
//

#include "VisualObjectDB.H"
#include "Image.H"
#include "Image_ColorOps.H"
#include <cstring>
#include <sstream>
#include <unistd.h>

using std::cin;

// #######################################################################
VisualObjectDB::VisualObjectDB()
{}

// #######################################################################
VisualObjectDB::VisualObjectDB(const char* filename, int sortOrder)
{
  FILE *file;
  int len;
    
  file = fopen (filename, "r");
  if (! file)
    LINFO("Could not open file: %s", filename);
  else
    {
      if(fscanf(file, "%i", &len) != 1)
        LINFO("invalid object database file beginning.");
      
      for(int i=0; i<len; i++)
        {
          VisualObject obj(file, sortOrder);
          objects.push_back(obj);
        }
    }
}

// #######################################################################
VisualObjectDB::~VisualObjectDB()
{
  objects.clear();
}

// #######################################################################
const VisualObject& VisualObjectDB::getObject(int index) const
{
  return objects[index];
}

// #######################################################################
const VisualObject& VisualObjectDB::getObject(char* objectName) const
{
  int index = -1;;
  for(uint i=0; i<objects.size(); i++)
    {
      if(strcmp(objects[i].getName(), objectName) == 0)
	{
	  index = i;
          break;
        }
    }
  return objects[index];
}

// #######################################################################
bool VisualObjectDB::addObject(VisualObject& obj)
{
  for(uint i=0; i<objects.size(); i++)
    {
      if(strcmp(obj.getName(),objects[i].getName()) == 0)
	return false;
    }
  objects.push_back(obj);
  return true;
}

// #######################################################################
int VisualObjectDB::getSize() const
{
  return (int)objects.size();
}

// #######################################################################
void VisualObjectDB::save(const char* filename) const
{
  FILE *stream;
  
  /* Open for write */
  if( (stream = fopen( filename, "w+" )) != NULL )
    {
      fprintf(stream, "%i\n", objects.size());
      for( uint i=0; i<objects.size(); i++)
        {
          objects[i].save(stream);
        }
    }
  fclose( stream );
}

// #######################################################################
const VisualObjectMatch VisualObjectDB::matchObject(VisualObject& obj,
                                                    Point2D& loc,
                                                    int recogMinMatch,
                                                    float recogMinMatchPercent,
                                                    char* matchingAlg,
                                                    bool sortKeys, 
                                                    bool sortObjects,
                                                    char* sortBy)
{
  VisualObjectMatch emptyMatch;
  std::vector<int> objDB;
  if(sortObjects)
    objDB = sortedIndices(obj, loc, sortBy);
  
  int numberComparisons = 0;
  // check each object in the db
  for(uint i=0; i<objects.size(); i++)
    {
      int index = i;
      if(sortObjects)
        index = objDB[i];

      // check object for match
      VisualObjectMatch match = objects[index].match(obj, matchingAlg, 
                                                     recogMinMatch,
                                                     recogMinMatchPercent,
                                                     sortKeys);
      match.incrementNumberKeyCompare(numberComparisons);

      if(match.getSuccess())
        {
          objects[index].incTimesRecognized();
          objects[index].setLastRecognized(loc);
          return match;
        }
      numberComparisons = match.getNumberKeyCompare();
    }
  return emptyMatch;
}

// #######################################################################
const vector<int> VisualObjectDB::sortedIndices(const VisualObject& obj,
                                                const Point2D& loc,
                                                char*& sortBy) const
{
  std::vector<int> sortedIndices;

  for(uint i=0; i<this->objects.size(); i++)
    {
      double val;
      if(strcmp(sortBy, "features") == 0)
        val = this->objects[i].getFeatureDistSq(obj);
      else if(strcmp(sortBy, "location") == 0)
        val = this->objects[i].getLastRecognized().distance(loc);
      int index = sortedIndices.size()-1;
      while(index >= 0 && 
            ((strcmp(sortBy, "features") == 0 && 
              val < this->objects[sortedIndices[index]].getFeatureDistSq(obj)) ||
             (strcmp(sortBy, "location") == 0 &&
              val < this->objects[sortedIndices[index]].getLastRecognized().distance(loc))))
        index--;
      sortedIndices.insert(sortedIndices.begin()+index+1, i);
    }

  
  return sortedIndices;
}

// #######################################################################
void VisualObjectDB::train(VisualObject& obj)
{
  std::string objName;
  std::stringstream s;
  s << objName << objects.size();
  // try to add object - will fail if obj name isnt unique
  obj.setName((char*)s.str().c_str());
  obj.setImageFilename(strcat((char*)s.str().c_str(), ".ppm"));
  addObject(obj);
}

// ######################################################################
void VisualObjectDB::filter(int recogThresh)
{
  for(uint i=0; i<objects.size(); i++)
    {
      if(objects[i].getTimesRecognized() < recogThresh)
        {
          objects.erase(objects.begin()+i);
          i--;
        }
    }
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
