/*!@file WTAwinner.H a winner-take-all winner (i.e., covert shift of attention) */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2000-2003   //
// by the University of Southern California (USC) and the iLab at USC.  //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: WTAwinner.H 2960 2003-11-03 20:31:37Z itti $
//

#ifndef WTAWINNER_H_DEFINED
#define WTAWINNER_H_DEFINED

#include "Point2D.H"
#include <cstdlib>

//! This is a an open class representing a WTA winner (covert attention shift)
/*! This class is an open (all members public) container for a
Point2D, time stamp, saliency voltage and possibly other things that
characterize a WTA winner, or covert shift of attention. Like Point2D,
this class is fully inlined, so there is no WTAwinner.C file. */

class WTAwinner
{
public:
  //! Default contructor, initializes to zeros
  inline WTAwinner();

  //! Constructor given a Point2D, time and other stuff
  inline WTAwinner(const Point2D& pp, const double tt,
                   const double svv, const double avv, const bool bor);

  //! Constructor given (i, j), time and other stuff
  inline WTAwinner(const int i, const int j, const double tt, const double svv,
                   const double avv, const bool bor);


  //! Return coordinates scaled down to scale of saliency map
  inline Point2D getSMcoords(const int smlevel) const;

  //! Build from coordinates given at the scale of the saliency map
  inline void buildFromSMcoords(const Point2D smcoords, const int smlevel,
                                const bool useRandom, const double tt,
                                const double svv, const double avv,
                                const bool bor);

  //! returns true if coordinates are not (-1, -1)
  inline bool isValid() const;

  // public data members:
  Point2D p;  //!< The spatial coordinates
  double t;   //!< The time, in seconds
  double sv;  //!< The saliency map voltage at winning location, in Volts
  double av;  //!< The Attention-Guidance-Map voltage at winning loc (Volts)
  bool boring;//!< True is this shift was made out of boredom
};

// ######################################################################
// #################### INLINED METHODS:
// ######################################################################
inline WTAwinner::WTAwinner() :
  p(0, 0), t(0.0), sv(0.0), av(0.0), boring(false)
{ }

// ######################################################################
inline WTAwinner::WTAwinner(const Point2D& pp, const double tt,
                            const double svv, const double avv,
                            const bool bor) :
  p(pp), t(tt), sv(svv), av(avv), boring(bor)
{ }

// ######################################################################
inline WTAwinner::WTAwinner(const int i, const int j,
                            const double tt, const double svv,
                            const double avv, const bool bor) :
  p(i, j), t(tt), sv(svv), av(avv), boring(bor)
{ }

// ######################################################################
inline Point2D WTAwinner::getSMcoords(const int smlevel) const
{
  // the 0.49 offset below is to eliminate possible random jitter
  // introduced by buildFromSMcoords():
  return Point2D(int(p.i / double(1 << smlevel) + 0.49),
                 int(p.j / double(1 << smlevel) + 0.49));
}

// ######################################################################
inline void WTAwinner::buildFromSMcoords(const Point2D smcoords,
                                         const int smlevel,
                                         const bool useRandom,
                                         const double tt,
                                         const double svv,
                                         const double avv,
                                         const bool bor)
{
  // scale coords up: By construction of our pyramids and of decX(),
  // decY() and friends, the coordinates of a saliency-map pixel
  // represent information around the top-left corner of that pixel,
  // not the center of tha pixel. So we simply scale the coordinates
  // up and do not attempt to center them over the extent of the
  // saliency map pixel:
  p.i = smcoords.i << smlevel;
  p.j = smcoords.j << smlevel;

  // add random jitter:
  if (useRandom)
    {
      int jitter = 1 << smlevel;  // saliency pixel size
      p.i += (int)(jitter * (drand48()*0.98 - 0.49));  // up to +/- SMpix*0.49
      p.j += (int)(jitter * (drand48()*0.98 - 0.49));
      if (p.i < 0) p.i = 0;  // make sure we don't become negative...
      if (p.j < 0) p.j = 0;
    }
  
  // initialize other members:
  t = tt; sv = svv; av = avv; boring = bor;
}

// ######################################################################
inline bool WTAwinner::isValid() const
{
  return p.i != -1 && p.j != -1;
}


// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
#endif
