/*!@file WinnerTakeAll.C 2D winner-take-all network */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: WinnerTakeAll.C 4185 2005-02-13 03:12:02Z itti $
//

#include "WinnerTakeAll.H"
#include "Image.H"
#include "log.H"
#include "ModelOptionDefs.H"
#include "ModelManager.H"
#include "FrameSeries.H"
#include "Brain.H"
#include "SaccadeController.H"
#include "Image_Transforms.H" // for chamfer34()

// ######################################################################
// ######################################################################
// ########## WinnerTakeAll implementation
// ######################################################################
// ######################################################################

WinnerTakeAll::WinnerTakeAll(ModelManager& mgr, const std::string& descrName,
                             const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsSaveResults(this, "WTAsaveResults", false), // see ModelOptionDefs.C
  itsWTAFactor(this, "WTAfactor", 50000.0),
  itsUseSaccadicSuppression(this, "WTAuseSacSuppress", true),
  itsUseBlinkSuppression(this, "WTAuseBlinkSuppress", true)
{ }

// ######################################################################
WinnerTakeAll::~WinnerTakeAll()
{ }

// ######################################################################
void WinnerTakeAll::exportOptions(const int whatToExport, const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE) {
    manager->requestOption(this, itsWTAFactor);
    manager->requestOption(this, itsUseSaccadicSuppression);
    manager->requestOption(this, itsUseBlinkSuppression);
  }

  if (whatToExport & OPTEXP_SAVE)
    manager->requestOption(this, itsSaveResults);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(whatToExport, recurse);
}

// ######################################################################
void WinnerTakeAll::saveResults(const SharedPtr<OutputFrameSeries>& ofs)
{
  if (itsSaveResults.getVal())
    ofs->writeFloat(getV(true), PNM, -1, true, "WTA");
}


// ######################################################################
// ######################################################################
// ########## WinnerTakeAllConfigurator implementation
// ######################################################################
// ######################################################################
WinnerTakeAllConfigurator::
WinnerTakeAllConfigurator(ModelManager& mgr,
                          const std::string& descrName,
                          const std::string& tagName) :
  ModelComponent(mgr, descrName, tagName),
  itsWTAtype(this, "WinnerTakeAllType", "None"),
  itsWTA(NULL), itsWhatToExport(OPTEXP_NONE)
{  }

// ######################################################################
WinnerTakeAllConfigurator::~WinnerTakeAllConfigurator()
{  }

// ######################################################################
void WinnerTakeAllConfigurator::exportOptions(const int whatToExport,
                                              const bool recurse)
{
  // keep a copy of whatToExport so that we can let our controller
  // know once we instantiate it:
  itsWhatToExport = whatToExport;

  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsWTAtype);

  // the base class will propagate this call to our subcomponents:
  ModelComponent::exportOptions(recurse);
}

// ######################################################################
SharedPtr<WinnerTakeAll> WinnerTakeAllConfigurator::getWTA() const
{ return itsWTA; }

// ######################################################################
bool WinnerTakeAllConfigurator::setModelParamString(const std::string& name,
                                                    const std::string& value,
                                                    const bool recurse,
                                                    const bool warn)
{
  // accept the parameter change:
  bool ret = ModelComponent::setModelParamString(name, value, recurse, warn);

  // was that a change of our baby's name?
  if (name.compare(itsWTAtype.getName()) == 0) {
    // if we had one, let's unregister it and destroy it; upon
    // destruction, it will unexport its command-line options:
    if (itsWTA.get()) { removeSubComponent(itsWTA); itsWTA.reset(NULL); }

    // instantiate a WTA of the appropriate type:
    if (value.compare("None") == 0)          // no SM
      itsWTA.reset(NULL);
    else if (value.compare("Std") == 0)      // standard
      itsWTA.reset(new WinnerTakeAllStd(*manager));
    else if (value.compare("Fast") == 0)     // fast
      itsWTA.reset(new WinnerTakeAllFast(*manager));
    else if (value.compare("Dummy") == 0)    // dummy
      itsWTA.reset(new WinnerTakeAllDummy(*manager));
    else if (value.compare("Greedy") == 0)    // greedy
      itsWTA.reset(new WinnerTakeAllGreedy(*manager));
    else
      LFATAL("Unknown WTA type %s -- ABORT", value.c_str());

    // add our baby as a subcomponent of us so that it will become
    // linked to the manager through us (hopefully we are registered
    // with the manager), which in turn will allow it to export its
    // command-line options and get configured:
    if (itsWTA.get()) {
      addSubComponent(itsWTA);

      // tell the controller to export its options:
      itsWTA->exportOptions(itsWhatToExport, true);
    }

    // some info message:
    LINFO("Selected WTA of type %s", value.c_str());
  }
  return ret;
}

// ######################################################################
// ######################################################################
// ########## WinnerTakeAllStd implementation
// ######################################################################
// ######################################################################

WinnerTakeAllStd::WinnerTakeAllStd(ModelManager& mgr,
                                   const std::string& descrName,
                                   const std::string& tagName) :
  WinnerTakeAll(mgr, descrName, tagName),
  Image<LeakyIntFire>(), itsInputCopy(),
  itsGIN(), itsT(0.0), itsGleak(1.0e-8F), itsGinh(1.0e-2F), itsGinput(5.0e-8F)
{
  itsGIN.setGleak(itsGleak);
}

// ######################################################################
WinnerTakeAllStd::~WinnerTakeAllStd()
{ }

// ######################################################################
void WinnerTakeAllStd::reset()
{ 
  freeMem(); itsInputCopy.freeMem();
  itsGleak = 1.0e-8F; itsGinh = 1.0e-2F; itsGinput = 5.0e-8F;
  itsT = 0.0;
}

// ######################################################################
void WinnerTakeAllStd::input(const Image<float>& in)
{
  if (initialized() == false) {  // first input, let's initialize our array
    resize(in.getDims(), ZEROS);

    Image<LeakyIntFire>::iterator nptr = beginw(), stop = endw();
    while (nptr != stop)
      {
        nptr->setGleak(itsGleak);
        nptr->setG(0.0F, itsGinh);
        ++nptr;
      }
  }

  // keep a copy of the input for use in evolve():
  itsInputCopy = in;
}

// ######################################################################
int WinnerTakeAllStd::getWidth() const
{ return Image<LeakyIntFire>::getWidth(); }

// ######################################################################
int WinnerTakeAllStd::getHeight() const
{ return Image<LeakyIntFire>::getHeight(); }

// ######################################################################
void WinnerTakeAllStd::evolve(const double t, Point2D& winner)
{
  winner.i = -1; int w = getWidth(), h = getHeight();

  // the array of neurons receive excitatory inputs from outside.
  // here we update the inputs and let the neurons evolve; here we
  // need to run this loop time step by time step, since this is a
  // feedback network, so we have code similar to that in
  // LeakyIntFire::evolve() to figure out how many time steps to run:
  double tstep = itsGIN.getTimeStep();  // time step of our neurons
  double dt, stopt = t - tstep * 0.01;
  if (t - itsT < 1.5 * tstep) dt = t - itsT;       // do it in one step
  else dt = (t - itsT) / int((t - itsT) / tstep);  // several steps

  for (double tt = itsT; tt < stopt; tt += dt)
    { 
      Image<LeakyIntFire>::iterator nptr = beginw();
      Image<float>::const_iterator inptr = itsInputCopy.begin();
      for (int j = 0 ; j < h; j ++)
        for (int i = 0; i < w; i ++)
          {
            nptr->input(itsGinput * (*inptr++));
            if (nptr->integrate(tt)) { winner.i = i; winner.j = j; }
            nptr->setG(0.0F, 0.0F);  // only leak conductance
            ++nptr;
          }

      // if there is a winner, the winner triggers the global inhibition:
      if (winner.i > -1) itsGIN.setG(itsGleak * 10.0F, 0.0F);

      // when the global inhibition fires, it triggers inhibitory
      // conductances for one time step in the array of excited
      // neurons, shuts off excitatory conductances, and turns itself
      // off:
      if (itsGIN.integrate(tt)) inhibit();
    }
  itsT = t;
}

// ######################################################################
Image<float> WinnerTakeAllStd::getV(const bool normalize) const
{
  Image<float> result(getDims(), NO_INIT);

  Image<float>::iterator dptr = result.beginw(), stop = result.endw();
  Image<LeakyIntFire>::const_iterator nptr = begin();
  while(dptr != stop) *dptr++ = (nptr++)->getV();

  if (normalize) {
    if (itsWTAFactor.getVal() > 0.0F) result *= itsWTAFactor.getVal();
    else result.normalize(0.0F, 255.0F);
  }

  return result;
}

// ######################################################################
float WinnerTakeAllStd::getV(const Point2D& p) const
{ return getVal(p).getV(); }

// ######################################################################
void WinnerTakeAllStd::inhibit()
{
  Image<LeakyIntFire>::iterator nptr = beginw(), stop = endw();
  while(nptr != stop) (nptr++)->setG(0.0F, itsGinh);
  itsGIN.setG(0.0F, 0.0F);
  LDEBUG("WTA inhibition firing...");
}

// ######################################################################
void WinnerTakeAllStd::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;
  if (on) inhibit();
  LINFO("------- WTA saccadic suppression %s -------", on ? "on":"off");
}

// ######################################################################
void WinnerTakeAllStd::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;
  if (on) inhibit();
  LINFO("------- WTA blink suppression %s -------", on ? "on":"off");
}

// ######################################################################
// ######################################################################
// ########## WinnerTakeAllFast implementation
// ######################################################################
// ######################################################################

WinnerTakeAllFast::WinnerTakeAllFast(ModelManager& mgr,
                                     const std::string& descrName,
                                     const std::string& tagName) :
  WinnerTakeAll(mgr, descrName, tagName), Image<float>()
{ }

// ######################################################################
WinnerTakeAllFast::~WinnerTakeAllFast()
{ }

// ######################################################################
void WinnerTakeAllFast::reset()
{ freeMem(); }

// ######################################################################
void WinnerTakeAllFast::input(const Image<float>& in)
{ Image<float> in2(in); swap(in2); }

// ######################################################################
int WinnerTakeAllFast::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int WinnerTakeAllFast::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
void WinnerTakeAllFast::evolve(const double t, Point2D& winner)
{
  float maxval;
  this->findMax(winner, maxval);
  if (maxval <= 1.0e-20F) { winner.i = -1; winner.j = -1; }
}

// ######################################################################
Image<float> WinnerTakeAllFast::getV(const bool normalize) const
{
  Image<float> ret(*this);
  if (normalize) {
    if (itsWTAFactor.getVal() > 0.0F) ret *= itsWTAFactor.getVal();
    else ret.normalize(0.0F, 255.0F);
  }

  return ret;
}

// ######################################################################
float WinnerTakeAllFast::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
void WinnerTakeAllFast::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;
  if (on) clear();
  LINFO("------- WTA saccadic suppression %s -------", on ? "on":"off");
}

// ######################################################################
void WinnerTakeAllFast::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;
  if (on) clear();
  LINFO("------- WTA blink suppression %s -------", on ? "on":"off");
}

// ######################################################################
// ######################################################################
// ########## WinnerTakeAllDummy implementation
// ######################################################################
// ######################################################################

WinnerTakeAllDummy::WinnerTakeAllDummy(ModelManager& mgr,
                                     const std::string& descrName,
                                     const std::string& tagName) :
  WinnerTakeAll(mgr, descrName, tagName), Image<float>()
{ }

// ######################################################################
WinnerTakeAllDummy::~WinnerTakeAllDummy()
{ }

// ######################################################################
void WinnerTakeAllDummy::reset()
{ freeMem(); }

// ######################################################################
void WinnerTakeAllDummy::input(const Image<float>& in)
{ Image<float> in2(in); swap(in2); }

// ######################################################################
int WinnerTakeAllDummy::getWidth() const
{ return Image<float>::getWidth(); }

// ######################################################################
int WinnerTakeAllDummy::getHeight() const
{ return Image<float>::getHeight(); }

// ######################################################################
void WinnerTakeAllDummy::evolve(const double t, Point2D& winner)
{ winner.i = -1; winner.j = -1; }

// ######################################################################
Image<float> WinnerTakeAllDummy::getV(const bool normalize) const
{
  Image<float> ret(*this);
  if (normalize) {
    if (itsWTAFactor.getVal() > 0.0F) ret *= itsWTAFactor.getVal();
    else ret.normalize(0.0F, 255.0F);
  }

  return ret;
}

// ######################################################################
float WinnerTakeAllDummy::getV(const Point2D& p) const
{ return getVal(p); }

// ######################################################################
void WinnerTakeAllDummy::saccadicSuppression(const bool on)
{
  if (itsUseSaccadicSuppression.getVal() == false) return;
  if (on) clear();
  LINFO("------- WTA saccadic suppression %s -------", on ? "on":"off");
}

// ######################################################################
void WinnerTakeAllDummy::blinkSuppression(const bool on)
{
  if (itsUseBlinkSuppression.getVal() == false) return;
  if (on) clear();
  LINFO("------- WTA blink suppression %s -------", on ? "on":"off");
}

// ######################################################################
// ######################################################################
// ########## WinnerTakeAllGreedy implementation
// ######################################################################
// ######################################################################

WinnerTakeAllGreedy::WinnerTakeAllGreedy(ModelManager& mgr,
                                     const std::string& descrName,
                                     const std::string& tagName) :
  WinnerTakeAllStd(mgr, descrName, tagName),
  itsThresholdFac(this, "WinnerTakeAllGreedyThreshFac", 0.5f),
  itsSC(NULL), itsSMlevel(-1)
{ }

// ######################################################################
WinnerTakeAllGreedy::~WinnerTakeAllGreedy()
{ }

// ######################################################################
void WinnerTakeAllGreedy::exportOptions(const int whatToExport,
                                        const bool recurse)
{
  // let's first do our exports:
  if (whatToExport & OPTEXP_CORE)
    manager->requestOption(this, itsThresholdFac);

  // the base class will propagate this call to our subcomponents:
  WinnerTakeAll::exportOptions(whatToExport, recurse);
}

// ######################################################################
void WinnerTakeAllGreedy::start2()
{
  // do our base start:
  WinnerTakeAllStd::start2();

  // now our own inits:
  if (manager->hasSubComponent("Brain"))
    {
      SharedPtr<Brain> brain;
      brain.dynCastFrom(manager->subComponent("Brain"));
      itsSC = brain->getSC();
      if (itsSC.get() == NULL)
        CLFATAL("I need a Brain that has an SC -- ABORT");
      LevelSpec ls; brain->getModelParamVal("LevelSpec", ls, true, true);
      itsSMlevel = ls.mapLevel();
    }
  else
    CLFATAL("I need a Brain -- ABORT");
}

// ######################################################################
void WinnerTakeAllGreedy::evolve(const double t, Point2D& winner)
{
  winner.i = -1; int w = getWidth(), h = getHeight();

  // the array of neurons receive excitatory inputs from outside.
  // here we update the inputs and let the neurons evolve; here we
  // need to run this loop time step by time step, since this is a
  // feedback network, so we have code similar to that in
  // LeakyIntFire::evolve() to figure out how many time steps to run:
  double tstep = itsGIN.getTimeStep();  // time step of our neurons
  double dt, stopt = t - tstep * 0.01;
  if (t - itsT < 1.5 * tstep) dt = t - itsT;       // do it in one step
  else dt = (t - itsT) / int((t - itsT) / tstep);  // several steps

  for (double tt = itsT; tt < stopt; tt += dt)
    { 
      Image<LeakyIntFire>::iterator nptr = beginw();
      Image<float>::const_iterator inptr = itsInputCopy.begin();
      for (int j = 0 ; j < h; j ++)
        for (int i = 0; i < w; i ++)
          {
            nptr->input(itsGinput * (*inptr++));
            if (nptr->integrate(tt)) { winner.i = i; winner.j = j; }
            nptr->setG(0.0F, 0.0F);  // only leak conductance
            ++nptr;
          }

      // if there is a winner, the winner triggers the global
      // inhibition; but before that let's replace our true WTA winner
      // by whoever is above threshold and closest to current eye
      // position:
      if (winner.i > -1)
        {
          // get our current voltages:
          Image<float> v = getV();

          // do not consider values below threshold:
          float thresh = v.getVal(winner) * itsThresholdFac.getVal();

          // get current eye position:
          Point2D eye = itsSC->getPreviousDecision().p;
          eye.i = int(eye.i / double(1 << itsSMlevel) + 0.49);
          eye.j = int(eye.j / double(1 << itsSMlevel) + 0.49);

          // if eye out of image (or unknown), we will not do our
          // special greedy processing here and instead will just act
          // like a normal WTA:
          if (v.coordsOk(eye) && (eye.i != 0 || eye.j != 0))
            {
              // create distance map from a single pixel at our eye position:
              Image<float> dmap(v.getDims(), ZEROS);
              dmap.setVal(eye, 100.0f);
              dmap = chamfer34(dmap, 1000.0f);

              // let's scan the dmap and find the closest location
              // that is a local max and is above threshold:
              const int w = v.getWidth(), h = v.getHeight();
              float bestd = 1000.0f; Point2D p;

              for (p.j = 0; p.j < h; p.j ++)
                for (p.i = 0; p.i < w; p.i ++)
                  if (v.getVal(p) >= thresh &&  // we are above threshold
                      dmap.getVal(p) < bestd && // we are closer to eye
                      v.isLocalMax(p) == true)  // we are a local max
                    {
                      winner.i = p.i;
                      winner.j = p.j;
                      bestd = dmap.getVal(p);
                    }
            }

          // trigger global inhibition:
          itsGIN.setG(itsGleak * 10.0F, 0.0F);
        }

      // when the global inhibition fires, it triggers inhibitory
      // conductances for one time step in the array of excited
      // neurons, shuts off excitatory conductances, and turns itself
      // off:
      if (itsGIN.integrate(tt)) inhibit();
    }
  itsT = t;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
