/*!@file WinnerTakeAll.H Header file for winner-take-all network */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: WinnerTakeAll.H 4179 2005-02-12 22:00:07Z itti $
//

#ifndef WINNERTAKEALL_H_DEFINED
#define WINNERTAKEALL_H_DEFINED

#include "Image.H"
#include "LeakyIntFire.H"
#include "ModelComponent.H"
class ModelManager;
class OutputFrameSeries;
class SaccadeController;

// ######################################################################
//! The winner-take-all base class
// ######################################################################
/*! This is a 2D winner-take-all, a.k.a., spatial maximum detector. It
  is just a base class with virtual function definitions. Various
  winner-take-all variants are available below, which provide
  implementations for these virtual functions. The WTA receives input
  through the input() function, and its current state (membrane
  potential of neurons) can be retrieved using getV(). Various
  inplementations will typically derive from class Image and inherit
  all the standard methods of class Image.  The user has the
  responsibility of calling evolve() often to ensure that possible
  difference equations in the WTA's neurons run well. */
class WinnerTakeAll : public ModelComponent
{
 public:
  // ######################################################################
  //! @name Constructor, destructor, and reset
  //@{

  //! Ininitialized constructor
  /*! The SM will be resized and initialized the first time input() is
    called */
  WinnerTakeAll(ModelManager& mgr,
                const std::string& descrName = "Winner-Take-All",
                const std::string& tagName = "WinnerTakeAll");
  
  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. A base
    implementation is provided that will only export the base class
    options (see the ModelParam definitions in the protected section
    below). */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);
  //! Destructor
  virtual ~WinnerTakeAll();

  //! Reset to initial state just after construction
  virtual void reset() = 0;

  //@}

  // ######################################################################
  //! @name Input, evolve and output
  //@{

  //! Set new input currents for all neurons
  /*! This should initialize and resize the network if the network is
    currently uninitialized (e.g., just after construction or reset()). */
  virtual void input(const Image<float>& in) = 0;

  //! Get width of the network
  virtual int getWidth() const = 0;

  //! Get height of the network
  virtual int getHeight() const = 0;

  //! Return all our membrane potential voltages as an Image<float>
  /*! The curent values are possibly normalized to 0..255. The
    normalization will use itsSalmapFactor if it is non-null */
  virtual Image<float> getV(const bool normalize = false) const = 0;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const = 0;

  //! Integrate inputs until time t (in s) and update membrane potentials
  /*! If new winner was selected, its coordinates are put in winner,
    otherwise, winner = (-1, -1). */
  virtual void evolve(const double t, Point2D& winner) = 0;

  //@}

  // ######################################################################
  //! @name Saccadic and blink suppression
  //@{

  //! Turn saccadic suppression on/off
  virtual void saccadicSuppression(const bool on) = 0;

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on) = 0;

  //@}

  // ######################################################################
  //! @name Results
  //@{

  //! Save our various results
  /*! A default implementation is provided, which just calls
      getV() and saves that map with a file name prefix of "WTA".
      @param ofs will figure out file name/format and save our
      results. */
  virtual void saveResults(const SharedPtr<OutputFrameSeries>& ofs);

  //@}

  // ######################################################################
protected:
  //! Save our internals when saveResults() is called?
  ModelParam<bool> itsSaveResults;

  //! Multiply WTA by a factor rather than normalize when doing getV()?
  ModelParam<float> itsWTAFactor;

  //! use saccadic suppression?
  ModelParam<bool> itsUseSaccadicSuppression;

  //! use blink suppression?
  ModelParam<bool> itsUseBlinkSuppression;

private:
  // forbid assignment and copy-construction:
  WinnerTakeAll& operator=(const WinnerTakeAll& wta);
  WinnerTakeAll(const WinnerTakeAll& wta);
};

// ######################################################################
//! WinnerTakeAll configurator
// ######################################################################
/*! This will export the --wta-type=XX command-line option and will
  instantiate a WTA of the desired type as the option gets assigned a
  value. As this happens, new options may become available in the
  command-line. To see them, use --help AFTER you have chosen the type
  to use. The current WTA may be retrieved using getWTA(). */
class WinnerTakeAllConfigurator : public ModelComponent
{
public:
  //! Constructor
  WinnerTakeAllConfigurator(ModelManager& mgr,
                            const std::string& descrName =
                            "Winner-Take-All Configurator",
                            const std::string& tagName =
                            "WinnerTakeAllConfigurator");

  //! destructor
  virtual ~WinnerTakeAllConfigurator();

  //! Export some of our ModelParam parameters as command-line-options
  /*! See the base function in ModelComponent.H for info. Actually,
    here we just keep a copy of whatToExport and we will call
    exportOptions() on our WTA once we know its type and
    instantiate it. */
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Get the chosen WTA
  /*! You should call this during start() of the ModelComponent that
    needs the WTA. CAUTION: this may be a SharedPtr(NULL) if no
    WTA type has been chosen. */
  SharedPtr<WinnerTakeAll> getWTA() const;

protected:
  ModelParam<std::string> itsWTAtype; //!< type of winner-take-all

  //! Intercept people changing our ModelParam
  /*! See ModelComponent.H; as parsing the command-line or reading a
    config file sets our name, we'll also here instantiate a
    controller of the proper type (and export its options) */
  virtual bool setModelParamString(const std::string& name,
                                   const std::string& value,
                                   const bool recurse = false,
                                   const bool warn = true);

private:
  SharedPtr<WinnerTakeAll> itsWTA; // the winner-take-all
  int itsWhatToExport;
};


// ######################################################################
//! The standard winner-take-all
// ######################################################################
/*! This is a trivial winner-take-all implementation, based on a 2D
  layer of LeakyIntFire neurons and a unique global inhibitory
  neuron. All neurons in the layer charge up in parallel; whenever one
  reaches threshold, it activates the inhibitory interneuron (which is
  connected to all neurons in the layer), which in turn resets all
  neurons in the layer.  WinnerTakeAllStd is an Image<LeakyIntFire>
  and inherits all the standard methods of class Image. To avoid
  confusion, we here add explicit input() and getV() methods (rather
  than providing conversion functions between LeakyIntFire and float,
  which could make the process more transparent but also possibly more
  confusing). The user has the responsibility of calling evolve()
  often to ensure that the difference equations in the LeakyIntegrator
  neurons run well. */
class WinnerTakeAllStd : public WinnerTakeAll, public Image<LeakyIntFire>
{
 public:
  //! Ininitialized constructor
  /*! The WTA will be resized and initialized the first time input() is
    called */
  WinnerTakeAllStd(ModelManager& mgr,
                   const std::string& descrName = "Winner-Take-All Std",
                   const std::string& tagName = "WinnerTakeAllStd");
  
  //! Destructor
  virtual ~WinnerTakeAllStd();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  /*! This will initialize and resize the network if the network is
    currently uninitialized (e.g., just after construction or reset()). */
  virtual void input(const Image<float>& in);

  //! Get width of the network
  virtual int getWidth() const;

  //! Get height of the network
  virtual int getHeight() const;

  //! Return all our membrane potential voltages as an Image<float>
  /*! The curent values are possibly normalized to 0..255. The
    normalization will use itsSalmapFactor if it is non-null */
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update membrane potentials
  virtual void evolve(const double t, Point2D& winner);

  //! Turn saccadic suppression on/off
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);

protected:  // (not private to allow WinnerTakeAllGreedy to mess with our guts)
  Image<float> itsInputCopy;
  LeakyIntFire itsGIN;               // global inhibition neuron
  double itsT;                       // keep track of time
  float itsGleak, itsGinh;           // in Siemens
  float itsGinput;                   // in Siemens
  void inhibit();                    // inhibit the whole layer
};

// ######################################################################
//! A fast winner-take-all
// ######################################################################
/*! This just does an Image<T>::findMax() on the last input array
  received. It hence returns a winner each time evolve() is called. */
class WinnerTakeAllFast : public WinnerTakeAll, public Image<float>
{
 public:
  //! Ininitialized constructor
  /*! The WTA will be resized and initialized the first time input() is
    called */
  WinnerTakeAllFast(ModelManager& mgr,
                    const std::string& descrName = "Winner-Take-All Fast",
                    const std::string& tagName = "WinnerTakeAllFast");
  
  //! Destructor
  virtual ~WinnerTakeAllFast();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  /*! This will initialize and resize the network if the network is
    currently uninitialized (e.g., just after construction or reset()). */
  virtual void input(const Image<float>& in);

  //! Get width of the network
  virtual int getWidth() const;

  //! Get height of the network
  virtual int getHeight() const;

  //! Return all our membrane potential voltages as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update membrane potentials
  virtual void evolve(const double t, Point2D& winner);

  //! Turn saccadic suppression on/off
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);
};

// ######################################################################
//! A dummy winner-take-all
// ######################################################################
/*! This will never win and just ignores its inputs. Useful to avoid
  wasting time evolving a more complex WTA in cases where we don't
  care about the model attending to anywhere (e.g., when comparing
  saliency maps to human eye movements). */
class WinnerTakeAllDummy : public WinnerTakeAll, public Image<float>
{
 public:
  //! Ininitialized constructor
  WinnerTakeAllDummy(ModelManager& mgr,
                    const std::string& descrName = "Winner-Take-All Dummy",
                    const std::string& tagName = "WinnerTakeAllDummy");
  
  //! Destructor
  virtual ~WinnerTakeAllDummy();

  //! Reset to initial state just after construction
  virtual void reset();

  //! Set new input currents for all neurons
  /*! This will initialize and resize the network if the network is
    currently uninitialized (e.g., just after construction or reset()). */
  virtual void input(const Image<float>& in);

  //! Get width of the network
  virtual int getWidth() const;

  //! Get height of the network
  virtual int getHeight() const;

  //! Return all our membrane potential voltages as an Image<float>
  virtual Image<float> getV(const bool normalize = false) const;

  //! Access function using Point2D
  virtual float getV(const Point2D& p) const;

  //! Integrate inputs until time t (in s) and update membrane potentials
  virtual void evolve(const double t, Point2D& winner);

  //! Turn saccadic suppression on/off
  virtual void saccadicSuppression(const bool on);

  //! Turn blink suppression on/off
  virtual void blinkSuppression(const bool on);
};

// ######################################################################
//! A greedy attention selection mechanism, not exactly winner-take-all
// ######################################################################
/*! This will select, among a set of possible candidate targets that
are above a threshold, the one that is closest to current eye
position. It hence relies on being able to find a Brain somewhere in
the model, and on querying its SaccadeController for current eye
position. Note that the returned winner could be exactly at our
current eye position if that is above threshold; we here do not
attempt to force it to be some distance away from current fixation,
and instead rely on IOR in the saliency map to enforce that if
desired. */
class WinnerTakeAllGreedy : public WinnerTakeAllStd
{
 public:
  //! Ininitialized constructor
  WinnerTakeAllGreedy(ModelManager& mgr,
                      const std::string& descrName = "Winner-Take-All Greedy",
                      const std::string& tagName = "WinnerTakeAllGreedy");
  
  //! Destructor
  virtual ~WinnerTakeAllGreedy();

  //! Export some of our ModelParam parameters as command-line-options
  virtual void exportOptions(const int whatToExport,
                             const bool recurse = true);

  //! Integrate inputs until time t (in s) and update membrane potentials
  /*! We operate like a WinnerTakeAllStd, except that when a winner is
    found, we look around for locations that are above threshold, and
    pick the closest one to our current eye position. */
  virtual void evolve(const double t, Point2D& winner);

protected:
  ModelParam<float> itsThresholdFac; //!< Threshold factor, in [0.0 .. 1.0]

  virtual void start2(); //! get started

private:
  SharedPtr<SaccadeController> itsSC; // pointer to our SC
  int itsSMlevel;  // level of saliency map
};

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
