/*!@file XWinManaged.C A window class with active window management */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: XWinManaged.C 3291 2004-05-19 18:12:45Z dhavale $
//

#include "XWinManaged.H"

#include "log.H"
#include "Pixels.H"
#include "Image.H"

#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <X11/Xutil.h>
#include <vector>

#define SLEEPTIME 100000

// ######################################################################
void XWinManaged::initHandling(void)
{
  XSetWindowBackground(display, window, WhitePixel(display,DefaultScreen(display)));

  XSelectInput(display, window, ExposureMask | ResizeRedirectMask | ButtonPressMask |
               KeyPressMask );

  deleteAtom = XInternAtom(display, "WM_DELETE_WINDOW", false);
  if (0 == XSetWMProtocols(display, window, &deleteAtom,1))
    LFATAL("Error setting the WM protocol");

  if (0 != pthread_create (&eventLoopThread, NULL, 
			   (thread_function_t)&event_loop_, this))
    LFATAL("Cannot create thread");  
}

// ######################################################################
XWinManaged::XWinManaged(const int w, const int h, const int x, const int y,
			 const char *title)
  : XWindow (w, h, x, y, title),
    windowPresent(true), toBeClosed(false)
{
  initHandling();
}


// ######################################################################
XWinManaged::XWinManaged(const Image< PixRGB<byte> >& img, const char *title)
  : XWindow(img.getWidth(), img.getHeight(), -1, -1, title),
    windowPresent(true), toBeClosed(false)
{
  initHandling();
  XWindow::drawImage(img);
}


// ######################################################################
XWinManaged::XWinManaged(const Image<byte>& img, const char *title)
  : XWindow(img.getWidth(), img.getHeight(), -1, -1, title),
    windowPresent(true), toBeClosed(false)
{
  initHandling();
  XWindow::drawImage(img);
}

// ######################################################################
XWinManaged::XWinManaged(const Image<float>& img, const char *title,
			 bool normalize)
  : XWindow(img.getWidth(), img.getHeight(), -1, -1, title),
    windowPresent(true), toBeClosed(false)
{
  initHandling();
  drawImage(img,0,0,normalize);
}


// ######################################################################
XWinManaged::~XWinManaged()
{
  // tell event loop thread to finish and wait for it to return
  windowPresent = false;
  if (0 != pthread_join(eventLoopThread, NULL))
    LFATAL("Thread is not returning");
}

// ######################################################################
void XWinManaged::drawImage(const Image<float>& img, 
			    const int x, const int y,
			    bool normalize)
{
  Image<float> image(img);

  if (normalize)
    image.normalize(0.0,255.0);
  else
    image.clamp(0.0,255.0);

  XWindow::drawImage((Image<byte>)image, x, y);
}

// ######################################################################
void XWinManaged::drawImage(const Image< PixRGB<byte> >& img, 
			    const int x, const int y)
{
  XWindow::drawImage(img, x, y);
}

// ######################################################################
void XWinManaged::drawImage(const Image<byte>& img, 
			    const int x, const int y)
{
  XWindow::drawImage(img, x, y);
}

// ######################################################################
void XWinManaged::event_loop (void)
{
  XEvent event;
  while (windowPresent)
    {
      XFlush(display);

      int num_events = XPending(display);
      
      while((num_events != 0))
	{
          num_events--;
          XNextEvent(display, &event);
          event_handler(&event);
	}
      
      // sleep before looking for the next event
      usleep (SLEEPTIME);
    }
  pthread_exit(NULL);
}

// ######################################################################
void XWinManaged::event_handler(XEvent *event)
{
  switch(event->type)
    {

      // user wants to close window - set the flag
    case ClientMessage:
      {
        if (event->xclient.data.l[0] == (int)deleteAtom)
          toBeClosed = true;
        
        break;
      }
      
      // redraw the image after Expose event
    case Expose:
	 {
           LINFO("Expose ...");
	   // draw image
	   while (XCheckMaskEvent(display, ExposureMask, event));
	   redraw();
	   break;
	 }

	 // handle resize request
    case ResizeRequest:
	 {
           LINFO("Resize...");
	   while (XCheckMaskEvent(display, ResizeRedirectMask, event));
	   XResizeWindow(display, window, iw, ih);
	   redraw();
	   break;
	 }
         
    case KeyPress:
      {
        lastKey.push_back(event->xbutton.button);
        if(lastKey.size()>queueLen){
          lastKey.pop_front();
        }
      break;
      }
      
    case ButtonPressMask:
      {
        lastPoint.push_back(Point2D(event->xbutton.x, event->xbutton.y));
        if(lastPoint.size() > queueLen) {
          lastPoint.pop_front();
        }
        break;
      }
      
    default:
      break;
    }
  // end switch (event->type)
}

// ######################################################################
void XWinManaged::redraw()
{
  if (image)
    {
	 if (shm_buffer)
	   XShmPutImage(display,window,gc,image,0,0,ix,iy,iw,ih,0);
	 else
	   XPutImage(display,window,gc,image,0,0,ix,iy,iw,ih);
    }
  else
    {
	 XClearWindow(display,window);
    }
  XFlush(display);
}

// ######################################################################
int XWinManaged::getLastKeyPress()
{
  int ch = -1;
  if(lastKey.size()){
    ch = lastKey.front();
    lastKey.pop_front();
  }
  redraw();
  usleep(SLEEPTIME);
  return ch;
}

// ######################################################################
Point2D XWinManaged::getLastMouseClick()
{
  Point2D p(-1, -1);

  if(lastPoint.size()){
    p = lastPoint.front();
    lastPoint.pop_front();
  }
  redraw();
  usleep(SLEEPTIME);
  return p;
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
