/*!@file XWinManaged.H A window class with active window management */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: XWinManaged.H 3291 2004-05-19 18:12:45Z dhavale $
//

#ifndef XWINMANAGED_H_DEFINED
#define XWINMANAGED_H_DEFINED

#include "XWindow.H"

#include <pthread.h>
#include <X11/Xlib.h>
#include <vector>
#include <algorithm>
#include <unistd.h>
#include <deque>

typedef void *(*thread_function_t)(void *);

template <class T> class PixRGB;
template <class T> class Image;

//! A window class with active window management
/*! This class is an extension of the XWindow class to allow event handling.
  In the current implementation, <I>exposure</I> and <I>resize</I> events
  are being handled. Instead of destroying the window, clicking the 
  <I>close</I> button will set a flag that can be accessed by calling
  pressedCloseButton().
*/
class XWinManaged : public XWindow
{
public:

  //! Constructor: create, display and map the window, start the event handling.
  /*! The call of to the constructor is identical to the parent class XWindow.
    In addition to creating the window, the event handling is being started.
    @param w Width of usable image area.
    @param h height of usable image area.
    @param x Horizontal position hint for window placement; use -1
    for automatic placement.
    @param y Vertical position hint for window placement; use -1
    for automatic placement.
    @param title String that will appear in the window title. */
  XWinManaged(const int w = 320, const int h = 240,
		    const int x = -1, const int y = -1,
		    const char *title = "USC Neuromorphic Vision");


  //! Constructor: create a window as large as img and display img
  XWinManaged(const Image< PixRGB<byte> >& img, 
	      const char *title = "USC Neuromorphic Vision");


  //! Constructor: create a window as large as img and display img
  XWinManaged(const Image<byte>& img,
	      const char *title = "USC Neuromorphic Vision");

  //! Constructor: create a window as large as img and display img
  XWinManaged(const Image<float>& img,
	      const char *title = "USC Neuromorphic Vision",
	      bool normalize = true);

  //! Destructor: destroy the window and stop the event handling
  virtual ~XWinManaged();

  //! Draw an image into the XWindow
  /*! @param img The image to draw. Image size must match the usable
    image area specified at construction of the XWindow.
    @param img is the image to display
    @param x if the x coordinate where the image should start in the window.
    @param y if the y coordinate where the image should start in the window.
  */
  void drawImage(const Image<float>& img, 
		 const int x = 0, const int y = 0,
		 bool normalize = true);
  //! Draw an image into the XWindow
  /*! @param img The image to draw. Image size must match the usable
    image area specified at construction of the XWindow.
    @param img is the image to display
    @param x if the x coordinate where the image should start in the window.
    @param y if the y coordinate where the image should start in the window.
  */
  void drawImage(const Image< PixRGB<byte> >& img, const int x = 0,
                 const int y = 0);

  //! Draw an image into the XWindow
  /*! @param img The image to draw. Image size must match the usable
    image area specified at construction of the XWindow.
    @param img is the image to display
    @param x if the x coordinate where the image should start in the window.
    @param y if the y coordinate where the image should start in the window.
  */
  void drawImage(const Image<byte>& img, const int x = 0,
                 const int y = 0);

  //! returns whether the user has clicked the <I>close</I> button.
  inline bool pressedCloseButton();
  

  //! returns the last mouse click at the front of the click buffer 
  //! does not distinguish between right and left buttons
  //! if no mouse clicks have occured since the last call
  //! returns (-1,-1)
  Point2D getLastMouseClick();

  //! returns the key press at the front
  //! of the keyboard buffer
  //! if no key press has occured since the last call
  //! returns -1
  int getLastKeyPress();
  
protected:

  //! initialize the event handling
  void initHandling(void);

  //! loop that periodically checks for events
  void event_loop(void);
  inline static void XWinManaged::event_loop_ (XWinManaged *thisobj)
  { thisobj->event_loop(); }

  //! deal with the events coming in - feel free to override
  virtual void event_handler (XEvent *event);

  //! redraw the image - may be overridden
  virtual void redraw();

  Atom deleteAtom;           // XServer atom for window closing action
  pthread_t eventLoopThread; // thread for the event handler loop
  bool windowPresent;        // flag whether event handling whould keep running
  bool toBeClosed;           // flag whether the user hashit the close button
  static const unsigned int queueLen=5;
  std::deque<int> lastKey;
  std::deque<Point2D> lastPoint;
};

// ######################################################################
// inlined methods

// ######################################################################
inline bool XWinManaged::pressedCloseButton()
{ return toBeClosed; }


// ######################################################################
// very useful listener for any close button action

//! watch for close button actions with all registered XWinManaged windows
class CloseButtonListener
{
public:

  //! add a window to the list by pointer
  inline void add(XWinManaged* win);

  //! add a window to the list by reference
  inline void add(XWinManaged& win);

  //! remove a window from the list by pointer
  inline void remove(XWinManaged* win);

  //! remove a window from the list by reference
  inline void remove(XWinManaged& win);

  //! has a close button been pressed at any of the windows?
  /*! This function will seg fault if the list contains pointers to
    windows that don't exist anymore. Make sure that you remove windows 
    from the CloseButtonListener before you explicitely delete them!*/
  inline bool pressedAnyCloseButton(void);

  //! wait for a close button to be pressed
  /*! run in an idle loop until a close button is pressed */
  inline void waitForClose();

private:
  std::vector<XWinManaged*> windows;
};

// ######################################################################
// inlined methods

// ######################################################################
inline void CloseButtonListener::add(XWinManaged* win)
{ windows.push_back(win); }

inline void CloseButtonListener::add(XWinManaged& win)
{ add(&win); }

// ######################################################################
inline void CloseButtonListener::remove(XWinManaged* win)
{
  std::vector<XWinManaged*>::iterator it = find(windows.begin(),windows.end(),win);
  if (it != windows.end())
    windows.erase(it);
}

inline void CloseButtonListener::remove(XWinManaged& win)
{ remove(&win); }

// ######################################################################
inline bool CloseButtonListener::pressedAnyCloseButton(void)
{
  for (uint i = 0; i < windows.size(); i++)
    if (windows[i]->pressedCloseButton()) return true;
  return false;
}

// ######################################################################
inline void CloseButtonListener::waitForClose()
{
  while (!(pressedAnyCloseButton())) usleep(30000);
}

#endif

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
