/*!@file XWindow.C Implementation for a simple window class */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: XWindow.C 2819 2003-09-04 05:45:48Z zhanshi $
//

#include "XWindow.H"

#include "log.H"
#include "Pixels.H"
#include "Image.H"

#include <stdlib.h>
#include <sys/shm.h>
#include <X11/Xutil.h>

namespace
{
  int fourByteAlign(int v)
  {
    const int remainder = v % 4;
    return (remainder == 0) ? v : v + (4-remainder);
  }
}

// ######################################################################
XWindow::XWindow(const int w, const int h, const int x, const int y,
                 const char *title)
{
  width = w; height = h; image = 0; iw = 0; ih = 0;

  // connect to X server:
  display = XOpenDisplay(NULL);
  if (display == NULL) LFATAL("Cannot connect to X server");

  // get default screen:
  screen = DefaultScreen(display);
  visual = DefaultVisual(display, screen);
  depth = DefaultDepth(display, screen);
  LDEBUG("Your screen depth is %d bpp", depth);
  byt_per_pix = 4;
  if (depth == 16) byt_per_pix = 2;

  // check if we are using automatic window placement:
  int x0 = x;
  int y0 = y;
  if( x0 == -1 && y0 == -1 )
    {
      x0 = 0;
      y0 = 0;
    }

  // open window:
  window = XCreateSimpleWindow(display,
                               DefaultRootWindow(display),
                               x0, y0, width, height, 2,
                               BlackPixel(display, screen),
                               WhitePixel(display, screen));
  if (window <= 0) LFATAL("Cannot open window");

  // set title:
  setTitle(title);

  /*
  // set title:
  XTextProperty wintitle;
  if (XStringListToTextProperty((char **)&title, 1, &wintitle) == 0)
    LFATAL("Error formatting window title");
  XSetWMName(display, window, &wintitle);
  //XSetWMIconName(display, win, &icon_name_property);
  */

  // pop this window up on the screen:
  XMapRaised(display, window);

  // set position of window if not using auto-placement:
  if( x != -1 || y != -1 )
    {
      setPosition( x, y );
    }

  // set events that we are interested in:
  XSelectInput(display, window, NoEventMask); // ExposureMask|ButtonPressMask

  // flush X request queue to server:
  XFlush(display);

  // alloc XImage in shared memory if posible (for faster display):
  int Shm_OK = XShmQueryExtension(display);

 // check if remote display, and disable shared memory if so:
  char *disp = getenv("DISPLAY");
  if (disp != NULL && disp[0] != ':')
    { Shm_OK = 0; LDEBUG("Not using shared memory for remote display"); }

  // Get a padded row width that is 4-byte aligned
  const int padded_width = fourByteAlign(width);

  // allocate an image buffer of the size of the window:
  if (Shm_OK)
    {
      shminfo.shmid = shmget(IPC_PRIVATE, padded_width*height*byt_per_pix,
                             IPC_CREAT | 0777);
      buffer = (char *)shmat(shminfo.shmid, NULL, 0);  // attach shared memory
      if (buffer == 0) LFATAL("Cannot get shared memory");
      shminfo.shmaddr = buffer;        // link buffer to shminfo structure
      shminfo.readOnly = False;
      XShmAttach(display, &shminfo);   // now attach to X
      shm_buffer = true;
    }
  else
    {
      buffer = (char *)malloc(padded_width * height * byt_per_pix);
      LDEBUG("buffer: %p", buffer);
      if (buffer == NULL) LFATAL("Cannot allocate image buffer");
      shm_buffer = false;
    }

  // create graphics context for later:
  gc = XCreateGC(display, window, 0, NULL);
}

// ######################################################################
void XWindow::setTitle(const char * title)
{
  // set title:
  XTextProperty wintitle;
  if (XStringListToTextProperty((char **)&title, 1, &wintitle) == 0)
    LFATAL("Error formatting window title");
  XSetWMName(display, window, &wintitle);
  //XSetWMIconName(display, win, &icon_name_property);
}

// ######################################################################
void XWindow::setPosition( const int x, const int y )
{
  XMoveWindow( display, window, x, y );
}

// ######################################################################
Point2D XWindow::getPosition( void )
{
  XWindowAttributes xwa;
  XGetWindowAttributes( display, window, &xwa );
  return( Point2D( xwa.x, xwa.y ) );
}

// ######################################################################
void XWindow::drawImage(const Image< PixRGB<byte> >& img, const int x,
                        const int y)
{
  ASSERT(x >= 0 && y >= 0 && x < width && y < height);
  int w = img.getWidth(); int h = img.getHeight();
  ASSERT(x + w <= width && y + h <= height);
  if (w != iw || h != ih) createImage(w, h);

  const byte* im = reinterpret_cast<const byte*>(img.getArrayPtr());
  byte* bu = reinterpret_cast<byte*>(buffer);

  if (byt_per_pix == 2)
    {
      const int bytes_per_row = fourByteAlign(byt_per_pix * w);

      // 16 bit format: 565, lowest byte first
      for (int j = 0; j < h; ++j)
        {
          byte* bu_row = bu + j*bytes_per_row;
          for (int i = 0; i < w; ++i)
            {
              *bu_row++ = ((im[1] & 0x1C)<<3) | ((im[2] & 0xF8)>>3);
              *bu_row++ = ((im[1] & 0xE0)>>5) | (im[0] & 0xF8);
              im += 3;
            }
        }
    }
  else
    {
      // 24 bit format with extra byte for padding
      for (int i = 0; i < w * h; i ++)
        { *bu++ = im[2]; *bu++ = im[1]; *bu++ = *im; *bu++ = 255; im += 3; }
    }

  if (shm_buffer)
    XShmPutImage(display, window, gc, image, 0, 0, x, y, w, h, 0);
  else
    XPutImage(display, window, gc, image, 0, 0, x, y, w, h);
  XFlush(display);

  ix = x;
  iy = y;

  // wait for completion:  (need a True as last param of XShmPutImage)
  //while(1) {
  //  XEvent xev; XNextEvent(display, &xev);
  //  LINFO("event %d", xev.type);
  //  if (xev.type == 70) break;
  //}
}

// ######################################################################
void XWindow::drawImage(const Image<byte>& img, const int x, const int y)
{
  ASSERT(x >= 0 && y >= 0 && x < width && y < height);
  int w = img.getWidth(); int h = img.getHeight();
  ASSERT(x + w <= width && y + h <= height);
  if (w != iw || h != ih) createImage(w, h);

  const byte* im = img.getArrayPtr();
  byte* bu = reinterpret_cast<byte*>(buffer);

  if (byt_per_pix == 2)
    {
      const int bytes_per_row = fourByteAlign(byt_per_pix * w);

      // 16 bit format: 565, lowest byte first
      for (int j = 0; j < h; ++j)
        {
          byte* bu_row = bu + j*bytes_per_row;
          for (int i = 0; i < w; ++i)
            {
              *bu_row++ = ((*im & 0x1C)<<3) | ((*im & 0xF8)>>3);
              *bu_row++ = ((*im & 0xE0)>>5) | (*im & 0xF8);
              im++;
            }
        }
    }
  else
    {
      // 24 bit format with extra byte for padding
      for (int i = 0; i < w * h; i ++)
        { *bu++ = *im; *bu++ = *im; *bu++ = *im++; *bu++ = 255; }
    }

  if (shm_buffer)
    XShmPutImage(display, window, gc, image, 0, 0, x, y, w, h, 0);
  else
    XPutImage(display, window, gc, image, 0, 0, x, y, w, h);
  XFlush(display);
  ix = x;
  iy = y;
}

// ######################################################################
XWindow::~XWindow()
{
  LDEBUG("Closing down...");
  if (shm_buffer)
    {
      XShmDetach(display, &shminfo);
      destroyImage();
      shmdt(shminfo.shmaddr);
      shmctl(shminfo.shmid, IPC_RMID, NULL);
    }
  else
    {
      destroyImage();
      free(buffer);
    }
  if (gc) XFreeGC(display, gc);
  if (display && window) XDestroyWindow(display, window);
  if (display) XCloseDisplay(display);     //segfaults???
}

// ######################################################################
void XWindow::createImage(const int w, const int h)
{
  destroyImage();

  uint depth;
  int pad;

  // make sure we use the correct depth values
  if (byt_per_pix == 2)
    { depth = 16; pad = 16;}
  else
    { depth = 24; pad = 32;}

  if (shm_buffer)
    // ok for shared memory; X will allocate the buffer:
    image = XShmCreateImage(display, visual, depth, ZPixmap,
                            buffer, &shminfo, w, h);
  else
    // cannot alloc in shared memory... do conventional alloc
    image = XCreateImage(display, visual, depth, ZPixmap , 0, buffer,
                         w, h, pad, fourByteAlign(byt_per_pix * w));

  iw = w; ih = h;
}

// ######################################################################
void XWindow::destroyImage()
{
  if (image)
    {
      // We have to set the XImage's data to zero before destroying the
      // XImage, because otherwise the XImage will want to free the data as
      // it is destroyed, which means that our 'buffer' would become a
      // dangling pointer, and that we'd be double-free'ing it when we try
      // to free() it ourselves later.
      image->data = 0;
      XDestroyImage(image);
      image = 0;
    }
}

// ######################################################################
Dims XWindow::getDims() const
{
  return Dims(width, height);
}

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */
