/*!@file XWindow.H Class definition for a simple window */

// //////////////////////////////////////////////////////////////////// //
// The iLab Neuromorphic Vision C++ Toolkit - Copyright (C) 2001 by the //
// University of Southern California (USC) and the iLab at USC.         //
// See http://iLab.usc.edu for information about this project.          //
// //////////////////////////////////////////////////////////////////// //
// Major portions of the iLab Neuromorphic Vision Toolkit are protected //
// under the U.S. patent ``Computation of Intrinsic Perceptual Saliency //
// in Visual Environments, and Applications'' by Christof Koch and      //
// Laurent Itti, California Institute of Technology, 2001 (patent       //
// pending; application number 09/912,225 filed July 23, 2001; see      //
// http://pair.uspto.gov/cgi-bin/final/home.pl for current status).     //
// //////////////////////////////////////////////////////////////////// //
// This file is part of the iLab Neuromorphic Vision C++ Toolkit.       //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is free software; you can   //
// redistribute it and/or modify it under the terms of the GNU General  //
// Public License as published by the Free Software Foundation; either  //
// version 2 of the License, or (at your option) any later version.     //
//                                                                      //
// The iLab Neuromorphic Vision C++ Toolkit is distributed in the hope  //
// that it will be useful, but WITHOUT ANY WARRANTY; without even the   //
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      //
// PURPOSE.  See the GNU General Public License for more details.       //
//                                                                      //
// You should have received a copy of the GNU General Public License    //
// along with the iLab Neuromorphic Vision C++ Toolkit; if not, write   //
// to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   //
// Boston, MA 02111-1307 USA.                                           //
// //////////////////////////////////////////////////////////////////// //
//
// Primary maintainer for this file: Laurent Itti <itti@usc.edu>
// $Id: XWindow.H 2820 2003-09-04 18:08:19Z zhanshi $
//

#ifndef XWINDOW_H_DEFINED
#define XWINDOW_H_DEFINED

#include "saliency.H"
#include "Point2D.H"

#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

class Dims;
template <class T> class PixRGB;
template <class T> class Image;

//! A simple X-window to display images.
/*! Use this class to create an window on your screen and display an image.

 It will use shared memory between us and the X server if supported by
 the X server and the display is local. Resizing is not allowed; so
 any image passed to drawImage() must have dimensions that match those
 specified at onstruction of the XWindow. */
class XWindow {
public:
  //! Constructor: create, display and map the window.
  /*!@param w Width of usable image area.
    @param h height of usable image area.
    @param x Horizontal position hint for window placement; use -1
    for automatic placement.
    @param y Vertical position hint for window placement; use -1
    for automatic placement.
    @param title String that will appear in the window title. */
  XWindow(const int w = 320, const int h = 240,
          const int x = -1, const int y = -1,
          const char *title = "USC Neuromorphic Vision");

  //! Destructor
  virtual ~XWindow();

  //! Set the position of an XWindow
  /*! @param x Horizontal position hint for window placement
    @param y Vertical position hint for window placement
  */
  void setPosition( const int x, const int y );

  //! Get the position of an XWindow
  /* Note: This may give you a bogus value if you used automatic window 
     placement */
  Point2D getPosition( void );

  //! Draw an image into the XWindow
  /*! @param img The image to draw. Image size must match the usable
    image area specified at construction of the XWindow.
    @param img is the image to display
    @param x if the x coordinate where the image should start in the window.
    @param y if the y coordinate where the image should start in the window.
  */
  void drawImage(const Image< PixRGB<byte> >& img, const int x = 0,
                 const int y = 0);

  //! Draw an image into the XWindow
  /*! @param img The image to draw. Image size must match the usable
    image area specified at construction of the XWindow.
    @param img is the image to display
    @param x if the x coordinate where the image should start in the window.
    @param y if the y coordinate where the image should start in the window.
  */
  void drawImage(const Image<byte>& img, const int x = 0,
                 const int y = 0);

  //! set the title bar to a new string
  virtual void setTitle(const char * title);

  //! return the dimensions of this window
  Dims getDims() const;

protected:
  void createImage(const int w, const int h);
  void destroyImage();

  int width, height;  // window size
  Display *display;   // the X display
  int screen;         // current X screen
  int depth;          // current X screen depth
  Window window;      // our window
  XImage *image;      // pointer to image data
  Visual *visual;     // pointer to current X visual
  GC gc;              // current graphic context
  XShmSegmentInfo shminfo; // info on shared memory segment
  char *buffer;       // shared memory buffer
  bool shm_buffer;    // true if we can use shared memory
  int iw, ih;         // XImage size
  int ix, iy;         // position of the image in the window
  int byt_per_pix;    // number of bytes per pixel
};

// ######################################################################
/* So things look consistent in everyone's emacs... */
/* Local Variables: */
/* indent-tabs-mode: nil */
/* End: */

#endif
